/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "arghints.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

GdamArgType    arg_hints_lookup(ArgHints*       arg_hints,
                                const char*     arg_name)
{
  int val;
  val = GPOINTER_TO_INT (g_hash_table_lookup (arg_hints->type_table,
					      arg_name));
  return (GdamArgType) val;
}

ArgHintsTable* arg_hints_table_new()
{
  ArgHintsTable *table;
  table = g_new (ArgHintsTable, 1);
  table->arg_table = g_hash_table_new (g_str_hash, g_str_equal);
  table->function_arg_table = g_hash_table_new (g_str_hash, g_str_equal);
  table->name_to_branch = g_hash_table_new (g_str_hash, g_str_equal);
  return table;
}
static ArgHints* arg_hints_new()
{
  ArgHints *rv;
  rv = g_new0 (ArgHints, 1);
  rv->type_table = g_hash_table_new (g_str_hash, g_str_equal);
  return rv;
}

static void
arg_hints_add_arg(ArgHints* hints, const char* name, GdamArgType type)
{
  g_hash_table_insert (hints->type_table,
		       g_strdup (name), GINT_TO_POINTER (type));
}

void                  arg_hints_table_add_arg(ArgHintsTable* table,
					      const char*    class_name,
                                              const char*    arg_name,
					      GdamArgType    arg_type)
{
  ArgHints *hints;
  hints = arg_hints_table_force_arg (table, class_name);
  arg_hints_add_arg (hints, arg_name, arg_type);
}
void                  arg_hints_table_add_func(ArgHintsTable* table,
					       const char*    class_name,
					       const char*    func_name,
                                               const char*    arg_name,
					       GdamArgType    arg_type)
{
  ArgHints *hints;
  hints = arg_hints_table_force_func (table, class_name, func_name);
  arg_hints_add_arg (hints, arg_name, arg_type);
}

#define SKIP_WHITESPACE(ptr)				\
		while (*(ptr) && isspace(*(ptr)))	\
			(ptr)++

static char* parse_nonws_string(char** p_start)
{
  char *start = *p_start;
  char *at = start;
  char *rv;
  while (*at != '\0' && isspace (*at))
    at++;
  rv = g_new (char, at - start + 1);
  memcpy (rv, start, at - start);
  rv[at - start] = '\0';
  return rv;
}
static GdamArgType parse_arg_type(char** p_start)
{
  char *start = *p_start;
  char *at = start;
  int len;
  while (*at != '\0' && isspace (*at))
    at++;
  len = at - start;

  if (len == 3 && strncasecmp (start, "int", len) == 0)
      return GDAM_ARG_TYPE_INT;
  else if (len == 6 && strncasecmp (start, "double", len) == 0)
      return GDAM_ARG_TYPE_DOUBLE;
  else if (len == 5 && strncasecmp (start, "int64", len) == 0)
      return GDAM_ARG_TYPE_INT64;
  else if (len == 6 && strncasecmp (start, "string", len) == 0)
      return GDAM_ARG_TYPE_STRING;
  else if (len == 3 + 6 && strncasecmp (start, "int_array", len) == 0)
      return GDAM_ARG_TYPE_INT_ARRAY;
  else if (len == 6 + 6 && strncasecmp (start, "double_array", len) == 0)
      return GDAM_ARG_TYPE_DOUBLE_ARRAY;
  else if (len == 5 + 6 && strncasecmp (start, "int64_array", len) == 0)
      return GDAM_ARG_TYPE_INT64_ARRAY;
  else if (len == 6 + 6 && strncasecmp (start, "string_array", len) == 0)
      return GDAM_ARG_TYPE_STRING_ARRAY;
  else
      return (GdamArgType) 0;
}
/*
 * syntax:  [as many lines as you like of the form]
 *    ARG class_name arg_name arg_type
 *    FUNC class_name func_name arg_name arg_type
 * [where ARG and FUNC are the words `arg' and `func', case insensitive]
 *
 * and where:
 *    class_name arg_name func_name := [A-Za-z_0-9]+
 *    arg_type := `int' | `string' | `int64' | `double' | `int_array' 
 *              | `string_array' | `int64_array' | `double_array'
 */
ArgHintsTable* arg_hints_table_load(const char*        filename)
{
  char buf[2048];
  FILE *fp;
  ArgHintsTable *table;

  fp = fopen (filename, "r");
  if (fp == NULL)
    {
      g_warning ("couldn't open %s: %s", filename, strerror (errno));
      return NULL;
    }

  table = arg_hints_table_new ();

  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      char *at;
      at = buf;
      SKIP_WHITESPACE (at);
      if (*at == '\0' || *at == '#')
	continue;
      if (strncasecmp (at, "arg", 3) == 0 && isspace (at[3]))
	{
	  char *class_name;
	  char *arg_name;
	  GdamArgType type;
	  at += 3;
	  class_name = parse_nonws_string (&at);
	  arg_name = parse_nonws_string (&at);
	  type = parse_arg_type (&at);
	  if (type == 0)
	    {
	      g_free (class_name);
	      g_free (arg_name);
	      goto error;
	    }
	  arg_hints_table_add_arg (table, class_name, arg_name, type);
	  g_free (class_name);
	  g_free (arg_name);
	  continue;
	}
      if (strncasecmp (at, "func", 4) == 0 && isspace (at[4]))
	{
	  char *class_name;
	  char *func_name;
	  char *arg_name;
	  GdamArgType type;
	  at += 4;
	  class_name = parse_nonws_string (&at);
	  func_name = parse_nonws_string (&at);
	  arg_name = parse_nonws_string (&at);
	  type = parse_arg_type (&at);
	  if (type == 0)
	    {
	      g_free (class_name);
	      g_free (func_name);
	      g_free (arg_name);
	      goto error;
	    }
	  arg_hints_table_add_func (table, class_name, func_name,
				    arg_name, type);
	  g_free (class_name);
	  g_free (func_name);
	  g_free (arg_name);
	  continue;
	}
    }

  return table;

error:
  arg_hints_table_destroy (table);
  return NULL;
}

static inline void maybe_try_parents(ArgHints* hints, const char* class_name,
                                     ArgHintsTable* table)
{
  g_return_if_fail (class_name != NULL);
  if (!hints->tried_parent)
    {
      const char *parent;
      hints->tried_parent = TRUE;
      parent = arg_hints_table_lookup_parent (table, class_name);
      if (parent == NULL)
	return;
      hints->parent_hints = arg_hints_table_force_arg (table, parent);
    }
}

ArgHints*      arg_hints_table_force_arg  (ArgHintsTable*     table,
                                           const char*        class_name)
{
  ArgHints *hints;
  g_return_val_if_fail (class_name != NULL, NULL);
  hints = g_hash_table_lookup (table->arg_table, class_name);
  if (hints == NULL)
    {
      hints = arg_hints_new ();
      maybe_try_parents (hints, class_name, table);
      g_hash_table_insert (table->arg_table, g_strdup (class_name), hints);
    }
  return hints;
}

ArgHints*      arg_hints_table_lookup_arg (ArgHintsTable*     table,
                                           const char*        class_name)
{
  ArgHints *hints;
  g_return_val_if_fail (class_name != NULL, NULL);
  hints = g_hash_table_lookup (table->arg_table, class_name);
  if (hints != NULL)
    maybe_try_parents (hints, class_name, table);
  return hints;
}
ArgHints*      arg_hints_table_force_func (ArgHintsTable*     table,
                                           const char*        class_name,
					   const char*        function_name)
{
  GHashTable *subtable;
  ArgHints *hints;
  g_return_val_if_fail (class_name != NULL, NULL);
  subtable = g_hash_table_lookup (table->function_arg_table, class_name);
  if (subtable == NULL)
    {
      subtable = g_hash_table_new (g_str_hash, g_str_equal);
      g_hash_table_insert (table->function_arg_table,
			   g_strdup (class_name), subtable);
      hints = NULL;
    }
  else
    {
      hints = g_hash_table_lookup (subtable, function_name);
    }
  if (hints == NULL)
    {
      hints = arg_hints_new ();
      g_hash_table_insert (subtable, g_strdup (function_name), hints);
    }
  return hints;
}

ArgHints*      arg_hints_table_lookup_func(ArgHintsTable*     table,
                                           const char*        class_name,
					   const char*        function_name)
{
  GHashTable *subtable;
  g_return_val_if_fail (class_name != NULL, NULL);
  subtable = g_hash_table_lookup (table->function_arg_table, class_name);
  if (subtable == NULL)
    return NULL;
  return g_hash_table_lookup (subtable, function_name);
}
const char*
arg_hints_table_lookup_parent(ArgHintsTable* table, const char* class_name)
{
  GSList *list;
  const char *parent;
  g_return_val_if_fail (class_name != NULL, NULL);
  list = g_hash_table_lookup (table->name_to_branch, class_name);
  if (list == NULL)
    return NULL;
  if (list->next == NULL)
    return NULL;
  parent = list->next->data;
  return parent;
}
void arg_hints_table_add_base_class(ArgHintsTable* table,
                                    const char* base)
{
  GSList *base_list;
  g_return_if_fail (base != NULL);
  base_list = g_hash_table_lookup (table->name_to_branch, base);
  g_return_if_fail (base_list == NULL);
  base_list = g_slist_prepend (NULL, g_strdup (base));
  g_hash_table_insert (table->name_to_branch, base_list->data, base_list);
}

void arg_hints_table_add_parent(ArgHintsTable* table,
                                const char* base,
				const char* derived)
{
  GSList *base_list;
  GSList *derived_list;
  g_return_if_fail (base != NULL);
  g_return_if_fail (derived != NULL);
  base_list = g_hash_table_lookup (table->name_to_branch, base);
  if (base_list == NULL)
    {
      g_warning ("%s not found ; trying to add derived class %s",
		 base, derived);
      return;
    }
  g_return_if_fail (base_list != NULL);
  derived_list = g_hash_table_lookup (table->name_to_branch, derived);
  if (derived_list != NULL)
    {
      g_warning ("class %s [base class %s] has already been entered",
		 derived, base);
      return;
    }
  g_return_if_fail (derived_list == NULL);
  derived_list = g_slist_prepend (base_list, g_strdup (derived));
  g_hash_table_insert (table->name_to_branch,
		       derived_list->data, derived_list);
}

static void free_type_table(GHashTable* table)
{
  g_hash_table_foreach (table, (GHFunc) g_free, NULL);
  g_hash_table_destroy (table);
}

static void free_sub_table(gpointer key, gpointer value, gpointer unused)
{
  g_free (key);
  free_type_table ((GHashTable *) value);
}

void arg_hints_table_destroy(ArgHintsTable* table)
{
  g_hash_table_foreach (table->function_arg_table, free_sub_table, NULL);
  g_hash_table_destroy (table->function_arg_table);
  free_type_table (table->arg_table);
  g_free (table);
}
