/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamchannel.h"


guint        gdam_channel_source_set_arg          (GdamChannel* channel,
                                                   int          source_id,
						   ...)
{
  GdamRequest *request;
  gboolean result;
  GdamArg *arg;
  va_list arg_list;
  va_start (arg_list, source_id);
  arg = gdam_arg_parse_from_vararg (1, &arg_list);
  va_end (arg_list);
  request = gdam_request_set_args (source_id,
				   1,
				   arg,
				   gdam_request_get_success,
				   gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return 0;
  gdam_arg_free_array (1, arg);
  return result;
}

guint        gdam_channel_create_sink             (GdamChannel* channel,
                                                   const char*  dev_name)
{
  GdamRequestGetIntStruct ret;
  GdamRequest *request;
  request = gdam_request_create_sink (dev_name,
				      gdam_request_get_int,
				      gdam_request_get_failure, &ret);
  if (!gdam_channel_process_request (channel, request))
    return 0;
  return ret.was_success ? ret.rv : 0;

}

gboolean     gdam_channel_mix_add                 (GdamChannel* channel,
                                                   guint        mix_id,
						   guint        subsource_id,
						   GdamTime     when)
{
  GdamRequest *request;
  gboolean result;
  request = gdam_request_add_mix (mix_id,
				  subsource_id,
				  when,
				  gdam_request_get_success,
				  gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}

gboolean     gdam_channel_detach_source           (GdamChannel* channel,
                                                   guint        source_id)
{
  GdamRequest *request;
  gboolean result;
  request = gdam_request_detach_source (source_id,
					gdam_request_get_success,
					gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}

gboolean     gdam_channel_seek_relative           (GdamChannel* channel,
                                                   guint        source_id,
						   GdamTime     when)
{
  GdamRequest *request;
  gboolean result;
  request = gdam_request_seek_relative (source_id,
					when,
					gdam_request_get_success,
					gdam_request_get_failure, &result);
  if (gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}

gboolean     gdam_channel_feed_sink               (GdamChannel* channel,
                                                   guint        device_id,
                                                   guint        source_id)
{
  GdamRequest *request;
  gboolean result;
  request = gdam_request_feed_sink (device_id,
				    source_id,
				    gdam_request_get_success,
				    gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}



gboolean     gdam_channel_synchronize             (GdamChannel* channel,
                                                   int          mobile_id,
                                                   int          fixed_id,
					           gdouble      mobile_bpm,
					           gdouble      mobile_start,
					           gdouble      fixed_bpm,
					           gdouble      fixed_start)
{
  gboolean result;
  GdamRequest *request;
  request = gdam_request_synchronize (mobile_id,
				      fixed_id,
				      mobile_bpm,
				      mobile_start,
				      fixed_bpm,
				      fixed_start,
				      gdam_request_get_success,
				      gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}

gboolean     gdam_channel_sync_preview            (GdamChannel* channel,
                                                   int          mobile_id,
                                                   int          fixed_id,
					           gdouble      mobile_bpm,
					           gdouble      mobile_start,
					           gdouble      fixed_bpm,
					           gdouble      fixed_start,
						   GdamTime*    time_rv)
{
  GdamRequestGetTimeStruct ret;
  GdamRequest *request;
  request = gdam_request_sync_preview (mobile_id,
				       fixed_id,
				       mobile_bpm, mobile_start,
				       fixed_bpm, fixed_start,
				       gdam_request_get_time,
				       gdam_request_get_failure, &ret);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  if (ret.was_success)
    {
      *time_rv = ret.rv;
      return TRUE;
    }
  return FALSE;
}

guint        gdam_channel_start_caching           (GdamChannel* channel,
                                                   int          to_drain_id,
						   GHookFunc    done_caching,
						   gpointer     done_data)
{
  GdamRequest *request;
  guint buffer_id, event_id;
  GdamRequestGet2IntStruct result;
  GdamChannelEventHandler *handler;
  request = gdam_request_start_caching (to_drain_id,
					gdam_request_get_2int,
					gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  if (!result.was_success)
    return 0;

  buffer_id = result.rv[0];
  event_id = result.rv[1];

  gdam_debug ("start_caching: buffer_id=%d, event_id=%d\n",
	      buffer_id, event_id);

  /* Connect that event... */
  handler = gdam_channel_simple_handler_new (event_id,
					     TRUE, done_caching, done_data);
  gdam_channel_register_handler (channel, handler);

  return buffer_id;
}


gboolean gdam_channel_get_server_side_int(GdamChannel* channel,
                                          const char* namespac,
					  const char* name,
					  gint* output)
{
  GdamRequestGetIntStruct ret;
  GdamRequest *request;
  request = gdam_request_get_server_side_int (namespac,
					      name,
					      gdam_request_get_int,
					      gdam_request_get_failure, &ret);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  if (ret.was_success)
    {
      *output = ret.rv;
      return TRUE;
    }
  return FALSE;
}

guint        gdam_channel_subscribe_source_ended  (GdamChannel* channel,
                                                   guint        source_id,
						   GHookFunc    on_ended,
						   gpointer     on_ended_data)
{
  GdamRequestGetIntStruct ret;
  GdamRequest *request;
  request = gdam_request_subscribe_source_ended (source_id,
						 gdam_request_get_int,
						 gdam_request_get_failure,
						 &ret);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  if (ret.was_success)
    {
      guint connection_id = ret.rv;
      gdam_channel_register_handler (channel,
				     gdam_channel_simple_handler_new
				     (connection_id, TRUE, on_ended,
				      on_ended_data));
      return connection_id;
    }
  return 0;
}
guint        gdam_channel_subscribe_source_timer  (GdamChannel* channel,
                                                   guint        source_id,
						   int          period,
						   GdamRequestGotTimeFunc
						                periodic,
						   gpointer     periodic_data)
{
  GdamRequestGetIntStruct ret;
  GdamRequest *request;
  request = gdam_request_subscribe_source_timer (source_id,
						 period,
						 gdam_request_get_int,
						 gdam_request_get_failure,
						 &ret);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  if (ret.was_success)
    {
      guint connection_id = ret.rv;
      gdam_channel_register_handler (channel,
				     gdam_channel_time_handler_new
				     (connection_id, FALSE, periodic,
				      periodic_data));
      return connection_id;
    }
  return 0;
}
gboolean gdam_channel_unsubscribe(GdamChannel* channel, guint subscription_id)
{
  GdamRequest *request;
  gboolean result;
  request = gdam_request_unsubscribe (subscription_id,
				      gdam_request_get_success,
				      gdam_request_get_failure, &result);
  if (!gdam_channel_process_request (channel, request))
    return FALSE;
  return result;
}

