/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamchannel.h"
#include <gdam/common/gdamconnect.h>
#include <stdlib.h>
#include <gtk/gtkmain.h>
#include <gdam/common/gdam-protocol.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <gdam/common/gdampoll.h>

/* XXX: add a configure check for socketpair, and disallow
 *      launching of the server if it doesn't exist.
 *
 *      It should exist on any civilised system...
 */


/* just like pipe, except resulting pipes are bidirectional.     */
/* a.k.a. stream pipes, see Advanced Programming, Stevens, Ch 15 */
#define s_pipe(fds) socketpair(AF_UNIX, SOCK_STREAM, 0, fds)

/* Fork and exec a gdam-server, returning a socket
 * to talk to the server with. */
static int launch_gdam_server(gboolean secure)
{
  int fds[2];
  int pid;
  if (s_pipe (fds) < 0)
    g_error ("error creating socketpair to talk to server with");

retry_fork:
  pid = fork ();
  if (pid < 0)
    {
      if (errno == EAGAIN)
	{
	  g_message ("supposedly temporary problem forking... "
		     "waiting then retrying.");
	  sleep (2);
	  goto retry_fork;
	}
      g_error ("error forking gdam-server");
    }

  if (pid == 0)
    {
      /* Child. */
      char fd_as_str[255];
      char *args[128];
      int num_args;
      const char *server_path = NULL;
      struct stat sbuf;
      sprintf (fd_as_str, "%d", fds[1]);
      close (fds[0]);

      /* Don't close fds[1] on exec: */
      fcntl (fds[1], F_SETFD, 0);

      /* HACK: if the file ../server/gdam-server exists,
       *       chdir and use it explicitly. */
      if (stat ("../server/gdam-server", &sbuf) == 0)
	{
	  chdir ("../server");
	  server_path = "./gdam-server";
	}

      num_args = 0;
      args[num_args++] = "gdam-server";
      if (gdam_debugging_on)
	args[num_args++] = "--debug";
      args[num_args++] = "--control-fd";
      args[num_args++] = fd_as_str;
      if (secure)
	args[num_args++] = "--dont-bind";
      args[num_args++] = "--guess-config-file";
      args[num_args++] = NULL;
      if (server_path != NULL)
	{
	  execv (server_path, args);
	}
      execvp ("gdam-server", args);
      g_warning ("couldn't find gdam-server to fork");
      _exit (0);
    }
  close (fds[1]);
  return fds[0];
}
static gboolean on_connect_failure(GdamChannel* channel)
{
  GdamChannelServerLaunch launch_val = channel->launch_server_val;
  if (channel->launch_server_func != NULL)
    {
      GdamChannelServerLaunchFunc func = channel->launch_server_func;
      gpointer data = channel->launch_server_data;
      launch_val = (*func) (channel, data);
    }
  switch (launch_val)
    {
    case GDAM_CHANNEL_SERVER_LAUNCH_ABORT:
      {
	g_message ("Couldn't connect to gdam-server: aborting");
	exit (1);
	break;
      }

    case GDAM_CHANNEL_SERVER_LAUNCH_INSECURE:
      channel->poll_fd.fd = launch_gdam_server (FALSE);
      break;

    case GDAM_CHANNEL_SERVER_LAUNCH_SECURE:
      channel->poll_fd.fd = launch_gdam_server (TRUE);
      break;

    case GDAM_CHANNEL_SERVER_LAUNCH_DEFER:
      channel->launch_server_deferred = TRUE;
      channel->is_connecting = FALSE;
      break;

    default:
      g_assert_not_reached ();
    }
  return FALSE;
}

static void default_fatal_func(gpointer unused)
{
  fprintf (stderr, "The server has died!\n\n" "ABORTING...\n");
  abort ();
}

static gboolean default_auth_handler(GdamChannel* channel, GdamBuffer* resp)
{
  gdam_debug ("default_auth_handler:  proceeding");
  return TRUE;
}
GdamChannel* gdam_channel_new_full(const char* addr_spec,
                                   GdamBuffer* auth_info,
				   GdamChannelServerLaunchFunc func,
				   gpointer    func_data,
				   GdamChannelOnConnectFunc on_connect_func,
				   gpointer    on_connect_data)
{
  GdamChannel *channel;
  int fd;
  int is_connecting = 0;

  switch (gdam_connect_generic (addr_spec, &fd, gdam_debugging_on))
    {
    case GDAM_CONNECT_CONNECTING:
      is_connecting = 1;
      break;
    case GDAM_CONNECT_CONNECTED:
      is_connecting = 0;
      break;
    case GDAM_CONNECT_FAILED:
      fd = -1;
      is_connecting = 0;
      break;
    }

  channel = g_new0 (GdamChannel, 1);
  gdam_buffer_construct (&channel->outgoing);
  gdam_buffer_construct (&channel->incoming);
  channel->first_request = channel->last_request = NULL;
  channel->first_pending = channel->last_pending = NULL;
  channel->num_pending = 0;
  channel->max_pending = 5;
  channel->poll_fd.fd = fd;
  channel->source_id = 0;
  channel->ref_count = 1;
  channel->is_connecting = is_connecting;
  channel->quit_main = 0;
  channel->launch_server_func = func;
  channel->launch_server_data = func_data;
  channel->launch_server_val = GDAM_CHANNEL_SERVER_LAUNCH_ABORT;
  channel->auth_handler = default_auth_handler;
  channel->event_to_info_map = g_hash_table_new (NULL, NULL);
  channel->fatal_func = default_fatal_func;
  channel->on_connect_func = on_connect_func;
  channel->on_connect_data = on_connect_data;
  g_hook_list_init (&channel->chat_handlers,
		    sizeof (GHook) + sizeof (char *) * 2);
  {
    int length;
    g_assert (gdam_buffer_peek_int (auth_info, &length));
    g_assert (auth_info->size == length);
  }
  gdam_buffer_append_buffer (&channel->outgoing, auth_info, FALSE);

  if (fd < 0)
    g_idle_add ((GSourceFunc) on_connect_failure, channel);

  /* TODO: register auth handshake (as pending request) */
  return channel;
}


static gboolean gdam_channel_prepare(gpointer  source_data, 
			             GTimeVal *current_time,
			             gint     *timeout,
			             gpointer  user_data)
{
  GdamChannel *channel = (GdamChannel *) user_data;
  if (channel->is_connecting)
    {
      channel->poll_fd.events = G_IO_IN | G_IO_OUT;
    }
  else if (channel->outgoing.size > 0)
    {
      channel->poll_fd.events = G_IO_IN | G_IO_OUT;
    }
  else
    {
      channel->poll_fd.events = G_IO_IN;
    }

  if (channel->poll_fd.fd < 0 || channel->launch_server_deferred)
    return FALSE;
  *timeout = -1;
  channel->poll_fd.revents = 0;
  return FALSE;
}
static gboolean gdam_channel_check(gpointer  source_data, 
			           GTimeVal *current_time,
			           gpointer  user_data)
{
  GdamChannel *channel = (GdamChannel *) user_data;
  if (channel->poll_fd.fd < 0 || channel->launch_server_deferred)
    return FALSE;
  if (channel->poll_fd.revents)
    return TRUE;
  return FALSE;
}
static gboolean is_event_type(int type)
{
  if (type > GDAM_EVENT_FIRST && type < GDAM_MESSAGE_LAST_MESSAGE)
    return TRUE;
  return FALSE;
}

void gdam_channel_register_handler(GdamChannel* channel, 
                                  GdamChannelEventHandler* handler)
{
  g_hash_table_insert (channel->event_to_info_map,
		       GUINT_TO_POINTER (handler->connection_id), handler);
}
void gdam_channel_unregister_handler(GdamChannel* channel,
                                     guint connection_id)
{
  g_hash_table_remove (channel->event_to_info_map,
		       GUINT_TO_POINTER (connection_id));
}

static gboolean one_shot_simple_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
  handler->callback (handler->user_data);
  return TRUE;
}
static gboolean repeating_simple_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
  handler->callback (handler->user_data);
  return FALSE;
}

GdamChannelEventHandler*
gdam_channel_simple_handler_new(guint connection_id,
                                gboolean one_shot,
                                GHookFunc func,
				gpointer func_data)
{
  GdamChannelEventHandler *handler;
  handler = g_new (GdamChannelEventHandler, 1);
  handler->connection_id = connection_id;
  handler->callback = func;
  handler->user_data = func_data;
  if (one_shot)
    handler->marshal = one_shot_simple_marshal;
  else
    handler->marshal = repeating_simple_marshal;
  return handler;
}

static gboolean one_shot_time_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
  GdamRequestGotTimeFunc func = (GdamRequestGotTimeFunc) (handler->callback);
  GdamTime time;
  if (!gdam_buffer_parse_time (buffer, &time))
    return TRUE;
  (*func) (time, handler->user_data);
  return TRUE;
}
static gboolean repeating_time_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
  GdamRequestGotTimeFunc func = (GdamRequestGotTimeFunc) (handler->callback);
  GdamTime time;
  if (!gdam_buffer_parse_time (buffer, &time))
    {
      g_warning ("error parsing time in periodic timer");
      return TRUE;		/* causes it to stop happening... */
    }
  (*func) (time, handler->user_data);
  return FALSE;
}

GdamChannelEventHandler* gdam_channel_time_handler_new
                                                  (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotTimeFunc func,
						   gpointer     func_data)
{
  GdamChannelEventHandler *handler;
  handler = g_new (GdamChannelEventHandler, 1);
  handler->connection_id = connection_id;
  handler->callback = (GHookFunc) func;
  handler->user_data = func_data;
  if (one_shot)
    handler->marshal = one_shot_time_marshal;
  else
    handler->marshal = repeating_time_marshal;
  return handler;
}


/* INT */
static gboolean one_shot_int_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
  GdamRequestGotIntFunc func = (GdamRequestGotIntFunc) (handler->callback);
  int i;
  if (!gdam_buffer_parse_int (buffer, &i))
    return TRUE;
  (*func) (i, handler->user_data);
  return TRUE;
}
static gboolean repeating_int_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
  GdamRequestGotIntFunc func = (GdamRequestGotIntFunc) (handler->callback);
  int i;
  if (!gdam_buffer_parse_int (buffer, &i))
    {
      g_warning ("error parsing int in periodic timer");
      return TRUE;		/* causes it to stop happening... */
    }
  (*func) (i, handler->user_data);
  return FALSE;
}

GdamChannelEventHandler* gdam_channel_int_handler_new
                                                  (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotIntFunc func,
						   gpointer     func_data)
{
  GdamChannelEventHandler *handler;
  handler = g_new (GdamChannelEventHandler, 1);
  handler->connection_id = connection_id;
  handler->callback = (GHookFunc) func;
  handler->user_data = func_data;
  if (one_shot)
    handler->marshal = one_shot_int_marshal;
  else
    handler->marshal = repeating_int_marshal;
  return handler;
}
/* STRING */
static gboolean one_shot_string_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
  GdamRequestGotStringFunc func = (GdamRequestGotStringFunc)
    (handler->callback);
  char *string;
  string = gdam_buffer_parse_string0 (buffer);
  if (string == NULL)
    return TRUE;
  (*func) (string, handler->user_data);
  g_free (string);
  return TRUE;
}
static gboolean repeating_string_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
  GdamRequestGotStringFunc func = (GdamRequestGotStringFunc)
    (handler->callback);
  char *string;
  string = gdam_buffer_parse_string0 (buffer);
  if (string == NULL)
    {
      g_warning ("error parsing string in periodic timer");
      return TRUE;		/* causes it to stop happening... */
    }
  (*func) (string, handler->user_data);
  g_free (string);
  return FALSE;
}

GdamChannelEventHandler* gdam_channel_string_handler_new
					  (guint        connection_id,
					   gboolean     one_shot,
					   GdamRequestGotStringFunc func,
					   gpointer     func_data)
{
  GdamChannelEventHandler *handler;
  handler = g_new (GdamChannelEventHandler, 1);
  handler->connection_id = connection_id;
  handler->callback = (GHookFunc) func;
  handler->user_data = func_data;
  if (one_shot)
    handler->marshal = one_shot_string_marshal;
  else
    handler->marshal = repeating_string_marshal;
  return handler;
}

static void
marshal_chat_messages (GHook     *hook,
		       gpointer   strings)
{
  char **strs = strings;
  char **requested = (char **) (hook + 1);
  GdamChannelChatFunc chat_func = (GdamChannelChatFunc) hook->func;
  if ((requested[0] == NULL || strcmp (requested[0], strs[0]) == 0)
   && (requested[1] == NULL || strcmp (requested[1], strs[1]) == 0))
    (*chat_func) (strs[0], strs[1], strs[2], hook->data);
}

static gboolean
process_chat_message (GdamChannel   *channel,
		      GdamBuffer    *buffer)
{
  char *strings[3];
  strings[0] = gdam_buffer_parse_string0 (buffer);
  strings[1] = gdam_buffer_parse_string0 (buffer);
  if (strings[1] == NULL)
    {
      g_free (strings[0]);
      return FALSE;
    }
  strings[2] = strchr (strings[1], ' ');
  if (strings[2] == NULL)
    {
      g_free (strings[0]);
      g_free (strings[1]);
      return FALSE;
    }
  strings[2][0] = '\0';
  strings[2] = strings[2] + 1;
  g_hook_list_marshal (&channel->chat_handlers,
		       FALSE,
		       marshal_chat_messages,
		       &strings[0]);
  g_free (strings[0]);
  g_free (strings[1]);
  return TRUE;
}

static gboolean gdam_channel_dispatch_event(GdamChannel* channel,
                                            GdamBuffer*  command)
{
  /* Lookup handler. */
  GdamChannelEventHandler *info;
  int type;
  guint id;
  GHashTable *event_to_info_map = channel->event_to_info_map;

  if (!gdam_buffer_parse_int (command, &type)
      || !gdam_buffer_parse_int (command, &id))
    {
      //gdam_debug("dispatch_event: too short");
      /* XXX: these 2 debugs needn't be repeated as warnings.  */
      g_warning ("dispatch_event: too short");
      return FALSE;
    }

  if (id == 0 && type == GDAM_EVENT_CHAT_MESSAGE)
    return process_chat_message (channel, command);

  info = g_hash_table_lookup (channel->event_to_info_map,
			      GUINT_TO_POINTER (id));

  if (!info)
    {
      //gdam_debug("event %d of type %d occurred when not expected",
      //id, type);
      g_warning ("event %d of type %d occurred when not expected", id, type);
      return FALSE;
    }
  /* Invoke handler. */
  if (info->marshal (info, command, id))
    {
      g_free (info);
      g_hash_table_remove (event_to_info_map, GUINT_TO_POINTER (id));
    }
  return TRUE;
}

static gboolean gdam_channel_dispatch_response(GdamChannel* channel,
                                               GdamBuffer*  command)
{
  GdamRequest *pending = (GdamRequest *) channel->first_pending->data;
  channel->first_pending = g_slist_remove (channel->first_pending, pending);
  channel->num_pending--;
  if (!channel->first_pending)
    channel->last_pending = NULL;
  if (!pending->marshal (pending, command))
    return FALSE;
  gdam_request_destroy (pending);
  return TRUE;
}

/* If there is enough data in the buffer to dispatch,
 * we do so... */
static gboolean gdam_channel_maybe_dispatch_input(GdamChannel* channel)
{
  GdamBuffer *incoming = &channel->incoming;
  while (1)
    {
      GdamBuffer cmd;
      int length, type;
      if (!gdam_buffer_peek_int (incoming, &length))
	break;
      if (incoming->size < length)
	break;
      gdam_buffer_construct (&cmd);
      gdam_buffer_send_buffer (&cmd, incoming, length, TRUE);
      g_assert (length == cmd.size);
      gdam_buffer_parse_int (&cmd, &length);
      if (channel->auth_handler)
	{
	  if (!channel->auth_handler (channel, &cmd))
	    {
	      gdam_buffer_destruct (&cmd);
	      g_warning ("channel unauthorized, cannot " "dispatch input.");

	      return FALSE;
	    }
	  if (channel->on_connect_func != NULL)
	    {
	      (*channel->on_connect_func) (channel, channel->on_connect_data);

	      /* no real reason to do this... */
	      channel->on_connect_func = NULL;
	      channel->on_connect_data = NULL;
	    }
	  channel->auth_handler = NULL;
	  gdam_buffer_destruct (&cmd);
	  continue;
	}
      gdam_buffer_peek_int (&cmd, &type);
      if (is_event_type (type))
	{
	  if (!gdam_channel_dispatch_event (channel, &cmd))
	    {
	      g_warning ("couldn't dispatch event, cannot "
			 "dispatch input.");
	      gdam_buffer_destruct (&cmd);
	      return FALSE;
	    }
	}
      else
	{
	  if (!gdam_channel_dispatch_response (channel, &cmd))
	    {
	      g_warning ("couldn't dispatch response, cannot "
			 "dispatch input.");
	      gdam_buffer_destruct (&cmd);
	      return FALSE;
	    }
	}
      gdam_buffer_destruct (&cmd);
    }
  return TRUE;
}

static void maybe_write_to_outgoing(GdamChannel* channel)
{
  while (channel->num_pending < channel->max_pending
	 && channel->first_request != NULL)
    {
      GSList *old_first = channel->first_request;
      GdamRequest *request = (GdamRequest *) (old_first->data);
      GdamBuffer *req_buf = &request->request_command;
      int length;

      g_assert (gdam_buffer_peek_int (req_buf, &length));
      g_assert (req_buf->size == length);

      gdam_buffer_append_buffer (&channel->outgoing, req_buf, TRUE);

      /* Now, horrible first manipulation to remove
       * `first_request' and append it as `last_pending'...
       * (This looks horrible but it is efficient in that
       *  it doesn't create or destroy a list node. */
      channel->first_request = old_first->next;
      if (!channel->first_request)
	channel->last_request = NULL;
      old_first->next = NULL;
      if (channel->last_pending)
	channel->last_pending->next = old_first;
      else
	channel->first_pending = old_first;
      channel->last_pending = old_first;
      channel->num_pending++;
    }
}

static gboolean gdam_channel_dispatch(gpointer  source_data, 
			              GTimeVal *current_time,
			              gpointer  user_data)
{
  GdamChannel *channel = (GdamChannel *) user_data;
  int fd = channel->poll_fd.fd;
  if (channel->is_connecting)
    {
      gboolean verbose = (channel->launch_server_func == NULL);
      if (!gdam_connect_finish (fd, verbose))
	{
	  close (fd);
	  fd = channel->poll_fd.fd = -1;
	  channel->is_connecting = 0;
	  on_connect_failure (channel);
	  if (channel->launch_server_deferred)
	    return TRUE;
	  if (channel->poll_fd.fd < 0)
	    g_main_remove_poll (&channel->poll_fd);
	  return (fd >= 0);
	}
      channel->is_connecting = 0;
      return TRUE;
    }
  /*
     gdam_debug("dispatch io-event: fd=%d; ev=%d [in=%d, out=%d]", 
     channel->poll_fd.fd, channel->poll_fd.revents,
     G_IO_IN, G_IO_OUT);
   */
  if ((channel->poll_fd.revents & G_IO_IN) == G_IO_IN)
    {
      if (!gdam_buffer_read_one (&channel->incoming, fd))
	{
	  g_warning ("couldn't buffer_read_one");
	  goto channel_error;
	}
      if (!gdam_channel_maybe_dispatch_input (channel))
	{
	  g_warning ("couldn't maybe_dispatch_input");
	  goto channel_error;
	}
      maybe_write_to_outgoing (channel);
    }
  if ((channel->poll_fd.revents & G_IO_OUT) == G_IO_OUT)
    {
      if (!gdam_buffer_write_one (&channel->outgoing, fd))
	{
	  g_warning ("couldn't buffer_write_one");
	  goto channel_error;
	}
      maybe_write_to_outgoing (channel);
    }
  return TRUE;

channel_error:
  if (channel->fatal_func != NULL)
    (*channel->fatal_func) (channel->fatal_func_data);
  return FALSE;
}


static GSourceFuncs channel_source_funcs = {
	gdam_channel_prepare,
	gdam_channel_check,
	gdam_channel_dispatch,
	NULL
};


static void gdam_channel_detach_notify(gpointer chan)
{
  GdamChannel *channel = (GdamChannel *) chan;
  if (channel->source_id == 0)
    return;
  channel->source_id = 0;
  g_main_remove_poll (&channel->poll_fd);
  gdam_channel_unref (channel);
}

void gdam_channel_ref(GdamChannel* channel)
{
  ++channel->ref_count;
}

void gdam_channel_unref(GdamChannel* channel)
{
  if (--channel->ref_count == 0)
    gdam_channel_destroy (channel);
}

int gdam_channel_new_local_id(GdamChannel* channel)
{
  return --(channel->last_local_id);
}

/* Attach to the main loop... you have to do this for
 * events to get processed. */
void gdam_channel_attach(GdamChannel* channel)
{
  g_return_if_fail (channel->source_id == 0);
  channel->source_id = g_source_add (G_PRIORITY_DEFAULT,
				     TRUE,
				     &channel_source_funcs,
				     NULL,
				     channel, gdam_channel_detach_notify);
  g_main_add_poll (&channel->poll_fd, G_PRIORITY_DEFAULT);
  ++channel->ref_count;
}

void gdam_channel_detach(GdamChannel* channel)
{
  int source_id = channel->source_id;
  if (source_id == 0)
    return;
  channel->source_id = 0;
  g_source_remove (source_id);
  g_main_remove_poll (&channel->poll_fd);
}


/* Delete us (safe whether or not attached). */
void gdam_channel_destroy(GdamChannel* channel)
{
  gdam_channel_detach (channel);
  gdam_buffer_destruct (&channel->outgoing);
  gdam_buffer_destruct (&channel->incoming);
  g_slist_foreach (channel->first_request,
		   (GFunc) gdam_request_destroy, NULL);
  g_slist_foreach (channel->first_pending,
		   (GFunc) gdam_request_destroy, NULL);
  g_slist_free (channel->first_request);
  g_slist_free (channel->first_pending);
  if (channel->quit_main)
    exit (0);
  g_free (channel);
}


gboolean gdam_channel_append_request   (GdamChannel* channel,
                                        GdamRequest* request)
{
  GSList *new_last = g_slist_prepend (NULL, request);
  if (channel->last_request)
    channel->last_request->next = new_last;
  else
    channel->first_request = new_last;
  channel->last_request = new_last;
  maybe_write_to_outgoing (channel);
  return TRUE;
}


gboolean gdam_channel_prepend_request   (GdamChannel* channel,
                                        GdamRequest* request)
{
  GSList *new_first = g_slist_prepend (channel->first_request, request);
  if (!channel->last_request)
    channel->last_request = new_first;
  channel->first_request = new_first;
  maybe_write_to_outgoing (channel);
  return TRUE;
}

static gboolean flush_pending(GdamChannel* channel)
{
  GPollFD fds;
  fds.fd = channel->poll_fd.fd;
  while (channel->first_pending)
    {
      char buf[4096];
      int num_read;
      fds.events = G_IO_IN;
      if (channel->outgoing.size > 0)
	fds.events |= G_IO_OUT;
      if (gdam_poll (&fds, 1, -1) < 0)
	{
	  if (errno == EINTR)
	    continue;
	  perror ("polling in flush_pending");
	  return FALSE;
	}
      if (fds.revents & G_IO_IN)
	{
	  num_read = read (channel->poll_fd.fd, buf, sizeof (buf));
	  if (num_read == 0)
	    {
	      g_message ("flush_pending_requests: " "eof from server");
	      return FALSE;
	    }
	  if (num_read < 0)
	    {
	      if (errno == EINTR)
		continue;
	      if (errno == EWOULDBLOCK)
		{
		  g_warning ("non-blocking socket going into "
			     "busy loop (flush-pending)");
		  continue;
		}
	      perror ("reading doing synchronous command");
	      return FALSE;
	    }
	  gdam_buffer_append (&channel->incoming, buf, num_read);
	  if (!gdam_channel_maybe_dispatch_input (channel))
	    return FALSE;
	}
      if (fds.revents & G_IO_OUT)
	if (!gdam_buffer_write_one (&channel->outgoing, fds.fd))
	  return FALSE;
    }
  return TRUE;
}
	
static void poll_for_readwrite(int fd)
{
  GPollFD fds;
  fds.fd = fd;
  fds.events = G_IO_IN | G_IO_OUT;
  gdam_poll (&fds, 1, -1);
}

gboolean gdam_channel_process_request(GdamChannel* channel, 
                                      GdamRequest* request)
{
  static gboolean processing_request = FALSE;

  gboolean rv;

  if (processing_request)
    {
      g_error ("process_request was reinvoked from a request "
	       "terminate handler");
      return FALSE;
    }

  processing_request = TRUE;
  if (channel->is_connecting)
    {
      gdam_debug ("[process_request] connecting...");
      poll_for_readwrite (channel->poll_fd.fd);
      if (!gdam_connect_finish (channel->poll_fd.fd, TRUE))
	{
	  if (channel->quit_main)
	    exit (0);
	  return FALSE;
	}
      channel->is_connecting = 0;
    }
  if (channel->auth_handler)
    gdam_channel_append_request (channel, request);
  else
    gdam_channel_prepend_request (channel, request);
  rv = flush_pending (channel);
  processing_request = FALSE;
  return rv;
}

void gdam_channel_set_launch_server       (GdamChannel* channel,
                                           GdamChannelServerLaunch mode)
{
  g_return_if_fail (mode != GDAM_CHANNEL_SERVER_LAUNCH_DEFER);
  switch (mode)
    {
    case GDAM_CHANNEL_SERVER_LAUNCH_ABORT:
      g_message ("Sorry couldn't open server... quitting");
      exit (1);
      break;
    case GDAM_CHANNEL_SERVER_LAUNCH_INSECURE:
      channel->poll_fd.fd = launch_gdam_server (FALSE);
      break;
    case GDAM_CHANNEL_SERVER_LAUNCH_SECURE:
      channel->poll_fd.fd = launch_gdam_server (TRUE);
      break;
    default:
      g_assert_not_reached ();
      break;
    }
  channel->launch_server_deferred = 0;
}

/* chat */
guint        gdam_channel_trap_chat_messages      (GdamChannel *channel,
						   const char  *username,
						   const char  *command,
						   GdamChannelChatFunc func,
						   gpointer     func_data,
						   GDestroyNotify destroy)
{
  GHook *n = g_hook_alloc (&channel->chat_handlers);
  char **requested = (char **) (n + 1);
  n->func = (GHookFunc) func;
  n->data = func_data;
  requested[0] = g_strdup (username);
  requested[1] = g_strdup (command);
  n->destroy = destroy;
  g_hook_prepend (&channel->chat_handlers, n);
  return n->hook_id;
}

void         gdam_channel_untrap_chat_messages    (GdamChannel *channel,
						   guint        trap_id)
{
  GHook *hook = g_hook_get (&channel->chat_handlers, trap_id);
  char **requested;
  g_return_if_fail (hook != NULL);
  requested = (char **) (hook + 1);
  g_free (requested[0]);
  g_free (requested[1]);
  g_hook_destroy_link (&channel->chat_handlers, hook);
}

static void  
chat_success (gpointer recipient)
{
  g_message ("sent chat message to %s", (char*)recipient);
  g_free (recipient);
}
static void
chat_failure (gpointer recipient)
{
  g_message ("no user named %s", (char*)recipient);
  g_free (recipient);
}

/* command cannot have spaces... */
void
gdam_channel_send_chat_message     (GdamChannel *channel,
				    const char  *recipient,
				    const char  *command,
				    const char  *text)
{
  GdamRequest *request;
  char *message, *my_recipient;
  /* XXX: is this symmetric to chat unpacking? */
  message = g_strdup_printf ("%s %s", command, text);
  my_recipient = g_strdup (recipient);
  request = gdam_request_chat_message (my_recipient, 
      				       message,
				       chat_success,
				       chat_failure,
				       my_recipient);
  gdam_channel_process_request (channel, request);
  g_free (message);
}
