/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamrequest.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamsample.h>
#include <gdam/common/gdamarg.h>
#include <gdam/common/gdamalloca.h>

/* Tiny helper functions... */

void gdam_request_get_int(gint id, gpointer user_data)
{
  ((GdamRequestGetIntStruct *) user_data)->was_success = TRUE;
  ((GdamRequestGetIntStruct *) user_data)->rv = id;
}

void gdam_request_get_2int(gint id, gint id2, gpointer user_data)
{
  ((GdamRequestGet2IntStruct *) user_data)->was_success = TRUE;
  ((GdamRequestGet2IntStruct *) user_data)->rv[0] = id;
  ((GdamRequestGet2IntStruct *) user_data)->rv[1] = id2;
}


/* A GdamRequestGotTimeFunc that fills a GdamRequestGetIntStruct*
 * as its user data. */
void gdam_request_get_time(GdamTime time, gpointer user_data)
{
  ((GdamRequestGetTimeStruct *) user_data)->was_success = TRUE;
  ((GdamRequestGetTimeStruct *) user_data)->rv = time;
}

void gdam_request_get_data(int length, gpointer data, gpointer user_data)
{
  ((GdamRequestGetDataStruct *) user_data)->was_success = TRUE;
  ((GdamRequestGetDataStruct *) user_data)->data = data;
  ((GdamRequestGetDataStruct *) user_data)->length = length;
}
	

void gdam_request_get_failure(gpointer user_data)
{
  *(gboolean *) user_data = FALSE;
}

/* A ghookfunc to be used to mark failure in a gboolean*. */
void     gdam_request_get_success(gpointer user_data)
{
  *(gboolean *) user_data = TRUE;
}


/* build up requests (and their marshals) */

GdamRequest* gdam_request_new_generic(int type)
{
  GdamRequest *request = g_new (GdamRequest, 1);
  gdam_buffer_construct (&request->request_command);
  request->marshal = NULL;
  gdam_buffer_append_int (&request->request_command, type);
  request->num_subs = 0;
  request->subs = NULL;
  request->local_id = 0;
  return request;
}

static gboolean
gdam_marshal_generic(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_SUCCESS)
    {
      if (request->success)
	(*request->success) (request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	(*request->failure) (request->user_data);
    }
  else
    {
      g_warning ("response to `[generic_request]' was of "
		 "unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}

static gboolean
gdam_marshal_2int_on_success(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_2INT_SUCCESS
      || type == GDAM_RESPONSE_CACHING_STARTED)
    {
      int int_a, int_b;
      if (!gdam_buffer_parse_int (response, &int_a)
	  || !gdam_buffer_parse_int (response, &int_b))
	return FALSE;
      if (request->success)
	((GdamRequestGot2IntFunc) (request->success)) (int_a,
						       int_b,
						       request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	(*request->failure) (request->user_data);
    }
  else
    {
      g_warning ("response to two integer request was of "
		 "unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}

static gboolean
gdam_marshal_time_on_success(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_TIME_SUCCESS)
    {
      GdamTime time;
      if (!gdam_buffer_parse_time (response, &time))
	return FALSE;
      if (request->success)
	((GdamRequestGotTimeFunc) (request->success)) (time,
						       request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	(*request->failure) (request->user_data);
    }
  else
    {
      return FALSE;
    }
  return TRUE;
}

static gboolean
gdam_marshal_get_args(GdamRequest* request, GdamBuffer* response)
{
  int type, num_args;
  GdamArg *args;
  if (!gdam_buffer_parse_int (response, &type))
    {
      gdam_debug ("gdam_marshal_get_args: error getting type");
      return FALSE;
    }
  if (type == GDAM_RESPONSE_FAILED)
    {
      gdam_debug ("gdam_marshal_get_args: failure response");
      if (request->failure)
	(*request->failure) (request->user_data);
      return TRUE;
    }
  if (type != GDAM_RESPONSE_ARG_SUCCESS)
    {
      gdam_debug ("gdam_marshal_get_args: wrong response type");
      return FALSE;
    }
  if (!gdam_buffer_parse_int (response, &num_args))
    {
      gdam_debug ("gdam_marshal_get_args: error getting # of args");
      return FALSE;
    }
  //gdam_debug("num args is %d", num_args);
  args = gdam_arg_parse_from_buffer (response, num_args);
  if (num_args > 0 && args == NULL)
    {
      gdam_debug ("gdam_marshal_get_args: error parsing args");
      return FALSE;
    }
  if (request->success == NULL)
    return TRUE;

  (*(GdamRequestGotArgsFunc) (request->success)) (num_args,
						  args, request->user_data);
  gdam_arg_free_array (num_args, args);
  return TRUE;
}

/*
 *                      _             _       _
 *   ___ _ __ ___  __ _| |_ ___   ___(_)_ __ | | __
 *  / __| '__/ _ \/ _` | __/ _ \ / __| | '_ \| |/ /
 * | (__| | |  __/ (_| | ||  __/ \__ \ | | | |   <
 *  \___|_|  \___|\__,_|\__\___| |___/_|_| |_|_|\_\
 * 
 */

static gboolean
gdam_marshal_int_on_success(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_INT_SUCCESS)
    {
      int source_id;
      if (!gdam_buffer_parse_int (response, &source_id))
	return FALSE;
      if (request->success)
	((GdamRequestGotIntFunc) (request->success)) (source_id,
						      request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to `create_sink' was of " "unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}

GdamRequest*   gdam_request_create_sink  (const char*             device_name,
                                          GdamRequestGotIntFunc   callback,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_OPEN_DEVICE);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_int_on_success;
  gdam_buffer_append_string0 (&request->request_command, device_name);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *   __               _       _       _
 *  / _| ___  ___  __| |  ___(_)_ __ | | __
 * | |_ / _ \/ _ \/ _` | / __| | '_ \| |/ /
 * |  _|  __/  __/ (_| | \__ \ | | | |   <
 * |_|  \___|\___|\__,_| |___/_|_| |_|_|\_\
 * 
 */

GdamRequest*   gdam_request_feed_sink    (guint                   sink_id,
                                          guint                   source_id,
					  GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_FEED_DEVICE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, sink_id);
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *       _                      _       _
 *   ___| | ___  ___  ___   ___(_)_ __ | | __
 *  / __| |/ _ \/ __|/ _ \ / __| | '_ \| |/ /
 * | (__| | (_) \__ \  __/ \__ \ | | | |   <
 *  \___|_|\___/|___/\___| |___/_|_| |_|_|\_\
 * 
 */

GdamRequest*   gdam_request_close_sink   (guint                   sink_id,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_CLOSE_DEVICE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, sink_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *                                        _       _
 *  _ __ ___  ___  _ __   ___ _ __    ___(_)_ __ | | __
 * | '__/ _ \/ _ \| '_ \ / _ \ '_ \  / __| | '_ \| |/ /
 * | | |  __/ (_) | |_) |  __/ | | | \__ \ | | | |   <
 * |_|  \___|\___/| .__/ \___|_| |_| |___/_|_| |_|_|\_\
 *                |_|
 */
GdamRequest*   gdam_request_reopen_sink  (guint                   sink_id,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_REOPEN_DEVICE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, sink_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *                      _
 *   ___ _ __ ___  __ _| |_ ___   ___  ___  _   _ _ __ ___ ___
 *  / __| '__/ _ \/ _` | __/ _ \ / __|/ _ \| | | | '__/ __/ _ \
 * | (__| | |  __/ (_| | ||  __/ \__ \ (_) | |_| | | | (_|  __/
 *  \___|_|  \___|\__,_|\__\___| |___/\___/ \__,_|_|  \___\___|
 * 
 */
static gboolean
gdam_marshal_create_source(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_INT_SUCCESS)
    {
      int source_id;
      if (!gdam_buffer_parse_int (response, &source_id))
	return FALSE;
      if (request->success)
	((GdamRequestGotIntFunc) (request->success)) (source_id,
						      request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to `create_source' was of "
		 "unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}


/*
 *      _      _             _
 *   __| | ___| |_ __ _  ___| |__    ___  ___  _   _ _ __ ___ ___
 *  / _` |/ _ \ __/ _` |/ __| '_ \  / __|/ _ \| | | | '__/ __/ _ \
 * | (_| |  __/ || (_| | (__| | | | \__ \ (_) | |_| | | | (_|  __/
 *  \__,_|\___|\__\__,_|\___|_| |_| |___/\___/ \__,_|_|  \___\___|
 * 
 */
GdamRequest*   gdam_request_detach_source(guint                   source_id,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_DETACH_SOURCE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_seek_relative(guint                    source_id,
					  GdamTime                 when,
					  GHookFunc                success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_CONFIGURE_SOURCE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command,
			  GDAM_CFG_SOURCE_SEEK_RELATIVE);
  gdam_buffer_append_time (&request->request_command, when);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}


/*
 *            _     _   _                    _
 *   __ _  __| | __| | | |_ ___    _ __ ___ (_)_  __
 *  / _` |/ _` |/ _` | | __/ _ \  | '_ ` _ \| \ \/ /
 * | (_| | (_| | (_| | | || (_) | | | | | | | |>  <
 *  \__,_|\__,_|\__,_|  \__\___/  |_| |_| |_|_/_/\_\
 * 
 */
GdamRequest*   gdam_request_add_mix      (guint                   mix_id,
					  guint                   subsource_id,
					  GdamTime                when,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_CONFIGURE_SOURCE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, mix_id);
  gdam_buffer_append_int (&request->request_command, GDAM_CFG_MIX_ADD);
  gdam_buffer_append_int (&request->request_command, subsource_id);
  gdam_buffer_append_time (&request->request_command, when);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest* gdam_request_get_server_side_int(const char* namespac,
                                              const char* name,
                                              GdamRequestGotIntFunc callback,
					      GHookFunc failure,
					      gpointer user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_GET_SERVER_INT);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_int_on_success;
  gdam_buffer_append_string0 (&request->request_command, namespac);
  gdam_buffer_append_string0 (&request->request_command, name);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *             _                               _
 *   __ _  ___| |_   ___  ___  _   _ _ __   __| |  ___ _ __   ___  ___
 *  / _` |/ _ \ __| / __|/ _ \| | | | '_ \ / _` | / __| '_ \ / _ \/ __|
 * | (_| |  __/ |_  \__ \ (_) | |_| | | | | (_| | \__ \ |_) |  __/ (__
 *  \__, |\___|\__| |___/\___/ \__,_|_| |_|\__,_| |___/ .__/ \___|\___|
 *  |___/                                             |_|
 */
gboolean
gdam_marshal_get_data(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_DATA_SUCCESS)
    {
      int length;
      char *data;
      GdamRequestGotDataFunc got_data_func;
      if (!gdam_buffer_parse_int (response, &length))
	return FALSE;
      got_data_func = (GdamRequestGotDataFunc) (request->success);
      data = g_new (char, length);
      gdam_buffer_read (response, data, length);
      if (got_data_func)
	(*got_data_func) (length, data, request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to get-data request (eg get_sound_spec) "
		 "was of unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}

#if 0
GdamRequest*   gdam_request_get_sound_spec(int                    source_id,
                                           gboolean               recurse,
                                           GdamRequestGotDataFunc callback,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_GET_SOUND_SPEC);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_get_data;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, recurse);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest* gdam_request_set_sequencer (guint                   sequence_id,
					 int                     seq_len,
					 int                     num_els,
			    const GdamSoundSpecSequencerElement* elements,
                                         GHookFunc               success,
					 GHookFunc               failure,
					 gpointer                user_data)
{
  GdamRequest *request;
  int i;
  request = gdam_request_new_generic (GDAM_MESSAGE_CONFIGURE_SOURCE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, sequence_id);
  gdam_buffer_append_int (&request->request_command,
			  GDAM_CFG_SEQUENCE_CHANGE);
  gdam_buffer_append_int (&request->request_command, seq_len);
  gdam_buffer_append_int (&request->request_command, num_els);
  for (i = 0; i < num_els; i++)
    {
      const GdamSoundSpecSequencerElement *element = elements + i;
      gdam_buffer_append_int (&request->request_command, element->buffer_id);
      gdam_buffer_append_int (&request->request_command,
			      element->start_offset);
      gdam_buffer_append_int (&request->request_command,
			      element->resample_b16);
      gdam_buffer_append_int (&request->request_command, element->gain_b16);
      gdam_buffer_append_int (&request->request_command,
			      element->sample_offset);
      gdam_buffer_append_int (&request->request_command,
			      element->output_length);
    }
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
#endif

/*
 *      _             _                    _     _
 *  ___| |_ __ _ _ __| |_    ___ __ _  ___| |__ (_)_ __   __ _
 * / __| __/ _` | '__| __|  / __/ _` |/ __| '_ \| | '_ \ / _` |
 * \__ \ || (_| | |  | |_  | (_| (_| | (__| | | | | | | | (_| |
 * |___/\__\__,_|_|   \__|  \___\__,_|\___|_| |_|_|_| |_|\__, |
 *                                                       |___/
 */
GdamRequest*   gdam_request_start_caching (int                     source_id,
                                           GdamRequestGot2IntFunc  callback,
					   GHookFunc               failure,
					   gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_START_CACHING);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_2int_on_success;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest*   gdam_request_uncache       (int                     buffer_id,
                                           GHookFunc               success,
					   GHookFunc               failure,
					   gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_UNCACHE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, buffer_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}


gboolean
gdam_marshal_int_array(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_DATA_SUCCESS)
    {
      int length, i;
      gint32 *data;
      GdamRequestGotDataFunc got_data_func;
      if (!gdam_buffer_parse_int (response, &length))
	return FALSE;
      got_data_func = (GdamRequestGotDataFunc) (request->success);
      data = alloca (sizeof (int) * length);
      gdam_buffer_read (response, data, length * 4);
      for (i = 0; i < length; i++)
	data[i] = GINT32_FROM_BE (data[i]);

      /* TODO: fix this by conditionally copying `data' to
       *       an array of native ints. */
      g_assert (sizeof (int) == 4);

      if (got_data_func)
	(*got_data_func) (length, data, request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to get-data request (eg get_profile) "
		 "was of unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}


GdamRequest* gdam_request_get_buffer_profile(int                buffer_id,
                                             int                 profile_mode,
					     int                 start,
					     int                 length,
					     int                 num_output,
					     GdamRequestGetIntArrayFunc success,
					     GHookFunc           failure,
					     gpointer            user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_GET_CACHE_PROFILE);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_int_array;
  gdam_buffer_append_int (&request->request_command, buffer_id);
  gdam_buffer_append_int (&request->request_command, profile_mode);
  gdam_buffer_append_int (&request->request_command, start);
  gdam_buffer_append_int (&request->request_command, length);
  gdam_buffer_append_int (&request->request_command, num_output);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

static inline gint64 bpm_to_beat_period(gdouble bpm)
{
  gdouble rv = (double) (44100 << 6) / (bpm / 60.0);
  rv *= (GdamFixed_ONE >> 6);
  return (gint64) (rv);
}

void           gdam_request_helper_build_sync
                                          (int                    mobile_id,
                                           int                    fixed_id,
					   gdouble                mobile_bpm,
					   gdouble                mobile_start,
					   gdouble                fixed_bpm,
					   gdouble                fixed_start,
					   gint64*		  out)
{
  out[0] = mobile_id;
  out[1] = fixed_id;
  out[2] = bpm_to_beat_period (mobile_bpm);
  out[3] = (GdamTime) (44100.0 * mobile_start);
  out[4] = bpm_to_beat_period (fixed_bpm);
  out[5] = (GdamTime) (44100.0 * fixed_start);
}

GdamRequest*   gdam_request_synchronize   (int                    mobile_id,
                                           int                    fixed_id,
					   gdouble                mobile_bpm,
					   gdouble                mobile_start,
					   gdouble                fixed_bpm,
					   gdouble                fixed_start,
                                           GHookFunc              success,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  /* bpm => samplesb16/beat */

  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SYNCHRONIZE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;

  gdam_buffer_append_int (&request->request_command, mobile_id);
  gdam_buffer_append_int (&request->request_command, fixed_id);

  gdam_buffer_append_time (&request->request_command,
			   bpm_to_beat_period (mobile_bpm));
  gdam_buffer_append_time (&request->request_command,
			   (GdamTime) (44100.0 * mobile_start));

  gdam_buffer_append_time (&request->request_command,
			   bpm_to_beat_period (fixed_bpm));
  gdam_buffer_append_time (&request->request_command,
			   (GdamTime) (44100.0 * fixed_start));

  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_sync_preview  (int                    mobile_id,
                                           int                    fixed_id,
					   gdouble                mobile_bpm,
					   gdouble                mobile_start,
					   gdouble                fixed_bpm,
					   gdouble                fixed_start,
                                           GdamRequestGotTimeFunc success,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  /* bpm => samplesb16/beat */

  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SYNC_PREVIEW);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_time_on_success;

  gdam_buffer_append_int (&request->request_command, mobile_id);
  gdam_buffer_append_int (&request->request_command, fixed_id);

  gdam_buffer_append_time (&request->request_command,
			   bpm_to_beat_period (mobile_bpm));
  gdam_buffer_append_time (&request->request_command,
			   (GdamTime) (44100.0 * mobile_start));

  gdam_buffer_append_time (&request->request_command,
			   bpm_to_beat_period (fixed_bpm));
  gdam_buffer_append_time (&request->request_command,
			   (GdamTime) (44100.0 * fixed_start));

  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
/*
 *           _       _                _
 *  ___  ___| |_    | | _____   _____| |
 * / __|/ _ \ __|   | |/ _ \ \ / / _ \ |
 * \__ \  __/ |_    | |  __/\ V /  __/ |
 * |___/\___|\__|___|_|\___| \_/ \___|_|
 *             |_____|
 * set the logging level.
 */
GdamRequest*   gdam_request_set_log_level(int                      flags,
					  GHookFunc                success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SET_LOG_LEVEL);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, flags);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

/*
 *                      _
 *   ___ _ __ ___  __ _| |_ ___  
 *  / __| '__/ _ \/ _` | __/ _ \
 * | (__| | |  __/ (_| | ||  __/
 *  \___|_|  \___|\__,_|\__\___|
 * 
 */

GdamRequest* gdam_request_create_v       (int                     obj_type,
                                          const char*             type,
					  int                     local_id,
                                          int                     num_args,
					  GdamArg*                args,
					  GdamRequestGotIntFunc   success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  g_return_val_if_fail (local_id <= 0, NULL);
  request = gdam_request_new_generic (GDAM_MESSAGE_CREATE);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_create_source;
  request->local_id = local_id;
  gdam_buffer_append_int (&request->request_command, obj_type);
  gdam_buffer_append_string0 (&request->request_command, type);
  gdam_buffer_append_int (&request->request_command, local_id);
  gdam_buffer_append_int (&request->request_command, num_args);
  gdam_arg_write_to_buffer (&request->request_command, num_args, args);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest*   gdam_request_create       (int                     obj_type,
                                          const char*             type,
					  int                     local_id,
					  GdamRequestGotIntFunc   success,
					  GHookFunc               failure,
					  gpointer                user_data,
                                          int                     num_args,
					  ...)
{
  GdamRequest *request;
  va_list arg_list;
  GdamArg *args;
  if (num_args > 0)
    {
      va_start (arg_list, num_args);
      args = gdam_arg_parse_from_vararg (num_args, &arg_list);
      va_end (arg_list);
      if (!args)
	return NULL;
    }
  else
    {
      args = NULL;
    }
  request = gdam_request_create_v (obj_type, type, local_id,
				   num_args, args,
				   success, failure, user_data);
  gdam_arg_free_array (num_args, args);
  return request;
}

/*
 *                  _
 *  __ _ _ ___ __ _| |_ ___   ___ ___ _  _ _ _ __ ___
 * / _| '_/ -_) _` |  _/ -_) (_-</ _ \ || | '_/ _/ -_)
 * \__|_| \___\__,_|\__\___|_/__/\___/\_,_|_| \__\___|
 *                        |___|
 *
 *    [for backward compatibility]
 */
GdamRequest* gdam_request_create_source_v(const char*             type,
                                          int                     local_id,
                                          int                     num_args,
					  GdamArg*                args,
					  GdamRequestGotIntFunc   success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  return gdam_request_create_v (GDAM_PROTOCOL_SOURCE_ARG,
				type, local_id,
				num_args, args, success, failure, user_data);
}

GdamRequest*   gdam_request_create_source(const char*             type,
					  int                     local_id,
					  GdamRequestGotIntFunc   success,
					  GHookFunc               failure,
					  gpointer                user_data,
                                          int                     num_args,
					  ...)
{
  GdamRequest *request;
  va_list arg_list;
  GdamArg *args;
  if (num_args > 0)
    {
      va_start (arg_list, num_args);
      args = gdam_arg_parse_from_vararg (num_args, &arg_list);
      va_end (arg_list);
      if (!args)
	return NULL;
    }
  else
    {
      args = NULL;
    }
  request = gdam_request_create_v (GDAM_PROTOCOL_SOURCE_ARG,
				   type, local_id, num_args, args,
				   success, failure, user_data);
  gdam_arg_free_array (num_args, args);
  return request;
}


GdamRequest*   gdam_request_set_gen_args (int                      arena_id,
                                          int                      source_id,
					  int                      num_args,
                                          GdamArg*                 args,
                                          GHookFunc                success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SET_ARGS);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, num_args);
  gdam_arg_write_to_buffer (&request->request_command, num_args, args);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest*   gdam_request_set_args (int                      source_id,
				      int                      num_args,
                                      GdamArg*                 args,
                                      GHookFunc                success,
				      GHookFunc                failure,
				      gpointer                 user_data)
{
  return gdam_request_set_gen_args
    (GDAM_PROTOCOL_SOURCE_ARG,
     source_id, num_args, args, success, failure, user_data);
}

/*
 *             _
 *   __ _  ___| |_    __ _ _ __ __ _ ___
 *  / _` |/ _ \ __|  / _` | '__/ _` / __|
 * | (_| |  __/ |_  | (_| | | | (_| \__ \
 *  \__, |\___|\__|  \__,_|_|  \__, |___/
 *  |___/                      |___/
 */
GdamRequest*   gdam_request_get_gen_args (int                      arena_id,
                                          int                      source_id,
					  int                      num_args,
					  const char* const*       arg_names,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  int i;

  request = gdam_request_new_generic (GDAM_MESSAGE_GET_ARGS);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_get_args;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, num_args);
  for (i = 0; i < num_args; i++)
    gdam_buffer_append_string0 (&request->request_command, arg_names[i]);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest*   gdam_request_get_args     (int                      source_id,
					  int                      num_args,
					  const char* const*       arg_names,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  return gdam_request_get_gen_args
    (GDAM_PROTOCOL_SOURCE_ARG,
     source_id, num_args, arg_names, success, failure, user_data);
}

/*
 *             _           _ _
 *   __ _  ___| |_    __ _| | |   __ _ _ __ __ _ ___
 *  / _` |/ _ \ __|  / _` | | |  / _` | '__/ _` / __|
 * | (_| |  __/ |_  | (_| | | | | (_| | | | (_| \__ \
 *  \__, |\___|\__|  \__,_|_|_|  \__,_|_|  \__, |___/
 *  |___/                                  |___/
 */
GdamRequest*   gdam_request_get_all_gen_args
                                         (int                      arena_id,
                                          int                      source_id,
					  gboolean                 only_vital,
					  gboolean                 deprecated,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  guint flags = 0;
  request = gdam_request_new_generic (GDAM_MESSAGE_GET_ALL_ARGS);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_get_args;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, source_id);
  if (only_vital)
    flags |= GDAM_PROTOCOL_GET_ALL_ONLY_VITAL;
  if (deprecated)
    flags |= GDAM_PROTOCOL_GET_ALL_DEPRECATED;
  gdam_buffer_append_int (&request->request_command, flags);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_get_all_args (int                      source_id,
					  gboolean                 only_vital,
					  gboolean                 deprecated,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  return gdam_request_get_all_gen_args
    (GDAM_PROTOCOL_SOURCE_ARG,
     source_id, only_vital, deprecated, success, failure, user_data);
}

GdamRequest*   gdam_request_call_function(int                      arena_id,
                                          int                      source_id,
					  const char*              func_name,
					  int                      num_args,
					  GdamArg*                 args_in,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;

  request = gdam_request_new_generic (GDAM_MESSAGE_CALL_FUNCTION);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_get_args;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_string0 (&request->request_command, func_name);
  gdam_buffer_append_int (&request->request_command, num_args);
  gdam_arg_write_to_buffer (&request->request_command, num_args, args_in);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
GdamRequest*   gdam_request_call_class_function
                                         (const char*              class_name,
					  const char*              func_name,
					  int                      num_args,
					  GdamArg*                 args_in,
                                          GdamRequestGotArgsFunc   success,
					  GHookFunc                failure,
					  gpointer                 user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_CALL_CLASS_FUNCTION);
  request->user_data = user_data;
  request->success = (GHookFunc) success;
  request->failure = failure;
  request->marshal = gdam_marshal_get_args;
  gdam_buffer_append_string0 (&request->request_command, class_name);
  gdam_buffer_append_string0 (&request->request_command, func_name);
  gdam_buffer_append_int (&request->request_command, num_args);
  gdam_arg_write_to_buffer (&request->request_command, num_args, args_in);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
/*
 *                   _
 *  _ __   __ _  ___| | ____ _  __ _  ___
 * | '_ \ / _` |/ __| |/ / _` |/ _` |/ _ \
 * | |_) | (_| | (__|   < (_| | (_| |  __/
 * | .__/ \__,_|\___|_|\_\__,_|\__, |\___|
 * |_|                         |___/
 */
static void package_cleanup_subs(GdamRequest* package)
{
  int i;
  for (i = 0; i < package->num_subs; i++)
    {
      GdamRequest *sub = (GdamRequest *) (package->subs[i]);
      gdam_request_destroy (sub);
    }
  g_free (package->subs);
}

static gboolean
gdam_marshal_package(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    {
      package_cleanup_subs (request);
      return FALSE;
    }
  if (type == GDAM_RESPONSE_PACKAGE)
    {
      int i;
      int num_responses;
      gboolean success = TRUE;
      if (!gdam_buffer_parse_int (response, &num_responses))
	{
	  package_cleanup_subs (request);
	  return FALSE;
	}
      if (num_responses != request->num_subs)
	{
	  gdam_debug ("num_responses to package different than "
		      "number of requests");
	}
      for (i = 0; i < num_responses; i++)
	{
	  int length;
	  GdamBuffer tmp;
	  GdamRequest *sub;
	  gboolean success = TRUE;
	  if (!gdam_buffer_peek_int (response, &length))
	    {
	      package_cleanup_subs (request);
	      return FALSE;
	    }
	  gdam_buffer_construct (&tmp);
	  gdam_buffer_send_buffer (&tmp, response, length, TRUE);
	  gdam_buffer_parse_int (&tmp, &length);
	  sub = (GdamRequest *) (request->subs[i]);
	  if (!sub->marshal (sub, &tmp))
	    {
	      gdam_buffer_destruct (&tmp);
	      success = FALSE;
	    }
	  gdam_buffer_destruct (&tmp);
	}
      if (!success)
	{
	  package_cleanup_subs (request);
	  return FALSE;
	}
      if (request->success)
	((GdamRequestGotIntFunc) (request->success))
	  (num_responses, request->user_data);
    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to `PACKAGE' was of " "unknown type %d", type);
      package_cleanup_subs (request);
      return FALSE;
    }
  package_cleanup_subs (request);
  return TRUE;
}

/* Takes a GPtrArray of GdamRequests and makes it into a single
 * package request. */
GdamRequest*   gdam_request_package_v    (int                     num_requests,
					  GdamRequest**           requests,
                                          GdamRequestGotIntFunc   callback,
					  GHookFunc               failure,
                                          gpointer                user_data)
{
  int i;
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_PACKAGE);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_package;
  request->num_subs = num_requests;
  request->subs = (gpointer *) requests;
  gdam_buffer_append_int (&request->request_command, num_requests);
  for (i = 0; i < num_requests; i++)
    gdam_buffer_append_buffer (&request->request_command,
			       &(requests[i]->request_command), TRUE);
  gdam_buffer_prepend_length (&request->request_command);
  return request;

}

void gdam_request_destroy(GdamRequest* request)
{
  gdam_buffer_destruct (&request->request_command);
  g_free (request);
}

GdamRequest*   gdam_request_subscribe_source_ended
                                          (int                    source_id,
                                           GdamRequestGotIntFunc  callback,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SUBSCRIBE_SOURCE);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_int_on_success;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, GDAM_SUBSCRIPTION_ENDED);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_subscribe_source_timer
                                          (int                    source_id,
                                           int                    period,
                                           GdamRequestGotIntFunc  callback,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_SUBSCRIBE_SOURCE);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_int_on_success;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, GDAM_SUBSCRIPTION_TIMER);
  gdam_buffer_append_int (&request->request_command, period);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_unsubscribe   (int                    subscription,
                                           GHookFunc              success,
					   GHookFunc              failure,
					   gpointer               user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_UNSUBSCRIBE);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, subscription);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_ref          (guint                   arena_id,
					  guint                   object_id,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_REF);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, object_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest*   gdam_request_unref        (guint                   arena_id,
					  guint                   object_id,
                                          GHookFunc               success,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;
  request = gdam_request_new_generic (GDAM_MESSAGE_UNREF);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;
  gdam_buffer_append_int (&request->request_command, arena_id);
  gdam_buffer_append_int (&request->request_command, object_id);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}


GdamRequest*   gdam_request_source_time  (guint                   source_id,
                                          guint                   sub_id,
                                          GdamRequestGotTimeFunc  callback,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;

  g_return_val_if_fail (sub_id == 0, NULL);

  request = gdam_request_new_generic (GDAM_MESSAGE_CONFIGURE_SOURCE);
  request->user_data = user_data;
  request->success = (GHookFunc) callback;
  request->failure = failure;
  request->marshal = gdam_marshal_time_on_success;
  gdam_buffer_append_int (&request->request_command, source_id);
  gdam_buffer_append_int (&request->request_command, GDAM_CFG_GET_TIME);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest *
gdam_request_chat_message  (const char             *username,
			    const char             *message,
			    GHookFunc               success,
			    GHookFunc               failure,
			    gpointer                user_data)
{
  GdamRequest *request;
  g_return_val_if_fail (username != NULL && message != NULL, NULL);
  g_return_val_if_fail (strchr (message, ' ') != NULL, NULL);
  request = gdam_request_new_generic (GDAM_MESSAGE_CHAT);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;;
  gdam_buffer_append_string0 (&request->request_command, username);
  gdam_buffer_append_string0 (&request->request_command, message);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}

GdamRequest *
gdam_request_set_username  (const char             *username,
			    GHookFunc               success,
			    GHookFunc               failure,
			    gpointer                user_data)
{
  GdamRequest *request;
  g_return_val_if_fail (username != NULL, NULL);
  request = gdam_request_new_generic (GDAM_MESSAGE_SET_USERNAME);
  request->user_data = user_data;
  request->success = success;
  request->failure = failure;
  request->marshal = gdam_marshal_generic;;
  gdam_buffer_append_string0 (&request->request_command, username);
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}


static gboolean
gdam_marshal_str_array_on_success(GdamRequest* request, GdamBuffer* response)
{
  int type;
  if (!gdam_buffer_parse_int (response, &type))
    return FALSE;
  if (type == GDAM_RESPONSE_STRING_ARRAY_SUCCESS)
    {
      int num_strings;
      int i;
      char **strings;
      if (!gdam_buffer_parse_int (response, &num_strings))
	return FALSE;
      strings = g_new (char *, num_strings + 1);
      for (i = 0; i < num_strings; i++)
	{
	  strings[i] = gdam_buffer_parse_string0 (response);
	  if (strings[i] == NULL)
	    {
	      int j;
	      for (j = 0; j < i; j++)
		g_free (strings[j]);
	      g_free (strings);
	      return FALSE;
	    }
	}
      strings[num_strings] = NULL;
      if (request->success)
	(*(GdamRequestGotStringArrayFunc) (request->success))
	  (num_strings, strings, request->user_data);
      g_strfreev (strings);

    }
  else if (type == GDAM_RESPONSE_FAILED)
    {
      if (request->failure)
	request->failure (request->user_data);
    }
  else
    {
      g_warning ("response to `get_strings' was of " "unknown type %d", type);
      return FALSE;
    }
  return TRUE;
}


GdamRequest*   gdam_request_get_named_buffers
                                         (GdamRequestGotStringArrayFunc func,
					  GHookFunc               failure,
					  gpointer                user_data)
{
  GdamRequest *request;

  request = gdam_request_new_generic (GDAM_MESSAGE_GET_NAMED_BUFFERS);
  request->user_data = user_data;
  request->success = (GHookFunc) func;
  request->failure = failure;
  request->marshal = gdam_marshal_str_array_on_success;
  gdam_buffer_prepend_length (&request->request_command);
  return request;
}
