/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuffer.h"
#include "gdamalloca.h"
#include <glib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

struct _GdamBufferFragment
{
  int			alloc_length;
  int			data_start;
  int			data_length;
  GdamBufferFragment   *next;
  gpointer              data;
  gpointer              free_data;
  GDestroyNotify        free_func;

  unsigned              read_only : 1;
};
#define GDAM_BUFFER_FRAGMENT_DATA_START(fragment)	\
	((char*)((fragment)->data)) + (fragment)->data_start
#define GDAM_BUFFER_FRAGMENT_DATA_END(fragment)	\
	(GDAM_BUFFER_FRAGMENT_DATA_START(fragment) + (fragment)->data_length)

#define FRAGMENT_MULTIPLE	4096

 
void gdam_buffer_construct(GdamBuffer* buffer)
{
  buffer->first = buffer->last = 0;
  buffer->size = 0;
}

void gdam_buffer_destruct(GdamBuffer* buffer)
{
  GdamBufferFragment *at = buffer->first;
  while (at)
    {
      GdamBufferFragment *next = at->next;
      g_free (at);
      at = next;
    }
  buffer->first = buffer->last = NULL;
  buffer->size = 0;
}

GdamBuffer* gdam_buffer_new()
{
  GdamBuffer *buffer = g_new (GdamBuffer, 1);
  buffer->first = buffer->last = 0;
  buffer->size = 0;
  return buffer;
}
void gdam_buffer_destroy(GdamBuffer* buffer)
{
  if (buffer)
    {
      gdam_buffer_destruct (buffer);
      g_free (buffer);
    }
}

int gdam_buffer_read(GdamBuffer* buffer, void* out, int out_max)
{
  int rv = 0;
  while (buffer->first && out_max > 0)
    {
      GdamBufferFragment *at = buffer->first;
      GdamBufferFragment *next = at->next;
      if (at->data_length <= out_max)
	{
	  memcpy (out, GDAM_BUFFER_FRAGMENT_DATA_START (at), at->data_length);
	  out = (char *) out + at->data_length;
	  out_max -= at->data_length;
	  rv += at->data_length;
	  g_free (at);
	  if ((buffer->first = next) == 0)
	    buffer->last = 0;
	}
      else
	{
	  memcpy (out, GDAM_BUFFER_FRAGMENT_DATA_START (at), out_max);
	  at->data_length -= out_max;
	  at->data_start += out_max;
	  rv += out_max;
	  break;
	}
    }
  buffer->size -= rv;
  return rv;
}
int gdam_buffer_peek(const GdamBuffer* buffer, void* out, int out_max)
{
  int rv = 0;
  GdamBufferFragment *at = buffer->first;
  while (at && out_max > 0)
    {
      if (at->data_length <= out_max)
	{
	  memcpy (out, GDAM_BUFFER_FRAGMENT_DATA_START (at), at->data_length);
	  out = (char *) out + at->data_length;
	  out_max -= at->data_length;
	  rv += at->data_length;
	  at = at->next;
	}
      else
	{
	  memcpy (out, GDAM_BUFFER_FRAGMENT_DATA_START (at), out_max);
	  return rv + out_max;
	}
    }
  return rv;
}
static GdamBufferFragment* gdam_buffer_fragment_new(int len)
{
  GdamBufferFragment *fragment;
  len += sizeof (GdamBufferFragment);
  len += FRAGMENT_MULTIPLE - 1;
  len /= FRAGMENT_MULTIPLE;
  len *= FRAGMENT_MULTIPLE;
  fragment = (GdamBufferFragment *) g_malloc (len);
  fragment->alloc_length = len - sizeof (GdamBufferFragment);
  fragment->data_start = 0;
  fragment->data_length = 0;
  fragment->next = 0;
  fragment->data = fragment + 1;
  fragment->free_data = NULL;
  fragment->free_func = NULL;
  fragment->read_only = 0;
  return fragment;
}

void gdam_buffer_append(GdamBuffer* buffer, const void* in, int in_len)
{
  buffer->size += in_len;
  while (in_len > 0)
    {
      if (buffer->last)
	{
	  GdamBufferFragment *last = buffer->last;
	  int extra = last->alloc_length;
	  extra -= last->data_length;
	  extra -= last->data_start;
	  if (extra >= in_len)
	    {
	      memcpy (GDAM_BUFFER_FRAGMENT_DATA_END (last), in, in_len);
	      last->data_length += in_len;
	      return;
	    }
	  else
	    {
	      memcpy (GDAM_BUFFER_FRAGMENT_DATA_END (last), in, extra);
	      in_len -= extra;
	      in = (const char *) in + extra;
	      last->data_length += extra;
	    }
	}
      if (in_len <= 0)
	return;
      {
	GdamBufferFragment *new_fragment;
	new_fragment = gdam_buffer_fragment_new (in_len);
	if (buffer->last)
	  buffer->last->next = new_fragment;
	else
	  {
	    buffer->first = new_fragment;
	    /* HACK:  just put a little prepadding
	     * in the first buffer, so that prepend_int
	     * can go w/o allocating a new buffer. */
	    new_fragment->data_start = 8;

	    /* And this works safely b/c
	     * even if new_fragment is too small,
	     * the if statements above will detect
	     * it and loop again... */
	  }
	buffer->last = new_fragment;
      }
    }
}

void gdam_buffer_append_zeroes(GdamBuffer* buffer, int num)
{
  char *buf;
  buf = alloca (num);
  bzero (buf, num);
  gdam_buffer_append (buffer, buf, num);
}

#if 0				/* TODO:  see if this is better. */
void gdam_buffer_prepend(GdamBuffer* buffer, const void* in, int in_len)
{
  GdamBufferFragment *n;
  if (buffer->first && buffer->first->data_start > 0)
    {
      GdamBufferFragment *first = buffer->first;
      int clip = first->data_start;
      if (clip < in_len)
	{
	  memcpy ((char *) (first + 1),
		  (const char *) in + in_len - clip, clip);
	  first->data_start = 0;
	  first->data_length += clip;
	  in_len -= clip;
	}
      else
	{
	  memcpy (GDAM_BUFFER_FRAGMENT_DATA_START (first) - in_len,
		  in, in_len);
	  first->data_start -= in_len;
	  first->data_length += in_len;
	  return;
	}
    }
  n = gdam_buffer_fragment_new (in_len);
  n->data_start = n->alloc_length - in_len;
  memcpy (GDAM_BUFFER_FRAGMENT_DATA_START (n) - in_len, in, in_len);
  n->data_length = in_len;
  n->next = buffer->first;
  if (!buffer->last)
    buffer->last = n;
  buffer->first = n;
}
#endif

void gdam_buffer_prepend(GdamBuffer* buffer, const void* in, int in_len)
{
  GdamBufferFragment *first;
  int first_chunk = in_len;
  first = buffer->first;
  buffer->size += in_len;
  /* Try to squeeze in the front of the first
   * buffer. */
  if (first)
    {
      if (first->data_start < first_chunk)
	first_chunk = first->data_start;
      first->data_start -= first_chunk;
      first->data_length += first_chunk;
      memcpy (GDAM_BUFFER_FRAGMENT_DATA_START (first),
	      (const char *) in + in_len - first_chunk, first_chunk);
      in_len -= first_chunk;
      if (in_len <= 0)
	return;
    }

  /* Make a chunk to shove the rest of the data into. */
  first = gdam_buffer_fragment_new (in_len);
  first->data_start = first->alloc_length - in_len;
  memcpy (GDAM_BUFFER_FRAGMENT_DATA_START (first), in, in_len);
  first->data_length = in_len;
  first->next = buffer->first;
  if (!buffer->last)
    buffer->last = first;
  buffer->first = first;
}

int gdam_buffer_index_of(GdamBuffer* buffer, char to_find, int start_index)
{
  GdamBufferFragment *fragment;
  int rv = 0;
  for (fragment = buffer->first;
       fragment && start_index >= fragment->data_length;)
    {
      start_index -= fragment->data_length;
      rv += fragment->data_length;
      fragment = fragment->next;
    }
  if (!fragment)
    return -1;
  if (start_index)
    {
      void *found;
      char *data_start = GDAM_BUFFER_FRAGMENT_DATA_START (fragment);
      found = memchr (data_start + start_index, to_find,
		      fragment->data_length - start_index);
      if (found)
	{
	  rv += ((char *) found - data_start);
	  return rv;
	}
      rv += fragment->data_length;
      fragment = fragment->next;
      start_index = 0;
    }
  while (fragment)
    {
      void *found;
      char *data_start = GDAM_BUFFER_FRAGMENT_DATA_START (fragment);
      found = memchr (data_start, to_find, fragment->data_length);
      if (found)
	{
	  rv += ((char *) found - data_start);
	  return rv;
	}
      rv += fragment->data_length;
      fragment = fragment->next;
    }
  return -1;
}

int gdam_buffer_discard(GdamBuffer* buffer, int out_max)
{
  int rv = 0;
  while (buffer->first && out_max > 0)
    {
      GdamBufferFragment *at = buffer->first;
      GdamBufferFragment *next = at->next;
      if (at->data_length <= out_max)
	{
	  out_max -= at->data_length;
	  rv += at->data_length;
	  g_free (at);
	  if ((buffer->first = next) == 0)
	    buffer->last = 0;
	}
      else
	{
	  at->data_length -= out_max;
	  at->data_start += out_max;
	  rv += out_max;
	  break;
	}
    }
  buffer->size -= rv;
  return rv;
}

void     gdam_buffer_append_alloced (GdamBuffer        *buffer,
                                     gconstpointer      data,
                                     int                length,
                                     gpointer           free_data,
                                     GDestroyNotify     free_func)
{
  GdamBufferFragment *fragment;
  fragment = g_new (GdamBufferFragment, 1);
  fragment->alloc_length = length;
  fragment->data_start = 0;
  fragment->data_length = length;
  fragment->next = NULL;
  fragment->data = (gpointer) data;
  fragment->free_data = free_data;
  fragment->free_func = free_func;
  fragment->read_only = 1;
  if (buffer->first == NULL)
    {
      buffer->first = buffer->last = fragment;
      g_assert (buffer->size == 0);
      buffer->size = length;
    }
  else
    {
      buffer->last->next = fragment;
      buffer->last = fragment;
      buffer->size += length;
    }
}


#define MAX_IOVECTOR	64

gboolean gdam_buffer_write_one(GdamBuffer* buffer, int fd)
{
  int nfrag = 0;
  int nwritten;
  GdamBufferFragment *fragment;
  struct iovec *iov;
  for (fragment = buffer->first;
       fragment && nfrag < MAX_IOVECTOR; fragment = fragment->next, nfrag++)
    ;
  if (nfrag == 0)
    return TRUE;
  iov = (struct iovec *) alloca (sizeof (struct iovec) * nfrag);
  nfrag = 0;
  for (fragment = buffer->first;
       fragment && nfrag < MAX_IOVECTOR; fragment = fragment->next, nfrag++)
    {
      iov[nfrag].iov_base = GDAM_BUFFER_FRAGMENT_DATA_START (fragment);
      iov[nfrag].iov_len = fragment->data_length;
    }
retry_writev:
  nwritten = writev (fd, iov, nfrag);
  if (nwritten <= 0)
    {
      if (errno == EINTR)
	goto retry_writev;
      perror ("writev");
      return FALSE;
    }
  g_return_val_if_fail (nwritten > 0, FALSE);
  gdam_buffer_discard (buffer, nwritten);
  return TRUE;
}
gboolean gdam_buffer_read_one(GdamBuffer* buffer, int fd)
{
  int num_read;
  char buf[16 * 1024];
  num_read = read (fd, buf, sizeof (buf));
  if (num_read < 0)
    {
      if (errno == EINTR || errno == EWOULDBLOCK)
	return TRUE;
      g_warning ("gdam_buffer_read_one:  %s", strerror (errno));
      return FALSE;
    }
  if (num_read == 0)
    return FALSE;
  gdam_buffer_append (buffer, buf, num_read);
  return TRUE;
}

void gdam_buffer_append_buffer(GdamBuffer* dst,
                               GdamBuffer* src,
			       gboolean drain)
{
  if (!src->first)
    return;

  if (!drain)
    {
      GdamBufferFragment *at;
      for (at = src->first; at; at = at->next)
	{
	  gdam_buffer_append (dst,
			      GDAM_BUFFER_FRAGMENT_DATA_START (at),
			      at->data_length);
	}
    }
  else
    {
      if (dst->last)
	{
	  dst->last->next = src->first;
	  dst->last = src->last;
	  dst->size += src->size;
	  src->first = src->last = NULL;
	  src->size = 0;
	}
      else
	{
	  dst->first = src->first;
	  dst->last = src->last;
	  dst->size = src->size;
	  src->first = src->last = NULL;
	  src->size = 0;
	}
    }
}

void gdam_buffer_send_buffer(GdamBuffer*    buffer,
                             GdamBuffer*    source,
			     int            length,
			     gboolean       drain)
{
  GdamBufferFragment *at = source->first;
  int copied = 0;
  while (at && length > 0)
    {
      int to_copy = at->data_length;
      if (to_copy > length)
	to_copy = length;
      gdam_buffer_append (buffer,
			  GDAM_BUFFER_FRAGMENT_DATA_START (at), to_copy);
      copied += to_copy;
      length -= to_copy;
      at = at->next;
    }
  if (drain)
    gdam_buffer_discard (source, copied);
}

/*
 * Portable packing...
 */

#define INT_LENGTH 4
static inline void int2chararray(char* out, int i)
{
  gint32 gi = (gint32) i;
  gint32 swapped_gi = GINT32_TO_BE (gi);
  memcpy (out, &swapped_gi, INT_LENGTH);
}
static inline int chararray2int(const char* in)
{
  gint32 gi = *(const gint32 *) in;
  return (gint) GINT32_FROM_BE (gi);
}
#define TIME_LENGTH 8
static inline void time2chararray(char* out, GdamTime t)
{
  gint64 gi = (gint64) t;
  gint64 swapped_gi = GINT64_TO_BE (gi);
  memcpy (out, &swapped_gi, TIME_LENGTH);
}
static inline GdamTime chararray2time(const char* in)
{
  gint64 gi = *(const gint64 *) in;
  return (gint64) GINT64_FROM_BE (gi);
}
#define INT64_LENGTH 8
static inline void int642chararray(char* out, gint64 t)
{
  gint64 gi = (gint64) t;
  gint64 swapped_gi = GINT64_TO_BE (gi);
  memcpy (out, &swapped_gi, INT64_LENGTH);
}
static inline gint64 chararray2int64(const char* in)
{
	gint64 gi = *(const gint64*)in;
	return (gint64) GINT64_FROM_BE(gi);
}

#define DOUBLE_LENGTH sizeof(guint64)
static inline void double2chararray(char* out, double t)
{
  guint64 tmp = * (guint64 *) &t;
  * (guint64 *) out = GUINT64_TO_BE (tmp);
}
static inline double chararray2double(const char* in)
{
  guint64 tmp0 = * (const guint64 *) in;
  guint64 tmp1 = GUINT64_FROM_BE (tmp0);
  return * (gdouble *) &tmp1;
}

GdamBuffer* gdam_buffer_package(int prefix, int n, ...)
{
  GdamBuffer **v;
  va_list args;
  int i;
  if (n == 0)
    return gdam_buffer_package_v (prefix, 0, NULL);
  v = (GdamBuffer **) alloca (sizeof (GdamBuffer *) * n);
  va_start (args, n);
  for (i = 0; i < n; i++)
    v[i] = va_arg (args, GdamBuffer *);
  va_end (args);
  return gdam_buffer_package_v (prefix, n, v);
}

GdamBuffer* gdam_buffer_package_v(int prefix, int n, GdamBuffer** bufs)
{
  int total_length = INT_LENGTH * 2;	/* `prefix' + length. */
  int i;
  GdamBuffer *rv;
  for (i = 0; i < n; i++)
    total_length += bufs[i]->size;
  rv = gdam_buffer_new ();
  gdam_buffer_append_int (rv, total_length);
  gdam_buffer_append_int (rv, prefix);
  for (i = 0; i < n; i++)
    gdam_buffer_append_buffer (rv, bufs[i], TRUE);
  return rv;
}

void gdam_buffer_prepend_length(GdamBuffer* buffer)
{
  unsigned char i[INT_LENGTH];
  int len = buffer->size + INT_LENGTH;
  int2chararray (i, len);
  gdam_buffer_prepend (buffer, i, INT_LENGTH);
}
void gdam_buffer_append_string(GdamBuffer* buffer, const char* string)
{
  int len;
  len = string == NULL ? 0 : strlen (string);
  gdam_buffer_append (buffer, string, len);
}

void gdam_buffer_append_string0(GdamBuffer* buffer, const char* string)
{
  int len;
  g_return_if_fail (buffer != NULL);
  if (string == NULL)
    {
      string = "";
      len = 1;
    }
  else
    len = strlen (string) + 1;
  gdam_buffer_append (buffer, string, len);
}

char* gdam_buffer_parse_string0(GdamBuffer* buffer)
{
  int len = gdam_buffer_index_of (buffer, '\0', 0);
  char *rv;
  if (len < 0)
    return NULL;
  len++;			/* include NUL */
  rv = (char *) g_malloc (len);
  gdam_buffer_read (buffer, rv, len);
  return rv;
}
void gdam_buffer_append_int(GdamBuffer* buffer, int val)
{
  char i[INT_LENGTH];
  int2chararray (i, val);
  gdam_buffer_append (buffer, i, INT_LENGTH);
}
void gdam_buffer_append_time(GdamBuffer* buffer, GdamTime val)
{
  char t[TIME_LENGTH];
  time2chararray (t, val);
  gdam_buffer_append (buffer, t, TIME_LENGTH);
}
void gdam_buffer_append_int64(GdamBuffer* buffer, gint64 val)
{
  char i[INT64_LENGTH];
  int642chararray (i, val);
  gdam_buffer_append (buffer, i, INT64_LENGTH);
}
void gdam_buffer_append_double(GdamBuffer* buffer, gdouble val)
{
  char i[DOUBLE_LENGTH];
  double2chararray (i, val);
  gdam_buffer_append (buffer, i, DOUBLE_LENGTH);
}


gboolean gdam_buffer_parse_int(GdamBuffer* buffer, int* returned_i)
{
  char i[INT_LENGTH];
  if (buffer->size < INT_LENGTH)
    return FALSE;
  gdam_buffer_read (buffer, i, INT_LENGTH);
  *returned_i = chararray2int (i);
  return TRUE;
}
gboolean gdam_buffer_parse_time(GdamBuffer* buffer, GdamTime* returned_t)
{
  char t[TIME_LENGTH];
  if (buffer->size < TIME_LENGTH)
    return FALSE;
  gdam_buffer_read (buffer, t, TIME_LENGTH);
  *returned_t = chararray2time (t);
  return TRUE;
}
gboolean gdam_buffer_parse_double(GdamBuffer* buffer, double* returned_t)
{
  char t[DOUBLE_LENGTH];
  if (buffer->size < (int) DOUBLE_LENGTH)
    return FALSE;
  gdam_buffer_read (buffer, t, DOUBLE_LENGTH);
  *returned_t = chararray2double (t);
  return TRUE;
}
gboolean gdam_buffer_parse_int64(GdamBuffer* buffer, gint64* returned_t)
{
  char t[INT64_LENGTH];
  if (buffer->size < INT64_LENGTH)
    return FALSE;
  gdam_buffer_read (buffer, t, INT64_LENGTH);
  *returned_t = chararray2int64 (t);
  return TRUE;
}
gboolean gdam_buffer_peek_int(GdamBuffer* buffer, int* returned_i)
{
  char i[INT_LENGTH];
  if (buffer->size < INT_LENGTH)
    return FALSE;
  gdam_buffer_peek (buffer, i, INT_LENGTH);
  *returned_i = chararray2int (i);
  return TRUE;
}



