/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdebug.h"
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

gboolean      gdam_debugging_on = FALSE;
FILE*         gdam_debugging_fp = NULL;
const char   *gdam_debugging_names[128] = {NULL};

void gdam_debug(const char* format, ...)
{
  va_list args;
  FILE *fp;
  const char *section = NULL;

  if (!gdam_debugging_on)
    return;

  fp = gdam_debugging_fp;
  if (fp == NULL)
    fp = stderr;

  if (format[0] == '<')
    {
      guint c = (int) format[1] & 0x7f;
      section = gdam_debugging_names[c];
      if (section == NULL)
        return;
      format += 3;
    }

  va_start (args, format);
  if (section != NULL)
    fprintf (fp, "gdamdebug: %s: ", section);
  else
    fprintf (fp, "gdamdebug: ");
  vfprintf (fp, format, args);
  fprintf (fp, ".\n");
  va_end (args);
}

void gdam_debug_switch(int c, gboolean on, const char *str)
{
  if (on)
    g_return_if_fail (str != NULL);
  gdam_debugging_names[c & 0x7f] = on ? str : NULL;
}

static const char * const *gdam_debug_names_list = NULL;
void gdam_debug_register_names (const char * names [])
{
  if (gdam_debug_names_list != NULL)
    g_error ("gdam_debug_register_names called twice");
  gdam_debug_names_list = (const char * const *) names;
}

gboolean gdam_debug_turn_on_debug_flag (const char *flag)
{
  int i;
  const char *name;
  if (strncmp (flag, "--debug-", 8) != 0)
    return FALSE;
  name = flag + 8;
  for (i = 0; gdam_debug_names_list[i] != NULL; i++)
    if (strcmp (gdam_debug_names_list[i] + 2, name) == 0)
      {
	gdam_debugging_on = TRUE;
        gdam_debug_switch (gdam_debug_names_list[i][0], TRUE,
	                   gdam_debug_names_list[i] + 2);
	return TRUE;
      }
  g_error ("unknown debug flag: --debug-%s", name);
  return FALSE;
}


#define MAGIC_UNDERFLOW		0x34221da4
#define MAGIC_OVERFLOW		0x15e31a3e

typedef struct _MemHeader MemHeader;
struct _MemHeader {
	int			size;
	int			lineno;
	const char*		filename;
	MemHeader*		next;
	MemHeader*		prev;
	unsigned		magic_underflow;
};
static MemHeader* first_mem_header = NULL;
static MemHeader* last_mem_header  = NULL;
static FILE*      mem_log          = NULL;

static void first_time_setup()
{
  mem_log = stderr;
}
static void* raw_alloc(size_t size, int line, const char* filename)
{
  unsigned overflow_copy = MAGIC_OVERFLOW;
  char *rv;
  MemHeader *raw = (MemHeader *) malloc (sizeof (MemHeader)
					 + sizeof (unsigned) + size);
  g_assert (raw != NULL);
  rv = (char *) (raw + 1);
  raw->prev = NULL;
  raw->next = first_mem_header;
  if (first_mem_header)
    first_mem_header->prev = raw;
  else
    {
      static gboolean first_time = TRUE;
      if (first_time)
	{
	  first_time_setup ();
	}
      first_time = FALSE;
      last_mem_header = raw;
    }
  first_mem_header = raw;
  raw->magic_underflow = MAGIC_UNDERFLOW;
  raw->size = size;
  raw->lineno = line;
  raw->filename = filename;
  memcpy (rv + size, &overflow_copy, sizeof (unsigned));
  return rv;
}

static void raw_free(gpointer ptr, int line, const char* filename)
{
  MemHeader *header;
  unsigned overflow;
  char *tailer;
  if (!ptr)
    return;
  header = ((MemHeader *) ptr) - 1;
  tailer = (char *) ptr + (header->size);
  if (header->magic_underflow != MAGIC_UNDERFLOW)
    {
      g_warning ("underflow: %p free'd at line %d, file %s",
		 ptr, line, filename);
      g_error ("and alloc'd at %s, %d", header->filename, header->lineno);
    }
  memcpy (&overflow, tailer, sizeof (unsigned));
  if (overflow != MAGIC_OVERFLOW)
    {
      g_warning ("overflow: %p free'd at line %d, file %s",
		 ptr, line, filename);
      g_error ("and alloc'd at %s, %d", header->filename, header->lineno);
    }
  if (header->next)
    header->next->prev = header->prev;
  else
    {
      g_assert (header == last_mem_header);
      last_mem_header = header->prev;
    }
  if (header->prev)
    header->prev->next = header->next;
  else
    {
      g_assert (header == first_mem_header);
      first_mem_header = header->next;
    }

  memset (header, 0xef,
	  header->size + sizeof (MemHeader) + sizeof (unsigned));
}

void*  g_malloc_debug(size_t size, int line, const char* filename)
{
  void *rv = raw_alloc (size, line, filename);
  if (mem_log)
    fprintf (mem_log, "malloc(%ld) return %p called from %s:%d\n",
	     size, rv, filename, line);
  memset (rv, 0xed, size);
  return rv;
}
void*  g_realloc_debug(void* ptr, size_t size, int line, const char* filename)
{
  void *rv = raw_alloc (size, line, filename);
  if (mem_log)
    fprintf (mem_log,
	     "realloc(%p, %ld) return %p called from %s:%d\n",
	     ptr, size, rv, filename, line);
  memset (rv, 0xed, size);
  if (ptr)
    raw_free (ptr, line, filename);
  return rv;
}
void*  g_malloc0_debug(size_t size, int line, const char* filename)
{
  void *rv = raw_alloc (size, line, filename);
  if (mem_log)
    fprintf (mem_log,
	     "malloc0(%d) return %p called from %s:%d\n",
	     size, rv, filename, line);
  memset (rv, 0x0, size);
  return rv;
}
void  g_free_debug(void* ptr, int line, const char* filename)
{
  if (ptr)
    raw_free (ptr, line, filename);
  if (mem_log)
    fprintf (mem_log, "free(%p) called from %s:%d\n", ptr, filename, line);
}
