#include "gdamfdpoll.h"

typedef struct _GdamFdPollBase GdamFdPollBase;
typedef struct _GdamFdPollTimer GdamFdPollTimer;
typedef struct _GdamFdPollFd GdamFdPollFd;

struct _GdamFdPollFd
{
  /* GdamFdPollBase */
  guint is_timer : 1;
  guint is_running : 1;
  guint is_destroyed : 1;
  GdamFdPollFd *higher_priority;
  GdamFdPollFd *lower_priority;
  int priority;
  gpointer data;
  GDestroyNotify destroy;

  /* fd-specific fields */
  GPollFD poll_fd;
  GdamFdPollFunc func;
};

struct _GdamFdPollTimer
{
  /* GdamFdPollBase */
  guint is_timer : 1;
  guint is_running : 1;
  guint is_destroyed : 1;
  GdamFdPollTimer *prior_to_expire;
  GdamFdPollTimer *next_to_expire;
  int priority;
  gpointer data;
  GDestroyNotify destroy;

  /* timer-specific fields */
  GTimeVal timeout;
  guint interval;
  GdamFdPollTimerFunc func;
};

typedef union {
  GdamFdPollFd fd;
  GdamFdPollTimer timer;
} GdamFdPollEither;

static GMemChunk *gdam_fd_poll_chunk = NULL;
G_LOCK_DEFINE_STATIC (gdam_fd_poll_chunk);

static inline GdamFdPoll  *
gdam_fd_poll_alloc   ()
{
  GdamFdPoll *rv;
  G_LOCK (gdam_fd_poll_chunk);
  if (gdam_fd_poll_chunk == NULL)
    gdam_fd_poll_chunk = g_mem_chunk_create (GdamFdPollEither,
					     5, G_ALLOC_AND_FREE);
  rv = g_mem_chunk_alloc (gdam_fd_poll_chunk);
  G_UNLOCK (gdam_fd_poll_chunk);
  return rv;
}
static inline void
gdam_fd_poll_free (GdamFdPoll *gdam_fd_poll)
{
  G_LOCK (gdam_fd_poll_chunk);
  g_mem_chunk_free (gdam_fd_poll_chunk, gdam_fd_poll);
  G_UNLOCK (gdam_fd_poll_chunk);
}

#define GDAM_FD_POLL_PRIORITY		G_PRIORITY_DEFAULT

typedef struct _PollInfoGlobal PollInfoGlobal;
struct _PollInfoGlobal
{
  GdamFdPollFd *highest_priority;
  GdamFdPollFd *lowest_priority;
  GdamFdPollTimer *earliest_timer;
  GdamFdPollTimer *latest_timer;
};

static PollInfoGlobal global_poll_info = { NULL, NULL, NULL, NULL };
static guint          global_source_id = 0;

static gboolean 
poll_info_prepare  (gpointer  source_data, 
		    GTimeVal *current_time,
		    gint     *timeout,
		    gpointer  user_data)
{
  (void) source_data; (void) current_time; (void) timeout; (void) user_data;
  if (global_poll_info.earliest_timer != NULL)
    {
      GTimeVal *end_time = &global_poll_info.earliest_timer->timeout;
      int millis = (end_time->tv_sec - current_time->tv_sec) * 1000
	         + (end_time->tv_usec - current_time->tv_usec) / 1000;
      if (millis < 0)
	millis = 0;
      *timeout = millis;
    }
  return FALSE;
}

gboolean 
poll_info_check    (gpointer  source_data,
		    GTimeVal *current_time,
		    gpointer  user_data)
{
  GdamFdPollFd *tmp;
  (void) source_data; (void) user_data;

  g_get_current_time (current_time); /* HACK */

  if (global_poll_info.earliest_timer != NULL
   && (global_poll_info.earliest_timer->timeout.tv_sec < current_time->tv_sec
    || (global_poll_info.earliest_timer->timeout.tv_sec == current_time->tv_sec
     && global_poll_info.earliest_timer->timeout.tv_usec
	                           <= current_time->tv_usec)))
    {
      return TRUE;
    }
  for (tmp = global_poll_info.highest_priority;
       tmp != NULL;
       tmp = tmp->lower_priority)
    if (tmp->poll_fd.revents != 0)
      return TRUE;
  return FALSE;
}

static void
add_timer_to_global_list (GdamFdPollTimer *timer)
{
  GdamFdPollTimer *tmp;
  for (tmp = global_poll_info.earliest_timer;
       tmp != NULL
   && (  (tmp->timeout.tv_sec  < timer->timeout.tv_sec)
     ||  (tmp->timeout.tv_sec  == timer->timeout.tv_sec
       && tmp->timeout.tv_usec <  timer->timeout.tv_usec)
     || ( tmp->timeout.tv_sec  == timer->timeout.tv_sec
       && tmp->timeout.tv_usec == timer->timeout.tv_usec
       && tmp->priority        >  timer->priority)  );
       tmp = tmp->next_to_expire);
    ;
  if (tmp == NULL)
    {
      timer->next_to_expire = NULL;
      timer->prior_to_expire = global_poll_info.latest_timer;
      global_poll_info.latest_timer = timer;
      if (global_poll_info.earliest_timer == NULL)
	global_poll_info.earliest_timer = timer;
      else
	timer->prior_to_expire->next_to_expire = timer;
    }
  else
    {
      timer->prior_to_expire = tmp->prior_to_expire;
      timer->next_to_expire = tmp;
      if (timer->prior_to_expire != NULL)
	timer->prior_to_expire->next_to_expire = timer;
      else
	global_poll_info.earliest_timer = timer;
      timer->next_to_expire->prior_to_expire = timer;
    }
}

/* returns the next valid source of lower priority */
static GdamFdPollFd *
gdam_fd_poll_destroy_now (GdamFdPollFd      *tmp)
{
  GdamFdPollFd *next;
  if (tmp->destroy != NULL)
    (*tmp->destroy) (tmp->data);
  next = tmp->lower_priority;
  if (tmp->lower_priority != NULL)
    tmp->lower_priority->higher_priority = tmp->higher_priority;
  else
    global_poll_info.lowest_priority = tmp->higher_priority;
  if (tmp->higher_priority != NULL)
    tmp->higher_priority->lower_priority = tmp->lower_priority;
  else
    global_poll_info.highest_priority = tmp->lower_priority;
  g_main_remove_poll (&tmp->poll_fd);
  gdam_fd_poll_free ((GdamFdPoll *) tmp);
  return next;
}

static GdamFdPollTimer *
gdam_fd_poll_timer_destroy_now (GdamFdPollTimer      *tmp)
{
  GdamFdPollTimer *next;
  if (tmp->destroy != NULL)
    (*tmp->destroy) (tmp->data);
  next = tmp->next_to_expire;
  if (tmp->next_to_expire != NULL)
    tmp->next_to_expire->prior_to_expire = tmp->prior_to_expire;
  else
    global_poll_info.latest_timer = tmp->prior_to_expire;
  if (tmp->prior_to_expire != NULL)
    tmp->prior_to_expire->next_to_expire = tmp->next_to_expire;
  else
    global_poll_info.earliest_timer = tmp->next_to_expire;
  gdam_fd_poll_free ((GdamFdPoll *) tmp);
  return next;
}

gboolean 
poll_info_dispatch (gpointer  source_data, 
		    GTimeVal *dispatch_time,
		    gpointer  user_data)
{
  (void) source_data; (void) user_data;
  g_get_current_time (dispatch_time);		/* HACK */
  {
    GdamFdPollFd *tmp;
    GdamFdPollFd *next = NULL;
    for (tmp = global_poll_info.highest_priority; tmp != NULL; tmp = next)
      {
	if (tmp->poll_fd.revents != 0)
	  {
	    g_assert (!tmp->is_destroyed && !tmp->is_running);
	    tmp->is_running = 1;
	    if (!(*tmp->func) ((GdamFdPoll *) tmp,
			       tmp->poll_fd.fd,
			       tmp->poll_fd.revents,
			       tmp->data))
	      tmp->is_destroyed = 1;
	    if (tmp->is_destroyed)
	      {
		next = gdam_fd_poll_destroy_now (tmp);
	      }
	    else
	      {
		tmp->is_running = 0;
		next = tmp->lower_priority;
	      }
	  }
	else
	  {
	    next = tmp->lower_priority;
	  }
      }
  }
  {
    GdamFdPollTimer *tmp;
    GdamFdPollTimer *next = NULL;
    GdamFdPollTimer *to_reorder;
    for (tmp = global_poll_info.earliest_timer; tmp != NULL; tmp = next)
      {
	g_assert (!tmp->is_destroyed && !tmp->is_running);
	if (dispatch_time->tv_sec < tmp->timeout.tv_sec
	 || (dispatch_time->tv_sec == tmp->timeout.tv_sec
	  && dispatch_time->tv_usec < tmp->timeout.tv_usec))
	  break;
	tmp->is_running = 1;
	if (!(*tmp->func) ((GdamFdPoll *) tmp, tmp->data))
	  tmp->is_destroyed = 1;
	if (tmp->is_destroyed)
	  {
	    next = gdam_fd_poll_timer_destroy_now (tmp);
	  }
	else
	  {
	    tmp->is_running = 0;
	    next = tmp->next_to_expire;
	  }
      }

    /* All timers from global_poll_info.earliest_timer to tmp (non-inclusive)
     * must be reordered. [Note: it might be a good idea to re-query the
     * system timer here!]
     */
    to_reorder = global_poll_info.earliest_timer;
    if (tmp == NULL)
      global_poll_info.earliest_timer = global_poll_info.latest_timer = NULL;
    else
      {
	global_poll_info.earliest_timer = tmp;
	tmp->prior_to_expire = NULL;
      }
    while (to_reorder != tmp)
      {
	/* update timer */
	to_reorder->timeout.tv_usec = dispatch_time->tv_usec
			            + to_reorder->interval % 1000 * 1000;
	to_reorder->timeout.tv_sec = dispatch_time->tv_sec
	                           + to_reorder->interval / 1000;
	if (to_reorder->timeout.tv_usec >= 1000 * 1000)
	  {
	    to_reorder->timeout.tv_usec -= 1000 * 1000;
	    to_reorder->timeout.tv_sec += 1;
	  }

	/* add to the global list of timers */
	next = to_reorder->next_to_expire;
	add_timer_to_global_list (to_reorder);
	to_reorder = next;
      }
  }
  return TRUE;
}

static GSourceFuncs source_funcs =
{
  poll_info_prepare,
  poll_info_check,
  poll_info_dispatch,
  NULL
};


GdamFdPoll *
gdam_fd_poll_new  (int            fd,
		   int            priority,
		   GIOCondition   events,
		   GdamFdPollFunc func,
		   gpointer       data,
		   GDestroyNotify destroy)
{
  GdamFdPollFd *tmp;
  GdamFdPollFd *fd_poll;

  /* don't allow the same file-descriptor to be added twice */
  for (tmp = global_poll_info.highest_priority;
       tmp != NULL;
       tmp = tmp->lower_priority)
    g_return_val_if_fail (tmp->is_destroyed || tmp->poll_fd.fd != fd, NULL);

  fd_poll = (GdamFdPollFd *) gdam_fd_poll_alloc ();

  fd_poll->is_timer = 0;
  fd_poll->poll_fd.fd = fd;
  fd_poll->poll_fd.events = events;
  fd_poll->poll_fd.revents = 0;
  fd_poll->is_running = 0;
  fd_poll->is_destroyed = 0;
  fd_poll->func = func;
  fd_poll->data = data;
  fd_poll->destroy = destroy;
  fd_poll->priority = priority;

  for (tmp = global_poll_info.highest_priority;
       tmp != NULL && tmp->priority > priority;
       tmp = tmp->lower_priority)
    ;
  if (tmp == NULL)
    {
      fd_poll->lower_priority = NULL;
      fd_poll->higher_priority = global_poll_info.lowest_priority;
      global_poll_info.lowest_priority = fd_poll;
      if (global_poll_info.highest_priority == NULL)
	global_poll_info.highest_priority = fd_poll;
      else
	fd_poll->higher_priority->lower_priority = fd_poll;
    }
  else
    {
      fd_poll->higher_priority = tmp->higher_priority;
      fd_poll->lower_priority = tmp;
      if (fd_poll->higher_priority != NULL)
	fd_poll->higher_priority->lower_priority = fd_poll;
      else
	global_poll_info.highest_priority = fd_poll;
      fd_poll->lower_priority->higher_priority = fd_poll;
    }
  g_main_add_poll (&fd_poll->poll_fd, GDAM_FD_POLL_PRIORITY);
  if (global_source_id == 0)
    global_source_id = g_source_add (GDAM_FD_POLL_PRIORITY,
				     FALSE,		/* can_recurse */
				     &source_funcs,
				     NULL, NULL, NULL);
  return (GdamFdPoll *) fd_poll;
}

GdamFdPoll *
gdam_fd_poll_new_timeout  (int            millis_timeout,
		           int            priority,
		           GdamFdPollTimerFunc func,
		           gpointer       data,
		           GDestroyNotify destroy)
{
  GdamFdPollTimer *timer;
  GdamFdPollTimer *tmp;
  timer = (GdamFdPollTimer *) gdam_fd_poll_alloc ();

  timer->is_timer = 1;
  timer->is_running = 0;
  timer->is_destroyed = 0;
  timer->func = func;
  timer->data = data;
  timer->destroy = destroy;
  timer->priority = priority;
  g_get_current_time (&timer->timeout);
  timer->interval = millis_timeout;
  timer->timeout.tv_sec += millis_timeout / 1000;
  timer->timeout.tv_usec += millis_timeout % 1000 * 1000;
  if (timer->timeout.tv_usec >= 1000 * 1000)
    {
      timer->timeout.tv_usec -= 1000 * 1000;
      timer->timeout.tv_sec += 1;
    }

  add_timer_to_global_list (timer);

  if (global_source_id == 0)
    global_source_id = g_source_add (GDAM_FD_POLL_PRIORITY,
				     FALSE,		/* can_recurse */
				     &source_funcs,
				     NULL, NULL, NULL);
  return (GdamFdPoll *) timer;
}

void
gdam_fd_poll_set_events (GdamFdPoll    *poll_object,
			 GIOCondition   events)
{
  GdamFdPollFd *fd = (GdamFdPollFd *) poll_object;;
  g_return_if_fail (fd->is_timer == 0);
  fd->poll_fd.events = events;
}

void
gdam_fd_poll_destroy (GdamFdPoll    *fd_poll)
{
  GdamFdPollFd *fd = (GdamFdPollFd *) fd_poll;
  if (fd->is_timer)
    {
      GdamFdPollTimer *timer = (GdamFdPollTimer *) fd_poll;
      if (timer->is_running)
	timer->is_destroyed = 1;
      else
	(void) gdam_fd_poll_timer_destroy_now (timer);
    }
  else
    {
      if (fd->is_running)
	fd->is_destroyed = 1;
      else
	(void) gdam_fd_poll_destroy_now (fd);
    }
}
