#include "gdammathfunction.h"
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

typedef struct _GdamMathFunctionClassOperator GdamMathFunctionClassOperator;

#define LEFT_PAREN_STRING    "("
#define RIGHT_PAREN_STRING   ")"

#ifndef M_PI
#define M_PI      3.14159265358979323846
#endif

typedef struct _ParseNode ParseNode;
struct _ParseNode
{
  gboolean is_text;
  union {
    char             *text;
    GdamMathFunction *function;
  } d;
};

#if 0
static void print_parse_node(ParseNode* node)
{
  if (node->is_text)
    g_message ("TEXT: %s", node->d.text);
  else
    {
      char *func;
      func = gdam_math_function_to_string (node->d.function);
      g_message ("FUNCTION: %s", func);
      g_free (func);
    }
}
#endif

typedef struct _Operator Operator;
struct _Operator
{
        const char *normal;
        const char *inverted;
        GdamMathFunctionClass *func_class;
        int precedence;
};

static GSList* parse_functions_and_operators(GSList* tokens);

/*
 * Various mathfunctionclasses.
 */
struct _GdamMathFunctionClassOperator
{
        GdamMathFunctionClass   base;
        Operator               *operator;
};

/*
 * Global tables.
 */
static GHashTable      *names_to_operators = NULL;
static GHashTable      *names_to_function_classes = NULL;
static void initialize_hash_table()
{
  names_to_operators = g_hash_table_new (g_str_hash, g_str_equal);
  names_to_function_classes = g_hash_table_new (g_str_hash, g_str_equal);
}

static void func_operator_append_to_string(GdamMathFunction*   function,
                                           GString*            string)
{
  GdamMathFunctionOperator *operator_func;
  Operator *operator;
  int i;

  operator_func = (GdamMathFunctionOperator *) function;
  g_return_if_fail (operator_func->num_args > 0);
  operator = operator_func->operator;

  gdam_math_function_append (operator_func->args[0],
			     string, operator->precedence);
  for (i = 1; i < operator_func->num_args; i++)
    {
      g_string_append_c (string, ' ');
      if (!operator_func->whether_inverted
	  || !operator_func->whether_inverted[i])
	{
	  g_string_append (string, operator->normal);
	}
      else
	{
	  g_string_append (string, operator->inverted);
	}
      g_string_append_c (string, ' ');
      gdam_math_function_append (operator_func->args[i],
				 string, operator->precedence);
    }
}

void gdam_math_function_class_new_binary
				(const char*	major_op,
				 const char*    minor_op,
				 int            precedence,
				 GdamEvalFunc   evaluator)
{
  Operator *operator;
  GdamMathFunctionClassOperator *operator_class;
  GdamMathFunctionClass *func_class;
  if (names_to_operators == NULL)
    {
      initialize_hash_table ();
    }

  g_return_if_fail (g_hash_table_lookup (names_to_operators, major_op) ==
		    NULL);
  g_return_if_fail (minor_op == NULL
		    || g_hash_table_lookup (names_to_operators,
					    minor_op) == NULL);
  operator = g_new (Operator, 1);
  operator_class = g_new0 (GdamMathFunctionClassOperator, 1);
  func_class = (GdamMathFunctionClass *) operator_class;

  /* set up operator. */
  operator->normal = major_op;
  operator->inverted = minor_op;
  operator->func_class = func_class;
  operator->precedence = precedence;

  /* set up FunctionClass. */
  operator_class->operator = operator;
  func_class->category = GDAM_MATH_FUNCTION_OPERATOR;
  func_class->append_to_string = func_operator_append_to_string;
  func_class->evaluator = evaluator;
  func_class->class_name = major_op;
  func_class->precedence = precedence;

  /* add them to tables. */
  g_hash_table_insert (names_to_operators, (gpointer) major_op, operator);
  g_hash_table_insert (names_to_function_classes,
		       (gpointer) major_op, func_class);
  if (minor_op != NULL)
    g_hash_table_insert (names_to_operators, (gpointer) minor_op, operator);

}

static void named_append_to_string(GdamMathFunction*     function,
                                   GString*              string)
{
  GdamMathFunctionNamed *named_func;

  named_func = (GdamMathFunctionNamed *) function;
  g_string_append (string, named_func->klass->class_name);
  g_string_append_c (string, '(');
  gdam_math_function_append (named_func->argument, string, 0);
  g_string_append_c (string, ')');
}


void gdam_math_function_class_new_named_function
				(const char*	named_function,
				 GdamEvalFunc   evaluator)
{
  GdamMathFunctionClass *func_class;

  if (names_to_function_classes == NULL)
    initialize_hash_table ();

  func_class = g_new0 (GdamMathFunctionClass, 1);
  func_class->class_name = named_function;
  func_class->category = GDAM_MATH_FUNCTION_NAMED;
  func_class->append_to_string = named_append_to_string;
  func_class->evaluator = evaluator;

  g_hash_table_insert (names_to_function_classes,
		       (gpointer) named_function, func_class);
}

GdamMathFunctionClass* gdam_math_function_class_get_by_name
			        (const char*    name)
{
  if (names_to_function_classes == NULL)
    return NULL;
  return g_hash_table_lookup (names_to_function_classes, name);
}
static Operator* lookup_operator(const char* name)
{
  if (names_to_operators == NULL)
    return NULL;
  return g_hash_table_lookup (names_to_operators, name);
}
	
static gboolean is_operator_char(char op_char)
{
  return strchr ("*/-+", op_char) != NULL;
}
static int get_operator_length(const char* operator)
{
  return 1;
}

#define SKIP_CHAR_TYPE(ptr, condition, index)			\
	while ( (ptr)[(index)] != '\0'				\
	     && ( condition ((ptr)[(index)]) ) )		\
	     		index++

#define DOUBLE_CHARS "0123456789" ".eE"
#define DOUBLE_FIRST_CHARS DOUBLE_CHARS "+-"

static GSList* tokenize(const char* string)
{
  GSList *list = NULL;
  gboolean error = FALSE;
  char *new_string;
  ParseNode *node;
  gboolean operator_allowed = FALSE;
  gboolean is_number = FALSE;

  while (*string != '\0')
    {
      int length;
      gboolean new_operator_allowed = FALSE;
      while (*string != '\0' && isspace (*string))
	string++;

      if (*string == '\0')
	break;

      length = 0;
      is_number = FALSE;
      if (strchr (operator_allowed ? DOUBLE_CHARS : DOUBLE_FIRST_CHARS,
		  *string) != NULL)
	{
	  length = 1 + strspn (string + 1, DOUBLE_CHARS);
	  new_operator_allowed = TRUE;
	}
      else if (*string == 'x' && !isalpha (string[1]))
	{
	  length = 1;
	  SKIP_CHAR_TYPE (string, isspace, length);
	  if (string[length] == '[')
	    {
	      length++;
	      SKIP_CHAR_TYPE (string, isspace, length);
	      SKIP_CHAR_TYPE (string, isdigit, length);
	      SKIP_CHAR_TYPE (string, isspace, length);
	      if (string[length] != ']')
		{
		  g_warning ("expected ']'");
		  error = TRUE;
		}
	      else
		{
		  length++;
		}
	    }
	  new_operator_allowed = TRUE;
	}
      else if (*string == '(' || *string == ')')
	{
	  length = 1;
	}
      else if (isalpha (*string))
	{
	  SKIP_CHAR_TYPE (string, isalpha, length);
	}
      else if (is_operator_char (*string))
	{
	  length = get_operator_length (string);
	}
      else
	{
	  g_warning ("unexpected character: `%c'", *string);
	  error = TRUE;
	}

      operator_allowed = new_operator_allowed;

      if (error)
	{
	  g_slist_foreach (list, (GFunc) g_free, NULL);
	  g_slist_free (list);
	  return NULL;
	}

      g_return_val_if_fail (length > 0, NULL);

      new_string = g_new (char, length + 1);
      memcpy (new_string, string, length);
      new_string[length] = '\0';
      node = g_new (ParseNode, 1);
      node->is_text = TRUE;
      node->d.text = new_string;
      list = g_slist_prepend (list, node);
      string += length;
    }
  return g_slist_reverse (list);
}

static void parse_node_free(ParseNode* node)
{
  if (node->is_text)
    g_free (node->d.text);
  else
    gdam_math_function_destroy (node->d.function);
  g_free (node);
}

static void find_constants(ParseNode* node)
{
  int length;
  char *text;
  gdouble val;
  char *endp;
  if (!node->is_text)
    return;

  text = node->d.text;
  if (strcmp (text, "pi") == 0)
    {
      val = M_PI;
    }
  else
    {
      length = strspn (text, DOUBLE_CHARS);
      if (length == 0)
	return;

      if (text[length] != '\0')
	{
	  g_warning ("trailing garbage after number?");
	  return;
	}
      val = strtod (text, &endp);
    }
  g_free (text);

  node->is_text = FALSE;
  node->d.function = gdam_math_function_constant_new (val);
}
static void
find_variables(ParseNode            *node,
	       GdamMathParseContext *context)
{
  char *text;
  guint i;
  int index = -1;			/* which input variable */
  if (!node->is_text)
    return;

  text = node->d.text;
  if (context != NULL && context->has_var_names)
    {
      /* TODO: use a hash-table */
      for (i = 0; i < context->num_var_names; i++)
	if (strcmp (text, context->var_names[i]) == 0)
	  break;
      if (i < context->num_var_names)
	index = i;
    }
  if (index == -1)
    {
      if (text[0] != 'x' || isalpha (text[1]))
	return;
      i = 1;
      SKIP_CHAR_TYPE (text, isspace, i);
      if (text[i] == '[')
	{
	  i++;
	  SKIP_CHAR_TYPE (text, isspace, i);
	  index = atoi (text + i);
	}
      else
	{
	  index = 0;
	}
    }
  g_free (text);
  node->is_text = FALSE;
  node->d.function = gdam_math_function_variable_new (index);
}

/* Take a list of tokens, and recurse on each pair
 * of parentheses (leaving a parenthesis-free expression).
 *
 * NULL is returned on error and the inputs are deleted.
 */
static GSList* condense_parentheses(GSList* tokens)
{
  GSList **plist = &tokens;
  GSList **start = NULL;
  int paren_level = 0;
  gboolean error = FALSE;

  while (*plist != NULL)
    {
      ParseNode *node = (*plist)->data;
      int new_paren_level = paren_level;
      if (node->is_text)
	{
	  char *text = node->d.text;
	  if (strcmp (text, LEFT_PAREN_STRING) == 0)
	    {
	      new_paren_level++;
	    }
	  else if (strcmp (text, RIGHT_PAREN_STRING) == 0)
	    {
	      new_paren_level--;
	    }
	}
      if (paren_level == 0 && new_paren_level == 1)
	{
	  start = plist;
	}
      else if (paren_level == 1 && new_paren_level == 0)
	{
	  GSList *list_end = *plist;
	  GSList *next_el = list_end->next;
	  GSList *expr_list = (*start)->next;
	  GSList *func_list;
	  GSList *end_func_list;
	  g_free (list_end->data);
	  g_slist_free_1 (list_end);
	  *plist = NULL;
	  g_free ((*start)->data);
	  g_slist_free_1 (*start);
	  func_list = parse_functions_and_operators (expr_list);
	  if (func_list == NULL)
	    {
	      error = TRUE;
	      *start = next_el;
	      g_warning ("couldn't parse parenthesized sub");
	      break;
	    }
	  end_func_list = g_slist_last (func_list);
	  plist = &end_func_list->next;
	  *plist = next_el;
	  *start = func_list;
	  start = NULL;
	  paren_level = 0;
	  continue;
	}

      paren_level = new_paren_level;
      plist = &((*plist)->next);
    }
  if (paren_level != 0)
    {
      g_warning ("mismatched parenthesis");
      error = TRUE;
    }
  if (error)
    {
      g_slist_foreach (tokens, (GFunc) parse_node_free, NULL);
      g_slist_free (tokens);
      return NULL;
    }
  return tokens;
}

static Operator* find_operator_of_lowest_precedence(GSList* tokens)
{
  GSList *at;
  Operator *rv = NULL;
  for (at = tokens; at != NULL; at = at->next)
    {
      ParseNode *parse_node = at->data;
      Operator *tmp;

      if (!parse_node->is_text)
	continue;

      tmp = lookup_operator (parse_node->d.text);
      if (tmp == NULL)
	continue;

      if (rv == NULL || tmp->precedence < rv->precedence)
	rv = tmp;
    }
  return rv;
}

/* Scan through the token list, looking for operators
 * matching `operator'.
 *
 * The expression is divided into sublists of tokens,
 * and a list of operators, which separated them.
 */
static void reorganize_lists_for_operator(GSList*      tokens,
                                          Operator*    operator,
                                          GSList**     sublists_out,
					  GSList**     operators_out)
{
  GSList *at;
  GSList *first_token = NULL;
  GSList *last = NULL;

  *sublists_out = NULL;
  *operators_out = NULL;
  for (at = tokens; at != NULL;)
    {
      ParseNode *parse_node = at->data;
      gboolean matched_op = FALSE;

      if (parse_node->is_text)
	{
	  const char *text = parse_node->d.text;
	  if (strcmp (text, operator->normal) == 0)
	    matched_op = TRUE;
	  else if (operator->inverted != NULL
		   && strcmp (text, operator->inverted) == 0)
	    matched_op = TRUE;
	}

      if (matched_op)
	{
	  GSList *next = at->next;
	  at->next = *operators_out;
	  *operators_out = at;
	  at = next;
	  last->next = NULL;
	  *sublists_out = g_slist_prepend (*sublists_out, first_token);
	  first_token = NULL;
	  last = NULL;
	}
      else
	{
	  if (first_token == NULL)
	    first_token = at;
	  last = at;
	  at = at->next;
	}
    }
  if (first_token != NULL)
    *sublists_out = g_slist_prepend (*sublists_out, first_token);

  *sublists_out = g_slist_reverse (*sublists_out);
  *operators_out = g_slist_reverse (*operators_out);
  g_return_if_fail (g_slist_length (*sublists_out)
		    == g_slist_length (*operators_out) + 1);
}

static GSList* condense_named(GSList* list, GdamMathFunctionClass* klass)
{
  ParseNode *tmp;
  ParseNode *name_node;
  GdamMathFunction *arg;
  GdamMathFunction *new_function;
  GSList *rv;
  g_return_val_if_fail (list->next != NULL, NULL);
  name_node = list->data;
  rv = list->next;
  tmp = list->next->data;
  if (tmp->is_text)
    {
      g_warning ("function name %s was not followed by (arg)",
		 klass->class_name);
      return NULL;
    }
  arg = tmp->d.function;
  new_function = gdam_math_function_named_new (klass, arg);

  tmp->d.function = new_function;
  g_free (name_node->d.text);
  g_free (name_node);
  g_slist_free_1 (list);
  return rv;

}

static GSList* parse_functions_and_operators(GSList* tokens)
{
  GSList **plist;
  Operator *operator = NULL;
  tokens = condense_parentheses (tokens);
  if (tokens == NULL)
    return NULL;

  /* Parse function names. */
  for (plist = &tokens; *plist != NULL;)
    {
      ParseNode *node = (*plist)->data;

      if (node->is_text)
	{
	  GdamMathFunctionClass *klass;
	  const char *name = node->d.text;
	  klass = gdam_math_function_class_get_by_name (name);
	  if (klass != NULL && klass->category == GDAM_MATH_FUNCTION_NAMED)
	    {
	      GSList *tmp;
	      tmp = condense_named (*plist, klass);
	      if (tmp == NULL)
		{
		  /* XXX: leaks. */
		  return NULL;
		}
	      *plist = tmp;
	      continue;
	    }
	}

      plist = &((*plist)->next);
    }

  while ((operator = find_operator_of_lowest_precedence (tokens)) != NULL)
    {
      GSList *subs;
      GSList *ops;
      GSList *at;
      reorganize_lists_for_operator (tokens, operator, &subs, &ops);
      for (at = subs; at != NULL; at = at->next)
	{
	  GSList *sub_list = at->data;
	  at->data = parse_functions_and_operators (sub_list);
	  sub_list = at->data;

	  /* XXX: leaks... */
	  g_return_val_if_fail (sub_list != NULL, NULL);

	  /* this really should be an internal error. */
	  g_return_val_if_fail (sub_list->next == NULL, NULL);
	}

      /* Aggregate the lists into a single
       * GdamMathFunctionOperator node. */
      {
	GdamMathFunction *rv_func;
	ParseNode *rv;
	int num_args;
	GdamMathFunction **args;
	gboolean *inverted = NULL;
	GSList *tmp;
	int index;

	num_args = g_slist_length (subs);
	args = g_new (GdamMathFunction *, num_args);
	if (operator->inverted != NULL)
	  {
	    inverted = g_new (gboolean, num_args);
	    inverted[0] = FALSE;
	  }
	index = 0;
	for (tmp = subs; tmp != NULL; tmp = tmp->next)
	  {
	    GSList *sublist = tmp->data;
	    ParseNode *parse_node = sublist->data;

	    /* leaks... */
	    g_return_val_if_fail (!parse_node->is_text, NULL);
	    args[index++] = parse_node->d.function;
	  }
	if (inverted)
	  {
	    index = 1;
	    for (tmp = ops; tmp != NULL; tmp = tmp->next)
	      {
		ParseNode *node = tmp->data;
		gboolean is_inv;
		g_assert (node->is_text);
		is_inv = strcmp (node->d.text, operator->inverted) == 0;
		inverted[index++] = is_inv;
	      }
	  }
	rv_func = gdam_math_function_operator_new
	  (operator->func_class, num_args, args, inverted);
	rv = g_new (ParseNode, 1);
	rv->is_text = FALSE;
	rv->d.function = rv_func;
	g_slist_foreach (subs, (GFunc) g_slist_free, NULL);
	g_slist_free (subs);
	g_slist_free (ops);
	return g_slist_prepend (NULL, rv);
      }
    }

  g_return_val_if_fail (tokens->next == NULL, tokens);
  return tokens;
}

GdamMathFunction* gdam_math_function_parse(const char* string,
					   GdamMathParseContext *context)
{
  GSList *tokens;
  ParseNode *parse_node;
  tokens = tokenize (string);
  g_slist_foreach (tokens, (GFunc) find_constants, NULL);
  g_slist_foreach (tokens, (GFunc) find_variables, context);
  tokens = parse_functions_and_operators (tokens);
  if (g_slist_length (tokens) != 1)
    {
      g_warning ("error parsing function: didn't reduce to singleton");
      g_slist_foreach (tokens, (GFunc) parse_node_free, NULL);
      g_slist_free (tokens);
      return NULL;
    }

  parse_node = tokens->data;
  if (parse_node->is_text)
    {
      g_warning ("error parsing function: text node: %s", parse_node->d.text);
      g_slist_foreach (tokens, (GFunc) parse_node_free, NULL);
      g_slist_free (tokens);
      return NULL;
    }

  g_slist_free (tokens);
  return parse_node->d.function;
}

/*
 *                      _                   _   _
 *   ___ ___  _ __  ___| |_ _ __ _   _  ___| |_(_)_ __   __ _
 *  / __/ _ \| '_ \/ __| __| '__| | | |/ __| __| | '_ \ / _` |
 * | (_| (_) | | | \__ \ |_| |  | |_| | (__| |_| | | | | (_| |
 *  \___\___/|_| |_|___/\__|_|   \__,_|\___|\__|_|_| |_|\__, |
 *                                                      |___/
 *   __                  _   _
 *  / _|_   _ _ __   ___| |_(_) ___  _ __  ___
 * | |_| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 * |  _| |_| | | | | (__| |_| | (_) | | | \__ \
 * |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 * 
 */

GdamMathFunction*  gdam_math_function_named_new
				(GdamMathFunctionClass* klass,
				 GdamMathFunction*      argument)
{
  GdamMathFunctionNamed *named;

  named = g_new0 (GdamMathFunctionNamed, 1);
  named->category = GDAM_MATH_FUNCTION_NAMED;
  named->klass = klass;
  named->evaluator = klass->evaluator;
  named->argument = argument;

  return (GdamMathFunction *) named;
}


GdamMathFunction*  gdam_math_function_operator_new
				(GdamMathFunctionClass* klass,
				 int                 num_args,
				 GdamMathFunction**  args,
				 gboolean*           whether_inverted)
{
  GdamMathFunctionClassOperator *func_class_operator;
  GdamMathFunctionOperator *func_operator;

  func_class_operator = (GdamMathFunctionClassOperator *) klass;
  func_operator = g_new0 (GdamMathFunctionOperator, 1);
  func_operator->category = GDAM_MATH_FUNCTION_OPERATOR;
  func_operator->klass = klass;
  func_operator->evaluator = klass->evaluator;
  func_operator->operator = func_class_operator->operator;
  func_operator->whether_inverted = whether_inverted;
  func_operator->num_args = num_args;
  func_operator->args = args;

  return (GdamMathFunction *) func_operator;
}

static void constant_append_to_string(GdamMathFunction* function,
                                      GString*          string)
{
  GdamMathFunctionConstant *func_constant;
  func_constant = (GdamMathFunctionConstant *) function;
  g_string_sprintfa (string, "%.9g", func_constant->return_value);
}

static gdouble constant_evaluator(GdamMathFunction*  function,
                                  gdouble*           input)
{
  return ((GdamMathFunctionConstant *) function)->return_value;
}

static GdamMathFunctionClass* gdam_math_function_constant_class_get()
{
  static GdamMathFunctionClass the_class;
  static gboolean inited = FALSE;
  if (inited)
    return &the_class;
  the_class.category = GDAM_MATH_FUNCTION_CONSTANT;
  the_class.precedence = 100;
  the_class.append_to_string = constant_append_to_string;
  the_class.evaluator = constant_evaluator;
  the_class.class_name = "constant";
  the_class.destroy = NULL;
  inited = TRUE;
  return &the_class;
}

GdamMathFunction* gdam_math_function_constant_new(gdouble    val)
{
  GdamMathFunctionConstant *func_constant;
  func_constant = g_new0 (GdamMathFunctionConstant, 1);
  func_constant->category = GDAM_MATH_FUNCTION_CONSTANT;
  func_constant->klass = gdam_math_function_constant_class_get ();
  func_constant->evaluator = constant_evaluator;
  func_constant->return_value = val;
  return (GdamMathFunction *) func_constant;
}

static void variable_append_to_string(GdamMathFunction* function,
                                      GString*          string)
{
  int index;
  GdamMathFunctionVariable *variable;
  variable = (GdamMathFunctionVariable *) function;
  index = variable->index;
  if (index == 0)
    g_string_append_c (string, 'x');
  else
    g_string_sprintfa (string, "x[%d]", index);
}

static gdouble variable_evaluator(GdamMathFunction*  function,
                                  gdouble*           input)
{
  GdamMathFunctionVariable *variable;
  variable = (GdamMathFunctionVariable *) function;
  return input[variable->index];
}

static GdamMathFunctionClass* gdam_math_function_variable_class_get()
{
  static GdamMathFunctionClass the_class;
  static gboolean inited = FALSE;
  if (inited)
    return &the_class;
  the_class.category = GDAM_MATH_FUNCTION_VARIABLE;
  the_class.precedence = 100;
  the_class.append_to_string = variable_append_to_string;
  the_class.evaluator = variable_evaluator;
  the_class.class_name = "variable";
  the_class.destroy = NULL;
  inited = TRUE;
  return &the_class;
}
GdamMathFunction* gdam_math_function_variable_new(int index)
{
  GdamMathFunctionVariable *func_variable;
  func_variable = g_new0 (GdamMathFunctionVariable, 1);
  func_variable->category = GDAM_MATH_FUNCTION_VARIABLE;
  func_variable->klass = gdam_math_function_variable_class_get ();
  func_variable->evaluator = variable_evaluator;
  func_variable->index = index;
  return (GdamMathFunction *) func_variable;
}



char* gdam_math_function_to_string(GdamMathFunction*   function)
{
  GString *string;
  char *return_val;

  string = g_string_new ("");
  gdam_math_function_append (function, string, 0);
  return_val = string->str;
  g_string_free (string, FALSE);
  return return_val;
}

void   gdam_math_function_append(GdamMathFunction*     function,
                                 GString*              string,
				 int                   precedence)
{
  GdamMathFunctionClass *klass;
  gboolean parenthesize;
  klass = function->any.klass;

  parenthesize = (klass->precedence <= precedence);
  if (parenthesize)
    g_string_append_c (string, '(');
  (*klass->append_to_string) (function, string);
  if (parenthesize)
    g_string_append_c (string, ')');
}

gdouble   gdam_math_function_eval(GdamMathFunction*     function,
                                  gdouble*              input)
{
  return (*function->any.evaluator) (function, input);
}

void    gdam_math_function_destroy(GdamMathFunction* function)
{
  GdamMathFunctionClass *func_class = function->any.klass;

  if (func_class->destroy)
    (*func_class->destroy) (function);

  g_free (function);
}

