/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsearchpath.h"
#include "gdamtildeexpand.h"
#include <sys/stat.h>

static gboolean search_path_initialized = FALSE;
static char** search_path = NULL;


static void set_search_path(const char* new_search_path)
{
  if (search_path != NULL)
    g_strfreev (search_path);
  if (new_search_path == NULL)
    search_path = NULL;
  else
    search_path = g_strsplit (new_search_path, ":", 0);
}
#ifndef DEST_PREFIX
#define DEST_PREFIX "/usr/local"
#endif

char *gdam_search_path_config = NULL;

void gdam_search_path_initialize()
{
  char *path;
  if (search_path_initialized)
    return;
  search_path_initialized = TRUE;
  path = g_strdup_printf ("%s/.gdam:/etc/gdam:%s/lib/gdam/local:.",
			  g_get_home_dir (), DEST_PREFIX);
  set_search_path (path);
  g_free (path);

  gdam_search_path_config = g_strdup_printf ("%s/.gdam:%s/etc/gdam:/etc/gdam",
                                             g_get_home_dir (), DEST_PREFIX);
}

/* num_filenames_out should be set to maximum number of return values
 * before calling. */
void 
gdam_search_path_all (const char*          filename,
                      const char*          path,
		      guint                flags,
		      char**               filenames_out,
		      int*		   num_filenames_out)
{
  char **local_paths = NULL;
  char **at;
  int max_filenames = *num_filenames_out;
  int num_filenames = 0;
  if (search_path_initialized == FALSE)
    gdam_search_path_initialize ();
  if (path != NULL)
    {
      local_paths = g_strsplit (path, ":", 0);
      for (at = local_paths; *at != NULL; at++)
	{
	  if (**at == '~')
	    {
	      char *old_path = *at;
	      *at = gdam_tilde_expand (old_path);
	      if (*at)
		g_free (old_path);
	      else
		*at = old_path;
	    }
	}
    }
  else
    local_paths = search_path;

  /* XXX: look at 'flags' in this case?? */
  if (*filename == '/')
    {
      filenames_out[0] = g_strdup (filename);
      *num_filenames_out = 1;
      return;
    }

  at = local_paths;
  while (*at != NULL && num_filenames < max_filenames)
    {
      char *fullpath = g_strjoin ("/", *at, filename, NULL);
      struct stat stat_buf;
      if (stat (fullpath, &stat_buf) < 0)
	{
	  g_free (fullpath);
	  at++;
	  continue;
	}
      if ((flags & GDAM_SEARCH_PATH_FIND_DIRECTORY)
	  && (S_ISDIR (stat_buf.st_mode)))
	{
	  filenames_out[num_filenames] = fullpath;
	  num_filenames++;
	  at++;
	  continue;
	}
      if ((flags & GDAM_SEARCH_PATH_FIND_FILE)
	  && (S_ISREG (stat_buf.st_mode)))
	{
	  filenames_out[num_filenames] = fullpath;
	  num_filenames++;
	  at++;
	  continue;
	}
      g_free (fullpath);
      at++;
    }
  *num_filenames_out = num_filenames;
  return;
}

char* gdam_search_path(const char*          filename,
                       const char*          path,
		       guint                flags)
{
  char *filenames[1];
  int num = 1;
  gdam_search_path_all (filename, path, flags, filenames, &num);
  return num > 0 ? filenames[0] : NULL;
}

