#include "gdamspline.h"
#include <math.h>

/*
 * TODO:  currently, each segment is evaluated from 0 ... end - start.
 *        we might get slightly better numerical precision from
 *        using
 *                   (start - end) / 2 ... (end - start) / 2
 *       ie centered around 0.
 */

typedef struct _GdamSplineSegment GdamSplineSegment;
struct _GdamSplineSegment {
	/* private members */
	gfloat			a,b,c,d;
	gfloat			start;
	gfloat			end;
};

struct _GdamSpline {
	/* private members */
	int			num_segments;
	GdamSplineSegment*	segments;
};

/*
 * Solve for x,y in these equations simulataneously:
 *	x1 * x + y1 * y = out1		(1)
 *	x2 * x + y2 * y = out2		(2)
 *
 * Find the ratio closest to 1, and subtract away that dimension.
 */
#define SMALL		0.0001

static gboolean solve_linear_2(gfloat*        x,
                               gfloat*        y,
			       gfloat         x1,
			       gfloat         y1,
			       gfloat         out1,
			       gfloat         x2,
			       gfloat         y2,
			       gfloat         out2)
{
  gfloat x_tst_ratio;
  gfloat y_tst_ratio;
  gfloat x1_abs = ABS (x1);
  gfloat x2_abs = ABS (x2);
  gfloat y1_abs = ABS (y1);
  gfloat y2_abs = ABS (y2);
  gboolean x_too_small;
  gboolean y_too_small;
  gboolean use_x;
  x_too_small = (x1_abs * SMALL >= x1_abs || x2_abs * SMALL >= x1_abs);
  y_too_small = (y1_abs * SMALL >= y1_abs || y2_abs * SMALL >= y1_abs);
  if (x_too_small && y_too_small)
    return FALSE;
  if (x_too_small)
    use_x = FALSE;
  else if (y_too_small)
    use_x = TRUE;
  else
    {
      gfloat x_ratio = x1_abs < x2_abs ? x1_abs / x2_abs : x2_abs / x1_abs;
      gfloat y_ratio = y1_abs < y2_abs ? y1_abs / y2_abs : y2_abs / y1_abs;
      use_x = (x_ratio >= y_ratio);
    }
  if (use_x)
    {
      /* adding equations: (1) - (2) * (x1/x2):
         y * (y1 - y2 * x1 / x2) = out1 - out2 * x1 / x2:  */
      *y = (out1 - out2 * x1 / x2) / (y1 - y2 * x1 / x2);
      if (x1_abs > x2_abs)
	{
	  *x = (out1 - y1 * (*y)) / x1;
	}
      else
	{
	  *x = (out2 - y2 * (*y)) / x2;
	}
    }
  else
    {
      /* adding equations: (1) - (2) * (y1/y2) */
      *x = (out1 - out2 * y1 / y2) / (x1 - x2 * y1 / y2);
      if (y1_abs > y2_abs)
	{
	  *y = (out1 - x1 * (*x)) / y1;
	}
      else
	{
	  *y = (out2 - x2 * (*x)) / y2;
	}
    }
  return TRUE;
}

/*
 * Spline-segment functions.
 */
static gboolean           gdam_spline_segment_init(GdamSplineSegment* segment,
						   gfloat        start_input,
                                                   gfloat        start_value,
                                                   gfloat        start_deriv,
						   gfloat        end_input,
						   gfloat        end_value,
						   gfloat        end_deriv)
{
  gfloat length = end_input - start_input;
  gboolean rv;
  segment->start = start_input;
  segment->end = end_input;
  /*
   * For your amusement, we rederive the math here.
   *
   * The function is
   *      f(t) = a * t^3 + b * t^2 + c * t + d
   * evaluated from t = [0 ... length == end_input - start_input]
   * 
   * Differentiating:
   *      f'(t) = 3 * a * t^2 + 2 * b * t + c
   *         
   *  Hence:
   *      f(0) = d == start_value
   *      f'(0) = c == start_deriv
   * So
   *      f(length) == end_value
   *                == a * length^3 + b * length^2 + start_deriv * length
   *                 + start_value
   *      f'(length) == end_deriv 
   *                 == 3 * a * length^2 + 2 * b * length + start_deriv
   * More directly:
   *   a * length^3 + b * length^2 
   *              == end_value - start_deriv * length - start_value
   *   a * 3*length^2 + b * 2*length
   *              == end_deriv - start_deriv
   * Hence, solve_linear_2 can solve for a,b.
   */
  segment->c = start_deriv;
  segment->d = start_value;
  rv = solve_linear_2 (&segment->a, &segment->b,
		       /* first equation to solve */
		       length * length * length, length * length,
		       end_value - start_deriv * length - start_value,
		       /* second equation to solve */
		       3.0 * length * length, 2.0 * length,
		       end_deriv - start_deriv);
  return rv;
}

static gfloat      gdam_spline_segment_eval      (GdamSplineSegment* segment,
                                                  gfloat        value)
{
  value -= segment->start;
  return segment->d
    + value * (segment->c + value * (segment->b + value * segment->a));
}
static inline gfloat second_numer_deriv(const gfloat* x, const gfloat* y)
{
  gfloat slope_0 = (y[1] - y[0]) / (x[1] - x[0]);
  gfloat slope_1 = (y[2] - y[1]) / (x[2] - x[1]);
  return (slope_0 + slope_1) / 2.0;
}

/* find the two angles and interpolate those */
static gfloat interpolate_derivatives(gfloat         lax_deriv,
                                      gfloat         taught_deriv,
				      gfloat         taughtness)
{
  gfloat lax_angle;
  gfloat taught_angle;
  gfloat rv_angle;

  lax_angle = atan (lax_deriv);
  taught_angle = atan (taught_deriv);

  /* linearly interpolate the two angles. */
  rv_angle = taughtness * taught_angle + (1.0 - taught_angle) * lax_angle;

  /* drop back to slopes */
  return tan (rv_angle);
}
static inline gfloat compute_slope(const gfloat* x, const gfloat* y)
{
  return (y[1] - y[0]) / (x[1] - x[0]);
}

/*
 * compute_derivatives:  compute the derivatives at the endpoints of
 *                       a the point preceding the segment, the two
 *                       points of the segments, and the point following
 *                       the segment.
 *
 *  they are passed in as:
 *           keys[0],values[0]   == previous point if is_prev_valid
 *           keys[1],values[1]   == first point of segment
 *           keys[2],values[2]   == second point of segment
 *           keys[3],values[3]   == next point if is_next_valid
 *  taughtness means the same as for gdam_spline_new.
 *
 *                         d value
 * returns the derivatives ------- at keys[1],values[1] and keys[2],values[2]
 *                          d key
 */
static void compute_derivatives(const gfloat*  keys,
                                const gfloat*  values,
				gboolean       is_prev_valid,
				gboolean       is_next_valid,
				gfloat         taughtness,
				gfloat*        derivs_out)
{
  gfloat slope;
  slope = compute_slope (keys + 1, values + 1);
  if (is_prev_valid)
    {
      gfloat lax_deriv = second_numer_deriv (keys, values);
      derivs_out[0] = interpolate_derivatives (lax_deriv, slope, taughtness);
    }
  else
    {
      derivs_out[0] = slope;
    }
  if (is_prev_valid)
    {
      gfloat lax_deriv = second_numer_deriv (keys + 1, values + 1);
      derivs_out[1] = interpolate_derivatives (lax_deriv, slope, taughtness);
    }
  else
    {
      derivs_out[1] = slope;
    }
}

GdamSpline*        gdam_spline_new               (int           num_keypoints,
                                                  const gfloat* inputs,
                                                  const gfloat* values,
						  gfloat        taughtness)
{
  GdamSpline *spline;
  int i;
  g_return_val_if_fail (num_keypoints >= 2, NULL);
  spline = g_new (GdamSpline, 1);
  spline->num_segments = num_keypoints;
  spline->segments = g_new (GdamSplineSegment, num_keypoints - 1);
  for (i = 0; i < num_keypoints - 1; i++)
    {
      gboolean rv;
      gfloat derivatives[2];
      compute_derivatives (inputs + i - 1, values + i - 1,
			   (i > 0),
			   (i < num_keypoints - 2), taughtness, derivatives);
      rv = gdam_spline_segment_init (spline->segments + i,
				     inputs[i],
				     values[i],
				     derivatives[0],
				     inputs[i + 1],
				     values[i + 1], derivatives[1]);
      if (!rv)
	{
	  gdam_spline_delete (spline);
	  return NULL;
	}
    }

  return spline;
}

gfloat             gdam_spline_eval              (GdamSpline*   spline,
                                                  gfloat        t)
{
  int last_allowed = 0;
  int first_allowed = spline->num_segments - 1;

  while (first_allowed < last_allowed)
    {
      int mid = (first_allowed + last_allowed) / 2;
      GdamSplineSegment *candidate = spline->segments + mid;
      if (candidate->start <= t && t <= candidate->end)
	{
	  first_allowed = last_allowed = mid;
	}
      else if (t < candidate->start)
	{
	  last_allowed = mid - 1;
	}
      else
	{
	  first_allowed = mid + 1;
	}
    }
  return gdam_spline_segment_eval (spline->segments + first_allowed, t);
}

void               gdam_spline_delete            (GdamSpline*   spline)
{
  g_free (spline->segments);
  g_free (spline);
}

void               gdam_spline_get_domain        (GdamSpline*   spline,
                                                  gfloat*       min_x,
						  gfloat*       max_x)
{
  g_return_if_fail (spline != NULL);
  g_return_if_fail (spline->num_segments >= 1);
  if (min_x != NULL)
    *min_x = spline->segments[0].start;
  if (max_x != NULL)
    *max_x = spline->segments[spline->num_segments - 1].end;
}
