/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamwaitpid.h"
#include <gdam/common/gdamdebug.h>
#include <signal.h>
#include <sys/wait.h>

typedef struct _GdamWaitHandler GdamWaitHandler;
struct _GdamWaitHandler {
	GdamWaitPidFunc     wait_func;
	gpointer            wait_func_data;
	GDestroyNotify      destroy;
};

static gboolean    gdam_wait_prepare              (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean    gdam_wait_check                (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean    gdam_wait_dispatch             (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static GdamWaitHandler*
gdam_wait_handler_new(GdamWaitPidFunc     wait_func,
                      gpointer            wait_func_data,
		      GDestroyNotify      destroy)
{
  GdamWaitHandler *handler;
  handler = g_new (GdamWaitHandler, 1);
  handler->wait_func = wait_func;
  handler->wait_func_data = wait_func_data;
  handler->destroy = destroy;
  return handler;
}

/* By pid. */
static GHashTable*    waiting_handlers = NULL;

/* A gsource that is invoking wait and looking for SIGCHLD. */
static guint          wait_source_id = 0;

static void start_wait_source();

void       gdam_wait_pid          (int                 pid,
                                   GdamWaitPidFunc     wait_func,
                                   gpointer            wait_func_data,
				   GDestroyNotify      destroy)
{
  GSList *handlers;
  GdamWaitHandler *handler;
  if (waiting_handlers == NULL)
    {
      waiting_handlers = g_hash_table_new (NULL, NULL);
      start_wait_source ();
    }

  handlers = g_hash_table_lookup (waiting_handlers, GINT_TO_POINTER (pid));
  handler = gdam_wait_handler_new (wait_func, wait_func_data, destroy);
  handlers = g_slist_prepend (handlers, handler);
  g_hash_table_insert (waiting_handlers, GINT_TO_POINTER (pid), handlers);
}

static void destroy_handler(GdamWaitHandler* handler)
{
  if (handler->destroy != NULL)
    (*handler->destroy) (handler->wait_func_data);
  g_free (handler);
}

void       gdam_wait_pid_remove   (int                 pid)
{
  GSList *handlers;
  if (waiting_handlers == NULL)
    return;

  handlers = g_hash_table_lookup (waiting_handlers, GINT_TO_POINTER (pid));
  g_slist_foreach (handlers, (GFunc) destroy_handler, NULL);
  g_slist_free (handlers);
  g_hash_table_remove (waiting_handlers, GINT_TO_POINTER (pid));
}

/*
 *                           gdam_wait_do_waiting
 *
 * Wait until there are no pids to wait for...
 */
static void run_handler(gpointer handler_ptr, gpointer info_ptr)
{
  GdamWaitHandler *handler = handler_ptr;
  GdamWaitPidInfo *info = info_ptr;
  (*handler->wait_func) (handler->wait_func_data, info);
}
static void gdam_wait_do_waiting()
{
  for (;;)
    {
      int status;
      int pid = waitpid (-1, &status, WNOHANG);
      GdamWaitPidInfo info;
      GSList *handlers;
      if (pid <= 0)
	break;
      info.pid = pid;
      info.exited = WIFEXITED (status);
      if (info.exited)
	{
	  info.d.exit_status = WEXITSTATUS (status);
	  info.dumped_core = FALSE;
	}
      else
	{
	  info.d.signal = WTERMSIG (status);
#ifdef WCOREDUMP
	  info.dumped_core = WCOREDUMP (status);
#else
	  info.dumped_core = FALSE;
#endif
	}
      handlers = g_hash_table_lookup (waiting_handlers,
				      GINT_TO_POINTER (pid));
      g_slist_foreach (handlers, run_handler, &info);
      g_slist_foreach (handlers, (GFunc) destroy_handler, &info);
      g_slist_free (handlers);
      g_hash_table_remove (waiting_handlers, GINT_TO_POINTER (pid));
    }
}

/*=====================================================================
 * The ``source'' that actively checks for terminated processes.
 */
static GSourceFuncs source_funcs = {
	gdam_wait_prepare,
	gdam_wait_check,
	gdam_wait_dispatch,
	NULL
};

static volatile gboolean got_sig_child = FALSE;

static void handle_sig_child(int signal_number)
{
  got_sig_child = TRUE;
}

static void start_wait_source()
{
  g_return_if_fail (wait_source_id == 0);

  signal (SIGCHLD, handle_sig_child);
  wait_source_id = g_source_add (G_PRIORITY_DEFAULT,
				 FALSE, &source_funcs, NULL, NULL, NULL);
}


static gboolean    gdam_wait_prepare              (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data)
{
  return got_sig_child;
}

static gboolean    gdam_wait_check                (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  return got_sig_child;
}

static gboolean    gdam_wait_dispatch             (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  got_sig_child = FALSE;
  gdam_wait_do_waiting ();
  return TRUE;
}


/*=====================================================================
 * Gracefully (?) killing a child process.
 *
 * Kill a process -TERM, then if millis_til_sigkill elapses and
 * we have not reaped the process, kill it -KILL and wait again.
 * Finally, print a warning if that fails.
 */
typedef struct _KillPidInfo KillPidInfo;
struct _KillPidInfo
{
  int pid;
  guint timer_id;
  GdamWaitHandler *handler;
};

static gboolean sigkill_failed(KillPidInfo *pid_info)
{
  pid_info->timer_id = 0;
  g_warning ("pid %d did not die, even with the KILL signal", pid_info->pid);
  return FALSE;
}

static gboolean kill_pid_info_timer_expired(KillPidInfo *pid_info)
{
  pid_info->timer_id = 0;
  gdam_debug ("pid %d did not die:  sending the KILL signal", pid_info->pid);
  kill (pid_info->pid, SIGKILL);
  pid_info->timer_id = g_timeout_add (3000, 
				      (GSourceFunc) sigkill_failed,
                                      pid_info);
  return FALSE;
}
static void done_waiting_for_killed_process(gpointer           user_data,
                                            GdamWaitPidInfo*   wait_info)
{
  KillPidInfo *pid_info = user_data;
  if (pid_info->timer_id != 0)
    {
      g_source_remove (pid_info->timer_id);
      pid_info->timer_id = 0;
    }
  g_free (pid_info);
}

void       gdam_wait_pid_kill_pid (int                 pid,
                                   int                 millis)
{
  KillPidInfo *pid_info;
  pid_info = g_new (KillPidInfo, 1);
  pid_info->pid = pid;
  pid_info->timer_id = g_timeout_add (millis,
                                      (GSourceFunc) kill_pid_info_timer_expired,
				      pid_info);
  pid_info->handler = gdam_wait_handler_new (done_waiting_for_killed_process,
                                             pid_info,
					     NULL);
}

