/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambeatinfo.h"
#include <glib.h>
#include <math.h>
#include <string.h>

GdamBeatInfo*	gdam_beat_info_new	()
{
  return g_new0 (GdamBeatInfo, 1);
}


static int internal_find_range (const GdamBeatInfo* bi, gdouble time,
				gboolean require_bpm)
{
  gdouble newfb;
  gdouble deffb;
  gboolean have_match;
  GList *currange;
  int num_ranges;
  int beatinfo_index;
  int new_beatinfo_index;
  num_ranges = g_list_length (bi->range_list);

  currange = bi->range_list;
  new_beatinfo_index = beatinfo_index = 0;
  newfb = 0;
  have_match = FALSE;

  /* Find the last (furthest into song) range which has a firstbeat
   * time before the new time.  Assumes that the range list is time-ordered.
   */
  /* TODO: take into account range end? */
  while (currange)
    {
      GdamBeatInfoRange *bir = (GdamBeatInfoRange *) currange->data;
      beatinfo_index++;
      /* If the range doesn't have a bpm (this usually means the
       * range is used as an index point) don't consider it.
       */
      if (bir->bpm <= 0 && require_bpm)
	{
	  currange = currange->next;
	  continue;
	}
      //g_return_val_if_fail (bir->bpm > 0, 0);
      /* If the beatinfo range has a range_start defined, and our time is
       * after it, then it is a candidate. 
       */
      if (bir->range_start > 0 
	  && bir->range_start <= time 
	  && bir->range_start > newfb)
	{
	  newfb = bir->first_beat;
	  new_beatinfo_index = beatinfo_index;
	  have_match = TRUE;
	}
      /* If the current time is after this beatinfo's firstbeat, it is 
       * a candidate.
       */
      if (bir->first_beat <= time && bir->first_beat > newfb)
	{
	  newfb = bir->first_beat;
	  new_beatinfo_index = beatinfo_index;
	  have_match = TRUE;
	}
      currange = currange->next;
    }
  /* Check whether default fb is a better candidate, or the last
   * resort.
   */ 
  deffb = bi->afirst_beat;
  /* XXX: handle range start of zero? */
  if (bi->arange_start > 0)
    deffb = bi->arange_start;
  if ((deffb <= time && deffb > newfb) || !have_match) 
    {
	  newfb = deffb;
	  new_beatinfo_index = 0;
    }
  return new_beatinfo_index;
}


/*=====================================================================*
 *                     gdam_beat_info_guess_range_index                *
 *								       *
 *  Given a time, this functions scans the extended beat info and      *
 *  determines the index of the beatinfo range that time is in.        *
 *                                                                     *
 * Parameters:                                                         *
 *     bi                the info to adjust.                           *
 *     time              time within the song, in seconds              *
 *=====================================================================*/
int gdam_beat_info_guess_range_index(const GdamBeatInfo* bi, gdouble time)
{
  return internal_find_range (bi, time, TRUE);
}
 
int gdam_beat_info_guess_index_index(const GdamBeatInfo* bi, gdouble time)
{
  return internal_find_range (bi, time, FALSE);
}

/* Find the next index point or beatinfo range start.  Assumes the
 * the beatinfo ranges are stored in the order of their firstbeat times.
 */
int gdam_beat_info_guess_next_index (const GdamBeatInfo* bi, gdouble time)
{
  gdouble newfb;
  gdouble rvfb;
  GList *currange;
  int rv;
  int num_ranges;
  int beatinfo_index;
  num_ranges = g_list_length (bi->range_list);

  currange = bi->range_list;
  beatinfo_index = 0;
  newfb = gdam_beat_info_get_first_beat (bi, 0);
  if (bi->arange_start > 0)
    newfb = bi->arange_start;

  rv = -1;
  rvfb = 0;

  /* If the first beat occurs after our time arg, then it is a candidate. */
  if (newfb > time)
    {
      rv = 0;
      rvfb = newfb;
    }

  while (currange)
    {
      GdamBeatInfoRange *bir = (GdamBeatInfoRange *) currange->data;
      gdouble bfb; 
      beatinfo_index++;
      bfb = bir->first_beat;
      if (bir->range_start > 0)
	bfb = bir->range_start;
      /* If the this beatinfo's firstbeat is after the current time, and
       * comes before any candidate, it is the new choice.
       */
      if ((bfb > time) && (rv < 0 || bfb < rvfb))
	{
	   rv = beatinfo_index;
	   rvfb = bfb;
	}
      currange = currange->next;
    }
  return rv;
}

gdouble
gdam_beat_info_get_next (const GdamBeatInfo    *beat_info,
			 gdouble                time)
{
  const GList *list;
  gdouble rv = time - 1.0;
  for (list = beat_info->range_list; list != NULL; list = list->next)
    {
      const GdamBeatInfoRange *range = list->data;
      if (range->range_end >= time)
	{
	  gdouble beat_period = 60.0 / range->bpm;
	  gdouble i = (time - range->first_beat) / beat_period;
	  gint next_beat_index = ceil (i);
	  gdouble beat_time;
	  if (next_beat_index < 0)
	    next_beat_index = 0;

	  /* TODO: clip next_beat_index to num_beats - 1 (which is annoying to compute) */

	  beat_time = range->first_beat + beat_period * next_beat_index;

	  if (next_beat_index >= 0
	   && range->range_start < beat_time 
	   && beat_time < range->range_end
	   && beat_time >= time)   /* XXX: this last condition is unnecessary */
	    {
	      if (rv < time || rv > beat_time)
		rv = beat_time;
	    }
	}
    }
  return rv;
}


void gdam_beat_info_set_range_first_beat (GdamBeatInfo* beat_info, 
					  int 		range_index,
					  gdouble	first_beat)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_if_fail (range_index <= num_ranges);
  g_return_if_fail (range_index >= 0);

  if (range_index == 0)
    {
      beat_info->afirst_beat = first_beat;
      return;
    }

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_if_fail (nbir != NULL);
  nbir->first_beat = first_beat;
}

gdouble gdam_beat_info_get_range_first_beat (const GdamBeatInfo* beat_info, 
					     int 		 range_index)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_val_if_fail (range_index <= num_ranges, -1);
  g_return_val_if_fail (range_index >= 0, -1);

  if (range_index == 0)
    {
      return beat_info->afirst_beat;
    }

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_val_if_fail (nbir != NULL, 0);
  return nbir->first_beat;
}

gdouble gdam_beat_info_get_range_bpm (const GdamBeatInfo* beat_info, 
				      int 		  range_index)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_val_if_fail (range_index <= num_ranges, -1);
  g_return_val_if_fail (range_index >= 0, -1);

  if (range_index == 0)
      return beat_info->abpm;

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_val_if_fail (nbir != NULL, 0);
  return nbir->bpm;
}

void    gdam_beat_info_set_range_bpm (GdamBeatInfo*       beat_info, 
				      int 		  range_index,
				      gdouble 		  bpm)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_if_fail (range_index <= num_ranges);
  g_return_if_fail (range_index >= 0);

  if (range_index == 0)
    {
      beat_info->abpm = bpm;
      return;
    }

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_if_fail (nbir != NULL);
  nbir->bpm = bpm;
}


const char* 	gdam_beat_info_get_range_name (const GdamBeatInfo*    beat_info,
					       int 		   range_index)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_val_if_fail (range_index <= num_ranges, NULL);
  g_return_val_if_fail (range_index >= 0, NULL);

  if (range_index == 0)
    {
      return beat_info->aname;
    }

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_val_if_fail (nbir != NULL, NULL);
  return nbir->name;
}

void        	gdam_beat_info_set_range_name (GdamBeatInfo *beat_info,
					       int 	     range_index,
					       const char   *name)
{
  gpointer next_data;
  GdamBeatInfoRange *nbir;
  int num_ranges = 0;
  int cur_idx;
  if (beat_info->range_list)
    num_ranges = g_list_length (beat_info->range_list);

  g_return_if_fail (name != NULL);
  g_return_if_fail (range_index <= num_ranges);
  g_return_if_fail (range_index >= 0);

  if (range_index == 0)
    {
      beat_info->aname = g_strdup (name);
      return;
    }

  next_data = g_list_nth_data (beat_info->range_list, range_index - 1);
  nbir = (GdamBeatInfoRange *) next_data;
  g_return_if_fail (nbir != NULL);
  nbir->name = g_strdup (name);
}


/*  TODO: take into account (and ignore) index points */
gdouble gdam_beat_info_guess_range_end (const GdamBeatInfo     *bi, 
					int 			index)
{
  int num_ranges = 0;
  int cur_idx;
  gdouble next_first_beat;
  next_first_beat = 0;
  if (bi->range_list)
    num_ranges = g_list_length (bi->range_list);
  g_return_val_if_fail (index >= 0, -1);
  cur_idx = index;

  /* If there is a range after us, find the beginning of the next */
  while (cur_idx < num_ranges && next_first_beat == 0)
    {
      gpointer next_data;
      GdamBeatInfoRange *nbir;
      next_data = g_list_nth_data (bi->range_list, cur_idx);
      nbir = (GdamBeatInfoRange *) next_data;
      g_return_val_if_fail (nbir != NULL, 0);
      /* If cur range is a full range (not an index point) we're done. */
      if (nbir->bpm > 0) 
	{
	  if (nbir->range_start > 0)
	    next_first_beat = nbir->range_start;
	  else
	    next_first_beat = nbir->first_beat;
	}
      cur_idx++;
    }
  return next_first_beat;
}


double		gdam_beat_info_get_bpm	(const GdamBeatInfo*	beat_info,
				       	 float			position)
{
  return beat_info ? beat_info->abpm : 0;
}

double		gdam_beat_info_get_first_beat	
					(const GdamBeatInfo*	beat_info,
				       	 float			position)
{
  return beat_info ? beat_info->afirst_beat : 0;
}

double		gdam_beat_info_get_beat_length	
					(const GdamBeatInfo*	beat_info,
				       	 float			position)
{
  gdouble rv = beat_info ? (60.0 / beat_info->abpm) : 0;
  if (rv < 0.01)
    {
      g_warning ("beat_length unreasonably small...");
      return 0;
    }
  return rv;
}

void		gdam_beat_info_set_bpm	(GdamBeatInfo*		beat_info,
					 double		new_bpm)
{
  g_return_if_fail (beat_info != NULL);
  beat_info->abpm = new_bpm;
}

void		gdam_beat_info_set_first_beat	
					(GdamBeatInfo*		beat_info,
					 double		new_first_beat)
{
  g_return_if_fail (beat_info != NULL);
  beat_info->afirst_beat = new_first_beat;
}

static void copy_one_bir(gpointer data, gpointer udata)
{
  GdamBeatInfoRange *bir;
  GdamBeatInfoRange *newbir;
  GList **newrl;
  //g_return_if_fail(data != NULL);
  //g_return_if_fail(udata != NULL);
  bir = (GdamBeatInfoRange *) data;
  newrl = (GList **) (udata);

  newbir = g_new0 (GdamBeatInfoRange, 1);
  *newbir = *bir;
  newbir->name = g_strdup (bir->name);
  *newrl = g_list_append (*newrl, newbir);
}

static void copy_one_loop(gpointer data, gpointer udata)
{
  GdamLoopInfo *li;
  GdamLoopInfo *newli;
  GList **newll;
  //g_return_if_fail(data != NULL);
  //g_return_if_fail(udata != NULL);
  li = (GdamLoopInfo *) data;
  newll = (GList **) (udata);

  newli = g_new0 (GdamLoopInfo, 1);
  *newli = *li;
  newli->name = g_strdup (li->name);
  *newll = g_list_append (*newll, newli);
}

GdamBeatInfo*	gdam_beat_info_copy	(const GdamBeatInfo*	beat_info)
{
  GdamBeatInfo *rv = g_new0 (GdamBeatInfo, 1);
  GList *new_rl = NULL;
  GList *new_ll = NULL;
  g_return_val_if_fail (beat_info != NULL, NULL);
  rv->abpm = beat_info->abpm;
  rv->afirst_beat = beat_info->afirst_beat;
  rv->arange_start = beat_info->arange_start;
  rv->arange_end = beat_info->arange_end;
  rv->aname = beat_info->aname ? g_strdup (beat_info->aname) : NULL;
  if (beat_info->range_list)
    {
      g_list_foreach (beat_info->range_list, copy_one_bir, &new_rl);
      rv->range_list = new_rl;
    }
  if (beat_info->loop_list)
    {
      g_list_foreach (beat_info->loop_list, copy_one_loop, &new_ll);
      rv->loop_list = new_ll;
    }
  return rv;
}

void free_loop(gpointer data, gpointer udata)
{
  GdamLoopInfo* li;
  li = (GdamLoopInfo*)data;
  g_free (li->name);
  g_free (li);
}

void free_range(gpointer data, gpointer udata)
{
  GdamBeatInfoRange* bir;
  bir = (GdamBeatInfoRange*)data;
  /* TODO: need to free this, but it is causing a seg fault...*/
  g_free (bir->name);
  g_free (bir);
}

/* XXX: need to free loops. */
void		gdam_beat_info_destroy	(GdamBeatInfo*		beat_info)
{
  if (beat_info->range_list)
    {
      g_list_foreach (beat_info->range_list, free_range, NULL);
      g_list_free (beat_info->range_list);
      beat_info->range_list = NULL;
    }
  if (beat_info->loop_list)
    {
      g_list_foreach (beat_info->loop_list, free_loop, NULL);
      g_list_free (beat_info->loop_list);
      beat_info->loop_list = NULL;
    }
  if (beat_info->aname)
    {
      g_free (beat_info->aname);
    }
  g_free (beat_info);
}

#define GDAM_IS_IN_DELTA(a,b,delta)		\
	( (a) >= (b) - (delta)			\
	  && (a) <= (b) + (delta) )


/*=====================================================================*
 *                     gdam_beat_info_add_loop                         *
 *                                                                     *
 * Parameters:                                                         *
 *     info              the info to adjust.                           *
 *     name              name of the loop.                             *
 *     start             time within the song where the loop begins,   *
 *     			 in seconds.                                   *
 *     length		 length of the loop, in seconds.               *
 *=====================================================================*/
gboolean             gdam_beat_info_add_loop(GdamBeatInfo*     info,
                                             const char*       name,
					     gdouble           start,
					     gdouble           length)
{
  GList *at;
  GdamLoopInfo *name_match = NULL;
  GdamLoopInfo *info_match = NULL;

  /*  Check if we have a loop by that name or info...  */
  for (at = info->loop_list; at != NULL; at = at->next)
    {
      GdamLoopInfo *li;
      li = (GdamLoopInfo *) at->data;
      if (li->name && name && strcmp (li->name, name) == 0 && !name_match)
	{
	  name_match = li;

	}
      else if ((GDAM_IS_IN_DELTA (li->start, start, 0.01)
		&& GDAM_IS_IN_DELTA (li->length, length, 0.01))
	       && !info_match)
	{
	  info_match = li;
	}
    }
  if (name_match)
    {
      name_match->start = start;
      name_match->length = length;
    }
  else if (info_match)
    {
      if (info_match->name)
	g_free (info_match->name);
      info_match->name = g_strdup (name);
    }
  else
    {
      GdamLoopInfo *loop_info;
      loop_info = g_new0 (GdamLoopInfo, 1);
      loop_info->name = g_strdup (name);
      loop_info->start = start;
      loop_info->length = length;
      info->loop_list = g_list_append (info->loop_list, loop_info);
    }
  /* TODO: emit songinfo_changed?? */
  return TRUE;
}

const GList* gdam_beat_info_get_loops(const GdamBeatInfo* beat_info)
{
  return beat_info->loop_list;
}





