/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <stdlib.h>
#include "gdambounce.h"
#include "gdamutil.h"
#include <gdam/common/gdamsample.h>
#include "gdammodel.h"

#define MY_ACTION_LIST (GdkDragAction) GDK_ACTION_COPY

typedef struct _delayinfo {
    GdamBounce* bounce;
    float amt;
} delayinfo;


/*
 *     ____ _                 ___       _ _   
 *    / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *   | |   | |/ _` / __/ __|  | || '_ \| | __|
 *   | |___| | (_| \__ \__ \  | || | | | | |_ 
 *    \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */                                            

/* Functions for Model class */
static gboolean gdam_bounce_match_beat_info(GdamModel* model, 
					    const GdamBeatInfo* match_info);
/* Functions for EchoBase class */
static gboolean gdam_bounce_refresh_echo_array(GdamEchoBase* echo_base);

static void gdam_bounce_init(GdamBounce* bounce)
{
  bounce->beat_factor = 0.5;
}

static void gdam_bounce_class_init(GdamBounceClass* klass);

GtkType gdam_bounce_get_type()
{
  static GtkType bounce_type = 0;
  if (!bounce_type)
    {
      static const GtkTypeInfo bounce_info = {
	"GdamBounce",
	sizeof (GdamBounce),
	sizeof (GdamBounceClass),
	(GtkClassInitFunc) gdam_bounce_class_init,
	(GtkObjectInitFunc) gdam_bounce_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      bounce_type = gtk_type_unique (GDAM_TYPE_ECHO_BASE, &bounce_info);
    }
  return bounce_type;
}


/*  _____                 _   _                 
 * |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 * |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */                                            

static gboolean gdam_bounce_refresh_echo_array(GdamEchoBase* echo_base)
{
  float delay = 0;
  float bpm;
  GdamBounce *bounce = GDAM_BOUNCE (echo_base);
  bounce->echo_base.num_echoes = 2;
  bounce->echo_base.echoes = g_renew (gdouble,
				      bounce->echo_base.echoes,
				      bounce->echo_base.num_echoes * 2);

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bounce)->beat_info, 0);
  if (bpm >= 1 && bounce->bounced)
    {
      //gdam_debug("setting bounce delay for %f bpm", bpm);
      delay = bounce->beat_factor * 60.0 / bpm;
    }
  else if (bounce->bounced)
    {
      delay = bounce->beat_factor * 1.0;
    }

  bounce->echo_base.echoes[0] = 1.0;
  bounce->echo_base.echoes[1] = (delay * 44100.0);
  bounce->echo_base.echoes[2] = 0.0;
  bounce->echo_base.echoes[3] = 44100;

  /*gdam_debug("bounce:refresh_array: [%d %d] [%d %d]",
     (int)(delay * 44100.0), (int)1.0, 44100, 0); */
  return TRUE;
}

static gboolean gdam_bounce_match_beat_info(GdamModel* model, 
					    const GdamBeatInfo* match_info)
{
  GdamBounce *bounce = GDAM_BOUNCE (model);
  GdamEchoBaseClass *echo_base_class;
  float match_bpm;
  echo_base_class = GDAM_ECHO_BASE_CLASS (GTK_OBJECT (model)->klass);
  match_bpm = gdam_beat_info_get_bpm (match_info, 0);
  gdam_beat_info_set_bpm (&model->beat_info, match_bpm);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (bounce));
  return TRUE;
}

static void gdam_bounce_arg_set_bounced(GtkObject* object, GdamArg* arg)
{
  GdamBounce *bounce;
  int old_bounced;
  bounce = GDAM_BOUNCE (object);
  old_bounced = bounce->bounced;
  bounce->bounced = GDAM_ARG_INT (arg);

  if ((bounce->bounced && old_bounced) || (!old_bounced && !bounce->bounced))
    return;

  gdam_debug ("gdam_bounce_arg_set_bounce: %s", bounce->bounced ?
	      "bounced" : "normal");
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (bounce));
}

static void gdam_bounce_arg_get_bounced(GtkObject* object, GdamArg* arg)
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (object);
  GDAM_ARG_INT (arg) = bounce->bounced;
}

static void gdam_bounce_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (object);
  bounce->beat_factor = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (bounce));
}

static void gdam_bounce_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (object);
  GDAM_ARG_DOUBLE (arg) = bounce->beat_factor;
}

static void 
gdam_bounce_arg_set_bounce_by_beat_factor (GtkObject* object, GdamArg* arg)
{
  GdamBounce *bounce = GDAM_BOUNCE (object);
  gdam_arg_set (object, "beat_factor", GDAM_ARG_DOUBLE (arg));
  gdam_arg_set (object, "bounced", bounce->bounced + 1);
}

static void gdam_bounce_arg_set_eighth_beat_factor(GtkObject* object, 
						  GdamArg* arg)
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (object);
  bounce->beat_factor = GDAM_ARG_DOUBLE (arg) / 8;
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (bounce));
}

static void gdam_bounce_arg_get_eighth_beat_factor(GtkObject* object, 
						  GdamArg* arg)
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (object);
  GDAM_ARG_DOUBLE (arg) = bounce->beat_factor * 8;
}
 
GdamBounce* gdam_bounce_new()
{
  GdamBounce *bounce;
  bounce = GDAM_BOUNCE (gtk_type_new (GDAM_TYPE_BOUNCE));
  return bounce;
}
 
static void gdam_bounce_class_init(GdamBounceClass* klass)
{
  GdamEchoBaseClass *echo_base_class = GDAM_ECHO_BASE_CLASS (klass);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  echo_base_class->refresh_echo_array = gdam_bounce_refresh_echo_array;
  model_class->match_beat_info = gdam_bounce_match_beat_info;
  gdam_arg_register (object_class,
		     "bounced",
		     GDAM_ARG_TYPE_INT,
		     gdam_bounce_arg_set_bounced,
		     gdam_bounce_arg_get_bounced, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bounce_arg_set_beat_factor,
		     gdam_bounce_arg_get_beat_factor,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bounce_by_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bounce_arg_set_bounce_by_beat_factor,
		     NULL,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "eighth_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bounce_arg_set_eighth_beat_factor,
		     gdam_bounce_arg_get_eighth_beat_factor,
		     NULL, NULL, GDAM_ARG_VITAL);
}
