/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <stdlib.h>
#include "gdamecho.h"
#include <gdam/common/gdamsample.h>
#include "gdammodel.h"
#include "gdamutil.h"

#define MY_ACTION_LIST (GdkDragAction) GDK_ACTION_COPY

/*
 *     ____ _                 ___       _ _   
 *    / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *   | |   | |/ _` / __/ __|  | || '_ \| | __|
 *   | |___| | (_| \__ \__ \  | || | | | | |_ 
 *    \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */                                            

enum {
	UPDATE_DELAY,
	LAST_SIGNAL
};

static guint echo_signals[LAST_SIGNAL];

static void gdam_echo_class_init(GdamEchoClass* klass);

static void gdam_echo_init(GdamEcho* echo)
{
  echo->gain = 1.0;
  echo->decay = 0.5;
  echo->echoes = 1;
}

GtkType gdam_echo_get_type()
{
  static GtkType echo_type = 0;
  if (!echo_type)
    {
      static const GtkTypeInfo echo_info = {
	"GdamEcho",
	sizeof (GdamEcho),
	sizeof (GdamEchoClass),
	(GtkClassInitFunc) gdam_echo_class_init,
	(GtkObjectInitFunc) gdam_echo_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      echo_type = gtk_type_unique (GDAM_TYPE_ECHO_BASE, &echo_info);
    }
  return echo_type;
}



/*
 *     ____      _ _ _                _        
 *    / ___|__ _| | | |__   __ _  ___| | _____ 
 *   | |   / _` | | | '_ \ / _` |/ __| |/ / __|
 *   | |__| (_| | | | |_) | (_| | (__|   <\__ \
 *    \____\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 */

static gboolean gdam_echo_refresh_echo_array(GdamEchoBase* echo_base)
{
  int i;
  GdamEcho *echo = GDAM_ECHO (echo_base);
  gfloat cur_delay = 0;
  gfloat fade = echo->decay;
  gfloat gain = echo->gain;
  gfloat delay = echo->delay;
  gfloat bpm, dummy_beat;
  echo->echo_base.num_echoes = echo->echoes;

  if (!echo->echoed)
    echo->echo_base.num_echoes = 0;

  echo->echo_base.num_echoes += 2;

  echo->echo_base.echoes = g_renew (gdouble, echo->echo_base.echoes,
				    echo->echo_base.num_echoes * 2);

  for (i = 0; i < (echo_base->num_echoes - 1); i++)
    {
      /*g_message("echo %d: volume %f, delay %f", i,gain,cur_delay); */
      echo->echo_base.echoes[2 * i + 0] = gain;
      echo->echo_base.echoes[2 * i + 1] = 44100.0 * cur_delay;
      cur_delay += delay;
      gain *= fade;
    }

  dummy_beat = GDAM_ECHO_BASE (echo)->dummy_beat;
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (echo)->beat_info, 0);
  if (bpm < 0.1)
    bpm = 60.0;
  echo->echo_base.echoes[2 * i + 0] = 0;
  echo->echo_base.echoes[2 * i + 1] = 44100.0 * dummy_beat * (60.0 / bpm);

  gdam_debug ("set %d echoes with delay %f, decay %f",
	      echo_base->num_echoes, delay, fade);
  return TRUE;
}

static gboolean gdam_echo_match_beat_info(GdamModel*          model,
					  const GdamBeatInfo* match_info)
{
  GdamEcho *echo = GDAM_ECHO (model);
  GdamEchoBaseClass *echo_base_class;
  float match_bpm;
  match_bpm = gdam_beat_info_get_bpm (match_info, 0);
  echo_base_class = GDAM_ECHO_BASE_CLASS (GTK_OBJECT (model)->klass);
  gdam_beat_info_set_bpm (&model->beat_info, match_bpm);
  gdam_echo_set_delay (echo, echo->beat_factor * match_bpm);
  return TRUE;
}

GdamEcho* 	gdam_echo_new()
{
  GdamEcho *echo;
  echo = GDAM_ECHO (gtk_type_new (GDAM_TYPE_ECHO));
  return echo;
}

static void gdam_echo_arg_set_echoed(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo;
  gboolean was_echoed;
  echo = GDAM_ECHO (object);
  was_echoed = echo->echoed;
  echo->echoed = GDAM_ARG_INT (arg);

  if ((was_echoed && !echo->echoed) || (!was_echoed && echo->echoed))
    gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
}

static void gdam_echo_arg_get_echoed(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo;
  echo = GDAM_ECHO (object);
  GDAM_ARG_INT (arg) = echo->echoed;
}

static void gdam_echo_arg_set_num_echoes(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo;
  echo = GDAM_ECHO (object);
  if (GDAM_ARG_INT (arg) == echo->echoes)
    return;
  echo->echoes = GDAM_ARG_INT (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
}

static void gdam_echo_arg_get_num_echoes(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo;
  echo = GDAM_ECHO (object);
  GDAM_ARG_INT (arg) = echo->echoes;
}

static void gdam_echo_arg_set_gain(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  if (echo->gain == GDAM_ARG_DOUBLE (arg))
    return;
  echo->gain = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
}

static void gdam_echo_arg_get_gain(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  GDAM_ARG_DOUBLE (arg) = echo->gain;
}

static void gdam_echo_arg_set_decay(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  if (echo->decay == GDAM_ARG_DOUBLE (arg))
    return;
  echo->decay = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
}

static void gdam_echo_arg_get_decay(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  GDAM_ARG_DOUBLE (arg) = echo->decay;
}

static void gdam_echo_arg_set_delay(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  if (echo->delay == GDAM_ARG_DOUBLE (arg))
    return;
  echo->delay = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
}

static void gdam_echo_arg_get_delay(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  GDAM_ARG_DOUBLE (arg) = echo->delay;
}

static void gdam_echo_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  float bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (echo)->beat_info, 0);
  gdouble delay;
  echo->beat_factor = GDAM_ARG_DOUBLE (arg);
  if (bpm > 0.1)
    delay = echo->beat_factor * 60.0 / bpm;
  else 
    delay = echo->beat_factor * 1.0;
  gdam_echo_set_delay (echo, delay);
}

static void gdam_echo_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamEcho *echo = GDAM_ECHO (object);
  GDAM_ARG_DOUBLE (arg) = echo->beat_factor;
}

static void real_set_delay(GdamEcho* echo, double delay)
{
  //echo->delay = delay;
  //gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (echo));
  g_warning ("echo: deprecated method 'real_set_delay'.");
}

void gdam_echo_set_delay(GdamEcho* echo, double delay)
{
  GdamArg arg;
  arg.name = "delay";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = delay;
  gdam_arg_set_object_arg (GTK_OBJECT (echo), &arg);
}

static void gdam_echo_class_init(GdamEchoClass* klass)
{
  GdamEchoBaseClass *echo_base_class = GDAM_ECHO_BASE_CLASS (klass);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  model_class->match_beat_info = gdam_echo_match_beat_info;
  echo_base_class->refresh_echo_array = gdam_echo_refresh_echo_array;
  klass->update_delay = real_set_delay;
  gdam_arg_register (object_class,
		     "echoed",
		     GDAM_ARG_TYPE_INT,
		     gdam_echo_arg_set_echoed,
		     gdam_echo_arg_get_echoed,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "num_echoes",
		     GDAM_ARG_TYPE_INT,
		     gdam_echo_arg_set_num_echoes,
		     gdam_echo_arg_get_num_echoes,
		     NULL, "0,1000", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_arg_set_gain,
		     gdam_echo_arg_get_gain, NULL, "-10,10", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "decay",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_arg_set_decay,
		     gdam_echo_arg_get_decay,
		     NULL, "-1.5,1.5", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "delay",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_arg_set_delay,
		     gdam_echo_arg_get_delay, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_echo_arg_set_beat_factor,
		     gdam_echo_arg_get_beat_factor,
		     NULL, "0,", GDAM_ARG_VITAL);
  /* TODO: remove this signal. */
  echo_signals[UPDATE_DELAY] =
    gtk_signal_new ("update_delay",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamEchoClass, update_delay),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, echo_signals, LAST_SIGNAL);
}
