/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaminfomanager.h"
#include <gtk/gtksignal.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>


typedef struct _GdamInfomanager GdamInfomanager;

struct _GdamInfomanager {
	GHashTable*		model_to_info;
	GdamModel*		lead_model;
	GHashTable*		radio_buttons;
};

typedef struct _GdamInfo GdamInfo;

struct _GdamInfo {
	GdamModel*		model;
	guint                   model_destroy_id;

	float			bpm;
	GdamModel*		bpm_driver;
	guint                   bpm_destroy_id;

	GdamModel*		beat_fraction_driver;
	guint                   beat_fraction_destroy_id;
	float			volume_req;

	GHashTable*		referers;
};
static void gdam_info_set_model(GdamInfo* info, GdamModel* model);
static void gdam_info_set_bpm_driver(GdamInfo* info, GdamModel* model);
static void gdam_info_set_beat_fraction_driver(GdamInfo* info, GdamModel*);

/** system infoman **/
static GdamInfomanager* infomanager = NULL;

static GdamInfomanager* gdam_infomanager_new()
{
  GdamInfomanager *rv;
  rv = g_new (GdamInfomanager, 1);
  rv->lead_model = NULL;
  rv->radio_buttons = g_hash_table_new (NULL, NULL);
  rv->model_to_info = g_hash_table_new (NULL, NULL);
  return rv;
}

static GdamInfo* gdam_info_new()
{
  GdamInfo *info;
  info = g_new0 (GdamInfo, 1);
  info->volume_req = -1;
  return info;
}
static void
submodel_destroy(GdamModel* model, GdamInfomanager* infomanager)
{
  GdamInfo *info;
  info = g_hash_table_lookup (infomanager->model_to_info, model);
  g_return_if_fail (info != NULL);
  info->model_destroy_id = 0;
  gdam_info_set_bpm_driver (info, NULL);
  gdam_info_set_beat_fraction_driver (info, NULL);
  g_hash_table_remove (infomanager->model_to_info, model);
  if (infomanager->lead_model == model)
    infomanager->lead_model = NULL;
  g_free (info);
}

static void gdam_info_set_model(GdamInfo* info, GdamModel* model)
{
  if (infomanager == NULL)
    infomanager = gdam_infomanager_new ();
  if (info->model)
    {
      gtk_signal_disconnect (GTK_OBJECT (info->model),
			     info->model_destroy_id);
      info->model_destroy_id = 0;
    }
  info->model = model;
  if (model != NULL)
    {
      info->model_destroy_id
	= gtk_signal_connect (GTK_OBJECT (model),
			      "destroy",
			      GTK_SIGNAL_FUNC (submodel_destroy),
			      infomanager);
    }
}
static void
bpm_driver_destroy(GdamModel* model, GdamInfo* info)
{
  info->bpm_driver = NULL;
  info->bpm_destroy_id = 0;
}
static void gdam_info_set_bpm_driver(GdamInfo* info, GdamModel* model)
{
  if (info->bpm_driver)
    {
      gtk_signal_disconnect (GTK_OBJECT (info->bpm_driver),
			     info->bpm_destroy_id);
      info->bpm_destroy_id = 0;
    }
  info->bpm_driver = model;
  if (info->bpm_driver != NULL)
    {
      info->bpm_destroy_id
	= gtk_signal_connect (GTK_OBJECT (info->bpm_driver),
			      "destroy",
			      GTK_SIGNAL_FUNC (bpm_driver_destroy), info);
    }
}

static void
beat_fraction_driver_destroy(GdamModel* model, GdamInfo* info)
{
  info->beat_fraction_driver = NULL;
  info->beat_fraction_destroy_id = 0;
}

static void
gdam_info_set_beat_fraction_driver(GdamInfo* info, GdamModel* model)
{
  if (info->beat_fraction_driver)
    {
      gtk_signal_disconnect (GTK_OBJECT (info->beat_fraction_driver),
			     info->beat_fraction_destroy_id);
      info->beat_fraction_destroy_id = 0;
    }
  info->beat_fraction_driver = model;
  if (info->beat_fraction_driver != NULL)
    {
      info->beat_fraction_destroy_id
	= gtk_signal_connect (GTK_OBJECT (info->beat_fraction_driver),
			      "destroy",
			      GTK_SIGNAL_FUNC (beat_fraction_driver_destroy),
			      info);
    }
}

static inline GdamInfo* lookup_info(GdamInfomanager* infomanager,
                                    GdamModel*   model)
{
  GHashTable *table;
  g_return_val_if_fail (infomanager != NULL, NULL);
  table = infomanager->model_to_info;
  return (GdamInfo *) g_hash_table_lookup (table, model);
}

static inline GdamInfo* force_lookup_info(GdamInfomanager* infomanager,
                                    GdamModel*   model)
{
  GdamInfo *info;
  info = lookup_info (infomanager, model);
  if (!info)
    {
      info = gdam_info_new ();
      gdam_info_set_model (info, model);
    }
  return info;
}

/* Add a model to the complist.  If an entry with the same model pointer
 *   already exists, we do not add and return false.
 */
void gdam_infomanager_add(GdamModel* model)
{
  GdamInfo *info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, model);
  if (info)
    return;
  info = gdam_info_new ();
  gdam_info_set_model (info, model);
  g_hash_table_insert (infomanager->model_to_info, model, info);
  return;
}

float gdam_infomanager_get_desired_bpm(GdamModel*   model)
{
  GdamInfo *info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, model);
  if (info && info->bpm_driver)
    {
      return gdam_beat_info_get_bpm (&info->bpm_driver->beat_info, 0);
    }
  else if (info && info->bpm > 0.0)
    {
      return info->bpm;
    }
  else if (infomanager->lead_model)
    {
      return gdam_beat_info_get_bpm (&infomanager->lead_model->beat_info, 0);
    }
  else
    {
      return -1;
    }
}

void gdam_infomanager_set_desired_bpm(GdamModel*   model,
			              float        bpm)
{
  GdamInfo *info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = force_lookup_info (infomanager, model);
  info->bpm = bpm;
}


/* Add a model to the model list.  If an entry with the same model pointer
 *   already exists, we do not add and return false.  Else, we try and inherit
 *   attributes from an existing model's entry.
 */
void gdam_infomanager_add_clone
		       (GdamModel*   model,
			GdamModel*   source_model)
{
  GdamInfo *info;
  GdamInfo *source_info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, model);
  if (info)
    return;
  source_info = lookup_info (infomanager, source_model);
  info = gdam_info_new ();
  gdam_info_set_model (info, model);
  if (source_info)
    {
      gdam_info_set_beat_fraction_driver (info,
					  source_info->beat_fraction_driver);
      gdam_info_set_bpm_driver (info, source_info->bpm_driver);
    }
  g_hash_table_insert (infomanager->model_to_info, model, info);
}

void gdam_infomanager_set_bpm_driver(GdamModel*       model,
				     GdamModel*       bpm_driver)
{
  GdamInfo *info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, model);
  if (!info)
    return;
  gdam_info_set_bpm_driver (info, bpm_driver);
  return;
}

void gdam_infomanager_set_beat_fraction_driver
                       (GdamModel*   model,
			GdamModel*   beat_fraction_driver)
{
  GdamInfo *info;
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, model);
  if (info)
    gdam_info_set_beat_fraction_driver (info, beat_fraction_driver);
  return;
}

static void print_gdam_info(gpointer key, gpointer data, gpointer user_data)
{
  GdamInfo *info = (GdamInfo *) data;
  fprintf (stderr, "COMP: %p  ", info->model);
  fprintf (stderr, "BPMDRIVER: %p  ", info->bpm_driver);
  fprintf (stderr, "BEATPERCENTDRIVER: %p\n", info->beat_fraction_driver);
}

/*  Prints the list of registered models.
 */
void gdam_infomanager_print_list()
{
  if (!infomanager)
    infomanager = gdam_infomanager_new ();
  g_hash_table_foreach (infomanager->model_to_info, print_gdam_info, 0);
  fprintf (stderr, "\n");
}

void gdam_infomanager_set_lead_model(GdamModel* lead_model)
{
  GdamInfo *info;
  if (infomanager == NULL)
    infomanager = gdam_infomanager_new ();
  info = lookup_info (infomanager, lead_model);
  if (!info)
    gdam_infomanager_add (lead_model);
  if (!infomanager->lead_model)
    {
      gdam_arg_set (GTK_OBJECT (lead_model), "anchored", 1);
    }
  else if (infomanager->lead_model != lead_model)
    {
      gdam_arg_set (GTK_OBJECT (infomanager->lead_model), "anchored", 0);
    }
  infomanager->lead_model = lead_model;
}


gboolean	gdam_infomanager_get_sync_info	(GdamModel*   requester,
						 guint*		  driver_id,
						 GdamBeatInfo*	  driver_info)
{
  float lead_fb;
  float lead_bpm;
  if (infomanager == NULL)
    infomanager = gdam_infomanager_new ();
  if (!infomanager->lead_model || !infomanager->lead_model->source_id)
    return FALSE;
  *driver_id = infomanager->lead_model->source_id;
  /* TODO: generalize this */
  lead_fb =
    gdam_beat_info_get_first_beat (&infomanager->lead_model->beat_info, 0);
  lead_bpm = gdam_beat_info_get_bpm (&infomanager->lead_model->beat_info, 0);
  gdam_beat_info_set_first_beat (driver_info, lead_fb);
  gdam_beat_info_set_bpm (driver_info, lead_bpm);
  return TRUE;
}
