/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdaminputdevice.h"
#include <gtk/gtksignal.h>
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdamplugin.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

static GtkObjectClass* parent_class = NULL;

typedef struct _GdamInputDeviceConfiguration GdamInputDeviceConfiguration;
struct _GdamInputDeviceConfiguration 
{
  char                 *device_name;
  char                 *patch_file;
  char                 *device_type;
  GdamInputDevice      *input_device;
};
static gboolean gdam_input_device_real_event(GdamInputDevice  *input_device, 
                                             GdamInputEvent   *event,
                                             GdamInputTrap    *input_trap);
static void gdam_input_device_real_new_trap(GdamInputDevice  *input_device, 
                                            GdamInputTrap    *input_trap);

static GHashTable *config_table = NULL;
static GHashTable *config_type_table = NULL;


typedef gboolean (*GtkSignal_BOOL__POINTER_POINTER) (GtkObject * object,
                                                     gpointer arg1,
                                                     gpointer arg2,
                                                     gpointer user_data);
void
gdam_marshal_BOOL__POINTER_POINTER (GtkObject * object,
                                    GtkSignalFunc func,
                                    gpointer func_data, GtkArg * args)
{
  GtkSignal_BOOL__POINTER_POINTER rfunc;
  gboolean *return_val;
  return_val = GTK_RETLOC_BOOL (args[2]);
  rfunc = (GtkSignal_BOOL__POINTER_POINTER) func;
  *return_val = (*rfunc) (object,
                          GTK_VALUE_POINTER (args[0]),
                          GTK_VALUE_POINTER (args[1]),
                          func_data);
}


enum 
{
  NEW_TRAP,
  EVENT,
  LAST_SIGNAL
};

static guint input_signals[LAST_SIGNAL] = { 0 };

typedef struct _GdamInputHook GdamInputHook;

static void gdam_input_device_init(GdamInputDevice* input_device)
{
}
static void gdam_input_device_class_init(GdamInputDeviceClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  input_signals[NEW_TRAP]
    = gtk_signal_new ("new-trap",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamInputDeviceClass,
					 new_trap),
		      gtk_marshal_NONE__POINTER,
		      GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  input_signals[EVENT]
    = gtk_signal_new ("event",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamInputDeviceClass,
					 event),
		      gdam_marshal_BOOL__POINTER_POINTER,
		      GTK_TYPE_BOOL, 2, GTK_TYPE_POINTER, GTK_TYPE_POINTER);
  gtk_object_class_add_signals (object_class, input_signals, LAST_SIGNAL);
  klass->event = gdam_input_device_real_event;
  klass->new_trap = gdam_input_device_real_new_trap;
}

/* Map from GdamInputEvent*  =>  GdamInputTrap*  */
static GHashTable* trap_table = NULL;
static GHashTable* trap_id_table = NULL;
static guint       last_trap_id = 0;


GdamInputTrap*   gdam_input_trap_find        (GdamInputEvent*    input_event)
{
  if (trap_table == NULL)
    return NULL;

  return g_hash_table_lookup (trap_table, input_event);
}

GdamInputTrap*   gdam_input_trap_find_or_create(GdamInputEvent*    input_event)
{
  GdamInputTrap *trap;
  GdamInputDevice *device = input_event->input_device;
  g_return_val_if_fail (device != NULL, NULL);
  if (trap_table == NULL)
    {
      trap_table = g_hash_table_new (
				     (GHashFunc) gdam_input_event_hash,
				     (GCompareFunc) gdam_input_event_equal);
      trap_id_table = g_hash_table_new (NULL, NULL);
    }

  trap = g_hash_table_lookup (trap_table, input_event);
  if (trap != NULL)
    return trap;

  trap = g_new0 (GdamInputTrap, 1);
  trap->event = gdam_input_event_canonicalize (input_event);
  if (trap->event == NULL)
    {
      g_free (trap);
      g_warning ("error canonicalized input-event");
      return NULL;
    }
  trap->hooks = NULL;
  trap->trap_id = ++last_trap_id;
  g_hash_table_insert (trap_id_table, GUINT_TO_POINTER (last_trap_id), trap);
  g_hash_table_insert (trap_table, trap->event, trap);
  device->traps = g_slist_prepend (device->traps, trap);
  gtk_signal_emit (GTK_OBJECT (input_event->input_device),
		   input_signals[NEW_TRAP], trap);
  return trap;
}

GdamInputTrap* gdam_input_trap_get_by_id(guint id)
{
  if (trap_id_table == NULL)
    return NULL;
  return g_hash_table_lookup (trap_id_table, GUINT_TO_POINTER (id));
}

struct _GdamInputHook {
	GdamInputTrap*		trap;
	GdamInputHookFunc       hook_func;
	GDestroyNotify          destroy;
	gpointer                user_data;
	GtkObject*              alive_object;
	guint			alive_object_destroy;
	guint			hook_id;
	gint 			block_count;
};

static GHashTable* hook_id_table = NULL;
static guint       last_hook_id  = 0;

static guint gdam_input_hook_alloc_id(GdamInputHook* hook)
{
  guint id;
  if (hook_id_table == NULL)
    hook_id_table = g_hash_table_new (NULL, NULL);
  id = ++last_hook_id;
  hook->hook_id = id;
  g_hash_table_insert (hook_id_table, GINT_TO_POINTER (id), hook);
  return id;
}
static GdamInputHook* gdam_input_hook_find_by_id(guint hook)
{
  return g_hash_table_lookup (hook_id_table, GINT_TO_POINTER (hook));
}

static void gdam_input_hook_destroy(GdamInputHook* input_hook)
{
  if (input_hook->trap != NULL)
    {
      GdamInputTrap *trap = input_hook->trap;
      trap->hooks = g_slist_remove (trap->hooks, input_hook);
    }
  if (input_hook->alive_object != NULL)
    gtk_signal_disconnect (input_hook->alive_object,
			   input_hook->alive_object_destroy);
  if (input_hook->destroy != NULL)
    (*input_hook->destroy) (input_hook->user_data);
  g_hash_table_remove (hook_id_table, GINT_TO_POINTER (input_hook->hook_id));
  g_free (input_hook);
}

static void alive_object_destroyed(GtkObject* object, GdamInputHook* hook)
{
  hook->alive_object = NULL;
  hook->alive_object_destroy = 0;
  gdam_input_hook_destroy (hook);
}

guint            gdam_input_hook_connect     (GdamInputTrap     *trap,
                                              GdamInputHookFunc  hook_func,
                                              GDestroyNotify     destroy,
                                              gpointer           user_data,
                                              GtkObject         *alive_object)
{
  GdamInputHook *hook;
  g_return_val_if_fail (trap != NULL, 0);
  hook = g_new0 (GdamInputHook, 1);
  hook->trap = trap;
  hook->hook_func = hook_func;
  hook->destroy = destroy;
  hook->user_data = user_data;
  hook->block_count = 0;
  hook->alive_object = alive_object;
  hook->alive_object_destroy
    = gtk_signal_connect (alive_object,
			  "destroy",
			  GTK_SIGNAL_FUNC (alive_object_destroyed), hook);
  trap->hooks = g_slist_append (trap->hooks, hook);
  return gdam_input_hook_alloc_id (hook);
}

void             gdam_input_hook_disconnect  (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);

  if (hook == NULL)
    return;
  gdam_input_hook_destroy (hook);
}
void             gdam_input_hook_block       (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);
  g_return_if_fail (hook != NULL);
  hook->block_count++;
}
void             gdam_input_hook_unblock     (guint              hook_id)
{
  GdamInputHook *hook;
  hook = gdam_input_hook_find_by_id (hook_id);
  g_return_if_fail (hook != NULL);
  g_return_if_fail (hook->block_count > 0);
  hook->block_count--;

}

/*
 * The callstack for emit is basically:
 *
 *       gdam_input_trap_emit
 *            GdamInputEvent::event
 *                  hook_funcs
 *
 * ie everything goes through the virtual `event',
 * so that perhaps derivers can act as event filters.
 * Also allows use gtk_signal_emission_stop.
 */
static gboolean gdam_input_device_real_event(GdamInputDevice*  input_device, 
	                                     GdamInputEvent*   event,
	                                     GdamInputTrap*    input_trap)
{
  GSList *list;
  gboolean rv = FALSE;
  for (list = input_trap->hooks; list != NULL; list = list->next)
    {
      GdamInputHook *hook;
      hook = list->data;
      if (hook->block_count == 0)
	if ((*hook->hook_func) (event, hook->user_data))
	  rv = TRUE;
    }
  return rv;
}
gboolean         gdam_input_trap_emit        (GdamInputTrap*     input_trap,
                                              GdamInputEvent*    input_event)
{

  gboolean rv;
  if (input_trap == NULL)
    input_trap = gdam_input_trap_find_or_create (input_event);
  g_return_val_if_fail (input_trap != NULL, FALSE);
  rv = FALSE;
  gtk_signal_emit (GTK_OBJECT (input_event->input_device),
		   input_signals[EVENT], input_event, input_trap, &rv);
  return rv;
}

static void gdam_input_device_real_new_trap(GdamInputDevice*  input_device, 
	                                    GdamInputTrap*    input_trap)
{
  gdam_debug ("gdam_input_device_real_new_trap [id %d]", input_trap->trap_id);
}

GtkType gdam_input_device_get_type()
{
  static GtkType input_device_type = 0;
  if (!input_device_type)
    {
      static const GtkTypeInfo input_device_info = {
	"GdamInputDevice",
	sizeof (GdamInputDevice),
	sizeof (GdamInputDeviceClass),
	(GtkClassInitFunc) gdam_input_device_class_init,
	(GtkObjectInitFunc) gdam_input_device_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      parent_class = gtk_type_class (parent);
      input_device_type = gtk_type_unique (parent, &input_device_info);

      gdam_input_event_get_type ();
      gdam_input_event_range_get_type ();
      gdam_input_event_int_get_type ();
      gdam_input_event_string_get_type ();

    }

  return input_device_type;
}

/*
 * Global default trap.
 */
static GdamInputTrap* default_trap = NULL;
void gdam_input_trap_set_default(GdamInputTrap* trap)
{
  default_trap = trap;
}
GdamInputTrap* gdam_input_trap_get_default()
{
  return default_trap;
}


/*
 * Naming the traps.
 */
void             gdam_input_trap_set_name    (GdamInputTrap*     trap,
                                              const char*        name)
{
  if (trap->name)
    g_free (trap->name);
  trap->name = g_strdup (name);
  trap->name_is_generated = FALSE;
}
const char*      gdam_input_trap_get_name    (GdamInputTrap*     trap)
{
  if (trap->name == NULL)
    {
      trap->name = gdam_input_event_stringify (trap->event);
      trap->name_is_generated = TRUE;
    }
  return trap->name;
}

/*
 * Iterating the Traps.
 */

typedef struct _IterateData IterateData;
struct _IterateData {
	gpointer iterate_data;
	GdamInputTrapIterator iterate_func;
};

static void for_one_device(const char*       device_name,
                            GdamInputDeviceConfiguration* configuration,
			    IterateData*      data)
{
  GdamInputDevice *device = configuration->input_device;
  if (device != NULL)
    g_slist_foreach (device->traps,
		     (GFunc) data->iterate_func, data->iterate_data);
}

void             gdam_input_trap_forall      (GdamInputDevice*   device,
					      GdamInputTrapIterator iter_func,
					      gpointer           iter_data)
{
  IterateData iterate_data;
  iterate_data.iterate_func = iter_func;
  iterate_data.iterate_data = iter_data;
  if (device == NULL)
    {
      g_hash_table_foreach (config_table,
			    (GHFunc) for_one_device, &iterate_data);
    }
  else if (device->traps != NULL)
    {
      g_slist_foreach (device->traps, (GFunc) iter_func, iter_data);
    }
}

/*=====================================================================
 * Emitting signals when a event is tripped.
 */

typedef struct _GdamInputSignal GdamInputSignal;
struct _GdamInputSignal {
	GdamInputTrap*	trap;

	GtkObject*      object;

	guint		connect_id;

	char*		signal_name;
	char*		handler_name;
};

static void gdam_input_signal_free(GdamInputSignal* sig)
{
  g_free (sig->signal_name);
  g_free (sig->handler_name);
  g_free (sig);
}



/*=====================================================================
 * Setting arguments when a event is tripped.
 */

typedef struct _GdamInputArg GdamInputArg;
struct _GdamInputArg {
	GdamInputTrap*	trap;
	GdamArg		arg;

	GtkObject*      object;

	guint		connect_id;

	/* For range arguments. */
	gdouble		min_value, max_value;
	gint		is_logarithmic;
	gint		is_exponent;

	/* For variable arguments. */
	gdouble 	factor;
	/* For constant and variable arguments. */
	gboolean	is_delta;
	/* For constant arguments. */
	gboolean	is_factor;
};

static void gdam_input_arg_free(GdamInputArg* arg)
{
  gdam_arg_destruct (1, &arg->arg);
  g_free (arg);
}

static gboolean on_event_set_signal(GdamInputEvent*     input_event,
			            gpointer            user_data)
{
  GdamInputSignal *sig;
  sig = (GdamInputSignal *) user_data;
  gdam_debug ("emitting signal %s on object of type %s", sig->signal_name,
              gtk_type_name (sig->object->klass->type));
  gtk_signal_emit_by_name (sig->object, sig->signal_name);
  return TRUE;
}


static gdouble translate_range_value (GdamInputEventRange *range,
                                      GdamInputArg *input_arg)
{
  gdouble value;
  if (range->min_value == range->max_value)
    {
      value = range->min_value;
    }
  else if (input_arg->is_logarithmic)
    {
      gdouble offset;
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value);
      offset = 1.0 - input_arg->min_value;
      value = value * (log (input_arg->max_value + offset) 
	              - log (input_arg->min_value + offset));
      value += log (input_arg->min_value + offset);
      value = exp (value) - offset;
    }
  else if (input_arg->is_exponent) 
    {
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value)
	* (input_arg->max_value - input_arg->min_value) + input_arg->min_value;
      value = ldexp (1, value);
    }
  else 
    {
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value)
	* (input_arg->max_value - input_arg->min_value) + input_arg->min_value;
    }
  return value;
}



/* 
 * `event_type' must be a GdamInputEventRange.
 */
static gboolean on_range_set_arg(GdamInputEvent*     input_event,
			         gpointer            user_data)
{
  GdamInputEventRange *range;
  GdamInputArg *input_arg = user_data;
  gdouble value;
  range = (GdamInputEventRange *) input_event;
  value = translate_range_value (range, input_arg);
  GDAM_ARG_DOUBLE (&input_arg->arg) = value;
  gdam_arg_set_object_arg (input_arg->object, &input_arg->arg);
  return TRUE;
}

static gboolean on_range_set_arg_int(GdamInputEvent*     input_event,
			             gpointer            user_data)
{
  GdamInputEventRange *range;
  GdamInputArg *input_arg = user_data;
  gdouble value;
  range = (GdamInputEventRange *) input_event;
  value = translate_range_value (range, input_arg);
  GDAM_ARG_INT (&input_arg->arg) = (int) value;
  gdam_arg_set_object_arg (input_arg->object, &input_arg->arg);
  return TRUE;
}

/* Given a trap corresponding to an EventRange, add a hook which
 * sets a GdamArg on an object. */
/* XXX: better way to pass extra args w/o changing function call */
guint gdam_input_hook_connect_range(GdamInputTrap*   trap,
                                    GtkObject*       object,
				    const char*      arg_name,
				    gdouble          min_value,
				    gdouble          max_value,
				    gboolean         is_logarithmic,
				    gboolean         is_exponent)
{
  GdamInputArg *input_arg;
  gboolean is_double;

  if (!gdam_input_event_type_is_a (trap->event->event_type,
				   gdam_input_event_range_get_type ()))
    {
      g_warning ("only GdamInputEventRange's (not %s) may be connected"
		 " through %s",
		 gdam_input_event_type_name (trap->event->event_type),
		 __FUNCTION__);
      return 0;
    }

  g_return_val_if_fail (trap != NULL, 0);

  input_arg = g_new0 (GdamInputArg, 1);
  input_arg->trap = trap;
  input_arg->arg.name = g_strdup (arg_name);
  switch (gdam_arg_query_type (object->klass->type, arg_name))
    {
      case GDAM_ARG_TYPE_INT:
	is_double = FALSE;
	break;
      case GDAM_ARG_TYPE_DOUBLE:
	is_double = TRUE;
	break;
      default:
	g_warning ("%s::%s was neither double nor int in %s",
	    	gtk_type_name (object->klass->type), arg_name, __FUNCTION__);
	return 0;
    }
  if (is_double)
    {
      input_arg->arg.type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (&input_arg->arg) = min_value;
    }
  else
    {
      input_arg->arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&input_arg->arg) = (int) min_value;
    }
  input_arg->object = object;
  input_arg->min_value = min_value;
  input_arg->max_value = max_value;
  input_arg->is_logarithmic = is_logarithmic;
  input_arg->is_exponent = is_exponent;
  input_arg->connect_id = gdam_input_hook_connect
    (trap,
     is_double ? on_range_set_arg : on_range_set_arg_int,
     (GDestroyNotify) gdam_input_arg_free, input_arg, object);

  return input_arg->connect_id;
}


guint gdam_input_hook_connect_signal(GdamInputTrap*   trap,
                                     GtkObject*       object,
				     const char*      signal_name,
				     const char*      handler_name)
{
  GdamInputSignal *input;
  input = g_new0 (GdamInputSignal, 1);
  input->trap = trap;
  input->signal_name = g_strdup(signal_name);
  input->handler_name = g_strdup(handler_name);
  input->object = object;
  input->connect_id = gdam_input_hook_connect
    (trap,
     on_event_set_signal,
     (GDestroyNotify) gdam_input_signal_free, input, object);
  return input->connect_id;
}

static gdouble arg_value_as_float (GdamArg* arg)
{
  if (arg->type == GDAM_ARG_TYPE_DOUBLE)
    return GDAM_ARG_DOUBLE (arg);
  else if (arg->type == GDAM_ARG_TYPE_INT)
    return GDAM_ARG_INT (arg);
  else
    {
      g_warning ("type mismatch in arg_value, name %s", arg->name);
      return 0;
    }
}


/*
 * Setting constant arguments when a GdamInputEvent is trapped.
 */
static gboolean on_const_set_arg(GdamInputEvent*     input_event,
			         gpointer            user_data)
{
  GdamInputArg *input_arg = user_data;
  int num_out;
  const char* arg_name;
  GdamArg *args;
  arg_name = input_arg->arg.name;

  if (input_arg->is_delta || input_arg->is_factor)
    {
      gdouble val = 0;
      GdamArg new_arg;

      /* Get modifying value */
      val = arg_value_as_float (&input_arg->arg);

      /* Add to or multiply with current value */
      gdam_arg_get_by_name (input_arg->object, 1, &arg_name, &num_out, &args);
      if (num_out == 1)
	{
	  if (input_arg->is_delta)
	    val += arg_value_as_float (args);
	  else
	    val *= arg_value_as_float (args);
	}
      else 
	{
	  g_warning ("got %d matches for arg name %s", num_out, arg_name);
	}

      /* Set new value */
      new_arg.name = g_strdup (arg_name);
      new_arg.type = input_arg->arg.type;
      if (input_arg->arg.type == GDAM_ARG_TYPE_DOUBLE)
	GDAM_ARG_DOUBLE (&new_arg) = val;
      else if (input_arg->arg.type == GDAM_ARG_TYPE_INT)
	GDAM_ARG_INT (&new_arg) = (gint)val;
      gdam_debug ("constant binding mod: setting arg %s in object %s to %f",
		  input_arg->arg.name, 
		  gtk_type_name (input_arg->object->klass->type), val);
      gdam_arg_set_object_arg (input_arg->object, &new_arg);
      g_free (new_arg.name);
      return TRUE;
    }
  gdam_debug ("constant binding: setting arg of name %s in object %s to %f",
	      input_arg->arg.name, 
	      gtk_type_name (input_arg->object->klass->type), 
	      arg_value_as_float (&input_arg->arg));
  gdam_arg_set_object_arg (input_arg->object, &input_arg->arg);
  return TRUE;
}

guint gdam_input_hook_connect_const(GdamInputTrap*   trap,
                                    GtkObject*       object,
				    GdamArg*         arg)
{
  return gdam_input_hook_connect_const_full(trap, object, arg, 0, 0);
}    

guint gdam_input_hook_connect_const_full(GdamInputTrap*   trap,
                                         GtkObject*       object,
				         GdamArg*         arg,
					 gboolean	  is_delta,
					 gboolean         is_factor)
{
  GdamInputArg *input_arg;

  input_arg = g_new0 (GdamInputArg, 1);
  input_arg->trap = trap;
  input_arg->arg = *arg;
  input_arg->object = object;
  input_arg->is_delta = is_delta;
  input_arg->is_factor = is_factor;
  input_arg->connect_id = gdam_input_hook_connect
    (trap,
     on_const_set_arg,
     (GDestroyNotify) gdam_input_arg_free, input_arg, object);
  gdam_debug ("new constant binding: arg %s val %f obj %s delta %d factor %d",
              input_arg->arg.name, arg_value_as_float (arg),
	      gtk_type_name (object->klass->type),
	      is_delta, is_factor);
  return input_arg->connect_id;
}    


/*
 * Setting variable arguments when a GdamInputEventInt is trapped.
 */
static gboolean on_variable_set_arg(GdamInputEvent*     input_event,
			            gpointer            user_data)
{
  GdamInputArg *input_arg = user_data;
  GdamInputIntEvent* int_event = (GdamInputIntEvent*)input_event;
  gdouble val;
  GdamArg* arg;
  arg = &input_arg->arg;

  val = int_event->number * input_arg->factor;
  if (input_arg->is_delta)
    /* XXX: get val, add, set val */
    {
      int num_out;
      const char* arg_name;
      GdamArg *args;
      arg_name = arg->name;
      gdam_arg_get_by_name (input_arg->object, 1, &arg_name, &num_out, &args);
      if (num_out == 1)
	{
	  if (args->type == GDAM_ARG_TYPE_DOUBLE)
	    val += GDAM_ARG_DOUBLE (args);
	  else if (args->type == GDAM_ARG_TYPE_INT)
	    val += GDAM_ARG_INT (args);
	  else
	    {
	      g_warning ("type mismatch in var_set");
	      return FALSE;
	    }
	}
    }

  if (arg->type == GDAM_ARG_TYPE_INT)
    GDAM_ARG_INT (arg) = (int) (val + 0.5);
  else if (arg->type == GDAM_ARG_TYPE_DOUBLE)
    GDAM_ARG_DOUBLE (arg) = val;
  else
    {
      g_warning ("type mismatch in var_set");
      return FALSE;
    }

  gdam_debug ("variable event, val %f from number %d", val, int_event->number);

  gdam_arg_set_object_arg (input_arg->object, arg);
  return TRUE;
}

guint gdam_input_hook_connect_variable(GdamInputTrap*   trap,
                                       GtkObject*       object,
				       GdamArg*         arg,
				       gdouble		factor,
				       int		is_delta)
{
  GdamInputArg *input_arg;

  g_return_val_if_fail (trap != NULL, 0);

  if (!gdam_input_event_type_is_a (trap->event->event_type,
				   gdam_input_event_int_get_type ()))
    {
      g_warning ("only GdamInputEventInt's (not %s) may be connected"
		 " through %s",
		 gdam_input_event_type_name (trap->event->event_type),
		 __FUNCTION__);
      return 0;
    }

  input_arg = g_new0 (GdamInputArg, 1);
  input_arg->trap = trap;
  input_arg->arg = *arg;
  input_arg->object = object;
  input_arg->factor = factor;
  input_arg->is_delta = is_delta;
  input_arg->connect_id = gdam_input_hook_connect
    (trap,
     on_variable_set_arg,
     (GDestroyNotify) gdam_input_arg_free, input_arg, object);
  return input_arg->connect_id;
}    


/*
 *      ``Master Patch File''
 *
 * Store collections of triplets:
 *         device_name		patch_file	Device
 *
 * So that the devices can be opened on demand.
 */
void gdam_input_device_configure(const char* device_name,
                                 const char* patch_file,
				 const char* device_type)
{
  GdamInputDeviceConfiguration *configuration;
  g_return_if_fail (device_name != NULL);
  g_return_if_fail (device_type != NULL);
  configuration = g_new0 (GdamInputDeviceConfiguration, 1);
  configuration->device_name = g_strdup (device_name);
  configuration->patch_file = g_strdup (patch_file);
  configuration->device_type = g_strdup (device_type);
  configuration->input_device = NULL;
  if (config_table == NULL)
    {
      config_table = g_hash_table_new (g_str_hash, g_str_equal);
      config_type_table = g_hash_table_new (g_str_hash, g_str_equal);
    }
  g_hash_table_insert (config_table,
		       configuration->device_name, configuration);
  g_hash_table_insert (config_type_table,
		       configuration->device_type, configuration);
}

/* Read input.devices:
 *
 *         /dev/midi???        patch-file-name GdamMidiDevice
 */
gboolean gdam_input_device_load_database(const char* filename)
{
  FILE *fp;
  gboolean must_free_filename = FALSE;
  char buf[4096];
  int lineno = 0;

  /* Find ~/.gdam/midi.devices */
  if (filename == NULL)
    {
      gdam_search_path_initialize ();
      filename = gdam_search_path ("input.devices",
				   gdam_search_path_config,
       				   GDAM_SEARCH_PATH_FIND_FILE);
      must_free_filename = TRUE;
    }

  if (filename == NULL)
    {
      g_message ("no input.devices file -- "
		 "no input patch information files loaded");
      return FALSE;
    }

  fp = fopen (filename, "r");
  if (fp == NULL)
    {
      g_warning ("error opening input device database file %s: %s",
		 filename, strerror (errno));
      return FALSE;
    }
  while (fgets (buf, sizeof (buf), fp))
    {
      const char *start_dev_name;
      const char *start_path_fname;
      const char *tmp;
      char *module_name;
      char *path_fname;
      char *dev_name;
      ++lineno;
      start_dev_name = buf;
      while (*start_dev_name != '\0' && isspace (*start_dev_name))
	start_dev_name++;
      if (*start_dev_name == '\0' || *start_dev_name == '#')
	continue;
      tmp = start_dev_name;
      while (*tmp != '\0' && !isspace (*tmp))
	tmp++;
      if (tmp == start_dev_name || *tmp == '\0')
	{
	  g_warning ("too few arguments on line %d", lineno);
	  continue;
	}
      dev_name = alloca ((tmp - start_dev_name) + 1);
      memcpy (dev_name, start_dev_name, tmp - start_dev_name);
      dev_name[tmp - start_dev_name] = '\0';

      start_path_fname = tmp;
      while (*start_path_fname != '\0' && isspace (*start_path_fname))
	start_path_fname++;
      tmp = start_path_fname;
      while (*tmp != '\0' && !isspace (*tmp))
	tmp++;
      if (tmp == start_path_fname)
	{
	  g_warning ("no path on line %d", lineno);
	  continue;
	}
      path_fname = alloca ((tmp - start_path_fname) + 1);
      memcpy (path_fname, start_path_fname, tmp - start_path_fname);
      path_fname[tmp - start_path_fname] = '\0';

      while (*tmp != '\0' && isspace (*tmp))
	tmp++;

      if (*tmp)
	{
	  const char *end = tmp;
	  while (*end != '\0' && !isspace (*end))
	    end++;
	  module_name = alloca (end + 1 - tmp);
	  memcpy (module_name, tmp, end - tmp);
	  module_name[end - tmp] = '\0';
	}
      else
	{
	  module_name = NULL;
	  g_warning ("line %d in %s didn't have "
		     "device_type for %s", lineno, filename, dev_name);
	  continue;
	}
      gdam_input_device_configure (dev_name, path_fname, module_name);
    }

  fclose (fp);
  return TRUE;
}

GdamInputDevice* gdam_input_device_get_by_name(const char* device_name)
{
  GdamInputDeviceConfiguration *configuration;
  GdamInputDevice *input_device;
  GtkType type;
  if (config_table == NULL)
    return NULL;
  configuration = g_hash_table_lookup (config_table, device_name);
  if (configuration == NULL)
    return NULL;
  if (configuration->input_device != NULL)
    return configuration->input_device;
  type = gdam_plugin_get_type_by_name ("model", configuration->device_type);
  if (type == 0)
    {
      g_warning ("couldn't load input device module %s",
		 configuration->device_type);
      return NULL;
    }
  if (!gtk_type_is_a (type, GDAM_TYPE_INPUT_DEVICE))
    {
      g_warning ("%s was not an input device", configuration->device_type);
      return NULL;
    }
  input_device = (GdamInputDevice *) gtk_type_new (type);
  if (!gdam_input_device_open (input_device, configuration->device_name))
    {
      g_warning ("error opening %s", configuration->device_name);
      gtk_object_sink ((GtkObject *) input_device);
      return NULL;
    }
  configuration->input_device = input_device;
  gtk_object_ref ((GtkObject *) configuration->input_device);
  gtk_object_sink ((GtkObject *) configuration->input_device);
  return input_device;
}

GdamInputDevice* gdam_input_device_get_by_type(const char* type_name)
{
  GdamInputDeviceConfiguration *configuration;
  if (config_type_table == NULL)
    return NULL;
  configuration = g_hash_table_lookup (config_type_table, type_name);
  if (configuration == NULL)
    return NULL;
  return gdam_input_device_get_by_name (configuration->device_name);
}

gboolean gdam_input_device_open(GdamInputDevice* input_device,
                                const char*      device_name)
{
  GdamInputDeviceClass *klass;
  klass = GDAM_INPUT_DEVICE_CLASS (GTK_OBJECT (input_device)->klass);
  g_free (input_device->device_name);
  input_device->device_name = g_strdup (device_name);
  g_return_val_if_fail (klass->open != NULL, FALSE);
  return (*klass->open) (input_device);
}
