/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdaminputdevicefile.h"
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

static void gdam_input_device_file_init(GdamInputDeviceFile* input_device_file)
{
}
void            gdam_input_device_file_close(GdamInputDevice* input_device)
{
  GdamInputDeviceFile *device_file = GDAM_INPUT_DEVICE_FILE (input_device);

  if (device_file->io_channel)
    {
      /* XXX: verify that the watch we added is destroyed
         automatically. */
      int fd = g_io_channel_unix_get_fd  (device_file->io_channel);
      (*GDAM_INPUT_DEVICE_FILE_GET_CLASS (input_device)->close_fd) 
			(device_file, fd);
      g_io_channel_unref (device_file->io_channel);
      device_file->io_channel = NULL;
    }
}

static gboolean gdam_input_device_file_parse_raw (GIOChannel*   io_channel,
                                          	  GIOCondition  condition,
					  	  gpointer      user_data)
{
  GdamInputDeviceFile *device = GDAM_INPUT_DEVICE_FILE (user_data);
  GdamInputDeviceFileClass *file_class;
  guint8 midi_buf[1024];
  guint num_read;
  guint index;
  GIOError io_error;
  const char *filename = GDAM_INPUT_DEVICE (device)->device_name;
  io_error = g_io_channel_read (io_channel,
				(char *) midi_buf,
				sizeof (midi_buf), &num_read);
  if (io_error == G_IO_ERROR_AGAIN)
    return TRUE;
  if (io_error != G_IO_ERROR_NONE)
    {
      g_warning ("error reading from midi device %s", filename);
      return FALSE;
    }
  if (num_read <= 0)
    {
      g_warning ("MIDI device closed...");
      return FALSE;
    }
  file_class = GDAM_INPUT_DEVICE_FILE_GET_CLASS (device);
  g_return_val_if_fail (file_class->process != NULL, FALSE);
  return (*file_class->process) (device, midi_buf, num_read);
}


static int   gdam_input_device_file_open_fd (GdamInputDeviceFile* device_file)
{
  int fd;
  GdamInputDevice *device;
  device = GDAM_INPUT_DEVICE (device_file);
  fd = open (device->device_name, O_RDONLY);
  if (fd < 0)
    {
      g_warning ("couldn't open %s: %s",
		 device->device_name, strerror (errno));
      return -1;
    }
  /* TODO: need we always specify this?? */
  fcntl (fd, F_SETFL, fcntl (fd, F_GETFL) | O_NONBLOCK);
  return fd;
}

static void gdam_input_device_file_close_fd (GdamInputDeviceFile *device,
    					     int		  fd)
{
  close (fd);
}


gboolean        gdam_input_device_file_open(GdamInputDevice* input_device)
{
  int fd;
  GdamInputDeviceFile *device_file = GDAM_INPUT_DEVICE_FILE (input_device);
  GdamInputDeviceFileClass *klass;
  klass = GDAM_INPUT_DEVICE_FILE_CLASS (GTK_OBJECT (input_device)->klass);

  if (device_file->io_channel != NULL)
    gdam_input_device_file_close (input_device);

  fd = (*GDAM_INPUT_DEVICE_FILE_GET_CLASS (input_device)->open_fd) 
                   (device_file);

  if (fd < 0)
    return FALSE;

  g_message ("device %s opened on fd %d", input_device->device_name, fd);
  device_file->io_channel = g_io_channel_unix_new (fd);
  g_io_add_watch (device_file->io_channel,
		  G_IO_IN, gdam_input_device_file_parse_raw, input_device);
  return TRUE;
}

static GIOError gdam_input_device_file_read (GdamInputDeviceFile* device, 
    				             char *buf,
				             guint max_read,
				             guint *num_read)
{
  return g_io_channel_read (device->io_channel,
			    buf,
			    max_read,
			    num_read);
}

static void
gdam_input_device_file_class_init(GdamInputDeviceClass* input_device_class)
{
  GdamInputDeviceFileClass* klass;
  klass = GDAM_INPUT_DEVICE_FILE_CLASS (input_device_class);
  input_device_class->open = gdam_input_device_file_open;
  input_device_class->close = gdam_input_device_file_close;
  klass->close_fd = gdam_input_device_file_close_fd;
  klass->open_fd = gdam_input_device_file_open_fd;
}

GtkType gdam_input_device_file_get_type()
{
  static GtkType input_device_file_type = 0;
  if (!input_device_file_type)
    {
      static const GtkTypeInfo input_device_file_info = {
	"GdamInputDeviceFile",
	sizeof (GdamInputDeviceFile),
	sizeof (GdamInputDeviceFileClass),
	(GtkClassInitFunc) gdam_input_device_file_class_init,
	(GtkObjectInitFunc) gdam_input_device_file_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_INPUT_DEVICE;
      input_device_file_type = gtk_type_unique (parent,
						&input_device_file_info);
    }
  return input_device_file_type;
}
