#include "gdaminputevent.h"
#include <stdlib.h>

typedef struct _GdamInputEventType GdamInputEventType;
struct _GdamInputEventType {
	guint			type;
	GdamInputEventTypeInfo	type_info;
	guint			parent_type;
};

static GArray* event_types = NULL;
static GHashTable* event_types_by_name = NULL;

#define EVENT_FIRST	30000

guint gdam_input_event_get_unique(guint               event_parent,
                                  GtkType             device_type,
                                  GdamInputEventTypeInfo* type)
{
  GdamInputEventType event_type;
  guint id;
  GdamInputEventType *parent;
  if (event_types == NULL)
    {
      event_types = g_array_new (FALSE, FALSE, sizeof (GdamInputEventType));
      event_types_by_name = g_hash_table_new (g_str_hash, g_str_equal);
      g_return_val_if_fail (event_parent == 0, 0);
      parent = NULL;
    }
  else
    {
      guint index = event_parent - EVENT_FIRST;
      /*g_return_val_if_fail (index >= 0, 0); */
      g_return_val_if_fail (index < event_types->len, 0);
      parent = &g_array_index (event_types, GdamInputEventType, index);
    }
  id = event_types->len + EVENT_FIRST;
  event_type.type = id;
  event_type.type_info = *type;
  event_type.parent_type = event_parent;
  if (parent != NULL)
    {
      GdamInputEventTypeInfo *info = &event_type.type_info;
      if (info->stringify == NULL)
	info->stringify = parent->type_info.stringify;
      if (info->hash == NULL)
	info->hash = parent->type_info.hash;
      if (info->equal == NULL)
	info->equal = parent->type_info.equal;
      if (info->to_keyvalues == NULL)
	info->to_keyvalues = parent->type_info.to_keyvalues;
      if (info->from_keyvalues == NULL)
	info->from_keyvalues = parent->type_info.from_keyvalues;
      if (info->canonicalize == NULL)
	info->canonicalize = parent->type_info.canonicalize;
      if (info->destroy == NULL)
	info->destroy = parent->type_info.destroy;
    }
  g_array_append_val (event_types, event_type);
  g_hash_table_insert (event_types_by_name,
		       (char *) event_type.type_info.event_type_name,
		       GINT_TO_POINTER (id));
  return id;
}

static GdamInputEventType* get_type(guint type)
{
  int index;
  if (type < EVENT_FIRST)
    return NULL;
  if (event_types == NULL)
    return NULL;
  if (type >= EVENT_FIRST + event_types->len)
    return NULL;
  index = type - EVENT_FIRST;
  return &g_array_index (event_types, GdamInputEventType, index);
}

GdamInputEventTypeInfo*
        gdam_input_event_get_type_info_by_name(const char* name)
{
  guint index;
  GdamInputEventType *private_type;
  if (event_types_by_name == NULL)
    return NULL;
  index = (int) g_hash_table_lookup (event_types_by_name, name);
  if (index < EVENT_FIRST)
    return NULL;
  if (index >= EVENT_FIRST + event_types->len)
    return NULL;
  index -= EVENT_FIRST;
  private_type = &g_array_index (event_types, GdamInputEventType, index);
  return &private_type->type_info;
}
int     gdam_input_event_get_type_id_by_name  (const char* name)
{
  if (event_types_by_name == NULL)
    return 0;
  return GPOINTER_TO_INT (g_hash_table_lookup (event_types_by_name, name));
}
gboolean gdam_input_event_type_is_a(guint child, guint parent)
{
  g_return_val_if_fail (EVENT_FIRST <= child, FALSE);
  g_return_val_if_fail (child < EVENT_FIRST + event_types->len, FALSE);
  g_return_val_if_fail (EVENT_FIRST <= parent, FALSE);
  g_return_val_if_fail (parent < EVENT_FIRST + event_types->len, FALSE);
  while (child != 0 && child != parent)
    child = g_array_index (event_types,
			   GdamInputEventType,
			   child - EVENT_FIRST).parent_type;
  if (child == parent)
    return TRUE;
  return FALSE;
}

/*
 * Base InputEvent.
 */

static char*	gdam_input_event_real_stringify(GdamInputEvent* event)
{
  g_warning ("gdam_input_event_stringify called");
  return NULL;
}
static unsigned	gdam_input_event_real_hash(GdamInputEvent* event)
{
  return GPOINTER_TO_INT (event->input_device)
    ^ (event->event_type << 16) ^ (event->input_id << 8);
}
static gboolean	gdam_input_event_real_equal(GdamInputEvent* event1,
	                                    GdamInputEvent* event2)
{
  return event1->input_device == event2->input_device
    && event1->event_type == event2->event_type
    && event1->input_id == event2->input_id;
}
gboolean gdam_input_event_parse_device(GdamInputEvent* event,
                                       const char* device_name)
{
  if (device_name == NULL)
    {
      event->input_device = NULL;
      return FALSE;
    }
  event->input_device = gdam_input_device_get_by_name (device_name);
  return (event->input_device != NULL);
}

static GHashTable* gdam_input_event_real_to_keyvalues(GdamInputEvent* event)
{
  GHashTable *table;
  const char *device_name;
  table = g_hash_table_new (g_str_hash, g_str_equal);
  gdam_input_store_int (table, "input_id", event->input_id);

  device_name = event->input_device->device_name;
  gdam_input_store_string (table, "device", device_name);
  return table;
}
static GdamInputEvent* gdam_input_event_real_from_keyvalues(int type,
                                                            GHashTable* table)
{
  GdamInputEvent *event;
  char *device_name;
  event = g_new (GdamInputEvent, 1);
  event->event_type = type;
  event->input_id = gdam_input_fetch_int (table, "input_id");

  device_name = g_hash_table_lookup (table, "device");
  gdam_input_event_parse_device (event, device_name);
  return event;
}

static GdamInputEvent* gdam_input_event_real_canonicalize(GdamInputEvent* event)
{
  g_warning ("gdam_input_event_real_canonicalize called");
  return NULL;
}
static GdamInputEventTypeInfo input_event_info = {
	"GdamInputEvent",
	gdam_input_event_real_stringify,
	gdam_input_event_real_hash,
	gdam_input_event_real_equal,
	gdam_input_event_real_to_keyvalues,
	gdam_input_event_real_from_keyvalues,
	gdam_input_event_real_canonicalize,
	(void(*)(GdamInputEvent*))g_free
};

guint gdam_input_event_get_type()
{
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique (0,
					GDAM_TYPE_INPUT_DEVICE,
					&input_event_info);
  return type;
}

/*
 * Range InputEvent.
 */

static char* gdam_input_event_range_stringify(GdamInputEvent* event)
{
  return g_strdup_printf ("%s, range %d",
			  event->input_device->device_name, event->input_id);
}

static GHashTable* gdam_input_event_range_to_keyvalues(GdamInputEvent* event)
{
  GHashTable *table;
  GdamInputEventRange *range = (GdamInputEventRange *) event;
  const char *device_name;
  table = g_hash_table_new (g_str_hash, g_str_equal);
  gdam_input_store_int (table, "input_id", event->input_id);

  device_name = event->input_device->device_name;
  gdam_input_store_string (table, "device", device_name);
  gdam_input_store_double (table, "min_value", range->min_value);
  gdam_input_store_double (table, "max_value", range->max_value);
  gdam_input_store_double (table, "value", range->value);
  return table;
}
static GdamInputEvent* gdam_input_event_range_from_keyvalues(int type,
                                                            GHashTable* table)
{
  GdamInputEvent *event;
  GdamInputEventRange *range;
  char *device_name;
  range = g_new0 (GdamInputEventRange, 1);
  event = (GdamInputEvent *) range;
  event->event_type = type;
  event->input_id = gdam_input_fetch_int (table, "input_id");
  range->min_value = gdam_input_fetch_double (table, "min_value");
  range->max_value = gdam_input_fetch_double (table, "max_value");
  range->value = gdam_input_fetch_double (table, "value");
  device_name = g_hash_table_lookup (table, "device");
  gdam_input_event_parse_device (event, device_name);
  return event;
}

static GdamInputEvent*
gdam_input_event_range_canonicalize(GdamInputEvent* event)
{
  GdamInputEventRange *range = (GdamInputEventRange *) event;
  GdamInputEventRange *new_range;
  new_range = g_new (GdamInputEventRange, 1);
  *new_range = *range;
  return (GdamInputEvent *) new_range;
}

static GdamInputEventTypeInfo input_event_range_info = {
	"GdamInputEventRange",
	gdam_input_event_range_stringify,
	NULL,
	NULL,
	gdam_input_event_range_to_keyvalues,
	gdam_input_event_range_from_keyvalues,
	gdam_input_event_range_canonicalize,
	NULL
};

guint gdam_input_event_range_get_type()
{
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique (gdam_input_event_get_type (),
					GDAM_TYPE_INPUT_DEVICE,
					&input_event_range_info);
  return type;
}


/*
 * Signal InputEvent
 */
static char* gdam_input_event_signal_stringify(GdamInputEvent* event)
{
  return g_strdup_printf ("%s, signal, input_id = %d",
			  event->input_device->device_name, event->input_id);
}

static GdamInputEvent*
gdam_input_event_signal_canonicalize(GdamInputEvent* event)
{
  GdamInputEventSignal *signal_event = (GdamInputEventSignal *) event;
  GdamInputEventSignal *new_signal;
  new_signal = g_new (GdamInputEventSignal, 1);
  *new_signal = *signal_event;
  new_signal->signal_length = 0;
  new_signal->signal_name = NULL;
  new_signal->handler_length = 0;
  new_signal->handler_name = NULL;
  g_message ("signal_canon called, e_type=%s", 
             gdam_input_event_type_name (event->event_type));
  return (GdamInputEvent *) new_signal;
}

static GdamInputEventTypeInfo input_event_signal_info = {
	"GdamInputEventSignal",
	gdam_input_event_signal_stringify,
	NULL,
	NULL,
	NULL,
	NULL,
	gdam_input_event_signal_canonicalize,
	NULL
};

guint gdam_input_event_signal_get_type()
{
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique (gdam_input_event_get_type (),
					GDAM_TYPE_INPUT_DEVICE,
					&input_event_signal_info);
  return type;
}


/*
 * Int InputEvent
 */
static char* gdam_input_event_int_stringify(GdamInputEvent* event)
{
  return g_strdup_printf ("%s, discrete %d",
			  event->input_device->device_name, event->input_id);
}
static GdamInputEvent*
gdam_input_event_int_canonicalize(GdamInputEvent* event)
{
  GdamInputIntEvent *int_event = (GdamInputIntEvent *) event;
  GdamInputIntEvent *new_int;
  new_int = g_new (GdamInputIntEvent, 1);
  *new_int = *int_event;
  new_int->enum_name = NULL;
  return (GdamInputEvent *) new_int;
}

static GdamInputEventTypeInfo input_event_int_info = {
	"GdamInputEventInt",
	gdam_input_event_int_stringify,
	NULL,
	NULL,
	NULL,			/* perhaps we should store value ??? */
	NULL,			/* perhaps we should fetch value ??? */
	gdam_input_event_int_canonicalize,
	NULL
};

guint gdam_input_event_int_get_type()
{
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique (gdam_input_event_get_type (),
					GDAM_TYPE_INPUT_DEVICE,
					&input_event_int_info);
  return type;
}


/*
 * String InputEvent
 */
static char* gdam_input_event_string_stringify(GdamInputEvent* event)
{
  return g_strdup_printf ("%s, string data %d",
			  event->input_device->device_name, event->input_id);
}
static GdamInputEvent*
gdam_input_event_string_canonicalize(GdamInputEvent* event)
{
  GdamInputStringEvent *string_event = (GdamInputStringEvent *) event;
  GdamInputStringEvent *new_string;
  new_string = g_new (GdamInputStringEvent, 1);
  *new_string = *string_event;
  new_string->length = 0;
  new_string->data = NULL;
  return (GdamInputEvent *) new_string;
}

static GdamInputEventTypeInfo input_event_string_info = {
	"GdamInputEventString",
	gdam_input_event_string_stringify,
	NULL,
	NULL,
	NULL,			/* perhaps we should store value ??? */
	NULL,			/* perhaps we should fetch value ??? */
	gdam_input_event_string_canonicalize,
	NULL
};

guint gdam_input_event_string_get_type()
{
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique (gdam_input_event_get_type (),
					GDAM_TYPE_INPUT_DEVICE,
					&input_event_string_info);
  return type;
}

/*
 * Wrappers to safely call the virtual methods.
 */


unsigned        gdam_input_event_hash        (GdamInputEvent* event)
{
  GdamInputEventType *type;
  g_return_val_if_fail (event != NULL, 0);
  type = get_type (event->event_type);
  g_return_val_if_fail (type != NULL, 0);
  return (*type->type_info.hash) (event);
}
gboolean        gdam_input_event_equal       (GdamInputEvent* event1,
                                              GdamInputEvent* event2)
{
  GdamInputEventType *type;
  if (event1->event_type != event2->event_type)
    return FALSE;
  type = get_type (event1->event_type);
  g_return_val_if_fail (type != NULL, FALSE);
  return (*type->type_info.equal) (event1, event2);
}
char*           gdam_input_event_stringify   (GdamInputEvent* event)
{
  GdamInputEventType *type;
  g_return_val_if_fail (event != NULL, NULL);
  type = get_type (event->event_type);
  g_return_val_if_fail (type != NULL, NULL);
  return (*type->type_info.stringify) (event);
}

GdamInputEvent* gdam_input_event_canonicalize(GdamInputEvent* event)
{
  GdamInputEventType *type;
  type = get_type (event->event_type);
  g_return_val_if_fail (type != NULL, NULL);
  return (*type->type_info.canonicalize) (event);
}

const char* gdam_input_event_type_name(int ev_type)
{
  GdamInputEventType *type = get_type (ev_type);
  g_return_val_if_fail (type != NULL, "unknown");
  return type->type_info.event_type_name;
}

void            gdam_input_event_destroy     (GdamInputEvent* event)
{
  GdamInputEventType *type;
  type = get_type (event->event_type);
  g_return_if_fail (type != NULL);
  (*type->type_info.destroy) (event);
}


int     gdam_input_fetch_int     (GHashTable* str_table,
                                  const char* name)
{
  char *value;
  value = g_hash_table_lookup (str_table, name);
  if (value == NULL)
    return -1;
  return atoi (value);
}
char*   gdam_input_fetch_string  (GHashTable* str_table,
                                  const char* name)
{
  char *value;
  value = g_hash_table_lookup (str_table, name);
  return g_strdup ((value == NULL) ? "" : value);
}
gdouble gdam_input_fetch_double  (GHashTable* str_table,
                                  const char* name)
{
  char *value;
  value = g_hash_table_lookup (str_table, name);
  if (value == NULL)
    return -1;
  return strtod (value, NULL);
}
void    gdam_input_store_int     (GHashTable* str_table,
                                  const char* name,
				  int value)
{
  char buf[255];
  g_snprintf (buf, sizeof (buf), "%d", value);
  gdam_input_store_string (str_table, name, buf);
}
void    gdam_input_store_double  (GHashTable* str_table,
                                  const char* name,
				  gdouble value)
{
  char buf[255];
  g_snprintf (buf, sizeof (buf), "%6.6g", value);
  gdam_input_store_string (str_table, name, buf);
}
void    gdam_input_store_string  (GHashTable* str_table,
                                  const char* name,
				  const char* value)
{
  char *old;
  old = g_hash_table_lookup (str_table, name);
  if (old)
    {
      g_hash_table_insert (str_table, (char *) name, g_strdup (value));
      g_free (old);
    }
  else
    {
      g_hash_table_insert (str_table, g_strdup (name), g_strdup (value));
    }
}
