/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdammix.h"
#include <string.h>
#include <gtk/gtksignal.h>

enum {
	ELEMENT_DELETED,
	ELEMENT_ADDED,
	ELEMENT_CHANGED,
	ELEMENTS_CHANGED,
	LAST_SIGNAL
};

static guint mix_signals[LAST_SIGNAL] = { 0 };
static void gdam_mix_init(GdamMix* mix)
{
  mix->beat_length = 1.0;
}
static int gdam_mix_play_request(GdamProducer*	      producer,
				 GdamPlayInfo*	      play_info,
				 int                  subsource_id,
				 GdamRequestBuilder*  builder)
{
  GSList *list;
  GdamMix *mix;
  GdamRequest *request;
  gint64 *time_sources;
  int num_elements;
  int index;
  GdamPlayInfo subplayinfo;
  GdamArg args[3];
  guint nargs;
  GdamModel *model = GDAM_MODEL (producer);

  g_return_val_if_fail (subsource_id == 0, 0);

  mix = GDAM_MIX (producer);
  num_elements = g_slist_length (mix->elements);
  time_sources = g_new (gint64, num_elements * 2);

  if (play_info == NULL)
    memset (&subplayinfo, 0, sizeof (GdamPlayInfo));
  else
    subplayinfo = *play_info;

  /* for now, just add all the submodels to the mix.

   * TODO: it would be good to not add sources that
   * will never make sound b/c we are starting after them,
   * and we are not looping.
   */
  index = 0;
  for (list = mix->elements; list != NULL; list = list->next)
    {
      GdamMixElement *element = list->data;
      int subsource_id;
      subsource_id = gdam_producer_play_request (element->producer,
						 &subplayinfo, 0, builder);
      if (subsource_id == 0)
	return 0;
      time_sources[index * 2 + 0] = element->start_time;
      time_sources[index * 2 + 1] = subsource_id;
      index++;
    }

  nargs = 0;
  args[nargs].name = "mixed";
  args[nargs].type = GDAM_ARG_TYPE_INT64_ARRAY;
  GDAM_ARG_INT64_ARRAY (args + nargs).length = num_elements * 2;
  GDAM_ARG_INT64_ARRAY (args + nargs).array = time_sources;
  nargs++;

  if (mix->is_looping)
    {
      args[nargs].name = "loop_length";
      args[nargs].type = GDAM_ARG_TYPE_INT64;
      GDAM_ARG_INT64 (args + nargs) = (gint64) (mix->loop_length * 44100.0);
      nargs++;
    }

  args[nargs].name = "mode";
  args[nargs].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + nargs) = mix->is_looping ? "loop" : "eternal";
  nargs++;

  gtk_object_ref (GTK_OBJECT (mix));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source_v ("GdasMix",
					  model->source_id,
					  nargs,
					  args,
					  gdam_model_got_subsource,
					  gdam_model_didnt_get_subsource,
					  mix);
  g_free (time_sources);
  return gdam_request_builder_append_create (builder, request);
}

static void
mix_arg_set_looping (GtkObject *object, GdamArg *arg)
{
  GDAM_MIX (object)->is_looping = GDAM_ARG_INT (arg) ? TRUE : FALSE;
}

static void
mix_arg_get_looping (GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_INT (arg) = GDAM_MIX (object)->is_looping;
}

static gdouble
gdam_mix_get_length (GdamModel *model)
{
  GdamMix *mix = GDAM_MIX (model);
  if (mix->loop_length > 0.0)
    return mix->loop_length;
  return -1;
}

static void gdam_mix_class_init(GdamMixClass* mix_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (mix_class);
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (mix_class);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (mix_class);
  producer_class->play_request = gdam_mix_play_request;
  model_class->get_length = gdam_mix_get_length;
  gdam_arg_register (object_class,
		     "looping",
		     GDAM_ARG_TYPE_INT,
		     mix_arg_set_looping,
		     mix_arg_get_looping,
		     NULL, "boolean",
		     GDAM_ARG_VITAL);

  mix_signals[ELEMENT_CHANGED]
    = gtk_signal_new
    ("element-changed",
     GTK_RUN_LAST,
     object_class->type,
     GTK_SIGNAL_OFFSET (GdamMixClass, element_changed),
     gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  mix_signals[ELEMENT_DELETED]
    = gtk_signal_new
    ("element-deleted",
     GTK_RUN_LAST,
     object_class->type,
     GTK_SIGNAL_OFFSET (GdamMixClass, element_deleted),
     gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  mix_signals[ELEMENT_ADDED]
    = gtk_signal_new
    ("element-added",
     GTK_RUN_LAST,
     object_class->type,
     GTK_SIGNAL_OFFSET (GdamMixClass, element_added),
     gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
  mix_signals[ELEMENTS_CHANGED]
    = gtk_signal_new
    ("elements-changed",
     GTK_RUN_LAST,
     object_class->type,
     GTK_SIGNAL_OFFSET (GdamMixClass, elements_changed),
     gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, mix_signals, LAST_SIGNAL);
}

GtkType gdam_mix_get_type()
{
  static GtkType mix_type = 0;
  if (!mix_type)
    {
      static const GtkTypeInfo mix_info = {
	"GdamMix",
	sizeof (GdamMix),
	sizeof (GdamMixClass),
	(GtkClassInitFunc) gdam_mix_class_init,
	(GtkObjectInitFunc) gdam_mix_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_PRODUCER;
      mix_type = gtk_type_unique (parent, &mix_info);
    }
  return mix_type;
}

GdamMixElement* gdam_mix_element_find(GdamMix* mix, GdamProducer* producer)
{
  GSList *elements = mix->elements;
  while (elements != NULL)
    {
      GdamMixElement *element = elements->data;
      if (element->producer == producer)
	return element;
      elements = elements->next;
    }
  return NULL;
}

GdamMixElement*     gdam_mix_add      (GdamMix*         mix,
                                       GdamProducer*    producer,
			               gdouble          start_time)
{
  GdamMixElement *element;
  GdamRequestBuilder *builder;
  int sub_id;
  GdamProducer *sub;
  GdamRequest *request;
  GdamModel *mix_model;
  GdamPlayInfo play_info;

  mix_model = GDAM_MODEL (mix);

  element = gdam_mix_element_find (mix, producer);
  g_return_val_if_fail (element == NULL, NULL);

  element = g_new (GdamMixElement, 1);
  element->producer = producer;
  element->start_time = start_time;
  gtk_object_ref ((GtkObject *) producer);
  gtk_object_sink ((GtkObject *) producer);
  GDAM_MODEL (producer)->parent = mix_model;
  element->length = gdam_model_get_length ((GdamModel *) producer);

  mix->elements = g_slist_prepend (mix->elements, element);

  if (mix_model->source_id == 0)
    {
      return element;
    }

  /* Update server notion... */
  builder = gdam_request_builder_new ();
  sub = element->producer;
  bzero (&play_info, sizeof (play_info));
  sub_id = gdam_producer_play_request (sub, &play_info, 0, builder);
  g_return_val_if_fail (sub_id != 0, NULL);
  gdam_request_builder_append_generic
    (builder,
     gdam_request_add_mix
     (mix_model->source_id,
      sub_id, (gint64) (start_time * 44100.0), NULL, NULL, NULL));
  {
    GdamArg arg;
    arg.name = "child_dont_remove_on_eof";
    arg.type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (&arg) = sub_id;
    gdam_request_builder_append_generic
      (builder,
       gdam_request_set_args (mix_model->source_id,
			    1, &arg, NULL, NULL, NULL));
  }

  request = gdam_request_builder_package (builder);
  gdam_request_builder_destroy (builder);
  gdam_channel_append_request (mix_model->channel, request);

  return element;
}

void        gdam_mix_remove   (GdamMix*         mix,
                               GdamProducer*    producer)
{
  GdamProducer *subproducer;
  GdamModel *submodel;
  GSList **plist;
  GdamMixElement *element = NULL;
  for (plist = &mix->elements; *plist != NULL;)
    {
      GSList *list = *plist;
      GdamMixElement *tmp_element = list->data;
      if (tmp_element->producer == producer)
	{
	  *plist = list->next;
	  g_slist_free_1 (list);
	  element = tmp_element;
	}
      else
	{
	  plist = &((*plist)->next);
	}
    }
  g_return_if_fail (element != NULL);
  mix->elements = g_slist_remove (mix->elements, element);
  subproducer = element->producer;
  submodel = GDAM_MODEL (subproducer);
  if (submodel->source_id != 0)
    {
      gdam_channel_append_request (submodel->channel,
				   gdam_request_detach_source (submodel->
							       source_id,
							       NULL, NULL,
							       NULL));
    }
  gtk_object_unref (GTK_OBJECT (submodel));
}

void        gdam_mix_set_start(GdamMix*         mix,
			       GdamProducer*    producer,
                               gdouble          start_time)
{
  GSList *list = mix->elements;
  while (list != NULL)
    {
      GdamMixElement *element = list->data;
      if (element->producer == producer)
	{
	  GdamModel *model = GDAM_MODEL (producer);
	  GdamModel *mix_model = GDAM_MODEL (mix);
	  if (element->start_time == start_time)
	    return;
	  element->start_time = start_time;
	  gtk_signal_emit ((GtkObject *) mix,
			   mix_signals[ELEMENT_CHANGED], element);
	  if (mix_model->source_id == 0)
	    return;

	  gdam_channel_append_request
	    (GDAM_MODEL (producer)->channel,
	     gdam_request_add_mix
	     (mix_model->source_id,
	      model->source_id,
	      (gint64) (start_time * 44100.0), NULL, NULL, NULL));
	  return;
	}
      list = list->next;
    }
  return;
}
void gdam_mix_set_beat_length(GdamMix*         mix,
			      gdouble          beat_length)
{
  GdamBeatInfo *beat_info = gdam_beat_info_new ();
  gdam_beat_info_set_bpm (beat_info, 60.0 / (beat_length));
}

void
gdam_mix_set_loop_length (GdamMix   *mix,
			  gdouble    loop_length)
{
  GdamModel *model = GDAM_MODEL (mix);
  mix->loop_length = loop_length;
  if (mix->is_looping && model->source_id != 0)
    {
      GdamArg args[2];
      args[0].name = "loop_length";
      args[0].type = GDAM_ARG_TYPE_INT64;
      GDAM_ARG_INT64 (args+0) = (gint64) (44100.0 * loop_length);
      args[1].name = "mode";
      args[1].type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (args+1) = "loop";
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args (model->source_id,
							  2, args,
							  NULL, NULL, NULL));
    }
  gdam_producer_update_length (GDAM_PRODUCER (mix), loop_length);
}

void        gdam_mix_freeze   (GdamMix*         mix)
{
  mix->freeze_count++;
}

void        gdam_mix_thaw     (GdamMix*         mix)
{
  mix->freeze_count--;
}
