/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamrespeed.h"
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include "gdammodel.h"
#include "gdamfilter.h"
#include "gdamutil.h"

static void gdam_respeed_class_init(GdamRespeedClass* klass);
static void              on_child_set_beat_info(GdamFilter*         filter,
						const GdamBeatInfo* child_bi);

static int gdam_respeed_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);
static void gdam_respeed_tempo_changed(GdamRespeed* respeed, float newtempo);

static void gdam_respeed_init(GdamRespeed* respeed)
{
  respeed->rate = 1.0;
  respeed->input_block_size = 1024;
}

GtkType gdam_respeed_get_type()
{
  static GtkType respeed_type = 0;
  if (!respeed_type)
    {
      static const GtkTypeInfo respeed_info = {
	"GdamRespeed",
	sizeof (GdamRespeed),
	sizeof (GdamRespeedClass),
	(GtkClassInitFunc) gdam_respeed_class_init,
	(GtkObjectInitFunc) gdam_respeed_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      respeed_type = gtk_type_unique (GDAM_TYPE_FILTER, &respeed_info);
    }
  return respeed_type;
}

static inline void maybe_dispense_request(GdamRespeed* respeed)
{
  GdamArg arg[2];

  arg[0].name = "input_size";
  arg[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg[0]) = respeed->input_block_size;

  arg[1].name = "rate";
  arg[1].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg[1]) = respeed->rate;

  gdam_model_set_args (GDAM_MODEL (respeed), 0, 2, arg);
}

static int gdam_respeed_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamRespeed *respeed = GDAM_RESPEED (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  float rate;
  rate = respeed->rate;
  gtk_object_ref (GTK_OBJECT (respeed));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasRespeed",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					respeed,
					4,
					"input_size",
					GDAM_ARG_TYPE_INT,
					respeed->input_block_size,
					"rate",
					GDAM_ARG_TYPE_DOUBLE,
					rate,
					"smoothing",
					GDAM_ARG_TYPE_INT,
					1,
					"filtered",
					GDAM_ARG_TYPE_INT, new_source_id);
  return gdam_request_builder_append_create (builder, request);
}

static void gdam_respeed_arg_set_block_size(GtkObject* object, GdamArg* arg)
{
  GdamRespeed *respeed = GDAM_RESPEED (object);
  respeed->input_block_size = GDAM_ARG_INT (arg);
  gdam_debug ("block size set to %d", respeed->input_block_size);
  maybe_dispense_request (respeed);
}

static void gdam_respeed_arg_get_block_size(GtkObject* object, GdamArg* arg)
{
  GdamRespeed *respeed = GDAM_RESPEED (object);
  GDAM_ARG_INT (arg) = respeed->input_block_size;
}

static void gdam_respeed_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  GdamRespeed *respeed = GDAM_RESPEED (object);
  GdamModel *child;
  GdamBeatInfo child_bi;
  float new_bpm;
  float new_first_beat;
  child = GDAM_FILTER (respeed)->child;
  respeed->rate = GDAM_ARG_DOUBLE (arg);
  if (child != NULL)
    {
      GdamBeatInfo *new_beat;
      new_beat = gdam_beat_info_new ();
      child_bi = child->beat_info;
      new_bpm = gdam_beat_info_get_bpm (&child_bi, 0) * respeed->rate;
      new_first_beat = gdam_beat_info_get_first_beat (&child_bi, 0)
	/ respeed->rate;
      gdam_beat_info_set_bpm (new_beat, new_bpm);
      gdam_beat_info_set_first_beat (new_beat, new_first_beat);
      gdam_model_set_beat_info (GDAM_MODEL (respeed), new_beat);
      gdam_respeed_tempo_changed (respeed, new_bpm);
      gdam_beat_info_destroy (new_beat);
    }
  maybe_dispense_request (respeed);
}

static void gdam_respeed_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  GdamRespeed *respeed;
  respeed = GDAM_RESPEED (object);
  GDAM_ARG_DOUBLE (arg) = respeed->rate;
}

enum {
	TEMPO_CHANGED,
	LAST_SIGNAL 
};

static guint respeed_signals[LAST_SIGNAL];

static void gdam_respeed_class_init(GdamRespeedClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  model_class->num_arg_sets = 1;
  filter_class->append_start_request = gdam_respeed_append_start_request;
  filter_class->on_child_set_beat_info = on_child_set_beat_info;
  gdam_arg_register (object_class,
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_respeed_arg_set_rate,
		     gdam_respeed_arg_get_rate,
		     NULL, "0.5,2", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "block_size",
		     GDAM_ARG_TYPE_INT,
		     gdam_respeed_arg_set_block_size,
		     gdam_respeed_arg_get_block_size,
		     NULL, "16,16384", GDAM_ARG_VITAL);
  respeed_signals[TEMPO_CHANGED] =
    gtk_signal_new ("tempo_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamRespeedClass, tempo_changed),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, respeed_signals, LAST_SIGNAL);
}

static void gdam_respeed_tempo_changed(GdamRespeed* respeed, float newtempo)
{
  GtkObject *obj = GTK_OBJECT (respeed);
  if (gtk_signal_n_emissions (obj, respeed_signals[TEMPO_CHANGED]))
    return;
  gtk_signal_emit (obj, respeed_signals[TEMPO_CHANGED], newtempo);
}

void gdam_respeed_set_rate(GdamRespeed* respeed, gdouble rate)
{
  GdamArg arg;
  arg.name = "rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = rate;
  gdam_arg_set_object_arg (GTK_OBJECT (respeed), &arg);
}

static void              on_child_set_beat_info(GdamFilter*         filter,
						const GdamBeatInfo* child_bi)
{
  GdamRespeed *respeed = GDAM_RESPEED (filter);
  GdamBeatInfo *new_beat;
  float new_bpm;
  float new_first_beat;
  new_beat = gdam_beat_info_new ();
  new_bpm = gdam_beat_info_get_bpm (child_bi, 0) * respeed->rate;
  new_first_beat = gdam_beat_info_get_first_beat (child_bi, 0)
    / respeed->rate;
  gdam_beat_info_set_bpm (new_beat, new_bpm);
  gdam_beat_info_set_first_beat (new_beat, new_first_beat);
  gdam_model_set_beat_info (GDAM_MODEL (respeed), new_beat);
  gdam_respeed_tempo_changed (respeed, new_bpm);
  gdam_beat_info_destroy (new_beat);
}
