/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencer.h"
#include "gdambuffermanager.h"
#include <gdam/common/gdamtildeexpand.h>
#include <gdam/component/gdamproducermp3.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtksignal.h>

#define MAX_SPLIT 64

typedef struct _GdamSequencerLoadThread GdamSequencerLoadThread;
struct _GdamSequencerLoadThread {
	GdamSequencer*			sequencer;
	GdamChannel*			channel;
	GList*				elements;
	FILE*				fp;
	int				destroy_connect_id;
	gboolean			is_mp3;

	char*				filename;
	GdamSequencerElement*		under_construction;
	GdamLocalSampleInfo*		local_under_construction;
};

static void continue_processing_thread(GdamSequencerLoadThread* load_thread);
static void on_sequencer_destroy(GdamSequencer* sequencer,
                                 GdamSequencerLoadThread* thread)
{
  thread->destroy_connect_id = 0;
  thread->sequencer = NULL;
}

static void free_load_thread(GdamSequencerLoadThread* load_thread)
{
  if (load_thread->fp)
    fclose (load_thread->fp);
  g_list_foreach (load_thread->elements, (GFunc) g_free, NULL);
  g_list_free (load_thread->elements);

  if (load_thread->sequencer && load_thread->destroy_connect_id)
    {
      gtk_signal_disconnect (GTK_OBJECT (load_thread->sequencer),
			     load_thread->destroy_connect_id);
      load_thread->destroy_connect_id = 0;
      load_thread->sequencer = NULL;
    }
  g_free (load_thread);
}

/* Whichever type of element we were building, complete it. */
static void finished_constructing_element(GdamSequencerLoadThread* thread)
{
  if (thread->under_construction)
    {
      thread->elements = g_list_prepend (thread->elements,
					 thread->under_construction);
      thread->under_construction = NULL;
    }
  if (thread->local_under_construction)
    {
      thread->local_under_construction = NULL;
    }

  g_free (thread->filename);
  thread->filename = NULL;
}


static gboolean match_filename_extension (const char* filename,
    					  const char* ext)
{
  gboolean rv;
  const char *extension;
  if (filename == NULL)
      return FALSE;
  extension = strrchr (filename, '.');
  if (extension != NULL && strcasecmp (extension, ext) == 0)
    {
      if (!g_dirname (filename))
	  return FALSE;
      if (g_basename (filename))
	  return TRUE;
    }
  return FALSE;
}

static gboolean is_mp3_filename(const char* filename)
{
  return match_filename_extension (filename, ".mp3");
}

static gboolean is_wav_filename(const char* filename)
{
  return match_filename_extension (filename, ".wav");
}

static void buffer_manager_cached (int buffer_id, gpointer udata)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) udata;

  if (load_thread->under_construction)
    {
      load_thread->under_construction->buffer_id = buffer_id;
    }
  else if (load_thread->local_under_construction)
    {
      load_thread->local_under_construction->buffer_id = buffer_id;
    }
  if (load_thread->sequencer)
    {
      finished_constructing_element (load_thread);
      continue_processing_thread (load_thread);
    }
  else
    {
      g_warning ("no sequencer??");
      free_load_thread (load_thread);
    }
}

static void buffer_manager_failed (gpointer udata)
{
  GdamSequencerLoadThread *load_thread;
  load_thread = (GdamSequencerLoadThread *) udata;

  g_warning ("buffer manager failed to load %s", load_thread->filename);

  /* Continue processing??? */
  /*continue_processing_thread(load_thread); */
  /* Or not??? */
  free_load_thread (load_thread);
}

static void append_cache_request(GdamSequencerLoadThread* load_thread,
                                 char*                    filename,
				 gdouble		  start,
				 gdouble		  len)
{
  GdamRequestBuilder *builder = gdam_request_builder_new ();
  int local_id;
  //const char* source_type;
  char* description;
  GdamChannel *channel = load_thread->channel;
  GdamArg args[4];
  int num_args = 0;

  g_assert (load_thread->filename == NULL);
  gdam_debug ("append_cache_request: loading %s", filename);

  g_return_if_fail (filename != NULL);
  load_thread->filename = filename;

  if (len > 0)
    {
      /* TODO: start w/o len?? */
      description = g_strdup_printf (" %3.5f %3.5f %s", start, len, filename);
    }
  /* TODO: fix, or inherit, description */
  else if (is_mp3_filename (filename))
    {
      description = g_strdup_printf ("Mp3 %s", filename);
    }
  else
    {
      description = g_strdup_printf ("Wav %s", filename);
    }

  args[num_args].type = GDAM_ARG_TYPE_STRING;
  args[num_args].name = "filename";
  GDAM_ARG_STRING (args + num_args) = filename;
  num_args++;

  args[num_args].type = GDAM_ARG_TYPE_INT64;
  args[num_args].name = "position";
  GDAM_ARG_INT64 (args + num_args) = (gint64) (start * 44100.0);
  num_args++;

  if (len> 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT64;
      args[num_args].name = "length";
      GDAM_ARG_INT64 (args + num_args) = (gint64) (len * 44100.0);
      num_args++;
    }

  args[num_args].name = "exec_command";
  args[num_args].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + num_args)
    = (char *) gdam_mp3_exec_cmd_from_filename (filename);
  num_args++;

  local_id = gdam_request_builder_append_create 
    (builder,
     gdam_request_create_source_v ("GdasSeekableMp3", 
                                   gdam_channel_new_local_id (channel),
				   num_args, args, NULL, NULL, NULL));

  gdam_buffer_manager_start_caching (builder, local_id, channel, TRUE,
      				     description,
                                     buffer_manager_cached,
				     buffer_manager_failed,
				     load_thread);
}


/* If we set *filename_out, that file will be cached for us. */
static GdamLocalSampleInfo* 
parse_local_sample (GdamSequencerLoadThread *load_thread,
    		    char **split, int num_split, char **filename_out,
		    gdouble *start_out, gdouble *len_out,
		    int line_version, char *buf_start, char *buf_copy)
{
  GdamLocalSampleInfo *linfo;
  GdamBufferManagerInfo *info;
  char *filename = NULL;
  char *my_filename = NULL;
  char *description = NULL;
  gdouble start = -1000.0;
  gdouble len = -1000.0;
  int li_idx;

  g_return_val_if_fail (num_split >= 7, NULL);

  li_idx = gdam_sequencer_add_local_info (load_thread->sequencer, 0, 0, 0);
  linfo = &g_array_index (load_thread->sequencer->local_infos, 
      		          GdamLocalSampleInfo, li_idx);
  if (line_version == 1)
    {
      filename = split[1];
      linfo->local_id = atoi (split[2]);
	/* temporary, buf_id will be set based on local_id */
      linfo->buffer_id = atoi (split[2]);  
      linfo->rate = atof (split[3]);
      linfo->gain = atof (split[4]);
      linfo->sample_offset = atoi (split[5]);
      linfo->output_length = atoi (split[6]);

      if (num_split >= 8)
	linfo->trigger_rate = atof (split[7]);
    }
  else
    {
      g_return_val_if_fail (num_split >= 10, NULL);
      linfo->local_id = atoi (split[1]);
      linfo->buffer_id = atoi (split[1]);
      linfo->rate = atof (split[2]);
      linfo->gain = atof (split[3]);
      linfo->sample_offset = atoi (split[4]);
      linfo->output_length = atoi (split[5]);
      linfo->trigger_rate = atof (split[6]);
      linfo->trigger_gain = atof (split[7]);
      g_return_val_if_fail (strcmp (split[8], "FILENAME") == 0 ||
                            strcmp (split[8], "FILENAME_RANGE") == 0, NULL);
      /* Extract original unbroken filename */
      my_filename = g_strdup (buf_copy + (split[9] - buf_start));
      if (strcmp (split[8], "FILENAME") == 0)
	filename = my_filename;
      else if (strcmp (split[8], "FILENAME_RANGE") == 0)
	{
	  char **fragments = g_strsplit (my_filename, ":", 3);
	  g_return_val_if_fail (fragments != NULL && fragments[0] != NULL && 
	                          fragments[1] != NULL && fragments[2] != NULL,
				NULL);
	  start = atof (fragments[0]);
	  len = atof (fragments[1]);
	  filename = g_strdup (fragments[2]);
	  description = g_strdup_printf (" %3.5f  %3.5f  %s",
			 start, len, filename);
	  *start_out = start;
	  *len_out = len;
	  g_strfreev (fragments);
        }
    }

  if (description) {}
  else if (is_mp3_filename (filename))
    description = g_strdup_printf ("Mp3 %s", filename);
  else if (filename)
    description = g_strdup_printf ("Wav %s", filename);
  info = gdam_buffer_manager_info_by_desc (description);
  if (info != NULL)
    linfo->buffer_id = info->buffer_id;
  else
    {
      linfo->buffer_id = 0;
      *filename_out = g_strdup (filename);
    }
  g_free (description);
  if (my_filename)
    g_free (my_filename);
  return linfo;
}


/* Parses a line of input until an element or local sample info is encountered.
 */
static gboolean
parse_element (GdamSequencerLoadThread* load_thread, char** filename_out,
    	       gdouble *start_out, gdouble *len_out)
{
  GdamSequencerElement *rv;
  GdamBufferManagerInfo *info;
  char buf[1024];
  FILE* fp;
  fp = load_thread->fp; 
  while (fgets (buf, sizeof (buf), fp) != NULL)
    {
      char *at = buf;
      char *split[128];
      char *description = "";
      char *filename;
      char *original_line = g_strndup (buf, sizeof (buf));
      int num_split = 0;
      while (*at && isspace (*at))
	at++;
      /* Skip comment lines. */
      if (*at == 0 || *at == '#')
	continue;

      g_strchomp (at);
      g_strchomp (original_line);

      /* Split the line into whitespace delimited pieces. */
      /* TODO: handle filenames with spaces, by ignoring spaces within quotes */
      while (*at && num_split < MAX_SPLIT)
	{
	  char *end_word = at;
	  while (*end_word && !isspace (*end_word))
	    end_word++;
	  split[num_split++] = at;
	  if (*end_word == '\0')
	    break;
	  *end_word = 0;
	  at = end_word + 1;
	  while (*at && isspace (*at))
	    at++;
	}
      if (num_split == 2 && strcmp(split[0], "bpm:") == 0)
        {
	  GdamBeatInfo* new_beat_info;
	  float bpm = atof (split[1]);
	  new_beat_info = gdam_beat_info_copy 
			  (&GDAM_MODEL(load_thread->sequencer)->beat_info);
	  gdam_beat_info_set_bpm (new_beat_info, bpm);
	  gdam_model_set_beat_info (GDAM_MODEL(load_thread->sequencer), 
				    new_beat_info);
	  gdam_beat_info_destroy(new_beat_info);
	  continue;
        }
      else if (num_split == 2 && strcmp(split[0], "beats:") == 0)
        {
	  float beats = atoi (split[1]);
	  gdam_sequencer_set_num_beats(load_thread->sequencer, beats);
	  continue;
        }
      /* This is for backward compatability only. */
      else if (strcmp(split[0], "GdamLocalSampleInfo") == 0)
        {
	  GdamLocalSampleInfo *linfo;
	  linfo = parse_local_sample (load_thread, split, num_split, 
	      			      filename_out, start_out, len_out, 1,
				      buf, original_line);
	  if (linfo)
	    {
	      load_thread->local_under_construction = linfo;
	      return TRUE;
	    }
	  g_warning ("error parsing local sample.");
	  continue;
        }
      else if (strcmp(split[0], "GdamLocalSampleInfo2") == 0)
        {
	  GdamLocalSampleInfo *linfo;
	  linfo = parse_local_sample (load_thread, split, num_split, 
	      			      filename_out, start_out, len_out, 2,
				      buf, original_line);
	  if (linfo)
	    {
	      load_thread->local_under_construction = linfo;
	      return TRUE;
	    }
	  g_warning ("error parsing local sample version 2: '%s'", buf);
	  continue;
        }
      else if (num_split != 6 && num_split != 7 && num_split != 8)
	{
	  g_warning ("bad number of fields (%d) in line %s parsing sequence",
	      	     num_split, buf);
	  continue;
	}

      g_return_val_if_fail (num_split >= 6, FALSE);
      g_return_val_if_fail (split[5] != NULL, FALSE);

      rv =  gdam_sequencer_new_element ();
      filename = split[0];
      rv->beat_offset = atof (split[1]);
      rv->rate = atof (split[2]);
      rv->gain = atof (split[3]);
      rv->sample_offset = atoi (split[4]);
      rv->output_length = atoi (split[5]);
      if (num_split >= 7) 
	rv->pattern_number = atoi (split[6]);
      if (num_split == 8) 
	rv->local_id = atoi (split[7]);

      /* If this sequence element is based on a local id, we needn't worry
       * about the buffer_id.
       */
      if (strcmp (filename, "none") == 0 || strcmp (filename, "local") == 0)
	{
	  if (num_split < 8 || rv->local_id < 0)
	    {
	      g_warning ("missing filename and local id");
	      g_free (rv);
	      continue;
	    }
	  /* TODO: breaks tracking of samples in sequencer. */
	  rv->buffer_id = 0;
	}
      /* If we need a buffer_id, decide whether we need to cache any file. */
      else
	{
	  if (is_mp3_filename (filename))
	    description = g_strdup_printf ("Mp3 %s", filename);
	  else
	    description = g_strdup_printf ("Wav %s", filename);
	  info = gdam_buffer_manager_info_by_desc (description);
	  if (info != NULL)
	    rv->buffer_id = info->buffer_id;
	  else
	    {
	      rv->buffer_id = 0;
	      *filename_out = g_strdup (filename);
	    }
	  g_free (description);
	}

      load_thread->under_construction = rv;
      return TRUE;
    }
  return FALSE;
}

/* Do what needs to be done in the buffer-loading thread.  This is called
 * whenever we are in a position to cache a new buffer... ie we have
 * finished caching a buffer or we are beginning to cache the first. 
 */
static void continue_processing_thread(GdamSequencerLoadThread* load_thread)
{
  char *filename;
  gdouble start = 0;
  gdouble len = -1;
  filename = NULL;

  /* We shouldn't be waiting on any sort of sample. */
  g_assert (load_thread->under_construction == NULL);
  g_assert (load_thread->local_under_construction == NULL);

  gdam_debug ("processing thread...");

  /* Check if the sequencer has been destroyed. */
  if (load_thread->sequencer == NULL)
    {
      free_load_thread (load_thread);
      return;
    }

  /* While we can read an element or local info from the file we're loading */
  while (parse_element (load_thread, &filename, &start, &len))
    {
      /* If we have a filename to cache, start it caching */
      if (filename)
	{
	  append_cache_request (load_thread, filename, start, len);
	  return;
	}
      finished_constructing_element (load_thread);
    }
  if (load_thread->elements)
    {
      gdam_sequencer_set_elements (load_thread->sequencer,
				   load_thread->elements);
      load_thread->elements = NULL;
    }
  free_load_thread (load_thread);
}

void                    gdam_sequencer_load       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamSequencerLoadThread *loader;
  FILE *fp;
  char *expanded_filename;
  /* Open the file */
  expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "r");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_load: error loading %s", filename);
      return;
    }
  gdam_debug ("loading file %s", expanded_filename);
  /* Create a thread to load buffers. */
  loader = g_new0 (GdamSequencerLoadThread, 1);
  loader->fp = fp;
  loader->sequencer = sequencer;
  loader->channel = GDAM_MODEL (sequencer)->channel;
  loader->destroy_connect_id
    = gtk_signal_connect (GTK_OBJECT (sequencer),
			  "destroy",
			  GTK_SIGNAL_FUNC (on_sequencer_destroy), loader);
  loader->elements = NULL;
  loader->under_construction = NULL;
  loader->local_under_construction = NULL;
  loader->filename = NULL;
  /* Start the buffer-loading thread */
  continue_processing_thread (loader);
}

static void write_local_sample (GdamLocalSampleInfo *linfo, FILE* fp)
{
  GdamBufferManagerInfo *info;
  const char *filename = "local";
  char *filename_tag = "FILENAME";
  info = gdam_buffer_manager_info_by_id (linfo->buffer_id);
  g_return_if_fail (info != NULL);

  if (strncmp (info->description, "Wav ", 4)    == 0
      || strncmp (info->description, "Mp3 ", 4) == 0)
    filename = info->description + 4;
  else
    {
      char **tokens = g_strsplit (info->description, " ", 0);
      gdouble start = -1000.0;
      gdouble length = -1000.0;
      char *my_filename = NULL;
      int i = 0;

      while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	i++;
      if (tokens[i])
	{
	  start = atof (tokens[i]);
	  i++;
	}

      while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	i++;
      if (tokens[i])
	{
	  length = atof (tokens[i]);
	  i++;
	}

      /* Collect all remaining tokens into the filename */
      while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	i++;
      if (tokens[i])
	{
	  my_filename = tokens[i];
	  i++;
	  while (tokens[i])
	    {
	      char *old_my_f = my_filename;
	      my_filename = g_strconcat (my_filename, " ", tokens[i], NULL);
	      g_free (old_my_f);
	      i++;
	    }
	}

      if (length > 0 && my_filename != NULL)
	{
	  filename_tag = "FILENAME_RANGE";
	  filename = g_strdup_printf ("%0.4g:%0.4g:%s", start, length, 
				      my_filename);
	}
      else
	{
	  g_warning ("couldn't save element with description %s", 
		     info->description);
	  return;
	}
    }
  fprintf (fp, "GdamLocalSampleInfo2 %d %4.4g %4.4g %d %d %4.4g %4.4g %s %s\n",
	   linfo->local_id,
	   linfo->rate,
	   linfo->gain,
	   linfo->sample_offset,
	   linfo->output_length,
	   linfo->trigger_rate,
	   linfo->trigger_gain,
	   filename_tag,
	   filename);
}

static void write_element(GdamSequencerElement* element, FILE* fp)
{
  GdamBufferManagerInfo *info;
  const char *filename = "none";
  if (element->local_id < 0)
    {
      info = gdam_buffer_manager_info_by_id (element->buffer_id);
      g_return_if_fail (info != NULL);
      if (strncmp (info->description, "Wav ", 4)    == 0
	  || strncmp (info->description, "Mp3 ", 4) == 0)
	filename = info->description + 4;
      else
	{
	  char **tokens = g_strsplit (info->description, " ", 0);
	  gdouble start = -1000.0;
	  gdouble length = -1000.0;
	  char *my_filename = NULL;
	  int i = 0;

	  while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	    i++;
	  if (tokens[i])
	    {
	      start = atof (tokens[i]);
	      i++;
	    }

	  while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	    i++;
	  if (tokens[i])
	    {
	      length = atof (tokens[i]);
	      i++;
	    }

	  while (tokens[i] != NULL && strcmp (tokens[i], "") == 0)
	    i++;
	  if (tokens[i])
	    my_filename = tokens[i];

	  if (length > 0 && my_filename != NULL)
	    filename = g_strdup_printf ("%4.4g:%4.4g:%s", start, length, 
		                        my_filename);
	  else
	    {
	      g_warning ("couldn't save element with description %s", 
	                 info->description);
	      return;
	    }
	}
    }
  fprintf (fp, "%s %4.4g %4.4g %4.4g %d %d %d %d\n",
	   filename,
	   element->beat_offset,
	   element->rate,
	   element->gain,
	   element->sample_offset,
	   element->output_length, 
	   element->pattern_number,
	   element->local_id);
}
	

void                    gdam_sequencer_save       (GdamSequencer* sequencer,
                                                   const char*    filename)
{
  GdamBeatInfo* bi;
  guint i;
  FILE *fp;
  char *expanded_filename = gdam_tilde_expand (filename);
  fp = fopen (expanded_filename, "w");
  g_free (expanded_filename);
  if (!fp)
    {
      g_warning ("gdam_sequencer_save: error creating %s", filename);
      return;
    }
  bi = &(GDAM_MODEL (sequencer)->beat_info);
  fprintf (fp, "beats: %4.4g\n", sequencer->num_beats);
  fprintf (fp, "bpm: %4.4g\n", gdam_beat_info_get_bpm(bi, 0));
  g_list_foreach (sequencer->elements, (GFunc) write_element, fp);
  if (sequencer->local_infos)
    {
      for (i = 0; i < sequencer->local_infos->len; i++)
	{
	  GdamLocalSampleInfo *linfo;
	  linfo = &g_array_index (sequencer->local_infos, 
	      			  GdamLocalSampleInfo,
				  i);
	  write_local_sample (linfo, fp);
	}
    }
  fclose (fp);
}
