/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamtoplevel.h"
#include <gtk/gtksignal.h>
#include "gdamrequestbuilder.h"

enum {
	DEVICE_LIST_CHANGED,
	LAST_SIGNAL
};

typedef struct _PendingRequest PendingRequest;
typedef struct _GdamModelInfo GdamModelInfo;

static guint toplevel_signals[LAST_SIGNAL] = {0};

static void gdam_toplevel_new_request(GdamModel* Model,
				      GdamModel* child,
				      GdamPlayInfo*  play_info,
				      int            local_src_id,
				      GdamRequestBuilder* builder);

struct _GdamModelInfo {
	GdamToplevel*		toplevel;
	guint			destroy_connect_id;
	GdamModel*		model;
	gboolean		misses_data;
	GdamToplevelDevice*	device;
};

struct _PendingRequest {
	int			local_src_id;
	GdamRequestBuilder*	builder;
};

static void gdam_toplevel_class_init(GdamModelClass* toplevel_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (toplevel_class);
  toplevel_class->new_request = gdam_toplevel_new_request;
  ((GdamToplevelClass *) toplevel_class)->device_list_changed = NULL;
  toplevel_signals[DEVICE_LIST_CHANGED]
    = gtk_signal_new ("device-list-changed",
		      GTK_RUN_FIRST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamToplevelClass,
					 device_list_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, toplevel_signals, LAST_SIGNAL);
}

static void device_ref(GdamToplevelDevice* device)
{
  g_return_if_fail ((device != NULL));
  ++(device->ref_count);
}
static void device_unref(GdamToplevelDevice* device)
{
  if (--(device->ref_count) == 0)
    g_error ("cannot free device");	/* TODO: fix this */
}

static void gdam_toplevel_init(GdamToplevel* toplevel)
{
  toplevel->model_to_info = g_hash_table_new (NULL, NULL);
  toplevel->devices_by_name = g_hash_table_new (g_str_hash, g_str_equal);
}

GtkType gdam_toplevel_get_type()
{
  static GtkType toplevel_type = 0;
  if (!toplevel_type)
    {
      static const GtkTypeInfo toplevel_info = {
	"GdamToplevel",
	sizeof (GdamToplevel),
	sizeof (GdamToplevelClass),
	(GtkClassInitFunc) gdam_toplevel_class_init,
	(GtkObjectInitFunc) gdam_toplevel_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      toplevel_type = gtk_type_unique (GDAM_TYPE_MODEL, &toplevel_info);
    }
  return toplevel_type;
}

GdamToplevel*  gdam_toplevel_new(GdamChannel* channel)
{
  GdamToplevel *toplevel;
  g_return_val_if_fail (channel != 0, NULL);

  toplevel = GDAM_TOPLEVEL (gtk_type_new (GDAM_TYPE_TOPLEVEL));
  GDAM_MODEL (toplevel)->channel = channel;

  return toplevel;
}


GdamToplevel*  gdam_toplevel_get_global(GdamChannel* channel)
{
  static GdamToplevel *global = NULL;
  if (global == NULL)
    global = gdam_toplevel_new (channel);
  return global;
}

/*=====================================================================*
 *                       gdam_toplevel_add_device                      *
 *                                                                     *
 * Create a record for an already opened device (the device must       *
 * also already have a mixer attached.)                                *
 *                                                                     *
 * Parameters:                                                         *
 *     toplevel          the model that maintains the device table.    *
 *     device_name       the name of the device which has been opened. *
 *     device_id         the server-side identifying number of the     *
 *                       device.                                       *
 *     mixer_id          the source-id of the mixer which is feeding   *
 *                       the device.                                   *
 *=====================================================================*/
static void make_add_args(GdamArg* args, int local_src_id)
{
  args[0].name = "add_with_latency";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&args[0]) = local_src_id;
  args[1].name = "child_remove_on_eof";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&args[1]) = local_src_id;
}
static void append_feed_request(PendingRequest*     pending_request,
                                GdamToplevelDevice* device,
				gboolean            was_final)
{
  GdamRequestBuilder *builder = pending_request->builder;
  GdamRequest *package;
  GdamArg args[2];
  make_add_args (args, pending_request->local_src_id);
  gdam_request_builder_append_generic (builder,
				       gdam_request_set_args (device->
							      mixer_id, 2,
							      args, NULL,
							      NULL, NULL));
  if (!was_final)
    {
      gdam_debug ("toplevel append_feed_req DETACHING source id %d",
	  	  pending_request->local_src_id);
      gdam_request_builder_append_generic (builder,
					   gdam_request_detach_source
					   (pending_request->local_src_id, NULL,
					    NULL, NULL));
    }
  package = gdam_request_builder_package (builder);
  gdam_request_builder_destroy (builder);
  gdam_channel_append_request (GDAM_MODEL (device->toplevel)->channel,
			       package);
}

	
void          gdam_toplevel_add_device  (GdamToplevel*   toplevel,
                                         const char*     device_name,
					 guint           device_id,
					 guint           mixer_id)
{
  GdamToplevelDevice *device;
  device = g_hash_table_lookup (toplevel->devices_by_name, device_name);
  if (device)
    {
      if (device->mixer_id == 0 && device->device_id == 0)
	{
	  GSList *waiting = device->waiting_requests;
	  GSList *at;
	  device->waiting_requests = NULL;
	  device->device_id = device_id;
	  device->mixer_id = mixer_id;
	  waiting = g_slist_reverse (waiting);
	  for (at = waiting; at != NULL; at = at->next)
	    append_feed_request (
				 (PendingRequest *) (at->data),
				 device, at->next == NULL ? TRUE : FALSE);
	  toplevel->devices = g_slist_append (toplevel->devices, device);
	  gtk_signal_emit (GTK_OBJECT (toplevel),
			   toplevel_signals[DEVICE_LIST_CHANGED]);
	  return;
	}

      g_warning ("add_device: device '%s' already exists", device_name);
      return;
    }

  device = g_new (GdamToplevelDevice, 1);
  device->ref_count = 1;
  device->device_name = g_strdup (device_name);
  device->device_id = device_id;
  device->mixer_id = mixer_id;
  device->toplevel = toplevel;
  device->waiting_requests = NULL;
  g_hash_table_insert (toplevel->devices_by_name,
		       device->device_name, device);
  if (device_id != 0)
    {
      g_assert (mixer_id != 0);
      toplevel->devices = g_slist_append (toplevel->devices, device);
      gtk_signal_emit (GTK_OBJECT (toplevel),
		       toplevel_signals[DEVICE_LIST_CHANGED]);
    }
  else
    {
      g_assert (mixer_id == 0);
    }
  if (toplevel->default_device == NULL)
    toplevel->default_device = device;
}

/*=====================================================================*
 *                      gdam_toplevel_open_device                      *
 *                                                                     *
 * Open a device, and call a function when that succeeds or fails.     *
 *                                                                     *
 * The handlers may be called immediately,                             *
 * (ie before this function returns) or they will be called            *
 * when we get word back from the server.                              *
 *                                                                     *
 * Parameters:                                                         *
 *     toplevel          the model that maintains the device table.    *
 *     device_name       the name of the device which has been opened. *
 *     succeeded         called if the device was opened successfully. *
 *     failed            called if the device open failed.             *
 *     user_data         last parameter to the succeeded and failed    *
 *                       functions.                                    *
 *=====================================================================*/
typedef struct _DeviceOpening DeviceOpening;
struct _DeviceOpening {
	GdamToplevel*		toplevel;
	char*			device_name;
	GdamRequestGetIntStruct	device_id;
	GdamRequestGetIntStruct	mixer_id;
	int			recorder_id;

	GdamToplevelOpened      succeeded;
	GdamToplevelOpenFailed  failed;
	gpointer                user_data;
};
static void device_opening_destroy(DeviceOpening* opening)
{
  g_free (opening->device_name);
  g_free (opening);
}
static void feed_package_succeeded(int num, gpointer user_data)
{
  DeviceOpening *opening = user_data;
  gdam_toplevel_add_device (opening->toplevel,
			    opening->device_name,
			    opening->device_id.rv, opening->mixer_id.rv);
  device_opening_destroy (opening);
}
static void feed_package_failed(gpointer user_data)
{
  DeviceOpening *opening = user_data;
  g_warning ("feeding device '%s' from mixer failed", opening->device_name);
  device_opening_destroy (opening);
}
static void created_recorder(int id, gpointer user_data)
{
  DeviceOpening *opening = user_data;
  gdam_toplevel_set_recorder (opening->toplevel, opening->device_name, id);
  opening->recorder_id = id;
}
	
static void failed_to_create_recorder(gpointer user_data)
{
  g_message ("failed to create recorder");
}

const char* gdam_toplevel_server_mix_class_name = "GdasMix";

static void device_open_succeed(int device_id, gpointer user_data)
{
  DeviceOpening *opening = user_data;
  GdamRequest **make_mixer;
  GdamChannel *channel = GDAM_MODEL (opening->toplevel)->channel;
  int mix_id;
  int recorder_id;
  gdam_debug ("device_open_succeeded: %d", device_id);
  make_mixer = g_new (GdamRequest *, 3);
  opening->device_id.rv = device_id;
  mix_id = gdam_channel_new_local_id (channel);
  recorder_id = gdam_channel_new_local_id (channel);
  make_mixer[0] =
    gdam_request_create_source (gdam_toplevel_server_mix_class_name, mix_id,
				gdam_request_get_int,
				gdam_request_get_failure, &opening->mixer_id,
				1, "mode", GDAM_ARG_TYPE_STRING, "eternal");
  make_mixer[1] =
    gdam_request_create_source ("GdasRecorderExternal", recorder_id,
				created_recorder, failed_to_create_recorder,
				opening, 1, "filtered", GDAM_ARG_TYPE_INT,
				mix_id);
  make_mixer[2] =
    gdam_request_feed_sink (device_id, recorder_id, NULL, NULL, "eternal");
  gdam_channel_append_request (channel,
			       gdam_request_package_v (3, make_mixer,
						       feed_package_succeeded,
						       feed_package_failed,
						       opening));
}

static void device_open_failed(gpointer user_data)
{
  DeviceOpening *opening = (DeviceOpening *) user_data;
  g_warning ("opening device '%s' failed", opening->device_name);
  device_opening_destroy (opening);
}

static void got_device_recorder(int num_args, GdamArg* args, gpointer opening)
{
  DeviceOpening *dev_opening = opening;
  if (num_args < 1 || GDAM_ARG_INT (args) == 0)
    {
      g_warning ("error getting recording device");
      return;
    }
  gdam_toplevel_set_recorder (dev_opening->toplevel,
			      dev_opening->device_name, GDAM_ARG_INT (args));
  device_opening_destroy (dev_opening);
}
static void didnt_get_device_recorder(gpointer opening)
{
  DeviceOpening *dev_opening = opening;
  g_warning ("couldn't get recording device");
  device_opening_destroy (dev_opening);
}

static void get_server_side_ints_succeeded(int num, gpointer user_data)
{
  DeviceOpening *opening = (DeviceOpening *) user_data;
  GdamChannel *channel = GDAM_MODEL (opening->toplevel)->channel;

  gdam_debug ("get server-side ints: %d (opening=%p, name=%s)", num, opening,
	      opening->device_name);

  if (opening->device_id.was_success && opening->mixer_id.was_success)
    {
      GdamRequest *get_parent;
      const char *names = "parent";
      gdam_debug ("get server-side ints: success");
      gdam_toplevel_add_device (opening->toplevel,
				opening->device_name,
				opening->device_id.rv, opening->mixer_id.rv);
      if (opening->succeeded)
	(*opening->succeeded) (opening->device_id.rv,
			       opening->mixer_id.rv, opening->user_data);
      get_parent = gdam_request_get_args (opening->mixer_id.rv,
					  1,
					  &names,
					  got_device_recorder,
					  didnt_get_device_recorder, opening);
      gdam_channel_append_request (channel, get_parent);
      return;
    }
  if (opening->device_id.was_success)
    {
      g_warning ("device '%s' was already open - but no mixer",
		 opening->device_name);
      device_opening_destroy (opening);
      return;
    }
  if (opening->mixer_id.was_success)
    {
      g_warning ("device '%s' already had open mixer - but no device",
		 opening->device_name);
      device_opening_destroy (opening);
      return;
    }
  gdam_debug ("get server-side ints: failure");
  gdam_debug ("about to request create sink for device %s",
	      opening->device_name);
  gdam_channel_append_request (channel,
			       gdam_request_create_sink (opening->device_name,
							 device_open_succeed,
							 device_open_failed,
							 opening));
}

static void get_server_side_ints_failed(gpointer user_data)
{
  DeviceOpening *opening = (DeviceOpening *) user_data;
  g_error ("error with package request to open device %s",
	   opening->device_name);
  device_opening_destroy (opening);
}


gboolean          gdam_toplevel_is_device_open 
					(GdamToplevel*          toplevel,
                                         const char*            device_name)
{
  GdamToplevelDevice *toplevel_device;

  toplevel_device = g_hash_table_lookup (toplevel->devices_by_name,
					 device_name);
  return toplevel_device ? TRUE : FALSE;
}


void          gdam_toplevel_open_device (GdamToplevel*          toplevel,
                                         const char*            device_name,
					 GdamToplevelOpened     succeeded,
					 GdamToplevelOpenFailed failed,
					 gpointer               user_data)
{
  GdamRequest **get_requests;
  GdamRequest *package;
  DeviceOpening *device_opening;
  GdamToplevelDevice *toplevel_device;

  get_requests = g_new (GdamRequest *, 2);

  toplevel_device = g_hash_table_lookup (toplevel->devices_by_name,
					 device_name);
  if (toplevel_device != NULL)
    {
      guint mixer_id = toplevel_device->mixer_id;
      guint device_id = toplevel_device->device_id;
      if (mixer_id == 0 || device_id == 0)
	{
	  g_warning ("tried to open device '%s' while it was already "
		     "being opened", device_name);
	  if (failed)
	    (*failed) (user_data);
	  return;
	}
      if (succeeded)
	(*succeeded) (device_id, mixer_id, user_data);
      return;
    }

  device_opening = g_new (DeviceOpening, 1);

  device_opening->toplevel = toplevel;
  device_opening->device_name = g_strdup (device_name);
  device_opening->succeeded = succeeded;
  device_opening->failed = failed;
  device_opening->user_data = user_data;

  /* TODO: use get_set_server_side. */
  get_requests[0] = gdam_request_get_server_side_int
    ("devices",
     device_name,
     gdam_request_get_int,
     gdam_request_get_failure, &device_opening->device_id);
  get_requests[1] = gdam_request_get_server_side_int
    ("device-feeds",
     device_name,
     gdam_request_get_int,
     gdam_request_get_failure, &device_opening->mixer_id);
  package = gdam_request_package_v (2,
				    get_requests,
				    get_server_side_ints_succeeded,
				    get_server_side_ints_failed,
				    device_opening);
  gdam_channel_append_request (GDAM_MODEL (toplevel)->channel, package);

  gdam_toplevel_add_device (device_opening->toplevel,
			    device_opening->device_name, 0, 0);
}

void          gdam_toplevel_set_recorder(GdamToplevel*   toplevel,
                                         const char*     device_name,
					 int             recorder_id)
{
  GdamToplevelDevice *device;
  device = g_hash_table_lookup (toplevel->devices_by_name, device_name);
  g_return_if_fail (device != NULL);
  device->recorder_id = recorder_id;
}

/*
 * handler for when a model is destroyed.
 */
static void on_destroy(GdamModel* model, GdamModelInfo* info)
{
  g_hash_table_remove (info->toplevel->model_to_info, model);
  gtk_object_unref (GTK_OBJECT (model));
}

/*=====================================================================*
 *                      gdam_toplevel_new_request                      *
 *                                                                     *
 * GdamModel method to deal with a child offering a new subsource.     *
 *                                                                     *
 * Parameters:                                                         *
 *     model         the toplevel.                                     *
 *     child             the model which is going to assume            *
 *                       local_src_id.                                 *
 *     local_src_id      the negative number assigned by the builder   *
 *                       for the uninstantiated source.                *
 *     builder           a structure maintaining the requests.         *
 *=====================================================================*/
static void gdam_toplevel_new_request(GdamModel* model,
				      GdamModel* child,
				      GdamPlayInfo*  play_info,
				      int            local_src_id,
				      GdamRequestBuilder* builder)
{
  GdamToplevel *toplevel = GDAM_TOPLEVEL (model);
  GdamModelInfo *info;
  GdamToplevelDevice *device;
  gboolean was_synchronized = FALSE;
  info = g_hash_table_lookup (toplevel->model_to_info, child);
  if (info == NULL)
    {
      g_warning ("gdam_toplevel_new_request: couldnt find model info for "
	  	 "supposed child of type %s.... ABORTING  (lid=%d, cid=%d)",
		 gtk_type_name (GTK_OBJECT (child)->klass->type),
		 local_src_id, child->source_id);
      return;
    }
  device = info->device;
  g_return_if_fail (device != NULL);

  gdam_debug ("gdam_toplevel_new_request: child's localid=%d type=%s",
	      local_src_id, gtk_type_name (GTK_OBJECT (child)->klass->type));

  if (info->device->device_id == 0)
    {
      PendingRequest *pending_request;
      pending_request = g_new (PendingRequest, 1);
      pending_request->builder = builder;
      pending_request->local_src_id = local_src_id;
      device->waiting_requests =
	g_slist_prepend (device->waiting_requests, pending_request);
      return;
    }

  /*  If the child model who is requesting a new subsource already has a 
   *  source playing, detach it.  As a model may request no subsource, 
   *  toplevel_new_request can be used to place stop requests.
   */
  if (child->source_id && child->source_id != local_src_id)
    {
      if (local_src_id) 
	{
	  g_warning ("gdam_toplevel_new_request: child of type %s is already "
		     "playing, DETACHING source %d",
		     gtk_type_name (GTK_OBJECT (child)->klass->type),
		     child->source_id);
	}
#if 0
      else
	{
	  g_message ("gdam_toplevel_new_request: issuing stop request on "
	             "child of type %s, DETACHING source %d",
		     gtk_type_name (GTK_OBJECT (child)->klass->type),
		     child->source_id);
	}
#endif
      gdam_request_builder_append_generic (builder,
					   gdam_request_detach_source 
					   		(child->source_id,
							 NULL, NULL, NULL));
      gdam_model_set_source_id (child, 0);
    }
  /* If this is a stop request, but our child had no source id, then
   * we cannot stop... emit a warning, someone is confused.
   */
  else if (local_src_id == 0)
    {
      g_warning ("gdam_toplevel_new_request: stop request on child of type %s "
		 "but source id is already zero....  cannot detach.",
		  gtk_type_name (GTK_OBJECT (child)->klass->type));
    }
#if 0
  else
    {
       g_message ("gdam_toplevel_new_request: harmless lack of source id");
    }
#endif


  /*  If this is a play request, and we are playing in a synchronized fashion,
   *  retrieve the necessary timing info and append the synch request.
   */
  if (local_src_id && play_info != NULL && play_info->is_synchronized 
      && play_info->lead_id && play_info->lead_beat_info)
    {
      gdouble our_bpm, our_first_beat;
      gdouble lead_bpm, lead_first_beat;
      GdamArg sync_arg;
      GdamTime synchronization_data[6];
      our_bpm = gdam_beat_info_get_bpm (&child->beat_info, 0);
      our_first_beat = gdam_beat_info_get_first_beat (&child->beat_info, 0);
      lead_bpm = gdam_beat_info_get_bpm (play_info->lead_beat_info, 0);
      lead_first_beat =
	gdam_beat_info_get_first_beat (play_info->lead_beat_info, 0);
      gdam_debug ("toplevel: sync: us: bpm=%.6g, first=%.6g; "
		  "play_info: bpm=%.6g, first=%.6g; ", our_bpm,
		  our_first_beat, lead_bpm, lead_first_beat);

      gdam_request_helper_build_sync (local_src_id,
				      play_info->lead_id,
				      our_bpm,
				      our_first_beat,
				      lead_bpm,
				      lead_first_beat, 
				      synchronization_data);
      sync_arg.name = "synchronize";
      sync_arg.type = GDAM_ARG_TYPE_INT64_ARRAY;
      GDAM_ARG_INT64_ARRAY (&sync_arg).length = 6;
      GDAM_ARG_INT64_ARRAY (&sync_arg).array = synchronization_data;
      gdam_request_builder_append_generic (builder,
					   gdam_request_set_args
					   (local_src_id, 1, 
					    &sync_arg, NULL,
					    NULL, NULL));
      was_synchronized = TRUE;
    }
  /*  If we have a stop request which is asking for synchronization, then
   *  someone is confused.
   */
  else if (!local_src_id &&play_info != NULL && play_info->is_synchronized 
  	   && play_info->lead_id && play_info->lead_beat_info)
    {
      g_warning ("toplevel_new_request: trying to synchronize a stop request!  "
		 "child type: %s  is confused.",
		 gtk_type_name (GTK_OBJECT (child)->klass->type));
    }


  /* If this is a play request, attach. */
  if (local_src_id != 0)
    {
      guint mixer_id = info->device->mixer_id;
      GdamArg args[2];
      make_add_args (args, local_src_id);
      //g_message ("issuing play request on child of type %s local_src_id %d",
	  	 //gtk_type_name (GTK_OBJECT (child)->klass->type), local_src_id);
      gdam_request_builder_append_generic (builder,
					   gdam_request_set_args (mixer_id,
								  2, args,
								  NULL, NULL,
								  NULL));
    }
  
  /* If this is a play request, and our child doesn't want to miss data,
   * append source_doesnt_miss request.
   */
  if (local_src_id && !was_synchronized && !info->misses_data)
    {
      GdamArg arg;
      arg.name = "source_doesnt_miss";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = local_src_id;
      gdam_request_builder_append_generic 
			    (builder,
			     gdam_request_set_args (info->device->mixer_id,
						    1, &arg, 
						    NULL, NULL, NULL));
    }

  /* Post the requests to the channel, and free the builder. */
  gdam_channel_append_request (model->channel,
			       gdam_request_builder_package (builder));
  gdam_request_builder_destroy (builder);
}


/* BUG:   should these signals be connected to model? */
void gdam_toplevel_add_child   (GdamToplevel*   toplevel,
                                GdamModel*  model)
{
  GdamModelInfo *info;
  info = (GdamModelInfo *) g_hash_table_lookup (toplevel->model_to_info,
						model);
  if (info)
    {
      g_warning ("model %p [%s] was added twice", model,
		 gtk_type_name (GTK_OBJECT (model)->klass->type));
      return;
    }

  info = g_new (GdamModelInfo, 1);
  info->misses_data = TRUE;
  info->toplevel = toplevel;
  info->model = model;
  info->device = toplevel->default_device;
  g_return_if_fail ((info->device != NULL));
  device_ref (info->device);

  g_return_if_fail (model->parent != (GdamModel *) toplevel);
  g_return_if_fail (model->parent == NULL);


  /* Trap the destroy notify. */
  info->destroy_connect_id = gtk_signal_connect (GTK_OBJECT (model),
						 "destroy",
						 GTK_SIGNAL_FUNC (on_destroy),
						 info);

  /* Reference the model. */
  gtk_object_ref (GTK_OBJECT (model));
  gtk_object_sink (GTK_OBJECT (model));
  model->parent = GDAM_MODEL (toplevel);

  g_hash_table_insert (toplevel->model_to_info, model, info);
}

void gdam_toplevel_remove_child(GdamToplevel*   toplevel,
                                GdamModel*  model)
{

  GtkObject *object;
  GdamModelInfo *info;
  info = (GdamModelInfo *) g_hash_table_lookup (toplevel->model_to_info,
						model);
  g_return_if_fail (info != NULL);

  model->parent = NULL;
  object = GTK_OBJECT (model);
  gtk_signal_disconnect (object, info->destroy_connect_id);
  gtk_object_unref (object);
  g_hash_table_remove (toplevel->model_to_info, model);
  device_unref (info->device);
  g_free (info);
}
void          gdam_toplevel_set_child_device (GdamToplevel*   toplevel,
					      GdamModel*      model,
					      const char*     device_name)
{
  GdamModelInfo *model_info;
  GdamToplevelDevice *device;
  model_info = g_hash_table_lookup (toplevel->model_to_info, model);
  g_return_if_fail (model_info != NULL);
  device = g_hash_table_lookup (toplevel->devices_by_name, device_name);
  if (device == NULL)
    {
      g_warning ("toplevel: cannot set model to device %s", device_name);
      return;
    }
  if (model_info->device == device)
    {
      return;
    }
  model_info->device = device;
  if (model->source_id != 0)
    {
      GdamArg args[2];
      make_add_args (args, model->source_id);
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args (device->mixer_id,
							  2, args,
							  NULL, NULL, NULL));
    }
}

void    gdam_toplevel_set_misses_data   (GdamToplevel*   toplevel,
					 GdamModel*      model,
					 gboolean        misses_data)
{
  GdamModelInfo *model_info;
  model_info = g_hash_table_lookup (toplevel->model_to_info, model);
  g_return_if_fail (model_info != NULL);
  model_info->misses_data = misses_data;
}
