/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamturntable.h"
#include <stdlib.h>
#include <gdk/gdk.h>
#include <gtk/gtksignal.h>
#include <gdam/common/gdamsample.h>
#include <gdam/client/gdamrequest.h>
#include "gdamutil.h"
#include <math.h>
#include <string.h>

/*
 * Strategy:
 *       producing->beat_info the beat information for the underlying
 *                            producing model.
 *       this->beat_info      the resampled beat info.
 */

static void gdam_turntable_add_xml  (GdamModel*        model,
                                     xmlNode*          parent,
				     GdamXmlInfo*      xml_info);
static void gdam_turntable_parse_xml(GdamModel*        model,
                                     xmlNode*          node,
				     GdamXmlInfo*      xml_info);

/*   ____ _                 ___       _ _   
 *  / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 * | |   | |/ _` / __/ __|  | || '_ \| | __|
 * | |___| | (_| \__ \__ \  | || | | | | |_ 
 *  \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */                                        


#define PARTIAL 1

enum {
	UPDATE_ANCHORED,
	UPDATE_BEATNUM,
	UPDATE_SYNC_PREVIEW,
	UPDATE_JUMP_PREVIEW,
	LAST_SIGNAL
};

static guint turntable_signals[LAST_SIGNAL];

/* Model class functions. */
static gboolean gdam_turntable_match_beat_info(GdamModel* model,
					       const GdamBeatInfo* match_info);
/* Producer class functions */
static void     gdam_turntable_stop_request      (GdamProducer* producer,
                                                  GdamRequestBuilder* builder);
static int      gdam_turntable_play_request      (GdamProducer* producer,
					          GdamPlayInfo* play_info,
						  int           source_id,
                                                  GdamRequestBuilder* builder);
static void     gdam_turntable_real_set_position (GdamProducer*  producer,
					          gdouble	   position);
#if 0
static void     handle_producer_set_source_id    (GdamProducer*  producer,
			                          int       	   new_id,
			                          GdamTurntable* turntable);
#endif
static void 	resample_rate_changed		 (GdamTurntable* turntable);
static void     append_unsubscribe_events        (GdamTurntable* turntable,  
                                                  GdamRequestBuilder* builder);

static void gdam_turntable_arg_set_set_fb(GtkObject* object, GdamArg* arg)
{
  GdamTurntable *turntable;
  GdamProducer *producer;
  GdamBeatInfo *beat_info;
  gdouble position;

  turntable = GDAM_TURNTABLE (object);
  g_return_if_fail (turntable->producer != NULL);
  producer = turntable->producer;
  position = producer->stopped_position;

  beat_info = gdam_beat_info_copy (&GDAM_MODEL (producer)->beat_info);
  gdam_beat_info_set_first_beat (beat_info, position);
  gdam_model_set_beat_info (GDAM_MODEL (producer), beat_info);
  gdam_beat_info_destroy (beat_info);
}

static void 
gdam_turntable_arg_set_cue_point (GtkObject *object, 
                                  GdamArg   *arg)
{
  int iarg = GDAM_ARG_INT(arg);
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  g_return_if_fail (iarg >= 0 && iarg < GDAM_TURNTABLE_NUM_CUE_POINTS);

  /* set index point arg to current stopped position */
  /* TODO: if we are playing, get a more accurate position */
  turntable->cue_points[iarg] 
    = GDAM_PRODUCER (turntable)->stopped_position * turntable->rate;
}

static void 
gdam_turntable_arg_set_clear_cue_point (GtkObject *object, 
                                        GdamArg   *arg)
{
  int iarg = GDAM_ARG_INT(arg);
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  g_return_if_fail (iarg >= 0 && iarg < GDAM_TURNTABLE_NUM_CUE_POINTS);
  
  turntable->cue_points[iarg] = 0;
}

static void 
seek_to_cue_point (GdamTurntable *turntable,
		   int iarg,
		   gboolean in_beat)
{
  if (!turntable->producer)
    return;
  
  g_return_if_fail (iarg >= 0 && iarg < GDAM_TURNTABLE_NUM_CUE_POINTS);

  if (turntable->cue_points[iarg] > 0) 
    {
      if (GDAM_MODEL (turntable->producer)->source_id)
	{
          if (in_beat)
	    gdam_turntable_seek_in_beat (turntable, turntable->cue_points[iarg]);
	  else
	    gdam_producer_seek (GDAM_PRODUCER (turntable),
			        turntable->cue_points[iarg] / turntable->rate, 0);
	}
      else
	gdam_producer_set_position (turntable->producer,
				    turntable->cue_points[iarg]);
    }
}

static void 
gdam_turntable_arg_set_seek_to_cue_point (GtkObject *object, 
                                          GdamArg   *arg)
{
  seek_to_cue_point (GDAM_TURNTABLE (object), GDAM_ARG_INT (arg), 0);
}

static void 
gdam_turntable_arg_set_seek_in_beat_to_cue_point (GtkObject *object, 
                                                  GdamArg   *arg)
{
  seek_to_cue_point (GDAM_TURNTABLE (object), GDAM_ARG_INT (arg), 1);
}

static void gdam_turntable_arg_set_seek(GtkObject* object, GdamArg* arg)
{
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (object);
  gdam_turntable_seek (turntable, GDAM_ARG_DOUBLE (arg));
}

static void gdam_turntable_arg_set_seek_by_beats(GtkObject* object, 
						 GdamArg* arg)
{
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (object);
  gdam_turntable_seek_by_beats (turntable, GDAM_ARG_DOUBLE (arg));
}

/* TODO: get beats/bar from beatinfo structure?? */
static void gdam_turntable_arg_set_seek_by_bars(GtkObject* object, 
						GdamArg* arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  gdam_turntable_seek_by_beats (turntable, GDAM_ARG_DOUBLE (arg) * 4.0);
}

/*  This argument is used as a wrapper for the producer
 *  'playing' argument to add a few extra modes, ie
 *
 *  -1 		        => stop_or_first_beat
 *  n > 0 while playing => sync playing source
 *
 *    other values are forwarded to producer's playing arg.
 */
static void gdam_turntable_arg_set_play_or_synch(GtkObject* object, 
					         GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GdamProducer *producer = turntable->producer;
  GdamModel *model = GDAM_MODEL (turntable);
  GdamModel *pmodel = GDAM_MODEL (producer);
  int argval = GDAM_ARG_INT (arg);
  g_return_if_fail (producer != NULL);

  /*  If the arg is a stop-type code */
  if (argval == -1)
    {
      /* TODO: rearrange logic here? */
      if ((model->source_id != 0))
	{
	  GdamArg new_arg;
	  new_arg.type = GDAM_ARG_TYPE_INT;
	  new_arg.name = "playing";
	  GDAM_ARG_INT (&new_arg) = 0;
	  gdam_debug ("play_or_synch: setting playing to 0");
	  gdam_arg_set_object_arg (GTK_OBJECT (producer), &new_arg);
	  gdam_arg_set_object_arg (object, &new_arg);
	}
      else if ((pmodel->source_id != 0))
	{
	  GdamArg new_arg;
	  new_arg.type = GDAM_ARG_TYPE_INT;
	  new_arg.name = "playing";
	  GDAM_ARG_INT (&new_arg) = 0;
	  gdam_debug ("play_or_synch: setting playing to 0");
	  gdam_arg_set_object_arg (GTK_OBJECT (producer), &new_arg);
	  gdam_arg_set_object_arg (object, &new_arg);
	}
      else if (producer != NULL)
	{
	  gdouble first_beat;
	  first_beat = gdam_turntable_get_current_producer_fb (turntable);
	  gdam_debug ("play_or_synch: setting position to fb");
	  gdam_producer_set_position (producer, first_beat);
	}
      else
	gdam_debug ("play_or_synch: doing nothing");
    }
  /*  If the arg is a play-type arg and we're playing  */
  else if (((GDAM_MODEL (turntable)->source_id != 0)) && argval > 0)
    {
      turntable_do_sync (turntable);
    }
  /*  If the arg is a play-type arg and we're not playing,
   *  pass it on.  */
  else
    {
      GdamArg new_arg;
      new_arg.type = GDAM_ARG_TYPE_INT;
      new_arg.name = "playing";
      GDAM_ARG_INT (&new_arg) = argval;
      /* XXX: both really needed?  should 2nd set_arg be conditional? */
      gdam_arg_set_object_arg (GTK_OBJECT (producer), &new_arg);
      gdam_arg_set_object_arg (object, &new_arg);
    }
}

static void gdam_turntable_arg_set_resample_rate(GtkObject* object, 
					         GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  gdam_debug ("gdam turntable arg set rate");
  turntable->rate = GDAM_ARG_DOUBLE (arg);
  resample_rate_changed (turntable);
}

static void gdam_turntable_arg_get_resample_rate(GtkObject* object, 
					         GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GDAM_ARG_DOUBLE (arg) = turntable->rate;
}

static void gdam_turntable_arg_set_seek_factor(GtkObject* object, 
					       GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  turntable->seek_factor = GDAM_ARG_DOUBLE (arg);
}

static void gdam_turntable_arg_get_seek_factor(GtkObject* object, 
					         GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GDAM_ARG_DOUBLE (arg) = turntable->seek_factor;
}

static void gdam_turntable_arg_set_tempo_locked(GtkObject* object, 
					       GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  turntable->tempo_locked = GDAM_ARG_INT (arg);
}

static void gdam_turntable_arg_get_tempo_locked(GtkObject* object, 
					         GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GDAM_ARG_INT (arg) = turntable->tempo_locked;
}


static void gdam_turntable_arg_set_beatinfo_locked(GtkObject* object, 
					           GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  turntable->beatinfo_locked = GDAM_ARG_INT (arg);
}

static void gdam_turntable_arg_get_beatinfo_locked(GtkObject* object, 
					           GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GDAM_ARG_INT (arg) = turntable->beatinfo_locked;
}



/*  Recomputes turntable beat info based on producer beat info, resample rate,
 *  bpm_factor, and extended beat info.  The _bi argument is the producer's beat
 *  info.  If _bi is NULL, the producer beatinfo is retrieved from the producer
 *  model.  */
/* TODO: emit update_beatnum signal?? */
static void recompute_beat_info(GdamTurntable* turntable, 
				const GdamBeatInfo* _bi)
{
  GdamProducer *producer = turntable->producer;
  const GdamBeatInfo *bi;
  GdamBeatInfo *beat_info;
  int num_ranges = 0;
  float bpm = 0;
  float first_beat = 0;
  float next_first_beat = 0;
  float prev_last_beat = 0;
  g_return_if_fail (producer != NULL);
  bi = _bi ? _bi : &(GDAM_MODEL (producer)->beat_info);
  if (bi->range_list)
    num_ranges = g_list_length (bi->range_list);

  bpm = gdam_beat_info_get_range_bpm (bi, turntable->beatinfo_index);
  first_beat = gdam_beat_info_get_range_first_beat (bi, 
						    turntable->beatinfo_index);
  prev_last_beat = first_beat;

  /* If our current range is only an index point, then we need to find out
   * which full range describes our position and determine the bpm from that.
   */
  if (bpm < 1.0)
    {
      int valid_idx;
      valid_idx = gdam_beat_info_guess_range_index (bi, first_beat);
      bpm = gdam_beat_info_get_range_bpm (bi, valid_idx);
    }

  /* Determine the end of our range */
  next_first_beat = gdam_beat_info_guess_range_end (bi, 
      						    turntable->beatinfo_index);

  /* Translate new beatinfo range boundary points. */
  next_first_beat = next_first_beat / turntable->rate;
  turntable->next_first_beat = next_first_beat;
  prev_last_beat = prev_last_beat / turntable->rate;
  turntable->prev_last_beat = prev_last_beat;

  /* Construct a simple beatinfo from the data of the appropriate range 
   * of the extended beatinfo.
   */
  /* TODO: perform tempo translation on complex beatinfo structure?? */
  beat_info = gdam_beat_info_new ();
  bpm = bpm * turntable->rate / turntable->bpm_factor;
  first_beat = first_beat / turntable->rate;
  gdam_beat_info_set_bpm (beat_info, bpm);
  gdam_beat_info_set_first_beat (beat_info, first_beat);
  gdam_model_set_beat_info (GDAM_MODEL (turntable), beat_info);
  gdam_beat_info_destroy (beat_info);

  /* Check for sanity. */
  if (next_first_beat && next_first_beat <= first_beat)
    {
      g_warning ("problems with extended beatinfo... out of order?");
    }
}


void gdam_turntable_cycle_beat_info(GdamTurntable* turntable)
{
  GdamProducer *producer = turntable->producer;
  gdouble position;
  int num_ranges = 0;
  const GdamBeatInfo *bi;
  bi = &(GDAM_MODEL (producer)->beat_info);
  if (bi->range_list)
    num_ranges = g_list_length (bi->range_list);
  turntable->beatinfo_index++;
  if (turntable->beatinfo_index > num_ranges)
    turntable->beatinfo_index = 0;
  recompute_beat_info (turntable, NULL);
  position = GDAM_PRODUCER (turntable)->stopped_position;
  gtk_signal_emit (GTK_OBJECT (turntable),
		   turntable_signals[UPDATE_BEATNUM],
		   gdam_turntable_get_beat_number (turntable, position));
}



/* Given a time, we determine which beatinfo range we are in.  */
/*  TODO:  does this take producer position or turntable position??  */
void gdam_turntable_guess_beat_info(GdamTurntable* turntable, gdouble new_time)
{
  GdamProducer *producer = turntable->producer;
  int old_beatindex;
  int new_beatinfo_index;
  const GdamBeatInfo *bi;
  bi = &(GDAM_MODEL (producer)->beat_info);
  old_beatindex = turntable->beatinfo_index;

  /* Find the last (furthest into song) range which has a firstbeat
   * time before the new time. 
   */
  new_beatinfo_index = gdam_beat_info_guess_range_index (bi, new_time);

  /* If we have moved into a new range... */
  if (new_beatinfo_index != old_beatindex)
    {
      turntable->beatinfo_index = new_beatinfo_index;
      recompute_beat_info (turntable, NULL);
    }
}




static void gdam_turntable_arg_set_nudge_bpm_factor(GtkObject* object, 
					            GdamArg*   arg)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (object);
  GdamBeatInfo new_beat;
  float bpm;
  float fb;
  int val = GDAM_ARG_INT (arg);
  g_return_if_fail (turntable->producer != NULL);
  if (val == 1)
    {
      turntable->bpm_factor *= 2;
    }
  else if (val == -1)
    {
      turntable->bpm_factor /= 2;
    }
  else
    {
      g_warning ("turntable_arg_set_nudge_bpm_fac: %d bad value", val);
    }
  if (turntable->bpm_factor <= 0)
    {
      g_warning ("%f bad value for bpm_factor", turntable->bpm_factor);
      turntable->bpm_factor = 1;
    }
  recompute_beat_info (turntable, NULL);
}

static void _gdam_turntable_class_init(GdamTurntableClass* klass)
{
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  model_class->match_beat_info = gdam_turntable_match_beat_info;
  model_class->add_xml = gdam_turntable_add_xml;
  model_class->parse_xml = gdam_turntable_parse_xml;
  producer_class->stop_request = gdam_turntable_stop_request;
  producer_class->play_request = gdam_turntable_play_request;
  producer_class->set_position = gdam_turntable_real_set_position;

  gdam_arg_register (object_class,
		     "seek",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_turntable_arg_set_seek, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "seek_by_bars",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_turntable_arg_set_seek_by_bars,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "seek_by_beats",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_turntable_arg_set_seek_by_beats,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "resample_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_turntable_arg_set_resample_rate,
		     gdam_turntable_arg_get_resample_rate,
		     NULL, "0,8", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "seek_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_turntable_arg_set_seek_factor,
		     gdam_turntable_arg_get_seek_factor,
		     NULL, "0,8", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "tempo_locked",
		     GDAM_ARG_TYPE_INT,
		     gdam_turntable_arg_set_tempo_locked,
		     gdam_turntable_arg_get_tempo_locked,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "beatinfo_locked",
		     GDAM_ARG_TYPE_INT,
		     gdam_turntable_arg_set_beatinfo_locked,
		     gdam_turntable_arg_get_beatinfo_locked,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "nudge_bpm_factor",
		     GDAM_ARG_TYPE_INT,
		     gdam_turntable_arg_set_nudge_bpm_factor,
		     NULL, NULL, "-1,1", 0);
  gdam_arg_register (object_class,
		     "play_or_synch",
		     GDAM_ARG_TYPE_INT,
		     gdam_turntable_arg_set_play_or_synch,
		     NULL, NULL, "-2,3", 0);
  /* When this arg is set, we make the current position the first beat.
   * arg value unused. */
  gdam_arg_register (object_class,
		     "set_first_beat",
		     GDAM_ARG_TYPE_INT,
	 	     gdam_turntable_arg_set_set_fb, NULL, NULL, NULL, 0);
  /* Cue Points are temporary markers in the song which can be set and 
   * can be seeked to. */
  gdam_arg_register (object_class,
                     "seek_in_beat_to_cue_point",
                     GDAM_ARG_TYPE_INT,
                     gdam_turntable_arg_set_seek_in_beat_to_cue_point, 
                     NULL, NULL, "0,", 0);
  gdam_arg_register (object_class,
                     "seek_to_cue_point",
                     GDAM_ARG_TYPE_INT,
                     gdam_turntable_arg_set_seek_to_cue_point, 
                     NULL, NULL, "0,", 0);
  gdam_arg_register (object_class,
                     "set_cue_point",
                     GDAM_ARG_TYPE_INT,
                     gdam_turntable_arg_set_cue_point, NULL, NULL, "0,", 0);
  gdam_arg_register (object_class,
                     "clear_cue_point",
                     GDAM_ARG_TYPE_INT,
                     gdam_turntable_arg_set_clear_cue_point, 
                     NULL, NULL, "0,", 0);

  turntable_signals[UPDATE_BEATNUM] =
    gtk_signal_new ("update_beatnum",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamTurntableClass, update_beatnum),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  turntable_signals[UPDATE_SYNC_PREVIEW] =
    gtk_signal_new ("update_sync_preview",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamTurntableClass,
				       update_sync_preview),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  turntable_signals[UPDATE_JUMP_PREVIEW] =
    gtk_signal_new ("update_jump_preview",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamTurntableClass,
				       update_jump_preview),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, turntable_signals, LAST_SIGNAL);
}

static void 
_gdam_turntable_init(GdamTurntable* turntable)
{
  turntable->gain = 1.0;
  turntable->rate = 1.0;
  turntable->bpm_factor = 1.0;
}

GtkType gdam_turntable_get_type()
{
  static GtkType turntable_type = 0;
  if (!turntable_type)
    {
      static const GtkTypeInfo turntable_info = {
	"GdamTurntable",
	sizeof (GdamTurntable),
	sizeof (GdamTurntableClass),
	(GtkClassInitFunc) _gdam_turntable_class_init,
	(GtkObjectInitFunc) _gdam_turntable_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      turntable_type = gtk_type_unique (GDAM_TYPE_PRODUCER, &turntable_info);
    }
  return turntable_type;
}

GdamTurntable* gdam_turntable_new    (GdamChannel*        channel) 
{
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (gtk_type_new (GDAM_TYPE_TURNTABLE));
  ((GdamModel *) turntable)->channel = channel;
  return turntable;
}

/*
 *            _ _ _                _
 *   ___ __ _| | | |__   __ _  ___| | _____
 *  / __/ _` | | | '_ \ / _` |/ __| |/ / __|
 * | (_| (_| | | | |_) | (_| | (__|   <\__ \
 *  \___\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 * 
 */

/* Called when the turntable->rate variable has changed. */
/* TODO: scale length??? */
static void resample_rate_changed(GdamTurntable* turntable)
{
  GdamProducer *producer = turntable->producer;
  GdamModel *model = GDAM_MODEL (turntable);
  GdamModel *producer_model = producer ? GDAM_MODEL (producer) : NULL;
  GdamChannel *channel = model->channel;

  if (turntable->resample_id)
    {
      GdamArg arg;
      GdamRequest *request;
      arg.name = "rate";
      arg.type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (&arg) = turntable->rate;
      request = gdam_request_set_args (turntable->resample_id,
				       1,
				       &arg,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (channel, request);
    }
  if (producer != NULL)
    {
      recompute_beat_info (turntable, NULL);
      if (producer_model->source_id == 0)
	{
	  float stop_pos;
	  /* TODO: how to best handle rate of zero? */
	  stop_pos = turntable->rate ? 
	             producer->stopped_position / turntable->rate :
		     producer->stopped_position;
	  gdam_producer_set_position (GDAM_PRODUCER (turntable), stop_pos);
	}
    }
  return;
}

static void sync_preview_success(GdamTime preview_time, gpointer data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (data);
  gtk_signal_emit (GTK_OBJECT (turntable),
		   turntable_signals[UPDATE_SYNC_PREVIEW],
		   (double) preview_time / 44100.0);
  turntable->making_synch_preview_request = FALSE;
  gtk_object_unref (GTK_OBJECT (turntable));
}

static void sync_preview_failure(gpointer data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (data);
  g_warning ("server-side sync preview failure!!!");
  turntable->making_synch_preview_request = FALSE;
  gtk_object_unref (GTK_OBJECT (turntable));
}

void turntable_do_sync_preview(GdamTurntable* turntable)
{
  GdamBeatInfo new_beat_info;
  guint leader_id;
  GdamModel *model = (GdamModel *) turntable;
  GdamRequest *request;
  float comp_bpm;
  float comp_first_beat;
  float lead_bpm;
  float lead_first_beat;
  int tt_id;
  g_return_if_fail (model != NULL);
  if (turntable == NULL)
    {
      g_warning ("missing needed pointers in do_sync_preview");
    }
  if (!model->source_id)
    return;
  if (turntable->making_synch_preview_request)
    return;
  if (GDAM_MODEL (turntable)->anchored)
    {
      /*
         gtk_signal_emit(GTK_OBJECT(turntable),
         turntable_signals[UPDATE_SYNC_PREVIEW],
         0.0);
       */
      return;
    }
  if (!gdam_infomanager_get_sync_info (model, &leader_id, &new_beat_info))
    {
      gdam_debug ("turntable didnt get sync info, not previewing.");
      return;
    }
  comp_bpm = gdam_beat_info_get_bpm (&model->beat_info, 0);
  comp_first_beat = gdam_beat_info_get_first_beat (&model->beat_info, 0);
  lead_bpm = gdam_beat_info_get_bpm (&new_beat_info, 0);
  lead_first_beat = gdam_beat_info_get_first_beat (&new_beat_info, 0);
  if (lead_bpm < 1.0 || comp_bpm < 1.0)
    {
      g_warning ("missing bpm value: lead=%f comp=%f fb=%f,%f\n\n",
		 lead_bpm, comp_bpm, lead_first_beat, comp_first_beat);
      return;
    }
  if (turntable->producer->position_source_id)
    tt_id = turntable->producer->position_source_id;
  else
    tt_id = model->source_id;
  request = gdam_request_sync_preview (tt_id,
				       leader_id,
				       comp_bpm,
				       comp_first_beat,
				       lead_bpm,
				       lead_first_beat,
				       sync_preview_success,
				       sync_preview_failure, turntable);
  gdam_channel_append_request (model->channel, request);
  turntable->making_synch_preview_request = TRUE;
  gtk_object_ref (GTK_OBJECT (turntable));
  return;
}

void turntable_do_sync(GdamTurntable* turntable)
{
  GdamRequest *request;
  GdamBeatInfo static_beat_info;
  GdamModel *model = GDAM_MODEL (turntable);
  guint leader_id;
  float comp_bpm;
  float comp_first_beat;
  float static_bpm;
  float static_first_beat;
  int tt_id;
  if (turntable == NULL || model->source_id == 0)
    return;
  if (GDAM_MODEL (turntable)->anchored)
    {
      gtk_signal_emit (GTK_OBJECT (turntable),
		       turntable_signals[UPDATE_SYNC_PREVIEW], 0.0);
      return;
    }
  if (!gdam_infomanager_get_sync_info (model, &leader_id, &static_beat_info))
    return;
  comp_bpm = gdam_beat_info_get_bpm (&model->beat_info, 0),
    comp_first_beat = gdam_beat_info_get_first_beat (&model->beat_info, 0);
  static_bpm = gdam_beat_info_get_bpm (&static_beat_info, 0),
    static_first_beat = gdam_beat_info_get_first_beat (&static_beat_info, 0);
  gdam_debug ("sync: mobile id: %d bpm: %f fb: %f",
	      model->source_id, comp_bpm, comp_first_beat);
  gdam_debug ("sync: static id: %d bpm: %f fb: %f",
	      leader_id, static_bpm, static_first_beat);
  if (turntable->producer->position_source_id)
    tt_id = turntable->producer->position_source_id;
  else
    tt_id = GDAM_MODEL (turntable)->source_id;
  request = gdam_request_synchronize (tt_id,
				      leader_id,
				      comp_bpm,
				      comp_first_beat,
				      static_bpm,
				      static_first_beat,
				      gdam_do_nothing, gdam_do_nothing, NULL);
  gdam_channel_append_request (model->channel, request);
  gtk_signal_emit (GTK_OBJECT (turntable),
		   turntable_signals[UPDATE_SYNC_PREVIEW], 0.0);
}

gdouble gdam_turntable_get_current_producer_bpm(GdamTurntable* turntable){
	gdouble bpm = 0;
	GdamBeatInfo* bi = &GDAM_MODEL(turntable->producer)->beat_info;
	if (turntable->beatinfo_index == 0)
		bpm = gdam_beat_info_get_bpm(bi, 0);
	else {
		GdamBeatInfoRange* bir;
		gpointer data;
		data = g_list_nth_data(bi->range_list, 
				       turntable->beatinfo_index-1);
		bir = (GdamBeatInfoRange*)data;
		bpm = bir->bpm;
	}

	return bpm;
}

gdouble gdam_turntable_get_current_producer_fb(GdamTurntable* turntable){
	gdouble rv = 0;
	GdamBeatInfo* bi = &GDAM_MODEL(turntable->producer)->beat_info;
	if (turntable->beatinfo_index == 0)
		rv = gdam_beat_info_get_first_beat(bi, 0);
	else {
		GdamBeatInfoRange* bir;
		gpointer data;
		data = g_list_nth_data(bi->range_list, 
				       turntable->beatinfo_index-1);
		bir = (GdamBeatInfoRange*)data;
		rv = bir->first_beat;
	}
	return rv;
}

/*  Returns the bpm value of the turntable's output.  */
gdouble gdam_turntable_get_current_bpm(GdamTurntable* turntable){
	gdouble bpm = 0;
	bpm = gdam_turntable_get_current_producer_bpm(turntable);
	bpm = bpm * turntable->rate / turntable->bpm_factor;
	return bpm;
}

/*  Returns the firstbeat value of the turntable's output.  */
gdouble gdam_turntable_get_current_fb(GdamTurntable* turntable){
	gdouble fb = 0;
	fb = gdam_turntable_get_current_producer_fb(turntable);
	fb = fb / turntable->rate;
	return fb;
}


/* Seeks the turntable's producer by an even number of beats. */
void gdam_turntable_seek_by_beats(GdamTurntable* turntable, gdouble numBeats)
{
  gdouble bpm;
  /* TODO: check for unreasonable values? */
  gdam_debug ("tt:seek_by_beats - %f", numBeats);
  if (!turntable->producer)
    {
      gdam_debug ("no producer, aborting seek_by_beats");
      return;
    }
  bpm = gdam_turntable_get_current_producer_bpm (turntable);
  if (bpm < 1.0)
    gdam_turntable_seek (turntable, numBeats * 1.0);
  else
    {
      float length;
      length = numBeats * 60.0 / bpm;
      if (turntable->bpm_factor > 0)
	length *= turntable->bpm_factor;
      if (turntable->seek_factor > 0)
	length *= turntable->seek_factor;
      gdam_turntable_seek (turntable, length);
    }
}



/* Seeks the turntable's producer by an even number of beats....  target time
 * is in producer time.  */
void gdam_turntable_seek_in_beat(GdamTurntable* turntable, gdouble target_time)
{
  gdouble prod_bpm;
  gdouble current_prod_time;
  gdouble jump_length;
  const GdamBeatInfo *prod_beat;
  int source_range_index;
  int target_range_index;
  g_return_if_fail (turntable->producer != NULL);

  current_prod_time = GDAM_PRODUCER (turntable)->stopped_position;
  current_prod_time *= turntable->rate;
  jump_length = target_time - current_prod_time;
  prod_beat = &(GDAM_MODEL (turntable->producer)->beat_info);
  source_range_index = turntable->beatinfo_index;
  target_range_index = gdam_beat_info_guess_range_index (prod_beat,target_time);
  /* TODO: check for unreasonable values? */
  gdam_debug ("tt:seek_in_beat to %f", target_time);
  prod_bpm = gdam_turntable_get_current_producer_bpm (turntable);

  /* TODO: should we only seek if we can do it in time?? */
  if (prod_bpm < 1.0)
    {
      gdam_turntable_seek (turntable, jump_length);
      return;
    }
  /* TODO: if tt is not playing (or, always?) let the
   * target time be <= rather than < */
  /* If we are working within a single tempo, we need merely ensure that
   * the jump length is an even number of beats/bars.
   */
  if (source_range_index == target_range_index)
    {
      gdouble unit_length;
      int units;
      unit_length = 4 * 60.0 / prod_bpm;
      /*  If we are seeking forward, round small so our even jump
       *  lands us before the target time.  */
      if (target_time > current_prod_time)
	{
	  units = (int) (jump_length / unit_length);
	}
      /*  If we are seeking backward, round big so our even jump
       *  lands us before the target time.  */
      else
	{
	  units = floor (jump_length / unit_length);
	}
      jump_length = units * unit_length;

      gdam_turntable_seek (turntable, jump_length);
    }
  /* If we are working across tempos, we must ensure that we stay in
   * phase in a trickier fashion.
   */
  else
    {
      gdouble unit_length;
      gdouble target_unit_length;
      gdouble target_bpm;
      gdouble delta_time;
      gdouble delta_phase;
      gdouble source_phase;
      gdouble target_phase;
      gdouble new_target_time;
      gdouble prod_fb;
      int units;

      /* Get beat data for current position */
      prod_fb = gdam_turntable_get_current_producer_fb (turntable);
      unit_length = 4 * 60.0 / prod_bpm;
      source_phase = current_prod_time - prod_fb;
      source_phase = fmod (source_phase, unit_length);
      source_phase /= unit_length;

      /* Get beat data for target position. */
      prod_fb = gdam_beat_info_get_range_first_beat (prod_beat, 
		      				     target_range_index);
      target_bpm = gdam_beat_info_get_range_bpm (prod_beat, target_range_index);
      target_unit_length = 4 * 60.0 / target_bpm;
      target_phase = target_time - prod_fb;
      target_phase = fmod (target_phase, target_unit_length);
      target_phase /= target_unit_length;

      /* Calculate the nudge we must apply to stay in synch. */
      delta_phase = source_phase - target_phase;
      delta_time = (1.0 - delta_phase) * unit_length * -1.0;

      /* XXX: Give ourselves a little leeway for latency, to ensure
       * we don't seek to just after the target point because current_time
       * is a little old... */
      if (delta_time > -0.15)
	{
	  g_message ("Too little room in delta_time %f", delta_time);
	  delta_time -= unit_length;
	}
      jump_length += delta_time;
      gdam_turntable_seek (turntable, jump_length);
    }
}





/*  Seeks the turntable's PRODUCER by jump_amount.   jump_amount is in
 *  producer time, ie before the resample filter.  */
void gdam_turntable_seek(GdamTurntable* turntable, gdouble jump_amount)
{
  gdam_debug ("tt:seek - %f prod_id: %d", jump_amount,
	      GDAM_MODEL (turntable->producer)->source_id);
  if (turntable->producer && GDAM_MODEL (turntable->producer)->source_id)
    {
      gdam_producer_seek (turntable->producer, jump_amount, 1);
    }
  else
    {
      gdouble cur_prod_pos;
      gdouble new_prod_pos;
      cur_prod_pos = GDAM_PRODUCER (turntable)->stopped_position;
      cur_prod_pos *= turntable->rate;
      new_prod_pos = cur_prod_pos + jump_amount;
      gdam_producer_set_position (turntable->producer, new_prod_pos);
    }
}


static void zero_source_id(gpointer data)
{
  GdamModel *model;
  model = GDAM_MODEL (data);
  gdam_debug ("turntable_stop: about to set source_id to 0");
  gdam_model_set_source_id (model, 0);
}

static void gdam_turntable_stop_request(GdamProducer* producer,
                                        GdamRequestBuilder* builder)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (producer);
  GdamModel *model = GDAM_MODEL (producer);
  int source_id = model->source_id;
  GdamArg arg;
  GdamRequest *request;


  gdam_debug ("turntable_stop_req called, source_id=%d", source_id);

  if (!turntable->producer)
    {
      gdam_debug ("turntable_stop_req aborting, no producer.");
      return;
    }
  gdam_debug ("turntable_stop_req...  calling set_source_id 0");
  /* XXX: cannot set source_id to zero because if we are added to a toplevel,
   * the toplevel needs our model->source_id to post our requests...
   * however, if we reset source_id in a callback (as in producer)
   * we might get caught in an in-between state.
   */
  gdam_model_set_source_id (model, 0);
  if (source_id)
    append_unsubscribe_events (turntable, builder);
  arg.type = GDAM_ARG_TYPE_INT;
  arg.name = "filtered";
  GDAM_ARG_INT (&arg) = 0;
  request = gdam_request_set_args (source_id, 1, &arg, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);
}

static void producer_ended_callback(gpointer user_data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (user_data);
  GdamProducer *producer = turntable->producer;
  GdamChannel *channel = GDAM_MODEL (turntable)->channel;
  GdamArg new_arg;
  /* TODO: emit stopped signal */

  gdam_debug ("turntable: producer ended.");
  g_return_if_fail (producer != NULL);
  if (turntable->pos_sub_id)
    {
      gdam_channel_append_request (channel,
				   gdam_request_unsubscribe (turntable->
							     pos_sub_id, NULL,
							     NULL, NULL));
      turntable->pos_sub_id = 0;
    }
  if (turntable->eof_sub_id)
    {
      gdam_channel_append_request (channel,
				   gdam_request_unsubscribe (turntable->
							     eof_sub_id, NULL,
							     NULL, NULL));
      turntable->eof_sub_id = 0;
    }
  new_arg.type = GDAM_ARG_TYPE_INT;
  new_arg.name = "playing";
  GDAM_ARG_INT (&new_arg) = 0;
  gdam_arg_set_object_arg (GTK_OBJECT (producer), &new_arg);
  /* XXX do we want this?? */
  //gdam_model_set_source_id(GDAM_MODEL(turntable), 0);
}

static void producer_position_callback(GdamTime time, gpointer user_data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (user_data);
  float pos = (double) time / 44100.0;
  pos /= turntable->rate;
  //gdam_debug("turntable_producer_pos_callback: %f .", pos);
  if (!turntable->user_dragging_position)
    gdam_producer_set_position (GDAM_PRODUCER (turntable), pos);
}


static void on_resample_create_func(int resample_id, gpointer turntable_ptr)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (turntable_ptr);
  turntable->resample_id = resample_id;
  gdam_debug ("gdam_turntable: resample created successfully, id %d",
	      resample_id);
}

static void on_resample_create_fail_func(gpointer turntable_ptr)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (turntable_ptr);
  gdam_debug ("gdam_turntable [%p]: resample created failed", turntable);
}

static GdamRequest*
build_resample_request(GdamTurntable* turntable, guint subsource_id)
{
  GdamChannel *channel = GDAM_MODEL (turntable)->channel;
  return gdam_request_create_source ("GdasResample",
				     gdam_channel_new_local_id (channel),
				     on_resample_create_func,
				     on_resample_create_fail_func,
				     turntable,
				     2,
				     "rate",
				     GDAM_ARG_TYPE_DOUBLE,
				     turntable->rate,
				     "filtered",
				     GDAM_ARG_TYPE_INT, subsource_id);
}

static void on_gain_create_func(int gain_id, gpointer turntable_ptr)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (turntable_ptr);
  gdam_model_set_source_id (GDAM_MODEL (turntable), gain_id);
  gdam_debug ("gdam_turntable: gain created successfully, id %d", gain_id);
}

static void on_gain_create_fail_func(gpointer turntable_ptr)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (turntable_ptr);
  gdam_debug ("gdam_turntable [%p]: gain created failed", turntable);
}


static GdamRequest*
build_gain_request(GdamTurntable* turntable, guint tmp_resample_id)
{
  GdamChannel *channel = GDAM_MODEL (turntable)->channel;
  return gdam_request_create_source ("GdasGain",
				     gdam_channel_new_local_id (channel),
				     on_gain_create_func,
				     on_gain_create_fail_func,
				     turntable,
				     2,
				     "gain",
				     GDAM_ARG_TYPE_DOUBLE,
				     turntable->gain,
				     "filtered",
				     GDAM_ARG_TYPE_INT, tmp_resample_id);
}

typedef struct _TempoInfo {
  GdamTurntable* turntable;
  gdouble target_bpm;
} TempoInfo;


static int idle_set_tempo(gpointer data) 
{
  TempoInfo* ti;
  ti = (TempoInfo *) data;
  turntable_set_tempo (ti->turntable, ti->target_bpm);
  return 0;
}


static void     handle_producer_set_beat_info      (GdamProducer*  producer,
			                            const GdamBeatInfo* bi,
			                            GdamTurntable* turntable)
{
  gdouble last_bpm;
  gdouble new_bpm;
  last_bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (turntable)->beat_info, 0);
  recompute_beat_info (turntable, bi);
  new_bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (turntable)->beat_info, 0);
  if (last_bpm > 0.1 && new_bpm > 0 && new_bpm != last_bpm)
    {
      if (turntable->tempo_locked)
	{
	  TempoInfo* ti;
	  ti = g_new (TempoInfo, 1);
	  ti->turntable = turntable;
	  ti->target_bpm = last_bpm;
	  g_idle_add (idle_set_tempo, ti);
	}
    }
}

static void     handle_producer_set_position       (GdamProducer*  producer,
			                            double	   prod_pos,
			                            GdamTurntable* turntable)
{
  float tt_position;
  tt_position = prod_pos / turntable->rate;
  gdam_producer_set_position (GDAM_PRODUCER (turntable), tt_position);
}

static void     handle_producer_update_length      (GdamProducer*  producer,
			                            double	   prod_len,
			                            GdamTurntable* turntable)
{
  float tt_len;
  tt_len = prod_len / turntable->rate;
  //turntable->track_length = tt_len;
  turntable->track_length = prod_len;
  turntable->beatinfo_index = 0;
  gdam_producer_update_length (GDAM_PRODUCER (turntable), tt_len);
}

static void append_unsubscribe_events(GdamTurntable* turntable,  
                                      GdamRequestBuilder* builder)
{

  if (turntable->eof_sub_id)
    {
      gdam_request_builder_append_generic (builder,
					   gdam_request_unsubscribe
					   (turntable->eof_sub_id, NULL, NULL,
					    NULL));
      turntable->eof_sub_id = 0;
    }
  if (turntable->pos_sub_id)
    {
      gdam_request_builder_append_generic (builder,
					   gdam_request_unsubscribe
					   (turntable->pos_sub_id, NULL, NULL,
					    NULL));
      turntable->pos_sub_id = 0;
    }
}


static void eof_connect_success(int eof_sub_id, gpointer user_data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (user_data);
  turntable->eof_sub_id = eof_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (turntable)->channel,
				 gdam_channel_simple_handler_new (eof_sub_id,
								  TRUE,
								  producer_ended_callback,
								  turntable));
}

static void 
position_callback_connect_success(int pos_sub_id, gpointer user_data)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (user_data);
  turntable->pos_sub_id = pos_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (turntable)->channel,
				 gdam_channel_time_handler_new (pos_sub_id,
								FALSE,
								producer_position_callback,
								turntable));
}

static void append_subscribe_events(GdamTurntable*      turntable,
                                    GdamRequestBuilder* builder,
				    int                 sub_source_id)
{
  g_return_if_fail (turntable->producer != NULL);
  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_ended
				       (sub_source_id, eof_connect_success,
					NULL, turntable));

  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_timer
				       (sub_source_id, 200,
					position_callback_connect_success,
					NULL, turntable));
  gdam_debug ("turntable: appending subscribe_timer request");
}


static int  gdam_turntable_play_request(GdamProducer*  producer,
				        GdamPlayInfo*  play_info,
					int            source_id,
					GdamRequestBuilder* builder)
{
  GdamTurntableClass *tt_class;
  GdamTurntable *turntable = GDAM_TURNTABLE (producer);
  GdamRequest *request;
  int resample_id_tmp, gain_id_tmp;
  GdamPlayInfo *a_play_info = play_info
    ? play_info : g_new0 (GdamPlayInfo, 1);
  tt_class = GDAM_TURNTABLE_CLASS (GTK_OBJECT (turntable)->klass);

  /* If we are already playing, sync and return.
   * TODO: really here in play_request?? or just in arg_set / _play */
  if (GDAM_MODEL (turntable)->source_id != 0)
    {
      turntable_do_sync (turntable);
      return 0;
    }
  if (turntable->producer == NULL)
    {
      g_warning ("no producer... aborting turntable_play_song");
      return 0;
    }
  append_unsubscribe_events (turntable, builder);

  if (source_id == 0)
    source_id = gdam_producer_play_request (turntable->producer,
					    a_play_info, 0, builder);

  /* If our producer couldn't play, then neither can we. */
  if (source_id == 0)
    {
      if (a_play_info && a_play_info != play_info)
	g_free (a_play_info);
      return 0;
    }

  request = build_resample_request (turntable, source_id);
  resample_id_tmp = gdam_request_builder_append_create (builder, request);
  request = build_gain_request (turntable, resample_id_tmp);
  gain_id_tmp = gdam_request_builder_append_create (builder, request);

  if (turntable->producer->position_source_id)
    append_subscribe_events (turntable, builder, 
	                     turntable->producer->position_source_id);
  else
    append_subscribe_events (turntable, builder, source_id);
  /* ??? */
  a_play_info->position = producer->stopped_position * turntable->rate;
  GDAM_MODEL (turntable)->source_id = gain_id_tmp;
  return gain_id_tmp;
}

/* TODO: forward stopped_pos to sub producer???  (need to distinguish
 *       between producer having been set and getting stopped position?) */
static void gdam_turntable_real_set_position(GdamProducer* ttproducer,
					     gdouble	   position)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (ttproducer);
  GdamProducer *sub_prod = turntable->producer;
  float sub_pos;
  ttproducer->stopped_position = position;
  if (turntable->next_first_beat > 0 
      && position > turntable->next_first_beat)
    {
      if (!turntable->beatinfo_locked)
	gdam_turntable_guess_beat_info (turntable, position);
    }
  else if (turntable->prev_last_beat > 0
	   && position < turntable->prev_last_beat
	   && !turntable->beatinfo_locked)
    {
      gdam_turntable_guess_beat_info (turntable, position);
    }
  sub_pos = position * turntable->rate;
  gdam_producer_set_position (sub_prod, sub_pos);
  gtk_signal_emit (GTK_OBJECT (turntable),
		   turntable_signals[UPDATE_BEATNUM],
	   gdam_turntable_get_beat_number (turntable, position));
}

/* Changes the turntable's resampling rate so that the apparent tempo of the
 *   output sound matches the argument.
 */
void turntable_set_tempo(GdamTurntable* turntable, gdouble new_tempo)
{
  float rate;
  GdamChannel *channel = ((GdamModel *) turntable)->channel;
  GdamModel *model = GDAM_MODEL (turntable);
  GdamModel *prod_model = GDAM_MODEL (turntable->producer);
  GdamBeatInfo *new_beat;
  float prod_bpm;
  float prod_fb;
  float bpm_factor = turntable->bpm_factor;
  prod_fb = gdam_turntable_get_current_producer_fb (turntable);
  prod_bpm = gdam_turntable_get_current_producer_bpm (turntable);
  if (prod_bpm <= 0.0 || turntable->producer == NULL)
    return;
  rate = new_tempo * bpm_factor / prod_bpm;

  gdam_debug ("turntable: matching to %f prod bpm %f new rate %f", new_tempo,
	      prod_bpm, rate);

  if (turntable->resample_id)
    {
      GdamArg arg;
      GdamRequest *request;
      arg.name = "rate";
      arg.type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (&arg) = rate;
      request = gdam_request_set_args (turntable->resample_id,
				       1,
				       &arg,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (channel, request);
    }
  if (!turntable->producer)
    {
      g_warning ("no producing submodel");
      return;
    }

  turntable->rate = rate;

  new_beat = gdam_beat_info_new ();
  gdam_beat_info_set_bpm (new_beat, prod_bpm * rate);
  gdam_beat_info_set_first_beat (new_beat, prod_fb / rate);
  gdam_model_set_beat_info (model, new_beat);
  gdam_turntable_set_rate (turntable, rate);
  gdam_beat_info_destroy (new_beat);
}

GdamProducer* gdam_turntable_get_producer(GdamTurntable*      turntable)
{
  return turntable->producer;
}

void gdam_turntable_set_producer(GdamTurntable*      turntable,
                                 GdamProducer*       producer)
{
  if (turntable->producer)
    {
      GtkObject *obj_producer = GTK_OBJECT (turntable->producer);
      //gtk_signal_disconnect(obj_producer,
      //turntable->sub_set_source_handle);
      GDAM_MODEL (turntable->producer)->parent = NULL;
      gtk_signal_disconnect (obj_producer,
			     turntable->sub_set_beat_info_handle);
      gtk_signal_disconnect (obj_producer,
			     turntable->sub_set_position_handle);
      gtk_signal_disconnect (obj_producer,
			     turntable->sub_update_length_handle);
      gtk_object_unref (obj_producer);
    }
  turntable->producer = producer;
  if (producer)
    {
      GtkSignalFunc handle_func;
      GDAM_MODEL (producer)->parent = GDAM_MODEL (turntable);
      gtk_object_ref (GTK_OBJECT (producer));
      gtk_object_sink (GTK_OBJECT (producer));
      handle_func = GTK_SIGNAL_FUNC (handle_producer_set_position);
      turntable->sub_set_position_handle
	= gtk_signal_connect (GTK_OBJECT (producer),
			      "set_position", handle_func, turntable);
      handle_func = GTK_SIGNAL_FUNC (handle_producer_update_length);
      turntable->sub_update_length_handle
	= gtk_signal_connect (GTK_OBJECT (producer),
			      "update_length", handle_func, turntable);
      handle_func = GTK_SIGNAL_FUNC (handle_producer_set_beat_info);
      turntable->sub_set_beat_info_handle
	= gtk_signal_connect (GTK_OBJECT (producer),
			      "set_beat_info", handle_func, turntable);
    }
}


float gdam_turntable_get_beat_number(GdamTurntable* turntable, float _netpos)
{
  GdamModel *model;
  double since_first_beat;
  float netpos = _netpos;
  //double bpm, first_beat;
  //gdam_debug("turntable: getting beat number for time %f.", netpos);
  if (!turntable || !turntable->producer)
    {
      gdam_debug ("get_beat_number aborting");
      return 0.0;
    }
  model = GDAM_MODEL (turntable);
  since_first_beat = netpos -
    gdam_beat_info_get_first_beat (&model->beat_info, 0);
  return since_first_beat *
    gdam_beat_info_get_bpm (&model->beat_info, 0) / 60.0;
}

void gdam_turntable_zero(GdamTurntable* turntable)
{
  GdamModel *producer = GDAM_MODEL (turntable->producer);
  if (!producer)
    {
      g_warning ("gdam_turntable_zero: no producer ");
      return;
    }
  if (producer)
    {
      if (producer->source_id)
	{
	  GdamPlayInfo *play_info = g_new0 (GdamPlayInfo, 1);
	  play_info->position = 0.0;
	  gdam_producer_play (turntable->producer, play_info);
	  g_free (play_info);
	}
      else
	{
	  GDAM_PRODUCER (turntable)->stopped_position = 0.0;
	}
    }
  gdam_producer_set_position (GDAM_PRODUCER (turntable), 0.0);
}

/* Seeks to the first beat of the song. */
void gdam_turntable_start(GdamTurntable* turntable)
{
  GdamModel *ttcomp = GDAM_MODEL (turntable);
  gdouble first_beat;
  gdouble producer_first_beat;
  gdam_debug ("gdam_turntable_start");
  if (!turntable->producer)
    {
      g_warning ("gdam_turntable_start: no producer: cannot start");
      return;
    }
  first_beat = gdam_beat_info_get_first_beat (&ttcomp->beat_info, 0);
  producer_first_beat = gdam_turntable_get_current_producer_fb (turntable);
  if (turntable->producer)
    {
      if (GDAM_MODEL (turntable->producer)->source_id)
	{
	  GdamPlayInfo *play_info = g_new0 (GdamPlayInfo, 1);
	  play_info->position = producer_first_beat;
	  gdam_producer_play (turntable->producer, play_info);
	  g_free (play_info);
	}
      else
	{
	  gdam_producer_set_position (turntable->producer,
				      producer_first_beat);
	}
    }
  gdam_producer_set_position (GDAM_PRODUCER (turntable), first_beat);
}

static gboolean gdam_turntable_match_beat_info(GdamModel* model,
					       const GdamBeatInfo* match_info)
{
  GdamTurntable *turntable = GDAM_TURNTABLE (model);
  float bpm = gdam_beat_info_get_bpm (match_info, 0);
  if (!turntable || !match_info)
    g_warning ("turntable_match_beat_info: NULL argument.");
  turntable_set_tempo (turntable, bpm);
  return TRUE;
}

void gdam_turntable_set_anchored(GdamTurntable* turntable, gboolean anchored)
{
  GdamArg arg;
  arg.name = "anchored";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = (int) anchored;
  gdam_arg_set_object_arg (GTK_OBJECT (turntable), &arg);
}

void gdam_turntable_set_rate(GdamTurntable* turntable, float newrate)
{
  GdamArg arg;
  arg.name = "resample_rate";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = newrate;
  gdam_arg_set_object_arg (GTK_OBJECT (turntable), &arg);
}

void gdam_turntable_update_beatnum(GdamTurntable* turntable, float newbn)
{
  GtkObject *this_object = GTK_OBJECT (turntable);

  if (gtk_signal_n_emissions (this_object, turntable_signals[UPDATE_BEATNUM]))
    return;

  gtk_signal_emit (this_object, turntable_signals[UPDATE_BEATNUM], newbn);
}


void gdam_turntable_update_jump_preview(GdamTurntable* turntable, float newval)
{
  GtkObject *this_object = GTK_OBJECT (turntable);

  if (gtk_signal_n_emissions (this_object,
			      turntable_signals[UPDATE_JUMP_PREVIEW]))
    return;

  gtk_signal_emit (this_object,
		   turntable_signals[UPDATE_JUMP_PREVIEW], newval);
}



static void gdam_turntable_add_xml(GdamModel*     model,
                                   xmlNode*       parent,
				   GdamXmlInfo*   xml_info)
{
  GdamTurntable *turntable;
  GdamProducer *producer;
  turntable = GDAM_TURNTABLE (model);

  producer = turntable->producer;
  if (producer != NULL
      && gdam_xml_info_lookup_by_object (xml_info, producer) == 0)
    {
      xmlNode *node;
      xmlNode *stack_node;
      GdamModel *producer = GDAM_MODEL (turntable->producer);
      node = gdam_model_to_xml (producer, xml_info);
      stack_node = xmlNewNode (NULL, "producer");
      xmlAddChild (stack_node, node);
      xmlAddChild (parent, stack_node);
    }
}

static void gdam_turntable_parse_xml(GdamModel*     model,
                                     xmlNode*       node,
				     GdamXmlInfo*   xml_info)
{
  GdamModel *submodel;
  GdamTurntable *turntable;
  turntable = GDAM_TURNTABLE (model);
  if (strcmp (node->name, "producer") == 0)
    {
      if (node->childs == NULL || node->childs->next != NULL)
	{
	  g_warning ("tag <filter> must have exactly one subtag");
	  return;
	}
      submodel = gdam_model_from_xml (node->childs, model->channel, xml_info);
      if (!GDAM_IS_PRODUCER (submodel))
	{
	  GtkType type = GTK_OBJECT (submodel)->klass->type;
	  g_warning ("the turntable's producer was set to %s, "
		     "which does not derive from GdamProducer",
		     gtk_type_name (type));
	  return;
	}
      gdam_turntable_set_producer (turntable, GDAM_PRODUCER (submodel));
      return;
    }

  g_warning ("node <%s> unhandler in GdamBuilder", node->name);
}

