/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamview.h"
#include <gdam/common/gdamplugin.h>
#include <gtk/gtksignal.h>
#include <stdlib.h>
#include <string.h>

static void gdam_view_init(GdamView* view)
{
}
static void gdam_view_class_init(GdamViewClass* view_class)
{
}

GtkType gdam_view_get_type()
{
  static GtkType view_type = 0;
  if (!view_type)
    {
      static const GtkTypeInfo view_info = {
	"GdamView",
	sizeof (GdamView),
	sizeof (GdamViewClass),
	(GtkClassInitFunc) gdam_view_class_init,
	(GtkObjectInitFunc) gdam_view_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      view_type = gtk_type_unique (parent, &view_info);
    }
  return view_type;
}

xmlNode* gdam_view_to_xml(GdamView* skin, GdamXmlInfo* info)
{
  GdamViewClass *klass;
  GtkObjectClass *object_class;
  xmlNode *rv = NULL;
  int id;
  gboolean built_info = FALSE;
  object_class = GTK_OBJECT (skin)->klass;
  klass = GDAM_VIEW_CLASS (object_class);
  if (info == NULL)
    {
      built_info = TRUE;
      info = gdam_xml_info_new ();
    }

  g_return_val_if_fail (gdam_xml_info_lookup_by_object (info, skin) == 0, 0);
  id = gdam_xml_info_alloc_id (info, skin);
  if (klass->make_xml != NULL)
    {
      rv = xmlNewNode (NULL, gtk_type_name (object_class->type));
      if ((*klass->make_xml) (skin, rv, info) == FALSE)
	{
	  xmlFreeNode (rv);
	  rv = NULL;
	}
      {
	char id_buf[255];
	snprintf (id_buf, sizeof (id_buf), "%d", id);
	xmlSetProp (rv, "id", id_buf);
      }
    }
  if (rv == NULL)
    g_warning ("error making skin for %s",
	       gtk_type_name (object_class->type));
  if (built_info)
    gdam_xml_info_destroy (info);
  return rv;
}

gboolean gdam_view_parse_xml(GdamView*          skin,
                             xmlNode*           node,
			     GdamXmlInfo*       xml_info)
{
  GdamViewClass *klass;
  GtkObjectClass *object_class;
  gboolean must_destroy_xml = FALSE;
  object_class = GTK_OBJECT (skin)->klass;
  klass = GDAM_VIEW_CLASS (object_class);
  gdam_debug ("gdam_view_parse_xml invoked on <%s>", node->name);
  if (klass->parse_xml != NULL)
    {
      gboolean rv = FALSE;
      xmlNode *at;

      if (xml_info == NULL)
	{
	  xml_info = gdam_xml_info_new ();
	  must_destroy_xml = TRUE;
	}

      /* XXX: this is total hackery, but some skins (mininetwork)
       * assume that the model is always built first.
       */
      for (at = node->childs; at != NULL; at = at->next)
	{
	  if (g_strcasecmp (at->name, "model") == 0)
	    if ((*klass->parse_xml) (skin, at, xml_info))
	      rv = TRUE;
	}
      for (at = node->childs; at != NULL; at = at->next)
	{
	  gdam_debug ("at->parse_xml invoked on <%s>", at->name);
	  if (g_strcasecmp (at->name, "model") != 0)
	    if ((*klass->parse_xml) (skin, at, xml_info))
	      rv = TRUE;
	}
      if (rv)
	gdam_view_synchronize (skin);
      if (must_destroy_xml)
	gdam_xml_info_destroy (xml_info);
      return rv;
    }
  g_warning ("no parse_xml method in class %s",
	     gtk_type_name (object_class->type));
  return FALSE;
}
GdamView* gdam_view_from_xml(xmlNode* node, GdamXmlInfo* xml_info)
{
  const char *type_name;
  GtkType type;
  GdamView *view;
  char *id_str;

  gboolean must_destroy_xml = FALSE;

  type_name = node->content;

  if (strcmp (type_name, "id") == 0 && xml_info != NULL
      && node->childs != NULL && node->childs->next == NULL)
    {
      int id = atoi (node->childs->content);
      return GDAM_VIEW (gdam_xml_info_lookup_by_id (xml_info, id));
    }

  type = gdam_plugin_get_type_by_name ("skin", type_name);
  if (type == 0)
    {
      g_warning ("no type `%s': cannot load xml file", type_name);
      return NULL;
    }
  if (!gtk_type_is_a (type, GDAM_TYPE_VIEW))
    {
      g_warning ("%s is not a GdamView... cannot continue", type_name);
      return NULL;
    }
  if (xml_info == NULL)
    {
      xml_info = gdam_xml_info_new ();
      must_destroy_xml = TRUE;
    }
  view = GDAM_VIEW (gtk_type_new (type));
  id_str = xmlGetProp (node, "id");
  if (id_str != NULL)
    gdam_xml_info_assoc (xml_info, atoi (id_str), view);
  if (!gdam_view_parse_xml (view, node->childs, xml_info))
    {
      gtk_object_sink (GTK_OBJECT (view));
      if (must_destroy_xml)
	gdam_xml_info_destroy (xml_info);
      return NULL;
    }
  if (must_destroy_xml)
    gdam_xml_info_destroy (xml_info);
  return view;
}

static void on_skin_destroy(GtkObject* skin, GdamModel* model)
{
  g_return_if_fail (model->skin_list != NULL);
  model->skin_list = g_slist_remove (model->skin_list, skin);
  if (model->skin_list == NULL)
    gtk_object_destroy (GTK_OBJECT (model));
}

void           gdam_view_add (GdamView*   skin,
                                    GdamModel*  model)
{
  GdamViewClass *view_class;
  view_class = GDAM_VIEW_CLASS (GTK_OBJECT (skin)->klass);
  model->skin_list = g_slist_prepend (model->skin_list, skin);
  gtk_signal_connect (GTK_OBJECT (skin),
		      "destroy", GTK_SIGNAL_FUNC (on_skin_destroy), model);
}

void gdam_view_synchronize(GdamView* skin)
{
  GdamViewClass *cs_class;
  cs_class = GDAM_VIEW_CLASS (GTK_OBJECT (skin)->klass);
  if (cs_class->synchronize != NULL)
    (*cs_class->synchronize) (skin);
}
