/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gdam/common/gdamplugin.h>
#include "gdambalance.h"


enum {
	VOLUME_LL_CHANGED,
	VOLUME_LR_CHANGED,
	VOLUME_RL_CHANGED,
	VOLUME_RR_CHANGED,
	LAST_SIGNAL
};

static gboolean sent_balance_array(gpointer user_data)
{
  GdamBalance *balance = GDAM_BALANCE (user_data);
  GdamModel *model = GDAM_MODEL (user_data);
  GdamArg arg;
  if (model->source_id == 0)
    {
      balance->idle_id = 0;
      return FALSE;
    }
  arg.name = "volumes";
  arg.type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (&arg).length = 4;
  GDAM_ARG_DOUBLE_ARRAY (&arg).array = balance->volumes;
  gdam_channel_append_request (model->channel,
			       gdam_request_set_args (model->source_id,
						      1,
						      &arg,
						      NULL, NULL, NULL));
  balance->idle_id = 0;
  return FALSE;
}

static void indicate_balance_gains_changed(GdamBalance* balance)
{
  if (balance->idle_id == 0)
    balance->idle_id = g_idle_add (sent_balance_array, balance);
}

static guint balance_signals[LAST_SIGNAL];

/* Gain class function */

static int gdam_balance_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);
static void gdam_balance_class_init(GdamBalanceClass* klass);

static void gdam_balance_init(GdamBalance* balance)
{
  balance->volumes[0] = 1.0;
  balance->volumes[1] = 0.0;
  balance->volumes[2] = 0.0;
  balance->volumes[3] = 1.0;
}

GtkType gdam_balance_get_type()
{
  static GtkType balance_type = 0;
  if (!balance_type)
    {
      static const GtkTypeInfo balance_info = {
	"GdamBalance",
	sizeof (GdamBalance),
	sizeof (GdamBalanceClass),
	(GtkClassInitFunc) gdam_balance_class_init,
	(GtkObjectInitFunc) gdam_balance_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      balance_type = gtk_type_unique (GDAM_TYPE_FILTER, &balance_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return balance_type;
}

static int gdam_balance_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamBalance *balance = GDAM_BALANCE (filter);
  GdamRequest *request;
  GdamModel *model;
  model = GDAM_MODEL (filter);
  gtk_object_ref (GTK_OBJECT (balance));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasBalance",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					balance,
					2,
					"volumes",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					4,
					balance->volumes,
					"filtered",
					GDAM_ARG_TYPE_INT, new_source_id);
  return gdam_request_builder_append_create (builder, request);
}

void		gdam_balance_set_volume(GdamBalance*		balance,
					int                     input_channel,
					int                     output_channel,
                                        gdouble                 new_vol)
{
  int index = (input_channel == 0 ? 0 : 1) * 2
    + (output_channel == 0 ? 0 : 1);
  GtkObject *object;
  guint signal_id;
  switch (index)
    {
    case 0:
      signal_id = balance_signals[VOLUME_LL_CHANGED];
      break;
    case 1:
      signal_id = balance_signals[VOLUME_LR_CHANGED];
      break;
    case 2:
      signal_id = balance_signals[VOLUME_RL_CHANGED];
      break;
    case 3:
      signal_id = balance_signals[VOLUME_RR_CHANGED];
      break;
    default:
      {
	signal_id = 0;
	g_assert_not_reached ();
      }
    }
  balance->volumes[index] = new_vol;

  object = GTK_OBJECT (balance);
  if (gtk_signal_n_emissions (object, signal_id) != 0)
    return;
  gtk_signal_emit (object, signal_id);
}

static void gdam_balance_arg_set_volume_ll(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  balance->volumes[0] = GDAM_ARG_DOUBLE (arg);
  indicate_balance_gains_changed (balance);
}
static void gdam_balance_arg_get_volume_ll(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  GDAM_ARG_DOUBLE (arg) = balance->volumes[0];
}

static void gdam_balance_arg_set_volume_lr(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  balance->volumes[1] = GDAM_ARG_DOUBLE (arg);
  indicate_balance_gains_changed (balance);
}
static void gdam_balance_arg_get_volume_lr(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  GDAM_ARG_DOUBLE (arg) = balance->volumes[1];
}

static void gdam_balance_arg_set_volume_rl(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  balance->volumes[2] = GDAM_ARG_DOUBLE (arg);
  indicate_balance_gains_changed (balance);
}
static void gdam_balance_arg_get_volume_rl(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  GDAM_ARG_DOUBLE (arg) = balance->volumes[2];
}

static void gdam_balance_arg_set_volume_rr(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  balance->volumes[3] = GDAM_ARG_DOUBLE (arg);
  indicate_balance_gains_changed (balance);
}
static void gdam_balance_arg_get_volume_rr(GtkObject* object, GdamArg* arg)
{
  GdamBalance *balance = GDAM_BALANCE (object);
  GDAM_ARG_DOUBLE (arg) = balance->volumes[3];
}

static void gdam_balance_class_init(GdamBalanceClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  klass->volume_ll_changed = indicate_balance_gains_changed;
  klass->volume_lr_changed = indicate_balance_gains_changed;
  klass->volume_rl_changed = indicate_balance_gains_changed;
  klass->volume_rr_changed = indicate_balance_gains_changed;
  filter_class->append_start_request = gdam_balance_append_start_request;
  balance_signals[VOLUME_LL_CHANGED] =
    gtk_signal_new ("volume_ll_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBalanceClass, volume_ll_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  balance_signals[VOLUME_LR_CHANGED] =
    gtk_signal_new ("volume_lr_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBalanceClass, volume_lr_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  balance_signals[VOLUME_RL_CHANGED] =
    gtk_signal_new ("volume_rl_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBalanceClass, volume_rl_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  balance_signals[VOLUME_RR_CHANGED] =
    gtk_signal_new ("volume_rr_changed",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBalanceClass, volume_rr_changed),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, balance_signals, LAST_SIGNAL);
  gdam_arg_register (object_class,
		     "volume_ll",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_balance_arg_set_volume_ll,
		     gdam_balance_arg_get_volume_ll,
		     NULL, "-100,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "volume_lr",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_balance_arg_set_volume_lr,
		     gdam_balance_arg_get_volume_lr,
		     NULL, "-100,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "volume_rl",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_balance_arg_set_volume_rl,
		     gdam_balance_arg_get_volume_rl,
		     NULL, "-100,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "volume_rr",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_balance_arg_set_volume_rr,
		     gdam_balance_arg_get_volume_rr,
		     NULL, "-100,100", GDAM_ARG_VITAL);
}
