/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambender.h"
#include <gdam/component/gdampitch.h>
#include <gdam/common/gdamsample.h>
#include <gdam/common/gdamplugin.h>
#include <gtk/gtksignal.h>
#include <gdam/component/gdammodel.h>
#include <gdam/component/gdamfilter.h>
#include <gdam/component/gdamutil.h>
#include <gdam/common/gdam-protocol.h>
#include <math.h>

enum {
	SLOW_TO_STALL,
	LAST_SIGNAL
};

static guint bender_signals[LAST_SIGNAL];

/* Bender class function */
static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time);
/* Filter class function */
static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);
static int get_respeed_args (GdamBender *bender,
			     int        mode,
			     GdamArg    *args);

static void gdam_bender_init(GdamBender* bender)
{
  bender->backbuffer = 8.0;
  bender->beat_factor = 0.5;
  bender->rate_factor = 1.0;
  bender->pitch_factor = 1.0;
  bender->respeed_rate = 1.0;
  bender->bendable = 1;
  bender->mode = GDAM_BENDER_KEEP_TIME;
  bender->scale_bend_amount = 1.0;
}

static void gdam_bender_class_init(GdamBenderClass* klass);
GtkType gdam_bender_get_type()
{
  static GtkType bender_type = 0;
  if (!bender_type)
    {
      static const GtkTypeInfo bender_info = {
	"GdamBender",
	sizeof (GdamBender),
	sizeof (GdamBenderClass),
	(GtkClassInitFunc) gdam_bender_class_init,
	(GtkObjectInitFunc) gdam_bender_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      bender_type = gtk_type_unique (GDAM_TYPE_FILTER, &bender_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return bender_type;
}

static void made_request(gpointer data)
{
  GdamBender *bender = GDAM_BENDER (data);
  gdam_debug ("made request");
  if (bender->pending_request != NULL)
    {
      gdam_debug ("handling pending request");
      gdam_channel_append_request (GDAM_MODEL (bender)->channel,
				   bender->pending_request);
      bender->pending_request = NULL;
    }
  else
    {
      bender->making_request = FALSE;
    }
}

static void bender_real_slow_to_stall(GdamBender* bender, gdouble slow_time)
{
  g_warning ("deprecated method: slow_to_stall");
}

GdamBender* 	gdam_bender_new		()
{
  GdamBender *bender;
  bender = GDAM_BENDER (gtk_type_new (GDAM_TYPE_BENDER));
  return bender;
}

static gdouble
get_block_size_seconds (GdamBender *bender)
{
  gdouble block_size;
  gdouble bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  gdouble rate = bender->respeed_rate;
  bpm /= rate;  /* convert to input tempo */
  if (bpm > 1)
    block_size = bender->beat_factor * 60.0 / bpm;
  else
    block_size = bender->beat_factor;
  if (block_size < 0.01)
    block_size = 1.0;
  return block_size;
}

static char *
get_mode_string (GdamBender *bender)
{
  if (bender->mode == GDAM_BENDER_KEEP_TIME)
    return g_strdup ("keep_time");
  else if (bender->mode == GDAM_BENDER_SNAP_TIME)
    return g_strdup ("snap_time");
  else if (bender->mode == GDAM_BENDER_CONTINUOUS)
    return g_strdup ("continuous");
  else if (bender->mode == GDAM_BENDER_RETURN)
    return g_strdup ("return");
  else if (bender->mode == GDAM_BENDER_KEEP_IN_BEAT)
    return g_strdup ("keep_in_beat");
  else if (bender->mode == GDAM_BENDER_KEEP_INPUT_CONSTANT)
    return g_strdup ("keep_input_constant");
  else if (bender->mode == GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID)
    return g_strdup ("keep_input_constant_valid");
  g_warning ("unknown mode %d", bender->mode);
  return NULL;
}
 
static void 
got_args (int num_args, GdamArg *args, gpointer user_data)
{
  gdouble time;
  gdouble fraction;
  g_return_if_fail (num_args == 2);
  GDAM_BENDER (user_data)->request_pending = 0;
  gtk_object_unref (GTK_OBJECT (user_data));
  time = (gdouble) GDAM_ARG_INT (args) / 44100.0;
  gdam_arg_set (GTK_OBJECT (user_data), "playback_position", time);
  time = (gdouble) GDAM_ARG_INT (args + 1) / 44100.0;
  gdam_arg_set (GTK_OBJECT (user_data), "playback_offset", time);

  fraction = time / GDAM_BENDER (user_data)->backbuffer;
  fraction = 1.0 - fraction;
  if (fraction < 0.0)
    fraction = 0.0;
  if (fraction > 1.0)
    fraction = 1.0;
  gdam_arg_set (GTK_OBJECT (user_data), "playback_fraction", fraction);
}

static void 
didnt_get_args (gpointer user_data)
{
  g_warning ("error getting position from bender");
  gtk_object_unref (GTK_OBJECT (user_data));
}

/*  This function is called regularly to keep the position up to date
 */ 
static int 
position_timeout (GdamBender *bender)
{
  GdamRequest *request;
  const char *name_array[2] = { "playback_position", "playback_offset" };
  const char* const* arg_names = name_array;
  int num_args = 2;
  GdamRequestGotArgsFunc arg_func = got_args;

  /* if the source has stopped, or the user is no longer interested in position,
   * then cancel the timeout */
  if (GDAM_MODEL (bender)->source_id == 0 || !bender->follow_position)
    {
      bender->position_timeout_id = 0;
      return FALSE;
    }
  /* if we're already checking the position, leave the callout registered */
  if (bender->request_pending)
    return TRUE;
  bender->request_pending = TRUE;
  gtk_object_ref (GTK_OBJECT (bender));
  request = gdam_request_get_args (GDAM_MODEL (bender)->source_id,
				   num_args, 
				   arg_names,
				   arg_func, didnt_get_args, bender);
  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
  return TRUE;
}

/* Helper which determines the segment size in seconds of a bend effect, 
 * taking into account the beat info of the current song. */
static gdouble
get_rub_width 		(GdamBender 	*bender)
{
  float bpm;
  GdamModel *child = GDAM_FILTER (bender)->child;

  if (bender->beat_factor <= 0)
    return 1.0;

  if (child != NULL)
    bpm = gdam_beat_info_get_bpm (&child->beat_info, 0);
  else
    bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);

  if (bpm > 1)
    return bender->beat_factor * 60.0 / bpm;
  return bender->beat_factor;
}


/* create the change array for independent speed / pitch control, taking 
 * into account respeed_rate and pitch_factor.  returns a pointer to
 * a new int array (which should be freed by caller at some point)
 * and writes the number of pairs into num_out */
static int *
get_respeed_changes (GdamBender *bender, 
                     gdouble r_width, 
		     int *num_out)
{
  float r_length = 180.0;
  float r_f = bender->respeed_rate;
  float p_f = bender->pitch_factor;
  int num_points = 1;
  int *drag_data;
  int n;

  g_return_val_if_fail (r_width > 0, NULL);

  num_points = r_length / r_width;
  if (num_points % 2) num_points++;  /* ensure an even number of points */
  drag_data = g_new (int, num_points * 2);

  /* TODO: position jump to be offbeat so we don't double all 
   * hits which occur on the downbeats. */

  /* play 100% speed forever. */

  if (r_f == 1.0 && p_f == 1.0)
    {
      drag_data[0] = r_length * 44100;
      drag_data[1] = 1.0 * GdamFixed_ONE;
      num_points = 2;
    }
#if 0
  else
    {
      for (n = 0; n < num_points * 2; n += 4)
	{
	  /* This pair says to play at 100% speed for r_width seconds */
	  drag_data[n + 0] = r_width * 44100;
	  drag_data[n + 1] = p_f * 1.0 * GdamFixed_ONE;
	  /* The following pair, with a time value of zero, indicates a 
	   * jump in samples. */
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = -1.0 * (p_f * r_width - r_width) * -44100;
	}
    }
#endif

  /* If we're slowing down, make backwards jumps.
   * for instance, at 90% we jump back 0.1 second for every second played.
   * scaled to back 0.1111 for every 1.1111 we play.  */
  else if (r_f < 1.0)
    {
      for (n = 0; n < num_points * 2; n += 4)
	{
	  /* This pair says to play at 100% speed for r_width seconds */
	  drag_data[n + 0] = r_width * 44100 / p_f;
	  drag_data[n + 1] = p_f * 1.0 * GdamFixed_ONE;
	  /* The following pair, with a time value of zero, indicates a 
	   * jump in samples. */
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = p_f * (r_width - r_f * r_width) * -44100;
	}
    }
  /* If we're speeding up, make forward jumps.
   * for instance, at 110% we jump forward 0.1 second for every second played.
   * but, scaled to 0.91 + 0.091 = 1.0 beats of input time */
  else
    {
      gdouble play_size = r_width * 1.0 / r_f;  /* i.e. r_w * 0.91 */
      gdouble jump_size = r_width - play_size;
      for (n = 0; n < num_points * 2; n += 4)
	{
	  /* This pair says to play at 100% speed */
	  drag_data[n + 0] = play_size * 44100 / p_f;
	  drag_data[n + 1] = p_f * 1.0 * GdamFixed_ONE;
	  /* jump forward */
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = p_f * jump_size * 44100;
	}
    }

  *num_out = num_points;
  return drag_data;
}


static gboolean
bender_change_verify (int *change_array, int num_pairs)
{
  int i;
  for (i = 0; i < num_pairs; i++)
    {
      /* if first value is <= 0 then second value is a number of samples
       * to jump */
      if (change_array[2*i + 0] <= 0)
	{
	  /* ... but we always use exactly zero */
	  if (change_array[i + 0] < 0)
	    {
	      g_warning ("unexpected negative output time %d in change array pair %d", change_array[2*i + 0], i);
	      return FALSE;
	    }
	  if (change_array[2*i + 1] > 44100 * 4 ||
	      change_array[2*i + 1] < 44100 * -4)
	    {
	      g_warning ("suspiciously large jump %d in change array, pair %d", change_array[2*i + 1], i);
	      return FALSE;
	    }
	}
      /* first value is number of samples to output, second is output rate */
      else
	{
	  if (change_array[2*i + 1] > (int)(10.0 * GdamFixed_ONE) ||
	      change_array[2*i + 1] < (int)(-10.0 * GdamFixed_ONE))
	    {
	      g_warning ("suspiciously extreme rate %f in change array, pair %d", (gdouble)change_array[2*i + 1]/(gdouble)GdamFixed_ONE, i);
	      return FALSE;
	    }
	}
    }
  return TRUE;
}


static int gdam_bender_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamBender *bender = GDAM_BENDER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  char *mode_string = get_mode_string (bender);
  gdouble block_size;
  GdamArg args[14];
  int num_args;

  if (!mode_string)
    {
      g_warning ("unknown mode %d", bender->mode);
      mode_string = g_strdup ("keep time");
    }

  block_size = get_block_size_seconds (bender);
  args[0].name = "filtered";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 0) = new_source_id;
  args[1].name = "buffer_length";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = (int) (bender->backbuffer * 44100.0);
  args[2].name = "beat_length";
  args[2].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 2) = (int) (block_size * 44100.0);
  args[3].name = "change_mode";
  args[3].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 3) = mode_string;
  num_args = 4;

  /* If we are resampling, start at appropriate speed */
  if (bender->is_resampling)
    num_args += get_respeed_args (bender, 0, args + num_args);

  gtk_object_ref (GTK_OBJECT (bender));
  gdam_debug ("append_start_request: building bender from %d", new_source_id);
  model->source_id = gdam_channel_new_local_id (model->channel);

  request = gdam_request_create_source_v ("GdasBender",
					  model->source_id,
					  num_args,
					  args,
					  gdam_model_got_subsource,
					  gdam_model_didnt_get_subsource,
					  bender);

  if (bender->position_timeout_id == 0 && bender->follow_position)
    {
      bender->position_timeout_id
	= gtk_timeout_add (500,
			   (GtkFunction) position_timeout,
			   bender);
    }

  return gdam_request_builder_append_create (builder, request);
}

void gdam_bender_slow_to_stall(GdamBender* bender, float slow_time)
{
  GtkObject *this_object = GTK_OBJECT (bender);

  if (gtk_signal_n_emissions (this_object, bender_signals[SLOW_TO_STALL]))
    return;

  gtk_signal_emit (this_object, bender_signals[SLOW_TO_STALL], slow_time);
}

void		gdam_bender_set_mode	(GdamBender*		bender,
					 GDAM_BENDER_MODE	mode)
{
  GdamArg args[2];
  int num_args = 1;
  char *mode_string;

  g_return_if_fail (mode < GDAM_BENDER_NUM_MODES);

  bender->mode = mode;
  gdam_debug ("Bender set_mode to %d", mode);

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;

  mode_string = get_mode_string (bender);

  if (mode_string == NULL)
    return;

  GDAM_ARG_STRING (args) = mode_string;

  if (mode == GDAM_BENDER_KEEP_IN_BEAT)
    {
      gdouble block_size = get_block_size_seconds (bender);
      num_args = 2;
      args[1].name = "beat_length";
      args[1].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args+1) = (int)(44100.0 * block_size);
      GDAM_ARG_STRING (args) = g_strdup ("keep_in_beat");
    }

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       num_args, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


/* Posts a set_args request, avoiding congestion by metering out one
 * request at a time, and keeping only the most recent update. */
static void
bender_post_arg_requests 	(GdamBender 	*bender,
                                 GdamArg        *args,
				 int             num_args)
{
  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       num_args,
				       args,
				       made_request, made_request, bender);
      if (bender->making_request)
	{
	  if (bender->pending_request != NULL)
	    gdam_request_destroy (bender->pending_request);
	  bender->pending_request = request;
	}
      else
	{
	  bender->making_request = TRUE;
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }
  else
    {
      gdam_debug ("skipping bender arg requests, no source id");
    }
}

/* Set the bender's changes to a list of [time_in_samples_output,rate] pairs. */
void		gdam_bender_set_changes	(GdamBender*		bender,
					 int			num_points,
					 int*			changes)
{
  int *drag_data = changes;
  GdamArg args[2];

  /* TODO: fix this memory leak */
  /*memcpy(drag_data, changes, sizeof(int)*num_points*2); */

#if 0
  int n;
  gdam_debug ("Bender set_changes, %d points", num_points);
  fprintf (stderr, "Bender array:");
  for (n = 0; n < num_points * 2; n += 2)
    {
      fprintf (stderr, "  %f %f", (float) drag_data[n] / 44100,
	       (float) drag_data[n + 1] / GdamFixed_ONE);
    }
#endif

  g_return_if_fail (bender_change_verify (changes, num_points));

  /* TODO: copy this data? */
  args[0].name = "changes";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args).length = num_points ? num_points * 2 : 0;
  GDAM_ARG_INT_ARRAY (args).array = drag_data;

  bender_post_arg_requests (bender, args, 1);
}



/*  If we are in KEEP_TIME mode, where the underlying song continues to
 *  play while the output is bent, then snap back to where we should
 *  have been if no bending had occurred.
 */
void		gdam_bender_jump_to_mark(GdamBender*		bender)
{
  GdamArg args[2];
  int *changes;

  gdam_debug ("Bender jump_to_mark");

  if (bender->mode != GDAM_BENDER_KEEP_TIME &&
      bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT &&
      bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID &&
      bender->mode != GDAM_BENDER_KEEP_IN_BEAT)
    {
      return;
    }

  args[0].name = "jump_to_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = 0;


  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 0;
  GDAM_ARG_INT_ARRAY (args + 1).array = NULL;
#if 0
  changes = g_new (int, 2);
  changes[0] = 100 * 44100;
  changes[1] = bender->last_rate * GdamFixed_ONE;
  GDAM_ARG_INT_ARRAY (args + 1).length = 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = changes;
#endif

  bender_post_arg_requests (bender, args, 2);
}

/* Set the bender to play at 100% speed. */
void		gdam_bender_passthrough (GdamBender*		bender)
{
  GdamArg args[4];

  args[0].name = "spinning_between_position_markers";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 0) = 0;

  args[1].name = "sliding";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = 0;

  args[2].name = "changes";
  args[2].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 2).length = 0;
  GDAM_ARG_INT_ARRAY (args + 2).array = NULL;

  args[3].name = "jump_to_mark";
  args[3].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 3) = 0;

  bender_post_arg_requests (bender, args, 4);
}



void		gdam_bender_set_back_buffer	
					(GdamBender*		bender,
					 float			length)
{
  GdamArg args[1];

  gdam_debug ("Bender set_back_buffer");

  args[0].name = "back_buffer";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = length * 44100.0;

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
  bender->backbuffer = length;
}

void		gdam_bender_rewind	(GdamBender*		bender,
					 float			rewind_length)
{
  int num_points = 1;
  int *drag_data = g_new (int, num_points * 2);
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  GdamArg args[2];

  if (bender->bendable != 1)
    return;

  drag_data[0] = 1000.0 * 44100.0 / r_f;
  drag_data[1] = -1 * r_f * GdamFixed_ONE;

  g_return_if_fail (bender_change_verify (drag_data, num_points));

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}

static int delta_millis(GTimeVal* older, GTimeVal* younger)
{
  int millis = (older->tv_sec - younger->tv_sec) * 1000
    + (older->tv_usec - younger->tv_usec) / 1000;
  return millis;
}

static void
bender_seek (GdamBender *bender, gdouble seconds)
{
  int *drag_data;
  GdamArg arg;

  drag_data = g_new (int, 2);

  drag_data[0] = 0;
  drag_data[1] = seconds * 44100;
  arg.name = "changes";
  arg.type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (&arg).length = 2;
  GDAM_ARG_INT_ARRAY (&arg).array = drag_data;

  bender_post_arg_requests (bender, &arg, 1);
}

static void 
arg_seek_by_seconds (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (bender->bendable != 1)
    return;
  if (GDAM_ARG_DOUBLE (arg) != 0)
    bender_seek (bender, GDAM_ARG_DOUBLE (arg));
}

static void 
arg_seek_by_beats (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdouble bpm, seconds;

  if (bender->bendable != 1)
    return;

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    seconds = GDAM_ARG_DOUBLE (arg) * 60.0 / bpm;
  else
    seconds = GDAM_ARG_DOUBLE (arg);

  if (seconds != 0)
    bender_seek (bender, seconds);
}

static void gdam_bender_arg_jump_to_mark(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (bender->bendable != 1)
    return;
  if (GDAM_ARG_INT (arg) >= 1)
    gdam_bender_jump_to_mark (bender);
}


static void 
arg_set_default_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->default_position_marker = GDAM_ARG_INT (arg);
}

static void 
arg_get_default_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->default_position_marker;
}

static void 
arg_set_scale_note_idx (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdouble hertz = gdam_hz_from_midi_note_idx (GDAM_ARG_INT (arg));
  gdouble rate = hertz / 440;
  gdam_arg_set (object, "scale_bend_amount", rate);
}

static void 
arg_play_default_position_marker_by_note_idx (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdouble hertz = gdam_hz_from_midi_note_idx (GDAM_ARG_INT (arg));
  gdouble rate = hertz / 440;
  gdam_arg_set (object, "scale_bend_amount", rate);
  gdam_arg_set (object, "jump_to_position_marker", 
                bender->default_position_marker);
}

static void 
gdam_bender_arg_set_extend_spin_end (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_args[1];
  new_args[0].name = "extend_spin_endpoint";
  new_args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (new_args + 0) = 4410000;

  bender_post_arg_requests (bender, new_args, 1);
}

static void 
gdam_bender_arg_set_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[1];

  if (bender->bendable != 1)
    return;

  args[0].name = "set_position_marker";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = GDAM_ARG_INT (arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


static void 
gdam_bender_arg_jump_to_position_marker (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[1];

  if (bender->bendable != 1)
    return;

  args[0].name = "jump_to_position_marker";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = GDAM_ARG_INT (arg);

  bender->last_position_mark = GDAM_ARG_INT (arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}


static void 
gdam_bender_arg_jump_to_nearest_mark (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[3];
  gdouble block_size;
  int block_samples;

  if (bender->bendable != 1)
    return;

  g_return_if_fail (bender->beat_factor > 0);

  block_size = get_block_size_seconds (bender);
  block_samples = (int)(44100.0 * block_size);

  gdam_debug ("Bender jump_to_nearest_mark");

  if (bender->mode != GDAM_BENDER_KEEP_TIME &&
      bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT &&
      bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID &&
      bender->mode != GDAM_BENDER_KEEP_IN_BEAT)
    {
      return;
    }

  args[0].name = "jump_to_nearest_mark";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args) = block_samples;

  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = 0;
  GDAM_ARG_INT_ARRAY (args + 1).array = NULL;

  args[2].name = "change_mode";
  args[2].type = GDAM_ARG_TYPE_STRING;
  if (bender->mode == GDAM_BENDER_KEEP_TIME)
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_time");
  else if (bender->mode == GDAM_BENDER_KEEP_INPUT_CONSTANT)
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_input_constant");
  else if (bender->mode == GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID)
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_input_constant_valid");
  else
    GDAM_ARG_STRING (args+ 2) = g_strdup ("keep_in_beat");
  /* XXX: use get_mode_string ?? */

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       3,
				       args,
				       gdam_do_nothing,
				       gdam_do_nothing, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }

}

static void gdam_bender_arg_set_autotune(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->autotune = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_autotune(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->autotune;
}


/* used for various free bends where pitch is interactively controlled
 * by user.  bend_amount is used to set the pitch, the response to
 * bend_amount is determined by the bend mode */
#define GDAM_BEND_DELTA_CEILING 200  
static void 
bend_amount_changed (GdamBender *bender)
{
  GTimeVal real_time;
  int num_points = 10;
  float total_time = 0.2;
  int *change_array = g_new (int, num_points * 2);
  int n, delmil;
  float last_rate = bender->last_rate;
  float new_rate;
  float cur_rate = last_rate;
  float step_rate;
  float step_time;
  int pitch = (int) (bender->bend_amount * 8192.0 * 2);
  gboolean jump_to_mark = FALSE;
  //int flat_spot_radius = 800;  some joysticks have a huge flat spot
  int flat_spot_radius = 50;

  if (bender->bendable != 1)
    return;

  /* XXX: for now, out-of-range value means not bending. */
  if ((pitch < 0 || pitch > 16383))
    {
      if (!bender->is_bending)
	{
	  return;
	}
      gdam_debug ("HANDS OFF - no longer bending...");
      /*  TODO: re-enable this once it doesn't crash server. */
      jump_to_mark = TRUE;
      /*gdam_bender_jump_to_mark(bender); */
      new_rate = 1.0;
      bender->is_bending = FALSE;
    }
  else
    {
      bender->is_bending = TRUE;
      /* We fake (for now) raw pitch range: 0 <- 8192 -> 16383 */
      /* Pitch Wheel Mode:  0 <- 1 -> 2 */
      if (bender->bend_mode == 0)
	{
	  new_rate = (float) pitch / 8192;
	}
#if 0
      /* Pitch Wheel Mode:  -2 <- 0 |1| 0-> 2 */
      else if (bender->bend_mode == 1)
	{
	  if (pitch == 8192)
	    new_rate = 1.0;
	  else if (pitch < 8192)
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	  else
	    new_rate = (float) (8192 - pitch) * -2 / 8192.0;
	}
#endif
      /* Pitch Wheel Mode:  -2 <- 0 |1| 0-> 2 */
      else if (bender->bend_mode == 1)
	{
	  int fsr = flat_spot_radius;
	  if (pitch > 8192 - flat_spot_radius && 
	      pitch < 8192 + flat_spot_radius)
	    pitch = 8192;
	  if (pitch == 8192)
	    new_rate = 1.0;
	  else if (pitch < 8192)
	    new_rate = (float) ((8192 - fsr) - pitch) * -2 / (8192.0 - fsr);
	  else
	    new_rate = (float) ((8192 + fsr) - pitch) * -2 / (8192.0 - fsr);
	}
      /* Pitch Wheel Mode:  -2 <- 1 -> 2 */
      else if (bender->bend_mode == 2)
	{
	  if (pitch >= 8192)
	    new_rate = (float) pitch / 8192.0;
	  else if (pitch > 5000)
	    new_rate = (float) (pitch - 5000) * 1 / 3192.0;
	  else
	    new_rate = (float) (pitch - 5000) * 2 / 5000.0;
	}
      /*  Pitch Wheel Mode:  -4 <- -1 <- 0 |1| 0 -> 1 -> 4  */
      else if (bender->bend_mode == 3)
	{
	  int fsr = flat_spot_radius;
	  gdouble quarter = 4096.0 - (gdouble)fsr/2;
	  gdouble lhalf = 8192.0 - (gdouble)fsr;
	  gdouble uhalf = 8192.0 + (gdouble)fsr;
	  gdouble threequarter = 8192.0 + 4096.0 + (gdouble)fsr/2;
	  if (pitch > 8192 - flat_spot_radius && 
	      pitch < 8192 + flat_spot_radius)
	    pitch = 8192;
	  if (pitch == 8192)
	    new_rate = 1.0;
	  else if (pitch < quarter)
	    new_rate = (float) (-1 + (quarter - pitch) * -3.0 / quarter);
	  else if (pitch < 8192)
	    new_rate = (float) (lhalf - pitch) * -1.0 / quarter;
	  else if (pitch < threequarter)
	    new_rate = (float) (uhalf - pitch) * -1.0 / quarter;
	  else
	    new_rate = (float) (1 + (threequarter - pitch) * -3.0 / quarter);
	}
      /* TODO: Pitch Wheel Mode:  -3 <- -1 |1| 1 -> 0 */
      else
	{
	  g_warning ("Confused in bend.");
	  new_rate = 1.0;
	}
    }

  new_rate *= bender->scale_bend_amount;

  if (bender->autotune)
    new_rate = gdam_pitch_get_magic_rate (new_rate);

  step_rate = (new_rate - last_rate) / num_points;

  g_get_current_time (&real_time);
  delmil = delta_millis (&real_time, &bender->bend_adj_time);
  g_get_current_time (&bender->bend_adj_time);

  if (delmil < GDAM_BEND_DELTA_CEILING)
    {
      total_time = 0.001 * delmil;
    }
  if (delmil == 0)
    {
      total_time = 0.001;
    }

  step_time = total_time / num_points;
  bender->last_rate = new_rate;

  gdam_debug ("midi bend wheel changed!!!");
  gdam_debug ("delmil: %d", delmil);

  /* Old method, build a bend array */
#if 1
  {
    for (n = 0; n < num_points * 2; n += 2)
      {
	cur_rate += step_rate;
	change_array[n] = step_time * 44100;
	if (n == num_points * 2 - 2)
	  change_array[n] = 1000.0 * 44100;
	change_array[n + 1] = cur_rate * GdamFixed_ONE;
      }
    gdam_bender_set_changes (bender, num_points, change_array);
  }

#else

  /* New method: use server-side sliding capability */
  {
    GdamArg args[8];
    int i = 0;

    args[i].name = "slide_step_size_seconds";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = step_time;
    i++;

    args[i].name = "slide_rate_delta";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = step_rate;
    i++;

    args[i].name = "slide_rate_factor";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = 0;
    i++;

    args[i].name = "slide_num_steps";
    args[i].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (args + i) = num_points;
    i++;

    args[i].name = "slide_jump_size";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = 0;
    i++;

    args[i].name = "slide_current_rate";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = cur_rate;
    i++;

    args[i].name = "sliding";
    args[i].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (args + i) = 1;
    i++;

    /* In case we are already sliding with a large step size,
     * clear the changes now to force our new setting to take effect */
    args[i].name = "changes";
    args[i].type = GDAM_ARG_TYPE_INT_ARRAY;
    GDAM_ARG_INT_ARRAY (args + i).length = 0;
    GDAM_ARG_INT_ARRAY (args + i).array = NULL;
    i++;

    bender_post_arg_requests (bender, args, i);
  }
#endif

  if (jump_to_mark)
    gdam_bender_jump_to_mark (bender);
}

static void 
gdam_bender_arg_set_bend_amount(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->bend_amount = GDAM_ARG_DOUBLE (arg);
  bend_amount_changed (bender);
}

static void 
gdam_bender_arg_set_scale_bend_amount (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->scale_bend_amount = GDAM_ARG_DOUBLE (arg);
  if (bender->autotune_scale_bend_amount)
    bender->scale_bend_amount 
      = gdam_pitch_get_magic_rate (bender->scale_bend_amount);
  bend_amount_changed (bender);
}

static void
gdam_bender_arg_get_scale_bend_amount (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->scale_bend_amount;
}

static void 
gdam_bender_arg_set_autotune_scale_bend_amount (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->autotune_scale_bend_amount = GDAM_ARG_INT (arg);
  if (bender->autotune_scale_bend_amount)
    bender->scale_bend_amount
      = gdam_pitch_get_magic_rate (bender->scale_bend_amount);
}

static void
gdam_bender_arg_get_autotune_scale_bend_amount (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->autotune_scale_bend_amount;
}

/* Slide from one pitch to another over time.  factor is multiplied
 * with the beat_factor to determine length of slide.
 * TODO: slide logarithmically, hold/reset/slide on last pitch.   */
static void
gdam_bender_slide (GdamBender *bender, gdouble start_rate, gdouble
    		   end_rate, gdouble factor)
{
  float cur_rate = start_rate;
  float step_rate, step_time;
  float bpm;
  float total_time = 0.2;
  int num_points = 300;
  int n;

  if (bender->beat_factor <= 0)
    return;

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm > 1)
    total_time = factor * bender->beat_factor * 60.0 / bpm;
  else
    total_time = factor * bender->beat_factor;

  num_points = total_time / (0.001);  /* change speed 1000 times per sec */

  step_rate = (end_rate - start_rate) / num_points;
  step_time = total_time / num_points;
  bender->last_rate = end_rate;

  {
    GdamArg args[8];
    int i = 0;

    args[i].name = "slide_step_size_seconds";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = step_time;
    i++;

    args[i].name = "slide_rate_delta";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = step_rate;
    i++;

    args[i].name = "slide_rate_factor";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = 0;
    i++;

    args[i].name = "slide_num_steps";
    args[i].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (args + i) = num_points;
    i++;

    args[i].name = "slide_jump_size";
    args[i].type = GDAM_ARG_TYPE_DOUBLE;
    GDAM_ARG_DOUBLE (args + i) = 0;
    i++;

    if (bender->jump_to_mark_on_slide)
      {
	args[i].name = "jump_to_mark";
	args[i].type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT (args + i) = 0;
	i++;
      }

    // if (bender->enforce_slide_start_rate)
    if (TRUE)
      {
	args[i].name = "slide_current_rate";
	args[i].type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE (args + i) = start_rate;
	i++;
      }

    args[i].name = "sliding";
    args[i].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (args + i) = 1;
    i++;

    bender_post_arg_requests (bender, args, i);
  }
}

/* TODO: option to jump to true zero (stops at 0.1 to avoid clicking)
 * or hold or jump to 100% at end of stall */
static void 
gdam_bender_stall (GdamBender *bender, gdouble factor)
{
  /* factor * 4 is a hack because stalls always 'feel' too short compared
   * to other bend effects. */
  gdam_bender_slide (bender, 1.0, 0.1, factor * 4);
}


void		gdam_bender_rub		(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length)
{
  float r_length = 0;
  float r_width = 0;
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  int num_points = 1;
  int *drag_data;
  int n;
  GdamArg args[2];

  if (!rub_width)
    {
      float bpm;
      if (bender->beat_factor <= 0)
	return;
      bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
      if (bpm > 1)
	r_width = bender->beat_factor * 60.0 / bpm;
      else
	r_width = bender->beat_factor;
    }
  else if (rub_width < 0)
    {
      /* TODO: what to do? */
      return;
    }
  else
    r_width = rub_width;
  if (!rub_length)
    {
      r_length = 180.0;
    }
  else if (rub_length < 0)
    {
      return;
    }
  else
    r_length = rub_length;

  g_return_if_fail (r_width > 0);
  g_return_if_fail (r_length > 0);


  gdam_debug ("Bender rub, width: %f length: %f", r_width, r_length);


  /*  Rub rates weighted +/- rate */
  if (bender->rub_mode == 1)
    {
      float rate0 = r_f;
      float rate1 = (1.0 + (1.0 - r_f));
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  /*  Rub rates weighted +rate, normal */
  else if (bender->rub_mode == 2)
    {
      float rate0 = r_f;
      float rate1 = 1.0;
      float time0 = rate1 * 2 * r_width / (rate0 + rate1);
      float time1 = 2 * r_width - time0;

      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = time0 * 44100;
	  drag_data[n + 1] = -1.0 * rate0 * GdamFixed_ONE;
	  drag_data[n + 2] = time1 * 44100;
	  drag_data[n + 3] = rate1 * GdamFixed_ONE;
	}
    }
  else if (bender->rub_mode == 3)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 3);
      for (n = 0; n < num_points * 3; n += 6)
	{
	  drag_data[n + 0] = num_samples / 2;
	  drag_data[n + 1] = -2.0 * cur_rate;
	  drag_data[n + 2] = num_samples * 5 / 10;
	  drag_data[n + 3] = 0;
	  drag_data[n + 4] = num_samples;
	  drag_data[n + 5] = 1.0 * cur_rate;
	  cur_rate *= 0.9;
	}
    }
  else if (bender->rub_mode == 4)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  cur_rate *= 1.12;
	}
    }
  else if (bender->rub_mode == 5)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  cur_rate *= 1.12;
	  num_samples *= 0.75;
	}
    }
  else if (bender->rub_mode == 6)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = -1.0 * cur_rate;
	  if (n == 12)
	    {
	      cur_rate *= 1.12;
	    }
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate;
	  if (n == 4)
	    {
	      num_samples /= 3;
	    }
	  else if (n == 8)
	    {
	      num_samples *= 3;
	      num_samples /= 4;
	      cur_rate *= 1.12;
	    }
	  else if (n == 12)
	    {
	      cur_rate *= 1.12;
	    }
	  else if (n == 16)
	    {
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	      cur_rate /= 1.12;
	    }
	  else if (n == 20)
	    {
	      cur_rate *= 1.12;
	    }
	}
    }
  else if (bender->rub_mode == 7)
    {
      float cur_rate = r_f * GdamFixed_ONE;
      float num_samples = r_width * 44100 / r_f;
      float factor = 1.12;
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = num_samples;
	  drag_data[n + 1] = 1.0 * cur_rate;
	  drag_data[n + 2] = num_samples;
	  drag_data[n + 3] = 1.0 * cur_rate * 0;
	  cur_rate *= factor;
	  if ((n == 16) || (n == 32) || (n == 48))
	    {
	      factor = 1 / factor;
	    }
	  if (n == 32) 
	    {
	      num_samples /= 2;
	    }
	  else if (n == 40) 
	    {
	      num_samples *= 2;
	    }
	}
    }
  /* loops the recent past in reverse, jumping forward when we reach the
   * beginning of the sample. */
  else if (bender->rub_mode == 8)
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }
  /* spin via backwards jumps */
  else if (bender->rub_mode == 9)
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = 1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * -44100 / r_f;
	}
    }
  /*  Rub rates even.  */
  else
    {
      num_points = r_length / r_width;
      if (num_points % 2)
	num_points++;
      drag_data = g_new (int, num_points * 2);
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = r_width * 44100 / r_f;
	  drag_data[n + 3] = r_f * GdamFixed_ONE;
	}
    }

  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  g_return_if_fail (bender_change_verify (drag_data, num_points));
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}

/* Implement 'spin', on-the-fly looping of the recent past or near future
 * with variable speed and direction.   While spinning, the underlying
 * song does not advance in time.
 *
 * Rub Width:  size of looped segment, in seconds
 * Rub Length: how long the looping should last
 * Mode:   1:  spin recent past forward 
 *         2:  near future forward
 *         3:  recent past in reverse 
 *         4:  near future reverse
 */
void		gdam_bender_spin	(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length,
                                         int                    spin_mode)
{
  float r_length = 0;
  float r_width = 0;
  float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  int num_points = 1;
  int *drag_data;
  int n;
  GdamArg args[2];

  if (!rub_width)
    {
      r_width = get_rub_width (bender);
    }
  else if (rub_width < 0)
    {
      /* TODO: what to do? */
      g_warning ("negative rub width, aborting bender spin");
      return;
    }
  else
    r_width = rub_width;
  if (!rub_length)
    {
      r_length = 180.0;
    }
  else if (rub_length < 0)
    {
      return;
    }
  else
    r_length = rub_length;

  g_return_if_fail (r_width > 0);
  g_return_if_fail (r_length > 0);

  num_points = r_length / r_width;
  if (num_points % 2)
    num_points++;
  drag_data = g_new (int, num_points * 2);

  if (spin_mode == 1)
    {
      /* spin the recent past by making backwards jumps of r_width seconds
       * every r_width seconds. */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  /* The following pair, with a time value of zero, indicates a 
	   * jump of -r_width seconds. */
	  drag_data[n + 0] = 0;
	  drag_data[n + 1] = r_width * -44100 / r_f;
	  /* This pair says to play at 100% speed for r_width seconds */
	  drag_data[n + 2] = r_width * 44100 / r_f;
	  drag_data[n + 3] = 1.0 * r_f * GdamFixed_ONE;
	}
    }
  else if (spin_mode == 2)
    {
      /* spin the near future.  as above, but the jump comes after the forward
       * play. */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = 1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * -44100 / r_f;
	}
    }
  else if (spin_mode == 3)
    {
      /* spin the recent past in reverse by playing at -100% speed and making 
       * forward jumps.  */
      for (n = 0; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }
  else if (spin_mode == 4)
    {
      /* spin the near future in reverse by playing the near future at 100% 
       * speed, then repeatedly at -100% speed with forward jumps.  */
      /* TODO: initial forward jump rather than initial forward play?? */
      drag_data[0] = r_width * 44100 / r_f;
      drag_data[1] = 1.0 * r_f * GdamFixed_ONE;
      for (n = 2; n < num_points * 2; n += 4)
	{
	  drag_data[n + 0] = r_width * 44100 / r_f;
	  drag_data[n + 1] = -1.0 * r_f * GdamFixed_ONE;
	  drag_data[n + 2] = 0;
	  drag_data[n + 3] = r_width * 44100 / r_f;
	}
    }

  /* XXX: set component's mode arg */
  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("continuous");
  g_return_if_fail (bender_change_verify (drag_data, num_points));
  args[1].name = "changes";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;

  bender_post_arg_requests (bender, args, 2);
}

void gdam_bender_spin_between_markers (GdamBender* bender)
{
  GdamArg args[4];
  /* position marker 0 should have already been set to the starting point */
  args[0].name = "set_position_marker";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args+0) = 1;
  args[1].name = "spin_start_marker";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args+1) = 0;
  args[2].name = "spin_end_marker";
  args[2].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args+2) = 1;
  args[3].name = "spinning_between_position_markers";
  args[3].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args+3) = 1;

  bender_post_arg_requests (bender, args, 4);
}

#if 0
/* Implement respeed:  achieve a certain overall rate by making little
 * jumps every beat or half beat.  rub_width is the number of beats to
 * play between jumps.  rate_factor is the overall rate. */
void		gdam_bender_respeed	(GdamBender*		bender,
					 float			rub_width,
					 float			rub_length,
                                         int                    spin_mode)
{
  //float r_length = 180.0;  /* XXX: infinite length! */
  float r_length = 10.0;  /* XXX: infinite length! */
  float r_width = get_rub_width (bender);
  //float r_f = bender->rate_factor ? bender->rate_factor : 1.0;
  float r_f = bender->respeed_rate;
  int num_points = 1;
  int *drag_data;
  GdamArg args[3];
             
  g_return_if_fail (r_width > 0);

  drag_data = get_respeed_changes (bender, r_width, &num_points);

  /* XXX: set component's mode arg */
  args[0].name = "change_mode";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args) = g_strdup ("keep_time");
  g_return_if_fail (bender_change_verify (drag_data, num_points));
  args[1].name = "changes";
  //args[1].name = "replace_changes_at_jump";
  args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 1).length = num_points * 2;
  GDAM_ARG_INT_ARRAY (args + 1).array = drag_data;
  args[2].name = "rate";
  args[2].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 2) = (int)((1 << 16) * r_f) / (int)(1 << 16);

  bender_post_arg_requests (bender, args, 3);
}
#endif


/* Helper which fills in the args to accomplish respeed.  */
static int get_respeed_args (GdamBender *bender,
			     int        mode,
			     GdamArg    *args)
{
  float r_width = get_rub_width (bender);
  float r_f = bender->respeed_rate;
  float step_size = r_width;
  float jump_size = 0;
  int i = 0;

  if (r_f >= 1.0)
  {
    step_size = r_width * (1.0 / r_f);
    jump_size = r_width * (1.0 - (1.0 / r_f));
  }
  else if (r_f >= 0.5)
  {
    /* TODO: rather than playing past the r_width boundary, how about reversing
     * at r_width and making up the extra time in backspin? */
    step_size = r_width * (1.0 / r_f);
    jump_size = r_width * (1.0 - (1.0 / r_f));
  }
  /* TODO: sound gets twice as choppy when we drop below 0.5.... */
  else     /* r_f < 0.5 */
  {
    step_size = r_width;
    jump_size = -1.0 * r_width * (1.0 - r_f);
  }
             
  g_return_val_if_fail (r_width > 0, 0);

  args[i].name = "slide_step_size_seconds";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = step_size;
  i++;

  args[i].name = "slide_rate_delta";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = 0;
  i++;

  args[i].name = "slide_rate_factor";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = 0;
  i++;

  args[i].name = "slide_num_steps";
  args[i].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + i) = 10000000;  /* XXX: should have 'forever' */
  i++;

  args[i].name = "slide_jump_size";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = jump_size;
  i++;

  args[i].name = "slide_current_rate";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = 1.0;
  i++;

  args[i].name = "sliding";
  args[i].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + i) = 1;
  i++;

  return i;
}

/* Implement respeed using server-side facilities.  rub_width is the 
 * number of beats to play between jumps.  rate_factor is the overall rate. */
/* TODO: this implementation doesn't allow independent pitch control */
void		gdam_bender_respeed	(GdamBender*		bender,
                                         int                    mode)
{
  GdamArg args[8];
  int i = get_respeed_args (bender, mode, args);
  bender_post_arg_requests (bender, args, i);
}


static void gdam_bender_arg_set_resample_rate(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->pitch_factor = GDAM_ARG_DOUBLE (arg);
  if (bender->is_resampling)
    gdam_bender_respeed (bender, 0);
}
static void gdam_bender_arg_get_resample_rate(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->pitch_factor;
}

static void gdam_bender_arg_set_respeed_rate(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->respeed_rate = GDAM_ARG_DOUBLE (arg);
  gdam_arg_set (object, "overall_rate", bender->respeed_rate);
}
static void gdam_bender_arg_get_respeed_rate(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->respeed_rate;
}

static void gdam_bender_arg_set_is_resampling(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->is_resampling = GDAM_ARG_INT (arg);
}
static void gdam_bender_arg_get_is_resampling(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->is_resampling;
}

static void gdam_bender_arg_set_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->rate_factor = GDAM_ARG_DOUBLE (arg);
}
static void gdam_bender_arg_get_rate_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->rate_factor;
}

static void gdam_bender_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (bender->bendable)
    gdam_bender_set_mode (bender, GDAM_ARG_INT (arg));
}
static void gdam_bender_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (arg->type == GDAM_ARG_TYPE_INT)
    {
      GDAM_ARG_INT (arg) = bender->mode;
    }
  else
    {
      GDAM_BENDER_MODE mode = bender->mode;
      if (mode == GDAM_BENDER_KEEP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_time");
      else if (mode == GDAM_BENDER_KEEP_IN_BEAT)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_in_beat");
      else if (mode == GDAM_BENDER_KEEP_INPUT_CONSTANT)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_input_constant");
      else if (mode == GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID)
	GDAM_ARG_STRING (arg) = g_strdup ("keep_input_constant_valid");
      else if (mode == GDAM_BENDER_SNAP_TIME)
	GDAM_ARG_STRING (arg) = g_strdup ("snap_time");
      else if (mode == GDAM_BENDER_CONTINUOUS)
	GDAM_ARG_STRING (arg) = g_strdup ("continuous");
      else if (mode == GDAM_BENDER_RETURN)
	GDAM_ARG_STRING (arg) = g_strdup ("return");
    }
}

static void gdam_bender_arg_set_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (bender->bendable)
    bender->bend_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_get_bend_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->bend_mode;
}

static void gdam_bender_arg_set_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (bender->bendable)
    bender->rub_mode = GDAM_ARG_INT (arg);
}

static void gdam_bender_arg_set_rub_by_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  arg->name = "rub_mode";
  gdam_arg_set_object_arg (object, arg);
  arg->name = "rub_by_mode";
  gdam_bender_rub (bender, 0, 0);
}


static void gdam_bender_arg_set_passthrough (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  gdam_bender_passthrough (bender);
}

static void gdam_bender_arg_set_spin (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  gdam_bender_spin (bender, 0, 0, GDAM_ARG_INT (arg));
}


static void 
gdam_bender_arg_set_jump_to_spin_start (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_args[1];

  if (GDAM_ARG_INT (arg))
    {
      new_args[0].name = "jump_to_spin_start";
      new_args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (new_args+0) = 1;
      bender_post_arg_requests (bender, new_args, 1);
    }
}

static void 
gdam_bender_arg_set_spin_between_markers (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gboolean was_spinning;

  if (bender->bendable != 1)
    return;

  was_spinning = bender->spinning_between_position_markers;
  bender->spinning_between_position_markers = GDAM_ARG_INT (arg);
  if (bender->spinning_between_position_markers && !was_spinning)
    gdam_bender_spin_between_markers (bender);
  else if (!bender->spinning_between_position_markers && was_spinning)
    gdam_bender_passthrough (bender);
}

static void 
gdam_bender_arg_get_spin_between_markers (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->spinning_between_position_markers;
}

static void gdam_bender_arg_set_respeed (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  gdam_bender_respeed (bender, GDAM_ARG_INT (arg));
}

static void 
gdam_bender_arg_set_stall_by_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  gdam_bender_stall (bender, GDAM_ARG_DOUBLE (arg));
}

static void
arg_get_bent (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->bent;
}

static void
arg_set_bent (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamRequest *request;
  GdamArg args[3];
  int num_args = 2;

  bender->bent = GDAM_ARG_INT (arg);

  if (bender->bent)
    return;
  if (!GDAM_MODEL (bender)->source_id)
    return;

  gdam_bender_passthrough (bender);
}

static void 
gdam_bender_arg_set_dj_rewind_by_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);

  if (bender->bendable != 1)
    return;

  /* factor * 4 is a hack because stalls always 'feel' too short compared
   * to other bend effects. */
  gdam_bender_slide (bender, -4.0, -0.1, GDAM_ARG_DOUBLE (arg) * 4);
}

static void
arg_get_position_bending (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->position_bending;
}

static void
arg_set_position_bending (GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[3];
  int num_args = 2;

  if (GDAM_ARG_INT (arg) == 0)
    {
      bender->position_bending = FALSE;
      /* jump to mark */
      args[0].name = "jump_to_mark";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args) = 0;
      /* erase changes */
      args[1].name = "changes";
      args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
      GDAM_ARG_INT_ARRAY (args + 1).length = 0;
      GDAM_ARG_INT_ARRAY (args + 1).array = NULL;
    }
  else
    {
      GTimeVal timeval;
      int change_array[2];
      g_get_current_time (&timeval);
      bender->position_bending = TRUE;
      bender->last_bend_timeval = timeval;
      bender->last_bend_position = 0;
      bender->max_bend_timeval = timeval;
      bender->max_bend_position = 0;
      bender->min_bend_timeval = timeval;
      bender->min_bend_position = 0;
      bender->last_bend_rate = 0;

      /* set the mark to the current position */
      /* XXX: don't keep setting mode, just use jump_to_mark arg */
      args[0].name = "change_mode";
      args[0].type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (args + 0) = get_mode_string (bender);

      /* set to 0% speed */
      change_array[0] = 1000.0 * 44100;  /*  play this rate indefinitely */
      change_array[1] = 0.0 * GdamFixed_ONE;
      g_return_if_fail (bender_change_verify (change_array, 1));
      args[1].name = "changes";
      args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
      GDAM_ARG_INT_ARRAY (args + 1).length = 2;
      GDAM_ARG_INT_ARRAY (args + 1).array = change_array;
    }

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       2, args, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
    }
}
static void 
gdam_bender_arg_get_bend_position(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->last_bend_position;
}

/* map  -/+1  to  -/+ (beat_factor / 2) 
 *
 *                       mark (bend start)
 *   mark - (bf/2)       |
 *   |-------------------+----------------------| mark + (bf/2)
 *   -1                  0                      1
 *
 * Bend By Position:  divide position movement into monotonic
 * blocks by detecting change in direction.  We start a block
 * on the server when we detect the end of the block in the 
 * position input.  This means we are effectively one block behind
 * what the user has indicated.  But being behind lets us anticipate
 * the end of each block we play.
 *
 * */
#define MIN_MILLIS_PER_UPDATE 1.0
static void 
gdam_bender_arg_set_bend_position(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GTimeVal timeval;
  gdouble millis, beat_factor, bpm, rate, cur_server_position;
  gdouble last_bend_position = bender->last_bend_position;
  gdouble bend_position = GDAM_ARG_DOUBLE (arg);
  GdamArg args[3];
  gboolean have_args = FALSE;
  int cur_server_offset = 0;
  int change_array[4];
  int num_change_pairs = 1;
  gdouble direction = (bend_position - last_bend_position) > 0 ? 1.0 : -1.0;
  gdouble jump_to_last_mark = FALSE;

  g_get_current_time (&timeval);
  /* millis is negative ?? */
  millis = delta_millis (&bender->last_bend_timeval, &timeval);
  millis *= -1;

  if (!bender->position_bending)
    return;

  if (millis < MIN_MILLIS_PER_UPDATE)
    return;

  /* calculate overall rate since last position event */
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);
  if (bpm < 1)
    bpm = 60.0;
  beat_factor = bender->beat_factor * 1000 * 60.0 / bpm; /* millis per beat */

#define BY_POSITION_CHANGES 1
#if BY_POSITION_CHANGES
{
  gfloat block_time;   /* time between direction changes */
  gfloat block_size;   /* distance between direction changes */

  /* TODO: block_time is negative */

  /* If we're moving back, then this is the new min position */
  if (direction < 0 && bender->last_direction < 0)
    {
      bender->min_bend_position = bend_position;
      bender->min_bend_timeval = timeval;
    }
  /* If we're moving forward, then this is the new min position */
  else if (direction > 0 && bender->last_direction > 0)
    {
      bender->max_bend_position = bend_position;
      bender->max_bend_timeval = timeval;
    }

  /* If we've changed direction then play the block up until that change... */
  else if ((direction < 0 && bender->last_direction > 0)  ||
           (direction >= 0 && bender->last_direction < 0))
    {

      gdouble block_start, block_center, block_shift;
      block_size = bender->max_bend_position - bender->min_bend_position;

      /* TODO:  this block is a subset within the available range.  keep
       * the block size the same, but translate it so the center of the
       * block we're playing is mirrored over the midpoint of the available
       * range.  this makes users happy; because what they hear is a cycle
       * behind their input, they hear the song play forward when they
       * change direction to reverse.  so, they expect to hear the audio
       * which is at the _end_ of the available range if they scrub a
       * tiny block at the _reverse_ end of the available range.
       */
      block_center = (bender->max_bend_position + bender->min_bend_position) 
		     / 2.0;
      block_shift = 2.0 * (0.5 - block_center);

      /* scale to honor beat factor setting (converts block_size to millis) */
      block_size *= beat_factor / 2.0;

      /* Direction change from reverse to forward means we play the entire 
       * reverse block.   */
      if (direction >= 0) /* && bender->last_direction < 0 */
	{
	  block_time = delta_millis (&bender->max_bend_timeval, &bender->min_bend_timeval);
	  block_start = bender->max_bend_position;
	  if (block_time > -10)
	    block_time = -100;
	  rate = 1.0 * block_size / block_time;  /* millis / millis */
	  cur_server_offset = (int)(44.1 * (bender->max_bend_position + block_shift) * beat_factor / 2.0);
	}
      /* Direction change from forward to reverse means we play the entire 
       * forward block.  */
      else  /* (direction < 0 && bender->last_direction > 0) */
	{
	  block_time = delta_millis (&bender->min_bend_timeval, &bender->max_bend_timeval);
	  if (block_time > -10)
	    block_time = -100;
	  rate = -1.0 * block_size / block_time;
	  cur_server_offset = (int)(44.1 * (bender->min_bend_position + block_shift) * beat_factor / 2.0);
	}

      /* This change array will be applied after we jump to the position
       * marker.  First, set initial position by jumping relative
       * to the position marker.  Then, play from that position at the
       * proper speed. */
      change_array[0] = 0;  /* position jump */
      change_array[1] = cur_server_offset;
      change_array[2] = (int)(1000.0 * 44100);  /*  play this rate forever */
      change_array[3] = (int)(rate * GdamFixed_ONE);
      num_change_pairs = 2;

      //g_message ("bend_position %f rate %f (block size %f mils [max %f min %f bf %f mils] / block time %f mils) preceeded by jump %d %s", bend_position, rate, block_size, bender->max_bend_position, bender->min_bend_position, beat_factor, block_time, cur_server_offset, direction >= 0 ? "rev->fwd" : "fwd->rev" );

      if (direction >= 0)
        {
	  bender->max_bend_position = bend_position;
	  bender->max_bend_timeval = timeval;
        }
      else
        {
	  bender->min_bend_position = bend_position;
	  bender->min_bend_timeval = timeval;
	}

      g_return_if_fail (bender_change_verify (change_array, num_change_pairs));

      have_args = 1;
      jump_to_last_mark = TRUE;
    }
}
#else
  rate = (bend_position - last_bend_position) * beat_factor / 2.0;
  rate /= millis;  /* millis change in position over millis elapsed time */
  //g_message ("bend by position: rate %f from input %f millis %f", rate, bend_position, millis);

  /* use our knowledge of the server's state to estimate current position
   * and rate 
   * force our estimate of current offset to be correct (this prevents
   * position drift over time) */
  cur_server_position = last_bend_position * beat_factor / 2.0; /* last start */
  cur_server_position += millis * bender->last_bend_rate;
  cur_server_offset = cur_server_position * 44.1; /* convert to samples */

  change_array[0] = (int)(1000.0 * 44100);  /*  play this rate indefinitely */
  change_array[1] = (int)(rate * GdamFixed_ONE);
  num_change_pairs = 1;
  have_args = 1;
  g_return_if_fail (bender_change_verify (change_array, num_change_pairs));

  bender->last_bend_rate = rate;
#endif

  if (have_args)
    {
#if 0
      args[0].name = "jump_to_mark";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args) = cur_server_offset;
#endif
      g_return_if_fail (bender_change_verify (change_array, num_change_pairs));
      args[0].name = "jump_to_position_marker";
      args[0].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args) = bender->last_position_mark;
      args[1].name = "changes";
      args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
      GDAM_ARG_INT_ARRAY (args + 1).length = 2 * num_change_pairs;
      GDAM_ARG_INT_ARRAY (args + 1).array = change_array;

      if (GDAM_MODEL (bender)->source_id)
	{
	  GdamRequest *request;
	  request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
					   2, args, NULL, NULL, NULL);
	  gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
	}
    }

  bender->last_bend_timeval = timeval;
  bender->last_direction = direction;
  bender->last_bend_position = bend_position;
}

static void gdam_bender_arg_set_rewind_by_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_rewind (bender, -1);
}

static void gdam_bender_arg_get_rub_mode(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->rub_mode;
}

static void 
gdam_bender_arg_get_playback_position (GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = -1;
}

static void 
gdam_bender_arg_set_playback_position (GtkObject *object, GdamArg *arg)
{
}

static void 
gdam_bender_arg_get_playback_offset (GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = -1;
}

static void 
gdam_bender_arg_set_playback_offset (GtkObject *object, GdamArg *arg)
{
}

/* offset as a fraction of buffer_size,  0.0 - 1.0 */
static void 
gdam_bender_arg_get_playback_fraction (GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = 1.0;
}
static void 
gdam_bender_arg_set_playback_fraction (GtkObject *object, GdamArg *arg)
{
}

static void
do_slide (GdamBender *bender, int slide_from_idx)
{
  gdouble rate, start_rate;
  int from_idx = slide_from_idx;

  if (!from_idx)
    from_idx = 57;

  start_rate = gdam_hz_from_midi_note_idx (from_idx) / 440.0;
  if (bender->slide_delta_absolute)
    rate = gdam_hz_from_midi_note_idx (57 + bender->slide_half_steps) / 440.0;
  else
    rate = gdam_hz_from_midi_note_idx (from_idx + bender->slide_half_steps) / 440.0;

  gdam_bender_slide (bender, start_rate, rate, 1);
}

static void
arg_set_do_slide (GtkObject *object, GdamArg *arg)
{
  do_slide (GDAM_BENDER (object), 0);
}

static void
arg_set_slide_half_steps (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->slide_half_steps = GDAM_ARG_INT (arg);
}

static void 
arg_get_slide_half_steps (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->slide_half_steps;
}

static void
arg_set_follow_position (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->follow_position = GDAM_ARG_INT (arg);

  if (bender->position_timeout_id == 0 && bender->follow_position
      && GDAM_MODEL (bender)->source_id != 0)
    {
      bender->position_timeout_id
	= gtk_timeout_add (500,
			   (GtkFunction) position_timeout,
			   bender);
    }
}

static void
arg_get_follow_position (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->follow_position;
}

  /*  step_size_beats,rate_delta,rate_factor,num_steps,jump_size_beats,[current_rate], */
static void
arg_set_morph_array (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg args[8];
  int i = 0;
  gdouble bpm;
  gdouble *morph_array = GDAM_ARG_DOUBLE_ARRAY (arg).array;

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (bender)->beat_info, 0);

  if (bpm < 1.0)
    bpm = 60;

  g_return_if_fail (GDAM_ARG_DOUBLE_ARRAY (arg).length >= 5);

  args[i].name = "slide_step_size_seconds";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = morph_array[0] * 60 / bpm;
  i++;

  args[i].name = "slide_rate_delta";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = morph_array[1];
  i++;

  args[i].name = "slide_rate_factor";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = morph_array[2];
  i++;

  args[i].name = "slide_num_steps";
  args[i].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + i) = (int)morph_array[3];
  i++;

  args[i].name = "slide_jump_size";
  args[i].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (args + i) = morph_array[4] * 60 / bpm;
  i++;

  if (GDAM_ARG_DOUBLE_ARRAY (arg).length >= 6)
    {
      args[i].name = "slide_current_rate";
      args[i].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + i) = morph_array[5];
      i++;
    }

  args[i].name = "sliding";
  args[i].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + i) = 1;
  i++;

  /* In case we are already sliding with a large step size,
   * clear the changes now to force our new setting to take effect */
  args[i].name = "changes";
  args[i].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + i).length = 0;
  GDAM_ARG_INT_ARRAY (args + i).array = NULL;
  i++;

  bender_post_arg_requests (bender, args, i);
}


static void
arg_set_backbuffer (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  gdam_bender_set_back_buffer (bender, GDAM_ARG_DOUBLE (arg));
}

static void
arg_get_backbuffer (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->backbuffer;
}

static void
arg_set_slide_from_note_idx (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->sliding++;
  if (bender->sliding > 0)
    do_slide (bender, GDAM_ARG_INT (arg));
}


static void
arg_set_sliding (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  int old_count = bender->sliding;
  bender->sliding = GDAM_ARG_INT (arg);
  if (bender->sliding > 0 && bender->sliding > old_count)
    do_slide (bender, 0);
  else if (bender->sliding <= 0)
  {
    /* Assume we don't want to track negative values of active count */
    bender->sliding = 0;
    gdam_bender_passthrough (bender);
  }
}

static void 
arg_get_sliding (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->sliding;
}

static void 
gdam_bender_arg_get_bendable (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_INT (arg) = bender->bendable;
}

/* TODO: if not bendable, then ignore bend_amount, *_mark, etc */
static void 
gdam_bender_arg_set_bendable (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  bender->bendable = GDAM_ARG_INT (arg);
  if (bender->bendable != 1)
    gdam_bender_passthrough (bender);
}

static void 
gdam_bender_arg_set_is_bendable (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  if (GDAM_ARG_INT (arg) == 1 && !bender->bendable)
    gdam_arg_set (object, "bendable", 1);
  else if (GDAM_ARG_INT (arg) != 1 && bender->bendable)
    gdam_arg_set (object, "bendable", 0);
}

static void
gdam_bender_arg_get_overall_rate (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->respeed_rate;
}
static void
gdam_bender_arg_set_overall_rate (GtkObject *object, GdamArg *arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamModel *child = GDAM_FILTER (bender)->child;
  gdouble rate = bender->respeed_rate;
  if (child != NULL)
    {
      GdamBeatInfo *new_beat = gdam_beat_info_new ();
      GdamBeatInfo child_bi = child->beat_info;
      float new_bpm = gdam_beat_info_get_bpm (&child_bi, 0) * rate;
      float new_first_beat;
      new_first_beat = gdam_beat_info_get_first_beat (&child_bi, 0) / rate;
      gdam_beat_info_set_bpm (new_beat, new_bpm);
      gdam_beat_info_set_first_beat (new_beat, new_first_beat);
      gdam_model_set_beat_info (GDAM_MODEL (bender), new_beat);
      gdam_beat_info_destroy (new_beat);
    }
  if (bender->is_resampling)
    gdam_bender_respeed (bender, 0);
}

static void gdam_bender_arg_set_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_arg;
  if (bender->beat_factor == GDAM_ARG_DOUBLE (arg) / 8)
    return;
  new_arg.name = "beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = GDAM_ARG_DOUBLE (arg) / 8;
  gdam_arg_set_object_arg (object, &new_arg);
}
static void gdam_bender_arg_get_eighth_beat_factor(GtkObject* object, 
						   GdamArg*   arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = 8 * bender->beat_factor;
}

static void gdam_bender_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GdamArg new_arg;
  if (bender->beat_factor == GDAM_ARG_DOUBLE (arg))
    return;
  bender->beat_factor = GDAM_ARG_DOUBLE (arg);
  new_arg.name = "eighth_beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = bender->beat_factor * 8;
  gdam_arg_set_object_arg (object, &new_arg);

  if (GDAM_MODEL (bender)->source_id)
    {
      GdamRequest *request;
      GdamArg beat_length_arg;
      gdouble block_size = get_block_size_seconds (bender);
      beat_length_arg.name = "beat_length";
      beat_length_arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&beat_length_arg) = (int) (block_size * 44100.0);
      request = gdam_request_set_args (GDAM_MODEL (bender)->source_id,
				       1, &beat_length_arg, NULL, NULL, NULL);
      gdam_channel_append_request (GDAM_MODEL (bender)->channel, request);
      if (bender->is_resampling)
	gdam_bender_respeed (bender, 0);
    }
}

static void gdam_bender_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamBender *bender = GDAM_BENDER (object);
  GDAM_ARG_DOUBLE (arg) = bender->beat_factor;
}


/* TODO: generalize for all rate filters. */
static void              on_child_set_beat_info(GdamFilter*         filter,
						const GdamBeatInfo* child_bi)
{
  GdamBender *bender = GDAM_BENDER (filter);
  GdamBeatInfo *new_beat = gdam_beat_info_new ();
  gdouble rate = bender->respeed_rate;
  float new_bpm = gdam_beat_info_get_bpm (child_bi, 0) * rate;
  float new_first_beat = gdam_beat_info_get_first_beat (child_bi, 0) / rate;
  gdam_beat_info_set_bpm (new_beat, new_bpm);
  gdam_beat_info_set_first_beat (new_beat, new_first_beat);
  gdam_model_set_beat_info (GDAM_MODEL (bender), new_beat);
  if (bender->is_resampling)
    gdam_bender_respeed (bender, 0);
  gdam_beat_info_destroy (new_beat);
}

static void gdam_bender_class_init(GdamBenderClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  klass->slow_to_stall = bender_real_slow_to_stall;
  filter_class->append_start_request = gdam_bender_append_start_request;
  filter_class->on_child_set_beat_info = on_child_set_beat_info;
  gdam_arg_register (object_class,
		     "resample_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_resample_rate,
		     gdam_bender_arg_get_resample_rate,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "respeed_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_respeed_rate,
		     gdam_bender_arg_get_respeed_rate,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "is_resampling",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_is_resampling,
		     gdam_bender_arg_get_is_resampling,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "overall_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_overall_rate,
		     gdam_bender_arg_get_overall_rate,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "rate_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_rate_factor,
		     gdam_bender_arg_get_rate_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_beat_factor,
		     gdam_bender_arg_get_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "eighth_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_eighth_beat_factor,
		     gdam_bender_arg_get_eighth_beat_factor,
		     NULL, ",", GDAM_ARG_VITAL);
  /* Whether the filter should keep time, bend freely, snap to beat etc.  */
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_mode,
		     gdam_bender_arg_get_mode, NULL, NULL, GDAM_ARG_VITAL);
  /* The response curve to the 'bend amount' arg */
  gdam_arg_register (object_class,
		     "bend_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_bend_mode,
		     gdam_bender_arg_get_bend_mode,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bend_amount",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_bend_amount, NULL, NULL, ",", 0);
  gdam_arg_register (object_class,
		     "scale_bend_amount",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_scale_bend_amount, 
		     gdam_bender_arg_get_scale_bend_amount, 
		     NULL, ",", 0);
  gdam_arg_register (object_class,
		     "autotune_scale_bend_amount",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_autotune_scale_bend_amount, 
		     gdam_bender_arg_get_autotune_scale_bend_amount, 
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "jump_to_mark",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "jump_to_nearest_mark",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_nearest_mark, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "set_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_position_marker, 
		     NULL, NULL, "0,3", 0);
  gdam_arg_register (object_class,
		     "jump_to_position_marker",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_jump_to_position_marker, 
		     NULL, NULL, "0,3", 0);
  gdam_arg_register (object_class,
		     "default_position_marker",
		     GDAM_ARG_TYPE_INT,
		     arg_set_default_position_marker, 
		     arg_get_default_position_marker, 
		     NULL, "0,3", 0);
  gdam_arg_register (object_class,
		     "play_default_position_marker_by_note_idx",
		     GDAM_ARG_TYPE_INT,
		     arg_play_default_position_marker_by_note_idx,
		     NULL, NULL, "8,128", 0);
  gdam_arg_register (object_class,
		     "scale_note_idx",
		     GDAM_ARG_TYPE_INT,
		     arg_set_scale_note_idx,
		     NULL, NULL, "8,128", 0);
  gdam_arg_register (object_class,
		     "extend_spin_endpoint",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_extend_spin_end,
		     NULL, NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "autotune",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_autotune,
		     gdam_bender_arg_get_autotune,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "rub_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rub_mode,
		     gdam_bender_arg_get_rub_mode,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "rub_by_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rub_by_mode,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "rewind_by_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_rewind_by_mode,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "stall_by_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_stall_by_factor,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "dj_rewind_by_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_dj_rewind_by_factor,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "spin",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_spin,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "freeloop",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_spin_between_markers,
		     gdam_bender_arg_get_spin_between_markers,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "jump_to_spin_start",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_jump_to_spin_start,
		     NULL, NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "respeed",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_respeed,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "passthrough",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_passthrough,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "bend_position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_bend_position,
		     gdam_bender_arg_get_bend_position,
		     NULL, "-1.0,1.0", 0);
  gdam_arg_register (object_class,
		     "position_bending",
		     GDAM_ARG_TYPE_INT,
		     arg_set_position_bending,
		     arg_get_position_bending,
		     NULL, "0,1", 0);
  /* simply track arg value, and when it drops to zero, jump to mark */
  gdam_arg_register (object_class,
		     "bent",
		     GDAM_ARG_TYPE_INT,
		     arg_set_bent,
		     arg_get_bent,
		     NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "playback_position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_playback_position,
		     gdam_bender_arg_get_playback_position,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "playback_offset",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_playback_offset,
		     gdam_bender_arg_get_playback_offset,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "playback_fraction",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_bender_arg_set_playback_fraction,
		     gdam_bender_arg_get_playback_fraction,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "is_bendable",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_is_bendable,
		     NULL,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "do_slide",
		     GDAM_ARG_TYPE_INT,
		     arg_set_do_slide,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "bendable",
		     GDAM_ARG_TYPE_INT,
		     gdam_bender_arg_set_bendable,
		     gdam_bender_arg_get_bendable,
		     NULL, "0,1", 0);
  gdam_arg_register (object_class,
		     "slide_half_steps",
		     GDAM_ARG_TYPE_INT,
		     arg_set_slide_half_steps,
		     arg_get_slide_half_steps,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "sliding",
		     GDAM_ARG_TYPE_INT,
		     arg_set_sliding,
		     arg_get_sliding,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "slide_from_note_idx",
		     GDAM_ARG_TYPE_INT,
		     arg_set_slide_from_note_idx,
		     NULL,
		     NULL, "1,576", 0);
  gdam_arg_register (object_class,
		     "seek_by_seconds",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_seek_by_seconds , NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "seek_by_beats",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_seek_by_beats , NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "backbuffer",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_backbuffer, 
		     arg_get_backbuffer, 
		     NULL, "0,300", 1);
  gdam_arg_register (object_class,
		     "follow_position",
		     GDAM_ARG_TYPE_INT,
		     arg_set_follow_position, 
		     arg_get_follow_position, 
		     NULL, "boolean", 1);
  /*  step_size_beats,rate_delta,rate_factor,num_steps,jump_size_beats,[current_rate], */
  gdam_arg_register (object_class,
		     "morph_array",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     arg_set_morph_array, 
		     NULL, 
		     NULL, NULL, 0);


  bender_signals[SLOW_TO_STALL] =
    gtk_signal_new ("slow_to_stall",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamBenderClass, slow_to_stall),
		    gtk_marshal_NONE__DOUBLE,
		    GTK_TYPE_NONE, 1, GTK_TYPE_DOUBLE);
  gtk_object_class_add_signals (object_class, bender_signals, LAST_SIGNAL);
}

void gdam_bender_set_mark(GdamBender* bender){
	if (bender->mode != GDAM_BENDER_KEEP_TIME &&
	    bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT &&
	    bender->mode != GDAM_BENDER_KEEP_INPUT_CONSTANT_VALID &&
	    bender->mode != GDAM_BENDER_KEEP_IN_BEAT){
		return;
	}
	gdam_bender_set_mode(bender, bender->mode);
}

