/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamcoremididevice.h"

#if HAVE_COREMIDI_COREMIDI_H

#include <gdam/common/gdamplugin.h>
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdamplugin.h>
#include <gdam/common/gdamdebug.h>
#include <math.h>
#include <gtk/gtksignal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

/* Maximum number of errors we may encounter in a midi file
 * before stopping processing. */
#define MAX_ERRORS	20

static int   
gdam_core_midi_device_open_fd (GdamInputDeviceFile *input_device_file);
static void
gdam_core_midi_device_close_fd (GdamInputDeviceFile *input_device_file, int fd);
static gboolean   gdam_core_midi_device_open (GdamInputDevice *input_device);
static void       gdam_core_midi_device_close (GdamInputDevice *input_device);

static void 
gdam_core_midi_device_init(GdamCoreMidiDevice* core_midi_device)
{
}

/* XXX: make sure various core service devices don't derive from device
 * file class.... or maybe they should... */
static void
gdam_core_midi_device_class_init(GdamCoreMidiDeviceClass* core_midi_device_class)
{
  GdamInputDeviceFileClass* klass;
  GdamInputDeviceClass *device_class;
  device_class = GDAM_INPUT_DEVICE_CLASS (core_midi_device_class);
  klass = GDAM_INPUT_DEVICE_FILE_CLASS (core_midi_device_class);
  //device_file_class->process = gdam_midi_device_process;
  klass->open_fd = gdam_core_midi_device_open_fd;
  klass->close_fd = gdam_core_midi_device_close_fd;
  //device_class->open = gdam_core_midi_device_open;
  //device_class->close = gdam_core_midi_device_close;
}

GtkType gdam_core_midi_device_get_type()
{
  static GtkType core_midi_device_type = 0;
  if (!core_midi_device_type)
    {
      static const GtkTypeInfo core_midi_device_info = {
	"GdamCoreMidiDevice",
	sizeof (GdamCoreMidiDevice),
	sizeof (GdamCoreMidiDeviceClass),
	(GtkClassInitFunc) gdam_core_midi_device_class_init,
	(GtkObjectInitFunc) gdam_core_midi_device_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_MIDI_DEVICE;
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
      core_midi_device_type = gtk_type_unique (parent, &core_midi_device_info);
    }
  return core_midi_device_type;
}

/* Called whenever the input port has midi packets for us to process.  This
 * is called asynchronously by OS X so we have to be careful when
 * accessing data.  */
static void
core_midi_read_proc (const MIDIPacketList *pktlist, void *read_proc_ref_con, 
                     void *src_conn_ref_con)
{
  GdamCoreMidiDevice *core_midi_device;
  GdamInputDeviceFile *file_device;
  GdamInputDeviceFileClass *device_file_class;
  const MIDIPacket *my_packet = &pktlist->packet[0];
  int i;

  core_midi_device = (GdamCoreMidiDevice *)read_proc_ref_con;
  file_device = (GdamInputDeviceFile *)read_proc_ref_con;
  device_file_class = (GdamInputDeviceFileClass *) (GTK_OBJECT (file_device)->klass);

  /* TODO: parse midi packets and emit an event for each. */
  for (i = 0; i < pktlist->numPackets; i++)
    {
      write (core_midi_device->data_write_fd, 
             my_packet->data, 
             my_packet->length);
      //(*device_file_class->process) 
      //  (file_device, 
      //   g_memdup (my_packet->data, my_packet->length), 
      //   my_packet->length);
      my_packet = MIDIPacketNext (my_packet);
    }
}

#if 0
GdamCoreMidiDevice* gdam_core_midi_device_new ()
{
  GdamCoreMidiDevice *core_midi_device;
  core_midi_device = GDAM_CORE_MIDI_DEVICE (gtk_type_new (GDAM_TYPE_CORE_MIDI_DEVICE));

  return core_midi_device;
}
#endif

static gboolean   
gdam_core_midi_device_open (GdamInputDevice *input_device)
{
  GdamCoreMidiDevice *core_midi_device = GDAM_CORE_MIDI_DEVICE (input_device);
  ItemCount num_sources = MIDIGetNumberOfSources();
  int i;

  g_message ("core midi device new with %ld sources", num_sources);

  /* For every midi source (hardware device, software midi translator etc)
   * open it and attach a read func to parse incoming data.  */
  for (i = 0; i < num_sources; i++)
    {
      MIDIEndpointRef src = MIDIGetSource (i);
      /* identifies source of data */
      void *src_conn_ref_con = src;
      MIDIClientRef client;
      MIDIReadProc read_proc = core_midi_read_proc;
      void *ref_con = core_midi_device;
      MIDIPortRef in_port;

      g_message ("core midi device opening source %d", i);

      //CFStringRefkMIDIPropertyUniqueID
      //CFStringRefkMIDIPropertyDeviceID

      /* OS X will create a high priority thread to call our read proc
       * with incoming data */
      MIDIClientCreate (CFSTR("Gdam Default Client"), NULL, NULL, &client);
      MIDIInputPortCreate (client, CFSTR("Gdam Input Port"), read_proc, ref_con, &in_port);
      MIDIPortConnectSource (in_port, src, src_conn_ref_con);

      core_midi_device->in_port = in_port;
      core_midi_device->endpoint_ref = src;
    }
  return TRUE;
}

static void
gdam_core_midi_device_close_fd (GdamInputDeviceFile *input_device_file,
			        int		     fd)
{
  GdamInputDevice *device = GDAM_INPUT_DEVICE (input_device_file);
  gdam_core_midi_device_close (device);
}

static int   
gdam_core_midi_device_open_fd (GdamInputDeviceFile *input_device_file)
{
  int pipe_fds[2];
  GdamCoreMidiDevice *core_midi_device;
  core_midi_device  = GDAM_CORE_MIDI_DEVICE (input_device_file);

  if (pipe (pipe_fds) < 0)
    {
      g_warning ("pipe failed.");
      return -1;
    }
  core_midi_device->data_read_fd = pipe_fds[0];
  core_midi_device->data_write_fd = pipe_fds[1];

  if (!gdam_core_midi_device_open (GDAM_INPUT_DEVICE (input_device_file)))
    {
      g_warning ("core_midi_device_open failed");
      return -1;
    }

  fcntl (pipe_fds[1], F_SETFL, fcntl (pipe_fds[1], F_GETFL) | O_NONBLOCK);
  fcntl (pipe_fds[0], F_SETFL, fcntl (pipe_fds[0], F_GETFL) | O_NONBLOCK);
  return pipe_fds[0];
}

static void
gdam_core_midi_device_close (GdamInputDevice *input_device)
{
  GdamCoreMidiDevice *core_midi_device = GDAM_CORE_MIDI_DEVICE (input_device);
  if (core_midi_device->in_port)
    {
      MIDIPortDisconnectSource (core_midi_device->in_port,
				core_midi_device->endpoint_ref);
      MIDIPortDispose (core_midi_device->in_port);
    }
}


#endif /* HAVE_COREMIDI_COREMIDI_H */
