/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfader.h"
#include <stdlib.h>		/* for qsort. */
#include <gdam/common/gdamplugin.h>
#include <gtk/gtksignal.h>

static GtkObjectClass* parent_class = NULL;

enum {
	FADES_CHANGED,
	LAST_SIGNAL
};

static guint fader_signals[LAST_SIGNAL] = { 0 };

static int gdam_fader_append_start_request      (GdamFilter*       filter,
						 int               sub_id,
						 GdamRequestBuilder* builder);
static void         gdam_fader_destroy             (GtkObject*        fader);

static void gdam_fader_beat_changed (GdamFader *fader, gdouble bpm);




static gboolean gdam_fader_get_args(GdamFader* fader, 
                                int*       num_args,
				GdamArg**  args)
{
  gint64 *array;
  gdouble *phase_array;
  gdouble firstbeat, bpm, beatlen;
  GdamArg *rv;
  GdamBeatInfo *bi;
  int i;

  if (!fader->faded)
    {
      int period_length = 100;
      array = g_new (gint64, 4);
      rv  = g_new (GdamArg, 2);
      *num_args = 2;
      array[0] = 0;
      array[1] = GdamFixed_ONE;
      array[2] = period_length - 1;
      array[3] = GdamFixed_ONE;
      rv[0].name = g_strdup ("keypoints");
      rv[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
      GDAM_ARG_INT64_ARRAY (rv + 0).length = 4;
      GDAM_ARG_INT64_ARRAY (rv + 0).array = array;
      rv[1].name = g_strdup ("mode");
      rv[1].type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (rv + 1) = g_strdup ("loops");
      *args = rv;
      return TRUE;
    }

  array = g_new (gint64, fader->num_fades * 2);
  phase_array = g_new (gdouble, 2);
  rv  = g_new (GdamArg, 3);
  bi = (&GDAM_MODEL (fader)->beat_info);
  for (i = 0; i < fader->num_fades; i++)
    {
      array[i * 2 + 0] = fader->fade_array[i].time;
      array[i * 2 + 1] = fader->fade_array[i].volume;
    }
  gdam_debug ("component: gdamfader: num_fades=%d", fader->num_fades);
  g_return_val_if_fail (fader->num_fades > 1, FALSE);
  rv[0].name = g_strdup ("keypoints");
  rv[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
  GDAM_ARG_INT64_ARRAY (rv + 0).length = fader->num_fades * 2;
  GDAM_ARG_INT64_ARRAY (rv + 0).array = array;

  rv[1].name = g_strdup ("mode");
  rv[1].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (rv + 1) = g_strdup ("loops");


  firstbeat = gdam_beat_info_get_first_beat (bi, 0);
  bpm = gdam_beat_info_get_bpm (bi, 0);
  beatlen = 60.0 / bpm;

  *num_args = 2;

  if (bpm > 0.001)
    {
      *num_args += 1;
      phase_array[0] = firstbeat;
      phase_array[1] = beatlen;
      rv[2].name = g_strdup ("phase_sync");
      rv[2].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
      GDAM_ARG_DOUBLE_ARRAY (rv + 2).length = 2;
      GDAM_ARG_DOUBLE_ARRAY (rv + 2).array = phase_array;
    }

  *args = rv;
  return TRUE;
}

static void fades_changed(GdamFader* fader)
{
  GdamArg *args;
  int num_args;
  GdamRequest *request;
  guint source_id = GDAM_MODEL (fader)->source_id;
  gtk_signal_emit (GTK_OBJECT (fader), fader_signals[FADES_CHANGED]);
  if (source_id == 0)
    return;
  if (!gdam_fader_get_args (fader, &num_args, &args))
    return;
  request = gdam_request_set_args (source_id,
				   num_args, args, NULL, NULL, NULL);
  gdam_arg_free_array (num_args, args);
  gdam_channel_append_request (GDAM_MODEL (fader)->channel, request);
}

static int gdam_fader_append_start_request      (GdamFilter*       filter,
						 int               sub_id,
						 GdamRequestBuilder* builder)
{
  GdamArg *args;
  GdamModel *model = GDAM_MODEL (filter);
  int num_args;
  gdam_fader_get_args (GDAM_FADER (filter), &num_args, &args);
  num_args++;
  args = g_renew (GdamArg, args, num_args);
  args[num_args - 1].name = g_strdup ("filtered");
  args[num_args - 1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + num_args - 1) = sub_id;
  gtk_object_ref (GTK_OBJECT (filter));
  model->source_id = gdam_channel_new_local_id (model->channel);
  return gdam_request_builder_append_create (builder,
					     gdam_request_create_source_v
					     ("GdasFader", model->source_id,
					      num_args, args,
					      gdam_model_got_subsource,
					      gdam_model_didnt_get_subsource,
					      filter));
}

static void         gdam_fader_destroy             (GtkObject*        fader)
{
  g_free (GDAM_FADER (fader)->fade_array);
  (*parent_class->destroy) (fader);
}

static void on_beatinfo_set(GtkObject* obj, GdamBeatInfo *beat_info) 
{
  GdamFader* fader;
  GdamBeatInfo* bi;
  gdouble arg_bpm;
  fader = GDAM_FADER (obj);
  bi = &GDAM_MODEL (fader)->beat_info;
  arg_bpm = gdam_beat_info_get_bpm (beat_info, 0);

  if (arg_bpm != fader->last_bpm)
      gdam_fader_beat_changed (fader, arg_bpm);

  fader->last_bpm = arg_bpm;
}

static void gdam_fader_init(GdamFader* fader)
{
  fader->beat_factor = 1;
  fader->peak_width = 0.10;
  fader->base_width = 0.30;
  fader->depth = 1.0;
  gtk_signal_connect (GTK_OBJECT(fader),
		      "set_beat_info",
		      GTK_SIGNAL_FUNC(on_beatinfo_set),
		      (gpointer)fader);
  gdam_fader_set_passthrough (fader, 44100);
}

static void gdam_fader_class_init(GdamFaderClass* fader_class);

GtkType gdam_fader_get_type()
{
  static GtkType fader_type = 0;
  if (!fader_type)
    {
      static const GtkTypeInfo fader_info = {
	"GdamFader",
	sizeof (GdamFader),
	sizeof (GdamFaderClass),
	(GtkClassInitFunc) gdam_fader_class_init,
	(GtkObjectInitFunc) gdam_fader_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      fader_type = gtk_type_unique (parent, &fader_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return fader_type;
}

gboolean gdam_fader_set_fades(GdamFader* fader,
                              int num_fades,
			      const GdamFaderElement* fades)
{
  g_free (fader->fade_array);
  fader->fade_array = g_new (GdamFaderElement, num_fades);
  memcpy (fader->fade_array, fades, sizeof (GdamFaderElement) * num_fades);
  fader->num_fades = num_fades;
  fades_changed (fader);
  return TRUE;
}

void gdam_fader_remove_fade(GdamFader* fader,
                            int        which_fade)
{
  g_return_if_fail (which_fade >= 0);
  g_return_if_fail (which_fade < fader->num_fades - 1);
  memmove (fader->fade_array + which_fade,
	   fader->fade_array + which_fade + 1,
	   sizeof (GdamFaderElement) * (fader->num_fades - which_fade - 1));
  fader->num_fades--;
  fader->fade_array = g_renew (GdamFaderElement,
			       fader->fade_array, fader->num_fades);
  fades_changed (fader);
}

static int fader_element_comparator(const void* a, const void* b)
{
  int time_a = ((GdamFaderElement *) a)->time;
  int time_b = ((GdamFaderElement *) b)->time;
  if (time_a < time_b)
    return -1;
  if (time_a > time_b)
    return +1;
  return 0;
}

void gdam_fader_sort_fade_array(int               num_fades,
                                GdamFaderElement* array)
{
  qsort (array,
	 num_fades, sizeof (GdamFaderElement), fader_element_comparator);
}

void     gdam_fader_set_passthrough (GdamFader*              fader,
                                     int                     period_length)
{
  g_free (fader->fade_array);
  fader->fade_array = g_new (GdamFaderElement, 2);
  fader->fade_array[0].time = 0;
  fader->fade_array[0].volume = GdamFixed_ONE;
  fader->fade_array[1].time = period_length - 1;
  fader->fade_array[1].volume = GdamFixed_ONE;
  fader->num_fades = 2;
}

static void 
scale_to 		(int                    num_fades,
		      	 GdamFaderElement*      fade_array,
		      	 int 		      	new_length)
{
  int i;
  int old_length = fade_array[num_fades - 1].time;
  g_return_if_fail (new_length > 0);
  for (i = 1; i < num_fades; i++)
    {
      gint64 temp;
      temp = (gint64) fade_array[i].time;
      temp *= new_length;
      temp /= old_length;
      fade_array[i].time = (int) temp;
    }
}

/*  This adjusts the volume level of the dummy fade points
 *  so that they match up at the ends of the fade period,
 *  and don't disrupt the fade pattern. */
static void 
adjust_dummy_points  (GdamFader*              fader,
		      int                     num_fades,
		      GdamFaderElement*       fade_array)
{
  if (num_fades < 2)
    return;
  if (num_fades == 2)
    {
      /* If there are no user points, set both to full height. */
      fade_array[0].volume = GdamFixed_ONE;
      fade_array[1].volume = GdamFixed_ONE;
    }
  else if (num_fades == 3)
    {
      /* If there is one user point, set both to that height. */
      fade_array[0].volume = fade_array[1].volume;
      fade_array[2].volume = fade_array[1].volume;
    }
  else
    {
      int tl, vl, tr, vr;
      int del_t, del_v, mid_t, v_int;
      int fader_len;
      fader_len = fader->fade_array[fader->num_fades - 1].time + 1;
      tl = fade_array[num_fades - 2].time;
      vl = fade_array[num_fades - 2].volume;
      tr = fade_array[1].time + fader_len;
      vr = fade_array[1].volume;
      del_t = tr - tl;
      del_v = vr - vl;
      mid_t = fader_len - tl;
      v_int = del_t ? ((gint64) mid_t * del_v) / del_t + vl : vl + vr / 2;
      fade_array[0].volume = v_int;
      fade_array[num_fades - 1].volume = v_int;
    }
}

static int fe_compare(gconstpointer a, gconstpointer b){
	GdamFaderElement* fea = (GdamFaderElement*)a;
	GdamFaderElement* feb = (GdamFaderElement*)b;
	return (int)fea->time - (int)feb->time;
}

static void do_perf(GdamFader* fader)
{
  GdamFaderElement *new_array;
  GdamFaderElement *orig_array;
  float cur_peak_pos = 0;
  float peak_width = .1;	/* Actually, radius */
  float base_width = .2;	/* Actually, radius */
  int array_size;
  int num_fades;
  int num_peaks;
  int n;
  int old_num_fades = fader->num_fades;
  double length;
  double wavelength;

  if (!fader->perforated)
    {
      gdam_debug ("perf not active, aborting...");
      return;
    }

  peak_width = fader->peak_width ? fader->peak_width : .1;
  base_width = fader->base_width ? fader->base_width : .2;
  length = (double) fader->fade_array[old_num_fades - 1].time / 44100.0;
  num_peaks = 1;
  num_fades = (num_peaks * 4) + 2;	/* 2 dummy points */
  wavelength = length / num_peaks;
  peak_width *= wavelength;
  base_width *= wavelength;

  array_size = sizeof (GdamFaderElement) * (num_fades);
  orig_array = new_array = (GdamFaderElement *) alloca (array_size);

  for (n = 1; n < num_peaks * 4 + 1; n += 4)
    {
      double pos1 = cur_peak_pos - base_width;
      double pos2 = cur_peak_pos - peak_width;
      double pos3 = cur_peak_pos + peak_width;
      double pos4 = cur_peak_pos + base_width;
      double min_vol = 1.0 - fader->depth;
      if (pos1 < 0)
	pos1 += length;
      if (pos2 < 0)
	pos2 += length;
      if (pos3 > length)
	pos3 -= length;
      if (pos4 > length)
	pos4 -= length;
      if (min_vol < 0)
	min_vol = 0;

      new_array[n + 0].time = (int) (pos1 * 44100.0);
      new_array[n + 1].time = (int) (pos2 * 44100.0);
      new_array[n + 2].time = (int) (pos3 * 44100.0);
      new_array[n + 3].time = (int) (pos4 * 44100.0);
      new_array[n + 0].volume = GdamFixed_ONE * min_vol;
      new_array[n + 1].volume = GdamFixed_ONE;
      new_array[n + 2].volume = GdamFixed_ONE;
      new_array[n + 3].volume = GdamFixed_ONE * min_vol;
      cur_peak_pos += wavelength;
    }

  /* Add Dummy points */
  orig_array[0].time = 0;
  orig_array[0].volume = GdamFixed_ONE;
  orig_array[num_fades - 1].time = fader->fade_array[old_num_fades - 1].time;
  orig_array[num_fades - 1].volume = GdamFixed_ONE;

  qsort (new_array, num_fades, sizeof (GdamFaderElement), fe_compare);

  adjust_dummy_points (fader, num_fades, new_array);

  gdam_fader_set_fades (fader, num_fades, new_array);
}

/* Called when either beat_factor or bpm has changed to rescale array */
static void gdam_fader_beat_changed (GdamFader *fader, gdouble bpm)
{
  float secs;
  int new_length;
  int array_size;
  int num_fades;
  GdamFaderElement *new_array;
  GdamFaderElement *fade_array;

  if (bpm < 1.0)
    bpm = 60;

  num_fades = fader->num_fades;
  fade_array = fader->fade_array;
  array_size = sizeof (GdamFaderElement) * (num_fades);
  new_array = (GdamFaderElement *) alloca (array_size);
  memcpy (new_array, fade_array, array_size);

  secs = 60.0 / bpm;
  secs *= fader->beat_factor;
  new_length = (int) (secs * 44100.0);

  scale_to (num_fades, new_array, new_length);

  adjust_dummy_points (fader, num_fades, new_array);
  gdam_fader_set_fades (fader, num_fades, new_array);
}

static void gdam_fader_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  gdouble bpm;
  fader->beat_factor = GDAM_ARG_DOUBLE (arg);
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (fader)->beat_info, 0);
  gdam_fader_beat_changed (fader, bpm);
}

static void gdam_fader_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_DOUBLE (arg) = fader->beat_factor;
}

static void gdam_fader_arg_set_peak_width(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  fader->peak_width = GDAM_ARG_DOUBLE (arg);
  do_perf (fader);
}
static void gdam_fader_arg_get_peak_width(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_DOUBLE (arg) = fader->peak_width;
}

static void gdam_fader_arg_set_base_width(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  fader->base_width = GDAM_ARG_DOUBLE (arg);
  do_perf (fader);
}
static void gdam_fader_arg_get_base_width(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_DOUBLE (arg) = fader->base_width;
}

static void gdam_fader_arg_set_depth(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  fader->depth = GDAM_ARG_DOUBLE (arg);
  do_perf (fader);
}
static void gdam_fader_arg_get_depth(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_DOUBLE (arg) = fader->depth;
}

static void gdam_fader_arg_set_faded(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  fader->faded = GDAM_ARG_INT (arg);
  fades_changed (fader);
}
static void gdam_fader_arg_get_faded(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_INT (arg) = fader->faded;
}

static void gdam_fader_arg_set_fade_array(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  int i;
  gint64 *array = GDAM_ARG_INT64_ARRAY (arg).array;
  fader->num_fades = GDAM_ARG_INT64_ARRAY (arg).length / 2;
  fader->fade_array = g_renew (GdamFaderElement,
			       fader->fade_array, fader->num_fades);
  for (i = 0; i < fader->num_fades; i++)
    {
      fader->fade_array[i].time = array[i * 2 + 0];
      fader->fade_array[i].volume = array[i * 2 + 1];
      g_message ("fade_array[%d]:  tv=%lld,%lld",
		  i, array[i * 2 + 0], array[i * 2 + 1]);
    }
  fades_changed (fader);
}

static void gdam_fader_arg_get_fade_array(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  gint64 *array = g_new (gint64, fader->num_fades * 2);
  int i;
  for (i = 0; i < fader->num_fades; i++)
    {
      array[i * 2 + 0] = fader->fade_array[i].time;
      array[i * 2 + 1] = fader->fade_array[i].volume;
    }
  g_return_if_fail (fader->num_fades > 1);
  GDAM_ARG_INT64_ARRAY (arg).length = fader->num_fades * 2;
  GDAM_ARG_INT64_ARRAY (arg).array = array;
}

void gdam_fader_phase_synch(GdamFader* fader)
{
  GdamArg arg;
  gdouble *array = g_new (gdouble, 2);
  gdouble firstbeat, beatlen;
  GdamRequest *request;
  GdamBeatInfo *bi = (&GDAM_MODEL (fader)->beat_info);
  guint source_id = GDAM_MODEL (fader)->source_id;

  if (source_id == 0)
    return;

  firstbeat = gdam_beat_info_get_first_beat (bi, 0);
  beatlen = 60.0 / gdam_beat_info_get_bpm (bi, 0);

  /* TODO: if we have a beat factor of >= 2, we might scale the
   * bpm we use for phase synching.
   */
  array[0] = firstbeat;
  array[1] = beatlen;
  arg.name = g_strdup ("phase_sync");
  arg.type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (&arg).length = 2;
  GDAM_ARG_DOUBLE_ARRAY (&arg).array = array;
  request = gdam_request_set_args (source_id, 1, &arg, NULL, NULL, NULL);
  /* TODO: free array? (other occurences as well) */
  gdam_channel_append_request (GDAM_MODEL (fader)->channel, request);
}

static void gdam_fader_arg_set_perfed(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  fader->perforated = GDAM_ARG_INT (arg);
  do_perf (fader);
}

static void gdam_fader_arg_get_perfed(GtkObject* object, GdamArg* arg)
{
  GdamFader *fader = GDAM_FADER (object);
  GDAM_ARG_INT (arg) = fader->perforated;
}

static gboolean validate_fade_array(GtkObject* object,
                                    GdamArg* arg,
				    gpointer unused)
{
  gint64 *array = GDAM_ARG_INT64_ARRAY (arg).array;
  int length = GDAM_ARG_INT64_ARRAY (arg).length;
  gint64 last_time = 0;
  int i;

  for (i = 0; i < length / 2; i++)
    {
      gint64 time = array[2 * i], volume = array[2 * i + 1];
      if (time < last_time)
	return FALSE;
      last_time = time;
      if (volume < -4 * GdamFixed_ONE)
	return FALSE;
      if (volume > +4 * GdamFixed_ONE)
	return FALSE;
    }
  return TRUE;
}

static void gdam_fader_class_init(GdamFaderClass* fader_class)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (fader_class);
  GDAM_FILTER_CLASS (fader_class)->append_start_request
    = gdam_fader_append_start_request;
  object_class->destroy = gdam_fader_destroy;
  gdam_arg_register (object_class,
		     "faded",
		     GDAM_ARG_TYPE_INT,
		     gdam_fader_arg_set_faded,
		     gdam_fader_arg_get_faded,
		     NULL, "0,", GDAM_ARG_VITAL);
		     //NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "fade_array",
		     GDAM_ARG_TYPE_INT64_ARRAY,
		     gdam_fader_arg_set_fade_array,
		     gdam_fader_arg_get_fade_array,
		     validate_fade_array, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_fader_arg_set_beat_factor,
		     gdam_fader_arg_get_beat_factor,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "perforated",
		     GDAM_ARG_TYPE_INT,
		     gdam_fader_arg_set_perfed,
		     gdam_fader_arg_get_perfed,
		     NULL, "boolean", GDAM_ARG_VITAL);
  /*   These may not be vital....   */
  gdam_arg_register (object_class,
		     "peak_width",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_fader_arg_set_peak_width,
		     gdam_fader_arg_get_peak_width,
		     NULL, "0,1", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "base_width",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_fader_arg_set_base_width,
		     gdam_fader_arg_get_base_width,
		     NULL, "0,1", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "depth",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_fader_arg_set_depth,
		     gdam_fader_arg_get_depth, NULL, "0,1", GDAM_ARG_VITAL);
  fader_signals[FADES_CHANGED]
    = gtk_signal_new ("fades-changed",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamFaderClass,
					 fades_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, fader_signals, LAST_SIGNAL);
}
