/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamfeedbackfilter.h"
#include <gdam/common/gdamplugin.h>
#include <gtk/gtksignal.h>
#include <gdam/common/gdamsample.h>

static int
gdam_feedback_filter_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_feedback_filter_init(GdamFeedbackFilter* feedback_filter)
{
  feedback_filter->input_gain = 1.0;
}
static void gdam_feedback_filter_class_init(GdamModelClass* model_class);

static guint get_source_id(GdamFeedbackFilter* filter)
{
  if (filter == NULL)
    return 0;
  return GDAM_MODEL (filter)->source_id;
}

static void feedback_filter_source_success_callback(int new_id, gpointer data)
{
  GdamFeedbackFilter *filter = GDAM_FEEDBACK_FILTER (data);
  GdamArg args[2];
  int num_args = 0;

  gdam_model_set_source_id (GDAM_MODEL (filter), new_id);

  if (filter->dest != NULL)
    {
      int dest_id = GDAM_MODEL (filter->dest)->source_id;
      if (dest_id != 0)
	{
	  args[num_args].name = "feedback_dest";
	  args[num_args].type = GDAM_ARG_TYPE_INT;
	  GDAM_ARG_INT (args + num_args) = dest_id;
	  num_args++;
	}
    }

  if (filter->source != NULL)
    {
      int source_id = GDAM_MODEL (filter->source)->source_id;
      if (source_id != 0)
	{
	  args[num_args].name = "feedback_source";
	  args[num_args].type = GDAM_ARG_TYPE_INT;
	  GDAM_ARG_INT (args + num_args) = source_id;
	  num_args++;
	}
    }

  if (num_args > 0)
    {
      gdam_channel_append_request (GDAM_MODEL (filter)->channel,
				   gdam_request_set_args (new_id,
							  num_args,
							  args,
							  NULL, NULL, NULL));
    }

  gtk_object_unref (GTK_OBJECT (filter));
}


static int
gdam_feedback_filter_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamFeedbackFilter *feedback_filter = GDAM_FEEDBACK_FILTER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  gdouble gain, input_gain, delay;
  int dest_source_id, source_source_id;
  gain = feedback_filter->gain;
  input_gain = feedback_filter->input_gain;
  delay = feedback_filter->delay;
  gdam_debug ("append_start_request: building feedback from %d",
	      new_source_id);

  dest_source_id = get_source_id (feedback_filter->dest);
  source_source_id = get_source_id (feedback_filter->source);
  gtk_object_ref (GTK_OBJECT (filter));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasFeedbackFilter",
					model->source_id,
					feedback_filter_source_success_callback,
					gdam_model_didnt_get_subsource,
					feedback_filter,
					6,
					"gain",
					GDAM_ARG_TYPE_DOUBLE,
					gain,
					"input_gain",
					GDAM_ARG_TYPE_DOUBLE,
					input_gain,
					"delay",
					GDAM_ARG_TYPE_INT,
					(int) (44100.0 * delay),
					"filtered",
					GDAM_ARG_TYPE_INT,
					new_source_id,
					"feedback_dest",
					GDAM_ARG_TYPE_INT,
					dest_source_id,
					"feedback_source",
					GDAM_ARG_TYPE_INT, source_source_id);
  return gdam_request_builder_append_create (builder, request);
}

GtkType gdam_feedback_filter_get_type()
{
  static GtkType feedback_filter_type = 0;
  if (!feedback_filter_type)
    {
      static const GtkTypeInfo feedback_filter_info = {
	"GdamFeedbackFilter",
	sizeof (GdamFeedbackFilter),
	sizeof (GdamFeedbackFilterClass),
	(GtkClassInitFunc) gdam_feedback_filter_class_init,
	(GtkObjectInitFunc) gdam_feedback_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      feedback_filter_type = gtk_type_unique (parent, &feedback_filter_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return feedback_filter_type;
}

static void dest_destroyed(GdamFeedbackFilter* filter)
{
  filter->dest = NULL;
  filter->dest_destroy_id = 0;
}

void gdam_feedback_filter_set_dest  (GdamFeedbackFilter* filter,
                                     GdamFeedbackFilter* dest)
{
  if (filter->dest == dest)
    return;
  if (filter->dest != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (filter->dest),
			     filter->dest_destroy_id);
    }
  filter->dest = dest;
  if (filter->dest != NULL)
    {
      filter->dest_destroy_id =
	gtk_signal_connect_object (GTK_OBJECT (filter->dest), "destroy",
				   GTK_SIGNAL_FUNC (dest_destroyed),
				   (GtkObject *) filter);
    }
  else
    {
      filter->dest_destroy_id = 0;
    }

  if (dest)
    gdam_feedback_filter_set_source (dest, filter);
}

static void source_destroyed(GdamFeedbackFilter* filter)
{
  filter->source = NULL;
  filter->source_destroy_id = 0;
}

void gdam_feedback_filter_set_source(GdamFeedbackFilter* filter,
                                     GdamFeedbackFilter* source)
{
  if (filter->source == source)
    return;

  if (filter->source != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (filter->source),
			     filter->source_destroy_id);
    }
  filter->source = source;
  if (filter->source != NULL)
    {
      filter->source_destroy_id =
	gtk_signal_connect_object (GTK_OBJECT (filter->source), "destroy",
				   GTK_SIGNAL_FUNC (source_destroyed),
				   (GtkObject *) filter);
    }
  else
    {
      filter->source_destroy_id = 0;
    }

  if (source)
    gdam_feedback_filter_set_dest (source, filter);
}

static void gdam_feedback_filter_set_input_gain (GdamFeedbackFilter* filter,
						 gdouble             input_gain)
{
  GdamModel *model = GDAM_MODEL (filter);
  if (filter->input_gain == input_gain)
    return;
  filter->input_gain = input_gain;
  if (model->source_id)
    {
      GdamArg arg;
      arg.name = "input_gain";
      arg.type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (&arg) = input_gain;
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args (model->source_id,
							  1, &arg,
							  NULL, NULL, NULL));
    }
}


static void gdam_feedback_filter_set_gain  (GdamFeedbackFilter* filter,
                                     	    gdouble             gain)
{
  GdamModel *model = GDAM_MODEL (filter);
  if (filter->gain == gain)
    return;
  filter->gain = gain;
  if (model->source_id)
    {
      GdamArg arg;
      arg.name = "gain";
      arg.type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (&arg) = gain;
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args (model->source_id,
							  1, &arg,
							  NULL, NULL, NULL));
    }
}

static void internal_gdam_feedback_filter_set_delay (GdamFeedbackFilter* filter,
                                     	    gdouble             delay)
{
  GdamModel *model = GDAM_MODEL (filter);
  int int_delay = (int) (delay * 44100.0);
  if (filter->delay == delay)
    return;
  filter->delay = delay;
  if (model->source_id)
    {
      GdamArg arg;
      arg.name = "delay";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = int_delay;
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args (model->source_id,
							  1, &arg,
							  NULL, NULL, NULL));
    }
}

void gdam_feedback_filter_set_delay (GdamFeedbackFilter* filter,
                                     gdouble             delay)
{
  GdamArg arg;
  arg.name = "delay";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = delay;
  gdam_arg_set_object_arg (GTK_OBJECT (filter), &arg);
}

static void gdam_feedback_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  float bpm;
  float newval;
  feedback->beat_factor = GDAM_ARG_DOUBLE (arg);
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (feedback)->beat_info, 0);
  /*  Consult infomanager for bpm? */
  if (bpm <= 1)
    {
      g_message ("aborting delay_beat");
      return;
    }
  newval = GDAM_ARG_DOUBLE (arg) * 60.0 / bpm;
  gdam_feedback_filter_set_delay (feedback, newval);
}

static void gdam_feedback_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  /* TODO: should this calculate real, current value? */
  GDAM_ARG_DOUBLE (arg) = feedback->beat_factor;
}

static void gdam_feedback_arg_set_delay(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  internal_gdam_feedback_filter_set_delay (feedback, GDAM_ARG_DOUBLE (arg));
}
static void gdam_feedback_arg_get_delay(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = feedback->delay;
}

static void gdam_feedback_arg_set_feedback_gain(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  gdam_feedback_filter_set_gain (feedback, GDAM_ARG_DOUBLE (arg));
}
static void gdam_feedback_arg_get_feedback_gain(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = feedback->gain;
}

static void gdam_feedback_arg_set_input_gain(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  gdam_feedback_filter_set_input_gain (feedback, GDAM_ARG_DOUBLE (arg));
}
static void gdam_feedback_arg_get_input_gain(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilter *feedback = GDAM_FEEDBACK_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = feedback->input_gain;
}

static void gdam_feedback_filter_class_init(GdamModelClass* model_class)
{
  GdamFilterClass *filter_class = GDAM_FILTER_CLASS (model_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (model_class);
  filter_class->append_start_request
    = gdam_feedback_filter_append_start_request;
  gdam_arg_register (object_class,
		     "feedback_gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_feedback_gain,
		     gdam_feedback_arg_get_feedback_gain,
		     NULL, "-1.5,1.5", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "input_gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_input_gain,
		     gdam_feedback_arg_get_input_gain,
		     NULL, "-1.5,1.5", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "delay",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_delay,
		     gdam_feedback_arg_get_delay,
		     NULL, "0,3600", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_beat_factor,
		     gdam_feedback_arg_get_beat_factor,
		     NULL, "0,3600", GDAM_ARG_VITAL);
}
