/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamflange.h"
#include <gdam/common/gdamplugin.h>
static void gdam_flange_factors_changed(GdamFlange* flange);

static void gdam_flange_init(GdamFlange* flange)
{
  GdamFirFilter *fir_filter = GDAM_FIR_FILTER (flange);
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (flange);
  flange->hertz = 1.0;
  flange->buffer_id = 1;
  flange->chorus = 1;
  flange->min_factor = 0.1;
  flange->max_factor = 3.1;
  fir_filter->grid_unit = 16.0;
  //g_free (fir_filter->elements);
  //fir_filter->elements = NULL;
  //fir_filter->num_elements = 0;
  fir_filter->echoed = 1;
  echo_base->ignore_passthrough = TRUE;
  echo_base->ignore_reverb = TRUE;
}

void gdam_flange_set_min_factor(GdamFlange* flange, double factor)
{
  GdamArg arg;
  arg.name = "min_factor";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = factor;
  gdam_arg_set_object_arg (GTK_OBJECT (flange), &arg);
}

void gdam_flange_set_max_factor(GdamFlange* flange, double factor)
{
  GdamArg arg;
  arg.name = "max_factor";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = factor;
  gdam_arg_set_object_arg (GTK_OBJECT (flange), &arg);
}

static void gdam_flange_arg_set_min_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble center_factor = (flange->min_factor + flange->max_factor) / 2.0;
  gdouble max_factor;
  flange->min_factor = GDAM_ARG_DOUBLE (arg);
  if (flange->min_factor > center_factor)
    flange->min_factor = center_factor;
  max_factor = center_factor + (center_factor - flange->min_factor);
  if (max_factor != flange->max_factor)
    gdam_flange_set_max_factor (flange, max_factor);
  /*flange->max_factor = max_factor; */
  gdam_flange_factors_changed (flange);
}
static void gdam_flange_arg_get_min_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  GDAM_ARG_DOUBLE (arg) = flange->min_factor;
}

static void gdam_flange_arg_set_mid_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble old_center_factor;
  gdouble delta;
  gdouble min_factor, max_factor;

  old_center_factor = (flange->min_factor + flange->max_factor) / 2.0;
  delta = GDAM_ARG_DOUBLE (arg) - old_center_factor;
  min_factor = flange->min_factor + delta;
  max_factor = flange->max_factor + delta;
  flange->min_factor = min_factor;
  flange->max_factor = max_factor;
  gdam_flange_set_min_factor (flange, min_factor);
  gdam_flange_set_max_factor (flange, max_factor);
  gdam_flange_factors_changed (flange);
}
static void gdam_flange_arg_get_mid_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  GDAM_ARG_DOUBLE (arg) = (flange->min_factor + flange->max_factor) / 2.0;
}

static void gdam_flange_arg_set_max_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble center_factor = (flange->min_factor + flange->max_factor) / 2.0;
  gdouble min_factor;
  flange->max_factor = GDAM_ARG_DOUBLE (arg);
  if (flange->max_factor < center_factor)
    flange->max_factor = center_factor;
  min_factor = center_factor - (flange->max_factor - center_factor);
  if (min_factor != flange->min_factor)
    gdam_flange_set_min_factor (flange, min_factor);
  /*flange->min_factor = min_factor; */
  gdam_flange_factors_changed (flange);
}
static void gdam_flange_arg_get_max_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  GDAM_ARG_DOUBLE (arg) = flange->max_factor;
}

static void wave_success(int id, gpointer data){
	GdamFlange* flange = GDAM_FLANGE(data);
	flange->wavetable_id = id;
	gtk_object_unref(GTK_OBJECT(flange));
}

static void wave_failure(gpointer data){
	GdamFlange* flange = GDAM_FLANGE(data);
	gtk_object_unref(GTK_OBJECT(flange));
}

static int
make_wavetable_request(GdamFlange* flange, GdamRequestBuilder* builder)
{
  GdamRequest *request;
  int local_id;
  gtk_object_ref (GTK_OBJECT (flange));
  local_id = gdam_channel_new_local_id (GDAM_MODEL (flange)->channel);
  request = gdam_request_create_source
    ("GdasWavetable",
     local_id,
     wave_success,
     wave_failure,
     flange,
     2,
     "buffer_id", GDAM_ARG_TYPE_INT, flange->buffer_id, 
     "hertz", GDAM_ARG_TYPE_DOUBLE, flange->hertz);
  flange->wavetable_id = local_id;
  return gdam_request_builder_append_create (builder, request);
}

static int gdam_flange_append_start_request(GdamFilter* filter,
				            int new_source_id,
				            GdamRequestBuilder* builder)
{
  GdamEchoBase *echo_base = GDAM_ECHO_BASE (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamRequest *request;
  int wavetable_id;
  gdouble min_max[2];
  GdamFlange *flange = GDAM_FLANGE (filter);
  gtk_object_ref (GTK_OBJECT (filter));
  min_max[0] = flange->min_factor;
  min_max[1] = flange->max_factor;
  wavetable_id = make_wavetable_request (flange, builder);
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasFlange",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					flange,
					6,
					"filtered",
					GDAM_ARG_TYPE_INT,
					new_source_id,
					"modulator",
					GDAM_ARG_TYPE_INT,
					wavetable_id,
					"echoes",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					echo_base->num_echoes * 2,
					echo_base->echoes,
					"min_max",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					2,
					min_max,
					"effect",
					GDAM_ARG_TYPE_DOUBLE,
					flange->effect,
					"chorus",
					GDAM_ARG_TYPE_INT, flange->chorus);
  return gdam_request_builder_append_create (builder, request);
}

static void reset_wavetable(GdamFlange* flange)
{
  GdamModel *model = (GdamModel *) flange;
  if (model->source_id != 0)
    {
      GdamRequestBuilder *builder;
      GdamArg arg;
      GdamRequest *package;
      int wavetable_id;
      builder = gdam_request_builder_new ();
      wavetable_id = make_wavetable_request (flange, builder);
      arg.name = "modulator";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = wavetable_id;
      gdam_request_builder_append_generic
	(builder,
	 gdam_request_set_args (model->source_id, 1, &arg, NULL, NULL, NULL));
      package = gdam_request_builder_package (builder);
      gdam_channel_append_request (model->channel, package);
      gdam_request_builder_destroy (builder);
    }
}
static void gdam_flange_arg_set_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  int buffer_id = GDAM_ARG_INT (arg);
  if (buffer_id != (int) flange->buffer_id)
    {
      flange->buffer_id = buffer_id;
      reset_wavetable (flange);
    }
}
static void gdam_flange_arg_get_buffer_id(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAM_FLANGE (object)->buffer_id;
}


static void gdam_flange_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble bpm, newval;
  GdamArg new_arg;
  flange->beat_factor = GDAM_ARG_DOUBLE (arg);
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (flange)->beat_info, 0);
  if (bpm < 1)
    bpm = 60.0;
  newval = flange->beat_factor * bpm / 60.0;
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  new_arg.name = "hertz";
  GDAM_ARG_DOUBLE (&new_arg) = newval;
  gdam_arg_set_object_arg (GTK_OBJECT (flange), &new_arg);
}

static void gdam_flange_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAM_FLANGE (object)->beat_factor;
}

static void gdam_flange_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble hertz = GDAM_ARG_DOUBLE (arg);
  if (hertz != flange->hertz)
    {
      GdamModel *model = GDAM_MODEL (flange);
      flange->hertz = hertz;
      if (GDAM_MODEL (flange)->source_id != 0)
	{
	  GdamArg arg;
	  arg.name = "hertz";
	  arg.type = GDAM_ARG_TYPE_DOUBLE;
	  GDAM_ARG_DOUBLE (&arg) = hertz;
	  gdam_channel_append_request (model->channel,
				       gdam_request_set_args
				       (flange->wavetable_id, 1, &arg, NULL,
					NULL, NULL));

	}
      reset_wavetable (flange);
    }
}

static void gdam_flange_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAM_FLANGE (object)->hertz;
}

static void gdam_flange_arg_set_effect(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gdouble effect = GDAM_ARG_DOUBLE (arg);
  if (effect != flange->effect)
    {
      flange->effect = effect;
      gdam_flange_factors_changed (flange);
    }
}
static void gdam_flange_arg_get_effect(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAM_FLANGE (object)->effect;
}

static void gdam_flange_arg_set_chorus(GtkObject* object, GdamArg* arg)
{
  GdamFlange *flange = GDAM_FLANGE (object);
  gboolean new_chorus = GDAM_ARG_INT (arg);
  if (new_chorus == flange->chorus)
    return;
  flange->chorus = new_chorus;
  gdam_flange_factors_changed (flange);
}
static void gdam_flange_arg_get_chorus(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAM_FLANGE (object)->chorus;
}


static void gdam_flange_class_init(GdamFlangeClass* flange_class)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (flange_class);
  gdam_arg_register (object_class,
		     "min_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_min_factor,
		     gdam_flange_arg_get_min_factor,
		     NULL, "0,30", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mid_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_mid_factor,
		     gdam_flange_arg_get_mid_factor,
		     NULL, "0,30", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "max_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_max_factor,
		     gdam_flange_arg_get_max_factor,
		     NULL, "0,30", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "chorus",
		     GDAM_ARG_TYPE_INT,
		     gdam_flange_arg_set_chorus,
		     gdam_flange_arg_get_chorus,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdam_flange_arg_set_buffer_id,
		     gdam_flange_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_hertz,
		     gdam_flange_arg_get_hertz,
		     NULL, "0.001,10000", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_beat_factor,
		     gdam_flange_arg_get_beat_factor,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "effect",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_flange_arg_set_effect,
		     gdam_flange_arg_get_effect, NULL, "0,3", GDAM_ARG_VITAL);
  GDAM_FILTER_CLASS (flange_class)->append_start_request
    = gdam_flange_append_start_request;
}

GtkType gdam_flange_get_type()
{
  static GtkType flange_type = 0;
  if (!flange_type)
    {
      static const GtkTypeInfo flange_info = {
	"GdamFlange",
	sizeof (GdamFlange),
	sizeof (GdamFlangeClass),
	(GtkClassInitFunc) gdam_flange_class_init,
	(GtkObjectInitFunc) gdam_flange_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FIR_FILTER;
      flange_type = gtk_type_unique (parent, &flange_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return flange_type;
}

static void gdam_flange_factors_changed(GdamFlange* flange)
{
  GdamModel *model = GDAM_MODEL (flange);
  guint source_id = model->source_id;
  if (source_id != 0)
    {
      GdamArg args[3];
      gdouble min_max[2];
      args[0].name = "min_max";
      args[0].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
      GDAM_ARG_DOUBLE_ARRAY (args + 0).array = min_max;
      GDAM_ARG_DOUBLE_ARRAY (args + 0).length = 2;
      min_max[0] = flange->min_factor;
      min_max[1] = flange->max_factor;
      args[1].name = "effect";
      args[1].type = GDAM_ARG_TYPE_DOUBLE;
      GDAM_ARG_DOUBLE (args + 1) = flange->effect;
      args[2].name = "chorus";
      args[2].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + 2) = flange->chorus;
      gdam_channel_append_request (model->channel,
				   gdam_request_set_args
				   (source_id, 3, args, NULL, NULL, NULL));
    }
}
