/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamladspafilter.h"
#include <gdam/common/gdamsearchpath.h>
#include <gdam/component/gdambindingfile.h>
#include <gdam/common/gdam-protocol.h>
#include <gtk/gtksignal.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>
#include <gdam/component/gdamhistory.h>

gboolean 
parse_range_ladspa_parameter_xml_node    (xmlNode       *laspa_node,
				          GtkObject     *object);

enum {
          METERS_UPDATED,
          CONTROL_UPDATED,
	  LAST_SIGNAL
};

static guint ladspa_signals[LAST_SIGNAL];


/*
 *        _             _         _
 *  _ __ | |_   _  __ _(_)_ __   | |_ _   _ _ __   ___  ___
 * | '_ \| | | | |/ _` | | '_ \  | __| | | | '_ \ / _ \/ __|
 * | |_) | | |_| | (_| | | | | | | |_| |_| | |_) |  __/\__ \
 * | .__/|_|\__,_|\__, |_|_| |_|  \__|\__, | .__/ \___||___/
 * |_|            |___/               |___/|_|
 */

/* Here is a sample section from the xml file: */
/*
  <plugin>
    <name>...</name>
    <author>...</author>
    <unique_id>...</unique_id>
    <module>...</module>
    <module_index>...</module_index>
    <control>
      <name>...</name>
      <logarithmic>1</logarithmic>
      <min_value>0</min_value>
      <max_value>1</max_value>
    </control>
    <meter>
      <name>...</name>
      <logarithmic>1</logarithmic>
      <min_value>0</min_value>
      <max_value>1</max_value>
    </meter>
    <num_inputs>1</num_inputs>
    <num_outputs>1</num_outputs>
  </plugin>
*/

/* Random note:
 *
 * Although a LADSPA plugin may have any number of
 * inputs and outputs, here we only support 1 of them.
 * Furthermore we require that there be at least one of each.
 */

/* Hashtable mapping ``database names'' to PluginTypes. */
static GHashTable* plugin_hashtable = NULL;
static GPtrArray*  plugin_names = NULL;

static gint
xml_node_to_boolean(xmlNode* node)
{
  const char *contents = (const char *) node->content;
  if (contents == NULL && node->childs != NULL)
    contents = (const char *) node->childs->content;
  g_return_val_if_fail (contents != NULL, 0);
  return atoi (contents);
}

static gdouble
xml_node_to_double(xmlNode* node)
{
  const char *contents = (const char *) node->content;
  if (contents == NULL && node->childs != NULL)
    contents = (const char *) node->childs->content;
  g_return_val_if_fail (contents != NULL, 0);
  return strtod (contents, NULL);
}

char** gdam_ladspa_plugin_type_get_names  ()
{
  if (plugin_names == NULL)
    return NULL;
  if (plugin_names->pdata == NULL)
    return NULL;
  if (plugin_names->len <= 0)
    return NULL;

  if (plugin_names->pdata[plugin_names->len-1])
    {
      g_ptr_array_add (plugin_names, NULL);
    }
  return (char **) plugin_names->pdata;
}

GdamLadspaPluginType *
gdam_ladspa_plugin_type_get_by_name (const char *name)
{
  return g_hash_table_lookup (plugin_hashtable, name);
}

static gboolean
parse_control_or_meter(xmlNode*                  children,
                       GdamLadspaParameterType*  param_out)
{
  xmlNode *node;
  memset (param_out, 0, sizeof (GdamLadspaParameterType));
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "name") == 0)
	{
	  if (param_out->name != NULL)
	    {
	      g_warning ("control had multiple <name> tags");
	      return FALSE;
	    }
	  param_out->name = (char *) xmlNodeGetContent (node);
	}
      else if (strcmp (node->name, "logarithmic") == 0)
	{
	  param_out->is_logarithmic = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "sample_rate_ratio") == 0)
	{
	  param_out->is_sample_rate_ratio = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "integer") == 0)
	{
	  param_out->is_integer = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "toggle") == 0)
	{
	  param_out->is_toggle = xml_node_to_boolean (node);
	}
      else if (strcmp (node->name, "min_value") == 0)
	{
	  param_out->min_value = xml_node_to_double (node);
	  param_out->has_min_value = 1;
	}
      else if (strcmp (node->name, "max_value") == 0)
	{
	  param_out->max_value = xml_node_to_double (node);
	  param_out->has_max_value = 1;
	}
      else
	{
	  g_warning ("unknown node type for ladspa control: %s", node->name);
	}
    }
  return TRUE;
}

static void gdam_ladspa_plugin_type_destroy(GdamLadspaPluginType* type)
{
  int i;
  g_free (type->database_name);
  g_free (type->module_name);
  g_free (type->author);
  for (i = 0; i < type->num_meters; i++)
    g_free (type->meters[i].name);
  g_free (type->meters);
  for (i = 0; i < type->num_controls; i++)
    g_free (type->controls[i].name);
  g_free (type->controls);
  g_free (type);
}


static GdamLadspaPluginType* gdam_ladspa_plugin_type_from_xml(xmlNode* node)
{
  GdamLadspaPluginType *rv;
  GArray *control_array;
  GArray *meter_array;
  xmlNode *child;

  rv = g_new0 (GdamLadspaPluginType, 1);
  control_array = g_array_new (FALSE, FALSE,
			       sizeof (GdamLadspaParameterType));
  meter_array = g_array_new (FALSE, FALSE, sizeof (GdamLadspaParameterType));

  /* XXX: still need to check that the string returned by
   *      xmlNodeGetContent should be g_free'd !!!!!!!!! */
  for (child = node->childs; child != NULL; child = child->next)
    {
      if (strcmp (child->name, "name") == 0)
	{
	  rv->database_name = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "author") == 0)
	{
	  rv->author = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "unique_id") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->unique_id = id;
	}
      else if (strcmp (child->name, "module") == 0)
	{
	  rv->module_name = (char *) xmlNodeGetContent (child);
	}
      else if (strcmp (child->name, "module_index") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->module_index = id;
	}
      else if (strcmp (child->name, "control") == 0)
	{
	  GdamLadspaParameterType param_type;
	  if (!parse_control_or_meter (child->childs, &param_type))
	    {
	      g_warning ("couldn't parse control: skipping");
	      continue;
	    }
	  g_array_append_val (control_array, param_type);
	}
      else if (strcmp (child->name, "meter") == 0)
	{
	  GdamLadspaParameterType param_type;
	  if (!parse_control_or_meter (child->childs, &param_type))
	    {
	      g_warning ("couldn't parse meter: skipping");
	      continue;
	    }
	  g_array_append_val (meter_array, param_type);
	}
      else if (strcmp (child->name, "num_inputs") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->num_inputs = id;
	}
      else if (strcmp (child->name, "num_outputs") == 0)
	{
	  int id;
	  char *id_str;
	  id_str = (char *) xmlNodeGetContent (child);
	  id = atoi (id_str);
	  g_free (id_str);
	  rv->num_outputs = id;
	}
      else
	{
	  g_warning ("unknown node type for ladspa plugin: %s"
		     "... ignoring it", child->name);
	}
    }

  /* Shimmy the array we just constructed into the PluginType
   * structure. */
  rv->num_controls = control_array->len;
  rv->controls = (GdamLadspaParameterType *) control_array->data;
  rv->num_meters = meter_array->len;
  rv->meters = (GdamLadspaParameterType *) meter_array->data;
  g_array_free (control_array, FALSE);
  g_array_free (meter_array, FALSE);

  if (rv->database_name == NULL
      || rv->module_name == NULL || rv->module_index < 0)
    {
      g_warning ("incomplete XML-LADSPA description");
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  if (rv->num_inputs == 0 || rv->num_inputs > 2)
    {
      gdam_debug ("skipping LADSPA plugin with %d inputs [%s]",
		  rv->num_inputs, rv->database_name);
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  if (rv->num_outputs == 0)
    {
      gdam_debug ("skipping LADSPA plugin with no outputs [%s]",
		  rv->database_name);
      gdam_ladspa_plugin_type_destroy (rv);
      return NULL;
    }
  return rv;
}

static void initialize_ladspa_database()
{
  char *filename;
  xmlDoc *doc;
  xmlNode *root;
  xmlNode *child;

  if (plugin_hashtable == NULL)
    {
      plugin_hashtable = g_hash_table_new (g_str_hash, g_str_equal);
      plugin_names = g_ptr_array_new ();
    }

  filename = gdam_search_path ("ladspa.xml",
			       "/usr/share/ladspa:/usr/local/share/ladspa"
			       ":/usr/lib/ladspa:/usr/local/lib/ladspa"
			       ":/usr/share/gdam:/usr/local/share/gdam"
			       ":/usr/lib/gdam:/usr/local/lib/gdam"
			       ":~/.gdam", GDAM_SEARCH_PATH_FIND_FILE);
  if (filename == NULL)
    {
      g_warning ("couldn't find ladspa.xml -- "
		 "you need to run gdamladspapluginscanner");
      return;
    }


  doc = xmlParseFile (filename);
  if (doc == NULL)
    {
      g_warning ("xml in %s (ladspa database) malformed", filename);
      return;
    }

  root = doc->root;
  if (strcmp (root->name, "LADSPA-Plugin-Descriptions") != 0)
    {
      g_warning ("LADSPA plugin database didn't have appropriate "
		 "initial xml-tag <LADSPA-Plugin-Descriptions>");
      xmlFreeDoc (doc);
      return;
    }
  for (child = root->childs; child != NULL; child = child->next)
    {
      GdamLadspaPluginType *plugin_type;
      plugin_type = gdam_ladspa_plugin_type_from_xml (child);
      if (plugin_type == NULL)
	{
	  /*g_message("error parsing ladspa plugin- skipping"); */
	  continue;
	}
      if (g_hash_table_lookup (plugin_hashtable,
			       plugin_type->database_name) != NULL)
	{
	  g_warning ("multiple plugins named `%s' were found -- "
		     "ignoring duplicates", plugin_type->database_name);
	  gdam_ladspa_plugin_type_destroy (plugin_type);
	  continue;
	}
      g_hash_table_insert (plugin_hashtable,
			   plugin_type->database_name, plugin_type);
      g_ptr_array_add (plugin_names, plugin_type->database_name);
    }
  xmlFreeDoc (doc);
}

void 
gdam_ladspa_filter_set_all_controls (GdamLadspaFilter *ladspa);

static void
control_success (int num_args, GdamArg* args, gpointer user_data)
{
  GdamLadspaFilter *ladspa_filter;
  ladspa_filter = (GdamLadspaFilter *)user_data;
  ladspa_filter->making_arg_request--;
  if (ladspa_filter->controls_pending > 0)
    gdam_ladspa_filter_set_all_controls (ladspa_filter);
}

static void
gdam_ladspa_filter_controls_to_xml (GdamLadspaFilter *ladspa, xmlNode *parent)
{
  GdamLadspaPluginType *plugin_type = ladspa->plugin_type;
  xmlNode *plugin_node;
  char buf[1024];
  int num_controls = plugin_type->num_controls;
  int i;

  plugin_node = xmlNewChild (parent, NULL, "Plugin", NULL);
  snprintf (buf, sizeof (buf), "%d", plugin_type->unique_id);
  xmlSetProp (plugin_node, "plugin_id", buf);
  if (plugin_type->database_name)
    xmlSetProp (plugin_node, "name", plugin_type->database_name);

  /*  For each control port, save indentification and values */
  for (i = 0; i < num_controls; i++)
    {
      xmlNode *control_node, *lvalue_node, *rvalue_node;

      /* Create a node to identify the control port. */
      control_node = xmlNewChild (plugin_node, NULL, "Control-Port", NULL);
      snprintf (buf, sizeof (buf), "%d", i);
      xmlSetProp (control_node, "index", buf);
      if (plugin_type->controls[i].name)
	xmlSetProp (control_node, "name", plugin_type->controls[i].name);

#define STEREO_VALUE 0

#if STEREO_VALUE
      /* Store values for each channel  (explicitly left and right for now) */
      snprintf (buf, sizeof (buf), "%f", ladspa->controls[i * 2 + 0]);
      lvalue_node = xmlNewChild (control_node, NULL, "Value", buf);
      xmlSetProp (lvalue_node, "channel_idx", "0");
      snprintf (buf, sizeof (buf), "%f", ladspa->controls[i * 2 + 1]);
      lvalue_node = xmlNewChild (control_node, NULL, "Value", buf);
      xmlSetProp (lvalue_node, "channel_idx", "1");
#else
      snprintf (buf, sizeof (buf), "%f", ladspa->controls[i * 2]);
      xmlSetProp (control_node, "value", buf);
#endif
    }
}

static gboolean
gdam_ladspa_filter_controls_from_xml (GdamLadspaFilter *ladspa, xmlNode *node)
{
  GdamLadspaPluginType *plugin_type = ladspa->plugin_type;
  xmlNode *plugin_node = node;
  xmlNode *child;
  char buf[1024];
  const char *property;
  int num_controls = plugin_type->num_controls;
  int i;

  if (g_strcasecmp (plugin_node->name, "Plugin") != 0)
    {
      g_warning ("node type %s expected Plugin", plugin_node->name);
      return FALSE;
    }
  property = xmlGetProp (plugin_node, "plugin_id");
  if (!property)
    {
      g_warning ("missing unique id when loading control values from xml");
      return FALSE;
    }
  else if (atoi (property) != (int)plugin_type->unique_id)
    {
      return FALSE;
    }

  for (child = plugin_node->childs; child != NULL; child = child->next)
    {
      xmlNode *control_node = child;
      gdouble value, left_value, right_value;
      int index;
      if (g_strcasecmp (control_node->name, "Control-Port") != 0)
	{
	  g_warning ("unknown node type %s when loading preset", 
	             control_node->name);
	  continue;
	}
      property = xmlGetProp (control_node, "index");
      if (!property)
	{
	  g_warning ("missing control index when loading values from xml");
	  continue;
	}
      index = atoi (property);
#if STEREO_VALUE
      {
	xmlNode *lvalue_node, *rvalue_node;
	g_warning ("unfinished");
      }
#else
      property = xmlGetProp (control_node, "value");
      if (!property)
	{
	  g_warning ("missing control index when loading values from xml");
	  continue;
	}
      value = atof (property);
      left_value = right_value = value;
#endif
      gdam_ladspa_filter_set_control (ladspa, index, left_value, right_value);
    }
  return TRUE;
}



/*=====================================================================
 *          Server correspondence
 */
static int gdam_ladspa_filter_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamLadspaFilter *ladspa = GDAM_LADSPA_FILTER (filter);
  GdamModel *model = GDAM_MODEL (filter);
  GdamLadspaPluginType *plugin_type = ladspa->plugin_type;
  GdamRequest *request;
  gint *index_array;
  int i, local_id;
  int num_controls;
  GdamArg args[2];
  gtk_object_ref (GTK_OBJECT (ladspa));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasLadspaPlugin",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					ladspa, 0);
  local_id = gdam_request_builder_append_create (builder, request);

  args[0].name = "filename";
  args[0].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + 0) = plugin_type->module_name;
  args[1].name = "index";
  args[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 1) = plugin_type->module_index;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					local_id,
					"set_plugin",
					2, args, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);

  num_controls = plugin_type->num_controls;
  index_array = alloca (sizeof (int) * num_controls);
  for (i = 0; i < num_controls; i++)
    index_array[i] = i;
  args[0].name = "control_indices";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = num_controls;
  GDAM_ARG_INT_ARRAY (args + 0).array = index_array;
  args[1].name = "controls";
  args[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).length = num_controls * 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).array = ladspa->controls;
  g_return_val_if_fail (ladspa->making_arg_request <= 0, 0);
  ladspa->making_arg_request++;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					local_id,
					"set_controls",
					2, args, control_success, NULL, ladspa);
  gdam_request_builder_append_generic (builder, request);

  args[0].name = "filtered";
  args[0].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + 0) = new_source_id;
  request = gdam_request_set_args (local_id, 1, args, NULL, NULL, NULL);
  gdam_request_builder_append_generic (builder, request);

  return local_id;
}

static void gdam_ladspa_filter_reset(GtkObject*        object,
                                     GdamArg*          arg)
{
  GdamRequest *request;
  GdamModel *model = GDAM_MODEL (object);
  int source_id = model->source_id;
  request = gdam_request_set_args (source_id, 1, arg, NULL, NULL, NULL);
  gdam_channel_append_request (model->channel, request);
}

static GPtrArray *num_array = NULL;

static const char *
get_number_as_str (guint n)
{
  static GStringChunk *chunk;
  char buf[64];
  if (num_array == NULL)
    {
      num_array = g_ptr_array_new ();
      chunk = g_string_chunk_new (640);
    }
  if (num_array->len <= n)
    g_ptr_array_set_size (num_array, n + 1);
  if (num_array->pdata[n] == NULL)
    {
      g_snprintf (buf, sizeof (buf), "%d", n);
      num_array->pdata[n] = g_string_chunk_insert (chunk, buf);
    }
  return num_array->pdata[n];
}

/* for keeping track of per-parameter argument registration */
static int num_registered_args = 0;

static void
arg_set_parameter (GtkObject* object, GdamArg*   arg)
{
  GdamLadspaFilter *ladspa_filter = GDAM_LADSPA_FILTER (object);
  int idx = atoi (arg->name);
  gdouble val = GDAM_ARG_DOUBLE (arg);
  gdam_ladspa_filter_set_control (ladspa_filter, idx, val, val);
}

static void 
arg_get_parameter (GtkObject* object, GdamArg*   arg)
{
  GdamLadspaFilter *ladspa_filter = GDAM_LADSPA_FILTER (object);
  int idx = atoi (arg->name);
  GDAM_ARG_DOUBLE (arg) = ladspa_filter->controls[idx * 2];
}

static void
arg_set_save_preset (GtkObject* object, GdamArg*   arg)
{
  GdamLadspaFilter *ladspa_filter = GDAM_LADSPA_FILTER (object);
  gdam_ladspa_filter_save_preset (ladspa_filter, GDAM_ARG_STRING (arg));
}

static void
arg_set_load_preset (GtkObject* object, GdamArg*   arg)
{
  GdamLadspaFilter *ladspa_filter = GDAM_LADSPA_FILTER (object);
  gdam_ladspa_filter_load_preset (ladspa_filter, GDAM_ARG_STRING (arg));
}

static void gdam_ladspa_filter_init(GdamLadspaFilter* ladspa_filter)
{
}
static void gdam_ladspa_filter_class_init(GdamFilterClass* filter_class)
{
  GdamBindingNodeTypeInfo *info;
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (filter_class);
  guint i;
  info = g_new (GdamBindingNodeTypeInfo, 1);
  info->node_name = g_strdup ("range_ladspa_parameter");
  info->parse = parse_range_ladspa_parameter_xml_node;
  gdam_binding_file_node_type_register (info);
  filter_class->append_start_request
    = gdam_ladspa_filter_append_start_request;

  gdam_arg_register (object_class,
		     "save_preset",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_save_preset,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "load_preset",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_load_preset,
		     NULL, NULL, NULL, 0);

  ladspa_signals[CONTROL_UPDATED] =
    gtk_signal_new ("control_updated",
		    GTK_RUN_LAST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GdamLadspaFilterClass, control_updated),
		    gtk_marshal_NONE__INT,
		    GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  gtk_object_class_add_signals (object_class, 
				ladspa_signals, 
				LAST_SIGNAL);
}

GtkType gdam_ladspa_filter_get_type()
{
  static GtkType ladspa_filter_type = 0;
  if (!ladspa_filter_type)
    {
      static const GtkTypeInfo ladspa_filter_info = {
	"GdamLadspaFilter",
	sizeof (GdamLadspaFilter),
	sizeof (GdamLadspaFilterClass),
	(GtkClassInitFunc) gdam_ladspa_filter_class_init,
	(GtkObjectInitFunc) gdam_ladspa_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      initialize_ladspa_database ();
      ladspa_filter_type = gtk_type_unique (parent, &ladspa_filter_info);
    }
  return ladspa_filter_type;
}

static void
gdam_ladspa_filter_set_plugin_type(GdamLadspaFilter*     filter,
                                   GdamLadspaPluginType* plugin_type)
{
  GtkObjectClass *object_class;
  int num_controls = plugin_type->num_controls;
  int num_meters = plugin_type->num_meters;
  int i;

  object_class = GTK_OBJECT_CLASS (GTK_OBJECT (filter)->klass);

  filter->controls = g_new0 (gdouble, 2 * num_controls);
  filter->meters = g_new0 (gdouble, 2 * num_meters);
  filter->which_meters_to_query = g_new0 (gboolean, num_meters);
  filter->plugin_type = plugin_type;

  for (i = num_registered_args; i < num_controls; i++)
    {
      gdam_arg_register (object_class,
			 get_number_as_str (i),
			 GDAM_ARG_TYPE_DOUBLE,
			 arg_set_parameter,
			 arg_get_parameter,
			 NULL, NULL, 0);
      num_registered_args++;
    }
}

GdamLadspaFilter* gdam_ladspa_filter_new(const char* name)
{
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *filter;
  GtkType type;

  type = GDAM_TYPE_LADSPA_FILTER;

  plugin_type = g_hash_table_lookup (plugin_hashtable, name);
  if (plugin_type == NULL)
    {
      g_warning ("Sorry, no plugin in the database named %s", name);
      return NULL;
    }

  filter = GDAM_LADSPA_FILTER (gtk_type_new (type));
  gdam_ladspa_filter_set_plugin_type (filter, plugin_type);
  gdam_ladspa_filter_load_preset (filter, "default");
  return filter;
}
void gdam_ladspa_filter_set_control     (GdamLadspaFilter* filter,
                                         int               control_index,
                                         gdouble           left_value,
                                         gdouble           right_value)
{
  GdamModel *model;
  GdamArg args[2];
  gdouble lr_values[2];
  GdamRequest *request;

  model = GDAM_MODEL (filter);
  filter->controls[control_index * 2 + 0] = left_value;
  filter->controls[control_index * 2 + 1] = right_value;

  gdam_arg_set (GTK_OBJECT (filter), get_number_as_str (control_index),
                left_value);

  gtk_signal_emit (GTK_OBJECT (filter),
		   ladspa_signals[CONTROL_UPDATED],
		   control_index);

  if (model->source_id == 0)
    return;

  if (filter->making_arg_request > 0)
    {
      filter->controls_pending++;
      return;
    }
  filter->making_arg_request++;

  args[0].name = "control_indices";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = 1;
  GDAM_ARG_INT_ARRAY (args + 0).array = &control_index;
  lr_values[0] = left_value;
  lr_values[1] = right_value;
  args[1].name = "controls";
  args[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).length = 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).array = lr_values;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					model->source_id,
					"set_controls",
					2, args, control_success, NULL, filter);
  gdam_channel_append_request (model->channel, request);
}

void 
gdam_ladspa_filter_set_all_controls (GdamLadspaFilter *ladspa)
{
  GdamLadspaPluginType *plugin_type = ladspa->plugin_type;
  GdamModel *model = (GdamModel *)ladspa;
  GdamArg args[2];
  gint *index_array;
  GdamRequest *request;
  int num_controls, i;

  num_controls = plugin_type->num_controls;
  index_array = alloca (sizeof (int) * num_controls);
  for (i = 0; i < num_controls; i++)
    index_array[i] = i;
  args[0].name = "control_indices";
  args[0].type = GDAM_ARG_TYPE_INT_ARRAY;
  GDAM_ARG_INT_ARRAY (args + 0).length = num_controls;
  GDAM_ARG_INT_ARRAY (args + 0).array = index_array;
  args[1].name = "controls";
  args[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).length = num_controls * 2;
  GDAM_ARG_DOUBLE_ARRAY (args + 1).array = ladspa->controls;
  if (ladspa->making_arg_request > 0)
    {
      ladspa->controls_pending++;
      return;
    }
  ladspa->controls_pending = 0;
  ladspa->making_arg_request++;
  request = gdam_request_call_function (GDAM_PROTOCOL_SOURCE_ARG,
					model->source_id,
					"set_controls",
					2, args, control_success, NULL, ladspa);
  gdam_channel_append_request (model->channel, request);
}

/* 
 *  _                   _     _     _           _ _             
 * (_)_ __  _ __  _   _| |_  | |__ (_)_ __   __| (_)_ __   __ _ 
 * | | '_ \| '_ \| | | | __| | '_ \| | '_ \ / _` | | '_ \ / _` |
 * | | | | | |_) | |_| | |_  | |_) | | | | | (_| | | | | | (_| |
 * |_|_| |_| .__/ \__,_|\__| |_.__/|_|_| |_|\__,_|_|_| |_|\__, |
 *         |_|                                            |___/ 
 */

typedef struct _GdamInputLadspaParameter GdamInputLadspaParameter;
struct _GdamInputLadspaParameter {
	GdamInputTrap*	trap;
	int		parameter_index;

	GtkObject*      object;

	guint		connect_id;

	/* For range arguments. */
	gdouble		min_value, max_value;
	gint		is_logarithmic;
	gint		is_exponent;

	/* For variable arguments. */
	gdouble 	factor;
	/* For constant and variable arguments. */
	gboolean	is_delta;
	/* For constant arguments. */
	gboolean	is_factor;
	gboolean	is_int;
};

static void 
ladspa_parameter_free(GdamInputLadspaParameter *arg)
{
  g_free (arg);
}

static gdouble translate_range_value (GdamInputEventRange *range,
                                      GdamInputLadspaParameter *input_arg)
{
  gdouble value;
  if (input_arg->is_logarithmic) 
    {
      gdouble offset;
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value);
      offset = 1.0 - input_arg->min_value;
      value = value * (log (input_arg->max_value + offset) 
	              - log (input_arg->min_value + offset));
      value += log (input_arg->min_value + offset);
      value = exp (value) - offset;
    }
  else if (input_arg->is_exponent) 
    {
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value)
	* (input_arg->max_value - input_arg->min_value) + input_arg->min_value;
      value = ldexp (1, value);
    }
  else 
    {
      value = (range->value - range->min_value)
	/ (range->max_value - range->min_value)
	* (input_arg->max_value - input_arg->min_value) + input_arg->min_value;
    }
  if (input_arg->is_int)
    value = (int)value;
  return value;
}


/* 
 * `event_type' must be a GdamInputEventRange.
 */
static gboolean 
on_param_set (GdamInputEvent*     input_event,
	      gpointer            user_data)
{
  GdamInputEventRange *range;
  GdamInputLadspaParameter *input_param = user_data;
  gdouble value;
  range = (GdamInputEventRange *) input_event;
  value = translate_range_value (range, input_param);
  gdam_ladspa_filter_set_control (GDAM_LADSPA_FILTER (input_param->object),
				 input_param->parameter_index, value, value);
  return TRUE;
}


/* Given a trap corresponding to an EventRange, add a hook which
 * sets a GdamArg on an object. */
guint input_hook_connect_range_ladspa (GdamInputTrap*   trap,
                                       GtkObject*       object,
				       gboolean	        parameter_index,
				       gdouble          min_value,
				       gdouble          max_value,
				       gboolean         is_logarithmic,
				       gboolean         is_exponent)
{
  GdamInputLadspaParameter *input_param;
  gboolean is_double;

  if (!gdam_input_event_type_is_a (trap->event->event_type,
				   gdam_input_event_range_get_type ()))
    {
      g_warning ("only GdamInputEventRange's (not %s) may be connected"
		 " through %s",
		 gdam_input_event_type_name (trap->event->event_type),
		 __FUNCTION__);
      return 0;
    }

  g_return_val_if_fail (trap != NULL, 0);

  input_param = g_new0 (GdamInputLadspaParameter, 1);
  input_param->trap = trap;
  input_param->parameter_index = parameter_index;
  input_param->object = object;
  input_param->min_value = min_value;
  input_param->max_value = max_value;
  input_param->is_logarithmic = is_logarithmic;
  input_param->is_exponent = is_exponent;
  input_param->is_int = FALSE;
  input_param->connect_id = gdam_input_hook_connect
    (trap, on_param_set,
     (GDestroyNotify) ladspa_parameter_free, input_param, object);

  return input_param->connect_id;
}

static const char* get_string_content(xmlNode* node)
{
  if (node->content)
    return node->content;
  if (node->childs == NULL)
    return NULL;
  return node->childs->content;
}

gboolean 
parse_range_ladspa_parameter_xml_node    (xmlNode       *laspa_node,
				          GtkObject     *object)
{
  xmlNode *node;
  GdamArg arg;
  xmlNode *children = laspa_node->childs;
  const char *param_index = NULL;
  const char *min_value = NULL;
  const char *max_value = NULL;
  const char *is_logarithmic = "0";
  const char *is_exponent = "0";
  GdamInputEvent *input_event = NULL;
  guint input_hook;
  gboolean is_double;

  if (object == NULL)
    return FALSE;

  g_return_val_if_fail (GDAM_IS_LADSPA_FILTER (object), FALSE);

  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "parameter_index") == 0)
	{
	  param_index = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "min_value") == 0)
	{
	  min_value = get_string_content (node);
	  g_return_val_if_fail (min_value != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "max_value") == 0)
	{
	  max_value = get_string_content (node);
	  g_return_val_if_fail (max_value != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "is_logarithmic") == 0)
	{
	  is_logarithmic = get_string_content (node);
	  g_return_val_if_fail (is_logarithmic != NULL, FALSE);
	  continue;
	}
      if (strcmp (node->name, "is_exponent") == 0)
	{
	  is_exponent = get_string_content (node);
	  g_return_val_if_fail (is_exponent != NULL, FALSE);
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (input_event == NULL)
    {
      g_warning ("range must have GdamInputEvent or kin");
      return FALSE;
    }
  if (param_index == NULL)
    {
      g_warning ("<range_ladspa_parameter> was missing <parameter_index>");
      return FALSE;
    }
  if (min_value == NULL && max_value == NULL)
    {
      gdam_debug ("assuming range from 0.0 to 1.0");
      min_value = "0.0";
      max_value = "1.0";
    }
  if (min_value == NULL || max_value == NULL || is_logarithmic == NULL)
    {
      g_warning ("min_value or max_value was not specified");
      return FALSE;
    }

  input_hook = input_hook_connect_range_ladspa
			  (gdam_input_trap_find_or_create (input_event),
			   object,
			   strtod (param_index, NULL), 
			   strtod (min_value, NULL), 
			   strtod (max_value, NULL),
			   strtod (is_logarithmic, NULL),
			   strtod (is_exponent, NULL)
			   );
  return TRUE;
}

/* TODO: if !ladspa_presets/ && ~/.gdam/, then make dir ~/.gdam/ladspa_presets/
 */
static char *
find_ladspa_preset_dir ()
{
  return gdam_search_path ("ladspa_presets/",
                           NULL,
			   GDAM_SEARCH_PATH_FIND_DIRECTORY);
}

static char *
find_ladspa_preset_file (int unique_id)
{
  char buf[1024];
  snprintf (buf, sizeof (buf), "ladspa_presets/%d.xml", unique_id);
  return gdam_search_path (buf, NULL, GDAM_SEARCH_PATH_FIND_FILE);
}

/* Desired filename for saving presets:  may not exist, but is eligible for 
 * creation.  Returns NULL if no file exists and no location can be found to
 * create one.  Any string returned must be freed by the caller. */
static char *
get_ladspa_preset_filename (GdamLadspaFilter *filter, const char *preset_name)
{
  GdamLadspaPluginType *plugin_type = filter->plugin_type;
  /* The first ladspa preset directory found... this is the preferred location
   * to save a new file, although an existing file may be found in a
   * different directory.
   */
  char *dir;
  char *fname = NULL;

  /* As a special case, the "default" preset uses a different file lookup...
   * because default presets for all filters might be distributed in one file */
  if (g_strcasecmp (preset_name, "default") == 0)
    {
      const char *found;
      found =  gdam_search_path ("ladspa_presets/defaults.xml", 
                              NULL, GDAM_SEARCH_PATH_FIND_FILE);
      if (found)
	return g_strdup (found);
      found =  gdam_search_path ("ladspa_defaults.xml",
                                 NULL, GDAM_SEARCH_PATH_FIND_FILE);
      if (found)
	return g_strdup (found);
      dir = find_ladspa_preset_dir ();
      if (dir)
	{
	  char *rv;
	  rv = g_strdup_printf ("%s/defaults.xml", dir);
	  g_free (dir);
	  return rv;
	}
      return NULL;
    }

  fname = find_ladspa_preset_file (plugin_type->unique_id);
  if (fname)
    return fname;

  /* If there is no existing file, create a new file in the
   * preferred location. */
  if (!fname)
    {
      char *rv;
      dir = find_ladspa_preset_dir ();
      if (!dir)
	return NULL;
      rv = g_strdup_printf ("%s/%d.xml", dir, plugin_type->unique_id);
      g_free (dir);
      return rv;
    }
  return NULL;
}

static xmlDoc *
find_or_create_xml_doc (const char *fname)
{
  xmlDoc *document = xmlParseFile (fname);

  /* If we don't have an existing doc, but have a valid filename... */
  if (!document && fname)
    {
      xmlNode *root;
      document = xmlNewDoc (XML_DEFAULT_VERSION);
      root = xmlNewDocNode (document, NULL, "Ladspa-Plugin-Settings", NULL);
      document->root = root;
    }
  return document;
}

static char* ladspa_preset_directory = NULL;

static void
initialize_preset_directory()
{
  static gboolean inited = FALSE;
  struct stat stat_buf;
  const char *gdam_dir;

  if (inited)
    return;

  inited = TRUE;
  initialize_root_gdam_directory ();

  gdam_dir = gdam_get_user_dir ();

  if (!gdam_dir)
    return;

  ladspa_preset_directory = g_strdup_printf ("%s/ladspa_presets", gdam_dir);

  if (stat (ladspa_preset_directory, &stat_buf) < 0)
    {
      g_warning ("no ladspa_presets directory -- I will create one");
      if (mkdir (ladspa_preset_directory, 0755) < 0)
	{
	  g_warning ("I cannot make the directory `%s'...",
		     ladspa_preset_directory);
	  g_free (ladspa_preset_directory);
	  ladspa_preset_directory = NULL;
	  return;
	}
      return;
    }
  if (!S_ISDIR (stat_buf.st_mode))
    {
      g_warning ("currently %s exists but isn't a directory, "
		 "guess I won't save your presets",
		 ladspa_preset_directory);
      g_free (ladspa_preset_directory);
      ladspa_preset_directory = NULL;
      return;
    }
  return;
}





void
gdam_ladspa_filter_save_preset (GdamLadspaFilter *filter, 
                                const char       *preset_name)
{
  GdamLadspaPluginType *plugin_type;
  char *fname = NULL;
  char *dir;
  char *property;
  gboolean done = FALSE;
  xmlDoc *document;
  xmlNode *root, *child;
  xmlNode *preset_node = NULL;
  plugin_type = filter->plugin_type;

  initialize_preset_directory();

  fname = get_ladspa_preset_filename (filter, preset_name);
  if (fname == NULL)
    {
      g_warning ("couldn't find ladspa preset file, consider "
	         "creating a directory ~/.gdam/ladspa_presets/");
      return;
    }
  document = find_or_create_xml_doc (fname);
  root = document->root;

  /* If this doc already has a node for our preset, check whether
   * our plugin type is represented... if so, clear the old data. */
  for (child = root->childs; child != NULL && !done; child = child->next)
    {
      if (g_strcasecmp (child->name, "Ladspa-Plugin-Settings") == 0 &&
	  (property = xmlGetProp (child, "preset_name")) &&
	  g_strcasecmp (property, preset_name) == 0)
	{
	  gboolean done2 = FALSE;
	  xmlNode *subchild;
	  /* This is the node for our preset */
	  preset_node = child;
	  /* Look for old data */
	  for (subchild = child->childs; 
	       subchild != NULL && !done2;
	       subchild = subchild->next)
	    {
	      const char *name = subchild->name;
	      if (g_strcasecmp (subchild->name, "Plugin") == 0 &&
		  (property = xmlGetProp (subchild, "plugin_id")) &&
		  atoi (property) == (int)plugin_type->unique_id)
		{
		  xmlUnlinkNode (subchild);
		  xmlFreeNode (subchild);
		  done2 = TRUE;
		}
	    }
	  done = TRUE;
	}
    }

  if (!preset_node)
    {
      preset_node = xmlNewChild (root, NULL, "Ladspa-Plugin-Settings", NULL);
      xmlSetProp (preset_node, "preset_name", preset_name);
    }
  gdam_ladspa_filter_controls_to_xml (filter, preset_node);

  xmlSaveFile (fname, document);
  xmlFreeDoc (document);
  g_free (fname);
}

static void 
handle_settings_node (GdamLadspaFilter *filter, 
                      xmlNode          *node,
		      const char       *preset_name)
{
  xmlNode *child;
  gboolean handled_node = FALSE;

  for (child = node->childs; child != NULL; child = child->next)
    {
      if (g_strcasecmp (child->name, "Ladspa-Plugin-Settings") == 0)
	{
          const char *property = xmlGetProp (child, "preset_name");
	  /* skip presets whose names don't match */
	  if (property && g_strcasecmp (property, preset_name) != 0)
	    continue;
	  handle_settings_node (filter, child, preset_name);
	  handled_node = TRUE;
	}
      else if (g_strcasecmp (child->name, "Plugin") == 0)
	{
	  gboolean rv;
	  rv = gdam_ladspa_filter_controls_from_xml (filter, child);
	  if (rv)
	    handled_node = TRUE;
	}
    }
  if (!handled_node)
    g_warning ("couldn't handle any ladspa filter settings nodes");
}

void
gdam_ladspa_filter_load_preset (GdamLadspaFilter *filter, 
                                const char       *preset_name)
{
  char *fname;
  xmlDoc *document;
  xmlNode *root, *child;
  fname = get_ladspa_preset_filename (filter, preset_name);
  if (fname == NULL)
    {
      g_warning ("couldn't find ladspa preset file");
      return;
    }
  document = xmlParseFile (fname);
  if (document == NULL)
    {
      g_warning ("couldn't parse xml from %s", fname);
      g_free (fname);
      return;
    }
  g_free (fname);
  root = document->root;
  if (g_strcasecmp (root->name, "Ladspa-Plugin-Settings") != 0)
    {
      g_warning ("node %s expected Ladspa-Plugin-Settings", root->name);
      return;
    }
  handle_settings_node (filter, root, preset_name);
  xmlFreeDoc (document);
}

