/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_MIDI_DEVICE_H_
#define __GDAM_MIDI_DEVICE_H_


typedef struct _GdamMidiDevice GdamMidiDevice;
typedef struct _GdamMidiDeviceClass GdamMidiDeviceClass;
typedef struct _GdamMidiTrap GdamMidiTrap;

typedef struct _GdamMidiEventNote GdamMidiEventNote;
typedef struct _GdamMidiEventPitch GdamMidiEventPitch;
typedef struct _GdamMidiEventParam GdamMidiEventParam;
typedef struct _GdamMidiEventSystemExclusive GdamMidiEventSystemExclusive;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "../gdaminputdevicefile.h"

typedef union _GdamMidiEvent GdamMidiEvent;

GtkType gdam_midi_device_get_type();
#define GDAM_TYPE_MIDI_DEVICE			(gdam_midi_device_get_type ())
#define GDAM_MIDI_DEVICE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_MIDI_DEVICE, GdamMidiDevice))
#define GDAM_MIDI_DEVICE_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_MIDI_DEVICE, GdamMidiDeviceClass))
#define GDAM_IS_MIDI_DEVICE(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_MIDI_DEVICE))
#define GDAM_IS_MIDI_DEVICE_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_MIDI_DEVICE))

/* For use with midi piano keys. */
struct _GdamMidiEventNote {
	GdamInputIntEvent	input_event;
	int			velocity;
	int			midi_index;
	const char*		note_name;
	int			octave;
	gdouble			hertz;

	gboolean		matches_any;
};

/* For use with pitch wheel. */
struct _GdamMidiEventPitch {
	GdamInputEventRange	input_event;
};

/* For use with "data entry" aka "parameter" midi events. */
struct _GdamMidiEventParam {
	GdamInputEventRange	input_event;
	int			parameter;
};

struct _GdamMidiEventSystemExclusive {
	GdamInputEvent		input_event;
	const char*		make_name;
	const char*		model_name;
	int			message_length;
	guint8*			message;
};

struct _GdamMidiDeviceClass {
	GdamInputDeviceFileClass	input_device_class;

};


/*=====================================================================*
 *                            GdamMidiDevice                           *
 *                                                                     *
 * A device, in the sense it is a single /dev/midi* file.              *
 *=====================================================================*/
struct _GdamMidiDevice {
	GdamInputDeviceFile	object;

	/* Iff we are in the middle of a system exclusive
	 * then this is non-NULL. */
	GString*		system_exclusive;

	/* List of patch files (by filename) we've loaded */
	GSList*			patch_list;

	/* MIDI make and model id numbers. */
	int			make_id;
	int			model_id;

	int			triplet_length;
	unsigned char		triplet[3];
};

/*
 * Load a `map file'  of human readable names for various midi controls.
 * (We made this format up ourselves:
 *     pitchwheel 0 name
 *     parameter 0 0 name
 *     sysex length ... ... ... name
 *     etc.)
 */
gboolean        gdam_midi_device_load_map   (GdamMidiDevice*    midi_dev,
                                             const char*        db_filename);


/* For derivers, or those who want to simulate device inputs... */
guint gdam_midi_event_note_on_get_type();
guint gdam_midi_event_note_off_get_type();
guint gdam_midi_event_any_note_on_get_type();
guint gdam_midi_event_any_note_off_get_type();
guint gdam_midi_event_pitch_wheel_get_type();
guint gdam_midi_event_parameter_get_type();
guint gdam_midi_event_sysex_get_type();

GdamInputEvent* gdam_midi_event_program_new (GdamMidiDevice* device,
                                             int             channel,
				             int             program);
GdamInputEvent* gdam_midi_event_channel_pressure_new
                                            (GdamMidiDevice* device,
                                             int             channel,
					     int             pressure);

/* if `matches_any' is TRUE, this event is intended to match any
 * pitch.  otherwise, each pitch is trapped separately.
 */
GdamInputEvent* gdam_midi_event_note_new    (GdamMidiDevice* device,
					     int             instrument,
                                             int             pitch,
					     int             volume,
					     gboolean        on,
					     gboolean        matches_any);
GdamInputEvent* gdam_midi_event_key_pressure_new
                                            (GdamMidiDevice* device,
					     int             instrument,
                                             int             pitch,
					     int             volume);
GdamInputEvent* gdam_midi_event_parameter_new(GdamMidiDevice* device,
                                             int             channel,
					     int             which_param,
					     int             value);
GdamInputEvent* gdam_midi_event_pitch_wheel_new
                                            (GdamMidiDevice* device,
                                             int             channel,
					     int             arg);
GdamInputEvent* gdam_midi_event_sysex_new   (GdamMidiDevice* device,
				             const char*     maker,
				             const char*     model,
                                             int             length,
				             const guint8*   bytes);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
