/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamphaser.h"
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gdam/component/gdamutil.h>

enum {
	DELAY_RANGE,
	DEPTH,
	FEEDBACK,
	LFO_HERTZ,
	LAST_ARG_SET
};


static void gdam_phaser_class_init(GdamPhaserClass* klass);

static int gdam_phaser_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_phaser_init(GdamPhaser* phaser)
{
  phaser->min_hertz = 440.0;
  phaser->max_hertz = 1600.0;
  phaser->feedback = 0.4;
  phaser->depth = 0.4;
}

GtkType gdam_phaser_get_type()
{
  static GtkType phaser_type = 0;
  if (!phaser_type)
    {
      static const GtkTypeInfo phaser_info = {
	"GdamPhaser",
	sizeof (GdamPhaser),
	sizeof (GdamPhaserClass),
	(GtkClassInitFunc) gdam_phaser_class_init,
	(GtkObjectInitFunc) gdam_phaser_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      phaser_type = gtk_type_unique (GDAM_TYPE_FILTER, &phaser_info);
    }
  return phaser_type;
}

static inline void maybe_dispense_delay_range_request(GdamPhaser* phaser)
{
  GdamArg arg[2];

  arg[0].name = "min_delay";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 0) = phaser->min_hertz;
  arg[1].name = "max_delay";
  arg[1].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 1) = phaser->max_hertz;

  gdam_model_set_args (GDAM_MODEL (phaser), DELAY_RANGE, 2, arg);
}
static inline void maybe_dispense_feedback_request(GdamPhaser* phaser)
{
  GdamArg arg[1];

  arg[0].name = "feedback";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 0) = phaser->feedback;

  gdam_model_set_args (GDAM_MODEL (phaser), FEEDBACK, 1, arg);
}
static inline void maybe_dispense_depth_request(GdamPhaser* phaser)
{
  GdamArg arg[1];

  arg[0].name = "depth";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 0) = phaser->depth;

  gdam_model_set_args (GDAM_MODEL (phaser), DEPTH, 1, arg);
}

static inline void maybe_dispense_lfo_hertz_request(GdamPhaser* phaser)
{
  GdamArg arg[1];

  arg[0].name = "rate";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 0) = phaser->lfo_hertz;

  gdam_model_set_args (GDAM_MODEL (phaser), LFO_HERTZ, 1, arg);
}



static int gdam_phaser_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
  GdamPhaser *phaser = GDAM_PHASER (filter);
  GdamRequest *request;
  float volume;
  GdamModel *model = GDAM_MODEL (filter);
  gtk_object_ref (GTK_OBJECT (phaser));
  model->source_id = gdam_channel_new_local_id (model->channel);
  request = gdam_request_create_source ("GdasPhaser",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					phaser,
					6,
					"min_delay",
					GDAM_ARG_TYPE_DOUBLE,
					phaser->min_hertz,
					"max_delay",
					GDAM_ARG_TYPE_DOUBLE,
					phaser->max_hertz,
					"feedback",
					GDAM_ARG_TYPE_DOUBLE,
					phaser->feedback,
					"depth",
					GDAM_ARG_TYPE_DOUBLE,
					phaser->depth,
					"rate",
					GDAM_ARG_TYPE_DOUBLE,
					phaser->lfo_hertz,
					"filtered",
					GDAM_ARG_TYPE_INT, new_source_id);
  return gdam_request_builder_append_create (builder, request);
}


static void gdam_phaser_arg_set_min_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  phaser->min_hertz = GDAM_ARG_DOUBLE (arg);
  maybe_dispense_delay_range_request (phaser);
}
static void gdam_phaser_arg_get_min_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  GDAM_ARG_DOUBLE (arg) = phaser->min_hertz;
}
static void gdam_phaser_arg_set_max_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  phaser->max_hertz = GDAM_ARG_DOUBLE (arg);
  maybe_dispense_delay_range_request (phaser);
}
static void gdam_phaser_arg_get_max_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  GDAM_ARG_DOUBLE (arg) = phaser->max_hertz;
}
static void gdam_phaser_arg_set_feedback(GtkObject* object, GdamArg* arg)
{
  gdouble feedback;
  GdamPhaser *phaser = GDAM_PHASER (object);
  feedback = GDAM_ARG_DOUBLE (arg);
  feedback = CLAMP (feedback, 0.0, 1.0);
  phaser->feedback = feedback;
  maybe_dispense_feedback_request (phaser);
}
static void gdam_phaser_arg_get_feedback(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  gdouble feedback;
  GDAM_ARG_DOUBLE (arg) = phaser->feedback;
}

static void gdam_phaser_arg_set_depth(GtkObject* object, GdamArg* arg)
{
  gdouble depth;
  GdamPhaser *phaser = GDAM_PHASER (object);
  depth = GDAM_ARG_DOUBLE (arg);
  depth = CLAMP (depth, 0.0, 1.0);
  phaser->depth = depth;
  maybe_dispense_depth_request (phaser);
}
static void gdam_phaser_arg_get_depth(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  GDAM_ARG_DOUBLE (arg) = phaser->depth;
}
static void gdam_phaser_arg_set_lfo_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  phaser->lfo_hertz = GDAM_ARG_DOUBLE (arg);
  maybe_dispense_lfo_hertz_request (phaser);
}

static void gdam_phaser_arg_get_lfo_hertz(GtkObject* object, GdamArg* arg)
{
  GdamPhaser *phaser = GDAM_PHASER (object);
  GDAM_ARG_DOUBLE (arg) = phaser->lfo_hertz;
}


static void gdam_phaser_class_init(GdamPhaserClass* klass)
{
  GdamFilterClass *filter_class;
  GdamModelClass *model_class;
  GtkObjectClass *object_class;
  filter_class = GDAM_FILTER_CLASS (klass);
  model_class = GDAM_MODEL_CLASS (klass);
  object_class = GTK_OBJECT_CLASS (klass);
  model_class->num_arg_sets = LAST_ARG_SET;
  filter_class->append_start_request = gdam_phaser_append_start_request;
  gdam_arg_register (object_class,
		     "min_hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_phaser_arg_set_min_hertz,
		     gdam_phaser_arg_get_min_hertz,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "max_hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_phaser_arg_set_max_hertz,
		     gdam_phaser_arg_get_max_hertz,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "feedback",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_phaser_arg_set_feedback,
		     gdam_phaser_arg_get_feedback,
		     NULL, "-5,5", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "depth",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_phaser_arg_set_depth,
		     gdam_phaser_arg_get_depth, NULL, "0,1", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "lfo_hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_phaser_arg_set_lfo_hertz,
		     gdam_phaser_arg_get_lfo_hertz,
		     NULL, "0.01,440", GDAM_ARG_VITAL);
}

