/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducermp3click.h"
#include "../gdamproducermp3.h"
#include "../gdamsongdatabase.h"
#include "../gdambeatinfo.h"
#include <gdam/common/gdamplugin.h>
#include <math.h>
#include <gtk/gtksignal.h>
#include <gdam/common/gdamsample.h>
#include <stdlib.h>

static void     gdam_producer_mp3_click_real_set_filename
					   (GdamProducerMp3Click* mp3_click,
					    const char*      filename);
static int
gdam_producer_mp3_click_play_request (GdamProducer*       producer,
			              GdamPlayInfo*       play_info,
				      int                 subsource_id,
				      GdamRequestBuilder* builder);
static void
gdam_producer_mp3_click_stop_request (GdamProducer*       producer,
				      GdamRequestBuilder* builder);
static void     gdam_producer_mp3_click_destroy  (GtkObject*       object);

static GtkObjectClass* parent_class = NULL;

static void build_time_vols(gint64* time_vols,
			    gdouble beat_period,
			    gdouble click_volume);
static void refresh_fade_array(GdamProducerMp3Click* mp3);

enum {
	KEYPOINTS,
	LAST_ARG
};

#define ATTACK_SAMPLES		50
#define DECAY_SAMPLES		300


static void beat_info_to_data(GdamBeatInfo* beat_info, double* beat_data,
			      gdouble position)
{
  gdouble bpm, first_beat;
  bpm = gdam_beat_info_get_bpm (beat_info, position);
  first_beat = gdam_beat_info_get_first_beat (beat_info, position);
  g_return_if_fail (bpm > 0.001);
  beat_data[0] = first_beat;
  beat_data[1] = 60.0 / bpm;
}

void gdam_producer_mp3_click_set_beat_times
                                   (GdamProducerMp3Click* producer_mp3_click,
                                    int                   num_beats,
                                    const gdouble*        beats)
{
  gdouble *beat_times;
  beat_times = g_memdup (beats, num_beats * sizeof (gdouble));

  g_free (producer_mp3_click->beat_times);
  producer_mp3_click->beat_times = beat_times;
  producer_mp3_click->num_beats = num_beats;

  refresh_fade_array (producer_mp3_click);
}

static void gdam_producer_mp3_click_arg_set_volume(GtkObject* object, 
						   GdamArg* arg)
{
  GDAM_PRODUCER_MP3_CLICK (object)->click_volume = GDAM_ARG_DOUBLE (arg);
  refresh_fade_array (GDAM_PRODUCER_MP3_CLICK (object));
}

static void gdam_producer_mp3_click_arg_set_playing_pos(GtkObject* object, 
						        GdamArg* arg)
{
}

static void gdam_producer_mp3_click_arg_get_playing_pos(GtkObject* object, 
						        GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = 0;
}


static void gdam_producer_mp3_click_arg_get_volume(GtkObject* object, 
						   GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAM_PRODUCER_MP3_CLICK (object)->click_volume;
}

static void gdam_producer_mp3_click_arg_set_filename(GtkObject* object, GdamArg* arg)
{
  gdam_producer_mp3_click_real_set_filename (GDAM_PRODUCER_MP3_CLICK (object),
					     GDAM_ARG_STRING (arg));
}
static void gdam_producer_mp3_click_arg_get_filename(GtkObject* object, GdamArg* arg)
{
  GdamProducerMp3Click *producer_mp3_click = GDAM_PRODUCER_MP3_CLICK (object);
  GDAM_ARG_STRING (arg) = g_strdup (producer_mp3_click->filename);
}

static gboolean gdam_producer_mp3_click_real_set_beatinfo (GdamModel* model,
							const GdamBeatInfo* new)
{
  gboolean rv = GDAM_MODEL_CLASS (parent_class)->set_beat_info (model, new);
  refresh_fade_array (GDAM_PRODUCER_MP3_CLICK (model));
  return rv;
}

static void gdam_producer_mp3_click_class_init(GdamProducerMp3ClickClass* klass)
{
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  producer_class->play_request = gdam_producer_mp3_click_play_request;
  producer_class->stop_request = gdam_producer_mp3_click_stop_request;
  model_class->set_beat_info = gdam_producer_mp3_click_real_set_beatinfo;
  GTK_OBJECT_CLASS (klass)->destroy = gdam_producer_mp3_click_destroy;
  parent_class = gtk_type_class (GDAM_TYPE_PRODUCER);
  gdam_arg_register (object_class,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdam_producer_mp3_click_arg_set_filename,
		     gdam_producer_mp3_click_arg_get_filename,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "click_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_mp3_click_arg_set_volume,
		     gdam_producer_mp3_click_arg_get_volume,
		     NULL, "0,2", GDAM_ARG_VITAL);
  /* XXX: if this is useful, it should be put in producer */
  gdam_arg_register (object_class,
		     "playing_position",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_producer_mp3_click_arg_set_playing_pos,
		     gdam_producer_mp3_click_arg_get_playing_pos,
		     NULL, ",", 0);
  model_class->num_arg_sets = LAST_ARG;
}

static void gdam_producer_mp3_click_init(GdamProducerMp3Click* producer_mp3_click)
{
  producer_mp3_click->fade_envelope = 64;
  producer_mp3_click->click_volume = 1.0;
}

GtkType gdam_producer_mp3_click_get_type()
{
  static GtkType mp3_click_type = 0;
  if (!mp3_click_type)
    {
      static const GtkTypeInfo mp3_click_info = {
	"GdamProducerMp3Click",
	sizeof (GdamProducerMp3Click),
	sizeof (GdamProducerMp3ClickClass),
	(GtkClassInitFunc) gdam_producer_mp3_click_class_init,
	(GtkObjectInitFunc) gdam_producer_mp3_click_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      mp3_click_type = gtk_type_unique (GDAM_TYPE_PRODUCER, &mp3_click_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return mp3_click_type;
}

GdamProducerMp3Click* gdam_producer_mp3_click_new(GdamChannel*	channel)
{
  GdamProducerMp3Click *rv;
  rv = GDAM_PRODUCER_MP3_CLICK (gtk_type_new (GDAM_TYPE_PRODUCER_MP3_CLICK));
  ((GdamModel *) rv)->channel = channel;
  return rv;
}

static void build_time_vols(gint64* time_vols,
			    gdouble beat_period,
			    gdouble click_volume)
{
  int pulse_left = ATTACK_SAMPLES;
  int pulse_right = DECAY_SAMPLES;
  int beat_length = (int) (44100.0 * beat_period);
  GdamFixed peak_vol;
  peak_vol = (GdamFixed) ((double) GdamFixed_ONE * click_volume);
  time_vols[0] = 0;
  time_vols[1] = peak_vol;
  time_vols[2] = pulse_right;
  time_vols[3] = 0;
  time_vols[4] = beat_length - pulse_left;
  time_vols[5] = 0;
  time_vols[6] = beat_length;
  time_vols[7] = peak_vol;
}


static void got_mp3_subsource(int id, gpointer user_data)
{
  GDAM_PRODUCER_MP3_CLICK (user_data)->mp3_id = id;
}

static void didnt_get_mp3_subsource(gpointer user_data)
{
  GDAM_PRODUCER_MP3_CLICK (user_data)->mp3_id = 0;
  g_message ("error creating mp3");
}

static void fader_failure(gpointer user_data)
{
  g_warning ("fader failure!");
  GDAM_PRODUCER_MP3_CLICK (user_data)->fader_id = 0;
}

static void fader_success(int id, gpointer user_data)
{
  gdam_debug ("fader success, id=%d", id);
  GDAM_PRODUCER_MP3_CLICK (user_data)->fader_id = id;
}

static void wavetable_failure(gpointer user_data)
{
  g_warning ("wavetable failure!");
}

static void wavetable_success(int id, gpointer user_data)
{
  gdam_debug ("wavetable success, id=%d", id);
}

static gint sort_doubles(gconstpointer a, gconstpointer b)
{
  gdouble val_a = *(const gdouble *) a;
  gdouble val_b = *(const gdouble *) b;
  return (val_a < val_b) ? (-1) : (val_a > val_b) ? (+1) : 0;
}
static gint sort_int64s(gconstpointer a, gconstpointer b)
{
  gint64 val_a = *(const gint64 *) a;
  gint64 val_b = *(const gint64 *) b;
  return (val_a < val_b) ? (-1) : (val_a > val_b) ? (+1) : 0;
}


static gint64* build_arbitrary_time_vols(GdamProducerMp3Click* click,
                                         int*                  num_out)
{
  int num_pair = click->num_beats * 3 + 1;
  gint64 *rv = g_new (gint64, num_pair * 2);
  int i, at, first_index;
  GSList *peaks;
  /* Strategy:  make the keypoints (with just high/low),
     sort them,
     and add if they are overlapping... */

  /* Build the sorted list of peaks. */
  peaks = NULL;
  for (i = 0; i < click->num_beats; i++)
    peaks = g_slist_prepend (peaks, &click->beat_times[i]);
  peaks = g_slist_sort (peaks, sort_doubles);

  /* Build the sorted keypoints array. */
  at = 0;
  rv[at++] = 0;
  rv[at++] = 0;
  for (i = 0; i < click->num_beats; i++)
    {
      gint64 time = click->beat_times[i] * 44100.0;
      rv[at++] = time - ATTACK_SAMPLES;
      rv[at++] = 0;
      rv[at++] = time;
      rv[at++] = 0;
      rv[at++] = time + DECAY_SAMPLES;
      rv[at++] = 0;
    }
  qsort (rv, at / 2, sizeof (gint64) * 2, sort_int64s);
  *num_out = at;

  /* Now scan through each peak and add the volumes that peak
   * contributes to `rv' */
  first_index = 0;
  while (peaks)
    {
      gdouble val = *(gdouble *) peaks->data;
      gint64 time = val * 44100.0;
      peaks = g_slist_remove (peaks, peaks->data);

      for (; first_index < num_pair; first_index++)
	if (rv[first_index * 2] >= time - ATTACK_SAMPLES)
	  break;
      g_message ("for beat at time %0.5g, using index %d", val, first_index);
      for (i = first_index; i < num_pair; i++)
	{
	  int vol;
	  int offset;
	  offset = (int) (rv[i * 2] - time);
	  if (offset > DECAY_SAMPLES)
	    break;
	  if (offset < 0)
	    {
	      /* Attack contribution. */
	      GdamSampleProduct val;
	      val = -offset;
	      val *= GdamFixed_ONE;
	      val /= ATTACK_SAMPLES;
	      vol = GdamFixed_ONE - val;
	    }
	  else
	    {
	      /* Decay contribution. */
	      GdamSampleProduct val;
	      val = offset;
	      val *= GdamFixed_ONE;
	      val /= DECAY_SAMPLES;
	      vol = GdamFixed_ONE - val;
	    }
	  g_return_val_if_fail (vol >= 0, NULL);
	  rv[i * 2 + 1] += vol;
	}
      g_message ("stopped at %d", i);
    }

  return rv;
}

static void
mp3_click_position_callback(GdamTime time, gpointer user_data)
{
  GdamProducerMp3Click *mp3_click;
  float pos;
  GdamArg arg;
  arg.name = "playing_position";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  mp3_click = GDAM_PRODUCER_MP3_CLICK (user_data);
  pos = (double) time / 44100.0;
  //pos += mp3_click->last_play_position;
  gdam_producer_set_position (GDAM_PRODUCER (mp3_click), pos);
  GDAM_ARG_DOUBLE (&arg) = pos;
  gdam_arg_set_object_arg (GTK_OBJECT (mp3_click), &arg);
}

static void 
position_callback_connect_success(int pos_sub_id, gpointer user_data)
{
  GdamProducerMp3Click *mp3_click;
  mp3_click = GDAM_PRODUCER_MP3_CLICK (user_data);
  mp3_click->position_sub_id = pos_sub_id;
  gdam_channel_register_handler (GDAM_MODEL (mp3_click)->channel,
				 gdam_channel_time_handler_new (pos_sub_id,
								FALSE,
								mp3_click_position_callback,
								mp3_click));
}

static void
gdam_producer_mp3_click_stop_request (GdamProducer*       producer,
				      GdamRequestBuilder* builder)
{
  GdamProducerMp3Click *mp3_click;
  mp3_click = GDAM_PRODUCER_MP3_CLICK (producer);
  if (mp3_click->position_sub_id)
    {
      gdam_request_builder_append_generic (builder,
					   gdam_request_unsubscribe
					   (mp3_click->position_sub_id, NULL,
					    NULL, NULL));
      mp3_click->position_sub_id = 0;
    }
  GDAM_PRODUCER_CLASS (parent_class)->stop_request (producer, builder);
}

static int
gdam_producer_mp3_click_play_request (GdamProducer*       producer,
			              GdamPlayInfo*       play_info,
				      int                 subsource_id,
				      GdamRequestBuilder* builder)
{
  GdamProducerMp3Click *mp3_click = GDAM_PRODUCER_MP3_CLICK (producer);
  GdamModel *model;
  GdamChannel *channel;
  int num_args;
  int mp3_id = 0, wavetable_id = 0;
  GdamArg args[5];
  gdouble position = play_info ? play_info->position : 0.0;

  g_return_val_if_fail (mp3_click != NULL, 0);
  if (!mp3_click->filename)
    {
      g_warning ("gdam_producer_mp3_click_play: no filename");
      return 0;
    }
  model = GDAM_MODEL (mp3_click);
  channel = model->channel;

  mp3_click->mp3_id = 0;
  mp3_click->fader_id = 0;
  mp3_click->last_play_position = position;
  gdam_debug ("producermp3_click: not in sync.");

  args[0].type = GDAM_ARG_TYPE_STRING;
  args[0].name = "filename";
  GDAM_ARG_STRING (args + 0) = mp3_click->filename;
  args[1].type = GDAM_ARG_TYPE_INT64;
  args[1].name = "position";
  GDAM_ARG_INT64 (args + 1) = (gint64) (position * 44100.0);

  num_args = 2;
  if (mp3_click->buffer_size > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "buffer_size";
      GDAM_ARG_INT (args + num_args) = mp3_click->buffer_size;
      num_args++;
    }
  if (mp3_click->fade_envelope > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT;
      args[num_args].name = "fade_envelope";
      GDAM_ARG_INT (args + num_args) = mp3_click->fade_envelope;
      num_args++;
    }
  if (play_info->length > 0)
    {
      args[num_args].type = GDAM_ARG_TYPE_INT64;
      args[num_args].name = "length";
      GDAM_ARG_INT64 (args + num_args) =
	(gint64) (play_info->length * 44100.0);
      num_args++;
    }
  args[num_args].name = "exec_command";
  args[num_args].type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (args + num_args) 
    = (char *)gdam_mp3_exec_cmd_from_filename (mp3_click->filename);
  num_args++;

  mp3_id = gdam_channel_new_local_id (channel);
  gdam_request_builder_append_create  (builder,
				       gdam_request_create_source_v
				       ("GdasSeekableMp3",
					mp3_id,
					num_args, args,
					got_mp3_subsource,
					didnt_get_mp3_subsource,
					mp3_click));
  g_return_val_if_fail (mp3_id != 0, 0);
  gdam_request_builder_append_generic (builder,
				       gdam_request_subscribe_source_timer
				       (mp3_id, 200,
					position_callback_connect_success,
					NULL, mp3_click));
  if (mp3_click->beat_times != NULL
      || gdam_beat_info_get_bpm (&model->beat_info, position) > 0)
    {
      gint64 *time_vols;
      /* XXX: make beat_info have num beat information */
      int num_beats = 100;
      int num_time_vols = (3 * num_beats + 2) * 2;
      gdouble beat_data[2];
      gdouble bpm;
      GdamBeatInfo *beat_info = &model->beat_info;
      int num_args;
      GdamArg args[4];
      char *mode;

      wavetable_id = gdam_channel_new_local_id (channel);
      gdam_request_builder_append_create 
				(builder,
				 gdam_request_create_source
				 ("GdasWavetable",
				  wavetable_id,
				  wavetable_success,
				  wavetable_failure,
				  NULL, 4,
				  "buffer_id",
				  GDAM_ARG_TYPE_INT,
				  1, "hertz",
				  GDAM_ARG_TYPE_DOUBLE,
				  440.0,
				  "use_overall_position",
				  GDAM_ARG_TYPE_INT,
				  1, "position",
				  GDAM_ARG_TYPE_INT64,
				  (gint64) (position *
					    44100.0)));
      g_return_val_if_fail (wavetable_id != 0, 0);

      bpm = gdam_beat_info_get_bpm (beat_info, position);
      if (mp3_click->num_beats <= 0 && bpm > 0.001)
	{
	  num_time_vols = 8;
	  time_vols = g_new (gint64, num_time_vols);
	  beat_info_to_data (beat_info, beat_data, position);
	  build_time_vols (time_vols, beat_data[1], mp3_click->click_volume);
	  mode = "loops";
	}
      else
	{
	  time_vols = build_arbitrary_time_vols (mp3_click, &num_time_vols);
	  mode = "shelf";
	  beat_data[0] = 0;
	  beat_data[1] = 100000;
	}

      num_args = 3;
      args[0].name = "keypoints";
      args[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
      GDAM_ARG_INT64_ARRAY (args + 0).length = num_time_vols;
      GDAM_ARG_INT64_ARRAY (args + 0).array = time_vols;
      args[1].name = "mode";
      args[1].type = GDAM_ARG_TYPE_STRING;
      GDAM_ARG_STRING (args + 1) = mode;
      args[2].name = "filtered";
      args[2].type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (args + 2) = wavetable_id;
      if (bpm > 0.001)
	{
	  num_args = 4;
	  args[3].name = "phase_sync";
	  args[3].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
	  GDAM_ARG_DOUBLE_ARRAY (args + 3).length = 2;
	  GDAM_ARG_DOUBLE_ARRAY (args + 3).array = beat_data;
	}

      mp3_click->fader_id = gdam_channel_new_local_id (channel);
      gdam_request_builder_append_create (builder,
					  gdam_request_create_source_v
					  ("GdasFader",
					   mp3_click->fader_id,
					   num_args, args,
					   fader_success, fader_failure,
					   mp3_click));
      g_return_val_if_fail (mp3_click->fader_id != 0, 0);
      g_free (time_vols);
    }
  {
    gint64 mixed[4];
    GdamTime tim = (GdamTime) (44100.0 * position);

    mixed[0] = 0;
    //mixed[0] = (gint64) (position * 44100.0);
    mixed[1] = mp3_id;
    mixed[2] = 0;
    //mixed[2] = (gint64) (position * 44100.0);
    mixed[3] = mp3_click->fader_id;

    if (mp3_id == 0) 
	    g_warning("missing mp3 id in mp3_click_play_request");

    model->source_id = gdam_channel_new_local_id (channel);
    gdam_request_builder_append_create 
			  (builder,
			   gdam_request_create_source
			   ("GdasMix", 
			    model->source_id,
			    gdam_model_got_subsource,
			    gdam_model_didnt_get_subsource,
			    mp3_click, 2, "mixed",
			    GDAM_ARG_TYPE_INT64_ARRAY,
			    mp3_click->fader_id == 0 ? 2 : 4,
			    mixed, "mode",
			    GDAM_ARG_TYPE_STRING,
			    "first_ended"));
    //gdam_request_builder_append_generic 
      			//(builder, 
			 //gdam_request_seek_relative (model->source_id,
						     //tim,
						     //NULL, NULL, NULL));

  }

  gdam_debug ("Creating producing mp3_click with position %f length %f",
	      position, play_info->length);
  gtk_object_ref (GTK_OBJECT (mp3_click));

  return model->source_id;
}

static void refresh_fade_array(GdamProducerMp3Click* mp3)
{
  GdamModel *model = GDAM_MODEL (mp3);
  gint64 *time_vols;
  GdamArg arg[2];

  if (model->source_id)
    {
      GdamBeatInfo *beat_info;
      gdouble beat_data[2];
      int num_time_vols;
      beat_info = &model->beat_info;
      beat_info_to_data (beat_info, beat_data, 0.0);
      if (!mp3->fader_id)
	{
	  g_warning ("missing fader id in refresh_fade_array!");
	  return;
	}
      if (mp3->num_beats <= 0)
	{
	  num_time_vols = 8;
	  time_vols = g_new (gint64, num_time_vols);
	  build_time_vols (time_vols, beat_data[1], mp3->click_volume);
	}
      else
	{
	  time_vols = build_arbitrary_time_vols (mp3, &num_time_vols);
	  beat_data[0] = 0;
	  beat_data[1] = 100000;
	}
      arg[0].name = "keypoints";
      arg[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
      GDAM_ARG_INT64_ARRAY (arg + 0).length = num_time_vols;
      GDAM_ARG_INT64_ARRAY (arg + 0).array = time_vols;
      arg[1].name = "phase_sync";
      arg[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
      GDAM_ARG_DOUBLE_ARRAY (arg + 1).length = 2;
      GDAM_ARG_DOUBLE_ARRAY (arg + 1).array = beat_data;
      gdam_model_set_args_full (model, mp3->fader_id, KEYPOINTS, 2, arg);
    }
}


static void gdam_producer_mp3_click_real_set_filename(GdamProducerMp3Click* mp3_click,
                                                const char*      filename)
{
  GdamSongDatabase *database = gdam_song_database_system ();
  GdamSongInfo *info = NULL;
  GdamBeatInfo *beat_info;
  gboolean was_playing = FALSE;
  gdouble position = 0;
  /* TODO: might this permit multiple requests? */
  if (GDAM_MODEL (mp3_click)->source_id != 0)
    {
      was_playing = TRUE;
      gdam_producer_stop (GDAM_PRODUCER (mp3_click));
    }
  if (mp3_click->filename)
    g_free (mp3_click->filename);
  mp3_click->filename = g_strdup (filename);
  if (database)
    info = gdam_song_database_find (database, filename);
  if (info)
    {
      gdam_producer_update_length (GDAM_PRODUCER (mp3_click), info->length);
      position = gdam_beat_info_get_first_beat (info->beat_info, 0);
      beat_info = gdam_beat_info_copy (info->beat_info);
    }
  else
    {
      beat_info = gdam_beat_info_new ();
      /* TODO: get rid of this */
      gdam_beat_info_set_first_beat (beat_info, 0);
      gdam_beat_info_set_bpm (beat_info, 90);
    }
  gdam_producer_set_position (GDAM_PRODUCER (mp3_click), position);
  gdam_model_set_beat_info (GDAM_MODEL (mp3_click), beat_info);
  gdam_beat_info_destroy (beat_info);
}

static void gdam_producer_mp3_click_destroy(GtkObject* object)
{
  GdamProducerMp3Click *mp3_click = GDAM_PRODUCER_MP3_CLICK (object);
  g_free (mp3_click->filename);
  (*parent_class->destroy) (object);
}

void gdam_producer_mp3_click_set_filename(GdamProducerMp3Click*  	producer_mp3_click,
				    const char* 	filename)
{
  GdamArg arg;
  arg.name = "filename";
  arg.type = GDAM_ARG_TYPE_STRING;
  GDAM_ARG_STRING (&arg) = (char *) filename;
  gdam_arg_set_object_arg (GTK_OBJECT (producer_mp3_click), &arg);
}

void gdam_producer_mp3_click_set_buffer_size(GdamProducerMp3Click* producer_mp3_click,
                                       int              buffer_size)
{
  int source_id;
  g_return_if_fail (producer_mp3_click != NULL);
  source_id = producer_mp3_click->mp3_id;
  if (GDAM_MODEL (producer_mp3_click)->source_id && source_id && buffer_size)
    {
      GdamArg *arg = g_new (GdamArg, 1);
      arg->name = g_strdup ("buffer_size");
      GDAM_ARG_INT (arg) = buffer_size;
      arg->type = GDAM_ARG_TYPE_INT;
      gdam_channel_append_request (GDAM_MODEL (producer_mp3_click)->channel,
				   gdam_request_set_args (source_id,
							  1,
							  arg,
							  NULL, NULL, NULL));
    }
  producer_mp3_click->buffer_size = buffer_size;
}
