/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamproducerwalker.h"
#include <gdam/common/gdamplugin.h>
#include "../gdambuffermanager.h"
#include "../gdambeatinfo.h"
#include <gtk/gtksignal.h>
#include <gdam/common/gdamplugin.h>
#include "../gdamutil.h"

static int gdam_producer_walker_play_request(GdamProducer*       producer,
					        GdamPlayInfo*       play_info,
					        int                 source_id,
					        GdamRequestBuilder* builder);

static void
gdam_producer_walker_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id);
static void 
gdam_producer_walker_init	(GdamProducerWalker* producer_walker)
{
  producer_walker->mean_length = 1.0;
  producer_walker->buffer_id = -1;
  producer_walker->overall_rate = 0.0;
}

static void maybe_dispense_request(GdamProducerWalker* walker)
{
  GdamArg arg[6];
  GdamRequest *request;
  GdamModel *component = GDAM_MODEL (walker);
  gdouble bpm;
  gdouble firstbeat;
  gdouble forward_likelihood;
  if (!component->source_id)
    return;
  bpm = gdam_beat_info_get_bpm (&component->beat_info, 0);
  firstbeat = gdam_beat_info_get_first_beat (&component->beat_info, 0);
  forward_likelihood = 0.5 + walker->overall_rate / 2;

  arg[0].name = "mean_length";
  arg[0].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 0) = walker->mean_length * 44100.0;
  arg[1].name = "buffer_id";
  arg[1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_DOUBLE (arg + 1) = walker->mean_length * 44100.0;
  arg[2].name = "mode";
  arg[2].type = GDAM_ARG_TYPE_STRING;
  if (walker->mode == 0)
    GDAM_ARG_STRING (arg + 2) = g_strdup ("random");
  else if (walker->mode == 1)
    GDAM_ARG_STRING (arg + 2) = g_strdup ("nonsync");
  else if (walker->mode == 2)
    GDAM_ARG_STRING (arg + 2) = g_strdup ("sync");
  else
    g_warning ("invalid mode %d for walker", walker->mode);
  arg[3].name = "forward_likelihood";
  arg[3].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 3) = forward_likelihood;
  arg[4].name = "bpm";
  arg[4].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 4) = bpm;
  arg[5].name = "first_beat_time";
  arg[5].type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (arg + 5) = firstbeat;

  request = gdam_request_set_args (component->source_id, bpm > 1 ? 6 : 4,
				   arg,
				   gdam_do_nothing, gdam_do_nothing, NULL);

  gdam_channel_append_request (component->channel, request);
}

static void gdam_walker_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GdamArg newarg;
  gdouble delay;
  gdouble bpm;
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (walker)->beat_info, 0);
  if (bpm >= 1)
    delay = GDAM_ARG_DOUBLE (arg) * 60.0 / bpm;
  else
    delay = 0;
  newarg.name = "mean_length";
  newarg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&newarg) = delay;
  gdam_arg_set_object_arg (object, &newarg);
}


#define GDAM_MIN_WALK 0.001
static void gdam_walker_arg_set_mean_length(GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  walker->mean_length = GDAM_ARG_DOUBLE (arg);
  if (walker->mean_length < GDAM_MIN_WALK)
    {
      g_warning ("limiting walk length to %f secs minimum.", GDAM_MIN_WALK);
      walker->mean_length = GDAM_MIN_WALK;
    }
  maybe_dispense_request (walker);
}
static void gdam_walker_arg_get_mean_length(GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GDAM_ARG_DOUBLE (arg) = walker->mean_length;
}

static void 
gdam_walker_arg_set_overall_rate (GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  walker->overall_rate = GDAM_ARG_DOUBLE (arg);
  g_message ("overall rate %f", walker->overall_rate);
  maybe_dispense_request (walker);
}
static void 
gdam_walker_arg_get_overall_rate (GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GDAM_ARG_DOUBLE (arg) = walker->overall_rate;
}

static void 
gdam_walker_arg_set_impose_rate (GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  walker->impose_overall_rate = GDAM_ARG_INT (arg);
  maybe_dispense_request (walker);
}
static void
gdam_walker_arg_get_impose_rate (GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GDAM_ARG_INT (arg) = walker->impose_overall_rate;
}

static void gdam_walker_arg_set_buffer_id(GtkObject* object, GdamArg* arg)
{
  /*GdamBeatInfo beat_info; */
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GdamBufferManagerInfo *info;
  GdamModel *model = GDAM_MODEL (walker);
  gint new_buffer_id = GDAM_ARG_INT (arg);
  //gdam_producer_stop(GDAM_PRODUCER(walker));

  gdam_buffer_manager_ref (new_buffer_id, model);
  if (walker->buffer_id > 0)
    gdam_buffer_manager_unref (walker->buffer_id, model);

  walker->buffer_id = new_buffer_id;
  info = gdam_buffer_manager_info_by_id (new_buffer_id);

  g_return_if_fail (info != NULL);
  walker->buffer_length = info->length;
  if (info->beat_info)
    {
      gdam_model_set_beat_info (GDAM_MODEL (walker), info->beat_info);
    }
  else
    {
      GdamBeatInfo *bi;
      bi = gdam_beat_info_new ();
      gdam_model_set_beat_info (GDAM_MODEL (walker), bi);
      gdam_beat_info_destroy (bi);
    }
  gdam_producer_update_length (GDAM_PRODUCER (walker),
			       (float) info->length / 44100.0);
  maybe_dispense_request (walker);
}

static void gdam_walker_arg_get_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  GDAM_ARG_INT (arg) = walker->buffer_id;
}

static void gdam_walker_arg_set_mode(GtkObject *object, GdamArg *arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  char *mode = GDAM_ARG_STRING (arg);
  if (strcmp (mode, "random") == 0)
    walker->mode = 0;
  else if (strcmp (mode, "nonsync") == 0)
    walker->mode = 1;
  else if (strcmp (mode, "sync") == 0)
    walker->mode = 2;
  else
    gdam_debug ("bad gdas_walker mode '%s'", mode);
  maybe_dispense_request (walker);
}

static void gdam_walker_arg_get_mode(GtkObject *object, GdamArg *arg)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (object);
  if (walker->mode == 0)
    GDAM_ARG_STRING (arg) = g_strdup ("random");
  else if (walker->mode == 1)
    GDAM_ARG_STRING (arg) = g_strdup ("nonsync");
  else if (walker->mode == 2)
    GDAM_ARG_STRING (arg) = g_strdup ("sync");
  else
    g_warning ("invalid mode %d for walker", walker->mode);
}


static void 
gdam_producer_walker_class_init	(GdamProducerWalkerClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamProducerClass *producer_class = GDAM_PRODUCER_CLASS (klass);
  producer_class->play_request = gdam_producer_walker_play_request;
  GDAM_MODEL_CLASS (klass)->detach_from_buffer
    = gdam_producer_walker_detach_from_buffer;
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdam_walker_arg_set_buffer_id,
		     gdam_walker_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mean_length",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_walker_arg_set_mean_length,
		     gdam_walker_arg_get_mean_length,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_walker_arg_set_beat_factor,
		     NULL,
		     NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_STRING,
		     gdam_walker_arg_set_mode,
		     gdam_walker_arg_get_mode, 
		     NULL,
		     "random,nonsync,sync",
		     GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "overall_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_walker_arg_set_overall_rate,
		     gdam_walker_arg_get_overall_rate,
		     NULL, "-1.0,1.0", 0);
  gdam_arg_register (object_class,
		     "impose_overall_rate",
		     GDAM_ARG_TYPE_INT,
		     gdam_walker_arg_set_impose_rate,
		     gdam_walker_arg_get_impose_rate,
		     NULL, "boolean", 0);
}

GtkType gdam_producer_walker_get_type()
{
  static GtkType producer_walker_type = 0;
  if (!producer_walker_type)
    {
      static const GtkTypeInfo producer_walker_info = {
	"GdamProducerWalker",
	sizeof (GdamProducerWalker),
	sizeof (GdamProducerWalkerClass),
	(GtkClassInitFunc) gdam_producer_walker_class_init,
	(GtkObjectInitFunc) gdam_producer_walker_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = gdam_plugin_get_type_by_name ("model",
						     "GdamProducer");
      producer_walker_type = gtk_type_unique (parent, &producer_walker_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return producer_walker_type;
}

static int gdam_producer_walker_play_request(GdamProducer*       producer,
					        GdamPlayInfo*       play_info,
					        int                 source_id,
					        GdamRequestBuilder* builder)
{
  GdamProducerWalker *walker = GDAM_PRODUCER_WALKER (producer);
  GdamChannel *channel;
  GdamModel *model;
  GdamRequest *request;
  gdouble position, bpm, firstbeat, forward_likelihood;
  const char* mode = "random";
  model = GDAM_MODEL (walker);
  bpm = gdam_beat_info_get_bpm (&model->beat_info, 0);
  firstbeat = gdam_beat_info_get_first_beat (&model->beat_info, 0);

  g_return_val_if_fail (walker != NULL, 0);

  if (walker->buffer_id < 0)
    {
      g_warning ("gdam_producer_walker_play: no buffer_id");
      return 0;
    }
  channel = model->channel;
  g_return_val_if_fail (channel != NULL, 0);

  position = play_info->position;

  gdam_debug ("Creating producing walker with position %f", position);
  gtk_object_ref (GTK_OBJECT (walker));
  model->source_id = gdam_channel_new_local_id (model->channel);
  if (walker->mode == 0)
    mode = "random";
  else if (walker->mode == 1)
    mode = "nonsync";
  else if (walker->mode == 2)
    mode = "sync";
  else
    g_warning ("invalid mode %d for walker", walker->mode);
  forward_likelihood = 0.5 + walker->overall_rate / 2;
  request = gdam_request_create_source ("GdasWalker",
					model->source_id,
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					walker,
					bpm > 1 ? 6 : 4,
					"buffer_id",
					GDAM_ARG_TYPE_INT,
					walker->buffer_id,
					"mean_length",
					GDAM_ARG_TYPE_DOUBLE,
					walker->mean_length * 44100.0,
					"mode",
					GDAM_ARG_TYPE_STRING,
					mode,
					"forward_likelihood",
					GDAM_ARG_TYPE_DOUBLE,
					forward_likelihood,
					"bpm",
					GDAM_ARG_TYPE_DOUBLE,
					bpm,
					"first_beat_time",
					GDAM_ARG_TYPE_DOUBLE,
					firstbeat);

  return gdam_request_builder_append_create (builder, request);
}

void gdam_producer_walker_set_buffer_id(GdamProducerWalker* walker, int id)
{
  GdamModel *model = GDAM_MODEL (walker);
  GdamArg arg;
  arg.name = "buffer_id";
  arg.type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (&arg) = id;
  gdam_arg_set_object_arg (GTK_OBJECT (model), &arg);
}

static void
gdam_producer_walker_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id)
{
  GdamProducerWalker *walker;
  walker = GDAM_PRODUCER_WALKER (model);
  if (walker->buffer_id != 0)
    gdam_producer_walker_set_buffer_id (walker, 0);
}
