/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamrangefinder.h"
#include <gdam/common/gdamplugin.h>
#include <ctype.h>

static void gdam_range_finder_init(GdamRangeFinder* range_finder)
{
  range_finder->device_number = -1;
}

static char* gdam_range_finder_event_stringify(GdamInputEvent* event)
{
  GdamInputEventRange *range = (GdamInputEventRange *) event;
  return g_strdup_printf ("rangefinder %d", range->input_id);
}

static guint gdam_range_finder_event_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamRangeFinderEvent",
    gdam_range_finder_event_stringify,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_range_get_type (),
       GDAM_TYPE_RANGE_FINDER, &type_info);
  gdam_plugin_check_version (__FILE__, GDAM_VERSION);
  return type;
}

static GdamInputEvent* gdam_range_finder_event(GdamRangeFinder* finder,
                                               int              device_number,
					       int              value)
{
  GdamInputEventRange *range;
  range = g_new (GdamInputEventRange, 1);
  range->event_type = gdam_range_finder_event_get_type ();
  range->input_device = GDAM_INPUT_DEVICE (finder);
  range->input_id = device_number;
  range->min_value = 0;
  range->max_value = (1 << 14);
  range->value = value;
  return (GdamInputEvent *) range;
}

static void got_number(GdamRangeFinder* range_finder)
{
  int cur_value = range_finder->cur_value;

  if (range_finder->device_number < 0)
    {
      range_finder->device_number = cur_value;
    }
  else
    {
      GdamInputEvent *event;
      event = gdam_range_finder_event (range_finder,
				       range_finder->device_number,
				       cur_value);
      gdam_input_trap_emit (NULL, event);
      range_finder->device_number = -1;
    }
}

static void parse_character(GdamRangeFinder* range_finder, int c)
{
  gboolean is_digit = isdigit (c);
  if (is_digit)
    {
      if (!range_finder->in_number)
	{
	  range_finder->in_number = TRUE;
	  range_finder->cur_value = 0;
	}
      range_finder->cur_value *= 10;
      range_finder->cur_value += (c - '0');
    }
  else
    {
      if (range_finder->in_number)
	{
	  got_number (range_finder);
	  range_finder->in_number = FALSE;
	}
    }
}

static gboolean gdam_range_finder_process(GdamInputDeviceFile *device_file,
                                          guint8              *midi_buf,
				          int		       num_read)
{
  GdamRangeFinder *range_finder = GDAM_RANGE_FINDER (device_file);
  int index;
  for (index = 0; index < num_read; index++)
    {
      int c = midi_buf[index];
      parse_character (range_finder, c);
    }
  return TRUE;
}


static void gdam_range_finder_class_init(GdamInputDeviceFileClass* klass)
{
  klass->process = gdam_range_finder_process;
}

GtkType gdam_range_finder_get_type()
{
  static GtkType range_finder_type = 0;
  if (!range_finder_type)
    {
      static const GtkTypeInfo range_finder_info = {
	"GdamRangeFinder",
	sizeof (GdamRangeFinder),
	sizeof (GdamRangeFinderClass),
	(GtkClassInitFunc) gdam_range_finder_class_init,
	(GtkObjectInitFunc) gdam_range_finder_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_INPUT_DEVICE_FILE;
      range_finder_type = gtk_type_unique (parent, &range_finder_info);
    }
  return range_finder_type;
}
