/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamspatialstereo.h"
#include <gdam/common/gdamarg.h>
#include <gdam/common/gdamplugin.h>

static void gdam_spatial_stereo_real_positions_change
					        (GdamSpatialStereo* spatial);
static void gdam_spatial_stereo_real_set_speed_of_sound     
                                                (GdamSpatialStereo* spatial,
	                                         gdouble            speed);
static void gdam_spatial_stereo_real_set_unit_gain_distance
                                                (GdamSpatialStereo* spatial,
	                                         gdouble            distance);
static int gdam_spatial_source_append_start_request
                                                (GdamFilter*       filter,
						 int               sub_id,
						 GdamRequestBuilder* builder);

static void     gdam_spatial_stereo_get_args(GdamSpatialStereo* spatial,
                                             int*               num_args_out,
					     GdamArg**          args_out)
{
  g_return_if_fail (GDAM_SPATIAL_STEREO (spatial) != NULL);
  *args_out = gdam_arg_new_from_valist (
					(*num_args_out = 6),
					"left_ear",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					spatial->left_ear_pos,
					"right_ear",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					spatial->right_ear_pos,
					"left_position",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					spatial->left_channel_pos,
					"right_position",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					3,
					spatial->right_channel_pos,
					"speed_of_sound",
					GDAM_ARG_TYPE_DOUBLE,
					spatial->speed_of_sound,
					"unit_gain_distance",
					GDAM_ARG_TYPE_DOUBLE,
					spatial->unit_gain_distance);
}

static void do_nothing(gpointer unused)
{
}

static void maybe_update_server_array(GdamSpatialStereo* spatial)
{
  GdamArg *args;
  int num_args;
  GdamRequest *request;
  int source_id = GDAM_MODEL (spatial)->source_id;
  if (source_id == 0)
    return;
  gdam_spatial_stereo_get_args (spatial, &num_args, &args);
  request = gdam_request_set_args (source_id,
				   num_args,
				   args, do_nothing, do_nothing, NULL);
  gdam_channel_append_request (GDAM_MODEL (spatial)->channel, request);
  gdam_arg_free_array (num_args, args);
}

static void gdam_spatial_stereo_init(GdamSpatialStereo* spatial)
{
  spatial->left_channel_pos[0] = -3.0;
  spatial->left_channel_pos[1] = 0.0;
  spatial->left_channel_pos[2] = 0.0;
  spatial->right_channel_pos[0] = +3.0;
  spatial->right_channel_pos[1] = 0.0;
  spatial->right_channel_pos[2] = 0.0;
  spatial->left_ear_pos[0] = -0.1;
  spatial->left_ear_pos[1] = 0.0;
  spatial->left_ear_pos[2] = 0.0;
  spatial->right_ear_pos[0] = +0.1;
  spatial->right_ear_pos[1] = 0.0;
  spatial->right_ear_pos[2] = 0.0;

  spatial->speed_of_sound = 332.0;
  spatial->unit_gain_distance = 1.0;
}

static void
gdam_spatial_stereo_class_init(GdamSpatialStereoClass* klass)
{
  klass->positions_change = gdam_spatial_stereo_real_positions_change;
  klass->set_speed_of_sound = gdam_spatial_stereo_real_set_speed_of_sound;
  klass->set_unit_gain_distance
    = gdam_spatial_stereo_real_set_unit_gain_distance;
  GDAM_FILTER_CLASS (klass)->append_start_request
    = gdam_spatial_source_append_start_request;
}

GtkType gdam_spatial_stereo_get_type()
{
  static GtkType spatial_stereo_type = 0;
  if (!spatial_stereo_type)
    {
      static const GtkTypeInfo spatial_stereo_info = {
	"GdamSpatialStereo",
	sizeof (GdamSpatialStereo),
	sizeof (GdamSpatialStereoClass),
	(GtkClassInitFunc) gdam_spatial_stereo_class_init,
	(GtkObjectInitFunc) gdam_spatial_stereo_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_FILTER;
      spatial_stereo_type = gtk_type_unique (parent, &spatial_stereo_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return spatial_stereo_type;
}

static int gdam_spatial_source_append_start_request
                                                (GdamFilter*       filter,
						 int               sub_id,
						 GdamRequestBuilder* builder)
{
  GdamArg *args;
  int num_args;
  GdamModel *model = GDAM_MODEL (filter);
  gdam_spatial_stereo_get_args (GDAM_SPATIAL_STEREO (filter),
				&num_args, &args);
  num_args++;
  args = g_renew (GdamArg, args, num_args);
  args[num_args - 1].name = g_strdup ("filtered");
  args[num_args - 1].type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (args + num_args - 1) = sub_id;
  gtk_object_ref (GTK_OBJECT (filter));
  model->source_id = gdam_channel_new_local_id (model->channel);
  /* TODO:  args is a memory leak here. */
  return gdam_request_builder_append_create (builder,
					     gdam_request_create_source_v
					     ("GdasSpatialStereo",
					      model->source_id, num_args,
					      args, gdam_model_got_subsource,
					      gdam_model_didnt_get_subsource,
					      filter));
}

static void gdam_spatial_stereo_real_positions_change
					        (GdamSpatialStereo* spatial)
{
  maybe_update_server_array (spatial);
}
static void gdam_spatial_stereo_real_set_speed_of_sound     
                                                (GdamSpatialStereo* spatial,
	                                         gdouble            speed)
{
  spatial->unit_gain_distance = speed;
  maybe_update_server_array (spatial);
}
static void gdam_spatial_stereo_real_set_unit_gain_distance
                                                (GdamSpatialStereo* spatial,
	                                         gdouble            distance)
{
  spatial->unit_gain_distance = distance;
  maybe_update_server_array (spatial);
}


void gdam_spatial_stereo_positions_change (GdamSpatialStereo* spatial)
{
  GdamSpatialStereoClass *klass;
  klass = GDAM_SPATIAL_STEREO_CLASS (GTK_OBJECT (spatial)->klass);
  (*klass->positions_change) (spatial);
}
void gdam_spatial_stereo_set_speed_of_sound     (GdamSpatialStereo* spatial,
	                                         gdouble            speed)
{
  GdamSpatialStereoClass *klass;
  klass = GDAM_SPATIAL_STEREO_CLASS (GTK_OBJECT (spatial)->klass);
  (*klass->set_speed_of_sound) (spatial, speed);
}
void gdam_spatial_stereo_set_unit_gain_distance (GdamSpatialStereo* spatial,
	                                         gdouble            distance)
{
  GdamSpatialStereoClass *klass;
  klass = GDAM_SPATIAL_STEREO_CLASS (GTK_OBJECT (spatial)->klass);
  (*klass->set_unit_gain_distance) (spatial, distance);
}


