/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <stdlib.h>
#include <gdam/common/gdamplugin.h>
#include "gdamspin.h"
#include <gdam/component/gdamutil.h>
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdammodel.h>
#include <gdam/component/gdaminfomanager.h>

#define MY_ACTION_LIST (GdkDragAction) GDK_ACTION_COPY

static void on_set_source(GdamSpin* spin, guint new_source_id);

/*
 *     ____ _                 ___       _ _   
 *    / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *   | |   | |/ _` / __/ __|  | || '_ \| | __|
 *   | |___| | (_| \__ \__ \  | || | | | | |_ 
 *    \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */                                            

/* Functions for Component class */
static gboolean gdam_spin_match_beat_info(GdamModel* component, 
					  const GdamBeatInfo* match_info);
/* Functions for EchoBase class */
static gboolean gdam_spin_refresh_echo_array(GdamEchoBase* echo_base);

static void gdam_spin_class_init(GdamSpinClass* klass);

static void gdam_spin_init(GdamSpin* spin)
{
  spin->beat_factor = 2;

  /* Trap our child's set_source signal so we know when it is changed */
  gtk_signal_connect_object_while_alive (GTK_OBJECT (spin),
					 "set_source",
					 GTK_SIGNAL_FUNC (on_set_source),
					 GTK_OBJECT (spin));
}

GtkType gdam_spin_get_type()
{
  static GtkType spin_type = 0;
  if (!spin_type)
    {
      static const GtkTypeInfo spin_info = {
	"GdamSpin",
	sizeof (GdamSpin),
	sizeof (GdamSpinClass),
	(GtkClassInitFunc) gdam_spin_class_init,
	(GtkObjectInitFunc) gdam_spin_init,
	/* reserved 1 */ NULL,
	/* reserved 2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      spin_type = gtk_type_unique (GDAM_TYPE_ECHO_BASE, &spin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return spin_type;
}


/*  _____                 _   _                 
 * |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 * |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */                                            

static gboolean gdam_spin_refresh_echo_array(GdamEchoBase* echo_base)
{
  float delay = 1.0;
  float bpm;
  GdamSpin *spin = GDAM_SPIN (echo_base);
  spin->echo_base.num_echoes = 2;
  spin->echo_base.echoes = g_renew (gdouble,
				    spin->echo_base.echoes,
				    spin->echo_base.num_echoes * 2);

  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (spin)->beat_info, 0);
  if (bpm >= 1)
    {
      delay = spin->beat_factor * 60.0 / bpm;
    }
  else 
    {
      delay = spin->beat_factor * 1.0;
    }
  /* Delays */
  spin->echo_base.echoes[1] = (int) 0;
  spin->echo_base.echoes[3] = (int) (delay * 44100.0);
  /* Volumes */
  spin->echo_base.echoes[0] = spin->spun ? 0.0 : 1.0;
  spin->echo_base.echoes[2] = spin->spun ? 1.0 : 0.0;
  /* Reverb? */
  spin->echo_base.do_reverb = spin->spun ? TRUE : FALSE;

  return TRUE;
}

static void on_set_source(GdamSpin* spin, guint new_source_id)
{
  if (new_source_id == 0)
    {
      GdamArg arg;
      arg.name = "spun";
      arg.type = GDAM_ARG_TYPE_INT;
      GDAM_ARG_INT (&arg) = 0;
      gdam_arg_set_object_arg (GTK_OBJECT (spin), &arg);
    }
}


static gboolean gdam_spin_match_beat_info(GdamModel*          model, 
					  const GdamBeatInfo* match_info)
{
  GdamSpin *spin = GDAM_SPIN (model);
  GdamEchoBaseClass *echo_base_class;
  float bpm;
  echo_base_class = GDAM_ECHO_BASE_CLASS (GTK_OBJECT (model)->klass);
  bpm = gdam_beat_info_get_bpm (match_info, 0);
  gdam_beat_info_set_bpm (&model->beat_info, bpm);
  echo_base_class->refresh_echo_array (GDAM_ECHO_BASE (spin));
  return TRUE;
}

static void gdam_spin_arg_set_spun(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  spin->spun = GDAM_ARG_INT (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (spin));
}

static void gdam_spin_arg_get_spun(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  GDAM_ARG_INT (arg) = spin->spun;
}

void gdam_spin_set_beat_factor(GdamSpin* spin, double delay)
{
  GdamArg arg;
  arg.name = "beat_factor";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = delay;
  gdam_arg_set_object_arg (GTK_OBJECT (spin), &arg);
}

static void gdam_spin_arg_set_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  spin->beat_factor = GDAM_ARG_DOUBLE (arg);
  gdam_echo_base_refresh_echo_array (GDAM_ECHO_BASE (spin));
  /* TODO: update sub_beat? */
}

static void gdam_spin_arg_get_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  GDAM_ARG_DOUBLE (arg) = spin->beat_factor;
}


/*  Sub Beat Factor:  varies from -3 to 1 to 3 where beat factor
 *  varies from 1/3 to 1 to 3.  That is, sub-bf = -n => bf = 1/n 
 */
/* TODO: is this /should this be treated as an int?? */
static void gdam_spin_arg_set_sub_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  double bf;
  bf = GDAM_ARG_DOUBLE (arg);
  if (bf < 0)
    bf = -1.0 / bf;
  gdam_spin_set_beat_factor (spin, bf);
}

static void gdam_spin_arg_get_sub_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  if (spin->beat_factor == 0)
    GDAM_ARG_DOUBLE (arg) = spin->beat_factor;
  else if (spin->beat_factor < 1)
    GDAM_ARG_DOUBLE (arg) = -1.0 / spin->beat_factor;
  else
    GDAM_ARG_DOUBLE (arg) = spin->beat_factor;
}

static void gdam_spin_arg_set_nudge_beat_factor(GtkObject* object, GdamArg* arg)
{
  GdamSpin *spin = GDAM_SPIN (object);
  GdamArg new_arg;
  int val = GDAM_ARG_INT (arg);
  double sbf;
  /*  Calculate current sub beat factor.  */
  if (spin->beat_factor == 0)
    sbf = spin->beat_factor;
  else if (spin->beat_factor < 1)
    sbf = (-1.0 / spin->beat_factor);
  else
    sbf = spin->beat_factor;
  /*  Apply nudge  */
  if (sbf <= 1 && sbf >= -1)
    {
      if (val > 0)
	sbf = 2;
      else
	sbf = -2;
    }
  else
    {
      sbf += val;
      if (sbf == -1)
	sbf = 1;
    }
  new_arg.name = "sub_beat_factor";
  new_arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&new_arg) = sbf;
  gdam_arg_set_object_arg (object, &new_arg);
}

 
GdamSpin* gdam_spin_new()
{
  GdamSpin *spin;
  spin = GDAM_SPIN (gtk_type_new (GDAM_TYPE_SPIN));
  return spin;
}

static void gdam_spin_class_init(GdamSpinClass* klass)
{
  GdamEchoBaseClass *echo_base_class = GDAM_ECHO_BASE_CLASS (klass);
  GdamModelClass *model_class = GDAM_MODEL_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  echo_base_class->refresh_echo_array = gdam_spin_refresh_echo_array;
  model_class->match_beat_info = gdam_spin_match_beat_info;
  gdam_arg_register (object_class,
		     "spun",
		     GDAM_ARG_TYPE_INT,
		     gdam_spin_arg_set_spun,
		     gdam_spin_arg_get_spun, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_spin_arg_set_beat_factor,
		     gdam_spin_arg_get_beat_factor,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "sub_beat_factor",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_spin_arg_set_sub_beat_factor,
		     gdam_spin_arg_get_sub_beat_factor,
		     NULL, NULL, GDAM_ARG_VITAL);
  /* XXX: no longer neede, replace with normal arg bindings */
  gdam_arg_register (object_class,
		     "nudge_beat_factor",
		     GDAM_ARG_TYPE_INT,
		     gdam_spin_arg_set_nudge_beat_factor,
		     NULL, NULL, NULL, 0);
}

