/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamusbhid.h"

#if HAVE_LINUX_INPUT_H

#include <gdam/common/gdamplugin.h>
#include <linux/input.h>
#include <string.h>
#include <stdlib.h>

typedef struct input_event usb_input_event;

static void gdam_usbevent_init(GdamUsbHid* usbevent)
{
}

/* XXX: figure out how not to leak memory here */
static char *
translate_usb_input_range_to_name (GdamUsbHidEventRange *evt)
{
  if (evt->type  == EV_ABS)
    return g_strdup_printf ("c:%d", evt->code);
  return g_strdup_printf ("t:%d c:%d v:%d (unknown range_type %x)",
                          evt->type, evt->code, evt->value, evt->type);
}

/* XXX: rename variable to relative */

/* XXX: figure out how not to leak memory here */
static char *
translate_usb_input_to_name (GdamUsbHidEvent *evt)
{
  if (evt->type  == EV_RST)
    return g_strdup_printf ("c:%d v:%d rst", evt->code, evt->value);
  if (evt->type  == EV_KEY)
    return g_strdup_printf ("c:%d v:%d", evt->code, evt->value);
  if (evt->type  == EV_REL)
    return g_strdup_printf ("c:%d", evt->code);
  if (evt->type  == EV_ABS)
    return g_strdup_printf ("c:%d", evt->code);
  return g_strdup_printf ("t:%d c:%d v:%d (unknown type %x)", 
                          evt->type, evt->code, evt->value, evt->type);
}

/* Event Ranges are a different type, and one EventRange is defined
 * for each code,type pair, and covers all values */
static void 
gdam_usbevent_parse_event_range (GdamUsbHid* hid, usb_input_event* evt)
{
  GdamUsbHidEventRange *event;
  event = g_new0 (GdamUsbHidEventRange, 1);
  event->event.input_device = (GdamInputDevice *) (hid);
  event->event.event_type = gdam_usbhid_event_absolute_get_type ();
  event->code = evt->code;
  event->type = evt->type;
  event->value = evt->value;
  event->event.value = evt->value;
  /* XXX: are these always the correct values?? */
  event->event.min_value = 0;
  event->event.max_value = 256;
  event->name = translate_usb_input_range_to_name (event);
  //g_message ("parsed event range, code %d type %d value %d", 
             //evt->code, evt->type, evt->value);
  gdam_input_trap_emit (NULL, (GdamInputEvent *) event);
}


/* On usb data input, we take a usb_input_event struct from the linux 
 * input core and create a GdamUsbHidEvent.  We then emit the
 * event to anyone who had trapped it. */
static void 
gdam_usbevent_parse_event(GdamUsbHid* hid, usb_input_event* evt)
{
  GdamUsbHidEvent *event;

  if (evt->type  == EV_ABS) 
    {
      gdam_usbevent_parse_event_range (hid, evt);
      //return;  /* not returning -> also emit usbhidevent */
    }

  event = g_new0 (GdamUsbHidEvent, 1);
  event->event.input_device = (GdamInputDevice *) (hid);
  event->event.event_type = gdam_usbhid_event_get_type ();
  if (evt->type  == EV_REL)
    event->event.event_type = gdam_usbhid_event_variable_get_type ();
  if (evt->type  == EV_KEY)
    event->event.event_type = gdam_usbhid_event_key_get_type ();
  event->code = evt->code;
  event->type = evt->type;
  event->value = evt->value;
  event->event.number = evt->value;
  event->name = translate_usb_input_to_name (event);
  //g_message ("parsed hid event, code %d type %d value %d", 
             //evt->code, evt->type, evt->value);
  gdam_input_trap_emit (NULL, (GdamInputEvent *) event);
}

static gboolean gdam_usbevent_process  (GdamInputDeviceFile *device_file,
                                            guint8              *hid_buf,
					    int                  num_read)
{
  GdamUsbHid *hid = GDAM_USB_HID (device_file);
  int index;
  int num_to_process = num_read;
  guint8* buf_ptr = hid_buf;

  if (num_read <= 0)
    {
      g_warning ("usb device closed...");
      return FALSE;
    }
  while (num_to_process > 0) {
    int evtsize = sizeof (usb_input_event);
    if (num_read >= evtsize) 
    {
      usb_input_event* evt;
      evt = (usb_input_event*)buf_ptr;
      gdam_usbevent_parse_event (hid, evt);
      num_to_process -= evtsize;
      buf_ptr += evtsize;
    }
    else 
      {
	g_warning ("usbhid, had only %d bytes of %d", num_to_process, evtsize);
	/* TODO: does this close device?  we probably should... */
	return FALSE;
      }
  }
  return TRUE;
}


static void gdam_usbevent_class_init(GdamInputDeviceFileClass* klass)
{
  klass->process = gdam_usbevent_process;
}

/* A device file which gets hid events from the linux kernel input core */
GtkType gdam_usb_hid_get_type()
{
  static GtkType usbevent_type = 0;
  if (!usbevent_type)
    {
      static const GtkTypeInfo usbevent_info = {
	"GdamUsbHid",
	sizeof (GdamUsbHid),
	sizeof (GdamUsbHidClass),
	(GtkClassInitFunc) gdam_usbevent_class_init,
	(GtkObjectInitFunc) gdam_usbevent_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_INPUT_DEVICE_FILE;
      usbevent_type = gtk_type_unique (parent, &usbevent_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
      (void) gdam_usbhid_event_get_type ();
      (void) gdam_usbhid_event_variable_get_type ();
      (void) gdam_usbhid_event_absolute_get_type ();
      (void) gdam_usbhid_event_key_get_type ();
    }
  return usbevent_type;
}

static char* 
gdam_usb_event_stringify (GdamInputEvent *event)
{
  GdamUsbHidEvent *hid_event = (GdamUsbHidEvent *) event;
  return g_strdup_printf ("UsbHidEvent: %s", hid_event->name);
}

static char *
gdam_usb_event_key_stringify (GdamInputEvent *event)
{
  GdamUsbHidEvent *hid_event = (GdamUsbHidEvent *) event;
  return g_strdup_printf ("UsbHidEventKey: %s", hid_event->name);
}

static char *
gdam_usb_event_variable_stringify (GdamInputEvent *event)
{
  GdamUsbHidEvent *hid_event = (GdamUsbHidEvent *) event;
  return g_strdup_printf ("UsbHidEventVariable: %s", hid_event->name);
}

static char *
gdam_usb_event_absolute_stringify (GdamInputEvent *event)
{
  GdamUsbHidEventRange *hid_event = (GdamUsbHidEventRange *) event;
  return g_strdup_printf ("UsbHidEventAbsolute: %s", hid_event->name);
}

static unsigned  
gdam_usb_event_variable_hash (GdamInputEvent *event)
{
  return GPOINTER_TO_UINT (event->input_device)
    ^ (((GdamUsbHidEvent *) event)->code << 16)
    ^ (((GdamUsbHidEvent *) event)->type << 6)
    ^ (event->event_type << 8);
}

static unsigned 
gdam_usb_event_range_hash (GdamInputEvent *event)
{
  return GPOINTER_TO_UINT (event->input_device)
    ^ (((GdamUsbHidEventRange *) event)->code << 16)
    ^ (((GdamUsbHidEventRange *) event)->type << 6)
    ^ (event->event_type << 8);
}

static unsigned gdam_usb_event_hash(GdamInputEvent* event)
{
  return GPOINTER_TO_UINT (event->input_device)
    ^ (((GdamUsbHidEvent *) event)->code << 16)
    ^ (((GdamUsbHidEvent *) event)->type << 6)
    ^ (((GdamUsbHidEvent *) event)->value << 4)
    ^ (event->event_type << 8);
}

static GHashTable* gdam_usb_event_to_keyvalues(GdamInputEvent* event)
{
  GHashTable *table = g_hash_table_new (g_str_hash, g_str_equal);
  const char *device_name;
  GdamUsbHidEvent *usb_event = (GdamUsbHidEvent *) event;
  table = g_hash_table_new (g_str_hash, g_str_equal);

  device_name = event->input_device->device_name;
  gdam_input_store_string (table, "device", device_name);

  gdam_input_store_int (table, "type", usb_event->type);
  gdam_input_store_int (table, "code", usb_event->code);
  gdam_input_store_int (table, "value", usb_event->value);
  gdam_input_store_string (table, "name", usb_event->name);
  return table;
}

static GdamInputEvent * 
gdam_usb_event_from_keyvalues	(int         type,
                                 GHashTable *table)
{

  GdamUsbHidEvent *event;
  char *device_name;
  int htype, code, value;

  event = g_new0 (GdamUsbHidEvent, 1);
  event->event.event_type = type;
  event->event.input_id = 0;
  device_name = g_hash_table_lookup (table, "device");
  gdam_input_event_parse_device ((GdamInputEvent *) event, device_name);
  htype = gdam_input_fetch_int (table, "type");
  code = gdam_input_fetch_int (table, "code");
  value = gdam_input_fetch_int (table, "value");
  if (type == gdam_usbhid_event_key_get_type ())
    event->type = EV_KEY;
  else if (htype == -1)
    {
      g_warning ("no `type' in usb event");
      g_free (event);
      return NULL;
    }
  else
    event->type = htype;
  if (code == -1)
    {
      g_warning ("no `code' in usb event");
      g_free (event);
      return NULL;
    }
  event->code = code;
  event->value = value;
  event->event.number = value;
  event->name = translate_usb_input_to_name (event);
  return (GdamInputEvent *) event;
}

static GHashTable * 
gdam_usb_event_range_to_keyvalues (GdamInputEvent *event)
{
  GHashTable *table = g_hash_table_new (g_str_hash, g_str_equal);
  const char *device_name;
  GdamUsbHidEventRange *usb_event = (GdamUsbHidEventRange *) event;
  table = g_hash_table_new (g_str_hash, g_str_equal);

  device_name = event->input_device->device_name;
  gdam_input_store_string (table, "device", device_name);

  gdam_input_store_int (table, "type", usb_event->type);
  gdam_input_store_int (table, "code", usb_event->code);
  gdam_input_store_int (table, "value", usb_event->event.value);
  gdam_input_store_string (table, "name", usb_event->name);
  return table;
}

static GdamInputEvent *
gdam_usb_event_range_from_keyvalues     (int 	      type,  
                                         GHashTable   *table)
{

  GdamUsbHidEventRange *event;
  char *device_name;
  int htype, code, value;

  event = g_new0 (GdamUsbHidEventRange, 1);
  event->event.event_type = type;
  event->event.input_id = 0;
  device_name = g_hash_table_lookup (table, "device");
  gdam_input_event_parse_device ((GdamInputEvent *) event, device_name);
  htype = gdam_input_fetch_int (table, "type");
  code = gdam_input_fetch_int (table, "code");
  value = gdam_input_fetch_int (table, "value");
  if (htype == -1)
    {
      g_warning ("no `type' in usb event");
      g_free (event);
      return NULL;
    }
  if (code == -1)
    {
      g_warning ("no `code' in usb event");
      g_free (event);
      return NULL;
    }
  event->type = htype;
  event->code = code;
  event->value = value;
  event->event.value = value;
  event->event.min_value = 0;
  event->event.max_value = 256;
  event->name = translate_usb_input_range_to_name (event);
  return (GdamInputEvent *) event;
}

/* Events for devices which emit variable values.  Because 'value' can
 * cover a wide range, we don't match against it.
 */
static gboolean 
gdam_usb_event_variable_equal (GdamInputEvent *event1,
                               GdamInputEvent *event2)
{
  GdamUsbHidEvent *kevent1 = (GdamUsbHidEvent *) event1;
  GdamUsbHidEvent *kevent2 = (GdamUsbHidEvent *) event2;
  return event1->input_device == event2->input_device
    && event1->event_type == event2->event_type
    && kevent1->type == kevent2->type
    && kevent1->code == kevent2->code;
}

static gboolean  
gdam_usb_event_range_equal (GdamInputEvent *event1,
			    GdamInputEvent *event2)
{
  GdamUsbHidEventRange *kevent1 = (GdamUsbHidEventRange *) event1;
  GdamUsbHidEventRange *kevent2 = (GdamUsbHidEventRange *) event2;
  return event1->input_device == event2->input_device
    && event1->event_type == event2->event_type
    && kevent1->type == kevent2->type
    && kevent1->code == kevent2->code;
}

static gboolean 
gdam_usb_event_equal (GdamInputEvent *event1,
                      GdamInputEvent *event2)
{
  GdamUsbHidEvent *kevent1 = (GdamUsbHidEvent *) event1;
  GdamUsbHidEvent *kevent2 = (GdamUsbHidEvent *) event2;
  return event1->input_device == event2->input_device
    && event1->event_type == event2->event_type
    && kevent1->type == kevent2->type
    && kevent1->code == kevent2->code
    && kevent1->value == kevent2->value;
}

static GdamInputEvent* gdam_usb_event_canonicalize(GdamInputEvent* event)
{
  return g_memdup (event, sizeof (GdamUsbHidEvent));
}

static GdamInputEvent* gdam_usb_event_range_canonicalize(GdamInputEvent* event)
{
  return g_memdup (event, sizeof (GdamUsbHidEventRange));
}

/* The generic usb hid event, as presented by the linux kernel input core.
 * This is used for keys and other event which emit one or more constant values.
 * An event is defined for each value emitted.  This type is used for any
 * HID event type which isn't being parsed yet. */
int gdam_usbhid_event_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamUsbHidEvent",
    gdam_usb_event_stringify,
    gdam_usb_event_hash,
    gdam_usb_event_equal,
    gdam_usb_event_to_keyvalues,
    gdam_usb_event_from_keyvalues,
    gdam_usb_event_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_int_get_type (), GDAM_TYPE_USB_HID, &type_info);
  return type;
}

/* An HID event which includes a variable value, such as a relative position
 * event (mice, trackball, touchpad)  In contrast to a normal HidEvent,
 * there is only one event for all values. */
int gdam_usbhid_event_variable_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamUsbHidEventVariable",
    gdam_usb_event_stringify,
    gdam_usb_event_variable_hash,
    gdam_usb_event_variable_equal,
    gdam_usb_event_to_keyvalues,
    gdam_usb_event_from_keyvalues,
    gdam_usb_event_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_int_get_type (), GDAM_TYPE_USB_HID, &type_info);
  return type;
}

/* XXX: don't parse type from events whose type is implicit from name */

/* An HID event which includes an absolute value, such as a joystick axis
 * or midi slider. */
int gdam_usbhid_event_absolute_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamUsbHidEventAbsolute",
    gdam_usb_event_absolute_stringify,
    gdam_usb_event_range_hash,
    gdam_usb_event_range_equal,
    gdam_usb_event_range_to_keyvalues,
    gdam_usb_event_range_from_keyvalues,
    gdam_usb_event_range_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_range_get_type (), GDAM_TYPE_USB_HID, &type_info);
  return type;
}

/* Key press... generally 1 for press, 2 for release. */
int 
gdam_usbhid_event_key_get_type()
{
  static GdamInputEventTypeInfo type_info = {
    "GdamUsbHidEventKey",
    gdam_usb_event_key_stringify,
    gdam_usb_event_hash,
    gdam_usb_event_equal,
    gdam_usb_event_to_keyvalues,
    gdam_usb_event_from_keyvalues,
    gdam_usb_event_canonicalize,
    NULL
  };
  static guint type = 0;
  if (type == 0)
    type = gdam_input_event_get_unique
      (gdam_input_event_int_get_type (), GDAM_TYPE_USB_HID, &type_info);
  return type;
}


#endif
