/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gdam/client/gdamchannel.h>
#include <gdam/component/gdamtoplevel.h>
#include <gdam/component/gdamturntable.h>
#include <gdam/component/gdamgain.h>
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdammp3info.h>
#include <gdam/common/gdamwaitpid.h>
#include <gdam/component/gdamproducerqueue.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <gdam/common/gdamautoconf.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "getline.h"


#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>

/* XXX: the following are fairly non-portable and should
 *      be ifdef'd appropriately. */
#include <dirent.h>
#include <sys/ioctl.h>

/* List of GdamSongInfo pointers. */
guint             ended_subscription = 0;
guint             mixer_id           = 0;
guint             device_id          = 0;
GdamChannel*      channel            = NULL;
GdamProducer*	  producer	     = NULL;
GdamProducerQueue* queue	     = NULL;
GdamTurntable*	  turntable	     = NULL;
GdamGain*	  gain		     = NULL;
GdamToplevel*	  toplevel	     = NULL;
gboolean          quiet              = FALSE;
gboolean          do_exit            = FALSE;
gboolean          use_short_names    = FALSE;
gboolean          use_termios        = TRUE;
gboolean          running_clock      = TRUE;
gboolean          display_secs       = FALSE;
volatile gboolean ctrl_c_hit         = FALSE;
gboolean          never_miss_data    = FALSE;
gdouble           volume_delta       = 0.04;
gdouble           last_position      = 0.0;

/* Requests to clear the line later. */
static int        clear_line_delay = 5000;
static GTimeVal   clear_request_time;
static gboolean   clear_line_requested = FALSE;

static void on_song_ended(gpointer unused);
static gboolean do_quit(gpointer unused);

static void print_help();
static void print_playlist();
static void print_canon_titles();
static void maybe_clear_line();
static void set_echo(gboolean echo_state);
static void request_clear_line();
static char* title_canon(const char* title);
static gboolean source_ex_file(const char* filename);
static void keyboard_io_watch_suspend ();
static void keyboard_io_watch_release ();



/*      ____  _             _ _     _     ____  _          __  __ 
 *     |  _ \| | __ _ _   _| (_)___| |_  / ___|| |_ _   _ / _|/ _|
 *     | |_) | |/ _` | | | | | / __| __| \___ \| __| | | | |_| |_ 
 *     |  __/| | (_| | |_| | | \__ \ |_   ___) | |_| |_| |  _|  _|
 *     |_|   |_|\__,_|\__, |_|_|___/\__| |____/ \__|\__,_|_| |_|  
 *                    |___/                                       
 */

static int compare_smart_filename(gconstpointer p1, gconstpointer p2);

static int
compare_alpha (gconstpointer p1, gconstpointer p2)
{
  int rv;
  GdamSongInfo *song1 = (GdamSongInfo *) p1;
  GdamSongInfo *song2 = (GdamSongInfo *) p2;
  rv = strcmp (song1->filename, song2->filename);
  return rv;
}

static int
match_title (gconstpointer p1, gconstpointer p2)
{
  int rv;
  char *song1 = (char *) p1;
  char *song2 = (char *) p2;
  rv = strcmp (song1, song2);
  return rv;
}

static int
compare_smart_songinfo (gconstpointer p1, gconstpointer p2)
{
  GdamSongInfo *song1 = (GdamSongInfo *) p1;
  GdamSongInfo *song2 = (GdamSongInfo *) p2;
  return compare_smart_filename (song1->filename, song2->filename);
}

static int
compare_smart_filename (gconstpointer p1, gconstpointer p2)
{
  int pos1p = -1;
  int pos1c = -1;
  int pos2p = -1;
  int pos2c = -1;
  char f1[2048] = "\0";
  char f2[2048] = "\0";
  char *fn;
  char *fn2;
  char *last_slash;
  int rv = 0;
  GList *found;
  fn = g_strdup ((char *) p1);
  fn2 = g_strdup ((char *) p2);

  //fprintf(stderr, "compare_smart: %s \r\n          %s\r\n", fn, fn2);

  last_slash = strrchr (fn, '/');
  if (last_slash)
    fn = last_slash + 1;
  if (isdigit (fn[0]) && isdigit (fn[1]) && (fn[2] == '-'))
    strncpy (f1, fn + 3, strlen (fn) - 7);
  else
    strncpy (f1, fn, strlen (fn) - 4);
  found = g_list_find_custom (queue->playlist_canon_titles, title_canon (f1),
			      match_title);
  pos1p = g_list_position (queue->playlist_canon_titles, found);
  found = g_list_find_custom (queue->cddb_canon_titles, title_canon (f1),
			      match_title);
  pos1c = g_list_position (queue->cddb_canon_titles, found);
  //fprintf(stderr, "got index %d for %s\r\n", pos1p, f1);

  last_slash = strrchr (fn2, '/');
  if (last_slash)
    fn2 = last_slash + 1;
  if (isdigit (fn2[0]) && isdigit (fn2[1]) && (fn2[2] == '-'))
    strncpy (f2, fn2 + 3, strlen (fn2 + 3) - 4);
  else
    strncpy (f2, fn2, strlen (fn2) - 4);
  found = g_list_find_custom (queue->playlist_canon_titles, title_canon (f2),
			      match_title);
  pos2p = g_list_position (queue->playlist_canon_titles, found);
  found = g_list_find_custom (queue->cddb_canon_titles, title_canon (f2),
			      match_title);
  pos2c = g_list_position (queue->cddb_canon_titles, found);
  //fprintf(stderr, "got index %d for %s\r\n", pos2p, f2);

  if (pos1p >= 0 && pos2p >= 0)
    {
      rv = pos1p - pos2p;
    }
  else if (pos1c >= 0 && pos2c >= 0)
    {
      //fprintf(stderr, "compare_smart returning %d ", pos1c - pos2c);
      rv = pos1c - pos2c;
    }
  /* if neither list contained both... */
  else if (pos1p >= 0 || pos1c >= 0)
    rv = -1;
  else if (pos2p >= 0 || pos2c >= 0)
    rv = 1;
  /* if neither list contained either... */
  else
    /* TODO: sort alphabetically by directory? */
    rv = strcmp (fn, fn2);
  //fprintf(stderr, "compare_smart returning 0: %d %d %d %d", 
  //pos1p, pos2p, pos1c, pos2c);
  return rv;
}

static void sort_alpha()
{
  GList *playlist = g_list_copy (queue->songinfo_list);
  playlist = g_list_sort (playlist, compare_alpha);
  gdam_producer_queue_set_queue_list (GDAM_PRODUCER_QUEUE (producer),
				      playlist);
}

static void sort_smart()
{
  GList *playlist = g_list_copy (queue->songinfo_list);
  gdam_debug ("sorting  %d songs smartly.",
	      g_list_length (queue->songinfo_list));
  playlist = g_list_sort (playlist, compare_smart_songinfo);
  gdam_producer_queue_set_queue_list (GDAM_PRODUCER_QUEUE (producer),
				      playlist);
}

/* TODO: treat special cases like the word 'and' */
static char* title_canon(const char* title)
{
  char rv[2048];
  int i;
  int j = 0;
  //fprintf(stderr, "canon: %s\r\n", title);
  for (i = 0; title[i]; i++)
    {
      if (isalnum (title[i]))
	{
	  if (isupper (title[i]))
	    rv[j] = tolower (title[i]);
	  else
	    rv[j] = title[i];
	  j++;
	}
    }
  rv[j] = '\0';
  //fprintf(stderr, "canon: %s \r\n into: %s \r\n", title, rv);
  return strdup (rv);
}

static void on_song_ended(gpointer unused)
{
  //g_warning("Song ended, quitting gdam123\r\n\r\n");
  do_quit (NULL);
}

static void print_canon_title(gpointer list_data, gpointer unused){
	fprintf(stderr, "%s\r\n", ((char*)list_data));
}

static void
print_canon_titles ()
{
  fprintf (stderr, "Cddb Canon Titles: \r\n");
  g_list_foreach (queue->cddb_canon_titles, print_canon_title, NULL);
  fprintf (stderr, "Playlist Canon Titles: \r\n");
  g_list_foreach (queue->playlist_canon_titles, print_canon_title, NULL);
}


/*     ____                _                     
 *    |  _ \ _ __ ___   __| |_   _  ___ ___ _ __ 
 *    | |_) | '__/ _ \ / _` | | | |/ __/ _ \ '__|
 *    |  __/| | | (_) | (_| | |_| | (_|  __/ |   
 *    |_|   |_|  \___/ \__,_|\__,_|\___\___|_|   
 *                                               
 *      ____            _             _ 
 *     / ___|___  _ __ | |_ _ __ ___ | |
 *    | |   / _ \| '_ \| __| '__/ _ \| |
 *    | |__| (_) | | | | |_| | | (_) | |
 *     \____\___/|_| |_|\__|_|  \___/|_|
 */

#if 0
#endif

/* Callback to exit program */
static gboolean do_quit(gpointer unused){
	exit(0);
}

/*  Registers callback to quit after producer has had chance to stop.  */
static void quit_prog(){
	gdam_producer_stop(GDAM_PRODUCER(producer));
	g_idle_add(do_quit, NULL);
}

#define TIME_LABEL_WIDTH 12
static void print_indented_filename()
{
  int n;
  char filename[1024];
  int column_width = 80;
  int free_width;
  struct winsize size;

  if (ioctl (STDIN_FILENO, TIOCGWINSZ, (char *) &size) >= 0)
    column_width = size.ws_col;

  free_width = column_width - TIME_LABEL_WIDTH - 1;

  if (use_short_names)
    strcpy (filename,
	    g_basename (queue->current_song->filename));
  else
    strcpy (filename, queue->current_song->filename);
  if ((int) strlen (filename) > free_width)
    filename[free_width] = '\0';
  fprintf (stderr, "\r");
  for (n = 0; n < TIME_LABEL_WIDTH; n++)
    fprintf (stderr, " ");
  fprintf (stderr, " %s", filename);
}

static void on_set_filename(gpointer unused)
{
  fprintf (stderr, "\033[K");
  print_indented_filename ();
}

static int last_col_width = 0;
static void
refresh_position ()
{
  if (running_clock)
    {
      int column_width;
      struct winsize size;
      maybe_clear_line ();
      if (ioctl (STDIN_FILENO, TIOCGWINSZ, (char *) &size) >= 0)
	{
	  column_width = size.ws_col;
	  if (column_width != last_col_width)
	    {
	      last_col_width = column_width;
	      print_indented_filename ();
	    }
	}

      if (display_secs)
	{
	  fprintf (stderr, "\r [ %3.0f   ] ", last_position);
	}
      else
	{
	  int secs = (int) (last_position);
	  if (last_position < 0)
	    fprintf (stderr, "\r [ -%02d:%02d ] ",
		     secs / 60, abs (secs % 60));
	  else
	    fprintf (stderr, "\r [ %03d:%02d ] ", secs / 60, secs % 60);
	}
    }
  else
    fprintf (stderr, "\r ");
}

static void
refresh_position_and_filename ()
{
  const char *name;
  refresh_position ();

  if (queue->current_song == NULL)
    name = "[nothing]";
  else if (use_short_names)
    name = g_basename (queue->current_song->filename);
  else
    name = queue->current_song->filename;

  fprintf (stderr, "%s", name);
}
  

static void on_set_position(GtkObject* object,
			    double new_val)
{
  last_position = new_val;
  refresh_position ();
}



/*
void cleanup_curses()
{
  nocbreak ();
  nl ();
  //echo();
  //fprintf(stderr, "cleaning up.... \r\n");
}
*/

static void seek_song(gdouble amt, int whence)
{
  gdam_producer_seek (GDAM_PRODUCER (producer), amt, whence);
}

static void seek_song_beats(gdouble amt)
{
  /* fprintf(stderr, "seeking %f beats\r\n", amt); */
  gdam_turntable_seek_by_beats (turntable, amt);
}

static void toggle_repeat()
{
  int n;
  gboolean do_repeat;

  if (queue->do_repeat)
    do_repeat = FALSE;
  else
    do_repeat = TRUE;

  gdam_queue_set_repeat (queue, do_repeat);

  fprintf (stderr, "\r");
  for (n = 0; n < TIME_LABEL_WIDTH; n++)
    fprintf (stderr, " ");
  fprintf (stderr, " repeat %s ", queue->do_repeat ? "on" : "off");
  fprintf (stderr, "\033[K\r");
  request_clear_line ();
}

static void
set_vol (float newvol)
{
  int n;
  if (!gain)
    return;
  fprintf (stderr, "\r");
  for (n = 0; n < TIME_LABEL_WIDTH; n++)
    fprintf (stderr, " ");
  fprintf (stderr, " volume: %3d%% ", (int) (newvol * 100 + 0.5));
  fprintf (stderr, "\033[K\r");
  request_clear_line ();
  gdam_gain_set_volume (gain, newvol);
}

static void
nudge_vol (float amt)
{
  float newvol;
  int n;
  if (!gain)
    return;
  newvol = gain->volume;
  newvol += amt;
  if (newvol < 0.0)
    newvol = 0.0;
  set_vol (newvol);
}

static void
toggle_pause ()
{
  GdamModel *component = GDAM_MODEL (producer);

  if (component->source_id != 0)
    gdam_producer_stop (producer);
  else
    gdam_producer_play_now (producer);
}


/* Print a range of songs.  counts from 1. */
static void print_range(guint lower, guint upper)
{
  int n;
  g_return_if_fail (lower <= upper);
  fprintf (stderr, "\r\n");
  for (n = lower; n <= (int) upper; n++)
    {
      GList *to_del;
      to_del = g_list_nth (queue->songinfo_list, n - 1);

      if (to_del)
	{
	  char filename[1024];
	  GdamSongInfo* doomed;
	  doomed = ((GdamSongInfo *)(to_del->data));
	  if (to_del->data == queue->current_song)
	    fprintf (stderr, "%3d=>", n);
	  else
	    fprintf (stderr, "%3d  ", n);
	  if (use_short_names)
	    strcpy (filename, g_basename (doomed->filename));
	  else
	    strcpy (filename, doomed->filename);
	  fprintf (stderr, "%s\n\r", filename);
	}
    }
}

/*
 *          _
 *  ___   _(_)
 * (_) \ / / |
 *  _ \ V /| |
 * (_) \_/ |_|
 * 
 */
typedef struct _ViInfo ViInfo;
struct _ViInfo
{
  char *tmp_filename;
  int pid;
};

static void
vi_info_free (gpointer vi_info_ptr)
{
  ViInfo *vi_info = vi_info_ptr;
  g_free (vi_info->tmp_filename);
  g_free (vi_info);
}

static void
handle_vi_done (gpointer user_data,
		GdamWaitPidInfo *waitpid_info)
{
  ViInfo *vi_info = user_data;
  GList *songinfo_list = NULL;

  /* start selecting on the keyboard again */
  keyboard_io_watch_release ();

  /* read the new playlist into songinfo_list */
  {
    FILE *fp;
    char buf[2048];
    GdamSongDatabase *song_database;
    fp = fopen (vi_info->tmp_filename, "r");
    if (fp == NULL)
      {
	g_warning ("unable to read temporary file `%s'", vi_info->tmp_filename);
	return;
      }
    song_database = gdam_song_database_system ();
    while (fgets (buf, sizeof (buf), fp))
      {
	GdamSongInfo *song_info;
	g_strchomp (buf);
	song_info = gdam_song_database_find (song_database, buf);
	if (song_info == NULL)
	  {
	    g_warning ("couldn't find song-database entry for %s: "
		       "making new default entry", buf);

	    /* XXX: 0,0: are those the right default values ? */
	    song_info = gdam_song_info_new (buf, 0.0, 0.0);

	    if (song_info == NULL)
	      continue;
	  }
	songinfo_list = g_list_prepend (songinfo_list, song_info);
      }
    unlink (vi_info->tmp_filename);
    fclose (fp);
  }
  songinfo_list = g_list_reverse (songinfo_list);

  /* set the new playlist */
  gdam_producer_queue_set_queue_list (queue, songinfo_list);

  /* free the songlist */
  g_list_free (songinfo_list);
}

static void
write_filename_by_song_info (gpointer info_ptr, gpointer fp_ptr)
{
  GdamSongInfo *song_info = info_ptr;
  FILE *fp = fp_ptr;
  fprintf (fp, "%s\n", song_info->filename);
}

static char *
my_tmpnam (const char *name)
{
  static const char *prefix = NULL;
  static guint count = 0;
  if (prefix == NULL)
    {
      prefix = getenv ("TMPDIR");
      if (prefix == NULL)
	prefix = getenv ("TEMPDIR");
      if (prefix == NULL)
	prefix = "/tmp";
    }
  return g_strdup_printf ("%s%c%s-%05x", prefix,
			  G_DIR_SEPARATOR, name, count++);
}

FILE *
gdam_tmp_file_fp_with_name (char **filename_out)
{
  FILE *fp;
  char *nam;
  while ((nam = my_tmpnam ("gdam")) != NULL)
    {
      int fd = open (nam, O_RDWR | O_CREAT | O_EXCL, 0644);
      if (fd < 0)
	{
	  g_free (nam);
	  continue;
	}
      fp = fdopen (fd, "r+");
      g_assert (fp != NULL);
      if (filename_out != NULL)
	*filename_out = g_strdup (nam);
      return fp;
    }
  g_error ("tmpnam failed");
  return NULL;
}

static void
launch_vi_process ()
{
  FILE *fp;
  ViInfo *vi_info = g_new0 (ViInfo, 1);

  /* create the temporary playlist */
  fp = gdam_tmp_file_fp_with_name (&vi_info->tmp_filename);
  if (fp == NULL)
    {
      g_warning ("launch_vi_process: couldn't create tmp file");
      return;
    }
  g_list_foreach (queue->songinfo_list, write_filename_by_song_info, fp);
  fclose (fp);

  /* stop selecting on the keyboard. */
  keyboard_io_watch_suspend ();

  /* launch the vi process */
  vi_info->pid = fork ();
  if (vi_info->pid < 0)
    {
      g_warning ("error forking %s", g_strerror (errno));
      return;
    }
  if (vi_info->pid == 0)
    {
      const char *editor = getenv ("VISUAL");
      if (editor == NULL)
	editor = getenv ("EDITOR");
      if (editor == NULL)
	editor = "vi";
      execlp (editor, editor, vi_info->tmp_filename, NULL);
      _exit (127);
    }

  gdam_wait_pid (vi_info->pid, handle_vi_done, vi_info, vi_info_free);
}

/*
 *      _____          ____  _         _      
 *     | ____|_  __   / ___|| |_ _   _| | ___ 
 *     |  _| \ \/ /___\___ \| __| | | | |/ _ \
 *     | |___ >  <_____|__) | |_| |_| | |  __/
 *     |_____/_/\_\   |____/ \__|\__, |_|\___|
 *                               |___/        
 *    ____                                          _     
 *   / ___|___  _ __ ___  _ __ ___   __ _ _ __   __| |___ 
 *  | |   / _ \| '_ ` _ \| '_ ` _ \ / _` | '_ \ / _` / __|
 *  | |__| (_) | | | | | | | | | | | (_| | | | | (_| \__ \
 *   \____\___/|_| |_| |_|_| |_| |_|\__,_|_| |_|\__,_|___/
 */

static void process_set_options(char** words)
{
  if (words[0] == NULL)
    {
      printf ("volume_delta %.4g\nbuffer_size %d\nnever_miss_data %d\n",
	      volume_delta * 100.0, queue->buffer_size, 
	      gdam_producer_queue_get_never_miss_data (queue));
      return;
    }
  if (strcmp (words[0], "volume_delta") == 0)
    {
      float delta;
      if (words[0] == NULL)
	{
	  g_message ("usage: volume_delta #");
	  return;
	}
      delta = atof (words[1]);
      if (delta <= 0.00001)
	{
	  g_warning ("won't set volume_delta to <=0");
	  return;
	}
      volume_delta = 0.01 * delta;
      gdam_debug ("volume_delta set to %4.4g", volume_delta);
      return;
    }
  if (strcmp (words[0], "buffer_size") == 0)
    {
      gdouble size;
      char *endp;
      if (words[0] == NULL)
	{
	  g_message ("usage: buffer_size #");
	  return;
	}
      size = strtod (words[1], &endp);
      if (*endp == 'k')
	{
	  size *= 1024;
	}
      else if (*endp == 's')
	{
	  size *= 44100;
	}
      if (size <= 32768)
	{
	  g_warning ("buffer size is dangerously low! "
		     "(must be >32768 or so)");
	  return;
	}
      queue->buffer_size = (int) size;
      return;
    }
  if (strcmp (words[0], "never_miss_data") == 0)
    {
      //gdam_producer_queue_set_never_miss_data (queue, TRUE);
      return;
    }
  if (strcmp (words[0], "miss_data") == 0)
    {
      //gdam_producer_queue_set_never_miss_data (queue, FALSE);
      return;
    }
  g_warning ("unknown `:set' option: %s", words[0]);
}

static char** split_into_words(const char* line)
{
  const char *at = line;
  char **rv;
  int num_words = 0;
  while (*at)
    {
      while (*at && isspace (*at))
	at++;
      if (*at)
	{
	  gboolean done = FALSE;
	  num_words++;
	  /* ignore backslashed spaces... */
	  while (!done)
	    {
	      while (*at && !isspace (*at))
		at++;
	      if (*at && isspace (*at) && (at != line) && (*(at - 1) == '\\'))
		{
		  at++;
		}
	      else
		{
		  done = TRUE;
		}

	    }
	}
    }
  if (num_words == 0)
    return NULL;
  rv = g_new (char *, num_words + 1);
  num_words = 0;
  at = line;
  while (*at)
    {
      while (*at && isspace (*at))
	at++;
      if (*at)
	{
	  const char *start_word = at;
	  gboolean done = FALSE;
	  /* ignore backslashed spaces... */
	  while (!done)
	    {
	      while (*at && !isspace (*at))
		at++;
	      if (*at && isspace (*at) && at != line && *(at - 1) == '\\')
		{
		  at++;
		}
	      else
		{
		  done = TRUE;
		}

	    }
	  rv[num_words] = g_new (char, at - start_word + 1);
	  memcpy (rv[num_words], start_word, at - start_word);
	  rv[num_words][at - start_word] = '\0';
	  num_words++;
	}
    }
  rv[num_words] = NULL;
  return rv;
}

static void parse_ex(const char* line)
{
  gboolean do_forward = FALSE;
  gboolean do_back = FALSE;
  int ex_num1 = 0;
  int ex_num2 = 0;
  gboolean ex_had_num1 = FALSE;
  gboolean ex_had_num2 = FALSE;

  char **words;

  if (isdigit (line[0]))
    {
      int numtotal = 0;
      for (; isdigit (line[0]); line++)
	{
	  int key = line[0] - 48;
	  numtotal = 10 * numtotal + key;
	}
      if (!ex_had_num1)
	{
	  ex_had_num1 = TRUE;
	  ex_num1 = numtotal;
	}
      if (line[0] == ',')
	{
	  line++;
	}
      /* Reset for second number */
      do_forward = FALSE;
      do_back = FALSE;
      numtotal = 0;
      /* check for relative range */
      if (line[0] == '+')
	{
	  do_forward = TRUE;
	  line++;
	}
      else if (line[0] == '-')
	{
	  do_back = TRUE;
	  line++;
	}
      /* Try to read second number */
      if (isdigit (line[0]))
	{
	  for (; isdigit (line[0]); line++)
	    {
	      int key = line[0] - 48;
	      numtotal = 10 * numtotal + key;
	    }
	  if (do_forward)
	    {
	      ex_num2 = ex_num1 + numtotal;
	    }
	  else if (do_back)
	    {
	      ex_num2 = ex_num1 - numtotal;
	    }
	  else
	    {
	      ex_num2 = numtotal;
	    }
	  ex_had_num2 = TRUE;
	}
      else if (line[0] == '$')
	{
	  ex_num2 = g_list_length (queue->songinfo_list);
	  ex_had_num2 = TRUE;
	  line++;
	}
      for (; isspace (line[0]); line++)
	{
	}
    }

  words = split_into_words (line);
  if (words == NULL || *words == NULL)
    {
      if (words)
	g_strfreev (words);
      if (ex_had_num1)
	{
	  gdam_producer_queue_skip_to (queue, ex_num1);
	}
      return;
    }
  if (strncmp (words[0], "quit", 1) == 0)
    {
      quit_prog (NULL);
    }
  else if (strncmp (words[0], "print", 1) == 0)
    {
      if (ex_had_num1)
	{
	  int lower = ex_had_num1 ? ex_num1 :
	    g_list_index (queue->songinfo_list,
			  queue->current_song) + 1;
	  int upper = ex_had_num2 ? ex_num2 : lower;
	  print_range (lower, upper);
	}
      else
	print_playlist ();
    }
  else if (strncmp (words[0], "list", 1) == 0)
    {
      int cur_index;
      int lower;
      int upper;
      int radius = ex_had_num1 ? ex_num1 : 4;
      cur_index = g_list_index (queue->songinfo_list,
				queue->current_song) + 1;
      lower = cur_index - radius;
      if (lower < 1)
	lower = 1;
      upper = cur_index + radius;
      if (upper > (int) g_list_length (queue->songinfo_list))
	upper = g_list_length (queue->songinfo_list);
      print_range (lower, upper);
    }
  else if (strncmp (words[0], "clock", 5) == 0)
    {
      running_clock = TRUE;
    }
  else if (strncmp (words[0], "noclock", 7) == 0)
    {
      running_clock = FALSE;
    }
  else if (strncmp (words[0], "nohidedir", 8) == 0)
    {
      use_short_names = FALSE;
    }
  else if (strncmp (words[0], "hidedir", 6) == 0)
    {
      use_short_names = TRUE;
    }
  else if (strncmp (words[0], "help", 1) == 0)
    {
      print_help ();
    }
  else if (strncmp (words[0], "read", 1) == 0)
    {
      if (words[1] != NULL)
	{
	  gdam_producer_queue_source_file (queue, words[1]);
	}
    }
  else if (strncmp (words[0], "append", 1) == 0)
    {
      int i = 1;
      while (words[i])
	{
	  gdam_producer_queue_source_file (queue, words[i]);
	  i++;
	}
    }
  else if (strncmp (words[0], "delete", 1) == 0)
    {
      int lower = ex_had_num1 ? ex_num1 : g_list_index (queue->songinfo_list,
							queue->current_song) +
	1;
      int upper = ex_had_num2 ? ex_num2 : lower;
      gdam_producer_queue_delete_range (GDAM_PRODUCER_QUEUE (producer), lower,
					upper);
    }
  else if (strncmp (words[0], "set", 2) == 0)
    {
      process_set_options (words + 1);
    }
  else if (strncmp (words[0], "move", 1) == 0)
    {
      int moveto = 0;
      int lower = ex_had_num1 ? ex_num1 : g_list_index (queue->songinfo_list,
							queue->current_song) +
	1;
      int upper = ex_had_num2 ? ex_num2 : lower;
      if (words[1] != NULL)
	moveto = (int) strtol (words[1], NULL, 10);
      if (moveto <= ex_num2 && moveto >= ex_num1)
	;
      else if (moveto)
	gdam_producer_queue_move_range (GDAM_PRODUCER_QUEUE (producer),
					lower, upper, moveto);
    }
  else if (strncmp (words[0], "source", 6) == 0)
    {
      if (words[1] != NULL && !source_ex_file (words[1]))
	g_warning ("error sourcing %s", words[1]);
    }
  else if (strncmp (words[0], "volume", 3) == 0)
    {
      const char *num = words[1];
      char *endp;
      gdouble volume;
      if (num == NULL)
	{
	  g_message ("volume currently %.6g", GDAM_GAIN (gain)->volume * 100);
	  return;
	}
      volume = strtod (num, &endp);
      if (num == endp)
	{
	  g_warning ("error parsing percentile argument "
		     "to volume command");
	}
      else
	{
	  gdam_gain_set_volume (gain, volume / 100.0);
	}
    }
  else if (strcmp (words[0], "vi") == 0)
    {
      launch_vi_process ();
    }
  else
    {
      fprintf (stderr, "ex: couldn't parse: '%s'\r\n", line);
    }
  g_strfreev (words);
}
static gboolean source_ex_file(const char* filename)
{
  FILE *fp = fopen (filename, "r");
  char buf[4096];
  if (fp == NULL)
    return FALSE;
  while (fgets (buf, sizeof (buf), fp))
    parse_ex (buf);
  fclose (fp);
  return TRUE;
}

static void get_ex()
{
  char *line;
  char *prompt = ":";
  set_echo (TRUE);
  fprintf (stderr, "\r\033[K");
  if ((line = getline (prompt)) != NULL)
    {
      char *oline = line;
      fprintf (stderr, "\r\033M");
      if (strlen (line) > 0 && line[strlen (line) - 1] == '\n')
	line[strlen (line) - 1] = '\0';
      parse_ex (line);
      gl_histadd (oline);
    }
  set_echo (FALSE);
  refresh_position_and_filename ();
}

static char* last_search_term = NULL;
static int   last_search_direction = 1;

static void get_search(int direction)
{
  char *line;
  char *prompt = (direction > 0) ? "/" : "?";
  last_search_direction = direction;
  set_echo (TRUE);
  fprintf (stderr, "\r\033[K");
  if ((line = getline (prompt)) != NULL)
    {
      char *oline = line;
      fprintf (stderr, "\r\033M");
      if (strlen (line) > 0 && line[strlen (line) - 1] == '\n')
	line[strlen (line) - 1] = '\0';
      if (strlen (line) > 0)
	{
	  if (last_search_term)
	    free (last_search_term);
	  last_search_term = strdup (line);
	  gdam_producer_queue_search_by_string (queue, line, direction);
	  /* TODO: memory leak?? caused seg fault... 
	   * i think getline doesn't require this to be freed?? */
	  //free(oline);
	}
      else if (last_search_term)
	{
	  gdam_producer_queue_search_by_string (queue,
						last_search_term, direction);
	}
    }
  set_echo (FALSE);
  refresh_position_and_filename ();
}

/*
 *   ____       _       _     _____                      
 *  |  _ \ _ __(_)_ __ | |_  |  ___|   _ _ __   ___ ___  
 *  | |_) | '__| | '_ \| __| | |_ | | | | '_ \ / __/ __| 
 *  |  __/| |  | | | | | |_  |  _|| |_| | | | | (__\__ \ 
 *  |_|   |_|  |_|_| |_|\__| |_|   \__,_|_| |_|\___|___/ 
 */ 

static void print_extra_value(const char* key, const char* value, FILE* fp)
{
  fprintf (fp, "   %s => \"%s\"\n", key, value);
}


static void print_mp3_info()
{
  GdamSongInfo *song_info;
  GdamMp3Info *mp3_info;
  const char *filename;
  song_info = queue->current_song;
  if (song_info == NULL)
    {
      return;
    }
  filename = song_info->filename;
  mp3_info = gdam_mp3_info_new (filename);
  if (mp3_info == NULL)
    {
      g_warning ("%s was not an mp3 file", filename);
      return;
    }
  fprintf (stderr, "\r\033[K");
  fprintf (stderr,
	   "%s: %4.4gsec, %dbps, %d samples/sec.\n",
	   filename,
	   mp3_info->length, mp3_info->bitrate, mp3_info->sampling_frequency);
  if (mp3_info->extra_values != NULL)
    g_hash_table_foreach (mp3_info->extra_values,
			  (GHFunc) print_extra_value, stderr);
  //song_info->length = mp3_info->length;
  gdam_mp3_info_destroy (mp3_info);
}


static int printcount = 0;

static void print_song(gpointer list_data, gpointer unused)
{
  char filename[1024];
  int column_width = 80;
  int free_width;
  struct winsize size;

  printcount++;

  if (ioctl (STDIN_FILENO, TIOCGWINSZ, (char *) &size) >= 0)
    column_width = size.ws_col;
  free_width = column_width - 5;

  if (use_short_names)
    strcpy (filename, g_basename (((GdamSongInfo *) list_data)->filename));
  else
    strcpy (filename, ((GdamSongInfo *) list_data)->filename);

  if ((int) strlen (filename) > free_width)
    filename[free_width] = '\0';


  //if (list_data == queue->current_song)
  if (queue->current_idx == printcount - 1)
    fprintf (stderr, "%3d=>", printcount);
  else if (list_data == queue->queued_song)
    fprintf (stderr, "%3d= ", printcount);
  else
    fprintf (stderr, "%3d  ", printcount);

  fprintf (stderr, "%s\r\n", filename);
}


static void print_playlist()
{
  fprintf (stderr, "\r\n");
  printcount = 0;
  g_list_foreach (queue->songinfo_list, print_song, NULL);
}

static void print_help(){
	fprintf(stderr,

"Gdam123 Help:\r\n"
"  Changing Songs:  j,k move to next/prev song, nG to nth song, G to last song"
									"\r\n"
"  Seeking Songs:   h,l move back/forward one beat;  [,] move back/forward a "
								"bar\r\n"
"                   ^ seeks to beginning of song;  n| seeks to nth second\r\n"
"  Pausing:         i pauses and unpauses song\r\n"
"  Volume Control:  o,p nudge volume down/up by delta%% [default:delta=10]\r\n"
"                   O,P set volume to 0%%, 100%%\r\n"
"                   \":volume 0..100%%\" changes the volume to that percent\r\n"
"                   \":set volume_delta percent\" changes the effect of o,p\r\n"
"  Playlist:        \":print\" prints playlist\r\n"
"                   \":append filenames\" adds to playlist\r\n"
"                   \":n delete\" removes the nth song from the playlist\r\n"
"                   \":n,+m move o\" moves songs n through n+m to position o"
									"\r\n"
"		    z shuffles the playlist;  r toggles repeat mode\r\n"
"  Searching:       /,? search forward/back by song title.\r\n"
"                   n,N repeat last search in same/opposite direction.\r\n"
"  Misc:            q or \":quit\" exits program\r\n"
"                   ^G (control-g) prints info on current track\r\n"
"                   \":help\" prints this messge\r\n"
"                   \":source filename\" parses the file as a "
                                                   "series of command\r\n"
"                   \":hidedir\" \":nohidedir\" toggle short filename "
								"display\r\n"
"                   \":set buffer_size time\" buffer => `time' (44100, 1s, "
								"1024k)\r\n"

	       );
}

static gboolean usage_printed = FALSE;
static void usage()
{
  if (usage_printed)
    return;
  fprintf (stderr, "Command-Line Mp3 Player for GDAM\n"
	   "\n"
	   "usage:   gdam123 [mp3file(s)] [mp3dir(s)]\n"
	   "options:\n"
	   "   --gdam-server host:port [localhost:6003]\n"
	   "   --device device [/dev/dsp, $GDAM_DEVICE]\n"
	   "   -z 	            	    shuffle playlist\n"
	   "type \":help<ENTER>\" for online help.\n");
  usage_printed = TRUE;
}



/*
 *   ____                                _  __              
 *  |  _ \ _ __ ___   ___ ___  ___ ___  | |/ /___ _   _ ___ 
 *  | |_) | '__/ _ \ / __/ _ \/ __/ __| | ' // _ \ | | / __|
 *  |  __/| | | (_) | (_|  __/\__ \__ \ | . \  __/ |_| \__ \
 *  |_|   |_|  \___/ \___\___||___/___/ |_|\_\___|\__, |___/
 *                                                |___/     
 */

static int read_key()
{
  char c;
  int result;
  result = read (0, &c, 1);
  if (result == 1)
    return c;
  if (result == 0)
    return EOF;
  if (errno == EINTR)
    return read_key ();
  /*
     g_warning("read_key: %s", strerror(errno));
   */
  return EOF;
}

static void refresh_screen()
{
  fprintf (stderr, "\033c");
  print_indented_filename ();
}


static int count = 0;
static gboolean delete_pending = FALSE;
static gboolean process_keystroke(GIOChannel* kbd_channel,
			          GIOCondition condition,
			          gpointer data)
{
  int key;
  guint num;
  gboolean hadcount;
  if ((key = read_key ()) == EOF)
    return FALSE;
  if (isdigit (key))
    {
      count = 10 * count + key - '0';
      return TRUE;
    }
  hadcount = count;
  num = count ? (guint) count : 1;
  count = 0;
  switch (key)
    {
    case 'z':
      gdam_producer_queue_shuffle_list (queue);
      break;
    case 'j':
      gdam_producer_queue_skip_songs (queue, num);
      break;
    case 'k':
      gdam_producer_queue_skip_songs (queue, -num);
      break;
    case 'G':
      if (hadcount)
	gdam_producer_queue_skip_to (queue, num);
      else
	gdam_producer_queue_skip_to (queue,
				     g_list_length (queue->songinfo_list));
      break;
    case 'h':
      seek_song_beats (-1.0 * num);
      break;
    case 'l':
      seek_song_beats (1.0 * num);
      break;
    case 12:
      {
	refresh_screen ();
	break;
      }
    case '[':
      seek_song_beats (-4.0 * num);
      break;
    case ']':
      seek_song_beats (4.0 * num);
      break;
    case '{':
      seek_song_beats (-16.0 * num);
      break;
    case '}':
      seek_song_beats (16.0 * num);
      break;
    case '^':
      seek_song (0.0, 0);
      break;
    case '|':
      seek_song (1.0 * num, 0);
      break;
    case 'q':
      quit_prog ();
      break;
    case 'i':
      toggle_pause ();
      break;
#if 0
    case 'e':
      gdam_producer_stop (producer);
      break;
      //case 'r' : 
      //if (!producer->making_play_request)
      //gdam_producer_play_now(producer);
      //break;
#endif
    case 'r':
      toggle_repeat ();
      break;
    case 'o':
      nudge_vol (-volume_delta * num);
      break;
    case 'p':
      nudge_vol (volume_delta * num);
      break;
    case 'O':
      set_vol (0);
      break;
    case 'P':
      set_vol (1);
      break;
    case ':':
      get_ex ();
      break;
    case '/':
      get_search (1);
      break;
    case '?':
      get_search (-1);
      break;
    case 'n':
      if (last_search_term)
	gdam_producer_queue_search_by_string (queue,
					      last_search_term,
					      last_search_direction);
      break;
    case 'N':
      if (last_search_term)
	gdam_producer_queue_search_by_string (queue,
					      last_search_term,
					      -1 * last_search_direction);
      break;
    case 'd':
      if ((key = read_key ()) == EOF)
	return FALSE;
      switch (key)
	{
	case 'd':
	  {
	    int todel;
	    todel = g_list_index (queue->songinfo_list,
				  queue->current_song) + 1;
	    gdam_producer_queue_delete_range (GDAM_PRODUCER_QUEUE (producer),
					      todel, todel + num - 1);
	  }
	  break;
	case 'G':
	  {
	    int lower;
	    int upper;
	    lower = g_list_index (queue->songinfo_list,
				  queue->current_song) + 1;
	    upper = g_list_length (queue->songinfo_list) + 1;
	    gdam_producer_queue_delete_range (GDAM_PRODUCER_QUEUE (producer),
					      lower, upper);
	  }
	  break;
	default:
	  break;
	}
      break;
    case 'a':
      sort_alpha ();
      break;
    case 's':
      sort_smart ();
      break;
    case 'c':
      print_canon_titles ();
      break;
    case 7:
      print_mp3_info ();
      break;
    case 27:
      count = 0;
      delete_pending = FALSE;
      break;
    default:
      //fprintf(stderr, "key: %d\r\n", key);
      break;
    }
  /* value of 'key' may have changed. */
  return TRUE;
}

/* start/stop keyboard io watch */
static guint keyboard_io_watch = 0;
static guint keyboard_io_watch_suspend_count = 1; /* we release it on startup */
static GIOChannel *keyboard_channel = NULL;
static gboolean keyboard_suspend_running_clock = TRUE;

static void
keyboard_io_watch_suspend ()
{
  if (keyboard_io_watch != 0)
    {
      g_assert (keyboard_io_watch_suspend_count == 0);
      if (keyboard_io_watch != 0)
	{
	  g_source_remove (keyboard_io_watch);
	  keyboard_io_watch = 0;
	  keyboard_suspend_running_clock = running_clock;
	  running_clock = FALSE;
	}
    }
  ++keyboard_io_watch_suspend_count;
}

static void
keyboard_io_watch_release ()
{
  g_assert (keyboard_io_watch_suspend_count != 0);
  --keyboard_io_watch_suspend_count;
  if (keyboard_io_watch_suspend_count == 0)
    {
      keyboard_io_watch = g_io_add_watch (keyboard_channel,
					  G_IO_IN, process_keystroke, NULL);
      running_clock = keyboard_suspend_running_clock;
    }
}

/* --- ctrl-c source -- */
static void ctrl_c_handler(int signo)
{
  ctrl_c_hit = TRUE;
}
static gboolean    ctrl_c_prepare          (gpointer         src_data,
                                            GTimeVal*        curtime,
                                            gint*            timeout,
		                            gpointer         user_data)
{
  return ctrl_c_hit;
}
static gboolean    ctrl_c_check            (gpointer         src_data,
                                            GTimeVal*        curtime,
		                            gpointer         user_data)
{
  return ctrl_c_hit;
}

/* Number of millis before bringing up the double-ctrl-c interface. */
static int      ctrl_c_period = 1000;
static gboolean has_last_time = FALSE;
static GTimeVal last_time_val;
static gboolean skipping_double_ctrl_c = FALSE;
static int delta_millis(GTimeVal* older, GTimeVal* younger)
{
  int millis = (older->tv_sec - younger->tv_sec) * 1000
    + (older->tv_usec - younger->tv_usec) / 1000;
  return millis;
}

static gboolean    ctrl_c_double_dispatch()
{
  int c;
  fprintf (stderr, "[Q]uit running,\r\n"
	   "[S]kip double-ctrl-c processing for a while,\r\n"
	   "[K]eep this song running, but exit.\r\n"
	   "[A]bort this song and start the next.\r\n"
	   "[C]ontinue as though ctrl-c hadn't been pressed.\r\n");

  while ((c = read_key ()) != EOF)
    {
      if (c == 'k' || c == 'K')
	{
	  exit (0);
	}
      if (c == 'q' || c == 'Q')
	{
	  quit_prog ();
	  return 1;
	}
      if (c == 'c' || c == 'C')
	{
	  return 1;
	}
      if (c == 'a' || c == 'A')
	{
	  return -1;
	}
      if (c == 's' || c == 'S')
	{
	  skipping_double_ctrl_c = TRUE;
	  return 1;
	}
    }
  return -1;
}
static gboolean    ctrl_c_dispatch         (gpointer         src_data,
                                            GTimeVal*        curtime,
		                            gpointer         user_data)
{
  ctrl_c_hit = FALSE;
  if (has_last_time)
    {
      int timeout;
      GTimeVal real_current_time;
      g_get_current_time (&real_current_time);
      timeout = delta_millis (&real_current_time, &last_time_val);
      last_time_val = real_current_time;
      if (!skipping_double_ctrl_c && timeout < ctrl_c_period)
	{
	  int code = ctrl_c_double_dispatch ();
	  if (code == 0 || code == 1)
	    return code;
	}
    }
  else
    {
      has_last_time = TRUE;
      last_time_val = *curtime;
    }
  gdam_producer_queue_skip_songs (queue, 1);
  return TRUE;
}

static GSourceFuncs ctrl_source_funcs = {
	ctrl_c_prepare,
	ctrl_c_check,
	ctrl_c_dispatch,
	NULL
};


/*    _____                   _             _   ____  _          __  __ 
 *   |_   _|__ _ __ _ __ ___ (_)_ __   __ _| | / ___|| |_ _   _ / _|/ _|
 *     | |/ _ \ '__| '_ ` _ \| | '_ \ / _` | | \___ \| __| | | | |_| |_ 
 *     | |  __/ |  | | | | | | | | | | (_| | |  ___) | |_| |_| |  _|  _|
 *     |_|\___|_|  |_| |_| |_|_|_| |_|\__,_|_| |____/ \__|\__,_|_| |_|  
 */


/* Request to have line cleared after time has elapsed. */
static void   request_clear_line(){
	g_get_current_time(&clear_request_time);
	clear_line_requested = TRUE;
}

/*  Clear the line if a request has matured.  */
static void maybe_clear_line()
{
  if (clear_line_requested)
    {
      int timeout;
      GTimeVal real_current_time;
      g_get_current_time (&real_current_time);
      timeout = delta_millis (&real_current_time, &clear_request_time);
      if (timeout > clear_line_delay)
	{
	  print_indented_filename ();
	  fprintf (stderr, "\033[K");
	  clear_line_requested = FALSE;
	}
    }
}

static void set_echo(gboolean echo_state)
{
  int fd = STDIN_FILENO;
  struct termios tattr;
  if (!use_termios)
    return;
  /* Set the funny terminal modes. */
  tcgetattr (fd, &tattr);
  if (echo_state)
    tattr.c_lflag |= ECHO;
  else
    tattr.c_lflag &= ~ECHO;
  tcsetattr (fd, TCSAFLUSH, &tattr);
}

struct termios saved_attributes;

static void reset_terminal_mode()
{
  tcsetattr (STDIN_FILENO, TCSANOW, &saved_attributes);
}

static void set_raw()
{
  int fd = STDIN_FILENO;
  struct termios tattr;

  if (!isatty (fd))
    {
      g_warning ("controlling device is not a terminal...");
      use_termios = FALSE;
      return;
    }

  /* Save the terminal attributes so we can restore them later. */
  tcgetattr (fd, &saved_attributes);
  atexit (reset_terminal_mode);

  /* Set the funny terminal modes. */
  tcgetattr (fd, &tattr);
  tattr.c_lflag &= ~(ICANON | ECHO);
  tattr.c_cc[VMIN] = 1;
  tattr.c_cc[VTIME] = 0;
  tcsetattr (fd, TCSAFLUSH, &tattr);
}

static void print_device_open_succeeded(guint device_id,
                                        guint mixer_id,
					gpointer unused)
{
  gdam_debug ("device_open: succeeded: device_id=%d, mixer_id=%d",
	      device_id, mixer_id);
}
static void print_device_open_failed(gpointer unused)
{
  g_error ("device_open: failed");
}

static void do_open_device(gpointer name, gpointer toplevel)
{
  gdam_debug ("gdam123 -- do_open_device: schedule %s", (char *) name);
  gdam_toplevel_open_device (GDAM_TOPLEVEL (toplevel),
			     (const char *) name,
			     print_device_open_succeeded,
			     print_device_open_failed, NULL);
}

//static void sig_winch(int signo){
//write(2, "SIGWINCH\n", 9);
	//fprintf(stderr, "\r\nSIG WINCH!!\r\n");
	//print_indented_filename();
	//return;
//}

static gboolean
play_now_idle_func (gpointer data)
{
  gdam_producer_play_now (producer);
  return FALSE;
}

static void
on_connect_func (GdamChannel *channel, gpointer data)
{
  g_idle_add (play_now_idle_func, NULL);
}

/*  __  __       _       
 * |  \/  | __ _(_)_ __  
 * | |\/| |/ _` | | '_ \ 
 * | |  | | (_| | | | | |
 * |_|  |_|\__,_|_|_| |_|
 */                     

int main(int argc, char** argv)
{
  gboolean do_shuffle;
  int i;
  //struct sigaction sigact;
  char *gdam123rc;
  GTimeVal timeval;
  g_set_prgname (*argv);
  signal (SIGINT, ctrl_c_handler);
  gtk_type_init ();
  gtk_signal_init ();
  gl_tab_completion (gl_completion_proc);

  keyboard_channel = g_io_channel_unix_new (STDIN_FILENO);

  g_get_current_time (&timeval);
  srand (timeval.tv_usec);

  if (argc == 1)
    {
      gdam_debug ("gdam123: not enough args");
      usage ();
      exit (0);
    }

  if (!gdam_channel_init_full (&argc, &argv, &channel,
			       NULL, NULL,
			       on_connect_func, NULL))
    return 0;

  do_shuffle = FALSE;

  queue = gdam_producer_queue_new (channel);
  queue->terminates = TRUE;
  producer = GDAM_PRODUCER (queue);
  turntable = gdam_turntable_new (channel);
  gain = GDAM_GAIN (gtk_type_new (GDAM_TYPE_GAIN));
  GDAM_MODEL (gain)->channel = channel;
  toplevel = gdam_toplevel_new (channel);
  gtk_object_ref (GTK_OBJECT (toplevel));

  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] == '-')
	{
	  if (strcmp (argv[i], "-v") == 0
	      || strcmp (argv[i], "--version") == 0)
	    {
	      printf ("gdam123 version %s.\n"
		      "(C) Copyright Dave Benson, "
		      "Geoff Matters. 2000.\n"
		      "\n"
		      "This program comes WITH NO WARRANTY! "
		      "USE AT YOUR OWN RISK.\n", VERSION);
	      exit (0);
	    }
	  if (strcmp (argv[i] + 1, "-help") == 0)
	    {
	      usage ();
	      continue;
	    }
	  if (strcmp (argv[i] + 1, "-device") == 0)
	    {
	      if (argv[++i] == NULL)
		usage ();
	        channel->device_list = g_list_append (channel->device_list, 
						       argv[i]);
	      continue;
	    }
	  if (strcmp (argv[i] + 1, "z") == 0)
	    {
	      do_shuffle = TRUE;
	      continue;
	    }
	  g_warning ("invalid option: %s", argv[i]);
	  continue;
	}
      gdam_producer_queue_source_file (queue, argv[i]);
    }

  if (channel->device_list == NULL)
    {
      char *device_name = getenv ("GDAM_DEVICE");
      if (device_name)
	channel->device_list = g_list_append (channel->device_list, 
	    				       device_name);
    }

  if (channel->device_list == NULL)
    {
      g_message ("Using /dev/dsp since no --device was given");
      channel->device_list = g_list_append (channel->device_list, "/dev/dsp");
    }

  if (queue->songinfo_list == NULL)
    {
      gdam_debug ("gdam123: NULL songinfo_list");
      usage ();
      exit (0);
    }

  g_list_foreach (channel->device_list, do_open_device, toplevel);

  gdam_turntable_set_producer (turntable, producer);
  gdam_filter_set_child (GDAM_FILTER (gain), GDAM_MODEL (turntable));
  gdam_toplevel_add_child (toplevel, GDAM_MODEL (gain));
  gtk_signal_connect (GTK_OBJECT (producer),
		      "set_filename", on_set_filename, NULL);
  gtk_signal_connect (GTK_OBJECT (producer),
		      "source_ended", on_song_ended, NULL);
  gtk_signal_connect (GTK_OBJECT (turntable),
		      "set_position",
		      GTK_SIGNAL_FUNC (on_set_position), NULL);

  //bzero(&sigact, sizeof(struct sigaction));
  //sigact.sa_flags = SA_RESTART;
  //sigact.sa_handler = sig_winch;
  //if (sigaction(SIGWINCH, &sigact, NULL) < 0)
  //g_error("signal error");

  if (use_termios)
    set_raw ();

  g_source_add (G_PRIORITY_DEFAULT, TRUE, &ctrl_source_funcs, NULL, NULL,
		NULL);

  keyboard_io_watch_release ();

  gdam123rc = gdam_search_path ("gdam123rc", gdam_search_path_config,
				GDAM_SEARCH_PATH_FIND_FILE);
  if (gdam123rc)
    {
      gdam_debug ("using rc file %s", gdam123rc);
      if (!source_ex_file (gdam123rc))
	{
	  g_warning ("error sourcing rc file %s", gdam123rc);
	}
      g_free (gdam123rc);
      gdam123rc = NULL;
    }
  //sort_smart();
  //gdam_producer_queue_print_queue(GDAM_PRODUCER_QUEUE(producer));

  if (do_shuffle)
    gdam_producer_queue_shuffle_list (queue);
  gdam_producer_queue_skip_to (queue, 1);
  while (g_main_iteration (TRUE))
    ;
  return 0;
}
