/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdammp3info.h>
#include <gdam/component/gdamsongdatabase.h>
#include <gtk/gtksignal.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <gdam/common/gdamautoconf.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

/* Strategy:  we should take a bunch of filters, possibly inverted, and apply
 * all that are given on the command line. */

gboolean print_as_file_list = FALSE;
gboolean warn_missing = FALSE;
gboolean warn_unused = FALSE;
gboolean warn_missing_inverted = FALSE;
gboolean warn_unused_inverted = FALSE;
GSList *match_list = NULL;
GSList *miss_list = NULL;


static inline void 
list_file (const char *format, const char *filename)
{
  if (print_as_file_list)
    fprintf (stdout, "%s ", filename);
  else
    fprintf (stdout, format, filename);
}

static int 
traverse_func (gpointer key, gpointer value, gpointer data)
{
  GdamSongInfo *si = (GdamSongInfo *) value;
  int match = 0;
  if (warn_missing)
    {
      struct stat stat_struct;
      int rc = stat (si->filename, &stat_struct);
      if ((rc == 0 && warn_missing_inverted) || 
	  (rc < 0  && !warn_missing_inverted))
	{
	  match = TRUE;
	}
      else 
	{
	  gboolean wrong_type;
	  wrong_type = !S_ISREG (stat_struct.st_mode) && 
	               !S_ISLNK (stat_struct.st_mode);
	  if ((!wrong_type  && warn_missing_inverted) ||
	      (wrong_type && !warn_missing_inverted))
	    match = TRUE;
	  else
	    goto failed;
	}
    }
  if (warn_unused)
    {
      gboolean no_bpm, unused;
      GHashTable *props = (GHashTable *) si->song_properties;
      gpointer val = g_hash_table_lookup (props, "backup");
      gdouble bpm;
      bpm = si->beat_info ? gdam_beat_info_get_bpm (si->beat_info, 0) : 0;
      no_bpm = (bpm <= 0);
      unused = (no_bpm && !val);

      if ((unused && !warn_unused_inverted) || 
	  (!unused  && warn_unused_inverted))
	match = TRUE;
      else
	goto failed;
    }
  if (!match)
    goto failed;

  list_file ("song %s matched\n", si->filename);
  match_list = g_slist_append (match_list, si);
  return 0;

failed:
  miss_list = g_slist_append (miss_list, si);
  return 0;
}

static gboolean usage_printed = FALSE;
static void usage()
{
  if (usage_printed)
    return;
  fprintf (stderr, "Command-Line Song Database Tool for GDAM, used for\n"
                   "listing songs which meet certain requirements and\n"
                   "general song database management.\n"
	   "\n"
	   "usage:   gdamdbtool [options] [--output_filename] songdb_file\n"
	   "filter options:\n"
	   "   --missing        list songs cannot be found on disk\n"
	   "   --unused         songs with no beat info and no 'backup' tag\n"
	   "filter options can be inverted, i.e.  --not-missing\n"
	   "\n"
	   "output options:\n"
	   "   --filename-list  output is a list of filenames (for scripting)\n"
	   "   --output fname   output filename of output database\n"
	   "   --remove         remove listed songs from output database\n"
	   "   --keep           remove non-listed songs from output database\n"
	   );
  usage_printed = TRUE;
}



int main(int argc, char** argv)
{
  int i;
  char *song_db_filename = NULL;
  GIOChannel *kbd_channel = g_io_channel_unix_new (0);
  GdamSongDatabase *database;

  g_set_prgname (*argv);
  gtk_type_init ();
  gtk_signal_init ();

  if (argc == 1)
    {
      usage ();
      exit (0);
    }

  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] == '-')
	{
	  if (strcmp (argv[i], "-v") == 0
	      || strcmp (argv[i], "--version") == 0)
	    {
	      printf ("gdamdbtool version %s.\n"
		      "(C) Copyright Dave Benson, "
		      "Geoff Matters. 2001.\n"
		      "\n"
		      "This program comes WITH NO WARRANTY! "
		      "USE AT YOUR OWN RISK.\n", VERSION);
	      exit (0);
	    }
	  if (strcmp (argv[i] + 1, "-usage") == 0)
	    {
	      usage ();
	      continue;
	    }
	  /* List any song in the database which cannot be found on disk */
	  if (strcmp (argv[i] + 1, "-missing") == 0)
	    {
	      warn_missing = TRUE;
	      continue;
	    }
	  if (strcmp (argv[i] + 1, "-not-missing") == 0)
	    {
	      warn_missing = TRUE;
	      warn_missing_inverted = TRUE;
	      continue;
	    }
	  /* Anything with beat info or a "backup" tag is good... rest listed */
	  if (strcmp (argv[i] + 1, "-unused") == 0)
	    {
	      warn_unused = TRUE;
	      continue;
	    }
	  if (strcmp (argv[i] + 1, "-not-unused") == 0)
	    {
	      warn_unused = TRUE;
	      warn_unused_inverted = TRUE;
	      continue;
	    }
	  if (strcmp (argv[i] + 1, "-filename-list") == 0)
	    {
	      print_as_file_list = TRUE;
	      continue;
	    }
	  g_warning ("invalid option: %s", argv[i]);
	  continue;
	}
      song_db_filename = argv[i];
    }

  if (!song_db_filename)
    {
      usage();
      return 1;
    }

  database = gdam_song_database_new_from_file (song_db_filename);

  if (!database)
    {
      g_warning ("couldn't open song database %s", song_db_filename);
    }

  g_tree_traverse (database->by_order, traverse_func, G_IN_ORDER, NULL);

  return 0;
}
