/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 *
 * The ALSA driver request the alsa-library header to
 * be installed.
 *
 * see
 *      http://www.alsa-project.org/
 *
 *
 *
 * Under debian you can get the necessary packages via apt:
 *
 *       apt-get install libasound0.4-dev 
 *
 * [daveb: 07 May 2000]
 */

#include <gdam/common/gdamautoconf.h>

#if  HAVE_ALSA

#if HAVE_ALSA9
#include <alsa/asoundlib.h>
#else
#include <sys/asoundlib.h>
#endif
#include "gdamdevicealsa.h"
#include <sys/types.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>


#if HAVE_ALSA9
/* ALSA 0.9.x API */
/* added by Takashi Iwai <tiwai@suse.de> */

static void gdam_device_alsa_init(GdamDeviceAlsa* device_alsa)
{
}

/* Convert output_device->options => SND_PCM_FORMAT_*. */
static snd_pcm_format_t options_to_alsa_fmt(const GdamOutputDeviceOptions* options)
{
  guint num_bytes_per_sample, is_big_endian, is_signed;
  /* flags[num_bytes_per_sample][is_big_endian][is_signed] */
  static snd_pcm_format_t flags[4][2][2] = {
    {{SND_PCM_FORMAT_U8, SND_PCM_FORMAT_S8},
     {SND_PCM_FORMAT_U8, SND_PCM_FORMAT_S8}},
    {{SND_PCM_FORMAT_U16_LE, SND_PCM_FORMAT_S16_LE},
     {SND_PCM_FORMAT_U16_BE, SND_PCM_FORMAT_S16_BE}},
    {{SND_PCM_FORMAT_U24_LE, SND_PCM_FORMAT_S24_LE},
     {SND_PCM_FORMAT_U24_BE, SND_PCM_FORMAT_S24_BE}},
    {{SND_PCM_FORMAT_U32_LE, SND_PCM_FORMAT_S32_LE},
     {SND_PCM_FORMAT_U32_BE, SND_PCM_FORMAT_S32_BE}}
  };
  if (options->bits_per_sample % 8 != 0
      || options->bits_per_sample <= 0 || options->bits_per_sample > 32)
    {
      g_warning ("cannot use %d bits/sample; using 16 bits/sample",
		 options->bits_per_sample);
      num_bytes_per_sample = 2;
    }
  else
    {
      num_bytes_per_sample = options->bits_per_sample / 8;
    }

  is_big_endian = (G_BIG_ENDIAN == G_BYTE_ORDER) ? 1 : 0;
  if (options->is_wrong_endian)
    is_big_endian = 1 - is_big_endian;

  is_signed = options->is_signed ? 1 : 0;

  return flags[num_bytes_per_sample - 1][is_big_endian][is_signed];
}

static gboolean gdam_device_alsa_open_fd(GdamOutputDevice* output_device)
{
  struct pollfd pfd;
  const char *device_name;
  snd_pcm_format_t format;
  int err;
  int period_size, buffer_size, rate;
  snd_pcm_t *alsa_handle;
  snd_pcm_hw_params_t *hw;
  GdamOutputDeviceOptions *options = &output_device->options;
  GdamDeviceAlsa *device_alsa;

  device_name = GDAM_SINK (output_device)->device_name;
  device_alsa = GDAM_DEVICE_ALSA (output_device);

  if ((err = snd_pcm_open (&alsa_handle, device_name, SND_PCM_STREAM_PLAYBACK,
			   SND_PCM_NONBLOCK)) < 0)
    {
      g_warning ("pcm device %s couldn't be opened", device_name);
      return FALSE;
    }

  snd_pcm_hw_params_alloca(&hw);
  snd_pcm_hw_params_any(alsa_handle, hw);

  format = options_to_alsa_fmt(options);

  if (snd_pcm_hw_params_set_access(alsa_handle, hw, SND_PCM_ACCESS_RW_INTERLEAVED) < 0) {
    g_warning ("ALSA: cannot set interleaved");
    return FALSE;
  }
  if (snd_pcm_hw_params_set_format(alsa_handle, hw, format) < 0) {
    g_warning ("ALSA: cannot set format %d", format);
    return FALSE;
  }
  if ((rate = snd_pcm_hw_params_set_rate_near(alsa_handle, hw, options->sampling_rate, 0)) < 0) {
    g_warning ("ALSA: cannot set rate %d", options->sampling_rate);
    return FALSE;
  }
  if (snd_pcm_hw_params_set_channels(alsa_handle, hw, options->num_channels) < 0) {
    g_warning ("ALSA: cannot set channels %d", options->num_channels);
    return FALSE;
  }

  period_size = (1 << options->log_frag_size);
  if ((period_size = snd_pcm_hw_params_set_period_size_near(alsa_handle, hw, period_size, 0)) < 0) {
    g_warning ("ALSA: cannot set period size");
    return FALSE;
  }
  buffer_size = period_size * options->num_fragments;
  if ((buffer_size = snd_pcm_hw_params_set_buffer_size_near(alsa_handle, hw, buffer_size)) < 0) {
    g_warning ("ALSA: cannot set buffer size");
    return FALSE;
  }

  if ((err = snd_pcm_hw_params(alsa_handle, hw)) < 0) {
    g_warning ("ALSA: cannot set hw_params: %s", snd_strerror(err));
    return FALSE;
  }

  if (snd_pcm_poll_descriptors_count(alsa_handle) > 1)
    g_warning("ALSA: more than 1 poll descriptors; using the first one..");

  if (snd_pcm_poll_descriptors(alsa_handle, &pfd, 1) < 0) {
    g_warning("ALSA: cannot find poll descriptor");
    return FALSE;
  }

  options->sampling_rate = rate;
  options->block_size = period_size;
  output_device->device_fd = pfd.fd;
  device_alsa->alsa_handle = alsa_handle;
  return TRUE;
}

static void gdam_device_alsa_close_fd(GdamOutputDevice* output_device)
{
  GdamDeviceAlsa *device_alsa = GDAM_DEVICE_ALSA (output_device);
  if (device_alsa->alsa_handle)
    {
      snd_pcm_close(device_alsa->alsa_handle);
      device_alsa->alsa_handle = NULL;
    }
}

int    gdam_device_alsa_write_pcm     (GdamOutputDevice*  device, 
				       gconstpointer      buffer, 
				       int                length)
{
  snd_pcm_t *handle = GDAM_DEVICE_ALSA (device)->alsa_handle;
  int rv;

  length = snd_pcm_bytes_to_frames(handle, length);
  rv = snd_pcm_writei (handle, buffer, length);

  if (rv < length)
    {
      if (rv < 0) {
	if (rv == -EPIPE) {
	  /* underrun */
	  if (snd_pcm_prepare(handle) < 0)
	    g_warning("ALSA: prepare failed");
	} else if (rv == -ESTRPIPE) {
	  while ((rv = snd_pcm_resume(handle)) == -EAGAIN)
	    sleep(1);	/* wait until suspend flag is released */
	  if (rv < 0)
	    if (snd_pcm_prepare(handle) < 0)
	      g_warning("ALSA: prepare failed");
	} else
	  g_message ("alsa: pcm_write returned error '%s'", snd_strerror (rv));
      }
      else
	g_message ("alsa: wrote only %d of %d samples", rv, length);
      rv = length;
    }

  return snd_pcm_frames_to_bytes(handle, rv);
}


#else
/* ALSA 0.5.x API */

typedef struct _GdamAlsaCard GdamAlsaCard;

static void gdam_device_alsa_init(GdamDeviceAlsa* device_alsa)
{
}


/* Global state.
 * 
 * We save this data around, since it is unlikely to change
 * over a server's lifetime.
 */
struct _GdamAlsaCard {
	int			alsa_index;
	snd_ctl_t*		device_handle;
	struct snd_ctl_hw_info	hw_info;
	int			num_pcm_devices;
	snd_pcm_t**		pcm_devices;
	int*			num_subdevices;
	snd_pcm_t***		pcm_subdevices;

	/* perhaps an option when the are error parsing the device_id,
	 * but unused and should probably be deleted. */
	snd_pcm_t*		default_pcm_device;
};


static GdamAlsaCard* gdam_alsa_card_new_from_index(int card_index)
{
  GdamAlsaCard *alsa_card;
  int ret_code;
  int num_devices;
  int device_index;

  alsa_card = g_new0 (GdamAlsaCard, 1);

  /* Open the card. */
  ret_code = snd_ctl_open (&alsa_card->device_handle, card_index);
  if (ret_code < 0)
    {
      g_warning ("error opening alsa card %d: %s",
		 card_index, snd_strerror (ret_code));
      g_free (alsa_card);
      return NULL;
    }

  /* Query the hardware description. */
  ret_code = snd_ctl_hw_info (alsa_card->device_handle, &alsa_card->hw_info);
  if (ret_code < 0)
    {
      g_warning ("error querying alsa card %d: %s",
		 card_index, snd_strerror (ret_code));
      g_free (alsa_card);
      return NULL;
    }
  snd_ctl_close (alsa_card->device_handle);
  alsa_card->device_handle = NULL;

  /* Find the pcm devices on that card. */
  num_devices = alsa_card->hw_info.pcmdevs;
  g_return_val_if_fail (num_devices != 0, NULL);

  alsa_card->num_pcm_devices = num_devices;
  alsa_card->pcm_devices = g_new0 (snd_pcm_t *, num_devices);
  alsa_card->num_subdevices = g_new0 (int, num_devices);
  alsa_card->pcm_subdevices = g_new0 (snd_pcm_t **, num_devices);
  return alsa_card;
}

static GdamAlsaCard* gdam_alsa_card_get(int card_id)
{
  static int num_cards = 0;
  static GdamAlsaCard **alsa_cards = NULL;
  static gboolean inited = FALSE;
  int i;

  guint alsa_mask;

  if (inited)
    {
      if (card_id >= num_cards || card_id < 0)
	{
	  return NULL;
	}
      return alsa_cards[card_id];
    }

  inited = TRUE;

  alsa_mask = snd_cards_mask ();
  if (alsa_mask == 0)
    {
      g_warning ("no alsa sound cards were found");
      return NULL;
    }

  num_cards = sizeof (guint) * 8;
  alsa_cards = g_new (GdamAlsaCard *, num_cards);

  for (i = 0; i < num_cards; i++)
    {
      if (((1 << i) & alsa_mask) != 0)
	{
	  GdamAlsaCard *alsa_card;
	  alsa_card = gdam_alsa_card_new_from_index (i);
	  if (alsa_card == NULL)
	    g_warning ("error opening alsa card %d", i);
	  else
	    gdam_debug ("got alsa card %d", i);
	  alsa_cards[i] = alsa_card;
	}
      else
	{
	  alsa_cards[i] = NULL;
	  /*g_message("no mask for card %d", i);*/
	}
    }
  return alsa_cards[card_id];
}

static snd_pcm_t* gdam_alsa_pcm_device_get(int card_index, 
					   int device_index,
					   int subdevice_index)
{
  int ret_code;
  snd_pcm_t *handle;
  GdamAlsaCard *alsa_card;
  snd_pcm_info_t pcm_info;
  int playback = -1;
  handle = NULL;

  alsa_card = gdam_alsa_card_get (card_index);
  g_return_val_if_fail (alsa_card != NULL, NULL);
  if (device_index >= alsa_card->num_pcm_devices)
    {
      g_warning ("device %d out of range", device_index);
      return NULL;
    }

  /*  If we already have the needed device, return it. */
  if (subdevice_index < 0 && alsa_card->pcm_devices[device_index] != NULL)
    {
      /*g_message("returning C%dP%d from array", card_index, device_index);*/
      return alsa_card->pcm_devices[device_index];
    }
  else if (subdevice_index >= 0
	   && alsa_card->pcm_subdevices[device_index] != NULL
	   && alsa_card->num_subdevices != NULL
	   && subdevice_index < alsa_card->num_subdevices[device_index]
	   && alsa_card->pcm_subdevices[device_index][subdevice_index]
	   != NULL)
    {
      /*g_message("returning C%dP%d#%d from array", card_index, device_index, subdevice_index);*/
      return alsa_card->pcm_subdevices[device_index][subdevice_index];
    }

  /*  Query device info to see if subdevice is available.  */
  ret_code = snd_ctl_open (&alsa_card->device_handle, card_index);
  ret_code = snd_ctl_pcm_info (alsa_card->device_handle,
			       device_index, &pcm_info);
  if (ret_code < 0)
    g_warning ("error querying info for device %d on card %d: %s",
	       device_index, card_index, snd_strerror (ret_code));
  else
    playback = pcm_info.playback;
  snd_ctl_close (alsa_card->device_handle);
  alsa_card->device_handle = NULL;

  /*  If device cannot play back, return 0. */
  alsa_card->num_subdevices[device_index] = pcm_info.playback + 1;
  if ((int) pcm_info.playback < 0)
    {
      alsa_card->pcm_devices[device_index] = NULL;
      return NULL;
    }
  alsa_card->pcm_subdevices[device_index] =
    g_new0 (snd_pcm_t *, pcm_info.playback + 1);

  /*  If we are simply opening the device.  */
  if (subdevice_index < 0)
    {
      ret_code = snd_pcm_open (&handle,
			       card_index,
			       device_index, SND_PCM_OPEN_PLAYBACK);
      if (ret_code < 0)
	{
	  g_warning ("error opening alsa device %d on card %d: %s",
		     device_index, card_index,
		     //snd_strerror(ret_code));
		     strerror (-1 * ret_code));
	  handle = NULL;
	}
      /*g_message("pcm_opened C%dP%d ", card_index, device_index);*/
    }
  else
    {
      ret_code = snd_pcm_open_subdevice (&handle,
					 card_index,
					 device_index,
					 subdevice_index,
					 SND_PCM_OPEN_PLAYBACK);
      if (ret_code < 0)
	{
	  g_warning ("error opening alsa subdevice %d on device %d"
		     " on card %d: %s",
		     subdevice_index, device_index, card_index,
		     strerror (-1 * ret_code));
	  handle = NULL;
	}
      /*g_message("pcm_opened C%dP%d#%d ", card_index, device_index, subdevice_index);*/
    }

  /*  If we have opened [sub]device, flush it.  If this isn't done, 
   *  some cards seem to hang if they experience an underrun as
   *  they are first opened. 
   */
  if (handle)
    {
      struct snd_pcm_channel_info cinfo;
      int rc;

      memset (&cinfo, 0, sizeof (cinfo));
      cinfo.channel = SND_PCM_CHANNEL_PLAYBACK;
      if ((rc = snd_pcm_channel_info (handle, &cinfo)) < 0)
	{
	  g_warning ("channel %p info error: %s\n",
		     handle, snd_strerror (rc));
	}

      rc = snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_PLAYBACK);
      if (rc < 0)
	{
	  gdam_debug ("error flushing pcm playback channel"
		      " %p : %s", handle, snd_strerror (rc));
	}
    }

  /*  Store device for easy retrieval on future calls  */
  if (subdevice_index < 0)
    {
      alsa_card->pcm_devices[device_index] = handle;
    }
  else
    {
      g_return_val_if_fail (alsa_card->pcm_subdevices != NULL, NULL);
      g_return_val_if_fail (alsa_card->pcm_subdevices[device_index] != NULL,
			    NULL);
      alsa_card->pcm_subdevices[device_index][subdevice_index] = handle;
    }
  return handle;
}




/* XXX: this should parse the two LAST numbers in the name!
 *      [not the two FIRST as it does now]; so it screws up:
 *
 *           /mnt/dev1/snd/pcmC0D0
 *      unlikely as that is...
 *
 *      Of course, in such a case you REALLY should just say
 *
 *           pcmC0D0
 *
 *           CHANGE by geoff:
 *           pcmC0D0#0 means subdevice (#) 0 on device (D) 0 on card (C) 0
 *
 */
static gboolean 
gdam_device_alsa_get_card_and_device(const char* device_name,
				     int*        card_num_out,
				     int*        device_num_out,
				     int*        subdevice_num_out)
{
  int id;

  *subdevice_num_out = -1;

  while (*device_name && !isdigit (*device_name))
    device_name++;
  *card_num_out = atoi (device_name);
  if (*card_num_out >= 8 * (int) sizeof (int))
    return FALSE;

  /* skip past the number. */
  while (*device_name && isdigit (*device_name))
    device_name++;

  /* and the subsequent nonnumber. */
  while (*device_name && !isdigit (*device_name))
    device_name++;

  *device_num_out = atoi (device_name);
  if (*device_num_out < 0)
    return FALSE;

  /* skip past the number. */
  while (*device_name && isdigit (*device_name))
    device_name++;

  /* and the subsequent nonnumber. */
  while (*device_name && !isdigit (*device_name))
    device_name++;

  if (*device_name)
    *subdevice_num_out = atoi (device_name);

  return TRUE;
}

/* Convert output_device->options => SND_PCM_SFMT_*. */
static guint options_to_alsa_fmt(const GdamOutputDeviceOptions* options)
{
  guint num_bytes_per_sample, is_big_endian, is_signed;
  /* flags[num_bytes_per_sample][is_big_endian][is_signed] */
  static guint flags[4][2][2] = {
    {{SND_PCM_SFMT_U8, SND_PCM_SFMT_S8},
     {SND_PCM_SFMT_U8, SND_PCM_SFMT_S8}},
    {{SND_PCM_SFMT_U16_LE, SND_PCM_SFMT_S16_LE},
     {SND_PCM_SFMT_U16_BE, SND_PCM_SFMT_S16_BE}},
    {{SND_PCM_SFMT_U24_LE, SND_PCM_SFMT_S24_LE},
     {SND_PCM_SFMT_U24_BE, SND_PCM_SFMT_S24_BE}},
    {{SND_PCM_SFMT_U32_LE, SND_PCM_SFMT_S32_LE},
     {SND_PCM_SFMT_U32_BE, SND_PCM_SFMT_S32_BE}}
  };
  if (options->bits_per_sample % 8 != 0
      || options->bits_per_sample <= 0 || options->bits_per_sample > 32)
    {
      g_warning ("cannot use %d bits/sample; using 16 bits/sample",
		 options->bits_per_sample);
      num_bytes_per_sample = 2;
    }
  else
    {
      num_bytes_per_sample = options->bits_per_sample / 8;
    }

  is_big_endian = (G_BIG_ENDIAN == G_BYTE_ORDER) ? 1 : 0;
  if (options->is_wrong_endian)
    is_big_endian = 1 - is_big_endian;

  is_signed = options->is_signed ? 1 : 0;

  return flags[num_bytes_per_sample][is_big_endian][is_signed];
}

typedef struct _AlsaFormat AlsaFormat;
struct _AlsaFormat {
	/* alsa_format must be first. */
	guint		alsa_format;
	guint		bits_per_sample;
	guint		bytes_per_sample;
	gboolean        is_big_endian;
	gboolean        is_signed;
};

static int compare_uints(gconstpointer ptr1, gconstpointer ptr2)
{
  guint u1 = *(const guint *) ptr1;
  guint u2 = *(const guint *) ptr2;

  if (u1 < u2)
    return -1;
  if (u1 > u2)
    return +1;
  return 0;
}

static gboolean    alsa_fmt_to_options(guint                    format,
                                       GdamOutputDeviceOptions* options)
{
  AlsaFormat *format_out;
  static AlsaFormat formats[] = {
    {SND_PCM_SFMT_S8, 8, 1, FALSE, TRUE},
    {SND_PCM_SFMT_U8, 8, 1, FALSE, FALSE},
    {SND_PCM_SFMT_S16_BE, 16, 2, TRUE, TRUE},
    {SND_PCM_SFMT_U16_BE, 16, 2, TRUE, FALSE},
    {SND_PCM_SFMT_S16_LE, 16, 2, FALSE, TRUE},
    {SND_PCM_SFMT_U16_LE, 16, 2, FALSE, FALSE},
    {SND_PCM_SFMT_S24_BE, 24, 4, TRUE, TRUE},
    {SND_PCM_SFMT_U24_BE, 24, 4, TRUE, FALSE},
    {SND_PCM_SFMT_S24_LE, 24, 4, FALSE, TRUE},
    {SND_PCM_SFMT_U24_LE, 24, 4, FALSE, FALSE},
    {SND_PCM_SFMT_S32_BE, 32, 4, TRUE, TRUE},
    {SND_PCM_SFMT_U32_BE, 32, 4, TRUE, FALSE},
    {SND_PCM_SFMT_S32_LE, 32, 4, FALSE, TRUE},
    {SND_PCM_SFMT_U32_LE, 32, 4, FALSE, FALSE},
  };
  static gboolean initialized = FALSE;

  if (!initialized)
    {
      initialized = TRUE;
      qsort (formats,
	     sizeof (formats) / sizeof (formats[0]),
	     sizeof (AlsaFormat), compare_uints);
    }

  format_out = bsearch (&format,
			formats,
			sizeof (formats) / sizeof (formats[0]),
			sizeof (AlsaFormat), compare_uints);
  if (format_out == NULL)
    {
      g_warning ("couldn't find SND_PCM_* for %d", format);
      return FALSE;
    }
  options->is_signed = format_out->is_signed;
  options->bytes_per_sample = format_out->bytes_per_sample;
  options->bits_per_sample = format_out->bits_per_sample;
  options->is_wrong_endian = format_out->is_big_endian;
  if (G_LITTLE_ENDIAN == G_BYTE_ORDER)
    options->is_wrong_endian = !options->is_wrong_endian;
  options->is_signed = format_out->is_signed;
  return TRUE;
}

static void print_channel_status(int status){
	char* status_str;
	status_str = NULL;
	if (status == SND_PCM_STATUS_NOTREADY)
		status_str = "channel is not ready ";
	if (status == SND_PCM_STATUS_READY)
		status_str = "channel is ready for prepare call ";
	if (status == SND_PCM_STATUS_PREPARED)
		status_str = "channel is ready to go ";
	if (status == SND_PCM_STATUS_RUNNING)
		status_str = "channel is running ";
	if (status == SND_PCM_STATUS_UNDERRUN)
		status_str = "channel reached an underrun and it is not ready ";
	if (status == SND_PCM_STATUS_OVERRUN)
		status_str = "channel reached an overrun and it is not ready ";
	if (status == SND_PCM_STATUS_PAUSED)
		status_str = "channel is paused ";
}

static void print_params(snd_pcm_channel_params_t* params, int bytes_per_sample)
{
#if 0
  g_message ("channel_params: channel=%d, mode=%d, "
	     "start_mode=%d, stop_mode=%d, time=%d, ust_time=%d\n",
	     params->channel, params->mode, params->start_mode,
	     params->stop_mode, params->time, params->ust_time);
  g_message ("channel_params: format: interleave=%d, format=%d, "
	     "rate=%d, voices=%d, special=%d\n",
	     params->format.interleave, params->format.format,
	     params->format.rate, params->format.voices,
	     params->format.special);

#endif
  if (params->mode == SND_PCM_MODE_BLOCK)
    {
#if 0
      g_message ("channel_params: block: frag_size=%d, frags_min/max=%d/%d", 
                 params->buf.block.frag_size,
                 params->buf.block.frags_min,
                 params->buf.block.frags_max);
#endif
      gdam_debug ("Set soundcard fragments:  "
		  "%d to %d frags of %d bytes; %3.3fs",
		  params->buf.block.frags_min,
		  params->buf.block.frags_max,
		  params->buf.block.frag_size,
		  (float) params->buf.block.frag_size *
		          params->buf.block.frags_max
			  / 44100.0 / bytes_per_sample / 2);
    }
  else
    {
#if 0
      g_message ("channel_params: stream: queue_size=%d, "
		 "fill=%d, max_fill=%d\n",
		 params->buf.stream.queue_size,
		 params->buf.stream.fill, params->buf.stream.max_fill);
#endif
    }
}


static gboolean gdam_device_alsa_open_fd(GdamOutputDevice* output_device)
{
  int fd;
  const char *device_name;
  int card_index;
  int device_index;
  int subdevice_index;
  guint card_mask;
  int ret_code;

  snd_pcm_t *alsa_handle;
  snd_pcm_format_t format;
  snd_pcm_channel_setup_t setup;

  snd_pcm_channel_params_t channel_params;
  snd_pcm_channel_info_t channel_info;
  snd_pcm_channel_status_t status;

  GdamOutputDeviceOptions *options = &output_device->options;
  GdamDeviceAlsa *device_alsa;
  GdamAlsaCard *card;

  device_name = GDAM_SINK (output_device)->device_name;
  device_alsa = GDAM_DEVICE_ALSA (output_device);

  if (!gdam_device_alsa_get_card_and_device (device_name,
					     &card_index,
					     &device_index, &subdevice_index))
    {
      g_warning ("error parsing device_ids");
    }

  card = gdam_alsa_card_get (card_index);
  if (card == NULL)
    {
      g_warning ("couldn't get alsa card %d", card_index);
      return FALSE;
    }

  if (device_index < 0 || device_index >= card->num_pcm_devices)
    {
      g_warning ("device index %d was out of range", device_index);
      return FALSE;
    }

  g_return_val_if_fail (card->pcm_devices != NULL, FALSE);
  alsa_handle = gdam_alsa_pcm_device_get (card_index,
					  device_index, subdevice_index);
  if (alsa_handle == NULL)
    {
      g_warning ("pcm device C%dD%d#%d couldn't be opened",
		 card_index, device_index, subdevice_index);
      return FALSE;
    }
  else
    {
      /*g_message("opened pcm device C%dD%d#%d", card_index, device_index, subdevice_index);*/
    }


  /* XXX: need error handling for options_to_alsa_fmt */
  bzero (&format, sizeof (format));
  //format.format = options_to_alsa_fmt(options);
  format.format = SND_PCM_SFMT_S16_LE;
  format.rate = 44100;
  format.interleave = 1;
  format.voices = options->num_channels;

  /* duplex/record support eventually ... */


  memset (&channel_info, 0, sizeof (channel_info));
  channel_info.channel = SND_PCM_CHANNEL_PLAYBACK;
  ret_code = snd_pcm_channel_info (alsa_handle, &channel_info);
  if (ret_code < 0)
    {
      g_warning ("error setting channel info: %s", snd_strerror (ret_code));
    }

  /*  Seems to return < 0 normally, ignore result  */
  snd_pcm_channel_flush (alsa_handle, SND_PCM_CHANNEL_PLAYBACK);

  snd_pcm_munmap (alsa_handle, SND_PCM_CHANNEL_PLAYBACK);

  /*  Seems to return < 0 normally, ignore result  */
  snd_pcm_channel_flush (alsa_handle, SND_PCM_CHANNEL_PLAYBACK);

  memset (&channel_params, 0, sizeof (channel_params));
  channel_params.channel = SND_PCM_CHANNEL_PLAYBACK;
  channel_params.mode = SND_PCM_MODE_BLOCK;
  channel_params.format = format;
/*snd_pcm_digital_t digital; *//* digital setup */

  channel_params.start_mode = SND_PCM_START_DATA;
#ifdef SND_PCM_STOP_STOP
  channel_params.stop_mode = SND_PCM_STOP_STOP;
#endif
  channel_params.time = 0;
  channel_params.ust_time = 0;
/*snd_pcm_sync_t sync; *//* sync group */
  channel_params.buf.block.frag_size = (1 << options->log_frag_size);
  channel_params.buf.block.frags_min = 1;
  channel_params.buf.block.frags_max = options->num_fragments;

  print_params (&channel_params, 2);
  ret_code = snd_pcm_channel_params (alsa_handle, &channel_params);
  if (ret_code < 0)
    {
      g_warning ("error setting sound card parameters: %s",
		 snd_strerror (ret_code));
    }

  fd = snd_pcm_file_descriptor (alsa_handle, SND_PCM_CHANNEL_PLAYBACK);

  if (snd_pcm_plugin_prepare (alsa_handle, SND_PCM_CHANNEL_PLAYBACK) < 0)
    {
      g_warning ("snd_pcm_plugin_prepare failed");
    }

  memset (&setup, 0, sizeof (setup));
  setup.mode = SND_PCM_MODE_BLOCK;
  setup.channel = SND_PCM_CHANNEL_PLAYBACK;
  if (snd_pcm_channel_setup (alsa_handle, &setup) < 0)
    {
      g_warning ("snd_pcm_channel_setup failed");
    }


  /*
   * XXX: should we read the data back out of the device,
   *      as we do for deviceoss [which needs it]?
   */

  options->block_size = (1 << options->log_frag_size)
    * 8 / options->bits_per_sample;
  output_device->device_fd = fd;
  device_alsa->alsa_handle = alsa_handle;
  return TRUE;
}

static void gdam_device_alsa_close_fd(GdamOutputDevice* output_device)
{
  if (output_device->device_fd >= 0)
    {
    }
}

int    gdam_device_alsa_write_pcm     (GdamOutputDevice*  device, 
				       gconstpointer      buffer, 
				       int                length)
{
  snd_pcm_t *handle = GDAM_DEVICE_ALSA (device)->alsa_handle;
  snd_pcm_channel_status_t status;
  int rv;
  int ret_code;
  rv = snd_pcm_write (handle, buffer, length);

  if (rv < length)
    {
      if (rv < 0)
	g_message ("alsa: pcm_write returned error '%s'", snd_strerror (rv));
      else
	g_message ("alsa: wrote only %d of %d samples", rv, length);
      rv = length;
    }

  /* Always check status */
  memset (&status, 0, sizeof (status));
  status.channel = SND_PCM_CHANNEL_PLAYBACK;
  if ((ret_code = snd_pcm_channel_status (handle, &status)) < 0)
    {
      g_warning ("alsa: error checking channel status: %s",
		 snd_strerror (ret_code));
    }

#if 0
  /* If we had an underrun, prepare the channel (old-style) */
  if (status.status == SND_PCM_STATUS_UNDERRUN)
    {
      if (snd_pcm_plugin_prepare (handle, SND_PCM_CHANNEL_PLAYBACK) < 0)
	g_warning ("underrun: playback channel prepare error");
      g_message ("underrun");
    }
#endif

  /* If we had an underrun, flush the channel and rewrite the sound
   * to the device (alsaplayer-style) 
   */
  if (status.underrun)
    {
      /*  Flush channel */
      /* XXX  do we want to flush the channel, or does that screw
       * up our device timing ?
       */
      ret_code = snd_pcm_channel_flush (handle, SND_PCM_CHANNEL_PLAYBACK);
      if (ret_code < 0)
	g_warning ("alsa: underrun, playback flush error: %s",
		   snd_strerror (ret_code));

      /*  Prepare the plugin  */
      /* snd_pcm_playback_prepare(sound_handle);  from aplayer */
      ret_code = snd_pcm_playback_prepare (handle);
      if (ret_code < 0)
	g_warning ("alsa: underrun, playback playback_prepare error: %s",
		   snd_strerror (ret_code));
      if ((ret_code = snd_pcm_write (handle, buffer, length)) < 0)
	g_warning ("alsa: underrun, playback re-write error: %s",
		   snd_strerror (ret_code));

      /* Recheck channel status */
      memset (&status, 0, sizeof (status));
      status.channel = SND_PCM_CHANNEL_PLAYBACK;
      if ((ret_code = snd_pcm_channel_status (handle, &status)) < 0)
	g_warning ("alsa: error rechecking channel status: %s",
		   snd_strerror (ret_code));
      if (status.underrun)
	g_warning ("ALSA: write error!!!");
      else
	g_message ("ALSA: underrun, recovered");
    }
  return rv;
}

#endif /* HAVE_ALSA9 */

static void gdam_device_alsa_class_init(GdamDeviceAlsaClass* device_alsa_class)
{
  GdamOutputDeviceClass *klass;
  klass = GDAM_OUTPUT_DEVICE_CLASS (device_alsa_class);
  klass->open_fd = gdam_device_alsa_open_fd;
  klass->close_fd = gdam_device_alsa_close_fd;
  klass->write_pcm = gdam_device_alsa_write_pcm;
}


GtkType gdam_device_alsa_get_type()
{
  static GtkType device_alsa_type = 0;
  if (!device_alsa_type)
    {
      static const GtkTypeInfo device_alsa_info = {
	"GdamDeviceAlsa",
	sizeof (GdamDeviceAlsa),
	sizeof (GdamDeviceAlsaClass),
	(GtkClassInitFunc) gdam_device_alsa_class_init,
	(GtkObjectInitFunc) gdam_device_alsa_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_OUTPUT_DEVICE;
      device_alsa_type = gtk_type_unique (parent, &device_alsa_info);
    }
  return device_alsa_type;
}

/* 
 * This function is mandatory for plugin devices.
 */
GdamSink* gdam_device_alsa_new(gpointer      device_data,
                               const char*   device_name,
			       const char*   device_options)
{
  GtkType type = gdam_device_alsa_get_type ();
  GdamDeviceAlsa *dev_alsa = (GdamDeviceAlsa *) gtk_type_new (type);
  if (!gdam_sink_open (GDAM_SINK (dev_alsa), device_name, device_options))
    {
      gtk_object_unref (GTK_OBJECT (dev_alsa));
      return NULL;
    }
  return GDAM_SINK (dev_alsa);
}

#else /* ! HAVE_ALSA */

#include "gdamdevicealsa.h"

GdamSink* gdam_device_alsa_new(gpointer      device_data,
                               const char*   device_name,
			       const char*   device_options)
{

  g_warning (" Gdam ALSA Driver was not compiled because alsa wasn't\n"
	     "\t\t\tfound on your system when configure was run. \n"
	     "\t\t\tPlease make sure that asoundlib \n"
	     "\t\t\t[in particular sys/asoundlib.h] is installed\n"
	     "\t\t\tdont forget to delete config.cache");
  return NULL;
}

#endif /* ! HAVE_ALSA */
