/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if NOT, WRITe to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamdevicedummy.h"
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>

#define MAX_ATOMIC_SOURCE_READ		2048

static GtkObjectClass *parent_class = NULL;

/* GdamSink virtuals. */
static gboolean gdam_device_dummy_real_open     (GdamSink         *sink,
                                                 const char       *name,
				                 const char       *opts);
static void     gdam_device_dummy_real_close    (GdamSink         *sink);
static guint    gdam_device_dummy_real_get_latency(GdamSink       *sink);
static gboolean gdam_device_dummy_set_source    (GdamSink         *sink,
                                                 GdasSource       *source);
static void     gdam_device_dummy_remove_source (GdamSink         *sink);

/* GtkObject virtual. */
static void     gdam_device_dummy_destroy       (GtkObject        *to_kill);

/* Add/remove the sink from the main loop. */
static void     gdam_device_dummy_stop          (GdamDeviceDummy  *device);
static void     gdam_device_dummy_play          (GdamDeviceDummy  *device);

/* And this function consumes one fragment of audio */
static gboolean gdam_device_dummy_write_one_frag(GdamDeviceDummy* device);


static void 
gdam_device_dummy_options_clear(GdamDeviceDummyOptions *options)
{
  options->sampling_rate = 44100;
  options->num_channels = 2;
  options->bits_per_sample = 16;
  options->block_size = 4096;
  options->log_frag_size = 12;
  options->num_fragments = 4;
}

static void 
gdam_device_dummy_init(GdamDeviceDummy *device_dummy)
{
  gdam_device_dummy_options_clear (&device_dummy->options);
  device_dummy->sample_size = 4;
}

static void 
gdam_device_dummy_class_init 	(GdamDeviceDummyClass *klass)
{
  GdamSinkClass *sink_class;
  sink_class = GDAM_SINK_CLASS (klass);

  sink_class->set_source = gdam_device_dummy_set_source;
  sink_class->remove_source = gdam_device_dummy_remove_source;
  sink_class->open = gdam_device_dummy_real_open;
  sink_class->close = gdam_device_dummy_real_close;
  sink_class->get_latency = gdam_device_dummy_real_get_latency;
  GTK_OBJECT_CLASS (sink_class)->destroy = gdam_device_dummy_destroy;
}

GtkType gdam_device_dummy_get_type()
{
  static GtkType device_dummy_type = 0;
  if (!device_dummy_type)
    {
      static const GtkTypeInfo device_dummy_info = {
	"GdamDeviceDummy",
	sizeof (GdamDeviceDummy),
	sizeof (GdamDeviceDummyClass),
	(GtkClassInitFunc) gdam_device_dummy_class_init,
	(GtkObjectInitFunc) gdam_device_dummy_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SINK;
      parent_class = gtk_type_class (parent);
      device_dummy_type = gtk_type_unique (parent, &device_dummy_info);
    }
  return device_dummy_type;
}

static gboolean 
parse_options		     (const char*              options_string,
                              GdamDeviceDummyOptions*  options)
{
  char **option_strings;
  char **at;
  option_strings = g_strsplit (options_string, ":", 0);
  gdam_device_dummy_options_clear (options);
  for (at = option_strings; *at != NULL; at++)
    {
      char *opt = *at;
      if (strncmp (opt, "logfragsize=", 12) == 0)
	{
	  int log_frag_size = atoi (opt + 12);
	  if (log_frag_size < 6 || log_frag_size > 100)
	    {
	      g_warning ("log_frag_size was out of bounds...");
	    }
	  else
	    {
	      options->log_frag_size = log_frag_size;
	    }
	}
      else if (strcmp (opt, "mono") == 0)
	{
	  options->num_channels = 1;
	}
      else if (strcmp (opt, "stereo") == 0)
	{
	  options->num_channels = 2;
	}
      else if (strncmp (opt, "rate=", 5) == 0)
	{
	  options->sampling_rate = atoi (opt + 5);
	}
      else
	{
	  g_warning ("unknown option to output device: %s", opt);
	}
      g_free (opt);
    }
  g_free (option_strings);
  return TRUE;
}

/*
 * GdamSink virtuals
 */

static void 
gdam_device_dummy_options_changed	(GdamDeviceDummy *device)
{
  GdamDeviceDummyOptions *options = &device->options;
  if (options->bytes_per_sample == 0)
    {
      if (options->bits_per_sample == 24)
	options->bytes_per_sample = 4;
      else
	options->bytes_per_sample = options->bits_per_sample / 8;
    }
  {
    int input_size = sizeof (GdamSample);
    int num_channels = options->num_channels;
    int mid_size = sizeof (GdamSampleValue) * num_channels;
    int output_size = options->bytes_per_sample * num_channels;
    int i;
    i = MAX (input_size, mid_size);
    i = MAX (i, output_size);
    device->conversion_sample_size = i;
  }
  device->total_sample_size = options->bytes_per_sample * options->num_channels;
}


static gboolean 
gdam_device_dummy_real_open     (GdamSink         *sink,
                                 const char       *name,
				 const char       *opts)
{
  GdamDeviceDummy *device_dummy;
  GdamDeviceDummyClass *klass;

  device_dummy = GDAM_DEVICE_DUMMY (sink);
  klass = GDAM_DEVICE_DUMMY_CLASS (GTK_OBJECT (sink)->klass);

  if (name != sink->device_name)
    {
      g_free (sink->device_name);
      sink->device_name = g_strdup (name);
    }

  if (!parse_options (opts, &device_dummy->options))
    {
      return FALSE;
    }

  gdam_debug ("Device set at %d bits/sample, %d channels, rate %d",
	      device_dummy->options.bits_per_sample,
	      device_dummy->options.num_channels,
	      device_dummy->options.sampling_rate);
  gdam_device_dummy_options_changed (device_dummy);
  return TRUE;
}

static void     
gdam_device_dummy_real_close     (GdamSink         *sink)
{
  GdamDeviceDummy *device_dummy;
  device_dummy = GDAM_DEVICE_DUMMY (sink);
  gdam_device_dummy_stop (device_dummy);
}

static guint
gdam_device_dummy_real_get_latency(GdamSink *sink)
{
  (void) sink;
  return 0;
}

static gboolean 
gdam_device_dummy_set_source     (GdamSink         *sink,
                                  GdasSource       *source)
{
  sink->feed = source;
  source->device = sink;
  gtk_object_ref (GTK_OBJECT (source));
  gdam_device_dummy_play (GDAM_DEVICE_DUMMY (sink));
  return TRUE;
}

static void     
gdam_device_dummy_remove_source  (GdamSink         *sink)
{
  gdam_device_dummy_stop (GDAM_DEVICE_DUMMY (sink));
  gtk_object_unref (GTK_OBJECT (sink->feed));
}

static void     
gdam_device_dummy_destroy        (GtkObject        *to_kill)
{
  GdamDeviceDummy *device_dummy;
  GdamSink *sink;
  device_dummy = GDAM_DEVICE_DUMMY (to_kill);
  sink = GDAM_SINK (to_kill);
  gdam_device_dummy_stop (device_dummy);
  gdam_sink_close (sink);
  (*parent_class->destroy) (to_kill);
}

/* Remove the sink from the main loop. */
static void     
gdam_device_dummy_stop         (GdamDeviceDummy *device)
{
  if (device->timeout_sub_id)
    {
      g_source_remove (device->timeout_sub_id);
      device->timeout_sub_id = 0;
    }
}

/* Add the sink to the main loop. */
static void     
gdam_device_dummy_play         (GdamDeviceDummy *device)
{
  int fragment_size = device->options.block_size;
  float fragment_samples = fragment_size / device->total_sample_size;
  float fraction = fragment_samples * 2 / 44100;
  int millis = 1000 * fraction;
  int id;
  id = g_timeout_add (millis, 
		      (GSourceFunc)gdam_device_dummy_write_one_frag, 
		      device);
  device->timeout_sub_id = id;
}


/* XXX: isolate read-from-prebuffer code into an OutputDevice function? */

/* Consume a single fragment of audio data. */
static gboolean
gdam_device_dummy_write_one_frag(GdamDeviceDummy *device_dummy)
{
  GdamDeviceDummyClass *klass;
  GdamSample *fragment_buffer;
  GdamSink *sink = (GdamSink *) device_dummy;
  GdasSource *feed = sink->feed;
  gboolean was_prebuffer = FALSE;
  int num_read, to_read;
  int block_size, frag_buffer_size;
  int old_pre_buffer_start, old_pre_buffer_amount;
  int fragment_size = device_dummy->options.block_size;
  char *tmp_buffer;

  if (feed == NULL)
    return FALSE;
  if (!gdas_source_is_ready_to_read (feed))
    return FALSE;

  to_read = device_dummy->options.block_size * 2;
  if (to_read > MAX_ATOMIC_SOURCE_READ)
    to_read = MAX_ATOMIC_SOURCE_READ;

  tmp_buffer = alloca (device_dummy->conversion_sample_size * to_read);
  num_read = gdas_source_read (feed, (GdamSample *) tmp_buffer, to_read);
  g_return_val_if_fail (num_read <= to_read, FALSE);

  block_size = device_dummy->options.block_size;
  frag_buffer_size = device_dummy->total_sample_size * block_size;
  fragment_buffer = alloca (frag_buffer_size);

  return TRUE;
}

GdamSink *
gdam_device_dummy_new       (gpointer      device_data,
    			     const char   *device_name,
			     const char   *device_options)
{
  GtkType type = gdam_device_dummy_get_type();
  GdamDeviceDummy* dev_dummy = (GdamDeviceDummy*)gtk_type_new(type);
  if (!gdam_sink_open(GDAM_SINK(dev_dummy), device_name, device_options)) {
	  gtk_object_unref(GTK_OBJECT(dev_dummy));
	  return NULL;
  }
  return GDAM_SINK(dev_dummy);
}
