/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 * Unconfuse the autopluginlister: NOT A PLUGIN.
 */

#include "gdamdeviceoss.h"

#if HAVE_OSS

#include <sys/soundcard.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

static void gdam_device_oss_init(GdamDeviceOss* device_oss)
{
}
static gboolean gdam_device_oss_open_fd(GdamOutputDevice* output_device)
{
	int fd;
	const char* device_name;
	int t;
	int r;
	int num_channels;
	GdamOutputDeviceOptions* options = &output_device->options;
	device_name = GDAM_SINK(output_device)->device_name;
	fd = open(device_name, O_WRONLY, 0);
	if (fd < 0) {
		g_warning("couldn't open sound device %s: %s",
			device_name, strerror(errno));
		goto err;
	}
	/*if (!device_options->do_ioctls)
		return fd;*/

	/* Close the device on exec. */
	fcntl(fd, F_SETFD, FD_CLOEXEC);

	if (! options->no_ioctls) {
		t = (options->log_frag_size & 0xffff)
		  | ((options->num_fragments & 0xffff)<<16);
		if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &t) < 0) {
			g_warning("error going to correct amount of frag (%s)",
					strerror(errno));
		} else {
			int total_bytes;
			int sample_size;
			sample_size = options->bits_per_sample / 8 * 
							options->num_channels;
			total_bytes = options->num_fragments 
					<< options->log_frag_size;
			g_message("Set soundcard fragments:  "
				   "%d frags of %d bytes; %3.3fs",
					options->num_fragments,
					(1<<options->log_frag_size),
				(float)total_bytes/44100.0/sample_size);
		}
		t = options->bits_per_sample;
		if (ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &t) < 0) {
			g_warning("error going to %d bits per sample", t);
			goto err;
		}
		t = (options->num_channels == 2) ? 1 : 0;
			
		if (ioctl(fd, SNDCTL_DSP_STEREO, &t) < 0) {
			g_warning("error setting device to stereo: %s",
				  strerror(errno));
			goto err;
		}
#ifdef SNDCTL_DSP_PROFILE
		t = APF_NETWORK;
		if (ioctl(fd, SNDCTL_DSP_PROFILE, &t) < 0) {
			gdam_debug("unable to tell soundcard bufferer "
				"to expect pops from us");
		}
#else
		g_warning("not compiled with DSP_PROFILE support");
#endif
		r = options->sampling_rate;
#if 0
		if (r != 44100) {
			g_warning("only supported output rate is %d, "
				  "you are using %lu,"
				  " but go ahead...", 44100, r);
		}
#endif
		if (ioctl(fd, SNDCTL_DSP_SPEED, &r) < 0) {
			g_warning("error setting dev to %d samples per sec", t);
			goto err;
		}
		if (ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &r) < 0) {
			g_warning("error getting dev block size");
			goto err;
		}
		gdam_debug("DSP Block size is %lu", r);
		/*sound_device_block_size = r;*/

		options->sampling_rate = options->bits_per_sample = -1;
		num_channels = -1;
		if (ioctl(fd, SOUND_PCM_READ_RATE, &options->sampling_rate) < 0)
		  {
		    g_warning ("error querying sapmling rate from device");
		    g_warning("you may need to manually specify the "
			    "soundcard rate if it does not support "
			    "the defaults");
		  }
		if (ioctl(fd, SOUND_PCM_READ_BITS, &options->bits_per_sample)<0)
		    g_warning ("error querying bits/channel from device");
		if (ioctl(fd, SOUND_PCM_READ_CHANNELS, &num_channels) < 0)
		    g_warning ("error reading num-channels from device");
		if (options->sampling_rate == -1)
		  {
		    g_warning ("bad sapmling rate value from device");
		    g_warning("you may need to manually specify the "
			    "soundcard rate if it does not support "
			    "the defaults");
		  }
		if (options->bits_per_sample == -1)
		    g_warning ("bad bit/channel value from device");
		if (num_channels == -1)
		    g_warning ("bad num-channels value from device");

                if (num_channels <= 0)
			num_channels = 2;
		if (options->bits_per_sample <= 0)
			options->bits_per_sample = 16;
		if (options->sampling_rate <= 0)
			options->sampling_rate = 44100;
		options->block_size = (1 << options->log_frag_size)
				    * 8 / options->bits_per_sample;
		options->num_channels = num_channels;
	}
	output_device->device_fd = fd;
	return TRUE;

err:
	if (fd >= 0)
		close(fd);
	return FALSE;
}

static void gdam_device_oss_class_init(GdamDeviceOssClass* device_oss_class)
{
	GdamOutputDeviceClass* klass;
	klass = GDAM_OUTPUT_DEVICE_CLASS(device_oss_class);
	klass->open_fd = gdam_device_oss_open_fd;
}


GtkType gdam_device_oss_get_type()
{
	static GtkType device_oss_type = 0;
	if (!device_oss_type) {
		static const GtkTypeInfo device_oss_info = {
			"GdamDeviceOss",
			sizeof(GdamDeviceOss),
			sizeof(GdamDeviceOssClass),
			(GtkClassInitFunc) gdam_device_oss_class_init,
			(GtkObjectInitFunc) gdam_device_oss_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_OUTPUT_DEVICE;
		device_oss_type = gtk_type_unique(parent,
		                                 &device_oss_info);
	}
	return device_oss_type;
}

/* 
 * This function is mandatory for plugin devices.
 */
GdamSink* gdam_device_oss_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
	GtkType type = gdam_device_oss_get_type();
	GdamDeviceOss* dev_oss = (GdamDeviceOss*)gtk_type_new(type);
	if (!gdam_sink_open(GDAM_SINK(dev_oss), device_name, device_options)) {
		gtk_object_unref(GTK_OBJECT(dev_oss));
		return NULL;
	}
	return GDAM_SINK(dev_oss);
}

#else /* HAVE_OSS */

GdamSink* gdam_device_oss_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
  g_warning (" Gdam OSS Driver was not compiled because oss wasn't\n"
	     "\t\t\tfound on your system when configure was run. \n"
	     "\t\t\tPlease make sure that oss \n"
	     "\t\t\t[in particular sys/soundcard.h] is installed\n"
	     "\t\t\tdont forget to delete config.cache");
  return NULL;
}

#endif /* !HAVE_OSS */
