/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamacl.h"

/*=====================================================================*
 *                   gdam_acl_rule_deny_user(int uid)                  *
 *                                                                     *
 * Create a new rule, which denies a particular user-id.               *
 *                                                                     *
 * Parameters:                                                         *
 *     uid               the user-id to disallow.                      *
 *=====================================================================*/
static int gdam_acl_user_deny(GdamPermissions* permissions,
	                      gpointer         acl_data)
{
  int acl_user_id = GPOINTER_TO_INT (acl_data);
  int user_id = permissions->user_id;
  if (user_id == -1 || acl_user_id != user_id)
    return 0;
  return -1;
}

GdamAclRule* gdam_acl_rule_deny_user(int uid)
{
  GdamAclRule *new_rule = g_new (GdamAclRule, 1);
  new_rule->check_permissions = gdam_acl_user_deny;
  new_rule->destroy_acl_data = NULL;
  new_rule->acl_data = GINT_TO_POINTER (uid);
  return new_rule;
}

/*=====================================================================*
 *                  gdam_acl_rule_grant_user(int uid)                  *
 *                                                                     *
 * Create a new rule, which allows a particular user-id.               *
 *                                                                     *
 * Parameters:                                                         *
 *     uid               the user-id to allow.                         *
 *=====================================================================*/
static int gdam_acl_user_grant(GdamPermissions* permissions,
	                       gpointer         acl_data)
{
  int acl_user_id = GPOINTER_TO_INT (acl_data);
  int user_id = permissions->user_id;
  if (user_id == -1 || acl_user_id != user_id)
    return 0;
  return 1;
}
GdamAclRule* gdam_acl_rule_grant_user(int uid)
{
  GdamAclRule *new_rule = g_new (GdamAclRule, 1);
  new_rule->check_permissions = gdam_acl_user_grant;
  new_rule->destroy_acl_data = NULL;
  new_rule->acl_data = GINT_TO_POINTER (uid);
  return new_rule;
}

/*=====================================================================*
 *                             gdam_acl_new                            *
 *                                                                     *
 * Create a new access control list (ACL) which                        *
 * the default permission (grant/deny) as given.                       *
 *                                                                     *
 * Parameters:                                                         *
 *     default_grant     whether to grant a connections with no        *
 *                       matching rules.                               *
 *=====================================================================*/
GdamAcl* gdam_acl_new(gboolean default_grant)
{
  GdamAcl *acl = g_new (GdamAcl, 1);
  acl->first_rule = acl->last_rule = NULL;
  acl->default_grant = default_grant ? 1 : 0;
  return acl;
}

/*=====================================================================*
 *                         gdam_acl_is_granted                         *
 *                                                                     *
 * Parameters:                                                         *
 *     acl               the access-control-list to check against.     *
 *     user_perms        the permissions to consider.                  *
 *=====================================================================*/
gboolean gdam_acl_is_granted(GdamAcl* acl, GdamPermissions* user_perms)
{
  GSList *at;
  for (at = acl->first_rule; at; at = at->next)
    {
      GdamAclRule *rule = (GdamAclRule *) (at->data);
      switch (rule->check_permissions (rule->acl_data, user_perms))
	{
	case -1:
	  return FALSE;
	case +1:
	  return TRUE;
	case 0:
	  break;

	  /* Triggering the next assertion means
	   * a `check_permissions' function has gone amuck. */
	default:
	  g_assert_not_reached ();
	}
    }
  return acl->default_grant;
}


/*=====================================================================*
 *                        gdam_acl_prepend_rule                        *
 *                                                                     *
 * Adds a rule to the acl so that it is the first rule checked.        *
 *                                                                     *
 * Parameters:                                                         *
 *     acl               the access-control-list to add a rule to.     *
 *     acl_rule          the rule to prepend.                          *
 *=====================================================================*/
void gdam_acl_prepend_rule(GdamAcl* acl, GdamAclRule* acl_rule)
{
  acl->first_rule = g_slist_prepend (acl->first_rule, acl_rule);
  if (!acl->last_rule)
    acl->last_rule = acl->first_rule;
}

/*=====================================================================*
 *                        gdam_acl_prepend_rule                        *
 *                                                                     *
 * Adds a rule to the acl so that it is the first rule checked.        *
 *                                                                     *
 * Parameters:                                                         *
 *     acl               the access-control-list to add a rule to.     *
 *     acl_rule          the rule to prepend.                          *
 *=====================================================================*/
void gdam_acl_append_rule(GdamAcl* acl, GdamAclRule* acl_rule)
{
  GSList *new_tail = g_slist_append (NULL, acl_rule);
  if (acl->first_rule)
    {
      acl->last_rule->next = new_tail;
      acl->last_rule = new_tail;
    }
  else
    {
      acl->first_rule = acl->last_rule = new_tail;
    }
}

/*=====================================================================*
 *                           gdam_acl_destroy                          *
 *                                                                     *
 * Parameters:                                                         *
 *     to_kill           the access-control-list to delete.            *
 *=====================================================================*/
void gdam_acl_destroy(GdamAcl* to_kill)
{
  GSList *at = to_kill->first_rule;
  while (at)
    {
      GdamAclRule *rule = (GdamAclRule *) (at->data);
      if (rule->destroy_acl_data)
	(*rule->destroy_acl_data) (rule->acl_data);
      g_free (rule);
      at = at->next;
    }
  g_slist_free (to_kill->first_rule);
}
