#include "gdamcachebuffer.h"

/* --- beat detection --- */
static int zero_threshold = (int)(0.005 * GDAM_SAMPLE_ONE);
static inline gboolean 
is_zero (int sample)
{
  if (sample < zero_threshold && sample > -1 * zero_threshold)
    return TRUE;
  return FALSE;
}

gdouble *
gdam_cache_buffer_guess_beat_times (GdamCacheBuffer *buffer,
				    gdouble window_size,
				    int amplitude_threshold,
				    gdouble amplitude_ratio,
				    gdouble min_sample_length,
				    int *num_beat_times)
{
  gdouble *rv;
  GList *beat_times = NULL;
  gboolean are_over = FALSE;
  gboolean descending = TRUE;
  gboolean positive = TRUE;
  gdouble last_time_over_threshold = 0;
  gdouble length = (gdouble)(buffer->buffer_length) / 44100.0;
  /* beginning of our current window, in seconds */
  gdouble last_window_break = 0;
  gdouble last_beat_time = -100;
  /* track peak values within each window */
  int current_max_peak = 0;
  int last_max_peak = 0;
  /* amplitude which triggers a beat */
  int i;
  guint k;
  int num_samples = buffer->buffer_length;
  int audible_threshold = (int)(0.02 * GDAM_SAMPLE_ONE);

  /* Walk through the buffer from the beginning */
  for (i=0; i < num_samples; i++)
    {
      gdouble time = length * (gdouble)i / (gdouble)num_samples;
      gboolean zero_crossing = FALSE;
      gboolean is_beat = FALSE;
      int peak = 0;
      GdamSample sa1 = gdam_cache_buffer_index (buffer, i);
      GdamSample sa2;
      if (i > 0)
	sa2 = gdam_cache_buffer_index (buffer, i-1);

      /* Track the direction of the waveform, and peak values.  a peak is
       * the point within a region bounded by zero crossings which is 
       * furthest from zero amplitude.
       */
      if (i > 0 && sa1.s[0] < sa2.s[0])
	{
	  if (!descending)
	    peak = abs (sa2.s[0]);
	  descending = TRUE;
	}
      else if (i > 0)
	{
	  if (descending)
	    peak = abs (sa2.s[0]);
	  descending = FALSE;
	}
      if (sa1.s[0] > 0)
	{
	  if (!positive)
	    zero_crossing = TRUE;
	  positive = TRUE;
	}
      else
	{
	  if (positive)
	    zero_crossing = TRUE;
	  positive = FALSE;
	}

      /* If we have reached a new window, cycle variables */
      if (time - last_window_break > window_size)
	{
	  last_max_peak = current_max_peak;
	  current_max_peak = 0;
	  last_window_break = time;
	}

      if (peak > current_max_peak)
	current_max_peak = peak;

      /* If this sample is over the threshhold */
      if (sa1.s[0] > amplitude_threshold 
	  || sa1.s[0] < -1 * amplitude_threshold)
	{
	  /* If we were in a low-amplitude state, then this is a beat */
	  if (!are_over)
	    {
	      if (time - last_beat_time  > min_sample_length)
		{
		  gdam_debug ("is_beat by threshold at %f", time);
		  is_beat = TRUE;
		}
	      are_over = TRUE;
	    }
	  last_time_over_threshold = time;
	}
      /* If this sample is under */
      else
	{
	  gdouble delta_time = time - last_time_over_threshold;
	  /* If we were in a high-amplitude state...  */
	  if (are_over && delta_time > window_size)
	    are_over = FALSE;
	}

      /* A sharp increase in peak amplitude between windows triggers a beat */
      if ((sa1.s[0] > audible_threshold || sa1.s[0] < -1 * audible_threshold) 
	  && peak > amplitude_ratio * last_max_peak 
	  && time - last_beat_time > min_sample_length)
	{
	  gdam_debug ("is_beat by max_peak at %f", time);
	  is_beat = TRUE;
	}

      /* If a beat was triggered... */
      if (is_beat)
	{
	  GdamSample sac;
	  gdouble beat_time, amp_thold;
	  gboolean last_was_positive = FALSE;
	  int closest_sample, j;

	  amp_thold = (gdouble)amplitude_threshold / GDAM_SAMPLE_ONE;

	  /* Jump back to estimated beginning of attack curve */
	  beat_time = time - (amp_thold / 50);
	  closest_sample = (int)(beat_time * (gdouble)num_samples / length);

	  /* Seek backwards to zero crossing */
	  j = closest_sample;
	  sac = gdam_cache_buffer_index (buffer, j);
	  if (sac.s[0] > 0)
	    last_was_positive = TRUE;
	  while (!is_zero (sac.s[0]) && j > 0)
	    {
	      gboolean is_positive = FALSE;
	      if (sac.s[0] > 0)
		is_positive = TRUE;
	      if (is_positive && !last_was_positive)
		gdam_debug ("missed zero crossing at sample %d", j);
	      j--;
	      if (j >= 0)
		sac = gdam_cache_buffer_index (buffer, j);
	    }

	  while (is_zero (sac.s[0]))
	    {
	      j++;
	      sac = gdam_cache_buffer_index (buffer, j);
	    }

	  /* Adopt the location of the last zero crossing. */
	  beat_time = length * (gdouble)j / (gdouble)num_samples;
	  if (beat_time < 0)
	    {
	      gdam_debug ("repairing negative beat time %f", beat_time);
	      beat_time = 0;
	    }

	  /* are_over ??? */
	  are_over = TRUE;
	  last_beat_time = beat_time;
	    {
	      gdouble *newdouble = g_new (gdouble, 1);
	      *newdouble = beat_time;
	      beat_times = g_list_append (beat_times, newdouble);
	    }
	}
    }
  rv = g_new (gdouble, g_list_length (beat_times));
  for (k = 0; k < g_list_length (beat_times); k++)
    {
      gdouble *btime = (gdouble *)(g_list_nth_data (beat_times, k));
      rv[k] = *btime;
      g_free (btime);
    }
  *num_beat_times = g_list_length (beat_times);
  g_list_free (beat_times);
  return rv;
}

