/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamcachebuffer.h"
#include "gdamfft.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamautoconf.h>
#include <stdlib.h>

	
/*
 * For displaying a GdamCacheBuffer on the client side.
 */

/* compute NUM-CHANNELS-OUT ints of output in OUT
 * in an index-style-specific manner:
 * left, right, stereo, average, fft, min-max are the current examples.
 *
 * The data is to be taken from a window centered around
 * INDEX in BUFFER.  RANGE gives the spacing between adjacent samples.
 */
typedef void (*ProfileSampleFunc)(int             *out,
			          guint            num_channels_out,
			          GdamCacheBuffer *buffer,
			          guint            index,
			          guint            range);


static void
profile_left_sample (int             *out,
		     guint            num_channels_out,
		     GdamCacheBuffer *buffer,
		     guint            index,
		     guint            range)
{
  (void) num_channels_out;
  (void) range;
  out[0] = gdam_cache_buffer_index (buffer, index).s[0];
}

static void
profile_right_sample (int             *out,
		      guint            num_channels_out,
		      GdamCacheBuffer *buffer,
		      guint            index,
		      guint            range)
{
  (void) num_channels_out;
  (void) range;
  out[1] = gdam_cache_buffer_index (buffer, index).s[1];
}

static void
profile_average_sample (int             *out,
		        guint            num_channels_out,
		        GdamCacheBuffer *buffer,
		        guint            index,
		        guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out = ((int) s.s[0] + s.s[1]) / 2;
}

static void
profile_stereo_sample (int             *out,
		       guint            num_channels_out,
		       GdamCacheBuffer *buffer,
		       guint            index,
		       guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out++ = s.s[0];
  *out++ = s.s[1];
}

static void
profile_fft_sample (int             *out,
		    guint            num_channels_out,
		    GdamCacheBuffer *buffer,
		    guint            index,
		    guint            range)
{
  gfloat *samps = alloca (sizeof (gfloat) * num_channels_out * 2);
  guint i;
  gint cur_index = index - num_channels_out;
  const gfloat *window = gdam_fft_get_hamming_window_f (num_channels_out * 2);
  (void) range;
  for (i = 0; i < num_channels_out * 2; i++)
    {
      GdamSample t = gdam_cache_buffer_index (buffer, cur_index);
      int s = t.s[0] + t.s[1];
      cur_index++;
      samps[i] = (gfloat) s * window[i] / (GDAM_SAMPLE_ONE * 2);
    }
  gdam_real_fft_f (samps, num_channels_out, TRUE);
  /* TODO: doesn't handle samps[1] correctly (see gdamfft.h) */
  for (i = 0; i < num_channels_out; i++)
    {
      gfloat p = samps[2 * i + 0] * samps[2 * i + 0]
	       + samps[2 * i + 1] * samps[2 * i + 1];
      p = CLAMP (p, -2.0, +2.0);
      out[i] = (gint) (p * GDAM_SAMPLE_ONE);
    }
}

#define MAX_MINMAX_TEST		16
static void
profile_minmax_sample (int             *out,
		       guint            num_channels_out,
		       GdamCacheBuffer *buffer,
		       guint            index,
		       guint            range)
{
  gint minimum_value, maximum_value;

  g_assert (num_channels_out == 2);

#define INIT_MIN_MAX(index)					\
  G_STMT_START{							\
    GdamSample s;						\
    s = gdam_cache_buffer_index (buffer, (index));		\
    if (s.s[0] < s.s[1])					\
      {								\
	minimum_value = s.s[0];					\
	maximum_value = s.s[1];					\
      }								\
    else							\
      {								\
	minimum_value = s.s[1];					\
	maximum_value = s.s[0];					\
      }								\
  }G_STMT_END
#define UPDATE_MIN_MAX(index)					\
  G_STMT_START{							\
    GdamSample s;						\
    s = gdam_cache_buffer_index (buffer, (index));		\
    if (s.s[0] < s.s[1])					\
      {								\
	if (minimum_value > s.s[0])				\
	  minimum_value = s.s[0];				\
	if (maximum_value > s.s[1])				\
	  maximum_value = s.s[1];				\
      }								\
    else							\
      {								\
	if (minimum_value > s.s[1])				\
	  minimum_value = s.s[1];				\
	if (maximum_value > s.s[0])				\
	  maximum_value = s.s[0];				\
      }								\
  }G_STMT_END
	
  if (range > MAX_MINMAX_TEST)
    {
      /* just take MAX_MINMAX_TEST samples for speed */
      gint i;
      gint at = index - range / 2;
      INIT_MIN_MAX (at);
      for (i = -MAX_MINMAX_TEST / 2 + 1; i <= MAX_MINMAX_TEST / 2; i++)
	{
	  /* XXX: use bresenheim's */
	  at = index + (gint64) i * range / MAX_MINMAX_TEST;
	  UPDATE_MIN_MAX (at);
	}
    }
  else
    {
      /* scan (index - range/2 .. index + range/2) */
      gint at = index - range / 2;
      gint end = index + range / 2;
      INIT_MIN_MAX (at);
      at++;
      while (at <= end)
	{
	  UPDATE_MIN_MAX (at);
	  at++;
	}
    }
  out[0] = minimum_value;
  out[1] = maximum_value;

#undef INIT_MIN_MAX
#undef UPDATE_MIN_MAX
}

static inline ProfileSampleFunc
get_profile_info_for_mode (int mode,
			   gint *num_channels_out)
{
  switch (mode)
    {
    case GDAM_PROTOCOL_PROFILE_LEFT:
      *num_channels_out = 1;
      return profile_left_sample;
    case GDAM_PROTOCOL_PROFILE_RIGHT:
      *num_channels_out = 1;
      return profile_right_sample;
    case GDAM_PROTOCOL_PROFILE_STEREO:
      *num_channels_out = 2;
      return profile_stereo_sample;
    case GDAM_PROTOCOL_PROFILE_AVERAGE:
      *num_channels_out = 1;
      return profile_average_sample;
    case GDAM_PROTOCOL_PROFILE_MIN_MAX:
      *num_channels_out = 1;
      return profile_minmax_sample;
    default:
      {
	if (GDAM_PROTOCOL_PROFILE_FFT <= mode
	 && mode <= GDAM_PROTOCOL_PROFILE_FFT_LAST)
	  {
	    int log_fft_size = mode - GDAM_PROTOCOL_PROFILE_FFT;
	    if (log_fft_size < 3)
	      log_fft_size = 3;
	    *num_channels_out = 1 << log_fft_size;
	    return profile_fft_sample;
	  }
	else
	  {
	    g_warning ("cache_profile mode not known: %d", mode);
	    return NULL;
	  }
      }
    }
}

int
gdam_cache_buffer_profile_mode_get_n_channels (int profile_mode)
{
  gint rv = -1;
  (void) get_profile_info_for_mode (profile_mode, &rv);
  return rv;
}

gint *
gdam_cache_buffer_profile (GdamCacheBuffer  *buffer,
			   int               mode,
			   int               sample_start,
			   int               sample_length,
			   int               num_outputs)
{
  int prepad_zeroes = 0, postpad_zeroes = 0;
  int orig_num_outputs = num_outputs;
  int channels = 0;
  ProfileSampleFunc profile_func = NULL;
  gint *outputs;
  int *at;
  int i;
  int num_ints;
  gint range;

  profile_func = get_profile_info_for_mode (mode, &channels);
  g_return_val_if_fail (channels > 0, NULL);

  /* Does it overrun the sample's start? */
  if (sample_start < 0)
    {
      prepad_zeroes = (-sample_start) * num_outputs / sample_length;
      sample_length += sample_start;
      sample_start = 0;
      num_outputs -= prepad_zeroes;
    }

  /* Allocate the output area... */
  num_ints = channels * orig_num_outputs;
  outputs = g_new (gint, num_ints);

  if (sample_length <= 0 || prepad_zeroes < 0)
    {
      bzero (outputs, sizeof (gint) * num_ints);
      return outputs;
    }

  /* Does it overrun the sample's end? */
  if (sample_start + sample_length > buffer->buffer_length)
    {
      int cut = sample_start + sample_length - buffer->buffer_length;
      postpad_zeroes = (gint64) cut *num_outputs / sample_length;
      num_outputs -= postpad_zeroes;
      sample_length -= cut;
    }

  /* If they are hopelessly out of range, fill the buffer
   * with silence and return. */
  if (num_outputs <= 0)
    {
      bzero (outputs, sizeof (gint) * num_ints);
      return outputs;
    }

  /* Zero any padding we had to insert. */
  bzero (outputs, channels * prepad_zeroes * sizeof (gint));

  at = outputs + channels * prepad_zeroes;
  range = sample_length / num_outputs;
  if (range < 0)
    range = -range;
  for (i = 0; i < num_outputs; i++)
    {
      gint64 index = i;
      index *= sample_length;
      index /= num_outputs;
      index += sample_start;
      (*profile_func) (at, channels, buffer, index, range);
      at += channels;
    }
  /* Zero any padding we had to insert. */
  bzero (at, channels * postpad_zeroes * sizeof (gint));
  return outputs;
}
