/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamcachebuffer.h"
#include "gdamfft.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamautoconf.h>
#include <stdlib.h>

	
/*
 * For displaying a GdamCacheBuffer on the client side.
 */

/* compute NUM-CHANNELS-OUT ints of output in OUT
 * in an index-style-specific manner:
 * left, right, stereo, average, fft, min-max are the current examples.
 *
 * The data is to be taken from a window centered around
 * INDEX in BUFFER.  RANGE gives the spacing between adjacent samples.
 */
typedef void (*ProfileSampleFuncDouble)(gdouble         *out,
					guint            num_channels_out,
			                GdamCacheBuffer *buffer,
			                guint            index,
			                guint            range);
typedef void (*ProfileSampleFuncInt)   (gint            *out,
					guint            num_channels_out,
			                GdamCacheBuffer *buffer,
			                guint            index,
			                guint            range);
typedef void (*ProfileSampleFunc)      (gpointer         out,
					guint            num_channels_out,
			                GdamCacheBuffer *buffer,
			                guint            index,
			                guint            range);
typedef struct _ProfileSampleFuncs ProfileSampleFuncs;
struct _ProfileSampleFuncs
{
  ProfileSampleFuncDouble double_func;
  ProfileSampleFuncInt int_func;
};


/* --- left --- */
static void
profile_left_sample_double (gdouble         *out,
			    guint            num_channels_out,
			    GdamCacheBuffer *buffer,
			    guint            index,
			    guint            range)
{
  GdamSampleValue value = gdam_cache_buffer_index (buffer, index).s[0];
  out[0] = GDAM_SAMPLE_TO_FLOAT (value);
}

static void
profile_left_sample_int    (gint            *out,
			    guint            num_channels_out,
			    GdamCacheBuffer *buffer,
			    guint            index,
			    guint            range)
{
  GdamSampleValue value = gdam_cache_buffer_index (buffer, index).s[0];
  out[0] = value;
}

static ProfileSampleFuncs profile_left_sample_funcs =
{
  profile_left_sample_double,
  profile_left_sample_int
};

/* --- right --- */
static void
profile_right_sample_double (gdouble         *out,
			    guint            num_channels_out,
			    GdamCacheBuffer *buffer,
			    guint            index,
			    guint            range)
{
  GdamSampleValue value = gdam_cache_buffer_index (buffer, index).s[1];
  out[0] = GDAM_SAMPLE_TO_FLOAT (value);
}

static void
profile_right_sample_int   (gint            *out,
			    guint            num_channels_out,
			    GdamCacheBuffer *buffer,
			    guint            index,
			    guint            range)
{
  GdamSampleValue value = gdam_cache_buffer_index (buffer, index).s[1];
  out[0] = value;
}

static ProfileSampleFuncs profile_right_sample_funcs =
{
  profile_right_sample_double,
  profile_right_sample_int
};

/* --- average (aka mono) --- */
static void
profile_average_sample_int (int             *out,
		            guint            num_channels_out,
		            GdamCacheBuffer *buffer,
		            guint            index,
		            guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out = ((int) s.s[0] + s.s[1]) / 2;
}

static void
profile_average_sample_double (gdouble         *out,
		               guint            num_channels_out,
		               GdamCacheBuffer *buffer,
		               guint            index,
		               guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out = GDAM_SAMPLE_TO_FLOAT(s.s[0] + s.s[1]) * 0.5;
}

static ProfileSampleFuncs profile_average_sample_funcs =
{
  profile_average_sample_double,
  profile_average_sample_int
};

/* --- stereo --- */
static void
profile_stereo_sample_int (int             *out,
		           guint            num_channels_out,
		           GdamCacheBuffer *buffer,
		           guint            index,
		           guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out++ = s.s[0];
  *out++ = s.s[1];
}

static void
profile_stereo_sample_double (gdouble         *out,
		              guint            num_channels_out,
		              GdamCacheBuffer *buffer,
		              guint            index,
		              guint            range)
{
  GdamSample s = gdam_cache_buffer_index (buffer, index);
  (void) num_channels_out;
  (void) range;
  *out++ = GDAM_SAMPLE_TO_FLOAT (s.s[0]);
  *out++ = GDAM_SAMPLE_TO_FLOAT (s.s[1]);
}

static ProfileSampleFuncs profile_stereo_sample_funcs =
{
  profile_stereo_sample_double,
  profile_stereo_sample_int
};


/* --- fft --- */
static void
profile_fft_sample_int (int             *out,
		        guint            num_channels_out,
		        GdamCacheBuffer *buffer,
		        guint            index,
		        guint            range)
{
  gfloat *samps = alloca (sizeof (gfloat) * num_channels_out * 2);
  guint i;
  gint cur_index = index - num_channels_out;
  const gfloat *window = gdam_fft_get_hamming_window_f (num_channels_out * 2);
  (void) range;
  for (i = 0; i < num_channels_out * 2; i++)
    {
      GdamSample t = gdam_cache_buffer_index (buffer, cur_index);
      int s = t.s[0] + t.s[1];
      cur_index++;
      samps[i] = (gfloat) s * window[i] / (GDAM_SAMPLE_ONE * 2);
    }
  gdam_real_fft_f (samps, num_channels_out, TRUE);
  /* TODO: doesn't handle samps[1] correctly (see gdamfft.h) */
  for (i = 0; i < num_channels_out; i++)
    {
      gfloat p = samps[2 * i + 0] * samps[2 * i + 0]
	       + samps[2 * i + 1] * samps[2 * i + 1];
      if (p > 2.0)
	p = 2.0;
      out[i] = (gint) (p * GDAM_SAMPLE_ONE);
    }
}

static void
profile_fft_sample_double (gdouble         *out,
		           guint            num_channels_out,
		           GdamCacheBuffer *buffer,
		           guint            index,
		           guint            range)
{
  gdouble *samps = alloca (sizeof (gdouble) * num_channels_out * 2);
  guint i;
  gint cur_index = index - num_channels_out;
  const gfloat *window = gdam_fft_get_hamming_window_f (num_channels_out * 2);
  (void) range;
  for (i = 0; i < num_channels_out * 2; i++)
    {
      GdamSample t = gdam_cache_buffer_index (buffer, cur_index);
      int s = t.s[0] + t.s[1];
      cur_index++;
      samps[i] = (gfloat) s * window[i] / (GDAM_SAMPLE_ONE * 2);
    }
  gdam_real_fft (samps, num_channels_out, TRUE);
  /* TODO: doesn't handle samps[1] correctly (see gdamfft.h) */
  for (i = 0; i < num_channels_out; i++)
    {
      out[i] = samps[2 * i + 0] * samps[2 * i + 0]
	     + samps[2 * i + 1] * samps[2 * i + 1];
    }
}

static ProfileSampleFuncs profile_fft_sample_funcs = 
{
  profile_fft_sample_double,
  profile_fft_sample_int
};


/* --- min/max pairs --- */
#define MAX_MINMAX_TEST		128
static void
profile_minmax_sample_int (int             *out,
		           guint            num_channels_out,
		           GdamCacheBuffer *buffer,
		           guint            index,
		           guint            range)
{
  gint minimum_value, maximum_value;

  g_assert (num_channels_out == 2);

#define INIT_MIN_MAX(index)					\
  G_STMT_START{							\
    GdamSample s;						\
    s = gdam_cache_buffer_index (buffer, (index));		\
    if (s.s[0] < s.s[1])					\
      {								\
	minimum_value = s.s[0];					\
	maximum_value = s.s[1];					\
      }								\
    else							\
      {								\
	minimum_value = s.s[1];					\
	maximum_value = s.s[0];					\
      }								\
  }G_STMT_END
#define UPDATE_MIN_MAX(index)					\
  G_STMT_START{							\
    GdamSample s;						\
    s = gdam_cache_buffer_index (buffer, (index));		\
    if (s.s[0] < s.s[1])					\
      {								\
	if (minimum_value > s.s[0])				\
	  minimum_value = s.s[0];				\
	if (maximum_value > s.s[1])				\
	  maximum_value = s.s[1];				\
      }								\
    else							\
      {								\
	if (minimum_value > s.s[1])				\
	  minimum_value = s.s[1];				\
	if (maximum_value > s.s[0])				\
	  maximum_value = s.s[0];				\
      }								\
  }G_STMT_END
	
  if (range > MAX_MINMAX_TEST)
    {
      /* just take MAX_MINMAX_TEST samples for speed */
      gint i;
      gint at = index - range / 2;
      INIT_MIN_MAX (at);
      for (i = -MAX_MINMAX_TEST / 2 + 1; i <= MAX_MINMAX_TEST / 2; i++)
	{
	  /* XXX: use bresenheim's */
	  at = index + (gint64) i * range / MAX_MINMAX_TEST;
	  UPDATE_MIN_MAX (at);
	}
    }
  else
    {
      /* scan (index - range/2 .. index + range/2) */
      gint at = index - range / 2;
      gint end = index + range / 2;
      INIT_MIN_MAX (at);
      at++;
      while (at <= end)
	{
	  UPDATE_MIN_MAX (at);
	  at++;
	}
    }
  out[0] = minimum_value;
  out[1] = maximum_value;

#undef INIT_MIN_MAX
#undef UPDATE_MIN_MAX
}

static void
profile_minmax_sample_double (gdouble         *out,
		              guint            num_channels_out,
		              GdamCacheBuffer *buffer,
		              guint            index,
		              guint            range)
{
  int i[2];
  profile_minmax_sample_int (i, num_channels_out, buffer, index, range);
  out[0] = GDAM_SAMPLE_TO_FLOAT (i[0]);
  out[1] = GDAM_SAMPLE_TO_FLOAT (i[1]);
}

static ProfileSampleFuncs profile_minmax_sample_funcs =
{
  profile_minmax_sample_double,
  profile_minmax_sample_int
};

#define LEFT_BRACE_CHAR		'{'
#define RIGHT_BRACE_CHAR	'}'

/* --- string => profile sampling info --- */
static inline ProfileSampleFuncs *
get_profile_info_for_mode (const char *name,
			   gint *num_channels_out)
{
  switch (name[0])
    {
    case 'f':
      if (name[1] == 'f' && name[2] == 't')
	{
	  int log_level = -1;
	  if (name[3] == LEFT_BRACE_CHAR)
	    log_level = atoi (name + 4);
	  if (log_level < 0)
	    log_level = 3;
	  *num_channels_out = 1 << log_level;
	  return &profile_fft_sample_funcs;
	}
      break;
    case 'l':
      if (strcmp (name, "left") == 0)
	{
	  *num_channels_out = 1;
	  return &profile_left_sample_funcs;
	}
      break;
    case 'r':
      if (strcmp (name, "right") == 0)
	{
	  *num_channels_out = 1;
	  return &profile_right_sample_funcs;
	}
      break;
    case 's':
      if (strcmp (name, "stereo") == 0)
	{
	  *num_channels_out = 2;
	  return &profile_stereo_sample_funcs;
	}
      break;
    case 'm':
      if (strcmp (name, "mono") == 0)
	{
	  *num_channels_out = 1;
	  return &profile_average_sample_funcs;
	}
      else if (strcmp (name, "minmax") == 0)
	{
	  *num_channels_out = 2;
	  return &profile_minmax_sample_funcs;
	}
      break;
    case 'a':
      if (strcmp (name, "average") == 0)
	{
	  *num_channels_out = 1;
	  return &profile_average_sample_funcs;
	}
      break;
    }
  return NULL;
}

/* INPUT:
 *    mode         - string - one of left, right, stereo, mono, minmax, fft%d
 *    decimation   - int    - for fft, the decimation factor
 *    use_doubles  - int    - boolean: return doubles or integers
 *    start        - int64
 *    length       - int64
 *    num_outputs  - int
 * OUTPUT:
 *    start        - int64  - initial sample
 *    length       - int64  - num samples covered
 *    num_outputs  - int    - num sample points in values
 *    num_channels - int    - values per sample
 *    values       - int* or double* - the values
 */
/* NOTE: the start/length/num_outputs may be clipped before they are returned
 *       to you.
 */
void
gdam_cache_buffer_func_profile(GtkObject      *object,
			       int             num_args_in,
			       GdamArg        *arguments_in,
			       int            *num_args_out,
			       GdamArg       **arguments_out)
{
  GdamCacheBuffer *cache_buffer = GDAM_CACHE_BUFFER (object);
  ProfileSampleFuncs *profile_funcs;
  ProfileSampleFunc profile_func = NULL;
  int channels = 0;
  gpointer outputs;
  gint64 start = 0;
  gint64 length = -1;
  gint num_outputs = 16;
  const char *mode = "mono";
  int i;
  int decimation = 1;
  gboolean use_doubles = FALSE;
  int arg_type;
  guint bytes_per_output;
  char *at;
  gint range;
  GdamArg *args;
  int valid_output_start, valid_output_end;

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arguments_in + i;
      const char *name = arg->name;
      if (strcmp (name, "mode") == 0 && arg->type == GDAM_ARG_TYPE_STRING)
	mode = GDAM_ARG_STRING (arg);
      else if (strcmp (name, "decimation") == 0
	    && arg->type == GDAM_ARG_TYPE_INT)
	decimation = GDAM_ARG_INT (arg);
      else if (strcmp (name, "use_doubles") == 0
	    && arg->type == GDAM_ARG_TYPE_INT)
	use_doubles = GDAM_ARG_INT (arg);
      else if (strcmp (name, "start") == 0
	    && arg->type == GDAM_ARG_TYPE_INT64)
	start = GDAM_ARG_INT64 (arg);
      else if (strcmp (name, "length") == 0
	    && arg->type == GDAM_ARG_TYPE_INT64)
	length = GDAM_ARG_INT64 (arg);
      else if (strcmp (name, "num_outputs") == 0
	    && arg->type == GDAM_ARG_TYPE_INT)
	num_outputs = GDAM_ARG_INT (arg);
      else
	{
	  g_warning ("unknown parameter/type %s", arg->name);
	  continue;
	}
    }
  profile_funcs = get_profile_info_for_mode (mode, &channels);
  if (channels == 0 || profile_funcs == NULL)
    {
      char buf[256];
      g_snprintf (buf, sizeof (buf), "invalid mode '%s'", mode);
      *arguments_out = gdam_arg_new_from_valist (1,
						 "error",
						 GDAM_ARG_TYPE_STRING,
						 buf);
      *num_args_out = 1;
      return;
    }
  profile_func = use_doubles ? (ProfileSampleFunc) profile_funcs->double_func
                             : (ProfileSampleFunc) profile_funcs->int_func;

  /* Allocate the output area... */
  if (use_doubles)
    {
      arg_type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
      bytes_per_output = channels * sizeof (gdouble);
    }
  else
    {
      arg_type = GDAM_ARG_TYPE_INT_ARRAY;
      bytes_per_output = channels * sizeof (gint);
    }
  outputs = g_malloc (bytes_per_output * num_outputs);

  /* Do the profiling */
  at = outputs;
  range = length / num_outputs;
  if (range < 0)
    range = -range;
  for (i = 0; i < num_outputs; i++)
    {
      gint64 index = i;
      index *= length;
      index /= num_outputs;
      index += start;
      (*profile_func) (at, channels, cache_buffer, index, range);
      at += bytes_per_output;
    }

  if (start < 0)
    {
      valid_output_start = -start * num_outputs / length;
      if (valid_output_start > num_outputs)
	valid_output_start = num_outputs;
    }
  else
    valid_output_start = 0;
  if (start + length > cache_buffer->buffer_length)
    {
      valid_output_end = (cache_buffer->buffer_length - start) * num_outputs / length;
      if (valid_output_end < valid_output_start)
	valid_output_end = valid_output_start;
    }
  else
    valid_output_end = num_outputs;

  /* Set up the return arguments. */
  *num_args_out = 7;
  args = gdam_arg_new_from_valist (*num_args_out,
	   "start", GDAM_ARG_TYPE_INT64, (gint64) start,
	   "length", GDAM_ARG_TYPE_INT64, (gint64) length,
	   "num_outputs", GDAM_ARG_TYPE_INT, (int) num_outputs,
	   "num_channels", GDAM_ARG_TYPE_INT, (int) channels,
	   "valid_output_start", GDAM_ARG_TYPE_INT, (int) valid_output_start,
	   "valid_output_end", GDAM_ARG_TYPE_INT, (int) valid_output_end,
	   "values", arg_type, 0, NULL);
  *arguments_out = args;
  if (use_doubles)
    {
      GDAM_ARG_DOUBLE_ARRAY (args + 6).length = channels * num_outputs;
      GDAM_ARG_DOUBLE_ARRAY (args + 6).array = outputs;
    }
  else
    {
      GDAM_ARG_INT_ARRAY (args + 6).length = channels * num_outputs;
      GDAM_ARG_INT_ARRAY (args + 6).array = outputs;
    }
}

