/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 * Unconfuse the autopluginlister: NOT A PLUGIN.
 */

#define DO_FORCED_READ 0

#include "gdamcallbackdevice.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <gdam/server/gdamconfigfile.h>
#include <gdam/server/gdamarctan.h>
#include <gdam/common/gdamfdpoll.h>

static void gdam_callback_device_init(GdamCallbackDevice* callback_device)
{
  GdamOutputDevice *output_device;
  GdamSample big_one;
  GdamSample16 _one[1];

  output_device = GDAM_OUTPUT_DEVICE (callback_device);
  if (DO_FORCED_READ)
    callback_device->use_semaphore = 1;
  output_device->missing_device_fd = 1;

  /* TODO:  figure out which systems don't allow named semaphores
   * at the configure stage */
#define USE_UNNAMED_SEMAPHORES 0

  if (callback_device->use_semaphore) 
    {

#if USE_UNNAMED_SEMAPHORES
      sem_t *newsem;

      newsem = g_new (sem_t, 1);
      sem_init (newsem, 0, 1);
      callback_device->semaphore = newsem;

      newsem = g_new (sem_t, 1);
      sem_init (newsem, 0, 1);
      callback_device->buffer_semaphore = newsem;
#else
      /* Find an unused semaphore  (from mosx-mpg123) */
      char s[10];
      gboolean not_done = TRUE;
      int attempts = 0;
      strcpy(s, "/gdamca-0000");
      do
      {
	      int m;
	      for(m = 11; m >= 8; m--)
		      if( (s[m]++) <= '9')
			      break;
		      else
			      s[m] = '0';
	callback_device->semaphore = sem_open(s, O_CREAT | O_EXCL, 0644, 0);
	not_done = (callback_device->semaphore == NULL);
	attempts++;
	if (attempts > 100 && not_done)
	  {
	    fprintf (stderr, "FAILED TO CREATE SEMAPHORE by %s\n", s);
	    not_done = FALSE;
	  }
      } while (not_done);

      attempts = 0;
      not_done = TRUE;
      do
      {
	      int m;
	      for(m = 11; m >= 8; m--)
		      if( (s[m]++) <= '9')
			      break;
		      else
			      s[m] = '0';
	callback_device->buffer_semaphore = sem_open(s, O_CREAT | O_EXCL, 0644, 0);
	not_done = (callback_device->buffer_semaphore == NULL);
	attempts++;
	if (attempts > 100 && not_done)
	  {
	    fprintf (stderr, "FAILED TO CREATE SEMAPHORE by %s\n", s);
	    not_done = FALSE;
	  }
      } while (not_done);

#endif  /* UnNamed Semaphores */

      g_return_if_fail (callback_device->semaphore != NULL);
      g_return_if_fail (callback_device->buffer_semaphore != NULL);

      /* semaphore is up for grabs */
      sem_post (callback_device->semaphore);
      sem_post (callback_device->buffer_semaphore);

      fprintf (stderr, "CallbackDevice: created and posted semaphores\n");
    }
}

#define GSTF(sv) ((float)(sv) / GDAM_SAMPLE_ONE);
static void 
convert_to_float        (void*       dst_ptr,
                         const void* src_ptr,
                         int         num_samples)
{
  GdamSample *samp_src;
  float *float_dest;
  gint32 biggest_in = -1;
  int i;
  samp_src = (GdamSample *)src_ptr;
  float_dest = (float *)dst_ptr;
/* For each stereo sample, convert to interleaved float */
  for (i = 0; i < num_samples - 1; i+=2) 
    {
      float_dest[0] = GSTF (samp_src->s[0]);
      float_dest[1] = GSTF (samp_src->s[1]);
      float_dest += 2;
      samp_src++;
    }
}

static void do_limit(int           limit_mode,
                     GdamSample*   buffer,
		     int           num_read)
{
#if (GDAM_SAMPLE_BITS == 24)
  if (limit_mode == 1)
    {
      while (num_read > 0)
	{
	  int t;
	  t = buffer->s[0];
	  buffer->s[0] = GDAM_SAMPLE_CLAMP (t);
	  t = buffer->s[1];
	  buffer->s[1] = GDAM_SAMPLE_CLAMP (t);
	  num_read--;
	  buffer++;
	}
    }
  else if (limit_mode == 2)
    {
      while (num_read > 0)
	{
	  int t;
	  t = buffer->s[0];
	  buffer->s[0] = gdam_arctan_b24 (t);
	  t = buffer->s[1];
	  buffer->s[1] = gdam_arctan_b24 (t);

	  num_read--;
	  buffer++;
	}
    }
#endif
}

/* virtual for adding data to the prebuffer */
static gboolean
real_fill_prebuffer (GdamOutputDevice *device)
{
  GdasSource *feed = GDAM_SINK (device)->feed;
  GdamCallbackDevice *callback_device = GDAM_CALLBACK_DEVICE (device);
  char *pre_buffer = device->pre_buffer;
  char *tmp_buffer;
  int fragment_size;
  int num_read, to_read;
  int sample_size;
  int end_index;
  int my_pre_buffer_amount, pre_buffer_amt, pre_buffer_start, pre_buffer_end;

  /* This entire function is optionally protected by a semaphore, to protect
   * against reentrance if it is called in an asynchronous fashion. */
  if (callback_device->use_semaphore)
    {
      sem_wait (callback_device->buffer_semaphore);
    }

  if ((feed == NULL) || (!gdas_source_is_ready_to_read (feed)))
    {
      if (callback_device->use_semaphore)
	{
	  sem_post (callback_device->buffer_semaphore);
	}
      return FALSE;
    }

/* Look at volatile values once, they may be changed under us by another thread.
 * although, start will only increase... so to_read is actually min_to_read
 */
  pre_buffer_end = device->pre_buffer_end;
  pre_buffer_start = device->pre_buffer_start;

  if (pre_buffer_end < pre_buffer_start) 
    pre_buffer_amt = pre_buffer_end + device->pre_buffer_alloced 
			   - pre_buffer_start;
  else
    pre_buffer_amt = pre_buffer_end - pre_buffer_start;

  //if (my_pre_buffer_amount != pre_buffer_amt)
  //  g_warning ("my_pba %d != pba %d (start %d end %d)", my_pre_buffer_amount, pre_buffer_amt, pre_buffer_start, pre_buffer_end);

  fragment_size = device->options.block_size;
  sample_size = device->total_sample_size;

  to_read = device->pre_buffer_alloced - pre_buffer_amt;
  /* Leave one buffer of empty space (this is just a hack to avoid ambiguity
   * when start == end... enables lockless operation. */
  to_read -= device->options.block_size;

  if (to_read <= 0 && FALSE)
    {
      fprintf (stderr, "deferring fill, amount %d\n", pre_buffer_amt);
      return TRUE;  /* TRUE -> don't let g_main_iteration_do block */
    }

  //fprintf (stderr, "TO_READ %d, amount %d\n", to_read, pre_buffer_amt);

  if (to_read > GDAM_MAX_ATOMIC_SOURCE_READ)
    to_read = GDAM_MAX_ATOMIC_SOURCE_READ;
  /* make sure we leave at least one whole block of data */
  if (to_read + pre_buffer_amt < device->options.block_size)
    to_read = device->options.block_size - pre_buffer_amt;

  if (to_read <= 0)
    { 
      if (callback_device->use_semaphore)
	{
	  sem_post (callback_device->buffer_semaphore);
	}
      return FALSE;
    }

  tmp_buffer = callback_device->fill_tmp_buffer;

  num_read = gdas_source_read (feed, (GdamSample *) tmp_buffer, to_read);
  if (num_read > to_read)
    {
      fprintf (stderr, "WARNING, got too many samples in fill_prebuffer\n");
      return FALSE;
    }

  if (num_read <= 0)
    { 
      if (callback_device->use_semaphore)
	{
	  sem_post (callback_device->buffer_semaphore);
	}
      return FALSE;
    }

  do_limit (device->limit_mode, (GdamSample *) tmp_buffer, num_read);
  if (device->to_native_channels != NULL)
    (*device->to_native_channels) (tmp_buffer, tmp_buffer, num_read);

    {
      int channel_samps;
      channel_samps = device->options.num_channels * num_read;
      convert_to_float (tmp_buffer, tmp_buffer, channel_samps);
    }

  /* Copy the samples into pre_buffer */
  end_index = pre_buffer_end;
  if (num_read + end_index >= device->pre_buffer_alloced)
    {
      int p1 = device->pre_buffer_alloced - end_index;
      int p2 = num_read - p1;
      /*g_message("p1/p2=%d/%d",p1,p2); */
      memcpy (pre_buffer + end_index * sample_size,
	      tmp_buffer, p1 * sample_size);
      memcpy (pre_buffer, tmp_buffer + p1 * sample_size, p2 * sample_size);
    }
  else
    {
      /*g_message("num_read=%d",num_read); */
      memcpy (pre_buffer + end_index * sample_size,
	      tmp_buffer, num_read * sample_size);
    }

    end_index += num_read;
    if (end_index >= device->pre_buffer_alloced)
      end_index -= device->pre_buffer_alloced;
    device->pre_buffer_end = end_index;

  //fprintf (stderr, "FILLED PREBUFFER %d samples to %d (start idx %d end idx %d)\n", num_read, pre_buffer_amt + num_read, device->pre_buffer_start, device->pre_buffer_end);

  if (callback_device->use_semaphore)
    {
      sem_post (callback_device->buffer_semaphore);
    }

  return FALSE;
}


/*  Fills a buffer with one fragment of audio.  This is called asynchronously
 *  by the driver callback function, so we avoid glib functions and use
 *  a semaphore to protect our data structs. */
gboolean
gdam_callback_device_fill_one_frag (GdamCallbackDevice *callback_device,
				    GdamSample     *fragment_buffer_out,
				    int            frag_buffer_size_out,
                                    int                     _samples_out)
{
  GdamSample *fragment_buffer;
  GdamOutputDevice *output_device = GDAM_OUTPUT_DEVICE (callback_device);
  GdamOutputDeviceClass *klass;
  GdamSink *sink;
  GdasSource *feed;
  int fragment_size, block_size, frag_buffer_size, sample_size, num_samples;
  int old_pre_buffer_start, old_pre_buffer_amount, pre_buffer_end;
  int new_pre_buffer_start;
  gboolean was_prebuffer, buffer_has_data;
  was_prebuffer = FALSE;
  buffer_has_data = TRUE;

  sink = (GdamSink *) output_device;
  feed = sink->feed;

  block_size = _samples_out;
  sample_size = output_device->total_sample_size;
  num_samples = block_size;

  frag_buffer_size = sample_size * block_size;
  if (frag_buffer_size != frag_buffer_size_out) 
    {
      fprintf (stderr, "fsize %d does not match fsizeout %d (sample size %d block size %d)\n", 
	       frag_buffer_size, frag_buffer_size_out,
 	       sample_size, block_size);
      if (frag_buffer_size != frag_buffer_size_out)
	{
	  fprintf (stderr, "WARNING: frag sizes don't match, aborting\n");
	  return 0;
	}
    }

  fragment_buffer = callback_device->temp_buffer;

  old_pre_buffer_start = output_device->pre_buffer_start;
  pre_buffer_end = output_device->pre_buffer_end;

  if (pre_buffer_end < old_pre_buffer_start)
    old_pre_buffer_amount = pre_buffer_end + output_device->pre_buffer_alloced 
			   - old_pre_buffer_start;
  else
    old_pre_buffer_amount = pre_buffer_end - old_pre_buffer_start;

#if 0
  fprintf (stderr, "frag_size %d from block %d stereo_sample_size %d total_sample_size %d (requested num_samples: %d)\n",
	     frag_buffer_size, block_size, sample_size, 
             output_device->total_sample_size, _samples_out);
#endif

/* XXX: we see constant underruns all the time when there are no sources feeding
 * the mix... and yet it seems sink->feed is always ready to read?? */

/* If we don't have sufficient data in the buffer... */
  if (old_pre_buffer_amount < block_size)
    {
#if DO_FORCED_READ
      /* TODO: is_ready_to_read is not thread safe */
      gboolean ready = gdas_source_is_ready_to_read (feed);

      /* if sink->feed is ready to be read from, then force the buffer full */
      if (ready && DO_FORCED_READ) 
	{
	  GdamOutputDeviceClass *device_class;
	  device_class = GDAM_OUTPUT_DEVICE_CLASS (GTK_OBJECT (output_device)->klass);
	  //gboolean will_wait = sem_trywait (output_device->buffer_semaphore);
          /* If we would wait on the semaphore, then we know someone is already
           * working to fill the pre_buffer.  Wait until we get the lock, and
           * data will be there. */
          //if (will_wait) {
	  //    sem_wait (output_device->buffer_semaphore);
	  //    sem_post (output_device->buffer_semaphore);
	  //}
          (*device_class->fill_buffer) (output_device);
	  fprintf (stderr, "forced_read\n");

	  pre_buffer_end = output_device->pre_buffer_end;
	  if (pre_buffer_end < old_pre_buffer_start)
	    old_pre_buffer_amount = pre_buffer_end 
                                    + output_device->pre_buffer_alloced 
				    - old_pre_buffer_start;
	  else
	    old_pre_buffer_amount = pre_buffer_end - old_pre_buffer_start;

          if (old_pre_buffer_amount < block_size) 
            {
	      fprintf (stderr, "double underrun!\n");
	      bzero (fragment_buffer, frag_buffer_size);
	      buffer_has_data = FALSE;
	    }
	}
      else
#endif
	/* we have an underrun */
        {
	  fprintf (stderr, "!callback underrun %d in buffer need %d\n",
	           old_pre_buffer_amount, block_size);
	  bzero (fragment_buffer, frag_buffer_size);
          buffer_has_data = FALSE;
	}
    }

  new_pre_buffer_start = old_pre_buffer_start;

  if (buffer_has_data)
    {
      if (old_pre_buffer_start + block_size
	  >= output_device->pre_buffer_alloced)
	{
	  int p1 = output_device->pre_buffer_alloced - old_pre_buffer_start;
	  int p2 = block_size - p1;
	  int p1_bytes = p1 * sample_size;
	  memcpy (fragment_buffer,
		  output_device->pre_buffer
		  + old_pre_buffer_start * sample_size, p1_bytes);
	  memcpy (fragment_buffer + p1_bytes,
		  output_device->pre_buffer, p2 * sample_size);
	  new_pre_buffer_start = p2;
	}
      else
	{
	  fragment_buffer = output_device->pre_buffer
	    + sample_size * old_pre_buffer_start;
	  new_pre_buffer_start = old_pre_buffer_start + block_size;
	}

      /* request a buffer fill. */
      write (callback_device->ipc_write_fd, "1", 1);

      output_device->pre_buffer_start = new_pre_buffer_start;
      
      //fprintf (stderr, "CallbackDevice: emptied prebuffer %d stereo samples to %d stereo samples (start idx %d end idx %d)\n", block_size, old_pre_buffer_amount - block_size, output_device->pre_buffer_start, output_device->pre_buffer_end);
    }

  output_device->has_written = TRUE;

  //*fragment_buffer_out = (GdamSample*) malloc (frag_buffer_size);
  memcpy (fragment_buffer_out, fragment_buffer, frag_buffer_size);

  return TRUE;
}


#if 0
/* we provide this function to service the audio device. 
   output_data should be filled with data before returning. */
static OSStatus 
audio_io_proc (AudioDeviceID device,
               const AudioTimeStamp  *now,
               const AudioBufferList *input_data,
               const AudioTimeStamp  *in_time,
               AudioBufferList       *output_data,
               const AudioTimeStamp  *out_time,
               void                  *client_data)
{
  GdamCallbackDevice *callback_device;
  GdamOutputDevice *output_device;
  GdamSink *sink;
  int num_buffers, num_channels;
  int i;

  callback_device = (GdamCallbackDevice *) client_data;
  output_device = (GdamOutputDevice *) callback_device;
  sink = (GdamSink *) output_device;
  num_buffers = output_data->mNumberBuffers;

  if (!sink->feed) {
    bzero (output_data->mBuffers[0].mData, 
           output_data->mBuffers[0].mDataByteSize);
    return;
  }


  for (i=0; i < num_buffers; i++)
    {
      long samples_out;
      gboolean rv;
      samples_out = output_data->mBuffers[i].mDataByteSize / sizeof(float);
      num_channels = output_data->mBuffers[i].mNumberChannels;

      //fprintf (stderr, "audio_io_proc on %d buffers of %d bytes %d samples %d channels\n", num_buffers, output_data->mBuffers[i].mDataByteSize, samples_out, num_channels);

      rv = one_frag (output_device, 
                     output_data->mBuffers[i].mData, 
		     callback_device->frag_buffer_size,
		     samples_out / num_channels);
      if (!rv) {
	g_warning ("couldn't read one frag");
      }
    }
}
#endif
               

static int    
gdam_callback_device_write_pcm     (GdamOutputDevice*  device, 
				      gconstpointer      buffer, 
				      int                length)
{
  fprintf (stderr, "CALLBACK DEVICE WRITE PCM??\n");
  return 0;
}

#define DEFAULT_PREBUFFER_MULTIPLE	4

static void
real_device_options_changed (GdamOutputDevice *device)
{
  GdamOutputDeviceOptions *options = &device->options;
  GdamCallbackDevice *callback_device = GDAM_CALLBACK_DEVICE (device);
  gpointer old_to_native = device->to_native_samples;
  gboolean is_big_endian;
  int pbmultiple = DEFAULT_PREBUFFER_MULTIPLE;
  is_big_endian = (G_BYTE_ORDER == G_BIG_ENDIAN);
  if (options->is_wrong_endian)
    is_big_endian = !is_big_endian;
  device->to_native_samples =
    gdam_sample_converter_get (GDAM_SAMPLE_BITS, TRUE,
			       (G_BYTE_ORDER == G_BIG_ENDIAN),
			       options->bits_per_sample,
			       options->is_signed, is_big_endian);
  device->to_native_channels =
    gdam_sample_converter_get_channel_changer (2, options->num_channels);
  if (options->bytes_per_sample == 0)
    {
      if (options->bits_per_sample == 24)
	options->bytes_per_sample = 4;
      else
	options->bytes_per_sample = options->bits_per_sample / 8;
    }
  {
    int input_size = sizeof (GdamSample);
    int num_channels = options->num_channels;
    int mid_size = sizeof (GdamSampleValue) * num_channels;
    int output_size = options->bytes_per_sample * num_channels;
    int i;
    i = MAX (input_size, mid_size);
    i = MAX (i, output_size);
    device->conversion_sample_size = i;
    callback_device->fill_tmp_buffer = (char *) malloc (i * GDAM_MAX_ATOMIC_SOURCE_READ);
  }

  if (callback_device->use_semaphore) 
    {
      while (sem_wait (callback_device->buffer_semaphore)) {}
    }

  if (gdam_prebuffer_multiple > 0)
    pbmultiple = gdam_prebuffer_multiple;

  device->total_sample_size = options->bytes_per_sample
    * options->num_channels;
  device->total_sample_size = sizeof (GdamSample);
  device->pre_buffer_alloced = device->total_sample_size
    * options->block_size * pbmultiple;

  /* Leave one buffer of empty space */
  device->pre_buffer_alloced += options->block_size;
  device->pre_buffer_start = 0;
  device->pre_buffer_end = 0;
  device->pre_buffer = g_realloc (device->pre_buffer,
				  device->pre_buffer_alloced
				  * device->total_sample_size);
  g_message ("alloced %d bytes from sampsize %d block size %d samples, multiple 2.0", device->pre_buffer_alloced, device->total_sample_size, options->block_size);
  /* Unlock the semaphore */
  if (callback_device->use_semaphore)
    {
      sem_post(callback_device->buffer_semaphore);
    }
}

static void
callback_device_destroy (GtkObject *object)
{
  GdamCallbackDevice *callback_device = GDAM_CALLBACK_DEVICE (object);
  if (callback_device->fragment_buffer_out)
    g_free (callback_device->fragment_buffer_out);
  if (callback_device->temp_buffer)
    g_free (callback_device->temp_buffer);
  if (callback_device->semaphore)
    sem_destroy (callback_device->semaphore);
  if (callback_device->buffer_semaphore)
    sem_destroy (callback_device->buffer_semaphore);
}

static void 
gdam_callback_device_class_init (GdamCallbackDeviceClass *callback_device_class)
{
  GdamOutputDeviceClass* klass;
  klass = GDAM_OUTPUT_DEVICE_CLASS(callback_device_class);
  klass->fill_buffer = real_fill_prebuffer;
  klass->options_changed = real_device_options_changed;
  //klass->write_pcm = gdam_callback_device_write_pcm;
}


GtkType gdam_callback_device_get_type()
{
	static GtkType callback_device_type = 0;
	if (!callback_device_type) {
		static const GtkTypeInfo callback_device_info = {
			"GdamCallbackDevice",
			sizeof(GdamCallbackDevice),
			sizeof(GdamCallbackDeviceClass),
			(GtkClassInitFunc) gdam_callback_device_class_init,
			(GtkObjectInitFunc) gdam_callback_device_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_OUTPUT_DEVICE;
		callback_device_type = gtk_type_unique(parent,
		                                 &callback_device_info);
	}
	return callback_device_type;
}

/* 
 * This function is mandatory for plugin devices.
 */
GdamSink* gdam_callback_device_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
	GtkType type = gdam_callback_device_get_type();
	GdamCallbackDevice* callback_dev = (GdamCallbackDevice*)gtk_type_new(type);
	if (!gdam_sink_open(GDAM_SINK(callback_dev), device_name, device_options)) {
		gtk_object_unref(GTK_OBJECT(callback_dev));
		return NULL;
	}
	return GDAM_SINK(callback_dev);
}

static gboolean 
handle_ipc_event (GdamFdPoll     *poll_object,
		  int             fd,
		  GIOCondition    events,
		  gpointer        data)
{
  GdamCallbackDevice *callback_dev = GDAM_CALLBACK_DEVICE (data);
  GdamOutputDeviceClass *device_class;
  device_class = GDAM_OUTPUT_DEVICE_CLASS (GTK_OBJECT (callback_dev)->klass);

  if ((events & (G_IO_ERR | G_IO_HUP)) != 0)
    {
      fprintf (stderr, "ipc pipe had problem.  terminating\n");
      return FALSE;  /* remove poll */
    }
  if ((events & G_IO_IN) == G_IO_IN)
    {
      char buf[8];
      int rv = read (callback_dev->ipc_read_fd, &buf, 1);
      (*device_class->fill_buffer) (GDAM_OUTPUT_DEVICE (callback_dev));
    }
  return TRUE;  /* keep polling */
}


static void
destroy_notify(gpointer user_data)
{
  GdamCallbackDevice *callback_dev = GDAM_CALLBACK_DEVICE (user_data);
  callback_dev->ipc_fd_poll = NULL;
  fprintf (stderr, "IPC PIPE DESTROYED\n");
}

void
gdam_callback_device_add_ipc_poll (GdamCallbackDevice   *callback_dev)
{
  int pipe_fds[2];

  if (pipe (pipe_fds) < 0)
    {
      fprintf (stderr, "WARNING: pipe failed: %s\n", strerror (errno));
      return;
    }
  callback_dev->ipc_read_fd = pipe_fds[0];
  callback_dev->ipc_write_fd = pipe_fds[1];
  callback_dev->ipc_fd_poll
    = gdam_fd_poll_new (pipe_fds[0],
			GDAM_DEVICE_PRIORITY,
			G_IO_IN,
			handle_ipc_event,
			callback_dev,
			destroy_notify);
  fprintf (stderr, "added ipc poll\n");
}
