/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamdevice.h"
#include "gdamnamespace.h"

typedef struct _GdamDeviceInfo GdamDeviceInfo;

struct _GdamDeviceInfo {
	char* device_name;
	char* device_options;
	GdamDeviceOpenFunc open_func;
	gpointer device_data;
	GdamAcl* access_control;
	GdamSink* device;
	unsigned device_is_open : 1;
};

static GHashTable* device_table = NULL;
static GHashTable* get_device_table()
{
  if (!device_table)
    device_table = g_hash_table_new (g_str_hash, g_str_equal);
  return device_table;
}


gboolean gdam_device_may_open(const char* device, GdamPermissions* perm)
{
  GdamDeviceInfo *info;
  g_return_val_if_fail (device != NULL, FALSE);
  info = (GdamDeviceInfo *) g_hash_table_lookup (device_table, device);
  if (info == NULL)
    {
      g_warning ("gdam_device_may_open: device %s doesn't exist", device);
      return FALSE;
    }
  return gdam_acl_is_granted (info->access_control, perm);
}

GdamSink* gdam_device_is_open(const char* device)
{
  GdamDeviceInfo *info;
  g_return_val_if_fail (device != NULL, NULL);
  info = (GdamDeviceInfo *) g_hash_table_lookup (device_table, device);
  if (info != NULL && info->device != NULL)
    return info->device;
  return NULL;
}

void gdam_device_reopen(const char* device_name)
{
  GdamDeviceInfo *info;
  GdamSink *device;
  GdamSinkClass *sink_class;

  info = g_hash_table_lookup (device_table, device_name);
  if (info == NULL || info->device == NULL)
    {
      gdam_debug ("device_reopen(%s) called on non-open device", device_name);
      return;
    }
  device = info->device;
  sink_class = GDAM_SINK_CLASS (GTK_OBJECT (device)->klass);
  if (sink_class->close == NULL)
    g_warning ("device has no way to close...");
  else
    (*sink_class->close) (device);
  if (sink_class->open == NULL)
    g_warning ("device has no way to open...");
  else
    {
      if (!(*sink_class->open) (device, device_name, info->device_options))
	g_warning ("device %s couldn't be opened", device_name);
    }
}


GdamSink* gdam_device_open(const char* dev_name)
{
  GdamDeviceInfo *info;
  GdamSink *rv;
  g_return_val_if_fail (dev_name != NULL, NULL);

  if (!device_table)
    {
      g_warning ("gdam_device_open: no devices have been registered");
      return NULL;
    }
  if (!device_table
      || (!(info = (GdamDeviceInfo *) g_hash_table_lookup (device_table,
							   dev_name))))
    {
      g_warning ("gdam_device_open: device %s doesn't exist",
		 dev_name);
      return NULL;
    }
  if (info->device_is_open)
    {
      g_warning ("gdam_device_open: device %s cannot be opened twice",
		 dev_name);
      return NULL;
    }
  rv = info->open_func (info->device_data, dev_name, info->device_options);
  if (rv != NULL)
    {
      int id;
      g_message ("opened device %s", dev_name);
      info->device_is_open = 1;
      id = GDAM_SERVER_OBJECT (rv)->global_id;
      gdam_namespace_int_set ("devices", dev_name, id);
    }
  info->device = rv;
  return rv;
}

void gdam_device_register(const char*        device_name,
                          GdamDeviceOpenFunc open_func,
			  gpointer           device_data,
			  const char*        device_options,
		          GdamAcl*           access_control)
{
  GdamDeviceInfo *info;
  g_return_if_fail (device_name != NULL);
  if (g_hash_table_lookup (get_device_table (), device_name))
    {
      g_warning ("attempted to register sound device %s twice", device_name);
      return;
    }
  info = (GdamDeviceInfo *) g_new0 (GdamDeviceInfo, 1);
  info->device_name = g_strdup (device_name);
  info->open_func = open_func;
  info->device_data = device_data;
  info->access_control = access_control;
  info->device_is_open = 0;
  info->device_options = g_strdup (device_options);
  g_hash_table_insert (device_table, info->device_name, info);
}

void gdam_device_is_closing(GdamSink* device)
{
  GdamDeviceInfo *info;
  g_return_if_fail (device != NULL);
  info = (GdamDeviceInfo *) g_hash_table_lookup (device_table, device);
  g_return_if_fail (info != NULL);
  info->device_is_open = 1;
}
