#include "gdamintervalmap.h"

static GMemChunk *fencepost_chunk = NULL;


static GdamIntervalMapPost * fencepost_alloc ()
{
  if (fencepost_chunk == NULL)
    fencepost_chunk = g_mem_chunk_create ("GdamIntervalMapPost",
				          16,
				          G_ALLOC_AND_FREE);
  return g_chunk_new (GdamIntervalMapPost, fencepost_chunk);
}
static void fencepost_free (GdamIntervalMapPost *fencepost)
{
  g_chunk_free (fencepost, fencepost_chunk);
}
static int fencepost_compare (GdamIntervalMapPost *a, GdamIntervalMapPost *b)
{
  if (a->value < b->value)
    return -1;
  if (a->value > b->value)
    return +1;
  return 0;
}

struct _GdamIntervalMap
{
  GTree *by_value;
  GdamIntervalMapPost *first;
  GdamIntervalMapPost *last;
};

typedef struct _SearchInfo SearchInfo;
struct _SearchInfo
{
  int desired;
  GdamIntervalMapPost *last;
};

static gint gdam_interval_map_search_func (gpointer       key,
                                           gpointer       data)
{
  GdamIntervalMapPost *post = key;
  SearchInfo *info = data;
  info->last = post;
  if (info->desired < post->value)
    return -1;
  if (info->desired > post->value)
    return +1;
  return 0;
}

static GdamIntervalMapPost *
gdam_interval_map_try_fencepost (GdamIntervalMap *map,
                                 int              value)
{
  SearchInfo search_info;
  GdamIntervalMapPost *at;
  search_info.desired = value;
  search_info.last = NULL;
  g_tree_search (map->by_value, gdam_interval_map_search_func, &search_info);
  if (search_info.last == NULL)
    return NULL;

  at = search_info.last;
  while (at != NULL && at->value > value)
    at = at->prev;
  if (at != NULL)
    while (at->next != NULL && at->next->value <= value)
      at = at->next;
  return at;
}

static GdamIntervalMapPost *
gdam_interval_map_force_fencepost (GdamIntervalMap *map,
                                   int              value)
{
  GdamIntervalMapPost *prev = gdam_interval_map_try_fencepost (map, value);
  if (prev == NULL)
    { 
      /* need to add a new first element */
      GdamIntervalMapPost *new_first = fencepost_alloc ();
      new_first->value = value;
      new_first->prev = NULL;
      new_first->next = map->first;
      new_first->objects = NULL;

      if (map->first == NULL)
        map->first = map->last = new_first;
      else
        {
	  map->first->prev = new_first;
	  map->first = new_first;
	}

      g_tree_insert (map->by_value, new_first, new_first);
      return new_first;
    }
  else if (prev->value == value)
    return prev;
  else
    {
      GdamIntervalMapPost *exact = fencepost_alloc ();
      exact->value = value;
      exact->prev = prev;
      exact->next = prev->next;
      exact->objects = g_slist_copy (prev->objects);

      if (exact->next != NULL)
        exact->next->prev = exact;
      else
        map->last = exact;
      exact->prev->next = exact;

      g_tree_insert (map->by_value, exact, exact);
      return exact;
    }

  g_assert_not_reached ();
  return NULL;
}

GdamIntervalMap    *gdam_interval_map_new      ()
{
  GdamIntervalMap *map;
  map = g_new (GdamIntervalMap, 1);
  map->by_value = g_tree_new ((GCompareFunc) fencepost_compare);
  map->first = map->last = NULL;
  return map;
}

void                gdam_interval_map_add      (GdamIntervalMap *map,
                                                int              min_value,
                                                int              max_value,
						gpointer         object)
{
  GdamIntervalMapPost *start;
  GdamIntervalMapPost *end;
  GdamIntervalMapPost *at;

  g_return_if_fail (min_value <= max_value);
  g_return_if_fail (object != NULL);
  g_return_if_fail (map != NULL);

  start = gdam_interval_map_force_fencepost (map, min_value);
  end = gdam_interval_map_force_fencepost (map, max_value + 1);

  at = start;
  for (at = start; at != end; at = at->next)
    at->objects = g_slist_prepend (at->objects, object);
}

GdamIntervalMapPost *gdam_interval_map_lookup   (GdamIntervalMap *map,
                                                 int              value)
{
  GdamIntervalMapPost *at;
  at = gdam_interval_map_try_fencepost (map, value);
  if (at == NULL)
    return NULL;
  return at;
}
GdamIntervalMapPost *gdam_interval_map_first    (GdamIntervalMap *map)
{
  return map->first;
}

void                gdam_interval_map_destroy  (GdamIntervalMap *map)
{
  GdamIntervalMapPost *at;
  GdamIntervalMapPost *next = NULL;
  g_tree_destroy (map->by_value);
  for (at = map->first; at != NULL; at = next)
    {
      next = at->next;
      g_slist_free (at->objects);
      fencepost_free (at);
    }
  g_free (map);
}
void                gdam_interval_map_clear    (GdamIntervalMap *map)
{
  GdamIntervalMapPost *at;
  GdamIntervalMapPost *next = NULL;
  g_tree_destroy (map->by_value);
  map->by_value = g_tree_new ((GCompareFunc) fencepost_compare);
  for (at = map->first; at != NULL; at = next)
    {
      next = at->next;
      g_slist_free (at->objects);
      fencepost_free (at);
    }
  map->first = map->last = NULL;
}
