/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamauthmaker.h"
#include "gdamserver.h"
#include "gdamconnection.h"
#include "gdamserverdebug.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>

static gboolean 
gdam_server_handle_events        (GdamFdPoll     *poll_object,
				  int             fd,
				  GIOCondition    events,
				  gpointer        data)
{
  GdamServer *server;
  int listening_fd;
  int accepted_fd;
  struct sockaddr addr;
  int addr_size;

  server = (GdamServer *) (data);
  listening_fd = server->socket_fd;

  g_return_val_if_fail (server->fd_poll == poll_object, FALSE);
  if (listening_fd <= 0)
    return FALSE;
  if ((events & (G_IO_ERR | G_IO_HUP)) != 0)
    {
      g_message ("server incoming connection had problem.");
      return FALSE;
    }

  if ((events & (G_IO_IN)) == 0)
    {
      g_message ("server incoming connection missing data.");
      return FALSE;
    }

  /* Accept a new connection (at the socket layer.) */
  bzero (&addr, sizeof (addr));
  addr_size = sizeof (struct sockaddr);
  accepted_fd = accept (listening_fd, &addr, &addr_size);
  if (accepted_fd < 0)
    {
      if (errno == EWOULDBLOCK || errno == EINTR)
	return TRUE;
      g_warning ("gdam_server_dispatch: "
		 "error accepting new connection: %s", strerror (errno));
      return FALSE;
    }

  if (server->accept_check_func != NULL)
    {
      gpointer data = server->accept_check_data;
      gboolean ok;
      ok = (*server->accept_check_func) (accepted_fd, &addr, data);
      if (!ok)
	{
	  close (accepted_fd);
	  gdam_debug ("connection not accepted: failed check");
	  return TRUE;
	}
    }

  if (!(*server->new_conn_func) (accepted_fd,
				 server,
				 server->new_conn_data))
    {
      g_warning ("failed to make new connection");
      return FALSE;
    }
  return TRUE;
}

static gboolean
default_new_connection_func (gint              accepted_fd,
			     GdamServer       *server,
			     gpointer          unused)
{
  GdamAuthenticator *authenticator;
  GdamConnection *accepted_connection;

  g_assert (unused == NULL);

  /* Make a connection and attach it to the main loop. */
  authenticator = gdam_auth_maker_make (server->auth_maker);
  accepted_connection = gdam_connection_new (accepted_fd, authenticator);
  if (accepted_connection == NULL)
    return FALSE;

  gdam_debug ("connection %p on fd %d attached to mainloop",
	      accepted_connection, accepted_fd);

  gdam_connection_attach_to_main_loop (accepted_connection);
  return TRUE;
}


GdamServer *
gdam_server_new (int            sock_fd,
		 GdamAuthMaker *auth_maker)
{
  return gdam_server_new_full (sock_fd, auth_maker, NULL, NULL, NULL);
}

GdamServer *
gdam_server_new_full (int                         sock_fd,
		      GdamAuthMaker              *auth_maker,
		      GdamServerNewConnectionFunc new_conn_func,
		      gpointer                    new_conn_data,
		      GDestroyNotify              new_conn_destroy)
{
  GdamServer *server = g_new (GdamServer, 1);
  server->auth_maker = auth_maker;
  server->socket_fd = sock_fd;
  server->is_destructing = 0;
  server->accept_check_func = NULL;
  server->accept_check_data = NULL;
  server->accept_check_destroy = NULL;
  if (new_conn_func == NULL)
    {
      server->new_conn_func = default_new_connection_func;
      server->new_conn_data = NULL;
      server->new_conn_destroy = NULL;
    }
  else
    {
      server->new_conn_func = new_conn_func;
      server->new_conn_data = new_conn_data;
      server->new_conn_destroy = new_conn_destroy;
    }
  return server;
}

void
gdam_server_attach(GdamServer *server)
{
  server->fd_poll = gdam_fd_poll_new (server->socket_fd,
				      G_PRIORITY_LOW,
				      G_IO_IN,
				      gdam_server_handle_events,
				      server,
				      (GDestroyNotify) gdam_server_destroy);
}

void
gdam_server_detach(GdamServer *server)
{
  if (server->fd_poll != NULL)
    {
      gdam_fd_poll_destroy (server->fd_poll);
      server->fd_poll = NULL;
    }
}

void
gdam_server_destroy(GdamServer* server)
{
  if (server->is_destructing)
    return;
  server->is_destructing = 1;
  gdam_server_detach (server);
  if (server->socket_fd >= 0)
    {
      close (server->socket_fd);
      server->socket_fd = -1;
    }
  if (server->accept_check_destroy != NULL)
    (*server->accept_check_destroy) (server->accept_check_data);
  g_free (server);
}
