/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamserver.h"
#include "gdammain.h"
#include "gdamauthmaker.h"
#include "gdamconnection.h"
#include "gdamcleanup.h"
#include <gdam/common/gdamdebug.h>
#include <gdam/common/gdamarg.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <glib.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef __linux__
#include <linux/un.h>
#else
#include <sys/un.h>
#endif
#include <netdb.h>
#include <errno.h>
#include "gdamconfigfile.h"
#include "gdamrealtime.h"

#define DEFAULT_PORT 6003

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

static void
usage(const char* prog_name)
{
  fprintf (stderr, "usage: %s --port port_no --config config.file\n\n"
	   "Run a GDAM server (on the specified port).\n"
	   "\nOptions:\n"
	   "\t--port portno   : set the listening port number.\n"
	   "\t--config config : set the configuration file.\n"
	   "\t--realtime      : turn on realtime scheduling.\n"
	   "\n\nused by other programs only:\n"
	   "\t--control-fd fd : accept a connection on the "
	   "given file descriptor.\n"
	   "\t--dont-bind     : don't bind to any port.\n"
	   "\t--guess-config-file : try a bunch of common config\n"
	   "                        file locations\n"
	   "\t--debug-{SECTION} : debug some section of the code; SECTION\n"
	   "\t                    should be one of the following:\n"
	   "\t                       dispatch    Debug dispatching.\n"
	   "\t                       arg         Debug argument setting.\n"
	   "\t                       lifetime    Debug object creation/death.\n"
	   "\t                       connection  Debug the connection.\n"
	   "\t                       ref         Debug refcounts.\n"
	   , prog_name);
  exit (1);
}

/*
 *  _
 * | |_ ___ _ __    ___  ___ _ ____   _____ _ __
 * | __/ __| '_ \  / __|/ _ \ '__\ \ / / _ \ '__|
 * | || (__| |_) | \__ \  __/ |   \ V /  __/ |
 *  \__\___| .__/  |___/\___|_|    \_/ \___|_|
 *         |_|
 */

static int bind_tcp(int port_number)
{
  struct sockaddr_in sin;
  int rv;
  int one = 1;
  bzero (&sin, sizeof (sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (port_number);
  /*sin.sin_addr = htonl(INADDR_ANY); */
  rv = socket (PF_INET, SOCK_STREAM, 0);
  if (rv < 0)
    {
      perror ("error creating listening socket??");
      return -1;
    }
  fcntl (rv, F_SETFD, FD_CLOEXEC);
  if (setsockopt (rv, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (one)) < 0)
    perror ("warning: couldn't make sockaddr reuseable");
  if (gdam_bind_localhost)
    {
      struct hostent *hent;
      hent = gethostbyname ("localhost");
      memcpy (&sin.sin_addr, hent->h_addr_list[0], hent->h_length);
    }
  if (bind (rv, (struct sockaddr *) (&sin), sizeof (struct sockaddr)) < 0)
    {
      fprintf (stderr, "bind: failed to bind to port %d: %s.\n\n",
	       port_number, strerror (errno));
      if (port_number < 1024 && getuid ())
	{
	  fprintf (stderr, "A likely reason is that only root "
		   "may bind to low-numbered ports.\n");
	}
      close (rv);
      return -1;
    }
  if (listen (rv, SOMAXCONN) < 0)
    {
      fprintf (stderr, "error listening: %s\n", strerror (errno));
      close (rv);
      return -1;
    }
  return rv;
}

/* XXX: TODO: add ipv6 support */
static gboolean check_tcp_connect(int              fd,
                                  gpointer         sockaddr_ptr,
                                  gpointer         accept_check_data)
{
  int port_no = GPOINTER_TO_INT (accept_check_data);
  struct sockaddr_in *in_addr = sockaddr_ptr;
  char *ip_addr = (char *) &(in_addr->sin_addr);
  GSList *allowed_addrs;

  g_return_val_if_fail (in_addr->sin_family == AF_INET, FALSE);

  allowed_addrs = gdam_config_lookup_tcp_auth (port_no);
  if (allowed_addrs == NULL)
    return TRUE;
  while (allowed_addrs != NULL)
    {
      if (memcmp (allowed_addrs->data, ip_addr, 4) == 0)
	return TRUE;
      allowed_addrs = allowed_addrs->next;
    }
  return FALSE;
}

static gboolean make_server_tcp(int port_no)
{
  GdamServer *server;
  int fd = bind_tcp (port_no);
  if (fd < 0)
    {
      g_warning ("error binding to port %d", port_no);
      return FALSE;
    }
  server = gdam_server_new (fd, gdam_auth_maker_root_new ());

  if (!server)
    {
      g_warning ("error creating server");
      return FALSE;
    }

  server->accept_check_func = check_tcp_connect;
  server->accept_check_data = GINT_TO_POINTER (port_no);
  gdam_server_attach (server);
  return TRUE;
}

/*
 *              _               _                       _
 *  _   _ _ __ (_)_  __      __| | ___  _ __ ___   __ _(_)_ __
 * | | | | '_ \| \ \/ /____ / _` |/ _ \| '_ ` _ \ / _` | | '_ \
 * | |_| | | | | |>  <_____| (_| | (_) | | | | | | (_| | | | | |
 *  \__,_|_| |_|_/_/\_\     \__,_|\___/|_| |_| |_|\__,_|_|_| |_|
 * 
 * 
 *  ___  ___ _ ____   _____ _ __
 * / __|/ _ \ '__\ \ / / _ \ '__|
 * \__ \  __/ |   \ V /  __/ |
 * |___/\___|_|    \_/ \___|_|
 * 
 */

static int bind_unix_domain(const char* unix_path)
{
  struct sockaddr_un un_addr;
  int rv;
  int one = 1;
  bzero (&un_addr, sizeof (un_addr));
  un_addr.sun_family = AF_UNIX;
  strncpy (un_addr.sun_path, unix_path, UNIX_PATH_MAX);
  rv = socket (PF_UNIX, SOCK_STREAM, 0);
  if (rv < 0)
    {
      perror ("error creating listening socket??");
      return -1;
    }
  fcntl (rv, F_SETFD, FD_CLOEXEC);
  if (bind (rv, (struct sockaddr *) (&un_addr), sizeof (struct sockaddr)) < 0)
    {
      fprintf (stderr, "bind: failed to bind to %s: %s.\n\n",
	       unix_path, strerror (errno));
      close (rv);
      return -1;
    }
  if (listen (rv, SOMAXCONN) < 0)
    {
      fprintf (stderr, "error listening: %s\n", strerror (errno));
      close (rv);
      return -1;
    }
  return rv;
}
#if HAS_UNIX_DOMAIN_CREDENTIALS
/* TODO: Figure this stuff out.
 *       It's broken [and thus unused always]
 *   Issues:
 *      - the exact syntax/call-sequence is tricky -- find an example
 *      - the cmsg should be received back in the main loop ... as
 *        it is, it could hang the server forever waiting 
 *        for the credentials.
 */
static int get_credential_uid(int fd)
{
  /* uh, this is *client's side* */
  struct cmsghdr auth_header;
  auth_header.cmsg_level = SOL_SOCKET;	/* historical */
  auth_header.cmsg_type = SCM_CREDENTIALS;
...}
#endif

/* XXX: TODO: add ipv6 support */
static gboolean check_unix_domain_connect(int              fd,
                                          gpointer         sockaddr_ptr,
                                          gpointer         accept_check_data)
{
  int port_no = GPOINTER_TO_INT (accept_check_data);
  struct sockaddr_un *un_addr = sockaddr_ptr;
  GSList *allowed_uids;

  g_return_val_if_fail (un_addr->sun_family == AF_UNIX, FALSE);

  allowed_uids = gdam_config_lookup_unix_auth (un_addr->sun_path);
  if (allowed_uids == NULL)
    return TRUE;
#if HAS_UNIX_DOMAIN_CREDENTIALS
  socket_uid = get_credential_uid (fd);
  if (socket_uid < 0)
    {
      g_warning ("error establishing user identity");
      return FALSE;
    }
  while (allowed_uids != NULL)
    {
      int uid = GPOINTER_TO_INT (allowed_uids->data);
      if (uid == socket_uid)
	return TRUE;
      allowed_uids = allowed_uids->next;
    }
#endif
  return FALSE;
}

static gboolean make_server_unix(const char* unix_path)
{
  GdamServer *server;
  int fd = bind_unix_domain (unix_path);
  if (fd < 0)
    {
      g_warning ("error binding to unix-domain %s", unix_path);
      return FALSE;
    }
  server = gdam_server_new (fd, gdam_auth_maker_root_new ());
  if (!server)
    {
      g_warning ("error creating server");
      return FALSE;
    }

  server->accept_check_func = check_unix_domain_connect;
  server->accept_check_data = g_strdup (unix_path);
  server->accept_check_destroy = g_free;

  gdam_server_attach (server);
  return TRUE;
}

static const char *debug_flags[] = {
  "a:arg",
  "d:dispatch",
  "l:lifetime",
  "c:connection",
  "r:ref",
  NULL
};

int main(int argc, char** argv)
{
  int i;
  gboolean made_server = FALSE;
  gboolean try_realtime = TRUE;
  gboolean dont_make_server = FALSE;
  int control_fd = -1;
  gdam_debug_register_names (debug_flags);
  if (argv[1] == NULL)
    usage (argv[0]);
  if (argv[1] != NULL)
    {
      if (strcmp (argv[1], "-v") == 0 || strcmp (argv[1], "--version") == 0)
	{
	  printf ("gdam-server, version %s.\n"
		  "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
		  "\n"
		  "THIS PROGRAM COMES WITH NO WARRANTY! "
		  "USE AT YOUR OWN RISK.\n", VERSION);
	  exit (0);
	}
      if (strcmp (argv[1], "-h") == 0 || strcmp (argv[1], "--help") == 0)
	{
	  usage (argv[0]);
	}
    }
  signal (SIGPIPE, SIG_IGN);
  gdam_init (&argc, &argv);
  g_set_prgname (argv[0]);
  gdam_cleanup_init ();
  for (i = 1; i < argc; i++)
    {
      if (strcmp (argv[i], "--help") == 0
	  || strcmp (argv[i], "-help") == 0 || strcmp (argv[i], "-h") == 0)
	{
	  usage (argv[0]);
	}
      if (strcmp (argv[i], "--port") == 0)
	{
	  const char *port_str = argv[++i];
	  int port_no;
	  if (!port_str)
	    {
	      fprintf (stderr, "need argument after " "--port.\n");
	      usage (argv[0]);
	    }
	  port_no = atoi (port_str);
	  if (!make_server_tcp (port_no))
	    exit (1);
	  made_server = TRUE;
	  continue;
	}
      if (strcmp (argv[i], "--unix") == 0)
	{
	  const char *unix_path = argv[++i];
	  if (unix_path == NULL)
	    {
	      fprintf (stderr, "need argument after " "--unix.\n");
	      usage (argv[0]);
	    }
	  if (!make_server_unix (unix_path))
	    exit (1);
	  made_server = TRUE;
	  continue;
	}
      if (strcmp (argv[i], "--dont-bind") == 0)
	{
	  dont_make_server = TRUE;
	  continue;
	}
      if (gdam_debug_turn_on_debug_flag (argv[i]))
	continue;
      if (strcmp (argv[i], "--control-fd") == 0)
	{
	  const char *control_fd_str = argv[++i];
	  if (control_fd != -1)
	    g_error ("only one --control-fd may be " "specified");
	  control_fd = atoi (control_fd_str);
	  if (control_fd < 0)
	    g_error ("control fd must be nonnegative");
	  continue;
	}
      if (strcmp (argv[i], "--no-realtime") == 0)
	{
	  try_realtime = FALSE;
	  continue;
	}
      fprintf (stderr, "unexpected option: %s\n", argv[i]);
      usage (argv[0]);
    }
  if (control_fd >= 0)
    {
      GdamConnection *connection;
      connection = gdam_connection_new (control_fd,
					gdam_authenticator_root ());
      gdam_connection_attach_to_main_loop (connection);
      if (!made_server)
	{
	  /* HACK: if there is no other server,
	     then probably we should exit
	     when this connection terminates.

	     I guess setting made_server to TRUE
	     is a hack as well. */
	  connection->exit_upon_close = 1;
	}
    }
  if (!made_server && !dont_make_server)
    {
      fprintf (stderr, "no port specified, using default %d.\n",
	       DEFAULT_PORT);
      if (!make_server_tcp (DEFAULT_PORT))
	{
	  fprintf (stderr, "no server created... aborting...\n");
	  exit (1);
	}
    }
  if (try_realtime)
    gdam_realtime_set ();

  gdam_main ();

  return 0;
}
