/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasecho.h"
#include "gdamdispatch.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamalloca.h>
#include <stdlib.h>
#include "gdamcachebuffer.h"

#define ECHO_BUFFER_GRANULARITY		16384

static int         gdas_echo_read               (GdasSource*	  echo,
                                       	         GdamSample*	  buf,
				       	         int		  n);
/* TODO:  fix these!!!
static void        gdas_echo_miss_data          (GdasSource*      input,
	                                         int              lost);
static gboolean    gdas_echo_is_ready_to_read   (GdasSource*      source);
*/
static void        gdas_echo_class_init         (GdasEchoClass*   klass);
static void        gdas_echo_init               (GdasEcho*        echo);
static void        gdas_echo_destroy            (GtkObject*       object);
static void 	   gdas_echo_finalize 		(GtkObject* 	  object);

static GtkObjectClass* parent_class;

/* an hour is the max echo we allow--
   even that is basically insanely large... */
#define MAX_ECHO_DELAY		44100 * 3600

GtkType gdas_echo_get_type()
{
  static GtkType echo_type = 0;
  if (!echo_type)
    {
      static const GtkTypeInfo echo_info = {
	"GdasEcho",
	sizeof (GdasEcho),
	sizeof (GdasEchoClass),
	(GtkClassInitFunc) gdas_echo_class_init,
	(GtkObjectInitFunc) gdas_echo_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_FILTER);
      echo_type = gtk_type_unique (GDAS_TYPE_FILTER, &echo_info);
    }
  return echo_type;
}

#define MAX_DELAY_SAMPLES		(1000000.0)

static void gdas_echo_arg_set_echoes(GtkObject* object, GdamArg* arg)
{
  int num_echoes = GDAM_ARG_DOUBLE_ARRAY (arg).length / 2;
  GdasEchoElement *elements = g_new (GdasEchoElement, num_echoes * 2);
  gdouble *array = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  int i;
  int out_index = 0;
  for (i = 0; i < num_echoes; i++)
    {
      gdouble delay = array[2 * i + 1];
      if (delay < 0.0 || delay > MAX_DELAY_SAMPLES)
	{
	  gdam_debug ("echo %d was out of bounds... skipping", i);
	  continue;
	}
      elements[out_index].volume = array[2 * i + 0] * GdamFixed_ONE;
      elements[out_index].delay = delay;
      out_index++;
    }
  gdas_echo_set_echoes (GDAS_ECHO (object), out_index, elements);
}

static void gdas_echo_arg_get_echoes(GtkObject* object, GdamArg* arg)
{
  GdasEcho *echo = GDAS_ECHO (object);
  int num_echoes = echo->num_echoes;
  gdouble *array = g_new (gdouble, num_echoes * 2);
  int i;
  for (i = 0; i < num_echoes; i++)
    {
      array[2 * i + 0] = (gdouble) echo->echoes[i].volume / GdamFixed_ONE;
      array[2 * i + 1] = echo->echoes[i].delay;
    }
  GDAM_ARG_DOUBLE_ARRAY (arg).length = num_echoes * 2;
  GDAM_ARG_DOUBLE_ARRAY (arg).array = array;
}
static void gdas_echo_arg_set_reverb(GtkObject* object, GdamArg* arg)
{
  GDAS_ECHO (object)->reverb = GDAM_ARG_INT (arg);
}
	
static void gdas_echo_arg_get_reverb(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_ECHO (object)->reverb;
}
static void gdas_echo_arg_set_passthrough_volume(GtkObject* object,
                                                 GdamArg* arg)
{
  int vol = (int) (GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE);
  GDAS_ECHO (object)->passthrough_volume = vol;
}
	
static void gdas_echo_arg_get_passthrough_volume(GtkObject* object,
                                                 GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = (gdouble) GDAS_ECHO (object)->passthrough_volume
    / GdamFixed_ONE;
}

static void gdas_echo_arg_set_wet_volume(GtkObject* object,
                                         GdamArg* arg)
{
  int vol = (int) (GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE);
  GDAS_ECHO (object)->wet_volume = vol;
}
	
static void gdas_echo_arg_get_wet_volume(GtkObject* object,
                                         GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = (gdouble) GDAS_ECHO (object)->wet_volume
    / GdamFixed_ONE;
}

static void gdas_echo_arg_make_buffer(GtkObject* object, GdamArg* arg)
{
  GdamCacheBuffer *cache_buffer;
  GdamSample *buffer;
  GdasEcho *echo = GDAS_ECHO (object);
  int start_offset = echo->past_buf_next_output - echo->max_delay;

  g_assert (echo->past_buf_length >= echo->max_delay);

  buffer = g_new (GdamSample, echo->max_delay);
  if (start_offset < 0)
    {
      gdam_sample_copy (buffer,
			echo->past_buf + echo->past_buf_length
			+ start_offset, -start_offset);
      gdam_sample_copy (buffer - start_offset,
			echo->past_buf, echo->max_delay + start_offset);
    }
  else
    {
      gdam_sample_copy (buffer,
			echo->past_buf + start_offset, echo->max_delay);
    }
  cache_buffer = gdam_cache_buffer_memory (echo->max_delay, buffer, g_free,
                                           FALSE);
  {
    GdamConnection *connection = gdam_connection_get_top ();
    gdam_connection_ref (connection, GTK_OBJECT (cache_buffer));
  }
  GDAM_ARG_INT (arg) = gdam_cache_buffer_get_id (cache_buffer);
}

static void gdas_echo_class_init(GdasEchoClass* klass)
{
  GdasSourceClass *source_class = (GdasSourceClass *) klass;
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gdas_echo_destroy;
  object_class->finalize = gdas_echo_finalize;
  source_class->read = gdas_echo_read;
  klass->add_echo = gdam_sample_add_scaled;
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "echoes",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     gdas_echo_arg_set_echoes,
		     gdas_echo_arg_get_echoes, NULL, ",", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "reverb",
		     GDAM_ARG_TYPE_INT,
		     gdas_echo_arg_set_reverb,
		     gdas_echo_arg_get_reverb,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "passthrough_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_echo_arg_set_passthrough_volume,
		     gdas_echo_arg_get_passthrough_volume,
		     NULL, "0,100", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "wet_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_echo_arg_set_wet_volume,
		     gdas_echo_arg_get_wet_volume,
		     NULL, "0,100", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "make_buffer",
		     GDAM_ARG_TYPE_INT,
		     NULL,
		     gdas_echo_arg_make_buffer, NULL, NULL, GDAM_ARG_TRIGGER);
}

static void gdas_echo_init(GdasEcho* echo)
{
  echo->num_echoes = 1;
  echo->echoes = g_new (GdasEchoElement, 1);
  echo->echoes->delay = 0;
  echo->echoes->volume = GdamFixed_ONE;
  echo->wet_volume = GdamFixed_ONE;
  echo->past_buf_length = 0;
  echo->past_buf_next_output = 0;
  echo->past_buf = NULL;
  echo->fade_window = 64;
  echo->reverb = FALSE;
}

static void gdas_echo_do_echo(GdasEcho*        echo,
			      GdasEchoElement* echo_array,
			      int              num_echoes,
                              GdamSample*      recent_past,
			      GdamSample*      output,
			      int              num_in_out)
{
  int which_echo;
  void (*adder) (GdamSample *, const GdamSample *, int, GdamFixed);
  GdasEchoClass *echo_class;
  echo_class = (GdasEchoClass *) (((GtkObject *) echo)->klass);
  adder = echo_class->add_echo;

  /* For each echo, add the revolumized
   * delayed samples.  */
  for (which_echo = 0; which_echo < num_echoes; which_echo++)
    {
      /*
       * The data comes from three contiguous blocks,
       *       index - delay >= 0            : from recent past
       *       index - delay >= -past_buf_next_output
       *                                     : from first contiguous
       *                                       part of buffer.
       *       otherwise                     : second contiguous
       *                                       part of buffer.
       */

      GdamFixed scale = echo_array[which_echo].volume;

      /* First index in output where souce is derived from
       * recent_past. */
      int index_recent_past = echo_array[which_echo].delay;

      /* First index in output where source is derived from
       * echo->past_buf's first portion (it comes in two
       * chunks, being a circular buffer.). */
      int index_past_start = index_recent_past - echo->past_buf_next_output;
      if (index_past_start > 0)
	{
	  int copy_len = index_past_start;
	  if (copy_len > num_in_out)
	    copy_len = num_in_out;
	  (*adder) (output,
		    echo->past_buf + echo->past_buf_length
		    - index_past_start, copy_len, scale);
	}

      if (index_recent_past > 0 && index_past_start < num_in_out)
	{
	  int dst_index = index_past_start;
	  int src_index = 0;
	  int copy_len;
	  if (dst_index < 0)
	    {
	      src_index = -dst_index;
	      dst_index = 0;
	    }
	  copy_len = echo->past_buf_next_output - src_index;
	  if (copy_len + dst_index > num_in_out)
	    copy_len = num_in_out - dst_index;
	  g_assert (copy_len >= 0);
	  (*adder) (output + dst_index, echo->past_buf + src_index,
		    copy_len, scale);
	}

      if (index_recent_past < num_in_out)
	{
	  (*adder) (output + index_recent_past, recent_past,
		    num_in_out - index_recent_past, scale);
	}
    }
}
static void do_fade(GdamSample*        fading_in,
                    const GdamSample*  fading_out,
		    int                fade_index,
		    int                length,
		    int                fade_length)
{
  guint32 vol_b15, delta_vol_b15;
  if (fade_index < 0)
    {
      length += fade_index;
      fade_index = 0;
    }
  if (fade_index + length > fade_length)
    {
      int cut = fade_index + length - fade_length;
      length -= cut;
    }
  if (length < 0)
    return;
  vol_b15 = (guint32) (((((guint64) fade_index) << 15)) / fade_length);
  delta_vol_b15 = (guint32) (((1LL << 15)) / fade_length);

  while (length-- > 0)
    {
      GdamSampleProduct in, out;
      in = fading_in->s[0];
      in *= vol_b15;
      out = fading_out->s[0];
      out *= (1 << 15) - vol_b15;
      fading_in->s[0] = (in + out) >> 15;

      in = fading_in->s[1];
      in *= vol_b15;
      out = fading_out->s[1];
      out *= (1 << 15) - vol_b15;
      fading_in->s[1] = (in + out) >> 15;

      fading_in++;
      fading_out++;
      vol_b15 += delta_vol_b15;
    }
}


static int gdas_echo_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasEcho *echo = GDAS_ECHO (source);
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  int num_read;
  GdamSample *recent_past;

  g_return_val_if_fail (echo != NULL, 0);
  filtered = filter->filtered;
  if (filtered == NULL)
    return 0;

  if (echo->recent_past_buf_size < n * sizeof (GdamSample))
    {
      if (echo->recent_past_buf)
	free (echo->recent_past_buf);
      echo->recent_past_buf = (GdamSample *) malloc (n * sizeof (GdamSample));
      echo->recent_past_buf_size = n * sizeof (GdamSample);
    }
  recent_past = echo->recent_past_buf;

  num_read = gdas_source_read (filtered, recent_past, n);

  gdam_sample_zero (buf, num_read);
  gdas_echo_do_echo (echo, echo->echoes, echo->num_echoes,
		     recent_past, buf, num_read);
  if (echo->fade_time > 0 && echo->old_echoes != NULL)
    {
      int fade_time = MIN (echo->fade_time, num_read);
      GdamSample *tmpbuf;
      tmpbuf = alloca (sizeof (GdamSample) * num_read);
      gdam_sample_zero (tmpbuf, num_read);
      gdas_echo_do_echo (echo,
			 echo->old_echoes,
			 echo->old_num_echoes, recent_past, tmpbuf, num_read);

      do_fade (buf,
	       tmpbuf,
	       echo->fade_window - echo->fade_time,
	       fade_time, echo->fade_window);

      if ((echo->fade_time -= fade_time) <= 0)
	{
	  g_assert (echo->fade_time == 0);
	  g_free (echo->old_echoes);
	  echo->old_echoes = 0;
	  echo->old_num_echoes = 0;
	}
    }

  /* Now copy the recent past to past... */
  {
    int p1 = echo->past_buf_length - echo->past_buf_next_output;
    GdamSample *start_out;
    GdamSample *src_start = echo->reverb ? buf : recent_past;
    start_out = echo->past_buf + echo->past_buf_next_output;

    if (p1 > num_read)
      {
	gdam_sample_copy (start_out, src_start, num_read);
	echo->past_buf_next_output += num_read;
      }
    else if (p1 == num_read)
      {
	gdam_sample_copy (start_out, src_start, p1);
	echo->past_buf_next_output = 0;
      }
    else if (echo->past_buf_length > num_read)
      {
	gdam_sample_copy (start_out, src_start, p1);
	gdam_sample_copy (echo->past_buf, src_start + p1, num_read - p1);
	echo->past_buf_next_output = num_read - p1;
      }
    else
      {
	gdam_sample_copy (echo->past_buf,
			  src_start + num_read
			  - echo->past_buf_length, echo->past_buf_length);
	echo->past_buf_next_output = 0;
      }
  }
  if (echo->wet_volume != GdamFixed_ONE)
    {
      gdam_sample_scale (buf, num_read, echo->wet_volume);
    }
  if (echo->passthrough_volume != 0)
    {
      gdam_sample_add_scaled (buf, recent_past, num_read,
			      echo->passthrough_volume);
    }
  return num_read;
}
		

void gdas_echo_set_echoes(GdasEcho* echo,
                          int       num_echoes,
			  GdasEchoElement* elements)
{
  int max_delay = 0;
  int i;
  GdamSample *new_past;
  for (i = 0; i < num_echoes; i++)
    if (elements[i].delay > max_delay)
      max_delay = elements[i].delay;
  if (echo->fade_window > 0)
    for (i = 0; i < echo->num_echoes; i++)
      if (echo->echoes[i].delay > max_delay)
	max_delay = echo->echoes[i].delay;
  echo->max_delay = max_delay;
  /* The following rounding makes tons of small changes
   * faster... */
#if defined(ECHO_BUFFER_GRANULARITY) && ECHO_BUFFER_GRANULARITY > 1
  max_delay += ECHO_BUFFER_GRANULARITY - 1;
  max_delay /= ECHO_BUFFER_GRANULARITY;
  max_delay *= ECHO_BUFFER_GRANULARITY;
#endif

  if (echo->fade_window > 0)
    {
      g_free (echo->old_echoes);
      echo->old_echoes = echo->echoes;
      echo->old_num_echoes = echo->num_echoes;
    }
  else
    {
      g_free (echo->echoes);
      echo->echoes = echo->old_echoes = NULL;
    }
  if (echo->old_echoes != NULL)
    echo->fade_time = echo->fade_window;
  echo->num_echoes = num_echoes;
  echo->echoes = elements;

  if (max_delay == echo->past_buf_length)
    {
      return;
    }
  /* TODO:  check this and tidy this!!!!!!!!!!!** */
  new_past = g_new (GdamSample, max_delay);
  if (echo->past_buf)
    {
      /* Length of older piece. */
      int p1 = echo->past_buf_length - echo->past_buf_next_output;
      /* Length of younger piece. */
      int p2 = echo->past_buf_next_output;
      GdamSample *new_past_end = new_past + max_delay;
      if (p2 >= max_delay)
	{
	  gdam_sample_copy (new_past,
			    echo->past_buf + (p2 - max_delay), max_delay);
	}
      else if (max_delay <= echo->past_buf_length)
	{
	  int from_p1 = (max_delay - p2);
	  /* Buffer getting shorter, but still longer
	   * than p2. */
	  gdam_sample_copy (new_past_end - p2, echo->past_buf, p2);
	  gdam_sample_copy (new_past,
			    echo->past_buf + p2 + (p1 - from_p1), from_p1);
	}
      else
	{
	  gdam_sample_copy (new_past_end - p2, echo->past_buf, p2);
	  gdam_sample_copy (new_past_end - p2 - p1, echo->past_buf + p2, p1);
	  gdam_sample_zero (new_past, max_delay - echo->past_buf_length);
	}
      g_free (echo->past_buf);
    }
  else
    {
      gdam_sample_zero (new_past, max_delay);
    }

  echo->past_buf = new_past;
  if (max_delay > echo->past_buf_length)
    echo->past_buf_next_output = echo->past_buf_length;
  else
    echo->past_buf_next_output = 0;
  echo->past_buf_length = max_delay;
}

static void gdas_echo_destroy(GtkObject* object)
{
  GdasEcho *echo = GDAS_ECHO (object);
  g_free (echo->echoes);
  g_free (echo->old_echoes);
  echo->echoes = NULL;
  echo->old_echoes = NULL;
  if (echo->recent_past_buf)
    g_free (echo->recent_past_buf);
  echo->recent_past_buf = NULL;
  echo->recent_past_buf_size = 0;
  (*parent_class->destroy) (object);
}

static void gdas_echo_finalize (GtkObject *object)
{
  GdasEcho *echo = GDAS_ECHO (object);
  g_free (echo->past_buf);
  (*parent_class->finalize) (object);
}
