/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasfilter.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"


static gboolean    gdas_filter_is_ready_to_read (GdasSource*      source);
static int         gdas_filter_read             (GdasSource*	  filter,
                                       	         GdamSample*	  buf,
				       	         int		  n);
static gboolean    gdas_filter_seek             (GdasSource*      source,
                                                 GdamTime         offset,
		                                 int              whence);
static GdamTime    gdas_filter_get_position     (GdasSource*      source);

static void        gdas_filter_miss_data        (GdasSource*      input,
					         int              lost);

static gboolean    gdas_filter_detach_child     (GdasSource*      source,
		                                 GdasSource*      sub);
static void        gdas_filter_class_init       (GdasFilterClass* klass);
static void        gdas_filter_init             (GdasFilter*      filter);
static void        gdas_filter_destroy          (GdasFilter*      filter);
static void        gdas_filter_get_subsources   (GdasSource*      source,
                                                 int*             num_subs,
						 GdasSource***    subs);
static GtkObjectClass* parent_class;

GtkType gdas_filter_get_type()
{
  static GtkType filter_type = 0;
  if (!filter_type)
    {
      static const GtkTypeInfo filter_info = {
	"GdasFilter",
	sizeof (GdasFilter),
	sizeof (GdasFilterClass),
	(GtkClassInitFunc) gdas_filter_class_init,
	(GtkObjectInitFunc) gdas_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      filter_type = gtk_type_unique (GDAS_TYPE_SOURCE, &filter_info);
    }
  return filter_type;
}


static void gdas_filter_arg_set_filtered(GtkObject* object, GdamArg* arg)
{
  int source_id = GDAM_ARG_INT (arg);
  GdasSource *sub;
  if (source_id == 0)
    sub = NULL;
  else
    sub = gdam_connection_get_source (NULL, source_id);
  gdam_debug ("set_filtered: filter=%d, child=%d [source_id=%d]",
	      GDAM_SERVER_OBJECT (object)->global_id,
	      sub ? GDAM_SERVER_OBJECT (sub)->global_id : 0, source_id);
  gdas_filter_set_source (GDAS_FILTER (object), sub);
  if (sub != NULL)
    gtk_object_unref ((GtkObject *) sub);
}
static void gdas_filter_arg_get_filtered(GtkObject* object, GdamArg* arg)
{
  GdasSource *filtered = GDAS_FILTER (object)->filtered;
  GDAM_ARG_INT (arg) = (filtered != NULL)
    ? GDAM_SERVER_OBJECT (filtered)->global_id : 0;
}

static void gdas_filter_class_init(GdasFilterClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  GdasSourceClass *source_class = (GdasSourceClass *) klass;
  klass->attach_filtered = NULL;
  source_class->read = gdas_filter_read;
  source_class->seek = gdas_filter_seek;
  source_class->get_position = gdas_filter_get_position;
  source_class->detach_child = gdas_filter_detach_child;
  source_class->miss_data = gdas_filter_miss_data;
  source_class->is_ready_to_read = gdas_filter_is_ready_to_read;
  source_class->get_subsources = gdas_filter_get_subsources;
  object_class->destroy = (void (*)(GtkObject *)) gdas_filter_destroy;
  gdam_arg_register (object_class,
		     "filtered",
		     GDAM_ARG_TYPE_INT,
		     gdas_filter_arg_set_filtered,
		     gdas_filter_arg_get_filtered,
		     NULL, NULL, GDAM_ARG_VITAL);
}

static void gdas_filter_init(GdasFilter* filter)
{
  filter->filtered = NULL;
}

void gdas_filter_set_source(GdasFilter* filter, GdasSource* sub)
{
  GtkObjectClass *object_class = ((GtkObject *) filter)->klass;
  GdasFilterClass *filter_class = (GdasFilterClass *) object_class;
  g_return_if_fail ((GdasSource *) filter != sub);

  {
    int filt_id = GDAM_SERVER_OBJECT (filter)->global_id;
    int sub_id = sub ? GDAM_SERVER_OBJECT (sub)->global_id : 0;
    gdam_debug ("gdas_filter_set_source: filter=%d, source=%d",
		filt_id, sub_id);}

  /* sub->ref_count ++ */
  if (sub)
    gtk_object_ref (GTK_OBJECT (sub));

  if (filter->filtered)
    gdas_source_detach_child (GDAS_SOURCE (filter), filter->filtered);
  if (!sub)
    return;

  /* sub->ref_count ++ */
  if (filter_class->attach_filtered)
    filter_class->attach_filtered (filter, sub);
  else
    gdas_source_orphan (sub);	/* increases sub's ref cnt */
  if (gdas_source_is_ancestor (GDAS_SOURCE (filter), sub))
    g_warning ("%s is a parent of %s; err...",
	       gtk_type_name (GTK_OBJECT (filter)->klass->type),
	       gtk_type_name (GTK_OBJECT (sub)->klass->type));

  g_return_if_fail (!gdas_source_is_ancestor (GDAS_SOURCE (filter), sub));
  gtk_object_sink (GTK_OBJECT (sub));
  filter->filtered = sub;
  sub->parent = GDAS_SOURCE (filter);

  /* sub->ref_count -- */
  gtk_object_unref (GTK_OBJECT (sub));
}

GdasSource* gdas_filter_detach_source(GdasFilter* filter)
{
  GtkObjectClass *object_class = ((GtkObject *) filter)->klass;
  GdasSourceClass *source_class = (GdasSourceClass *) object_class;
  GdasSource *sub = filter->filtered;
  if (source_class->detach_child)
    {
      gtk_object_ref (GTK_OBJECT (sub));
      source_class->detach_child (GDAS_SOURCE (filter), sub);
    }
  filter->filtered = NULL;
  return sub;
}

void gdas_filter_kill_source(GdasFilter* filter)
{
  GtkObjectClass *object_class = ((GtkObject *) filter)->klass;
  GdasSourceClass *source_class = (GdasSourceClass *) object_class;
  GdasSource *sub = filter->filtered;
  //filter->filtered = NULL;
  source_class->detach_child (GDAS_SOURCE (filter), sub);
}

static gboolean gdas_filter_is_ready_to_read(GdasSource* source)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  g_return_val_if_fail (filter != NULL, FALSE);
  filtered = filter->filtered;
  return filtered ? gdas_source_is_ready_to_read (filtered) : TRUE;
}
static int gdas_filter_read(GdasSource*	source, GdamSample* buf, int n)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  g_return_val_if_fail (filter != NULL, 0);
  filtered = filter->filtered;
  if (filtered == NULL)
    return 0;
  return gdas_source_read (filtered, buf, n);
}
static gboolean gdas_filter_seek(GdasSource*      source,
                                 GdamTime         offset,
		                 int              whence)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  g_return_val_if_fail (filter != NULL, FALSE);
  filtered = filter->filtered;
  if (filtered == NULL)
    {
      g_warning ("filter-seek on filter without subsource");
      return FALSE;
    }
  return gdas_source_seek (filtered, offset, whence);
}

static GdamTime gdas_filter_get_position(GdasSource*      source)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  g_return_val_if_fail (filter != NULL, FALSE);
  filtered = filter->filtered;
  if (filtered == NULL)
    {
      g_warning ("filter-getposition on filter without subsource");
      return FALSE;
    }

  /* Check for circular dependency. */
  g_return_val_if_fail (filtered != source, -1);
  return gdas_source_get_position (filtered);
}
static void gdas_filter_miss_data(GdasSource* source, int lost)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered;
  g_return_if_fail (filter != NULL);
  filtered = filter->filtered;
  if (!filtered)
    return;
  gdas_source_miss_data (filtered, lost);
}

static gboolean gdas_filter_detach_child(GdasSource* source, GdasSource* sub)
{
  GdasFilter *filter = GDAS_FILTER (source);
  GdasSource *filtered = filter->filtered;
  g_return_val_if_fail (filter != NULL, FALSE);

  if (filter->filtered == NULL)
    return FALSE;

  g_return_val_if_fail (filter->filtered == sub, FALSE);

  filter->filtered = NULL;
  filtered->parent = NULL;
  gtk_object_unref (GTK_OBJECT (filtered));

  return TRUE;
}
static void        gdas_filter_get_subsources   (GdasSource*      source,
                                                 int*             num_subs,
						 GdasSource***    subs)
{
  GdasFilter *filter = GDAS_FILTER (source);
  g_return_if_fail (filter != NULL);
  if (filter->filtered)
    {
      *num_subs = 1;
      *subs = g_new (GdasSource *, 1);
      (*subs)[0] = filter->filtered;
    }
  else
    {
      *num_subs = 0;
      *subs = NULL;
    }
}

static void gdas_filter_destroy(GdasFilter* filter)
{
  if (filter->filtered)
    gdas_filter_kill_source (filter);
  parent_class->destroy ((GtkObject *) filter);
}

int         gdas_filter_sub_read       (GdasFilter* filter, 
                                        GdamSample* buf,
					int         n)
{
  if (filter->filtered == NULL)
    return 0;
  return gdas_source_read (filter->filtered, buf, n);
}
