/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasmix.h"
#include "gdamnamespace.h"
#include <stdlib.h>
#include <gdam/common/gdam-protocol.h>
#include "gdamdispatch.h"

#if 1
#define debug_assert(x) g_assert(x)
#else
#define debug_assert(x) 
#endif


static void gdas_mix_element_free(GdasMixElement* element);
static GdasMixElement* gdas_mix_element_new(GdasSource* source, GdamTime time);
static GtkObjectClass* parent_class  = NULL;
static GMemChunk*      element_chunk = NULL;

static void gdas_mix_init(GdasMix* mix)
{
  mix->mode = GDAS_MIX_ETERNAL;
}

static gint compare_mix_elements(gconstpointer a, gconstpointer b)
{
  const GdasMixElement *element_a = a;
  const GdasMixElement *element_b = b;
  const GdasMix *mix = element_a->mix;
  gint64 position = mix->position;
  gint64 a_start = element_a->start_offset;
  gint64 b_start = element_b->start_offset;
  if (a_start < position && b_start >= position)
    return (+1);
  if (b_start < position && a_start >= position)
    return (-1);
  return (a_start < b_start) ? (-1) : (a_start > b_start) ? (+1) : 0;
}

/* Called to move an element to either the waiting list or
 * the ended list. */
static void source_ended(GSList** plist)
{
  GSList *list = *plist;
  GdasMixElement *element = list->data;
  GdasMix *mix = element->mix;
  GSList **dest;

  gdam_debug ("mix: %s ended",
	      gtk_type_name (GTK_OBJECT (element->source)->klass->type));

  *plist = g_slist_remove (*plist, element);
  if (element->remove_when_ended)
    {
      GdamSample tmp[20];
      if (gdas_source_read (element->source, tmp, 20) > 0)
	g_warning ("bogus eof event in mixer");
      gdas_mix_element_free (element);
      return;
    }
  dest = element->ended ? &mix->ended : &mix->waiting;
  *dest = g_slist_insert_sorted (*dest, element, compare_mix_elements);
}

/* Assumes: it just needs to read without regard to
 *          mixer-based end conditions (ie looping-position wraps)
 *
 *          Returning less than the number of samples requested
 *          means the mix has ended!!!
 */
static int
gdas_mix_read_one(GdasMix*        mix,
                   GdamSample*      buffer,
		   int              num_samples)
{
  GSList **plist;
  GdamSample *tmp_buffer;
  GdasMixClass *mix_class = GDAS_MIX_CLASS (GTK_OBJECT (mix)->klass);
  GdasMixSampleFunc init_func = mix_class->init_samples;
  GdasMixSampleFunc add_func = mix_class->add_samples;
  int length_init = 0;
  gint64 mix_position = mix->position;
  gint64 end_position = mix_position + num_samples;
  int min_end_time = num_samples;
  int max_end_time = 0;
  int rv;

  if (mix->tmp_buffer_size < sizeof (GdamSample) * num_samples)
    {
      if (mix->tmp_buffer)
	free (mix->tmp_buffer);
      mix->tmp_buffer = malloc (sizeof (GdamSample) * num_samples);
      mix->tmp_buffer_size = sizeof (GdamSample) * num_samples;
    }
  tmp_buffer = mix->tmp_buffer;

/* Copy NUM_READ samples from tmp_buffer into
 * buffer + OFFSET, initializing or adding as needed.
 *
 * Initialize `buffer' out of `tmp_buffer' as follows:
 *
 *  If length_init < offset,
 *
 *     0      length_init          offset      (offset+num_read)
 *     +-----------+------------------+----------------+-----------....
 *     |           |     zero_out     |   init_func    |
 *     +-----------+------------------+----------------+-----------....
 *                                    |tmp_buffer
 * If length_init >= offset,
 *     0       offset            length_init   (offset+num_read)
 *     +-----------+------------------+----------------+-----------....
 *     |           |    add_func      |   init_func    |
 *     +-----------+------------------+----------------+-----------....
 *                 |tmp_buffer                         ^
 *                                               new length_init
 */
#define AUGMENT_BUFFER(num_read, offset)                                  \
	{								  \
        if (length_init < (offset)) {                                     \
		gdam_sample_zero(buffer + length_init,                    \
		                 (offset) - length_init);                 \
                (*init_func)(mix,                                         \
                             buffer + (offset),                           \
                             tmp_buffer,                                  \
                             num_read);                                   \
        } else {                                                          \
                int tmp;                                                  \
                if (length_init > (offset)) {                             \
                        (*add_func)(mix,                                  \
                                    buffer + (offset),                    \
                                    tmp_buffer,                           \
                                    length_init - (offset));              \
                }                                                         \
                tmp = (offset) + (num_read) - length_init;                \
                if (tmp > 0) {                                            \
                        (*init_func)(mix,                                 \
                                     buffer + length_init,                \
                                     tmp_buffer + length_init - (offset), \
				     tmp);                                \
                }                                                         \
        }                                                                 \
	if (length_init < (offset) + num_read)				  \
		length_init = (offset) + num_read;			  \
	}

/* Does ELEMENT's start_offset occur in the samples we 
 * are going to output? */
#define STARTS_IN_RANGE(element)					    \
		(mix->position           <  (element)->start_offset 	    \
              && (element)->start_offset <= end_position)


  for (plist = &mix->playing; *plist != NULL;)
    {
      GSList *list = *plist;
      GdasMixElement *element = list->data;
      GdasSource *sub = element->source;

      debug_assert (element->position == mix->position);

      /* Check to see if we are going to need to 
       * seek to the beginning over the course of
       * this read.  Because we never wrap
       * mix's position in this function,
       * we can ignore wrapping; only element restarting.
       */
      if (mix->mode == GDAS_MIX_LOOPS && STARTS_IN_RANGE (element))
	{
	  int p1 = element->start_offset - element->position;
	  int p2;
	  int r1, r2;
	  int amt_read;
	  if (p1 < 0)
	    p1 += mix->loop_length;
	  p2 = num_samples - p1;
	  if (gdas_source_is_ready_to_read (sub))
	    {
	      r1 = gdas_source_read (sub, tmp_buffer, p1);
	      if (r1 < 0)
		r1 = 0;
	      if (r1 < p1)
		gdam_sample_zero (tmp_buffer + r1, p1 - r1);
	    }
	  else
	    r1 = 0;
	  gdas_source_seek (sub, -(p2 + r1), 1);
	  if (gdas_source_is_ready_to_read (sub))
	    {
	      r2 = gdas_source_read (sub, tmp_buffer + p1, p2);
	      if (r2 < 0)
		r2 = 0;
	    }
	  else
	    r2 = 0;

	  /* XXX: initializes samples to zero,
	   *      then add_funcs subsequent.
	   *      Not acceptable for fast looping modulators.
	   *
	   *      To fix, do just one read here and
	   *      reinsert on the waiting list...
	   */
	  if (r1 < p1)
	    gdam_sample_zero (tmp_buffer + r1, p1 - r1);

	  amt_read = r2 + p1;
	  AUGMENT_BUFFER (amt_read, 0);
	  element->position = element->start_offset + r2;
	  if (amt_read > max_end_time)
	    max_end_time = amt_read;

	}
      else
	{
	  int amt_read;
	  gint64 buffer_offset = element->position - mix_position;
	  int to_read = num_samples - buffer_offset;
	  debug_assert (buffer_offset == 0);
	  debug_assert (to_read >= 0);
	  if (gdas_source_is_ready_to_read (element->source) && to_read > 0)
	    amt_read = gdas_source_read (element->source,
					 tmp_buffer, to_read);
	  else
	    amt_read = 0;
	  if (amt_read > max_end_time)
	    max_end_time = amt_read;
	  if (amt_read > 0)
	    {
	      AUGMENT_BUFFER (amt_read, (int) buffer_offset);
	      length_init = amt_read;
	    }
	  element->position += amt_read;
	}
      if (element->position < end_position)
	{
	  if (gdas_source_is_ready_to_read (sub))
	    {
	      /* The source has ended,
	       * b/c it is ready to read but it
	       * didn't return more samples. */
	      gint64 offset;
	      int amt_read;
	      /*gdam_debug ("source %d [%s] was behind but readytoread",
			  GDAM_SERVER_OBJECT (sub)->global_id,
			  gtk_type_name (GTK_OBJECT (sub)->klass->type));*/
	      amt_read = element->position - mix_position;
	      if (amt_read < min_end_time)
		min_end_time = amt_read;
	      if (mix->mode == GDAS_MIX_LOOPS)
		{
		  /* the offset is always negative */
		  offset = element->start_offset - element->position;
		  element->position += offset;
		  gdas_source_seek (sub, offset, 1);
		}
	      source_ended (plist);
	      if (mix->mode == GDAS_MIX_FIRST_ENDED)
		mix->mix_should_end = TRUE;
	      continue;
	    }
	  else
	    {
	      int lost;
	      lost = (int) (end_position - element->position);
	      gdas_source_miss_data (sub, lost);
	      element->position += lost;
	    }
	}
      plist = &((*plist)->next);
    }
  for (plist = &mix->waiting; *plist != NULL;)
    {
      GSList *list = *plist;
      GdasMixElement *element = list->data;
      GdasSource *sub = element->source;
      int offset, to_read, num_read;

      /* We get to break b/c of the sorted start times. */
      if (!STARTS_IN_RANGE (element))
	break;

      offset = (int) (element->start_offset - mix_position);
      to_read = num_samples - offset;
      num_read = gdas_source_read (sub, tmp_buffer, to_read);

      if (num_read < to_read)
	{
	  int r = num_read;
	  if (r < 0)
	    r = 0;
	  if (gdas_source_is_ready_to_read (sub))
	    {
	      /* If the element ended, rewind it. */
	      int end_time = offset + num_read;
	      gint64 offset;
	      offset = element->start_offset - element->position;	/* negative */
	      if (mix->mode == GDAS_MIX_LOOPS)
		gdas_source_seek (sub, offset, 1);
	      else
		element->ended = TRUE;
	      element->position = element->start_offset;

	      source_ended (plist);
	      if (mix->mode == GDAS_MIX_FIRST_ENDED)
		{
		  mix->mix_should_end = TRUE;
		}

	      if (end_time > max_end_time)
		max_end_time = end_time;
	      continue;
	    }
	  else
	    {
	      gdas_source_miss_data (sub, to_read - r);
	    }
	}
      element->position = end_position;

      /* Move element to playing list. */
      *plist = g_slist_remove (*plist, element);
      mix->playing = g_slist_prepend (mix->playing, element);
    }
  if (mix->mode == GDAS_MIX_FIRST_ENDED)
    rv = min_end_time;
  else if (mix->mode == GDAS_MIX_ALL_ENDED
	   && mix->waiting == NULL && mix->playing == NULL)
    rv = max_end_time;
  else
    {
      /* if mode == LOOPS or ETERNAL */
      rv = num_samples;
    }

  if (length_init < rv)
    {
      gdam_sample_zero (buffer + length_init, rv - length_init);
    }

  return rv;
}

static int gdas_mix_read(GdasSource*     source,
			 GdamSample*     samples,
                         int             max_read)
{
  GdasMix *mix = GDAS_MIX (source);
  int num_read = 0;

  g_return_val_if_fail (0 <= mix->position, 0);
  if (mix->mix_should_end)
    {
      gdam_debug ("gdas_mix_read: mix_should_end");
      return 0;
    }

  /* Loop around the position count, calling read_one until 
   * it returns eof or the buffer is full. */
  if (mix->mode == GDAS_MIX_LOOPS)
    {
      g_return_val_if_fail (mix->position < mix->loop_length, 0);
      while (num_read < max_read)
	{
	  gint64 remaining = mix->loop_length - mix->position;
	  int r;
	  if (remaining > max_read - num_read)
	    remaining = max_read - num_read;
	  r = (int) remaining;
	  r = gdas_mix_read_one (mix, samples + num_read, r);
	  g_return_val_if_fail (r == remaining, num_read + r);
	  mix->position += r;
	  if (mix->loop_length == mix->position)
	    mix->position = 0;
	  num_read += r;
	}
      return num_read;
    }
  else
    {
      int rv;
      rv = gdas_mix_read_one (mix, samples, max_read);
      if (rv < max_read)
	mix->mix_should_end = 1;
      mix->position += rv;
      return rv;
    }
}

static void gdas_mix_arg_set_mixed(GtkObject* object, GdamArg* arg)
{
  GdasMix *mix = GDAS_MIX (object);
  GdasSource **sources;
  gint64 *times;
  int num_subs = GDAM_ARG_INT64_ARRAY (arg).length / 2;
  int i;
  gint64 *array = GDAM_ARG_INT64_ARRAY (arg).array;
  sources = (GdasSource **) alloca (sizeof (GdasSource *) * num_subs);
  times = (gint64 *) alloca (sizeof (gint64) * num_subs);
  for (i = 0; i < num_subs; i++)
    {
      int source_id = array[2 * i + 1];
      sources[i] = gdam_connection_get_source (NULL, source_id);
      if (sources[i] == NULL)
	{
	  gdam_debug ("couldn't find source %d", source_id);
	  continue;
	}
      times[i] = array[2 * i];
    }
  gdas_mix_freeze (mix);
  gdas_mix_clear (mix);
  for (i = 0; i < num_subs; i++)
    {
      if (sources[i] != NULL)
	{
	  gdas_mix_add (mix, sources[i], times[i]);
	  gtk_object_unref (GTK_OBJECT (sources[i]));
	}
    }
  gdas_mix_thaw (mix);
}

static gboolean    gdas_mix_seek             (GdasSource*       source,
	                                       GdamTime          offset,
					       int               whence)
{
  GdasMix *mix = GDAS_MIX (source);
  GSList *elements;
  GSList *last;
  gint64 new_pos;
  new_pos = whence == 0 ? offset : offset + mix->position;
  if (new_pos == mix->position)
    return TRUE;

  if (mix->mode == GDAS_MIX_LOOPS)
    mix->position %= mix->loop_length;

  elements = mix->playing;
  last = g_slist_last (elements);
  if (last)
    last->next = mix->waiting;
  else
    elements = last = mix->waiting;
  last = g_slist_last (last);
  if (last)
    last->next = mix->ended;
  else
    elements = last = mix->ended;
  mix->playing = mix->waiting = mix->ended = NULL;
  mix->position = new_pos;

  /* Distribute the sources by seeking them,
   * then added them to the right list. */
  while (elements)
    {
      GSList *next = elements->next;
      GdasMixElement *element = elements->data;
      GSList **plist;

      /* Figure out which list to add it to. */
      if (element->start_offset <= new_pos)
	{
	  gint64 subpos = new_pos - element->start_offset;
	  if (!gdas_source_seek (element->source, subpos, 0))
	    {
	      if (mix->mode == GDAS_MIX_LOOPS)
		plist = &mix->waiting;
	      else
		plist = &mix->ended;
	    }
	  else
	    {
	      plist = &mix->playing;
	      element->position = new_pos;
	    }
	}
      else
	{
	  plist = &mix->waiting;
	}

      element->ended = (plist == &mix->ended);

      /* Add to that list. */
      elements->next = *plist;
      *plist = elements;
      elements = next;
    }

  /* Sort the waiting list. */
  mix->waiting = g_slist_sort (mix->waiting, compare_mix_elements);
  return TRUE;
}
static gboolean    gdas_mix_is_ready_to_read (GdasSource*       source)
{
  return TRUE;
}
static GdamBuffer* gdas_mix_interpret_command(GdasSource*       source,
					       GdamBuffer*       command,
					       gboolean*         was_ok)
{
  GdasSourceClass *par_source_class = (GdasSourceClass *) parent_class;
  int type;
  if (gdam_buffer_peek_int (command, &type))
    {
      if (type == GDAM_CFG_MIX_ADD)
	{
	  GdasMix *mix;
	  GdasSource *subsource;
	  int source_id;
	  GdamTime when;
	  GdasMixElement *element;
	  mix = GDAS_MIX (source);
	  gdam_buffer_parse_int (command, &type);
	  if (!gdam_buffer_parse_int (command, &source_id))
	    {
	      *was_ok = FALSE;
	      gdam_debug ("mix_add: no source_id ");
	      return NULL;
	    }
	  subsource = gdam_connection_get_source (NULL, source_id);
	  if (subsource == NULL)
	    {
	      *was_ok = FALSE;
	      gdam_debug ("mix_add: bad source id %d", source_id);
	      return NULL;
	    }

	  if (!gdam_buffer_parse_time (command, &when))
	    {
	      *was_ok = FALSE;
	      gdam_debug ("mix_add: no time (too short)");
	      return NULL;
	    }
	  gdam_dispatch_log ("add %d %d %lld",
			     GDAM_SERVER_OBJECT (source)->global_id,
			     source_id, when);
	  element = gdas_mix_add (mix, subsource, when);
	  element->remove_when_ended = TRUE;
	  *was_ok = TRUE;
	  gtk_object_unref (GTK_OBJECT (subsource));
	  return NULL;
	}
    }
  return par_source_class->interpret_command (source, command, was_ok);
}

static GdamTime    gdas_mix_get_position     (GdasSource*       source)
{
  return GDAS_MIX (source)->position;
}
static void        gdas_mix_get_latency_info (GdasSource*       source_mix,
                                               GdasSource*       sub,
				               GdamLatencyInfo*  info)
{
  info->latency = 0.0;
  info->resample = GdamFixed_ONE;
}
static void gdas_mix_loops_miss_data(GdasMix* mix, int lost)
{
  GSList **plist;
  for (plist = &mix->playing; *plist != NULL;)
    {
      GdasMixElement *element = (*plist)->data;
      gint64 final_offset;
      gint64 offset;
      int seek;
      offset = element->position - element->start_offset;
      final_offset = mix->position + lost - element->start_offset;
      if (offset < 0)
	offset += mix->loop_length;
      if (final_offset < 0)
	final_offset += mix->loop_length;
      seek = offset + lost - mix->loop_length;
      if (seek > 0)
	gdas_source_miss_data (element->source, seek);
      else if (seek < 0)
	{
	  if (!gdas_source_seek (element->source, seek, 1))
	    {
	      /* Move element on to waiting queue. */
	      source_ended (plist);
	      if (mix->mode == GDAS_MIX_FIRST_ENDED)
		mix->mix_should_end = TRUE;
	      continue;
	    }
	}
      element->position += lost;
      if (element->position >= mix->loop_length)
	element->position -= mix->loop_length;
      plist = &(*plist)->next;
    }
  for (plist = &mix->waiting; *plist != NULL;)
    {
      GSList *list = *plist;
      GdasMixElement *element = list->data;
      gint64 final_offset;
      gint64 offset;
      int seek;
      offset = element->position - element->start_offset;
      final_offset = mix->position + lost - element->start_offset;
      if (offset < 0)
	offset += mix->loop_length;
      if (final_offset < 0)
	final_offset += mix->loop_length;
      seek = offset + lost - mix->loop_length;
      if (seek > 0)
	gdas_source_miss_data (element->source, seek);
      else if (seek < 0)
	{
	  break;
	}
      element->position += lost;
      if (element->position >= mix->loop_length)
	element->position -= mix->loop_length;

      /* Put *plist on playing queue. */
      *plist = list->next;
      list->next = mix->playing;
      mix->playing = list;
    }
}
static void gdas_mix_nonlooping_miss_data(GdasMix* mix, int lost)
{
  GSList *list;
  gint64 end_position = mix->position + lost;
  for (list = mix->playing; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      gdas_source_miss_data (element->source, lost);
      element->position += lost;
    }
  while (mix->waiting)
    {
      GdasMixElement *element;
      gint64 to_miss;
      list = mix->waiting;
      element = list->data;
      to_miss = end_position - element->start_offset;
      if (to_miss <= 0)
	break;
      gdas_source_miss_data (element->source, to_miss);
      element->position += to_miss;
      mix->waiting = mix->waiting->next;
      list->next = mix->playing;
      mix->playing = list;
    }
}
static void        gdas_mix_miss_data        (GdasSource*       source,
					       int               lost)
{
  GdasMix *mix = GDAS_MIX (source);
  if (mix->mode == GDAS_MIX_LOOPS)
    {
      lost %= mix->loop_length;
      if (lost == 0)
	return;
      gdas_mix_loops_miss_data (mix, lost);
      mix->position += lost;
      if (mix->position > mix->loop_length)
	mix->position -= mix->loop_length;
    }
  else
    {
      gdas_mix_nonlooping_miss_data (mix, lost);
      mix->position += lost;
    }
}
static GdasMixElement*
gdas_mix_element_new(GdasSource* source, GdamTime time)
{
  GdasMixElement *element;
  GtkObject *object = (GtkObject *) source;
  gtk_object_ref (object);
  gtk_object_sink (object);
  element = g_mem_chunk_alloc0 (element_chunk);
  element->source = source;
  element->start_offset = time;
  element->position = time;
  element->remove_when_ended = TRUE;
  return element;
}


static void gdas_mix_element_free(GdasMixElement* element)
{
  element->source->parent = NULL;
  gtk_object_unref (GTK_OBJECT (element->source));
  g_mem_chunk_free (element_chunk, element);
}

/* XXX: this is really repetitive... */
static gboolean    gdas_mix_detach_child     (GdasSource*       source,
                                               GdasSource*       sub_source)
{
  GSList **plist;
  GdasMix *mix = GDAS_MIX (source);
  for (plist = &mix->playing; *plist != NULL; plist = &((*plist)->next))
    {
      GdasMixElement *element = (*plist)->data;
      if (element->source == sub_source)
	{
	  *plist = g_slist_remove (*plist, element);
	  gdas_mix_element_free (element);
	  return TRUE;
	}
    }
  for (plist = &mix->waiting; *plist != NULL; plist = &((*plist)->next))
    {
      GdasMixElement *element = (*plist)->data;
      if (element->source == sub_source)
	{
	  *plist = g_slist_remove (*plist, element);
	  gdas_mix_element_free (element);
	  return TRUE;
	}
    }
  for (plist = &mix->ended; *plist != NULL; plist = &((*plist)->next))
    {
      GdasMixElement *element = (*plist)->data;
      if (element->source == sub_source)
	{
	  *plist = g_slist_remove (*plist, element);
	  gdas_mix_element_free (element);
	  return TRUE;
	}
    }
  return FALSE;
}
static void        gdas_mix_get_subsources   (GdasSource*       source,
                                               int*              num_subs,
					       GdasSource***     subs)
{
  GdasMix *mix = GDAS_MIX (source);
  int i = 0;
  GSList *list;
  *num_subs = g_slist_length (mix->ended)
    + g_slist_length (mix->waiting) + g_slist_length (mix->playing);
  *subs = g_new (GdasSource *, *num_subs);
  for (list = mix->waiting; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      (*subs)[i++] = element->source;
    }
  for (list = mix->playing; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      (*subs)[i++] = element->source;
    }
  for (list = mix->ended; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      (*subs)[i++] = element->source;
    }
}
static void        gdas_mix_destroy          (GtkObject*        object)
{
  GdasMix *mix = GDAS_MIX (object);
  gdas_mix_clear (mix);
  if (mix->tmp_buffer)
    free (mix->tmp_buffer);
  mix->tmp_buffer = NULL;
  mix->tmp_buffer_size = 0;
  (*parent_class->destroy) (object);
}
static void        gdas_mix_init_samples      (GdasMix*          mix,
                                               GdamSample*       dest,
				               const GdamSample* source,
				               int               num_samples)
{
  gdam_sample_copy (dest, source, num_samples);
}
static void        gdas_mix_add_samples       (GdasMix*          mix,
                                               GdamSample*       dest,
				               const GdamSample* source,
				               int               num_samples)
{
  gdam_sample_addto (dest, source, num_samples);
}

static GdasMixElement*
gdas_mix_get_element_by_source(GdasMix* mix, GdasSource* source)
{
  GSList *list;
  for (list = mix->playing; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      if (element->source == source)
	return element;
    }
  for (list = mix->waiting; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      if (element->source == source)
	return element;
    }
  for (list = mix->ended; list != NULL; list = list->next)
    {
      GdasMixElement *element = list->data;
      if (element->source == source)
	return element;
    }
  return NULL;
}

static GdasMixElement*
gdas_mix_get_element_by_id(GdasMix* mix, gint source_id)
{
  GdasSource *source;
  source = gdam_connection_get_source (NULL, source_id);
  if (source == NULL)
    return NULL;
  return gdas_mix_get_element_by_source (mix, source);
}

/*
 *     _                                         _
 *    / \   _ __ __ _ _   _ _ __ ___   ___ _ __ | |_ ___
 *   / _ \ | '__/ _` | | | | '_ ` _ \ / _ \ '_ \| __/ __|
 *  / ___ \| | | (_| | |_| | | | | | |  __/ | | | |_\__ \
 * /_/   \_\_|  \__, |\__,_|_| |_| |_|\___|_| |_|\__|___/
 *              |___/
 */
static void append_timeval(gpointer element_ptr, gpointer ptr_int64array)
{
  gint64 **p_array = ptr_int64array;
  gint64 *array = *p_array;
  GdasMixElement *element = element_ptr;
  array[0] = element->start_offset;
  array[1] = GDAM_SERVER_OBJECT (element->source)->global_id;
  *p_array = array + 2;
}

static gint sort_int64(gconstpointer a_ptr, gconstpointer b_ptr)
{
  gint64 a = *(gint64 *) a_ptr;
  gint64 b = *(gint64 *) b_ptr;
  return (a < b) ? -1 : (a > b) ? +1 : 0;
}

static void gdas_mix_arg_get_mixed(GtkObject* object, GdamArg* arg)
{
  GdasMix *mix = GDAS_MIX (object);
  int num_sources;
  gint64 *rv;
  gint64 *at;
  num_sources = g_slist_length (mix->ended)
    + g_slist_length (mix->waiting) + g_slist_length (mix->playing);
  at = rv = g_new (gint64, num_sources * 2);
  g_slist_foreach (mix->ended, append_timeval, &at);
  g_slist_foreach (mix->waiting, append_timeval, &at);
  g_slist_foreach (mix->playing, append_timeval, &at);
  g_assert (at == rv + num_sources * 2);
  qsort (rv, num_sources, sizeof (gint64) * 2, sort_int64);
  GDAM_ARG_INT64_ARRAY (arg).array = rv;
  GDAM_ARG_INT64_ARRAY (arg).length = num_sources * 2;
}

static void gdas_mix_set_terminates(GtkObject* object, GdamArg* arg)
{
}
static void gdas_mix_get_terminates(GtkObject* object, GdamArg* arg)
{
}
static void gdas_mix_arg_set_source_does_miss(GtkObject* object, GdamArg* arg)
{
  GdasMixElement *element;
  int source_id = GDAM_ARG_INT (arg);
  element = gdas_mix_get_element_by_id (GDAS_MIX (object), source_id);
  if (element == NULL)
    {
      gdam_debug ("%s: no subsource of id %d", arg->name, source_id);
      return;
    }
  element->may_miss_data = TRUE;
}
static void
gdas_mix_arg_set_source_doesnt_miss(GtkObject* object, GdamArg* arg)
{
  GdasMixElement *element;
  int source_id = GDAM_ARG_INT (arg);
  element = gdas_mix_get_element_by_id (GDAS_MIX (object), source_id);
  if (element == NULL)
    {
      gdam_debug ("%s: no subsource of id %d", arg->name, source_id);
      return;
    }
  element->may_miss_data = FALSE;
}
static void
gdas_mix_arg_add_with_latency(GtkObject* object, GdamArg* arg)
{
  GdasMix *mix = GDAS_MIX (object);
  GdamTime when = mix->position;
  GdamLatencyInfo latency;
  GdasSource *sub = gdam_connection_get_source (NULL, GDAM_ARG_INT (arg));
  if (sub == NULL)
    {
      gdam_debug ("add_with_latency(%d): source didn't exist",
		  GDAM_ARG_INT (arg));
      return;
    }
  gdas_source_find_total_latency (sub, &latency);
  when += latency.latency;
  gdas_mix_add (mix, sub, when);

  gtk_object_unref ((GtkObject *) sub);
}

static void gdas_mix_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdasMix *mix = GDAS_MIX (object);
  const char *mode = GDAM_ARG_STRING (arg);
  if (strcmp (mode, "all_ended") == 0)
    {
      mix->mode = GDAS_MIX_ALL_ENDED;
    }
  else if (strcmp (mode, "eternal") == 0)
    {
      mix->mode = GDAS_MIX_ETERNAL;
    }
  else if (strcmp (mode, "first_ended") == 0)
    {
      mix->mode = GDAS_MIX_FIRST_ENDED;
    }
  else if (strcmp (mode, "loop") == 0)
    {
      if (mix->loop_length > 0)
	mix->mode = GDAS_MIX_LOOPS;
      else
	gdam_debug ("gdas_mix_set_mode: cannot set mode to "
		    "loop before setting loop length");
    }
  else
    {
      gdam_debug ("gdas_mix_set_mode: unknown mode %s", mode);
    }
}
static void gdas_mix_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  const char *mode = NULL;
  switch (GDAS_MIX (object)->mode)
    {
    case GDAS_MIX_ETERNAL:
      mode = "eternal";
      break;
    case GDAS_MIX_LOOPS:
      mode = "loop";
      break;
    case GDAS_MIX_FIRST_ENDED:
      mode = "first_ended";
      break;
    case GDAS_MIX_ALL_ENDED:
      mode = "all_ended";
      break;
    default:;
    }
  GDAM_ARG_STRING (arg) = mode ? g_strdup (mode) : NULL;
}


static void set_child_remove_on_eof(GdasMix* mix, int source_id, gboolean val)
{
  GdasMixElement *element;
  GdasSource *source = gdam_connection_get_source (NULL, source_id);
  if (source == NULL)
    {
      gdam_debug ("setting child_remove_on_eof: "
		  "source %d wasn't found", source_id);
      return;
    }

  element = gdas_mix_get_element_by_source (mix, source);
  if (element == NULL)
    {
      gdam_debug ("setting child_remove_on_eof: "
		  "source %d wasn't in mix",
		  GDAM_SERVER_OBJECT (source)->global_id);
      gtk_object_unref ((GtkObject *) source);
      return;
    }
  element->remove_when_ended = val;
  gtk_object_unref ((GtkObject *) source);
}

static void gdas_mix_arg_child_remove_on_eof(GtkObject* object, GdamArg* arg)
{
  set_child_remove_on_eof (GDAS_MIX (object), GDAM_ARG_INT (arg), TRUE);
}
static void gdas_mix_arg_child_dont_remove_on_eof(GtkObject*object,GdamArg*arg)
{
  set_child_remove_on_eof (GDAS_MIX (object), GDAM_ARG_INT (arg), FALSE);
}

static void
gdas_mix_arg_set_loop_length (GtkObject *object,
			      GdamArg   *arg)
{
  GDAS_MIX (object)->loop_length = GDAM_ARG_INT64 (arg);
}
static void
gdas_mix_arg_get_loop_length (GtkObject *object,
			      GdamArg   *arg)
{
  GDAM_ARG_INT64 (arg) = GDAS_MIX (object)->loop_length;
}
static void gdas_mix_class_init(GdasMixClass* mix_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (mix_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (mix_class);
  source_class->read = gdas_mix_read;
  source_class->seek = gdas_mix_seek;
  source_class->get_position = gdas_mix_get_position;

  source_class->miss_data = gdas_mix_miss_data;
  source_class->interpret_command = gdas_mix_interpret_command;
  source_class->is_ready_to_read = gdas_mix_is_ready_to_read;
  source_class->detach_child = gdas_mix_detach_child;
  source_class->get_latency_info = gdas_mix_get_latency_info;
  source_class->get_position = gdas_mix_get_position;
  source_class->get_subsources = gdas_mix_get_subsources;
  mix_class->init_samples = gdas_mix_init_samples;
  mix_class->add_samples = gdas_mix_add_samples;
  object_class->destroy = gdas_mix_destroy;

  gdam_arg_register (object_class,
		     "mixed",
		     GDAM_ARG_TYPE_INT64_ARRAY,
		     gdas_mix_arg_set_mixed,
		     gdas_mix_arg_get_mixed, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "terminates",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_set_terminates,
		     gdas_mix_get_terminates, NULL, "boolean",
		     GDAM_ARG_DEPRECATED);
  gdam_arg_register (object_class,
		     "source_does_miss",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_arg_set_source_does_miss, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "source_doesnt_miss",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_arg_set_source_doesnt_miss,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "add_with_latency",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_arg_add_with_latency, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "child_remove_on_eof",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_arg_child_remove_on_eof, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "child_dont_remove_on_eof",
		     GDAM_ARG_TYPE_INT,
		     gdas_mix_arg_child_dont_remove_on_eof,
		     NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_STRING,
		     gdas_mix_arg_set_mode,
		     gdas_mix_arg_get_mode,
		     NULL, "eternal,loop,first_ended,all_ended",
		     GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "loop_length",
		     GDAM_ARG_TYPE_INT64,
		     gdas_mix_arg_set_loop_length,
		     gdas_mix_arg_get_loop_length,
		     NULL, "1,", GDAM_ARG_VITAL);

  if (element_chunk == NULL)
    element_chunk = g_mem_chunk_create (GdasMixElement, 16, G_ALLOC_AND_FREE);
}

GtkType gdas_mix_get_type()
{
  static GtkType mix_type = 0;
  if (!mix_type)
    {
      static const GtkTypeInfo mix_info = {
	"GdasMix",
	sizeof (GdasMix),
	sizeof (GdasMixClass),
	(GtkClassInitFunc) gdas_mix_class_init,
	(GtkObjectInitFunc) gdas_mix_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_SOURCE;
      parent_class = gtk_type_class (parent);
      mix_type = gtk_type_unique (parent, &mix_info);
    }
  return mix_type;
}

GdasMix*      gdas_mix_new          (GdasMixMode    mode)
{
  GdasMix *mix = GDAS_MIX (gtk_type_new (GDAS_TYPE_MIX));
  mix->mode = mode;
  return mix;
}
void           gdas_mix_freeze       (GdasMix*       mix)
{
  mix->freeze_count++;
}
void           gdas_mix_thaw         (GdasMix*       mix)
{
  mix->freeze_count--;
  if (mix->freeze_count == 0)
    {
      mix->waiting = g_slist_sort (mix->waiting, compare_mix_elements);
    }
}
void           gdas_mix_clear        (GdasMix*       mix)
{
  g_slist_foreach (mix->playing, (GFunc) gdas_mix_element_free, NULL);
  g_slist_foreach (mix->waiting, (GFunc) gdas_mix_element_free, NULL);
  g_slist_foreach (mix->ended, (GFunc) gdas_mix_element_free, NULL);
  g_slist_free (mix->playing);
  g_slist_free (mix->waiting);
  g_slist_free (mix->ended);
  mix->playing = mix->waiting = mix->ended = NULL;
}
GdasMixElement* gdas_mix_add          (GdasMix*       mix,
                                       GdasSource*     source,
				       gint64          time)
{
  GdasMixElement *element;
  gboolean must_unref = FALSE;
  if (time == -1)
    time = mix->position;
  element = gdas_mix_element_new (source, time);
  if (source->parent != NULL)
    {
      gdas_source_orphan (source);
      must_unref = TRUE;
    }
  source->parent = (GdasSource *) mix;
  element->mix = mix;
  if (time < mix->position)
    {
      gint64 skip = mix->position - time;
      gdas_source_miss_data (source, skip);
      element->position = mix->position;
    }

  if (time <= mix->position)
    {
      mix->playing = g_slist_prepend (mix->playing, element);
    }
  else
    {
      if (mix->freeze_count > 0)
	mix->waiting = g_slist_prepend (mix->waiting, element);
      else
	mix->waiting
	  = g_slist_insert_sorted (mix->waiting,
				   element, compare_mix_elements);
    }
  mix->mix_should_end = FALSE;
  if (must_unref)
    gtk_object_unref (GTK_OBJECT (source));
  return element;
}
GdasMixElement* gdas_mix_add_named    (GdasMix*       mix,
                                       GdasSource*    source,
				       gint64         time,
				       const char*    name)
{
  GdasSource *old_named;
  GdasMixElement *element;
  old_named = gdas_mix_get_named (mix, name);
  if (old_named != NULL)
    (void) gdas_source_detach_child (GDAS_SOURCE (mix), old_named);
  if (source != NULL)
    {
      element = gdas_mix_add (mix, source, time);
      element->name = g_strdup (name);
      return element;
    }
  return NULL;
}

static GdasSource* get_named_from_list(GSList* list, const char* name)
{
  while (list != NULL)
    {
      GdasMixElement *element = list->data;
      if (element->name != NULL && strcmp (element->name, name) == 0)
	return element->source;
      list = list->next;
    }
  return NULL;
}
GdasSource*    gdas_mix_get_named     (GdasMix*       mix,
				       const char*    name)
{
  GdasSource *named;
  named = get_named_from_list (mix->playing, name);
  if (named != NULL)
    return named;
  named = get_named_from_list (mix->waiting, name);
  if (named != NULL)
    return named;
  named = get_named_from_list (mix->ended, name);
  if (named != NULL)
    return named;
  return NULL;
}
