/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdassource.h"
#include "gdasmp3.h"
#include "gdasexternal.h"
#include <gdam/common/gdambuffer.h>
#include <gdam/common/gdamarg.h>
#include <gdam/common/gdam-protocol.h>
#include <gdam/server/gdamconfigfile.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>

static GtkObjectClass*  parent_class = NULL;

static void gdas_mp3_start(GdasMp3* mp3, GdamTime start, GdamTime length);

static void gdas_mp3_set_filename(GtkObject* object, GdamArg* arg)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  if (mp3->filename)
    g_free (mp3->filename);
  mp3->filename = g_strdup (GDAM_ARG_STRING (arg));
}
static void gdas_mp3_get_filename(GtkObject* object, GdamArg* arg)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  GDAM_ARG_STRING (arg) = g_strdup (mp3->filename);
}

static void gdas_mp3_set_position(GtkObject* object, GdamArg* arg)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  mp3->start_time = GDAM_ARG_INT64 (arg);
}
static void gdas_mp3_get_position(GtkObject* object, GdamArg* arg)
{
  GdasExternal *external = GDAS_EXTERNAL (object);
  GdasMp3 *mp3 = GDAS_MP3 (object);
  if (external->fd_poll == 0)
    GDAM_ARG_INT64 (arg) = mp3->start_time;
  else
    GDAM_ARG_INT64 (arg) = external->samples_output;
}

static void gdas_mp3_set_downsample(GtkObject* object, GdamArg* arg)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  int downsample = GDAM_ARG_INT (arg);
  if (downsample == 1 || downsample == 2 || downsample == 4)
    {
      mp3->downsample = downsample;
    }
  else
    {
      gdam_debug ("invalid downsample %d, must be 1,2,or 4", downsample);
    }
}
static void gdas_mp3_get_downsample(GtkObject* object, GdamArg* arg)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  GDAM_ARG_INT (arg) = mp3->downsample;
}


static void gdas_mp3_done_constructing(GdasSource* source)
{
  GdasMp3 *mp3 = GDAS_MP3 (source);
  gdas_mp3_start (mp3,
		  mp3->start_time, GDAS_EXTERNAL (mp3)->samples_remaining);
}

static void 
gdas_mp3_finalize (GtkObject *object)
{
  GdasMp3 *mp3 = GDAS_MP3 (object);
  if (mp3->filename)
    g_free (mp3->filename);
  (*parent_class->finalize) (object);
}

static void gdas_mp3_class_init(GtkObjectClass* object_class)
{
  GDAS_SOURCE_CLASS (object_class)->done_constructing
    = gdas_mp3_done_constructing;
  object_class->finalize = gdas_mp3_finalize;
  gdam_arg_register (object_class,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdas_mp3_set_filename,
		     gdas_mp3_get_filename, NULL, NULL, GDAM_ARG_VITAL);
  /* TODO: does position belong in some superclass, perhaps where the
   * 'seek' member function is declared? */
  gdam_arg_register (object_class,
		     "position",
		     GDAM_ARG_TYPE_INT64,
		     gdas_mp3_set_position,
		     gdas_mp3_get_position, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "downsample",
		     GDAM_ARG_TYPE_INT,
		     gdas_mp3_set_downsample,
		     gdas_mp3_get_downsample, NULL, "1,4", GDAM_ARG_VITAL);
}

static void gdas_mp3_init(GdasMp3* mp3)
{
  mp3->downsample = 1;
}

GtkType gdas_mp3_get_type()
{
  static GtkType mp3_type = 0;
  if (!mp3_type)
    {
      static const GtkTypeInfo mp3_info = {
	"GdasMp3",
	sizeof (GdasMp3),
	sizeof (GdasMp3Class),
	(GtkClassInitFunc) gdas_mp3_class_init,
	(GtkObjectInitFunc) gdas_mp3_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_EXTERNAL);
      mp3_type = gtk_type_unique (GDAS_TYPE_EXTERNAL, &mp3_info);
    }
  return mp3_type;
}

/* Come command-line players don't offer sample-accurate starting points,
 * so we start before and as near as possible, then drop samples to make 
 * up the difference */
#define MPG123_SAMPLES_PER_FRAME 1152
#define OGG123_SAMPLES_PER_FRAME 44100

static void
exec_mpg123_start(GdasMp3* mp3, GdamTime start, GdamTime length)
{
  int start_frame;
  int miss;
  char start_frame_str[128];
  char **command;
  int i = 0;
  GdasExternal *external = GDAS_EXTERNAL (mp3);
  if (start < 0)
    {
      external->zero_time = (int) (-start);
      start = 0;
    }
  start_frame = (int) (start / MPG123_SAMPLES_PER_FRAME);
  sprintf (start_frame_str, "%d", start_frame);
  command = g_new (char *, 32);
  command[i++] = "mpg123";
  command[i++] = "-q";
  command[i++] = "-r";
  command[i++] = "44100";
  command[i++] = "--stereo";
  command[i++] = "-s";
  command[i++] = "-k";
  command[i++] = start_frame_str;
  command[i++] = mp3->filename;
  command[i++] = NULL;
  /* XXX: aren't args after this NULL lost? */
#if 0
  if (length != -1)
    {
      char len_str[128];
      command[i++] = "-n";
      sprintf (len_str, "%d", (int) (length / MPG123_SAMPLES_PER_FRAME) + 2);
      command[i++] = len_str;
    }
  if (mp3->downsample == 2)
    {
      command[i++] = "-2";
    }
  else if (mp3->downsample == 4)
    {
      command[i++] = "-4";
    }
  command[i++] = mp3->filename;
  command[i] = NULL;
#endif
  gdas_external_run_command (external, command);
  miss = (int) (start - (GdamTime) start_frame * MPG123_SAMPLES_PER_FRAME);
  g_assert (miss >= 0);
  if (miss > 0)
    gdas_source_miss_data (GDAS_SOURCE (mp3), miss);

  external->samples_output = start;

  /* Since samples_remaining happens to be the number used
   * for `unlimited' in both cases, ... */
  external->samples_remaining = length;
}

static void
exec_decoder_start (GdasMp3* mp3, GdamTime start, GdamTime length)
{
  gdouble start_secs = 0;
  int start_frame = 0;
  int miss = 0;
  int i = 0;
  int j = 0;
  char **command;
  char *cur_arg;
  char start_frame_str[128];
  char start_sec_str[128];
  GdasExternal *external = GDAS_EXTERNAL (mp3);
  GdamConfigExternalType *external_type;

  if (start < 0)
    {
      external->zero_time = (int) (-start);
      start = 0;
    }

  external_type = gdam_external_type_from_type_name (mp3->exec_cmd_string);
  g_return_if_fail (external_type != NULL);
  g_return_if_fail (external_type->command_args != NULL);

  if (external_type->frame_size > 1)
    {
      int start_samples = 0;
      start_frame = (int) (start / external_type->frame_size);
      start_samples = (start / external_type->frame_size) * external_type->frame_size;
      start_secs = (gdouble)start_samples / 44100.0;
      miss = (int) (start - (GdamTime) start_frame * external_type->frame_size);
    }
  else
    {
      start_frame = start;
      start_secs = (gdouble)start / 44100.0;
    }

  command = g_new (char *, external_type->num_command_args + 1);
  for (j = 0; external_type->command_args[j] != NULL; j++)
    {
      cur_arg = external_type->command_args[j];
      if (cur_arg[0] == '%')
	{
	  switch (cur_arg[1])
	    {
	      case 'f':
		command[i++] = mp3->filename;
		break;
	      case 's':
		sprintf (start_frame_str, "%d", start_frame);
		command[i++] = start_frame_str;
		break;
	      case 'S':
		sprintf (start_sec_str, "%f", start_secs);
		command[i++] = start_sec_str;
		break;
	      default:
		g_warning ("unknown percent-prefixed code %c", cur_arg[1]);
		command[i++] = cur_arg;
	    }
	}
      /*  handle args with embedded percent codes, eg --start=%S+%s/44100 */
      else if (strchr (cur_arg, '%'))
	{
	  char **bits = g_strsplit (cur_arg, "%", 0);
	  char **my_bits = g_new (char *, 16);  /* XXX: hard limit */
	  int k;
	  int l = 0;

	  my_bits[l++] = bits[0];

	  for (k = 1; bits[k] != NULL; k++)
	    {
	      char *cur_bit = bits[k];
	      switch (cur_bit[0])
		{
		  case 'f':
		    my_bits[l++] = mp3->filename;
		    break;
		  case 's':
		    sprintf (start_frame_str, "%d", start_frame);
		    my_bits[l++] = start_frame_str;
		    my_bits[l++] = cur_bit+1;
		    break;
		  case 'S':
		    sprintf (start_sec_str, "%f", start_secs);
		    my_bits[l++] = start_sec_str;
		    my_bits[l++] = cur_bit+1;
		    break;
		  default:
		    g_warning ("unknown embedded percent-prefixed code %c bit %d '%s' arg %d '%s'", cur_bit[1], k, cur_bit, j, cur_arg);
		    my_bits[l++] = cur_bit+1;
		}
	    }
	  my_bits[l++] = NULL;
	  command[i++] = g_strjoinv (NULL, my_bits);
	  g_strfreev (bits);
	  g_free (my_bits);
	}
      else
	command[i++] = cur_arg;
    }
  command[i++] = NULL;

  gdas_external_run_command (external, command);
  g_assert (miss >= 0);
  if (miss > 0)
    gdas_source_miss_data (GDAS_SOURCE (mp3), miss);

  external->samples_output = start;

  /* Since samples_remaining happens to be the number used
   * for `unlimited' in both cases, ... */
  external->samples_remaining = length;
}

static void
exec_ogg123_start (GdasMp3* mp3, GdamTime start, GdamTime length)
{
  int start_frame;
  int miss;
  char **command;
  char start_str[128];
  int i = 0;
  GdasExternal *external = GDAS_EXTERNAL (mp3);
  if (start < 0)
    {
      external->zero_time = (int) (-start);
      start = 0;
    }
  start_frame = (int) (start / OGG123_SAMPLES_PER_FRAME);
  command = g_new (char *, 32);
  command[i++] = "ogg123";
  command[i++] = "-q";
  command[i++] = "--device=raw";
  command[i++] = "--file=-";
  command[i++] = "-k";
  sprintf (start_str, "%d", start_frame);
  command[i++] = start_str;
  /* XXX: support output length??  stereo?? 44100 rate?  */
  command[i++] = mp3->filename;
  command[i++] = NULL;
  gdas_external_run_command (external, command);
  miss = (int) (start - (GdamTime) start_frame * OGG123_SAMPLES_PER_FRAME);
  g_assert (miss >= 0);
  if (miss > 0)
    gdas_source_miss_data (GDAS_SOURCE (mp3), miss);

  external->samples_output = start;

  /* Since samples_remaining happens to be the number used
   * for `unlimited' in both cases, ... */
  external->samples_remaining = length;
}


static void
exec_madplay_start (GdasMp3 *mp3, GdamTime start, GdamTime length)
{
  char **command;
  char start_str[128];
  int i = 0;
  int secs, samps;
  GdasExternal *external = GDAS_EXTERNAL (mp3);

  if (start < 0)
    {
      external->zero_time = (int) (-start);
      start = 0;
    }

  secs = start / 44100;
  samps = start % 44100;
  command = g_new (char *, 32);
  command[i++] = "madplay";
  command[i++] = "-q";
  command[i++] = "--stereo";
  command[i++] = "--output=raw:-";
  sprintf (start_str, "--start=%d+%d/44100", secs, samps);
  command[i++] = start_str;
  command[i++] = mp3->filename;
  command[i++] = NULL;
  /* XXX: support output length??  stereo?? 44100 rate?  */
  gdas_external_run_command (external, command);

  external->samples_output = start;

  /* Since samples_remaining happens to be the number used
   * for `unlimited' in both cases, ... */
  external->samples_remaining = length;
}

static void
gdas_mp3_start (GdasMp3 *mp3, GdamTime start, GdamTime length)
{
  if (mp3->exec_cmd == GDAS_EXEC_MADPLAY)
    exec_madplay_start (mp3, start, length);
  else if (mp3->exec_cmd == GDAS_EXEC_OGG123)
    exec_ogg123_start (mp3, start, length);
  else if (mp3->exec_cmd == GDAS_EXEC_TYPE_FROM_STRING)
    exec_decoder_start (mp3, start, length);
  else
    exec_mpg123_start (mp3, start, length);
}


GdasMp3 * 
gdas_mp3_new_full (const char *filename, GdamTime start, 
                   GdamTime len, GdasExecCmd exec_cmd,
		   const char *exec_cmd_string)
{
  GdasMp3 *mp3 = GDAS_MP3 (gtk_type_new (GDAS_TYPE_MP3));
  if (start < 0)
    {
      mp3->external.zero_time = -start;
      start = 0;
    }
  if (mp3->filename)
    g_free (mp3->filename);
  mp3->filename = g_strdup (filename);
  mp3->exec_cmd = exec_cmd;
  mp3->exec_cmd_string = g_strdup (exec_cmd_string);
  gdas_mp3_start (mp3, start, len);
  return mp3;
}

GdasMp3 *
gdas_mp3_new (const char *filename, GdamTime start, GdamTime len)
{
  return gdas_mp3_new_full (filename, start, len, GDAS_EXEC_MPG123, NULL);
}

