/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasproxysource.h"
#include <gdam/common/gdam-protocol.h>
#include "gdamconnection.h"

static void gdas_proxy_source_class_init       (GdasSourceClass*  klass);
static void gdas_proxy_source_init             (GdasProxySource*  proxy);
static void gdas_proxy_source_destroy          (GtkObject*        proxy);
static void gdas_proxy_source_finalize         (GtkObject*        proxy);
static GdamTime 
            gdas_proxy_source_get_position     (GdasSource*       input);
static int  gdas_proxy_source_read             (GdasSource*       input,
                                                GdamSample*       samples,
		                                int               max_read);
static void gdas_proxy_source_miss_data        (GdasSource*       input,
                                                int               samples_lost);
static gboolean
            gdas_proxy_source_is_ready_to_read (GdasSource* proxy_source);

static GtkObjectClass* parent_class = NULL;

GtkType gdas_proxy_source_get_type()
{
  static GtkType proxy_source_type = 0;
  if (!proxy_source_type)
    {
      static const GtkTypeInfo proxy_source_info = {
	"GdasProxySource",
	sizeof (GdasProxySource),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_proxy_source_class_init,
	(GtkObjectInitFunc) gdas_proxy_source_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      proxy_source_type = gtk_type_unique (GDAS_TYPE_SOURCE,
					   &proxy_source_info);
    }
  return proxy_source_type;
}

static void
gdas_proxy_source_get_fade_envelope(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_PROXY_SOURCE (object)->max_envelope;
}
static void
gdas_proxy_source_set_fade_envelope(GtkObject* object, GdamArg* arg)
{
  gdas_proxy_source_set_envelope (GDAS_PROXY_SOURCE (object),
				  GDAM_ARG_INT (arg));
}
static void gdas_proxy_source_get_delegate(GtkObject* object, GdamArg* arg)
{
  GdasSource *source = GDAS_PROXY_SOURCE (object)->delegate;
  int source_id = (source != NULL)
    ? GDAM_SERVER_OBJECT (source)->global_id : 0;
  GDAM_ARG_INT (arg) = source_id;
}
static void gdas_proxy_source_get_cueing(GtkObject* object, GdamArg* arg)
{
  GdasSource *source = GDAS_PROXY_SOURCE (object)->cueing;
  int source_id = (source != NULL)
    ? GDAM_SERVER_OBJECT (source)->global_id : 0;
  GDAM_ARG_INT (arg) = source_id;
}


static void gdas_proxy_source_class_init(GdasSourceClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  object_class->destroy = gdas_proxy_source_destroy;
  object_class->finalize = gdas_proxy_source_finalize;
  klass->read = gdas_proxy_source_read;
  klass->miss_data = gdas_proxy_source_miss_data;
  klass->is_ready_to_read = gdas_proxy_source_is_ready_to_read;
  klass->get_position = gdas_proxy_source_get_position;
  gdam_arg_register (object_class,
		     "fade_envelope",
		     GDAM_ARG_TYPE_INT,
		     gdas_proxy_source_set_fade_envelope,
		     gdas_proxy_source_get_fade_envelope,
		     NULL, "0,16384", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "delegate",
		     GDAM_ARG_TYPE_INT,
		     NULL,
		     gdas_proxy_source_get_delegate,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "cueing",
		     GDAM_ARG_TYPE_INT,
		     gdas_proxy_source_get_cueing,
		     NULL, NULL, NULL, GDAM_ARG_VITAL);
}

static void gdas_proxy_source_init(GdasProxySource* proxy_source)
{
}

static void gdas_proxy_source_destroy(GtkObject* proxy)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (proxy);
  if (proxy_source->delegate)
    {
      GtkObject *delegate = GTK_OBJECT (proxy_source->delegate);
      proxy_source->delegate = NULL;
      gtk_object_unref (delegate);
    }
  if (proxy_source->cueing)
    {
      GtkObject *cueing = GTK_OBJECT (proxy_source->cueing);
      proxy_source->cueing = NULL;
      gtk_object_unref (cueing);
    }
  (*parent_class->destroy) (proxy);
}

static void gdas_proxy_source_finalize(GtkObject* proxy)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (proxy);
  if (proxy_source->fading_out)
    g_free (proxy_source->fading_out);
  (*parent_class->finalize) (proxy);
}

static int gdas_proxy_source_read(GdasSource* input,
                                  GdamSample* samples,
		                  int max_read)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (input);
  GdasSource *delegate = proxy_source->delegate;
  int rv;
  int fade_index = proxy_source->fading_out_index;
  int max_fade_index = proxy_source->max_envelope;
  if (!delegate)
    return 0;
  rv = gdas_source_read (delegate, samples, max_read);
  if (rv > 0 && proxy_source->cueing)
    {
      gdas_source_miss_data (proxy_source->cueing, rv);
    }
  if (fade_index < max_fade_index)
    {
      ///int i = 0;
      GdamFixed start_vol, delta_vol;
      GdamSample *src = proxy_source->fading_out + fade_index;
      GdamSample *dst = samples;
      start_vol = fade_index * (1 << 14) / max_fade_index;
      delta_vol = (1 << 14) / max_fade_index;
      {
	int to_fade = (max_fade_index - fade_index);
	if (to_fade > rv)
	  to_fade = rv;
	while (to_fade > 0)
	  {
	    gint64 tmp64;
	    int tmp;
	    tmp64 = ((gint64) start_vol * dst->s[0])
		  + ((gint64) ((1 << 14) - start_vol) * src->s[0]);
	    tmp = (tmp64 >> 14);
	    dst->s[0] = CLAMP (tmp, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
	    tmp64 = ((gint64) start_vol * dst->s[1])
		  + ((gint64) ((1 << 14) - start_vol) * src->s[1]);
	    tmp = (tmp64 >> 14);
	    dst->s[1] = CLAMP (tmp, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
	    dst++;
	    src++;
	    to_fade--;
	    fade_index++;
	    start_vol += delta_vol;
	  }
      }
      proxy_source->fading_out_index = fade_index;
    }
  return rv;
}
static GdamTime gdas_proxy_source_get_position(GdasSource* input)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (input);
  GdasSource *delegate = proxy_source->delegate;
  if (!delegate)
    return -1;
  if (proxy_source->cueing)
    return gdas_source_get_position (proxy_source->cueing);
  return gdas_source_get_position (delegate);
}

static void gdas_proxy_source_miss_data(GdasSource* input,
                                        int samples_lost)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (input);
  GdasSource *delegate = proxy_source->delegate;
  if (delegate)
    gdas_source_miss_data (delegate, samples_lost);
  if (proxy_source->cueing)
    gdas_source_miss_data (proxy_source->cueing, samples_lost);
}

static gboolean gdas_proxy_source_is_ready_to_read(GdasSource* proxy)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (proxy);
  GdasSource *delegate = proxy_source->delegate;
  GdasSource *cueing = proxy_source->cueing;
  if (cueing && gdas_source_is_ready_to_read (cueing))
    {
      if (delegate)
	{
	  proxy_source->delegate = NULL;
	  if (proxy_source->max_envelope > 0)
	    {
	      int num_read;
	      int slide;
	      num_read = gdas_source_read
		(delegate,
		 proxy_source->fading_out, proxy_source->max_envelope);
	      if (num_read < 0)
		num_read = 0;
	      slide = proxy_source->max_envelope - num_read;
	      if (slide > 0)
		memmove (proxy_source->fading_out + slide,
			 proxy_source->fading_out,
			 sizeof (GdamSample) * num_read);
	      proxy_source->fading_out_index = slide;
	    }
	  gtk_object_unref (GTK_OBJECT (delegate));
	}
      gdam_debug ("cued ss becoming delegate");
      proxy_source->delegate = cueing;
      proxy_source->cueing = NULL;
      return TRUE;
    }
  if (!delegate)
    return TRUE;
  return gdas_source_is_ready_to_read (delegate);
}


void   gdas_proxy_source_set    (GdasProxySource* proxy_source,
                                 GdasSource*      delegate)
{
  g_return_if_fail (proxy_source != NULL);
  g_return_if_fail (delegate != NULL);
  gtk_object_ref (GTK_OBJECT (delegate));
  if (proxy_source->delegate)
    {
      GtkObject *old_delegate = GTK_OBJECT (proxy_source->delegate);
      proxy_source->delegate = NULL;
      gtk_object_unref (old_delegate);
    }
  gtk_object_sink (GTK_OBJECT (delegate));
  proxy_source->delegate = delegate;
}

void   gdas_proxy_source_set_cue(GdasProxySource* proxy_source,
                                 GdasSource*      to_cue)
{
  g_return_if_fail (proxy_source != NULL);
  g_return_if_fail (to_cue != NULL);
  gtk_object_ref (GTK_OBJECT (to_cue));
  if (proxy_source->cueing)
    {
      GtkObject *old_cueing = GTK_OBJECT (proxy_source->cueing);
      proxy_source->cueing = NULL;
      gtk_object_unref (old_cueing);
    }
  gtk_object_sink (GTK_OBJECT (to_cue));
  proxy_source->cueing = to_cue;
}
void   gdas_proxy_source_set_envelope(GdasProxySource* proxy_source,
                                      int              max_envelope)
{
  proxy_source->max_envelope = max_envelope;
  proxy_source->fading_out = g_renew (GdamSample,
				      proxy_source->fading_out, max_envelope);
  proxy_source->fading_out_index = max_envelope;

}
