/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasrawfile.h"
#include <gdam/common/gdamsampleconverter.h>
#include <stdlib.h>

typedef struct _GdamSubRawFile GdamSubRawFile;
struct _GdamSubRawFile {
	GdasSource		source;
	FILE*			fp;
	int			sample_size;
	gint64			position;
	gint64			bytes_remaining;
	  /* A number of empty samples which will be output before data
	   * from our FILE*   this is used for seeks and starting in 
	   * negative time.  position does not include zero_time */
	gint64			zero_time;
	int                     num_channels;
	GdamSampleConverter     sample_converter;
	GdamSampleConverter     channel_converter;
	void*			read_buffer;
	guint 			read_buffer_size;
	void                  (*close_func)(FILE*);
};

static GdamSubRawFile*     gdam_sub_raw_file_new      (FILE*       fp,
	                                               int	   sample_size,
	                                               GdamSampleConverter,
						       int         channels,
					               void (*closer)(FILE*));
static int                 gdam_sub_raw_file_read     (GdasSource* source,
                                                       GdamSample* buf,
						       int         n);
static void                gdam_sub_raw_file_miss_data(GdasSource* source,
                                                       int         n);
static void                gdam_sub_raw_file_destroy  (GtkObject*  object);

#define GDAM_TYPE_SUB_RAW_FILE              (gdam_sub_raw_file_get_type ())
#define GDAM_SUB_RAW_FILE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_SUB_RAW_FILE, GdamSubRawFile))
static GtkType gdam_sub_raw_file_get_type();

static GtkObjectClass* resample_class = NULL;
static GtkObjectClass* source_class = NULL;

/*
 *                        __ _ _
 *  _ __ __ ___      __  / _(_) | ___
 * | '__/ _` \ \ /\ / / | |_| | |/ _ \
 * | | | (_| |\ V  V /  |  _| | |  __/
 * |_|  \__,_| \_/\_/   |_| |_|_|\___|
 * 
 */

static GdamTime    
gdam_sub_raw_get_position     (GdasSource*      source)
{
  return GDAM_SUB_RAW_FILE (source)->position
    - GDAM_SUB_RAW_FILE (source)->zero_time;
}
 
static void gdam_sub_raw_file_class_init(GdasSourceClass* source_class)
{
  source_class->read = gdam_sub_raw_file_read;
  source_class->miss_data = gdam_sub_raw_file_miss_data;
  source_class->get_position = gdam_sub_raw_get_position;
  ((GtkObjectClass *) source_class)->destroy = gdam_sub_raw_file_destroy;
}
 
GtkType gdam_raw_file_get_type()
{
  static GtkType source_type = 0;
  if (!source_type)
    {
      static const GtkTypeInfo source_info = {
	"GdamRawFile",
	sizeof (GdamRawFile),
	sizeof (GdasResampleClass),
	(GtkClassInitFunc) NULL,
	(GtkObjectInitFunc) NULL,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      resample_class = gtk_type_class (GDAS_TYPE_RESAMPLE);
      source_type = gtk_type_unique (GDAS_TYPE_RESAMPLE, &source_info);
    }
  return source_type;
}


void gdam_raw_file_configure          (GdamRawFile*  raw_file,
                                       FILE*         fp,
			               void        (*close_func)(FILE*),
			               int           input_hertz,
			               int           bits_per_sample,
                                       int           is_big_endian,
			               int           is_signed,
			               int           num_channels)

{
  GdasResample *resample = GDAS_RESAMPLE (raw_file);
  gdouble rate;
  GdasFilter *filter = GDAS_FILTER (raw_file);
  GdamSampleConverter converter;
  int length;
  GdasSource *subsource;
  GdamSubRawFile *sub_raw_file;
  g_return_if_fail (resample != NULL && filter != NULL);
  g_return_if_fail (fp != NULL);

  converter = gdam_sample_converter_get (bits_per_sample,
					 is_signed,
					 is_big_endian,
					 GDAM_SAMPLE_BITS,
					 TRUE, G_BYTE_ORDER == G_BIG_ENDIAN);
  if (converter == NULL)
    {
      gdam_debug ("no converter for input sample type found");
      return;
    }

  /* XXX: hmm... if (bits_per_sample == 24) this is broken!
   *      [not much of a concern for now since i've never heard
   *       of such a .wav file]
   */
  length = (bits_per_sample / 8);

  sub_raw_file = gdam_sub_raw_file_new (fp,
					length,
					converter, num_channels, close_func);
  sub_raw_file->bytes_remaining = raw_file->bytes_remaining;
  subsource = GDAS_SOURCE (sub_raw_file);
  g_return_if_fail (subsource);
  gdas_filter_set_source (filter, subsource);
  rate = (gdouble) input_hertz / 44100.0 * (65536.0 * 65536.0);
  resample->rate = rate;

  if (raw_file->to_seek != 0)
    {
      int to_seek = raw_file->to_seek;
      raw_file->to_seek = 0;
      gdam_raw_file_seek_relative (raw_file, to_seek);
    }
  if (raw_file->samples_remaining > 0)
    {
      sub_raw_file->bytes_remaining = raw_file->samples_remaining * 
	                              sub_raw_file->sample_size * 
				      sub_raw_file->num_channels;
    }
}

static void 
gdam_sub_raw_file_miss_data (GdasSource* source, int n)
{
  GdamSubRawFile *sub = GDAM_SUB_RAW_FILE (source);
  int bytes = n * sub->sample_size * sub->num_channels;

  g_return_if_fail (n >= 0);

  sub->bytes_remaining -= bytes;

  if (sub->zero_time > 0)
    {
      int i = MIN (n, sub->zero_time);
      sub->zero_time -= i;
      n -= i;
    }

  if (n >= 0)
    {
      bytes = n * sub->sample_size * sub->num_channels;
      if (fseek (sub->fp, bytes, SEEK_CUR) < 0)
	{
	  /* XXX: onetime allocataion?? */
	  void *buf = alloca (bytes);
	  fread (buf, sub->sample_size, n * sub->num_channels, sub->fp);
	}
      sub->position += n;
    }
}


/*
 *            _                             __ _ _
 *  ___ _   _| |__    _ __ __ ___      __  / _(_) | ___
 * / __| | | | '_ \  | '__/ _` \ \ /\ / / | |_| | |/ _ \
 * \__ \ |_| | |_) | | | | (_| |\ V  V /  |  _| | |  __/
 * |___/\__,_|_.__/  |_|  \__,_| \_/\_/   |_| |_|_|\___|
 * 
 */

static void gdam_sub_raw_file_init(GdamSubRawFile* sub_raw_file)
{
  sub_raw_file->sample_size = sizeof (GdamSample);
  sub_raw_file->bytes_remaining = -1;
  sub_raw_file->sample_converter = NULL;
  sub_raw_file->channel_converter = NULL;
}

static int gdam_sub_raw_file_read(GdasSource* source, GdamSample* buf, int n)
{
  int i;
  GdamSubRawFile *sub_raw_file = GDAM_SUB_RAW_FILE (source);
  void *input;
  int scratch_size;
  int rv = 0;

  int num_channels = sub_raw_file->num_channels;
  {
    int input_size = sub_raw_file->num_channels * sub_raw_file->sample_size;
    int mid_size = sub_raw_file->num_channels * sizeof (GdamSampleValue);
    int output_size = sizeof (GdamSample);
    scratch_size = MAX (input_size, mid_size);
    scratch_size = MAX (scratch_size, output_size);
  }
  if ((int)sub_raw_file->read_buffer_size < n * scratch_size)
    {
      if (sub_raw_file->read_buffer)
	free (sub_raw_file->read_buffer);
      sub_raw_file->read_buffer = malloc (n * scratch_size);
      sub_raw_file->read_buffer_size = n * scratch_size;
    }
  input = sub_raw_file->read_buffer;

  /* if we are nearing the end of the file, limit the amount we try to
   * read to the amount available */
  if (sub_raw_file->bytes_remaining >= 0)
    {
      gint64 samples_remaining = sub_raw_file->bytes_remaining;
      samples_remaining /= sub_raw_file->sample_size;
      samples_remaining /= num_channels;
      if (samples_remaining < n)
	n = samples_remaining;
    }
  if (n <= 0)
    return n;

  /* if we have zero_time (indicating that we are playing in negative_time) */
  if (sub_raw_file->zero_time > 0)
    {
      int j = MIN (n, sub_raw_file->zero_time);
      sub_raw_file->zero_time -= j;
      n -= j;
      gdam_sample_zero (buf, j);
      rv += j;
      buf += j;
      /* time remaining variable to account for zero_time */
      if (sub_raw_file->bytes_remaining > 0)
	sub_raw_file->bytes_remaining 
	  -= j * sub_raw_file->sample_size * num_channels;
    }
  /* if the zero-time covered the entire read, we can return. */
  if (n <= 0)
    return rv;

  if ((i = fread (input,
		  sub_raw_file->sample_size,
		  n * num_channels, sub_raw_file->fp)) <= 0)
    return rv;
  if (sub_raw_file->bytes_remaining > 0)
    sub_raw_file->bytes_remaining -= i * sub_raw_file->sample_size;

  if (sub_raw_file->sample_converter != NULL)
    (*sub_raw_file->sample_converter) (input, input, i);
  if (num_channels > 1)
    i /= num_channels;
  if (sub_raw_file->channel_converter != NULL)
    {
      (*sub_raw_file->channel_converter) (buf, input, i);
    }
  else
    {
      memcpy (buf, input, i * sizeof (GdamSampleValue) * num_channels);
    }

  rv += i;
  sub_raw_file->position += i;
  return rv;
}

void
gdam_raw_file_seek_relative (GdamRawFile *raw_file, int n)
{
  GdasSource *child = GDAS_FILTER (raw_file)->filtered;

  /* If we have a sub_raw_file, then we can seek using fseek, fread,
   * and sub_raw_file->zero_time as appropriate */
  if (child && GDAM_SUB_RAW_FILE (child)->fp)
    {
      GdamSubRawFile *sub = GDAM_SUB_RAW_FILE (child);
      if (n > 0)
	{
	  int bytes = n * sub->sample_size * sub->num_channels;
	  if (fseek (sub->fp, bytes, SEEK_CUR) < 0)
	    {
	      /* XXX: sub reads to minimize allocations? onetime allocation? */
	      void *buf = alloca (bytes);
	      fread (buf, sub->sample_size, n * sub->num_channels, sub->fp);
	    }
	  sub->position += n;
	}
      else
	{
	  sub->zero_time += -1 * n;
	}
      sub->bytes_remaining -= n * sub->sample_size * sub->num_channels;
    }
  /* If we don't have a sub_raw_file, add this seek to the pending seek */
  else
    raw_file->to_seek += n;
}

static GtkType gdam_sub_raw_file_get_type()
{
  static GtkType sub_raw_type = 0;
  if (!sub_raw_type)
    {
      static const GtkTypeInfo sub_raw_info = {
	"GdamSubRawFile",
	sizeof (GdamSubRawFile),
	sizeof (GdasResampleClass),
	(GtkClassInitFunc) gdam_sub_raw_file_class_init,
	(GtkObjectInitFunc) gdam_sub_raw_file_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      source_class = gtk_type_class (GDAS_TYPE_SOURCE);
      sub_raw_type = gtk_type_unique (GDAS_TYPE_SOURCE, &sub_raw_info);
    }
  return sub_raw_type;
}
static void gdam_sub_raw_file_destroy(GtkObject*  object)
{
  GdamSubRawFile *sub_raw_file = (GdamSubRawFile *) object;
  if (sub_raw_file->fp)
    {
      (*sub_raw_file->close_func) (sub_raw_file->fp);
      sub_raw_file->fp = NULL;
    }
  if (sub_raw_file->read_buffer)
    free (sub_raw_file->read_buffer);
  sub_raw_file->read_buffer = NULL;
  sub_raw_file->read_buffer_size = 0;
  (*resample_class->destroy) (object);
}

static GdamSubRawFile* gdam_sub_raw_file_new(FILE*               fp,
	                                     int	         sample_size,
	                                     GdamSampleConverter sample_conv,
					     int                 num_channels,
					     void    (*closer)(FILE*))
{
  GdamSubRawFile *file;
  file = GDAM_SUB_RAW_FILE (gtk_type_new (gdam_sub_raw_file_get_type ()));
  file->fp = fp;
  file->sample_size = sample_size;
  file->sample_converter = sample_conv;
  file->num_channels = num_channels;
  if (num_channels != 2)
    {
      GdamSampleConverter con;
      con = gdam_sample_converter_get_channel_changer (num_channels, 2);
      file->channel_converter = con;
    }
  file->close_func = closer;
  return file;
}

void gdam_raw_file_set_bytes_remaining(GdamRawFile*    raw_file,
                                       gint64          bytes_remaining)
{
  GdasSource *subsource = GDAS_FILTER (raw_file)->filtered;
  raw_file->bytes_remaining = bytes_remaining;
  if (subsource != NULL)
    GDAM_SUB_RAW_FILE (subsource)->bytes_remaining = bytes_remaining;
}

void 
gdam_raw_file_set_samples_remaining (GdamRawFile    *raw_file,
                                     gint64          samples_remaining)
{
  GdasSource *subsource = GDAS_FILTER (raw_file)->filtered;
  if (subsource != NULL)
    GDAM_SUB_RAW_FILE (subsource)->bytes_remaining 
      = samples_remaining * 
	GDAM_SUB_RAW_FILE (subsource)->sample_size * 
	GDAM_SUB_RAW_FILE (subsource)->num_channels;
  else
    raw_file->samples_remaining = samples_remaining;
}

