/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasseekablemp3.h"
#include "gdaswav.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamarg.h>

static void gdas_seekable_mp3_destroy(GtkObject* seekable_mp3);
static gboolean    gdas_seekable_mp3_seek             (GdasSource*   source,
					               GdamTime      when,
					               int           whence);
static GdamTime    gdas_seekable_mp3_get_position     (GdasSource*   source);

static GtkObjectClass* parent_class = NULL;

static void gdas_seekable_mp3_set_filename(GtkObject* object, GdamArg* arg)
{
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (object);
  mp3->filename = g_strdup (GDAM_ARG_STRING (arg));
  mp3->stopped_length = -1;
}
static void gdas_seekable_mp3_get_filename(GtkObject* object, GdamArg* arg)
{
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (object);
  GDAM_ARG_STRING (arg) = g_strdup (mp3->filename);
}

static void gdas_seekable_mp3_set_length(GtkObject* object, GdamArg* arg)
{
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (object);
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (object);
  GdamTime length = GDAM_ARG_INT64 (arg);
  mp3->stopped_length = length;
  if (proxy->cueing != NULL)
    {
      if (mp3->use_gdaswav)
	gdam_raw_file_set_samples_remaining (GDAM_RAW_FILE (proxy->cueing),
					     length);
      else
	GDAS_EXTERNAL (proxy->cueing)->samples_remaining = length;
    }
  if (proxy->delegate != NULL)
    {
      if (mp3->use_gdaswav)
	gdam_raw_file_set_samples_remaining (GDAM_RAW_FILE (proxy->delegate),
					     length);
      else
	GDAS_EXTERNAL (proxy->delegate)->samples_remaining = length;
    }
}

static void gdas_seekable_mp3_get_length(GtkObject* object, GdamArg* arg)
{
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (object);
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (object);
  GdasSource *real_source = NULL;
  if (proxy->cueing)
    real_source = proxy->cueing;
  else if (proxy->delegate)
    real_source = proxy->delegate;
  else
    {
      GDAM_ARG_INT64 (arg) = mp3->stopped_length;
      return;
    }
  if (mp3->use_gdaswav)
    GDAM_ARG_INT64 (arg) = GDAM_RAW_FILE (real_source)->samples_remaining;
  else
    GDAM_ARG_INT64 (arg) = GDAS_EXTERNAL (real_source)->samples_remaining;
}

static void
gdas_seekable_mp3_set_exec_cmd (GtkObject* object, GdamArg* arg)
{
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (object);
  const char *string = GDAM_ARG_STRING (arg);
  seekable_mp3->exec_cmd_string = g_strdup (string);
  if (strcmp (string, "mpg123") == 0)
    {
      seekable_mp3->exec_cmd = GDAS_EXEC_MPG123;
    }
  else if (strcmp (string, "madplay") == 0)
    {
      seekable_mp3->exec_cmd = GDAS_EXEC_MADPLAY;
    }
  else if (strcmp (string, "ogg123") == 0)
    {
      seekable_mp3->exec_cmd = GDAS_EXEC_OGG123;
    }
  else if (strcmp (string, "default_mp3") == 0)
    {
      seekable_mp3->exec_cmd = GDAS_EXEC_DEFAULT_MP3;
    }
  else if (strcmp (string, "gdaswav") == 0)
    {
      seekable_mp3->use_gdaswav = TRUE;
    }
  else
    {
      seekable_mp3->exec_cmd = GDAS_EXEC_TYPE_FROM_STRING;
      gdam_debug ("gdas_seekable_mp3 no such exec_cmd %s, setting FROM_STRING:", string);
    }
}


static void
gdas_seekable_mp3_get_exec_cmd (GtkObject* object, GdamArg* arg)
{
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (object);
  const char *string = NULL;
  if (seekable_mp3->use_gdaswav)
    {
      GDAM_ARG_STRING (arg) = "gdaswav";
      return;
    }
  switch (seekable_mp3->exec_cmd)
    {
    case GDAS_EXEC_MPG123:
      string = "mpg123";
      break;
    case GDAS_EXEC_MADPLAY:
      string = "madplay";
      break;
    case GDAS_EXEC_OGG123:
      string = "ogg123";
      break;
    case GDAS_EXEC_DEFAULT_MP3:
      string = "default_mp3";
      break;
    case GDAS_EXEC_TYPE_FROM_STRING:
      string = seekable_mp3->exec_cmd_string;
      break;
    default:
      g_warning ("exec_cmd was invalid in seekable_mp3");
      return;
    }
  /* XXX: strdup?? */
  GDAM_ARG_STRING (arg) = g_strdup (string);
}

static void gdas_seekable_mp3_set_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasProxySource *proxy_source = GDAS_PROXY_SOURCE (object);
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (object);
  int buffer_size = GDAM_ARG_INT (arg);
  if (buffer_size == 0)
    buffer_size = gdas_external_get_default_buffer_size ();
  if (buffer_size <= 16)
    {
      gdam_debug ("buffer size set too short");
      return;
    }
  seekable_mp3->buffer_size = buffer_size;
  if (seekable_mp3->use_gdaswav)
    {
      g_message ("sorry, can't set buffer size on wave files");
      return;
    }
  if (proxy_source->cueing != NULL)
    {
      GdasExternal *external = GDAS_EXTERNAL (proxy_source->cueing);
      external->max_buffer_size = buffer_size;
    }
  if (proxy_source->delegate != NULL)
    {
      GdasExternal *external = GDAS_EXTERNAL (proxy_source->delegate);
      external->max_buffer_size = buffer_size;
    }
}
static void gdas_seekable_mp3_get_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (object);
  GDAM_ARG_INT (arg) = seekable_mp3->buffer_size;
}

static void 
gdas_seekable_mp3_set_never_miss_data(GtkObject* object, GdamArg* arg)
{
  GDAS_SEEKABLE_MP3 (object)->never_miss_data = GDAM_ARG_INT (arg) ? 1 : 0;
}

static void 
gdas_seekable_mp3_get_never_miss_data(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_SEEKABLE_MP3 (object)->never_miss_data;
}

static GdasSource *
get_source (GdasSeekableMp3 *mp3, gint64 position, gint64 len)
{
  GdasSource *new_source;
  if (mp3->use_gdaswav)
    {
      new_source = (GdasSource *) (gtk_type_new (gdas_wav_get_type ()));
      gdam_raw_file_seek_relative (GDAM_RAW_FILE (new_source), position);
      gdam_raw_file_set_samples_remaining (GDAM_RAW_FILE (new_source), len);
      gdam_arg_set (GTK_OBJECT (new_source), "filename", mp3->filename);
    }
  else
    {
      new_source = (GdasSource *) gdas_mp3_new_full (mp3->filename,
	  					     position,
						     len,
						     mp3->exec_cmd,
						     mp3->exec_cmd_string);
      GDAS_EXTERNAL (new_source)->max_buffer_size = mp3->buffer_size;
      GDAS_EXTERNAL (new_source)->ignore_miss_data = mp3->never_miss_data;
    }
  return new_source;
}

static void gdas_seekable_mp3_done_constructing(GdasSource* source)
{
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (source);
  GdasSource *new_source;
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (source);
  g_return_if_fail (proxy->delegate == NULL);
  gdam_debug ("seekable_mp3: creating new song at %4.6gs",
	      (gdouble) mp3->stopped_position / 44100.0);
  new_source = get_source (mp3, mp3->stopped_position, mp3->stopped_length);
  gdas_proxy_source_set (proxy, GDAS_SOURCE (new_source));
}


static void gdas_seekable_mp3_class_init(GtkObjectClass* object_class)
{
  GdasSourceClass *source_class = (GdasSourceClass *) object_class;
  object_class->destroy = gdas_seekable_mp3_destroy;
  source_class->seek = gdas_seekable_mp3_seek;
  source_class->done_constructing = gdas_seekable_mp3_done_constructing;
  source_class->get_position = gdas_seekable_mp3_get_position;

  gdam_arg_register (object_class,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdas_seekable_mp3_set_filename,
		     gdas_seekable_mp3_get_filename,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT64,
		     gdas_seekable_mp3_set_length,
		     gdas_seekable_mp3_get_length,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_size",
		     GDAM_ARG_TYPE_INT,
		     gdas_seekable_mp3_set_buffer_size,
		     gdas_seekable_mp3_get_buffer_size,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "never_miss_data",
		     GDAM_ARG_TYPE_INT,
		     gdas_seekable_mp3_set_never_miss_data,
		     gdas_seekable_mp3_get_never_miss_data,
		     NULL, "0,1", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "exec_command",
		     GDAM_ARG_TYPE_STRING,
		     gdas_seekable_mp3_set_exec_cmd,
		     gdas_seekable_mp3_get_exec_cmd,
		     NULL, NULL, GDAM_ARG_VITAL);
}

static void gdas_seekable_mp3_init(GdasSeekableMp3* mp3)
{
  mp3->stopped_length = -1;
  mp3->buffer_size = gdas_external_get_default_buffer_size ();
}
GtkType gdas_seekable_mp3_get_type()
{
  static GtkType source_type = 0;
  if (!source_type)
    {
      static const GtkTypeInfo source_info = {
	"GdasSeekableMp3",
	sizeof (GdasSeekableMp3),
	sizeof (GdasProxySourceClass),
	(GtkClassInitFunc) gdas_seekable_mp3_class_init,
	(GtkObjectInitFunc) gdas_seekable_mp3_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent_type = GDAS_TYPE_PROXY_SOURCE;
      parent_class = gtk_type_class (parent_type);
      source_type = gtk_type_unique (parent_type, &source_info);
    }
  return source_type;
}

static void gdas_seekable_mp3_destroy(GtkObject* object)
{
  GdasSeekableMp3 *mp3 = GDAS_SEEKABLE_MP3 (object);
  g_free (mp3->filename);
  mp3->filename = NULL;
  parent_class->destroy (object);
}

void             gdas_seekable_mp3_seek_absolute(GdasSeekableMp3* mp3,
                                                 GdamTime      samples)
{
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (mp3);
  GdasSource *delegate = proxy->delegate;
  GdasSource *new_source;
  if (!delegate)
    {
      mp3->stopped_position = samples;
      return;
    }
  new_source = get_source (mp3, samples, -1);
  gdas_proxy_source_set_cue (proxy, new_source);
}

void             gdas_seekable_mp3_seek_relative(GdasSeekableMp3* mp3,
                                                 GdamTime      samples)
{
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (mp3);
  GdamTime at;
  GdasSource *delegate = proxy->delegate;
  GdasSource *new_source;
  if (!delegate)
    {
      mp3->stopped_position += samples;
      return;
    }
  at = gdas_source_get_position (proxy->delegate);
  new_source = get_source (mp3, at + samples, -1);
  gdas_proxy_source_set_cue (proxy, new_source);
}

static GdamTime gdas_seekable_mp3_get_position(GdasSource* source)
{
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (source);
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (source);
  if (proxy->delegate == NULL)
    return seekable_mp3->stopped_position;
  return gdas_source_get_position (proxy->delegate);
}

static gboolean gdas_seekable_mp3_seek(GdasSource*   source,
				      GdamTime      when,
				      int           whence)
{
  GdasSeekableMp3 *seekable_mp3 = GDAS_SEEKABLE_MP3 (source);
  GdasProxySource *proxy = GDAS_PROXY_SOURCE (source);
  if (proxy->delegate == NULL)
    {
      switch (whence)
	{
	case 0:
	  seekable_mp3->stopped_position = when;
	  return TRUE;
	case 1:
	  seekable_mp3->stopped_position += when;
	  gdam_debug ("seekable_mp3... nudging before "
		      "starting %.4gs", (gdouble) when / 44100.0);
	  return TRUE;
	}
    }
  if (whence == 1)
    {
      gdas_seekable_mp3_seek_relative (GDAS_SEEKABLE_MP3 (source), when);
      return TRUE;
    }
  if (whence == 0)
    {
      gdas_seekable_mp3_seek_absolute (GDAS_SEEKABLE_MP3 (source), when);
      return TRUE;
    }
  gdam_debug ("couldn't seek mp3 from end-of-file");
  return FALSE;
}
