/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdassequencer.h"
#include <gdam/common/gdamarg.h>
#include <stdio.h>

static int         gdas_sequencer_read         (GdasSource*    input,
	                                        GdamSample*    samples,
						int            max_read);
static void        gdas_sequencer_miss_data    (GdasSource*    input,
	                                        int            samples_lost);
static gboolean    gdas_sequencer_seek         (GdasSource*    source,
                                                GdamTime       offset,
						int            whence);
static GdamTime    gdas_sequencer_get_position (GdasSource*    source);
static void        gdas_sequencer_destroy      (GtkObject*     object);
static void        clear_sequence              (GdasSequencer* sequencer);
static gboolean    prepend_sequencer_element   (GdasSequencer* sequencer,
		                                int            buffer_id,
		                                int            start_offset,
		                                int            resample,
		                                int            gain,
		                                int            sample_offset,
		                                int            output_length);

static GtkObjectClass* parent_class = NULL;
static GdasSourceClass* parent_source_class = NULL;

static void gdas_sequencer_set_elements(GtkObject* object, GdamArg* arg)
{
  int num_el = GDAM_ARG_INT_ARRAY (arg).length;
  int *elements = GDAM_ARG_INT_ARRAY (arg).array;
  int i;
  GdasSequencer *sequencer = GDAS_SEQUENCER (object);
  if (num_el % 6 != 0)
    {
      gdam_debug ("gdas_sequencer_set_elements, num ints was not "
		  "a multiple of 6");
    }
  num_el /= 6;

  clear_sequence (sequencer);
  for (i = 0; i < num_el; i++)
    {
      if (!prepend_sequencer_element (sequencer,
				      elements[i * 6 + 0],
				      elements[i * 6 + 1],
				      elements[i * 6 + 2],
				      elements[i * 6 + 3],
				      elements[i * 6 + 4],
				      elements[i * 6 + 5]))
	{
	  gdam_debug ("error creating sequencer subelement %d", i);
	}
    }
}
static void gdas_sequencer_get_elements(GtkObject* object, GdamArg* arg)
{
  GdasSequencer *sequencer = GDAS_SEQUENCER (object);
  int *elements = g_new (int, g_slist_length (sequencer->elements) * 6);
  int *rv = elements;
  GSList *at;
  for (at = sequencer->elements; at; at = at->next)
    {
      GdasSequencerElement *element;
      element = (GdasSequencerElement *) (at->data);
      *rv++ = gdam_cache_buffer_get_id (element->buffer);
      *rv++ = element->start_offset;
      *rv++ = element->resample;
      *rv++ = element->volume;
      *rv++ = element->in_buffer_offset;
      *rv++ = element->sample_output_length;
    }
  GDAM_ARG_INT_ARRAY (arg).length = rv - elements;
  GDAM_ARG_INT_ARRAY (arg).array = elements;
}
static void gdas_sequencer_get_length(GtkObject* object, GdamArg* arg)
{
  GdasSequencer *sequencer = GDAS_SEQUENCER (object);
  GDAM_ARG_INT (arg) = sequencer->sequence_length;
}
static void gdas_sequencer_set_length(GtkObject* object, GdamArg* arg)
{
  int sequence_length = GDAM_ARG_INT (arg);
  GdasSequencer *sequencer = GDAS_SEQUENCER (object);
  if (sequence_length <= 0 || sequence_length > 1000 * 1000 * 1000)
    {
      gdam_debug ("sequence_length ridiculous... sorry");
      return;
    }
  sequencer->sequence_length = sequence_length;
}

static void gdas_sequencer_class_init(GdasSourceClass* to_init)
{
  GtkObjectClass *object_class = (GtkObjectClass *) to_init;
  to_init->read = gdas_sequencer_read;
  to_init->miss_data = gdas_sequencer_miss_data;
  to_init->get_position = gdas_sequencer_get_position;
  to_init->seek = gdas_sequencer_seek;
  object_class->destroy = gdas_sequencer_destroy;
  gdam_arg_register (object_class,
		     "elements",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     gdas_sequencer_set_elements,
		     gdas_sequencer_get_elements, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "sequence_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_sequencer_set_length,
		     gdas_sequencer_get_length, NULL, "0,", GDAM_ARG_VITAL);
}

static void gdas_sequencer_init(GdasSequencer* to_init)
{
  to_init->elements = NULL;
  to_init->sequence_length = 44100 * 4;
}


GtkType gdas_sequencer_get_type ()
{
  static GtkType sequencer_type = 0;
  if (!sequencer_type)
    {
      static const GtkTypeInfo sequencer_info = {
	"GdasSequencer",
	sizeof (GdasSequencer),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_sequencer_class_init,
	(GtkObjectInitFunc) gdas_sequencer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      parent_source_class = (GdasSourceClass *) parent_class;
      sequencer_type = gtk_type_unique (GDAS_TYPE_SOURCE, &sequencer_info);
    }
  return sequencer_type;
}

static void free_element(gpointer el, gpointer unused)
{
  GdasSequencerElement *element = (GdasSequencerElement *) el;
  if (element->buffer)
    gtk_object_unref ((GtkObject *) element->buffer);
  g_free (element);
}

static void clear_sequence(GdasSequencer* sequencer)
{
  g_slist_foreach (sequencer->elements, free_element, NULL);
  g_slist_free (sequencer->elements);
  sequencer->elements = NULL;
}

static gboolean prepend_sequencer_element(GdasSequencer* sequencer,
		                          int            buffer_id,
		                          int            start_offset,
		                          int            resample,
		                          int            gain,
		                          int            sample_offset,
		                          int            output_length)
{
  GdasSequencerElement *element;
  int max_output_length;
  GdamCacheBuffer *buffer;
  int sequence_size = sequencer->sequence_length;

  buffer = gdam_cache_buffer_get_by_id (buffer_id);
  if (!buffer)
    return FALSE;

  if (start_offset >= sequence_size)
    {
      gdam_debug ("sequencer: element started after end-of-sequence");
      return FALSE;
    }
  if (output_length > sequence_size)
    {
      output_length = sequence_size;
      gdam_debug ("sequence: sample_output_length exceeded " "sequence size");
    }
  if (buffer->buffer_length <= sample_offset)
    {
      gdam_debug ("sequence: start offset in cachebuffer "
		  "exceeded sequence size");
      return FALSE;
    }
  if (resample < 1)
    {
      gdam_debug ("sequence: resample too slow");
      resample = 1;
    }
  max_output_length = (buffer->buffer_length - sample_offset);
  max_output_length = (int) (((gint64) max_output_length * GdamFixed_ONE)
			     / resample);
  if (output_length == -1)
    {
      output_length = max_output_length;
      if (output_length > sequence_size)
	output_length = sequence_size;
    }
  else if (output_length > max_output_length)
    {
      gdam_debug ("sequence: crimping %d samples desired "
		  "sample output length to %d samples",
		  output_length, max_output_length);
      output_length = max_output_length;
    }

  element = g_new (GdasSequencerElement, 1);
  element->start_offset = start_offset;
  element->resample = resample;
  element->volume = gain;
  element->in_buffer_offset = sample_offset;
  element->sample_output_length = output_length;
  element->buffer = buffer;
  sequencer->elements = g_slist_prepend (sequencer->elements, element);
  gtk_object_ref (GTK_OBJECT (buffer));
  return TRUE;
}

typedef struct _ReadInfo {
	GdamSample* 	to_fill;
	int         	length;
        GdasSequencer*	sequencer;
} ReadInfo;
static inline 
GdamSampleValue scale_add(GdamSampleValue orig,
                          GdamSampleValue incr,
			  int scale,
			  int shift)
{
  int rv = (int) orig + (((GdamSampleProduct) scale * incr) >> shift);
  return CLAMP (rv, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
}

static void do_one_read(gpointer el, gpointer info)
{
  GdasSequencerElement *element = (GdasSequencerElement *) el;
  ReadInfo *read_info = (ReadInfo *) info;
  GdamSample *output = read_info->to_fill;
  int to_do = read_info->length;
  int pos = read_info->sequencer->position;
  int cycle = read_info->sequencer->sequence_length;
  int sample_length = element->sample_output_length;
  int element_end = element->start_offset + sample_length;
  int skip = cycle - sample_length;
  GdamFixed scale = element->volume;
  int level_shift = GdamFixed_LOG_ONE;

  {
    int needed, avail, buffer_length;
    needed = (int) ((gint64) sample_length * element->resample
		    / GdamFixed_ONE);
    buffer_length = element->buffer->buffer_length;
    avail = buffer_length - element->in_buffer_offset;
    if (avail < needed)
      {
	int new_len, cut;
	new_len = (gint64) avail *GdamFixed_ONE / element->resample;
	cut = sample_length - new_len;
	element_end -= cut;
	sample_length -= cut;
	element->sample_output_length -= cut;
	skip += cut;
      }
  }

  if (sample_length <= 0 && cycle <= 0)
    return;

  while (to_do > 0)
    {
      int in_sample = -1;
      int end_index;
      if (element->start_offset <= pos && pos < element_end)
	in_sample = pos - element->start_offset;
      else if (element->start_offset <= pos + cycle
	       && pos + cycle < element_end)
	in_sample = pos + cycle - element->start_offset;

      /* 
       * This is where we render samples...
       */
      if (in_sample >= 0)
	{
	  int to_output = sample_length - in_sample;

	  /* Incrementor for whole/fraction parts of the index */
	  unsigned incr_frac;
	  int incr_whole;

	  /* Accumulator for index */
	  int index_whole;
	  unsigned index_frac;

	  GdamSample *input;

	  if (to_output > to_do)
	    to_output = to_do;

	  incr_frac = element->resample;
	  index_whole = (int) ((gint64) element->resample
			       * in_sample / GdamFixed_ONE);
	  index_frac = (incr_frac * in_sample) / GdamFixed_ONE;
	  incr_whole = incr_frac / GdamFixed_ONE;
	  incr_frac <<= (sizeof (unsigned) * 8 - GdamFixed_LOG_ONE);

	  to_do -= to_output;

#if 1
	  gdam_cache_buffer_read (element->buffer,
	                          output,
				  element->in_buffer_offset + index_whole,
				  index_frac,
				  &end_index,
				  &index_frac,
				  to_output,
				  element->resample,
				  element->volume,
				  TRUE);
#else
	  while (to_output-- > 0)
	    {
	      output->s[0] = scale_add (output->s[0],
					input->s[0], scale, level_shift);
	      output->s[1] = scale_add (output->s[1],
					input->s[1], scale, level_shift);
	      output++;
	      input += incr_whole;
	      {
		guint32 new_frac;
		new_frac = incr_frac + index_frac;
		if (new_frac < index_frac)
		  input++;
		index_frac = new_frac;
	      }
	    }
#endif
	}

      /* This is where we skip blank space... */
      if (to_do > 0)
	{
	  to_do -= skip;
	  output += skip;
	}
    }
}

static int       gdas_sequencer_read           (GdasSource      *input,
                                                GdamSample      *samples,
                                                int              max_read)
{
  ReadInfo info;

  info.to_fill = samples;
  info.length = max_read;
  info.sequencer = GDAS_SEQUENCER (input);

  gdam_sample_zero (samples, max_read);

  g_slist_foreach (info.sequencer->elements, do_one_read, &info);
  info.sequencer->position += max_read;
  info.sequencer->position %= info.sequencer->sequence_length;
  return max_read;
}

static void      gdas_sequencer_miss_data      (GdasSource*      input,
	                                        int              samples_lost)
{
  GdasSequencer *sequencer = GDAS_SEQUENCER (input);
  sequencer->position += samples_lost;
  sequencer->position %= sequencer->sequence_length;
}

static void      gdas_sequencer_destroy            (GtkObject*       object)
{
  GdasSequencer *sequencer = GDAS_SEQUENCER (object);
  clear_sequence (sequencer);
  parent_class->destroy (object);
}

static GdamTime gdas_sequencer_get_position(GdasSource* source)
{
  return GDAS_SEQUENCER (source)->position;
}

static gboolean    gdas_sequencer_seek         (GdasSource*    source,
                                                GdamTime       offset,
						int            whence)
{
  GdasSequencer *sequencer = GDAS_SEQUENCER (source);
  int sequence_length = sequencer->sequence_length;
  if (sequence_length <= 0)
    return FALSE;
  if (whence == 1)
    offset += sequencer->position;
  else if (whence == 2)
    offset += sequence_length;
  offset %= sequence_length;
  while (offset < 0)
    offset += sequence_length;
  sequencer->position = offset;
  return TRUE;
}
