/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
/*
    Portions of GdasWalker by Mike Radford <mrad@blorf.com>
*/
#include "gdaswalker.h"
#include <gdam/common/gdam-protocol.h>
#include <stdlib.h>
#include <math.h>

typedef struct _GdasWalkerConstraint GdasWalkerConstraint;

struct _GdasWalkerConstraint
{
  int forward_budget, backward_budget;
};

static int         gdas_walker_read             (GdasSource   *input,
                                                 GdamSample   *samples,
                                                 int           max_read);
static void        gdas_walker_miss_data        (GdasSource   *input,
                                                 int           samples_lost);
static void        gdas_walker_destroy          (GtkObject    *object);
static gboolean    gdas_walker_setup            (GdasWalker   *walker,
                                                 guint         buffer_id);
static GtkObjectClass *parent_class = NULL;
static GdasSourceClass *parent_source_class = NULL;

static double powers_of_two[61];
static inline double power_of_two(int exponent)
{
  g_return_val_if_fail (exponent >= -30 && exponent <= 30, 0.0);
  return powers_of_two[exponent + 30];
}
static void init_powers_of_two(void)
{
  int exponent;
  for (exponent = -30; exponent <= 30; ++exponent)
    powers_of_two[exponent + 30] = pow (2.0, exponent);
}

static inline double gdas_walker_drand48(GdasWalker *this)
{
#if HAVE_ERAND48
  return erand48 (this->random_state);
#else
  gdouble val;
  val = rand ();
  val /= RAND_MAX;
  return val;
#endif
}
static inline void gdas_walker_srand48(GdasWalker *this, int seed)
{
  this->random_state[0] = 0x330e;
  *((int *) &this->random_state[1]) = seed;
}
static inline int gdas_walker_get_seed(GdasWalker *this)
{
  return *((int *) &this->random_state[1]);
}

static void gdas_walker_clear_constraints (GdasWalker *walker)
{
  if (walker->constraints)
    {
      guint i;
      for (i = 0; i < walker->constraints->len; ++i)
	g_free (g_ptr_array_index (walker->constraints, i));
      g_ptr_array_free (walker->constraints, TRUE);
    }
}

static void gdas_walker_add_constraint (GdasWalker *walker,
					int         forward_budget,
					int         backward_budget)
{
  GdasWalkerConstraint *constraint;
  constraint = g_new (GdasWalkerConstraint, 1);
  constraint->forward_budget = forward_budget;
  constraint->backward_budget = backward_budget;
  if (!walker->constraints)
    walker->constraints = g_ptr_array_new ();
  g_ptr_array_add (walker->constraints, constraint);
}

static void gdas_walker_set_mean_length(GtkObject *object, GdamArg *arg)
{
  GDAS_WALKER (object)->mean_length = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_mean_length(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->mean_length;
}
static void gdas_walker_set_bpm(GtkObject *object, GdamArg *arg)
{
  GDAS_WALKER (object)->bpm = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_bpm(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->bpm;
}
static void gdas_walker_set_first_beat_time(GtkObject *object, GdamArg *arg)
{
  GDAS_WALKER (object)->t_first_beat = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_first_beat_time(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->t_first_beat;
}
static void gdas_walker_set_sync_likelihood(GtkObject *object, GdamArg *arg)
{
  GDAS_WALKER (object)->sync_likelihood = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_sync_likelihood(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->sync_likelihood;
}
static void gdas_walker_set_forward_likelihood(GtkObject *object, GdamArg *arg)
{
  GDAS_WALKER (object)->forward_likelihood = GDAM_ARG_DOUBLE (arg);
}
static void gdas_walker_get_forward_likelihood(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_WALKER (object)->forward_likelihood;
}
static void gdas_walker_set_random_seed(GtkObject *object, GdamArg *arg)
{
  gdas_walker_srand48 (GDAS_WALKER (object), GDAM_ARG_INT (arg));
}
static void gdas_walker_get_random_seed(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_INT (arg) = gdas_walker_get_seed (GDAS_WALKER (object));
}
static void gdas_walker_set_clear_constraints(GtkObject *object, GdamArg *arg)
{
  gdas_walker_clear_constraints (GDAS_WALKER (object));
}
static void gdas_walker_get_clear_constraints(GtkObject *object, GdamArg *arg)
{
  GDAM_ARG_INT (arg) = 0;
}
static void gdas_walker_set_mode(GtkObject *object, GdamArg *arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  char *mode = GDAM_ARG_STRING (arg);
  if (strcmp (mode, "random") == 0)
    {
      walker->beat_mode = FALSE;
      walker->sync_mode = FALSE;
    }
  else if (strcmp (mode, "nonsync") == 0)
    {
      walker->beat_mode = TRUE;
      walker->sync_mode = FALSE;
    }
  else if (strcmp (mode, "sync") == 0)
    {
      walker->beat_mode = TRUE;
      walker->sync_mode = TRUE;
    }
  else
    gdam_debug ("bad gdas_walker mode '%s'", mode);
}
static void gdas_walker_get_mode(GtkObject *object, GdamArg *arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (!walker->beat_mode)
    GDAM_ARG_STRING (arg) = g_strdup ("random");
  else if (!walker->sync_mode)
    GDAM_ARG_STRING (arg) = g_strdup ("nonsync");
  else
    GDAM_ARG_STRING (arg) = g_strdup ("sync");
}
static void gdas_walker_set_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (!gdas_walker_setup (walker, GDAM_ARG_INT (arg)))
    {
      gdam_debug ("walker_setup failed (buffer_id=%d)", GDAM_ARG_INT (arg));
      return;
    }
  if (walker->position >= walker->cache->buffer_length)
    walker->position = walker->cache->buffer_length - 1;
}
static void gdas_walker_get_buffer_id(GtkObject* object, GdamArg* arg)
{
  GdamCacheBuffer *buffer = GDAS_WALKER (object)->cache;
  GDAM_ARG_INT (arg) = buffer ? gdam_cache_buffer_get_id (buffer) : 0;
}
static gboolean gdas_walker_seek(GdasSource      *input,
                                 GdamTime         offset,
                                 int              whence)
{
  GdasWalker *walker = GDAS_WALKER (input);
  int buffer_len = 0;
  int position = walker->position;
  int new_position;
  if (walker->cache != NULL)
    buffer_len = walker->cache->buffer_length;
  if (whence == 2)
    new_position = buffer_len + offset;
  else if (whence == 1)
    new_position = position + offset;
  else
    new_position = offset;
  if (buffer_len > 0)
    walker->position = new_position % buffer_len;
  return TRUE;
}
static GdamTime gdas_walker_get_position(GdasSource *source)
{
  return GDAS_WALKER (source)->position;
}

static void gdas_walker_set_add_constraints(GtkObject* object, GdamArg* arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  int i;
  if (GDAM_ARG_INT_ARRAY (arg).length % 2)
    {
      gdam_debug ("set_add_constraints called with odd array length");
      return;
    }
  for (i = 0; i < GDAM_ARG_INT_ARRAY (arg).length; i += 2)
    {
      gdas_walker_add_constraint (walker,
				  GDAM_ARG_INT_ARRAY (arg).array[i],
				  GDAM_ARG_INT_ARRAY (arg).array[i+1]);
    }
}

static void gdas_walker_get_add_constraints(GtkObject* object, GdamArg* arg)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (walker->constraints)
    {
      int len = walker->constraints->len;
      int *array;
      int i, j;
      array = g_new (int, 2 * len);
      for (i = j = 0; i < len; ++i)
	{
	  GdasWalkerConstraint *constraint =
	    g_ptr_array_index (walker->constraints, i);
	  array[j++] = constraint->forward_budget;
	  array[j++] = constraint->backward_budget;
	}
      GDAM_ARG_INT_ARRAY (arg).length = 2 * len;
      GDAM_ARG_INT_ARRAY (arg).array = array;
    }
  else
    {
      GDAM_ARG_INT_ARRAY (arg).length = 0;
/* XXX: ??? */
      GDAM_ARG_INT_ARRAY (arg).array = g_new (int, 0);
    }
}

static void gdas_walker_class_init(GdasSourceClass* to_init)
{
  GtkObjectClass *object_class = (GtkObjectClass *) to_init;
  to_init->read = gdas_walker_read;
  to_init->miss_data = gdas_walker_miss_data;
  to_init->seek = gdas_walker_seek;
  to_init->get_position = gdas_walker_get_position;
  object_class->destroy = gdas_walker_destroy;
  init_powers_of_two ();

  gdam_arg_register (object_class,
		     "mean_length",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_mean_length,
		     gdas_walker_get_mean_length, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_STRING,
		     gdas_walker_set_mode,
		     gdas_walker_get_mode, NULL,
		     "random,nonsync,sync",
		     GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bpm",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_bpm,
		     gdas_walker_get_bpm, NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "first_beat_time",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_first_beat_time,
		     gdas_walker_get_first_beat_time,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "sync_likelihood",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_sync_likelihood,
		     gdas_walker_get_sync_likelihood,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "forward_likelihood",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_walker_set_forward_likelihood,
		     gdas_walker_get_forward_likelihood,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "random_seed",
		     GDAM_ARG_TYPE_INT,
		     gdas_walker_set_random_seed,
		     gdas_walker_get_random_seed, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdas_walker_set_buffer_id,
		     gdas_walker_get_buffer_id, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "clear_constraints",
		     GDAM_ARG_TYPE_INT,
		     gdas_walker_set_clear_constraints,
		     gdas_walker_get_clear_constraints,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "add_constraints",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     gdas_walker_set_add_constraints,
		     gdas_walker_get_add_constraints,
		     NULL, "0,", GDAM_ARG_VITAL);
}

static void gdas_walker_init(GdasWalker* to_init)
{
  to_init->mean_length = 44100.0/10.0;
  to_init->beat_mode = TRUE;
  to_init->sync_mode = TRUE;
  to_init->bpm = 120.0;
  to_init->t_first_beat = 0.0;
  to_init->sync_likelihood = 0.5;
  to_init->forward_likelihood = 0.5;
  gdas_walker_srand48 (to_init, 0xdeadbeef);
}

GtkType gdas_walker_get_type ()
{
  static GtkType walker_type = 0;
  if (!walker_type)
    {
      static const GtkTypeInfo walker_info = {
	"GdasWalker",
	sizeof (GdasWalker),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_walker_class_init,
	(GtkObjectInitFunc) gdas_walker_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      parent_source_class = (GdasSourceClass *) parent_class;
      walker_type = gtk_type_unique (GDAS_TYPE_SOURCE, &walker_info);
    }
  return walker_type;
}

static gboolean    gdas_walker_setup(GdasWalker    *walker,
                                     guint          buffer_id)
{
  GdamCacheBuffer *cache_buf;
  cache_buf = gdam_cache_buffer_get_by_id (buffer_id);
  if (cache_buf == NULL)
    return FALSE;
  if (cache_buf->buffer_length <= 0)
    {
      gdam_debug ("walker %u: still loading...", buffer_id);
      return FALSE;
    }
  gtk_object_ref (GTK_OBJECT (cache_buf));
  if (walker->cache)
    gtk_object_unref (GTK_OBJECT (walker->cache));
  walker->cache = cache_buf;
  return TRUE;
}

GdasWalker* gdas_walker_new(guint buffer_id)
{
  GdasWalker *walker;
  walker = GDAS_WALKER (gtk_type_new (GDAS_TYPE_WALKER));
  if (!gdas_walker_setup (walker, buffer_id))
    {
      gtk_object_destroy (GTK_OBJECT (walker));
      return NULL;
    }
  return walker;
}

static int       gdas_walker_read           (GdasSource      *input,
                                             GdamSample      *samples,
                                             int              max_read)
{
  GdasWalker *walker = GDAS_WALKER (input);
  GdamCacheBuffer *cache_buffer = walker->cache;

  int position = walker->position;
  int direction = walker->direction;
  int length = walker->length;
  gint64 t_samples = walker->t_samples;

  gboolean beat_mode = walker->beat_mode;
  gboolean sync_mode = walker->sync_mode;
  double mean_length = walker->mean_length;
  double beat_length = 44100.0 * 60.0 / walker->bpm;
  double first_beat_position = 44100.0 * walker->t_first_beat;
  double sync_likelihood = walker->sync_likelihood;
  double forward_likelihood = walker->forward_likelihood;
  GdasWalkerConstraint *constraint = NULL;

  int buffer_len;
  int to_read = max_read;

  if (cache_buffer == NULL)
    return 0;

  buffer_len = cache_buffer->buffer_length;

  while (to_read > 0)
    {
      /* Find next constraint with nonzero budget, if any. */
      for (;;)
	{
	  if (constraint == NULL)
	    {
	      if (!walker->constraints || walker->constraints->len == 0)
		break;
	      constraint = g_ptr_array_index (walker->constraints, 0);
	    }
	  if (constraint->forward_budget + constraint->backward_budget > 0)
	    break;
	  g_ptr_array_remove_index (walker->constraints, 0);
	  g_free (constraint);
	  constraint = NULL;
	}

      while (length <= 0)
	{
	  /* Pick a random length from an exponential distribution. */
	  double L;
	  {
	    double x;
	    do
	      x = gdas_walker_drand48 (walker);
	    while (x == 0.0);
	    L = -mean_length * log (x);
	  }

	  /* Pick a random direction. */
	  {
	    double p_forward;
	    if (constraint)
	      {
		double sum =
		  constraint->forward_budget + constraint->backward_budget;
		p_forward = sum > 0 ?
		  constraint->forward_budget / sum : forward_likelihood;
	      }
	    else
	      p_forward = forward_likelihood;
	    direction =
	      gdas_walker_drand48 (walker) < p_forward ? 1 : -1;
	  }

	  if (beat_mode)
	    {
	      double p = (double) position - first_beat_position;
	      double L_trunc;
	      int exponent;

	      /* Find even powers of 2 in beats between p and p+L (or p-L),
	       * moving from most even to least even.  (I.e., an even
	       * quarter note is considered before an even eighth note).
	       */
	      if (direction == 1)
		{
		  /* Is there an integral multiple of 2^exponent beats (in
		   * song position) between p and p+L?  (Sync mode and
		   * nonsync mode are identical when moving forward, since
		   * there will be no difference in phase.)
		   */
		  double p_after = p + L;
		  for (exponent = 30; exponent > -30; --exponent)
		    {
		      double power_len = beat_length * power_of_two (exponent);
		      L_trunc = power_len * floor (p_after/power_len) - p;
		      if (L_trunc <= 0.0 || L_trunc > L)
			continue;
		      if (gdas_walker_drand48 (walker) < sync_likelihood)
			break;
		    }
		}
	      else if (sync_mode)
		{
		  /* Sync mode--can we adjust the phase difference to
		   * integral multiple of 2^exponent beats?  (Is there a
		   * multiple of 2^exponent between phase(p) = t-p and
		   * phase(p-L) = phase(p) + 2L ?)
		   */
		  double t = (1.0/44100.0) * (double) t_samples;
		  double phase1 = t - p;
		  double phase2 = phase1 + 2.0 * L;
		  for (exponent = 30; exponent > -30; --exponent)
		    {
		      double power_len = beat_length * power_of_two (exponent);
		      double phase_trunc = power_len * floor (phase2/power_len);
		      L_trunc = 0.5 * (phase_trunc - phase1);
		      if (L_trunc <= 0.0 || L_trunc > L)
			continue;
		      if (gdas_walker_drand48 (walker) < sync_likelihood)
			break;
		    }
		}
	      else
		{
		  /* Nonsync mode--is there an integral multiple of
		   * 2^exponent beats (in song position) between p-L and p?
		   */
		  double p_after = p - L;
		  for (exponent = 30; exponent > -30; --exponent)
		    {
		      double power_len = beat_length * power_of_two (exponent);
		      L_trunc = p - power_len * ceil (p_after/power_len); 
		      if (L_trunc <= 0.0 || L_trunc > L)
			continue;
		      if (gdas_walker_drand48 (walker) < sync_likelihood)
			break;
		    }
		}
	      L = L_trunc;
	    }
	  length = (int) L;
	}
      if (direction == -1)
	{
	  if (position < 1 ||
	      (constraint && constraint->backward_budget == 0))
	    direction = +1;
	}
      else
	{
	  if (position >= buffer_len - 1 ||
	      (constraint && constraint->forward_budget == 0))
	    direction = -1;
	}
      position += direction;

      *samples++ = gdam_cache_buffer_index (cache_buffer, position);
      --to_read;
      --length;
      ++t_samples;

      if (constraint)
	{
	  if (direction == 1)
	    --constraint->forward_budget;
	  else
	    --constraint->backward_budget;
	}
    }
  walker->position = position;
  walker->direction = direction;
  walker->length = length;
  walker->t_samples = t_samples;

  return max_read;
}

static void      gdas_walker_miss_data      (GdasSource      *input,
                                             int              samples_lost)
{
  //GdasWalker* walker = GDAS_WALKER (input);
}

static void      gdas_walker_destroy            (GtkObject       *object)
{
  GdasWalker *walker = GDAS_WALKER (object);
  if (walker->cache)
    {
      GtkObject *buf = GTK_OBJECT (walker->cache);
      walker->cache = NULL;
      gtk_object_unref (buf);
    }
  gdas_walker_clear_constraints (walker);
  parent_class->destroy (object);
}
