/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaswav.h"
#include <gdam/common/gdam-protocol.h>

/*
struct _WaveFormatHeader {
	short         		wFormatTag;
	unsigned short		wChannels;
	unsigned long 		dwSamplesPerSec;
	unsigned long 		dwAvgBytesPerSec;
	unsigned short		wBlockAlign;
	unsigned short		wBitsPerSample;
};
*/


typedef struct _IFFHeader IFFHeader;
struct _IFFHeader {
	char		name[5];		/* last byte is always nul */
	unsigned	length;
};

static gboolean read_iff_header(FILE* fp, IFFHeader* iff_header)
{
  char buf[8];
  guint32 raw_int;
  if (fread (buf, 8, 1, fp) != 1)
    return FALSE;
  memcpy (iff_header->name, buf, 4);
  iff_header->name[4] = '\0';
  raw_int = *(guint32 *) (buf + 4);
  iff_header->length = GUINT32_FROM_LE (raw_int);
  return TRUE;
}

static inline gboolean read_le_short(FILE* fp, int* out)
{
  char s[2];
  gint16 int16_raw;
  if (fread (s, 2, 1, fp) != 1)
    return FALSE;
  int16_raw = *(gint16 *) (s);
  *out = (int) GINT16_FROM_LE (int16_raw);
  return TRUE;
}

static inline gboolean read_le_long(FILE* fp, int* out)
{
  char s[4];
  gint32 int32_raw;
  if (fread (s, 4, 1, fp) != 1)
    return FALSE;
  int32_raw = *(gint32 *) (s);
  *out = (int) GINT32_FROM_LE (int32_raw);
  return TRUE;
}

static gboolean discard(FILE* fp, int to_discard)
{
  char buf[1024];
  while (to_discard > 0)
    {
      int to_read = sizeof (buf);
      if (to_read > to_discard)
	to_read = to_discard;
      if (fread (buf, to_read, 1, fp) != 1)
	return FALSE;
      to_discard -= to_read;
    }
  return TRUE;
}

static gboolean
process_wave_format_header(FILE* fp,
                           int* rv_hertz,
			   int* rv_bits_per_sample,
			   int* rv_num_channels,
			   int  fmt_length)
{
  int format_tag, channels, samples_per_sec,
    avg_bytes_per_sec, block_align, bits_per_sample;
  int to_discard;
  int format = 0;
#define HEADER_PROCESSED           16
  if (!read_le_short (fp, &format_tag)
      || !read_le_short (fp, &channels)
      || !read_le_long (fp, &samples_per_sec)
      || !read_le_long (fp, &avg_bytes_per_sec)
      || !read_le_short (fp, &block_align)
      || !read_le_short (fp, &bits_per_sample))
    return FALSE;

  gdam_debug ("bits/sample=%d; hz=%d; channels=%d",
	      bits_per_sample, samples_per_sec, channels);
  if (format_tag != 1)
    {
      gdam_debug ("cannot deal with compressed wave file");
      return FALSE;
    }

  if (channels <= 0 || channels > 100)
    {
      gdam_debug ("invalid number of channels: %d", channels);
      return FALSE;
    }

  *rv_num_channels = channels;

  if (bits_per_sample % 8 != 0 || bits_per_sample <= 0)
    {
      gdam_debug ("sample size must be either 8 or 16 bits");
      return FALSE;
    }

  gdam_debug ("wav header: %d bit, %s sound", bits_per_sample,
	      (format & GDAM_RAW_FILE_SIGNED) ? "signed" : "unsigned");
  if (rv_bits_per_sample)
    *rv_bits_per_sample = bits_per_sample;
  if (rv_hertz)
    *rv_hertz = samples_per_sec;
  to_discard = fmt_length - HEADER_PROCESSED;
  if (to_discard > 0)
    {
      gdam_debug ("read_wave_header: discarding %d bytes of "
		  "crud after known header", to_discard);
      discard (fp, to_discard);
    }
  return TRUE;
}
gboolean gdas_wav_load(GdasWav* wav, const char* filename)
{
  FILE *fp = fopen (filename, "rb");
  IFFHeader iff_header;
  char wave[4];
  gboolean got_format = FALSE;
  int hertz;
  int bits_per_sample, num_channels;
  if (!fp)
    {
      gdam_debug ("error opening wav file %s", filename);
      return FALSE;
    }
  wav->filename = g_strdup (filename);

  if (!read_iff_header (fp, &iff_header)
      || fread (wave, 4, 1, fp) < 1 || memcmp (wave, "WAVE", 4) != 0)
    {
      fclose (fp);
      gdam_debug ("error reading wav header of %s", filename);
      return FALSE;
    }

  for (;;)
    {
      if (!read_iff_header (fp, &iff_header))
	{
	  fclose (fp);
	  gdam_debug ("error reading wav chunk of %s", filename);
	  return FALSE;
	}
      if (strcmp (iff_header.name, "fmt\040") == 0)
	{
	  if (!process_wave_format_header (fp,
					   &hertz,
					   &bits_per_sample,
					   &num_channels, iff_header.length))
	    {
	      fclose (fp);
	      gdam_debug ("error parsing wav `fmt' header "
			  "for %s", filename);
	      return FALSE;
	    }
	  got_format = TRUE;
	}
      else if (strcmp (iff_header.name, "data") == 0)
	{
	  break;
	}
      else
	{
	  discard (fp, iff_header.length);
	}
    }
  if (!got_format)
    {
      gdam_debug ("no format header in wav file");
      return FALSE;
    }
  if (iff_header.length > 0)
    {
      gdam_raw_file_set_bytes_remaining (GDAM_RAW_FILE (wav),
					 iff_header.length);
    }
  gdam_raw_file_configure (GDAM_RAW_FILE (wav), fp, (void (*)(FILE *)) fclose,
			   hertz, bits_per_sample, FALSE,	/* always little-endian */
			   bits_per_sample == 8 ? FALSE : TRUE, num_channels);
  return TRUE;
}

static void gdas_wav_file_set_filename(GtkObject* object, GdamArg* argument)
{
  if (!gdas_wav_load (GDAS_WAV (object), GDAM_ARG_STRING (argument)))
    gdam_debug ("wav_file_set_filename: error loading `%s'",
		GDAM_ARG_STRING (argument));
}
static void gdas_wav_file_get_filename(GtkObject* object, GdamArg* argument)
{
  GDAM_ARG_STRING (argument) = g_strdup (GDAS_WAV (object)->filename);
}

static void gdas_wav_file_class_init(GtkObjectClass* klass)
{
  gdam_arg_register (klass,
		     "filename",
		     GDAM_ARG_TYPE_STRING,
		     gdas_wav_file_set_filename,
		     gdas_wav_file_get_filename, NULL, NULL, GDAM_ARG_VITAL);
}

GtkType gdas_wav_get_type()
{
  static GtkType wav_type = 0;
  if (!wav_type)
    {
      static const GtkTypeInfo wav_info = {
	"GdasWav",
	sizeof (GdasWav),
	sizeof (GdamRawFileClass),
	(GtkClassInitFunc) gdas_wav_file_class_init,
	(GtkObjectInitFunc) NULL,	/*gdas_wav_file_init, */
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      wav_type = gtk_type_unique (GDAM_TYPE_RAW_FILE, &wav_info);
    }
  return wav_type;
}

