/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaswavetable.h"
#include "gdamnamespace.h"
#include <math.h>

#define SINE_TABLE_SIZE 16384
#define TRIANGLE_TABLE_SIZE 4096
#define SQUARE_TABLE_SIZE 1024

static void compute_sine_table(GdamSample* out, int size)
{
  double at = 0.0;
  double incr = M_PI * 2.0 / size;
  while (size-- > 0)
    {
      out->s[0] = out->s[1] = (int) (sin (at) * GDAM_SAMPLE_ONE);
      out++;
      at += incr;
    }
}
#define GDAM_SAMPLE_RANGE (GDAM_SAMPLE_ONE * 2)
static void compute_triangle_table(GdamSample* out, int size)
{
  GdamSample *out_end = out + size;
  int index = 0;
  size = (size + 1) / 2;
  while (index < size)
    {
      GdamSampleValue value;

      out_end--;
      value = ((GdamSampleProduct) index * GDAM_SAMPLE_RANGE) / size;
      value -= GDAM_SAMPLE_ONE;
      out->s[0] = out->s[1] = out_end->s[0] = out_end->s[1] = value;
      out++;
      index++;
    }
}

static GdamCacheBuffer* sine_id()
{
  GdamSample *buffer = g_new (GdamSample, SINE_TABLE_SIZE);
  compute_sine_table (buffer, SINE_TABLE_SIZE);
  return gdam_cache_buffer_memory (SINE_TABLE_SIZE, buffer, g_free, FALSE);
}


static GdamCacheBuffer* square_id()
{
  GdamSample *buffer = g_new (GdamSample, SQUARE_TABLE_SIZE);
  GdamCacheBuffer *cache_buffer;
  int i;
  for (i = 0; i < SQUARE_TABLE_SIZE / 2; i++)
    buffer[i].s[0] = buffer[i].s[1] = GDAM_SAMPLE_ONE;
  for (; i < SQUARE_TABLE_SIZE; i++)
    buffer[i].s[0] = buffer[i].s[1] = 0;
  cache_buffer = gdam_cache_buffer_memory (SQUARE_TABLE_SIZE, buffer, g_free,
                                           FALSE);
  return cache_buffer;
}

static GdamCacheBuffer* triangle_id()
{
  GdamSample *buffer = g_new (GdamSample, TRIANGLE_TABLE_SIZE);
  GdamCacheBuffer *cache_buffer;
  compute_triangle_table (buffer, TRIANGLE_TABLE_SIZE);
  cache_buffer =
    gdam_cache_buffer_memory (TRIANGLE_TABLE_SIZE, buffer, g_free, FALSE);
  return cache_buffer;
}

static void rotate_right_90(GdamCacheBuffer* cache)
{
  int i;
  int q_len = cache->buffer_length / 4;
  GdamSample *buffer = cache->_buffer;
  for (i = 0; i < q_len; i++)
    {
      GdamSampleValue tmp = buffer[i].s[1];
      buffer[i].s[1] = buffer[i + 1 * q_len].s[1];
      buffer[i + q_len].s[1] = buffer[i + 2 * q_len].s[1];
      buffer[i + 2 * q_len].s[1] = buffer[i + 3 * q_len].s[1];
      buffer[i + 3 * q_len].s[1] = tmp;
    }
}

static GdamCacheBuffer* stereo_sine_id()
{
  GdamCacheBuffer *cache = sine_id ();
  rotate_right_90 (cache);
  return cache;
}


typedef struct _GdamCommonBufferInfo GdamCommonBufferInfo;
struct _GdamCommonBufferInfo {
	const char*		name;
	GdamCacheBuffer*	(*maker)();
};

static GdamCommonBufferInfo buffer_infos[] = {
	{ "sine", sine_id },
	{ "square", square_id },
	{ "triangle", triangle_id },
	{ "stereo-sine", stereo_sine_id },
};

#define N_BUFFER_INFOS		(sizeof(buffer_infos) / sizeof(buffer_infos[0]))
static GHashTable* buffers_by_name = NULL;
static GPtrArray*  buffers_by_id   = NULL;

static int  add_common_buffer(const char*       name,
			      GdamCacheBuffer*  cache_buffer)
{
  int id;
  char *copy_name;
  if (buffers_by_name == NULL)
    {
      buffers_by_name = g_hash_table_new (g_str_hash, g_str_equal);
      buffers_by_id = g_ptr_array_new ();
      g_ptr_array_add (buffers_by_id, NULL);
    }
  id = buffers_by_id->len;
  if (id != gdam_cache_buffer_get_id (cache_buffer))
    {
      g_error ("internal error: "
	       "got buffer id %d for %s; expected %d",
	       gdam_cache_buffer_get_id (cache_buffer), name, id);
    }
  g_assert (g_hash_table_lookup (buffers_by_name, name) == NULL);
  copy_name = g_strdup (name);
  g_ptr_array_add (buffers_by_id, copy_name);
  g_hash_table_insert (buffers_by_name, copy_name, cache_buffer);
  gtk_object_ref (GTK_OBJECT (cache_buffer));
  gdam_namespace_int_set ("wavetables", copy_name, id);
  return id;
}

void gdas_wavetable_init_common()
{
  static gboolean has_inited = FALSE;
  guint i;
  if (has_inited)
    return;
  has_inited = TRUE;

  for (i = 0; i < N_BUFFER_INFOS; i++)
    {
      (void) add_common_buffer (buffer_infos[i].name,
				(*buffer_infos[i].maker) ());
    }
}
char** gdas_wavetable_get_list_of_common()
{
  char **rv;
  gdas_wavetable_init_common ();

  /* The array we maintain has a NULL at the beginning,
   * the array we return has a NULL at the end.
   * They are the same length. */
  rv = g_new (char *, buffers_by_id->len);

  memcpy (rv,
	  (char **) buffers_by_id->pdata + 1,
	  sizeof (char *) * (buffers_by_id->len - 1));
  rv[buffers_by_id->len - 1] = NULL;
  return rv;
}
