/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaswavetable.h"
#include <gdam/common/gdam-protocol.h>

static int         gdas_wavetable_read             (GdasSource*   input,
	                                            GdamSample*   samples,
						    int           max_read);
static void        gdas_wavetable_miss_data        (GdasSource*   input,
	                                            int           samples_lost);
static GdamTime    gdas_wavetable_get_position     (GdasSource*   source);
static gboolean    gdas_wavetable_seek             (GdasSource*   source,
                                                    GdamTime      time,
						    int           whence);

static void        gdas_wavetable_destroy          (GtkObject*    object);

static GtkObjectClass* parent_class = NULL;
static GdasSourceClass* parent_source_class = NULL;

static void gdas_wavetable_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  int buffer_len;
  double hertz = GDAM_ARG_DOUBLE (arg);

  if (wavetable->cache == NULL)
    return;

  buffer_len = wavetable->cache->buffer_length;
  hertz *= (1 << 16);
  hertz *= (1 << 16);
  hertz /= 44100.0;
  hertz *= buffer_len;
  wavetable->velocity_32 = (gint64) (hertz);
  gdam_debug ("wavetable_set_hertz: velocity=%lld", wavetable->velocity_32);
}
static void gdas_wavetable_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  int buffer_len = wavetable->cache->buffer_length;
  double hertz = wavetable->velocity_32;
  hertz /= (1 << 16);
  hertz /= (1 << 16);
  hertz *= 44100.0;
  if (buffer_len <= 0)
    hertz = 0;
  else
    hertz /= buffer_len;
  GDAM_ARG_DOUBLE (arg) = hertz;
}
static void gdas_wavetable_arg_set_velocity(GtkObject* object, GdamArg* arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  double velocity = GDAM_ARG_DOUBLE (arg);
  velocity *= (1 << 16);
  velocity *= (1 << 16);
  wavetable->velocity_32 = (gint64) velocity;
}
static void gdas_wavetable_arg_get_velocity(GtkObject* object, GdamArg* arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  double velocity = wavetable->velocity_32;
  velocity /= (1 << 16);
  velocity /= (1 << 16);
  GDAM_ARG_DOUBLE (arg) = velocity;
}

static void gdas_wavetable_arg_set_buffer_id(GtkObject*        object,
                                             GdamArg*          arg)
{
  GdamCacheBuffer *cache_buf;
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  int buffer_id = GDAM_ARG_INT (arg);
  cache_buf = gdam_cache_buffer_get_by_id (buffer_id);
  if (cache_buf == NULL)
    {
      gdam_debug ("no buffer id %d", buffer_id);
      return;
    }
  if (cache_buf->buffer_length <= 0)
    {
      gdam_debug ("wavetable %u: still loading...", buffer_id);
      return;
    }
  gtk_object_ref (GTK_OBJECT (cache_buf));
  if (wavetable->cache)
    gtk_object_unref (GTK_OBJECT (wavetable->cache));
  wavetable->cache = cache_buf;
  {
    gint64 buf_len_32 = (gint64) cache_buf->buffer_length << 32;
    if (wavetable->index_32 > buf_len_32)
      wavetable->index_32 %= buf_len_32;
  }
}
static void gdas_wavetable_arg_get_buffer_id(GtkObject*        object,
                                             GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  if (wavetable->cache)
    GDAM_ARG_INT (arg) = gdam_cache_buffer_get_id (wavetable->cache);
}
static void gdas_wavetable_arg_set_interpolating(GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  wavetable->interpolating = GDAM_ARG_INT (arg);
}
static void gdas_wavetable_arg_get_interpolating(GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  GDAM_ARG_INT (arg) = wavetable->interpolating;
}

static void gdas_wavetable_arg_set_terminates(GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  wavetable->terminates = GDAM_ARG_INT (arg) ? 1 : 0;
}
static void gdas_wavetable_arg_get_terminates(GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  GDAM_ARG_INT (arg) = wavetable->terminates;
}
static void gdas_wavetable_arg_set_use_overall_position
                                                (GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  wavetable->use_overall_position = GDAM_ARG_INT (arg) ? 1 : 0;
}
static void gdas_wavetable_arg_get_use_overall_position
                                                (GtkObject*        object,
                                                 GdamArg*          arg)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  GDAM_ARG_INT (arg) = wavetable->use_overall_position;
}

static void gdas_wavetable_class_init(GdasSourceClass* to_init)
{
  GtkObjectClass *object_class = (GtkObjectClass *) to_init;
  to_init->read = gdas_wavetable_read;
  to_init->miss_data = gdas_wavetable_miss_data;
  to_init->get_position = gdas_wavetable_get_position;
  to_init->seek = gdas_wavetable_seek;
  object_class->destroy = gdas_wavetable_destroy;

  gdam_arg_register (object_class,
		     "hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_wavetable_arg_set_hertz,
		     gdas_wavetable_arg_get_hertz, NULL, "0,", 0);
  gdam_arg_register (object_class,
		     "velocity",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_wavetable_arg_set_velocity,
		     gdas_wavetable_arg_get_velocity,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "use_overall_position",
		     GDAM_ARG_TYPE_INT,
		     gdas_wavetable_arg_set_use_overall_position,
		     gdas_wavetable_arg_get_use_overall_position,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "buffer_id",
		     GDAM_ARG_TYPE_INT,
		     gdas_wavetable_arg_set_buffer_id,
		     gdas_wavetable_arg_get_buffer_id,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "interpolating",
		     GDAM_ARG_TYPE_INT,
		     gdas_wavetable_arg_set_interpolating,
		     gdas_wavetable_arg_get_interpolating,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "terminates",
		     GDAM_ARG_TYPE_INT,
		     gdas_wavetable_arg_set_terminates,
		     gdas_wavetable_arg_get_terminates,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

static void gdas_wavetable_init(GdasWavetable* to_init)
{
  to_init->index_32 = 0;
  to_init->velocity_32 = (1LL << 32);
  to_init->interpolating = 1;
}

GtkType gdas_wavetable_get_type ()
{
  static GtkType wavetable_type = 0;
  if (!wavetable_type)
    {
      static const GtkTypeInfo wavetable_info = {
	"GdasWavetable",
	sizeof (GdasWavetable),
	sizeof (GdasSourceClass),
	(GtkClassInitFunc) gdas_wavetable_class_init,
	(GtkObjectInitFunc) gdas_wavetable_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
      parent_source_class = (GdasSourceClass *) parent_class;
      wavetable_type = gtk_type_unique (GDAS_TYPE_SOURCE, &wavetable_info);
    }
  return wavetable_type;
}

static inline GdamSample wavetable_index   (GdamSample* buffer,
                                            gint64      index_32,
				            gint        buflen)
{
  return buffer[index_32 >> 32];
}

/* TODO: try to write inline assembly for this... */
static inline GdamSample wavetable_lerp   (GdamSample* buffer,
                                           gint64      index_32,
				           gint        buflen)
{
  GdamSample out;
  int index = (int) (index_32 >> 32);
  guint32 fraction = (guint32) (index_32);
  const GdamSample *a = buffer + index;
  const GdamSample *b;
  if (index == buflen - 1)
    b = buffer;
  else
    b = a + 1;
  out.s[0] = (GdamSampleValue)
    ((((gint64) (~fraction) * a->s[0])
      + ((gint64) (fraction) * b->s[0])) >> 32);
  out.s[1] = (GdamSampleValue)
    ((((gint64) (~fraction) * a->s[1])
      + ((gint64) (fraction) * b->s[1])) >> 32);
  return out;
}

static inline GdamSample wavetable_index_16(GdamSample16 *buffer,
                                            gint64        index_32,
				            gint          buflen)
{
  GdamSample rv;
  buffer += (index_32 >> 32);
  rv.s[0] = (GdamSampleValue) (buffer->s[0]) << (GDAM_SAMPLE_BITS - 16);
  rv.s[1] = (GdamSampleValue) (buffer->s[1]) << (GDAM_SAMPLE_BITS - 16);
  return rv;
}



/* TODO: try to write inline assembly for this... */
static inline GdamSample wavetable_lerp_16(GdamSample16 *buffer,
                                           gint64        index_32,
				           gint          buflen)
{
  GdamSample out;
  int index = (int) (index_32 >> 32);
  guint32 fraction = (guint32) (index_32);
  const GdamSample16 *a = buffer + index;
  const GdamSample16 *b;
  if (index == buflen - 1)
    b = buffer;
  else
    b = a + 1;
  out.s[0] = (GdamSampleValue)
    ((((gint64) (~fraction) * a->s[0])
      + ((gint64) (fraction) * b->s[0])) >> (32 - (GDAM_SAMPLE_BITS - 16)));
  out.s[1] = (GdamSampleValue)
    ((((gint64) (~fraction) * a->s[1])
      + ((gint64) (fraction) * b->s[1])) >> (32 - (GDAM_SAMPLE_BITS - 16)));
  return out;
}

#define	MAX_WAVETABLE_AVERAGE		3
static int       gdas_wavetable_read           (GdasSource*      input,
	                                        GdamSample*      samples,
						int              max_read)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (input);
  int buffer_len;
  gpointer buffer;
  gboolean is_16bit;
  gint64 index_32 = wavetable->index_32;
  gint64 velocity_32 = wavetable->velocity_32;
  gint64 buflen_32;
  gboolean terminates = wavetable->terminates;
  int rv = max_read;

  if (wavetable->cache == NULL)
    return 0;

  if (input->ended)
    return 0;

  is_16bit = wavetable->cache->is_16bit;

  buffer_len = wavetable->cache->buffer_length;
  buflen_32 = (gint64) buffer_len << 32;
  buffer = wavetable->cache->_buffer;
  g_assert (0 <= index_32);
  g_assert (index_32 < ((gint64) buffer_len << 32));
#define do_wavetable_read(lerp_function, end_condition, reset_op)	\
  G_STMT_START {							\
    while (max_read-- > 0)						\
      {									\
        *samples++ = lerp_function (buffer, index_32, buffer_len);	\
	index_32 += velocity_32;					\
	if (end_condition)						\
	  {								\
	    if (terminates)						\
	      {								\
		gdas_source_has_ended (input);				\
		break;							\
	      }								\
	    index_32 reset_op buflen_32;				\
	  }								\
      }									\
  } G_STMT_END
#define do_wavetable_read_forward(lerp_function)			\
  do_wavetable_read(lerp_function, index_32 >= buflen_32, -=)
#define do_wavetable_read_backward(lerp_function)			\
  do_wavetable_read(lerp_function, index_32 < 0, +=)
  if (velocity_32 < 0)
    {
      if (is_16bit)
	if (wavetable->interpolating)
	  do_wavetable_read_backward (wavetable_lerp_16);
	else
	  do_wavetable_read_backward (wavetable_index_16);
      else
	if (wavetable->interpolating)
	  do_wavetable_read_backward (wavetable_lerp);
	else
	  do_wavetable_read_backward (wavetable_index);
    }
  else if (velocity_32 > 0)
    {
      if (is_16bit)
	if (wavetable->interpolating)
	  do_wavetable_read_forward (wavetable_lerp_16);
	else
	  do_wavetable_read_forward (wavetable_index_16);
      else
	if (wavetable->interpolating)
	  do_wavetable_read_forward (wavetable_lerp);
	else
	  do_wavetable_read_forward (wavetable_index);
    }
  else
    {
      /* um, velocity == 0??? */
      gdam_sample_zero (samples, rv);
      max_read = 0;
    }

  wavetable->index_32 = index_32;
  if (max_read <= 0)
    {
      wavetable->position += rv;
      return rv;
    }
  wavetable->position += (rv - max_read);
  return rv - max_read;
}

static void      gdas_wavetable_miss_data      (GdasSource      *input,
	                                        int              samples_lost)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (input);
  gint64 new_index = wavetable->index_32;
  new_index += (gint64) samples_lost *wavetable->velocity_32;
  new_index %= (((gint64) wavetable->cache->buffer_length) << 32);
  wavetable->index_32 = new_index;
  wavetable->position += samples_lost;
}

static GdamTime    gdas_wavetable_get_position     (GdasSource   *source)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (source);
  if (wavetable->use_overall_position)
    return wavetable->position;
  if (wavetable->velocity_32 >= wavetable->index_32)
    return 0;
  if (wavetable->velocity_32 == 0)
    return 0;
  return wavetable->index_32 / wavetable->velocity_32;
}

static gboolean    gdas_wavetable_seek             (GdasSource   *source,
                                                    GdamTime      time,
						    int           whence)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (source);
  int buffer_length;
  if (wavetable->use_overall_position)
    {
      source->ended = 0;
      if (whence == 0)
	wavetable->position = time;
      else
	wavetable->position += time;
      return TRUE;
    }
  if (wavetable->cache == NULL)
    return FALSE;
  buffer_length = wavetable->cache->buffer_length;
  if (buffer_length >= wavetable->velocity_32)
    return FALSE;
  if (buffer_length <= 0)
    return FALSE;

  switch (whence)
    {
    case 0:
    case 2:
      {
	wavetable->index_32 = time * wavetable->velocity_32;
	wavetable->index_32 %= ((gint64) buffer_length << 32);
	break;
      }
    case 1:
      {
	wavetable->index_32 += time * wavetable->velocity_32;
	wavetable->index_32 %= ((gint64) buffer_length << 32);
	break;
      }
    default:
      g_warning ("wavetable_seek: bad whence");
      break;
    }
  while (wavetable->index_32 < 0)
    wavetable->index_32 += ((gint64) buffer_length << 32);
  source->ended = 0;
  return TRUE;
}

static void      gdas_wavetable_destroy            (GtkObject*       object)
{
  GdasWavetable *wavetable = GDAS_WAVETABLE (object);
  if (wavetable->cache)
    {
      GtkObject *buf = GTK_OBJECT (wavetable->cache);
      wavetable->cache = NULL;
      gtk_object_unref (buf);
    }
  parent_class->destroy (object);
}


#if 0
void gdas_wavetable_set_velocity(GdasWavetable*   wavetable,
                                 gint64           velocity_32)
{
  wavetable->velocity_32 = velocity_32;
}
void gdas_wavetable_set_hertz   (GdasWavetable*   wavetable,
                                 gdouble          hertz)
{
  int buffer_len = wavetable->cache->buffer_length;
  /* Broken down to prevent overflow. */
  gint64 velo = (gint64) (hertz * buffer_len / 44100.0);
  wavetable->velocity_32 = velo;
}
#endif
