#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <glib.h>

/*
 * This program can write stdin to a wavfile or
 * and mp3 file.
 */

static void usage()
{
  fprintf (stderr, "usage: %s filename [options]\n\n"
                   "Record to a file.\n"
		   "used internally by gdam.\n",
		   g_get_prgname ());
  exit (1);
}

typedef enum _FileType {
  FILE_TYPE_MP3,
  FILE_TYPE_WAV,
  FILE_TYPE_RAW
} FileType;

static FileType guess_file_type (const char       *filename,
                                 const char       *options)
{
  const char *extension;
  extension = strrchr (filename, '.');
  if (extension == NULL)
    {
      g_message ("No extension on %s.  Writing raw audio data.", filename);
      return FILE_TYPE_RAW;
    }
  if (strcasecmp (extension, ".wav") == 0)
    return FILE_TYPE_WAV;
  if (strcasecmp (extension, ".mp3") == 0)
    return FILE_TYPE_MP3;
  g_warning ("No recognized extension for %s.  Writing raw audio data.",
             filename);
  return FILE_TYPE_RAW;
}

static int   open_fd (const char *filename)
{
  int fd;
  fd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  if (fd < 0)
    g_error ("couldn't create %s:  %s", filename, strerror (errno));
  return fd;
}

static int copied_length = 0;
static void copy_data (int fd)
{

}

static int   process_raw (const char *filename, const char *options)
{
  int fd;
  fd = open_fd (filename);
  copy_data (fd);
  return 0;
}
#define WAV_RIFF   0x46464952   /* "RIFF" */
#define WAV_WAVE   0x45564157   /* "WAVE" */
#define WAV_LIST   0x5453494c   /* "LIST" */
#define WAV_INFO   0x4f464e49   /* "INFO" */
#define WAV_ISFT   0x54465349   /* "ISFT" */
#define WAV_ICRD   0x44524349   /* "ICRD" */
#define WAV_FMT    0x20746d66   /* "fmt " */
#define WAV_DATA   0x61746164   /* "data" */
#define WAV_MONO   0x00010001   /* 0x0001 0x0001 */
#define WAV_STEREO 0x00020001   /* 0x0001 0x0002 */

static gboolean write_wave_file_header(int fd, int length)
{
  guint32 byte_length;
  guint32 h[11];
  guint32 fmt = (16 << 16) | 4;

  byte_length = length * 4;
  memcpy (h + 0, "RIFF", 4);
  h[1] = byte_length + 36;
  memcpy (h + 2, "WAVE", 4);
  memcpy (h + 3, "fmt ", 4);
  h[4] = GINT32_TO_LE (16);
  h[5] = GINT32_TO_LE (WAV_STEREO);
  h[6] = 44100;
  h[7] = 44100 * 4;
  h[8] = fmt;                   /* TODO: see what this field should be */
  memcpy (h + 9, "data", 4);
  h[10] = byte_length;
  if (write (fd, h, 44) != 44)
    return FALSE;
  return TRUE;
}

static int   process_wav (const char *filename, const char *options)
{
  int fd;
  fd = open_fd (filename);

  /* write wav header */
  if (! write_wave_file_header (fd, 0))
    {
      g_warning ("error writing wav file header to %s", filename);
      return 1;
    }

  copy_data (fd);
  if (lseek (fd, 0, SEEK_SET) < 0)
    g_warning ("couldn't not seek back to write length; "
                 "perhaps output is a pipe");
  else
    write_wave_file_header (fd, copied_length);
  return 0;
}
static int   process_mp3 (const char *filename, const char *options)
{
  const char *encoder = "lame";
  const char *encoder_base;
  if (options != NULL)
    encoder = options;
  encoder_base = strrchr (encoder, '/');
  if (encoder_base == NULL)
    encoder_base = encoder;
  else
    encoder_base++;

  if (strcmp (encoder_base, "gogo") == 0)
    execlp (encoder, encoder_base, "stdin", filename, NULL);
  else
    execlp (encoder, encoder_base, "-", filename, NULL);
  g_error ("sorry, couldn't run encoder %s", encoder);
  return 0;
}

int main (int argc, char *argv[])
{
  const char *filename = argv[1];
  const char *options;
  FileType file_type;

  g_set_prgname (argv[0]);
  if (filename == NULL || strcmp (filename, "--help") == 0)
    usage ();
  
  options = argv[1];

  file_type = guess_file_type (filename, options);
  switch (file_type)
    {
    case FILE_TYPE_MP3:
      return process_mp3 (filename, options);
    case FILE_TYPE_WAV:
      return process_wav (filename, options);
    case FILE_TYPE_RAW:
      return process_raw (filename, options);
    default:
      g_assert_not_reached ();
      break;
    }
  return 0;
}
