#include <sys/asoundlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <errno.h>
#include <fcntl.h>

static void usage(const char* prog_name)
{
	fprintf(stderr, "usage: %s [options] device_name\n\n"
	                "Play from stdin to an alsa device.\n",
			prog_name);
	exit(1);
}

static gboolean writen(int fd, const char* buf, int len)
{
	while (len > 0) {
		int n = write(fd, buf, len);
		if (n < 0) {
			if (errno == EINTR)
				continue;
			return FALSE;
		}
		buf += n; len -= n;
	}
	return TRUE;
}

static void print_params(snd_pcm_channel_params_t* params)
{
fprintf(stderr, "channel_params: channel=%d, mode=%d, start_mode=%d, stop_mode=%d, time=%d, ust_time=%d\n",
params->channel,
params->mode, params-> start_mode, params-> stop_mode, params-> time, params->ust_time);
fprintf(stderr, "channel_params: format: interleave=%d, format=%d, rate=%d, voices=%d, special=%d\n", params->format.interleave, params->format.format, params->format.rate, params->format.voices, params->format.special);

if(params->mode==SND_PCM_MODE_BLOCK) {
	fprintf(stderr, "channel_params: block: frag_size=%d, frags_min=%d, frags_max=%d\n", params->buf.block.frag_size,params->buf.block.frags_min,params->buf.block.frags_max);
}else{
	fprintf(stderr, "channel_params: stream: queue_size=%d, fill=%d, max_fill=%d\n", params->buf.stream.queue_size, params->buf.stream.fill,params->buf.stream.max_fill);
}
}


int main(int argc, char** argv)
{
  struct snd_pcm_channel_params params;
  struct snd_pcm_channel_setup setup;
  int arg;
  const char *device_name = NULL;
  int err;
  int fd;
  int t;
  int num_channels;
  int rate = 44100;
  int sample_size;
  int mode = SND_PCM_MODE_BLOCK;
  int log_frag_size = 11, num_fragments = 3;
  int channel = SND_PCM_CHANNEL_PLAYBACK;
  snd_pcm_format_t format;
  snd_pcm_t *pcm_handle;
  struct snd_pcm_channel_info cinfo;
  gboolean stereo = TRUE;
  for (arg = 1; arg < argc; arg++)
    {
      if (argv[arg][0] == '-')
	{
	  g_warning ("unknown option: %s", argv[arg]);
	  usage (argv[0]);
	}
      else
	{
	  if (device_name != NULL)
	    {
	      g_error ("only may specify one device: %s", argv[arg]);
	    }
	  device_name = argv[arg];
	}
    }

  if (device_name == NULL)
    {
      g_warning ("device name is mandatory");
      usage (argv[0]);
    }

  if ((err = snd_pcm_open (&pcm_handle, 0, 0, SND_PCM_OPEN_PLAYBACK)) < 0)
    {
      g_warning ("Error: audio open error: %s\n", snd_strerror (err));
    }
  memset (&cinfo, 0, sizeof (cinfo));
  cinfo.channel = channel;
  fprintf (stderr, "snd_pcm_channel_info\n");
  if ((err = snd_pcm_channel_info (pcm_handle, &cinfo)) < 0)
    {
      g_warning ("Error: channel info error: %s\n", snd_strerror (err));
    }

  snd_pcm_channel_flush (pcm_handle, SND_PCM_CHANNEL_PLAYBACK);

  bzero (&format, sizeof (format));
  format.format = SND_PCM_SFMT_S16_LE;
  format.voices = 2;
  format.interleave = 1;
  format.rate = 44100;

  memset (&params, 0, sizeof (params));
  params.mode = mode;
  params.channel = channel;
  memcpy (&params.format, &format, sizeof (format));

  if (channel == SND_PCM_CHANNEL_PLAYBACK)
    {
      params.start_mode = SND_PCM_START_FULL;
    }
  else
    {
      params.start_mode = SND_PCM_START_DATA;
    }
  params.stop_mode = SND_PCM_STOP_STOP;
  if (mode == SND_PCM_MODE_BLOCK)
    {
      //params.buf.block.frag_size = size;
      params.buf.block.frag_size = 1 << log_frag_size;
      params.buf.block.frags_max = -1;	/* little trick (playback only) */
      // params.buf.block.frags_max = 1;
      params.buf.block.frags_min = 1;
    }
  else
    {
      g_warning ("cant handle streaming mode");
    }
  print_params (&params);
  if ((err = snd_pcm_channel_params (pcm_handle, &params)) < 0)
    {
      g_message ("unable to set channel params: %s\n", snd_strerror (err));
      exit (1);
    }

  if (snd_pcm_plugin_prepare(pcm_handle, SND_PCM_CHANNEL_PLAYBACK) < 0) {
	  g_warning("snd_pcm_plugin_prepare failed");
  }

  {
    char* buf;
    int bufsize = 1 << log_frag_size;
    buf = alloca (bufsize);
    for (;;)
      {
	int num_read;
	int rv;

	num_read = read (STDIN_FILENO, buf, bufsize);
	if (num_read < 0)
	  {
	    if (errno == EINTR)
	      continue;
	    g_error ("reading from %s: %s", device_name, strerror (errno));
	  }
	if (num_read == 0)
	  break;
	rv = snd_pcm_write (pcm_handle, buf, num_read);

	if (rv < num_read)
	  {
	    if (rv < 0)
	      g_warning ("alsa: pcm_write returned error '%s'",
			 snd_strerror (rv));
	    else
	      g_message ("alsa: wrote only %d of %d samples", rv, num_read);
	  }
      }
  }
  return 0;
}
