#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <fcntl.h>

static void usage(const char* prog_name)
{
	fprintf(stderr, "usage: %s [options] device_name\n\n"
	                "Play from stdin to an oss sound device.\n"
			"  --logfragsize n   log base 2 of fragment size [12]\n"
			"  --numfragments n  number of fragments [3]\n"
			"  --rate n          sampling rate of device [44100]\n"
			"  --bits n          bits per sample [16]\n",
			prog_name);
	exit(1);
}
static gboolean writen(int fd, const char* buf, int len)
{
	while (len > 0) {
		int n = write(fd, buf, len);
		if (n < 0) {
			if (errno == EINTR)
				continue;
			return FALSE;
		}
		buf += n; len -= n;
	}
	return TRUE;
}


int main(int argc, char** argv)
{
	int arg;
	const char* device_name = NULL;
	int fd;
	int t;
	int num_channels;
	int sampling_rate = 44100;
	int bits = 16;
	int sample_size;
	int log_frag_size = 12;
	int num_fragments = 3;
	gboolean stereo = TRUE;
	for (arg = 1; arg < argc; arg++) {
		if (argv[arg][0] == '-') {
			if (strcmp(argv[arg]+1, "-help") == 0){
				usage(argv[0]);
				continue;
			}
			if (strcmp(argv[arg]+1, "-numfragments") == 0){
				if (argv[++arg] == NULL)
					usage(argv[0]);
				num_fragments = atoi(argv[arg]);
				continue;
			}
			if (strcmp(argv[arg]+1, "-rate") == 0){
				if (argv[++arg] == NULL)
					usage(argv[0]);
				sampling_rate = atoi(argv[arg]);
				continue;
			}
			if (strcmp(argv[arg]+1, "-bits") == 0){
				if (argv[++arg] == NULL)
					usage(argv[0]);
				bits = atoi(argv[arg]);
				continue;
			}
			if (strcmp(argv[arg]+1, "-logfragsize") == 0){
				if (argv[++arg] == NULL)
					usage(argv[0]);
				log_frag_size = atoi(argv[arg]);
				continue;
			}
			if (strcmp(argv[arg]+1, "-mono") == 0){
				if (argv[++arg] == NULL)
					usage(argv[0]);
				stereo = !(atoi(argv[arg]));
				continue;
			}
			g_warning("unknown option: %s", argv[arg]);
			usage(argv[0]);
		} else {
			if (device_name != NULL) {
				g_error("only may specify one device: %s",
					argv[arg]);
			}
			device_name = argv[arg];
		}
	}

	if (device_name == NULL) {
		g_warning("device name is mandatory");
		usage(argv[0]);
	}

	fd = open(device_name, O_WRONLY);
	if (fd < 0)
		g_error("opening %s: %s", device_name, strerror(errno));
	g_message("playoss opened device %s on fd %d", device_name, fd);
	g_message("playoss: %d fragments, logfragsize=%d", num_fragments, 
		  log_frag_size);
	/* Close the device on exec. */
	fcntl(fd, F_SETFD, FD_CLOEXEC);

	t = (log_frag_size & 0xffff) | ((num_fragments & 0xffff)<<16);
	if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &t) < 0) {
		g_warning("error going to correct amount of frag (%s)",
				strerror(errno));
	}
	t = bits;
	if (ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &t) < 0) {
		g_error("error going to %d bits per sample", t);
	}
	t = stereo ? 1 : 0;
	if (ioctl(fd, SNDCTL_DSP_STEREO, &t) < 0) {
		g_error("error setting device to stereo: %s",
			  strerror(errno));
	}
	if (ioctl(fd, SNDCTL_DSP_SPEED, &sampling_rate) < 0) {
		g_error("error setting dev to %d samples per sec", t);
	}

	sampling_rate = num_channels = sample_size = -1;
	if (ioctl(fd, SOUND_PCM_READ_RATE, &sampling_rate) < 0
	 || ioctl(fd, SOUND_PCM_READ_BITS, &sample_size) < 0
	 || ioctl(fd, SOUND_PCM_READ_CHANNELS, &num_channels) < 0
	 || sampling_rate == -1
	 || num_channels == -1
	 || sample_size == -1) {
                g_warning("error getting device parameters");
		g_warning("you may need to manually specify the soundcard "
			  "rate if it does not support the defaults");
	}
	if (num_channels == 1)          stereo = 0;
	else if (num_channels == 2)     stereo = 1;
	else {
		g_error("unsupported number of channels %d", num_channels);
	}
	fprintf(stderr, "PLAYOSS: device set at %d bits/sample, stereo %s, rate %d.\n",
			sample_size, stereo ? "on" : "off", sampling_rate);

	{
	  char *buf;
	  int buf_size = 1 << log_frag_size;
	  buf = alloca (buf_size);
	  for (;;) {
		  int num_read;
		  num_read = read(STDIN_FILENO, buf, buf_size);
		  if (num_read < 0) {
			  if (errno == EINTR)
				  continue;
			  g_error("reading from stdin: %s", strerror(errno));
		  }
		  if (num_read == 0)
			  break;
		  writen(fd, buf, num_read);
	  }
	}

	return 0;
}
