#include <gdam/common/gdamautoconf.h>
#include <glib.h>


#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sched.h>
#include <stdlib.h>
#include <gdam/common/gdampoll.h>

static void usage(const char* prog_name)
{
  fprintf (stderr, "usage: %s [options] stream_fd device_fd\n\n"
	   "Play from a file descriptor to the soundcard.\n"
	   "\nOptions:\n"
	   "\t--block-size #        Set the length of the data\n"
	   "\t                      we will try to write.\n"
	   "\t--num-blocks #        Set the number of blocks to\n"
	   "\t                      hold around.\n", prog_name);
  exit (1);
}


/* XXX: isolate all priority code in gdamrealtime.c ?? */
static gboolean
set_realtime_mode()
{
  int pid = getpid();

  if (geteuid () != 0 && getuid () != 0)
    {
      g_message ("gdamrtcat: you are not root, cannot run in realtime mode.");
      /* not root, don't even try to use realtime stuff */
      return FALSE;
    }
#ifdef  _SC_PRIORITY_SCHEDULING
  if (sysconf (_SC_PRIORITY_SCHEDULING) == -1)
    {
      g_warning ("WARNING: RR-scheduler not available, disabling.\n");
      return FALSE;
    }
#endif
#if HAVE_SCHED_SETSCHEDULER
  {
    int sched_fifo_min, sched_fifo_max;
    int rv;
    struct sched_param sched_parms;

    memset (&sched_parms, 0, sizeof (sched_parms));
    //sched_fifo_min = sched_get_priority_min (SCHED_FIFO);
    sched_fifo_max = sched_get_priority_max (SCHED_FIFO);
    sched_parms.sched_priority = sched_fifo_max;
    rv = sched_setscheduler (pid, SCHED_FIFO, &sched_parms);
    if (rv == -1)
      {
	if (errno == EPERM && getuid () != 0)
	  {
	    g_message ("Since you were not root, "
		       "could not set posix \"realtime\" mode in gdamrtcat");
	    return FALSE;
	  }
	g_warning ("gdamrtcat sched_setschedule error: %s", strerror (errno));
	return FALSE;
      }
    else if (rv < 0)
      {
	g_warning ("unexpected negative return value from realtime request");
	return FALSE;
      }
    return TRUE;
  }
#endif
#if HAVE_PTHREAD_ATTR_INIT
  {
    sched_param *param;
    pthread_attr_t pattr;
    int policy;

    pthread_attr_init (&pattr);
    pthread_attr_setschedparam (attr, param);
    pthread_attr_setschedpolicy (attr, policy);
    pthread_attr_destroy (&pattr);
    /* XXX: in this interface, attr can only be set at create??  figure out
       * how to change priority of _this_ thread */
    /* XXX: how do we tell if this call succeeded. */
    return TRUE;
  }
#endif
  return FALSE;
}

static gboolean
try_mlock_all ()
{
#if HAVE_MLOCKALL
  int kilobytes;
  int i;
  int pid = getpid();
  char** reserved_space;
  kilobytes = 1024;
  reserved_space = g_malloc (sizeof (char*) * kilobytes);

  /*  For each kilobyte reserved, alloc a new block and touch it.  */
  for (i = 0; i < kilobytes; i++)
    {
      char* new_block;
      reserved_space[i] = g_malloc (sizeof (char) * 1024);
      new_block = reserved_space[i];
      new_block[0] = 0;
      new_block[512] = 0;
    }

  mlockall(MCL_CURRENT);
  g_message ("set posix \"realtime\" mode for pid %d", pid);
  g_message ("locked current ram pages plus %dk extra", kilobytes);

  /*  Free each kilobyte reserved.  */
  for (i = 0; i < kilobytes; i++)
    {
      char* new_block;
      g_free (reserved_space[i]);
      reserved_space[i] = 0;
    }
  g_free (reserved_space);
  return TRUE;
#else
  return FALSE;
#endif /* !HAVE_MLOCKALL */
}

#if 0
static void set_realtime_mode()
{
  sched_setscheduler (getpid (), SCHED_FIFO, 0);
}
#endif

static void drop_setuid()
{
  int uid = getuid ();
  int euid = geteuid ();
  if (uid != 0 && euid != 0)
    return;
  if (uid != 0)
    {
      seteuid (uid);
      return;
    }
  g_warning ("TODO: drop priviledges to nobody...");
}

typedef struct _Buffer Buffer;
struct _Buffer {
	int		buffer_start;
	int		buffer_end;
	int		buffer_alloc;
	Buffer*		next;

	/* Buffer start right after structure */
};

static inline char* buffer_data(Buffer* buffer)
{
  return (char *) (buffer + 1);
}

/* Make a circular ring of Buffers. */
static Buffer* allocate_buffers(int buf_size, int num_buf)
{
  Buffer *rv;
  Buffer **p_last_next_buffer = &rv;
  g_assert (num_buf > 0);
  while (num_buf-- > 0)
    {
      Buffer *buffer = (Buffer *) malloc (sizeof (Buffer) + buf_size);
      buffer->buffer_start = buffer->buffer_end = 0;
      buffer->buffer_alloc = buf_size;
      *p_last_next_buffer = buffer;
      p_last_next_buffer = &buffer->next;
    }
  /* complete the circle */
  *p_last_next_buffer = rv;
  return rv;
}

#define IN_FD_READY		1
#define OUT_FD_READY		2
static int do_select(int in_fd, int out_fd)
{
  GPollFD fds[2];
  fds[0].fd = in_fd;
  fds[0].events = G_IO_IN;
  fds[1].fd = out_fd;
  fds[1].events = G_IO_OUT;
  if (gdam_poll (fds, 2, -1) < 0)
    {
      if (errno == EINTR)
	return 0;
      g_error ("error polling: %s", strerror (errno));
    }
  return (((fds[0].revents & G_IO_IN) != 0) ? IN_FD_READY : 0)
    | (((fds[1].revents & G_IO_OUT) != 0) ? OUT_FD_READY : 0);
}
static void do_read(Buffer** p_buffer, int fd)
{
  Buffer *buf = *p_buffer;
  int to_read = buf->buffer_alloc - buf->buffer_end;
  int num_read;
  if (to_read == 0)
    {
      return;
    }
  g_assert (to_read > 0);
  num_read = read (fd, buffer_data (buf) + buf->buffer_end, to_read);
  if (num_read < 0)
    {
      if (errno == EINTR)
	return;
      g_error ("error reading from file descriptor %d : %s", fd,
	       strerror (errno));
    }
  if (num_read == 0)
    {
      g_message ("end-of-file in do_read:  the device is closing");
      exit (0);
    }
  buf->buffer_end += num_read;
  if (num_read == to_read)
    *p_buffer = buf->next;
}

static void do_write(Buffer** p_buffer, int fd)
{
  Buffer *buf = *p_buffer;
  int to_write = buf->buffer_end - buf->buffer_start;
  int num_written;
  g_assert (to_write > 0);
  num_written = write (fd, buffer_data (buf) + buf->buffer_start, to_write);
  if (num_written < 0)
    {
      if (errno == EINTR)
	return;
      g_error ("error writing to file descriptor: %s", strerror (errno));
    }
  if (num_written == 0)
    {
      g_warning ("writing to sound device returned 0. huh?");
      exit (0);
    }
  buf->buffer_start += num_written;
  if (buf->buffer_start == buf->buffer_alloc)
    {
      buf->buffer_start = buf->buffer_end = 0;
      *p_buffer = buf->next;
    }
}

static void do_buffering(int in_fd,      int out_fd,
                         int block_size, int num_blocks)
{
  Buffer *in_buf;
  Buffer *out_buf;
  in_buf = out_buf = allocate_buffers (block_size, num_blocks);
  for (;;)
    {
      /*
       * Cases:
       *    no data in buffer
       *         may as well wait on in_fd, since we have 
       *         nothing to say to out_fd.
       *    buffer is full
       *         may as well wait on out_fd, since we cannot
       *         accept more data from in_fd.
       *    buffer has some data, but isn't full.
       *         select on in_fd and out_fd.
       */
      if (out_buf->buffer_end == out_buf->buffer_start)
	{
	  /* Buffer is empty. */
	  do_read (&in_buf, in_fd);
	}
      else if (in_buf->buffer_end == in_buf->buffer_alloc)
	{
	  /* Buffer is full. */
	  do_write (&out_buf, out_fd);
	}
      else
	{
	  int bits = do_select (in_fd, out_fd);
	  if ((bits & OUT_FD_READY) == OUT_FD_READY)
	    {
	      do_write (&out_buf, out_fd);
	    }
	  if ((bits & IN_FD_READY) == IN_FD_READY)
	    {
	      do_read (&in_buf, in_fd);
	    }
	}
    }

}

int main(int argc, char** argv)
{
  int arg;
  int stream_fd = -1;
  int device_fd = -1;
  int t;
  int block_size = 8192;
  int num_blocks = 3;
  gboolean stereo = TRUE;
  gboolean realtime = TRUE;
  for (arg = 1; arg < argc; arg++)
    {
      if (argv[arg][0] == '-')
	{
	  if (strcmp (argv[arg], "--no-realtime") == 0)
	    {
	      realtime = FALSE;
	      continue;
	    }
	  if (strcmp (argv[arg], "--block-size") == 0)
	    {
	      arg++;
	      if (argv[arg] == NULL)
		g_error ("need argument after --block-size");
	      block_size = atoi (argv[arg]);
	      continue;
	    }
	  if (strcmp (argv[arg], "--num-blocks") == 0)
	    {
	      arg++;
	      if (argv[arg] == NULL)
		g_error ("need argument after --num-blocks");
	      num_blocks = atoi (argv[arg]);
	      continue;
	    }
	  g_warning ("unknown option: %s", argv[arg]);
	  usage (argv[0]);
	}
      else
	{
	  if (stream_fd < 0)
	    {
	      stream_fd = atoi (argv[arg]);
	      continue;
	    }
	  if (device_fd < 0)
	    {
	      device_fd = atoi (argv[arg]);
	      continue;
	    }
	  g_warning ("too many arguments: %s", argv[arg]);
	  usage (argv[0]);
	}
    }

  if (device_fd < 0 || stream_fd < 0)
    {
      g_warning ("device and stream must be specified as "
		 "nonnegative integers");
      usage (argv[0]);
    }

  /* The actual program. */
  if (realtime)
    {
      set_realtime_mode ();
      try_mlock_all ();
    }
  drop_setuid ();
  do_buffering (stream_fd, device_fd, block_size, num_blocks);

  return 0;
}
