/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAS_BENDER_H_
#define __GDAS_BENDER_H_


typedef struct _GdasBender GdasBender;
typedef struct _GdasBenderClass GdasBenderClass;
typedef struct _GdasBenderChange GdasBenderChange;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*  Strategy:  Bender is a filter which allows time-domain distortion
 *  (changes in playback speed and seeks)  while optionally maintaining
 *  an overall 1:1 playback speed.  Each change is either a command to
 *  play at a certain speed or a command to change position.  
 *
 *  We keep a 'mark' which represents a location in the input or output
 *  stream.  In keep_time mode, the mark represents the point in the input
 *  where we would be if we had made no changes.  
 *
 *  We also keep an array of position markers which always represent a
 *  certain position within the input stream, regardless of mode.  It is
 *  possible that a position marker becomes invalid as data is discarded
 *  from the buffer, in which case it will be set to a negative number.
 */

#include <gdam/server/gdasbufferingfilter.h>
typedef enum {
	/* When the changes are over, keep playing from wherever we are. */
	GDAS_BENDER_CHANGE_CONTINUOUS,

	/* Keep a mark that can be jumped to ("jump_to_mark")
	 * and set ("set_mark").  the mark advances full speed,
	 * so jumping to the mark has the effect of 'catching up'
	 * to where we would have been without any bending.  if
	 * the current play position gets more than backbuffer_size
	 * behind the mark position, play position will be shifted
	 * forward to accomodate. */
	GDAS_BENDER_CHANGE_KEEP_TIME,

	/* After changes are over, snap to the mark.  This will
	 * stall the input until our changes require us to 
	 * advance. */
	/* XXX: what is the exact behaviour? */
	GDAS_BENDER_CHANGE_SNAP_TIME,

	/* When the changes are over, return to where we were
	 * when the changes started. */
	GDAS_BENDER_CHANGE_RETURN,

	/* Similar to "keep time" but automatically moves the end
	 * point back by a fraction of beat length to keep in beat */
	GDAS_BENDER_CHANGE_KEEP_IN_BEAT,

	/* Similar to "keep beat/time" but ensures that input always
	 * plays at 100% speed, and doesn't allow fast-forwarding
	 * into the future.  this is useful if our input is a
	 * microphone or such which produces input inflexibly.  Input
	 * data is always appended to the buffer, so one can't keep the
	 * playback position in the same area forever... eventually the
	 * position will become invalid as the audio it points to is
	 * displaced out the end of the buffer.  */
	GDAS_BENDER_CHANGE_KEEP_INPUT_CONSTANT,

	/*  Similar to input_constant, but the current playback position
	 *  never becomes invalid... if it comes within 1/2 (1?) beat
	 *  of the far end of the backbuffer, we start discarding incoming
	 *  data
	 */
	GDAS_BENDER_CHANGE_KEEP_INPUT_CONSTANT_VALID

	/* TODO: keep_in_beat which biases to forward buffering / back
	 * buffering / keeping even forward,backward */

} GdasBenderChangeMode;

GtkType gdas_bender_get_type();
#define GDAS_TYPE_BENDER			(gdas_bender_get_type ())
#define GDAS_BENDER(obj)              (GTK_CHECK_CAST ((obj), GDAS_TYPE_BENDER, GdasBender))
#define GDAS_BENDER_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAS_TYPE_BENDER, GdasBenderClass))
#define GDAM_IS_BENDER(obj)           (GTK_CHECK_TYPE ((obj), GDAS_TYPE_BENDER))
#define GDAM_IS_BENDER_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAS_TYPE_BENDER))

#define GDAS_BENDER_NUM_POSITION_MARKERS 4

struct _GdasBenderChange {
	int				num_samples;
	GdamFixed			new_rate;
	GdasBenderChange*		next;
	gboolean			is_jump;
};

struct _GdasBenderClass {
	GdasBufferingFilterClass	buffering_filter_class;
};

struct _GdasBender {
	GdasBufferingFilter		buffering_filter;

	/* List of change points. */
	GdasBenderChange*		changes;

	GdasBenderChangeMode		change_mode;

	/* Current position of playback. */
	/* In b16.
	 * it may be outside of the buffers range. */
	gint64				backbuffer_offset;

	/* The position we started from if it was set with
	 * `changes_temporary'.  If keep_time, this is the location
	 * we would have been sans (changes since set_mark or set_mode) 
	 * if keep_in_beat, keep_time position modulo beat_length   */
	gint64				end_offset;

	/* For fading... */
	gint				fade_window;
	gint				fade_remaining;
	GdamSample			fade_sample;
	GdamSample			last_sample;

	/* Number of samples further than current position to
	 * read from subsource. */
	int				readahead;

	/* Base resample rate. */
	GdamFixed			rate;

	/* If this is non-NULL then a fade is pending for the beginning
	 * of the next read. */
	gpointer			pending_fade_pointer;

	/* Number of samples in a beat. */
	gint64				beat_size;

	gint64		position_markers [GDAS_BENDER_NUM_POSITION_MARKERS];

	/* An implementation of spin using bender's seek capability */
	int             spin_start_marker;
	int             spin_end_marker;
	gboolean        spinning_between_position_markers;

	gboolean        dont_extend_change_list;

	/* Smooth rate slides.  Every step_size samples out,
	 * rate = rate * factor + delta */
	/* TODO: sub-sample step size, jumps */
	gint64				slide_step_size;
	gint64				slide_num_steps;
	gdouble				slide_rate_delta;
	gdouble				slide_rate_factor;
	gboolean			sliding;

	/* allow respeed-type morphing */
	gdouble				slide_jump_size;
	gdouble				slide_jump_accumulated_fraction;

	/* immediate resample rate. */
	GdamFixed			current_rate;
};


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
