/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdascompressor.h"

static void gdas_compressor_init(GdasCompressor* compressor)
{
  compressor->threshold = GDAM_SAMPLE_MAX * 7 / 10;
  compressor->ratio = GdamFixed_ONE;
}
static void gdas_compressor_process(GdasDynamicRange*   range,
                                    GdamSample*         out,
                                    const GdamSample*   in,
				    int                 num)
{
  GdasCompressor *compressor = GDAS_COMPRESSOR (range);
  GdamSampleValue max_sample = range->max_sample;
  GdamSampleValue last_max_sample = range->last_max_sample;
  GdamSampleValue threshold;
  gint64 cur_scale;
  gint64 scale_incr;
  GdamFixed scale_1, scale_2;
  GdamFixed ratio = compressor->ratio;
  int i;

  threshold = compressor->threshold;
  if (last_max_sample < threshold && max_sample < threshold)
    {
      gdam_sample_copy (out, in, num);
      return;
    }

  if (last_max_sample > threshold)
    {
      scale_1 = ((gint64) (last_max_sample - threshold) * ratio
		 + ((gint64) threshold * GdamFixed_ONE)) / last_max_sample;
    }
  else
    {
      scale_1 = GdamFixed_ONE;
    }
  if (max_sample > threshold)
    {
      scale_2 = ((gint64) (max_sample - threshold) * ratio
		 + ((gint64) threshold * GdamFixed_ONE)) / max_sample;
    }
  else
    {
      scale_2 = GdamFixed_ONE;
    }

  g_message ("process: scale %6.6g => %6.6g",
	     GDAM_FIXED_TO_DOUBLE (scale_1), GDAM_FIXED_TO_DOUBLE (scale_2));

  cur_scale = (gint64) scale_1 << 16;
  scale_incr = ((gint64) (scale_2 - scale_1) << 16) / num;
  for (i = 0; i < num; i++)
    {
      GdamFixed scale = (cur_scale >> 16);
      gint64 tmp;
      cur_scale += scale_incr;

      tmp = (gint64) in[i].s[0] * scale / GdamFixed_ONE;
      out[i].s[0] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);

      tmp = (gint64) in[i].s[1] * scale / GdamFixed_ONE;
      out[i].s[1] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);
    }
}

static void gdas_compressor_arg_set_ratio(GtkObject* object, GdamArg* arg)
{
  GdasCompressor *compressor = (GdasCompressor *) object;
  gdouble ratio = GDAM_ARG_DOUBLE (arg);
  if (ratio < 0.01)
    {
      gdam_debug ("ratio was too small");
      return;
    }
  compressor->ratio = (GdamFixed) ((gdouble) GdamFixed_ONE / ratio);
}
static void gdas_compressor_arg_get_ratio(GtkObject* object, GdamArg* arg)
{
  GdasCompressor *compressor = (GdasCompressor *) object;
  gdouble ratio = compressor->ratio > 0 ? compressor->ratio : 1;
  GDAM_ARG_DOUBLE (arg) = (gdouble) GdamFixed_ONE / ratio;
}

static void gdas_compressor_arg_set_threshold(GtkObject* object, GdamArg* arg)
{
  GdasCompressor *compressor = (GdasCompressor *) object;
  gdouble threshold = GDAM_ARG_DOUBLE (arg);
  if (threshold < 0.0 || threshold > 1.0)
    {
      gdam_debug ("threshold must be in [0,1]");
      return;
    }
  compressor->threshold = threshold * GDAM_SAMPLE_MAX;
}
static void gdas_compressor_arg_get_threshold(GtkObject* object, GdamArg* arg)
{
  GdasCompressor *compressor = (GdasCompressor *) object;
  gdouble threshold = compressor->threshold;
  threshold /= GDAM_SAMPLE_MAX;
  GDAM_ARG_DOUBLE (arg) = threshold;
}

static void gdas_compressor_class_init(GdasDynamicRangeClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  klass->process = gdas_compressor_process;
  gdam_arg_register (object_class,
		     "threshold",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_compressor_arg_set_threshold,
		     gdas_compressor_arg_get_threshold,
		     NULL, "0,10", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "ratio",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_compressor_arg_set_ratio,
		     gdas_compressor_arg_get_ratio,
		     NULL, "0,10", GDAM_ARG_VITAL);
}

GtkType gdas_compressor_get_type()
{
  static GtkType compressor_type = 0;
  if (!compressor_type)
    {
      static const GtkTypeInfo compressor_info = {
	"GdasCompressor",
	sizeof (GdasCompressor),
	sizeof (GdasCompressorClass),
	(GtkClassInitFunc) gdas_compressor_class_init,
	(GtkObjectInitFunc) gdas_compressor_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_DYNAMIC_RANGE;
      compressor_type = gtk_type_unique (parent, &compressor_info);
    }
  return compressor_type;
}
