/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasdistortion.h"

static void gdas_distortion_init(GdasDistortion* distortion)
{
  distortion->saturation = (GdamSampleProduct) GDAM_SAMPLE_ONE / 8
    * GdamFixed_ONE;
}

static int gdas_distortion_read(GdasSource* source,
                                GdamSample* buf,
				int         max_samples)
{
  GdasDistortion *distortion;
  GdamSample last_sample;
  int num_read;
  int change_whole;
  int change_part;
  int rv;
  int accum;

  distortion = GDAS_DISTORTION (source);
  last_sample = distortion->last_sample;
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buf, max_samples);
  rv = num_read;
  change_whole = (int) (distortion->saturation / GdamFixed_ONE);
  change_part = (int) (distortion->saturation % GdamFixed_ONE);
  accum = distortion->saturation_carry;

  while (num_read-- > 0)
    {
      int delta = change_whole;
      int v;
      int last;
      accum += change_part;
      if (accum >= GdamFixed_ONE)
	{
	  accum -= GdamFixed_ONE;
	  delta++;
	}

      v = buf->s[0];
      last = last_sample.s[0];
      if (v < last - delta)
	v = last - delta;
      else if (v > last + delta)
	v = last + delta;
      last_sample.s[0] = v;
      buf->s[0] = v;

      v = buf->s[1];
      last = last_sample.s[1];
      if (v < last - delta)
	v = last - delta;
      else if (v > last + delta)
	v = last + delta;
      last_sample.s[1] = v;
      buf->s[1] = v;

      buf++;
    }
  distortion->saturation_carry = accum;
  distortion->last_sample = last_sample;

  return rv;
}

static void gdas_distortion_arg_set_saturation(GtkObject* object, GdamArg* arg)
{
  gdouble saturation = GDAM_ARG_DOUBLE (arg);
  //saturation /= 44100.0;
  if (saturation > 1.0)
    saturation = 1.0;
  saturation *= GDAM_SAMPLE_ONE;
  if (saturation < 1.0)
    saturation = 1.0;
  GDAS_DISTORTION (object)->saturation = saturation * GdamFixed_ONE;
}
static void gdas_distortion_arg_get_saturation(GtkObject* object, GdamArg* arg)
{
  gdouble saturation = GDAS_DISTORTION (object)->saturation;
  saturation /= GdamFixed_ONE;
  saturation /= GDAM_SAMPLE_ONE;
  //saturation *= 44100.0;
  GDAM_ARG_DOUBLE (arg) = saturation;
}


static void gdas_distortion_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (source_class);
  source_class->read = gdas_distortion_read;
  gdam_arg_register (object_class,
		     "saturation",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_distortion_arg_set_saturation,
		     gdas_distortion_arg_get_saturation,
		     NULL, "0,10000", GDAM_ARG_VITAL);
}

GtkType gdas_distortion_get_type()
{
  static GtkType distortion_type = 0;
  if (!distortion_type)
    {
      static const GtkTypeInfo distortion_info = {
	"GdasDistortion",
	sizeof (GdasDistortion),
	sizeof (GdasDistortionClass),
	(GtkClassInitFunc) gdas_distortion_class_init,
	(GtkObjectInitFunc) gdas_distortion_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      distortion_type = gtk_type_unique (parent, &distortion_info);
    }
  return distortion_type;
}
