/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasequalizer.h"
#include <gdam/common/gdamplugin.h>
#include <math.h>
#include <string.h>

#define S	1.0			/* shelving parameter */

#ifndef M_2_PI
# define M_2_PI         0.63661977236758134308  /* 2/pi */
#endif
#ifndef M_LN2
# define M_LN2          0.69314718055994530942  /* log_e 2 */
#endif
#ifndef M_LN10
# define M_LN10         2.30258509299404568402  /* log_e 10 */
#endif


static void compute_coefficients(GdasEqualizer* equalizer)
{
  gdouble frequency = equalizer->frequency;
  gdouble bandwidth = equalizer->bandwidth;
  gdouble db_gain = equalizer->db_gain;
  int type = equalizer->type;
  gdouble A = 0.0, omega, sn, cs, alpha, beta = 0.0;
  gdouble a0, a1, a2, b0, b1, b2;

  omega = M_2_PI * frequency / 44100.0;
  sn = sin (omega);
  cs = cos (omega);
  alpha = sn * sinh (M_LN2 / 2.0 * bandwidth * omega / sn);

  if (type == GDAS_EQUALIZER_PEAKING_EQ
      || type == GDAS_EQUALIZER_LOW_SHELF
      || type == GDAS_EQUALIZER_HIGH_SHELF)
    {
      A = exp (db_gain * M_LN10 / 40.0);
      beta = sqrt ((A * A + 1.0) / S - ((A - 1.0) * (A - 1.0)));
    }

  switch (type)
    {
    case GDAS_EQUALIZER_LOW_PASS:
      b0 = (1.0 - cs) / 2.0;
      b1 = 1.0 - cs;
      b2 = (1.0 - cs) / 2.0;
      a0 = 1.0 + alpha;
      a1 = -2.0 * cs;
      a2 = 1.0 - alpha;
      break;
    case GDAS_EQUALIZER_HIGH_PASS:
      b0 = (1.0 + cs) / 2.0;
      b1 = -(1.0 + cs);
      b2 = (1.0 + cs) / 2.0;
      a0 = 1.0 + alpha;
      a1 = -2.0 * cs;
      a2 = 1.0 - alpha;
      break;
    case GDAS_EQUALIZER_BAND_PASS:
      b0 = alpha;
      b1 = 0;
      b2 = -alpha;
      a0 = 1.0 + alpha;
      a1 = -2.0 * cs;
      a2 = 1.0 - alpha;
      break;
    case GDAS_EQUALIZER_NOTCH:
      b0 = 1.0;
      b1 = -2.0 * cs;
      b2 = 1.0;
      a0 = 1.0 + alpha;
      a1 = -2.0 * cs;
      a2 = 1.0 - alpha;
      break;
    case GDAS_EQUALIZER_PEAKING_EQ:
      b0 = 1.0 + alpha * A;
      b1 = -2.0 * cs;
      b2 = 1.0 - alpha * A;
      a0 = 1.0 + alpha / A;
      a1 = -2.0 * cs;
      a2 = 1.0 - alpha / A;
      break;
    case GDAS_EQUALIZER_LOW_SHELF:
      b0 = A * ((A + 1.0) - (A - 1.0) * cs + beta * sn);
      b1 = 2.0 * A * ((A - 1.0) - (A + 1.0) * cs);
      b2 = A * ((A + 1.0) - (A - 1.0) * cs - beta * sn);
      a0 = (A + 1.0) + (A - 1.0) * cs + beta * sn;
      a1 = -2.0 * ((A - 1.0) + (A + 1.0) * cs);
      a2 = (A + 1.0) + (A - 1.0) * cs - beta * sn;
      break;
    case GDAS_EQUALIZER_HIGH_SHELF:
      b0 = A * ((A + 1.0) + (A - 1.0) * cs + beta * sn);
      b1 = -2.0 * A * ((A - 1.0) + (A + 1.0) * cs);
      b2 = A * ((A + 1.0) + (A - 1.0) * cs - beta * sn);
      a0 = (A + 1.0) - (A - 1.0) * cs + beta * sn;
      a1 = 2.0 * ((A - 1.0) - (A + 1.0) * cs);
      a2 = (A + 1.0) - (A - 1.0) * cs - beta * sn;
      break;
    default:
      gdam_debug ("bad equalizer type: %d", type);
      return;
    }

  gdam_debug ("a012=%g,%g,%g; b012=%g,%g,%g", a0, a1, a2, b0, b1, a2);

  equalizer->x_coeff[0] = (GdamFixed) (+b0 / a0 * GdamFixed_ONE);
  equalizer->x_coeff[1] = (GdamFixed) (+b1 / a0 * GdamFixed_ONE);
  equalizer->x_coeff[2] = (GdamFixed) (+b2 / a0 * GdamFixed_ONE);
  equalizer->y_coeff[1] = (GdamFixed) (-a1 / a0 * GdamFixed_ONE);
  equalizer->y_coeff[2] = (GdamFixed) (-a2 / a0 * GdamFixed_ONE);
  gdam_debug ("equalizer: x_coeffs = [%g, %g, %g], "
	      "y_coeffs = [%g, %g, %g]",
	      +b0 / a0, +b1 / a0, +b2 / a0, 0.0, -a1 / a0, -a2 / a0);

}

static void gdas_equalizer_arg_set_frequency(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  equalizer->frequency = GDAM_ARG_DOUBLE (arg);
  compute_coefficients (equalizer);
}
static void gdas_equalizer_arg_get_frequency(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  GDAM_ARG_DOUBLE (arg) = equalizer->frequency;
}
static void gdas_equalizer_arg_set_effect(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  equalizer->effect = GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE;
  equalizer->effect = CLAMP (equalizer->effect, 0, GdamFixed_ONE);
}
static void gdas_equalizer_arg_get_effect(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  GDAM_ARG_DOUBLE (arg) = (gdouble) equalizer->effect / GdamFixed_ONE;
}
static void gdas_equalizer_arg_set_bandwidth(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  equalizer->bandwidth = GDAM_ARG_DOUBLE (arg);
  compute_coefficients (equalizer);
}
static void gdas_equalizer_arg_get_bandwidth(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  GDAM_ARG_DOUBLE (arg) = equalizer->bandwidth;
}
static void gdas_equalizer_arg_set_db_gain(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  equalizer->db_gain = GDAM_ARG_DOUBLE (arg);
  compute_coefficients (equalizer);
}
static void gdas_equalizer_arg_get_db_gain(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  GDAM_ARG_DOUBLE (arg) = equalizer->db_gain;
}
static void gdas_equalizer_arg_set_mode(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  const char *mode = GDAM_ARG_STRING (arg);
  if (g_strcasecmp (mode, "lowpass") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_LOW_PASS;
    }
  else if (g_strcasecmp (mode, "highpass") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_HIGH_PASS;
    }
  else if (g_strcasecmp (mode, "bandpass") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_BAND_PASS;
    }
  else if (g_strcasecmp (mode, "peakingeq") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_PEAKING_EQ;
    }
  else if (g_strcasecmp (mode, "notch") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_NOTCH;
    }
  else if (g_strcasecmp (mode, "lowshelf") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_LOW_SHELF;
    }
  else if (g_strcasecmp (mode, "highshelf") == 0)
    {
      equalizer->type = GDAS_EQUALIZER_HIGH_SHELF;
    }
  else
    {
      g_warning ("unknowning equalizing mode %s", mode);
      return;
    }
  compute_coefficients (equalizer);
}

static void gdas_equalizer_arg_get_mode(GtkObject* object, GdamArg* arg)
{
  GdasEqualizer *equalizer = GDAS_EQUALIZER (object);
  const char *mode;

  switch (equalizer->type)
    {
    case GDAS_EQUALIZER_LOW_PASS:
      mode = "lowpass";
      break;
    case GDAS_EQUALIZER_HIGH_PASS:
      mode = "highpass";
      break;
    case GDAS_EQUALIZER_BAND_PASS:
      mode = "bandpass";
      break;
    case GDAS_EQUALIZER_PEAKING_EQ:
      mode = "peakingeq";
      break;
    case GDAS_EQUALIZER_NOTCH:
      mode = "notch";
      break;
    case GDAS_EQUALIZER_LOW_SHELF:
      mode = "lowshelf";
      break;
    case GDAS_EQUALIZER_HIGH_SHELF:
      mode = "highshelf";
      break;
    default:
      mode = "unknown";
    }
  GDAM_ARG_STRING (arg) = g_strdup (mode);
}

static inline void process_sample(GdasEqualizer* equalizer, GdamSample* in_out)
{
  int y = 0;
  int in0, in1;
  GdamSample *x_arr = equalizer->x;
  GdamSample *y_arr = equalizer->y;
  GdamFixed *x_coeff = equalizer->x_coeff;
  GdamFixed *y_coeff = equalizer->y_coeff;
  GdamSample out_eq;

  in0 = ((GdamSampleProduct) in_out->s[0] * equalizer->effect)
    / GdamFixed_ONE;
  in1 = ((GdamSampleProduct) in_out->s[1] * equalizer->effect)
    / GdamFixed_ONE;

  y = ((gint64) x_coeff[0] * in0
       + (gint64) x_coeff[1] * x_arr[0].s[0]
       + (gint64) x_coeff[2] * x_arr[1].s[0]
       + (gint64) y_coeff[1] * y_arr[0].s[0]
       + (gint64) y_coeff[2] * y_arr[1].s[0]) / GdamFixed_ONE;
  out_eq.s[0] = GDAM_SAMPLE_CLAMP (y);

  y += ((int) in_out->s[0] - in0);
  in_out->s[0] = GDAM_SAMPLE_CLAMP (y);

  y = ((gint64) x_coeff[0] * in1
       + (gint64) x_coeff[1] * x_arr[0].s[1]
       + (gint64) x_coeff[2] * x_arr[1].s[1]
       + (gint64) y_coeff[1] * y_arr[0].s[1]
       + (gint64) y_coeff[2] * y_arr[1].s[1]) / GdamFixed_ONE;
  out_eq.s[1] = GDAM_SAMPLE_CLAMP (y);

  y += ((int) in_out->s[1] - in1);
  in_out->s[1] = GDAM_SAMPLE_CLAMP (y);

  x_arr[1] = x_arr[0];
  y_arr[1] = y_arr[0];
  x_arr[0].s[0] = GDAM_SAMPLE_CLAMP (in0);
  x_arr[0].s[1] = GDAM_SAMPLE_CLAMP (in1);
  y_arr[0] = out_eq;
}

static int gdas_equalizer_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  GdasEqualizer *equalizer = GDAS_EQUALIZER (source);
  int i, num_read;
  if (filtered == NULL)
    return 0;
  num_read = gdas_source_read (filtered, buf, n);
  for (i = 0; i < num_read; i++)
    process_sample (equalizer, buf++);
  return num_read;
}

static void gdas_equalizer_class_init(GtkObjectClass* object_class)
{
  GDAS_SOURCE_CLASS (object_class)->read = gdas_equalizer_read;
  gdam_arg_register (object_class,
		     "frequency",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_equalizer_arg_set_frequency,
		     gdas_equalizer_arg_get_frequency,
		     NULL, "1,44100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "effect",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_equalizer_arg_set_effect,
		     gdas_equalizer_arg_get_effect,
		     NULL, "0,3", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "bandwidth",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_equalizer_arg_set_bandwidth,
		     gdas_equalizer_arg_get_bandwidth,
		     NULL, "-10,10", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "db_gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_equalizer_arg_set_db_gain,
		     gdas_equalizer_arg_get_db_gain,
		     NULL, "0,3", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mode",
		     GDAM_ARG_TYPE_STRING,
		     gdas_equalizer_arg_set_mode,
		     gdas_equalizer_arg_get_mode, NULL, NULL, GDAM_ARG_VITAL);
}
static void gdas_equalizer_init(GtkObject* object)
{
  GdasEqualizer *equalizer = (GdasEqualizer *) object;
  equalizer->type = GDAS_EQUALIZER_HIGH_PASS;
  equalizer->frequency = 440.0;
  equalizer->bandwidth = 1.0;
  equalizer->db_gain = 1.0;

  equalizer->effect = GdamFixed_ONE;

  /* Never used, but we zero it out initially then never touch it... */
  equalizer->y_coeff[0] = 0;
  equalizer->x_coeff[0] = GdamFixed_ONE;

  /*compute_coefficients(equalizer); */
}

GtkType gdas_equalizer_get_type()
{
  static GtkType equalizer_type = 0;
  if (!equalizer_type)
    {
      static const GtkTypeInfo equalizer_info = {
	"GdasEqualizer",
	sizeof (GdasEqualizer),
	sizeof (GdasFilterClass),
	(GtkClassInitFunc) gdas_equalizer_class_init,
	(GtkObjectInitFunc) gdas_equalizer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      equalizer_type = gtk_type_unique (GDAS_TYPE_FILTER, &equalizer_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return equalizer_type;
}
