/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasgate.h"

static void gdas_gate_init(GdasGate* gate)
{
  gate->threshold = GDAM_SAMPLE_ONE * 1 / 10;
  gate->last_passthrough = FALSE;
}
static void gdas_gate_process(GdasDynamicRange*   range,
                                    GdamSample*         out,
                                    const GdamSample*   in,
				    int                 num)
{
  GdasGate *gate = GDAS_GATE (range);
  GdamSampleValue max_sample = range->max_sample;
  GdamSampleValue last_max_sample = range->last_max_sample;
  GdamSampleValue threshold = gate->threshold;
  gint64 cur_scale;
  gint64 scale_incr;
  GdamFixed scale_1, scale_2;
  int i;
  /* We used to check (last_max_sample >= threshold), but last_max_sample
   * always had the same state as max_sample and we were missing state 
   * changes.  now we save our own state between calls to process. */
  gboolean gate_last_passthrough = gate->last_passthrough;
  gboolean last_passthrough = (last_max_sample >= threshold);
  gboolean passthrough = (max_sample >= threshold);

  if (gate->inverted)
    {
      last_passthrough = !last_passthrough;
      passthrough = !passthrough;
    }

  if (passthrough && last_passthrough)
    {
      gdam_sample_copy (out, in, num);
      return;
    }

  scale_1 = last_passthrough ? GdamFixed_ONE : 0;
  scale_2 = passthrough ? GdamFixed_ONE : 0;

  cur_scale = (gint64) scale_1 << 16;
  scale_incr = ((gint64) (scale_2 - scale_1) << 16) / num;

  for (i = 0; i < num; i++)
    {
      GdamFixed scale = (cur_scale >> 16);
      gint64 tmp;
      cur_scale += scale_incr;

      tmp = (gint64) in[i].s[0] * scale / GdamFixed_ONE;
      out[i].s[0] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);

      tmp = (gint64) in[i].s[1] * scale / GdamFixed_ONE;
      out[i].s[1] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);
    }
  gate->last_passthrough = passthrough;
}

static void gdas_gate_arg_set_threshold(GtkObject* object, GdamArg* arg)
{
  GdasGate *gate = (GdasGate *) object;
  gdouble threshold = GDAM_ARG_DOUBLE (arg);
  if (threshold < 0.0 || threshold > 1.0)
    {
      gdam_debug ("threshold must be in [0,1]");
      return;
    }
  gate->threshold = threshold * GDAM_SAMPLE_ONE;
}
static void gdas_gate_arg_get_threshold(GtkObject* object, GdamArg* arg)
{
  GdasGate *gate = (GdasGate *) object;
  gdouble threshold = gate->threshold;
  threshold /= GDAM_SAMPLE_ONE;
  GDAM_ARG_DOUBLE (arg) = threshold;
}
static void gdas_gate_arg_set_inverted(GtkObject* object, GdamArg* arg)
{
  GdasGate *gate = (GdasGate *) object;
  gate->inverted = GDAM_ARG_INT (arg);
}
static void gdas_gate_arg_get_inverted(GtkObject* object, GdamArg* arg)
{
  GdasGate *gate = (GdasGate *) object;
  GDAM_ARG_INT (arg) = gate->inverted;
}


static void gdas_gate_class_init(GdasDynamicRangeClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  klass->process = gdas_gate_process;
  gdam_arg_register (object_class,
		     "threshold",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_gate_arg_set_threshold,
		     gdas_gate_arg_get_threshold,
		     NULL, "0,10", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "inverted",
		     GDAM_ARG_TYPE_INT,
		     gdas_gate_arg_set_inverted,
		     gdas_gate_arg_get_inverted,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

GtkType gdas_gate_get_type()
{
  static GtkType gate_type = 0;
  if (!gate_type)
    {
      static const GtkTypeInfo gate_info = {
	"GdasGate",
	sizeof (GdasGate),
	sizeof (GdasGateClass),
	(GtkClassInitFunc) gdas_gate_class_init,
	(GtkObjectInitFunc) gdas_gate_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_DYNAMIC_RANGE;
      gate_type = gtk_type_unique (parent, &gate_info);
    }
  return gate_type;
}
