/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 * 
 * Code to implement the server-side of the gdamdsp
 * protocol [which encapsulates the data + ioctls of a raw audio
 * device]
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include "../gdassource.h"
#include "gdasgdamdsp.h"
#include "../gdamcleanup.h"
#include <gtk/gtksignal.h>
#include <gdam/common/gdambuffer.h>
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamsampleconverter.h>
#include "../gdamconnection.h"
#include "../gdamevent.h"
#include "../gdamserverdebug.h"
#include "../gdamrealtime.h"
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "../helper/gdamdsppreload.h"


static guint accept_signal_id = 0;
static guint accept_event_id = 0;

#ifndef MAX_UNIX_PATH
#define MAX_UNIX_PATH		108		/* er, linux constant */
#endif


#define MAX_PACKET_SIZE		(1024 * 1024)

/* Wait until 1/8 the maximum buffer is full,
 * or MAX_TO_WAIT_FOR samples are ready
 * before saying the source is ready to read. */
#define FRACTION_TO_WAIT_FOR	4
#define MAX_TO_WAIT_FOR		8192
#define GDAS_GDAM_DSP_PRIORITY       G_PRIORITY_DEFAULT

static void        gdas_gdam_dsp_class_init       (GdasSourceClass* klass);
static void        gdas_gdam_dsp_init             (GdasGdamDsp*    gdam_dsp);
static void        gdas_gdam_dsp_destroy          (GdasGdamDsp*    source);

static GdamTime    gdas_gdam_dsp_get_position     (GdasSource*      source);
static gboolean    gdas_gdam_dsp_is_ready_to_read (GdasSource*      source);
static int	   gdas_gdam_dsp_read             (GdasSource*      input,
					           GdamSample*      samples,
					           int              max_read);
static void        gdas_gdam_dsp_miss_data        (GdasSource*      input,
					           int              lost);
static gboolean    gdas_gdam_dsp_prepare          (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data);
static gboolean    gdas_gdam_dsp_check            (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);
static gboolean    gdas_gdam_dsp_dispatch         (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data);

static void gdas_gdam_dsp_func_begin_listening(GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out);
static void gdas_gdam_dsp_func_stop_listening (GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out);
static void gdas_gdam_dsp_func_subscribe_accept(GtkObjectClass* object_class,
				                int             num_args_in,
                                                GdamArg*        arguments_in,
				                int*            num_args_out,
				                GdamArg**       arguments_out);
static void gdas_gdam_dsp_func_maybe_accept   (GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out);

static inline void add_poll_fd(GdasGdamDsp* gdam_dsp)
{
  g_main_add_poll (&gdam_dsp->poll_fd, G_PRIORITY_DEFAULT);
  gdam_dsp->has_poll_fd = TRUE;
}
static inline void remove_poll_fd(GdasGdamDsp* gdam_dsp)
{
  g_main_remove_poll (&gdam_dsp->poll_fd);
  gdam_dsp->has_poll_fd = FALSE;
}



static GtkObjectClass*  parent_class = NULL;

GtkType gdas_gdam_dsp_get_type()
{
  static GtkType gdam_dsp_type = 0;
  if (!gdam_dsp_type)
    {
      static const GtkTypeInfo gdam_dsp_info = {
	"GdasGdamDsp",
	sizeof (GdasGdamDsp),
	sizeof (GdasGdamDspClass),
	(GtkClassInitFunc) gdas_gdam_dsp_class_init,
	(GtkObjectInitFunc) gdas_gdam_dsp_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      gdam_dsp_type = gtk_type_unique (GDAS_TYPE_SOURCE, &gdam_dsp_info);
    }
  return gdam_dsp_type;
}

static void gdas_gdam_dsp_set_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (object);
  gdam_dsp->max_buffer_size = GDAM_ARG_INT (arg);
}
static void gdas_gdam_dsp_get_buffer_size(GtkObject* object, GdamArg* arg)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (object);
  GDAM_ARG_INT (arg) = gdam_dsp->max_buffer_size;
}

static void gdas_gdam_dsp_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = (GtkObjectClass *) source_class;
  GdasGdamDspClass *gdam_dsp_class = GDAS_GDAM_DSP_CLASS (source_class);
  if (!parent_class)
    parent_class = gtk_type_class (GDAS_TYPE_SOURCE);
  source_class->read = gdas_gdam_dsp_read;
  source_class->miss_data = gdas_gdam_dsp_miss_data;
  source_class->is_ready_to_read = gdas_gdam_dsp_is_ready_to_read;
  source_class->get_position = gdas_gdam_dsp_get_position;
  object_class->destroy = (void (*)(GtkObject *)) gdas_gdam_dsp_destroy;
  gdam_arg_class_function_register (object_class,
				    "maybe_accept",
				    gdas_gdam_dsp_func_maybe_accept);
  gdam_arg_class_function_register (object_class,
				    "subscribe_accept",
				    gdas_gdam_dsp_func_subscribe_accept);
  gdam_arg_class_function_register (object_class,
				    "stop_listening",
				    gdas_gdam_dsp_func_stop_listening);
  gdam_arg_class_function_register (object_class,
				    "start_listening",
				    gdas_gdam_dsp_func_begin_listening);

  gdam_arg_register (object_class,
		     "buffer_size",
		     GDAM_ARG_TYPE_INT,
		     gdas_gdam_dsp_set_buffer_size,
		     gdas_gdam_dsp_get_buffer_size,
		     NULL, "0,", GDAM_ARG_VITAL);
}

static int default_buffer_size = 64 * 1024;

int gdas_gdam_dsp_get_default_buffer_size()
{
  return default_buffer_size;
}
void gdas_gdam_dsp_set_default_buffer_size(int buffer_size)
{
  default_buffer_size = buffer_size;
}

static void gdas_gdam_dsp_init(GdasGdamDsp* gdam_dsp)
{
  int src_bits_per_sample;
  gboolean is_big_endian = (G_BYTE_ORDER == G_BIG_ENDIAN);
  gdam_dsp->process_id = -1;
  gdam_dsp->poll_fd.fd = -1;
  gdam_dsp->poll_fd.events = G_IO_IN | G_IO_ERR | G_IO_HUP;
  gdam_dsp->reader_source = 0;
  gdam_dsp->skipped = 0;
  gdam_dsp->max_buffer_size = gdas_gdam_dsp_get_default_buffer_size ();
  gdam_dsp->num_channels = 2;
  gdam_dsp->bytes_per_sample = 2;
  src_bits_per_sample = gdam_dsp->bytes_per_sample * 8;
  gdam_dsp->sample_converter = gdam_sample_converter_get
    (src_bits_per_sample, TRUE, is_big_endian,
     GDAM_SAMPLE_BITS, TRUE, is_big_endian);
  gdam_buffer_construct (&gdam_dsp->buffer);
}

static GSourceFuncs gdam_dsp_source_funcs = {
	gdas_gdam_dsp_prepare,
	gdas_gdam_dsp_check,
	gdas_gdam_dsp_dispatch,
	NULL
};

/*=====================================================================*
 *                       gdas_gdam_dsp_has_opened                      *
 *                                                                     *
 * Tell the gdam_dsp to start reading off a given                      *
 * file descriptor.                                                    *
 *                                                                     *
 * Parameters:                                                         *
 *     gdam_dsp          the source to affect.                         *
 *     fd                the file descriptor to read from              *
 *=====================================================================*/
void gdas_gdam_dsp_has_opened(GdasGdamDsp* gdam_dsp, int fd)
{
  g_return_if_fail (gdam_dsp->poll_fd.fd == -1);
  g_return_if_fail (gdam_dsp->poll_fd.events ==
		    (G_IO_IN | G_IO_ERR | G_IO_HUP));
  g_return_if_fail (gdam_dsp->skipped == 0);
  g_return_if_fail (gdam_dsp->reader_source == 0);

  gdam_dsp->poll_fd.fd = fd;

  gdam_dsp->reader_source = g_source_add (GDAS_GDAM_DSP_PRIORITY,
					  TRUE,
					  &gdam_dsp_source_funcs,
					  NULL, gdam_dsp, NULL);
}
static gboolean set_packet_length(GdasGdamDsp* gdam_dsp, int new_len)
{
  if (new_len < 4)
    return FALSE;
  if (new_len > MAX_PACKET_SIZE)
    return FALSE;
  gdam_dsp->body_length = new_len - 4;
  if (gdam_dsp->body_length + 1 > gdam_dsp->body_max_size)
    {
      gdam_dsp->body_max_size = gdam_dsp->body_length + 1;
      gdam_dsp->body = g_renew (char, gdam_dsp->body,
				gdam_dsp->body_max_size);
    }
    return TRUE;
}

static void
set_input_format(GdasGdamDsp* gdam_dsp,
                 gboolean     is_signed,
		 int          num_bits,
		 gboolean     is_big_endian)
{
  gdam_dsp->bytes_per_sample = num_bits / 8;
  gdam_dsp->is_signed = is_signed ? 1 : 0;
  gdam_dsp->is_big_endian = is_big_endian ? 1 : 0;
  gdam_dsp->sample_converter =
    gdam_sample_converter_get (num_bits, is_signed, is_big_endian,
			       GDAM_SAMPLE_BITS,
			       TRUE, (G_BYTE_ORDER == G_BIG_ENDIAN));
}
static void
set_input_num_channels(GdasGdamDsp*  gdam_dsp,
                       int           num_channels)
{
  gdam_dsp->num_channels = num_channels;
  gdam_dsp->channel_converter
    = gdam_sample_converter_get_channel_changer (num_channels, 2);
}
static gboolean process_ioctl(GdasGdamDsp* gdam_dsp,
                              char**       args)
{
  if (strcmp (args[0], "reset") == 0)
    {
      return TRUE;
    }
  else if (strcmp (args[0], "post") == 0)
    {
      return TRUE;
    }
  else if (strcmp (args[0], "setfmt") == 0)
    {
      const char *fmt = args[1];
      gboolean is_signed;
      int num_bits;
      gboolean is_big_endian = (G_BYTE_ORDER == G_BIG_ENDIAN);
      if (fmt == NULL)
	return FALSE;
      is_signed = (fmt[0] == 's');
      fmt++;
      num_bits = atoi (fmt);
      if (num_bits != 8 && num_bits != 16)
	return FALSE;
      while (*fmt && isdigit (*fmt))
	fmt++;
      if (*fmt == 'b')
	is_big_endian = TRUE;
      else if (*fmt == 'l')
	is_big_endian = FALSE;
      set_input_format (gdam_dsp, is_signed, num_bits, is_big_endian);
      return TRUE;
    }
  else if (strcmp (args[0], "speed") == 0)
    {
      if (args[1] != NULL)
	gdam_dsp->rate = atoi (args[1]);
    }
  else if (strcmp (args[0], "channels") == 0)
    {
      int num_channels;
      if (args[1] == NULL)
	return FALSE;
      num_channels = atoi (args[1]);
      if (num_channels <= 0 || num_channels > 4)
	return FALSE;
      set_input_num_channels (gdam_dsp, num_channels);
      return TRUE;
    }
  else
    {
      gdam_debug ("unexpected error from gdamdsp client");
    }
  return FALSE;
}
static void process_write(GdasGdamDsp* gdam_dsp,
                          char*        body,
			  int          length)
{
  int num_samples;
  void *buf;
  int buf_size;
  int sample_size = gdam_dsp->bytes_per_sample;
  int num_channels = gdam_dsp->num_channels;
  GdamSampleConverter sample_converter = gdam_dsp->sample_converter;
  GdamSampleConverter channel_converter = gdam_dsp->channel_converter;
  int sc_size = sample_size * num_channels;
  num_samples = length / sc_size;
  buf_size = MAX ((int)sizeof (GdamSampleValue), sample_size)
    * MAX (2, num_channels) * num_samples;
  buf = alloca (buf_size);
  if (sample_converter != NULL)
    {
      (*sample_converter) (buf, body, num_samples * num_channels);
      if (channel_converter != NULL)
	(*channel_converter) (buf, buf, num_samples);
    }
  else
    {
      if (channel_converter != NULL)
	(*channel_converter) (buf, body, num_samples);
      else
	gdam_sample_copy (buf, body, num_samples);
    }
}

static gboolean process_packet(GdasGdamDsp* gdam_dsp)
{
  char *body = gdam_dsp->body;
  int length = gdam_dsp->body_length;
  gint32 packet_type;
  if (length < 4)
    {
      gdam_debug ("gdamdsp: process_packet: packet too short");
      return FALSE;
    }
  packet_type = *(gint32 *) body;
  packet_type = GINT32_FROM_BE (packet_type);

  body += 4;
  length -= 4;

  /* this is why the + 1 is packet_set_length is needed;
   * we often want the buffer NUL terminated */
  body[length] = '\0';

  switch (packet_type)
    {
    case GDAMDSP_OPEN:
      gdam_dsp->remote_open = TRUE;
      break;
    case GDAMDSP_IOCTL:
      {
	char **splitstrs;
	splitstrs = g_strsplit (body, ":", 0);
	if (!process_ioctl (gdam_dsp, splitstrs))
	  {
	    g_strfreev (splitstrs);
	    return FALSE;
	  }
	g_strfreev (splitstrs);
	break;
      }
    case GDAMDSP_WRITE:
      process_write (gdam_dsp, body, length);
      break;
    case GDAMDSP_CLOSE:
      gdam_dsp->remote_open = FALSE;
      break;
    default:
      {
	gdam_debug ("gdamdsp: process_packet: "
		    "unexpected packet type %d", packet_type);
	return FALSE;
      }
    }
  return TRUE;
}

static int process_piece     (GdasGdamDsp*      gdam_dsp,
                              char*             buf,
			      int               length)
{
  if (length <= 0)
    return length;

  if (gdam_dsp->header_length < 4)
    {
      int curlen = gdam_dsp->header_length;
      int remaining = 4 - curlen;
      int to_copy;
      to_copy = MIN (remaining, length);
      memcpy (&gdam_dsp->length_union.buf[curlen], buf, to_copy);
      gdam_dsp->header_length += to_copy;
      if (gdam_dsp->header_length == 4)
	{
	  guint32 new_len = gdam_dsp->length_union.big_endian_num;
	  new_len = GUINT32_FROM_BE (new_len);
	  /* sets `body_length' and ensures `body' is large
	   * enough */
	  if (!set_packet_length (gdam_dsp, new_len))
	    return -1;
	}
      return to_copy;
    }

  if (length > gdam_dsp->body_length - gdam_dsp->cur_length)
    length = gdam_dsp->body_length - gdam_dsp->cur_length;
  memcpy (gdam_dsp->body + gdam_dsp->cur_length, buf, length);
  gdam_dsp->cur_length += length;
  if (gdam_dsp->cur_length == gdam_dsp->body_length)
    {
      if (!process_packet (gdam_dsp))
	{
	  return -1;
	}
      gdam_dsp->cur_length = 0;
      gdam_dsp->header_length = 0;
    }

  return length;

}

static gboolean process_input(GdasGdamDsp*      gdam_dsp,
                              char*             buf,
			      int               length)
{
  int rv;
  while (length > 0)
    {
      rv = process_piece (gdam_dsp, buf, length);
      if (rv <= 0)
	return FALSE;
      buf += rv;
      length -= rv;
    }
  g_return_val_if_fail (length == 0, FALSE);
  return TRUE;
}

static gboolean do_one_read(GdasGdamDsp* gdam_dsp)
{
  int n;
  char buf[32 * 1024];
redo_read:
  n = read (gdam_dsp->poll_fd.fd, buf, sizeof (buf));
  if (n < 0)
    {
      if (errno == EINTR)
	goto redo_read;
      if (errno == EWOULDBLOCK)
	return TRUE;
      perror ("reading from gdam_dsp source");
      gdam_dsp->reader_source = 0;
      remove_poll_fd (gdam_dsp);
      return FALSE;
    }
  if (n == 0)
    {
      gdam_debug ("end-of-file from gdam_dsp source "
		  "[gsource %d, %p,%d]",
		  gdam_dsp->reader_source, gdam_dsp,
		  GDAM_SERVER_OBJECT (gdam_dsp)->global_id);
      gdam_dsp->reader_source = 0;
      remove_poll_fd (gdam_dsp);
      return FALSE;
    }

  return process_input (gdam_dsp, buf, n);
}

/* This function tells whether we know we need servicing 
 * before doing the poll.  It is also a good time to 
 * set up the PollFD.
 *
 * We always only need servicing if the fd is readable,
 * so we always return FALSE.
 *
 * But we do determine whether to poll at all via setting
 * gdam_dsp->poll_fd.events appropriately. */
static gboolean
gdas_gdam_dsp_prepare(gpointer  source_data,
                      GTimeVal* curtime,
                      gint*     timeout,
		      gpointer  user_data)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (user_data);

  g_return_val_if_fail (!GTK_OBJECT_DESTROYED (gdam_dsp), FALSE);

  /* We never have anything to do unless the source is ready. */
  *timeout = -1;

  /* Returning events == 0 means we have nothing to process. */
  gdam_dsp->poll_fd.events = gdam_dsp->poll_fd.revents = 0;

  /* We shouldn't be added as a source unless we have a valid
   * file descriptor... */
  g_return_val_if_fail (gdam_dsp->poll_fd.fd >= 0, FALSE);

  /* Set up the read fd only if our buffer isn't
   * too full already. */
  if (gdam_dsp->buffer.size >= gdam_dsp->max_buffer_size)
    {
      g_return_val_if_fail (gdam_dsp->skipped == 0, FALSE);
      gdam_dsp->poll_fd.events = 0;
      return FALSE;
    }

  if (!gdam_dsp->has_poll_fd)
    add_poll_fd (gdam_dsp);

  gdam_dsp->poll_fd.events = G_IO_IN | G_IO_ERR | G_IO_HUP;
  return FALSE;
}


/* Return whether we need servicing.
 * This is simple-- we just test revents. */
static gboolean
gdas_gdam_dsp_check(gpointer  source_data,
                    GTimeVal* curtime,
		    gpointer  user_data)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (user_data);
  g_return_val_if_fail (gdam_dsp->poll_fd.fd >= 0, FALSE);
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdas_gdam_dsp_check: revents=%d", gdam_dsp->poll_fd.revents);
  g_message ("gdas_gdam_dsp_check: revents=%d", gdam_dsp->poll_fd.revents);
  return gdam_dsp->poll_fd.revents ? TRUE : FALSE;
}

static gboolean
gdas_gdam_dsp_dispatch(gpointer  source_data,
                       GTimeVal* curtime,
		       gpointer  user_data)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (user_data);
  return do_one_read (gdam_dsp);
}

static gboolean
gdas_gdam_dsp_is_ready_to_read(GdasSource* source)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (source);
  int min_size = 1;
  int sample_size = gdam_dsp->num_channels * gdam_dsp->bytes_per_sample;

  /* Sources with no readers are either buffered,
   * or at eof.  Never blocking on anything at least... */
  if (gdam_dsp->reader_source == 0)
    return TRUE;

  if (gdam_dsp->buffer.size < sample_size)
    return FALSE;

  /* TEST: don't say we're ready to read unless we have 
   * a 1/FRACTION_TO_WAIT_FOR our buffer
   * or MAX_TO_WAIT_FOR samples, whichever is
   * lower. */
  if (!gdam_dsp->has_been_ready)
    {
      min_size = gdam_dsp->max_buffer_size / FRACTION_TO_WAIT_FOR;
      min_size = MIN (min_size, MAX_TO_WAIT_FOR);
      if (gdam_dsp->buffer.size >= sample_size * min_size)
	{
	  gdam_dsp->has_been_ready = 1;
	}
    }

  return (gdam_dsp->buffer.size >= sample_size * min_size);
}

static int
gdas_gdam_dsp_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (source);
  int num_channels;
  int rv = 0;
  int sample_size;

  if (gdam_dsp->buffer.size < (int) sizeof (GdamSample))
    {
      if (gdam_dsp->reader_source == 0)
	return rv;

      if (rv > 0)
	return rv;

      bzero (buf, sizeof (GdamSample));
      gdas_gdam_dsp_miss_data (source, 1);
      return rv + 1;
    }
  num_channels = gdam_dsp->num_channels;
  sample_size = gdam_dsp->num_channels * gdam_dsp->bytes_per_sample;
  if (gdam_dsp->buffer.size / sample_size < n)
    n = gdam_dsp->buffer.size / sample_size;
  gdam_buffer_read (&gdam_dsp->buffer, (char *) buf, n * sample_size);
  gdam_dsp->samples_output += n;

  if (gdam_dsp->sample_converter != NULL)
    (*gdam_dsp->sample_converter) (buf, buf, n * num_channels);

  return rv + n;
}

static void gdas_gdam_dsp_miss_data(GdasSource* source, int n)
{
  int discarded;
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (source);

  if (gdam_dsp->ignore_miss_data)
    return;

  gdam_dsp->samples_output += n;

  n *= gdam_dsp->num_channels * gdam_dsp->bytes_per_sample;
  discarded = gdam_buffer_discard (&gdam_dsp->buffer, n);
  n -= discarded;		/* remaining bytes we cannot yet discard... */
  gdam_dsp->skipped += n;
}

static void gdas_gdam_dsp_destroy(GdasGdamDsp* gdam_dsp)
{
  g_return_if_fail (gdam_dsp != NULL);
  if (gdam_dsp->poll_fd.fd >= 0)
    {
      close (gdam_dsp->poll_fd.fd);
      gdam_dsp->poll_fd.fd = -1;
    }

  gdam_buffer_destruct (&gdam_dsp->buffer);
  parent_class->destroy ((GtkObject *) gdam_dsp);
}

static GdamTime gdas_gdam_dsp_get_position(GdasSource* source)
{
  GdasGdamDsp *gdam_dsp = GDAS_GDAM_DSP (source);
  return (gint64) gdam_dsp->samples_output;
}

GdasGdamDsp* gdas_gdam_dsp_new(int fd)
{
  GdasGdamDsp *gdam_dsp;
  gdam_dsp = GDAS_GDAM_DSP (gtk_type_new (GDAS_TYPE_GDAM_DSP));
  gdas_gdam_dsp_has_opened (gdam_dsp, fd);
  return gdam_dsp;
}

/*
 *  _ _     _             _
 * | (_)___| |_ ___ _ __ (_)_ __   __ _
 * | | / __| __/ _ \ '_ \| | '_ \ / _` |
 * | | \__ \ ||  __/ | | | | | | | (_| |
 * |_|_|___/\__\___|_| |_|_|_| |_|\__, |
 *                                |___/
 */
GtkType gdam_dsp_listener_get_type();
#define GDAM_TYPE_DSP_LISTENER			(gdam_dsp_listener_get_type ())
#define GDAM_DSP_LISTENER(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_DSP_LISTENER, GdamDspListener))
#define GDAM_DSP_LISTENER_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_DSP_LISTENER, GdamDspListenerClass))
#define GDAM_IS_DSP_LISTENER(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_DSP_LISTENER))
#define GDAM_IS_DSP_LISTENER_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_DSP_LISTENER))

typedef struct _GdamDspListener GdamDspListener;
typedef struct _GdamDspListenerClass GdamDspListenerClass;
struct _GdamDspListenerClass {
	GtkObjectClass	klass;
	void          (*accept)(GdamDspListener*     listener,
	                        int                  fd);
};
struct _GdamDspListener {
	GtkObject	object;
	GPollFD		poll_fd;
	guint		gsource_id;
	int		auto_mix_id;
	int		accepted_fd;
	GSList*		first_accepted;
	GSList*		last_accepted;
};

static GHashTable* listening_socket_by_fd = NULL;


/* GSourceFuncs to accept new file-descriptors */
static gboolean gdamdsp_listener_prepare  (gpointer  source_data, 
			                   GTimeVal *current_time,
			                   gint     *timeout,
			                   gpointer  user_data)
{
  return FALSE;
}
static gboolean gdamdsp_listener_check    (gpointer  source_data,
			                   GTimeVal *current_time,
			                   gpointer  user_data)
{
  GdamDspListener *lis_socket = user_data;
  return (lis_socket->poll_fd.revents & G_IO_IN) == G_IO_IN;
}

static gboolean gdamdsp_listener_dispatch (gpointer  source_data, 
			                   GTimeVal *current_time,
			                   gpointer  user_data)
{
  GdamDspListener *lis_socket = user_data;
  int fd;
  struct sockaddr addr;
  int addr_length = sizeof (addr);
  fd = accept (lis_socket->poll_fd.fd, &addr, &addr_length);
  if (fd < 0)
    {
      if (errno == EINTR)
	return TRUE;
      g_warning ("error accepting from listening socket: %s",
		 strerror (errno));
      return FALSE;
    }

  lis_socket->accepted_fd = fd;
  lis_socket->last_accepted = g_slist_append (lis_socket->last_accepted,
					      GINT_TO_POINTER (fd));
  if (lis_socket->first_accepted == NULL)
    lis_socket->first_accepted = lis_socket->last_accepted;
  else
    lis_socket->last_accepted = lis_socket->last_accepted->next;

  /* announce that a new fd is ready??? */
  gtk_signal_emit (GTK_OBJECT (lis_socket), accept_signal_id);

  return TRUE;
}
static GSourceFuncs accept_gdam_dsp_funcs = {
	gdamdsp_listener_prepare,
	gdamdsp_listener_check,
	gdamdsp_listener_dispatch,
	NULL
};


static void gdam_dsp_listener_init(GdamDspListener* dsp_listener)
{
  dsp_listener->poll_fd.fd = -1;
  dsp_listener->poll_fd.events = G_IO_IN;
  dsp_listener->gsource_id = 0;
  dsp_listener->auto_mix_id = 0;
  dsp_listener->first_accepted = dsp_listener->last_accepted = NULL;
}
static void gdam_dsp_listener_set_fd(GdamDspListener* dsp_listener,
                                     int              fd)
{
  g_return_if_fail (dsp_listener->poll_fd.fd == -1);
  dsp_listener->poll_fd.fd = fd;
  g_main_add_poll (&dsp_listener->poll_fd, G_PRIORITY_DEFAULT);
  dsp_listener->gsource_id = g_source_add (G_PRIORITY_DEFAULT,
					   FALSE,
					   &accept_gdam_dsp_funcs,
					   NULL, dsp_listener, NULL);
}
static GdamDspListener* gdam_dsp_listener_new(int fd)
{
  GdamDspListener *listener;
  listener = GDAM_DSP_LISTENER (gtk_type_new (GDAM_TYPE_DSP_LISTENER));
  if (fd >= 0)
    gdam_dsp_listener_set_fd (listener, fd);
  return listener;
}
static void gdam_dsp_listener_destroy(GtkObject* object)
{
  GdamDspListener *listener;
  GSList *at;
  listener = GDAM_DSP_LISTENER (object);
  for (at = listener->first_accepted; at != NULL; at = at->next)
    close (GPOINTER_TO_INT (at->data));
  g_slist_free (listener->first_accepted);
  if (listener->poll_fd.fd >= 0)
    {
      close (listener->poll_fd.fd);
      listener->poll_fd.fd = -1;
    }
}

static void marshal_accept_event(GdamConnection*    registered,
                                 GtkObject*         actor,
				 int                connection_id,
				 gpointer           event_type_data,
				 gpointer           connection_data)
{
  GdamBuffer event;
  GdamDspListener *listener;
  listener = GDAM_DSP_LISTENER (actor);
  gdam_buffer_construct (&event);
  gdam_buffer_append_int (&event, GDAM_EVENT_SOURCE_CREATED);
  gdam_buffer_append_int (&event, connection_id);
  gdam_buffer_append_int (&event, listener->accepted_fd);
  gdam_connection_emit_event (registered, &event);
}
static void gdam_dsp_listener_class_init(GdamDspListenerClass* klass)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gdam_dsp_listener_destroy;
  accept_signal_id
    = gtk_signal_new ("accept",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamDspListenerClass, accept),
		      gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  gtk_object_class_add_signals (object_class, &accept_signal_id, 1);
  gdam_event_register (object_class->type,
		       "accept", marshal_accept_event, NULL, NULL);

  accept_event_id = gdam_event_register (object_class->type,
					 "accept",
					 marshal_accept_event, NULL, NULL);
}

GtkType gdam_dsp_listener_get_type()
{
  static GtkType dsp_listener_type = 0;
  if (!dsp_listener_type)
    {
      static const GtkTypeInfo dsp_listener_info = {
	"GdamDspListener",
	sizeof (GdamDspListener),
	sizeof (GdamDspListenerClass),
	(GtkClassInitFunc) gdam_dsp_listener_class_init,
	(GtkObjectInitFunc) gdam_dsp_listener_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      dsp_listener_type = gtk_type_unique (parent, &dsp_listener_info);
    }
  return dsp_listener_type;
}

static int create_listener(const char* path, int auto_mix_id)
{
  GdamDspListener *lis;
  int fd;
  struct sockaddr_un un_addr;
  fd = socket (PF_UNIX, SOCK_STREAM, 0);
  if (fd < 0)
    {
      gdam_debug ("error making unix-domain socket: %s", strerror (errno));
      return -1;
    }
  un_addr.sun_family = AF_UNIX;
  strncpy (un_addr.sun_path, path, MAX_UNIX_PATH);
  if (bind (fd, (struct sockaddr *) &un_addr, sizeof (un_addr)) < 0)
    {
      gdam_debug ("error binding to unix-path %s: %s", path,
		  strerror (errno));
      close (fd);
      return -1;
    }
  listen (fd, SOMAXCONN);

  lis = gdam_dsp_listener_new (fd);
  lis->auto_mix_id = auto_mix_id;
  if (listening_socket_by_fd == NULL)
    listening_socket_by_fd = g_hash_table_new (NULL, NULL);
  g_hash_table_insert (listening_socket_by_fd, GINT_TO_POINTER (fd), lis);
  return lis->poll_fd.fd;
}

static GdamDspListener* lookup_listener(int fd)
{
  if (listening_socket_by_fd == NULL)
    return NULL;
  return g_hash_table_lookup (listening_socket_by_fd, GINT_TO_POINTER (fd));
}

static void destroy_listener(int fd)
{
  GSList *at;
  GdamDspListener *sock;
  sock = lookup_listener (fd);
  g_return_if_fail (sock != NULL);
  g_hash_table_remove (listening_socket_by_fd, GINT_TO_POINTER (fd));
  gtk_object_sink (GTK_OBJECT (sock));
}

static void gdas_gdam_dsp_func_begin_listening(GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out)
{
  const char *path = NULL;
  int i;
  int fd;
  int auto_mix_id = 0;
  GdamArg *rv;
  g_message ("gdas_gdam_dsp_func_begin_listening");
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arguments_in + i;
      const char *name = arguments_in[i].name;
      if (strcmp (name, "path") == 0 && arg->type == GDAM_ARG_TYPE_STRING)
	{
	  path = GDAM_ARG_STRING (arg);
	  continue;
	}
      gdam_debug ("unexpected arg or bad type: %s", name);
      return;
    }
  if (path == NULL)
    {
      gdam_debug ("path is mandatory");
      return;
    }

  fd = create_listener (path, auto_mix_id);
  if (fd < 0)
    {
      return;
    }

  *num_args_out = 1;
  rv = gdam_arg_new_from_valist (1, "fd", GDAM_ARG_TYPE_INT, fd);
  *arguments_out = rv;
}

static void gdas_gdam_dsp_func_stop_listening (GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out)
{
  int fd = -1;
  int i;
  GdamArg *rv;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arguments_in + i;
      const char *name = arguments_in[i].name;
      if (strcmp (name, "fd") == 0 && arg->type == GDAM_ARG_TYPE_INT)
	{
	  fd = GDAM_ARG_INT (arg);
	  continue;
	}
      gdam_debug ("unexpected arg or bad type: %s", name);
      return;
    }
  if (fd < 0)
    {
      gdam_debug ("fd is mandatory");
      return;
    }
  destroy_listener (fd);
}
static void gdas_gdam_dsp_func_subscribe_accept(GtkObjectClass* object_class,
				                int             num_args_in,
                                                GdamArg*        arguments_in,
				                int*            num_args_out,
				                GdamArg**       arguments_out)
{
  int fd = -1;
  int i;
  GdamArg *rv;
  GdamConnection *connection;
  GdamDspListener *listener;
  guint connect_id;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arguments_in + i;
      const char *name = arguments_in[i].name;
      if (strcmp (name, "fd") == 0 && arg->type == GDAM_ARG_TYPE_INT)
	{
	  fd = GDAM_ARG_INT (arg);
	  continue;
	}
      gdam_debug ("unexpected arg or bad type: %s", name);
      return;
    }
  if (fd < 0)
    {
      gdam_debug ("fd is mandatory");
      return;
    }
  listener = lookup_listener (fd);
  if (listener == NULL)
    return;

  /* hack! */
  connection = gdam_connection_get_top ();
  connect_id = gdam_event_connect (GTK_OBJECT (listener), connection,
				   accept_event_id, NULL);
  *num_args_out = 1;
  *arguments_out = gdam_arg_new_from_valist (1, "subscribe_id",
					     GDAM_ARG_TYPE_INT,
					     (int) connect_id);
}

static void gdas_gdam_dsp_func_maybe_accept   (GtkObjectClass* object_class,
				               int             num_args_in,
                                               GdamArg*        arguments_in,
				               int*            num_args_out,
				               GdamArg**       arguments_out)
{
  GdamDspListener *lis;
  int fd = -1;
  int i;
  int accepted_fd;
  GdasGdamDsp *new_source;
  gpointer data;
  GdamArg *rv;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arguments_in + i;
      const char *name = arguments_in[i].name;
      if (strcmp (name, "fd") == 0 && arg->type == GDAM_ARG_TYPE_INT)
	{
	  fd = GDAM_ARG_INT (arg);
	  continue;
	}
      gdam_debug ("unexpected arg or bad type: %s", name);
      return;
    }
  if (fd < 0)
    {
      gdam_debug ("fd is mandatory");
      return;
    }

  lis = lookup_listener (fd);
  if (lis->first_accepted == NULL)
    return;

  data = lis->first_accepted->data;
  accepted_fd = GPOINTER_TO_INT (data);
  lis->first_accepted = g_slist_remove (lis->first_accepted, data);
  if (lis->first_accepted == NULL)
    lis->last_accepted = NULL;
  new_source = gdas_gdam_dsp_new (accepted_fd);
  rv = gdam_arg_new_from_valist (1,
				 "source_id",
				 GDAM_ARG_TYPE_INT,
				 GDAM_SERVER_OBJECT (new_source)->global_id);
  *arguments_out = rv;
  *num_args_out = 1;
}
