/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasgong.h"
#include <math.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

static void gdas_gong_init(GdasGong* gong)
{
}

#define TABLE_SIZE (8192)
static GdamSampleValue* sine_table = NULL;

#define INDEX_BITS	16
#define INDEX_ONE	(1 << INDEX_BITS)

typedef struct _GdasGongElement GdasGongElement;
struct _GdasGongElement {
	GdamFixed		table_index;
	GdamFixed		table_velocity;
	GdamFixed		volume;
	GdamFixed		decay;
	int			decay_length;
	int			until_decay;
};

static gboolean gdas_gong_read_one(GdasGongElement* element,
                                   GdamSample*      samples,
			           int              to_read)
{
  GdamFixed decay = element->decay;
  int until_decay = element->until_decay;
  GdamFixed volume = element->volume;
  gint64 index = element->table_index;
  gint64 velocity = element->table_velocity;
  while (to_read-- > 0)
    {
      int val;
      if (--until_decay == 0)
	{
	  volume = (gint64) volume *decay / GdamFixed_ONE;
	  until_decay = element->decay_length;
	  if (volume < 3)
	    return FALSE;
	}
      val = (GdamSampleProduct) volume
	* sine_table[index >> INDEX_BITS] / GdamFixed_ONE;
      val += samples->s[0];
      samples->s[0] = GDAM_SAMPLE_CLAMP (val);
      index += velocity;
      if (index > ((GdamSampleProduct) TABLE_SIZE << INDEX_BITS))
	index -= ((GdamSampleProduct) TABLE_SIZE << INDEX_BITS);
      samples++;
    }
  element->volume = volume;
  element->table_index = index;
  element->until_decay = until_decay;

  return TRUE;
}

static int  gdas_gong_read(GdasSource* source,
                           GdamSample* samples,
			   int         num_read)
{
  GSList **plist;
  GdasGong *gong;
  int i;
  gong = GDAS_GONG (source);
  gdam_sample_zero (samples, num_read);

  plist = &gong->gongs;
  while (*plist)
    {
      GdasGongElement *element = (*plist)->data;
      if (!gdas_gong_read_one (element, samples, num_read))
	{
	  GSList *next = (*plist)->next;
	  g_slist_free_1 (*plist);
	  *plist = next;
	  continue;
	}
      plist = &(*plist)->next;
    }
  for (i = 0; i < num_read; i++)
    samples[i].s[1] = samples[i].s[0];
  return num_read;
}

static void add_one(GdasGong* gong, gdouble* array)
{
  GdasGongElement *element;
  gdouble decay;
  gdouble from_half_life;
  element = g_new (GdasGongElement, 1);
  element->table_index = 0;
  element->table_velocity = array[0] * TABLE_SIZE / 44100.0 * INDEX_ONE;
  element->volume = GDAM_FIXED_FROM_DOUBLE (array[1]);
  element->decay = GDAM_FIXED_FROM_DOUBLE (0.999);
  decay = GDAM_FIXED_TO_DOUBLE (element->decay);
  from_half_life = log (decay) / log (0.5);
  element->decay_length = (from_half_life * array[2] * 44100.0);
  element->until_decay = element->decay_length;
  gong->gongs = g_slist_prepend (gong->gongs, element);
}

static void gdas_gong_set_strike(GtkObject* object, GdamArg* arg)
{
  GdasGong *gong = GDAS_GONG (object);
  int num_arg = GDAM_ARG_DOUBLE_ARRAY (arg).length;
  gdouble *array = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  int i;
  /* Hertz Init_Volume Halflife */
  if (num_arg == 0 || num_arg % 3 != 0)
    {
      gdam_debug ("strike required three arguments "
		  "(hertz,init_volume,halflife)");
      return;
    }
  for (i = 0; i < num_arg / 3; i++)
    add_one (gong, array + 3 * i);
}

static void gdas_gong_class_init(GdasGongClass* gong_class)
{
  GtkObjectClass *object_class;
  int i;
  object_class = GTK_OBJECT_CLASS (gong_class);
  GDAS_SOURCE_CLASS (gong_class)->read = gdas_gong_read;
  gdam_arg_register (object_class, "strike", GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     gdas_gong_set_strike, NULL, NULL, NULL,	/* XXX: needs validator! */
		     0);
  sine_table = g_new (GdamSampleValue, TABLE_SIZE);
  for (i = 0; i < TABLE_SIZE; i++)
    {
      gdouble val = M_PI * i * 2.0 / TABLE_SIZE;
      sine_table[i] = sin (val) * GDAM_SAMPLE_ONE;
    }
}

GtkType gdas_gong_get_type()
{
  static GtkType gong_type = 0;
  if (!gong_type)
    {
      static const GtkTypeInfo gong_info = {
	"GdasGong",
	sizeof (GdasGong),
	sizeof (GdasGongClass),
	(GtkClassInitFunc) gdas_gong_class_init,
	(GtkObjectInitFunc) gdas_gong_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_SOURCE;
      gong_type = gtk_type_unique (parent, &gong_info);
    }
  return gong_type;
}
