/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasgranularfilter.h"
#include <stdlib.h>
#include <math.h>

struct _GdasGranuleState {
	GdamFixed	value;
	gboolean	is_attacking;
	GdamFixed       max_volume;
	GdamFixed	attack_rate;
	guint32  	attack_frac;
	GdamFixed	decay_rate;
	guint32  	decay_frac;
	guint32		cur_frac;

	GdasGranuleState* next;
};

static GMemChunk* granule_chunk = NULL;

static GdasGranuleState* gdas_granular_state_new(GdasGranularFilter* filter)
{
  GdasGranuleState *grain;
  grain = g_chunk_new (GdasGranuleState, granule_chunk);
  grain->value = 0;
  grain->is_attacking = TRUE;
  grain->max_volume = filter->grain_volume;
  grain->attack_frac = filter->attack_rate_frac;
  grain->attack_rate = filter->attack_rate;
  grain->decay_frac = filter->decay_rate_frac;
  grain->decay_rate = filter->decay_rate;
  grain->cur_frac = 0;

  grain->next = NULL;
  return grain;
}
static void gdas_granular_state_free(GdasGranuleState* state)
{
  g_chunk_free (state, granule_chunk);
}

/* returns whether or not the grain is still alive. */
static gboolean gdas_granule_state_add(GdasGranuleState* grain, 
                                       GdamFixed*        buffer,
				       int               length)
{
  GdamFixed value = grain->value;
  guint32 cur_frac = grain->cur_frac;

  /* the attack side of the granule */
  if (grain->is_attacking)
    {
      GdamFixed attack_rate = grain->attack_rate;
      guint32 attack_frac = grain->attack_frac;
      GdamFixed max_value = grain->max_volume;

      while (value < max_value && length > 0)
	{
	  guint32 new_frac = cur_frac + attack_frac;
	  if (new_frac < cur_frac)
	    value++;
	  cur_frac = new_frac;
	  value += attack_rate;
	  *buffer++ += value;
	  length--;
	}

      if (length <= 0)
	{
	  grain->cur_frac = cur_frac;
	  grain->value = value;
	  return TRUE;
	}
      grain->is_attacking = FALSE;
    }

  /* the decay side of the granule */
  {
    GdamFixed decay_rate = grain->decay_rate;
    guint32 decay_frac = grain->decay_frac;
    while (value > 0 && length > 0)
      {
	guint32 new_frac = cur_frac + decay_frac;
	if (new_frac < cur_frac)
	  value++;
	cur_frac = new_frac;
	value += decay_rate;
	*buffer++ += value;
	length--;
      }

    if (value <= 0)
      {
	/* remove the granule */
	return FALSE;
      }
    grain->cur_frac = cur_frac;
    grain->value = value;
    return TRUE;
  }
}

static int generate_distance(GdasGranularFilter* granular)
{
  switch (granular->distribution)
    {
    case GDAS_GRANULAR_FILTER_POISSON:
      {
	int median_dist = granular->mean_granule_dist;
	g_warning ("POISSON not supported");
	return 10;
      }
    case GDAS_GRANULAR_FILTER_FLAT:
      {
	int mean = granular->mean_granule_dist;
	int dev = granular->stddev_granule_dist;
	gdouble dist;
	dist = (gdouble) rand () / RAND_MAX;
	dist -= 0.5;
	dist *= 2.0;
	dist *= dev;
	dist += mean;
	if (dist < 0.0)
	  return 0;
	return (int) dist;
      }
    case GDAS_GRANULAR_FILTER_GAUSSIAN:
      {
	int mean = granular->mean_granule_dist;
	int stddev = granular->stddev_granule_dist;
	g_warning ("GAUSSIAN not supported");
	return 10;
      }
    default:
      g_assert_not_reached ();
    }
  return 0;
}

static void gdas_granular_make_envelope(GdasGranularFilter* granular,
                                        GdamFixed*          values,
					int                 num)
{
  GdasGranuleState *grain;
  GdasGranuleState *last = NULL;
  memset (values, 0, sizeof (GdamFixed) * num);
  for (grain = granular->granules; grain != NULL;)
    {
      GdasGranuleState *next = grain->next;
      if (!gdas_granule_state_add (grain, values, num))
	{
	  if (last == NULL)
	    granular->granules = next;
	  else
	    last->next = next;
	  gdas_granular_state_free (grain);
	  granular->num_grains--;
	}
      else
	{
	  last = grain;
	}
      grain = next;
    }

  /* start new granules? */
  while (num > 0)
    {
      if (granular->next_granule_dist < num)
	{
	  /* add a new granule */
	  GdasGranuleState *grain;
	  grain = gdas_granular_state_new (granular);

	  /* add the samples from that grain; then
	   * maybe add the grain to our list. */
	  if (gdas_granule_state_add (grain, values, num))
	    {
	      /* add the grain to the list. */
	      grain->next = granular->granules;
	      granular->granules = grain;
	      granular->num_grains++;
	    }
	  else
	    {
	      /* the grain finished before the buffer
	       * ran out.  Just delete it and don't
	       * add it to any list. */
	      gdas_granular_state_free (grain);
	    }

	  num -= granular->next_granule_dist;
	  values += granular->next_granule_dist;
	  /* compute the next distance. */
	  granular->next_granule_dist = generate_distance (granular);
	}
      else
	{
	  granular->next_granule_dist -= num;
	  values += num;
	  num = 0;
	}
    }
}

static int  gdas_granular_filter_read  (GdasSource*         source,
                                        GdamSample*         sample,
					int                 max_read)
{
  int num_read;
  int i;
  GdamSampleValue *envelope;
  GdasGranularFilter *granular;
  granular = GDAS_GRANULAR_FILTER (source);
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), sample, max_read);
  if (num_read <= 0)
    return num_read;
  envelope = alloca (sizeof (GdamSampleValue) * num_read);
  gdas_granular_make_envelope (granular, envelope, num_read);
  for (i = 0; i < num_read; i++)
    {
      GdamFixed value = envelope[i];
      sample[i].s[0] = GDAM_FIXED_MUL (sample[i].s[0], value);
      sample[i].s[1] = GDAM_FIXED_MUL (sample[i].s[1], value);
    }
  return num_read;
}
static inline void set_attack_time(GdasGranularFilter* granular, gfloat time)
{
  gdouble attack_rate;
  attack_rate = (gdouble) granular->grain_volume / (time * 44100.0);
  granular->attack_rate = floor (attack_rate);
  attack_rate -= granular->attack_rate;
  attack_rate *= (gdouble) (1 << 16);
  attack_rate *= (gdouble) (1 << 16);
  granular->attack_rate_frac = attack_rate;
  granular->last_attack_time = time;
}
static inline gfloat get_attack_time(GdasGranularFilter* granular)
{
  gint64 attack_rate_int;
  gdouble attack_rate;
  attack_rate_int = ((gint64) granular->attack_rate << 32)
    + granular->attack_rate_frac;
  attack_rate = attack_rate_int / (gdouble) (1 << 16) / (gdouble) (1 << 16);
  return (granular->grain_volume / attack_rate) / 44100.0;
}
static inline void set_decay_time(GdasGranularFilter* granular, gfloat time)
{
  gdouble decay_rate;
  decay_rate = -(gdouble) granular->grain_volume / (time * 44100.0);
  granular->decay_rate = floor (decay_rate);
  decay_rate -= granular->decay_rate;
  decay_rate *= (gdouble) (1 << 16);
  decay_rate *= (gdouble) (1 << 16);
  granular->decay_rate_frac = decay_rate;
  granular->last_decay_time = time;
}
static inline gfloat get_decay_time(GdasGranularFilter* granular)
{
  gint64 decay_rate_int;
  gdouble decay_rate;
  decay_rate_int = ((gint64) granular->decay_rate << 32)
    + granular->decay_rate_frac;
  decay_rate = decay_rate_int / (gdouble) (1 << 16) / (gdouble) (1 << 16);
  return -(granular->grain_volume / decay_rate) / 44100.0;
}

static void gdas_granular_filter_init(GdasGranularFilter* granular)
{
  granular->mean_granule_dist = 11025;
  granular->distribution = GDAS_GRANULAR_FILTER_FLAT;
  granular->next_granule_dist = generate_distance (granular);
  granular->granules = NULL;
  granular->clamp = TRUE;
  granular->grain_volume = GdamFixed_ONE / 2;

  /* set the attack and decay rates */
  set_attack_time (granular, 0.25);
  set_decay_time (granular, 0.25);
}

static void gdas_granular_arg_set_attack_time(GtkObject* object, GdamArg* arg)
{
  set_attack_time (GDAS_GRANULAR_FILTER (object), GDAM_ARG_DOUBLE (arg));
}

static void gdas_granular_arg_get_attack_time(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = get_attack_time (GDAS_GRANULAR_FILTER (object));
}

static void gdas_granular_arg_set_decay_time(GtkObject* object, GdamArg* arg)
{
  set_decay_time (GDAS_GRANULAR_FILTER (object), GDAM_ARG_DOUBLE (arg));
}

static void gdas_granular_arg_get_decay_time(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = get_decay_time (GDAS_GRANULAR_FILTER (object));
}

static void gdas_granular_arg_set_mean_distance(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  filter->mean_granule_dist = (int) (GDAM_ARG_DOUBLE (arg) * 44100.0);
}

static void gdas_granular_arg_get_mean_distance(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = (gdouble) filter->mean_granule_dist / 44100.0;
}

static void gdas_granular_arg_set_stddev(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  filter->stddev_granule_dist = (int) (GDAM_ARG_DOUBLE (arg) * 44100.0);
}

static void gdas_granular_arg_get_stddev(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = (gdouble) filter->stddev_granule_dist / 44100.0;
}

static void gdas_granular_arg_set_grain_volume(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  filter->grain_volume = GDAM_FIXED_FROM_DOUBLE (GDAM_ARG_DOUBLE (arg));
  set_attack_time (filter, filter->last_attack_time);
  set_decay_time (filter, filter->last_decay_time);
}

static void gdas_granular_arg_get_grain_volume(GtkObject* object, GdamArg* arg)
{
  GdasGranularFilter *filter;
  filter = GDAS_GRANULAR_FILTER (object);
  GDAM_ARG_DOUBLE (arg) = GDAM_FIXED_TO_DOUBLE (filter->grain_volume);
}

static void gdas_granular_arg_set_distribution(GtkObject* object, GdamArg* arg)
{
  const char *mode = GDAM_ARG_STRING (arg);
  GdasGranularFilter *filter = GDAS_GRANULAR_FILTER (object);
  if (strcmp (mode, "gaussian") == 0)
    GDAM_ARG_INT (arg) = GDAS_GRANULAR_FILTER_GAUSSIAN;
  else if (strcmp (mode, "poisson") == 0)
    GDAM_ARG_INT (arg) = GDAS_GRANULAR_FILTER_POISSON;
  else if (strcmp (mode, "flat") == 0)
    GDAM_ARG_INT (arg) = GDAS_GRANULAR_FILTER_FLAT;
  else
    {
      g_warning ("no known distribution: %s", GDAM_ARG_STRING (arg));
    }
}

static void gdas_granular_arg_get_distribution(GtkObject* object, GdamArg* arg)
{
  const char *mode = NULL;
  switch (GDAS_GRANULAR_FILTER (object)->distribution)
    {
    case GDAS_GRANULAR_FILTER_FLAT:
      mode = "flat";
      break;
    case GDAS_GRANULAR_FILTER_POISSON:
      mode = "poisson";
      break;
    case GDAS_GRANULAR_FILTER_GAUSSIAN:
      mode = "gaussian";
      break;
    default:
      g_assert_not_reached ();
    }
  GDAM_ARG_STRING (arg) = g_strdup (mode);
}

static void gdas_granular_arg_set_clamp(GtkObject* object, GdamArg* arg)
{
  GDAS_GRANULAR_FILTER (object)->clamp = GDAM_ARG_INT (arg) ? 1 : 0;
}

static void gdas_granular_arg_get_clamp(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_GRANULAR_FILTER (object)->clamp;
}

static void gdas_granular_filter_class_init(GdasSourceClass* klass)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (klass);
  granule_chunk = g_mem_chunk_create (GdasGranuleState, 8, G_ALLOC_AND_FREE);
  klass->read = gdas_granular_filter_read;

  gdam_arg_register (object_class,
		     "attack_time",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_granular_arg_set_attack_time,
		     gdas_granular_arg_get_attack_time,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "decay_time",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_granular_arg_set_decay_time,
		     gdas_granular_arg_get_decay_time,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "mean_grain_distance",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_granular_arg_set_mean_distance,
		     gdas_granular_arg_get_mean_distance,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "stddev_grain_distance",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_granular_arg_set_stddev,
		     gdas_granular_arg_get_stddev,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "grain_volume",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_granular_arg_set_grain_volume,
		     gdas_granular_arg_get_grain_volume,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "distribution",
		     GDAM_ARG_TYPE_STRING,
		     gdas_granular_arg_set_distribution,
		     gdas_granular_arg_get_distribution,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "clamp",
		     GDAM_ARG_TYPE_INT,
		     gdas_granular_arg_set_clamp,
		     gdas_granular_arg_get_clamp,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

GtkType gdas_granular_filter_get_type()
{
  static GtkType granular_filter_type = 0;
  if (!granular_filter_type)
    {
      static const GtkTypeInfo granular_filter_info = {
	"GdasGranularFilter",
	sizeof (GdasGranularFilter),
	sizeof (GdasGranularFilterClass),
	(GtkClassInitFunc) gdas_granular_filter_class_init,
	(GtkObjectInitFunc) gdas_granular_filter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      granular_filter_type = gtk_type_unique (parent, &granular_filter_info);
    }
  return granular_filter_type;
}
