/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/*
 * Support for the LADSPA plugin format.
 * [currently in very early development]
 */

#include "gdasladspaplugin.h"

#if HAVE_LADSPA_H

#include <ladspa.h>
#include <gmodule.h>
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdamsample.h>
#include <gdam/common/gdamplugin.h>
#include <gdam/common/gdam-protocol.h>
#include "../gdamconnection.h"

/*
 * All the functions used to access the LADSPA characteristics
 * are wrapped are ``functions'' with the GdamArg system:
 *
 * Functions on the ``class'' structure:
 *
 *      query_plugin
 *          name                 == ladspa plugin name
 *          id                   == plugin id
 *        returns:
 *          input_line_names     == string array of input names
 *				    [the length gives the number of inputs]
 *          output_line_names    == string array of output names
 *			            [the length gives the number of outputs]
 *          control_names        == string array of controls that can be set
 *          meter_names          == string array of values that
 *                                  can be retrieved.
 *          control_flags        == ladspa control flags
 *          min_values           == minimum control values if applicable
 *          max_values           == maximum control values if applicable
 *
 * Functions on the source:
 *      [first set the plugin_name]
 *       
 *       set_controls
 *           control_indices`    == indices of controls to set
 *           controls            == left/right value pairs to set them to.
 *
 *       get_meters
 *           meter_indices       == indices of meters to get
 *         returns:
 *           controls            == left/right value pairs to set them to.
 *
 *       connect   [connect any GdasSource's output to a named input]
 *           input_index         == index of this plugin's input to connect.
 *           subsource           == source_id of the GdasSource to have
 *                                  as this plugin's input
 */

static inline void zero_float_buffer(gfloat* buffer, int len)
{
  bzero (buffer, len * sizeof (gfloat));
}

static GtkObjectClass* parent_class = NULL;

typedef struct _GdamLadspaModule GdamLadspaModule;

struct _GdamLadspaDescriptor {
	const LADSPA_Descriptor*	descriptor;
	GdamLadspaModule*	module;

	gboolean		native_stereo;

	int			num_input_ports;
	int			num_output_ports;
	int			num_control_ports;
	int			num_meter_ports;

	int*			input_port_ids;
	int*			output_port_ids;
	int*			control_port_ids;
	int*			meter_port_ids;
};

/* Information about a single connection into the plugin. */
struct _GdamLadspaInput {
	GdasSource*		input;
	GdamSample*		sample_buffer;
	gfloat*			left_buffer;
	gfloat*			right_buffer;
	int			num_read;
};

struct _GdamLadspaModule {
	char*			name;
	GModule*		module;
	int			ref_count;

	/* Array of descriptors that come from this module. */
	int			num_descriptors;
	GdamLadspaDescriptor*	descriptors;
};

static gboolean
gdam_ladspa_descriptor_init(GdamLadspaDescriptor* to_init,
                            GdamLadspaModule* module,
                            const LADSPA_Descriptor* ladspa_descriptor);

static gboolean starts_with(const char* full, const char* prefix)
{
  int prefix_length;
  prefix_length = strlen (prefix);
  if (strncmp (full, prefix, prefix_length) == 0)
    return TRUE;
  return FALSE;
}
static gboolean ends_with(const char* full, const char* suffix)
{
  int full_length, suffix_length;
  suffix_length = strlen (suffix);
  full_length = strlen (full);
  if (full_length < suffix_length)
    return FALSE;
  return strcmp (suffix, full + full_length - suffix_length) == 0;
}


/*
 *  _         _                               _      _
 * | |__ _ __| |____ __  __ _   _ __  ___  __| |_  _| |___ ___
 * | / _` / _` (_-< '_ \/ _` | | '  \/ _ \/ _` | || | / -_|_-<
 * |_\__,_\__,_/__/ .__/\__,_| |_|_|_\___/\__,_|\_,_|_\___/__/
 *                |_|
 */
static GHashTable* module_table = NULL;
static const char* ladspa_search_path = 
		"~/.ladspa:/usr/local/lib/ladspa:/usr/lib/ladspa";
/* Finding and referencing counting the actual binary modules.
 * Note that there may be multiple LADSPA_Descriptors per
 * module, so that's why we refcount. */
static GdamLadspaModule* gdam_ladspa_module_get(const char* plugin_name)
{
  GdamLadspaModule *rv;
  char *plugin_path;
  char *so_name;
  GModule *module;
  int i;
  LADSPA_Descriptor_Function function;
  GArray *descriptors;
  if (!module_table)
    module_table = g_hash_table_new (g_str_hash, g_str_equal);
  rv = g_hash_table_lookup (module_table, plugin_name);
  if (rv != NULL)
    {
      rv->ref_count++;
      return rv;
    }
  if (ends_with (plugin_name, ".so"))
    so_name = g_strdup (plugin_name);
  else
    so_name = g_strconcat (plugin_name, ".so", NULL);
  if (so_name[0] == '/')
    {
      if (!starts_with (so_name, "/usr/local/lib/ladspa")
	  && !starts_with (so_name, "/usr/lib/ladspa"))
	{
	  g_warning ("absolute plugin paths must be in "
		     "/usr/local/lib/ladspa or /usr/lib/ladspa");
	}
      plugin_path = so_name;
    }
  else
    {
      plugin_path = gdam_search_path (so_name, ladspa_search_path,
				      GDAM_SEARCH_PATH_FIND_FILE);
      g_free (so_name);
    }
  if (plugin_path == NULL)
    {
      g_warning ("couldn't find ladspa plugin named %s", plugin_name);
      return NULL;
    }
  module = g_module_open (plugin_path, 0);
  if (module == NULL)
    {
      char *err_msg;
      err_msg = g_module_error ();
      gdam_debug ("failed to open %s: %s", plugin_name, err_msg);
      g_free (err_msg);
      return NULL;
    }
  if (!g_module_symbol (module, "ladspa_descriptor", (gpointer *) & function))
    {
      gdam_debug ("LADSPA plugin format: "
		  "all LADSPA plugins must contain an exported function "
		  "`ladspa_descriptor'");
      if (!g_module_close (module))
	g_warning ("module close failed in error recovery");
      return NULL;
    }
  i = 0;
  rv = g_new (GdamLadspaModule, 1);
  descriptors = g_array_new (FALSE, FALSE, sizeof (GdamLadspaDescriptor));
  for (;;)
    {
      const LADSPA_Descriptor *descriptor;
      GdamLadspaDescriptor this_desc;
      descriptor = (*function) (i);
      if (descriptor == NULL)
	break;
      if (!gdam_ladspa_descriptor_init (&this_desc, rv, descriptor))
	{
	  g_warning ("descriptor_init failed!!!");
	  break;
	}
      g_array_append_val (descriptors, this_desc);
      i++;
    }
  rv->module = module;
  rv->ref_count = 1;
  rv->num_descriptors = descriptors->len;
  rv->descriptors = (GdamLadspaDescriptor *) (descriptors->data);
  return rv;
}

/*
 *  _         _                    _                _      _
 * | |__ _ __| |____ __  __ _   __| |___ ___ __ _ _(_)_ __| |_ ___ _ _ ___
 * | / _` / _` (_-< '_ \/ _` | / _` / -_|_-</ _| '_| | '_ \  _/ _ \ '_(_-<
 * |_\__,_\__,_/__/ .__/\__,_|_\__,_\___/__/\__|_| |_| .__/\__\___/_| /__/
 *                |_|       |___|                    |_|
 */
static gboolean
gdam_ladspa_descriptor_init(GdamLadspaDescriptor* to_init,
                            GdamLadspaModule* module,
                            const LADSPA_Descriptor* ladspa_descriptor)
{
  GArray *input_port_ids;
  GArray *output_port_ids;
  GArray *control_port_ids;
  GArray *meter_port_ids;
  guint i;

  to_init->descriptor = ladspa_descriptor;
  to_init->module = module;

  input_port_ids = g_array_new (FALSE, FALSE, sizeof (int));
  output_port_ids = g_array_new (FALSE, FALSE, sizeof (int));
  control_port_ids = g_array_new (FALSE, FALSE, sizeof (int));
  meter_port_ids = g_array_new (FALSE, FALSE, sizeof (int));

  for (i = 0; i < ladspa_descriptor->PortCount; i++)
    {
      int type = ladspa_descriptor->PortDescriptors[i];
      GArray *to_append;
      gboolean input, output, control, audio;

      input = LADSPA_IS_PORT_INPUT (type);
      output = LADSPA_IS_PORT_OUTPUT (type);
      control = LADSPA_IS_PORT_CONTROL (type);
      audio = LADSPA_IS_PORT_AUDIO (type);

      if ((!input && !output) || (input && output))
	{
	  g_warning ("exactly one of INPUT or OUTPUT must be set");
	  break;
	}
      if ((!control && !audio) || (control && audio))
	{
	  g_warning ("exactly 1 of CONTROL or AUDIO must be set");
	  break;
	}

      if (input)
	{
	  if (control)
	    to_append = control_port_ids;
	  else
	    to_append = input_port_ids;
	}
      else
	{
	  if (control)
	    to_append = meter_port_ids;
	  else
	    to_append = output_port_ids;
	}
      g_array_append_val (to_append, i);
    }
  if (i < ladspa_descriptor->PortCount)
    {
      g_warning ("ladspa-descriptor had bad flags -- cannot proceed");
      g_array_free (input_port_ids, TRUE);
      g_array_free (output_port_ids, TRUE);
      g_array_free (meter_port_ids, TRUE);
      g_array_free (control_port_ids, TRUE);
      /* XXX: no way to cleanup LADSPA_Descritor? */
      return FALSE;
    }

  to_init->num_input_ports = input_port_ids->len;
  to_init->input_port_ids = (int *) input_port_ids->data;
  g_array_free (input_port_ids, FALSE);

  to_init->num_output_ports = output_port_ids->len;
  to_init->output_port_ids = (int *) output_port_ids->data;
  g_array_free (output_port_ids, FALSE);

  to_init->num_control_ports = control_port_ids->len;
  to_init->control_port_ids = (int *) control_port_ids->data;
  g_array_free (control_port_ids, FALSE);

  to_init->num_meter_ports = meter_port_ids->len;
  to_init->meter_port_ids = (int *) meter_port_ids->data;
  g_array_free (meter_port_ids, FALSE);

  to_init->native_stereo = (to_init->num_input_ports == 2);

  return TRUE;
}

#if 0
static void
gdam_ladspa_descriptor_destruct(GdamLadspaDescriptor* descriptor)
{
  g_free (descriptor->meter_port_ids);
  g_free (descriptor->control_port_ids);
  g_free (descriptor->input_port_ids);
  g_free (descriptor->output_port_ids);
}
#endif

/* Small helper function to tidy up the argument type-checking below... */
static inline gboolean do_type_check(GdamArg* arg, guint type)
{
  if (arg->type != type)
    g_warning ("type mismatch for %s", arg->name);
  return arg->type == type;
}
#define CHECK_ARG(arg, type) do_type_check(&arg, type)

/*
 *                                         _             _
 *   __ _ _   _  ___ _ __ _   _      _ __ | |_   _  __ _(_)_ __
 *  / _` | | | |/ _ \ '__| | | |    | '_ \| | | | |/ _` | | '_ \
 * | (_| | |_| |  __/ |  | |_| |    | |_) | | |_| | (_| | | | | |
 *  \__, |\__,_|\___|_|   \__, |____| .__/|_|\__,_|\__, |_|_| |_|
 *     |_|                |___/_____|_|            |___/
 */
static void make_string_array(GdamArg* arg,
			      const LADSPA_Descriptor* descriptor,
                              int  num_indices,
			      const int* indices)
{
  char **copy;
  int i;
  const char *const *port_names = descriptor->PortNames;

  copy = g_new (char *, num_indices);
  for (i = 0; i < num_indices; i++)
    copy[i] = g_strdup (port_names[indices[i]]);

  arg->type = GDAM_ARG_TYPE_STRING_ARRAY;
  GDAM_ARG_STRING_ARRAY (arg).array = copy;
  GDAM_ARG_STRING_ARRAY (arg).length = num_indices;

}
static inline int
remap_ladspa_port_hint(LADSPA_PortRangeHintDescriptor range_hint_descriptor)
{
  int rv = 0;
  if (LADSPA_IS_HINT_BOUNDED_BELOW (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_HAS_MIN_VALUE;
  if (LADSPA_IS_HINT_BOUNDED_ABOVE (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_HAS_MAX_VALUE;
  if (LADSPA_IS_HINT_TOGGLED (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_IS_BOOLEAN;
  if (LADSPA_IS_HINT_SAMPLE_RATE (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_IS_SAMPLE_RATE;
  if (LADSPA_IS_HINT_LOGARITHMIC (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_IS_LOGARITHMIC;
  if (LADSPA_IS_HINT_INTEGER (range_hint_descriptor))
    rv |= GDAM_PROTOCOL_CONTROL_IS_INTEGER;
  return rv;
}
static void gdam_ladspa_class_query_plugin(GtkObjectClass*      klass,
                                           int                  num_args_in,
					   GdamArg*             arg_in,
					   int*                 num_args_out,
					   GdamArg**            args_out)
{
  guint32 id = 0;
  const char *plugin_name = NULL;
  gboolean has_id = FALSE;
  int i;
  GdamLadspaModule *module;
  GdamLadspaDescriptor *descriptor;
  GdamArg *rv;

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arg_in + i;
      if (strcmp (arg->name, "name") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_STRING))
	    continue;
	  plugin_name = GDAM_ARG_STRING (arg);
	  continue;
	}
      if (strcmp (arg->name, "id") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_INT))
	    continue;
	  id = GDAM_ARG_INT (arg);
	  has_id = TRUE;
	  continue;
	}
      gdam_debug ("unknown argument `%s' to query_plugin", arg->name);
    }
  if (plugin_name == NULL || !has_id)
    {
      gdam_debug ("plugin_name and id are required by query_plugin");
      return;
    }

  module = gdam_ladspa_module_get (plugin_name);
  if (module == NULL)
    {
      gdam_debug ("no module named %s found", plugin_name);
      return;
    }

  descriptor = module->descriptors + id;
  if (descriptor == NULL)
    {
      g_warning ("couldn't get plugin index %d from %s", id, plugin_name);
      return;
    }

  *num_args_out = 7;
  *args_out = rv = g_new (GdamArg, *num_args_out);
  rv[0].name = g_strdup ("input_line_names");
  rv[0].type = GDAM_ARG_TYPE_STRING_ARRAY;
  make_string_array (rv + 0,
		     descriptor->descriptor,
		     descriptor->num_input_ports, descriptor->input_port_ids);

  rv[1].name = g_strdup ("output_line_names");
  rv[1].type = GDAM_ARG_TYPE_STRING_ARRAY;
  make_string_array (rv + 1,
		     descriptor->descriptor,
		     descriptor->num_output_ports,
		     descriptor->output_port_ids);

  rv[2].name = g_strdup ("control_names");
  rv[2].type = GDAM_ARG_TYPE_STRING_ARRAY;
  make_string_array (rv + 2,
		     descriptor->descriptor,
		     descriptor->num_control_ports,
		     descriptor->control_port_ids);

  rv[3].name = g_strdup ("meter_names");
  rv[3].type = GDAM_ARG_TYPE_STRING_ARRAY;
  make_string_array (rv + 3,
		     descriptor->descriptor,
		     descriptor->num_meter_ports, descriptor->meter_port_ids);

  rv[4].name = g_strdup ("controls_flags");
  rv[4].type = GDAM_ARG_TYPE_INT_ARRAY;
  rv[5].name = g_strdup ("min_values");
  rv[5].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  rv[6].name = g_strdup ("max_values");
  rv[6].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
  GDAM_ARG_INT_ARRAY (rv + 4).length = descriptor->num_control_ports;
  GDAM_ARG_INT_ARRAY (rv + 5).length = descriptor->num_control_ports;
  GDAM_ARG_INT_ARRAY (rv + 6).length = descriptor->num_control_ports;
  for (i = 0; i < descriptor->num_control_ports; i++)
    {
      const LADSPA_PortRangeHint *range_hint;
      LADSPA_PortRangeHintDescriptor ladspa_hint_descriptor;
      int index = descriptor->control_port_ids[i];

      range_hint = descriptor->descriptor->PortRangeHints + index;
      ladspa_hint_descriptor = range_hint->HintDescriptor;
      GDAM_ARG_INT_ARRAY (rv + 4).array[i]
	= remap_ladspa_port_hint (ladspa_hint_descriptor);
      GDAM_ARG_DOUBLE_ARRAY (rv + 5).array[i] = range_hint->LowerBound;
      GDAM_ARG_DOUBLE_ARRAY (rv + 6).array[i] = range_hint->UpperBound;
    }
}

static void gdam_ladspa_func_set_controls(GtkObject*           object,
                                          int                  num_args_in,
					  GdamArg*             arg_in,
					  int*                 num_args_out,
					  GdamArg**            args_out)
{
  int num_control_indices = -1;
  int *control_indices = NULL;
  int num_control_values = -1;
  gdouble *control_values = NULL;
  int i;
  GdasLadspaPlugin *plugin;
  GdamLadspaDescriptor *descriptor;

  plugin = GDAS_LADSPA_PLUGIN (object);
  descriptor = plugin->descriptor;
  if (descriptor == NULL)
    {
      gdam_debug ("plugin has no type -- cannot set controls");
      return;
    }

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arg_in + i;
      if (strcmp (arg_in[i].name, "control_indices") == 0)
	{
	  if (!CHECK_ARG (arg_in[i], GDAM_ARG_TYPE_INT_ARRAY))
	    continue;
	  num_control_indices = GDAM_ARG_INT_ARRAY (arg).length;
	  control_indices = GDAM_ARG_INT_ARRAY (arg).array;
	  continue;
	}
      else if (strcmp (arg_in[i].name, "controls") == 0)
	{
	  if (!CHECK_ARG (arg_in[i], GDAM_ARG_TYPE_DOUBLE_ARRAY))
	    continue;
	  num_control_values = GDAM_ARG_DOUBLE_ARRAY (arg).length;
	  control_values = GDAM_ARG_DOUBLE_ARRAY (arg).array;
	  continue;
	}
      else
	{
	  gdam_debug ("unknown argument `%s' to set_controls",
		      arg_in[i].name);
	}
    }
  if (num_control_indices == -1)
    {
      gdam_debug ("`control_indices' required for " "ladspa::set_controls");
      return;
    }
  if (num_control_values != num_control_indices * 2)
    {
      gdam_debug ("`controls' must have twice has many args "
		  "as `control_indices' to ladspa::set_controls");
      return;
    }
  for (i = 0; i < num_control_indices; i++)
    {
      int control_index = control_indices[i];
      if (control_index < 0 || control_index >= descriptor->num_control_ports)
	{
	  gdam_debug ("control index %d out of range", control_index);
	  continue;
	}
      plugin->left_controls[control_index] = control_values[2 * i + 0];
      if (!plugin->native_stereo)
	{
	  plugin->right_controls[control_index] = control_values[2 * i + 1];
	}
    }
}
static void gdam_ladspa_func_get_meters   (GtkObject*           object,
                                           int                  num_args_in,
					   GdamArg*             arg_in,
					   int*                 num_args_out,
					   GdamArg**            args_out)
{
  int num_meter_indices = -1;
  int *meter_indices = NULL;
  int i;
  gdouble *meter_values;
  GdasLadspaPlugin *plugin;

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arg_in + i;
      if (strcmp (arg->name, "meter_indices") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_INT_ARRAY))
	    continue;
	  num_meter_indices = GDAM_ARG_INT_ARRAY (arg).length;
	  meter_indices = GDAM_ARG_INT_ARRAY (arg).array;
	  continue;
	}
      else
	{
	  gdam_debug ("unknown argument `%s' to get_meters", arg_in[i].name);
	}
    }
  if (num_meter_indices < 0)
    {
      gdam_debug ("`meter_names' required for ladspa::get_meters");
      return;
    }
  meter_values = g_new (gdouble, num_meter_indices * 2);
  plugin = GDAS_LADSPA_PLUGIN (object);
  for (i = 0; i < num_meter_indices; i++)
    {
      int meter_index = meter_indices[i];
      meter_values[2 * i + 0] = plugin->left_meters[meter_index];
      meter_values[2 * i + 1] = plugin->right_meters[meter_index];
    }
  *num_args_out = 1;
  *args_out = g_new (GdamArg, 1);
  GDAM_ARG_DOUBLE_ARRAY (*args_out).array = meter_values;
  GDAM_ARG_DOUBLE_ARRAY (*args_out).length = num_meter_indices * 2;
}

#if 0
/* Connect one subsource to the two inputs with left/right. */
static void gdas_ladspa_plugin_connect_stereo(GdasLadspaPlugin*   ladspa,
                                              int                 left_index,
                                              int                 right_index,
				              GdasSource*         subsource)
{
  GdamLadspaInput *input;
  g_return_if_fail (ladspa->native_stereo);

  input = ladspa->inputs + left_index;
  input->lnput = subsource;
  input->sample_buffer =...;
  input->left_buffer = g_new (gfloat,
			      GdasSource * input;
			      GdamSample * sample_buffer;
			      gfloat * left_buffer; gfloat * right_buffer;
			      int num_read;
			      }
#endif
static void gdas_ladspa_plugin_connect(GdasLadspaPlugin*   ladspa,
                                       int                 index,
                                       int                 aux_index,
				       GdasSource*         subsource)
{
  GdamLadspaInput *input;

  if (ladspa->descriptor == NULL)
    return;

  input = ladspa->inputs + index;

  if (input->input == subsource)
    return;

  /* Disconnect an existing input to that index. */
  if (subsource != NULL)
    {
      /* Reference subsource first
       * (in case input->subsource == subsource) */
      gtk_object_ref (GTK_OBJECT (subsource));
    }

  if (input->input)
    {
      gtk_object_unref (GTK_OBJECT (input->input));
    }

  /* Connect the new input (subsource) to input input_index. */
  input->input = subsource;


  /* Allocate gfloat slabs and connect the underlying foreign handle */
  if (subsource == NULL)
    {
      g_return_if_fail (input->left_buffer != NULL);
      zero_float_buffer (input->left_buffer, ladspa->buffer_size);
      zero_float_buffer (input->right_buffer, ladspa->buffer_size);
    }
  else if (input->left_buffer == NULL)
    {
      const LADSPA_Descriptor *desc = ladspa->descriptor->descriptor;
      int port_id = ladspa->descriptor->input_port_ids[index];
      int buffer_size = ladspa->buffer_size;

      input->left_buffer = g_new0 (gfloat, buffer_size);
      input->sample_buffer = g_new (GdamSample, buffer_size);
      (*desc->connect_port) (ladspa->left_handle,
			     port_id, input->left_buffer);
      if (!ladspa->native_stereo)
	{
	  input->right_buffer = g_new0 (gfloat, buffer_size);
	  (*desc->connect_port) (ladspa->right_handle,
				 port_id, input->right_buffer);
	}
    }

  if (aux_index >= 0)
    {
      const LADSPA_Descriptor *desc = ladspa->descriptor->descriptor;
      int port_id = ladspa->descriptor->input_port_ids[aux_index];
      int buffer_size = ladspa->buffer_size;
      g_return_if_fail (ladspa->native_stereo);
      g_return_if_fail (aux_index != index);
      input = ladspa->inputs + aux_index;
      if (subsource != NULL)
	gtk_object_ref (GTK_OBJECT (subsource));
      if (input->input != NULL)
	gtk_object_unref (GTK_OBJECT (input->input));
      input->input = subsource;
      input->left_buffer = g_new0 (gfloat, buffer_size);
      //g_free(ladspa->inputs[aux_index].right_buffer);
      ladspa->inputs[aux_index].right_buffer = NULL;
      ladspa->inputs[index].right_buffer = input->left_buffer;
      (*desc->connect_port) (ladspa->left_handle,
			     port_id, input->left_buffer);
    }
}

static void gdam_ladspa_func_connect      (GtkObject*           object,
                                           int                  num_args_in,
					   GdamArg*             arg_in,
					   int*                 num_args_out,
					   GdamArg**            args_out)
{
  GdasLadspaPlugin *plugin;
  int i;
  int index = -1;
  int subsource_id = 0;
  GdasSource *subsource;
  int has_subsource_id = FALSE;
  int num_inputs;

  plugin = GDAS_LADSPA_PLUGIN (object);

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arg_in + i;
      if (strcmp (arg->name, "input_index") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_INT))
	    continue;
	  index = GDAM_ARG_INT (arg);
	  continue;
	}
      if (strcmp (arg->name, "subsource") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_INT))
	    continue;
	  subsource_id = GDAM_ARG_INT (arg);
	  has_subsource_id = TRUE;
	  continue;
	}
      gdam_debug ("unknown argument `%s' to get_meters", arg->name);
    }

  num_inputs = plugin->descriptor->num_input_ports;
  if (index < 0 || index >= num_inputs)
    {
      gdam_debug ("bad `input_index' to ladspa::connect");
      return;
    }
  if (subsource_id == 0)
    {
      gdam_debug ("no `subsource' to ladspa::connect");
      return;
    }
  subsource = gdam_connection_get_source (NULL, subsource_id);
  if (subsource == NULL)
    {
      gdam_debug ("bad `subsource' to ladspa::connect");
      return;
    }

  gdas_ladspa_plugin_connect (plugin, index, -1, subsource);
  gtk_object_unref (GTK_OBJECT (subsource));

  return;
}

static void gdam_ladspa_func_set_plugin   (GtkObject*           object,
                                           int                  num_args_in,
					   GdamArg*             arg_in,
					   int*                 num_args_out,
					   GdamArg**            args_out)
{
  GdasLadspaPlugin *plugin;
  const LADSPA_Descriptor *desc;

  const char *filename = NULL;
  int index = -1;
  int i;
  GdamLadspaModule *module;
  gfloat *left_controls;
  gfloat *right_controls;

  plugin = GDAS_LADSPA_PLUGIN (object);
  if (plugin->descriptor != NULL)
    {
      gdam_debug ("set_plugin may only be called once");
      return;
    }

  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = arg_in + i;
      if (strcmp (arg->name, "filename") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_STRING))
	    continue;
	  filename = GDAM_ARG_STRING (arg);
	  continue;
	}
      if (strcmp (arg->name, "index") == 0)
	{
	  if (!CHECK_ARG (*arg, GDAM_ARG_TYPE_INT))
	    continue;
	  index = GDAM_ARG_INT (arg);
	  continue;
	}
      gdam_debug ("unknown argument `%s' to set_plugin", arg->name);
    }

  if (index < 0 || filename == NULL)
    {
      gdam_debug ("both `filename' and `index' must be "
		  "specified to `set_plugin'.");
      return;
    }

  module = gdam_ladspa_module_get (filename);
  if (module == NULL)
    {
      gdam_debug ("no module corresponding to %s was found", filename);
      return;
    }

  if (index >= module->num_descriptors)
    {
      gdam_debug ("index %d was too high for %s", index, filename);
      return;
    }

  plugin->descriptor = module->descriptors + index;
  desc = plugin->descriptor->descriptor;
  plugin->left_handle = (*desc->instantiate) (desc, 44100);
  if (desc->activate != NULL)
    (*desc->activate) (plugin->left_handle);

  plugin->inputs = g_new0 (GdamLadspaInput,
			   plugin->descriptor->num_input_ports);

  left_controls = g_new0 (gfloat, plugin->descriptor->num_control_ports);
  plugin->left_controls = left_controls;
  for (i = 0; i < plugin->descriptor->num_control_ports; i++)
    {
      int port_id = plugin->descriptor->control_port_ids[i];
      (*desc->connect_port) (plugin->left_handle, port_id, left_controls + i);
    }

  plugin->native_stereo = plugin->descriptor->native_stereo;

  if (plugin->native_stereo)
    {
      plugin->right_handle = NULL;
    }
  else
    {
      plugin->right_handle = (*desc->instantiate) (desc, 44100);
      if (desc->activate != NULL)
	(*desc->activate) (plugin->right_handle);

      right_controls = g_new0 (gfloat, plugin->descriptor->num_control_ports);
      for (i = 0; i < plugin->descriptor->num_control_ports; i++)
	{
	  int port_id;
	  port_id = plugin->descriptor->control_port_ids[i];
	  (*desc->connect_port) (plugin->right_handle,
				 port_id, right_controls + i);
	}
      plugin->right_controls = right_controls;
    }

  if (plugin->descriptor->num_output_ports > 0)
    {
      gfloat *left_buffer;
      gfloat *right_buffer;
      int output_port_id = plugin->descriptor->output_port_ids[0];
      plugin->left_output_lines = g_new (gfloat *, 1);
      left_buffer = g_new0 (gfloat, plugin->buffer_size);
      plugin->left_output_lines[0] = left_buffer;
      (*desc->connect_port) (plugin->left_handle,
			     output_port_id, left_buffer);

      plugin->right_output_lines = g_new (gfloat *, 1);
      right_buffer = g_new0 (gfloat, plugin->buffer_size);
      plugin->right_output_lines[0] = right_buffer;
      if (plugin->native_stereo)
	{
	  output_port_id = plugin->descriptor->output_port_ids[1];
	  (*desc->connect_port) (plugin->left_handle,
				 output_port_id, right_buffer);
	}
      else
	{
	  (*desc->connect_port) (plugin->right_handle,
				 output_port_id, right_buffer);
	}
    }
}

static inline void
reset_handle(const LADSPA_Descriptor* descriptor,
             void*                    handle)
{
  if (handle == NULL)
    return;

  if (descriptor->deactivate != NULL)
    (*descriptor->deactivate) (handle);
  if (descriptor->activate != NULL)
    (*descriptor->activate) (handle);
}

static void gdas_ladspa_plugin_set_reset  (GtkObject*           object,
                                           GdamArg*             arg)
{
  GdasLadspaPlugin *plugin = GDAS_LADSPA_PLUGIN (object);
  const LADSPA_Descriptor *descriptor = plugin->descriptor->descriptor;

  /* hmm.. */
  if (GDAM_ARG_INT (arg) == 0)
    return;

  reset_handle (descriptor, plugin->left_handle);
  reset_handle (descriptor, plugin->right_handle);
}

static void gdas_ladspa_plugin_set_filtered(GtkObject* object, GdamArg* arg)
{
  int subsource_id = GDAM_ARG_INT (arg);
  GdasSource *subsource;
  GdasLadspaPlugin *ladspa;

  if (subsource_id == 0)
    subsource = NULL;
  else
    {
      subsource = gdam_connection_get_source (NULL, subsource_id);
      if (subsource == NULL)
	{
	  g_warning ("couldn't find subsource %d", subsource_id);
	  return;
	}
    }

  ladspa = GDAS_LADSPA_PLUGIN (object);
  gdas_ladspa_plugin_connect (ladspa,
			      0, ladspa->native_stereo ? 1 : -1, subsource);

  if (subsource != NULL)
    gtk_object_unref (GTK_OBJECT (subsource));
}

static void gdas_ladspa_plugin_get_filtered(GtkObject* object, GdamArg* arg)
{
  GdasLadspaPlugin *plugin;
  plugin = GDAS_LADSPA_PLUGIN (object);
  if (plugin->inputs == NULL)
    {
      GDAM_ARG_INT (arg) = 0;
    }
  else if (plugin->inputs->input == NULL)
    {
      GDAM_ARG_INT (arg) = 0;
    }
  else
    {
      GdasSource *input = plugin->inputs->input;
      GDAM_ARG_INT (arg) = GDAM_SERVER_OBJECT (input)->global_id;
    }
}

/* Precondition: buffer_len <= buffer_size. */
static int  do_one_buffer_length(GdasLadspaPlugin*     plugin, 
                                 GdamSample*           buffer,
		                 int                   buffer_len)
{
  int max_read = -1;
  int i;
  GdamLadspaDescriptor *descriptor = plugin->descriptor;
  int total_max_read = 0;

  /* First, grab data from our underlying sources. */
  for (i = 0; i < descriptor->num_input_ports; i++)
    {
      GdamLadspaInput *input = plugin->inputs + i;
      GdasSource *subsource = input->input;
      int num_read;
      if (subsource == NULL)
	continue;
      if (input->right_buffer == NULL)
	{
	  g_return_val_if_fail (plugin->native_stereo, -1);
	  continue;
	}
      num_read = gdas_source_read (subsource,
				   input->sample_buffer, buffer_len);
      input->num_read = num_read;
      if (num_read < 0)
	num_read = 0;
      if (max_read == -1)
	max_read = num_read;
      else if (max_read > num_read)
	max_read = num_read;

      if (!gdas_source_is_ready_to_read (subsource))
	total_max_read = buffer_len;
    }

  if (total_max_read < max_read)
    total_max_read = max_read;

  /* Shuffle samples into the input buffers for this plugin. */
  for (i = 0; i < descriptor->num_input_ports; i++)
    {
      GdamLadspaInput *input = plugin->inputs + i;
      gfloat *left_buffer = input->left_buffer;
      gfloat *right_buffer = input->right_buffer;
      GdamSample *samples = input->sample_buffer;
      GdasSource *subsource = input->input;
      int num_read = input->num_read;
      int j;
      if (subsource == NULL)
	continue;
      if (right_buffer == NULL)
	continue;
      for (j = 0; j < num_read; j++)
	{
	  left_buffer[j] = GDAM_SAMPLE_TO_FLOAT (samples->s[0]);
	  right_buffer[j] = GDAM_SAMPLE_TO_FLOAT (samples->s[1]);
	  samples++;
	}
      for (; j < max_read; j++)
	{
	  left_buffer[j] = right_buffer[j] = 0.0;
	}
    }

  /* Run the left and right plugin handlers. */
  (*descriptor->descriptor->run) (plugin->left_handle, max_read);
  if (!plugin->native_stereo)
    (*descriptor->descriptor->run) (plugin->right_handle, max_read);

  /* Now translate the output back to ``buffer''. */
  {
    gfloat *left_output, *right_output;
    if (plugin->native_stereo)
      {
	int left_index = plugin->which_left_output_to_use;
	int right_index = plugin->which_right_output_to_use;
	left_output = plugin->left_output_lines[left_index];
	right_output = plugin->right_output_lines[right_index];
      }
    else
      {
	int output_line = plugin->which_left_output_to_use;
	left_output = plugin->left_output_lines[output_line];
	right_output = plugin->right_output_lines[output_line];
      }
    for (i = 0; i < max_read; i++)
      {
	gfloat val;

	//g_message("left_output[%d] = %.7g\n",i,left_output[i]);

	val = left_output[i];
	val = CLAMP (val, -1.0, +1.0);
	buffer[i].s[0] = GDAM_SAMPLE_FROM_FLOAT (val);

	val = right_output[i];
	val = CLAMP (val, -1.0, +1.0);
	buffer[i].s[1] = GDAM_SAMPLE_FROM_FLOAT (val);
      }
  }

  /* Pad with silence. */
  if (total_max_read > max_read)
    gdam_sample_zero (buffer + max_read, total_max_read - max_read);

  return total_max_read;
}

static int gdas_ladspa_plugin_read(GdasSource*         source,
                                   GdamSample*         buffer,
				   int                 buffer_len)
{

  GdasLadspaPlugin *plugin;
  int buffer_size;
  int num_read = 0;

  plugin = GDAS_LADSPA_PLUGIN (source);
  if (plugin->descriptor == NULL)
    {
      gdam_debug ("cannot read from ladspa_plugin: has no type");
      return 0;
    }
  buffer_size = plugin->buffer_size;

  while (num_read < buffer_len)
    {
      int to_read = buffer_len - num_read;
      int got;
      if (to_read > buffer_size)
	to_read = buffer_size;
      got = do_one_buffer_length (plugin, buffer + num_read, to_read);
      if (got <= 0)
	break;
      num_read += got;
    }
  return num_read;
}


static gboolean gdas_ladspa_plugin_detach_child(GdasSource*      parent, 
	                                        GdasSource*      child)
{
  GdasLadspaPlugin *plugin;
  int num_inputs;
  GdamLadspaInput *input;
  int i;

  g_return_val_if_fail (child != NULL, FALSE);

  plugin = GDAS_LADSPA_PLUGIN (parent);
  num_inputs = plugin->descriptor->num_input_ports;

  for (i = 0; i < num_inputs; i++)
    if (plugin->inputs[i].input == child)
      break;
  if (i == num_inputs)
    return FALSE;

  /* Disconnect the input at index `i'. */
  input = plugin->inputs + i;
  input->input = NULL;
  zero_float_buffer (input->left_buffer, plugin->buffer_size);
  zero_float_buffer (input->right_buffer, plugin->buffer_size);
  gtk_object_unref (GTK_OBJECT (child));

  return TRUE;
}

static void gdas_ladspa_plugin_destroy(GtkObject* object)
{
  GdasLadspaPlugin *plugin;
  int num_inputs;
  GdamLadspaInput *input;
  int i;
  const LADSPA_Descriptor *desc;

  plugin = GDAS_LADSPA_PLUGIN (object);
  num_inputs = plugin->descriptor->num_input_ports;
  input = plugin->inputs;

  desc = plugin->descriptor->descriptor;
  if (desc->deactivate != NULL)
    {
      (*desc->deactivate) (plugin->left_handle);
      (*desc->deactivate) (plugin->right_handle);
    }
  (*desc->cleanup) (plugin->left_handle);
  (*desc->cleanup) (plugin->right_handle);
  plugin->left_handle = NULL;
  plugin->right_handle = NULL;

  for (i = 0; i < num_inputs; i++)
    {
      if (input[i].input != NULL)
	{
	  GtkObject *object;
	  object = GTK_OBJECT (input[i].input);
	  input[i].input = NULL;
	  gtk_object_unref (object);
	}
    }

  (*parent_class->destroy) (object);
}

/* get_position: returns position of first input. */
/* XXX: what to do for no-input plugins? */
static GdamTime 
gdas_ladspa_plugin_get_position (GdasSource *source)
{
  GdasLadspaPlugin *ladspa = GDAS_LADSPA_PLUGIN (source);
  GdasSource *filtered;
  GdamLadspaInput *input;

  input = ladspa->inputs + 0;
  g_return_val_if_fail (input != NULL, FALSE);
  filtered = input->input;

  if (filtered == NULL)
    {
      g_warning ("ladsa_plugin_getposition on plugin without subsource");
      return FALSE;
    }

  /* Check for circular dependency. */
  g_return_val_if_fail (filtered != source, -1);

  return gdas_source_get_position (filtered);
}


/* Seeks all inputs.  returns true unless an input fails to seek. */
/* XXX: what to do if some seeks fail?? */
static gboolean 
gdas_ladspa_plugin_seek (GdasSource      *source,
                         GdamTime         offset,
		         int              whence)
{
  GdasLadspaPlugin *ladspa = GDAS_LADSPA_PLUGIN (source);
  GdamLadspaDescriptor *descriptor = ladspa->descriptor;
  gboolean rv = TRUE;
  int i;

  g_return_val_if_fail (descriptor != NULL, FALSE);

  for (i = 0; i < descriptor->num_input_ports; i++)
    {
      GdamLadspaInput *input = ladspa->inputs + i;
      GdasSource *filtered = input->input;
      if (filtered == NULL)
	continue;
      if (!gdas_source_seek (filtered, offset, whence))
	rv = FALSE;
    }

  return rv;
}

static void 
gdas_ladspa_plugin_miss_data (GdasSource *source, int lost)
{
  GdasLadspaPlugin *ladspa = GDAS_LADSPA_PLUGIN (source);
  GdamLadspaDescriptor *descriptor = ladspa->descriptor;
  int i;

  for (i = 0; i < descriptor->num_input_ports; i++)
    {
      GdamLadspaInput *input = ladspa->inputs + i;
      GdasSource *filtered = input->input;

      if (filtered == NULL)
	continue;
      gdas_source_miss_data (filtered, lost);
    }
}

/* is_ready_to_read if any input is ready to read. */
static gboolean 
gdas_ladspa_plugin_is_ready_to_read (GdasSource *source)
{
  GdasLadspaPlugin *ladspa = GDAS_LADSPA_PLUGIN (source);
  GdamLadspaDescriptor *descriptor = ladspa->descriptor;
  gboolean rv = FALSE;
  int i;

  g_return_val_if_fail (descriptor != NULL, FALSE);

  if (descriptor->num_input_ports == 0)
    rv = TRUE;

  for (i = 0; i < descriptor->num_input_ports; i++)
    {
      GdamLadspaInput *input = ladspa->inputs + i;
      GdasSource *filtered = input->input;
      if (filtered == NULL)
	continue;
      if (gdas_source_is_ready_to_read (filtered))
	rv = TRUE;
    }

  return rv;
}


static void gdas_ladspa_plugin_init(GdasLadspaPlugin* ladspa_plugin)
{
  ladspa_plugin->buffer_size = 4096;
}
static void gdas_ladspa_plugin_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = (GtkObjectClass *) source_class;
  source_class->read = gdas_ladspa_plugin_read;
  source_class->detach_child = gdas_ladspa_plugin_detach_child;
  source_class->get_position = gdas_ladspa_plugin_get_position;
  source_class->miss_data = gdas_ladspa_plugin_miss_data;
  source_class->is_ready_to_read = gdas_ladspa_plugin_is_ready_to_read;
  source_class->seek = gdas_ladspa_plugin_seek;
  object_class->destroy = gdas_ladspa_plugin_destroy;
  /*gdam_arg_class_function_register(object_class,
     "query_plugin",
     gdam_ladspa_class_query_plugin); */
  gdam_arg_function_register (object_class,
			      "set_plugin", gdam_ladspa_func_set_plugin);
  gdam_arg_function_register (object_class,
			      "set_controls", gdam_ladspa_func_set_controls);
  gdam_arg_function_register (object_class,
			      "get_meters", gdam_ladspa_func_get_meters);
  gdam_arg_function_register (object_class,
			      "connect", gdam_ladspa_func_connect);
  gdam_arg_register (object_class,
		     "filtered",
		     GDAM_ARG_TYPE_INT,
		     gdas_ladspa_plugin_set_filtered,
		     gdas_ladspa_plugin_get_filtered,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "reset",
		     GDAM_ARG_TYPE_INT,
		     gdas_ladspa_plugin_set_reset, NULL, NULL, NULL, 0);
}

GtkType gdas_ladspa_plugin_get_type()
{
  static GtkType ladspa_plugin_type = 0;
  if (!ladspa_plugin_type)
    {
      static const GtkTypeInfo ladspa_plugin_info = {
	"GdasLadspaPlugin",
	sizeof (GdasLadspaPlugin),
	sizeof (GdasLadspaPluginClass),
	(GtkClassInitFunc) gdas_ladspa_plugin_class_init,
	(GtkObjectInitFunc) gdas_ladspa_plugin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_SOURCE;
      parent_class = gtk_type_class (parent);
      ladspa_plugin_type = gtk_type_unique (parent, &ladspa_plugin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return ladspa_plugin_type;
}

#endif
