/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdaslimiter.h"

static void gdas_limiter_init(GdasLimiter* limiter)
{
  limiter->threshold = GDAM_SAMPLE_ONE;
}
static void gdas_limiter_process   (GdasDynamicRange*   range,
                                    GdamSample*         out,
                                    const GdamSample*   in,
				    int                 num)
{
  GdasLimiter *limiter = GDAS_LIMITER (range);
  GdamSampleValue max_sample = range->max_sample;
  GdamSampleValue last_max_sample = range->last_max_sample;
  GdamSampleValue threshold = limiter->threshold;
  gint64 cur_scale;
  gint64 scale_incr;
  GdamFixed scale_1, scale_2;
  int i;
  gboolean last_limit = (last_max_sample > threshold);
  gboolean limit = (max_sample > threshold);

  if (!last_limit && !limit)
    {
      gdam_sample_copy (out, in, num);
      return;
    }

  if (last_limit)
    scale_1 = threshold * GdamFixed_ONE / last_max_sample;
  else
    scale_1 = GdamFixed_ONE;
  if (limit)
    scale_2 = threshold * GdamFixed_ONE / max_sample;
  else
    scale_2 = GdamFixed_ONE;

  cur_scale = (gint64) scale_1 << 16;
  scale_incr = ((gint64) (scale_2 - scale_1) << 16) / num;
  for (i = 0; i < num; i++)
    {
      GdamFixed scale = (cur_scale >> 16);
      gint64 tmp;
      cur_scale += scale_incr;

      tmp = (gint64) in[i].s[0] * scale / GdamFixed_ONE;
      out[i].s[0] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);

      tmp = (gint64) in[i].s[1] * scale / GdamFixed_ONE;
      out[i].s[1] = (GdamSampleValue) GDAM_SAMPLE_CLAMP (tmp);
    }
}

static void gdas_limiter_arg_set_threshold(GtkObject* object, GdamArg* arg)
{
  GdasLimiter *limiter = (GdasLimiter *) object;
  gdouble threshold = GDAM_ARG_DOUBLE (arg);
  if (threshold < 0.0 || threshold > 1.0)
    {
      gdam_debug ("threshold must be in [0,1]");
      return;
    }
  limiter->threshold = threshold * GDAM_SAMPLE_ONE;
}
static void gdas_limiter_arg_get_threshold(GtkObject* object, GdamArg* arg)
{
  GdasLimiter *limiter = (GdasLimiter *) object;
  gdouble threshold = limiter->threshold;
  threshold /= GDAM_SAMPLE_ONE;
  GDAM_ARG_DOUBLE (arg) = threshold;
}


static void gdas_limiter_class_init(GdasDynamicRangeClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  klass->process = gdas_limiter_process;
  gdam_arg_register (object_class,
		     "threshold",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_limiter_arg_set_threshold,
		     gdas_limiter_arg_get_threshold,
		     NULL, "0,10", GDAM_ARG_VITAL);
}

GtkType gdas_limiter_get_type()
{
  static GtkType limiter_type = 0;
  if (!limiter_type)
    {
      static const GtkTypeInfo limiter_info = {
	"GdasLimiter",
	sizeof (GdasLimiter),
	sizeof (GdasLimiterClass),
	(GtkClassInitFunc) gdas_limiter_class_init,
	(GtkObjectInitFunc) gdas_limiter_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_DYNAMIC_RANGE;
      limiter_type = gtk_type_unique (parent, &limiter_info);
    }
  return limiter_type;
}
