/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasmininetwork.h"
#include <gdam/server/gdamcachebuffer.h>
#include <gdam/common/gdammathfunction.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

static GdasMiniNetworkNode*
find_node_by_number(GdasMiniNetwork* network, int id);
typedef struct _GdasMiniNetworkConstDelay GdasMiniNetworkConstDelay;

#define DC_BLOCK_LENGTH GDAS_MINI_NETWORK_DC_BLOCK_LENGTH

/* Find the start offsets (returned in ``output_array'')
 * from a colon-separated list (in ``string_to_search'')
 * giving up after ``max_entries''.
 *
 * If `duplicate_last' is TRUE, the extra values are set to the last value.
 *                   otherwise, the extra values are set to NULL.
 */
static void split_by_colon(const char** output_array,
                           const char*  string_to_search,
			   int          max_entries,
			   gboolean     duplicate_last)
{
  int output_index;
  const char *colon;

  /* The function beneath should segfault anyway. */
  g_assert (string_to_search != NULL);
  g_assert (max_entries > 0);

  output_array[0] = string_to_search;
  output_index = 1;
  colon = strchr (string_to_search, ':');

  while (output_index < max_entries && colon != NULL)
    {
      output_array[output_index++] = colon + 1;
      colon = strchr (colon + 1, ':');
    }

  {
    const char *pad;
    if (duplicate_last)
      pad = output_array[output_index - 1];
    else
      pad = NULL;
    while (output_index < max_entries)
      output_array[output_index++] = pad;
  }
}

/*
 *  _   _           _        _____
 * | \ | | ___   __| | ___  |_   _|   _ _ __   ___
 * |  \| |/ _ \ / _` |/ _ \   | || | | | '_ \ / _ \
 * | |\  | (_) | (_| |  __/   | || |_| | |_) |  __/
 * |_| \_|\___/ \__,_|\___|   |_| \__, | .__/ \___|
 *                                |___/|_|
 *  ____            _     _             _   _
 * |  _ \ ___  __ _(_)___| |_ _ __ __ _| |_(_) ___  _ __
 * | |_) / _ \/ _` | / __| __| '__/ _` | __| |/ _ \| '_ \
 * |  _ <  __/ (_| | \__ \ |_| | | (_| | |_| | (_) | | | |
 * |_| \_\___|\__, |_|___/\__|_|  \__,_|\__|_|\___/|_| |_|
 *            |___/
 */
typedef struct _GdasMiniNetworkNodeStyle GdasMiniNetworkNodeStyle;
struct _GdasMiniNetworkNodeStyle {
	char*                    name;
	GdasMiniNetworkNodeMaker maker;
	gpointer                 maker_data;
};
static GHashTable* styles_by_name = NULL;
	
void gdas_mini_network_node_type_register
			(const char*               node_type,
			 GdasMiniNetworkNodeMaker  maker,
			 gpointer                  maker_data)
{
  GdasMiniNetworkNodeStyle *style;
  if (styles_by_name == NULL)
    styles_by_name = g_hash_table_new (g_str_hash, g_str_equal);
  g_return_if_fail (g_hash_table_lookup (styles_by_name,
					 (char *) node_type) == NULL);
  style = g_new (GdasMiniNetworkNodeStyle, 1);
  style->name = g_strdup (node_type);
  style->maker = maker;
  style->maker_data = maker_data;
  g_hash_table_insert (styles_by_name, style->name, style);
}

/*
 *  _   _           _
 * | \ | | ___   __| | ___  ___
 * |  \| |/ _ \ / _` |/ _ \/ __|
 * | |\  | (_) | (_| |  __/\__ \
 * |_| \_|\___/ \__,_|\___||___/
 * 
 */
struct _GdasMiniNetworkMix {
	GdasMiniNetworkNode	node;
};
struct _GdasMiniNetworkMul {
	GdasMiniNetworkNode	node;
};
typedef struct _GdasMiniNetworkGain GdasMiniNetworkGain;
struct _GdasMiniNetworkGain {
	GdasMiniNetworkNode	node;
	gdouble			gain;
};

static void output_silence(GdasMiniNetworkNode* node)
{
  node->output[0] = node->output[1] = 0.0;
}
/*
 *   ____                      _        _   _           _
 *  / ___| ___ _ __   ___ _ __(_) ___  | \ | | ___   __| | ___
 * | |  _ / _ \ '_ \ / _ \ '__| |/ __| |  \| |/ _ \ / _` |/ _ \
 * | |_| |  __/ | | |  __/ |  | | (__  | |\  | (_) | (_| |  __/
 *  \____|\___|_| |_|\___|_|  |_|\___| |_| \_|\___/ \__,_|\___|
 * 
 *  ____  _          __  __
 * / ___|| |_ _   _ / _|/ _|
 * \___ \| __| | | | |_| |_
 *  ___) | |_| |_| |  _|  _|
 * |____/ \__|\__,_|_| |_|
 * 
 */
static void gdas_mini_network_node_generic_disconnect
					(GdasMiniNetworkNode* node,
					 guint                pipe_id)
{
  int i;
  for (i = 0; i < node->num_inputs; i++)
    if (node->pipe_ids[i] == pipe_id)
      break;
  g_return_if_fail (i < node->num_inputs);
  memmove (node->pipe_ids + i,
	   node->pipe_ids + i + 1,
	   sizeof (guint) * (node->num_inputs - i - 1));
  memmove (node->needed_inputs + i,
	   node->needed_inputs + i + 1,
	   sizeof (guint8) * (node->num_inputs - i - 1));
  memmove (node->input_nodes + i,
	   node->input_nodes + i + 1,
	   sizeof (GdasMiniNetworkNode *) * (node->num_inputs - i - 1));
  --(node->num_inputs);
}
static void gdas_mini_network_node_destroy(GdasMiniNetworkNode* node)
{
  if (node->destroy)
    (*node->destroy) (node);
  else
    {
      g_free (node->input_nodes);
      g_free (node->needed_inputs);
      g_free (node->pipe_ids);
      g_free (node);
    }
}

/*
 *            _
 *  _ __ ___ (_)_  __
 * | '_ ` _ \| \ \/ /
 * | | | | | | |>  <
 * |_| |_| |_|_/_/\_\
 * 
 * Use the name `input' for the single input type.
 */
static gint  node_mix_connect(GdasMiniNetworkNode* node,
                              const char*          input_name,
			      GdasMiniNetworkNode* new_input)
{
  int num_inputs = node->num_inputs;
  node->needed_inputs = g_renew (guint8, node->needed_inputs, num_inputs + 1);
  node->input_nodes = g_renew (GdasMiniNetworkNode *,
			       node->input_nodes, num_inputs + 1);
  node->pipe_ids = g_renew (guint, node->pipe_ids, num_inputs + 1);
  node->num_inputs++;

  node->input_nodes[num_inputs] = new_input;
  node->needed_inputs[num_inputs] = 1;
  /*g_message("mn: mix_connect: num_inputs: %d", node->num_inputs); */
  return num_inputs;
}
static void gdas_mini_network_node_mix_update(GdasMiniNetworkNode* node)
{
  gdouble left = 0.0, right = 0.0;
  int i;
  for (i = 0; i < node->num_inputs; i++)
    {
      /*g_message("updating from node at %p", node->input_nodes[i]); */
      left += node->input_nodes[i]->output[0];
      right += node->input_nodes[i]->output[1];
    }
  node->output[0] = left;
  node->output[1] = right;
}
static GdasMiniNetworkNode* gdas_mini_network_node_mix()
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "mix";
  node->connect = node_mix_connect;
  node->update = gdas_mini_network_node_mix_update;
  return node;
}

/*
 *                      _       _       _
 *  _ __ ___   ___   __| |_   _| | __ _| |_ ___  _ __
 * | '_ ` _ \ / _ \ / _` | | | | |/ _` | __/ _ \| '__|
 * | | | | | | (_) | (_| | |_| | | (_| | || (_) | |
 * |_| |_| |_|\___/ \__,_|\__,_|_|\__,_|\__\___/|_|
 * 
 */
static void gdas_mini_network_node_modulator_update(GdasMiniNetworkNode* node)
{
  gdouble left, right;
  int i;
  if (node->num_inputs == 0)
    {
      node->output[0] = node->output[1] = 0.0;
    }
  left = right = 1.0;
  for (i = 0; i < node->num_inputs; i++)
    {
      /*g_message("updating from node at %p", node->input_nodes[i]); */
      left *= node->input_nodes[i]->output[0];
      right *= node->input_nodes[i]->output[1];
    }
  node->output[0] = left;
  node->output[1] = right;
}
static GdasMiniNetworkNode* gdas_mini_network_node_modulator()
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "modulator";
  node->connect = node_mix_connect;
  node->update = gdas_mini_network_node_modulator_update;
  return node;
}


/*
 *              _
 *   __ _  __ _(_)_ __
 *  / _` |/ _` | | '_ \
 * | (_| | (_| | | | | |
 *  \__, |\__,_|_|_| |_|
 *  |___/
 */
static void gdas_mini_network_node_gain_update(GdasMiniNetworkNode* node)
{
  gdouble gain = ((GdasMiniNetworkGain *) node)->gain;
  g_return_if_fail (node->num_inputs == 1);
  node->output[0] = gain * node->input_nodes[0]->output[0];
  node->output[1] = gain * node->input_nodes[0]->output[1];
}

static gint  gain_connect(GdasMiniNetworkNode* node,
                          const char*          input_name,
			  GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return 0;
  node->update = gdas_mini_network_node_gain_update;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}
static void gain_set_params(GdasMiniNetworkNode* node,
			    const char*          description)
{
  gdouble gain = strtod (description, NULL);
  ((GdasMiniNetworkGain *) node)->gain = gain;
}
static void  gain_disconnect (GdasMiniNetworkNode* node,
			      guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}
static GdasMiniNetworkNode* gdas_mini_network_node_gain(const char* options,
                                                        gpointer    data)
{
  GdasMiniNetworkGain *gain_node;
  GdasMiniNetworkNode *node;
  gain_node = g_new0 (GdasMiniNetworkGain, 1);
  node = (GdasMiniNetworkNode *) gain_node;
  node->class_name = "gain";
  gain_node->gain = strtod (options, NULL);
  node->connect = gain_connect;
  node->disconnect = gain_disconnect;
  node->update = output_silence;
  node->set_params = gain_set_params;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  return node;
}

/*
 *  _           _
 * | |__   __ _| | __ _ _ __   ___ ___
 * | '_ \ / _` | |/ _` | '_ \ / __/ _ \
 * | |_) | (_| | | (_| | | | | (_|  __/
 * |_.__/ \__,_|_|\__,_|_| |_|\___\___|
 * 
 */
typedef struct _GdasMiniNetworkBalance GdasMiniNetworkBalance;
struct _GdasMiniNetworkBalance {
	GdasMiniNetworkNode	node;
	gdouble			gain_ll;
	gdouble			gain_lr;
	gdouble			gain_rl;
	gdouble			gain_rr;
};

static void gdas_mini_network_node_balance_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkBalance *balance = (GdasMiniNetworkBalance *) node;
  gdouble in0, in1;
  g_return_if_fail (node->num_inputs == 1);
  in0 = node->input_nodes[0]->output[0];
  in1 = node->input_nodes[0]->output[1];
  node->output[0] = balance->gain_ll * in0 + balance->gain_rl * in1;
  node->output[1] = balance->gain_lr * in0 + balance->gain_rr * in1;
}

static gint  balance_connect(GdasMiniNetworkNode* node,
                          const char*          input_name,
			  GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return 0;
  node->update = gdas_mini_network_node_balance_update;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}
static void balance_set_params(GdasMiniNetworkNode* node,
			       const char*          description)
{
  const char *strs[4] = { 0 };
  GdasMiniNetworkBalance *balance = (GdasMiniNetworkBalance *) node;
  split_by_colon (strs, description, 4, TRUE);
  balance->gain_ll = strtod (strs[0], NULL);
  balance->gain_lr = strtod (strs[1], NULL);
  balance->gain_rl = strtod (strs[2], NULL);
  balance->gain_rr = strtod (strs[3], NULL);
}
static void  balance_disconnect (GdasMiniNetworkNode* node,
			      guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}
static GdasMiniNetworkNode* gdas_mini_network_node_balance(const char* options,
                                                        gpointer    data)
{
  GdasMiniNetworkBalance *balance_node;
  GdasMiniNetworkNode *node;
  balance_node = g_new0 (GdasMiniNetworkBalance, 1);
  node = (GdasMiniNetworkNode *) balance_node;
  node->class_name = "balance";
  balance_set_params (node, options);
  node->connect = balance_connect;
  node->disconnect = balance_disconnect;
  node->update = output_silence;
  node->set_params = balance_set_params;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  return node;
}


/*
 *      _      _
 *   __| | ___| | __ _ _   _
 *  / _` |/ _ \ |/ _` | | | |
 * | (_| |  __/ | (_| | |_| |
 *  \__,_|\___|_|\__,_|\__, |
 *                     |___/
 */
/*
 * GdasMiniNetworkConstDelay
 */
struct _GdasMiniNetworkConstDelay {
	GdasMiniNetworkNode	node;

	int			delay;
	int			buffer_length;
	int			delay_1_index;

	/* Of size NUM_CHANNELS * (max_delay+1). */
	gdouble*		delay_buffer;
};

static void gdas_mini_network_node_const_delay_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  int read_index;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  gdouble *delay_buffer = delay->delay_buffer;
  read_index = delay->delay_1_index + 1 - delay->delay;
  if (read_index < 0)
    read_index += delay->buffer_length;
  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];
  node->output[0] = delay_buffer[2 * read_index + 0];
  node->output[1] = delay_buffer[2 * read_index + 1];
  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;
}
static void gdas_mini_network_node_const_delay_update_no_output
					(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  gdouble *delay_buffer = delay->delay_buffer;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];
  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;
  node->output[0] = node->output[1] = 0.0;
}

static inline void const_delay_pick_update_function(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay;
  delay = (GdasMiniNetworkConstDelay *) node;
  if (node->num_inputs == 0)
    node->update = output_silence;
  else if (delay->buffer_length < delay->delay)
    node->update = gdas_mini_network_node_const_delay_update_no_output;
  else
    node->update = gdas_mini_network_node_const_delay_update;
}

static gint  const_delay_connect(GdasMiniNetworkNode* node,
                                 const char*          input_name,
			         GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return 0;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  const_delay_pick_update_function (node);
  return 0;
}

/* XXX check this function carefully !!!!!! */
static void set_buffer_size(GdasMiniNetworkConstDelay* delay, int length)
{
  int copy_start;
  int copy_length = length;
  int buffer_length = delay->buffer_length;
  gdouble *old_buffer = delay->delay_buffer;
  gdouble *new_buffer = g_new (gdouble, length * 2);
  if (length > buffer_length)
    copy_length = buffer_length;
  copy_start = delay->delay_1_index - copy_length;
  if (copy_start < 0)
    {
      int oldest_data = -copy_start;
      memcpy (new_buffer,
	      old_buffer + 2 * (buffer_length - oldest_data),
	      oldest_data * sizeof (gdouble) * 2);
      memcpy (new_buffer + 2 * oldest_data,
	      old_buffer, (copy_length + copy_start) * sizeof (gdouble) * 2);
    }
  else if (copy_length > 0)
    {
      memcpy (new_buffer,
	      old_buffer + 2 * copy_start,
	      copy_length * sizeof (gdouble) * 2);
    }
  delay->delay_1_index = copy_length;
  while (copy_length < length)
    {
      new_buffer[2 * copy_length + 0] = 0.0;
      new_buffer[2 * copy_length + 1] = 0.0;
      copy_length++;
    }
  if (delay->delay_1_index == length)
    delay->delay_1_index = 0;
  g_free (old_buffer);
  delay->delay_buffer = new_buffer;
  delay->buffer_length = length;
}
static void const_delay_set_params(GdasMiniNetworkNode* node,
			           const char*          description)
{
  const char *colon;
  GdasMiniNetworkConstDelay *delay_node;
  delay_node = (GdasMiniNetworkConstDelay *) node;
  delay_node->delay = atoi (description);
  if (delay_node->delay < 1)
    delay_node->delay = 1;
  colon = strchr (description, ':');
  if (colon)
    {
      int buffer_length = atoi (colon + 1);
      if (buffer_length <= delay_node->delay)
	buffer_length = delay_node->delay + 1;
      set_buffer_size (((GdasMiniNetworkConstDelay *) node), buffer_length);
    }
  else
    {
      if (delay_node->buffer_length <= delay_node->delay)
	set_buffer_size (delay_node, delay_node->delay + 1);
    }

}
static void  const_delay_disconnect (GdasMiniNetworkNode* node,
			             guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}
static GdasMiniNetworkNode* gdas_mini_network_node_delay(const char* options,
                                                        gpointer    data)
{
  GdasMiniNetworkConstDelay *delay_node;
  GdasMiniNetworkNode *node;
  delay_node = g_new0 (GdasMiniNetworkConstDelay, 1);
  node = (GdasMiniNetworkNode *) delay_node;
  node->class_name = "delay";
  delay_node->delay = 1;
  node->connect = const_delay_connect;
  node->disconnect = const_delay_disconnect;
  node->update = output_silence;
  node->set_params = const_delay_set_params;
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 0;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  const_delay_set_params (node, options);
  return node;
}


/*
 *                  _       _     _            _      _
 * __   ____ _ _ __(_) __ _| |__ | | ___    __| | ___| | __ _ _   _
 * \ \ / / _` | '__| |/ _` | '_ \| |/ _ \  / _` |/ _ \ |/ _` | | | |
 *  \ V / (_| | |  | | (_| | |_) | |  __/ | (_| |  __/ | (_| | |_| |
 *   \_/ \__,_|_|  |_|\__,_|_.__/|_|\___|  \__,_|\___|_|\__,_|\__, |
 *                                                            |___/
 */
typedef struct _GdasMiniNetworkVariableDelay GdasMiniNetworkVariableDelay;
struct _GdasMiniNetworkVariableDelay
{
  GdasMiniNetworkConstDelay       const_delay;

  gdouble                         last_delays[2];

  /* These describe how to map the control input
   * on to the delay of the primary input. */
  gdouble                         min_delay;

  /* For each 1 unit moved by input, increase
   * the delay by this amount */
  gdouble                         delay_ratio;

  gboolean                        stereo_delay;
  gboolean                        average;

  gboolean                        has_control, has_parent;
};

static void var_delay_output_with_control(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  GdasMiniNetworkVariableDelay *var_delay;
  int read_index, old_read_index;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkNode *control_input = node->input_nodes[1];
  gdouble *delay_buffer = delay->delay_buffer;
  gdouble delay_amount;

  /* Fractions for linear interpolation. */
  gdouble older_frac;

  var_delay = (GdasMiniNetworkVariableDelay *) node;

  /* Compute the delay in samples. */
  delay_amount = (control_input->output[0] + control_input->output[1]) / 2;
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;

  if (delay_amount < 1.0)
    delay_amount = 1.0;
  else if (delay_amount >= delay->buffer_length - 1)
    {
      node->output[0] = node->output[1] = 0.0;
      return;
    }

  /* The next line should be restored, once
   * i figure out why it is noisy.
   *
   * currently we just don't do linear interpolation. */
  /*older_frac = (gdouble)((int) delay_amount + 1) - delay_amount; */
  older_frac = 0.0;

  read_index = delay->delay_1_index + 1 - (int) delay_amount;
  if (read_index < 0)
    read_index += delay->buffer_length;

  old_read_index = read_index - 1;
  if (old_read_index < 0)
    old_read_index += delay->buffer_length;
  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];
  node->output[0] = delay_buffer[2 * read_index + 0] * (1.0 - older_frac)
    + delay_buffer[2 * old_read_index + 0] * older_frac;
  node->output[1] = delay_buffer[2 * read_index + 1] * (1.0 - older_frac)
    + delay_buffer[2 * old_read_index + 1] * older_frac;
  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;

  var_delay->last_delays[0] = var_delay->last_delays[1] = delay_amount + 1;
}
static void var_delay_output_with_stereo_control(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  GdasMiniNetworkVariableDelay *var_delay;
  int read_index, old_read_index;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkNode *control_input = node->input_nodes[1];
  gdouble *delay_buffer = delay->delay_buffer;
  gdouble delay_amount;

  /* Fraction for linear interpolation. */
  gdouble older_frac;

  var_delay = (GdasMiniNetworkVariableDelay *) node;

  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];

  /* Compute the delay in samples. */

  /* Left channel... */
  delay_amount = control_input->output[0];
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;

  if (delay_amount < 1.0)
    delay_amount = 1.0;
  if (delay_amount >= delay->buffer_length)
    {
      node->output[0] = 0.0;
    }
  else
    {
      gdouble young_frac;
      older_frac = (gdouble) ((int) delay_amount + 1) - delay_amount;
      young_frac = 1.0 - older_frac;

      read_index = delay->delay_1_index + 1 - (int) delay_amount;
      if (read_index < 0)
	read_index += delay->buffer_length;

      old_read_index = read_index - 1;
      if (old_read_index < 0)
	old_read_index += delay->buffer_length;
      node->output[0] = delay_buffer[2 * read_index + 0] * young_frac
	+ delay_buffer[2 * old_read_index + 0] * older_frac;
    }

  var_delay->last_delays[0] = delay_amount + 1.0;

  /* Right channel... [go cut-n-paste !] */
  delay_amount = control_input->output[1];
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;

  if (delay_amount < 1.0)
    delay_amount = 1.0;
  if (delay_amount >= delay->buffer_length)
    {
      node->output[1] = 0.0;
    }
  else
    {
      gdouble young_frac;
      older_frac = (gdouble) ((int) delay_amount + 1) - delay_amount;
      young_frac = 1.0 - older_frac;

      read_index = delay->delay_1_index + 1 - (int) delay_amount;
      if (read_index < 0)
	read_index += delay->buffer_length;

      old_read_index = read_index - 1;
      if (old_read_index < 0)
	old_read_index += delay->buffer_length;
      node->output[1] = delay_buffer[2 * read_index + 1] * young_frac
	+ delay_buffer[2 * old_read_index + 1] * older_frac;
    }

  var_delay->last_delays[1] = delay_amount + 1.0;

  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;
}


/* XXX: this should average with a max number to average. */
static gdouble compute_averaged(gdouble* buffer,
			        int      delay_1_offset,
                                int      buffer_length,
			        int      channel,
			        gdouble  delay_1,
			        gdouble  delay_2)
{
  gdouble sum;
  gdouble *at;

  gdouble index_1_double, index_2_double;
  int index_1, index_2;
  int index;
  int step = 1;

  delay_1 = CLAMP (delay_1, 1.0, buffer_length - 1);
  delay_2 = CLAMP (delay_2, 1.0, buffer_length - 1);
  if (delay_1 < delay_2)
    {
      gdouble tmp = delay_1;
      delay_1 = delay_2;
      delay_2 = tmp;
    }
  index_1_double = -delay_1 + buffer_length + delay_1_offset;
  index_2_double = -delay_2 + buffer_length + delay_1_offset;
  index_1 = (int) index_1_double;
  index_2 = (int) index_2_double;
  at = buffer + 2 * index_1 + channel;

  index_1 %= buffer_length;
  index_2 %= buffer_length;

  if (index_1 == index_2)
    return *at;
  sum = (1.0 - index_1_double + index_1) * (*at);
  for (index = index_1 + 1; index != index_2; index += step)
    {
      sum += *at * step;
      at += 2 * step;
      if (index >= buffer_length)
	{
	  at -= 2 * buffer_length;
	  index -= buffer_length;
	}
    }
  sum = (index_2_double - index_2) * (*at);
  return sum / (delay_1 - delay_2);
}


static void var_delay_output_avg_with_control(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  GdasMiniNetworkVariableDelay *var_delay;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkNode *control_input = node->input_nodes[1];
  gdouble *delay_buffer = delay->delay_buffer;
  gdouble delay_amount;
  int buffer_length;
  int delay_1_index;

  var_delay = (GdasMiniNetworkVariableDelay *) node;
  buffer_length = delay->buffer_length;
  delay_1_index = delay->delay_1_index;


  /* Compute the delay in samples. */
  delay_amount = (control_input->output[0] + control_input->output[1]) / 2;
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;
  delay_buffer[delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay_1_index * 2 + 1] = input->last_output[1];
  node->output[0] = compute_averaged (delay_buffer,
				      delay_1_index,
				      buffer_length,
				      0,
				      var_delay->last_delays[0],
				      delay_amount);
  node->output[1] = compute_averaged (delay_buffer,
				      delay_1_index,
				      buffer_length,
				      1,
				      var_delay->last_delays[1],
				      delay_amount);
  var_delay->last_delays[1] = delay_amount + 1.0;
}

static void var_delay_output_avg_with_stereo_control(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  GdasMiniNetworkVariableDelay *var_delay;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkNode *control_input = node->input_nodes[1];
  gdouble *delay_buffer = delay->delay_buffer;
  gdouble delay_amount;
  int buffer_length;
  int delay_1_index;

  var_delay = (GdasMiniNetworkVariableDelay *) node;
  buffer_length = delay->buffer_length;
  delay_1_index = delay->delay_1_index;

  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];

  /* Compute the delay in samples. */

  /* Left channel... */
  delay_amount = control_input->output[0];
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;

  node->output[0] = compute_averaged (delay_buffer, delay_1_index,
				      buffer_length, 0,
				      var_delay->last_delays[0],
				      delay_amount);
  var_delay->last_delays[0] = delay_amount + 1.0;

  /* Right channel... [go cut-n-paste !] */
  delay_amount = control_input->output[1];
  delay_amount = var_delay->min_delay + var_delay->delay_ratio * delay_amount;
  node->output[1] = compute_averaged (delay_buffer, delay_1_index,
				      buffer_length, 1,
				      var_delay->last_delays[1],
				      delay_amount);
  var_delay->last_delays[1] = delay_amount + 1.0;

  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;
}


static void var_delay_output_without_control(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkConstDelay *delay = (GdasMiniNetworkConstDelay *) node;
  GdasMiniNetworkVariableDelay *var_delay;
  int read_index, old_read_index;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  gdouble *delay_buffer = delay->delay_buffer;
  gdouble delay_amount;
  gdouble older_frac;

  var_delay = (GdasMiniNetworkVariableDelay *) node;

  /* Compute the delay in samples. */
  delay_amount = var_delay->min_delay;

  if (delay_amount < 1.0)
    delay_amount = 1.0;
  else if (delay_amount >= delay->buffer_length)
    {
      node->output[0] = node->output[1] = 0.0;
      return;
    }

  older_frac = (gdouble) ((int) delay_amount + 1) - delay_amount;

  read_index = delay->delay_1_index + 1 - (int) delay_amount;
  if (read_index < 0)
    read_index += delay->buffer_length;

  old_read_index = read_index - 1;
  if (old_read_index < 0)
    old_read_index += delay->buffer_length;
  delay_buffer[delay->delay_1_index * 2 + 0] = input->last_output[0];
  delay_buffer[delay->delay_1_index * 2 + 1] = input->last_output[1];
  node->output[0] = delay_buffer[2 * read_index + 0] * (1.0 - older_frac)
    + delay_buffer[2 * old_read_index + 0] * older_frac;
  node->output[1] = delay_buffer[2 * read_index + 1]
    + delay_buffer[2 * old_read_index + 1] * older_frac;
  if (++(delay->delay_1_index) == delay->buffer_length)
    delay->delay_1_index = 0;
}
static int var_delay_connect(GdasMiniNetworkNode* node,
                             const char*          input_name,
			     GdasMiniNetworkNode* new_input)
{
  GdasMiniNetworkVariableDelay *var_delay;
  int index = 0;

  var_delay = (GdasMiniNetworkVariableDelay *) node;
  if (strcmp (input_name, "delay") == 0)
    {
      if (var_delay->has_control)
	return -1;
      var_delay->has_control = TRUE;
      if (var_delay->has_parent)
	{
	  node->input_nodes[1] = new_input;
	  index = 1;
	}
      else
	node->input_nodes[0] = new_input;
    }
  else if (strcmp (input_name, "input") == 0)
    {
      if (var_delay->has_parent)
	return -1;
      var_delay->has_parent = TRUE;
      if (var_delay->has_control)
	{
	  node->input_nodes[1] = node->input_nodes[0];
	  node->pipe_ids[1] = node->pipe_ids[0];
	  node->input_nodes[0] = new_input;
	}
      else
	{
	  node->input_nodes[0] = new_input;
	}
    }
  node->num_inputs++;
  if (!var_delay->has_parent)
    return index;

  /* TODO: we should make variants to play as though it
   *       had a silence input... */
  if (!var_delay->has_control)
    node->update = var_delay_output_without_control;
  else if (var_delay->average)
    {
      if (var_delay->stereo_delay)
	node->update = var_delay_output_avg_with_stereo_control;
      else
	node->update = var_delay_output_avg_with_control;
    }
  else
    {
      if (var_delay->stereo_delay)
	node->update = var_delay_output_with_stereo_control;
      else
	node->update = var_delay_output_with_control;
    }
  return index;
}
static void var_delay_disconnect(GdasMiniNetworkNode* node,
                                 guint                pipe_id)
{
  GdasMiniNetworkVariableDelay *var_delay;
  var_delay = (GdasMiniNetworkVariableDelay *) node;
  if (node->num_inputs >= 1 && node->pipe_ids[0] == pipe_id)
    {
      if (var_delay->has_parent)
	var_delay->has_parent = FALSE;
      else
	var_delay->has_control = FALSE;
      if (node->num_inputs == 2)
	{
	  node->input_nodes[0] = node->input_nodes[1];
	  node->pipe_ids[0] = node->pipe_ids[1];
	}

    }
  else if (node->num_inputs == 2 && node->pipe_ids[1] == pipe_id)
    {
      var_delay->has_control = FALSE;
    }
  else
    {
      /* Bad pipe id */
      return;
    }
  node->num_inputs--;
  if (!var_delay->has_parent)
    node->update = output_silence;
  else if (!var_delay->has_control)
    node->update = var_delay_output_without_control;
}
static void var_delay_set_params(GdasMiniNetworkNode* node,
                                 const char*          options)
{
  GdasMiniNetworkVariableDelay *var_delay;
  const char *strs[3];
  int buffer_length;

  var_delay = (GdasMiniNetworkVariableDelay *) node;
  split_by_colon (strs, options, 3, FALSE);
  var_delay->min_delay = strtod (strs[0], NULL);
  if (strs[1] == NULL)
    var_delay->delay_ratio = 1.0;
  else
    var_delay->delay_ratio = strtod (strs[1], NULL);
  if (strs[2] == NULL)
    {
      buffer_length = (var_delay->delay_ratio * 2 + var_delay->min_delay);
    }
  else
    {
      buffer_length = strtod (strs[2], NULL);
    }

  set_buffer_size ((GdasMiniNetworkConstDelay *) node, buffer_length);
}

static GdasMiniNetworkNode*
gdas_mini_network_node_variable_delay(const char* options,
                                      gpointer    data)
{
  GdasMiniNetworkVariableDelay *delay_node;
  GdasMiniNetworkNode *node;
  delay_node = g_new0 (GdasMiniNetworkVariableDelay, 1);
  node = (GdasMiniNetworkNode *) delay_node;
  node->class_name = "variable_delay";
  delay_node->min_delay = 1024;
  delay_node->delay_ratio = 512;
  node->connect = var_delay_connect;
  node->disconnect = var_delay_disconnect;
  node->update = output_silence;
  node->set_params = var_delay_set_params;
  node->needed_inputs = g_new (guint8, 2);
  delay_node->average = FALSE;
  node->needed_inputs[0] = 0;

  /* We actually insist on the control node having no delay,
   * a seemingly strict requirement but... */
  node->needed_inputs[1] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 2);
  node->pipe_ids = g_new0 (guint, 2);
  var_delay_set_params (node, options);
  return node;
}


/*
 *              _
 *  _ __   ___ (_)___  ___
 * | '_ \ / _ \| / __|/ _ \
 * | | | | (_) | \__ \  __/
 * |_| |_|\___/|_|___/\___|
 * 
 */

static void noise_update(GdasMiniNetworkNode* node)
{
  node->output[0] = (gdouble) (rand () - RAND_MAX / 2)
    / (gdouble) (RAND_MAX / 2);
  node->output[1] = (gdouble) (rand () - RAND_MAX / 2)
    / (gdouble) (RAND_MAX / 2);
}

static GdasMiniNetworkNode* gdas_mini_network_node_noise(const char* options,
                                                        gpointer    data)
{
	GdasMiniNetworkNode* node;
	node = g_new0(GdasMiniNetworkNode, 1);
	node->update = noise_update;
	node->class_name = "noise";
	return node;
}

/*
 *                      _              _
 *   ___ ___  _ __  ___| |_ __ _ _ __ | |_
 *  / __/ _ \| '_ \/ __| __/ _` | '_ \| __|
 * | (_| (_) | | | \__ \ || (_| | | | | |_
 *  \___\___/|_| |_|___/\__\__,_|_| |_|\__|
 * 
 */

static inline gboolean is_start_of_number(char c)
{
  return isdigit (c) || c == '+' || c == '-' || c == '.';
}

static void constant_set_params(GdasMiniNetworkNode* node,
			        const char*          description)
{
  char *colon;
  if (description == NULL)
    return;
  colon = strchr (description, ':');
  if (colon != NULL && is_start_of_number (colon[1]))
    {
      node->output[0] = strtod (description, NULL);
      node->output[1] = strtod (colon + 1, NULL);
    }
  else
    {
      node->output[0] = node->output[1] = strtod (description, NULL);
    }
}
static void control_set_params(GdasMiniNetworkNode* node,
			        const char*          description)
{
  if (description == NULL)
    return;
  node->output[0] = node->output[1] = strtod (description, NULL);
}
static void pulse_set_params(GdasMiniNetworkNode* node,
			        const char*          description)
{
  const char *num;
  if (description == NULL)
    return;
  num = strchr (description, ':');
  if (num == NULL)
    return;
  num++;

  if (description[0] == '1')
    {
      num = strchr (num, ':');
      if (num == NULL)
        return;
      num++;
    }

  node->output[0] = node->output[1] = strtod (num, NULL);
  g_message ("pulse_set_params: %s: num=%s [out=%.6g]", description, num, node->output[0]);
}
static GdasMiniNetworkNode* gdas_mini_network_node_constant(const char* options,
                                                            gpointer    data)
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "constant";
  node->set_params = constant_set_params;
  constant_set_params (node, (char *) options);
  return node;
}
static GdasMiniNetworkNode* gdas_mini_network_node_control(const char* options,
                                                           gpointer    data)
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "control";
  node->set_params = control_set_params;
  control_set_params (node, (char *) options);
  return node;
}
static GdasMiniNetworkNode* gdas_mini_network_node_pulse   (const char* options,
                                                            gpointer    data)
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "pulse";
  node->set_params = pulse_set_params;
  pulse_set_params (node, (char *) options);
  return node;
}


/*
 *      _
 *  ___(_)_ __
 * / __| | '_ \
 * \__ \ | | | |
 * |___/_|_| |_|
 * 
 */
typedef struct _GdasMiniNetworkSin GdasMiniNetworkSin;
struct _GdasMiniNetworkSin
{
  GdasMiniNetworkNode     node;

  gdouble                 left_radians;
  gdouble                 right_radians;
  gdouble                 zero_step;
  gdouble                 one_step;
};

static void sin_update_with_child(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkSin *sin_node = (GdasMiniNetworkSin *) node;
  gdouble new_radians;
  gdouble step;
  GdasMiniNetworkNode *input = node->input_nodes[0];

  step = sin_node->zero_step + sin_node->one_step * input->output[0];
  new_radians = sin_node->left_radians + step;
  if (new_radians < 0.0)
    new_radians += 2.0 * M_PI;
  else if (new_radians >= 2.0 * M_PI)
    new_radians -= 2.0 * M_PI;
  node->output[0] = sin (new_radians);
  sin_node->left_radians = new_radians;

  step = sin_node->zero_step + sin_node->one_step * input->output[1];
  new_radians = sin_node->right_radians + step;
  if (new_radians < 0.0)
    new_radians += 2.0 * M_PI;
  else if (new_radians >= 2.0 * M_PI)
    new_radians -= 2.0 * M_PI;
  node->output[1] = sin (new_radians);
  sin_node->right_radians = new_radians;
}

static void sin_update_ident_without_child(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkSin *sin_node = (GdasMiniNetworkSin *) node;
  gdouble new_radians;

  new_radians = sin_node->left_radians + sin_node->zero_step;
  if (new_radians < 0.0)
    new_radians += 2.0 * M_PI;
  else if (new_radians >= 2.0 * M_PI)
    new_radians -= 2.0 * M_PI;
  node->output[0] = node->output[1] = sin (new_radians);
  sin_node->left_radians = sin_node->right_radians = new_radians;
}

static void sin_update_without_child(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkSin *sin_node = (GdasMiniNetworkSin *) node;
  gdouble new_radians;

  new_radians = sin_node->left_radians + sin_node->zero_step;
  if (new_radians < 0.0)
    new_radians += 2.0 * M_PI;
  else if (new_radians >= 2.0 * M_PI)
    new_radians -= 2.0 * M_PI;
  node->output[0] = sin (new_radians);
  sin_node->left_radians = new_radians;

  new_radians = sin_node->right_radians + sin_node->zero_step;
  if (new_radians < 0.0)
    new_radians += 2.0 * M_PI;
  else if (new_radians >= 2.0 * M_PI)
    new_radians -= 2.0 * M_PI;
  node->output[1] = sin (new_radians);
  sin_node->left_radians = new_radians;
}
static inline void sin_pick_update_function(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkSin *sin_node;
  sin_node = (GdasMiniNetworkSin *) node;
  if (node->num_inputs == 0)
    {
      if (sin_node->left_radians == sin_node->right_radians)
	{
	  node->update = sin_update_ident_without_child;
	}
      else
	{
	  node->update = sin_update_without_child;
	}
    }
  else
    {
      node->update = sin_update_with_child;
    }
}

static gint  sin_connect(GdasMiniNetworkNode* node,
                         const char*          input_name,
			 GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return 0;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  sin_pick_update_function (node);
  return 0;
}

static void  sin_disconnect (GdasMiniNetworkNode* node,
			     guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  sin_pick_update_function (node);
}

static void sin_set_params(GdasMiniNetworkNode* node,
			   const char*          description)
{
  GdasMiniNetworkSin *sin_node = (GdasMiniNetworkSin *) node;
  gdouble zero_rate, max_rate;
  char *colon;
  colon = strchr (description, ':');
  if (colon)
    {
      zero_rate = strtod (description, NULL);
      max_rate = strtod (colon + 1, NULL);
    }
  else
    {
      zero_rate = strtod (description, NULL);
      max_rate = 2.0 * zero_rate;
    }
  max_rate -= zero_rate;
  sin_node->zero_step = 2.0 * M_PI * zero_rate / 44100.0;
  sin_node->one_step = 2.0 * M_PI * max_rate / 44100.0;
}

static GdasMiniNetworkNode* gdas_mini_network_node_sin(const char* options,
                                                       gpointer    data)
{
  GdasMiniNetworkSin *sin_node;
  GdasMiniNetworkNode *node;
  sin_node = g_new0 (GdasMiniNetworkSin, 1);
  node = (GdasMiniNetworkNode *) sin_node;
  node->class_name = "sine";
  node->connect = sin_connect;
  node->disconnect = sin_disconnect;
  node->set_params = sin_set_params;
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  sin_pick_update_function (node);
  sin_set_params (node, (char *) options);
  return node;
}


/*
 * 
 *  ___  __ _ _   _  __ _ _ __ ___
 * / __|/ _` | | | |/ _` | '__/ _ \
 * \__ \ (_| | |_| | (_| | | |  __/
 * |___/\__, |\__,_|\__,_|_|  \___|
 *         |_|
 */
typedef struct _GdasMiniNetworkSquare GdasMiniNetworkSquare;
struct _GdasMiniNetworkSquare
{
  GdasMiniNetworkNode     node;
  int                     on_cycle_length;
  int                     off_cycle_length;
  gdouble                 on_left_value;
  gdouble                 on_right_value;
  gdouble                 off_left_value;
  gdouble                 off_right_value;
  gboolean                is_on;
  int                     index;
};

static void square_set_params(GdasMiniNetworkNode* node,
			      const char*          description)
{
  GdasMiniNetworkSquare *square_node = (GdasMiniNetworkSquare *) node;
  const char *strs[6];

  split_by_colon (strs, description, 6, TRUE);
  square_node->on_cycle_length = strtod (strs[0], NULL);
  square_node->off_cycle_length = strtod (strs[1], NULL);
  square_node->on_left_value = strtod (strs[2], NULL);
  square_node->on_right_value = strtod (strs[3], NULL);
  square_node->off_left_value = strtod (strs[4], NULL);
  square_node->off_right_value = strtod (strs[5], NULL);
}

static void square_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkSquare *square_node = (GdasMiniNetworkSquare *) node;

  if (square_node->is_on)
    {
      if (square_node->index >= square_node->on_cycle_length)
	{
	  square_node->index = 0;
	  square_node->is_on = FALSE;
	}
    }
  else
    {
      if (square_node->index >= square_node->off_cycle_length)
	{
	  square_node->index = 0;
	  square_node->is_on = TRUE;
	}
    }
  if (square_node->is_on)
    {
      node->output[0] = square_node->on_left_value;
      node->output[1] = square_node->on_right_value;
    }
  else
    {
      node->output[0] = square_node->off_left_value;
      node->output[1] = square_node->off_right_value;
    }
  square_node->index++;
}

static GdasMiniNetworkNode* gdas_mini_network_node_square(const char* options,
                                                          gpointer    data)
{
  GdasMiniNetworkSquare *square_node;
  GdasMiniNetworkNode *node;
  square_node = g_new0 (GdasMiniNetworkSquare, 1);
  node = (GdasMiniNetworkNode *) square_node;
  node->class_name = "square";
  node->set_params = square_set_params;
  node->update = square_update;
  square_set_params (node, (char *) options);
  return node;
}



/*
 *  _        _                   _
 * | |_ _ __(_) __ _ _ __   __ _| | ___
 * | __| '__| |/ _` | '_ \ / _` | |/ _ \
 * | |_| |  | | (_| | | | | (_| | |  __/
 *  \__|_|  |_|\__,_|_| |_|\__, |_|\___|
 *                         |___/
 */
typedef struct _GdasMiniNetworkTriangle GdasMiniNetworkTriangle;
struct _GdasMiniNetworkTriangle
{
  GdasMiniNetworkNode     node;
  int                     up_cycle_length;
  int                     down_cycle_length;
  gdouble                 min_value;
  gdouble                 delta_value;
  gboolean                is_up;
  int                     index;
};

	/* max_time:min_time:min_value:max_value */
static void triangle_set_params(GdasMiniNetworkNode* node,
			      const char*          description)
{
  GdasMiniNetworkTriangle *triangle_node = (GdasMiniNetworkTriangle *) node;
  gdouble max_value;
  const char *strs[6];

  split_by_colon (strs, description, 6, TRUE);
  triangle_node->up_cycle_length = strtod (strs[0], NULL);
  triangle_node->down_cycle_length = strtod (strs[1], NULL);
  triangle_node->min_value = strtod (strs[2], NULL);
  max_value = strtod (strs[3], NULL);
  triangle_node->delta_value = max_value - triangle_node->min_value;
  if (triangle_node->up_cycle_length < 1)
    triangle_node->up_cycle_length = 1;
  if (triangle_node->down_cycle_length < 1)
    triangle_node->down_cycle_length = 1;
}

static void triangle_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkTriangle *triangle_node;
  gdouble val;
  triangle_node = (GdasMiniNetworkTriangle *) node;

  if (triangle_node->is_up)
    {
      if (triangle_node->index >= triangle_node->up_cycle_length)
	{
	  triangle_node->index = 0;
	  triangle_node->is_up = FALSE;
	}
    }
  else
    {
      if (triangle_node->index >= triangle_node->down_cycle_length)
	{
	  triangle_node->index = 0;
	  triangle_node->is_up = TRUE;
	}
    }
  if (triangle_node->is_up)
    {
      val = triangle_node->delta_value * triangle_node->index
	/ triangle_node->up_cycle_length + triangle_node->min_value;
    }
  else
    {
      gdouble frac;
      frac = (gdouble) triangle_node->index
	/ triangle_node->down_cycle_length;
      frac = 1.0 - frac;
      frac *= triangle_node->delta_value;
      frac += triangle_node->min_value;
      val = frac;

    }
  node->output[0] = node->output[1] = val;
  triangle_node->index++;
}

static GdasMiniNetworkNode* gdas_mini_network_node_triangle(const char* options,
                                                          gpointer    data)
{
  GdasMiniNetworkTriangle *triangle_node;
  GdasMiniNetworkNode *node;
  triangle_node = g_new0 (GdasMiniNetworkTriangle, 1);
  node = (GdasMiniNetworkNode *) triangle_node;
  node->class_name = "triangle";
  node->set_params = triangle_set_params;
  node->update = triangle_update;
  triangle_set_params (node, (char *) options);
  return node;
}



/*
 *                          _        _     _
 * __      ____ ___   _____| |_ __ _| |__ | | ___
 * \ \ /\ / / _` \ \ / / _ \ __/ _` | '_ \| |/ _ \
 *  \ V  V / (_| |\ V /  __/ || (_| | |_) | |  __/
 *   \_/\_/ \__,_| \_/ \___|\__\__,_|_.__/|_|\___|
 * 
 */
typedef struct _GdasMiniNetworkWavetable GdasMiniNetworkWavetable;
struct _GdasMiniNetworkWavetable
{
  GdasMiniNetworkNode node;
  /* Holds a reference to the cache buffer. */
  GdamCacheBuffer* cache_buffer;
  gdouble buffer_pos;
  gdouble rate;
};

static void wavetable_update_no_rate_lead(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkWavetable *wavetable;
  int length;
  GdamSample sample;
  wavetable = (GdasMiniNetworkWavetable *) node;
  length = wavetable->cache_buffer->buffer_length;
  if (length <= 0)
    return;
  if (wavetable->buffer_pos < 0.0 || wavetable->buffer_pos >= length)
    {
      wavetable->buffer_pos = fmod (wavetable->buffer_pos, length);
      if (wavetable->buffer_pos < 0.0)
	wavetable->buffer_pos += length;
    }
  sample = gdam_cache_buffer_index (wavetable->cache_buffer,
				    wavetable->buffer_pos);
  node->output[0] = (gdouble) sample.s[0] / GDAM_SAMPLE_ONE;
  node->output[1] = (gdouble) sample.s[1] / GDAM_SAMPLE_ONE;
  wavetable->buffer_pos++;
}

static void wavetable_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkWavetable *wavetable;
  GdasMiniNetworkNode *input;
  int length;
  GdamSample sample;
  wavetable = (GdasMiniNetworkWavetable *) node;
  length = wavetable->cache_buffer->buffer_length;
  if (length <= 0)
    return;
  input = node->input_nodes[0];
  if (wavetable->buffer_pos < 0.0 || wavetable->buffer_pos >= length)
    {
      wavetable->buffer_pos = fmod (wavetable->buffer_pos, length);
      if (wavetable->buffer_pos < 0.0)
	wavetable->buffer_pos += length;
    }
  sample = gdam_cache_buffer_index (wavetable->cache_buffer,
				    wavetable->buffer_pos);
  node->output[0] = (gdouble) sample.s[0] / GDAM_SAMPLE_ONE;
  node->output[1] = (gdouble) sample.s[1] / GDAM_SAMPLE_ONE;
  wavetable->buffer_pos += (input->output[0] + input->output[1]) / 2;
}
static void wavetable_pick_update_function(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkWavetable *wavetable;
  wavetable = (GdasMiniNetworkWavetable *) node;
  if (wavetable->cache_buffer == NULL)
    node->update = output_silence;
  else if (node->num_inputs == 0)
    node->update = wavetable_update_no_rate_lead;
  else
    node->update = wavetable_update;
}

static gint  wavetable_connect(GdasMiniNetworkNode* node,
                               const char*          input_name,
			       GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  wavetable_pick_update_function (node);
  return 0;
}

static void  wavetable_disconnect (GdasMiniNetworkNode* node,
			           guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  wavetable_pick_update_function (node);
}

static void wavetable_set_params(GdasMiniNetworkNode* node,
			         const char*          description)
{
  GdasMiniNetworkWavetable *wav = (GdasMiniNetworkWavetable *) node;
  int old_buffer_id;
  char *colon;
  if (wav->cache_buffer == NULL)
    old_buffer_id = 0;
  else
    old_buffer_id = gdam_cache_buffer_get_id (wav->cache_buffer);
  colon = strchr (description, ':');
  if (colon)
    {
      int buffer_id = atoi (colon + 1);
      gdouble rate;
      rate = strtod (description, NULL);
      if (buffer_id != old_buffer_id)
	{
	  GdamCacheBuffer *old_buf = wav->cache_buffer;
	  if (wav->cache_buffer)
	    {
	      gtk_object_unref (GTK_OBJECT (old_buf));
	      wav->cache_buffer = NULL;
	    }
	  if (buffer_id != 0)
	    {
	      GdamCacheBuffer *buffer;
	      buffer = gdam_cache_buffer_get_by_id (buffer_id);
	      if (buffer)
		{
		  gtk_object_ref (GTK_OBJECT (buffer));
		  wav->cache_buffer = buffer;
		}
	    }
	}
    }
  else
    {
      gdouble rate;
      rate = strtod (description, NULL);
      wav->rate = rate;
    }
  wavetable_pick_update_function (node);
}

static GdasMiniNetworkNode* gdas_mini_network_node_wavetable
                                                      (const char* options,
                                                       gpointer    data)
{
  GdasMiniNetworkWavetable *wavetable_node;
  GdasMiniNetworkNode *node;
  wavetable_node = g_new0 (GdasMiniNetworkWavetable, 1);
  node = (GdasMiniNetworkNode *) wavetable_node;
  node->class_name = "wavetable";
  node->connect = wavetable_connect;
  node->disconnect = wavetable_disconnect;
  node->set_params = wavetable_set_params;
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  wavetable_set_params (node, (char *) options);
  return node;
}

/*
 *  _       _                                       _        _    _
 * | |_ _ _(_)__ _ __ _ ___ _ ___ __ ____ ___ _____| |_ __ _| |__| |___
 * |  _| '_| / _` / _` / -_) '_\ V  V / _` \ V / -_)  _/ _` | '_ \ / -_)
 *  \__|_| |_\__, \__, \___|_|__\_/\_/\__,_|\_/\___|\__\__,_|_.__/_\___|
 *           |___/|___/     |___|
 */
typedef struct _SubTriggeredWavetable SubTriggeredWavetable;
typedef struct _GdasMiniNetworkTriggerWavetable GdasMiniNetworkTriggerWavetable;
struct _SubTriggeredWavetable {
	gdouble buffer_pos;
	gdouble rate;
};
	
struct _GdasMiniNetworkTriggerWavetable {
	GdasMiniNetworkNode node;
	GdamCacheBuffer* cache_buffer;
	GSList* earlier;
	GSList* latest;
	int max_wavetables;
	int num_wavetables;

	/* During the release period, the trigger will
	 * not go off again. */
	int trigger_release;

	gdouble threshold;
};

static void trigger_wavetable_update(GdasMiniNetworkNode* node)
{
}
static void trigger_wavetable_update_no_inputs(GdasMiniNetworkNode* node)
{
}
static gint  trigger_wavetable_connect(GdasMiniNetworkNode* node,
                                       const char*          input_name,
			               GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  node->update = trigger_wavetable_update;
  return 0;
}

static void  trigger_wavetable_disconnect (GdasMiniNetworkNode* node,
			                   guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = trigger_wavetable_update_no_inputs;
  return;
}

/*
 *      _ _     _        _ _           _   _
 *   __| (_)___| |_ _ __(_) |__  _   _| |_(_) ___  _ __  ___
 *  / _` | / __| __| '__| | '_ \| | | | __| |/ _ \| '_ \/ __|
 * | (_| | \__ \ |_| |  | | |_) | |_| | |_| | (_) | | | \__ \
 *  \__,_|_|___/\__|_|  |_|_.__/ \__,_|\__|_|\___/|_| |_|___/
 * 
 */
typedef struct _DistributionNode DistributionNode;
struct _DistributionNode
{
  GdasMiniNetworkNode node;
  int                 remaining;
  gfloat              mean_length;
  gfloat              stddev;		/* if applicable */
  gboolean            is_gaussian;
};

#define M_2PI   (M_PI * 2.0)

/* Implement the Box-Muller Transformation to make
 * pairs of gaussian variates from pairs of uniform variates.
 */
static gfloat get_gaussian ()
{
  static gboolean has_extra_gaussian = FALSE;
  static gfloat extra_gaussian;
  if (has_extra_gaussian)
    {
      has_extra_gaussian = FALSE;
      return extra_gaussian;
    }
  {
    gfloat a, b;
    gfloat common, angle;
    a = (gfloat) rand () / RAND_MAX;
    b = (gfloat) rand () / RAND_MAX;
    common = sqrt (-2.0 * log (a));
    angle = M_2PI * b;
    has_extra_gaussian = TRUE;
    extra_gaussian = cos (angle) * common;
    return sin (angle) * common;
  }
}

static gfloat get_poisson ()
{
  gfloat v = (gfloat) rand () / RAND_MAX;
  return -log (v);
}

static void distribution_update(GdasMiniNetworkNode *node)
{
  DistributionNode *dist_node = (DistributionNode *) node;;
  --(dist_node->remaining);
  if (dist_node->remaining <= 0)
    {
      gfloat val;
      node->output[0] = node->output[1] = 1.0;
      if (dist_node->is_gaussian)
	val = get_gaussian () * dist_node->stddev + dist_node->mean_length;
      else
        val = get_poisson () * dist_node->mean_length;
      dist_node->remaining = (int) CLAMP (val, 1, G_MAXINT);
    }
  else
    node->output[0] = node->output[1] = 0.0;
}

static void
distribution_set_params(GdasMiniNetworkNode *node, const char *params)
{
  DistributionNode *dist_node = (DistributionNode *) node;;
  dist_node->mean_length = strtod (params, NULL);
  if (dist_node->is_gaussian)
    {
      const char *colon = strchr (params, ':');
      if (colon != NULL)
        dist_node->stddev = strtod (colon + 1, NULL);
    }
}

static GdasMiniNetworkNode* gdam_mini_network_node_distribution_new
                                             (const char* options,
                                              gpointer    data)
{
  DistributionNode *dist_node;
  GdasMiniNetworkNode *node;
  dist_node = g_new0 (DistributionNode, 1);
  node = (GdasMiniNetworkNode *) dist_node;
  node->set_params = distribution_set_params;
  node->update = distribution_update;
  dist_node->is_gaussian = GPOINTER_TO_INT (data);
  if (dist_node->is_gaussian)
    node->class_name = "gaussian_distribution";
  else
    node->class_name = "paisson_distribution";
  distribution_set_params (node, options);
  return node;
}


/*
 *   __       _ _        __  __
 *  / _| __ _| | | ___  / _|/ _|
 * | |_ / _` | | |/ _ \| |_| |_
 * |  _| (_| | | | (_) |  _|  _|
 * |_|  \__,_|_|_|\___/|_| |_|
 * 
 */
typedef struct _GdasMiniNetworkFalloff GdasMiniNetworkFalloff;
struct _GdasMiniNetworkFalloff {
	GdasMiniNetworkNode	node;
	gdouble			falloff_per_sample;
};

static void falloff_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkFalloff *falloff;
  gdouble val;
  gdouble decay;
  gdouble tmp;
  GdasMiniNetworkNode *input = node->input_nodes[0];
  falloff = (GdasMiniNetworkFalloff *) node;
  decay = falloff->falloff_per_sample;

  val = abs (input->output[0]);
  tmp = node->output[0] * decay;
  node->output[0] = MAX (tmp, val);

  val = abs (input->output[1]);
  tmp = node->output[1] * decay;
  node->output[1] = MAX (tmp, val);
}
#define FALLOFF_TOO_SMALL 0.00001
static void falloff_update_no_input(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkFalloff *falloff;
  gdouble decay;
  falloff = (GdasMiniNetworkFalloff *) node;
  decay = falloff->falloff_per_sample;

  node->output[0] *= decay;
  node->output[1] *= decay;
  if (node->output[0] < FALLOFF_TOO_SMALL
      && node->output[1] < FALLOFF_TOO_SMALL)
    {
      node->output[0] = node->output[1] = 0.0;
      node->update = NULL;
    }
}


static gint  falloff_connect(GdasMiniNetworkNode* node,
                             const char*          input_name,
			     GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  node->update = falloff_update;
  return 0;
}

static void  falloff_disconnect (GdasMiniNetworkNode* node,
			         guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = falloff_update_no_input;
}

/* (decay ** (halflife * 44100)) == 0.5. */
/* decay = pow(0.5, 1 / (halflife * 44100)) */
static void falloff_set_params(GdasMiniNetworkNode* node,
			       const char*          description)
{
  GdasMiniNetworkFalloff *falloff = (GdasMiniNetworkFalloff *) node;
  gdouble halflife = strtod (description, NULL);
  if (halflife < 0.0000001)
    {
      falloff->falloff_per_sample = 0.0;
    }
  else
    {
      /* Testing the bounds of your systems pow function: */
      falloff->falloff_per_sample = pow (0.5, 1.0 / (halflife * 44100.0));
    }
}

static GdasMiniNetworkNode* gdas_mini_network_node_falloff
                                                      (const char* options,
                                                       gpointer    data)
{
  GdasMiniNetworkFalloff *falloff_node;
  GdasMiniNetworkNode *node;
  falloff_node = g_new0 (GdasMiniNetworkFalloff, 1);
  node = (GdasMiniNetworkNode *) falloff_node;
  node->class_name = "wavetable";
  node->connect = falloff_connect;
  node->disconnect = falloff_disconnect;
  node->set_params = falloff_set_params;
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  falloff_set_params (node, (char *) options);
  return node;
}

/*
 *  _                _                    _
 * | | _____   _____| |    _ __ ___   ___| |_ ___ _ __
 * | |/ _ \ \ / / _ \ |   | '_ ` _ \ / _ \ __/ _ \ '__|
 * | |  __/\ V /  __/ |   | | | | | |  __/ ||  __/ |
 * |_|\___| \_/ \___|_|___|_| |_| |_|\___|\__\___|_|
 *                   |_____|
 */
typedef struct _GdasMiniNetworkLevelMeter GdasMiniNetworkLevelMeter;
struct _GdasMiniNetworkLevelMeter {
	GdasMiniNetworkNode node;
	int analysis_length;
	int analysis_remain;
	gdouble last_last_left_max;
	gdouble last_left_max;
	gdouble cur_left_max;
	gdouble last_last_right_max;
	gdouble last_right_max;
	gdouble cur_right_max;
};

static void level_meter_set_params(GdasMiniNetworkNode* node,
			           const char*          description)
{
  GdasMiniNetworkLevelMeter *level;
  level = (GdasMiniNetworkLevelMeter *) node;
  level->analysis_length = atoi (description);
  if (level->analysis_length < 1)
    level->analysis_length = 1;
  if (level->analysis_remain > level->analysis_length)
    level->analysis_remain = level->analysis_length;
}
static void level_meter_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkLevelMeter *level;
  gdouble abs_left, abs_right;
  int analysis_length;
  int analysis_remain;

  level = (GdasMiniNetworkLevelMeter *) node;
  analysis_remain = level->analysis_remain;
  analysis_length = level->analysis_length;

  abs_left = ABS (input->output[0]);
  abs_right = ABS (input->output[1]);
  if (abs_left > level->cur_left_max)
    level->cur_left_max = abs_left;
  if (abs_right > level->cur_right_max)
    level->cur_right_max = abs_right;
  node->output[0]
    = ((level->last_last_left_max * analysis_remain)
       + (level->last_left_max * (analysis_length - analysis_remain)))
    / analysis_length;
  node->output[1]
    = ((level->last_last_right_max * analysis_remain)
       + (level->last_right_max * (analysis_length - analysis_remain)))
    / analysis_length;
  if (level->analysis_remain-- <= 0)
    {
      level->analysis_remain = analysis_length;
      level->last_last_left_max = level->last_left_max;
      level->last_left_max = level->cur_left_max;
      level->cur_left_max = 0.0;
      level->last_last_right_max = level->last_right_max;
      level->last_right_max = level->cur_right_max;
      level->cur_right_max = 0.0;
    }

}
static void level_meter_no_input_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkLevelMeter *level;
  int analysis_length;
  int analysis_remain;

  level = (GdasMiniNetworkLevelMeter *) node;
  analysis_remain = level->analysis_remain;
  analysis_length = level->analysis_length;

  node->output[0]
    = ((level->last_last_left_max * analysis_remain)
       + (level->last_left_max * (analysis_length - analysis_remain)))
    / analysis_length;
  node->output[1]
    = ((level->last_last_left_max * analysis_remain)
       + (level->last_right_max * (analysis_length - analysis_remain)))
    / analysis_length;
  if (level->analysis_remain-- <= 0)
    {
      level->analysis_remain = analysis_length;
      level->last_last_left_max = level->last_left_max;
      level->last_left_max = level->cur_left_max;
      level->last_last_right_max = level->last_right_max;
      level->last_right_max = level->cur_right_max;
      if (level->last_last_left_max == 0
	  && level->last_left_max == 0
	  && level->last_last_right_max == 0 && level->last_right_max == 0)
	{
	  node->update = output_silence;
	}
    }
}

static gint  level_meter_connect(GdasMiniNetworkNode* node,
                                 const char*          input_name,
			         GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  node->update = level_meter_update;
  return 0;
}

static void  level_meter_disconnect (GdasMiniNetworkNode* node,
			             guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = level_meter_no_input_update;
}


static GdasMiniNetworkNode*
gdas_mini_network_node_level_meter(const char* options,
                                   gpointer    data)
{
  GdasMiniNetworkLevelMeter *level_meter;
  GdasMiniNetworkNode *node;
  level_meter = g_new0 (GdasMiniNetworkLevelMeter, 1);
  node = (GdasMiniNetworkNode *) level_meter;
  node->class_name = "level_meter";
  node->connect = level_meter_connect;
  node->disconnect = level_meter_disconnect;
  node->set_params = level_meter_set_params;
  node->needed_inputs = g_new0 (guint8, 1);
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->update = output_silence;
  level_meter_set_params (node, (char *) options);
  return node;
}

/*
 *       _ _          ___          _
 *   ___| (_)_ __    ( _ )     ___| | __ _ _ __ ___  _ __
 *  / __| | | '_ \   / _ \/\  / __| |/ _` | '_ ` _ \| '_ \
 * | (__| | | |_) | | (_>  < | (__| | (_| | | | | | | |_) |
 *  \___|_|_| .__/   \___/\/  \___|_|\__,_|_| |_| |_| .__/
 *          |_|                                     |_|
 */
typedef struct _GdasMiniNetworkClip GdasMiniNetworkClip;
struct _GdasMiniNetworkClip
{
  GdasMiniNetworkNode	node;
  gdouble		lower_left_value;
  gdouble		lower_right_value;
  gdouble		upper_left_value;
  gdouble		upper_right_value;
  void (*update_func)(GdasMiniNetworkNode*);
};

static void clip_set_params(GdasMiniNetworkNode* node,
			      const char*          description)
{
  GdasMiniNetworkClip *clip_node = (GdasMiniNetworkClip *) node;
  const char *strs[4];

  split_by_colon (strs, description, 4, TRUE);
  clip_node->lower_left_value = strtod (strs[0], NULL);
  clip_node->upper_left_value = strtod (strs[1], NULL);
  clip_node->lower_right_value = strtod (strs[2], NULL);
  clip_node->upper_right_value = strtod (strs[3], NULL);
}

static void clamp_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkClip *clip_node = (GdasMiniNetworkClip *) node;
  node->output[0] = CLAMP (input->output[0],
			   clip_node->lower_left_value,
			   clip_node->upper_left_value);
  node->output[1] = CLAMP (input->output[1],
			   clip_node->lower_right_value,
			   clip_node->upper_right_value);
}
static void clip_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkClip *clip_node = (GdasMiniNetworkClip *) node;
  if (input->output[0] < clip_node->lower_left_value
      || input->output[0] > clip_node->upper_left_value)
    node->output[0] = 0.0;
  else
    node->output[0] = input->output[0];
  if (input->output[1] < clip_node->lower_right_value
      || input->output[1] > clip_node->upper_right_value)
    node->output[1] = 0.0;
  else
    node->output[1] = input->output[1];
}
static void trigger_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkClip *clip_node = (GdasMiniNetworkClip *) node;
  if (input->output[0] < clip_node->lower_left_value)
    node->output[0] = clip_node->lower_left_value;
  else if (input->output[0] > clip_node->upper_left_value)
    node->output[0] = clip_node->upper_left_value;
  else
    node->output[0] = 0.0;

  if (input->output[1] < clip_node->lower_right_value)
    node->output[1] = clip_node->lower_right_value;
  else if (input->output[1] > clip_node->upper_right_value)
    node->output[1] = clip_node->upper_right_value;
  else
    node->output[1] = 0.0;
}
static void gate_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkClip *clip_node = (GdasMiniNetworkClip *) node;
  if (input->output[0] < clip_node->lower_left_value
      || input->output[0] > clip_node->upper_left_value)
    node->output[0] = input->output[0];
  else
    node->output[0] = 0.0;
  if (input->output[1] < clip_node->lower_right_value
      || input->output[1] > clip_node->upper_right_value)
    node->output[1] = input->output[1];
  else
    node->output[1] = 0.0;
}


static gint  clip_connect(GdasMiniNetworkNode* node,
                          const char*          input_name,
			  GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->update = ((GdasMiniNetworkClip *) node)->update_func;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}
static void  clip_disconnect (GdasMiniNetworkNode* node,
			      guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}
static GdasMiniNetworkNode*
gdas_mini_network_node_clip_generic (const char* options,
				     const char* class_name,
                                     void (*update_func)(GdasMiniNetworkNode*))
{
  GdasMiniNetworkClip *clip_node;
  GdasMiniNetworkNode *node;
  clip_node = g_new0 (GdasMiniNetworkClip, 1);
  node = (GdasMiniNetworkNode *) clip_node;
  node->class_name = class_name;
  node->set_params = clip_set_params;
  node->update = output_silence;
  node->connect = clip_connect;
  node->disconnect = clip_disconnect;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  clip_node->update_func = update_func;
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  clip_set_params (node, (char *) options);
  return node;
}

static GdasMiniNetworkNode* gdas_mini_network_node_clip(const char* options,
                                                          gpointer    data)
{
  return gdas_mini_network_node_clip_generic (options, "clip", clip_update);
}

static GdasMiniNetworkNode* gdas_mini_network_node_clamp(const char* options,
                                                         gpointer    data)
{
  return gdas_mini_network_node_clip_generic (options, "clamp", clamp_update);
}
static GdasMiniNetworkNode* gdas_mini_network_node_trigger(const char* options,
                                                           gpointer    data)
{
  return gdas_mini_network_node_clip_generic (options,
					      "trigger", trigger_update);
}

static GdasMiniNetworkNode* gdas_mini_network_node_gate(const char* options,
                                                        gpointer    data)
{
  return gdas_mini_network_node_clip_generic (options, "gate", gate_update);
}

/*
 *            _         _ _           _ _
 *  _ __ __ _| |_ ___  | (_)_ __ ___ (_) |_ ___ _ __ ___
 * | '__/ _` | __/ _ \ | | | '_ ` _ \| | __/ _ \ '__/ __|
 * | | | (_| | ||  __/ | | | | | | | | | ||  __/ |  \__ \
 * |_|  \__,_|\__\___| |_|_|_| |_| |_|_|\__\___|_|  |___/
 * 
 */
typedef struct _GdasMiniNetworkRateLimiter GdasMiniNetworkRateLimiter;
struct _GdasMiniNetworkRateLimiter
{
  GdasMiniNetworkNode node;
  gdouble            up_rate;
  gdouble            down_rate;
  gboolean           is_exponential;
};

static void exponential_rate_limiter_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkRateLimiter *limiter;
  gdouble input, last_output, output;
  limiter = (GdasMiniNetworkRateLimiter *) node;

  /* left sample */
  input = node->input_nodes[0]->output[0];
  last_output = node->last_output[0];
  if (input < last_output)
    output = last_output + (input - last_output) * limiter->down_rate;
  else
    output = last_output + (input - last_output) * limiter->up_rate;
  node->output[0] = output;

  /* right sample */
  input = node->input_nodes[0]->output[1];
  last_output = node->last_output[1];
  if (input < last_output)
    output = last_output + (input - last_output) * limiter->down_rate;
  else
    output = last_output + (input - last_output) * limiter->up_rate;
  node->output[1] = output;
}

static void linear_rate_limiter_update(GdasMiniNetworkNode* node)
{
  gdouble delta;
  gdouble input, last_output, output;
  GdasMiniNetworkRateLimiter *limiter;
  limiter = (GdasMiniNetworkRateLimiter *) node;

  input = node->input_nodes[0]->output[0];
  last_output = node->last_output[0];
  delta = input - last_output;
  node->output[0] = last_output
                  + CLAMP (delta, -limiter->down_rate, limiter->up_rate);

  input = node->input_nodes[0]->output[1];
  last_output = node->last_output[1];
  delta = input - last_output;
  node->output[1] = last_output
                  + CLAMP (delta, -limiter->down_rate, limiter->up_rate);
}

static void rate_limiter_set_params(GdasMiniNetworkNode* node,
			            const char*          description)
{
  GdasMiniNetworkRateLimiter *limiter;
  const char *colon;

  limiter = (GdasMiniNetworkRateLimiter *) node;
  limiter->down_rate = strtod (description, NULL);
  colon = strchr (description, ':');
  if (colon != NULL)
    limiter->up_rate = strtod (colon + 1, NULL);
}

static int  rate_limiter_connect(GdasMiniNetworkNode* node,
                                 const char*          input_name,
			         GdasMiniNetworkNode* new_input)
{
  gboolean is_exp;
  if (node->num_inputs > 0)
    return -1;
  is_exp = ((GdasMiniNetworkRateLimiter *) node)->is_exponential;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  if (is_exp)
    node->update = exponential_rate_limiter_update;
  else
    node->update = linear_rate_limiter_update;
  return 0;
}

static GdasMiniNetworkNode* gdam_mini_network_node_rate_limiter_new
                                             (const char* options,
                                              gpointer    data)
{
  GdasMiniNetworkRateLimiter *limiter_node;
  GdasMiniNetworkNode *node;
  limiter_node = g_new0 (GdasMiniNetworkRateLimiter, 1);
  node = (GdasMiniNetworkNode *) limiter_node;
  node->set_params = rate_limiter_set_params;
  node->connect = rate_limiter_connect;
  node->disconnect = gdas_mini_network_node_generic_disconnect;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  limiter_node->is_exponential = GPOINTER_TO_INT (data);
  if (limiter_node->is_exponential)
    node->class_name = "exponential_rate_limiter";
  else
    node->class_name = "linear_rate_limiter";
  rate_limiter_set_params (node, options);
  return node;
}
/*
 *                                    _
 *   ___  _ __   ___      _ __   ___ | | ___
 *  / _ \| '_ \ / _ \    | '_ \ / _ \| |/ _ \
 * | (_) | | | |  __/    | |_) | (_) | |  __/
 *  \___/|_| |_|\___|____| .__/ \___/|_|\___|
 *                 |_____|_|
 */
typedef struct _GdasMiniNetworkOnePole GdasMiniNetworkOnePole;
struct _GdasMiniNetworkOnePole
{
  GdasMiniNetworkNode node;
  gdouble             last_in_coeff_double;
  gdouble             in_coeff_double;
  gdouble             out_coeff_double;


  /* Instead of using ``last_in'' -- use input->last_output. */
  /*gdouble             last_in[2];*/
};


static void one_pole_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode* input = node->input_nodes[0];
  GdasMiniNetworkOnePole* one_pole_node = (GdasMiniNetworkOnePole*) node;
  gdouble coeff_x0 = one_pole_node->last_in_coeff_double;
  gdouble coeff_x1 = one_pole_node->in_coeff_double;
  gdouble coeff_y1 = one_pole_node->out_coeff_double;

  node->output[0] = coeff_x0 * input->output[0]
		  + coeff_x1 * input->last_output[0]
		  + coeff_y1 * node->last_output[0];
  node->output[1] = coeff_x0 * input->output[1]
		  + coeff_x1 * input->last_output[1]
		  + coeff_y1 * node->last_output[1];
}

static gint  one_pole_connect(GdasMiniNetworkNode* node,
                              const char*          input_name,
			      GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->update = one_pole_update;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}
static void  one_pole_disconnect (GdasMiniNetworkNode* node,
			          guint                pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}

static GdasMiniNetworkNode*
gdas_mini_network_node_one_pole_generic(const char* options,
				        const char* class_name,
			     void (*set_params)(GdasMiniNetworkOnePole*,
				                       const char* options))
{
  GdasMiniNetworkOnePole *one_pole_node;
  GdasMiniNetworkNode *node;
  one_pole_node = g_new0 (GdasMiniNetworkOnePole, 1);
  node = (GdasMiniNetworkNode *) one_pole_node;
  node->class_name = class_name;
  node->set_params = (void (*)(GdasMiniNetworkNode*,const char*)) set_params;
  node->connect = one_pole_connect;
  node->disconnect = one_pole_disconnect;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  (*set_params) (one_pole_node, options);
  return node;
}

static void one_pole_hipass_set_params(GdasMiniNetworkOnePole *node,
                                       const char             *options)
{
  gdouble cutoff_freq = strtod (options, NULL) / 44100.0;
  gdouble param;
  if (cutoff_freq <= 0)
    return;

  param = exp (-2.0 * M_PI * cutoff_freq);
  node->last_in_coeff_double = -(1.0 + param) / 2.0;
  node->in_coeff_double = -node->last_in_coeff_double;
  node->out_coeff_double = param;;
}

static void one_pole_lopass_set_params(GdasMiniNetworkOnePole *node,
                                       const char             *options)
{
  gdouble cutoff_freq = strtod (options, NULL) / 44100.0;
  gdouble param;
  if (cutoff_freq <= 0)
    return;

  param = exp (-2.0 * M_PI * cutoff_freq);
  node->last_in_coeff_double = 0.0;
  node->in_coeff_double = 1.0 - param;
  node->out_coeff_double = param;;
}
static GdasMiniNetworkNode *
        gdam_mini_network_note_onepole(const char *options, gpointer data)
{
  if (GINT_TO_POINTER (data) != FALSE)
    return gdas_mini_network_node_one_pole_generic
                                   (options, "onepole_lopass",
                                    one_pole_lopass_set_params);
  else
    return gdas_mini_network_node_one_pole_generic
                                   (options, "onepole_hipass",
                                    one_pole_hipass_set_params);
}

typedef struct _GdasMiniNetworkTwoPole GdasMiniNetworkTwoPole;
struct _GdasMiniNetworkTwoPole {
	GdasMiniNetworkNode node;
	gdouble		    last_in_coeff_double;
	gdouble		    coeff_x0, coeff_x1, coeff_x2;
	gdouble		    coeff_y1, coeff_y2;


	/* For the sample before node->last_output */
	gdouble             last_out[2];

	/* For the sample before input->last_output */
	gdouble             last_in[2];
};


static void two_pole_update(GdasMiniNetworkNode* node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkTwoPole *two_pole_node = (GdasMiniNetworkTwoPole *) node;
  gdouble coeff_x0 = two_pole_node->coeff_x0;
  gdouble coeff_x1 = two_pole_node->coeff_x1;
  gdouble coeff_y1 = two_pole_node->coeff_y1;
  gdouble coeff_x2 = two_pole_node->coeff_x2;
  gdouble coeff_y2 = two_pole_node->coeff_y2;

  node->output[0] = coeff_x0 * input->output[0]
    + coeff_x1 * input->last_output[0]
    + coeff_y1 * node->last_output[0]
    + coeff_x2 * two_pole_node->last_in[0]
    + coeff_y2 * two_pole_node->last_out[0];
  node->output[1] = coeff_x0 * input->output[1]
    + coeff_x1 * input->last_output[1]
    + coeff_y1 * node->last_output[1]
    + coeff_x2 * two_pole_node->last_in[1]
    + coeff_y2 * two_pole_node->last_out[1];
  two_pole_node->last_in[0] = input->last_output[0];
  two_pole_node->last_in[1] = input->last_output[1];
  two_pole_node->last_out[0] = input->last_output[0];
  two_pole_node->last_out[1] = input->last_output[1];
}

static gint  two_pole_connect(GdasMiniNetworkNode* node,
                              const char*          input_name,
			      GdasMiniNetworkNode* new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->update = two_pole_update;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}

static GdasMiniNetworkNode*
gdas_mini_network_node_two_pole_generic(const char* options,
				        const char* class_name,
                                     void (*set_params)(GdasMiniNetworkNode*,
				                       const char* options))
{
  GdasMiniNetworkTwoPole *two_pole_node;
  GdasMiniNetworkNode *node;
  two_pole_node = g_new0 (GdasMiniNetworkTwoPole, 1);
  node = (GdasMiniNetworkNode *) two_pole_node;
  node->class_name = class_name;
  node->set_params = set_params;
  node->connect = two_pole_connect;

  /* one pole disconnect does just the right thing - reuse it */
  node->disconnect = one_pole_disconnect;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->needed_inputs = g_new (guint8, 1);
  node->needed_inputs[0] = 1;
  (*set_params) (node, options);
  return node;
}


/*
 *   __                                                         _    ___
 *  / _|_ __ ___  __ _ _   _  ___ _ __   ___ _   _     ___  ___| |_ / _ \
 * | |_| '__/ _ \/ _` | | | |/ _ \ '_ \ / __| | | |   / _ \/ __| __| | | |
 * |  _| | |  __/ (_| | |_| |  __/ | | | (__| |_| |  |  __/\__ \ |_| |_| |
 * |_| |_|  \___|\__, |\__,_|\___|_| |_|\___|\__, |___\___||___/\__|\___/
 *                  |_|                      |___/_____|
 */

typedef struct _GdasMiniNetworkFrequencyEst0 GdasMiniNetworkFrequencyEst0;
struct _GdasMiniNetworkFrequencyEst0
{
  GdasMiniNetworkNode node;
  int analysis_length;
  GSList* first_zero_crossing_time;
  GSList* last_zero_crossing_time;
  int num_zero_crossings;
  int cur_time;
  int old_state;
  gdouble threshold;

  /* Ratio of
   *      frequency_in_hertz
   *      ------------------
   *      num zero crossings
   */
  gdouble pitch_ratio;
};
static inline
void discard_old_zeroes(GdasMiniNetworkFrequencyEst0* frequency)
{
  int last_allowed;

  if (frequency->first_zero_crossing_time == NULL)
    return;

  last_allowed = frequency->cur_time - frequency->analysis_length;

  for (;;)
    {
      int t;
      GSList *list = frequency->first_zero_crossing_time;
      t = GPOINTER_TO_INT (list->data);
      if (t < last_allowed)
	{
	  frequency->num_zero_crossings--;
	  frequency->first_zero_crossing_time
	    = g_slist_remove (list, list->data);
	  if (frequency->first_zero_crossing_time == NULL)
	    {
	      frequency->last_zero_crossing_time = NULL;
	      return;
	    }
	}
      else
	return;

    }
}
static void
frequency_est0_set_params (GdasMiniNetworkNode * node,
			   const char *description)
{
  GdasMiniNetworkFrequencyEst0 *frequency;
  const char *strs[2];
  frequency = (GdasMiniNetworkFrequencyEst0 *) node;

  split_by_colon (strs, description, 2, FALSE);
  frequency->analysis_length = atoi (description);
  if (frequency->analysis_length < 1)
    frequency->analysis_length = 1;
  discard_old_zeroes (frequency);
  frequency->pitch_ratio = (gdouble) (44100 / 2) / frequency->analysis_length;

  if (strs[1] != NULL)
    {
      frequency->threshold = strtod (strs[1], NULL);
      if (frequency->threshold < 0.00001)
	frequency->threshold = 0.00001;
    }
  else
    {
      frequency->threshold = 0.001;
    }
}

static inline void
insert_zero_crossing (GdasMiniNetworkFrequencyEst0 * frequency)
{
  frequency->last_zero_crossing_time
    = g_slist_append (frequency->last_zero_crossing_time,
		      GINT_TO_POINTER (frequency->cur_time));
  if (frequency->first_zero_crossing_time == NULL)
    {
      frequency->first_zero_crossing_time
	= frequency->last_zero_crossing_time;
    }
  else
    {
      frequency->last_zero_crossing_time
	= frequency->last_zero_crossing_time->next;
    }
  frequency->num_zero_crossings++;
}
static void
frequency_est0_update (GdasMiniNetworkNode * node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  GdasMiniNetworkFrequencyEst0 *level;
  int analysis_length;
  int new_state;
  gdouble threshold;

  level = (GdasMiniNetworkFrequencyEst0 *) node;
  threshold = level->threshold;
  analysis_length = level->analysis_length;
  new_state = (input->output[0] < -threshold) ? -1
    : (input->output[0] > +threshold) ? +1 : 0;
  if (new_state * level->old_state == -1)
    insert_zero_crossing (level);
  level->old_state = new_state;
  level->cur_time++;
  discard_old_zeroes (level);

  node->output[0] = node->output[1]
    = level->pitch_ratio * level->num_zero_crossings;
}

static void
frequency_est0_no_input_update (GdasMiniNetworkNode * node)
{
  GdasMiniNetworkFrequencyEst0 *level;
  int analysis_length;
  gdouble threshold;

  level = (GdasMiniNetworkFrequencyEst0 *) node;
  threshold = level->threshold;
  analysis_length = level->analysis_length;
  level->cur_time++;
  discard_old_zeroes (level);

  node->output[0] = node->output[1]
    = level->pitch_ratio * level->num_zero_crossings;
  if (level->num_zero_crossings == 0)
    node->update = output_silence;
}

static gint
frequency_est0_connect (GdasMiniNetworkNode * node,
			const char *input_name,
			GdasMiniNetworkNode * new_input)
{
  if (node->num_inputs > 0)
    return -1;
  node->num_inputs = 1;
  node->input_nodes[0] = new_input;
  node->update = frequency_est0_update;
  return 0;
}

static void
frequency_est0_disconnect (GdasMiniNetworkNode * node, guint pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = frequency_est0_no_input_update;
}


static GdasMiniNetworkNode *
gdas_mini_network_node_frequency_est0 (const char *options, gpointer data)
{
  GdasMiniNetworkFrequencyEst0 *frequency_est0;
  GdasMiniNetworkNode *node;
  frequency_est0 = g_new0 (GdasMiniNetworkFrequencyEst0, 1);
  node = (GdasMiniNetworkNode *) frequency_est0;
  node->class_name = "frequency_estimator";
  node->connect = frequency_est0_connect;
  node->disconnect = frequency_est0_disconnect;
  node->set_params = frequency_est0_set_params;
  node->needed_inputs = g_new0 (guint8, 1);
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 1);
  node->pipe_ids = g_new0 (guint, 1);
  node->update = output_silence;
  frequency_est0_set_params (node, (char *) options);
  return node;
}

/*
 *                    _   _            _
 *   __ _ _   _  ___ | |_(_) ___ _ __ | |_
 *  / _` | | | |/ _ \| __| |/ _ \ '_ \| __|
 * | (_| | |_| | (_) | |_| |  __/ | | | |_
 *  \__, |\__,_|\___/ \__|_|\___|_| |_|\__|
 *     |_|
 */

typedef struct _GdasMiniNetworkQuotient GdasMiniNetworkQuotient;
struct _GdasMiniNetworkQuotient
{
  GdasMiniNetworkNode node;
  gboolean has_numer, has_denom;
};

static void
update_quotient_no_numer (GdasMiniNetworkNode * node)
{
  gdouble in[2];
  GdasMiniNetworkNode *input = node->input_nodes[0];
  in[0] = input->output[0];
  in[1] = input->output[1];
  if (-0.001 < in[0] && in[0] < 0.001)
    node->output[0] = 0.0;
  else
    node->output[0] = 1.0 / in[0];
  if (-0.001 < in[1] && in[1] < 0.001)
    node->output[1] = 0.0;
  else
    node->output[1] = 1.0 / in[1];
}
static void
update_quotient_no_denom (GdasMiniNetworkNode * node)
{
  GdasMiniNetworkNode *input = node->input_nodes[0];
  node->output[0] = input->output[0];
  node->output[1] = input->output[1];
}
static inline gdouble
safe_divide (gdouble n, gdouble d)
{
  /* Somewhat arbitrary divide-by-zero handling.
   * Should be safe, perhaps overzealous or slow. */
  gdouble min_d = (n < 0.0) ? (-n) : (n);
  min_d *= 0.0004;
  if (-min_d < d && d < min_d)
    d = (d < 0.0) ? min_d : (-min_d);
  return n / d;
}

static void
update_quotient (GdasMiniNetworkNode * node)
{
  GdasMiniNetworkNode *numer = node->input_nodes[0];
  GdasMiniNetworkNode *denom = node->input_nodes[1];
  node->output[0] = safe_divide (numer->output[0], denom->output[0]);
  node->output[1] = safe_divide (numer->output[1], denom->output[1]);
}

static void
find_quotient_update_function (GdasMiniNetworkQuotient * quotient)
{
  if (quotient->has_numer)
    {
      if (quotient->has_denom)
	quotient->node.update = update_quotient;
      else
	quotient->node.update = update_quotient_no_denom;
    }
  else
    {
      if (quotient->has_denom)
	quotient->node.update = update_quotient_no_numer;
      else
	quotient->node.update = output_silence;
    }
}

static gint
node_quotient_connect (GdasMiniNetworkNode * node,
		       const char *input_name,
		       GdasMiniNetworkNode * new_input)
{
  GdasMiniNetworkQuotient *quotient;
  quotient = (GdasMiniNetworkQuotient *) node;
  if (input_name[0] == 'n')
    {
      if (quotient->has_numer)
	return 0;
      if (quotient->has_denom)
	{
	  node->input_nodes[1] = node->input_nodes[0];
	  node->pipe_ids[1] = node->pipe_ids[0];
	}
      node->input_nodes[0] = new_input;
      quotient->has_numer = TRUE;
    }
  else
    {
      if (quotient->has_denom)
	return 0;
      if (quotient->has_numer)
	{
	  node->input_nodes[1] = new_input;
	}
      else
	{
	  node->input_nodes[0] = new_input;
	}
      quotient->has_denom = TRUE;
    }
  node->num_inputs++;
  find_quotient_update_function (quotient);
  return 0;
}
static void
node_quotient_disconnect (GdasMiniNetworkNode * node, guint pipe_id)
{
  GdasMiniNetworkQuotient *quotient;
  quotient = (GdasMiniNetworkQuotient *) node;
  if (node->num_inputs == 0)
    return;
  if (node->num_inputs == 1)
    {
      g_return_if_fail (node->pipe_ids[0] == pipe_id);
      node->num_inputs = 0;
      quotient->has_numer = quotient->has_denom = FALSE;
      node->update = output_silence;
      return;
    }

  if (node->pipe_ids[0] == pipe_id)
    {
      quotient->has_numer = FALSE;
      node->pipe_ids[0] = node->pipe_ids[1];
      node->input_nodes[0] = node->input_nodes[1];
    }
  else
    {
      g_return_if_fail (node->pipe_ids[1] == pipe_id);
      quotient->has_denom = FALSE;
    }
  node->num_inputs--;
  find_quotient_update_function (quotient);
}

static GdasMiniNetworkNode *
gdas_mini_network_node_quotient (const char *options, gpointer data)
{
  GdasMiniNetworkQuotient *rv;
  GdasMiniNetworkNode *node;
  rv = g_new0 (GdasMiniNetworkQuotient, 1);
  node = (GdasMiniNetworkNode *) rv;
  node->class_name = "quotient";
  node->disconnect = node_quotient_disconnect;
  node->connect = node_quotient_connect;
  node->needed_inputs = g_new (guint8, 2);
  node->needed_inputs[0] = 1;
  node->needed_inputs[1] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 2);
  node->pipe_ids = g_new0 (guint, 2);
  return node;

}

typedef struct _GdasMiniNetworkFunction GdasMiniNetworkFunction;
struct _GdasMiniNetworkFunction
{
  GdasMiniNetworkNode node;
  GdamMathFunction *function;
};
static void
function_update (GdasMiniNetworkNode * node)
{
  GdasMiniNetworkFunction *function;
  gdouble val;
  GdasMiniNetworkNode *input;
  GdamMathFunction *math_func;
  function = (GdasMiniNetworkFunction *) node;
  math_func = function->function;
  input = node->input_nodes[0];
  node->output[0] = gdam_math_function_eval (math_func, input->output + 0);
  node->output[1] = gdam_math_function_eval (math_func, input->output + 1);
}
static void
set_function_to_string (GdasMiniNetworkFunction * function, const char *str)
{
  GdasMiniNetworkNode *node;
  node = (GdasMiniNetworkNode *) function;
  if (function->function != NULL)
    gdam_math_function_destroy (function->function);
  function->function = gdam_math_function_parse (str, NULL);
  if (function->function != NULL && node->num_inputs > 0)
    node->update = function_update;
  else
    node->update = output_silence;
}

static gint
node_function_connect (GdasMiniNetworkNode * node,
		       const char *input_name,
		       GdasMiniNetworkNode * new_input)
{
  GdasMiniNetworkFunction *function;
  function = (GdasMiniNetworkFunction *) node;

  if (node->num_inputs > 0)
    return 0;
  node->update = function->function ? function_update : output_silence;
  node->input_nodes[0] = new_input;
  node->num_inputs = 1;
  return 0;
}
static void
node_function_disconnect (GdasMiniNetworkNode * node, guint pipe_id)
{
  if (node->num_inputs != 1)
    return;
  node->num_inputs = 0;
  node->update = output_silence;
}

static void
node_function_set_params (GdasMiniNetworkNode * node, const char *params)
{
  set_function_to_string ((GdasMiniNetworkFunction *) node, params);
}

static GdasMiniNetworkNode *
gdas_mini_network_node_function (const char *options, gpointer data)
{
  GdasMiniNetworkFunction *rv;
  GdasMiniNetworkNode *node;
  rv = g_new0 (GdasMiniNetworkFunction, 1);
  node = (GdasMiniNetworkNode *) rv;
  node->class_name = "function";
  node->disconnect = node_function_disconnect;
  node->connect = node_function_connect;
  node->set_params = node_function_set_params;
  node->needed_inputs = g_new (guint8, 2);
  node->needed_inputs[0] = 1;
  node->needed_inputs[1] = 1;
  node->input_nodes = g_new0 (GdasMiniNetworkNode *, 2);
  node->pipe_ids = g_new0 (guint, 2);
  set_function_to_string (rv, options);
  return node;
}


/*
 *  ___    _____    _   _           _
 * |_ _|  / / _ \  | \ | | ___   __| | ___  ___
 *  | |  / / | | | |  \| |/ _ \ / _` |/ _ \/ __|
 *  | | / /| |_| | | |\  | (_) | (_| |  __/\__ \
 * |___/_/  \___/  |_| \_|\___/ \__,_|\___||___/
 * 
 * There is one input and one output node per network.
 */
GdasMiniNetworkNode *
gdas_mini_network_node_input ()
{
  GdasMiniNetworkNode *node;
  node = g_new0 (GdasMiniNetworkNode, 1);
  node->class_name = "input";
  return node;
}

/*
 *  ____  ____   ____
 * |  _ \|  _ \ / ___|
 * | |_) | |_) | |
 * |  _ <|  __/| |___
 * |_| \_\_|    \____|
 * 
 */
static void
gdas_mini_network_func_add_node (GtkObject * object,
				 int num_in,
				 GdamArg * args_in,
				 int *num_out, GdamArg ** args_out)
{
  /*
   * Parameters:
   *    in:        type:STRING         the type of node
   *               options:STRING      options to create the node with
   *    out:       node_id:INT         the created node id.
   */
  const char *type = NULL;
  const char *options = NULL;
  int node_id;
  int i;
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (object);
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "type") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  type = GDAM_ARG_STRING (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "options") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  options = GDAM_ARG_STRING (args_in + i);
	  continue;
	}
      gdam_debug ("unknown option `%s' [to add_node]", args_in[i].name);
    }
  if (type == NULL)
    {
      g_warning ("``type'' is mandatory to add_node");
      network->last_node_id++;
      return;
    }
  node_id = gdas_mini_network_add_node (network, type, options);
  *num_out = 1;
  *args_out = gdam_arg_new_from_valist (1,
					"node_id",
					GDAM_ARG_TYPE_INT, node_id);
}
static void
gdas_mini_network_func_add_pipe (GtkObject * object,
				 int num_in,
				 GdamArg * args_in,
				 int *num_out, GdamArg ** args_out)
{
  /*
   * Parameters:
   *    in:        source:INT          source node id
   *               destination:INT     destination node id
   *               input_name:STRING   name of the input
   *    out:       node_id:INT         the created node id.
   */
  int src_id = -1, dst_id = -1;
  int pipe_id;
  int i;
  const char *input_name = NULL;
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (object);
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "source") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  src_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "destination") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  dst_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "input_name") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  input_name = GDAM_ARG_STRING (args_in + i);
	  continue;
	}
      gdam_debug ("unknown option `%s' [to add_pipe]", args_in[i].name);
    }
  if (src_id < 0 || dst_id < 0 || input_name == NULL)
    {
      g_warning ("source,destination,input_name are all mandatory "
		 "to add_pipe");
      network->last_pipe_id++;
      return;
    }
  pipe_id = gdas_mini_network_add_pipe (network, dst_id, input_name, src_id);
  *num_out = 1;
  *args_out = gdam_arg_new_from_valist (1,
					"pipe_id",
					GDAM_ARG_TYPE_INT, pipe_id);

}

static void gdas_mini_network_func_config_node
  (GtkObject * object,
   int num_in, GdamArg * args_in, int *num_out, GdamArg ** args_out)
{
  /*
   * Parameters:
   *    in:        node_id:INT          source node id
   *               options:STRING       destination node id
   */
  int node_id = -1;
  int i;
  const char *options = NULL;
  GdasMiniNetworkNode *node;
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (object);
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "node_id") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_INT)
	{
	  node_id = GDAM_ARG_INT (args_in + i);
	  continue;
	}
      if (strcmp (args_in[i].name, "options") == 0
	  && args_in[i].type == GDAM_ARG_TYPE_STRING)
	{
	  options = GDAM_ARG_STRING (args_in + i);
	  continue;
	}
      gdam_debug ("unknown option `%s' [to add_pipe]", args_in[i].name);
    }
  if (node_id < 0 || options == NULL)
    {
      g_warning ("node_id,options are all mandatory to add_pipe");
      return;
    }

  *num_out = 0;
  *args_out = NULL;

  node = find_node_by_number (network, node_id);
  if (node == 0)
    {
      gdam_debug ("asked to remove node %d which wasn't there");
      return;
    }
  gdam_debug ("config_node: %s / %s", node->class_name, options);
  if (node->set_params != NULL)
    (*node->set_params) (node, (char *) options);
}


/*
 *  ____
 * / ___|  ___  _   _ _ __ ___ ___
 * \___ \ / _ \| | | | '__/ __/ _ \
 *  ___) | (_) | |_| | | | (_|  __/
 * |____/ \___/ \__,_|_|  \___\___|
 * 
 *  _____                 _   _
 * |  ___|   _ _ __   ___| |_(_) ___  _ __  ___
 * | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 * |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 * 
 */



static void
gdas_mini_network_init (GdasMiniNetwork * mini_network)
{
  GdasMiniNetworkNode *node;
  int id;

  node = gdas_mini_network_node_input ();
  mini_network->input = node;
  node->node_id = 1;
  mini_network->all_nodes = g_slist_prepend (mini_network->all_nodes, node);

  node = gdas_mini_network_node_mix ();
  node->node_id = 2;
  mini_network->output = node;
  mini_network->all_nodes = g_slist_prepend (mini_network->all_nodes, node);
  mini_network->last_node_id = 2;
  mini_network->last_pipe_id = 0;
  mini_network->pipe_input = g_hash_table_new (NULL, NULL);
  mini_network->can_flow = FALSE;

  mini_network->dc_block_buffer = g_new0 (GdamSample, DC_BLOCK_LENGTH);

  mini_network->dc_block = TRUE;

  /*
     id = gdas_mini_network_add_pipe(mini_network, 2,
     "input", 1);
     g_return_if_fail(id == 1);
   */
}

#if GDAM_SAMPLE_BITS == 16
#define CLAMP_AT	1.0
#else /* 24-bits */
#define CLAMP_AT	3.0
#endif

static inline void
process_sample (GdasMiniNetwork * network, GdamSample * sample)
{
  GdasMiniNetworkNode *input_node = network->input;
  GdasMiniNetworkNode *output_node = network->output;
  GSList *list;
  gdouble val;
  for (list = network->eval_order; list != NULL; list = list->next)
    {
      GdasMiniNetworkNode *node = list->data;
      node->last_output[0] = node->output[0];
      node->last_output[1] = node->output[1];
    }
  input_node->output[0] = (gdouble) sample->s[0] / GDAM_SAMPLE_ONE;
  input_node->output[1] = (gdouble) sample->s[1] / GDAM_SAMPLE_ONE;
  /*g_message("input goes into node at %p", input_node); */
  for (list = network->eval_order; list != NULL; list = list->next)
    {
      GdasMiniNetworkNode *node = list->data;
      if (node->update != NULL)
	(*node->update) (node);
    }
  val = output_node->output[0];
  val = CLAMP (val, -CLAMP_AT, CLAMP_AT);
  sample->s[0] = (int) (val * GDAM_SAMPLE_ONE);
  val = output_node->output[1];
  val = CLAMP (val, -CLAMP_AT, CLAMP_AT);
  sample->s[1] = (int) (val * GDAM_SAMPLE_ONE);
}

static int
gdas_mini_network_read (GdasSource * source, GdamSample * buf, int max_read)
{
  int num_read;
  int i;
  int block_index;
  gboolean dc_block;
  GdamSample *block_buffer;
  int left_buffer_sum, right_buffer_sum;
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (source);

  dc_block = network->dc_block;
  block_index = network->dc_block_index;
  block_buffer = network->dc_block_buffer;
  left_buffer_sum = network->left_buffer_sum;
  right_buffer_sum = network->right_buffer_sum;

  if (!network->can_flow)
    {
      gdam_sample_zero (buf, max_read);
      return max_read;
    }
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buf, max_read);
  for (i = 0; i < num_read; i++)
    {
      process_sample (network, buf + i);
      if (dc_block)
	{
	  GdamSample *src_samp = buf + i;
	  GdamSample *buf_samp = block_buffer + block_index;
	  int tmp;

	  left_buffer_sum -= buf_samp->s[0];
	  right_buffer_sum -= buf_samp->s[1];
	  if (++block_index == DC_BLOCK_LENGTH)
	    block_index = 0;

	  tmp = src_samp->s[0] - left_buffer_sum / DC_BLOCK_LENGTH;
	  src_samp->s[0] = GDAM_SAMPLE_CLAMP (tmp);
	  tmp = src_samp->s[1] - right_buffer_sum / DC_BLOCK_LENGTH;
	  src_samp->s[1] = GDAM_SAMPLE_CLAMP (tmp);

	  buf_samp->s[0] = src_samp->s[0];
	  buf_samp->s[1] = src_samp->s[1];
	  left_buffer_sum += buf_samp->s[0];
	  right_buffer_sum += buf_samp->s[1];
	}
    }
  if (dc_block)
    {
      network->dc_block_index = block_index;
      network->left_buffer_sum = left_buffer_sum;
      network->right_buffer_sum = right_buffer_sum;
    }
  return num_read;
}

static void
gdas_mini_network_arg_set_dc_block (GtkObject * object, GdamArg * arg)
{
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (object);
  network->dc_block = GDAM_ARG_INT (arg);
}

static void
gdas_mini_network_arg_get_dc_block (GtkObject * object, GdamArg * arg)
{
  GdasMiniNetwork *network;
  network = GDAS_MINI_NETWORK (object);
  GDAM_ARG_INT (arg) = network->dc_block;
}

static void
gdas_mini_network_arg_remove_node (GtkObject * object, GdamArg * arg)
{
  GdasMiniNetwork *mini_network = GDAS_MINI_NETWORK (object);
  gdas_mini_network_remove_node (mini_network, GDAM_ARG_INT (arg));
}
static void
gdas_mini_network_arg_remove_pipe (GtkObject * object, GdamArg * arg)
{
  GdasMiniNetwork *mini_network = GDAS_MINI_NETWORK (object);
  gdas_mini_network_remove_pipe (mini_network, GDAM_ARG_INT (arg));
}

static void
gdas_mini_network_class_init (GdasMiniNetworkClass * klass)
{
  GtkObjectClass *object_class;
  GdasSourceClass *source_class;
  object_class = GTK_OBJECT_CLASS (klass);
  source_class = GDAS_SOURCE_CLASS (klass);
  source_class->read = gdas_mini_network_read;
  gdam_math_function_init ();
  gdam_arg_register (object_class,
		     "remove_node",
		     GDAM_ARG_TYPE_INT,
		     gdas_mini_network_arg_remove_node, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "dc_block",
		     GDAM_ARG_TYPE_INT,
		     gdas_mini_network_arg_set_dc_block,
		     gdas_mini_network_arg_get_dc_block,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "remove_pipe",
		     GDAM_ARG_TYPE_INT,
		     gdas_mini_network_arg_remove_pipe, NULL, NULL, NULL, 0);
  gdam_arg_function_register (object_class,
			      "add_node", gdas_mini_network_func_add_node);
  gdam_arg_function_register (object_class,
			      "add_pipe", gdas_mini_network_func_add_pipe);
  gdam_arg_function_register (object_class,
			      "configure_node",
			      gdas_mini_network_func_config_node);
  gdas_mini_network_node_type_register
    ("mix", (GdasMiniNetworkNodeMaker) gdas_mini_network_node_mix, NULL);
  gdas_mini_network_node_type_register
    ("modulator", gdas_mini_network_node_modulator, NULL);
  gdas_mini_network_node_type_register
    ("quotient", gdas_mini_network_node_quotient, NULL);

  gdas_mini_network_node_type_register
    ("function", gdas_mini_network_node_function, NULL);

  gdas_mini_network_node_type_register
    ("gain", gdas_mini_network_node_gain, NULL);
  gdas_mini_network_node_type_register
    ("balance", gdas_mini_network_node_balance, NULL);

  gdas_mini_network_node_type_register
    ("delay", gdas_mini_network_node_delay, NULL);
  gdas_mini_network_node_type_register
    ("variable_delay", gdas_mini_network_node_variable_delay, NULL);

  gdas_mini_network_node_type_register
    ("noise", gdas_mini_network_node_noise, NULL);
  gdas_mini_network_node_type_register
    ("sine", gdas_mini_network_node_sin, NULL);
  gdas_mini_network_node_type_register
    ("square", gdas_mini_network_node_square, NULL);
  gdas_mini_network_node_type_register
    ("triangle", gdas_mini_network_node_triangle, NULL);
  gdas_mini_network_node_type_register
    ("wavetable", gdas_mini_network_node_wavetable, NULL);
  gdas_mini_network_node_type_register
    ("constant", gdas_mini_network_node_constant, NULL);
  gdas_mini_network_node_type_register
    ("control", gdas_mini_network_node_control, NULL);
  gdas_mini_network_node_type_register
    ("pulse", gdas_mini_network_node_pulse, NULL);

  gdas_mini_network_node_type_register
    ("exponential_rate_limiter",
     gdam_mini_network_node_rate_limiter_new,
     GINT_TO_POINTER (TRUE));
  gdas_mini_network_node_type_register
    ("linear_rate_limiter",
     gdam_mini_network_node_rate_limiter_new,
     GINT_TO_POINTER (FALSE));
  gdas_mini_network_node_type_register
    ("gaussian_distribution",
     gdam_mini_network_node_distribution_new,
     GINT_TO_POINTER (TRUE));
  gdas_mini_network_node_type_register
    ("poisson_distribution",
     gdam_mini_network_node_distribution_new,
     GINT_TO_POINTER (FALSE));
  gdas_mini_network_node_type_register
    ("clip", gdas_mini_network_node_clip, NULL);
  gdas_mini_network_node_type_register
    ("clamp", gdas_mini_network_node_clamp, NULL);
  gdas_mini_network_node_type_register
    ("trigger", gdas_mini_network_node_trigger, NULL);
  gdas_mini_network_node_type_register
    ("gate", gdas_mini_network_node_gate, NULL);
  gdas_mini_network_node_type_register
    ("falloff", gdas_mini_network_node_falloff, NULL);
  gdas_mini_network_node_type_register
    ("level_meter", gdas_mini_network_node_level_meter, NULL);
  gdas_mini_network_node_type_register
    ("frequency_estimator", gdas_mini_network_node_frequency_est0, NULL);
  
  gdas_mini_network_node_type_register
    ("onepole_lopass", gdam_mini_network_note_onepole,
     GINT_TO_POINTER (TRUE));
  gdas_mini_network_node_type_register
    ("onepole_hipass", gdam_mini_network_note_onepole,
     GINT_TO_POINTER (FALSE));
}


GtkType gdas_mini_network_get_type ()
{
  static GtkType mini_network_type = 0;
  if (!mini_network_type)
    {
      static const GtkTypeInfo mini_network_info = {
	"GdasMiniNetwork",
	sizeof (GdasMiniNetwork),
	sizeof (GdasMiniNetworkClass),
	(GtkClassInitFunc) gdas_mini_network_class_init,
	(GtkObjectInitFunc) gdas_mini_network_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      mini_network_type = gtk_type_unique (parent, &mini_network_info);
    }
  return mini_network_type;
}

/*
 *  ____        _     _ _        ___       _             __
 * |  _ \ _   _| |__ | (_) ___  |_ _|_ __ | |_ ___ _ __ / _| __ _  ___ ___
 * | |_) | | | | '_ \| | |/ __|  | || '_ \| __/ _ \ '__| |_ / _` |/ __/ _ \
 * |  __/| |_| | |_) | | | (__   | || | | | ||  __/ |  |  _| (_| | (_|  __/
 * |_|    \__,_|_.__/|_|_|\___| |___|_| |_|\__\___|_|  |_|  \__,_|\___\___|
 * 
 */

guint
gdas_mini_network_add_node (GdasMiniNetwork * network,
			    const char *type_name, const char *options)
{
  int node_id;
  GdasMiniNetworkNodeStyle *style;
  GdasMiniNetworkNode *node;

  gdam_debug ("add_node: %s / %s", type_name, options);
  node_id = ++(network->last_node_id);

  style = g_hash_table_lookup (styles_by_name, (char *) type_name);
  if (style == NULL)
    {
      g_warning ("mininode type %s not found", type_name);
      return 0;
    }
  node = (*style->maker) (options, style->maker_data);
  if (node == NULL)
    {
      g_warning ("error making node of type %s", type_name);
      return 0;
    }
  node->node_id = node_id;
  network->all_nodes = g_slist_prepend (network->all_nodes, node);
  if (!gdas_mini_network_recompute_eval_order (network))
    {
      network->all_nodes = g_slist_remove (network->all_nodes, node);
      gdas_mini_network_node_destroy (node);
      return 0;
    }
  return node->node_id;
}
static GdasMiniNetworkNode *
find_node_by_number (GdasMiniNetwork * network, int id)
{
  GSList *list;
  for (list = network->all_nodes; list != NULL; list = list->next)
    {
      GdasMiniNetworkNode *node = list->data;
      if ((int) node->node_id == id)
	return node;
    }
  return NULL;
}

guint
gdas_mini_network_add_pipe (GdasMiniNetwork * network,
			    guint dst, const char *input_name, guint src)
{
  GdasMiniNetworkNode *dst_node;
  GdasMiniNetworkNode *src_node;
  guint pipe_id;
  int pipe_index;
  dst_node = find_node_by_number (network, dst);
  src_node = find_node_by_number (network, src);
  if (src_node == NULL)
    {
      gdam_debug ("source node %d not found", src);
      return 0;
    }
  if (dst_node == NULL)
    {
      gdam_debug ("destination node %d not found", dst);
      return 0;
    }
  if (dst_node->connect == NULL)
    {
      g_message ("%s has no connect method", dst_node->class_name);
      return 0;
    }
  pipe_index = (*dst_node->connect) (dst_node, input_name, src_node);
  if (pipe_index < 0)
    {
      gdam_debug ("error creating pipe %d=>%d (%s)", src, dst, input_name);
      return 0;
    }
  pipe_id = ++(network->last_pipe_id);
  //g_message("adding pipe %d from node %d to %d", pipe_id, src, dst);
  dst_node->pipe_ids[pipe_index] = pipe_id;
  src_node->output_pipes = g_slist_prepend (src_node->output_pipes,
					    GINT_TO_POINTER (pipe_id));
  g_hash_table_insert (network->pipe_input,
		       GINT_TO_POINTER (pipe_id), dst_node);
  if (!gdas_mini_network_recompute_eval_order (network))
    {
      gdas_mini_network_remove_pipe (network, pipe_id);
      return 0;
    }
  return pipe_id;
}

void
gdas_mini_network_remove_pipe (GdasMiniNetwork * network, guint pipe_id)
{
  GdasMiniNetworkNode *src_node;
  GdasMiniNetworkNode *dst_node;
  int pipe_index;

  dst_node = g_hash_table_lookup (network->pipe_input,
				  GINT_TO_POINTER (pipe_id));
  g_return_if_fail (dst_node != NULL);
  for (pipe_index = 0; pipe_index < dst_node->num_inputs; pipe_index++)
    if (dst_node->pipe_ids[pipe_index] == pipe_id)
      break;
  if (pipe_index == dst_node->num_inputs)
    {
      gdam_debug ("pipe id %d not found", pipe_id);
      return;
    }
  src_node = dst_node->input_nodes[pipe_index];
  src_node->output_pipes = g_slist_remove (src_node->output_pipes,
					   GINT_TO_POINTER (pipe_id));


  if (dst_node->disconnect != NULL)
    {
      (*dst_node->disconnect) (dst_node, pipe_id);
    }
  else
    {
      gdas_mini_network_node_generic_disconnect (dst_node, pipe_id);
    }

  /* Don't reevaluate here! It is unnecessary (b/c dependencies
   * are only decreased) and slow since this is called
   * in a loop from remove_node. */
}

void
gdas_mini_network_remove_node (GdasMiniNetwork * network, guint node_id)
{
  GSList *pipes_to_remove = NULL;
  int i;
  GSList *list;
  GdasMiniNetworkNode *node;
  g_return_if_fail (node_id >= 3);
  node = find_node_by_number (network, node_id);
  g_return_if_fail (node != NULL);
  pipes_to_remove = g_slist_copy (node->output_pipes);
  for (i = 0; i < node->num_inputs; i++)
    pipes_to_remove = g_slist_prepend (pipes_to_remove,
				       GINT_TO_POINTER (node->pipe_ids[i]));
  for (list = pipes_to_remove; list != NULL; list = list->next)
    gdas_mini_network_remove_pipe (network, GPOINTER_TO_INT (list->data));
  network->all_nodes = g_slist_remove (network->all_nodes, node);
  gdas_mini_network_recompute_eval_order (network);
}

/*
 *   ____                 _       _____                                   _
 *  / ___|_ __ __ _ _ __ | |__   |_   _| __ __ ___   _____ _ __ ___  __ _| |
 * | |  _| '__/ _` | '_ \| '_ \    | || '__/ _` \ \ / / _ \ '__/ __|/ _` | |
 * | |_| | | | (_| | |_) | | | |   | || | | (_| |\ V /  __/ |  \__ \ (_| | |
 *  \____|_|  \__,_| .__/|_| |_|   |_||_|  \__,_| \_/ \___|_|  |___/\__,_|_|
 *                 |_|
 */
static void
mark_untraversed (GdasMiniNetworkNode * node)
{
  node->has_traversed = FALSE;
}
static void
make_node_list (GdasMiniNetworkNode * node, GSList ** list)
{
  int i;
  if (node->has_traversed)
    return;
  *list = g_slist_prepend (*list, node);
  node->has_traversed = TRUE;
  for (i = 0; i < node->num_inputs; i++)
    make_node_list (node->input_nodes[i], list);
}

static gboolean
can_compute_node (GdasMiniNetworkNode * node)
{
  int i;
  for (i = 0; i < node->num_inputs; i++)
    {
      if (node->needed_inputs[i] && !node->input_nodes[i]->has_traversed)
	return FALSE;
    }
  return TRUE;
}
static GSList *
pass_through_nodes (GSList * list,
		    GdasMiniNetwork * network, gboolean * made_progress)
{
  GSList **plist;
  *made_progress = FALSE;

  for (plist = &list; *plist;)
    {
      GdasMiniNetworkNode *node;
      GSList *at = *plist;
      node = (GdasMiniNetworkNode *) ((*plist)->data);
      if (can_compute_node (node))
	{
	  node->has_traversed = TRUE;
	  *made_progress = TRUE;
	  network->eval_order = g_slist_prepend (network->eval_order, node);
	  *plist = at->next;
	}
      else
	{
	  plist = &(at->next);
	}
    }
  return list;
}

static void
print_node (GdasMiniNetworkNode * node)
{
  int i;
  fprintf (stderr, "node %d: %s\n", node->node_id, node->class_name);
  for (i = 0; i < node->num_inputs; i++)
    {
      fprintf (stderr, "    input from %d\n", node->input_nodes[i]->node_id);
    }
}

gboolean gdas_mini_network_recompute_eval_order (GdasMiniNetwork * network)
{
  GSList *to_eval = NULL;
  g_slist_foreach (network->all_nodes, (GFunc) mark_untraversed, NULL);

  g_slist_free (network->eval_order);
  network->eval_order = NULL;

  /* Find the list of nodes to evaluate. */
  make_node_list (network->output, &to_eval);

  g_slist_foreach (network->all_nodes, (GFunc) mark_untraversed, NULL);
  for (;;)
    {
      gboolean made_progress;
      to_eval = pass_through_nodes (to_eval, network, &made_progress);
      if (to_eval == NULL)
	break;
      if (!made_progress)
	{
	  gdam_debug ("couldn't evaluate network");
	  network->can_flow = FALSE;
	  return FALSE;
	}
    }
  network->can_flow = TRUE;
  network->eval_order = g_slist_reverse (network->eval_order);
  if (gdam_debugging_on)
    {
      //g_message("can flow: eval order = ");
      g_slist_foreach (network->eval_order, (GFunc) print_node, NULL);
    }
  return TRUE;
}
