/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasmoog.h"
#include <gdam/common/gdamsample.h>

static inline GdamSampleValue
do_pole(GdamSampleValue last, GdamSampleValue in, GdamSampleValue out, 
        GdamFixed one_minus_f)
{
  gint64 value;
  value = (GdamFixed_ONE * 3 / 10);
  value *= in;
  value += (GdamSampleProduct) out *one_minus_f;
  value += last;
  value /= GdamFixed_ONE;
  return (GdamSampleValue) value;
}

static int  gdas_moog_read(GdasSource*  source, 
                           GdamSample*  buf,
			   int          max_read)
{
  int num_read, remaining;
  GdamSample *at;
  GdasMoog *moog = GDAS_MOOG (source);
  GdamSample *out = moog->out;
  GdamSample *in = moog->in;
  GdamFixed input_scale = moog->input_scale;
  GdamFixed o_f = GdamFixed_ONE - moog->f, fb = moog->fb;
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buf, max_read);
  at = buf;
  remaining = num_read;
  g_message ("input_scale=%5.5g, moog->f = %5.5g, moog->fb=%5.5g",
	     GDAM_FIXED_TO_DOUBLE (input_scale),
	     GDAM_FIXED_TO_DOUBLE (moog->f), GDAM_FIXED_TO_DOUBLE (moog->fb));

  while (remaining-- > 0)
    {
#define DO_ONE_CHANNEL(ch)						    \
	{								    \
	gint64 input = at->s[ch];					    \
	input -= (out[3].s[ch] * fb) / GdamFixed_ONE;			    \
	input *= input_scale;						    \
	input >>= GdamFixed_LOG_ONE;					    \
	out[0].s[ch] = do_pole(input, in[0].s[ch], out[0].s[ch], o_f);	    \
	in[0].s[ch] = (gint) input;					    \
	out[1].s[ch] = do_pole(in[0].s[ch], in[1].s[ch], out[1].s[ch], o_f);\
	in[1].s[ch] = out[0].s[ch];				            \
	out[2].s[ch] = do_pole(in[1].s[ch], in[2].s[ch], out[2].s[ch], o_f);\
	in[2].s[ch] = out[1].s[ch];					    \
	out[3].s[ch] = do_pole(in[2].s[ch], in[3].s[ch], out[3].s[ch], o_f);\
	in[3].s[ch] = out[2].s[ch];				            \
	at->s[ch] = out[3].s[ch];					    \
	}
      DO_ONE_CHANNEL (0) DO_ONE_CHANNEL (1) at++;
    }
  return num_read;
}

#if 0
#define SAMPLING_RATE (44100.0)
#else
#define SAMPLING_RATE (2000.0)
#endif

static inline void update_private(GdasMoog* moog)
{
  gdouble freqcorrect = 1.16 / (double) (SAMPLING_RATE / 2.0);
  gdouble f;
  gdouble fb;

  f = GDAM_FIXED_TO_DOUBLE (moog->frequency) * freqcorrect;
  fb = GDAM_FIXED_TO_DOUBLE (moog->resonance) * (1.0 - 0.15 * f * f);

  moog->f = GDAM_FIXED_FROM_DOUBLE (f);
  moog->fb = GDAM_FIXED_FROM_DOUBLE (fb);
  moog->input_scale = GDAM_FIXED_FROM_DOUBLE (0.35013 * (f * f) * (f * f));
}
static void gdas_moog_arg_set_frequency(GtkObject* object, GdamArg* arg)
{
  GdasMoog *moog = GDAS_MOOG (object);
  moog->frequency = GDAM_FIXED_FROM_DOUBLE (GDAM_ARG_DOUBLE (arg));
  update_private (moog);
}
static void gdas_moog_arg_get_frequency(GtkObject* object, GdamArg* arg)
{
  GdasMoog *moog = GDAS_MOOG (object);
  GDAM_ARG_DOUBLE (arg) = GDAM_FIXED_TO_DOUBLE (moog->frequency);
}
static void gdas_moog_arg_set_resonance(GtkObject* object, GdamArg* arg)
{
  GdasMoog *moog = GDAS_MOOG (object);
  moog->resonance = GDAM_FIXED_FROM_DOUBLE (GDAM_ARG_DOUBLE (arg));
  moog->resonance = CLAMP (moog->resonance, 0, 4 * GdamFixed_ONE);
  update_private (moog);
}
static void gdas_moog_arg_get_resonance(GtkObject* object, GdamArg* arg)
{
  GdasMoog *moog = GDAS_MOOG (object);
  GDAM_ARG_DOUBLE (arg) = GDAM_FIXED_TO_DOUBLE (moog->resonance);
}
static void gdas_moog_init(GdasMoog* moog)
{
  moog->frequency = GdamFixed_ONE * 40;
  moog->resonance = 0;
  update_private (moog);
}

static void gdas_moog_class_init(GdasMoogClass* moog_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (moog_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (moog_class);
  source_class->read = gdas_moog_read;
  gdam_arg_register (object_class,
		     "frequency",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_moog_arg_set_frequency,
		     gdas_moog_arg_get_frequency,
		     NULL, "0,40000", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "resonance",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_moog_arg_set_resonance,
		     gdas_moog_arg_get_resonance,
		     NULL, "0,1", GDAM_ARG_VITAL);
}

GtkType gdas_moog_get_type()
{
  static GtkType moog_type = 0;
  if (!moog_type)
    {
      static const GtkTypeInfo moog_info = {
	"GdasMoog",
	sizeof (GdasMoog),
	sizeof (GdasMoogClass),
	(GtkClassInitFunc) gdas_moog_class_init,
	(GtkObjectInitFunc) gdas_moog_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      moog_type = gtk_type_unique (parent, &moog_info);
    }
  return moog_type;
}
