/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasoscilloscope.h"
#include <gdam/common/gdamplugin.h>

static GtkObjectClass* parent_class = NULL;

/* 
 * TODO: Put more love into this source:
 *
 *      - left/right/average options, maybe min/max
 *      - resample window options
 *      - fractional off-length
 */

static void
gdas_oscilloscope_arg_set_window_size(GtkObject* object, GdamArg* arg)
{
  GdasOscilloscope *oscilloscope;
  int window_size;
  window_size = GDAM_ARG_INT (arg);
  if (window_size <= 0)
    {
      g_warning ("window_size must be > 0");
      return;
    }
  oscilloscope = GDAS_OSCILLOSCOPE (object);
  oscilloscope->window_size = window_size;
  oscilloscope->window[0] = g_renew (GdamSample,
				     oscilloscope->window[0], window_size);
  oscilloscope->window[1] = g_renew (GdamSample,
				     oscilloscope->window[1], window_size);

  oscilloscope->cycle_index = 0;
  oscilloscope->filling = 0;
  oscilloscope->changed = FALSE;
}

static void
gdas_oscilloscope_arg_get_window_size(GtkObject* object, GdamArg* arg)
{
  GdasOscilloscope *oscilloscope;
  oscilloscope = GDAS_OSCILLOSCOPE (object);
  GDAM_ARG_INT (arg) = oscilloscope->window_size;
}

static void
gdas_oscilloscope_arg_set_off_length(GtkObject* object, GdamArg* arg)
{
  int length;
  length = GDAM_ARG_INT (arg);
  if (length < 0)
    {
      g_warning ("off cycle length must be >= 0");
      return;
    }
  GDAS_OSCILLOSCOPE (object)->off_length = length;
}

static void
gdas_oscilloscope_arg_get_off_length(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_OSCILLOSCOPE (object)->off_length;
}

/*
 * Compute the window.
 */
static void gdas_oscilloscope_fill_arg(GdasOscilloscope*    oscilloscope,
                                       GdamArg*             arg,
				       gboolean             only_if_mod)
{
  int length;
  int i;
  int *rv;
  GdamSample *window;

  if (only_if_mod && !oscilloscope->changed)
    {
      GDAM_ARG_INT_ARRAY (arg).length = 0;
      GDAM_ARG_INT_ARRAY (arg).array = NULL;
      return;
    }

  length = oscilloscope->window_size;
  GDAM_ARG_INT_ARRAY (arg).length = length;

  rv = g_new (int, length);
  window = oscilloscope->window[1 - oscilloscope->filling];
  for (i = 0; i < length; i++)
    rv[i] = (window[i].s[0]);

  GDAM_ARG_INT_ARRAY (arg).array = rv;
  oscilloscope->changed = FALSE;
}

static void
gdas_oscilloscope_arg_get_sample_data(GtkObject* object, GdamArg* arg)
{
  GdasOscilloscope *oscilloscope = GDAS_OSCILLOSCOPE (object);
  gdas_oscilloscope_fill_arg (oscilloscope, arg, FALSE);
}

static void
gdas_oscilloscope_arg_get_if_modified_sample_data
					(GtkObject* object, GdamArg* arg)
{
  GdasOscilloscope *oscilloscope = GDAS_OSCILLOSCOPE (object);
  gdas_oscilloscope_fill_arg (oscilloscope, arg, TRUE);
}

static int gdas_oscilloscope_read(GdasSource*         source,
                                  GdamSample*         buffer,
				  int                 max_read)
{
  int num_read;
  GdasOscilloscope *oscilloscope;
  GdamSample *at;
  int remain;
  int cycle_index;
  int filling;

  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buffer, max_read);

  /*
   * Now, figure out which part of the buffer to copy to where.
   *
   * Don't try to optimize the "multiple copies to the window
   * which overlap aren't needed" case yet;  much more common
   * under properly operating conditions is the "no copy" case,
   * which cannot be optimized...
   */
  oscilloscope = GDAS_OSCILLOSCOPE (source);
  at = buffer;
  remain = num_read;
  cycle_index = oscilloscope->cycle_index;
  filling = oscilloscope->filling;
  while (remain > 0)
    {
      /*g_return_val_if_fail(cycle_index <= oscilloscope->window_size,
         -1); */
      if (cycle_index <= -remain)
	{
	  cycle_index += remain;
	  remain = 0;
	}
      else if (cycle_index < 0)
	{
	  remain += cycle_index;
	  cycle_index = 0;
	}
      else
	{
	  int copy_amount;
	  GdamSample *window;
	  if (cycle_index + remain > oscilloscope->window_size)
	    {
	      copy_amount = oscilloscope->window_size - cycle_index;
	    }
	  else
	    {
	      copy_amount = remain;
	    }

	  window = oscilloscope->window[filling];
	  gdam_sample_copy (window + cycle_index, at, copy_amount);
	  remain -= copy_amount;
	  at += copy_amount;
	  cycle_index += copy_amount;
	  if (cycle_index == oscilloscope->window_size)
	    {
	      oscilloscope->changed = TRUE;
	      filling = 1 - filling;
	      cycle_index = -oscilloscope->off_length;
	    }
	}
    }
  oscilloscope->filling = filling;
  oscilloscope->cycle_index = cycle_index;
  return num_read;
}

/*
 * XXX: still need "miss data" virtual ...
 *      other skipping causes "window jitter"
 *      [probably completely undetectable]
 */

static void gdas_oscilloscope_init(GdasOscilloscope* oscilloscope)
{
  oscilloscope->window_size = 1024;
  oscilloscope->off_length = 1024;
  oscilloscope->changed = FALSE;
  oscilloscope->window[0] = g_new (GdamSample, oscilloscope->window_size);
  oscilloscope->window[1] = g_new (GdamSample, oscilloscope->window_size);
}

static void gdas_oscilloscope_destroy(GtkObject* object)
{
  GdasOscilloscope *oscilloscope;
  oscilloscope = GDAS_OSCILLOSCOPE (object);
  g_free (oscilloscope->window[0]);
  g_free (oscilloscope->window[1]);
  oscilloscope->window[0] = NULL;
  oscilloscope->window[1] = NULL;
  (*parent_class->destroy) (object);
}

static void gdas_oscilloscope_class_init(GdasOscilloscopeClass* oscilloscope_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (oscilloscope_class);
  GDAS_SOURCE_CLASS (oscilloscope_class)->read = gdas_oscilloscope_read;
  object_class->destroy = gdas_oscilloscope_destroy;
  gdam_arg_register (object_class,
		     "window_size",
		     GDAM_ARG_TYPE_INT,
		     gdas_oscilloscope_arg_set_window_size,
		     gdas_oscilloscope_arg_get_window_size,
		     NULL, "0,1000000", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "off_length",
		     GDAM_ARG_TYPE_INT,
		     gdas_oscilloscope_arg_set_off_length,
		     gdas_oscilloscope_arg_get_off_length,
		     NULL, "0,1000000", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "sample_data",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     NULL,
		     gdas_oscilloscope_arg_get_sample_data,
		     NULL, NULL, GDAM_ARG_TRIGGER);
  gdam_arg_register (object_class,
		     "if_modified_sample_data",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     NULL,
		     gdas_oscilloscope_arg_get_if_modified_sample_data,
		     NULL, NULL, GDAM_ARG_TRIGGER);
}

GtkType gdas_oscilloscope_get_type()
{
  static GtkType oscilloscope_type = 0;
  if (!oscilloscope_type)
    {
      static const GtkTypeInfo oscilloscope_info = {
	"GdasOscilloscope",
	sizeof (GdasOscilloscope),
	sizeof (GdasOscilloscopeClass),
	(GtkClassInitFunc) gdas_oscilloscope_class_init,
	(GtkObjectInitFunc) gdas_oscilloscope_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      oscilloscope_type = gtk_type_unique (parent, &oscilloscope_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return oscilloscope_type;
}

