/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdasphaser.h"
#include <math.h>

static inline
GdamSampleValue allpass_update(GdamSampleValue*     allpass,
                               GdamSampleValue      input,
			       GdamSampleValue      coeff)
{
  GdamSampleValue rv;
  rv = -GDAM_FIXED_MUL (input, coeff) + *allpass;
  *allpass = GDAM_FIXED_MUL (rv, coeff) + input;
  return rv;
}

static inline void process(GdasPhaser* phaser, GdamSample* sample)
{
  int s;
  GdamFixed a = phaser->a;
  GdamFixed feedback = phaser->feedback;
  GdamFixed depth = phaser->depth;

  s = sample->s[0];
  s += GDAM_FIXED_MUL (phaser->zm1.s[0], feedback);
  s = allpass_update (&phaser->allpass[0].s[0], s, a);
  s = allpass_update (&phaser->allpass[1].s[0], s, a);
  s = allpass_update (&phaser->allpass[2].s[0], s, a);
  s = allpass_update (&phaser->allpass[3].s[0], s, a);
  s = allpass_update (&phaser->allpass[4].s[0], s, a);
  s = allpass_update (&phaser->allpass[5].s[0], s, a);
  s = sample->s[0] + GDAM_FIXED_MUL (s, depth);
  sample->s[0] = GDAM_SAMPLE_CLAMP (s);


  s = sample->s[1];
  s += GDAM_FIXED_MUL (phaser->zm1.s[1], feedback);
  s = allpass_update (&phaser->allpass[0].s[1], s, a);
  s = allpass_update (&phaser->allpass[1].s[1], s, a);
  s = allpass_update (&phaser->allpass[2].s[1], s, a);
  s = allpass_update (&phaser->allpass[3].s[1], s, a);
  s = allpass_update (&phaser->allpass[4].s[1], s, a);
  s = allpass_update (&phaser->allpass[5].s[1], s, a);
  s = sample->s[1] + GDAM_FIXED_MUL (s, depth);
  sample->s[1] = GDAM_SAMPLE_CLAMP (s);
  phaser->zm1 = *sample;
}

static int gdas_phaser_read(GdasSource*      source,
                            GdamSample*      buffer,
			    int              max_read)
{
  GdasPhaser *phaser;
  int num_read;
  int i;
  gfloat phase, phase_delta;
  gfloat delay_min, delay_range;


  phaser = GDAS_PHASER (source);
  phase = phaser->lfo_phase;
  phase_delta = phaser->lfo_incr;
  delay_min = phaser->min_delay;
  delay_range = phaser->max_delay - phaser->min_delay;

  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buffer, max_read);
  for (i = 0; i < num_read; i++)
    {
      gfloat a;
      a = sin (phase);
      a = delay_min + delay_range * (a + 1.0) / 2.0;
      a = (1.0 - a) / (1.0 + a);
      phaser->a = GDAM_FIXED_FROM_DOUBLE (a);
      process (phaser, buffer + i);

      phase += phase_delta;
      if (phase >= 2.0 * M_PI)
	phase -= 2.0 * M_PI;
    }
  phaser->lfo_phase = phase;
  return num_read;
}

static void gdas_phaser_arg_set_min_delay(GtkObject* object, GdamArg* arg)
{
  GdasPhaser *phaser = GDAS_PHASER (object);
  gdouble delay = GDAM_ARG_DOUBLE (arg) / 44100.0 * 2.0;
  phaser->min_delay = GDAM_FIXED_FROM_DOUBLE (delay);
}
static void gdas_phaser_arg_get_min_delay(GtkObject* object, GdamArg* arg)
{
  gdouble hertz;
  GdasPhaser *phaser = GDAS_PHASER (object);
  hertz = GDAM_FIXED_TO_DOUBLE (phaser->min_delay);
  hertz /= 2.0;
  hertz *= 44100.0;
  GDAM_ARG_DOUBLE (arg) = hertz;
}
static void gdas_phaser_arg_set_max_delay(GtkObject* object, GdamArg* arg)
{
  GdasPhaser *phaser = GDAS_PHASER (object);
  gdouble delay = GDAM_ARG_DOUBLE (arg) / 44100.0 * 2.0;
  phaser->max_delay = GDAM_FIXED_FROM_DOUBLE (delay);
}
static void gdas_phaser_arg_get_max_delay(GtkObject* object, GdamArg* arg)
{
  gdouble hertz;
  GdasPhaser *phaser = GDAS_PHASER (object);
  hertz = GDAM_FIXED_TO_DOUBLE (phaser->max_delay);
  hertz /= 2.0;
  hertz *= 44100.0;
  GDAM_ARG_DOUBLE (arg) = hertz;
}
static void gdas_phaser_arg_set_feedback(GtkObject* object, GdamArg* arg)
{
  gdouble feedback;
  feedback = GDAM_ARG_DOUBLE (arg);
  feedback = CLAMP (feedback, 0.0, 1.0);
  GDAS_PHASER (object)->feedback = GDAM_FIXED_FROM_DOUBLE (feedback);
}
static void gdas_phaser_arg_get_feedback(GtkObject* object, GdamArg* arg)
{
  GdasPhaser *phaser = GDAS_PHASER (object);
  gdouble feedback;
  feedback = GDAM_FIXED_TO_DOUBLE (phaser->feedback);
  GDAM_ARG_DOUBLE (arg) = feedback;
}

static void gdas_phaser_arg_set_depth(GtkObject* object, GdamArg* arg)
{
  gdouble depth;
  depth = GDAM_ARG_DOUBLE (arg);
  depth = CLAMP (depth, 0.0, 1.0);
  GDAS_PHASER (object)->depth = GDAM_FIXED_FROM_DOUBLE (depth);
}
static void gdas_phaser_arg_get_depth(GtkObject* object, GdamArg* arg)
{
  GdasPhaser *phaser = GDAS_PHASER (object);
  gdouble depth;
  depth = GDAM_FIXED_TO_DOUBLE (phaser->depth);
  GDAM_ARG_DOUBLE (arg) = depth;
}
static void gdas_phaser_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  gdouble incr = M_PI * 2.0 * GDAM_ARG_DOUBLE (arg) / 44100.0;
  GDAS_PHASER (object)->lfo_incr = incr;
}

static void gdas_phaser_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  gdouble incr = GDAS_PHASER (object)->lfo_incr;
  incr /= M_PI * 2.0;
  incr *= 44100.0;
  GDAM_ARG_DOUBLE (arg) = incr;
}


static void gdas_phaser_init(GdasPhaser* phaser)
{
  GdamArg tmp;
  GtkObject *object = GTK_OBJECT (phaser);
  phaser->feedback = GDAM_FIXED_FROM_DOUBLE (0.4);
  phaser->depth = GDAM_FIXED_FROM_DOUBLE (0.5);
  tmp.name = "rate";
  tmp.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&tmp) = 0.5f;
  gdas_phaser_arg_set_rate (GTK_OBJECT (phaser), &tmp);

  tmp.name = "min_hertz";
  tmp.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&tmp) = 440.0;
  gdas_phaser_arg_set_min_delay (GTK_OBJECT (phaser), &tmp);

  tmp.name = "max_hertz";
  tmp.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&tmp) = 1600.0;
  gdas_phaser_arg_set_max_delay (GTK_OBJECT (phaser), &tmp);
}

static void gdas_phaser_class_init(GdasPhaserClass* phaser_class)
{
  GdasSourceClass *source_class;
  GtkObjectClass *object_class;
  source_class = GDAS_SOURCE_CLASS (phaser_class);
  object_class = GTK_OBJECT_CLASS (phaser_class);
  source_class->read = gdas_phaser_read;
  gdam_arg_register (object_class,
		     "min_hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_phaser_arg_set_min_delay,
		     gdas_phaser_arg_get_min_delay,
		     NULL, ".1,44100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "max_hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_phaser_arg_set_max_delay,
		     gdas_phaser_arg_get_max_delay,
		     NULL, ".1,44100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_phaser_arg_set_rate,
		     gdas_phaser_arg_get_rate,
		     NULL, ".01,100", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "depth",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_phaser_arg_set_depth,
		     gdas_phaser_arg_get_depth, NULL, "0,10", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "feedback",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_phaser_arg_set_feedback,
		     gdas_phaser_arg_get_feedback,
		     NULL, "0,1", GDAM_ARG_VITAL);
}

GtkType gdas_phaser_get_type()
{
  static GtkType phaser_type = 0;
  if (!phaser_type)
    {
      static const GtkTypeInfo phaser_info = {
	"GdasPhaser",
	sizeof (GdasPhaser),
	sizeof (GdasPhaserClass),
	(GtkClassInitFunc) gdas_phaser_class_init,
	(GtkObjectInitFunc) gdas_phaser_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      phaser_type = gtk_type_unique (parent, &phaser_info);
    }
  return phaser_type;
}
