/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdaspitchshift.h"
#include <math.h>

#define DEFAULT_DELAY	2048

static void gdas_pitch_shift_init(GdasPitchShift* pitch_shift)
{
  pitch_shift->delay_whole_incr = 0;
  pitch_shift->start_delay = DEFAULT_DELAY;
  pitch_shift->buffer_length = DEFAULT_DELAY;
  pitch_shift->buffer = g_new (GdamSample, DEFAULT_DELAY);
  pitch_shift->overlap_delta_delay = 64;

  pitch_shift->delay_0 = pitch_shift->start_delay;
}

static inline GdamSample get_sample(GdasPitchShift*     pitch_shift,
                                    int                 delay)
{
  int index = pitch_shift->insert_index;
  index -= delay;
  if (index < 0)
    index += pitch_shift->buffer_length;
  //g_message("get_sample: delay=%d, index=%d", delay, index);
  return pitch_shift->buffer[index];
}
static inline void insert_sample(GdasPitchShift* pitch_shift,
                                 GdamSample      sample)
{
  int index = pitch_shift->insert_index;
  pitch_shift->buffer[index] = sample;
  index++;
  if (index == pitch_shift->buffer_length)
    index = 0;
  pitch_shift->insert_index = index;
}
static inline GdamSample interp(GdamSample   a,
                                GdamSample   b,
				int          index,
				int          length)
{
  GdamSample rv;
  rv.s[0] = (gint16) (((gint64) a.s[0] * (length - index)
		       + (gint64) b.s[0] * (index)) / length);
  rv.s[1] = (gint16) (((gint64) a.s[1] * (length - index)
		       + (gint64) b.s[1] * (index)) / length);
  return rv;
}

static void process_buffer(GdasPitchShift*        pitch_shift,
			   GdamSample*            out,
                           GdamSample*            in,
			   int                    num_in)
{
  int delay_whole_incr = pitch_shift->delay_whole_incr;
  guint32 delay_frac_incr = pitch_shift->delay_frac_incr;
  int start_delay = pitch_shift->start_delay;
  int overlap_delta_delay = pitch_shift->overlap_delta_delay;
  int delay_0 = pitch_shift->delay_0;
  guint32 delay_0_frac = pitch_shift->delay_0_frac;
  int delay_1 = pitch_shift->delay_1;
  guint32 delay_1_frac = pitch_shift->delay_1_frac;
  while (num_in-- > 0)
    {
      int from_start = start_delay - delay_0;
      insert_sample (pitch_shift, *in);
      if (from_start < overlap_delta_delay)
	{
	  /* overlap mode */
	  GdamSample a = get_sample (pitch_shift, delay_0);
	  GdamSample b = get_sample (pitch_shift, delay_1);
	  guint32 tmp;
	  delay_0 -= delay_whole_incr;
	  tmp = delay_0_frac + delay_frac_incr;
	  if (tmp < delay_0_frac)
	    delay_0--;
	  delay_0_frac = tmp;
	  delay_1 -= delay_whole_incr;
	  tmp = delay_1_frac + delay_frac_incr;
	  if (tmp < delay_1_frac)
	    delay_1--;
	  delay_1_frac = tmp;

	  /* old fragment is ending */
	  if (delay_1 < 0)
	    {
	      delay_0 = delay_1;
	      delay_0_frac = delay_1_frac;
	    }
	  *out++ = interp (a, b, from_start, overlap_delta_delay);
	}
      else
	{
	  guint32 tmp;

	  /* just delay_0 */
	  *out++ = get_sample (pitch_shift, delay_0);
	  delay_0 -= delay_whole_incr;
	  tmp = delay_0_frac + delay_frac_incr;
	  if (tmp < delay_0_frac)
	    delay_0--;

	  if (delay_0 < overlap_delta_delay)
	    {
	      delay_1 = start_delay;
	      delay_1_frac = 0;
	    }
	}
    }
  pitch_shift->delay_0 = delay_0;
  pitch_shift->delay_0_frac = delay_0_frac;
  pitch_shift->delay_1 = delay_1;
  pitch_shift->delay_1_frac = delay_1_frac;
}

static inline gboolean try_fill_buffer(GdasPitchShift*  pitch_shift)
{
  int empty = pitch_shift->buffer_length - pitch_shift->cur_buffer_length;
  int got;
  GdamSample *buf;

  if (empty <= 0)
    return TRUE;

  buf = alloca (sizeof (GdamSample) * empty);
  got = gdas_filter_sub_read (GDAS_FILTER (pitch_shift), buf, empty);
  if (got + pitch_shift->insert_index >= pitch_shift->buffer_length)
    {
      int insert_index = pitch_shift->insert_index;
      int p1 = pitch_shift->buffer_length - insert_index;
      int p2 = got - p1;
      gdam_sample_copy (pitch_shift->buffer + insert_index, buf, p1);
      gdam_sample_copy (pitch_shift->buffer, buf + p1, p2);
      pitch_shift->insert_index = p2;
    }
  else
    {
      gdam_sample_copy (pitch_shift->buffer
			+ pitch_shift->insert_index, buf, got);
      pitch_shift->insert_index += got;
    }
  pitch_shift->cur_buffer_length += got;
  return (pitch_shift->cur_buffer_length == pitch_shift->buffer_length);
}

static int gdas_pitch_shift_read(GdasSource*       source,
                                 GdamSample*       buffer,
				 int               to_read)
{
  GdasPitchShift *pitch_shift;
  GdamSample *read_buffer;
  int num_read = 0;
  int desired_read;

  pitch_shift = GDAS_PITCH_SHIFT (source);

  /* try to fill the buffer if it isn't. */
  if (!try_fill_buffer (pitch_shift))
    {
      gdam_sample_zero (buffer, to_read);
      g_message ("try_fill_buffer: failed");
      return to_read;
    }


  /* estimate the amount of data needed */
  read_buffer = alloca (sizeof (GdamSample) * to_read);
  num_read = gdas_filter_sub_read (GDAS_FILTER (pitch_shift),
				   read_buffer, to_read);
  //g_message("process_buffer: num_read=%d", num_read);
  process_buffer (pitch_shift, buffer, read_buffer, num_read);
  return num_read;
}

static void gdas_pitch_shift_arg_set_rate(GtkObject* object, GdamArg* arg)
{
  GdasPitchShift *shift;
  gdouble rate = GDAM_ARG_DOUBLE (arg);
  shift = GDAS_PITCH_SHIFT (object);
  rate -= 1.0;
  shift->delay_whole_incr = (int) floor (rate);
  rate -= shift->delay_whole_incr;
  rate *= (1 << 16);
  rate *= (1 << 16);
  shift->delay_frac_incr = (guint32) rate;
}

static void gdas_pitch_shift_arg_get_rate(GtkObject* object, GdamArg* arg)
{
  GdasPitchShift *shift;
  gdouble rate;
  shift = GDAS_PITCH_SHIFT (object);
  rate = shift->delay_frac_incr;
  rate /= (1 << 16);
  rate /= (1 << 16);
  rate += shift->delay_whole_incr;
  GDAM_ARG_DOUBLE (arg) = rate + 1.0;
}


static void gdas_pitch_shift_class_init(GdasPitchShiftClass* pitch_shift_class)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (pitch_shift_class);
  gdam_arg_register (object_class,
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_pitch_shift_arg_set_rate,
		     gdas_pitch_shift_arg_get_rate,
		     NULL, NULL, GDAM_ARG_VITAL);
  GDAS_SOURCE_CLASS (pitch_shift_class)->read = gdas_pitch_shift_read;
}

GtkType gdas_pitch_shift_get_type()
{
  static GtkType pitch_shift_type = 0;
  if (!pitch_shift_type)
    {
      static const GtkTypeInfo pitch_shift_info = {
	"GdasPitchShift",
	sizeof (GdasPitchShift),
	sizeof (GdasPitchShiftClass),
	(GtkClassInitFunc) gdas_pitch_shift_class_init,
	(GtkObjectInitFunc) gdas_pitch_shift_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      pitch_shift_type = gtk_type_unique (parent, &pitch_shift_info);
    }
  return pitch_shift_type;
}
