/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdaspluckedstring.h"
#include <math.h>
#include <gdam/common/gdamplugin.h>
#include <stdlib.h>

static GtkObjectClass* parent_class;

static int gdas_plucked_string_read(GdasSource* source, GdamSample* buf, int n)
{
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  GdasPluckedString *plucked = GDAS_PLUCKED_STRING (source);
  GdamCircularBuffer *lowpass = &plucked->lowpass;
  GdamFixed a, gain;
  int num_read;
  GdamFixed feedback_div_4 = plucked->gain / 4;
  GdamSample last_allpass_out = plucked->last_allpass_out;
  GdamSample last_allpass_in = plucked->last_allpass_in;
  if (filtered == NULL)
    return 0;
  num_read = n = gdas_source_read (filtered, buf, n);
  if (n <= 0)
    return n;

  g_return_val_if_fail (lowpass->amount_buffered >= 2, -1);

  a = plucked->a;
  gain = plucked->gain;

  while (n-- > 0)
    {
      int next_out[2];
      int v;
      GdamSample lowpass_output, lowpass_input, tmp;

      lowpass_output = gdam_circular_buffer_read_one (lowpass);
      tmp = gdam_circular_buffer_peek_one (lowpass);
      if (plucked->filter_coeff != GdamFixed_ONE)
	{
	  tmp.s[0] = (int) (plucked->filter_coeff * tmp.s[0] / GdamFixed_ONE);
	  tmp.s[1] = (int) (plucked->filter_coeff * tmp.s[1] / GdamFixed_ONE);
	}
      lowpass_output.s[0] = ((int) tmp.s[0] + lowpass_output.s[0]) / 2;
      lowpass_output.s[1] = ((int) tmp.s[1] + lowpass_output.s[1]) / 2;

      next_out[0] = (int) (last_allpass_in.s[0])
	+ ((-(gint64) (last_allpass_out.s[0] * a)
	    + (gint64) (lowpass_output.s[0] * a)) / GdamFixed_ONE);
      next_out[1] = (int) (last_allpass_in.s[1])
	+ ((-(gint64) (last_allpass_out.s[1] * a)
	    + (gint64) (lowpass_output.s[1] * a)) / GdamFixed_ONE);

      last_allpass_in = lowpass_output;
      v = (gint64) next_out[0] * feedback_div_4 / (GdamFixed_ONE / 4)
	+ buf->s[0];
      lowpass_input.s[0] = GDAM_SAMPLE_CLAMP (v);
      v = (gint64) next_out[1] * feedback_div_4 / (GdamFixed_ONE / 4)
	+ buf->s[1];
      lowpass_input.s[1] = GDAM_SAMPLE_CLAMP (v);
      gdam_circular_buffer_append_one (lowpass, lowpass_input);
      buf->s[0] = GDAM_SAMPLE_CLAMP (next_out[0]);
      buf->s[1] = GDAM_SAMPLE_CLAMP (next_out[1]);
      last_allpass_out = *buf;
      buf++;
    }
  plucked->last_allpass_out = last_allpass_out;
  plucked->last_allpass_in = last_allpass_in;
  return num_read;
}
static void gdas_plucked_string_destroy(GtkObject* object)
{
  GdasPluckedString *plucked = GDAS_PLUCKED_STRING (object);
  gdam_circular_buffer_destruct (&plucked->lowpass);
  (*parent_class->destroy) (object);
}

static void gdas_plucked_string_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
  gdas_plucked_string_set_hertz (GDAS_PLUCKED_STRING (object),
				 GDAM_ARG_DOUBLE (arg));
}

static void gdas_plucked_string_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = GDAS_PLUCKED_STRING (object)->hertz;
}

static void gdas_plucked_string_arg_set_gain(GtkObject* object, GdamArg* arg)
{
  gdas_plucked_string_set_gain (GDAS_PLUCKED_STRING (object),
				GDAM_ARG_DOUBLE (arg));
}

static void gdas_plucked_string_arg_get_gain(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_DOUBLE (arg) = (gdouble) GDAS_PLUCKED_STRING (object)->gain
    / GdamFixed_ONE;
}

static void gdas_plucked_string_arg_set_filter_coeff
					(GtkObject* object, GdamArg* arg)
{
  GdamFixed coeff = (int) (GDAM_ARG_DOUBLE (arg) * GdamFixed_ONE);
  GDAS_PLUCKED_STRING (object)->filter_coeff = coeff;
}

static void gdas_plucked_string_arg_get_filter_coeff
                                        (GtkObject* object, GdamArg* arg)
{
  gdouble coeff = GDAS_PLUCKED_STRING (object)->filter_coeff;
  coeff /= GdamFixed_ONE;
  GDAM_ARG_DOUBLE (arg) = coeff;
}

static void gdas_plucked_string_arg_do_pluck(GtkObject* object, GdamArg* arg)
{
  gdas_plucked_string_pluck (GDAS_PLUCKED_STRING (object),
			     GDAM_ARG_DOUBLE (arg));
}

static void gdas_plucked_string_init(GdasPluckedString* plucked_string)
{
  plucked_string->filter_coeff = GdamFixed_ONE;
}
static void gdas_plucked_string_class_init(GdasPluckedStringClass* klass)
{
  GDAS_SOURCE_CLASS (klass)->read = gdas_plucked_string_read;
  GTK_OBJECT_CLASS (klass)->destroy = gdas_plucked_string_destroy;
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "hertz",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_plucked_string_arg_set_hertz,
		     gdas_plucked_string_arg_get_hertz,
		     NULL, "1,40000", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_plucked_string_arg_set_gain,
		     gdas_plucked_string_arg_get_gain,
		     NULL, "0,2", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "filter_coeff",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_plucked_string_arg_set_filter_coeff,
		     gdas_plucked_string_arg_get_filter_coeff,
		     NULL, "-1,2", GDAM_ARG_VITAL);
  gdam_arg_register (GTK_OBJECT_CLASS (klass),
		     "pluck",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_plucked_string_arg_do_pluck, NULL, NULL, "0,30", 0);
}

GtkType gdas_plucked_string_get_type()
{
  static GtkType plucked_string_type = 0;
  if (!plucked_string_type)
    {
      static const GtkTypeInfo plucked_string_info = {
	"GdasPluckedString",
	sizeof (GdasPluckedString),
	sizeof (GdasPluckedStringClass),
	(GtkClassInitFunc) gdas_plucked_string_class_init,
	(GtkObjectInitFunc) gdas_plucked_string_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      plucked_string_type = gtk_type_unique (parent, &plucked_string_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return plucked_string_type;
}

void gdas_plucked_string_set_hertz(GdasPluckedString*  string,
                                   gdouble             hertz)
{
  gdouble period = 44100.0 / hertz;
  int lowpass_len = floor (period + 0.5);
  gdouble allpass_delay = period - ((gdouble) lowpass_len - 0.5);
  gdouble half_frequency = 0.5 / period;
  gdouble alpha;
  g_assert (0.0 <= allpass_delay && allpass_delay < 1.0);
  alpha = sin ((1.0 - allpass_delay) * half_frequency)
    / sin ((1.0 + allpass_delay) * half_frequency);
  string->a = (GdamFixed) (alpha * GdamFixed_ONE);
  string->hertz = 440.0;
  gdam_circular_buffer_resize (&string->lowpass, lowpass_len, TRUE);
}

void gdas_plucked_string_set_gain (GdasPluckedString*  string,
                                   gdouble             gain)
{
  string->gain = (GdamFixed) (gain * GdamFixed_ONE);
}

static void add_noise(GdamSample* buf, int len, gdouble velo)
{
  int range = GDAM_SAMPLE_ONE * velo;
  if (range == 0)
    return;
  while (len-- > 0)
    {
      int add = rand () % range - range / 2;
      buf->s[0] += add;
      buf->s[1] += add;		/* TODO: clamp */
      buf++;
    }
}

void gdas_plucked_string_pluck           (GdasPluckedString*  filter,
                                          gdouble             velocity)
{
  GdamCircularBuffer *buffer = &filter->lowpass;
  add_noise (buffer->buffer, buffer->buffer_size, velocity);
}
