/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasqueue.h"
#include "../gdamconnection.h"
#include <gdam/common/gdamplugin.h>

static GtkObjectClass* parent_class = NULL;

static void     gdas_queue_class_init       (GdasSourceClass*  klass);
static void     gdas_queue_init             (GdasQueue*        queue);
static void     gdas_queue_destroy          (GtkObject*        queue);
static GdamTime gdas_queue_get_position     (GdasSource*       input);
static int      gdas_queue_read             (GdasSource*       input,
                                             GdamSample*       samples,
		                             int               max_read);
static gboolean gdas_queue_seek             (GdasSource*      source,
                                             GdamTime         offset,
		                             int              whence);
static void     gdas_queue_get_subsources   (GdasSource*      source,
                                             int*             num_subs,
					     GdasSource***    subs);

static void     gdas_queue_miss_data        (GdasSource*       input,
                                             int               samples_lost);
static gboolean gdas_queue_is_ready_to_read (GdasSource* queue_source);
static gboolean gdas_queue_detach_child     (GdasSource*      mix,
                                             GdasSource*      sub_source);

static void gdas_queue_init(GdasQueue* queue)
{
  queue->terminates = 1;
}

static void gdas_queue_arg_enqueue(GtkObject* object, GdamArg* arg)
{
  int source_id = GDAM_ARG_INT (arg);
  GdasSource *source;
  GdasQueue *queue = GDAS_QUEUE (object);
  g_return_if_fail (source_id != 0);
  g_return_if_fail (queue != NULL);
  gdam_debug ("en_queue: new num_subs=%d", GDAM_ARG_INT_ARRAY (arg).length);
  source = gdam_connection_get_source (NULL, source_id);
  g_return_if_fail (source != NULL);
  gdas_queue_enqueue (queue, source);
  gtk_object_unref (GTK_OBJECT (source));
}

static void gdas_queue_arg_get_queue(GtkObject* object, GdamArg* arg)
{
  int num_subsources;
  int *array;
  int i;
  GSList *at;
  GdasQueue *queue = GDAS_QUEUE (object);
  num_subsources = g_slist_length (queue->first_source);
  array = g_new (int, num_subsources);
  at = queue->first_source;
  for (i = 0; i < num_subsources; i++)
    {
      array[i] = GDAM_SERVER_OBJECT (at->data)->global_id;
      at = at->next;
    }
  GDAM_ARG_INT_ARRAY (arg).length = num_subsources;
  GDAM_ARG_INT_ARRAY (arg).array = array;
}
static void reparent(gpointer source, gpointer queue)
{
  GdasSource *child = GDAS_SOURCE (source);
  GdasSource *parent = GDAS_SOURCE (queue);
  if (child->parent != NULL)
    gdas_source_detach_child (child->parent, child);
  child->parent = parent;
}

static void gdas_queue_arg_set_queue(GtkObject* object, GdamArg* arg)
{
  GSList *new_queue = NULL;
  GdasQueue *queue = GDAS_QUEUE (object);
  int i;
  int length = GDAM_ARG_INT_ARRAY (arg).length;
  int *array = GDAM_ARG_INT_ARRAY (arg).array;
  for (i = 0; i < length; i++)
    {
      int id = array[i];
      GdasSource *new_source = gdam_connection_get_source (NULL, id);
      GtkObject *new_object;
      if (new_source == NULL)
	{
	  gdam_debug ("queue: no source of id %d found", id);
	  continue;
	}
      new_object = GTK_OBJECT (new_source);
      /* The reference is now done by gdam_connection_get_source */
      /*gtk_object_ref(new_object); */
      gtk_object_sink (new_object);
      new_queue = g_slist_prepend (new_queue, new_object);
    }
  while (queue->first_source)
    gdas_queue_dequeue (queue);
  queue->first_source = g_slist_reverse (new_queue);
  queue->last_source = g_slist_last (queue->first_source);
  g_slist_foreach (queue->first_source, reparent, queue);
}

static void gdas_queue_arg_set_terminates(GtkObject* object, GdamArg* arg)
{
  GDAS_QUEUE (object)->terminates = GDAM_ARG_INT (arg) ? 1 : 0;
}
static void gdas_queue_arg_get_terminates(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_QUEUE (object)->terminates;
}


static void gdas_queue_class_init(GdasSourceClass* source_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (source_class);
  object_class->destroy = gdas_queue_destroy;
  source_class->read = gdas_queue_read;
  source_class->miss_data = gdas_queue_miss_data;
  source_class->is_ready_to_read = gdas_queue_is_ready_to_read;
  source_class->detach_child = gdas_queue_detach_child;
  source_class->get_position = gdas_queue_get_position;
  source_class->get_subsources = gdas_queue_get_subsources;
  source_class->seek = gdas_queue_seek;
  gdam_arg_register (object_class,
		     "enqueue",
		     GDAM_ARG_TYPE_INT,
		     gdas_queue_arg_enqueue, NULL, NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "queue",
		     GDAM_ARG_TYPE_INT_ARRAY,
		     gdas_queue_arg_set_queue,
		     gdas_queue_arg_get_queue, NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "terminates",
		     GDAM_ARG_TYPE_INT,
		     gdas_queue_arg_set_terminates,
		     gdas_queue_arg_get_terminates,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

GtkType gdas_queue_get_type()
{
  static GtkType queue_type = 0;
  if (!queue_type)
    {
      static const GtkTypeInfo queue_info = {
	"GdasQueue",
	sizeof (GdasQueue),
	sizeof (GdasQueueClass),
	(GtkClassInitFunc) gdas_queue_class_init,
	(GtkObjectInitFunc) gdas_queue_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_SOURCE;
      parent_class = gtk_type_class (parent);
      queue_type = gtk_type_unique (parent, &queue_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return queue_type;
}

static void     gdas_queue_destroy          (GtkObject*        object)
{
  GdasQueue *queue = GDAS_QUEUE (object);
  while (queue->first_source)
    gdas_queue_dequeue (queue);
  (*parent_class->destroy) (object);
}


static int      gdas_queue_read             (GdasSource*       input,
                                             GdamSample*       samples,
		                             int               max_read)
{
  GdasQueue *queue = GDAS_QUEUE (input);
  int rv = 0;
  while (queue->first_source)
    {
      GdasSource *subsource;
      int num_read;
      subsource = queue->first_source->data;
      if (!gdas_source_is_ready_to_read (subsource))
	{
	  /* XXX: fix this typical hack... */
	  if (rv == 0)
	    {
	      gdam_sample_zero (samples, 1);
	      rv++;
	    }
	  return rv;
	}
      num_read = gdas_source_read (subsource, samples + rv, max_read - rv);
      if (num_read == 0)
	{
	  gdas_queue_dequeue (queue);
	  continue;
	}
      if (num_read == max_read - rv)
	return max_read;

      rv += num_read;
    }
  if (queue->first_source == NULL)
    {
      if (queue->terminates)
	return rv;
      gdam_sample_zero (samples + rv, max_read - rv);
      return max_read;
    }
  return rv;
}

static void     gdas_queue_miss_data        (GdasSource*       input,
                                             int               samples_lost)
{
  GdasQueue *queue = GDAS_QUEUE (input);
  GdasSource *subsource;
  if (queue == NULL || queue->first_source == NULL)
    return;
  subsource = GDAS_SOURCE (queue->first_source->data);
  gdas_source_miss_data (subsource, samples_lost);
}

static gboolean gdas_queue_is_ready_to_read (GdasSource* queue_source)
{
  GdasQueue *queue = GDAS_QUEUE (queue_source);
  GdasSource *first;
  g_return_val_if_fail (queue != NULL, FALSE);
  if (queue->first_source == NULL)
    return TRUE;
  first = GDAS_SOURCE (queue->first_source->data);
  return gdas_source_is_ready_to_read (first);
}

static gboolean gdas_queue_detach_child     (GdasSource*      source,
                                             GdasSource*      sub_source)
{
  GdasQueue *queue = GDAS_QUEUE (source);
  GSList *prev = NULL;
  GSList *at;
  for (at = queue->first_source; at != NULL; at = at->next)
    {
      if (at->data == sub_source)
	break;
      prev = at;
      //at = at->next;
    }
  if (at == NULL)
    return FALSE;
  if (prev)
    prev->next = at->next;
  else
    queue->first_source = at->next;
  if (at == queue->last_source)
    queue->last_source = prev;
  g_slist_free_1 (at);
  sub_source->parent = NULL;
  gtk_object_unref (GTK_OBJECT (sub_source));
  return TRUE;
}

GdasQueue* gdas_queue_new()
{
  GdasQueue *queue = GDAS_QUEUE (gtk_type_new (GDAS_TYPE_QUEUE));
  return queue;
}

void gdas_queue_enqueue(GdasQueue*    queue,
                        GdasSource*   source)
{
  GtkObject *sub_object = GTK_OBJECT (source);
  if (source->parent != NULL)
    {
      gdas_source_detach_child (source->parent, source);
    }
  source->parent = GDAS_SOURCE (queue);
  if (queue->first_source == NULL)
    queue->first_source = queue->last_source = g_slist_prepend (NULL, source);
  else
    {
      g_slist_append (queue->last_source, source);
      queue->last_source = queue->last_source->next;
    }
  g_return_if_fail (queue->first_source != NULL);
  g_return_if_fail (queue->last_source != NULL);
  gtk_object_ref (sub_object);
  gtk_object_sink (sub_object);
}

void gdas_queue_dequeue(GdasQueue*    queue)
{
  GSList *first = queue->first_source;
  GSList *next;
  GdasSource *subsource;
  g_return_if_fail (first != NULL);
  next = first->next;

  subsource = GDAS_SOURCE (queue->first_source->data);
  g_slist_free_1 (queue->first_source);
  queue->first_source = next;
  if (queue->first_source == NULL)
    queue->last_source = NULL;
  subsource->parent = NULL;
  gtk_object_unref (GTK_OBJECT (subsource));
}
static gboolean gdas_queue_seek             (GdasSource*      source,
                                             GdamTime         offset,
		                             int              whence)
{
  GdasQueue *queue = GDAS_QUEUE (source);
  GSList *first = queue->first_source;
  GdasSource *subsource;
  if (first == NULL)
    return FALSE;
  subsource = GDAS_SOURCE (first->data);
  return gdas_source_seek (subsource, offset, whence);
}

static GdamTime gdas_queue_get_position     (GdasSource*      source)
{
  GdasQueue *queue = GDAS_QUEUE (source);
  GSList *first = queue->first_source;
  GdasSource *subsource;
  if (first == NULL)
    return -1;
  subsource = GDAS_SOURCE (first->data);
  return gdas_source_get_position (subsource);
}
static void     gdas_queue_get_subsources   (GdasSource*      source,
                                             int*             num_subs,
					     GdasSource***    subs)
{
  GdasQueue *queue = GDAS_QUEUE (source);
  GSList *at = queue->first_source;
  int i;
  *num_subs = g_slist_length (at);
  *subs = g_new (GdasSource *, *num_subs);
  for (i = 0; at != NULL; at = at->next, i++)
    (*subs)[i] = at->data;
}
