/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasrecorder.h"
#include <gdam/server/gdamcachebuffer.h>
#include <gdam/common/gdamplugin.h>

static GtkObjectClass* parent_class = NULL;

/* TODO:  this really needs a 16-bit mode */

static int  gdas_recorder_read(GdasSource* source,
                               GdamSample* buffer,
			       int         max_read)
{
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  GdasRecorder *recorder = GDAS_RECORDER (source);
  int num_read;
  if (filtered == NULL)
    return 0;
  num_read = gdas_source_read (filtered, buffer, max_read);
  if (recorder->is_recording && num_read > 0)
    gdam_buffer_append (&recorder->buffer,
			buffer, sizeof (GdamSample) * max_read);
  return num_read;
}
static void  gdas_recorder_destroy(GtkObject* recorder)
{
  GdamBuffer *buffer = &(GDAS_RECORDER (recorder)->buffer);
  gdam_buffer_destruct (buffer);
  (*parent_class->destroy) (recorder);
}

static void gdas_recorder_init(GdasRecorder* recorder)
{
  gdam_buffer_construct (&recorder->buffer);
}
static void gdas_recorder_arg_set_recording(GtkObject* object, GdamArg* arg)
{
  GDAS_RECORDER (object)->is_recording = GDAM_ARG_INT (arg);
}

static void gdas_recorder_arg_get_recording(GtkObject* object, GdamArg* arg)
{
  GDAM_ARG_INT (arg) = GDAS_RECORDER (object)->is_recording;
}

static void gdas_recorder_arg_copy_buffer(GtkObject* object, GdamArg* arg)
{
  GdasRecorder *recorder = GDAS_RECORDER (object);
  GdamCacheBuffer *buffer = gdas_recorder_make_buffer (recorder, FALSE);
  GdamConnection *connection = gdam_connection_get_top ();
  if (connection != NULL)
    gdam_connection_ref (connection, GTK_OBJECT (buffer));
  GDAM_ARG_INT (arg) = gdam_cache_buffer_get_id (buffer);
}

static void gdas_recorder_arg_grab_buffer(GtkObject* object, GdamArg* arg)
{
  GdasRecorder *recorder = GDAS_RECORDER (object);
  GdamCacheBuffer *buffer = gdas_recorder_make_buffer (recorder, TRUE);
  GdamConnection *connection = gdam_connection_get_top ();
  if (connection != NULL)
    gdam_connection_ref (connection, GTK_OBJECT (buffer));
  GDAM_ARG_INT (arg) = gdam_cache_buffer_get_id (buffer);
}

static void gdas_recorder_arg_get_length(GtkObject* object, GdamArg* arg)
{
  int bytes_in_buffer = GDAS_RECORDER (object)->buffer.size;
  GDAM_ARG_INT (arg) = bytes_in_buffer / sizeof (GdamSample);
}

static void gdas_recorder_class_init(GdasRecorderClass* recorder_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (recorder_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (recorder_class);
  source_class->read = gdas_recorder_read;
  object_class->destroy = gdas_recorder_destroy;

  gdam_arg_register (object_class,
		     "recording",
		     GDAM_ARG_TYPE_INT,
		     gdas_recorder_arg_set_recording,
		     gdas_recorder_arg_get_recording,
		     NULL, "boolean", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "copy_buffer",
		     GDAM_ARG_TYPE_INT,
		     NULL,
		     gdas_recorder_arg_copy_buffer,
		     NULL, NULL, GDAM_ARG_TRIGGER);
  gdam_arg_register (object_class,
		     "grab_buffer",
		     GDAM_ARG_TYPE_INT,
		     NULL,
		     gdas_recorder_arg_grab_buffer,
		     NULL, NULL, GDAM_ARG_TRIGGER);
  gdam_arg_register (object_class,
		     "length",
		     GDAM_ARG_TYPE_INT,
		     NULL, gdas_recorder_arg_get_length, NULL, NULL, 0);
}

GtkType gdas_recorder_get_type()
{
  static GtkType recorder_type = 0;
  if (!recorder_type)
    {
      static const GtkTypeInfo recorder_info = {
	"GdasRecorder",
	sizeof (GdasRecorder),
	sizeof (GdasRecorderClass),
	(GtkClassInitFunc) gdas_recorder_class_init,
	(GtkObjectInitFunc) gdas_recorder_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      recorder_type = gtk_type_unique (parent, &recorder_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return recorder_type;
}

void gdas_recorder_stop(GdasRecorder* recorder)
{
  recorder->is_recording = FALSE;
}

void gdas_recorder_start(GdasRecorder* recorder)
{
  recorder->is_recording = TRUE;
}

void gdas_recorder_empty(GdasRecorder* recorder)
{
  gdam_buffer_destruct (&recorder->buffer);
}

GdamCacheBuffer*
gdas_recorder_make_buffer(GdasRecorder* recorder, gboolean drain)
{
  GdamSample *buf;
  int num_samples = recorder->buffer.size / sizeof (GdamSample);
  GdamCacheBuffer *rv;
  gdam_debug ("gdas_recorder_make_buffer: %d samples", num_samples);
  buf = g_new (GdamSample, num_samples);
  if (drain)
    gdam_buffer_read (&recorder->buffer,
		      buf, sizeof (GdamSample) * num_samples);
  else
    gdam_buffer_peek (&recorder->buffer,
		      buf, sizeof (GdamSample) * num_samples);
  rv = gdam_cache_buffer_memory (num_samples, buf, g_free, FALSE);
  g_return_val_if_fail (rv != NULL, NULL);
  return rv;
}

