#include <gdam/common/gdamautoconf.h>
#include "gdasrecorderexternal.h"
#include <gdam/common/gdamplugin.h>
#include "../gdamserverdebug.h"
#include <gdam/common/gdamsampleconverter.h>
#include <gdam/common/gdamconnect.h>
#include <gdam/common/gdamwaitpid.h>
#include <gdam/server/gdamrealtime.h>
#include "../gdamarctan.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

typedef struct _GdasRecorderOutput GdasRecorderOutput;

static GtkObjectClass* parent_class = NULL;

static void gdas_recorder_external_destroy(GtkObject* object);
static int  gdas_recorder_external_read(GdasSource* source,
                                        GdamSample* sample,
					int         max_read);

/* XXX: this should be dynamically determined instead of 
 *      #define'd and assert()d below */
#define MAX_BYTES_PER_SAMPLE		16

/*============================================================
 *                    GdasRecorderOutput
 *
 * One sink for the recorder.
 */
struct _GdasRecorderOutput {
	gboolean		is_paused;

	guint			id;

	/* The maximum length to extend buffer to, in samples. */
	int			max_buffer;

	GdamBuffer		buffer;

	gpointer                user_data;

	GPollFD			poll_fd;

	/* If reading from a process. */
	int			pid;

	/* A glib thread (ie a number you can use with g_source_remove),
	 * but actually for private use by start/stop_output_thread. */
	guint			output_thread;

	gboolean		has_thread;

	gboolean		had_error;
};

/* Create a pure virtual GdasRecorderOutput.
 * You must fill in the three function pointers before
 * you do anything with it. */
static GdasRecorderOutput* gdas_recorder_output_new()
{
  GdasRecorderOutput *output = g_new0 (GdasRecorderOutput, 1);
  gdam_buffer_construct (&output->buffer);
  return output;
}

static gboolean    gdas_recorder_output_prepare   (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data)
{
  GdasRecorderOutput *output = user_data;
  g_return_val_if_fail (output != NULL, FALSE);
  return FALSE;
}

static gboolean    gdas_recorder_output_check     (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  GdasRecorderOutput *output = user_data;
  if ((output->poll_fd.revents & G_IO_ERR) == G_IO_ERR)
    {
      output->had_error = TRUE;
      if (GDAM_SERVER_DEBUG_POLL)
	gdam_debug ("gdas_recorder_output_check: error writing");
      return TRUE;
    }
  if ((output->poll_fd.revents & G_IO_OUT) == G_IO_OUT)
    {
      if (GDAM_SERVER_DEBUG_POLL)
	gdam_debug ("gdas_recorder_output_check: writing");
      return TRUE;
    }
  if (GDAM_SERVER_DEBUG_POLL)
    gdam_debug ("gdas_recorder_output_check: not writing");
  return FALSE;
}

static gboolean    gdas_recorder_output_dispatch  (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
  GdasRecorderOutput *output = user_data;
  if (output->buffer.size == 0 || output->had_error)
    {
      g_main_remove_poll (&output->poll_fd);
      output->output_thread = 0;
      output->has_thread = FALSE;
      return FALSE;
    }
  if (!gdam_buffer_write_one (&output->buffer, output->poll_fd.fd))
    {
      output->had_error = TRUE;
      g_main_remove_poll (&output->poll_fd);
      output->output_thread = 0;
      output->has_thread = FALSE;
      return FALSE;
    }
  return TRUE;
}


static GSourceFuncs output_source_funcs = {
	gdas_recorder_output_prepare,
	gdas_recorder_output_check,
	gdas_recorder_output_dispatch,
	NULL
};

static void gdas_recorder_output_start(GdasRecorderOutput* output)
{
  if (output->buffer.size <= 0)
    return;
  if (output->had_error)
    return;

  output->poll_fd.events = G_IO_OUT | G_IO_ERR;
  g_assert (!output->has_thread);
  g_return_if_fail (output->output_thread == 0);
  g_main_add_poll (&output->poll_fd, G_PRIORITY_DEFAULT);
  output->output_thread = g_source_add (G_PRIORITY_DEFAULT,
					FALSE,
					&output_source_funcs,
					NULL, output, NULL);
  output->has_thread = TRUE;
}
static void gdas_recorder_output_stop(GdasRecorderOutput* output)
{
  g_assert (output->has_thread);
  g_source_remove (output->output_thread);
  g_main_remove_poll (&output->poll_fd);
  output->output_thread = 0;
  output->has_thread = FALSE;
}

static void gdas_recorder_output_destroy(GdasRecorderOutput* output)
{
  if (output->has_thread)
    {
      gdas_recorder_output_stop (output);
      output->has_thread = FALSE;
    }
  gdam_buffer_destruct (&output->buffer);
  close (output->poll_fd.fd);
  output->poll_fd.fd = -1;
  g_free (output);
}

static GdasRecorderOutput* gdas_recorder_output_new_fd(int fd)
{
  GdasRecorderOutput *output = gdas_recorder_output_new ();
  output->poll_fd.fd = fd;
  return output;
}

/*============================================================
 *                    GdasRecorderExternal
 */
static void gdas_recorder_external_init(GdasRecorderExternal* recorder_external)
{
  /* XXX: this should be configurable.... somewhere... */
  recorder_external->sample_size = sizeof (gint16);
  recorder_external->num_channels = 2;
  recorder_external->channel_converter
    = gdam_sample_converter_get_channel_changer (2, 2);
  /* convert from N-bit signed to 16-bits signed, native endian */
  recorder_external->sample_converter
    = gdam_sample_converter_get (GDAM_SAMPLE_BITS,
				 TRUE,
				 (G_BYTE_ORDER == G_BIG_ENDIAN),
				 16, TRUE, (G_BYTE_ORDER == G_BIG_ENDIAN));
}

static void gdas_recorder_external_class_init(GdasRecorderExternalClass* klass);
GtkType gdas_recorder_external_get_type()
{
  static GtkType recorder_external_type = 0;
  if (!recorder_external_type)
    {
      static const GtkTypeInfo recorder_external_info = {
	"GdasRecorderExternal",
	sizeof (GdasRecorderExternal),
	sizeof (GdasRecorderExternalClass),
	(GtkClassInitFunc) gdas_recorder_external_class_init,
	(GtkObjectInitFunc) gdas_recorder_external_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      recorder_external_type = gtk_type_unique (parent,
						&recorder_external_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return recorder_external_type;
}

static void do_limit(int           limit_mode,
                     GdamSampleValue* buf,
		     int           len)
{
#if (GDAM_SAMPLE_BITS == 24)
  if (limit_mode == 1)
    {
      while (len > 0)
	{
	  *buf = GDAM_SAMPLE_CLAMP (*buf);
	  buf++;
	  len--;
	}
    }
  else if (limit_mode == 2)
    {
      while (len > 0)
	{
	  *buf = gdam_arctan_b24 (*buf);
	  len--;
	  buf++;
	}
    }
#endif
}

static int  gdas_recorder_external_read(GdasSource* source,
                                        GdamSample* sample,
					int         max_read)
{
  GdasSource *filtered = GDAS_FILTER (source)->filtered;
  GdasRecorderExternal *recorder = GDAS_RECORDER_EXTERNAL (source);
  GSList *at;
  int num_read;
  char *buffer;
  int sample_size;
  int atom_size;

  /* the temporary storage size
   * for the channel + sample convertors */
  int alloc_atom_size;

  if (filtered == NULL)
    return 0;
  num_read = gdas_source_read (filtered, sample, max_read);
  if (num_read <= 0)
    {
      source->ended = 1;
      return num_read;
    }

  if (recorder->paused)
    return num_read;

  if (recorder->recorders == NULL)
    return num_read;

  sample_size = recorder->sample_size;
  atom_size = sample_size * recorder->num_channels;
  alloc_atom_size = recorder->num_channels * sizeof (GdamSampleValue);
  if (alloc_atom_size < atom_size)
    alloc_atom_size = atom_size;
  if ((int)recorder->read_buffer_size < max_read * alloc_atom_size)
    {
      if (recorder->read_buffer)
	free (recorder->read_buffer);
      recorder->read_buffer = malloc (max_read * alloc_atom_size);
      recorder->read_buffer_size = max_read * alloc_atom_size;
    }
  buffer = recorder->read_buffer;
  if (recorder->channel_converter != NULL)
    (*recorder->channel_converter) (buffer, sample, num_read);
  else
    gdam_sample_copy (buffer, sample, num_read);

  /* do limiting */
  do_limit (2, (GdamSampleValue *) buffer, recorder->num_channels * num_read);

  if (recorder->sample_converter != NULL)
    (*recorder->sample_converter) (buffer,
				   buffer, num_read * recorder->num_channels);
  for (at = recorder->recorders; at != NULL; at = at->next)
    {
      GdasRecorderOutput *output = at->data;
      int max_append;
      int discard;

      if (output->had_error)
	continue;

      /* we always want the most recent data,
       * so compute how many bytes of that there are. */
      max_append = MIN (output->max_buffer, num_read);

      /* discard data if we have too much */
      discard = output->buffer.size / atom_size
	      + max_append
	      - output->max_buffer;
      gdam_buffer_discard (&output->buffer, discard * atom_size);
      gdam_buffer_append (&output->buffer,
			  buffer
			  + atom_size * (num_read - max_append),
			  max_append * atom_size);
      if (!output->has_thread)
	{
	  gdas_recorder_output_start (output);
	}
    }
  return num_read;
}

static void gdas_recorder_external_destroy(GtkObject* object)
{
  GdasRecorderExternal *recorder = GDAS_RECORDER_EXTERNAL (object);
  GSList *outputs = recorder->recorders;
  recorder->recorders = NULL;
  g_slist_foreach (outputs, (GFunc) gdas_recorder_output_destroy, NULL);
  g_slist_free (outputs);
  if (recorder->read_buffer)
    free (recorder->read_buffer);
  recorder->read_buffer = NULL;
  recorder->read_buffer_size = 0;
  (*parent_class->destroy) (object);
}


guint gdas_recorder_external_add_fd     (GdasRecorderExternal* recorder,
                                         int                   fd,
					 int                   buffer_size,
					 int                   pid)
{
  GdasRecorderOutput *output;
  output = gdas_recorder_output_new_fd (fd);
  if (buffer_size > 0)
    output->max_buffer = buffer_size;
  recorder->recorders = g_slist_append (recorder->recorders, output);
  output->id = ++(recorder->last_output_id);
  output->pid = pid;
  return output->id;
}

/* --- our own private netcat implementation --- */

#define MAX_NETCAT_BUFFER		16384

static void
do_netcat_forever (const char *host,
		   int         port)
{
  guint8 ip_address[4];
  fd_set read_fds, write_fds, except_fds;
  GdamBuffer buffer;
  int remote_fd;
  switch (gdam_connect_tcp (host, port, &remote_fd, TRUE))
    {
    case GDAM_CONNECT_CONNECTED:
      break;
    case GDAM_CONNECT_CONNECTING:
      if (gdam_connect_finish (remote_fd, TRUE))
	break;
    case GDAM_CONNECT_FAILED:
      g_error ("failed to connect to remote host");
      break;
    }
  FD_ZERO (&read_fds);
  FD_ZERO (&write_fds);
  FD_ZERO (&except_fds);
  gdam_buffer_construct (&buffer);
  for (;;)
    {
      int rv;
      if (buffer.size > 0)
	FD_SET (remote_fd, &write_fds);
      if (buffer.size < MAX_NETCAT_BUFFER)
	FD_SET (0, &read_fds);
      FD_SET (0, &except_fds);
      FD_SET (remote_fd, &except_fds);
      rv = select (remote_fd + 1, &read_fds, &write_fds, &except_fds, NULL);
      if (rv < 0)
	{
	  if (errno == EINTR)
	    continue;
	  g_error ("do_netcat_forever: error running select()");
	}
      if (FD_ISSET (0, &except_fds))
	g_error ("do_netcat_forever: exception on stdin");
      if (FD_ISSET (remote_fd, &except_fds))
	g_error ("do_netcat_forever: exception on remote file descriptor");
      if (FD_ISSET (0, &read_fds))
	{
	  if (!gdam_buffer_read_one (&buffer, 0))
	    exit (0);
	  FD_CLR (0, &read_fds);
	}
      if (FD_ISSET (remote_fd, &write_fds))
	{
	  if (!gdam_buffer_write_one (&buffer, remote_fd))
	    exit (0);
	  FD_CLR (remote_fd, &write_fds);
	}
    }
  exit (1);
}

/* --- handle subprocess termination --- */
static void
handle_subprocess_exited (gpointer         user_data,
		          GdamWaitPidInfo *info)
{
  const char *program_name = user_data;
  if (info->exited)
    {
      if (info->d.exit_status != 0)
	g_warning ("process `%s' exited with nonzero termination status %d",
		 program_name, info->d.exit_status);
    }
  else
    {
      g_warning ("process `%s' was killed by signal %d",
		 program_name, info->d.signal);
    }
}

/* --- methods to connect / exec subprocesses which return an `output_id' --- */
guint
gdas_recorder_external_connect_remote(GdasRecorderExternal* recorder,
				      const char           *host,
				      guint                 port,
				      int                   buffer_size)
{
  int fds[2];
  int pid;
  if (pipe (fds) < 0)
    {
      g_warning ("error creating pipe for " "recorder_external_run_command");
      return 0;
    }
  /*  Set close-on-exec flag  */
  fcntl (fds[1], F_SETFD, 1);
  if ((pid = fork ()) < 0)
    {
      g_warning ("error forking process: %s", strerror (errno));
      return 0;
    }
  if (pid == 0)
    {
      gdam_realtime_unset ();
      close (fds[1]);
      dup2 (fds[0], STDIN_FILENO);
      do_netcat_forever (host, port);
    }
  close (fds[0]);
  gdam_wait_pid (pid, handle_subprocess_exited, "(builtin netcat)", NULL);
  return gdas_recorder_external_add_fd (recorder, fds[1], buffer_size, pid);
}
guint gdas_recorder_external_run_command(GdasRecorderExternal* recorder,
                                         const char* const*    command,
					 int                   buffer_size)
{
  int fds[2];
  int pid;
  if (pipe (fds) < 0)
    {
      g_warning ("error creating pipe for " "recorder_external_run_command");
      return 0;
    }
  /*  Set close-on-exec flag  */
  fcntl (fds[1], F_SETFD, 1);
  if ((pid = fork ()) < 0)
    {
      g_warning ("error forking process: %s", strerror (errno));
      return 0;
    }
  if (pid == 0)
    {
      gdam_realtime_unset ();
      close (fds[1]);
      dup2 (fds[0], STDIN_FILENO);
      execvp (command[0], (char **) command);
      _exit (127);
    }
  close (fds[0]);
  gdam_wait_pid (pid, handle_subprocess_exited, g_strdup (command[0]), g_free);
  return gdas_recorder_external_add_fd (recorder, fds[1], buffer_size, pid);
}

static gint has_id(gconstpointer output, gconstpointer id)
{
  return ((GdasRecorderOutput *) output)->id == GPOINTER_TO_UINT (id);
}

void  gdas_recorder_external_set_pause  (GdasRecorderExternal* recorder,
                                         guint                 sink_id,
					 gboolean              pause)
{
  GdasRecorderOutput *output;
  GSList *sink = g_slist_find_custom (recorder->recorders,
				      GUINT_TO_POINTER (sink_id),
				      has_id);
  if (sink == NULL)
    {
      gdam_debug ("recorder external: "
		  "error pausing sink %d: doesn't exist", sink_id);
      return;
    }
  output = sink->data;
  output->is_paused = pause;
}

void  gdas_recorder_external_kill_sink  (GdasRecorderExternal* recorder,
                                         guint                 sink_id)
{
  GSList *prev = NULL;
  GSList *at;
  GdasRecorderOutput *output = NULL;
  for (at = recorder->recorders; at != NULL; at = at->next)
    {
      output = at->data;
      if (output->id == sink_id)
	break;
      prev = at;
    }
  if (at == NULL)
    {
      gdam_debug ("recorder_external_kill_sink: no such sink %d", sink_id);
      return;
    }
  g_return_if_fail (output != NULL);
  if (prev == NULL)
    {
      g_assert (at == recorder->recorders);
      recorder->recorders = at->next;
    }
  else
    {
      prev->next = at->next;
    }
  g_slist_free_1 (at);

  /* Now, dispense with `output';
   * this will remove the glib polls and sources. */
  gdas_recorder_output_destroy (output);
}

static void gdas_recorder_external_record    (GtkObject *object,
				              int        num_in,
                                              GdamArg   *args_in,
				              int       *num_out,
				              GdamArg  **args_out)
{
  const char *filename = NULL;
  const char *options = NULL;
  int i;
  int buffer_size = (1 << 14);
  guint output_id;
  const char *command_array[4];
  GdasRecorderExternal *recorder;
  for (i = 0; i < num_in; i++)
    {
      if (strcmp (args_in[i].name, "filename") == 0
       && args_in[i].type == GDAM_ARG_TYPE_STRING)
        {
	  if (filename != NULL)
	    {
	      gdam_debug ("`filename' was specified twice to `record'");
	      return;
	    }
	  filename = GDAM_ARG_STRING (args_in + i);
	}
      else if (strcmp (args_in[i].name, "options") == 0
            && args_in[i].type == GDAM_ARG_TYPE_STRING)
        {
	  if (options != NULL)
	    {
	      gdam_debug ("`options' was specified twice to `record'");
	      return;
	    }
	  options = GDAM_ARG_STRING (args_in + i);
	}
      else
        {
	  gdam_debug ("unrecognized argument to record: %s", args_in[i].name);
	  return;
	}
    }
  if (filename == NULL)
    {
      gdam_debug ("no filename specified to `record'");
      return;
    }

  recorder = GDAS_RECORDER_EXTERNAL (object);

  /* TODO:  cannot record a file starting with - currently. */
  command_array[0] = "gdamexternalrecorderhelper";
  command_array[1] = filename;
  command_array[2] = options;
  command_array[3] = NULL;
  output_id = gdas_recorder_external_run_command (recorder,
                                                  command_array,
						  buffer_size);
  *num_out = 1;
  *args_out = gdam_arg_new_from_valist (1,
                                        "output_id",
					GDAM_ARG_TYPE_INT,
                                        output_id);
}

#define VERIFY_NO_OUTPUT_SPECIFIED()					\
	if (fd >= 0 || command != NULL || do_connect)			\
	  {								\
	    gdam_debug("two output flags were specified, "		\
		     "latest was %s", arg->name);			\
	    continue;							\
	  }

static void gdas_recorder_external_add_output(GtkObject* object,
				              int        num_in,
                                              GdamArg*   args_in,
				              int*       num_out,
				              GdamArg**  args_out)
{
  int fd = -1;
  int i;
  guint rv = 0;
  char **command = NULL;
  int buffer_size = 128 * 1024;
  gboolean do_connect = FALSE;
  char *connect_host = NULL;
  int connect_port = 6000;
  GdasRecorderExternal *recorder = GDAS_RECORDER_EXTERNAL (object);
  for (i = 0; i < num_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (strcmp (arg->name, "file") == 0
	  && arg->type == GDAM_ARG_TYPE_STRING)
	{
	  char *fname = GDAM_ARG_STRING (arg);
	  VERIFY_NO_OUTPUT_SPECIFIED ();
	  fd = open (fname, O_WRONLY | O_APPEND | O_CREAT, 0644);
	  if (fd < 0)
	    {
	      goto failed;
	    }
	  continue;
	}
      if (strcmp (arg->name, "truncate_file") == 0
	  && arg->type == GDAM_ARG_TYPE_STRING)
	{
	  char *fname = GDAM_ARG_STRING (arg);
	  VERIFY_NO_OUTPUT_SPECIFIED ();
	  fd = open (fname, O_WRONLY | O_TRUNC | O_CREAT, 0644);
	  if (fd < 0)
	    {
	      goto failed;
	    }
	  continue;
	}
      if (strcmp (arg->name, "connect") == 0
	  && arg->type == GDAM_ARG_TYPE_STRING)
	{
	  const char *spec = GDAM_ARG_STRING (arg);
	  const char *colon;
	  VERIFY_NO_OUTPUT_SPECIFIED ();
	  do_connect = TRUE;
	  colon = strchr (spec, ':');
	  if (colon == NULL)
	    {
	      connect_host = g_strdup (spec);
	    }
	  else
	    {
	      connect_host = g_new (char, colon - spec + 1);
	      memcpy (connect_host, spec, colon - spec);
	      connect_host[colon - spec] = '\0';
	      connect_port = atoi (colon + 1);
	    }
	  continue;
	}
      if (strcmp (arg->name, "buffer_size") == 0
	  && arg->type == GDAM_ARG_TYPE_INT)
	{
	  buffer_size = GDAM_ARG_INT (arg);
	  continue;
	}
    }
  if (do_connect)
    rv = gdas_recorder_external_connect_remote (recorder,
						connect_host, connect_port,
						buffer_size);
  else if (command != NULL)
    rv = gdas_recorder_external_run_command (recorder,
					     (const char *const *) command,
					     buffer_size);
  else if (fd >= 0)
    rv = gdas_recorder_external_add_fd (recorder, fd, buffer_size, 0);
  if (rv == 0)
    goto failed;

  *num_out = 1;
  *args_out = g_new (GdamArg, 1);
  (*args_out)->name = g_strdup ("output_id");
  (*args_out)->type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (*args_out) = rv;
  return;

failed:
  *args_out = NULL;
  *num_out = 0;
}
static void gdas_recorder_external_remove_output(GtkObject* object,
				                 int        num_in,
                                                 GdamArg*   args_in,
				                 int*       num_out,
				                 GdamArg**  args_out)
{
  int i;
  guint output_id = 0;
  GdasRecorderExternal *recorder = GDAS_RECORDER_EXTERNAL (object);
  for (i = 0; i < num_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (strcmp (arg->name, "output_id") == 0
	  && arg->type == GDAM_ARG_TYPE_INT)
	{
	  output_id = GDAM_ARG_INT (arg);
	  continue;
	}
      gdam_debug ("gdas_recorder_external_remove_output: "
		  "argument type %s unknown", arg->name);
    }
  if (output_id == 0)
    {
      *args_out = NULL;
      *num_out = 0;
      return;
    }
  gdas_recorder_external_kill_sink (recorder, output_id);
  *num_out = 1;
  *args_out = g_new (GdamArg, 1);
  (*args_out)->name = g_strdup ("success");
  (*args_out)->type = GDAM_ARG_TYPE_INT;
  GDAM_ARG_INT (*args_out) = 1;
  return;
}


static void gdas_recorder_external_class_init(GdasRecorderExternalClass* klass)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  source_class->read = gdas_recorder_external_read;
  object_class->destroy = gdas_recorder_external_destroy;
  gdam_arg_function_register (object_class,
			      "add_output",
			      gdas_recorder_external_add_output);
  gdam_arg_function_register (object_class,
			      "remove_output",
			      gdas_recorder_external_remove_output);
  gdam_arg_function_register (object_class,
			      "record",
			      gdas_recorder_external_record);
}

