/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasrespeed.h"
#include <gdam/common/gdamplugin.h>

static int 	gdas_respeed_process_block    (GdasBlockFilter*  filter,
	                                       GdamSample*       output,
					       const GdamSample* input);
 
static void     gdas_respeed_class_init       (GdasBlockFilterClass*);
static void     gdas_respeed_init             (GdasRespeed*      filter);
static void     gdas_respeed_destroy          (GdasRespeed*      respeed);


static GtkObjectClass* parent_class;

GtkType gdas_respeed_get_type()
{
  static GtkType block_filter_type = 0;
  if (!block_filter_type)
    {
      static const GtkTypeInfo filter_info = {
	"GdasRespeed",
	sizeof (GdasRespeed),
	sizeof (GdasRespeedClass),
	(GtkClassInitFunc) gdas_respeed_class_init,
	(GtkObjectInitFunc) gdas_respeed_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      parent_class = gtk_type_class (GDAS_TYPE_BLOCK_FILTER);
      block_filter_type =
	gtk_type_unique (GDAS_TYPE_BLOCK_FILTER, &filter_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return block_filter_type;
}


void     gdas_respeed_arg_set_rate     (GtkObject*	object,
                                        GdamArg*	rate)
{
  gdas_respeed_set_rate (GDAS_RESPEED (object), GDAM_ARG_DOUBLE (rate));
}

static gdouble get_rate(GdasBlockFilter* block_filter)
{
  return (gdouble) block_filter->output_block_size
    / (gdouble) block_filter->input_block_size;
}

void     gdas_respeed_arg_get_rate     (GtkObject*	object,
                                        GdamArg*	rate)
{
  GDAM_ARG_DOUBLE (rate) = get_rate (GDAS_BLOCK_FILTER (object));
}

static void gdas_respeed_arg_set_input_size(GtkObject* object, GdamArg* size)
{
  gdouble rate = get_rate (GDAS_BLOCK_FILTER (object));
  GdasRespeed *respeed = GDAS_RESPEED (object);
  respeed->input_size = GDAM_ARG_INT (size);
  gdas_respeed_set_rate (respeed, rate);
}
static void gdas_respeed_arg_get_input_size(GtkObject* object, GdamArg* size)
{
  GdasRespeed *respeed = GDAS_RESPEED (object);
  GDAM_ARG_INT (size) = respeed->input_size;
}


void     gdas_respeed_arg_set_smoothing(GtkObject*	object,
                                        GdamArg*	rate)
{
  GDAS_RESPEED (object)->smoothing = GDAM_ARG_INT (rate);
}

void     gdas_respeed_arg_get_smoothing(GtkObject*	object,
                                        GdamArg*	rate)
{
  GDAM_ARG_INT (rate) = GDAS_RESPEED (object)->smoothing;
}
	
static void gdas_respeed_class_init(GdasBlockFilterClass* klass)
{
  GtkObjectClass *object_class = (GtkObjectClass *) klass;
  object_class->destroy = (void (*)(GtkObject *)) gdas_respeed_destroy;
  klass->process_block = gdas_respeed_process_block;
  gdam_arg_register (object_class,
		     "rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_respeed_arg_set_rate,
		     gdas_respeed_arg_get_rate,
		     NULL, "0.5,1.5", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "input_size",
		     GDAM_ARG_TYPE_INT,
		     gdas_respeed_arg_set_input_size,
		     gdas_respeed_arg_get_input_size,
		     NULL, "1,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "smoothing",
		     GDAM_ARG_TYPE_INT,
		     gdas_respeed_arg_set_smoothing,
		     gdas_respeed_arg_get_smoothing,
		     NULL, "boolean", GDAM_ARG_VITAL);
}

typedef void (*GdamSampleFadeFunc)(GdamSample*          output, 
                                   const GdamSample*    input,
				   int                  length,
				   int                  volume_b24,
				   int                  volume_change_b24);

/* Triangular fades. */
static void gdam_sample_fade_copy_b24(GdamSample* output,
                                  const GdamSample* input,
				  int length,
				  int volume,
				  int volume_change)
{
  while (length--)
    {
      output->s[0] = (GdamSampleValue)
	(((gint64) volume * input->s[0]) >> 24);
      output->s[1] = (GdamSampleValue)
	(((gint64) volume * input->s[1]) >> 24);
      volume += volume_change;
      input++;
      output++;
    }
}
static void gdam_sample_fade_add_b24 (GdamSample* output,
                                  const GdamSample* input,
				  int length,
				  int volume,
				  int volume_change)
{
  gint64 vol = volume;
  while (length--)
    {
      output->s[0] += (GdamSampleValue) ((vol * input->s[0]) >> 24);
      output->s[1] += (GdamSampleValue) ((vol * input->s[1]) >> 24);
      volume += volume_change;
      input++;
      output++;
    }
}

/* Smoothing fade. */
#define SMOOTHING_TABLE_LOG_SIZE	11
#include "respeed-smoothing-table.inc"
static void gdam_sample_smoothing_fade_copy_b24(GdamSample* output,
                                                const GdamSample* input,
				                int length,
				                int volume,
				                int volume_change)
{
  while (length--)
    {
      GdamSampleProduct vol;
      int smoothing_index = volume >> (24 - SMOOTHING_TABLE_LOG_SIZE);
      vol = smoothing_table[smoothing_index] / 2;
      output->s[0] = (GdamSampleValue) ((vol * input->s[0]) >> 15);
      output->s[1] = (GdamSampleValue) ((vol * input->s[1]) >> 15);
      volume += volume_change;
      input++;
      output++;
    }
}
static void gdam_sample_smoothing_fade_add_b24 (GdamSample* output,
                                                const GdamSample* input,
				                int length,
				                int volume,
				                int volume_change)
{
  while (length--)
    {
      GdamSampleProduct vol;
      int smoothing_index = volume >> (24 - SMOOTHING_TABLE_LOG_SIZE);
      vol = smoothing_table[smoothing_index];
      output->s[0] += (GdamSampleValue) ((vol * input->s[0]) >> 16);
      output->s[1] += (GdamSampleValue) ((vol * input->s[1]) >> 16);
      volume += volume_change;
      input++;
      output++;
    }
}

static int 	gdas_respeed_process_block    (GdasBlockFilter*  filter,
	                                       GdamSample*       output,
					       const GdamSample* input)
{
  GdasRespeed *respeed = (GdasRespeed *) filter;
  int back_buf_size = respeed->back_buffer_size;
  int track_1_len = back_buf_size + respeed->replay_amount;
  int track_2_len = filter->input_block_size;
  int output_size = filter->output_block_size;
  int input_size = filter->input_block_size;
  int overlap;
  int fade_velocity_b24;
  int actual_fade_amount_b24, vol_1_b24, vol_2_b24;
  GdamSample *back_buf = respeed->back_buffer;
  GdamSample *output_at;
  int track_1_unoverlapped, track_2_unoverlapped;
  const GdamSample *input_at;

  GdamSampleFadeFunc init_func, add_func;

  if (respeed->smoothing)
    {
      init_func = gdam_sample_smoothing_fade_copy_b24;
      add_func = gdam_sample_smoothing_fade_add_b24;
    }
  else
    {
      init_func = gdam_sample_fade_copy_b24;
      add_func = gdam_sample_fade_add_b24;
    }

  if (track_2_len > output_size)
    track_2_len = output_size;
  overlap = track_1_len + track_2_len - filter->output_block_size;
  if (overlap <= 0)
    {
      gdam_sample_copy (output, back_buf, back_buf_size);
      gdam_sample_copy (output + back_buf_size,
			input, track_1_len - back_buf_size);
      gdam_sample_copy (output + output_size - track_2_len,
			input + input_size - track_2_len, track_2_len);
      return output_size;
    }

  fade_velocity_b24 = (1 << 24) / overlap;
  actual_fade_amount_b24 = fade_velocity_b24 * overlap;


  /* Copy (track_1_len - overlap) bytes verbatim from
   * from back_buf and start of input. */
  track_1_unoverlapped = track_1_len - overlap;
  output_at = output;
  if (back_buf_size >= track_1_unoverlapped)
    {
      /* All the unmixed bytes come from back buffer. */
      gdam_sample_copy (output_at,
			respeed->back_buffer, track_1_unoverlapped);
    }
  else
    {
      gdam_sample_copy (output_at,
			respeed->back_buffer, respeed->back_buffer_size);
      gdam_sample_copy (output_at + respeed->back_buffer_size,
			input,
			track_1_unoverlapped - respeed->back_buffer_size);
    }

  output_at += track_1_unoverlapped;
  vol_1_b24 = actual_fade_amount_b24;
  vol_2_b24 = (1 << 24) - actual_fade_amount_b24;

  /* Write overlapped region from track 1. */
  if (respeed->back_buffer_size >= track_1_len)
    {
      /* All overlapped samples come from the past-buffer. */
      g_assert (respeed->back_buffer_size == track_1_len);
      (*init_func) (output_at,
		    respeed->back_buffer,
		    track_1_len, vol_1_b24, -fade_velocity_b24);
    }
  else if (track_1_unoverlapped >= respeed->back_buffer_size)
    {
      /* All overlapped samples are replays from input */
      int start_index;
      start_index = track_1_unoverlapped - respeed->back_buffer_size;
      (*init_func) (output_at,
		    input + start_index,
		    overlap, vol_1_b24, -fade_velocity_b24);
    }
  else
    {
      /* The overlapped region that comes from track 1
       * straddles the back_buf/input boundary. */
      int from_back, from_replay;
      from_back = respeed->back_buffer_size - track_1_unoverlapped;
      from_replay = overlap - from_back;
      g_assert (from_back >= 0 && from_replay >= 0);
      (*init_func) (output_at,
		    respeed->back_buffer,
		    from_back, vol_1_b24, -fade_velocity_b24);
      (*init_func) (output_at + from_back,
		    input,
		    from_replay,
		    vol_1_b24 - fade_velocity_b24 * from_back,
		    -fade_velocity_b24);
    }

  /* Append overlapped region from track 2. */
  (*add_func) (output_at, input, overlap, vol_2_b24, fade_velocity_b24);

  output_at += overlap;

  /* Write (track_2_len - overlap) bytes from input to output,
   * without any fading. */
  track_2_unoverlapped = input_size - overlap;
  {
    int max_write = output + output_size - output_at;
    input_at = input + overlap;
    if (max_write < track_2_unoverlapped)
      {
	input_at = input + track_2_unoverlapped - max_write;
	track_2_unoverlapped = max_write;
      }
  }
  gdam_sample_copy (output_at, input_at, track_2_unoverlapped);

  /* Copy the appropriate amount from input into back_buf. */
  gdam_sample_copy (back_buf,
		    input + input_size - back_buf_size, back_buf_size);

  return output_size;
}

#define MIN_FADE_LENGTH 100
void	gdas_respeed_set_rate    (GdasRespeed*  respeed,
	                          gdouble       rate)
{
  GdasBlockFilter *block_filter = (GdasBlockFilter *) respeed;
  int input_size = respeed->input_size;
  int replay_amount;
  int back_buf_size;
  int output_size;
  input_size = input_size ? input_size : 1024;
  if (rate < 1.0)
    {
      output_size = (int) ((gdouble) input_size / rate);
      replay_amount = output_size - input_size + MIN_FADE_LENGTH;
      if (replay_amount > input_size)
	{
	  gdam_debug ("rate was too small for respeed");
	  replay_amount = input_size;
	  output_size = replay_amount + input_size - MIN_FADE_LENGTH;
	}
      if (replay_amount < MIN_FADE_LENGTH)
	back_buf_size = MIN_FADE_LENGTH - replay_amount;
      else
	back_buf_size = 0;
    }
  else
    {
      output_size = (int) ((gdouble) input_size / rate);
      back_buf_size = input_size - output_size;
      if (back_buf_size < MIN_FADE_LENGTH)
	replay_amount = MIN_FADE_LENGTH - back_buf_size;
      else
	replay_amount = 0;
    }

  /* TODO: make this shuffle the data and zero it better... */
  respeed->back_buffer = g_renew (GdamSample,
				  respeed->back_buffer, back_buf_size);
  if (respeed->back_buffer_size < back_buf_size)
    {
      memset (respeed->back_buffer + respeed->back_buffer_size,
	      0,
	      (back_buf_size - respeed->back_buffer_size)
	      * sizeof (GdamSample));
    }
  respeed->back_buffer_size = back_buf_size;
  respeed->replay_amount = replay_amount;
  gdas_block_filter_set_sizes (block_filter, input_size, output_size);
}

static void gdas_respeed_init(GdasRespeed* filter)
{
  filter->input_size = 1024;
}

static void gdas_respeed_destroy(GdasRespeed* respeed)
{
  g_free (respeed->back_buffer);
  parent_class->destroy ((GtkObject *) respeed);
}
