/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Mike Radford, Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
	mrad@blorf.com <Mike Radford>
*/
#include "gdasspatialstereo.h"
#include <gdam/common/gdamplugin.h>
#include <math.h>

#define SAMPLE_RATE (44100.0)

int gdas_spatial_stereo_read(GdasSource* input,
	                     GdamSample* buf,
			     int max_read);

static void gdas_spatial_stereo_init(GdasSpatialStereo* this)
{
	this->speed_of_sound = 332.0;
	this->lpos_y = this->lpos_z = 0.0;
	this->rpos_y = this->rpos_z = 0.0;
	this->lear_y = this->lear_z = 0.0;
	this->rear_y = this->rear_z = 0.0;

	this->lear_x = -100.0;
	this->lpos_x = -100.1;
	this->rear_x = 100.0;
	this->rpos_x = 100.1;
	this->unit_gain_dist = 0.1;

	gdas_buffering_filter_ensure_space(
		GDAS_BUFFERING_FILTER(this), (int)SAMPLE_RATE);
}

#define define_set_and_get(arg_name, prefix)				\
static void gdam_spatial_arg_set_##arg_name				\
				(GtkObject* object, GdamArg* arg)	\
{									\
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);	\
	spatial->##prefix##_x = GDAM_ARG_DOUBLE_ARRAY(arg).array[0];	\
	spatial->##prefix##_y = GDAM_ARG_DOUBLE_ARRAY(arg).array[1];	\
	spatial->##prefix##_z = GDAM_ARG_DOUBLE_ARRAY(arg).array[2];	\
}									\
									\
static void gdam_spatial_arg_get_##arg_name				\
				(GtkObject* object, GdamArg* arg)	\
{									\
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);	\
	GDAM_ARG_DOUBLE_ARRAY(arg).length = 3;				\
	GDAM_ARG_DOUBLE_ARRAY(arg).array = g_new(double, 3);		\
	GDAM_ARG_DOUBLE_ARRAY(arg).array[0] = spatial->##prefix##_x;	\
	GDAM_ARG_DOUBLE_ARRAY(arg).array[1] = spatial->##prefix##_y;	\
	GDAM_ARG_DOUBLE_ARRAY(arg).array[2] = spatial->##prefix##_z;	\
}

define_set_and_get(left_ear,       lear)
define_set_and_get(right_ear,      rear)
define_set_and_get(left_position,  lpos)
define_set_and_get(right_position, rpos)
static void gdam_spatial_arg_set_speed_of_sound(GtkObject* object,
                                                GdamArg* arg)
{
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);
	spatial->speed_of_sound = GDAM_ARG_DOUBLE(arg);
}
static void gdam_spatial_arg_get_speed_of_sound(GtkObject* object,
                                                GdamArg* arg)
{
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);
	GDAM_ARG_DOUBLE(arg) = spatial->speed_of_sound;
}
static void gdam_spatial_arg_set_unit_gain_distance(GtkObject* object,
                                                GdamArg* arg)
{
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);
	spatial->unit_gain_dist = GDAM_ARG_DOUBLE(arg);
}
static void gdam_spatial_arg_get_unit_gain_distance(GtkObject* object,
                                                GdamArg* arg)
{
	GdasSpatialStereo* spatial = GDAS_SPATIAL_STEREO(object);
	GDAM_ARG_DOUBLE(arg) = spatial->unit_gain_dist;
}

static void gdas_spatial_stereo_class_init(
		GdasSpatialStereoClass* spatial_stereo_class)
{
	GDAS_SOURCE_CLASS(spatial_stereo_class)->read =
		gdas_spatial_stereo_read;

	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "left_ear",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_spatial_arg_set_left_ear,
			  gdam_spatial_arg_get_left_ear,
			  NULL,
			  ",",
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "right_ear",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_spatial_arg_set_right_ear,
			  gdam_spatial_arg_get_right_ear,
			  NULL,
			  ",",
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "left_position",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_spatial_arg_set_left_position,
			  gdam_spatial_arg_get_left_position,
			  NULL,
			  ",",
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "right_position",
			  GDAM_ARG_TYPE_DOUBLE_ARRAY,
			  gdam_spatial_arg_set_right_position,
			  gdam_spatial_arg_get_right_position,
			  NULL,
			  ",",
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "speed_of_sound",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_spatial_arg_set_speed_of_sound,
			  gdam_spatial_arg_get_speed_of_sound,
			  NULL,
			  "0,",
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(spatial_stereo_class),
	                  "unit_gain_distance",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_spatial_arg_set_unit_gain_distance,
			  gdam_spatial_arg_get_unit_gain_distance,
			  NULL,
			  "0.00001,",
			  GDAM_ARG_VITAL);
}

GtkType gdas_spatial_stereo_get_type()
{
	static GtkType spatial_stereo_type = 0;
	if (!spatial_stereo_type) {
		static const GtkTypeInfo spatial_stereo_info = {
			"GdasSpatialStereo",
			sizeof(GdasSpatialStereo),
			sizeof(GdasSpatialStereoClass),
			(GtkClassInitFunc) gdas_spatial_stereo_class_init,
			(GtkObjectInitFunc) gdas_spatial_stereo_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		spatial_stereo_type = gtk_type_unique(GDAS_TYPE_BUFFERING_FILTER,
		                                 &spatial_stereo_info);
		gdam_plugin_check_version(__FILE__, GDAM_VERSION);
	}
	return spatial_stereo_type;
}

inline void source_to_ear_calc(GdasSpatialStereo *this,
			       double ex, double ey, double ez,
                               double px, double py, double pz,
                               int *delay,
                               double *gain)
{
	const double x = px - ex;
	const double y = py - ey;
	const double z = pz - ez;
	const double r = sqrt(x*x + y*y + z*z);
	*delay = rint(SAMPLE_RATE*r/this->speed_of_sound);
	*gain = this->unit_gain_dist/r;
}

int gdas_spatial_stereo_read(GdasSource* input,
	                     GdamSample* buf,
			     int max_read)
{
	GdasSpatialStereo *this = GDAS_SPATIAL_STEREO(input);
	GdasBufferingFilter *this_buffering_filter =
		GDAS_BUFFERING_FILTER(this);
	int delay_lcle, delay_lcre, delay_rcle, delay_rcre;
	double gain_lcle, gain_lcre, gain_rcle, gain_rcre;
	int amt_read;
	int i;
	if ((amt_read = gdas_buffering_filter_buffer_from_source(
			this_buffering_filter,
			GDAS_FILTER(this)->filtered, max_read)) <= 0) {
		gdam_debug("spatial_stereo: buffering_filter returned %d",
			amt_read);
		return amt_read;
	}
	source_to_ear_calc(this,
			   this->lpos_x, this->lpos_y, this->lpos_z,
			   this->lear_x, this->lear_y, this->lear_z,
		           &delay_lcle, &gain_lcle);
	source_to_ear_calc(this,
			   this->lpos_x, this->lpos_y, this->lpos_z,
			   this->rear_x, this->rear_y, this->rear_z,
		           &delay_lcre, &gain_lcre);
	source_to_ear_calc(this,
			   this->rpos_x, this->rpos_y, this->rpos_z,
			   this->lear_x, this->lear_y, this->lear_z,
		           &delay_rcle, &gain_rcle);
	source_to_ear_calc(this,
			   this->rpos_x, this->rpos_y, this->rpos_z,
			   this->rear_x, this->rear_y, this->rear_z,
		           &delay_rcre, &gain_rcre);


	delay_lcle += amt_read;
	delay_rcle += amt_read;
	delay_lcre += amt_read;
	delay_rcre += amt_read;
	for (i = 0; i < amt_read; ++i) {
		int left = rint(
			gain_lcle * (double)gdas_buffering_filter_sample_at(
				this_buffering_filter, delay_lcle--).s[0] +
			gain_rcle * (double)gdas_buffering_filter_sample_at(
				this_buffering_filter, delay_rcle--).s[1]
		);
		int right = rint(
			gain_lcre * (double)gdas_buffering_filter_sample_at(
				this_buffering_filter, delay_lcre--).s[0] +
			gain_rcre * (double)gdas_buffering_filter_sample_at(
				this_buffering_filter, delay_rcre--).s[1]
		);

		left = GDAM_SAMPLE_CLAMP(left);
		right = GDAM_SAMPLE_CLAMP(right);

		buf[i].s[0] = left;
		buf[i].s[1] = right;
	}
	return amt_read;
}

