/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdassplinemap.h"

static void
spline_solve (int n, gdouble x[], gdouble y[], gdouble y2[])
{
  gdouble p, sig, *u;
  gint i, k;

  u = g_malloc ((n - 1) * sizeof (u[0]));

  y2[0] = u[0] = 0.0;		/* set lower boundary condition to "natural" */

  for (i = 1; i < n - 1; ++i)
    {
      sig = (x[i] - x[i - 1]) / (x[i + 1] - x[i - 1]);
      p = sig * y2[i - 1] + 2.0;
      y2[i] = (sig - 1.0) / p;
      u[i] = ((y[i + 1] - y[i])
	      / (x[i + 1] - x[i]) - (y[i] - y[i - 1]) / (x[i] - x[i - 1]));
      u[i] = (6.0 * u[i] / (x[i + 1] - x[i - 1]) - sig * u[i - 1]) / p;
    }

  y2[n - 1] = 0.0;
  for (k = n - 2; k >= 0; --k)
    y2[k] = y2[k] * y2[k + 1] + u[k];

  g_free (u);
}

static void gdas_spline_map_solve(GdasSplineMap* spline_map)
{
  spline_solve (spline_map->num_points,
		spline_map->x, spline_map->y, spline_map->y2);
}


static gdouble
spline_eval (int n, gdouble x[], gdouble y[], gdouble y2[], gdouble val)
{
  gint k_lo, k_hi, k;
  gdouble h, b, a;

  /* do a binary search for the right interval: */
  k_lo = 0;
  k_hi = n - 1;
  while (k_hi - k_lo > 1)
    {
      k = (k_hi + k_lo) / 2;
      if (x[k] > val)
	k_hi = k;
      else
	k_lo = k;
    }

  h = x[k_hi] - x[k_lo];
  g_assert (h > 0.0);

  a = (x[k_hi] - val) / h;
  b = (val - x[k_lo]) / h;
  return a * y[k_lo] + b * y[k_hi] +
    ((a * a * a - a) * y2[k_lo] + (b * b * b - b) * y2[k_hi]) * (h * h) / 6.0;
}


static inline GdamSampleValue remap_by_spline(GdasSplineMap*  map, 
                                              GdamSampleValue value)
{
  gdouble x = GDAM_SAMPLE_TO_FLOAT (value);
  gdouble y;
  y = spline_eval (map->num_points, map->x, map->y, map->y2, x);
  return GDAM_SAMPLE_FROM_FLOAT (y);
}

static int  gdas_spline_map_read(GdasSource*       source,
                                 GdamSample*       buffer,
				 int               max_read)
{
  int num_read;
  int i;
  GdasSplineMap *map = GDAS_SPLINE_MAP (source);
  GdamSample *at = buffer;
  num_read = gdas_filter_sub_read (GDAS_FILTER (source), buffer, max_read);
  for (i = 0; i < num_read; i++)
    {
      at->s[0] = remap_by_spline (map, at->s[0]);
      at->s[1] = remap_by_spline (map, at->s[1]);
      at++;
    }
  return num_read;
}

static void gdas_spline_map_init(GdasSplineMap* spline_map)
{
  spline_map->x = g_new (gdouble, 6);
  spline_map->y = spline_map->x + 2;
  spline_map->y2 = spline_map->x + 4;
  spline_map->num_points = 2;
  spline_map->x[0] = -1;
  spline_map->y[0] = -1;
  spline_map->x[1] = 1;
  spline_map->y[1] = 1;
  gdas_spline_map_solve (spline_map);
}

static void gdas_spline_map_arg_set_spline(GtkObject* object, GdamArg* arg)
{
  GdasSplineMap *map;
  int num_points = GDAM_ARG_DOUBLE_ARRAY (arg).length;
  gdouble *points = GDAM_ARG_DOUBLE_ARRAY (arg).array;
  if (num_points % 2 != 0)
    {
      g_warning ("gdas_spline_map_arg_set_spline: array must be even");
      return;
    }
  num_points /= 2;

  map = GDAS_SPLINE_MAP (object);
  g_free (map->x);
  map->num_points = num_points;
  map->x = g_new (gdouble, num_points * 3);
  map->y = map->x + num_points;
  map->y2 = map->x + num_points * 2;
  memcpy (map->x, points, num_points * sizeof (gdouble) * 2);
  gdas_spline_map_solve (map);
}
static void gdas_spline_map_arg_get_spline(GtkObject* object, GdamArg* arg)
{
  GdasSplineMap *map;
  map = GDAS_SPLINE_MAP (object);
  GDAM_ARG_DOUBLE_ARRAY (arg).length = map->num_points * 2;
  GDAM_ARG_DOUBLE_ARRAY (arg).array
    = g_memdup (map->x, sizeof (gdouble) * 2 * map->num_points);
}

static void gdas_spline_map_class_init(GdasSplineMapClass* spline_map_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (spline_map_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (spline_map_class);
  gdam_arg_register (object_class,
		     "spline",
		     GDAM_ARG_TYPE_DOUBLE_ARRAY,
		     gdas_spline_map_arg_set_spline,
		     gdas_spline_map_arg_get_spline,
		     NULL, NULL, GDAM_ARG_VITAL);
  source_class->read = gdas_spline_map_read;
}

GtkType gdas_spline_map_get_type()
{
  static GtkType spline_map_type = 0;
  if (!spline_map_type)
    {
      static const GtkTypeInfo spline_map_info = {
	"GdasSplineMap",
	sizeof (GdasSplineMap),
	sizeof (GdasSplineMapClass),
	(GtkClassInitFunc) gdas_spline_map_class_init,
	(GtkObjectInitFunc) gdas_spline_map_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      spline_map_type = gtk_type_unique (parent, &spline_map_info);
    }
  return spline_map_type;
}
