/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdasvibrato.h"
#include <gdam/common/gdamplugin.h>
#include <gdam/server/gdamconnection.h>
#include <stdio.h>

#define MAX_AVERAGE_DIST	3

static void gdas_vibrato_destroy(GtkObject*);
static gboolean gdas_vibrato_detach_child(GdasSource* source, GdasSource* sub);
static GtkObjectClass* parent_class = NULL;
static void        gdas_vibrato_get_subsources  (GdasSource*      source,
                                                 int*             num_subs,
						 GdasSource***    subs);

/*
 * Attempts to read a certain number of samples
 * out of `filtered', with a few useful member
 * as parameters, since this function is actually
 * heavily inlined, believe it or not...
 * (If it is the fast case, if exits with just a gdam_buffer_read...
 *  which isn't *that* fast... but this is a complex effect.)
 *
 *
 * `frac' is the current fraction of a sample we have processed
 * from filtered -- `distance' is the number of samples to read (b32).
 */
static        gboolean read_distance(GdasVibrato* this,
		                     GdamBuffer*  filtered_buffer,
				     gint64       frac,
				     gint64       distance,
			             GdamSample*  out)
{
  GdasSource *filtered = GDAS_FILTER (this)->filtered;
  int skip_distance = (int) ((frac + distance) >> 32);

  /* These three numbers all have 12 bits of fraction,
   * `start_distance' is the initialize distance in samples (b12)
   * and left64/right64 and the intergral of the volume
   * over that distance.
   *
   * XXX: compute when this will overflow on 24-bit samples
   *      and see if it needs fixing.
   */
  GdamFixed start_distance = (distance >> 20);
  gint64 left64 = 0, right64 = 0;

  while ((int) (filtered_buffer->size / sizeof (GdamSample))
	 < skip_distance + 1)
    {
      if (gdas_source_is_ready_to_read (filtered))
	{
	  GdamSample buf[4096];
	  int num = sizeof (buf) / sizeof (GdamSample);
	  num = gdas_source_read (filtered, buf, num);
	  if (num > 0)
	    {
	      gdam_buffer_append (filtered_buffer,
				  buf, num * sizeof (GdamSample));
	      continue;
	    }
	}

      /* Too short... */
      gdam_buffer_destruct (filtered_buffer);
      return FALSE;
    }

  if (skip_distance > MAX_AVERAGE_DIST)
    {
      gdam_buffer_discard (filtered_buffer,
			   skip_distance * sizeof (GdamSample));
      gdam_buffer_read (filtered_buffer, out, sizeof (GdamSample));
      return TRUE;
    }

  while (distance >= (1LL << 32) - frac)
    {
      gint64 remaining = (1LL << 32) - frac;
      GdamSample at;
      gdam_buffer_read (filtered_buffer, &at, sizeof (GdamSample));
      left64 += (GdamSampleProduct) (remaining >> 20) * at.s[0];
      right64 += (GdamSampleProduct) (remaining >> 20) * at.s[1];
      distance -= remaining;
      frac = 0;
    }
  if (distance > 0)
    {
      GdamSample at;
      gdam_buffer_peek (filtered_buffer, &at, sizeof (GdamSample));
      left64 += (gint64) (distance >> 20) * at.s[0];
      right64 += (gint64) (distance >> 20) * at.s[1];
    }

  /* Without this +1 to start_distance, we wind up at exactly
   * (1<<15) sometimes, which is 1 greater than allowed by
   * GDAM_SAMPLE_{MIN,MAX}. */
  left64 /= (start_distance + 1);
  right64 /= (start_distance + 1);
  out->s[0] = (GdamSampleValue) left64;
  out->s[1] = (GdamSampleValue) right64;
  return TRUE;
}

static int  gdas_vibrato_read(GdasSource* this_source,
                              GdamSample* buf,
			      int         max_read)
{
  GdasVibrato *this_vibrato = GDAS_VIBRATO (this_source);
  gint64 frac_b32 = this_vibrato->frac_b32;
  GdamBuffer *vibrato_buffer = &this_vibrato->vibrato_buffer;
  GdamBuffer *filtered_buffer = &this_vibrato->filtered_buffer;

  /* The minimum resample rate (the resample rate for a source
   * whose samples are GDAM_MIN_SAMPLE). (In b16) */
  GdamFixed min_distance = this_vibrato->min_traverse;

  /* The amount over the minimum the resample rate will surge
   * to if at the maximum amplitude (GDAM_MAX_SAMPLE). */
  GdamFixed distance_range = this_vibrato->max_traverse - min_distance;

  GdasSource *vibrato = this_vibrato->vibrato;
  GdasSource *filtered = GDAS_FILTER (this_vibrato)->filtered;
  int num_vibrato_read, i;

  /* Reduce distance_range to b16. */
  distance_range >>= (GdamFixed_LOG_ONE - 16);
  min_distance >>= (GdamFixed_LOG_ONE - 16);

  if (filtered == NULL)
    {
      return 0;
    }
  if (vibrato == NULL)
    {
      /* TODO: resample. */
      return gdas_source_read (filtered, buf, max_read);
    }

  i = gdam_buffer_read (vibrato_buffer, buf, max_read * sizeof (GdamSample));
  i /= sizeof (GdamSample);
  if (max_read > i)
    {
      int max_to_read = max_read - i;
      int amt_read;
      if (vibrato != NULL)
	{
	  if (gdas_source_is_ready_to_read (vibrato))
	    {
	      amt_read = gdas_source_read (vibrato, buf + i, max_to_read);
	    }
	  else
	    {
	      gdam_sample_zero (buf + i, max_to_read);
	      amt_read = max_to_read;
	    }
	  if (amt_read > 0)
	    i += amt_read;
	}
    }
  num_vibrato_read = i;
  i = 0;

  g_return_val_if_fail (i == 0, 0);

  /* Iterate this loop at most once per output sample. */
  while (i < num_vibrato_read)
    {
      gint64 dist64;
      GdamFixed distance = ((int) buf[i].s[0] + buf[i].s[1]) / 2;
      distance -= GDAM_SAMPLE_ONE;

      /* XXX: just make this number the way it used to be,
         with 16 bits of fraction.  The above
         computations leave it with GDAM_SAMPLE_BITS bits...

         These few lines should be rewritten to use the
         full 24 bits of data. */
      distance >>= (GDAM_SAMPLE_BITS - 16);

      dist64 = ((gint64) distance * distance_range)
	+ ((gint64) min_distance << 16);
      g_assert (dist64 >= 0);
      if (!read_distance (this_vibrato,
			  filtered_buffer, frac_b32, dist64, buf + i))
	{
	  if (i == 0 && filtered && !filtered->ended)
	    {
	      /* HACK: prevent eof in this case... */
	      gdam_debug ("hack enabled in vibrato");
	      this_vibrato->frac_b32 = frac_b32;
	      buf[0] = this_vibrato->last_sample;
	      return 1;
	    }
	  this_vibrato->frac_b32 = frac_b32;
	  if (i > 0)
	    this_vibrato->last_sample = buf[i - 1];
	  break;
	}
      frac_b32 += dist64;
      frac_b32 %= (1LL << 32);
      i++;
    }
  gdam_buffer_append (vibrato_buffer,
		      buf + i, (num_vibrato_read - i) * sizeof (GdamSample));
  if (i == 0)
    return 0;
  this_vibrato->frac_b32 = frac_b32;
  this_vibrato->last_sample = buf[i - 1];
  return i;
}

static void gdas_vibrato_arg_set_min_rate(GtkObject* object, GdamArg* arg)
{
  gdouble rate = GDAM_ARG_DOUBLE (arg);
  if (rate < 0.0)
    rate = 0.0;
  GDAS_VIBRATO (object)->min_traverse = (int) (rate * GdamFixed_ONE);
}

static void gdas_vibrato_arg_get_min_rate(GtkObject* object, GdamArg* arg)
{
  gdouble rate;
  rate = (gdouble) GDAS_VIBRATO (object)->min_traverse / GdamFixed_ONE;
  GDAM_ARG_DOUBLE (arg) = rate;
}

static void gdas_vibrato_arg_set_max_rate(GtkObject* object, GdamArg* arg)
{
  gdouble rate = GDAM_ARG_DOUBLE (arg);
  if (rate < 0.0)
    rate = 0.0;
  GDAS_VIBRATO (object)->max_traverse = (int) (rate * GdamFixed_ONE);
}

static void gdas_vibrato_arg_get_max_rate(GtkObject* object, GdamArg* arg)
{
  gdouble rate;
  rate = (gdouble) GDAS_VIBRATO (object)->max_traverse / GdamFixed_ONE;
  GDAM_ARG_DOUBLE (arg) = rate;
}


static void gdas_vibrato_arg_set_vibrato(GtkObject* object, GdamArg* arg)
{
  int source_id = GDAM_ARG_INT (arg);
  GdasSource *vibrato = gdam_connection_get_source (NULL, source_id);
  if (vibrato == NULL)
    {
      gdam_debug ("set_vibrato: vibrato source id was invalid");
      return;
    }
  gdas_vibrato_set_vibrato (GDAS_VIBRATO (object), vibrato);
  gtk_object_unref (GTK_OBJECT (vibrato));
}

static void gdas_vibrato_arg_get_vibrato(GtkObject* object, GdamArg* arg)
{
  GdasSource *vibrato = GDAS_VIBRATO (object)->vibrato;
  GDAM_ARG_INT (arg) = (vibrato != NULL)
    ? GDAM_SERVER_OBJECT (vibrato)->global_id : 0;
}

static gboolean gdas_vibrato_is_ready_to_read(GdasSource* source)
{
  GdasVibrato *this_vibrato = GDAS_VIBRATO (source);
  GdasFilter *this_filter = GDAS_FILTER (source);
  gboolean is_vib_ready, is_filtered_ready;
  is_vib_ready = this_vibrato->vibrato_buffer.size
    || this_vibrato->vibrato == NULL
    || gdas_source_is_ready_to_read (this_vibrato->vibrato);
  is_filtered_ready = this_vibrato->filtered_buffer.size
    || this_filter->filtered == NULL
    || gdas_source_is_ready_to_read (this_filter->filtered);
  return is_vib_ready && is_filtered_ready;
}
static void gdas_vibrato_miss_data(GdasSource* source, int missed)
{
  GdasVibrato *this_vibrato = GDAS_VIBRATO (source);
  GdasFilter *this_filter = GDAS_FILTER (source);
  int m;
  {
    int discarded = missed * sizeof (GdamSample);
    discarded = gdam_buffer_discard (&this_vibrato->vibrato_buffer,
				     discarded);
    m = missed - discarded / sizeof (GdamSample);
  }
  if (m > 0 && this_vibrato->vibrato)
    gdas_source_miss_data (this_vibrato->vibrato, m);

  {
    int discarded = missed * sizeof (GdamSample);
    discarded = gdam_buffer_discard (&this_vibrato->filtered_buffer,
				     discarded);
    m = missed - discarded / sizeof (GdamSample);
  }
  if (m > 0 && this_filter->filtered)
    gdas_source_miss_data (this_filter->filtered, m);
}

static void gdas_vibrato_init(GdasVibrato* vibrato)
{
  gdam_buffer_construct (&vibrato->filtered_buffer);
  gdam_buffer_construct (&vibrato->vibrato_buffer);
  vibrato->min_traverse = vibrato->max_traverse = GdamFixed_ONE;
}
static void gdas_vibrato_class_init(GdasVibratoClass* vibrato_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (vibrato_class);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (vibrato_class);
  source_class->read = gdas_vibrato_read;
  source_class->is_ready_to_read = gdas_vibrato_is_ready_to_read;
  source_class->miss_data = gdas_vibrato_miss_data;
  source_class->detach_child = gdas_vibrato_detach_child;
  source_class->get_subsources = gdas_vibrato_get_subsources;
  object_class->destroy = gdas_vibrato_destroy;

  gdam_arg_register (object_class,
		     "vibrato",
		     GDAM_ARG_TYPE_INT,
		     gdas_vibrato_arg_set_vibrato,
		     gdas_vibrato_arg_get_vibrato,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "min_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_vibrato_arg_set_min_rate,
		     gdas_vibrato_arg_get_min_rate,
		     NULL, "0,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "max_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdas_vibrato_arg_set_max_rate,
		     gdas_vibrato_arg_get_max_rate,
		     NULL, "0,", GDAM_ARG_VITAL);
}

GtkType gdas_vibrato_get_type()
{
  static GtkType vibrato_type = 0;
  if (!vibrato_type)
    {
      static const GtkTypeInfo vibrato_info = {
	"GdasVibrato",
	sizeof (GdasVibrato),
	sizeof (GdasVibratoClass),
	(GtkClassInitFunc) gdas_vibrato_class_init,
	(GtkObjectInitFunc) gdas_vibrato_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_FILTER;
      parent_class = gtk_type_class (parent);
      vibrato_type = gtk_type_unique (parent, &vibrato_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return vibrato_type;
}

static gboolean gdas_vibrato_detach_child(GdasSource* source, GdasSource* sub)
{
  GdasVibrato *vibrato_filter = GDAS_VIBRATO (source);
  GdasSource *vibrato;

  g_return_val_if_fail (vibrato_filter != NULL, FALSE);

  vibrato = vibrato_filter->vibrato;

  if (vibrato != sub)
    {
      return (*GDAS_SOURCE_CLASS (parent_class)->detach_child) (source, sub);
    }

  if (vibrato == NULL)
    return FALSE;

  vibrato_filter->vibrato = NULL;
  vibrato->parent = NULL;
  gtk_object_unref (GTK_OBJECT (vibrato));

  return TRUE;
}
void gdas_vibrato_set_vibrato(GdasVibrato* vibrato, GdasSource* sub)
{
  g_return_if_fail ((GdasSource *) vibrato != sub);

  if (sub)
    g_return_if_fail (!gdas_source_is_ancestor (GDAS_SOURCE (vibrato), sub));

  /* sub->ref_count ++ */
  if (sub)
    gtk_object_ref (GTK_OBJECT (sub));

  if (vibrato->vibrato)
    gdas_source_detach_child (GDAS_SOURCE (vibrato), vibrato->vibrato);
  if (!sub)
    return;

  /* sub->ref_count ++ */
  gdas_source_orphan (sub);	/* increases sub's ref cnt */
  gtk_object_sink (GTK_OBJECT (sub));
  vibrato->vibrato = sub;
  sub->parent = GDAS_SOURCE (vibrato);

  /* sub->ref_count -- */
  gtk_object_unref (GTK_OBJECT (sub));
}
static void        gdas_vibrato_get_subsources  (GdasSource*      source,
                                                 int*             num_subs,
						 GdasSource***    subs)
{
  GdasVibrato *vibrato = GDAS_VIBRATO (source);
  GdasSource *sub_vibrato = vibrato->vibrato;
  GdasSource *sub_filtered = GDAS_FILTER (vibrato)->filtered;
  int i;
  *num_subs = (sub_filtered != NULL ? 1 : 0) + (sub_vibrato != NULL ? 1 : 0);
  i = 0;
  *subs = g_new (GdasSource *, *num_subs);
  if (sub_vibrato != NULL)
    (*subs)[i++] = sub_vibrato;
  if (sub_filtered != NULL)
    (*subs)[i++] = sub_filtered;
}

static void gdas_vibrato_destroy(GtkObject* object)
{
  GdasVibrato *vibrato = GDAS_VIBRATO (object);
  if (vibrato->vibrato != NULL)
    {
      GdasSource *subsource = vibrato->vibrato;
      vibrato->vibrato = NULL;
      subsource->parent = NULL;
      gtk_object_unref (GTK_OBJECT (subsource));
    }
  gdam_buffer_destruct (&vibrato->filtered_buffer);
  gdam_buffer_destruct (&vibrato->vibrato_buffer);
  (*parent_class->destroy) (object);
}
