#include "gdasxmmsinput.h"
#include <gdam/common/gdamwaitpid.h>
#include <gdam/common/gdamarg.h>
#include <gdam/common/gdamfullduplexpipe.h>
#include <gdam/common/gdamsearchpath.h>
#include <gdam/common/gdamutil.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

/* Priority of communication with the xmms-helper. */
#define XMMS_PRIORITY     		(G_PRIORITY_DEFAULT)

/* Milliseconds to wait for the TERM signal to take effect,
 * before using for the KILL signal.
 */
#define XMMS_INPUT_KILL_MILLIS		2000

static GtkObjectClass *parent_class = NULL;

typedef struct _GdasXmmsFormat GdasXmmsFormat;
struct _GdasXmmsFormat
{
  int num_bits;
  int is_big_endian;
  int is_signed;
  int rate;
  int num_channels;
};

/* GSourceFuncs for handing the file-descriptor. */
gboolean        gdas_xmms_input_prepare         (gpointer        source_data, 
			                         GTimeVal       *current_time,
			                         gint           *timeout,
			                         gpointer        user_data);
gboolean        gdas_xmms_input_check           (gpointer        source_data,
			                         GTimeVal       *current_time,
			                         gpointer        user_data);
gboolean        gdas_xmms_input_dispatch        (gpointer        source_data, 
			                         GTimeVal       *dispatch_time,
			                         gpointer        user_data);

/* GdasSource virtuals. */
static gboolean gdas_xmms_input_is_ready_to_read(GdasSource     *source);
static int      gdas_xmms_input_read            (GdasSource     *source,
                                                 GdamSample     *buffer,
				                 int             max_read);
static void     gdas_xmms_input_miss_data       (GdasSource     *source,
                                                 int             missed);

/* input processing. */
static gboolean gdas_xmms_input_process_incoming_data
                                                (GdasXmmsInput  *input,
                                                 const char     *buf,
                                                 int             length);
static gboolean gdas_xmms_input_set_format      (GdasXmmsInput  *input,
                                                 GdasXmmsFormat *format);

/* helper management functions. */
static gboolean gdas_xmms_input_launch_helper   (GdasXmmsInput  *input);
static void     gdas_xmms_input_kill_helper     (GdasXmmsInput  *input);

/* Utilities */
static void g_byte_array_append_n (GByteArray *array, const void *data, int len)
{
  int old_len;
  g_return_if_fail (array != NULL);
  old_len = array->len;
  g_byte_array_set_size (array, old_len + len);
  memcpy (array->data + old_len, data, len);
}

static gboolean skip_whitespace (const char **pstring)
{
  if (**pstring != '\0' && isspace (**pstring))
    ++(*pstring);
  else
    return FALSE;
  while (**pstring != '\0' && isspace (**pstring))
    ++(*pstring);
  return TRUE;
}

static void xmms_input_warning (GdasXmmsInput       *input,
                                const char          *format,
				...) G_GNUC_PRINTF (2, 3);

static void xmms_input_warning (GdasXmmsInput       *input,
                                const char          *format,
				...)
{
  va_list args;
  char *str;
  va_start (args, format);
  str = g_strdup_vprintf (format, args);
  va_end (args);

  if (input->plugin_type != NULL)
    g_warning ("XmmsInput: %s: %s", input->plugin_type, str);
  else
    g_warning ("XmmsInput: %s", str);
}

static void gdas_xmms_sample_buffer_free(GdasXmmsSampleBuffer *buffer)
{
  gdam_buffer_destruct (&buffer->buffer);
  g_free (buffer);
}

/* parse a whitespace-delimited int, swallowing any preceding
 * and trailing whitespace;  ints must be followed by whitespace.
 */
static gboolean parse_int (const char **pstring, int *rv)
{
  const char *end_string;
  skip_whitespace (pstring);

  /* cast to support lame libc's */
  *rv = (int) strtol ((char *) *pstring, (char **) &end_string, 10);

  if (end_string == *pstring)
    return FALSE;

  //g_message ("parsed... %s [%d]", *pstring, *end_string);
  *pstring = end_string;
  if (**pstring == '\0')
    return TRUE;

  if (!skip_whitespace (pstring))
    return FALSE;

  return TRUE;
}

static gboolean parse_int64 (const char **pstring, gint64 *rv)
{
  const char *end_string;
  skip_whitespace (pstring);

  /* cast to support lame libc's */
  *rv = (int) gdam_strtoll ((char *) *pstring, (char **) &end_string, 10);

  if (end_string == *pstring)
    return FALSE;

  //g_message ("parsed... %s [%d]", *pstring, *end_string);
  *pstring = end_string;
  if (**pstring == '\0')
    return TRUE;

  if (!skip_whitespace (pstring))
    return FALSE;

  return TRUE;
}

static GSourceFuncs xmms_source_funcs =
{
  gdas_xmms_input_prepare,
  gdas_xmms_input_check,
  gdas_xmms_input_dispatch,
  NULL
};

/*=====================================================================*
 *                            GSource Funcs                            *
 *=====================================================================*/
gboolean        gdas_xmms_input_prepare         (gpointer  source_data, 
			                         GTimeVal *current_time,
			                         gint     *timeout,
			                         gpointer  user_data)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (user_data);
  input->poll_fd.events = 0;

  if (input->outgoing.size > 0)
    input->poll_fd.events |= G_IO_OUT;
  if (input->current.buffer.size < input->buffer_size)
    input->poll_fd.events |= G_IO_IN;
  
  return FALSE;
}

gboolean        gdas_xmms_input_check           (gpointer  source_data,
			                         GTimeVal *current_time,
			                         gpointer  user_data)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (user_data);
  return (input->poll_fd.revents != 0);
}

gboolean        gdas_xmms_input_dispatch        (gpointer  source_data, 
			                         GTimeVal *dispatch_time,
			                         gpointer  user_data)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (user_data);
  gushort revents = input->poll_fd.revents;
  if ((revents & G_IO_IN) == G_IO_IN)
    {
      char buf[4096];
      int num_read;
      num_read = read (input->fd, buf, sizeof (buf));
      if (num_read < 0)
        {
          if (errno == EINTR || errno == EWOULDBLOCK)
            ;
          else
            {
              xmms_input_warning (input,
	                         "error reading from fd %d (in xmmsinput): %s",
                                  input->fd, strerror (errno));
              goto error;
            }
        }
      if (!gdas_xmms_input_process_incoming_data (input, buf, num_read))
        {
          xmms_input_warning (input, "protocol error from xmmsinput");
          goto error;
        }
    }

  if ((revents & G_IO_OUT) == G_IO_OUT)
    {
      gboolean rv;
      rv = gdam_buffer_write_one (&input->outgoing, input->fd);
      if (!rv)
        {
	  xmms_input_warning (input, "writing to fd %d (GdasXmmsInput) failed",
                              input->fd);
          goto error;
	}
    }

  /* XXX: could test if we need to be called again, and goto stop_dispatching
   *      if we have nothing more to do for now...
   *      (We could do this quite efficiently by testing poll_fd.events == 0.)
   */
  return TRUE;

error:
  close (input->fd);
  input->fd = -1;
stop_dispatching:
  input->gsource_id = 0;
  g_main_remove_poll (&input->poll_fd);
  return FALSE;
}

/*=====================================================================*
 *                           Protocol Parsing                          *
 *=====================================================================*/

/*---------------------------------------------------------------------
 *                  Specific Input Commands
 */
typedef gboolean   (*GdasXmmsCommandHandler)     (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder);

/* Process the `xmms open()' virtual when it is given to
 * the underlying Xmms OutputPlugin.  (We build this internally
 * within helper/gdamxmmsinput.c)
 *
 * Syntax:   OPEN num_bits is_big_endian is_signed rate num_channels
 */
static gboolean   gdas_xmms_input_process_OPEN   (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  const char *end;
  const char *start = remainder;
  GdasXmmsFormat format;
  if (input->virtual_device_open)
    xmms_input_warning (input, "multiply opening virtual output device");
  if (!parse_int (&remainder, &format.num_bits)
   || !parse_int (&remainder, &format.is_big_endian)
   || !parse_int (&remainder, &format.is_signed)
   || !parse_int (&remainder, &format.rate)
   || !parse_int (&remainder, &format.num_channels))
    {
      xmms_input_warning (input, "error parsing the five ints from %s", start);
      return FALSE;
    }

  if (!gdas_xmms_input_set_format (input, &format))
    {
      xmms_input_warning (input,
	"error setting format: %d bits, %s-endian, %s, %dhz, %d channels",
	format.num_bits, (format.is_big_endian ? "big" : "little"),
	(format.is_signed ? "signed" : "unsigned"),
        format.rate, format.num_channels);
      return FALSE;
    }

  input->virtual_device_open = TRUE;

  return TRUE;
}

/* Process a chunk of audio data.
 *
 * Syntax:  AUDIO num_samples sample_size_in_bytes
 */
static gboolean   gdas_xmms_input_process_AUDIO  (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  gint64 position;
  int num_samples, total_sample_size;

  if (!parse_int64 (&remainder, &input->sample_bytes_start_pos)
   || !parse_int (&remainder, &num_samples)
   || !parse_int (&remainder, &total_sample_size))
    return FALSE;
  
  input->num_sample_bytes_to_read = total_sample_size * num_samples;
  input->raw_samples = g_malloc (input->num_sample_bytes_to_read);
  return TRUE;
}

/* Close the audio device. */
static gboolean   gdas_xmms_input_process_CLOSE  (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  if (!input->virtual_device_open)
    xmms_input_warning (input, "closing unopened virtual output device");
  input->virtual_device_open = FALSE;
  return TRUE;
}

/* Flush audio data to the device.  (ignored) */
static gboolean   gdas_xmms_input_process_FLUSH  (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  return TRUE;
}

/* Information */
static gboolean   gdas_xmms_input_process_INFO   (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  g_message ("xmms: info: %s", remainder);
  return TRUE;
}

/* Unpause the output: this shouldn't happen. */
static gboolean   gdas_xmms_input_process_INFOTEXT(GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  g_message ("xmms: infotext: %s", remainder);
  return TRUE;
}

/* Pause the output: this shouldn't happen. */
static gboolean   gdas_xmms_input_process_PAUSE  (GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  xmms_input_warning (input, "didn't expect PAUSE to be called");
  return TRUE;
}

/* Unpause the output: this shouldn't happen. */
static gboolean   gdas_xmms_input_process_UNPAUSE(GdasXmmsInput   *input, 
                                                  const char      *name,
                                                  const char      *remainder)
{
  xmms_input_warning (input, "didn't expect UNPAUSE to be called");
  return TRUE;
}

/* XXX: threadsafety!! */
static GdasXmmsCommandHandler lookup_handler (const char *command)
{
  static GHashTable *table = NULL;
  if (table == NULL)
    {
      table = g_hash_table_new (g_str_hash, g_str_equal);
      g_hash_table_insert (table, "OPEN", gdas_xmms_input_process_OPEN);
      g_hash_table_insert (table, "AUDIO", gdas_xmms_input_process_AUDIO);
      g_hash_table_insert (table, "CLOSE", gdas_xmms_input_process_CLOSE);
      g_hash_table_insert (table, "FLUSH", gdas_xmms_input_process_FLUSH);
      g_hash_table_insert (table, "PAUSE", gdas_xmms_input_process_PAUSE);
      g_hash_table_insert (table, "UNPAUSE", gdas_xmms_input_process_UNPAUSE);
      g_hash_table_insert (table, "INFO", gdas_xmms_input_process_INFO);
      g_hash_table_insert (table, "INFOTEXT", gdas_xmms_input_process_INFOTEXT);
    }
  return (GdasXmmsCommandHandler) g_hash_table_lookup (table, command);
}

static gboolean gdas_xmms_input_process_command      (GdasXmmsInput  *input,
                                                      const char     *command)
{
  char command_name[256];
  const char *at;
  GdasXmmsCommandHandler handler;
  for (at = command; *at != '\0' && !isspace (*at); at++)
    ;
  if (at == command)
    {
      xmms_input_warning (input, "command started with a SPACE or had no name");
      return FALSE;
    }
  if (at - command > sizeof (command_name) - 1)
    {
      xmms_input_warning (input, "command name was too long");
      return FALSE;
    }
  memcpy (command_name, command, at - command);
  command_name[at - command] = '\0';

  for (  ; *at != '\0' && isspace (*at); at++)
    ;

  handler = lookup_handler (command_name);
  if (handler == NULL)
    {
      xmms_input_warning (input, "no handler for `%s' found", command_name);
      return FALSE;
    }
  
  return (*handler)(input, command_name, at);
}

/*---------------------------------------------------------------------
 *                  General input parsing
 */
static GdasXmmsSampleBuffer* find_buffer (GdasXmmsInput    *input,
                                          gint64            pos)
{
  GSList *tmp_list;
  GSList *list;
  GdasXmmsSampleBuffer *rv = NULL;
  int sample_size;
  g_return_val_if_fail (input->total_sample_size > 0, FALSE);
  sample_size = input->total_sample_size;
  tmp_list = g_slist_prepend (input->old_buffers, &input->current);

  for (list = tmp_list; list != NULL; list = list->next)
    {
      GdasXmmsSampleBuffer *buffer = list->data;
      gint64 end_pos;
      end_pos = buffer->start + buffer->buffer.size / sample_size;
      if (end_pos == pos)
        {
          rv = buffer;
          break;
        }
    }
  g_slist_free_1 (tmp_list);

  g_return_val_if_fail (rv != NULL, NULL);
  return rv;
}

static gboolean gdas_xmms_input_process_incoming_data(GdasXmmsInput  *input,
                                                      const char     *buf,
                                                      int             length)
{
  while (length > 0)
    {
      GByteArray *incoming = input->incoming_control;
      if (incoming == NULL)
        {
	  incoming = g_byte_array_new ();
	  input->incoming_control = incoming;
	}
      if (input->num_sample_bytes_to_read == 0)
        {
          /* Process a command. */
          char *newline;
          int cut_len;
          char *command;
          newline = memchr (buf, '\n', length);
          if (newline == NULL)
            {
              int old_len = incoming->len;
              g_byte_array_append_n (incoming, buf, length);
              return TRUE;
            }
          cut_len = newline + 1 - buf;
          g_byte_array_append_n (incoming, buf, cut_len);
          incoming->data[incoming->len - 1] = '\0';
          command = (char *) incoming->data;
          g_byte_array_free (incoming, FALSE);
          input->incoming_control = g_byte_array_new ();
          if (!gdas_xmms_input_process_command (input, command))
            {
              xmms_input_warning (input, "error parsing protocol: %s", command);
              g_free (command);
              return FALSE;
            }
          g_free (command);
          buf += cut_len;
          length -= cut_len;
        }
      else
        {
          /* Process audio data. */
          int to_copy = length;
          char *raw_sample_start;
          if (to_copy > input->num_sample_bytes_to_read)
            to_copy = input->num_sample_bytes_to_read;
          raw_sample_start = (char *) (input->raw_samples)
                           + input->sample_byte_index;
          memcpy (raw_sample_start, buf, to_copy);
          input->num_sample_bytes_to_read -= to_copy;
          input->sample_byte_index += to_copy;
	  g_return_val_if_fail (input->total_sample_size > 0, FALSE);
          if (input->num_sample_bytes_to_read == 0)
            {
              GdasXmmsSampleBuffer *buf;
              buf = find_buffer (input, input->sample_bytes_start_pos);
              if (buf == NULL)
                xmms_input_warning (input,
                                    "couldn't find buffer ending at %lld",
                                    input->sample_bytes_start_pos);
              else
                gdam_buffer_append_alloced (&buf->buffer, input->raw_samples,
                                            input->sample_byte_index,
                                            input->raw_samples, g_free);
              input->raw_samples = NULL;
              input->sample_byte_index = 0;

	      if (buf == &input->current)
	        {
		  if (buf->start < input->position)
		    {
		      /* drain samples, since we are behind. */
		      gint64 delta = input->position - buf->start;
		      int to_drain = (int) delta;
		      g_return_val_if_fail (delta == (gint64) to_drain, FALSE);
		      to_drain *= input->total_sample_size;
		      to_drain = gdam_buffer_discard (&buf->buffer, to_drain);
		      to_drain /= input->total_sample_size;
		      buf->start += to_drain;
		    }
		  
		  /* clear out other buffers in `old_buffers' */
		  if (input->old_buffers != NULL)
		    {
		      g_slist_foreach (input->old_buffers,
		                       (GFunc) gdas_xmms_sample_buffer_free,
				       NULL);
		      g_slist_free (input->old_buffers);
		      input->old_buffers = NULL;
		    }
		}
            }
          buf += to_copy;
          length -= to_copy;
        }
    }
  return TRUE;
}

/*---------------------------------------------------------------------
 *                         Changing formats.
 */
static gboolean gdas_xmms_input_set_format      (GdasXmmsInput  *input,
                                                 GdasXmmsFormat *format)
{
  g_return_val_if_fail (format->num_bits == 8 || format->num_bits == 16, FALSE);
  g_return_val_if_fail (format->rate == 44100, FALSE);
  g_return_val_if_fail (format->num_channels <= 4, FALSE);
  g_return_val_if_fail (format->num_channels >= 1, FALSE);

  input->sample_bits = format->num_bits;
  input->is_big_endian = format->is_big_endian;
  input->is_signed = format->is_signed;
  input->num_channels = format->num_channels;
  input->total_sample_size = format->num_bits * format->num_channels / 8;

  input->sample_converter = gdam_sample_converter_get
				    (format->num_bits, format->is_signed,
				     format->is_big_endian,
				     GDAM_SAMPLE_BITS, TRUE,
				     G_BYTE_ORDER == G_BIG_ENDIAN);
  input->channel_converter = gdam_sample_converter_get_channel_changer
				  (format->num_channels, 2);

  /* XXX: should we flush the buffers?
   *      That will require resetting ``current'' etc
   */

  return TRUE;
}

/*=====================================================================*
 *                         GdasSource Virtuals                         *
 *=====================================================================*/

static gboolean gdas_xmms_input_is_ready_to_read(GdasSource *source)
{
  GdasXmmsInput* xmms_input;
  xmms_input = GDAS_XMMS_INPUT (source);
  if (xmms_input->fd < 0)
    return TRUE;
  return xmms_input->current.buffer.size > 0
      || xmms_input->old_buffers != NULL;
}

static gboolean gdas_xmms_input_add_gsource (GdasXmmsInput *input)
{
  g_return_val_if_fail (input->gsource_id == 0, FALSE);
  g_return_val_if_fail (input->fd >= 0, FALSE);

  g_main_add_poll (&input->poll_fd, XMMS_PRIORITY);
  input->gsource_id = g_source_add (XMMS_PRIORITY,
                                    FALSE,
				    &xmms_source_funcs,
				    NULL,
				    input,
				    NULL);
  return TRUE;
}

static int  gdas_xmms_input_read (GdasSource     *source,
                                  GdamSample     *samples,
				  int             max_read)
{
  int num_read;
  GdasXmmsInput* xmms_input;
  GdasXmmsSampleBuffer *buffer;
  gpointer tmp_buf;

  xmms_input = GDAS_XMMS_INPUT (source);
  buffer = &xmms_input->current;
  g_return_val_if_fail (xmms_input->total_sample_size > 0, -1);
  if (buffer->buffer.size == 0)
    {
      GSList *list;
      for (list = xmms_input->old_buffers; list != NULL; list = list->next)
        {
	  buffer = list->data;
	  if (buffer->buffer.size > 0)
	    break;
	}
    }
  if (buffer == NULL)
    {
      g_warning ("gdas_xmms_input_read: couldn't find nonempty buffer");
      g_warning ("fd=%d", xmms_input->fd);
      return 0;
    }
  num_read = buffer->buffer.size / xmms_input->total_sample_size;
  if (num_read > max_read)
    num_read = max_read;
  if (xmms_input->total_sample_size < sizeof (GdamSample)
   && xmms_input->num_channels <= 2)
    tmp_buf = samples;
  else
    {
      int required_1 = xmms_input->num_channels * sizeof (GdamSampleValue);
      int required_2 = xmms_input->total_sample_size;
      int tmp_size = MAX (required_1, required_2);
      tmp_buf = alloca (num_read * tmp_size);
    }

  gdam_buffer_read (&xmms_input->current.buffer, tmp_buf,
                    num_read * xmms_input->total_sample_size);
  buffer->start += num_read;

  (*xmms_input->sample_converter) (tmp_buf, tmp_buf,
                                   num_read * xmms_input->num_channels);
  (*xmms_input->channel_converter) (samples, tmp_buf, num_read);

  if (xmms_input->gsource_id == 0
   && xmms_input->current.buffer.size < xmms_input->buffer_size
   && xmms_input->fd >= 0)
    {
      if (!gdas_xmms_input_add_gsource (xmms_input))
        return -1;
    }

  return num_read;
}
static void     gdas_xmms_input_miss_data       (GdasSource     *source,
                                                 int             missed)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (source);
  if (input->current.buffer.size > 0)
    {
      int to_drain = missed;
      int max_drain = input->current.buffer.size / input->total_sample_size;
      if (to_drain > max_drain)
        to_drain = max_drain;
      gdam_buffer_discard (&input->current.buffer,
                           to_drain * input->total_sample_size);
      input->current.start += to_drain;
    }
  input->position += missed;
}

/*=====================================================================*
 *                     helper management functions                     *
 *=====================================================================*/
static void     gdas_xmms_input_kill_helper     (GdasXmmsInput  *input)
{
  g_return_if_fail (input->xmms_helper_pid != 0);
  gdam_wait_pid_kill_pid (input->xmms_helper_pid, XMMS_INPUT_KILL_MILLIS);
  input->xmms_helper_pid = 0;

  if (input->gsource_id != 0)
    {
      g_source_remove (input->gsource_id);
      input->gsource_id = 0;
      g_main_remove_poll (&input->poll_fd);
    }
  
  if (input->fd >= 0)
    {
      close (input->fd);
      input->fd = -1;
      input->poll_fd.fd = -1;
    }
}

static gboolean gdas_xmms_input_launch_helper   (GdasXmmsInput  *input)
{
  int fds[2];
  const char *err;
  const char *plugin;
  char *free_plugin = NULL;
  char fd_as_string[256];
  g_return_val_if_fail (input->plugin_name != NULL, FALSE);

  if (input->plugin_name[0] == '/')
    plugin = input->plugin_name;
  else
    {
      free_plugin = gdam_search_path (input->plugin_name,
                        "/usr/local/lib/xmms/Input:"
                        "/usr/lib/xmms/Input:",
			GDAM_SEARCH_PATH_FIND_FILE);
      plugin = free_plugin;
    }

  if (plugin == NULL)
    {
      g_warning ("couldn't find plugin %s", input->plugin_name);
      return FALSE;
    }

  err = gdam_full_duplex_pipe (fds);
  if (err != NULL)
    {
      xmms_input_warning (input, "error making pipe: %s", err);
      g_free (free_plugin);
      return FALSE;
    }
  
retry_fork:
  input->xmms_helper_pid = fork ();
  input->fd = input->poll_fd.fd = fds[1];

  if (input->xmms_helper_pid < 0)
    {
      if (errno == EINTR)
        goto retry_fork;
      if (errno == EAGAIN)
        {
	  /* ugh, but, if you get EAGAIN you're probably screwed anyways... */
	  sleep (1);
	  goto retry_fork;
	}
      g_warning ("gdas_xmms_input_launch_helper: cannot fork: %s",
                 strerror (errno));
      g_free (free_plugin);
      close (fds[0]);
      return FALSE;
    }

  if (input->xmms_helper_pid > 0)
    {
      /* Parent process returns successfully. */
      g_free (free_plugin);
      close (fds[0]);
      return TRUE;
    }

  /* Child process: exec gdamxmmsinputhelper */
  /* fds[0] becomes standard input and standard output. */
  close (fds[1]);

  /* don't close this file descriptor on exec */
  fcntl (fds[0], F_SETFD, 0);

  g_snprintf (fd_as_string, sizeof (fd_as_string), "%d", fds[0]);

  /* exec gdamxmmsinputhelper */
  /* for running within the gdam source tree */
  execl ("helper/gdamxmmsinputhelper",
         "helper/gdamxmmsinputhelper",
	 plugin,
	 "--fd",
	 fd_as_string,
	 NULL);
  execl ("server/helper/gdamxmmsinputhelper",
         "server/helper/gdamxmmsinputhelper",
	 "--fd",
	 fd_as_string,
	 plugin,
	 NULL);

  /* for installed versions */
  execlp ("gdamxmmsinputhelper",
          "gdamxmmsinputhelper",
	  plugin,
	  "--fd",
	  fd_as_string,
	  NULL);

  _exit (127);
}

/*=====================================================================*
 *                  client-side interface "functions"                  *
 *=====================================================================*/

static gboolean gdas_xmms_input_run_command(GdasXmmsInput    *input,
                                            const char       *format,
                                            ...)
{
  va_list args;
  char *append;
  va_start (args, format);
  append = g_strdup_vprintf (format, args);
  va_end (args);

  gdam_buffer_append_string (&input->outgoing, append);
  g_free (append);

  if (input->gsource_id == 0 && input->fd >= 0)
    if (!gdas_xmms_input_add_gsource (input))
      {
        g_warning ("error starting the outgoing buffer thread");
        return FALSE;
      }
  return TRUE;
}

static void gdas_xmms_input_func_load(GtkObject      *object,
                                      int             num_args_in,
                                      GdamArg        *args_in,
                                      int            *num_args_out,
                                      GdamArg       **args_out)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (object);
  const char *plugin_name = NULL;
  gboolean rv = FALSE;
  int i;
  if (input->fd >= 0)
    {
      xmms_input_warning (input, "already loaded a xmms input plugin type");
      return;
    }
  
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (arg->type == GDAM_ARG_TYPE_STRING
       && strcmp (arg->name, "plugin_name") == 0)
        {
          plugin_name = GDAM_ARG_STRING (arg);
          continue;
        }

      g_warning ("unknown option (or of invalid type): %s", arg->name);
    }

  if (plugin_name == NULL)
    {
      g_warning ("no plugin_name specified to `load'");
      rv = FALSE;
    }

  /* Kill old XMMS helper. */

  if (input->xmms_helper_pid != 0)
    gdas_xmms_input_kill_helper (input);

  /* Start new XMMS helper. */
  g_free (input->plugin_name);
  input->plugin_name = g_strdup (plugin_name);
  rv = gdas_xmms_input_launch_helper (input);
  
end:
  *num_args_out = 1;
  *args_out = gdam_arg_new_from_valist (1, "success", GDAM_ARG_TYPE_INT, rv);
  return;
}
  
static void gdas_xmms_input_func_play(GtkObject      *object,
                                      int             num_args_in,
                                      GdamArg        *args_in,
                                      int            *num_args_out,
                                      GdamArg       **args_out)
{
  const char *filename = NULL;
  GdasXmmsInput *input = GDAS_XMMS_INPUT (object);
  gboolean rv = FALSE;
  int i;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = args_in + i;
      if (arg->type == GDAM_ARG_TYPE_STRING
       && strcmp (arg->name, "filename") == 0)
        {
          filename = GDAM_ARG_STRING (arg);
          continue;
        }

      g_warning ("unknown option (or of invalid type): %s", arg->name);
    }

  if (filename == NULL)
    {
      g_warning ("no filename specified to `play'");
      rv = FALSE;
    }
  else
    {
      rv = gdas_xmms_input_run_command (input, "START %s\n", filename);
    }
  
end:
  *num_args_out = 1;
  *args_out = gdam_arg_new_from_valist (1, "success", GDAM_ARG_TYPE_INT, rv);
  return;
}

static void gdas_xmms_input_func_stop(GtkObject      *object,
                                      int             num_args_in,
                                      GdamArg        *args_in,
                                      int            *num_args_out,
                                      GdamArg       **args_out)
{
  GdasXmmsInput *input = GDAS_XMMS_INPUT (object);
  gboolean rv = FALSE;
  int i;
  for (i = 0; i < num_args_in; i++)
    {
      GdamArg *arg = args_in + i;
      g_warning ("unknown option to stop (or of invalid type): %s", arg->name);
    }

  rv = gdas_xmms_input_run_command (input, "STOP\n");
  
end:
  *num_args_out = 1;
  *args_out = gdam_arg_new_from_valist (1, "success", GDAM_ARG_TYPE_INT, rv);
  return;
}


/*=====================================================================*
 *                            Initialization                           *
 *=====================================================================*/
static void gdas_xmms_input_init (GdasXmmsInput* xmms_input)
{
  xmms_input->buffer_size = 1024 * 128;
  xmms_input->fd = -1;
  xmms_input->poll_fd.fd = -1;
}

static void gdas_xmms_input_class_init (GdasXmmsInputClass* xmms_input_class)
{
  GdasSourceClass *source_class = GDAS_SOURCE_CLASS (xmms_input_class);

  source_class->read = gdas_xmms_input_read;
  source_class->is_ready_to_read = gdas_xmms_input_is_ready_to_read;
  source_class->miss_data = gdas_xmms_input_miss_data;

  gdam_arg_function_register (GTK_OBJECT_CLASS (xmms_input_class),
                              "load",
                              gdas_xmms_input_func_load);
  gdam_arg_function_register (GTK_OBJECT_CLASS (xmms_input_class),
                              "play",
                              gdas_xmms_input_func_play);
  gdam_arg_function_register (GTK_OBJECT_CLASS (xmms_input_class),
                              "stop",
                              gdas_xmms_input_func_stop);
}

GtkType gdas_xmms_input_get_type()
{
  static GtkType xmms_input_type = 0;
  if (!xmms_input_type) {
    static const GtkTypeInfo xmms_input_info =
    {
      "GdasXmmsInput",
      sizeof(GdasXmmsInput),
      sizeof(GdasXmmsInputClass),
      (GtkClassInitFunc) gdas_xmms_input_class_init,
      (GtkObjectInitFunc) gdas_xmms_input_init,
      /* reserved_1 */ NULL,
      /* reserved_2 */ NULL,
      (GtkClassInitFunc) NULL
    };
    GtkType parent = GDAS_TYPE_SOURCE;
    xmms_input_type = gtk_type_unique (parent, &xmms_input_info);
    parent_class = gtk_type_class (parent);
  }
  return xmms_input_type;
}
