#!/usr/bin/perl

# generate table of offsets for 16 bit signal limiting

# limit(x) = M tanh(x/M)
# where M = 32768
# table is symmetric: limit(-x) = -limit(x)
# for x < M, dx = x - limit(x) grows slower than limit(x), so use table of
#   x vs. dx (array index is dx, value is x)
# for x > M, limit(x) grows slower than dx, so use table of x vs. limit(x)
#   (array index is limit(x)-24955, value is x)

sub blorf($) {
  my ($x) = @_;
  my $u = $x/32768.0;
  my $a = exp($u);
  my $b = exp(-$u);
  return 32768.0 * ($a - $b) / ($a + $b);
}

my @low;
my $last = -1;

for ($in = 0; $in < 32768; ++$in) {
  my $lim = blorf($in);
  my $delta = int($in - $lim);
  if ($delta != $last) {
    push(@low, $in);
    $last = $delta;
  }
}
print "static int limit_low_s16[", scalar(@low), "] = {\n";
foreach (@low) {
  print "  $_,\n";
}
print "};\n";


my @high;
$last = -1;

for ( ; ; ++$in) {
  my $lim = blorf($in);
  my $out = int($lim);
  last if $out >= 32767;
  if ($out != $last) {
    push(@high, $in);
    $last = $out;
  }
}
print "static int limit_high_s16[", scalar(@high), "] = {\n";
foreach (@high) {
  print "  $_,\n";
}
print "};\n";

exit(0);

print <<"END";

/* limit a signed 16-bit signal to avoid overs
   input is an int which should be remapped into the 16-bit range */

signed short int limit_s16(int in) {
  int sign;

  if (in < 0) {
    in = -in;
    sign = -1;
  } else {
    sign = 1;
  }

  if (in >= 181705) {
    return sign*32767;

  } else {
    int lo = 0;
    int hi = 7811;
    int mid;

    if (in < 32768) {
      /* use table of (input-output) vs. input for lower range */
      while (lo < hi) {
        mid = (lo + hi)/2;
        if (limit_low_s16[mid] < in)
          lo = mid+1;
        else
          hi = mid;
      }
      return (signed short) sign*(in - mid);
    } else {
      /* use table of output vs. input for upper range */
      while (lo < hi) {
        mid = (lo + hi)/2;
        if (limit_high_s16[mid] < in)
          lo = mid+1;
        else
          hi = mid;
      }
      return (signed short) sign*(mid + 24955);
    }
  }
}

END

0;
