#include <glib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#include "synthlib.h"

gdouble synth_lib_sampling_rate = 44100;

static int global_argc = 0;
static char **global_argv = NULL;
static int at_arg;

#define input_buffer_length 2048
#define output_buffer_length 8192
static gdouble output_buffer[output_buffer_length];
static guint output_buffer_data_start = 0;
static guint output_buffer_data_length = 0;

void
synth_lib_init        (int argc,
		       char **argv)
{
  static gboolean has_inited = FALSE;
  if (has_inited)
    return;
  has_inited = TRUE;
  global_argc = argc;
  global_argv = argv;
  at_arg = 1;
  g_atexit (synth_lib_flush_output);
}

#define CHECK_ARG(fname)	\
  G_STMT_START{			\
    if (at_arg >= global_argc)	\
      g_error ("too few arguments in %s [%s]", #fname, global_argv[0]); \
  }G_STMT_END


void
synth_lib_get_double        (gdouble     *val_out)
{
  CHECK_ARG (synth_lib_get_double);
  *val_out = strtod (global_argv[at_arg++], NULL);
}

void
synth_lib_get_int (gint        *val_out)
{
  CHECK_ARG (synth_lib_get_double);
  *val_out = (int) strtol (global_argv[at_arg++], NULL, 10);
}

void
synth_lib_get_int_array  (guint       *len_out,
			  gint       **int_out)
{
  char **split;
  char *arg;
  int count;
  CHECK_ARG (synth_lib_get_int_array);
  arg = global_argv[at_arg++];
  split = g_strsplit (arg, ",", 0);
  count = 0;
  while (split[count] != NULL)
    count++;
  *len_out = count;
  *int_out = g_new (gint, count);
  count = 0;
  while (split[count] != NULL)
    {
      (*int_out)[count] = (int) strtol (split[count], NULL, 10);
      count++;
    }
  g_strfreev (split);
}

void
synth_lib_get_double_array  (guint       *len_out,
			     gdouble    **val_out)
{
  /* comma-separated list of doubles */
  char **split;
  char *arg;
  int count;
  CHECK_ARG (synth_lib_get_int_array);
  arg = global_argv[at_arg++];
  split = g_strsplit (arg, ",", 0);
  count = 0;
  while (split[count] != NULL)
    count++;
  *len_out = count;
  *val_out = g_new (gdouble, count);
  count = 0;
  while (split[count] != NULL)
    {
      (*val_out)[count] = strtod (split[count], NULL);
      count++;
    }
  g_strfreev (split);
}

void
synth_lib_get_string_array  (guint       *len_out,
			     gchar     ***val_out)
{
  guint i, count;
  CHECK_ARG (synth_lib_get_string_array);
  count = *len_out = atoi (global_argv[at_arg++]);
  *val_out = g_new (char *, count);
  for (i = 0; i < count; i++)
    {
      CHECK_ARG (synth_lib_get_string_array);
      (*val_out)[i] = g_strdup (global_argv[at_arg++]);
    }
}

void
synth_lib_get_stream        (SynthLibStream **stream_out)
{
  int fd;
  CHECK_ARG (synth_lib_get_stream);
  fd = atoi (global_argv[at_arg++]);
  *stream_out = synth_lib_stream_new (fd);
  g_assert (*stream_out != NULL);
}

void
synth_lib_get_stream_array  (guint       *len_out,
			     SynthLibStream ***streams_out)
{
  int *fds;
  guint i;
  synth_lib_get_int_array (len_out, &fds);
  *streams_out = g_new (SynthLibStream *, *len_out);
  for (i = 0; i < *len_out; i++)
    {
      (*streams_out)[i] = synth_lib_stream_new (fds[i]);
      g_assert ((*streams_out)[i] != NULL);
    }
  g_free (fds);
}


static gboolean writen(int fd, const void* out, int out_len)
{
  while (out_len > 0)
    {
      int w = write (fd, out, out_len);
      if (w < 0)
	{
	  if (errno == EINTR || errno == EAGAIN)
	    continue;
	  return FALSE;
	}
      if (w == 0)
	return FALSE;
      out = (const char *) out + w;
      out_len -= w;
    }
  return TRUE;
}


void
synth_lib_output (gdouble out)
{
  if (output_buffer_data_length > 128)
    {
      int len = output_buffer_data_length;
      if (len + output_buffer_data_start > output_buffer_length)
	len = output_buffer_length - output_buffer_data_start;
      if (!writen (STDOUT_FILENO, output_buffer + output_buffer_data_start,
	      len * sizeof (gdouble)))
	{
	  exit (1);
	}

      output_buffer_data_start += len;
      output_buffer_data_length -= len;
      if (output_buffer_data_start == output_buffer_length)
	output_buffer_data_start = 0;
    }
  if (output_buffer_data_length < output_buffer_length)
    {
      int at = output_buffer_data_length + output_buffer_data_start;
      if (at >= output_buffer_length)
	at -= output_buffer_length;
      output_buffer[at] = out;
      ++output_buffer_data_length;
    }
}

void
synth_lib_flush_output ()
{
  guint end = output_buffer_data_start + output_buffer_data_length;
  if (end > output_buffer_length)
    {
      if (!writen (STDOUT_FILENO,
		   output_buffer + output_buffer_data_start,
		   (output_buffer_length - output_buffer_data_start) * sizeof (gdouble)))
	{
	  exit (1);
	}
      output_buffer_data_start = 0;
      output_buffer_data_length -= (end - output_buffer_length);
    }
  if (!writen (STDOUT_FILENO,
	       output_buffer + output_buffer_data_start,
	       (output_buffer_data_length) * sizeof (gdouble)))
    exit (1);
  output_buffer_data_start = 0;
  output_buffer_data_length = 0;
}

/* --- stream (input stream) implementation --- */

#define input_buffer_bytes	(sizeof(gdouble) * input_buffer_length)
struct _SynthLibStream
{
  int fd;
  guint buffer_start;
  guint buffer_data_len;
  guint ended : 1;
  char buffer[input_buffer_bytes];

};

SynthLibStream *
synth_lib_stream_new   (int fd)
{
  SynthLibStream *stream = g_new (SynthLibStream, 1);
  stream->fd = fd;
  stream->buffer_start = 0;
  stream->buffer_data_len = 0;
  stream->ended = 0;
  return stream;
}

void
synth_lib_stream_close (SynthLibStream *stream)
{
  close (stream->fd);
  g_free (stream);
}

gboolean       
synth_lib_stream_get   (SynthLibStream *stream,
			gdouble        *out)
{
  while (stream->buffer_data_len < sizeof (gdouble))
    {
      guint start_index = stream->buffer_data_len + stream->buffer_start;
      gint max_read;
      int fd = stream->fd;
      int got;
      if (start_index >= input_buffer_bytes)
	start_index -= input_buffer_bytes;
      max_read = input_buffer_bytes - start_index;
      if (stream->buffer_data_len + max_read > input_buffer_bytes)
	max_read = input_buffer_bytes - stream->buffer_data_len;
      got = read (fd, stream->buffer + start_index, max_read);
      if (got < 0)
	{
	  if (errno == EINTR)
	    continue;
	  fprintf (stderr, "synth_lib_stream_get: %s.\n", g_strerror (errno));
	  stream->ended = 1;
	  return FALSE;
	}
      if (got == 0)
	{
	  stream->ended = 1;
	  /*fprintf (stderr, "synth_lib_stream_get: end-of-file.\n", g_strerror (errno));*/
	  return FALSE;
	}
      stream->buffer_data_len += got;
    }

  memcpy (out, stream->buffer + stream->buffer_start, sizeof (gdouble));
  stream->buffer_start += sizeof (gdouble);
  if (stream->buffer_start == input_buffer_bytes)
    stream->buffer_start = 0;
  stream->buffer_data_len -= sizeof (gdouble);
  return TRUE;
}

void
synth_lib_error  (const char *format,
		  ...)
{
  va_list args;
  va_start (args, format);
  fprintf (stderr, "ERROR: %s: ", global_argv[0]);
  vfprintf (stderr, format, args);
  va_end (args);
  fprintf (stderr, "!\n");
  exit (10);
}
