/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "testsource.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

static void test_source_init(TestSource* source)
{
}

static gboolean test_source_is_ready_to_read(GdasSource   *source)
{
  return TRUE;
}

static int test_source_read     (GdasSource     *source,
                                 GdamSample     *buffer,
				 int             max_read)
{
  TestSource * test_source;
  int rv = 0;
  test_source = TEST_SOURCE (source);
  while (rv < max_read && test_source->current_block != NULL)
    {
      GArray *samples = test_source->current_block->data;
      if (samples->len > test_source->block_index)
        {
	  int to_read = samples->len - test_source->block_index;
	  if (to_read > max_read)
	    to_read = max_read;
	  memcpy (buffer + rv,
	          &g_array_index(samples, GdamSample, test_source->block_index),
		  to_read * sizeof (GdamSample));
	  test_source->block_index += rv;
	}
      if (test_source->block_index >= samples->len)
	{
	  test_source->current_block = test_source->current_block->next;
	  test_source->block_index = 0;
	}
    }
  return rv;
}

static void test_source_class_init(GdasSourceClass* source_class)
{
  source_class->read = test_source_read;
  source_class->is_ready_to_read = test_source_is_ready_to_read;
}

GtkType test_source_get_type()
{
  static GtkType source_type = 0;
  if (!source_type)
    {
      static const GtkTypeInfo source_info = {
	"TestSource",
	sizeof (TestSource),
	sizeof (TestSourceClass),
	(GtkClassInitFunc) test_source_class_init,
	(GtkObjectInitFunc) test_source_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAS_TYPE_SOURCE;
      source_type = gtk_type_unique (parent, &source_info);
    }
  return source_type;
}

TestSourceData* test_source_data_new      ()
{
  TestSourceData *rv;
  rv = g_new0 (TestSourceData, 1);
  rv->ref_count = 1;
  return rv;
}
void test_source_data_add_array (TestSourceData   *data,
                                 GArray           *array)
{
  if (data->first_block == NULL)
    data->first_block = data->last_block = g_slist_prepend (NULL, array);
  else
    {
      g_slist_append (data->last_block, array);
      data->last_block = data->last_block->next;
    }
}

static gpointer test_source_data_add_uninit_chunk(TestSourceData*    data,
                                                 int                num_samples)
{
  GArray *array;
  array = g_array_new (FALSE, FALSE, sizeof (GdamSample));
  g_array_set_size (array, num_samples);
  test_source_data_add_array (data, array);
  return array->data;
}

void            test_source_data_add_chunk(TestSourceData*    test_source_data,
                                           int                num_samples,
				           GdamSample*        samples)
{
  gpointer data;
  data = test_source_data_add_uninit_chunk (test_source_data, num_samples);
  memcpy (data, samples, sizeof (GdamSample) * num_samples);
}

void            test_source_data_add_mono (TestSourceData*    test_source_data,
                                           int                num_samples,
					   GdamSampleValue*   values)
{
  GdamSample *data;
  data = test_source_data_add_uninit_chunk (test_source_data, num_samples);
  while (num_samples-- > 0)
    {
      data->s[0] = data->s[1] = *values;
      values++;
      data++;
    }
}

void            test_source_data_ref      (TestSourceData*       data)
{
  g_return_if_fail (data->ref_count > 0);
  data->ref_count++;
}
void            test_source_data_unref    (TestSourceData*       data)
{
  g_return_if_fail (data->ref_count > 0);
  --(data->ref_count);
  if (data->ref_count == 0)
    {
      g_slist_foreach (data->first_block,
		       (GFunc) g_array_free, GINT_TO_POINTER (TRUE));
      g_slist_free (data->first_block);
      g_free (data);
    }
}

GdasSource*     test_source_new           (TestSourceData*    test_source_data)
{
  TestSource *rv;
  test_source_data_ref (test_source_data);
  rv = gtk_type_new (TEST_TYPE_SOURCE);
  rv->source_data = test_source_data;
  return GDAS_SOURCE (rv);
}

static gboolean read_float(const char **pdesc, gfloat *pval)
{
  char *endp;
  *pval = strtod (*pdesc, &endp);
  if (*pdesc == endp)
    {
      return FALSE;
    }
  while (*endp && isspace (*endp))
    endp++;
  if (*endp != ':' && *endp != '\0')
    return FALSE;
  if (*endp == ':')
    endp++;
  *pdesc = endp;
  return TRUE;
}

static GArray  *parse_sample_spec(const char       *spec)
{
  gboolean stereo = TRUE;
  GArray *array;
  while (*spec != ':' && *spec != '\0')
    {
      switch (*spec)
	{
	case 'm':
	  stereo = FALSE;
	  break;
	case 's':
	  stereo = TRUE;
	  break;
	default:
	  g_warning ("invalid sample data for test");
	  return NULL;
	}
      spec++;
    }
  array = g_array_new (FALSE, FALSE, sizeof (GdamSample));
  if (*spec == '\0')
    return array;
  spec++;			/* skip past colon */

  /* now parse an array of floats => GdamSamples */
  if (stereo)
    {
      gfloat left, right;
      while (read_float (&spec, &left) && read_float (&spec, &right))
	{
	  GdamSample val;
	  val.s[0] = GDAM_SAMPLE_FROM_FLOAT (left);
	  val.s[1] = GDAM_SAMPLE_FROM_FLOAT (right);
	  g_array_append_val (array, val);
	}
    }
  else
    {
      gfloat float_val;
      while (read_float (&spec, &float_val))
	{
	  GdamSample val;
	  val.s[0] = val.s[1] = GDAM_SAMPLE_FROM_FLOAT (float_val);
	  g_array_append_val (array, val);
	}
    }
  return array;
}

GdasSource     *test_source_from_spec     (const char        *spec)
{
  TestSourceData *data;
  GArray *buf;
  buf = parse_sample_spec (spec);
  if (buf == NULL)
    return NULL;
  data = test_source_data_new ();
  test_source_data_add_array (data, buf);
  return test_source_new (data);
}
