
#include <stdio.h>

#if HAVE_XMMS

#include <glib.h>
#include <gmodule.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX		108
#endif
#define DEBUGPRINT(args) g_message args

typedef struct _VisPluginCommand VisPluginCommand;
typedef struct _VisPlugin
{
        void *handle;   /* Filled in by xmms */
        char *filename; /* Filled in by xmms */
        int xmms_session; /* The session ID for attaching to the control socket 
*/
        char *description; /* The description that is shown in the preferences b
ox */
        int num_pcm_chs_wanted; /* Numbers of PCM channels wanted in the call to
 render_pcm */
        int num_freq_chs_wanted; /* Numbers of freq channels wanted in the call 
to render_freq */
        void (*init)(void); /* Called when the plugin is enabled */
        void (*cleanup)(void); /* Called when the plugin is disabled */
        void (*about)(void); /* Show the about box */
        void (*configure)(void); /* Show the configure box */
        void (*disable_plugin)(struct _VisPlugin *); /* Call this with a pointer
 to your plugin to disable the plugin */
        void (*playback_start)(void); /* Called when playback starts */
        void (*playback_stop)(void); /* Called when playback stops */
        void (*render_pcm)(gint16 pcm_data[2][512]); /* Render the PCM data, don
't do anything time consuming in here */
        void (*render_freq)(gint16 freq_data[2][256]); /* Render the freq data, 
don't do anything time consuming in here */
} VisPlugin;

typedef struct _VisPluginData VisPluginData;
#define VISPLUGIN_CMD_BUFLEN          (2 * 512 * 2 + 1024)
struct _VisPluginData {
	VisPlugin* plugin;
	int buf_amount;
	char buffer[VISPLUGIN_CMD_BUFLEN];
};

static VisPlugin* load_vis_plugin(const char* filename)
{
	GModule* module;
	VisPlugin* (*get_vplugin_info)() = NULL;
	VisPlugin* rv;
	module = g_module_open(filename, 0);
	if (module == NULL) {
		char* error;
		error = g_module_error();
		g_warning("error opening %s: %s", filename, error);
		g_free(error);
		return NULL;
	}
	if (! g_module_symbol(module,
	                      "get_vplugin_info",
			      (gpointer*) &get_vplugin_info)) {
		char* error;
		error = g_module_error();
		g_warning("error getting get_vplugin_info in %s: %s",
			filename, error);
		g_free(error);
		return NULL;
	}
	g_return_val_if_fail(get_vplugin_info != NULL, NULL);
	rv = (*get_vplugin_info)();
	DEBUGPRINT(("loaded xmms plugin: num_pcm=%d, num_fft=%d",
			rv->num_pcm_chs_wanted, rv->num_freq_chs_wanted));
	return rv;
}

struct _VisPluginCommand {
	gboolean	is_freq_data;
	int		num_channels;
	gint16		data[1024];
};

static GMemChunk* chunk = NULL;

static VisPluginCommand* vis_plugin_command_alloc()
{
	if (chunk == NULL)
		chunk = g_mem_chunk_create(VisPluginCommand,
		                           2, G_ALLOC_AND_FREE);
	return g_chunk_new(VisPluginCommand, chunk);
}
static void vis_plugin_command_recycle(VisPluginCommand* command)
{
	g_chunk_free(command, chunk);
}

/* format:
 *   offset length
 *     0     1          0=is_pcm_data, 1=is_freq_data
 *     1     1          0=is_mono, 1=is_stereo
 *     2     1          0=is_big_endian, 1=is_little_endian
 *     3     1          0 for padding 
 *
 * If is_pcm_data, followed by num_channels * 512 16-bit, signed integers.
 * If is_freq_data, followed by num_channels * 256 16-bit, signed integers
 */

static VisPluginCommand* try_command(char* buffer, int length,
                                     int* used_length)
{
	guint8 preamble[4];
	VisPluginCommand* command;
	int num_shorts;
	if (length < sizeof(preamble))
		return NULL;
	memcpy(preamble, buffer, sizeof(preamble));

	num_shorts = (preamble[0] == 0 ? 512 : 256);

	/* stereo? */
	if (preamble[1])
		num_shorts *= 2;
	if (length < sizeof(preamble) + sizeof(gint16) * num_shorts)
		return NULL;
	command = vis_plugin_command_alloc();
	memcpy(command->data, buffer + 4, num_shorts * sizeof(gint16));
	if ((preamble[2] ? G_LITTLE_ENDIAN : G_BIG_ENDIAN) != G_BYTE_ORDER) {
		/* swap the endianness. */
		guint8* at = (guint8*) command->data;
		int num = num_shorts;
		while (--num >= 0) {
			guint8 tmp = at[0];
			at[0] = at[1];
			at[1] = tmp;
		}
	}
	command->is_freq_data = (preamble[0] == 1);
	command->num_channels = preamble[1] ? 2 : 1;
	*used_length = sizeof(preamble) + sizeof(gint16) * num_shorts;
	return command;
}

#if 0
static void print_nums(gint16* vals, int num)
{
	int i;
	for (i = 0; i < num - 1; i++)
		fprintf(stderr, "%d ", vals[i]);
	if (i < num)
		fprintf(stderr, "%d", vals[i]);
}
#endif
static void     run_command(VisPlugin*           plugin,
                            VisPluginCommand*    command)
{
	if (command->is_freq_data) {
		if (plugin->render_freq == NULL)
			return;
		(*plugin->render_freq)((void*)(command->data));
	} else {
		if (plugin->render_pcm == NULL)
			return;
#if 0
		fprintf(stderr, "pcm data: [");
		print_nums((gint16*)(command->data), 512);
		fprintf(stderr, "]\n");
#endif
		(*plugin->render_pcm)((void*)(command->data));
	}
}
static  gboolean handle_server_input(GIOChannel   *source,
			             GIOCondition  condition,
			             gpointer      data)
{
	VisPluginData* plugin_data;
	VisPlugin* plugin;
	VisPluginCommand* command;
	int allowed_amount;
	int num_read;
	char* buffer;
	GIOError err;
	plugin_data = data;
	plugin = plugin_data->plugin;
	allowed_amount = VISPLUGIN_CMD_BUFLEN - plugin_data->buf_amount;
	buffer = plugin_data->buffer;

	err = g_io_channel_read(source,
				buffer + plugin_data->buf_amount,
				allowed_amount,
				&num_read);
	if (err != G_IO_ERROR_NONE) {
		g_error("error [io error %d] reading from channel", err);
	}
	if (num_read == 0) {
		gtk_main_quit();
		return FALSE;
	}
	plugin_data->buf_amount += num_read;
	for (;;) {
		int num_used;
		command = try_command(buffer,
		                      plugin_data->buf_amount,
				      &num_used);
		if (command == NULL)
			break;
		g_assert(num_used <= plugin_data->buf_amount);
		plugin_data->buf_amount -= num_used;
		memmove(buffer,
			buffer + num_used,
			plugin_data->buf_amount);
		run_command(plugin, command);
		vis_plugin_command_recycle(command);
	}
	return TRUE;
}

static  gboolean handle_control_input(GIOChannel   *source,
			              GIOCondition  condition,
			              gpointer      data)
{
	char c;
	int num_got;
	VisPluginData* plugin_data = data;
	VisPlugin* plugin = plugin_data->plugin;
	if (g_io_channel_read(source, &c, 1, &num_got) != G_IO_ERROR_NONE)
		g_error("error reading from control socket");

	if (num_got == 0) {
		g_warning("eof from control socket");
		gtk_main_quit();
		return FALSE;
	}

	switch (c) {
		case 'c': {
			if (plugin->configure == NULL) {
				g_warning("cannot configure this plugin");
				return TRUE;
			}
			(*plugin->configure)();
			return TRUE;
		}
		case 'a': {
			if (plugin->about == NULL) {
				g_warning("cannot about box for this plugin");
				return TRUE;
			}
			(*plugin->about)();
			return TRUE;
		}


		default:
			g_warning("unexpected character on control fd");
			gtk_main_quit();
			return FALSE;
	}
	return TRUE;
}

static void run_vis_plugin(VisPlugin* plugin, int fd, int control_fd)
{
	VisPluginCommand* command;
	VisPluginData plugin_data;
	GIOChannel* io_channel;
	DEBUGPRINT(("initializing plugin"));
	if (plugin->init != NULL)
		(*plugin->init)();
	DEBUGPRINT(("done initializing plugin"));

	if (plugin->playback_start != NULL) {
		DEBUGPRINT(("calling playback_start"));
		(*plugin->playback_start)();
	}

	plugin_data.plugin = plugin;
	plugin_data.buf_amount = 0;

	io_channel = g_io_channel_unix_new(fd);
	g_io_add_watch(io_channel, G_IO_IN, handle_server_input, &plugin_data);

	if (control_fd >= 0) {
		io_channel = g_io_channel_unix_new(control_fd);
		g_io_add_watch(io_channel,
		               G_IO_IN,
			       handle_control_input,
			       &plugin_data);
	}

	gtk_main();
cleanup:
	if (plugin->cleanup)	
		(*plugin->cleanup)();

}

static void usage(const char* progname)
{
	fprintf(stderr, "usage: %s [--plugin plugin.so]\n\n"
	                "Load the given xmms visualization plugin.\n\n"
			"Options:\n"
	"\t--plugin plugin.so     Specify the plugin to run.\n"
	"\t--unix unix_path       Listen on a unix-domain socket for input.\n"
		, progname);
	exit(1);
}

/* copied from gdam/server/gdamservermain.c */
static int accept_fd(const char* unix_socket)
{
        struct sockaddr_un un_addr;
        int listen_fd;
        int accepted_fd;
        int one = 1;
	struct sockaddr accepted_addr;
	int addr_len = sizeof(struct sockaddr);
        bzero(&un_addr, sizeof(un_addr));
        un_addr.sun_family = AF_UNIX;
        strncpy(un_addr.sun_path, unix_socket, UNIX_PATH_MAX);
        listen_fd = socket(PF_UNIX, SOCK_STREAM, 0);
        if (listen_fd < 0) {
                perror("error creating listening socket??");
                return -1;
        }
        fcntl(listen_fd, F_SETFD, FD_CLOEXEC);
        if (bind(listen_fd,
                 (struct sockaddr*)(&un_addr),
                 sizeof(struct sockaddr_un)) < 0) {
                fprintf(stderr, "bind: failed to bind to %s: %s.\n\n",
                                unix_socket, strerror(errno));
                close(listen_fd);
		unlink(unix_socket);
                return -1;
        }
        if (listen(listen_fd, 1) < 0) {
                fprintf(stderr, "error listening: %s\n", strerror(errno));
                close(listen_fd);
		unlink(unix_socket);
                return -1;
        }
retry_accept:
	accepted_fd = accept(listen_fd, &accepted_addr, &addr_len);
	if (accepted_fd < 0) {
		if (errno == EINTR)
			goto retry_accept;
		fprintf(stderr, "error accepting connection: %s\n",
			strerror(errno));
		close(listen_fd);
		close(accepted_fd);
		unlink(unix_socket);
		return -1;
	}

	close(listen_fd);
	unlink(unix_socket);
	return accepted_fd;
}


#define GET_ARG(var)	G_STMT_START{				\
	if (argv[i + 1] == NULL)				\
		g_error("%s needs an argument", argv[i]);	\
	else							\
		var = (argv[++i]);				\
	}G_STMT_END

int main(int argc, char** argv)
{
	const char* plugin_name = NULL;
	const char* unix_socket = NULL;
	int i;
	VisPlugin* vis_plugin;
	int input_fd;
	int control_fd = -1;
	g_thread_init(NULL);
	gtk_init(&argc, &argv);
	gdk_rgb_init();
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--help") == 0
		 || strcmp(argv[i], "-help") == 0) {
		 	usage(argv[0]);
		}
		if (strcmp(argv[i], "--plugin") == 0) {
			GET_ARG(plugin_name);
			continue;
		}
		if (strcmp(argv[i], "--control-fd") == 0) {
			const char* str;
			GET_ARG(str);
			control_fd = atoi(str);
			continue;
		}
		if (strcmp(argv[i], "--unix") == 0) {
			GET_ARG(unix_socket);
			continue;
		}

		g_warning("unexpected argument %s", argv[i]);
		usage(argv[0]);
	}
	if (plugin_name == NULL) {
		g_warning("--plugin is mandatory");
		usage(argv[0]);
	}
	vis_plugin = load_vis_plugin(plugin_name);
	if (vis_plugin == NULL)
		g_error("error loading plugin %s", plugin_name);

	if (unix_socket != NULL)
		input_fd = accept_fd(unix_socket);
	else
		input_fd = 0;
	if (input_fd < 0)
		g_error("error accepting on %s", unix_socket);
	run_vis_plugin(vis_plugin, input_fd, control_fd);
	return 0;
}
#else
int main(int argc, char** argv)
{
	fprintf(stderr, "WARNING, compiled without xmms support.");
	return 1;
}
#endif  /*  HAVE_XMMS  */
