/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdambuffercombo.h"
#include <gtk/gtklist.h>
#include <gtk/gtkentry.h>
#include <gtk/gtksignal.h>
#include <ctype.h>
#include <stdlib.h>

enum {
	BUFFER_SELECTED,
	LAST_SIGNAL
};

static guint buffer_combo_signals[LAST_SIGNAL] = { 0 };

static GtkObjectClass* parent_class = NULL;


static void flush_new_items(GdamBufferCombo* combo)
{
  GList *new_items;
  new_items = g_list_reverse (combo->new_items);
  combo->new_items = NULL;
  gtk_list_append_items ((GtkList *) GTK_COMBO (combo)->list, new_items);
}

static void flush_old_items(GdamBufferCombo* combo)
{
  GList *old_items;
  old_items = combo->old_items;
  combo->old_items = NULL;
  gtk_list_remove_items ((GtkList *) GTK_COMBO (combo)->list, old_items);
  g_list_free (old_items);
}

static void buffer_combo_add_callback(GdamBufferManagerInfo* info,
	                              gpointer user_data)
{
  GdamBufferCombo *combo = user_data;
  GtkWidget *item;
  char text[1024];
  int buffer_id = info->buffer_id;

  g_snprintf (text, sizeof (text), "%d: %s", buffer_id, info->description);
  item = gtk_list_item_new_with_label (text);
  gtk_widget_show (item);
  combo->new_items = g_list_prepend (combo->new_items, item);
  g_hash_table_insert (combo->item_map, GINT_TO_POINTER (buffer_id), item);
  if (combo->freeze_count == 0)
    {
      flush_new_items (combo);
    }
}

static void buffer_combo_remove_callback(GdamBufferManagerInfo* info,
	                                 gpointer user_data)
{
  GdamBufferCombo *combo = user_data;
  GtkListItem *item;
  int buffer_id = info->buffer_id;
  item = g_hash_table_lookup (combo->item_map, GINT_TO_POINTER (buffer_id));
  g_return_if_fail (item != NULL);
  combo->old_items = g_list_prepend (combo->old_items, item);
  if (combo->freeze_count == 0)
    {
      flush_old_items (combo);
    }
}

void gdam_buffer_combo_freeze(GdamBufferCombo* combo)
{
  combo->freeze_count++;
}
void gdam_buffer_combo_thaw(GdamBufferCombo* combo)
{
  g_return_if_fail (combo->freeze_count >= 1);
  combo->freeze_count--;
  if (combo->freeze_count == 0)
    {
      flush_new_items (combo);
      flush_old_items (combo);
    }
}

static void gdam_buffer_combo_destroy(GtkObject* object)
{
  GdamBufferCombo *combo;
  combo = GDAM_BUFFER_COMBO (object);
  if (combo->notifier != NULL)
    {
      gdam_buffer_manager_unsubscribe (combo->notifier);
      combo->notifier = NULL;
    }
  if (combo->item_map != NULL)
    {
      g_hash_table_destroy (combo->item_map);
      combo->item_map = NULL;
    }
  (*parent_class->destroy) (object);
}
static void gdam_buffer_combo_entry_changed(GdamBufferCombo* combo)
{
  GtkEntry *entry;
  const char *text;
  int buffer_id;

  entry = GTK_ENTRY (GTK_COMBO (combo)->entry);
  g_return_if_fail (entry != NULL);

  text = gtk_entry_get_text (entry);
  if (!isdigit (*text))
    {
      g_warning ("cannot parse buffer id from combo box");
      return;
    }
  buffer_id = atoi (text);
  gtk_signal_emit (GTK_OBJECT (combo),
		   buffer_combo_signals[BUFFER_SELECTED], buffer_id);
}

static void gdam_buffer_combo_init(GdamBufferCombo* combo)
{
  GdamBufferManagerNotifier *notifier;

  combo->item_map = g_hash_table_new (NULL, NULL);

  notifier = g_new0 (GdamBufferManagerNotifier, 1);
  notifier->user_data = combo;
  notifier->buffer_add_callback = buffer_combo_add_callback;
  notifier->buffer_remove_callback = buffer_combo_remove_callback;
  combo->notifier = notifier;

  gdam_buffer_combo_freeze (combo);
  gdam_buffer_manager_subscribe (notifier);
  gdam_buffer_combo_thaw (combo);

  gtk_signal_connect_object
    (GTK_OBJECT (GTK_COMBO (combo)->entry),
     "changed",
     GTK_SIGNAL_FUNC (gdam_buffer_combo_entry_changed), GTK_OBJECT (combo));
}

static void gdam_buffer_combo_class_init(GdamBufferComboClass* klass)
{
  GtkObjectClass *object_class;
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gdam_buffer_combo_destroy;
  buffer_combo_signals[BUFFER_SELECTED]
    = gtk_signal_new ("buffer-selected",
		      GTK_RUN_LAST,
		      object_class->type,
		      GTK_SIGNAL_OFFSET (GdamBufferComboClass,
					 buffer_selected),
		      gtk_marshal_NONE__INT, GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  gtk_object_class_add_signals (object_class,
				buffer_combo_signals, LAST_SIGNAL);
}

GtkType gdam_buffer_combo_get_type()
{
  static GtkType buffer_combo_type = 0;
  if (!buffer_combo_type)
    {
      static const GtkTypeInfo buffer_combo_info = {
	"GdamBufferCombo",
	sizeof (GdamBufferCombo),
	sizeof (GdamBufferComboClass),
	(GtkClassInitFunc) gdam_buffer_combo_class_init,
	(GtkObjectInitFunc) gdam_buffer_combo_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = gtk_combo_get_type ();
      parent_class = gtk_type_class (parent);
      buffer_combo_type = gtk_type_unique (parent, &buffer_combo_info);
    }
  return buffer_combo_type;
}

GdamBufferCombo* gdam_buffer_combo_new()
{
  return GDAM_BUFFER_COMBO (gtk_type_new (GDAM_TYPE_BUFFER_COMBO));
}
