#include "gdambufferpixmap.h"
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdambuffermanager.h>

typedef struct _CacheInfo CacheInfo;
typedef struct _BufferPixmapCache BufferPixmapCache;
typedef struct _WaitNotifier WaitNotifier;
typedef struct _MinMax MinMax;

#define MIN_NUM_SAMPLES	1024

typedef enum
{
  CACHE_INFO_BLANK,
  CACHE_INFO_WAITING_FOR_LENGTH,		/* NOTE: currently unused */
  CACHE_INFO_WAITING,
  CACHE_INFO_VALID,
} CacheInfoState;

struct _WaitNotifier
{
  GdamBufferPixmapFunc hook;
  gpointer data;
  GDestroyNotify destroy;
  WaitNotifier *next;
};

struct _MinMax
{
  gint32 min_value;
  gint32 max_value;
};

struct _CacheInfo
{
  BufferPixmapCache *cache;
  guint buffer_id;
  gint64 length;
  CacheInfoState state : 3;
  guint was_uncached : 1;
  guint in_unlocked_list : 1;
  guint lock_count : 28;
  guint num_samples;
  MinMax *samples;

  CacheInfo *next_older_unlocked;
  CacheInfo *next_younger_unlocked;

  /* only if in WAITING or WAITING_FOR_LENGTH states */
  WaitNotifier *first_notifier;
  WaitNotifier *last_notifier;
};

#define cache_info_alloc()		g_new (CacheInfo, 1)
#define cache_info_free(info)		g_free (info)

struct _BufferPixmapCache
{
  GdamChannel *channel;
  GHashTable *id_to_cache_info;
  guint max_unlocked;
  guint num_locked;
  guint num_blank;
  guint num_blank_and_locked;
  CacheInfo *oldest_unlocked;
  CacheInfo *youngest_unlocked;
};

/* global table mapping GdamChannel => BufferPixmapCache */
static GHashTable *channel_to_cache = NULL;

#define DEFAULT_MAX_UNLOCKED		32
#define AUDIO_SAMPLES_PER_VISUAL_SAMPLE	1024
#define MAX_SAMPLINGS_PER_BUFFER	(1 << 18)

static BufferPixmapCache *
buffer_pixmap_cache_force (GdamChannel  *channel)
{
  BufferPixmapCache *rv;
  if (channel_to_cache == NULL)
    channel_to_cache = g_hash_table_new (NULL, NULL);
  rv = g_hash_table_lookup (channel_to_cache, channel);
  if (rv == NULL)
    {
      /* XXX: this needs to trap the destruction of the channel!!! */
      rv = g_new (BufferPixmapCache, 1);
      rv->channel = channel;
      rv->id_to_cache_info = g_hash_table_new (NULL, NULL);
      rv->max_unlocked = DEFAULT_MAX_UNLOCKED;
      rv->num_locked = rv->num_blank_and_locked = rv->num_blank = 0;
      rv->oldest_unlocked = rv->youngest_unlocked = NULL;

      g_hash_table_insert (channel_to_cache, channel, rv);
    }
  return rv;
}

static inline CacheInfo *
buffer_pixmap_cache_try_info (BufferPixmapCache *cache,
                              guint              buffer_id)
{
  gpointer buffer_id_ptr = GUINT_TO_POINTER (buffer_id);
  return g_hash_table_lookup (cache->id_to_cache_info, buffer_id_ptr);
}

static inline CacheInfo *
buffer_pixmap_cache_force_info (BufferPixmapCache *cache,
				guint              buffer_id)
{
  CacheInfo *cache_info;
  gpointer buffer_id_ptr = GUINT_TO_POINTER (buffer_id);
  cache_info = g_hash_table_lookup (cache->id_to_cache_info, buffer_id_ptr);
  if (cache_info == NULL)
    {
      cache_info = cache_info_alloc ();
      cache_info->buffer_id = buffer_id;
      cache_info->state = CACHE_INFO_BLANK;
      cache_info->was_uncached = 0;
      cache_info->in_unlocked_list = 0;
      cache_info->lock_count = 0;
      cache_info->num_samples = 0;
      cache_info->samples = NULL;
      cache_info->next_younger_unlocked = NULL;
      cache_info->next_older_unlocked = NULL;
      cache_info->cache = cache;
    }
  return cache_info;
}

static inline void
add_to_unlocked_list (BufferPixmapCache *cache,
                      CacheInfo         *cache_info)
{
  g_return_if_fail (!cache_info->in_unlocked_list);
  g_return_if_fail (cache_info->next_younger_unlocked == NULL);
  g_return_if_fail (cache_info->next_older_unlocked == NULL);

  cache_info->next_older_unlocked = cache->youngest_unlocked;
  if (cache_info->next_older_unlocked != NULL)
    cache_info->next_older_unlocked->next_younger_unlocked = cache_info;
  else
    cache->oldest_unlocked = cache_info;
  cache->youngest_unlocked = cache_info;
}

static inline void
remove_from_unlocked_list (BufferPixmapCache *cache,
                           CacheInfo         *cache_info)
{
  CacheInfo *younger = cache_info->next_younger_unlocked;
  CacheInfo *older = cache_info->next_older_unlocked;
  g_return_if_fail (cache_info->in_unlocked_list);
  if (older != NULL)
    older->next_younger_unlocked = younger;
  else
    cache->oldest_unlocked = younger;
  if (younger != NULL)
    younger->next_older_unlocked = older;
  else
    cache->youngest_unlocked = older;
  cache_info->next_older_unlocked = cache_info->next_younger_unlocked = NULL;
  cache_info->in_unlocked_list = 0;
}

static void
destroy_unlocked_cache_info (BufferPixmapCache *cache,
                             CacheInfo         *cache_info)
{
  g_return_if_fail (cache_info->lock_count == 0);
  if (cache_info->in_unlocked_list)
    remove_from_unlocked_list (cache, cache_info);
  g_free (cache_info->samples);
  cache_info_free (cache_info);
}

#define CACHE_GET_NUM_UNLOCKED_AND_NONBLANK(cache)		\
     ((int) g_hash_table_size (cache->id_to_cache_info)		\
    - cache->num_locked						\
    - cache->num_blank						\
    + cache->num_blank_and_locked)

static void
maybe_free_unlocked_cache_infos (BufferPixmapCache *cache)
{
  while (CACHE_GET_NUM_UNLOCKED_AND_NONBLANK (cache) > cache->max_unlocked)
    destroy_unlocked_cache_info (cache, cache->oldest_unlocked);
}

void
gdam_buffer_pixmap_set_cache_size (GdamChannel         *channel,
				   int                  size)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  cache->max_unlocked = size;
  maybe_free_unlocked_cache_infos (cache);
}

void
gdam_buffer_pixmap_uncache        (GdamChannel         *channel,
				   guint                buffer_id)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  CacheInfo *cache_info = buffer_pixmap_cache_try_info (cache, buffer_id);
  if (cache_info == NULL)
    return;
  if (cache_info->lock_count == 0)
    {
      destroy_unlocked_cache_info (cache, cache_info);
    }
  else
    {
      cache_info->was_uncached = 1;
    }
}

gboolean
gdam_buffer_pixmap_lock           (GdamChannel         *channel,
				   guint                buffer_id)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  CacheInfo *cache_info = buffer_pixmap_cache_try_info (cache, buffer_id);
  if (cache_info == NULL)
    return FALSE;
  ++(cache_info->lock_count);
  if (cache_info->lock_count == 1)
    {
      ++(cache->num_locked);
      if (cache_info->state == CACHE_INFO_VALID)
	remove_from_unlocked_list (cache, cache_info);
      else
        ++(cache->num_blank_and_locked);
    }
  return TRUE;
}

void
gdam_buffer_pixmap_unlock         (GdamChannel         *channel,
				   guint                buffer_id)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  CacheInfo *cache_info = g_hash_table_lookup (cache->id_to_cache_info,
                                               GUINT_TO_POINTER (buffer_id));
  g_return_if_fail (cache_info != NULL);
  g_return_if_fail (cache_info->lock_count > 0);
  --(cache_info->lock_count);
  if (cache_info->lock_count == 0)
    {
      if (cache_info->was_uncached)
        {
	  destroy_unlocked_cache_info (cache, cache_info);
	  return;
	}

      if (cache_info->state == CACHE_INFO_VALID)
	add_to_unlocked_list (cache, cache_info);
      else
        --(cache->num_blank_and_locked);

      --(cache->num_locked);
      maybe_free_unlocked_cache_infos (cache);
    }
}

gboolean
gdam_buffer_pixmap_get_cached     (GdamChannel         *channel,
				   guint                buffer_id,
				   GdamTime             start_time,
				   GdamTime             length,
				   gboolean             filled,
				   GdkPixmap           *dst_pixmap,
				   guint                left,
				   guint                top,
				   guint                width,
				   guint                height,
				   GdkGC               *gc)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  CacheInfo *cache_info = buffer_pixmap_cache_try_info (cache, buffer_id);
  /* indices in the MinMax array for which part of the buffer is covered */
  guint first_sample, last_sample, num_samples;

  /* iterate over each pixel in the x direction */
  guint x;

  /* current pixel being considered in the MinMax array */
  guint at_subsample;

  guint cached_num_samples;

  /* check if we have a valid cache_info */
  if (cache_info == NULL || cache_info->state != CACHE_INFO_VALID)
    return FALSE;
  cached_num_samples = cache_info->num_samples;

  /* interpret negative length as meaning "to the end of the sample" */
  if (length < 0)
    {
      length = cached_num_samples - start_time;
      if (length < 0)
	length = 0;
    }

  /* compute MinMax indices to scan */
  first_sample = start_time / AUDIO_SAMPLES_PER_VISUAL_SAMPLE;
  last_sample = (start_time + length) / AUDIO_SAMPLES_PER_VISUAL_SAMPLE;
  num_samples = last_sample - first_sample;

  /* figure out which x to start x (in case we start before the audio data) */
  if (start_time < 0)
    {
      x = (gint64) (-first_sample) * width / num_samples;
      at_subsample = 0;
    }
  else
    {
      x = 0;
      at_subsample = first_sample;
    }
  for (     ; x < width && at_subsample < cached_num_samples; x++)
    {
      guint last_subsample = first_sample + (guint64) num_samples * (x + 1) / width;
      gint h, min_y, max_y;
      MinMax min_max = cache_info->samples[at_subsample];

      /* clip to end of cached samples */
      if (last_subsample > cached_num_samples)
	last_subsample = cached_num_samples;

      if (at_subsample < last_subsample)
	{
	  at_subsample++;
	  while (at_subsample < last_subsample)
	    {
	      gint min_value = cache_info->samples[at_subsample].min_value;
	      gint max_value = cache_info->samples[at_subsample].max_value;
	      if (min_max.min_value > min_value)
		min_max.min_value = min_value;
	      if (min_max.max_value < max_value)
		min_max.max_value = max_value;
	      at_subsample++;
	    }
	}
      h = height / 2;
      min_y = (gint64) (min_max.min_value * h >> GDAM_SAMPLE_BITS) + h + top;
      max_y = (gint64) (min_max.max_value * h >> GDAM_SAMPLE_BITS) + h + top;
      if (filled)
	gdk_draw_line (dst_pixmap, gc, x + left, min_y, x + left, max_y);
      else
	{
	  gdk_draw_point (dst_pixmap, gc, x + left, min_y);
	  gdk_draw_point (dst_pixmap, gc, x + left, max_y);
	}
    }
  return TRUE;
}

static void
cache_info_handle_failure (gpointer   user_data)
{
  CacheInfo *cache_info = user_data;
  while (cache_info->first_notifier != NULL)
    {
      WaitNotifier *notifier = cache_info->first_notifier;
      cache_info->first_notifier = notifier->next;
      if (cache_info->first_notifier == NULL)
	cache_info->last_notifier = NULL;

      /* destroy the hook */
      if (notifier->destroy != NULL)
	(*notifier->destroy) (notifier->data);
      g_free (notifier);
    }

  /* TODO: anything else to do???
   *       we should delete the cache-info or set an error flag.
   */
}

static void
cache_info_handle_profile (int        num_args,
			   GdamArg   *args,
			   gpointer   user_data)
{
  CacheInfo *cache_info = user_data;
  guint num_ints;
  guint num_pairs;
  gint *int_array;
  guint i;
  GdamArg *arg;

  arg = gdam_arg_scan (num_args, args, "values", GDAM_ARG_TYPE_INT_ARRAY);
  if (arg == NULL)
    {
      if (num_args == 1 && args[0].type == GDAM_ARG_TYPE_STRING
	  && strcmp (args[0].name, "error") == 0)
	g_warning ("cache-buffer-profile: error: %s [buffer %d]",
		   GDAM_ARG_STRING (args + 0), cache_info->buffer_id);
      else
	g_warning ("cache-buffer-profile: couldn't find 'values' array "
		   "for buffer %d", cache_info->buffer_id);
      cache_info_handle_failure (cache_info);
      return;
    }
  num_ints = GDAM_ARG_INT_ARRAY (arg).length;
  int_array = GDAM_ARG_INT_ARRAY (arg).array;

  arg = gdam_arg_scan (num_args, args, "length", GDAM_ARG_TYPE_INT64);
  g_return_if_fail (arg != NULL);
  cache_info->length = GDAM_ARG_INT64 (arg);

  g_return_if_fail (num_ints % 2 == 0);
  num_pairs = num_ints / 2;

  /* initialize MinMax array from int_array. */
  cache_info->samples = g_new (MinMax, num_pairs);
  for (i = 0; i < num_pairs; i++)
    {
      cache_info->samples[i].min_value = int_array[2*i + 0];
      cache_info->samples[i].max_value = int_array[2*i + 1];
    }
  cache_info->state = CACHE_INFO_VALID;

  gdam_buffer_pixmap_lock (cache_info->cache->channel,
			   cache_info->buffer_id);

  /* chain off the various GdamBufferPixmapFunc's that are waiting */
  while (cache_info->first_notifier != NULL)
    {
      WaitNotifier *notifier = cache_info->first_notifier;
      cache_info->first_notifier = notifier->next;
      if (cache_info->first_notifier == NULL)
	cache_info->last_notifier = NULL;

      /* run the hook */
      (*notifier->hook) (cache_info->cache->channel,
			 cache_info->buffer_id,
			 notifier->data);
      if (notifier->destroy != NULL)
	(*notifier->destroy) (notifier->data);
      g_free (notifier);
    }

  gdam_buffer_pixmap_unlock (cache_info->cache->channel,
			     cache_info->buffer_id);
}

void
gdam_buffer_pixmap_get            (GdamChannel         *channel,
				   guint                buffer_id,
				   GdamBufferPixmapFunc func,
				   gpointer             data,
				   GDestroyNotify       destroy)
{
  BufferPixmapCache *cache = buffer_pixmap_cache_force (channel);
  CacheInfo *cache_info = buffer_pixmap_cache_force_info (cache, buffer_id);
  switch (cache_info->state)
    {
    case CACHE_INFO_VALID:
      {
	(*func) (channel, buffer_id, data);
	if (destroy)
	  (*destroy) (data);
	return;
      }
    case CACHE_INFO_BLANK:
      {
	GdamBufferManagerInfo *info;
	GdamRequest *request;
	gdouble num_pixels_double;
	gint num_pixels;
	info = gdam_buffer_manager_info_by_id (buffer_id);

	/* XXX: here we assume that the buffer-manager will
	 * 	always know about the buffer.
	 *
	 * 	We should go to CACHE_INFO_WAITING_FOR_LENGTH
	 * 	if this happens, but we just assume if won't instead.
	 */
	if (info == NULL)
	  {
	    g_warning ("gdam_buffer_pixmap_get: "
		       "buffer_id %d not tracked by buffer-manager", buffer_id);
	    if (destroy != NULL)
	      (*destroy) (data);
	    return;
	  }

	/* compute the number of min/max samplings to extract */
        num_pixels_double = (double) info->length * AUDIO_SAMPLES_PER_VISUAL_SAMPLE;
	if (num_pixels_double > MAX_SAMPLINGS_PER_BUFFER)
	  num_pixels = MAX_SAMPLINGS_PER_BUFFER;
	else
	  num_pixels = (int) num_pixels_double;

	{
#define NUM_PROFILE_ARGS 4
	GdamArg profile_args[NUM_PROFILE_ARGS];
	profile_args[0].name = "mode";
	profile_args[0].type = GDAM_ARG_TYPE_STRING;
	GDAM_ARG_STRING (profile_args+0) = "minmax";
	profile_args[1].name = "use_doubles";
	profile_args[1].type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT (profile_args+1) = FALSE;
	profile_args[2].name = "start";
	profile_args[2].type = GDAM_ARG_TYPE_INT64;
	GDAM_ARG_INT64 (profile_args+2) = 0;
	profile_args[2].name = "length";
	profile_args[2].type = GDAM_ARG_TYPE_INT64;
	GDAM_ARG_INT64 (profile_args+3) = -1;
	profile_args[3].name = "num_outputs";
	profile_args[3].type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT (profile_args+3) = num_pixels;

	/* make a profile request */
	request
	  = gdam_request_call_function (GDAM_PROTOCOL_BUFFER_ARG,
					buffer_id,
					"profile",
					NUM_PROFILE_ARGS,
					profile_args,
					cache_info_handle_profile,
					cache_info_handle_failure,
					cache_info);
#undef NUM_PROFILE_ARGS
	}

	/* now we have made a request so we're officially waiting for a reply */
	cache_info->state = CACHE_INFO_WAITING;

	/* fall-through to add to waiting notification queue */
      }
    case CACHE_INFO_WAITING_FOR_LENGTH:
    case CACHE_INFO_WAITING:
      {
	WaitNotifier *wait_notifier;

	/* add to waiting notification queue */
	wait_notifier = g_new (WaitNotifier, 1);
	wait_notifier->hook = func;
	wait_notifier->data = data;
	wait_notifier->destroy = destroy;
	wait_notifier->next = NULL;
	if (cache_info->last_notifier == NULL)
	  cache_info->first_notifier = wait_notifier;
	else
	  cache_info->last_notifier->next = wait_notifier;
	cache_info->last_notifier = wait_notifier;

	return;
      }
    }
  g_assert_not_reached ();
}
