/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdambuffersaveskin.h"
#include <gdam/component/gdambuffermanager.h>
#include "gdamskinutil.h"
#include <gdam/common/gdam-protocol.h>
#include <gtk/gtklabel.h>
#include <gtk/gtksignal.h>
#include <string.h>
#include <stdlib.h>

static void gdam_buffer_save_skin_init(GdamBufferSaveSkin* buffer_save_skin)
{
}

/*
 *     Arguments:  buffer_id
 */
static gboolean do_widget_init(GdamSkin* skin, gpointer user_data)
{
  GdamBufferSaveSkin *save_skin;
  save_skin = GDAM_BUFFER_SAVE_SKIN (skin);
  if (save_skin->id_widgets == NULL || save_skin->description_widgets == NULL)
    return FALSE;
  if (save_skin->saving_id == 0)
    return FALSE;
  gdam_buffer_save_skin_set_buffer_id (save_skin, save_skin->saving_id);
  if (save_skin->description != NULL)
    gdam_buffer_save_skin_set_description (save_skin, save_skin->description);
  return TRUE;
}

static gboolean gdam_buffer_save_build_new_model(GdamSkin*      skin,
						 GdamChannel*   channel,
						 int            num_options,
	                                         char**         options)
{
  GdamBufferSaveSkin *save_skin;
  save_skin = GDAM_BUFFER_SAVE_SKIN (skin);
  save_skin->saving_id = 0;

  if (num_options > 1)
    {
      save_skin->saving_id = atoi (options[0]);
    }
  else
    {
      save_skin->saving_id = gdam_buffer_get_default ();
    }
  gdam_skin_add_initializer (skin, do_widget_init, NULL, NULL);
  if (save_skin->saving_id != 0)
    {
      GdamBufferManagerInfo *info;
      info = gdam_buffer_manager_info_by_id (save_skin->saving_id);
      if (info != NULL)
	{
	  g_free (save_skin->description);
	  save_skin->description = g_strdup (info->description);
	}
    }

  return TRUE;
}

static gboolean gdam_buffer_save_filename_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamBufferSaveSkin *save_skin;
  if (!GTK_IS_ENTRY (object))
    {
      g_warning ("%s only valid for Entry's", handler_names[0]);
      return TRUE;
    }
  save_skin = GDAM_BUFFER_SAVE_SKIN (skin);
  if (save_skin->filename != NULL)
    {
      g_warning ("%s may only be specified once", handler_names[0]);
      return TRUE;
    }
  save_skin->filename = GTK_ENTRY (object);
  return TRUE;
}


#define gdam_buffer_save_description_handler	\
	   gdam_buffer_save_buffer_id_handler
static gboolean gdam_buffer_save_buffer_id_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamBufferSaveSkin *save_skin;
  GSList **list;
  save_skin = GDAM_BUFFER_SAVE_SKIN (skin);
  if (!GTK_IS_LABEL (object) && !GTK_IS_ENTRY (object))
    {
      g_warning ("%s only valid for Labels or Entrys", handler_names[0]);
      return TRUE;
    }
  if (strstr (handler_names[0], "buffer_id"))
    list = &save_skin->id_widgets;
  else
    list = &save_skin->description_widgets;
  *list = g_slist_prepend (*list, object);
  return TRUE;
}

static void on_save_success(int num, GdamArg* args, gpointer user_data)
{
  g_message ("successfully saved");
  gtk_object_unref (GTK_OBJECT (user_data));
}
static void on_save_failure(gpointer user_data)
{
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void do_save(GdamBufferSaveSkin* skin)
{
  GdamArg args[1];
  char *text;
  char *ext;
  args[0].name = "filename";
  args[0].type = GDAM_ARG_TYPE_STRING;

  text = gtk_entry_get_text (skin->filename);
  ext = strrchr (text, '.');
  if (ext == NULL || strcasecmp (ext, ".wav") != 0)
    {
      g_warning ("filename must end in .wav for security.");
      return;
    }
  GDAM_ARG_STRING (&args[0]) = text;

  gtk_object_ref (GTK_OBJECT (skin));
  gdam_channel_append_request
    (GDAM_SKIN (skin)->channel,
     gdam_request_call_function
     (GDAM_PROTOCOL_BUFFER_ARG,
      skin->saving_id,
      "save", 1, args, on_save_success, on_save_failure, skin));
}

static gboolean gdam_buffer_save_do_save_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object
    (object, signal_name, GTK_SIGNAL_FUNC (do_save), GTK_OBJECT (skin));
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] =
{
  {
  "gdam_buffer_save_filename", gdam_buffer_save_filename_handler}
  ,
  {
  "gdam_buffer_save_buffer_id", gdam_buffer_save_buffer_id_handler}
  ,
  {
  "gdam_buffer_save_description", gdam_buffer_save_description_handler}
  ,
  {
  "gdam_buffer_save_do_save", gdam_buffer_save_do_save_handler}
,};

static void gdam_buffer_save_skin_class_init(GdamSkinClass* skin_class)
{
  /*GTK_OBJECT_CLASS(skin_class)->destroy
   = gdam_buffer_manager_skin_destroy; *//* XXX */
  skin_class->build_new_model = gdam_buffer_save_build_new_model;
  gdam_skin_class_register_generic_handlers
    (skin_class, G_N_ELEMENTS (connect_infos), connect_infos);
}

GtkType gdam_buffer_save_skin_get_type()
{
  static GtkType buffer_save_skin_type = 0;
  if (!buffer_save_skin_type)
    {
      static const GtkTypeInfo buffer_save_skin_info = {
	"GdamBufferSaveSkin",
	sizeof (GdamBufferSaveSkin),
	sizeof (GdamBufferSaveSkinClass),
	(GtkClassInitFunc) gdam_buffer_save_skin_class_init,
	(GtkObjectInitFunc) gdam_buffer_save_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      buffer_save_skin_type = gtk_type_unique (parent,
					       &buffer_save_skin_info);
    }
  return buffer_save_skin_type;
}

static void set_text(GtkWidget* widget, const char* text)
{
  if (GTK_IS_LABEL (widget))
    gtk_label_set_text ((GtkLabel *) widget, text);
  else if (GTK_IS_ENTRY (widget))
    gtk_entry_set_text ((GtkEntry *) widget, text);
  else
    g_assert_not_reached ();
}
static void set_list_to_text(GSList* list_of_widgets, const char* text)
{
  g_slist_foreach (list_of_widgets, (GFunc) set_text, (gpointer) text);
}

void gdam_buffer_save_skin_set_description(GdamBufferSaveSkin*   skin,
                                           const char*           description)
{
  if (skin->description != description)
    {
      g_free (skin->description);
      skin->description = g_strdup (description);
    }
  set_list_to_text (skin->description_widgets, description);
}
void gdam_buffer_save_skin_set_buffer_id  (GdamBufferSaveSkin*   skin,
                                           int                   id)
{
  char buf[256];
  sprintf (buf, "%d", id);
  set_list_to_text (skin->id_widgets, buf);
}
