#include "gdambufferviewport.h"
#include "gdamskinutil.h"
#include <math.h>
#include <gdam/common/gdamalloca.h>

#define gdam_buffer_viewport_adornment_alloc()	g_new (GdamBufferViewportAdornment, 1)
#define gdam_buffer_viewport_adorn_hook_alloc()	g_new (GdamBufferViewportAdornHook, 1)
#define gdam_buffer_viewport_adornment_free	g_free
#define gdam_buffer_viewport_adorn_hook_free	g_free

static inline int
get_x_from_time(GdamBufferViewport *viewport, 
		gdouble             time)
{
  time -= viewport->time_start;
  if (viewport->time_length > 0.0001)
    time /= viewport->time_length;
  time *= viewport->draw_width;
  return (int) time;
}

static inline gdouble
get_time_from_x (GdamBufferViewport *viewport, 
	         int                 screen_x)
{
  gdouble time = screen_x;
  time /= viewport->draw_width;
  time *= viewport->time_length;
  time += viewport->time_start;
  return time;
}

/* --- generic adornments --- */
GdamBufferViewportAdornment *
gdam_buffer_viewport_add_adornment (GdamBufferViewport           *viewport,
				    gboolean			  is_shown,
				    gboolean                      is_under,
				    GdamBufferViewportAdornFuncs *funcs,
				    gpointer                      adorn_data)
{
  GdamBufferViewportAdornment *adornment;
  adornment = gdam_buffer_viewport_adornment_alloc ();
  adornment->viewport = viewport;
  adornment->funcs = funcs;
  adornment->data = adorn_data;
  adornment->prev = viewport->last_adornment;
  adornment->next = NULL;
  if (viewport->last_adornment != NULL)
    viewport->last_adornment->next = adornment;
  else
    viewport->first_adornment = adornment;
  viewport->last_adornment = adornment;
  adornment->first_hook = NULL;
  adornment->last_hook = NULL;
  adornment->is_shown = is_shown;
  adornment->is_under = is_under;
  if (is_shown)
    gdam_buffer_viewport_queue_redraw (viewport);
  return adornment;
}

void
gdam_buffer_viewport_trap_adornment_changed
                                  (GdamBufferViewportAdornment *adornment,
                                   GdamBufferViewportAdornmentChange func,
                                   gpointer                     data,
                                   GDestroyNotify               destroy)
{
  GdamBufferViewportAdornHook *hook;
  hook = gdam_buffer_viewport_adorn_hook_alloc ();
  hook->func = func;
  hook->data = data;
  hook->destroy = destroy;
  hook->next = NULL;
  if (adornment->last_hook == NULL)
    adornment->first_hook = hook;
  else
    adornment->last_hook->next = hook;
  adornment->last_hook = hook;
}

static void
adornment_run_change_hooks (GdamBufferViewportAdornment *adornment)
{
  GdamBufferViewportAdornHook *last = NULL;
  GdamBufferViewportAdornHook *hook = adornment->first_hook;
  while (hook != NULL)
    {
      if (! (*hook->func) (adornment, hook->data))
	{
	  GdamBufferViewportAdornHook *next;

	  /* delete the hook */
	  if (last != NULL)
	    last->next = hook->next;
	  else
	    adornment->first_hook = hook->next;
	  if (adornment->last_hook == hook)
	    adornment->last_hook = last;

	  next = hook->next;
	  if (hook->destroy != NULL)
	    (*hook->destroy) (hook->data);
	  gdam_buffer_viewport_adorn_hook_free (hook);
	  hook = next;
	}
      else
	{
	  last = hook;
	  hook = hook->next;
	}
    }
}


void
gdam_buffer_viewport_adornment_changed (GdamBufferViewportAdornment *adornment)
{
  adornment_run_change_hooks (adornment);
  if (adornment->is_shown)
    gdam_buffer_viewport_queue_redraw (adornment->viewport);
}

void
gdam_buffer_viewport_adornment_changed_part
                                  (GdamBufferViewportAdornment *adornment,
				   guint                        x,
				   guint                        y,
				   guint                        width,
				   guint                        height)
{
  adornment_run_change_hooks (adornment);
  if (adornment->is_shown)
    gdam_buffer_viewport_queue_redraw_part (adornment->viewport,
					    x, y, width, height);
}

/* --- return a list of GdamBufferViewportAdornment's of a specific type --- */
GSList *
gdam_buffer_viewport_get_adornments_by_func
                                    (GdamBufferViewport           *viewport,
                                     GdamBufferViewportAdornFuncs *funcs)
{
  GSList *rv = NULL;
  GdamBufferViewportAdornment *adornment;
  for (adornment = viewport->last_adornment;
       adornment != NULL;
       adornment = adornment->prev)
    if (adornment->funcs == funcs)
      rv = g_slist_prepend (rv, adornment);
  return rv;
}

void
gdam_buffer_viewport_destroy_all_adornments (GdamBufferViewport *viewport)
{
  while (viewport->first_adornment != NULL)
    {
      GdamBufferViewportAdornment *at = viewport->first_adornment;

      viewport->first_adornment = at->next;
      if (at->next == NULL)
	viewport->last_adornment = NULL;

      /* Destroy the adornments hooks. */
      while (at->first_hook != NULL)
	{
	  GdamBufferViewportAdornHook *hook = at->first_hook;
	  at->first_hook = at->last_hook = NULL;
	  while (hook != NULL)
	    {
	      GdamBufferViewportAdornHook *next = hook->next;
	      if (hook->destroy != NULL)
		(*hook->destroy) (hook->data);
	      gdam_buffer_viewport_adorn_hook_free (hook);
	      hook = next;
	    }
	}
      if (at->funcs->destroy != NULL)
	(*at->funcs->destroy) (at->data);
      gdam_buffer_viewport_adornment_free (at);
    }
}

/* --- Adornments: Beat Information --- */

/* XXX hackery to deal with bpm == 0, beat_period == inf */
#define GDAM_REALLY_BIG_BEAT_PERIOD 1000000000000

/* XXX things don't work well if beat_period==0, so hack... */
#define MIN_BEAT_PERIOD			(0.01)

static inline gdouble
beat_info_time_to_beat (GdamBufferViewportBeatInfo *beat_info,
			gdouble                     time)
{
  if (beat_info->beat_period <= 0.0)
    return -1;
  return (time - beat_info->start) / beat_info->beat_period;
}

static inline gdouble
beat_info_index_to_time (GdamBufferViewportBeatInfo *beat_info,
			 gint                        index)
{
  return beat_info->start + beat_info->beat_period * index;
}

static gboolean
beat_info_render  (GdamBufferViewport *viewport,
		   GdkDrawable        *drawable,
		   gpointer            data,
		   guint               x,
		   guint               y,
		   guint               width,
		   guint               height)
{
  gdouble time_per_pixel = viewport->time_length / viewport->draw_width;
  int x_end = x + width;
  int draw_width = viewport->draw_width;
  int right_clip = (draw_width < x_end) ? 0 : (draw_width - x_end);

  gdouble view_start = viewport->time_start + time_per_pixel * x;
  gdouble view_width = (width - right_clip) * time_per_pixel;
  gdouble view_end = view_start + view_width;

  GdamBufferViewportBeatInfo *beat_info_data = data;

  gdouble b_period = 0.0;  /* beat period hackery. */
  int which_beat = 0, last_beat_index = 0;
  int num_lines;
  int i;
  GdkSegment *segments;

  if (beat_info_data->num_beats <= 0)
    return FALSE;

  if (beat_info_data->beat_line_gc == NULL)
    {
      GtkWidget *widget = GDAM_SKIN_GET_WIDGET (viewport);
      beat_info_data->beat_line_gc = widget->style->fg_gc[GTK_STATE_PRELIGHT];
      gdk_gc_ref (beat_info_data->beat_line_gc);
    }

  /*  If we have a beat period (and therefore bpm) */
  if (beat_info_data->beat_period > 0 
      && beat_info_data->beat_period < GDAM_REALLY_BIG_BEAT_PERIOD)
    {
      gdouble length;
      gdouble end_in_beats;
      gdouble first_beat;
      gdouble last_beat;

      b_period = beat_info_data->beat_period;

      /*  Calculate which beats are visible, and the number of beat lines
       *  to draw.
       */
      length = gdam_buffer_viewport_get_sample_length (viewport);
      end_in_beats = beat_info_time_to_beat (beat_info_data, length);
      first_beat = beat_info_time_to_beat (beat_info_data, view_start);
      last_beat = beat_info_time_to_beat (beat_info_data, view_end);
      if (last_beat > end_in_beats)
	last_beat = end_in_beats;

      /* Calculate the first beat which is visible. */
      which_beat = rint (first_beat);
      if (which_beat < 0)
	which_beat = 0;

      /* Calculate the last beat which is visible. */
      last_beat_index = (int) ceil (last_beat);
      if (last_beat_index > beat_info_data->num_beats)
	last_beat_index = beat_info_data->num_beats;

      num_lines = last_beat_index - which_beat;
    }
  /* If the bpm is zero but the first beat is visible, draw the first
   * beat. 
   */ 
  else if (view_start < beat_info_data->start
        && beat_info_data->start < view_end)
    {
      b_period = 0.0;
      num_lines = 1;
      which_beat = 0;
      last_beat_index = 1;
    }
  else
    {
      num_lines = 0;
    }

  if (num_lines <= 0)
    return TRUE;

  segments = alloca (sizeof (GdkSegment) * num_lines);
  i = 0;
  for (; which_beat < last_beat_index; which_beat++)
    {
      gdouble t;
      gdouble time_offset;
      int x;
      t = beat_info_data->start + b_period * which_beat;
      time_offset = t - viewport->time_start;
      x = (int) (time_offset / viewport->time_length * viewport->draw_width);
      segments[i].x1 = x;
      segments[i].y1 = 0;
      segments[i].x2 = x;
      segments[i].y2 = viewport->draw_height;
      i++;
    }
  gdk_draw_segments (drawable, beat_info_data->beat_line_gc, segments, i);
  return TRUE;
}

static GdamCursorType
beat_info_get_cursor (GdamBufferViewport *viewport,
		      gpointer            data)
{
  GdamBufferViewportBeatInfo *beat_info = data;

  /* If viewport doesn't display beat markers */
  if (!beat_info->is_draggable)
    return viewport->is_selectable ? GDAM_SELECT_CURSOR : GDAM_MOUSE_CURSOR;

  /* If we use beat markers from the beat_times arrary */
  if (beat_info->num_beats > 0)
    return GDAM_MOUSE_CURSOR;

  return GDAM_CURSOR_INVALID;
}


static void
beat_info_adjust_to_drag_info(GdamBufferViewportAdornment *adornment,
		              GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatInfo *beat_info = adornment->data;
  gint drag_index = beat_info->dragging_beat_line;
  gdouble old_pos = beat_info_index_to_time (beat_info, drag_index);
  gdouble delta = drag_info->end_time - old_pos;

  if ((drag_info->modifiers & GDK_CONTROL_MASK) != 0)
    {
      beat_info->start = drag_info->end_time;
    }
  else if (drag_index >= 0 && (drag_info->modifiers & GDK_SHIFT_MASK) != 0)
    {
      beat_info->start += delta;
    }
  else if (drag_index >= 0)
    {
      beat_info->beat_period += delta / drag_index;
      if (beat_info->beat_period < MIN_BEAT_PERIOD)
	beat_info->beat_period = MIN_BEAT_PERIOD;
    }
  else if (beat_info->beat_period < 0)
    {
      /* BONUS HACK: if no beat period happens to be set,
       *             assume the user clicked to mark a single beat length.
       */
      beat_info->beat_period = drag_info->end_time - beat_info->start;
    }
  else
    return;

  gdam_buffer_viewport_adornment_changed (adornment);
}

/* Derive beat positions from bpm, fb and find the beat closest to 'time' */
static int
beat_info_get_line (GdamBufferViewportBeatInfo *beat_info,
	            gdouble                     time)
{
  return rint ((time - beat_info->start) / beat_info->beat_period);
}

/* start_drag */
static gboolean
beat_info_start_drag  (GdamBufferViewportAdornment *adornment,
		       GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatInfo *beat_info_data = adornment->data;
  int dragging_beat_line;

  if (!beat_info_data->is_draggable)
    return FALSE;
  if (beat_info_data->num_beats <= 0)
    return FALSE;
  if (drag_info->button != 1)
    return FALSE;

  dragging_beat_line = beat_info_get_line (beat_info_data, drag_info->end_time);
  if (dragging_beat_line < 0
   || dragging_beat_line >= beat_info_data->num_beats)
    dragging_beat_line = -1;
  beat_info_data->dragging_beat_line = dragging_beat_line;

  beat_info_adjust_to_drag_info (adornment, drag_info);
  return TRUE;
}

static gboolean
beat_info_keep_dragging  (GdamBufferViewportAdornment *adornment,
		          GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatInfo *beat_info_data = adornment->data;
  if (drag_info->button != 1)
    return FALSE;
  if (beat_info_data->dragging_beat_line >= 0)
    {
      beat_info_adjust_to_drag_info (adornment, drag_info);
      return TRUE;
    }
  return FALSE;
}

/* end_drag */
static gboolean
beat_info_end_drag       (GdamBufferViewportAdornment *adornment,
		          GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatInfo *beat_info_data = adornment->data;
  if (drag_info->button != 1)
    return FALSE;
  beat_info_adjust_to_drag_info (adornment, drag_info);
  beat_info_data->dragging_beat_line = -1;
  return TRUE;
}

static void
beat_info_destroy (gpointer data)
{
  GdamBufferViewportBeatInfo *beat_info_data = data;
  if (beat_info_data->beat_line_gc != NULL)
    gdk_gc_unref (beat_info_data->beat_line_gc);
  g_free (beat_info_data);
}

static GdamBufferViewportAdornFuncs beat_info_funcs =
{
  beat_info_render,
  beat_info_get_cursor,
  beat_info_start_drag,
  beat_info_keep_dragging,
  beat_info_end_drag,
  beat_info_destroy
};

GdamBufferViewportAdornment *
gdam_buffer_viewport_add_beat_info  (GdamBufferViewport *viewport,
				     gboolean            is_shown,
				     gboolean            is_under,
                                     gboolean            is_draggable,
				     guint               num_beats,
				     gdouble             start_beat,
				     gdouble             beat_period,
				     GdkGC              *beat_line_gc)
{
  GdamBufferViewportBeatInfo *beat_info;
  beat_info = g_new (GdamBufferViewportBeatInfo, 1);

  beat_info->is_draggable = is_draggable;
  beat_info->start = start_beat;
  beat_info->beat_period = beat_period;
  beat_info->num_beats = num_beats;
  beat_info->beat_line_gc = beat_line_gc;
  if (beat_line_gc != NULL)
    gdk_gc_ref (beat_line_gc);

  return gdam_buffer_viewport_add_adornment (viewport,
					     is_shown,
					     is_under,
					     &beat_info_funcs,
					     beat_info);
}

/* --- Adornments: Beat Arrays --- */
/* XXX: this should definitely be rewritten to use a binary-search. */
static void
beat_array_get_visible_beats(gdouble        start_time,
		             gdouble        duration,
		             int            num_beats, 
                             const gdouble *beats,
                             guint         *first_beat_out, 
		             guint         *num_beats_out)
{
  int first, num;
  gdouble end_time = start_time + duration;
  for (first = 0; first < num_beats; first++)
    /* the test for 0.0 == 0.0 commonly fails without the tiny delta value */
    if (beats[first] + 0.000001 >= start_time)
      break;
  for (num = 0; num + first < num_beats; num++)
    if (beats[first + num] > end_time)
      break;
  *first_beat_out = first;
  *num_beats_out = num;
}

static int
beat_array_find_index_by_pos (GdamBufferViewportAdornment *adornment,
		              int                          x,
		              int                          y)
{
  GdamBufferViewportBeatArray *array = adornment->data;
  GdamBufferViewport *viewport = adornment->viewport;
  int radius = array->beat_radius;
  int height = viewport->draw_height;
  int circle_y;
  int i;
  int tolerance;

  if (y > 8 * radius)
    return -1;
  if (array->num_beats == 0)
    return -1;

  circle_y = radius * 4;
  tolerance = 3 * radius * radius;

  g_return_val_if_fail (array->beats != NULL, 0);

  for (i = 0; i < array->num_beats; i++)
    {
      int circle_x = get_x_from_time (viewport, array->beats[i]);
      int dist_squared = (x - circle_x) * (x - circle_x)
	               + (y - circle_y) * (y - circle_y);
      if (dist_squared < tolerance)
	return i;
    }
  return -1;
}

/* XXX: very inefficient */
void
beat_array_delete_nth_beat_time (GdamBufferViewportAdornment *adornment, int n)
{
  GdamBufferViewportBeatArray *array = adornment->data;
  gdouble *beats;
  int i;
  g_return_if_fail (n < array->num_beats);
  beats = g_new (gdouble, array->num_beats - 1);
  for (i = 0; i < n; i++)
    {
      beats[i] = array->beats[i];
    }
  for (i = n; i < array->num_beats - 1; i++)
    {
      beats[i] = array->beats[i + 1];
    }
  g_free (array->beats);
  array->beats = beats;
  array->num_beats--;
  gdam_buffer_viewport_adornment_changed (adornment);
}

static gboolean
beat_array_render (GdamBufferViewport *viewport,
		   GdkDrawable        *drawable,
		   gpointer            data,
		   guint               x,
		   guint               y,
		   guint               width,
		   guint               height)
{
  GdamBufferViewportBeatArray *array = data;

  int first_beat, num_beat;
  gdouble start_time = get_time_from_x (viewport, x);
  gdouble duration;

  /* Draw beat markers from the beat_times array */
  if (array->beat_gc == NULL)
    {
      array->beat_gc = gdk_gc_new (drawable);
      /* Make the beat bars yellow */
      gdk_gc_set_foreground (array->beat_gc,
			     util_alloc_color_cached (65535, 65535, 0));
    }
  if (array->num_beats <= 0)
    return FALSE;

  duration = (gdouble) width *viewport->time_length
	   / viewport->draw_width;
  g_return_val_if_fail (array->beats != NULL, FALSE);
  beat_array_get_visible_beats (start_time, duration,
		                array->num_beats, array->beats,
		                &first_beat, &num_beat);
  if (num_beat > 0)
    {
      int i;
      int radius = array->beat_radius;
      int y = array->beat_radius * 4;
      int drag_index = array->dragging_beat;
      g_return_val_if_fail (array->beats != NULL, FALSE);
      for (i = 0; i < num_beat; i++)
	{
	  int pos;
	  gdouble time;
	  gboolean fill = (i + first_beat != drag_index);
	  time = array->beats[i + first_beat];
	  pos = get_x_from_time (viewport, time);
	  gdk_draw_arc (drawable,
			array->beat_gc,
			fill,
			pos - radius + 1,
			y, radius * 2 - 1, radius * 2 - 1, 0, 360 * 64);
	}
    }
  return TRUE;
}


static gboolean
beat_array_start_drag  (GdamBufferViewportAdornment *adornment,
			GdamBufferViewportDragInfo  *drag_info)
{
  /* If we clicked on a beat from the beat_times arrary, begin dragging it. */
  GdamBufferViewportBeatArray *beat_array = adornment->data;
  int dragging_beat = beat_array_find_index_by_pos (adornment,
					            drag_info->start_x,
					            drag_info->start_y);
  beat_array->dragging_beat = -1;
  if (dragging_beat >= 0)
    {
      if (drag_info->button == 3)
	{
	  /* Right click deletes beat marker */
	  beat_array_delete_nth_beat_time (adornment, dragging_beat);
	}
      else
	{
	  beat_array->dragging_beat = dragging_beat;
	  /* XXX: should rerender */
	}
      return TRUE;
    }
  return FALSE;
}

static gboolean
beat_array_keep_dragging  (GdamBufferViewportAdornment *adornment,
			   GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatArray *array = adornment->data;
  guint dragging_beat = array->dragging_beat;
  gdouble old_pos, new_pos;
  if (array->dragging_beat <= 0)
    return FALSE;
  g_return_val_if_fail (array->beats != NULL, FALSE);

  array->beats[dragging_beat] += drag_info->end_time - drag_info->last_time;

  /* TODO: Redraw just the relevant areas. */
  gdam_buffer_viewport_adornment_changed (adornment);
  return TRUE;
}

static gboolean
beat_array_end_drag       (GdamBufferViewportAdornment *adornment,
			   GdamBufferViewportDragInfo  *drag_info)
{
  GdamBufferViewportBeatArray *array = adornment->data;
  g_return_val_if_fail (array->beats != NULL, 0);
  if (array->dragging_beat >= 0)
    {
      int index = array->dragging_beat;
      array->beats[index] += drag_info->end_time - drag_info->last_time;
      array->dragging_beat = -1;

      /* TODO: only redraw a partial area */
      gdam_buffer_viewport_adornment_changed (adornment);
    }
  return TRUE;
}

static void
beat_array_destroy (gpointer data)
{
  GdamBufferViewportBeatArray *array = data;
  if (array->beat_gc != NULL)
    gdk_gc_unref (array->beat_gc);
  g_free (array->beats);
  g_free (array);
}

static GdamBufferViewportAdornFuncs beat_array_funcs =
{
  beat_array_render,
  NULL,				/* get_cursor */
  beat_array_start_drag,
  beat_array_keep_dragging,
  beat_array_end_drag,
  beat_array_destroy
};

GdamBufferViewportAdornment *
gdam_buffer_viewport_add_beat_array (GdamBufferViewport *viewport,
				     gboolean            is_shown,
				     gboolean            is_under,
                                     gboolean            is_draggable,
				     guint               num_beats,
				     const gdouble      *beats,
				     GdkGC              *beat_gc)
{
  GdamBufferViewportBeatArray *beat_array;
  beat_array = g_new (GdamBufferViewportBeatArray, 1);
  beat_array->is_draggable = is_draggable;
  beat_array->dragging_beat = -1;
  beat_array->num_beats = num_beats;
  beat_array->beats = g_memdup (beat_array->beats, sizeof (gdouble) * num_beats);
  beat_array->beat_gc = beat_gc;
  if (beat_gc != NULL)
    gdk_gc_ref (beat_gc);
  beat_array->beat_radius = 4;
  return gdam_buffer_viewport_add_adornment (viewport,
					     is_shown,
					     is_under,
					     &beat_array_funcs,
					     beat_array);
}


/* --- Adornments:  Markers --- */
/* Compute where the conveniently-spaced markers should be drawn.  */
static void
compute_marker_locations(GdamBufferViewport *viewport,
			 GdamBufferViewportMarkerData *marker_data)
{
  guint i;
  gdouble num_markers;
  gdouble marker_skip;
  gdouble start_in_marks, end_in_marks;
  if (marker_data->markers != NULL)
    {
      for (i = 0; i < marker_data->num_markers; i++)
	g_free (marker_data->markers[i].text);
      g_free (marker_data->markers);
      marker_data->markers = NULL;
    }

  if (marker_data->marker_font == NULL)
    return;

  marker_data->old_time_start = viewport->time_start;
  marker_data->old_time_length = viewport->time_length;
  marker_data->old_draw_width = viewport->draw_width;

  num_markers = viewport->time_length / marker_data->marker_unit;
  marker_skip = 0.001;
  while (num_markers / marker_skip > 20)
    {
      marker_skip *= 2;
      if (num_markers / marker_skip <= 20)
	break;
      marker_skip *= 5;
    }

  /* In even markers * marker_skip. */
  start_in_marks = viewport->time_start / marker_data->marker_unit;
  end_in_marks = start_in_marks + num_markers;
  start_in_marks = ceil (start_in_marks / marker_skip);
  end_in_marks = floor (end_in_marks / marker_skip);

  if (start_in_marks > end_in_marks)
    return;

  marker_data->num_markers = (int) (end_in_marks - start_in_marks + 1);
  marker_data->markers = g_new (GdamBufferViewportMarker, marker_data->num_markers);
  for (i = 0; i < marker_data->num_markers; i++)
    {
      GdamBufferViewportMarker *mark = marker_data->markers + i;
      gdouble mark_number = (start_in_marks + i) * marker_skip;
      gdouble t = marker_data->marker_unit * mark_number;
      const char *fmt;
      int cent_x;
      int text_width;
      char *str;
      t -= viewport->time_start;
      cent_x = t * viewport->draw_width / viewport->time_length;
      if (marker_skip < 0.01)
	fmt = "%7.03f";
      else if (marker_skip < 0.1)
	fmt = "%6.02f";
      else
	fmt = "%5.01f";
      str = g_strdup_printf (fmt, (gdouble) mark_number);
      text_width = gdk_string_width (marker_data->marker_font, str);
      mark->draw_x = cent_x - text_width / 2;
      mark->draw_y = viewport->draw_height - 5;
      mark->draw_width = text_width;
      mark->text = str;
    }
}

static gboolean
marker_render   (GdamBufferViewport *viewport,
		 GdkDrawable        *drawable,
		 gpointer            data,
		 guint               x,
		 guint               y,
		 guint               width,
		 guint               height)
{
  GdamBufferViewportMarkerData *marker_data = data;
  int i;
  int num_markers;
  g_return_val_if_fail (marker_data->marker_font != NULL, FALSE);
  if (marker_data->marker_gc == NULL)
    {
      marker_data->marker_gc = gdk_gc_new (drawable);
      gdk_gc_set_foreground (marker_data->marker_gc, marker_data->marker_color);
    }

  if (marker_data->num_markers == 0
   || marker_data->old_time_start != viewport->time_start
   || marker_data->old_time_length != viewport->time_length
   || marker_data->old_draw_width != viewport->draw_width)
    compute_marker_locations (viewport, marker_data);

  num_markers = marker_data->num_markers;
  for (i = 0; i < num_markers; i++)
    {
      GdamBufferViewportMarker *mark = marker_data->markers + i;
      if (mark->draw_x >= (int) (x + width)
	  || mark->draw_x + mark->draw_width <= (int) x)
	continue;
      gdk_draw_string (drawable,
		       marker_data->marker_font,
		       marker_data->marker_gc,
		       mark->draw_x, mark->draw_y, mark->text);
    }
  return TRUE;
}

static void
marker_destroy (gpointer data)
{
  GdamBufferViewportMarkerData *marker_data = data;
  guint i;
  for (i = 0; i < marker_data->num_markers; i++)
    g_free (marker_data->markers[i].text);
  g_free (marker_data->markers);
  if (marker_data->marker_font != NULL)
    gdk_font_unref (marker_data->marker_font);
  if (marker_data->marker_gc != NULL)
    gdk_gc_unref (marker_data->marker_gc);
  if (marker_data->marker_format != NULL)
    g_free (marker_data->marker_format);
  g_free (marker_data);
}

static GdamBufferViewportAdornFuncs marker_funcs =
{
  marker_render,
  NULL,			/* get_cursor */
  NULL,			/* start_drag */
  NULL,			/* keep_dragging */
  NULL,			/* end_drag */
  marker_destroy
};

GdamBufferViewportAdornment *
gdam_buffer_viewport_add_markers (GdamBufferViewport *viewport,
				  gboolean            is_shown,
				  gboolean            is_under,
                                  GdkFont            *marker_font,
                                  GdkGC              *marker_gc,
                                  GdkColor           *marker_color,
                                  const char         *marker_format,
                                  gdouble             marker_unit)
{
  GtkWidget *widget = GDAM_SKIN_GET_WIDGET (viewport);
  GdamBufferViewportMarkerData *marker_data;
  if (marker_font == NULL)
    marker_font = gdk_font_load ("fixed");
  if (marker_gc != NULL)
    gdk_gc_ref (marker_gc);
  marker_data = g_new0 (GdamBufferViewportMarkerData, 1);
  marker_data->marker_font = marker_font;
  marker_data->marker_gc = marker_gc;
  if (marker_color != NULL)
    marker_data->marker_color = marker_color;
  else
    marker_data->marker_color = util_alloc_color_cached (0, 65535, 0);
  marker_data->marker_format = g_strdup (marker_format);
  marker_data->marker_unit = marker_unit;

  return gdam_buffer_viewport_add_adornment (viewport,
					     is_shown,
					     is_under,
					     &marker_funcs,
					     marker_data);
}
