/* GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambufferviewport.h"
#include <gdam/common/gdam-protocol.h>
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdambuffermanager.h>
#include <gdam/component/gdamutil.h>
#include <gtk/gtkmain.h>
#include "gdamskinutil.h"
#include <gtk/gtkeventbox.h>
#include <math.h>
#include <stdlib.h>
#include <gdk/gdktypes.h>
#include <gdk/gdkkeysyms.h>

enum
{
  SELECTION_CHANGED,
  VIEW_CHANGED,
  LENGTH_CHANGED,
  BUFFER_CHANGED,
  LAST_SIGNAL
};

static void gdam_buffer_viewport_real_beat_lines_changed
                                              (GdamBufferViewport* viewport);

static void  gdam_buffer_viewport_view_changed(GdamBufferViewport* viewport);
static void gdam_buffer_viewport_selection_changed(GdamBufferViewport *view);
gdouble gdam_buffer_viewport_get_sample_length(GdamBufferViewport* viewport);
void       gdam_buffer_viewport_set_length    (GdamBufferViewport* viewport,
                                               gdouble             length);
static void maybe_update_cursor 	      (GdamBufferViewport* viewport,
					       guint               state);
static guint viewport_signals[LAST_SIGNAL] = {0};
static void gdam_buffer_viewport_destroy(GtkObject* object);
static void gdam_buffer_viewport_finalize(GtkObject* object);
static GtkObjectClass* parent_class = NULL;

static void
gdam_buffer_viewport_init(GdamBufferViewport* buffer_viewport)
{
  buffer_viewport->draw_width = 256;
  buffer_viewport->draw_height = 256;
  buffer_viewport->time_start = 0.0;
  buffer_viewport->time_length = 1.0;
  buffer_viewport->has_selection = FALSE;
  buffer_viewport->selection_start = 0.0;
  buffer_viewport->selection_length = 1.0;
  buffer_viewport->is_selectable = TRUE;
  buffer_viewport->use_buffer_manager = TRUE;
  buffer_viewport->sample_length = 0.0;
  buffer_viewport->samples_per_second = 44100;
  buffer_viewport->density = 5;
  buffer_viewport->mode = g_strdup ("left");
  gdam_buffer_viewport_set_render_waveform (buffer_viewport, NULL, NULL);
}

static void
zoom_changed (GtkAdjustment      *adjustment,
              GdamBufferViewport *viewport)
{
  gdouble zoom = exp (adjustment->value);
  if (zoom > 10000.0)
    {
      return;
    }
  gdam_buffer_viewport_set_zoom (viewport, zoom);
  return;
}
static void viewport_zoom_changed(GdamBufferViewport *viewport,
                                  GtkAdjustment      *adjustment)
{
  gdouble zoom = viewport->time_length;
  if (zoom < 0.0001)
    return;
  gtk_adjustment_set_value (adjustment, log (zoom));
}

static gboolean
gdam_buffer_viewport_log_zoom_adjustment_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GtkAdjustment *adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed", GTK_SIGNAL_FUNC (zoom_changed), skin);
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin), "view_changed", 
     GTK_SIGNAL_FUNC (viewport_zoom_changed), adj, GTK_OBJECT (adj));
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void center_changed(GtkAdjustment      *adjustment,
                           GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_set_center (viewport, adjustment->value);
}

static void
viewport_center_changed(GdamBufferViewport *viewport,
			GtkAdjustment      *adjustment)
{
  adjustment->page_size = viewport->time_length;
  if (viewport->time_length < viewport->sample_length)
    {
      adjustment->page_increment = viewport->time_length / 10.0;
      adjustment->step_increment = viewport->time_length / 100.0;
    }
  else
    {
      adjustment->page_increment = viewport->sample_length / 10.0;
      adjustment->step_increment = viewport->sample_length / 100.0;
    }
  adjustment->upper = viewport->sample_length + adjustment->page_size;
  adjustment->value = viewport->time_start + viewport->time_length / 2;
  gtk_adjustment_changed (adjustment);
}

static gboolean
gdam_buffer_viewport_center_adjustment_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GtkAdjustment *adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (center_changed), skin);
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin),
     "length_changed", GTK_SIGNAL_FUNC (viewport_center_changed), 
     adj, GTK_OBJECT (adj));
  gtk_signal_connect_while_alive
    (GTK_OBJECT (skin),
     "view_changed", GTK_SIGNAL_FUNC (viewport_center_changed), 
     adj, GTK_OBJECT (adj));
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void
on_buffer_id_received (const char       *text,
		       gpointer          data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (data);
  int id = atoi (text);
  gdam_debug ("buffer id %d dropped", id);
  gdam_buffer_viewport_set_buffer_id (viewport, id);
}

static gboolean
gdam_buffer_viewport_buffer_id_dropped_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_buffer_id_received,
			       "gdam-buffer-id-text", skin);
  return TRUE;
}


static void gdam_buffer_viewport_zoom_to_selection(GdamBufferViewport *viewport)
{
  if (viewport->has_selection)
    {
      viewport->time_start = viewport->selection_start;
      viewport->time_length = viewport->selection_length;
      gdam_buffer_viewport_changed (viewport);
    }
}

static void gdam_buffer_viewport_select_all (GdamBufferViewport *viewport)
{
  gdouble length;

  if (viewport->buffer_id == 0)
    return;

  length = gdam_buffer_viewport_get_sample_length (viewport);
  if (length < 0.0001)
    return;

  gdam_buffer_viewport_set_selection (viewport, 0.0, length);
}

static void gdam_buffer_viewport_select_none (GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_set_selection (viewport, -1.0, -1.0);
}

static void 
gdam_buffer_viewport_select_visible_area (GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_set_selection (viewport, viewport->time_start, 
                                      viewport->time_length);
}

static void 
gdam_buffer_viewport_extend_selection_to_first_sample 
                                      (GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_set_selection 
    (viewport, 0, viewport->selection_length + viewport->selection_start);
}

static void
gdam_buffer_viewport_extend_selection_to_last_sample (GdamBufferViewport *viewport)
{
  gdouble length;

  if (!viewport->has_selection)
    return;
  if (viewport->buffer_id == 0)
    return;

  length = gdam_buffer_viewport_get_sample_length (viewport);
  if (length < 0.0001)
    return;
  gdam_buffer_viewport_set_selection (viewport, viewport->selection_start, 
                                      (length - viewport->selection_start));
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_buffer_viewport_log_zoom_adjustment",
           gdam_buffer_viewport_log_zoom_adjustment_handler	},
	{ "gdam_buffer_viewport_center_adjustment",
           gdam_buffer_viewport_center_adjustment_handler	},
	{ "gdam_buffer_viewport_buffer_id_dropped",
           gdam_buffer_viewport_buffer_id_dropped_handler	},
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_buffer_viewport_whole_sample",
           (GdamSkinFunc)gdam_buffer_viewport_whole_sample },
	{ "gdam_buffer_viewport_select_visible_area",
           (GdamSkinFunc)gdam_buffer_viewport_select_visible_area },
	{ "gdam_buffer_viewport_zoom_to_selection",
           (GdamSkinFunc)gdam_buffer_viewport_zoom_to_selection },
	{ "gdam_buffer_viewport_select_all",
           (GdamSkinFunc)gdam_buffer_viewport_select_all },
	{ "gdam_buffer_viewport_select_none",
           (GdamSkinFunc)gdam_buffer_viewport_select_none },
	{ "gdam_buffer_viewport_extend_selection_to_last_sample",
           (GdamSkinFunc)gdam_buffer_viewport_extend_selection_to_last_sample },
	{ "gdam_buffer_viewport_extend_selection_to_first_sample",
           (GdamSkinFunc)gdam_buffer_viewport_extend_selection_to_first_sample },
};

static void 
arg_get_zoom (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  GDAM_ARG_DOUBLE (arg) = viewport->time_length;
}

static void 
arg_set_zoom (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  gdouble zoom = GDAM_ARG_DOUBLE (arg);
  gdouble center;
  center = viewport->time_start + viewport->time_length / 2.0;
  viewport->time_start = center - zoom / 2.0;
  viewport->time_length = zoom;
  gdam_buffer_viewport_changed (viewport);
}

static void 
arg_get_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  GDAM_ARG_DOUBLE (arg) = viewport->time_start;
}

static void 
arg_set_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  viewport->time_start = GDAM_ARG_DOUBLE (arg);
  gdam_buffer_viewport_changed (viewport);
}

static void 
arg_set_nudge_time_start (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  viewport->time_start += GDAM_ARG_DOUBLE (arg) * viewport->time_length;
  gdam_buffer_viewport_changed (viewport);
}

static void
arg_set_render_mode (GtkObject *object, GdamArg *arg)
{
  const char *str = GDAM_ARG_STRING (arg);
  if (strcmp (str, "2d") == 0)
    gdam_buffer_viewport_set_render_2d_tmp (GDAM_BUFFER_VIEWPORT (object));
  else if (strcmp (str, "waveform") == 0)
    gdam_buffer_viewport_set_render_waveform (GDAM_BUFFER_VIEWPORT (object),
					      NULL,		/* fg_gc */
					      NULL		/* bg_gc */
					      );
  else
    g_assert_not_reached();	/* should have been caught by validator */
}

static void
arg_get_render_mode (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  GDAM_ARG_STRING (arg) = g_strdup (viewport->render_mode);
}

static void
arg_set_profile_mode (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  const char *string = GDAM_ARG_STRING (arg);
  gdam_buffer_viewport_set_mode (viewport, string);
}

static void
arg_get_profile_mode (GtkObject *object, GdamArg *arg)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  const char *mode = gdam_buffer_viewport_get_mode (viewport);
  GDAM_ARG_STRING (arg) = g_strdup (mode);
}

static void gdam_buffer_viewport_class_init(GdamBufferViewportClass* klass)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  GtkType type = object_class->type;
  object_class->destroy = gdam_buffer_viewport_destroy;
  object_class->finalize = gdam_buffer_viewport_finalize;
  parent_class = gtk_type_class (GDAM_TYPE_SKIN);
  viewport_signals[SELECTION_CHANGED]
    = gtk_signal_new ("selection_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 selection_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[VIEW_CHANGED]
    = gtk_signal_new ("view_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 view_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[LENGTH_CHANGED]
    = gtk_signal_new ("length_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 length_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  viewport_signals[BUFFER_CHANGED]
    = gtk_signal_new ("buffer_changed",
		      GTK_RUN_FIRST,
		      type,
		      GTK_SIGNAL_OFFSET (GdamBufferViewportClass,
					 buffer_changed),
		      gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, viewport_signals, LAST_SIGNAL);
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  gdam_arg_register (object_class,
		     "zoom",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_zoom,
		     arg_get_zoom,
		     NULL, "0.01,", GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "time_start",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_time_start,
		     arg_get_time_start,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "nudge_time_start",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_nudge_time_start,
		     NULL,
		     NULL, NULL, 0);
  gdam_arg_register (object_class,
		     "render_mode",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_render_mode,
		     arg_get_render_mode,
		     NULL,
		     "waveform,2d",
		     GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
                     "profile_mode", 
		     GDAM_ARG_TYPE_STRING, 
		     arg_set_profile_mode, 
		     arg_get_profile_mode, 
		     NULL,
		     "left,right,average,stereo,minmax,fft",
		     GDAM_ARG_VITAL);
  klass->view_changed = gdam_buffer_viewport_view_changed;
}

/*=====================================================================*
 *                           get_x_from_time                           *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the viewport with the coordinate frame.       *
 *     time              the time in seconds to convert to screen      *
 *                       coordinates.                                  *
 *=====================================================================*/
static inline int     get_x_from_time(GdamBufferViewport* viewport, 
                                      gdouble             time)
{
  time -= viewport->time_start;
  if (viewport->time_length > 0.0001)
    time /= viewport->time_length;
  time *= viewport->draw_width;
  return (int) time;
}

/*=====================================================================*
 *                           get_time_from_x                           *
 *                                                                     *
 * Parameters:                                                         *
 *     viewport          the viewport.                                 *
 *     x                 the screen coordinate (in the window)         *
 *     return            the time into the sample in seconds.          *
 *=====================================================================*/
static inline gdouble get_time_from_x(GdamBufferViewport* viewport, 
                                      int                 screen_x)
{
  gdouble time = screen_x;
  time /= viewport->draw_width;
  time *= viewport->time_length;
  time += viewport->time_start;
  return time;
}

GtkType gdam_buffer_viewport_get_type()
{
  static GtkType buffer_viewport_type = 0;
  if (!buffer_viewport_type)
    {
      static const GtkTypeInfo buffer_viewport_info = {
	"GdamBufferViewport",
	sizeof (GdamBufferViewport),
	sizeof (GdamBufferViewportClass),
	(GtkClassInitFunc) gdam_buffer_viewport_class_init,
	(GtkObjectInitFunc) gdam_buffer_viewport_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      buffer_viewport_type = gtk_type_unique (parent, &buffer_viewport_info);
    }
  return buffer_viewport_type;
}

GdamBufferViewport *
gdam_buffer_viewport_new_shown(GdamChannel *channel)
{
  GdamBufferViewport *view;
  GtkType type = GDAM_TYPE_BUFFER_VIEWPORT;
  view = GDAM_BUFFER_VIEWPORT (gtk_type_new (type));
  GDAM_SKIN (view)->channel = channel;
  gdam_buffer_viewport_new_event_box (view);
  gtk_widget_show (GDAM_SKIN (view)->main_widget);
  return view;
}

static void
update_buffer_info(GdamBufferManagerInfo        *info,
		   gpointer                      viewport_ptr)
{
  GdamBufferViewport *viewport;
  gdouble length;
  viewport = GDAM_BUFFER_VIEWPORT (viewport_ptr);
  length = (gdouble) info->length / viewport->samples_per_second;
  gdam_buffer_viewport_set_length (viewport, length);
}

void
gdam_buffer_viewport_set_buffer_id (GdamBufferViewport *viewport,
				    guint               buffer_id)
{
  gdouble length;

  viewport->buffer_id = buffer_id;
  length = gdam_buffer_viewport_get_sample_length (viewport);
  gdam_buffer_viewport_set_length (viewport, length);
  gdam_buffer_viewport_view_changed (viewport);
  if (viewport->use_buffer_manager)
    {
      if (viewport->connection_id == 0)
	viewport->connection_id
	  = gdam_buffer_manager_connect (buffer_id,
					 update_buffer_info, viewport);
      else
	gdam_buffer_manager_reconnect (viewport->connection_id, buffer_id);
    }
  gtk_signal_emit (GTK_OBJECT (viewport), viewport_signals[BUFFER_CHANGED]);
}

void
gdam_buffer_viewport_set_view      (GdamBufferViewport *viewport,
				    gdouble             start,
				    gdouble             length)
{
  if (viewport->time_start == start && viewport->time_length == length)
    return;
  viewport->time_start = start;
  viewport->time_length = length;
  gdam_debug ("gdam_buffer_viewport_set_view: start=%4.4g, length=%4.4g",
	      start, length);

  /* Is this okay? */
  gdam_buffer_viewport_changed (viewport);
}

gdouble
gdam_buffer_viewport_get_sample_length(GdamBufferViewport *viewport)
{
  if (viewport->use_buffer_manager)
    {
      GdamBufferManagerInfo *info;
      info = gdam_buffer_manager_info_by_id (viewport->buffer_id);
      if (info == NULL)
	{
	  g_warning ("buffer_viewport: manager has no info "
		     "on buffer %d", viewport->buffer_id);
	  return 0;
	}
      return (gdouble) info->length / viewport->samples_per_second;
    }
  else
    {
      return (gdouble) viewport->sample_length;
    }
}

void
gdam_buffer_viewport_whole_sample  (GdamBufferViewport *viewport)
{
  gdouble length;

  if (viewport->buffer_id == 0)
    return;

  length = gdam_buffer_viewport_get_sample_length (viewport);
  if (length < 0.0001)
    return;
  gdam_buffer_viewport_set_view (viewport, 0.0, length);
}

void       
gdam_buffer_viewport_set_zoom (GdamBufferViewport *viewport,
                               gdouble             zoom)
{
  gdam_arg_set (GTK_OBJECT (viewport), "zoom", zoom);
}

void
gdam_buffer_viewport_set_center    (GdamBufferViewport *viewport,
				    gdouble             center)
{
  viewport->time_start = center - viewport->time_length / 2.0;
  gdam_buffer_viewport_changed (viewport);
  /* TODO: update position adjustment? */
}

void
gdam_buffer_viewport_set_mode (GdamBufferViewport *viewport,
                               const char         *string)
{
  char *new_mode;

  /* TODO: validate input */

  new_mode = g_strdup (string);
  g_free (viewport->mode);
  viewport->mode = new_mode;

  gdam_buffer_viewport_changed (viewport);
}

const char *
gdam_buffer_viewport_get_mode (GdamBufferViewport *viewport)
{
  return viewport->mode;
}

/* --- selections --- */
static void
gdam_buffer_viewport_selection_changed(GdamBufferViewport *viewport)
{
  GtkObject *object = GTK_OBJECT (viewport);
#if 0
  if (gtk_signal_n_emissions (object, viewport_signals[SELECTION_CHANGED]))
    return;
#endif
  gtk_signal_emit (object, viewport_signals[SELECTION_CHANGED]);

  /* TODO: do partial redraw here! */
  gdam_buffer_viewport_queue_redraw (viewport);
}

void
gdam_buffer_viewport_set_selection (GdamBufferViewport *viewport,
				    gdouble             start,
				    gdouble             length)
{
  if (start < 0 && length < 0)
    {
      viewport->has_selection = FALSE;
      viewport->selection_start = 0;
      viewport->selection_length = 0;
    }
  else
    {
      viewport->has_selection = TRUE;
      viewport->selection_start = start;
      viewport->selection_length = length;
    }
  gdam_buffer_viewport_selection_changed (viewport);
}

/* --- redrawing --- */
static inline gboolean
maybe_allocate_backbuffer_raw (GdamBufferViewport *viewport)
{
  if (viewport->backbuffer == NULL)
    {
      GtkWidget *widget = GDAM_SKIN_GET_WIDGET (viewport);
      if (widget == NULL || widget->window == NULL)
	return FALSE;
      viewport->backbuffer = gdk_pixmap_new (widget->window,
					     viewport->draw_width,
					     viewport->draw_height,
					     gdk_visual_get_best_depth ());
      return TRUE;
    }
  return FALSE;
}

static void
rerender_part_of_backbuffer (GdamBufferViewport *viewport,
			     guint               x,
			     guint               y,
			     guint               width,
			     guint               height)
{
  GdamBufferViewportAdornment *at;
  gdouble sel_time_end = viewport->selection_start + viewport->selection_length;
  gint selection_start = get_x_from_time (viewport, viewport->selection_start);
  gint selection_end = get_x_from_time (viewport, sel_time_end);
  GdamBufferViewportRenderInfo render_info;
  gpointer render_state = NULL;
  gboolean has_render_state = FALSE;
  gboolean try_render;

  maybe_allocate_backbuffer_raw (viewport);
  g_return_if_fail (viewport->backbuffer != NULL);

  render_info.drawable = viewport->backbuffer;
  render_info.data = viewport->render_data;
  render_info.viewport = viewport;
  render_info.is_double = viewport->last_samples_is_double;
  render_info.num_columns = viewport->last_num_samples;
  render_info.num_rows = viewport->last_num_channels;
  render_info.values = viewport->last_samples;
  render_info.x = x;
  render_info.y = y;
  render_info.width = width;
  render_info.height = height;
  render_info.draw_width = viewport->draw_width;
  render_info.draw_height = viewport->draw_height;
  render_info.has_selection = viewport->has_selection;
  render_info.selection_start = selection_start;
  render_info.selection_end = selection_end;
  render_info.valid_output_start = viewport->last_valid_output_start;
  render_info.valid_output_end = viewport->last_valid_output_end;

  if (viewport->last_num_channels == 0
   || viewport->last_num_samples == 0
   || viewport->last_samples == NULL)
    try_render = FALSE;
  else
    try_render = TRUE;

  if (try_render && viewport->render_init != NULL)
    {
      render_state = (*viewport->render_init) (&render_info);
      g_return_if_fail (render_state != NULL);
      has_render_state = TRUE;
    }

  if (try_render && viewport->render_bg != NULL)
    (*viewport->render_bg) (&render_info, render_state);
  else
    gdk_draw_rectangle (viewport->backbuffer,
			GDAM_SKIN_GET_WIDGET (viewport)->style->black_gc,
			TRUE,
			x, y, width, height);

  for (at = viewport->first_adornment;
       at != NULL;
       at = at->next)
    if (at->is_shown && at->is_under)
      (*at->funcs->render) (viewport,
			    viewport->backbuffer,
			    at->data,
			    x, y, width, height);

  if (try_render && viewport->render_fg != NULL)
      (*viewport->render_fg) (&render_info, render_state);

  for (at = viewport->first_adornment;
       at != NULL;
       at = at->next)
    if (at->is_shown && !at->is_under)
      (*at->funcs->render) (viewport,
			    viewport->backbuffer,
			    at->data,
			    x, y, width, height);

  if (has_render_state && viewport->render_state_destroy != NULL)
    (*viewport->render_state_destroy) (render_state);
}

static inline void
maybe_allocate_backbuffer (GdamBufferViewport *viewport)
{
  if (maybe_allocate_backbuffer_raw (viewport))
    rerender_part_of_backbuffer (viewport,
				 0, 0,
			         viewport->draw_width, viewport->draw_height);
}

static void
blt_backbuffer_rect(GdamBufferViewport *viewport,
		    guint               x,
		    guint               y,
		    guint               width,
		    guint               height)
{
  GtkWidget *widget = GDAM_SKIN_GET_WIDGET (viewport);
  if (widget->window == NULL)
    return;
  maybe_allocate_backbuffer (viewport);
  if (viewport->backbuffer == NULL)
    return;
  gdk_draw_pixmap (widget->window,
		   widget->style->fg_gc[0],
		   viewport->backbuffer,
		   x, y,
		   x, y,
		   width, height);
}

static void
rerender_backbuffer_to_main_widget (GdamBufferViewport *viewport)
{
  blt_backbuffer_rect (viewport,
		       0, 0, viewport->draw_width, viewport->draw_height);
}

static void
rerender_to_backbuffer (GdamBufferViewport *viewport)
{
  rerender_part_of_backbuffer (viewport, 0, 0,
			       viewport->draw_width, viewport->draw_height);
  rerender_backbuffer_to_main_widget (viewport);
}

static gboolean
redraw_idle_func (gpointer data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (data);
  viewport->redraw_idle_thread = 0;
  rerender_to_backbuffer (viewport);
  return FALSE;
}

void
gdam_buffer_viewport_queue_redraw (GdamBufferViewport *viewport)
{
  viewport->redraw_x = 0;
  viewport->redraw_y = 0;
  viewport->redraw_width = viewport->draw_width;
  viewport->redraw_height = viewport->draw_height;

  if (viewport->redraw_idle_thread != 0)
    return;
  viewport->redraw_idle_thread
    = g_idle_add_full (G_PRIORITY_LOW, redraw_idle_func, viewport, NULL);
}

void
gdam_buffer_viewport_queue_redraw_part (GdamBufferViewport *viewport,
					guint               x,
					guint               y,
					guint               width,
					guint               height)
{
  if (viewport->redraw_idle_thread != 0)
    {
      guint end_x = viewport->redraw_x + viewport->redraw_width;
      guint end_y = viewport->redraw_y + viewport->redraw_height;
      if (viewport->redraw_x > x)
	viewport->redraw_x = x;
      if (viewport->redraw_y > y)
	viewport->redraw_y = y;
      if (end_x < x + width)
	end_x = x + width;
      if (end_y < y + height)
	end_y = y + height;
      viewport->redraw_width = end_x - viewport->redraw_x;
      viewport->redraw_height = end_y - viewport->redraw_y;
    }
  else
    {
      viewport->redraw_x = x;
      viewport->redraw_y = y;
      viewport->redraw_width = width;
      viewport->redraw_height = height;
      viewport->redraw_idle_thread
	= g_idle_add_full (G_PRIORITY_LOW, redraw_idle_func, viewport, NULL);
    }
}

/* --- cursors --- */
static void
maybe_update_cursor (GdamBufferViewport *viewport,
		     guint               state)
{
  GdamCursorType cursor = GDAM_CURSOR_INVALID;

  GdamBufferViewportAdornment *adornment;
  for (adornment = viewport->last_adornment;
       adornment != NULL && cursor == GDAM_CURSOR_INVALID;
       adornment = adornment->prev)
    if (adornment->funcs->get_cursor != NULL)
      cursor = (*adornment->funcs->get_cursor)(viewport,
					       adornment->data);


  if (cursor == GDAM_CURSOR_INVALID)
    {
      /* Else beat locations are derived from bpm and firstbeat */
      if ((GDK_SHIFT_MASK & state) == GDK_SHIFT_MASK)
	cursor = GDAM_PHASE_CURSOR;
      else if ((GDK_CONTROL_MASK & state) == GDK_CONTROL_MASK)
	cursor = GDAM_FIRSTBEAT_CURSOR;
      else if ((GDK_MOD1_MASK & state) == GDK_MOD1_MASK)
	cursor = GDAM_INDEX_CURSOR;
      else
	cursor = GDAM_BEATLINE_CURSOR;
    }

  if (cursor == viewport->cursor)
    return;
  
  g_return_if_fail (GDAM_SKIN (viewport)->main_widget->window != NULL);
  gdam_set_window_cursor (GDAM_SKIN (viewport)->main_widget->window, cursor);
  viewport->cursor = cursor;
}


/* Callback for when the server responds to our request for samples. */
static void
failed_to_get_samples(gpointer user_data)
{
  g_warning ("failed to get buffer's samples");
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void 
got_samples (int      num_args,
	     GdamArg *args,
	     gpointer user_data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  GdamArg *arg;
  gpointer array;
  guint length;
  gpointer our_array;
  gboolean is_double;

  if (num_args == 1 && args[0].type == GDAM_ARG_TYPE_STRING
      && strcmp (args[0].name, "error") == 0)
    {
      g_warning ("got error from GdamCacheBuffer::profile: %s",
		 GDAM_ARG_STRING (args+0));
      failed_to_get_samples (viewport);
      return;
    }

  arg = gdam_arg_scan (num_args, args, "valid_output_start", GDAM_ARG_TYPE_INT);
  g_return_if_fail (arg != NULL);
  viewport->last_valid_output_start = GDAM_ARG_INT (arg);
  arg = gdam_arg_scan (num_args, args, "valid_output_end", GDAM_ARG_TYPE_INT);
  g_return_if_fail (arg != NULL);
  viewport->last_valid_output_end = GDAM_ARG_INT (arg);
  arg = gdam_arg_scan (num_args, args, "num_channels", GDAM_ARG_TYPE_INT);
  g_return_if_fail (arg != NULL);
  viewport->last_num_channels = GDAM_ARG_INT (arg);
  arg = gdam_arg_scan (num_args, args, "num_outputs", GDAM_ARG_TYPE_INT);
  g_return_if_fail (arg != NULL);
  viewport->last_num_samples = GDAM_ARG_INT (arg);
  arg = gdam_arg_scan (num_args, args, "values", GDAM_ARG_TYPE_INT_ARRAY);
  if (arg != NULL)
    {
      length = GDAM_ARG_INT_ARRAY (arg).length;
      array = GDAM_ARG_INT_ARRAY (arg).array;
      our_array = g_memdup (array, length * sizeof (int));
      is_double = FALSE;
    }
  else
    {
      arg = gdam_arg_scan (num_args, args, "values", GDAM_ARG_TYPE_DOUBLE_ARRAY);
      g_return_if_fail (arg != NULL);
      length = GDAM_ARG_DOUBLE_ARRAY (arg).length;
      array = GDAM_ARG_DOUBLE_ARRAY (arg).array;
      our_array = g_memdup (array, length * sizeof (gdouble));
      is_double = TRUE;
    }

  viewport->last_num_samples = GDAM_ARG_INT (arg);
  viewport->last_samples = our_array;
  viewport->last_samples_is_double = is_double;

  {
    gboolean must_redo = viewport->redo_query;
    viewport->getting_profile = FALSE;
    viewport->redo_query = FALSE;

    if (must_redo)
      gdam_buffer_viewport_changed (viewport);
  }

  gdam_buffer_viewport_queue_redraw (viewport);
  gtk_object_unref (GTK_OBJECT (viewport));
}

/* --- change notification --- */
static void 
gdam_buffer_viewport_view_changed (GdamBufferViewport *viewport)
{
  GdamRequest *request;
  int samps_per_sec, num_points;
  double start, length;
  if (viewport->buffer_id == 0)
    {
      gdam_buffer_viewport_queue_redraw (viewport);
      return;
    }
  if (viewport->getting_profile)
    {
      viewport->redo_query = TRUE;
      return;
    }
  gtk_object_ref (GTK_OBJECT (viewport));
  viewport->getting_profile = TRUE;
  samps_per_sec = viewport->samples_per_second;

  start = viewport->time_start;
  length = viewport->time_length;

  num_points = viewport->draw_width;
  if (viewport->mode == NULL || viewport->mode[0] != 'f'
   || viewport->mode[1] != 'f' || viewport->mode[2] != 't')
    num_points *= viewport->density;

#if 0
  /* assumes that start + length > 0 */
  if (start < 0)
    {
      double percent_neg = start / length;
      double percent_pos = 1.0 - percent_neg;
      num_points *= percent_pos;
      length *= percent_pos;
      start = 0;
    }
#endif
  {
    GdamArg profile_args[5];
    profile_args[0].name = "start";
    profile_args[0].type = GDAM_ARG_TYPE_INT64;
    GDAM_ARG_INT64 (profile_args+0) = (gint64) (start * samps_per_sec);
    profile_args[1].name = "length";
    profile_args[1].type = GDAM_ARG_TYPE_INT64;
    GDAM_ARG_INT64 (profile_args+1) = (gint64) (length * samps_per_sec);
    profile_args[2].name = "num_outputs";
    profile_args[2].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (profile_args+2) = num_points;
    profile_args[3].name = "mode";
    profile_args[3].type = GDAM_ARG_TYPE_STRING;
    GDAM_ARG_STRING (profile_args+3) = viewport->mode;
    profile_args[4].name = "use_doubles";
    profile_args[4].type = GDAM_ARG_TYPE_INT;
    GDAM_ARG_INT (profile_args+4) = 0;
    request = gdam_request_call_function (GDAM_PROTOCOL_BUFFER_ARG,
					  viewport->buffer_id,
					  "profile",
					  5, profile_args,
					  got_samples,
					  failed_to_get_samples,
					  viewport);
  }

  gdam_channel_append_request (GDAM_SKIN (viewport)->channel, request);
}

void
gdam_buffer_viewport_buffer_changed(GdamBufferViewport *viewport)
{
  gdam_buffer_viewport_view_changed (viewport);
}

void
gdam_buffer_viewport_changed(GdamBufferViewport *viewport)
{
  GtkObject *object = GTK_OBJECT (viewport);
#if 0
  if (gtk_signal_n_emissions (object, viewport_signals[VIEW_CHANGED]))
    return;
#endif
  gtk_signal_emit (object, viewport_signals[VIEW_CHANGED]);
}

/* --- window events --- */
static void
gdam_buffer_viewport_resized (GdamBufferViewport *viewport)
{
  if (viewport->backbuffer != NULL)
    {
      gdk_pixmap_unref (viewport->backbuffer);
      viewport->backbuffer = NULL;
    }
  maybe_allocate_backbuffer (viewport);
  rerender_backbuffer_to_main_widget (viewport);
}

static void
buffer_viewport_expose   (GtkEventBox        *event_box,
			  GdkEventExpose     *event,
		          GdamBufferViewport *viewport)
{
  rerender_backbuffer_to_main_widget (viewport);
}

static gint
buffer_viewport_configure      (GtkEventBox        *event_box,
				GdkEventConfigure  *config_event,
				GdamBufferViewport *viewport)
{
  if (viewport->draw_width == config_event->width
   && viewport->draw_height == config_event->height)
    return FALSE;
  viewport->draw_width = config_event->width;
  viewport->draw_height = config_event->height;
  gdam_buffer_viewport_resized (viewport);
  return FALSE;
}

/* --- Keyboard handling --- */
static int
key_to_state (gint key)
{
  switch (key)
    {
      case GDK_Alt_L:
      case GDK_Alt_R:
	return GDK_MOD1_MASK;
      case GDK_Shift_L:
      case GDK_Shift_R:
	return GDK_SHIFT_MASK;
      case GDK_Control_L:
      case GDK_Control_R:
	return GDK_CONTROL_MASK;
      default:
	return 0;
    }
}

static gint
buffer_viewport_key_press      (GtkEventBox        *event_box,
				GdkEventKey        *event,
				GdamBufferViewport *viewport)
{
  guint state = 0;
  state = event->state;

  if (event->keyval == GDK_Alt_L     || event->keyval == GDK_Alt_R ||
      event->keyval == GDK_Shift_L   || event->keyval == GDK_Shift_R ||
      event->keyval == GDK_Control_L || event->keyval == GDK_Control_R)
    {
      state |= key_to_state (event->keyval);
    }
  maybe_update_cursor (viewport, state);
  return 0;
}

static gint
buffer_viewport_key_release    (GtkEventBox        *event_box,
				GdkEventKey        *event,
				GdamBufferViewport *viewport)
{
  guint state = 0;
  state = event->state;

  if (event->keyval == GDK_Alt_L     || event->keyval == GDK_Alt_R ||
      event->keyval == GDK_Shift_L   || event->keyval == GDK_Shift_R ||
      event->keyval == GDK_Control_L || event->keyval == GDK_Control_R)
    {
      state &= ~key_to_state (event->keyval);
    }
  maybe_update_cursor (viewport, state);
  return TRUE;
}

/* --- Mouse handling --- */
static gint
buffer_viewport_enter          (GtkEventBox        *event_box,
				GdkEventCrossing   *event,
				GdamBufferViewport *viewport)
{
  gtk_widget_grab_focus (GTK_WIDGET (event_box));
  return 0;
}

static gint
buffer_viewport_press          (GtkEventBox        *event_box,
				GdkEventButton     *event,
				GdamBufferViewport *viewport)
{
  int beat_index;
  GdamBufferViewportAdornment *adornment;
  GdamBufferViewportDragInfo *drag_info = &viewport->drag_info;

  /* umm... don't we have to relinquish the grab??? */
  gtk_grab_add (GDAM_SKIN (viewport)->main_widget);

  /* ???: Historically, events with certain
   * buttons and masks are left for others to handle.
   */
  if (event->button == 1 && 
      (GDK_MOD1_MASK & event->state) == GDK_MOD1_MASK)
    return 0;

  /* ignore multiple buttons etc. */
  if (viewport->is_dragging)
    return 1;

  drag_info->start_x = drag_info->end_x = drag_info->last_x = event->x;
  drag_info->start_y = drag_info->end_y = drag_info->last_y = event->y;
  drag_info->modifiers = event->state;
  drag_info->button = event->button;
  drag_info->start_time = get_time_from_x (viewport, event->x);
  drag_info->last_time = drag_info->end_time = drag_info->start_time;

  /* Scan through the adornments in reverse order to see
   * if any of them are interested in the event.
   */
  for (adornment = viewport->last_adornment;
       adornment != NULL;
       adornment = adornment->prev)
    {
      if (adornment->funcs->start_drag != NULL
       && (*adornment->funcs->start_drag) (adornment, drag_info))
	{
	  viewport->is_dragging = TRUE;
	  viewport->dragging_adornment = adornment;
	  return 1;
	}
    }


  if (event->button != 1 && event->button != 2)
    return 0;

  if (viewport->is_selectable)
    {
      viewport->is_dragging = TRUE;
      viewport->dragging_adornment = NULL;
      viewport->has_selection = TRUE;
      viewport->selection_start = drag_info->start_time;
      viewport->selection_length = 0.0;
      gdam_buffer_viewport_selection_changed (viewport);
      return 1;
    }
  return 0;
}

static void
extend_selection_to_x (GdamBufferViewport *viewport,
		       gint                x)
{
  int start_x, end_x;
  GdamBufferViewportDragInfo *drag_info = &viewport->drag_info;
  gdouble start_time = drag_info->start_time;
  gdouble end_time = drag_info->end_time;
  if (start_time < end_time)
    {
      viewport->selection_start = start_time;
      viewport->selection_length = end_time - start_time;
    }
  else
    {
      viewport->selection_start = end_time;
      viewport->selection_length = start_time - end_time;
    }
  gdam_buffer_viewport_selection_changed (viewport);
}

static gboolean
scroll_thread (gpointer data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (data);
  GtkWidget *widget = GDAM_SKIN_GET_WIDGET (data);
  gint delta = 0;
  gint x, y;
  guint state;
  gdk_window_get_pointer (widget->window, &x, &y, &state);

  if (x < 0)
    delta = x;
  else if (x > viewport->draw_width)
    delta = x - viewport->draw_width;
  if ((state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) == 0
   || delta == 0)
    {
      viewport->scroll_timeout = 0;
      return FALSE;
    }
  viewport->time_start += viewport->time_length * delta / viewport->draw_width;
  /* XXX: a similar notification is needed for all adornments! */
  if (viewport->has_selection)
    {
      if (x < 0)
	{
	  gdouble d = viewport->selection_start - viewport->time_start;
	  viewport->selection_start -= d;
	  viewport->selection_length += d;
	}
      else
	{
	  viewport->selection_length = viewport->time_start
	                             + viewport->time_length
	                             - viewport->selection_start;
	}
      gdam_buffer_viewport_selection_changed (viewport);
    }
  gdam_buffer_viewport_changed (viewport);
  return TRUE;
}

static gint
buffer_viewport_motion         (GtkEventBox        *event_box,
				GdkEventMotion     *event,
				GdamBufferViewport *viewport)
{
  int x, y;
  int orig_x;
  guint state;
  GdamBufferViewportDragInfo *drag_info = &viewport->drag_info;

  maybe_update_cursor (viewport, event->state);
  
  if (!viewport->is_dragging)
    return 0;

  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
    }
  orig_x = x;
  x = CLAMP (x, 0, viewport->draw_width - 1);

  drag_info->last_x = drag_info->end_x;
  drag_info->last_y = drag_info->end_y;
  drag_info->last_time = drag_info->end_time;

  drag_info->end_x = x;
  drag_info->end_y = y;
  drag_info->end_time = get_time_from_x (viewport, x);

  drag_info->modifiers = event->state;

  if ((orig_x >= viewport->draw_width || orig_x < 0)
   && viewport->scroll_timeout == 0)
    viewport->scroll_timeout = g_timeout_add (100, scroll_thread, viewport);

  if (viewport->dragging_adornment != NULL)
    {
      GdamBufferViewportAdornment *adornment = viewport->dragging_adornment;
      if (adornment->funcs->keep_dragging == NULL)
	return 1;
      if (!(*adornment->funcs->keep_dragging) (adornment, drag_info))
	{
	  viewport->is_dragging = FALSE;
	  viewport->dragging_adornment = NULL;
	}
      return 1;
    }
  extend_selection_to_x (viewport, x);
  return 1;
}


static gint
buffer_viewport_release        (GtkEventBox        *event_box,
				GdkEventButton     *event,
				GdamBufferViewport *viewport)
{
  int x, y;
  GdamBufferViewportDragInfo *drag_info = &viewport->drag_info;
  gtk_grab_remove (GDAM_SKIN (viewport)->main_widget);

  if (!viewport->is_dragging)
    return 0;

  x = CLAMP (event->x, 0, viewport->draw_width - 1);
  y = event->y;

  drag_info->last_x = drag_info->end_x;
  drag_info->last_y = drag_info->end_y;
  drag_info->last_time = drag_info->end_time;

  drag_info->end_x = x;
  drag_info->end_y = y;
  drag_info->end_time = get_time_from_x (viewport, x);

  if (viewport->dragging_adornment != NULL)
    {
      GdamBufferViewportAdornment *adornment = viewport->dragging_adornment;
      if (adornment->funcs->end_drag != NULL)
	(*adornment->funcs->end_drag) (adornment, drag_info);
      viewport->is_dragging = FALSE;
      viewport->dragging_adornment = NULL;
      return 1;
    }

  if (event->button != 1 && event->button != 2)
    return 0;

  extend_selection_to_x (viewport, x);
  viewport->is_dragging = FALSE;
  return 1;
}

/* --- General window event handling --- */
static void
buffer_viewport_size_request   (GtkEventBox        *event_box,
				GtkRequisition     *requisition,
				GdamBufferViewport *viewport)
{
  requisition->width = viewport->draw_width;
  requisition->height = viewport->draw_height;
}

static void
buffer_viewport_size_allocate  (GtkEventBox        *event_box,
				GtkAllocation      *allocation,
				GdamBufferViewport *viewport)
{
  if (allocation->width == viewport->draw_width
   && allocation->height == viewport->draw_height)
    return;
  viewport->draw_width = allocation->width;
  viewport->draw_height = allocation->height;
  gdam_buffer_viewport_resized (viewport);
}

static void
buffer_viewport_realize        (GtkWidget          *widget,
                                GdamBufferViewport *viewport)
{
  g_return_if_fail (widget->window != NULL);
  gdam_set_window_cursor (widget->window, GDAM_SELECT_CURSOR);
}

void
gdam_buffer_viewport_new_event_box (GdamBufferViewport *viewport)
{
  GtkEventBox *event_box = GTK_EVENT_BOX (gtk_event_box_new ());
  GtkWidget *widget = GTK_WIDGET (event_box);
  GtkObject *object = GTK_OBJECT (event_box);
  gtk_widget_set_name (widget, "buffer_viewport_widget");
  gtk_widget_add_events (widget, GDK_EXPOSURE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_KEY_PRESS_MASK    | GDK_KEY_RELEASE_MASK
			 | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);

  gtk_signal_connect_after (object,
		      "expose_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_expose),
		      viewport);
  gtk_signal_connect (object,
		      "configure_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_configure), viewport);
  gtk_signal_connect (object,
		      "button_press_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_press), viewport);
  gtk_signal_connect (object,
		      "button_release_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_release), viewport);
  gtk_signal_connect (object,
		      "key_press_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_key_press), viewport);
  gtk_signal_connect (object,
		      "key_release_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_key_release), viewport);
  gtk_signal_connect (object,
		      "enter_notify_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_enter), viewport);
  gtk_signal_connect (object,
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (buffer_viewport_motion), viewport);
  gtk_signal_connect (object,
		      "size_request",
		      GTK_SIGNAL_FUNC (buffer_viewport_size_request),
		      viewport);
  gtk_signal_connect_object (object, "draw_focus",
			   GTK_SIGNAL_FUNC (rerender_backbuffer_to_main_widget),
			     (GtkObject *) viewport);
  gtk_signal_connect_object (object, "draw",
			   GTK_SIGNAL_FUNC (rerender_backbuffer_to_main_widget),
			     (GtkObject *) viewport);
  gtk_signal_connect (object,
		      "size_allocate",
		      GTK_SIGNAL_FUNC (buffer_viewport_size_allocate),
		      viewport);
  gtk_signal_connect (object,
		      "realize",
		      GTK_SIGNAL_FUNC (buffer_viewport_realize), viewport);

  GDAM_SKIN (viewport)->main_widget = widget;
}

static void
gdam_buffer_viewport_destroy(GtkObject *object)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);

  gdam_buffer_viewport_maybe_uncache (viewport);
  if (viewport->connection_id != 0)
    {
      gdam_buffer_manager_disconnect (viewport->connection_id);
      viewport->connection_id = 0;
    }

  gdam_buffer_viewport_destroy_all_adornments (viewport);
  if (viewport->backbuffer != NULL)
    {
      gdk_pixmap_unref (viewport->backbuffer);
      viewport->backbuffer = NULL;
    }
  if (viewport->scroll_timeout != 0)
    {
      g_source_remove (viewport->scroll_timeout);
      viewport->scroll_timeout = 0;
    }
  (*parent_class->destroy) (object);
}

static void
gdam_buffer_viewport_finalize(GtkObject *object)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (object);
  if (viewport->render_data_destroy != NULL)
    (*viewport->render_data_destroy) (viewport->render_data);
  g_free (viewport->render_mode);
  if (viewport->scroll_timeout != 0)
    g_source_remove (viewport->scroll_timeout);
  (*parent_class->finalize) (object);
}

void
gdam_buffer_viewport_set_length    (GdamBufferViewport *viewport,
				    gdouble             length)
{
  viewport->sample_length = length;
  gtk_signal_emit (GTK_OBJECT (viewport), viewport_signals[LENGTH_CHANGED]);
}

void
gdam_buffer_viewport_maybe_uncache(GdamBufferViewport *viewport)
{
  g_return_if_fail (viewport != NULL);
  if (viewport->buffer_id == 0)
    return;
  gdam_channel_append_request (viewport->skin.channel,
			       gdam_request_unref (GDAM_PROTOCOL_BUFFER_ARG,
						   viewport->buffer_id,
						   NULL, NULL, NULL));
}

