/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamchatskin.h"
#include <gdam/skin/gdamskinutil.h>
#include <stdlib.h>


/* Send synchronization info to another client. */
static void
send_sync_info (GdamChatSkin *skin, const char *recipient)
{
  GdamModel *model;
  GdamBeatInfo beat_info;
  guint lead_id;
  char *message;

  model = GDAM_SKIN (skin)->model;

  g_return_if_fail (skin->username != NULL);
  g_return_if_fail (strcmp ("", skin->username) != 0);
  g_return_if_fail (model != NULL);

  if (gdam_infomanager_get_sync_info (model, &lead_id, &beat_info))
    {
      message = g_strdup_printf ("%d:%f:%f", lead_id, 
	  			 gdam_beat_info_get_bpm (&beat_info, 0),
	  			 gdam_beat_info_get_first_beat (&beat_info, 0));
    }
  else
    {
      g_warning ("couldn't get sync info in chat skin");
      return;
    }

  gdam_channel_send_chat_message (GDAM_SKIN (skin)->channel, 
				  recipient,
				  "sync",
				  message);
  g_free (message);
}


static void
print_chat_messages (const char *sender,
		     const char *command,
		     const char *text,
		     gpointer    ptr)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (ptr);

  /* append new message to history window. */
  if (g_strcasecmp (command, "chat") == 0 && skin->log_entry)
    {
      gint pos;
      char *msg = g_strdup_printf ("%s: %s\n", sender, text);
      pos = gtk_editable_get_position (skin->log_entry);
      gtk_editable_insert_text (skin->log_entry, msg, strlen (msg), &pos);
    }
  else if (g_strcasecmp (command, "sync") == 0)
    {
      GdamBeatInfo *beat_info;
      GdamModel *model;
      char **chunks;
      model = GDAM_SKIN (skin)->model;
      chunks = g_strsplit (text, ":", 0);

      if (!chunks[0] || !chunks[1] || !chunks[2])
	{
	  g_warning ("missing data in sync");
	  return;
	}
      
      gdam_model_set_source_id (model, atoi (chunks[0])); 
      beat_info = gdam_beat_info_new ();
      gdam_beat_info_set_bpm (beat_info, atof (chunks[1]));
      gdam_beat_info_set_first_beat (beat_info, atof (chunks[2]));
      gdam_model_set_beat_info (model, beat_info);
      gdam_arg_set (GTK_OBJECT (skin), "lead_id", model->source_id);
    }
  else if (g_strcasecmp (command, "get_sync") == 0)
    {
      send_sync_info (skin, sender);
    }
  else 
    {
      g_message ("no use for message with command %s", command);
    }
}   

static gboolean
unset_username (gpointer ptr)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (ptr);
  gdam_arg_set (GTK_OBJECT (skin), "username", "");
  return FALSE;
}

static void
set_username_success (gpointer ptr)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (ptr);
}

static void
set_username_failure (gpointer ptr)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (ptr);
  /* TODO: send message to window */
  g_idle_add (unset_username, skin);
}


/* Send our sync info unasked to our default recipient */
static void
export_sync_info (GdamChatSkin *skin)
{
  g_return_if_fail (skin->recipient != NULL);
  g_return_if_fail (strcmp ("", skin->recipient) != 0);

  send_sync_info (skin, skin->recipient);
}

/* Send a message to another client. */
static void
do_chat (GdamChatSkin *skin)
{
  char *message;

  g_return_if_fail (skin->username != NULL);
  g_return_if_fail (strcmp ("", skin->username) != 0);
  g_return_if_fail (skin->recipient != NULL);
  g_return_if_fail (strcmp ("", skin->recipient) != 0);

  if (skin->message_entry)
    message = gtk_editable_get_chars (skin->message_entry, 0, -1);
  else
    message = g_strdup_printf ("test message from %s", skin->username);

  gdam_channel_send_chat_message (GDAM_SKIN (skin)->channel, 
				  skin->recipient,
				  "chat",
				  message);
  g_free (message);
}

/* Request sync info from another client. */
static void
do_get_sync (GdamChatSkin *skin)
{
  g_return_if_fail (skin->username != NULL);
  g_return_if_fail (strcmp ("", skin->username) != 0);
  g_return_if_fail (skin->recipient != NULL);
  g_return_if_fail (strcmp ("", skin->recipient) != 0);

  gdam_channel_send_chat_message (GDAM_SKIN (skin)->channel, 
				  skin->recipient,
				  "get_sync",
				  "please.");
}

/* Register our username with the server, and start listening for messages. */
static void
maybe_sign_on (GdamChatSkin *skin)
{
  GdamRequest *request; 
  int id;

  if (!skin->username || strcmp ("", skin->username) == 0)
    return;

  request = gdam_request_set_username (skin->username,
				       set_username_success,
				       set_username_failure,
				       skin);
  gdam_channel_process_request (GDAM_SKIN (skin)->channel, request);

  if (skin->chat_trap_id)
    gdam_channel_untrap_chat_messages (GDAM_SKIN (skin)->channel, 
	  			       skin->chat_trap_id);

  id = gdam_channel_trap_chat_messages (GDAM_SKIN (skin)->channel, NULL, NULL, 
				        print_chat_messages, skin, NULL);
  skin->chat_trap_id = id;
}

static void 
arg_set_username (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);

  if (skin->username && strcmp (skin->username, GDAM_ARG_STRING (arg)) == 0)
    return;

  if (skin->username)
    g_free (skin->username);

  skin->username = g_strdup (GDAM_ARG_STRING (arg));
  maybe_sign_on (skin);
}

static void 
arg_get_username (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);
  GDAM_ARG_STRING (arg) = g_strdup (skin->username);
}

static void 
arg_set_recipient (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);
  if (skin->recipient)
    g_free (skin->recipient);
  skin->recipient = g_strdup (GDAM_ARG_STRING (arg));
}

static void 
arg_get_recipient (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);
  GDAM_ARG_STRING (arg) = g_strdup (skin->recipient);
}

static void 
arg_set_lead_id (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);
  skin->lead_id = GDAM_ARG_INT (arg);
}

static void 
arg_get_lead_id (GtkObject *object, GdamArg *arg)
{
  GdamChatSkin *skin = GDAM_CHAT_SKIN (object);
  GDAM_ARG_INT (arg) = skin->lead_id;
}

/*  Enter a message to send another user. */
static gboolean 
message_entry_handler           (GdamSkin          *skin,
				 GtkObject         *object,
				 const char        *signal_name,
				 int                num_handler_names,
				 char             **handler_names)
{
  GdamChatSkin *chat_skin;
  chat_skin = GDAM_CHAT_SKIN (skin);
  chat_skin->message_entry = GTK_EDITABLE (object);
  return TRUE;
}

/*  Holds last n messages. */
static gboolean 
log_entry_handler               (GdamSkin          *skin,
				 GtkObject         *object,
				 const char        *signal_name,
				 int                num_handler_names,
				 char             **handler_names)
{
  GdamChatSkin *chat_skin;
  chat_skin = GDAM_CHAT_SKIN (skin);
  chat_skin->log_entry = GTK_EDITABLE (object);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_chat_skin_message_entry",
          message_entry_handler			},
	{ "gdam_chat_skin_log_entry",
          log_entry_handler			},
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_chat_skin_send_message",
           (GdamSkinFunc)do_chat },
	{ "gdam_chat_skin_get_sync",
           (GdamSkinFunc)do_get_sync },
	{ "gdam_chat_skin_export_sync_info",
           (GdamSkinFunc)export_sync_info },
};


static void 
gdam_chat_skin_init (GdamChatSkin *chat_skin)
{
}

static gboolean
build_new_model    (GdamSkin      *_skin,
		    GdamChannel   *channel,
		    int	           num_options,
		    char         **options)
{
  GdamModel *model;
  GdamChatSkin *skin = GDAM_CHAT_SKIN (_skin);
  char *user_name = getenv ("USER");

  model = gtk_type_new (GDAM_TYPE_MODEL);
  model->channel = channel;
  GDAM_SKIN (skin)->model = model;
  if (user_name)
    gdam_arg_set (GTK_OBJECT (skin), "username", user_name);
  return TRUE;
}


static void 
gdam_chat_skin_class_init (GdamChatSkinClass *klass)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  skin_class->build_new_model = build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  /* The name we are known by. */
  gdam_arg_register (object_class,
		     "username",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_username,
		     arg_get_username,
		     NULL, NULL, GDAM_ARG_VITAL);
  /* The username we are sending a message to. */
  gdam_arg_register (object_class,
		     "recipient",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_recipient,
		     arg_get_recipient,
		     NULL, NULL, GDAM_ARG_VITAL);
  /* Source id of a different client's lead component. */
  gdam_arg_register (object_class,
		     "lead_id",
		     GDAM_ARG_TYPE_INT,
		     arg_set_lead_id,
		     arg_get_lead_id,
		     NULL, NULL, GDAM_ARG_VITAL);
}

GtkType gdam_chat_skin_get_type()
{
	static GtkType chat_skin_type = 0;
	if (!chat_skin_type) {
		static const GtkTypeInfo chat_skin_info = {
			"GdamChatSkin",
			sizeof(GdamChatSkin),
			sizeof(GdamChatSkinClass),
			(GtkClassInitFunc) gdam_chat_skin_class_init,
			(GtkObjectInitFunc) gdam_chat_skin_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		chat_skin_type = gtk_type_unique(parent,
		                                 &chat_skin_info);
	}
	return chat_skin_type;
}
