/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamechobaseskin.h"
#include <gdam/component/gdambuffermanager.h>
#include <gdam/component/gdamechobase.h>
#include <gdam/client/gdamrequest.h>
#include <gtk/gtksignal.h>

#include "gdamskinutil.h"
static void gdam_echo_base_skin_class_init(GdamSkinClass* skin_class);

typedef struct _MadeBufferData MadeBufferData;
struct _MadeBufferData {
	GdamChannel* channel;
	char* description;
};

static void made_buffer(int        num_args,
                        GdamArg*   args,
                        gpointer   user_data)
{
  GdamBufferManagerInfo *info;
  MadeBufferData *buffer_data = user_data;
  g_return_if_fail (num_args == 1);
  g_return_if_fail (args->type == GDAM_ARG_TYPE_INT);
  g_return_if_fail (strcmp (args->name, "make_buffer") == 0);

  info = g_new0 (GdamBufferManagerInfo, 1);
  info->buffer_id = GDAM_ARG_INT (args);
  info->description = buffer_data->description;
  info->channel = buffer_data->channel;
  gdam_buffer_manager_add_buffer (info);
}

static void didnt_make_buffer(gpointer user_data)
{
  char *description = (char *) user_data;
  g_warning ("failed to make buffer out of echo source");
  g_free (description);
}

static char* get_buffer_description(GdamSkin* skin, GdamEchoBase* echo)
{
  return g_strdup_printf ("%s, source id %d",
			  gtk_type_name (GTK_OBJECT (skin)->klass->type),
			  GDAM_MODEL (echo)->source_id);
}

static void reverb_toggled(GdamEchoBaseSkin*	skin)
{
  GdamEchoBase *echo = GDAM_ECHO_BASE (GDAM_SKIN (skin)->model);
  if (!skin->reverb_tbutton)
    {
      g_warning ("error, missing reverb toggle button...");
      return;
    }
  echo->do_reverb = gtk_toggle_button_get_active (skin->reverb_tbutton);
  gdam_echo_base_refresh_echo_array (echo);
}

static void snag_echo_buffer(GdamSkin*    echo_skin)
{
  GdamEchoBase *echo = GDAM_ECHO_BASE (echo_skin->model);
  int source_id;
  const char *names[1];
  MadeBufferData *data;
  g_return_if_fail (echo != NULL);

  source_id = echo_skin->model->source_id;
  if (source_id == 0)
    {
      g_warning ("cannot get echo buffer out when source isn't playing");
      return;
    }

  names[0] = "make_buffer";
  data = g_new (MadeBufferData, 1);
  data->description = get_buffer_description (echo_skin, echo);
  data->channel = echo_skin->model->channel;
  gdam_channel_append_request (GDAM_MODEL (echo)->channel,
			       gdam_request_get_args (source_id,
						      1,
						      names,
						      made_buffer,
						      didnt_make_buffer,
						      data));
}

static gboolean gdam_echo_base_reverb_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamEchoBaseSkin *eb_skin = GDAM_ECHO_BASE_SKIN (skin);
  eb_skin->reverb_tbutton = GTK_TOGGLE_BUTTON (object);
  if (!eb_skin->reverb_tbutton)
    {
      g_warning ("gdam_echo_base_reverb_tbutton is only valid "
		 "for a toggle button.");
    }
  /* signal_name will usually be "toggled" */
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (reverb_toggled),
			     GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_echo_base_snag_buffer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (snag_echo_buffer),
			     GTK_OBJECT (skin));
  return TRUE;
}

static void gdam_echo_base_skin_init(GdamEchoBaseSkin* echo_base_skin)
{
}

GtkType gdam_echo_base_skin_get_type()
{
  static GtkType echo_base_skin_type = 0;
  if (!echo_base_skin_type)
    {
      static const GtkTypeInfo echo_base_skin_info = {
	"GdamEchoBaseSkin",
	sizeof (GdamEchoBaseSkin),
	sizeof (GdamEchoBaseSkinClass),
	(GtkClassInitFunc) gdam_echo_base_skin_class_init,
	(GtkObjectInitFunc) gdam_echo_base_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      echo_base_skin_type = gtk_type_unique (parent, &echo_base_skin_info);
    }
  return echo_base_skin_type;
}

static GdamSkinGenericConnect connect_infos[] =
{
  {
  "gdam_echo_base_snag_buffer", gdam_echo_base_snag_buffer_handler}
  ,
  {
  "gdam_echo_base_reverb_tbutton", gdam_echo_base_reverb_tbutton_handler}
};

static void gdam_echo_base_skin_class_init(GdamSkinClass* skin_class)
{
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}
