#include "gtkmessagequeue.h"
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <gdam/client/gdamchannel.h>
#include <gdam/client/gdamrequest.h>
#include <gdam/common/gdam-protocol.h>

typedef struct _MessageQueue MessageQueue;
struct _MessageQueue {
	GdamChannel*		channel;
	GtkMessageQueue*	message_queue;
	GtkWidget*		toplevel;
	gboolean		is_visible;
	int			log_flags;
};

static void set_queue_log_level(MessageQueue* queue, int new_level)
{
  GdamRequest *request;
  request = gdam_request_set_log_level (new_level, NULL, NULL, NULL);
  gdam_channel_append_request (queue->channel, request);
  queue->log_flags = new_level;
}

static void toggle_debugging_messages(GtkToggleButton*      toggle,
                                      MessageQueue*         queue)
{
  int new_state = queue->log_flags;
  if (toggle->active)
    new_state |= GDAM_LOG_LEVEL_DEBUG;
  else
    new_state &= ~GDAM_LOG_LEVEL_DEBUG;
  set_queue_log_level (queue, new_state);
}

static void toggle_error_messages    (GtkToggleButton*      toggle,
                                      MessageQueue*         queue)
{
  int new_state = queue->log_flags;
  if (toggle->active)
    new_state |= GDAM_LOG_LEVEL_ERROR;
  else
    new_state &= ~GDAM_LOG_LEVEL_ERROR;
  set_queue_log_level (queue, new_state);
}

/*
 *    +--------------------------------------------------------+
 *    | -GdamMessageQueue-                                     |
 *    |  +--------------------------------------------------+  |
 *    |  | some text                                        |  |
 *    |  | some text                                        |  |
 *    |  | some text                                        |  |
 *    |  +--------------------------------------------------+  |
 *    |                                                        |
 *    |   [ ] Debugging messages?     [ ] Error messages?      |
 *    |   [ ] View dispatch log?                               |
 *    +--------------------------------------------------------+
 */
static void append_log_message(const char* string, void* ptr)
{
  MessageQueue *queue = ptr;
  gtk_message_queue_append (queue->message_queue, string);
}
static MessageQueue* message_queue_new(GdamChannel*      channel)
{
  MessageQueue *queue;
  GtkBox *main_vbox;
  GtkBox *hbox;
  GtkWidget *scrollbar;
  GtkWidget *toggle;
  GdamChannelEventHandler *event_handler;

  queue = g_new (MessageQueue, 1);
  queue->channel = channel;
  queue->message_queue = gtk_message_queue_new (128 * 1024, 1024);
  queue->toplevel = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  queue->is_visible = FALSE;
  queue->log_flags = 0;

  event_handler = gdam_channel_string_handler_new (0, FALSE,
						   append_log_message, queue);
  gdam_channel_register_handler (channel, event_handler);

  /* create the widgets. */
  main_vbox = GTK_BOX (gtk_vbox_new (FALSE, 0));
  hbox = GTK_BOX (gtk_hbox_new (FALSE, 0));
  gtk_box_pack_start (hbox,
		      GTK_WIDGET (queue->message_queue->text), TRUE, TRUE, 0);
  scrollbar = gtk_vscrollbar_new (queue->message_queue->v_adjustment);
  gtk_box_pack_start (hbox, scrollbar, FALSE, FALSE, 0);
  gtk_box_pack_start_defaults (main_vbox, GTK_WIDGET (hbox));

  hbox = GTK_BOX (gtk_hbox_new (FALSE, 0));
  toggle = gtk_toggle_button_new_with_label ("View debugging messages?");
  gtk_box_pack_start_defaults (hbox, toggle);
  gtk_signal_connect (GTK_OBJECT (toggle),
		      "toggled",
		      GTK_SIGNAL_FUNC (toggle_debugging_messages), queue);

  toggle = gtk_toggle_button_new_with_label ("View error messages?");
  gtk_box_pack_start_defaults (hbox, toggle);
  gtk_signal_connect (GTK_OBJECT (toggle),
		      "toggled",
		      GTK_SIGNAL_FUNC (toggle_error_messages), queue);
  gtk_box_pack_start_defaults (main_vbox, GTK_WIDGET (hbox));
  gtk_widget_show_all (GTK_WIDGET (main_vbox));
  gtk_container_add (GTK_CONTAINER (queue->toplevel), GTK_WIDGET (main_vbox));

  return queue;
}

static void message_queue_set_visibility(MessageQueue* queue,
                                         gboolean      is_visible)
{
  if (queue->is_visible && !is_visible)
    {
      gtk_widget_hide (queue->toplevel);
      queue->is_visible = FALSE;
    }
  else if (!queue->is_visible && is_visible)
    {
      gtk_widget_show (queue->toplevel);
      queue->is_visible = TRUE;
    }
}

static GHashTable* channel_to_message_queue = NULL;

void      gdam_message_queue_toggle_visibility(GdamChannel*     channel)
{
  MessageQueue *message_queue;
  if (channel_to_message_queue == NULL)
    channel_to_message_queue = g_hash_table_new (NULL, NULL);
  message_queue = g_hash_table_lookup (channel_to_message_queue, channel);
  if (message_queue == NULL)
    {
      message_queue = message_queue_new (channel);
      g_hash_table_insert (channel_to_message_queue, channel, message_queue);
    }
  message_queue_set_visibility (message_queue, !message_queue->is_visible);
}
