/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamskin.h"
#include "gdampacker.h"
#include <gdam/component/gdamutil.h>
#include "string.h"
#include "gdamkeyboard.h"
#include "gdamskinutil.h"
#include <glade/glade.h>
#include <gdam/component/gdamfilter.h>
#include <gdam/component/gdamturntable.h>
#include <gdam/component/gdambuilder.h>
#include <stdlib.h>
#include <ctype.h>
#include <gtk/gtkwindow.h>


static void     gdam_packer_post_connecting_handler (GdamSkin*          skin);
static gboolean gdam_packer_parse_xml               (GdamView*          view,
	                                             xmlNode*           xml,
						     GdamXmlInfo*       info);
static gboolean gdam_packer_make_xml                (GdamView*          view,
	                                             xmlNode*           parent,
				                     GdamXmlInfo*       info);

static void gdam_packer_toplevel_connect(GdamSkin* skin, GdamToplevel* top)
{
  GdamPacker *packer = GDAM_PACKER (skin);
  packer->toplevel = top;
}

static void do_run_pending(gpointer key, gpointer val, gpointer packer)
{
  if (GDAM_IS_SKIN (val))
    {
      gdam_skin_run_pending_initializers (GDAM_SKIN (val));
    }
}

static gboolean run_children_pending (gpointer ptr)
{
  GdamPacker* packer;
  packer = GDAM_PACKER (ptr);
  if (packer->named_skins)
    g_hash_table_foreach (packer->named_skins, do_run_pending, packer);
  return FALSE;
}

static gboolean idle_try_assign (gpointer data);

static inline void try_assign_key_widget (GdamPacker *packer)
{
  GdamSkin* skin;
  GtkWidget* toplevel;
  skin = GDAM_SKIN (packer);

  if (packer->key_widget)
    return;
  if (!packer->had_keyboard_handler)
    return;
  if (skin->main_widget == NULL)
    return;

  /* TODO: move [if (!realized) idle;] logic to gdamkeyboard? */
  if (!GTK_WIDGET_REALIZED (skin->main_widget))
    {
      g_idle_add (idle_try_assign, packer);
      return;
    }

  /* TODO: duplicate code, suggests we should switch implementation to use
   * facilities in gdamskin.c  */
  toplevel = gtk_widget_get_toplevel (skin->main_widget);

  if (!toplevel)
    return;
  if (!GTK_IS_WINDOW (toplevel))
    return;

  if (!GTK_WIDGET_REALIZED (toplevel))
    {
      g_warning ("try_assign_key_widget: toplevel not realized, returning");
      return;
    }

  packer->key_widget = toplevel;
  gdam_keyboard_connect_widget (packer->key_widget);
}

/* an idle function which looks for a widget to trap hotkey presses. */
static gboolean idle_try_assign (gpointer data)
{
  GdamPacker *packer = GDAM_PACKER (data);
  try_assign_key_widget (packer);

  /* If we didn't get the key widget, try again. */
  if (!packer->key_widget)
    return TRUE;

  /* If we got the key widget, run any initializers which were waiting on
   * it. */
  gdam_skin_run_pending_initializers (GDAM_SKIN (packer));
  return FALSE;
}

static void gdam_packer_post_connecting_handler(GdamSkin* skin)
{
  GdamPacker *packer = GDAM_PACKER (skin);
  g_return_if_fail (packer != NULL);

  /* This assertion is so that no additional deferred handlers
   * will be added; should be done by gdam_skin_new() */
  g_return_if_fail (skin->finished_connecting);

  try_assign_key_widget (packer);
  g_idle_add (run_children_pending, packer);
}

/*
gdam_packer_set_filter_child:filter_name:child_name[:extra_name]
	Attach a GdamSkin/GdamModel as the child of
	a GdamSkin/GdamFilter.

	This handler can also be used to name the child of a
	filter/builder/turntable.
*/

typedef void (*ConnectChildFunc)(GdamModel* parent, GdamModel* child);
typedef GdamModel* (*GetChildFunc)(GdamModel* parent);

typedef struct _DeferredFilterSetChild DeferredFilterSetChild;
struct _DeferredFilterSetChild {
	char* child_name;
	char* filter_name;
	char* extra_name;
	ConnectChildFunc connect_func;
	GetChildFunc get_func;
};

static void init_child (const char* name,
    			gpointer model_or_skin,
			gpointer data)
{
  if (GDAM_IS_SKIN (model_or_skin))
    gdam_skin_run_pending_initializers (GDAM_SKIN (model_or_skin));
}

static GdamModel* model_from_child_ptr (gpointer child_ptr, gpointer user_data)
{
  GdamModel* child_model;
  DeferredFilterSetChild *set_child_info = user_data;
  child_model = NULL;

  if (child_ptr == NULL)
    g_warning ("missing child pointer in model_from_child_ptr!!");

  if (!GTK_IS_OBJECT (child_ptr))
    {
      g_warning ("child pointer %p is not an object in model_from_child_ptr!!",
	         child_ptr);
      return NULL;
    }

  if (GDAM_IS_MODEL (child_ptr))
    {
      child_model = GDAM_MODEL (child_ptr);
    }
  else if (GDAM_IS_PACKER (child_ptr))
    {
      gpointer subchild_ptr;
      if (!set_child_info->extra_name) 
	{
	  g_warning ("child was packer, and we had no name for a subchild");
	  return NULL;
	}
      subchild_ptr = g_hash_table_lookup (GDAM_PACKER (child_ptr)->named_skins,
					  set_child_info->extra_name);

      if (!subchild_ptr) 
	{
	  g_warning ("child was packer, and we could not find a child named %s",
	      	     set_child_info->extra_name);
	  return NULL;
	}
      if (!GTK_IS_OBJECT (subchild_ptr))
	{
	  g_warning ("subchild pointer %p is not an object in model_from_child_ptr!!",
		     subchild_ptr);
	  return NULL;
	}
      return model_from_child_ptr (subchild_ptr, user_data);
    }
  else if (GDAM_IS_SKIN (child_ptr))
    child_model = GDAM_SKIN (child_ptr)->model;
  else 
    {
      g_warning ("packer: don't know how to handle child of type %s ptr %p",
	         gtk_type_name (GTK_OBJECT (child_ptr)->klass->type),
		 child_ptr);
    }
  if (!child_model)
    {
      g_warning ("couldn't find child model from child pointer of type %s",
	         gtk_type_name (GTK_OBJECT (child_ptr)->klass->type));
    }
  return child_model;
}

static gboolean
do_set_filter_child_packing(GdamSkin* skin, gpointer user_data)
{
  DeferredFilterSetChild *set_child_info = user_data;
  gpointer child_ptr;
  gpointer filter_ptr;
  ConnectChildFunc cf;
  GdamModel *child_model;
  GdamModel *filter_model;
  char *child_name = set_child_info->child_name;
  char *filter_name = set_child_info->filter_name;
  GdamPacker *packer = GDAM_PACKER (skin);
  filter_ptr = g_hash_table_lookup (packer->named_skins, filter_name);
  child_ptr = g_hash_table_lookup (packer->named_skins, child_name);
  if (filter_ptr == NULL)
    {
      g_warning ("set_filter_child: %s not found", filter_name);
      return FALSE;
    }
  if (GDAM_IS_MODEL (filter_ptr))
    filter_model = GDAM_MODEL (filter_ptr);
  else
    filter_model = GDAM_SKIN (filter_ptr)->model;

  /* If we don't have a child by the given name, then name the child we
   * do have and return.
   * XXX: are we sure this is correct? this makes the behavior of
   * initialization dependent on the order callbacks are executed.
   * naming of packer pointers should occur before initialization. 
   */
  if (child_ptr == NULL)
    {
      g_warning ("SUSPECT CODE PATH!!");
      child_model = (*set_child_info->get_func) (filter_model);
      if (child_model == NULL)
	{
	  g_warning ("set_filter_child: %s not found", child_name);
	  return FALSE;
	}
      {
	GdamSkin *child_skin;
	child_skin = gdam_skin_find_from_model (child_model);
	if (child_skin != NULL)
	  child_ptr = child_skin;
	else
	  child_ptr = child_model;
	if (!GTK_IS_OBJECT (child_ptr))
	  {
	    g_warning ("child named %s was not an object, not naming (%p)",
		       child_name, child_ptr);
	    return TRUE;
	  }
	g_hash_table_insert (packer->named_skins,
			     g_strdup (child_name), child_ptr);
      }
      return TRUE;
    }
  else if (!GTK_IS_OBJECT (child_ptr))
    {
      g_warning ("set filter child: %s was found but was not object (%p)",
	         child_name, child_ptr);
      return TRUE;
    }
  child_model = model_from_child_ptr (child_ptr, set_child_info);
  g_return_val_if_fail (filter_model != NULL, FALSE);
  g_return_val_if_fail (child_model != NULL, FALSE);

  gdam_debug ("packer_filter_child_packing: %s is child of filter %s",
      	      gtk_type_name (GTK_OBJECT (child_model)->klass->type),
      	      gtk_type_name (GTK_OBJECT (filter_model)->klass->type));

  cf = set_child_info->connect_func;
  if (cf == (ConnectChildFunc) gdam_filter_set_child)
    g_return_val_if_fail (GDAM_IS_FILTER (filter_model), TRUE);
  if (cf == (ConnectChildFunc) gdam_builder_get_component)
    g_return_val_if_fail (GDAM_IS_BUILDER (filter_model), TRUE);
  if (cf == (ConnectChildFunc) gdam_turntable_get_producer)
    g_return_val_if_fail (GDAM_IS_TURNTABLE (filter_model), TRUE);

  (*set_child_info->connect_func) (filter_model, child_model);

  g_hash_table_foreach (packer->named_skins, (GHFunc)init_child, skin);
  return TRUE;
}

static void free_set_filter_child_info(gpointer user_data)
{
  DeferredFilterSetChild *set_child_info = user_data;
  g_free (set_child_info->child_name);
  g_free (set_child_info->filter_name);
  if (set_child_info->extra_name) 
    {
      g_free (set_child_info->extra_name);
    }
  g_free (set_child_info);
}

static gboolean gdam_packer_set_child_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  char *parent_name;
  char *child_name;
  DeferredFilterSetChild *set_child_info;
  if ((parent_name = command[1]) == NULL
      || (child_name = command[2]) == NULL)
    {
      g_warning ("too few arguments to %s", command[0]);
      return TRUE;
    }

  /* NULL the strings out so `cleanup_strings' doesn't
   * free them. */
  command[1] = command[2] = NULL;

  set_child_info = g_new0 (DeferredFilterSetChild, 1);
  set_child_info->child_name = child_name;
  set_child_info->filter_name = parent_name;

  if (command[3]) {
    set_child_info->extra_name = command[3];
    command[3] = NULL;
  }

  if (!strcmp (command[0], "gdam_packer_set_turntable_child")) 
    {
      set_child_info->connect_func =
	(ConnectChildFunc) gdam_turntable_set_producer;
      set_child_info->get_func = (GetChildFunc) gdam_turntable_get_producer;
    }
  else if (!strcmp (command[0], "gdam_packer_set_builder_child")) 
    {
      set_child_info->connect_func =
	(ConnectChildFunc) gdam_builder_set_component;
      set_child_info->get_func = (GetChildFunc) gdam_builder_get_component;
    }
  else if (!strcmp (command[0], "gdam_packer_set_filter_child")) 
    {
      set_child_info->connect_func = (ConnectChildFunc) gdam_filter_set_child;
      set_child_info->get_func = (GetChildFunc) gdam_filter_get_child;
    }
  else 
    {
      g_warning ("confused by handler %s", command[0]);
      return TRUE;
    }

  gdam_skin_add_initializer (skin,
			     do_set_filter_child_packing,
			     free_set_filter_child_info, set_child_info);
  return TRUE;
}


/*
         gdam_packer_add_toplevel:skin_name
*/
typedef struct _PackerToplevelInfo PackerToplevelInfo;
struct _PackerToplevelInfo {
	char* name;
	gboolean misses_data;
};
static void packer_toplevel_info_free(gpointer ptr)
{
  PackerToplevelInfo *info = ptr;
  g_free (info->name);
  g_free (info);
}
static gboolean do_toplevel_packer(GdamSkin* skin, gpointer user_data)
{
  PackerToplevelInfo *info = user_data;
  char *skin_name = info->name;
  GdamPacker *packer = GDAM_PACKER (skin);
  GdamSkin *subskin;
  GdamModel *model;

  subskin = g_hash_table_lookup (packer->named_skins, skin_name);
  if (subskin == NULL)
    {
      g_warning ("no skin named `%s': cannot add it to toplevel", skin_name);
      return TRUE;
    }
  if ((model = subskin->model) == NULL)
    {
      g_warning ("skin `%s' didn't have model: "
		 "cannot add to toplevel", skin_name);
      return TRUE;
    }
  if (packer->toplevel == NULL)
    {
      return FALSE;
    }

  gdam_debug ("packer_toplevel_add model of type %s",
      	      gtk_type_name (GTK_OBJECT (model)->klass->type));

  gdam_toplevel_add_child (packer->toplevel, model);
  gdam_toplevel_set_misses_data (packer->toplevel, model, info->misses_data);
  return TRUE;
}

typedef struct _ToggleInfo ToggleInfo;
struct _ToggleInfo {
	GdamPacker* packer;
	char* child_name;
};

static void do_toplevel_toggle_child(GtkObject* obj, gpointer user_data)
{
  ToggleInfo *toggle_info;
  char *skin_name;
  GdamSkin *subskin;
  GtkWidget *main_widget;
  GdamPacker *packer;
  toggle_info = (ToggleInfo *) user_data;
  packer = GDAM_PACKER (toggle_info->packer);
  skin_name = toggle_info->child_name;

  subskin = g_hash_table_lookup (packer->named_skins, skin_name);
  if (subskin == NULL)
    {
      g_warning ("no skin named `%s': cannot handle visibility toggle",
		 skin_name);
      return;
    }
  if ((main_widget = subskin->main_widget) == NULL)
    {
      g_warning ("skin `%s' didn't have main_widget: "
		 "cannot handle visibility toggle", skin_name);
      return;
    }
  if (GTK_WIDGET_VISIBLE (main_widget))
    gtk_widget_hide (main_widget);
  else
    gtk_widget_show (main_widget);
}

static gboolean gdam_packer_toggle_child_visibility_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  char *child_name;
  gboolean misses_data = TRUE;
  ToggleInfo *toggle_info;

  if ((child_name = command[1]) == NULL)
    {
      g_warning ("too few arguments to %s", command[0]);
      return TRUE;
    }

  /* NULL the strings out so `cleanup_strings' doesn't
   * free them. */
  command[1] = NULL;

  toggle_info = g_new (ToggleInfo, 1);
  toggle_info->packer = GDAM_PACKER (skin);
  toggle_info->child_name = child_name;

  gtk_signal_connect (object,
		      signal_name, do_toplevel_toggle_child, toggle_info);

  return TRUE;
}


static gboolean gdam_packer_add_toplevel_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  char *child_name;
  PackerToplevelInfo *info;
  gboolean misses_data = TRUE;

  if ((child_name = command[1]) == NULL)
    {
      g_warning ("too few arguments to %s", command[0]);
      return TRUE;
    }

  /* NULL the strings out so `cleanup_strings' doesn't
   * free them. */
  command[1] = NULL;
  if (command[2] != NULL && strcmp (command[2], "synchronized") == 0)
    {
      misses_data = FALSE;
    }

  info = g_new (PackerToplevelInfo, 1);
  info->name = child_name;
  info->misses_data = misses_data;

  gdam_skin_add_initializer (skin,
			     do_toplevel_packer,
			     packer_toplevel_info_free, info);
  return TRUE;
}


/*
         gdam_packer_create:local_name:skin_type:skin_filename:options
    		Create a skin as an element of this container.
		 (Usually a container for just one widget)
*/
/* XXX: check whether container is one, and has spare room. */
/* XXX: larger problem - need to monitor all named_skins for destroy */
static gboolean gdam_packer_create_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  char *local_name;
  char *skin_type;
  char *skin_filename;
  GtkContainer *container = GTK_CONTAINER (object);
  GdamPacker *packer;
  GdamSkin *sub_skin;
  GdamModel *sub_model;
  GtkWidget *main_widget;
  g_return_val_if_fail (container != NULL, TRUE);


  if ((local_name = command[1]) == NULL
      || (skin_type = command[2]) == NULL
      || (skin_filename = command[3]) == NULL)
    {
      g_warning ("not enough arguments to `packer_create'");
      return TRUE;
    }

  packer = GDAM_PACKER (skin);
  g_return_val_if_fail (packer != NULL, TRUE);

  sub_skin = g_hash_table_lookup (packer->named_skins, local_name);
  if (sub_skin != NULL)
    {
      gdam_debug ("not creating named skin %s, already exists", local_name);
      g_return_val_if_fail (GDAM_IS_SKIN (packer), TRUE);
      sub_model = sub_skin->model;
    }
  else
    {
      sub_skin = gdam_skin_new_by_type_name (skin_type,
					     skin_filename,
					     num_args - 4,
					     command + 4,
					     skin->channel, &sub_model);
      /* Set it to NULL to it isn't freed... */
      command[1] = NULL;
      if (!GTK_IS_OBJECT (sub_skin))
	{
	  g_warning ("newly created child named %s was not an object, not naming (%p)",
		     local_name, sub_skin);
	  return TRUE;
	}
      g_hash_table_insert (packer->named_skins, local_name, sub_skin);
      if (sub_model && packer->toplevel)
	{
	  gdam_toplevel_add_child (packer->toplevel, sub_model);
	}
    }
  g_return_val_if_fail (sub_skin != NULL, TRUE);


  main_widget = sub_skin->main_widget;
  g_return_val_if_fail (main_widget != NULL, TRUE);
  gdam_set_container (container, main_widget);
  gtk_widget_show (main_widget);

  /* Not sure if this assertion is really needed,
   * but it seems like a lot of later code is
   * ambiguous otherwise... */
  g_return_val_if_fail (sub_skin->model == sub_model, TRUE);

  return TRUE;
}


/*
	   gdam_packer_trap_key_press:key:skin:....

	   Connects a keypress like a button or something... */
typedef struct _DeferredKeyTrap DeferredKeyTrap;
struct _DeferredKeyTrap {
	char*			subskin_name;
	int                     keycode;
	gboolean                is_press;
	int                     num_args;
	char**                  args;
};
static gboolean deferred_key_trap_do_trap(GdamSkin* skin, gpointer user_data)
{
  GdamSkin *subskin;
  DeferredKeyTrap *def_trap = user_data;
  GdamKeyboardTrap *trap;
  GdamPacker *packer = GDAM_PACKER (skin);
  GtkWidget *toplevel;

  /* can't hook up the key trap until the packer has a valid key_widget to
   * listen for key presses */
  if (!packer->key_widget)
    {
      return FALSE;
    }

  /* duplicate code, suggests we should switch implementation to use
   * gdam_skin_trap_key_* */

  if (!GTK_WIDGET_REALIZED (packer->key_widget))
    {
      g_warning ("deferred_do_trap: key_widget not realized, deferring");
      return FALSE;
    }

  subskin = gdam_packer_get_skin (packer, def_trap->subskin_name);

  if (!subskin)
    return FALSE;

  gdam_debug ("deferred_do_trap, keywid %p %s key %d",
      packer->key_widget,
      gtk_type_name (GTK_OBJECT (packer->key_widget)->klass->type),
      def_trap->keycode);

  gdam_skin_do_key_trap (subskin, packer->key_widget, def_trap->keycode, 
                              def_trap->is_press, def_trap->num_args, 
			      def_trap->args);
  return TRUE;
}


static void deferred_key_trap_free(gpointer user_data)
{
  DeferredKeyTrap *def_trap = user_data;
  int i;
  for (i = 0; i < def_trap->num_args; i++)
    g_free (def_trap->args[i]);
  g_free (def_trap->args);
  g_free (def_trap->subskin_name);
  g_free (def_trap);
}


static gboolean gdam_packer_trap_key_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  GdamPacker *packer = GDAM_PACKER (skin);
  DeferredKeyTrap *deferred_trap;
  char *subskin_name;
  char *key_command;
  int key_code;
  if (num_args < 3)
    {
      g_warning ("need 3 args to trap_key_press");
      return TRUE;
    }
  key_command = command[1];
  key_code = gdam_keyboard_parse_key_code (key_command);
  if (key_code == 0)
    {
      g_warning ("error parsing key code: %s", command[1]);
      return TRUE;
    }
  subskin_name = command[2];
  command[2] = NULL;
  packer->had_keyboard_handler = TRUE;
  try_assign_key_widget (packer);
  deferred_trap = g_new (DeferredKeyTrap, 1);
  deferred_trap->subskin_name = subskin_name;
  deferred_trap->keycode = key_code;
  deferred_trap->is_press
    = strcmp ("gdam_packer_trap_key_release", command[0]) == 0 ? FALSE : TRUE;
  deferred_trap->num_args = num_args - 3;
  deferred_trap->args = gdam_strdupv (command + 3);
  gdam_skin_add_initializer (GDAM_SKIN (packer),
			     deferred_key_trap_do_trap,
			     deferred_key_trap_free, deferred_trap);
  return TRUE;
}

/*
	   gdam_packer_defer:skin:....

	   Cause a subskin to process the handler once it has been created. */
typedef struct _DeferredHandler DeferredHandler;
struct _DeferredHandler {
	char*			subskin_name;
	char*			signal_name;
	GtkObject*		object;
	int                     num_args;
	char**                  args;
};

static void deferred_handler_free(gpointer user_data)
{
  DeferredHandler *def_trap = user_data;
  int i;
  for (i = 0; i < def_trap->num_args; i++)
    g_free (def_trap->args[i]);
  g_free (def_trap->args);
  g_free (def_trap->signal_name);
  g_free (def_trap->subskin_name);
  g_free (def_trap);
}

static gboolean deferred_handler_do_handler(GdamSkin* skin, gpointer user_data)
{
  GdamSkin *subskin;
  DeferredHandler *def_trap = user_data;
  GdamPacker *packer = GDAM_PACKER (skin);
  subskin = gdam_packer_get_skin (packer, def_trap->subskin_name);

  g_return_val_if_fail (subskin != NULL, FALSE);

  if (GTK_OBJECT_DESTROYED (def_trap->object)) {
    gtk_object_unref (def_trap->object);
    return TRUE;
  }

  gtk_object_unref (def_trap->object);
  if (!gdam_skin_do_connect (subskin,
			     def_trap->object,
			     def_trap->signal_name,
			     def_trap->num_args, def_trap->args))
    {
      g_warning ("gdam_skin_do_connect failed on deferred handler: "
		 "num_args=%d, handler_hame=%s, skintype=%s",
		 def_trap->num_args,
		 def_trap->args[0],
		 gtk_type_name (GTK_OBJECT (subskin)->klass->type));
      return TRUE;
    }
  gdam_debug ("gdam_skin_do_connect succeeded on deferred handler: \n"
	      "num_args=%d handler_hame=%s skintype=%s skin %p subskin name %s",
	      def_trap->num_args,
	      def_trap->args[0],
	      gtk_type_name (GTK_OBJECT (subskin)->klass->type),
	      subskin,
	      def_trap->subskin_name);
  return TRUE;
}

static gboolean gdam_packer_defer_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  GdamPacker *packer = GDAM_PACKER (skin);
  DeferredHandler *deferred_handler;
  char *subskin_name;
  int key_code;
  if (num_args < 2)
    {
      g_warning ("need 3 args to defer_handler");
      return TRUE;
    }
  subskin_name = command[1];
  command[1] = NULL;
  //packer->had_keyboard_handler = TRUE;
  deferred_handler = g_new (DeferredHandler, 1);
  deferred_handler->subskin_name = subskin_name;
  //deferred_handler->is_press
  //= strcmp("gdam_packer_trap_key_release", command[0]) == 0
  //? FALSE : TRUE;
  deferred_handler->num_args = num_args - 2;
  deferred_handler->args = gdam_strdupv (command + 2);
  deferred_handler->object = object;
  deferred_handler->signal_name = g_strdup (signal_name);
  gtk_object_ref (object);
  gdam_skin_add_initializer (GDAM_SKIN (packer),
			     deferred_handler_do_handler,
			     deferred_handler_free, deferred_handler);
  return TRUE;
}

static void warn_ignored(gpointer unused) {
	g_warning("ignoring unhandled event");
}

static gboolean gdam_packer_ignore_handler
				(GdamSkin*          skin,
			         GtkObject*         object,
			         const char*        signal_name,
			         int                num_args,
			         char**             command)
{
  if (gtk_signal_lookup (signal_name, object->klass->type) == 0)
    {
      return TRUE;
    }
  gtk_signal_connect (object,
		      signal_name, GTK_SIGNAL_FUNC (warn_ignored), NULL);
  return TRUE;
}



static GdamSkinGenericConnect connect_infos[] =
{
  { "gdam_packer_set_filter_child", gdam_packer_set_child_handler} ,
  { "gdam_packer_set_builder_child", gdam_packer_set_child_handler} ,
  { "gdam_packer_set_turntable_child", gdam_packer_set_child_handler} ,
  { "gdam_packer_toggle_child_visibility",
     gdam_packer_toggle_child_visibility_handler} ,
  { "gdam_packer_add_toplevel", gdam_packer_add_toplevel_handler} ,
  { "gdam_packer_trap_key_press", gdam_packer_trap_key_handler} ,
  { "gdam_packer_trap_key_release", gdam_packer_trap_key_handler} ,
  { "gdam_packer_defer", gdam_packer_defer_handler} ,
  { "gdam_packer_ignore", gdam_packer_ignore_handler} ,
  { "gdam_packer_create", gdam_packer_create_handler}
};


static void gdam_packer_init(GdamPacker* packer)
{
  packer->named_skins = g_hash_table_new (g_str_hash, g_str_equal);
}



static void gdam_packer_class_init(GdamPackerClass* packer_class)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (packer_class);
  GdamViewClass *view_class;
  view_class = GDAM_VIEW_CLASS (packer_class);
  skin_class->post_connecting_handler = gdam_packer_post_connecting_handler;
  skin_class->toplevel_connect = gdam_packer_toplevel_connect;
  view_class->parse_xml = gdam_packer_parse_xml;
  view_class->make_xml = gdam_packer_make_xml;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

GtkType gdam_packer_get_type()
{
  static GtkType packer_type = 0;
  if (!packer_type)
    {
      static const GtkTypeInfo packer_info = {
	"GdamPacker",
	sizeof (GdamPacker),
	sizeof (GdamPackerClass),
	(GtkClassInitFunc) gdam_packer_class_init,
	(GtkObjectInitFunc) gdam_packer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      packer_type = gtk_type_unique (parent, &packer_info);
    }
  return packer_type;
}

GdamSkinClass* gdam_packer_class()
{
  static GdamSkinClass *class = NULL;
  if (class != NULL)
    return class;
  return class = GDAM_SKIN_CLASS (gtk_type_class (GDAM_TYPE_PACKER));
}

#if 0
typedef struct _DeferredEvent DeferredEvent;

struct _DeferredEvent {
	GdamPackerFunc      command;
	GDestroyNotify      done_with_command;
	gpointer            user_data;
};

static void run_and_delete_one(DeferredEvent* event, GdamPacker* packer)
{
  (*event->command) (packer, event->user_data);
  if (event->done_with_command)
    (*event->done_with_command) (event->user_data);
  g_free (event);
}

static void gdam_packer_run_deferred(GdamPacker* packer)
{
  GSList *deferred = packer->pending;
  packer->pending = NULL;
  deferred = g_slist_reverse (deferred);
  g_slist_foreach (deferred, (GFunc) run_and_delete_one, packer);
  g_slist_free (deferred);
}

void        gdam_packer_add_deferred_command(GdamPacker*         packer,
                                             GdamPackerFunc      command,
			                     GDestroyNotify      destroyer,
			                     gpointer            user_data)
{
  DeferredEvent *event;
  event = g_new (DeferredEvent, 1);
  event->command = command;
  event->done_with_command = destroyer;
  event->user_data = user_data;
  packer->pending = g_slist_prepend (packer->pending, event);
}
#endif

GdamModel* gdam_packer_get_model               (GdamPacker*         packer,
                                                const char*         name)
{
  gpointer ptr = g_hash_table_lookup (packer->named_skins, name);
  if (ptr == NULL)
    return NULL;
  if (GDAM_IS_MODEL (ptr))
    return GDAM_MODEL (ptr);
  else
    {
      GdamSkin *skin;
      skin = GDAM_SKIN (ptr);
      return skin->model;
    }
}

GdamSkin*      gdam_packer_get_skin            (GdamPacker*         packer,
			                        const char*         name)
{
  gpointer ptr = g_hash_table_lookup (packer->named_skins, name);
  if (ptr == NULL)
    {
      gdam_debug ("gdam_packer_get_skin: %s: unknown child named %s", 
		 GDAM_SKIN (packer)->filename, name);
      return NULL;
    }
  if (GDAM_IS_MODEL (ptr))
    {
      g_warning ("%s was a component, not a skin", name);
      return NULL;
    }
   else if (GDAM_IS_SKIN (ptr))
    {
      GdamSkin *skin;
      skin = GDAM_SKIN (ptr);
      return skin;
    }
  else 
    {
      g_warning ("couldn't get_skin named %s: not a skin", name);
      return NULL;
    }
}

/*
 * Parse/produce xml file.
 */

typedef struct _XmlMaker XmlMaker;
struct _XmlMaker {
	GdamPacker*		packer;
	xmlNode*		parent;
	GdamXmlInfo*            xml_info;
};

static void add_one_xml_skin(const char*        name,
                             gpointer           model_or_skin,
			     XmlMaker*          maker)
{
  xmlNode *obj_node = NULL;
  xmlNode *packer_node;
  gboolean was_skin = FALSE;
  int id;
  g_return_if_fail (model_or_skin != NULL);

  id = gdam_xml_info_lookup_by_object (maker->xml_info, model_or_skin);
  if (id != 0)
    {
      char id_buf[256];
      sprintf (id_buf, "%d", id);
      was_skin = GDAM_IS_VIEW (model_or_skin);
      packer_node = xmlNewNode (NULL, was_skin ? "skin" : "component");
      xmlNewProp (packer_node, "name", name);
      xmlAddChild (maker->parent, packer_node);
      xmlNewChild (packer_node, NULL, "id", id_buf);
      return;
    }
  else if (GDAM_IS_MODEL (model_or_skin))
    {
      GdamModel *comp = GDAM_MODEL (model_or_skin);
      obj_node = gdam_model_to_xml (comp, maker->xml_info);
      was_skin = FALSE;
    }
  else if (GDAM_IS_SKIN (model_or_skin))
    {
      GdamSkin *skin = GDAM_SKIN (model_or_skin);
      obj_node = gdam_skin_to_xml (skin, maker->xml_info);
      was_skin = TRUE;
    }
  else if (GTK_IS_OBJECT (model_or_skin))
    {
      g_warning ("%p [%s] is not a skin or component", model_or_skin,
		 gtk_type_name (GTK_OBJECT (model_or_skin)->klass->type));
      return;
    }
  else
    {
      g_warning ("gdam_packer: "
		 "add_one_xml: %p is not a gtkobject", model_or_skin);
      return;
    }
  if (obj_node == NULL)
    {
      g_warning ("error building xml [in packer]...");
      return;
    }

  packer_node = xmlNewNode (NULL, was_skin ? "skin" : "component");
  xmlNewProp (packer_node, "name", name);
  xmlAddChild (packer_node, obj_node);
  xmlAddChild (maker->parent, packer_node);
}

static gboolean gdam_packer_make_xml(GdamView* skin,
	                             xmlNode*           parent,
				     GdamXmlInfo*       xml_info)
{
  GdamPacker *packer = GDAM_PACKER (skin);
  if (packer->named_skins != NULL)
    {
      XmlMaker xml_maker;
      xml_maker.packer = packer;
      xml_maker.parent = parent;
      xml_maker.xml_info = xml_info;
      g_hash_table_foreach (packer->named_skins,
			    (GHFunc) add_one_xml_skin, &xml_maker);
    }

  /* XXX: this should be inherited??? */
  gdam_skin_add_filename_xml (GDAM_SKIN (skin), parent);

  /* XXX: store interconnect ??? */
  return TRUE;
}

static gboolean gdam_packer_parse_xml(GdamView*      view,
	                              xmlNode*       node,
				      GdamXmlInfo*   xml_info)
{
  GdamPacker *packer = GDAM_PACKER (view);
  GdamChannel *channel;
  channel = GDAM_SKIN (packer)->channel;

  /* XXX: get rid of this somehow... */
  packer->toplevel = gdam_toplevel_get_global (channel);

  if (strcmp (node->name, "filename") == 0)
    {
      gdam_skin_handle_filename_xml (GDAM_SKIN (packer), node);
      return TRUE;
    }

  if (strcmp (node->name, "skin") == 0
      || strcmp (node->name, "component") == 0)
    {
      char *name;
      GtkObject *obj;
      xmlNode *child;


      name = xmlGetProp (node, "name");
      if (name == NULL)
	{
	  g_warning ("<%s> must have name in packer", node->name);
	  return FALSE;
	}
      child = node->childs;
      if (child == NULL || child->next != NULL)
	{
	  g_warning ("<%s> must contain exactly one " "subnode", node->name);
	  return FALSE;
	}

      if (strcmp (node->name, "skin") == 0)
	{
	  GdamSkin *skin;
	  skin = gdam_skin_from_xml (child, channel, xml_info);
	  obj = GTK_OBJECT (skin);
	}
      else
	{
	  GdamModel *component;
	  component = gdam_model_from_xml (child, channel, xml_info);
	  obj = GTK_OBJECT (component);
	}
      if (obj != NULL)
	{
	  if (!GTK_IS_OBJECT (obj))
	    {
	      g_warning ("supposed object %p named %s is not an object.",
		         obj, name);
	      return FALSE;
	    }
	  g_hash_table_insert (packer->named_skins, name, obj);
	}
      return TRUE;
    }
  g_warning ("unhandled node type in packer: <%s>", node->content);
  /* Ignore it :) */
  return FALSE;
}
