/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdampixfont.h"
#include "pixmapcache.h"
#include <string.h>

/*
 *   ___    _            ___ _     ___        _  _____
 *  / __|__| |__ _ _ __ | _ (_)_ _| __|__ _ _| ||_   _|  _ _ __  ___
 * | (_ / _` / _` | '  \|  _/ \ \ / _/ _ \ ' \  _|| || || | '_ \/ -_)
 *  \___\__,_\__,_|_|_|_|_| |_/_\_\_|\___/_||_\__||_| \_, | .__/\___|
 *                                                    |__/|_|
 */

static GHashTable* pix_font_types_table = NULL;

static void initialize_stock_fonts()
{
  g_assert (pix_font_types_table == NULL);
  pix_font_types_table = g_hash_table_new (g_str_hash, g_str_equal);
  gdam_pix_font_type_register ("integer", 11, 1, "0123456789 \n");
  gdam_pix_font_type_register ("ascii",
			       32, 3,
			       " !\"#$%&'()*+,-./0123456789:;<=>?\n"
			       "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_\n"
			       "`abcdefghijklmnopqrstuvwxyz{|}~\n");
}

static int find_max_char(const char* str)
{
  int max_char = 0;
  while (*str)
    {
      guchar c = *str++;
      if (c > max_char)
	max_char = c;
    }
  return max_char;
}
void gdam_pix_font_type_register(const char*       type_name, 
                                 int               num_horizontal,
			         int               num_vertical,
			         const char*       text)
{
  int num_lines = 0;
  const char *tmp = text;
  char **strs;
  GdamPixFontType *type;
  const char *start;
  const char *start_line;

  if (pix_font_types_table == NULL)
    initialize_stock_fonts ();

  g_return_if_fail (g_hash_table_lookup (pix_font_types_table,
					 type_name) == NULL);

  while ((tmp = strchr (tmp, '\n')) != NULL)
    {
      num_lines++;
      tmp++;			/* skip past newline */
    }
  strs = g_new (char *, num_lines + 1);
  start = text;
  start_line = text;
  num_lines = 0;
  while ((tmp = strchr (start_line, '\n')) != NULL)
    {
      strs[num_lines] = g_new (char, tmp - start_line + 1);
      memcpy (strs[num_lines], start_line, tmp - start_line);
      strs[num_lines][tmp - start_line] = '\0';
      num_lines++;
      start_line = tmp + 1;
    }
  strs[num_lines] = NULL;

  type = g_new (GdamPixFontType, 1);
  type->name = g_strdup (type_name);
  type->num_horizontal = num_horizontal;
  type->num_vertical = num_vertical;
  type->character_layout = strs;
  type->max_character = find_max_char (start);
  g_hash_table_insert (pix_font_types_table, type->name, type);
}


GdamPixFontType* gdam_pix_font_type_stock(const char* layout_name)
{
  GdamPixFontType *type;
  if (pix_font_types_table == NULL)
    initialize_stock_fonts ();
  type = g_hash_table_lookup (pix_font_types_table, layout_name);
  if (type == NULL)
    g_warning ("gdam_pix_font_type_stock: unknown font layout: %s",
	       layout_name);
  return type;
}

static GMemChunk* element_chunk = NULL;
static GdamPixFontElement* gdam_pix_font_element_new(int         c,
                                                     int         image_left,
                                                     int         image_top,
                                                     int         image_width,
                                                     int         image_height)
{
  GdamPixFontElement *element;
  if (element_chunk == NULL)
    element_chunk = g_mem_chunk_create (GdamPixFontElement,
					64, G_ALLOC_AND_FREE);
  element = g_chunk_new (GdamPixFontElement, element_chunk);
  element->image_left = image_left;
  element->image_top = image_top;
  element->image_width = image_width;
  element->image_height = image_height;
  return element;
}
static void gdam_pix_font_element_free(GdamPixFontElement* element)
{
  g_chunk_free (element, element_chunk);
}
/*
 *   ____     _                 ____  _      _____           _
 *  / ___| __| | __ _ _ __ ___ |  _ \(_)_  _|  ___|__  _ __ | |_
 * | |  _ / _` |/ _` | '_ ` _ \| |_) | \ \/ / |_ / _ \| '_ \| __|
 * | |_| | (_| | (_| | | | | | |  __/| |>  <|  _| (_) | | | | |_
 *  \____|\__,_|\__,_|_| |_| |_|_|   |_/_/\_\_|  \___/|_| |_|\__|
 * 
 */
static GHashTable* registered_fonts = NULL;
static GtkObjectClass* parent_class = NULL;

static void         gdam_pix_font_register     (GdamPixFont*     pix_font);
static void         gdam_pix_font_unregister   (GdamPixFont*     pix_font);
static GdamPixFont* gdam_pix_font_lookup_cached(const char*      image_name,
                                                GdamPixFontType* type);

static void gdam_pix_font_destroy(GtkObject* object)
{
  GdamPixFont *pix_font;
  int i;
  pix_font = GDAM_PIX_FONT (object);
  if (pix_font->is_registered)
    gdam_pix_font_unregister (pix_font);

  g_free (pix_font->pixmap_name);
  pix_font->pixmap_name = NULL;

  for (i = 0; i < pix_font->num_elements; i++)
    {
      GdamPixFontElement *element;
      element = pix_font->font_elements[i];
      if (element)
	gdam_pix_font_element_free (element);
    }
  g_free (pix_font->font_elements);
  pix_font->font_elements = NULL;
  pix_font->num_elements = 0;

  (*parent_class->destroy) (object);
}

static void gdam_pix_font_init(GdamPixFont* pix_font)
{
}
static void gdam_pix_font_class_init(GdamPixFontClass* pix_font_class)
{
}

GtkType gdam_pix_font_get_type()
{
  static GtkType pix_font_type = 0;
  if (!pix_font_type)
    {
      static const GtkTypeInfo pix_font_info = {
	"GdamPixFont",
	sizeof (GdamPixFont),
	sizeof (GdamPixFontClass),
	(GtkClassInitFunc) gdam_pix_font_class_init,
	(GtkObjectInitFunc) gdam_pix_font_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_OBJECT;
      pix_font_type = gtk_type_unique (parent, &pix_font_info);
    }
  return pix_font_type;
}

gboolean gdam_pix_font_construct  (GdamPixFont*     pix_font,
				   GdkWindow*       window,
                                   const char*      image_name,
                                   GdamPixFontType* pix_font_type)
{
  GdkPixmap *pixmap;
  PixmapInfo *pixmap_info;
  char **at;
  int y;
  int tile_width, tile_height;

  g_return_val_if_fail (window != NULL, FALSE);

  pixmap_info = pixmap_cache_get_pixmap (window, image_name);
  if (pixmap_info == NULL)
    return FALSE;

  pix_font->pixmap_width = pixmap_info->width;
  pix_font->pixmap_height = pixmap_info->height;
  pix_font->pixmap = pixmap_info->pixmap;
  pix_font->pixmap_name = g_strdup (image_name);
  pix_font->font_type = pix_font_type;

  pix_font->num_elements = pix_font_type->max_character + 1;
  pix_font->font_elements = g_new0 (GdamPixFontElement *,
				    pix_font->num_elements);

  tile_width = pix_font->pixmap_width / pix_font_type->num_horizontal;
  tile_height = pix_font->pixmap_height / pix_font_type->num_vertical;

  at = pix_font_type->character_layout;
  y = 0;
  for (y = 0; y < pix_font_type->num_vertical && at[y] != NULL; y++)
    {
      int start_y = tile_height * y;
      char *line = at[y];
      int x;

      for (x = 0; x < pix_font_type->num_horizontal && line[x] != '\0'; x++)
	{
	  int c = (guchar) line[x];
	  int start_x = tile_width * x;
	  GdamPixFontElement *el;
	  el = gdam_pix_font_element_new (c, start_x, start_y,
					  tile_width, tile_height);
	  pix_font->font_elements[c] = el;
	}
    }
  return TRUE;
}


GdamPixFont* gdam_pix_font_new_cached(GdkWindow*       window,
                                      const char*      image_name,
                                      GdamPixFontType* font_type)
{
  GdamPixFont *rv;
  rv = gdam_pix_font_lookup_cached (image_name, font_type);
  if (rv == NULL)
    {
      GtkObject *object;
      object = gtk_type_new (GDAM_TYPE_PIX_FONT);
      rv = GDAM_PIX_FONT (object);
      if (!gdam_pix_font_construct (rv, window, image_name, font_type))
	{
	  gtk_object_sink (object);
	  return NULL;
	}
      gtk_object_ref (object);
      gtk_object_sink (object);
      gdam_pix_font_register (rv);
    }
  return rv;
}

void gdam_pix_font_get_string_size(GdamPixFont*     pix_font,
                                   const char*      string,
				   int*             width_out,
				   int*             height_out)
{
  int width = 0, height = 0;
  while (*string != '\0')
    {
      guchar c = (guchar) (*string++);
      GdamPixFontElement *element;
      if (c >= pix_font->num_elements)
	continue;
      element = pix_font->font_elements[c];
      if (element == NULL)
	continue;
      width += element->image_width;
      if (height < element->image_height)
	height = element->image_height;
    }
  *width_out = width;
  *height_out = height;
}

void gdam_pix_font_draw_string    (GdamPixFont*     pix_font,
				   GdkDrawable*     destination,
				   GdkGC*           gc,
				   int              dest_left,
				   int              dest_top,
                                   const char*      string)
{
  int x = dest_left;
  int y = dest_top;

  while (*string != '\0')
    {
      guchar c = (guchar) (*string++);
      GdamPixFontElement *element;
      if (c >= pix_font->num_elements)
	continue;
      element = pix_font->font_elements[c];
      if (element == NULL)
	continue;
      gdk_draw_pixmap (destination,
		       gc,
		       pix_font->pixmap,
		       element->image_left,
		       element->image_top,
		       x, y, element->image_width, element->image_height);
      x += element->image_width;
    }
}

/* Registering and unregistering fonts. */
typedef struct _GdamPixFontId GdamPixFontId;
struct _GdamPixFontId {
	char* image_name;
	GdamPixFontType* type;
};

static guint gdam_pix_font_id_hash(gconstpointer key)
{
  const GdamPixFontId *id = key;
  return g_str_hash (id->image_name) ^ GPOINTER_TO_UINT (id->type);
}
static gint  gdam_pix_font_id_equal(gconstpointer key1, gconstpointer key2)
{
  const GdamPixFontId *id1 = key1;
  const GdamPixFontId *id2 = key2;
  return g_str_equal (id1->image_name, id2->image_name)
    && id1->type == id2->type;
}

static GHashTable* font_cache = NULL;

static void         gdam_pix_font_register     (GdamPixFont*     pix_font)
{
  GdamPixFontId tmp;
  GdamPixFontId *key;
  if (font_cache == NULL)
    font_cache = g_hash_table_new (gdam_pix_font_id_hash,
				   gdam_pix_font_id_equal);
  tmp.image_name = (char *) pix_font->pixmap_name;
  tmp.type = pix_font->font_type;
  g_return_if_fail (g_hash_table_lookup (font_cache, &tmp) == NULL);

  key = g_new (GdamPixFontId, 1);
  key->image_name = g_strdup (pix_font->pixmap_name);
  key->type = pix_font->font_type;
  g_hash_table_insert (font_cache, key, pix_font);
}
static void         gdam_pix_font_unregister   (GdamPixFont*     pix_font)
{
  GdamPixFontId tmp;
  GdamPixFontId *id;
  GdamPixFont *font;
  gpointer key_pointer, value_pointer;
  tmp.image_name = (char *) pix_font->pixmap_name;
  tmp.type = pix_font->font_type;
  if (!g_hash_table_lookup_extended (font_cache,
				     &tmp, &key_pointer, &value_pointer))
    {
      g_warning ("gdam_pix_font_unregister: unregistered already?");
      return;
    }
  id = key_pointer;
  font = value_pointer;
  g_hash_table_remove (font_cache, id);
  g_free (id->image_name);
  g_free (id);
}
static GdamPixFont* gdam_pix_font_lookup_cached(const char*      image_name,
                                                GdamPixFontType* type)
{
  GdamPixFontId tmp;
  GdamPixFont *rv;
  tmp.image_name = (char *) image_name;
  tmp.type = type;
  if (font_cache == NULL)
    return NULL;
  rv = g_hash_table_lookup (font_cache, &tmp);
  if (rv == NULL)
    return NULL;
  return GDAM_PIX_FONT (rv);
}
