/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/* XXX: needs finalize/destroy signals */

#include "gdampixfontlabel.h"
#include <string.h>
static int  gdam_pix_font_label_expose_event   (GtkWidget*        widget,
                                                GdkEventExpose*   event);
static void gdam_pix_font_label_size_request   (GtkWidget*        widget,
                                                GtkRequisition*   requisition);

static gboolean try_to_get_font(GdamPixFontLabel* label, GdkWindow* window)
{
  if (label->font == NULL)
    {
      label->font = gdam_pix_font_new_cached (window,
					      label->font_name,
					      label->font_type);
    }
  return label->font != NULL;
}

static void redraw_label(GdamPixFontLabel* label)
{
  GtkWidget *widget;
  int width, height;
  int str_width, str_height;
  int slack_x, slack_y;
  int left, top;
  widget = GTK_WIDGET (label);

  if (widget->window == NULL && label->font == NULL)
    return;
  if (!try_to_get_font (label, widget->window))
    {
      g_warning ("redraw_label: couldn't get font for label");
      return;
    }

  width = widget->allocation.width;
  height = widget->allocation.height;
  gdam_pix_font_get_string_size (label->font,
				 label->text, &str_width, &str_height);
  slack_x = width - str_width;
  slack_y = height - str_height;
  left = slack_x * label->x_justify / 2;
  top = slack_y * label->y_justify / 2;

#if 0
  /* fill the background */
  /*gtk_style_set_background(widget->style,
     widget->window,
     GTK_STATE_NORMAL); */
  {
    GdkRectangle par_rect;
    par_rect.x = widget->allocation.x;
    par_rect.y = widget->allocation.y;
    par_rect.width = widget->allocation.width;
    par_rect.height = widget->allocation.height;
    gtk_widget_draw (widget->parent, &par_rect);
  }
#endif

  /* draw the string. */
  gdam_pix_font_draw_string (label->font,
			     widget->window,
			     widget->style->fg_gc[0], left, top, label->text);
}

static int  gdam_pix_font_label_expose_event   (GtkWidget*        widget,
                                                GdkEventExpose*   event)
{
  redraw_label (GDAM_PIX_FONT_LABEL (widget));
  return 0;
}
static void gdam_pix_font_label_size_request   (GtkWidget*        widget,
                                                GtkRequisition*   requisition)
{
  GdamPixFontLabel *label = GDAM_PIX_FONT_LABEL (widget);
  int x, y;
  if (widget->window == NULL && label->font == NULL)
    return;
  if (!try_to_get_font (label, widget->window))
    return;
  gdam_pix_font_get_string_size (label->font, label->text, &x, &y);
  requisition->width = x + label->misc.xpad * 2;
  requisition->height = y + label->misc.ypad * 2;
}

static void gdam_pix_font_label_init(GdamPixFontLabel* pix_font_label)
{
}

static void gdam_pix_font_label_class_init(GdamPixFontLabelClass* klass)
{
  GtkWidgetClass *widget_class;
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->expose_event = gdam_pix_font_label_expose_event;
  widget_class->size_request = gdam_pix_font_label_size_request;
}

GtkType gdam_pix_font_label_get_type()
{
  static GtkType pix_font_label_type = 0;
  if (!pix_font_label_type)
    {
      static const GtkTypeInfo pix_font_label_info = {
	"GdamPixFontLabel",
	sizeof (GdamPixFontLabel),
	sizeof (GdamPixFontLabelClass),
	(GtkClassInitFunc) gdam_pix_font_label_class_init,
	(GtkObjectInitFunc) gdam_pix_font_label_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_MISC;
      pix_font_label_type = gtk_type_unique (parent, &pix_font_label_info);
    }
  return pix_font_label_type;
}

GtkWidget*    gdam_pix_font_label_new     (const char*       font_name,
                                           GdamPixFontType*  font_type,
                                           const char*       text,
					   int               x_justify,
					   int               y_justify)
{
  GdamPixFontLabel *label;
  label = GDAM_PIX_FONT_LABEL (gtk_type_new (GDAM_TYPE_PIX_FONT_LABEL));
  label->text = g_strdup (text);
  label->font_name = g_strdup (font_name);
  label->font_type = font_type;
  label->x_justify = x_justify;
  label->y_justify = y_justify;
  return GTK_WIDGET (label);
}

void          gdam_pix_font_label_set_text(GdamPixFontLabel* label,
                                           const char*       text)
{
  char *old_text = label->text;
  label->text = g_strdup (text);
  g_free (old_text);
  if (GTK_WIDGET_MAPPED (label))
    redraw_label (label);
}

void          gdam_pix_font_label_set_font(GdamPixFontLabel* label,
					   const char*       font_name,
                                           GdamPixFontType*  font_type)
{
  char *old_font_name;

  if (label->font != NULL)
    gtk_object_unref ((GtkObject *) (label->font));

  label->font = NULL;
  g_free (label->font_name);
  old_font_name = label->font_name;
  label->font_name = g_strdup (font_name);
  g_free (old_font_name);
  label->font_type = font_type;

  if (GTK_WIDGET_MAPPED (label))
    redraw_label (label);
}

void          gdam_pix_font_label_set_just(GdamPixFontLabel* label,
                                           int               x_justify,
                                           int               y_justify)
{
  label->x_justify = x_justify;
  label->y_justify = y_justify;
  if (GTK_WIDGET_MAPPED (label))
    redraw_label (label);
}

const char*   gdam_pix_font_label_get_text(GdamPixFontLabel* label)
{
  return label->text;
}

/*
 * Format:
 *     string1:       filename:arrangement_name
 *     string2:       initial_text
 *     int1:          x-justification
 *     int2:          y-justification
 *
 * Arguments:
 *    filename            name of the pixmap file to ...
 *    arrangement_name    integer|ascii
 *    x-justification     0=left, 1=center, 2=right justification
 *    y-justification     0=top, 1=center, 2=bottom justification
 */
GtkWidget* gdam_pix_font_label_create_func (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char *pixmap_name;
  int length;
  const char *colon;
  const char *layout;
  GdamPixFontType *font_type;
  colon = strchr (string1, ':');
  g_return_val_if_fail (colon != NULL, NULL);
  pixmap_name = alloca (colon - string1 + 1);
  memcpy (pixmap_name, string1, colon - string1);
  pixmap_name[colon - string1] = '\0';
  layout = colon + 1;
  font_type = gdam_pix_font_type_stock (layout);
  g_return_val_if_fail (font_type != NULL, NULL);
  g_return_val_if_fail (string2 != NULL, NULL);
  g_return_val_if_fail (0 <= int1 && int1 <= 2, NULL);
  g_return_val_if_fail (0 <= int2 && int2 <= 2, NULL);
  return gdam_pix_font_label_new (pixmap_name, font_type,
				  string2, int1, int2);
}
