/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamskin-pixwidgets.h"
#include "gdamskin.h"
#include "gdamcursor.h"
#include "pixmapcache.h"
#include <string.h>
#include <stdlib.h>
#include <gtk/gtksignal.h>


/*  Bounce Label
 */

static void
bounce_label_load_image (gpointer ptr)
{
  GtkBounceLabel* bounce_label = GTK_BOUNCE_LABEL (ptr);
  GdkWindow *window = GTK_WIDGET (bounce_label)->window;
  PixmapInfo *info;
  char *xpm = bounce_label->background_file;

  if (xpm == NULL)
    {
      return;
    }
  info = pixmap_cache_get_pixmap (window, xpm);
  if (info == NULL)
    {
      g_warning ("error loading pixmap %s", xpm);
      return;
    }
  gtk_bounce_label_set_pixmap (bounce_label, info->pixmap, 
                               info->width, info->height);
}

#define DEFINE_IDLE_LOAD_FUNC(a,b)		               \
static gboolean                                                \
a (gpointer ptr)                                               \
{                                                              \
  GtkObject *object = GTK_OBJECT (ptr);                        \
  /* If our widget has been destroyed, abort. */               \
  if (object->klass == NULL || GTK_OBJECT_DESTROYED (object))  \
    {                                                          \
      gtk_widget_unref (GTK_WIDGET (object));                  \
      return FALSE;                                            \
    }                                                          \
  /* If we aren't ready, call again. */                        \
  if (!GTK_WIDGET (object)->window)                            \
    {							\
      g_warning ("widget %p %s type %s has no window",         \
	         object, gtk_widget_get_name (GTK_WIDGET (object)), \
		 gtk_type_name (object->klass->type));         \
      return FALSE;                                            \
    }							\
  b (object);                                                  \
  gtk_widget_unref (GTK_WIDGET (object));                      \
  return FALSE;  /* don't call again */                        \
}

DEFINE_IDLE_LOAD_FUNC(bounce_label_do_load_images, bounce_label_load_image)

void gtk_bounce_label_set_pixmaps_by_filename(GtkBounceLabel* label,
                                              const char*     filename)
{
  char *old_xpm_file;
  if (filename == label->background_file
      || (filename != NULL && label->background_file != NULL
	  && strcmp (filename, label->background_file) == 0))
    {
      return;
    }
  g_free (label->background_file);
  label->background_file = g_strdup (filename);
  if (GTK_WIDGET (label)->window)
    bounce_label_load_image (label);
  else
    {
      gtk_widget_ref (GTK_WIDGET (label));
      g_idle_add (bounce_label_do_load_images, label);
    }
}

GtkWidget* gtk_bounce_label_create_func      (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  GtkBounceLabel *bounce_label;
  bounce_label = GTK_BOUNCE_LABEL (gtk_type_new (GTK_TYPE_BOUNCE_LABEL));

  if (string1)
    {
      char **pixmaps;
      pixmaps = g_strsplit (string1, ":", 100);
      gtk_bounce_label_set_pixmaps_by_filename (bounce_label, pixmaps[0]);
      g_strfreev (pixmaps);
    }
  return GTK_WIDGET (bounce_label);
}



/*  Pix Button
 */

static void 
pix_button_load_images (gpointer ptr)
{
  GtkPixButton* pix_button = GTK_PIX_BUTTON (ptr);
  GdkWindow *window = GTK_WIDGET (pix_button)->window;
  GdkPixmap *pixmaps[4] = {NULL, NULL, NULL, NULL};
  GdkBitmap *masks[4] = {NULL, NULL, NULL, NULL};
  int i;

  for (i = 0; i < 4; i++)
    {
      PixmapInfo *info;
      char *xpm = pix_button->xpm_files[i];
      if (xpm == NULL)
	{
	  g_warning ("error: no file %d for pixbutton", i);
	  continue;
	}
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      pixmaps[i] = info->pixmap;
      masks[i] = info->mask;
    }
  gtk_pix_button_set_pixmaps (pix_button, 
                              pixmaps[0], masks[0],
                              pixmaps[1], masks[1],
                              pixmaps[2], masks[2],
                              pixmaps[3], masks[3]);
  gtk_widget_draw (GTK_WIDGET (pix_button), NULL);
}

DEFINE_IDLE_LOAD_FUNC(pixbutton_do_load_images, pix_button_load_images)

void gtk_pix_button_set_pixmaps_by_filename(GtkPixButton* button,
                                            const char*   normal,
                                            const char*   active,
					    const char*   prelight,
					    const char*   selected)
{
  button->xpm_files[GTK_STATE_NORMAL] = g_strdup (normal);
  button->xpm_files[GTK_STATE_ACTIVE] = g_strdup (active);
  button->xpm_files[GTK_STATE_PRELIGHT] = g_strdup (prelight);
  button->xpm_files[GTK_STATE_SELECTED] = g_strdup (selected);
  if (GTK_WIDGET (button)->window)
    pix_button_load_images (button);
  else
    {
      gtk_widget_ref (GTK_WIDGET (button));
      g_idle_add (pixbutton_do_load_images, button);
    }
}

GtkWidget* gtk_pix_button_create_func      (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char *pixmaps_colon_sep = g_strdup (string1);
  char **pixmaps = g_strsplit (pixmaps_colon_sep, ":", 100);
  GtkPixButton *pix_button;
  pix_button = GTK_PIX_BUTTON (gtk_type_new (GTK_TYPE_PIX_BUTTON));
  gtk_pix_button_set_pixmaps_by_filename (pix_button,
					  pixmaps[0],
					  pixmaps[1], pixmaps[2], pixmaps[3]);
  pix_button->is_toggle = int1;
  pix_button->is_display = int2;
  g_strfreev (pixmaps);
  g_free (pixmaps_colon_sep);
  return GTK_WIDGET (pix_button);
}



/*  Pix Scrollbar
 */

static void 
pixscroll_load_images (gpointer ptr)
{
  GtkPixScrollbar* pix_scrollbar = GTK_PIX_SCROLLBAR (ptr);
  GdkWindow *window = GTK_WIDGET (pix_scrollbar)->window;
  PixmapInfo *info;
  GdkPixmap *background = NULL;
  GdkPixmap *floater = NULL;
  GdkBitmap *background_mask = NULL;
  GdkBitmap *floater_mask = NULL;
  int fx = 0;
  int fy = 0;
  int bx = 0;
  int by = 0;
  char *xpm = pix_scrollbar->floater_file;
  if (xpm == NULL)
    {
      g_warning ("error: no filename for floater in pixscrollbar");
    }
  else
    {
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      else
	{
	  floater = info->pixmap;
	  floater_mask = info->mask;
          fx = info->width;
          fy = info->height;
	}
    }

  if (strcmp (pix_scrollbar->background_xpm_file, "none") != 0)
    {
      const char *fname = pix_scrollbar->background_xpm_file;
      PixmapInfo *info;
      info = pixmap_cache_get_pixmap (window, fname);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", fname);
	  return;
	}
      background = info->pixmap;
      background_mask = info->mask;
      bx = info->width;
      by = info->height;
    }

  gtk_pix_scrollbar_set_pixmaps (pix_scrollbar, floater, floater_mask, 
                                 background, background_mask);
  /* TODO: set_pixmaps should figure out these dimensions on its own... */
  pix_scrollbar->floater_size.x = fx;
  pix_scrollbar->floater_size.y = fy;
  pix_scrollbar->background_size.x = bx;
  pix_scrollbar->background_size.y = by;
  gtk_widget_draw (GTK_WIDGET (pix_scrollbar), NULL);
}

DEFINE_IDLE_LOAD_FUNC(pixscroll_do_load_images, pixscroll_load_images)

void gtk_pix_scrollbar_set_pixmaps_by_filename(GtkPixScrollbar* pix_scrollbar,
                                               const char*      background,
                                               const char*      floater)
{
  pix_scrollbar->floater_file = g_strdup (floater);
  pix_scrollbar->background_xpm_file = g_strdup (background);
  if (GTK_WIDGET (pix_scrollbar)->window)
    pixscroll_load_images (pix_scrollbar);
  else
    {
      gtk_widget_ref (GTK_WIDGET (pix_scrollbar));
      g_idle_add (pixscroll_do_load_images, pix_scrollbar);
    }
}

/*
 *   string1:    background:floater
 *   string2:    
 *   startx:starty:deltax:deltay:cur_value:min_value:max_value[:drag_to_pos]
 *   int1:       do_reset?
 *   int2:       trace_value?
 */

/*
 * To make a GtkPixScrollbar:
 *
 *   string1:    background:normal:active:prelight:selected
 *   string2:    startx:starty:deltax:deltay:cur_value:min_value:max_value:step
 *
 *
 * Here is a description of the fields.
 *
 *     background:       the xpm file to display in the background of
 *                       the scrollbar.
 *     normal:           the xpm file to use as the indicator normally.
 *
 *     start_x:          the x-coordinate for the minimum trough value.
 *     start_y:          the y-coordinate for the minimum trough value.
 *     delta_x:          the x-distance from min to max of the trough.
 *     delta_y:          the y-distance from min to max of the trough.
 *     cur_value:        initial value for the adjustment.
 *     min_value:        minimum value for the adjustment.
 *     max_value:        maximum value for the adjustment.
 *     step:             amount to step for other clicks
 */

GtkWidget* gtk_pix_scrollbar_create_func   (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char **pixmaps = g_strsplit (string1, ":", 100);
  char **coordinates;
  gfloat cur_value, min_value, max_value;
  gfloat step_increment, page_increment, page_size;
  GtkAdjustment *adjustment;
  GtkPixScrollbar *pix_scrollbar;
  gboolean do_reset;
  pix_scrollbar = GTK_PIX_SCROLLBAR (gtk_type_new (GTK_TYPE_PIX_SCROLLBAR));
  gtk_pix_scrollbar_set_pixmaps_by_filename (pix_scrollbar,
					     pixmaps[0], pixmaps[1]);
  g_strfreev (pixmaps);

  coordinates = g_strsplit (string2, ":", 100);
  pix_scrollbar->trough_start_x = atoi (coordinates[0]);
  pix_scrollbar->trough_start_y = atoi (coordinates[1]);
  pix_scrollbar->trough_delta_x = atoi (coordinates[2]);
  pix_scrollbar->trough_delta_y = atoi (coordinates[3]);

  pix_scrollbar->do_reset = int1;
  pix_scrollbar->trace_value = int2;

  cur_value = atof (coordinates[4]);
  min_value = atof (coordinates[5]);
  max_value = atof (coordinates[6]);
  if (coordinates[7] && atoi (coordinates[7]) != 0)
    {
      pix_scrollbar->drag_to_position = 0;
      if (!coordinates[8] || !coordinates[9] || !coordinates[10])
	{
	  g_warning ("missing step_inc page_inc and/or page_size after %s",
	              coordinates[7]);
	  step_increment = 0.01;
	  page_increment = 0.01;
	  page_size = 0.0;
	}
      else
	{
	  step_increment = atof (coordinates[8]);
	  page_increment = atof (coordinates[9]);
	  page_size = atof (coordinates[10]);
	}
    }
  else
    {
      pix_scrollbar->drag_to_position = 1;
      step_increment = 0.01;
      page_increment = 0.01;
      page_size = 0.0;
    }
  g_strfreev (coordinates);

  pix_scrollbar->orig_value = cur_value;

  adjustment = (GtkAdjustment *) gtk_adjustment_new (cur_value,
						     min_value,
						     max_value,
						     step_increment,
						     page_increment,
						     page_size);
  gtk_pix_scrollbar_set_adjustment (pix_scrollbar, adjustment);

  return GTK_WIDGET (pix_scrollbar);
}



/*  Pix Splitbar
 */

static void 
splitbar_load_images (gpointer ptr)
{
  GtkPixSplitbar* pix_splitbar = GTK_PIX_SPLITBAR (ptr);
  GdkWindow *window = GTK_WIDGET (pix_splitbar)->window;
  PixmapInfo *info;
  GdkPixmap *below_pixmap = NULL;
  GdkPixmap *above_pixmap = NULL;
  GdkBitmap *below_mask = NULL;
  GdkBitmap *above_mask = NULL;
  int fx = 0;
  int fy = 0;
  int bx = 0;
  int by = 0;
  char *xpm = pix_splitbar->above_file;
  if (xpm == NULL)
    {
      g_warning ("error: no filename for above in pixsplitbar");
    }
  else
    {
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      else
	{
	  above_pixmap = info->pixmap;
	  above_mask = info->mask;
          fx = info->width;
          fy = info->height;
	}
    }

  xpm = pix_splitbar->below_file;
  if (xpm == NULL)
    {
      g_warning ("error: no filename for below in pixsplitbar");
    }
  else
    {
      info = pixmap_cache_get_pixmap (window, xpm);
      if (info == NULL)
	{
	  g_warning ("error loading pixmap %s", xpm);
	  return;
	}
      else
	{
	  below_pixmap = info->pixmap;
	  below_mask = info->mask;
          bx = info->width;
          by = info->height;
	}
    }
  gtk_pix_splitbar_set_pixmaps (pix_splitbar, above_pixmap, above_mask, 
                                below_pixmap, below_mask);
  pix_splitbar->above_size.x = fx;
  pix_splitbar->above_size.y = fy;
  pix_splitbar->below_size.x = bx;
  pix_splitbar->below_size.y = by;
  gtk_widget_draw (GTK_WIDGET (pix_splitbar), NULL);
}

DEFINE_IDLE_LOAD_FUNC(splitbar_do_load_images, splitbar_load_images)

void 
gtk_pix_splitbar_set_pixmaps_by_filename (GtkPixSplitbar* pix_splitbar,
                                          const char*     above,
                                          const char*     below)
{
  pix_splitbar->above_file = g_strdup (above);
  pix_splitbar->below_file = g_strdup (below);
  if (GTK_WIDGET (pix_splitbar)->window)
    splitbar_load_images (pix_splitbar);
  else
    {
      gtk_widget_ref (GTK_WIDGET (pix_splitbar));
      g_idle_add (splitbar_do_load_images, pix_splitbar);
    }
}

static int string_to_orientation(const char* string)
{
  if (strcasecmp (string, "left") == 0)
    return GTK_PIX_SPLITBAR_LEFT;
  if (strcasecmp (string, "right") == 0)
    return GTK_PIX_SPLITBAR_RIGHT;
  if (strcasecmp (string, "top") == 0)
    return GTK_PIX_SPLITBAR_TOP;
  if (strcasecmp (string, "bottom") == 0)
    return GTK_PIX_SPLITBAR_BOTTOM;
  return -1;
}

/*
 * Create_func for gtk_pix_splitbar.
 *
 *   string1:    above:below
 *   string2:    orientation:cur_value:min_value:max_value
 *
 * Here is a description of the fields.
 *
 *     above:            the pixmap to display for the area above the
 *                       current value.
 *     below:            the pixmap to display for the area below the
 *                       current value.
 *                       the splitbar.
 *     orientation:      one of the following values:
 *          left              grow from right-to-left 
 *          right             grow from left-to-right 
 *          top               grow from bottom-to-top 
 *          bottom            grow from top-to-bottom 
 *     cur_value:        initial value for the adjustment.
 *     min_value:        minimum value for the adjustment.
 *     max_value:        maximum value for the adjustment.
 *
 *
 */
GtkWidget* gtk_pix_splitbar_create_func    (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char **pixmaps = g_strsplit (string1, ":", 100);
  char **coordinates;
  gfloat cur_value, min_value, max_value, step_incr;
  GtkAdjustment *adjustment;
  GtkPixSplitbar *pix_splitbar;
  int id;
  pix_splitbar = GTK_PIX_SPLITBAR (gtk_type_new (GTK_TYPE_PIX_SPLITBAR));
  gtk_pix_splitbar_set_pixmaps_by_filename (pix_splitbar,
					    pixmaps[0], pixmaps[1]);
  g_strfreev (pixmaps);

  coordinates = g_strsplit (string2, ":", 100);
  id = string_to_orientation (coordinates[0]);
  if (id < 0)
    {
      g_warning ("orientation %s not allowed to GtkPixSplitbar",
		 coordinates[0]);
      return NULL;
    }

  pix_splitbar->orientation = id;

  cur_value = atof (coordinates[1]);
  min_value = atof (coordinates[2]);
  max_value = atof (coordinates[3]);
  step_incr = coordinates[4] ? atof (coordinates[4]) : 0.01;
  g_strfreev (coordinates);

  adjustment = (GtkAdjustment *) gtk_adjustment_new (cur_value,
						     min_value,
						     max_value,
						     step_incr,
						     step_incr * 5,
						     step_incr);
  gtk_pix_splitbar_set_adjustment (pix_splitbar, adjustment);

  return GTK_WIDGET (pix_splitbar);
}




/*  Pix Dial
 */

static void 
dial_try_load_images (gpointer ptr)
{
  GtkPixDial *dial = GTK_PIX_DIAL (dial);
  GdkWindow *window = GTK_WIDGET (dial)->window;
  GdkPixmap *dial_pixmap = NULL;
  GdkBitmap *dial_mask = NULL;
  PixmapInfo *pixmap_info;

  g_return_if_fail (dial->num_tiles > 0);
  if (dial->pixmap != NULL)
    return;
  if (dial->image_filename == NULL)
    return;
  g_return_if_fail (window != NULL);
  pixmap_info = pixmap_cache_get_pixmap (window, dial->image_filename);
  if (pixmap_info == NULL)
    {
      g_warning ("dial: error loading image %s", dial->image_filename);
      return;
    }
  dial_pixmap = pixmap_info->pixmap;
  dial_mask = pixmap_info->mask;
  gtk_pix_dial_set_pixmap (dial, dial_pixmap, dial_mask);
}

DEFINE_IDLE_LOAD_FUNC(dial_do_load_images, dial_try_load_images)

#if 0
static gboolean 
dial_do_load_images (gpointer ptr)
{
  GtkObject *object = GTK_OBJECT (ptr);
  /* If our widget has been destroyed, abort and don't call again. */
  if (object->klass == NULL || GTK_OBJECT_DESTROYED (object))
    {
      gtk_widget_unref (GTK_WIDGET (object));
      return FALSE;
    }
  /* If we aren't ready, call again. */
  if (!GTK_WIDGET (object)->window)
    return TRUE;
  dial_try_load_images (GTK_PIX_DIAL (ptr));
  gtk_widget_unref (GTK_WIDGET (object));
  return FALSE;  /* don't call again */
}
#endif



void gtk_pix_dial_set_pixmap_by_filename(GtkPixDial*       dial,
                                      const char*     image_filename)
{
  dial->image_filename = g_strdup (image_filename);
  if (GTK_WIDGET (dial)->window)
    dial_try_load_images (dial);
  else
    {
      gtk_widget_ref (GTK_WIDGET (dial));
      g_idle_add (dial_do_load_images, dial);
    }
}

static GtkTooltips *dial_tooltips = NULL;

static void 
update_dial_tooltip (GtkPixDial* dial)
{
  GtkAdjustment *adj = dial->adjustment;
  gdouble value = adj->value;

  if (dial->is_int)
    value = (gdouble)((int)(adj->value + 0.5));

  g_return_if_fail (dial_tooltips != NULL);

  gtk_tooltips_set_tip(GTK_TOOLTIPS (dial_tooltips), 
      		       GTK_WIDGET (dial),
                       g_strdup_printf ("%f", value),
		       NULL);
}

static gboolean set_cursor 			 (gpointer        ptr)
{
  g_return_val_if_fail (GTK_WIDGET (ptr)->window != NULL, TRUE);
  gdam_set_window_cursor (GTK_WIDGET (ptr)->window, GDAM_HMOVE_CURSOR);
  gtk_widget_draw (GTK_WIDGET (ptr), NULL);
  return FALSE;
}

static void
pix_dial_realize (GtkWidget *widget)
{
  g_idle_add (set_cursor, widget);
}

/*
 * To make a GtkPixDial:
 *
 *   string1:    image_filename
 *   string2:    cur_value:min_value:max_value[:reversed][:reset][:pix_wrapcount]
 *   int1:       screen_range
 *   int2:       num_tiles
 *
 * Here is a description of the fields.
 *
 *     image_filename:   a vertically stacked array of pixmaps.
 *     cur_value:        initial value for the adjustment.
 *     min_value:        minimum value for the adjustment.
 *     max_value:        maximum value for the adjustment.
 *     screen_range:     number of pixels to move to go from
 *                       min_value to max_value.
 *     num_tiles:        number of vertically stacked images in the image file
 *     options:          flags configuring how to behave:
 *       reversed        the minimum value representation is at the bottom
 *       reset           jump to original value / pixmap on release
 */
GtkWidget* gtk_pix_dial_create_func        (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  char **coordinates;
  gfloat cur_value, min_value, max_value, step_incr;
  GtkAdjustment *adjustment;
  GtkPixDial *dial;
  int id;
  int i;
  dial = GTK_PIX_DIAL (gtk_type_new (GTK_TYPE_PIX_DIAL));
  gtk_pix_dial_set_pixmap_by_filename (dial, string1);

  coordinates = g_strsplit (string2, ":", 100);
  if (coordinates[0] == NULL
      || coordinates[1] == NULL || coordinates[2] == NULL)
    {
      g_warning ("gtk_pix_dial_create_func: too few arguments to "
		 "init adjustment");
      g_strfreev (coordinates);
      gtk_object_sink (GTK_OBJECT (dial));
      return NULL;
    }

  cur_value = atof (coordinates[0]);
  min_value = atof (coordinates[1]);
  max_value = atof (coordinates[2]);
  for (i = 3; coordinates[i] != NULL; i++)
    {
      const char *arg = coordinates[i];
      if (strcmp (arg, "reversed") == 0)
	{
	  dial->images_reversed = 1;
	  continue;
	}
      else if (strcmp (arg, "reset") == 0)
	{
	  dial->reset_on_release = 1;
	  continue;
	}
      else if (strcmp (arg, "is_int") == 0)
	{
	  dial->is_int = 1;
	  continue;
	}
      else if (strcmp (arg, "no_mouse") == 0)
	{
	  dial->no_mouse = 1;
	  continue;
	}
      else
	{
	  int wcount;
	  wcount = atoi (arg);
	  if (wcount > 0)
	    {
	      dial->wrapcount = wcount;
	    }
	  continue;
	}
    }
  //step_incr = 0.01;
  step_incr = 0.0;
  g_strfreev (coordinates);

  adjustment = (GtkAdjustment *) gtk_adjustment_new (cur_value,
						     min_value,
						     max_value,
						     step_incr,
						     step_incr * 5,
						     step_incr);
  gtk_pix_dial_set_adjustment (dial, adjustment);
  dial->screen_range = int1;
  dial->num_tiles = int2;
  dial->orig_value = cur_value;

  if (!dial->no_mouse)
    {
      gtk_signal_connect_object
	(GTK_OBJECT (dial),
	 "realize",
	 GTK_SIGNAL_FUNC (pix_dial_realize), 
	 GTK_OBJECT (dial));

      dial->connect_id = gtk_signal_connect_object
	(GTK_OBJECT (adjustment),
	 "value_changed",
	 GTK_SIGNAL_FUNC (update_dial_tooltip), 
	 GTK_OBJECT (dial));

      if (!dial_tooltips)
	{
	  dial_tooltips = gtk_tooltips_new ();
	  gtk_tooltips_set_delay (dial_tooltips, 0); /* millis till popup */
	}

      gtk_tooltips_set_tip(GTK_TOOLTIPS (dial_tooltips), 
			   GTK_WIDGET (dial),
			   g_strdup_printf ("%f", adjustment->value),
			   NULL);
    }

  return GTK_WIDGET (dial);
}

/* XXX: this is just a compatibility hack until we expunge all
 * references to gdam_dial from all xml files */
GtkWidget* gdam_dial_create_func           (const char*   name,
                                            const char*   string1,
                                            const char*   string2,
                                            int           int1,
                                            int           int2)
{
  return gtk_pix_dial_create_func (name, string1, string2, int1, int2);
}




