/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamskinbindingfile.h"
#include "gdampacker.h"
#include "gdamskin.h"
#include <gdam/component/gdaminputdevice.h>
#include <gdam/component/gdambindingfile.h>
#include <gdam/component/gdammodel.h>
#include <stdlib.h>
#include <string.h>

/*                     The binding interface documentation :)

<Gdam-Binding-Interface>
  <constant type="GdamGain">
    <argument>muted</argument>
    <value>1</value>
    <GdamMidiEventSysex>
      <sysex>0x01 0x02 0x03 0x04</sysex>
      <device>/dev/midi</device>
    </GdamMidiEventSysex>
  </constant>
  <range type="GdamGain">
    <argument>volume</argument>
    <GdamMidiEventPitchWheel>
      <input_id>3</input_id>
      <device>/dev/midi</device>
    </GdamMidiEventPitchWheel>
    <min_value>0.0</min_value>
    <max_value>2.0</max_value>
  </range>
</Gdam-Binding-Interface>

*/

gboolean gdam_skin_binding_file_process      (xmlNode*       node_list, 
                                              GdamSkin*      skin);

static const char* get_string_content(xmlNode* node)
{
  if (node->content)
    return node->content;
  if (node->childs == NULL)
    return NULL;
  return node->childs->content;
}

static gboolean parse_signal_xml_node (xmlNode*       const_node,
                                       GdamSkin*      skin)
{
  xmlNode *node;
  xmlNode *children = const_node->childs;
  GdamInputEvent *input_event = NULL;
  const char *signal_name = NULL;
  char *handlers[2];
  xmlNode *value_node = NULL;
  guint input_hook;
  const char *property;
  handlers[0] = NULL;
  handlers[1] = NULL;
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "signal") == 0)
	{
	  signal_name = get_string_content (node);
	  continue;
	}
      if (strcmp (node->name, "handler") == 0)
	{
	  handlers[0] = g_strdup (get_string_content (node));
	  continue;
	}
      input_event = parse_input_event (node);
      if (input_event == NULL)
	{
	  g_warning ("couldn't parse input event from "
		     "node named `%s'", node->name);
	  return FALSE;
	}
    }
  if (signal_name == NULL)
    {
      g_warning ("<signal> was missing <signal>");
      return FALSE;
    }
  if (handlers[0] == NULL)
    {
      g_warning ("<signal> was missing <handler>");
      return FALSE;
    }
  /* Always connect signal and handler */
  if (!gdam_skin_do_connect (skin, GTK_OBJECT (skin), signal_name, 1,
			     handlers))
    {
      g_warning ("error connecting signal=%s handler=%s", signal_name,
		 handlers[0]);
    }
  /* If we have an input event, emit that signal when the event happens. */
  if (input_event)
    {
      input_hook = gdam_input_hook_connect_signal
			    (gdam_input_trap_find_or_create (input_event), 
			     (GtkObject *) skin, 
			     signal_name,
			     handlers[0]);  
    }
  return TRUE;
}

static gboolean parse_bindingfile_xml_node (xmlNode*       xml_node,
                                            GdamSkin*      skin)
{
  xmlNode *node;
  xmlNode *children = xml_node->childs;
  const char *filename = NULL;
  for (node = children; node != NULL; node = node->next)
    {
      if (strcmp (node->name, "filename") == 0)
	{
	  filename = get_string_content (node);
	  continue;
	}
    }
  if (filename == NULL)
    {
      g_warning ("<bindingfile> was missing <filename>");
      return FALSE;
    }
  gdam_skin_parse_binding_file (skin, filename);
  return TRUE;
}



static gboolean parse_arg(GdamArg*    to_parse,
                          GtkObject*  object,
                          const char* arg_name,
			  xmlNode*    value_node)
{
  int type;
  const char *content;

  type = gdam_arg_query_type (object->klass->type, arg_name);
  if (type == 0)
    {
      g_warning ("no argument %s was found in %s",
		 arg_name, gtk_type_name (object->klass->type));
      return FALSE;
    }
  to_parse->name = g_strdup (arg_name);
  to_parse->type = type;
  if (value_node == NULL)
    {
      GDAM_ARG_DOUBLE (to_parse) = 0.0;
      return TRUE;
    }
  content = get_string_content (value_node);
  if (content == NULL)
    {
      g_warning ("<value> for <%s> had no string content", arg_name);
      return FALSE;
    }
  return gdam_arg_parse_val_from_string (to_parse, content) > 0;
}


static GdamSkin* type_property_to_skin(GdamSkin* skin, const char* property)
{
  GtkType allow_type;		/* type this tag allows */
  GtkType model_type;
  GtkType skin_type;
  const char *c;

  if (property == NULL) 
    {
      g_warning("NULL property in type_property_to_skin");
      return NULL;
    }

  skin_type = GTK_OBJECT (skin)->klass->type;

  c = strchr (property, ':');
  /*  If property is colonated list, parse list and handle. */
  if (c != NULL)
    {
      GdamSkin *newskin;
      char **tokens = g_strsplit (property, ":", 0);
      int i;
      newskin = NULL;
      allow_type = gtk_type_from_name (tokens[0]);
      if (tokens[0] && tokens[1]
	  && strcmp (tokens[0], "GdamPacker") == 0
	  && gtk_type_is_a (skin_type, allow_type))
	{
	  newskin = gdam_packer_get_skin (GDAM_PACKER (skin), tokens[1]);
	  if (!newskin)
	    {
	      g_warning ("bindingfile: packer had no skin named %s", tokens[1]);
	      return NULL;
	    }
	}
      else
	{
	  g_warning ("binding parse error: %s", property);
	  return NULL;
	}
      for (i = 0; tokens[i] != NULL; i++)
	{
	  g_free (tokens[i]);
	}
      g_free (tokens);
      return newskin;
    }
  else 
    {
      allow_type = gtk_type_from_name (property);
      if (gtk_type_is_a (skin_type, allow_type))
	{
	  return skin;
	}
      gdam_debug ("skin %s of type %s, not of type %s... skipping.", property, 
	  	  gtk_type_name (skin_type), gtk_type_name (allow_type));
    }
  return NULL;
}



	  /* We apply tags only to skins/models whose types match. */
static GdamModel* type_property_to_model(GdamSkin* skin, const char* property)
{
  GdamModel *model;
  GtkType allow_type;		/* type this tag allows */
  GtkType model_type;
  GtkType skin_type;
  GdamSkin *child_skin;
  const char *c;

  if (property == NULL)
    return NULL;

  child_skin = NULL;
  model = skin->model;
  skin_type = GTK_OBJECT (skin)->klass->type;
  c = strchr (property, ':');
  /*  If property is colonated list, parse list and handle. */
  if (c != NULL)
    {
      char **tokens = g_strsplit (property, ":", 0);
      int i;
      allow_type = gtk_type_from_name (tokens[0]);
      model = NULL;
      if (tokens[0] && tokens[1]
	  && strcmp (tokens[0], "GdamPacker") == 0
	  && gtk_type_is_a (skin_type, allow_type))
	{
	  model = gdam_packer_get_model (GDAM_PACKER (skin), tokens[1]);
	  if (!model)
	    {
	      gdam_debug ("packer had no model named %s", property);
	      return NULL;
	    }
	}
      else
	{
	  g_warning ("binding parse error: %s", property);
	  return NULL;
	}
      for (i = 0; tokens[i] != NULL; i++)
	{
	  g_free (tokens[i]);
	}
      g_free (tokens);
      return model;
    }
  /*  Else 'type' property must match type of model.  */
  else
    {
      if (model == NULL)
	return NULL;
      model_type = GTK_OBJECT (model)->klass->type;
      allow_type = gtk_type_from_name (property);
      if (allow_type == 0)
	return NULL;
      if (!gtk_type_is_a (model_type, allow_type))
	{
	  gdam_debug ("binding: skipping `%s': not a `%s' (prop=%s)",
		      gtk_type_name (model_type), 
		      gtk_type_name (allow_type), property);
	  return NULL;
	}
      return model;
    }
}

gboolean gdam_skin_binding_file_process      (xmlNode*       node_list, 
                                              GdamSkin*      skin)
{
  xmlNode *node;
  int processed = 0;

  for (node = node_list; node != NULL; node = node->next)
    {
      const char *name;
      const char *property;
      GdamModel *model;
      GdamSkin *subskin;
      GtkObject *mobj;
      GtkObject *sobj;
      subskin = NULL;

      name = node->name;

      property = xmlGetProp (node, "type");
      model = type_property_to_model (skin, property);
      subskin = type_property_to_skin (skin, property);

      mobj = model ? GTK_OBJECT (model) : NULL;
      sobj = subskin ? GTK_OBJECT (subskin) : NULL;

      if (!model && !subskin)
	{
	  gdam_debug ("couldn't get model or skin of type '%s' from parent %p"
	              "of type '%s' in xml node with name '%s'", property, skin,
		      gtk_type_name (GTK_OBJECT (skin)->klass->type), name);
	}
      else if (strcmp (name, "defer") == 0 && subskin)
	{
	  if (node->childs)
	    {
	      if (gdam_skin_binding_file_process (node->childs, subskin))
		processed++;
	    }
	}
      else if (strcmp (name, "constant") == 0)
	{
	  if (mobj && gdam_binding_parse_constant_xml_node (node, mobj))
	    processed++;
	  else if (sobj && gdam_binding_parse_constant_xml_node (node, sobj))
	    processed++;
	}
      else if (strcmp (name, "range") == 0)
	{
	  if (mobj && gdam_binding_parse_range_xml_node (node, mobj))
	    processed++;
	  else if (sobj && gdam_binding_parse_range_xml_node (node, sobj))
	    processed++;
	}
      else if (strcmp (name, "variable") == 0)
	{
	  if (mobj && gdam_binding_parse_variable_xml_node (node, mobj))
	    processed++;
	  else if (sobj && gdam_binding_parse_variable_xml_node (node, sobj))
	    processed++;
	}
      else if (strcmp (name, "signal") == 0 && subskin)
	{
	  if (parse_signal_xml_node (node, subskin))
	    processed++;
	}
      else if (strcmp (name, "bindingfile") == 0 && subskin)
	{
	  if (parse_bindingfile_xml_node (node, subskin))
	    processed++;
	}
      else
	{
	  /* Check for a registered type */
	  GList *current_type = gdam_binding_file_node_type_list ();
	  gboolean rv = FALSE;
	  while (current_type != NULL && !rv)
	    {
	      GdamBindingNodeTypeInfo *node_info;
	      node_info = (GdamBindingNodeTypeInfo *)current_type->data;
	      if (g_strcasecmp (name, node_info->node_name) == 0)
		{
		  if (mobj)
		    rv = (*node_info->parse) (node, mobj);
		  if (!rv && sobj)
		    rv = (*node_info->parse) (node, sobj);
		}
	      current_type = current_type->next;
	    }
	  if (rv)
	    processed++;
	  else 
	    {
	      g_warning ("bad skin binding file node!! name=%s model=%p "
		         "subskin=%p type=%s processed=%d", name, model, 
			 subskin, property, processed);
	    }
	}
    }
  if (processed <= 0)
    {
      g_warning ("got no bindings!");
      return FALSE;
    }
  gdam_debug ("Got %d bindings", processed);
  return TRUE;
}

