/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamsmoothfunction.h"

#define NUM_DIV			256

static void recompute_cached_points(GdamSmoothFunction* function)
{
  gfloat base_value;
  gfloat delta_value;
  gfloat end_value;
  GdamSpline *spline = function->function;
  int i;

  g_return_if_fail (spline != NULL);

  gdam_spline_get_domain (spline, &base_value, &end_value);
  delta_value = end_value - base_value;

  if (function->cached_points == NULL)
    {
      function->num_cached_points = NUM_DIV + 1;
      function->cached_points = g_new (GdkPoint, NUM_DIV + 1);
    }

  for (i = 0; i <= NUM_DIV; i++)
    {
      gfloat input_value = base_value + delta_value * i / NUM_DIV;
      gfloat output_value;
      GdkPoint *point = function->cached_points + i;
      output_value = gdam_spline_eval (spline, input_value);
      point->x = ((input_value - function->min_input)
		  / (function->max_input - function->min_input)
		  * function->draw_width);
      point->y = ((output_value - function->min_output)
		  / (function->max_output - function->min_output)
		  * function->draw_height);
    }
  function->cached_points_invalid = FALSE;
}

static void redraw_curve(GdamSmoothFunction* function,
                         GdkRectangle*       rectangle)
{
  int x, y, width, height;
  GtkWidget *widget;
  GdkWindow *window;
  GdkGC *bg;
  GdkGC *fg;
  widget = GTK_WIDGET (function);
  window = widget->window;
  bg = widget->style->bg_gc[widget->state];
  fg = widget->style->fg_gc[widget->state];

  /* Fill the background */
  if (rectangle != NULL)
    {
      x = rectangle->x;
      y = rectangle->y;
      width = rectangle->width;
      height = rectangle->height;
    }
  else
    {
      x = 0;
      y = 0;
      width = function->draw_width;
      height = function->draw_height;
    }
  gdk_draw_rectangle (window, bg, TRUE, x, y, width, height);

  /* Draw the curve */
  if (function->cached_points == NULL || function->cached_points_invalid)
    recompute_cached_points (function);
  gdk_draw_lines (window,
		  fg, function->cached_points, function->num_cached_points);
}

void gdam_smooth_function_set_spline(GdamSmoothFunction*  smooth_function,
                                     GdamSpline*          spline)
{
  GdamSpline *old_spline = smooth_function->function;
  if (old_spline != NULL && old_spline != spline)
    gdam_spline_delete (old_spline);
  smooth_function->function = spline;
  smooth_function->cached_points_invalid = TRUE;
  redraw_curve (smooth_function, NULL);
}

static inline gfloat
gdam_smooth_function_x_from_screen(GdamSmoothFunction* function,
                                   int                 x)
{
  return x * (function->max_input - function->min_input)
    / function->draw_width + function->min_input;
}
static inline gfloat
gdam_smooth_function_y_from_screen(GdamSmoothFunction* function,
                                   int                 y)
{
  return y * (function->max_output - function->min_output)
    / function->draw_height + function->min_output;
}
static inline int
gdam_smooth_function_x_to_screen(GdamSmoothFunction* function,
                                 gfloat              x)
{
  return (x - function->min_input)
    / (function->max_input - function->min_input) * function->draw_width;
}
static inline int
gdam_smooth_function_y_to_screen(GdamSmoothFunction* function,
                                 gfloat              y)
{
  return (y - function->min_output)
    / (function->max_output - function->min_output) * function->draw_height;
}
static void gdam_smooth_function_rebuild_spline(GdamSmoothFunction* function)
{
  /* Recompute a new GdamSpline from keypoints + taughtness */
  g_error ("XXX: unimplemented");
}

static void add_keypoint(GdamSmoothFunction*  function,
                         gfloat               x,
			 gfloat               y)
{
  g_error ("XXX: unimplemented");
}

static void remove_keypoint(GdamSmoothFunction* function,
                            GdamKeypointInfo*   keypoint)
{
  int index = keypoint - function->keypoints;
  g_error ("XXX: unimplemented");
}

static GdamKeypointInfo* gdam_smooth_function_search_keypoint
					(GdamSmoothFunction* function,
					 int                 screen_x,
					 int                 screen_y,
					 int*                drag_offset_x,
					 int*                drag_offset_y)
{
  int i;
  int radius_sqr;
  radius_sqr = function->keypoint_radius + 2;
  radius_sqr *= radius_sqr;
  for (i = 0; i < function->num_keypoints; i++)
    {
      int key_screen_x;
      int key_screen_y;
      int dist_sqr;
      GdamKeypointInfo *keypoint = &function->keypoints[i];
      key_screen_x = gdam_smooth_function_x_to_screen (function,
						       keypoint->pos[0]);
      key_screen_y = gdam_smooth_function_y_to_screen (function,
						       keypoint->pos[1]);

      /* find the offsets */
      key_screen_x -= screen_x;
      key_screen_y -= screen_y;

      /* find the squared distance */
      dist_sqr = key_screen_x * key_screen_x + key_screen_y * key_screen_y;

      if (dist_sqr <= radius_sqr)
	return keypoint;
    }
  return NULL;
}

static int gdam_smooth_function_motion_event      (GtkWidget*       widget,
                                                   GdkEventMotion*  motion)
{
  GdamSmoothFunction *function;
  GdamSpline *spline;
  int index;
  gfloat *keypoints;
  function = GDAM_SMOOTH_FUNCTION (widget);

  /* don't do anything unless we are dragging. */
  if (function->drag_keypoint < 0)
    return 0;
  index = function->drag_keypoint;

  /*
   * Relocate that keypoint.
   */
  {
    gfloat *keypoint = function->keypoints[index].pos;
    int x = motion->x;
    int y = motion->y;
    keypoint[0] = gdam_smooth_function_x_from_screen (function, x);
    keypoint[1] = gdam_smooth_function_y_from_screen (function, y);
  }

  /*
   * Recompute the spline and the rendered segments.
   */
  gdam_smooth_function_rebuild_spline (function);
  gdam_smooth_function_set_spline (function, function->function);
  return 0;
}

static int gdam_smooth_function_button_press_event(GtkWidget*       widget,
                                                   GdkEventButton*  event)
{
  GdamSmoothFunction *function;
  GdamSpline *spline;
  function = GDAM_SMOOTH_FUNCTION (widget);
  spline = function->function;

  switch (event->button)
    {
      /* button 1 pressed */
    case 1:
      {
	GdamKeypointInfo *keypoint;
	/*
	 * Scan through keypoints, looking for a hit.
	 */
	keypoint = gdam_smooth_function_search_keypoint
	  (function, event->x, event->y,
	   &function->drag_offset_x, &function->drag_offset_y);

	if (keypoint != NULL)
	  {
	    /*
	     * Found a hit; begin dragging it.
	     */
	    if (keypoint->is_movable)
	      function->drag_keypoint = keypoint - function->keypoints;
	  }
	else
	  {
	    /*
	     * Maybe add a new keypoint, if it's not
	     * too close to an existing one.
	     */
	    gfloat x, y;
	    x = gdam_smooth_function_x_from_screen (function, event->x);
	    y = gdam_smooth_function_y_from_screen (function, event->y);
	    add_keypoint (function, x, y);
	  }

	break;
      }
    case 3:
      {
	GdamKeypointInfo *keypoint;
	/*
	 * Scan through keypoints, looking for a hit.
	 */
	keypoint = gdam_smooth_function_search_keypoint
	  (function, event->x, event->y,
	   &function->drag_offset_x, &function->drag_offset_y);

	if (keypoint != NULL && keypoint->is_removable)
	  remove_keypoint (function, keypoint);
	break;
      }
    }

  return 0;
}


static void gdam_smooth_function_init(GdamSmoothFunction* smooth_function)
{
}
static void gdam_smooth_function_class_init(GdamSmoothFunctionClass* smooth_function_class)
{
}

GtkType gdam_smooth_function_get_type()
{
  static GtkType smooth_function_type = 0;
  if (!smooth_function_type)
    {
      static const GtkTypeInfo smooth_function_info = {
	"GdamSmoothFunction",
	sizeof (GdamSmoothFunction),
	sizeof (GdamSmoothFunctionClass),
	(GtkClassInitFunc) gdam_smooth_function_class_init,
	(GtkObjectInitFunc) gdam_smooth_function_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_DRAWING_AREA;
      smooth_function_type = gtk_type_unique (parent, &smooth_function_info);
    }
  return smooth_function_type;
}
