/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtkbouncelabel.h"
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>

static GtkObjectClass* parent_class = NULL;

static int  gtk_bounce_label_expose_event        (GtkWidget*        widget,
                                                  GdkEventExpose*   event);
static int  gtk_bounce_label_configure_event     (GtkWidget*        widget,
                                                  GdkEventConfigure*event);
static void gtk_bounce_label_draw                (GtkWidget*        widget,
                                                  GdkRectangle*     area);
static void gtk_bounce_label_draw_default        (GtkWidget*        widget);
static void gtk_bounce_label_draw_focus          (GtkWidget*        widget);
static void gtk_bounce_label_size_request        (GtkWidget*        widget,
                                                  GtkRequisition*   event);
static int  gtk_bounce_label_enter_notify_event  (GtkWidget*        widget,
                                                  GdkEventCrossing* event);
static int  gtk_bounce_label_leave_notify_event  (GtkWidget*        widget,
                                                  GdkEventCrossing* event);
static gboolean  gtk_bounce_label_visibility_notify_event  
						 (GtkWidget*        widget,
                                                  GdkEventVisibility* event);
static void gtk_bounce_label_destroy             (GtkObject*        object);
static void recompute_string_data(GtkBounceLabel* label);


static void gtk_bounce_label_init(GtkBounceLabel* bounce_label)
{

  GTK_WIDGET_SET_FLAGS (bounce_label, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
  gtk_widget_add_events (GTK_WIDGET (bounce_label),
			 GDK_EXPOSURE_MASK
			 | GDK_STRUCTURE_MASK
			 | GDK_ENTER_NOTIFY_MASK
			 | GDK_VISIBILITY_NOTIFY_MASK
			 | GDK_LEAVE_NOTIFY_MASK);
}
static void gtk_bounce_label_class_init(GtkBounceLabelClass* klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  widget_class->expose_event = gtk_bounce_label_expose_event;
  widget_class->configure_event = gtk_bounce_label_configure_event;
  widget_class->enter_notify_event = gtk_bounce_label_enter_notify_event;
  widget_class->leave_notify_event = gtk_bounce_label_leave_notify_event;
  widget_class->visibility_notify_event =
    gtk_bounce_label_visibility_notify_event;
  widget_class->size_request = gtk_bounce_label_size_request;
  widget_class->draw = gtk_bounce_label_draw;
  widget_class->draw_default = gtk_bounce_label_draw_default;
  widget_class->draw_focus = gtk_bounce_label_draw_focus;
  object_class->destroy = gtk_bounce_label_destroy;
}

GtkType gtk_bounce_label_get_type()
{
  static GtkType bounce_label_type = 0;
  if (!bounce_label_type)
    {
      static const GtkTypeInfo bounce_label_info = {
	"GtkBounceLabel",
	sizeof (GtkBounceLabel),
	sizeof (GtkBounceLabelClass),
	(GtkClassInitFunc) gtk_bounce_label_class_init,
	(GtkObjectInitFunc) gtk_bounce_label_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_EVENT_BOX;
      parent_class = gtk_type_class (parent);
      bounce_label_type = gtk_type_unique (parent, &bounce_label_info);
    }
  return bounce_label_type;
}

static void gtk_bounce_label_draw(GtkWidget* widget, GdkRectangle* area)
{
  GtkBounceLabel *label = GTK_BOUNCE_LABEL (widget);
  GdkWindow *window = widget->window;
  GdkFont *font;
  if (!GTK_WIDGET_DRAWABLE (widget))
    {
      return;
    }

  gdk_window_set_back_pixmap (widget->window, NULL, TRUE);
  gdk_window_clear_area (widget->window,
			 area->x, area->y, area->width, area->height);
  if (label->background == NULL)
    {
      gtk_paint_box (widget->style, widget->window,
		     GTK_STATE_NORMAL, GTK_SHADOW_IN,
		     area, widget, "buttondefault",
		     area->x, area->y, area->width, area->height);
    }
  else
    {
      gdk_window_copy_area (window,
			    widget->style->bg_gc[0],
			    0,
			    0,
			    label->background,
			    0,
			    0,
			    widget->allocation.width,
			    widget->allocation.height);
    }

  font = label->font ? label->font : widget->style->font;
  gdk_draw_string (window,
		   font,
		   widget->style->fg_gc[0],
		   label->centered_position.x + label->x_offset,
		   label->centered_position.y, label->text);
}
static void
gtk_bounce_label_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gtk_bounce_label_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

void gtk_bounce_label_set_pixmap(GtkBounceLabel* bounce_label,
                                 GdkPixmap*      pixmap,
				 int		 width,
				 int		 height)
{

  bounce_label->loaded_image = TRUE;

  bounce_label->background = pixmap;
  /*bounce_label->masks[i] = info->mask; */
  bounce_label->size.x = width;
  bounce_label->size.y = height;
}

static int gtk_bounce_label_expose_event(GtkWidget*       widget,
                                         GdkEventExpose*  event)
{
  GtkBounceLabel *button = GTK_BOUNCE_LABEL (widget);
  gtk_widget_draw_default (GTK_WIDGET (button));
  return FALSE;
}
static gboolean gtk_bounce_label_visibility_notify_event(GtkWidget*      widget,
                                                     GdkEventVisibility* event)
{
  GtkBounceLabel *label = GTK_BOUNCE_LABEL (widget);
  recompute_string_data (label);
  gtk_widget_draw_default (GTK_WIDGET (label));
  return FALSE;
}
static int gtk_bounce_label_configure_event       (GtkWidget*       widget,
                                                   GdkEventConfigure*event)
{
  GtkBounceLabel *label = GTK_BOUNCE_LABEL (widget);
  recompute_string_data (label);
  gtk_widget_draw_default (GTK_WIDGET (label));
  return FALSE;
}
static int gtk_bounce_label_enter_notify_event(GtkWidget*       widget,
                                               GdkEventCrossing*event)
{
  GtkBounceLabel *label = GTK_BOUNCE_LABEL (widget);
  GtkWidget *event_widget;
  event_widget = gtk_get_event_widget ((GdkEvent *) event);
  if (event_widget != widget || event->detail == GDK_NOTIFY_INFERIOR)
    return FALSE;
  gtk_bounce_label_start_bouncing (label);
  return FALSE;
}
static int gtk_bounce_label_leave_notify_event(GtkWidget*       widget,
                                               GdkEventCrossing*event)
{
  GtkBounceLabel *label = GTK_BOUNCE_LABEL (widget);
  GtkWidget *event_widget;
  event_widget = gtk_get_event_widget ((GdkEvent *) event);
  if (event_widget != widget || event->detail == GDK_NOTIFY_INFERIOR)
    return FALSE;
  gtk_bounce_label_stop_bouncing (label, FALSE);
  return FALSE;
}
static void gtk_bounce_label_size_request  (GtkWidget*       widget,
                                            GtkRequisition*  requisition)
{
  GtkBounceLabel *bounce_label = GTK_BOUNCE_LABEL (widget);
  requisition->width = bounce_label->size.x;
  requisition->height = bounce_label->size.y;
}

static void gtk_bounce_label_destroy(GtkObject* object)
{
  GtkBounceLabel *label;
  label = GTK_BOUNCE_LABEL (object);
  if (label->font != NULL && label->must_unref_font)
    {
      gdk_font_unref (label->font);
      label->font = NULL;
    }
  g_free (label->text);
  label->text = NULL;

  g_free (label->background_file);
  label->background_file = NULL;

  (*parent_class->destroy) (object);
}

 
static void recompute_string_data(GtkBounceLabel* label)
{
  GdkFont *font;
  GtkWidget *widget;
  widget = GTK_WIDGET (label);
  if (label->font != NULL)
    font = label->font;
  else
    font = widget->style->font;
  if (font != NULL)
    {
      int lbearing, rbearing, width, ascent, descent;
      int scr_top;
      int scr_width = widget->allocation.width;
      int scr_height = widget->allocation.height;

      gdk_string_extents (font, label->text,
			  &lbearing, &rbearing, &width, &ascent, &descent);
      scr_top = scr_height - ascent + descent;
      scr_top /= 2;

      label->centered_position.x = (scr_width - width) / 2;
      label->centered_position.y = scr_top + ascent;
      if (scr_width >= width)
	label->max_x_offset = 0;
      else
	label->max_x_offset = -label->centered_position.x + 3;
      if (label->max_x_offset < 0)
	label->max_x_offset = 0;
      label->x_offset = CLAMP (label->x_offset,
			       -label->max_x_offset, +label->max_x_offset);
    }
}

void gtk_bounce_label_set_text(GtkBounceLabel* label,
                               const char*     text)
{
  char *old_text = label->text;
  GdkFont *font = label->font;
  label->text = g_strdup (text);
  g_free (old_text);
  if (font == NULL)
    font = GTK_WIDGET (label)->style->font;
  recompute_string_data (label);
  gtk_widget_queue_draw (GTK_WIDGET (label));
}


static gboolean handle_timeout(GtkBounceLabel* label)
{
  gboolean was_negative = (label->x_offset < 0);
  gtk_widget_queue_draw (GTK_WIDGET (label));
  if (label->bouncing_right)
    {
      label->x_offset += 8;
      if (!label->is_bouncing && was_negative && label->x_offset >= 0)
	{
	  goto stop;
	}
      if (label->x_offset > +label->max_x_offset)
	{
	  label->x_offset = +label->max_x_offset;
	  label->bouncing_right = 0;
	}
    }
  else
    {
      label->x_offset -= 8;
      if (!label->is_bouncing && !was_negative && label->x_offset < 0)
	{
	  goto stop;
	}
      if (label->x_offset < -label->max_x_offset)
	{
	  label->x_offset = -label->max_x_offset;
	  label->bouncing_right = 1;
	}
    }
  return TRUE;

stop:
  label->x_offset = 0;
  label->bounce_timer_id = 0;
  return FALSE;
}

void       gtk_bounce_label_start_bouncing        (GtkBounceLabel* label)
{
  label->is_bouncing = 1;
  if (label->bounce_timer_id != 0)
    return;

  label->bounce_timer_id
    = g_timeout_add (100, (GSourceFunc) handle_timeout, label);
}

void       gtk_bounce_label_stop_bouncing         (GtkBounceLabel* label,
                                                   gboolean        now)
{
  label->is_bouncing = 0;
  if (now && label->bounce_timer_id != 0)
    {
      g_source_remove (label->bounce_timer_id);
      label->bounce_timer_id = 0;
      label->x_offset = 0;
    }
}
