/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtkmessagequeue.h"
#include <string.h>

GtkMessageQueue*   gtk_message_queue_new    (int              max_bytes,
                                             int              max_entries)
{
  GtkMessageQueue *queue;
  queue = g_new (GtkMessageQueue, 1);
  queue->max_entries = max_entries;
  queue->max_bytes = max_bytes;
  queue->num_entries = 0;
  queue->num_bytes = 0;
  queue->v_adjustment = (GtkAdjustment *)
    gtk_adjustment_new (0.0, 0.0, 1.1, 0.1, 0.1, 0.1);
  queue->text = (GtkText *) gtk_text_new (NULL, queue->v_adjustment);
  queue->first_bytes = NULL;
  queue->last_bytes = NULL;
  gtk_widget_show (GTK_WIDGET (queue->text));
  gtk_object_ref (GTK_OBJECT (queue->text));
  return queue;
}

void               gtk_message_queue_append (GtkMessageQueue* message_queue,
                                             const char*      string)
{
  GtkText *text = message_queue->text;
  int bytes = strlen (string);
  char *with_nl;
  int new_num_bytes;
  int new_num_entries = 1 + message_queue->num_entries;
  int to_chop;

  /* Make sure the string has a newline on it,
   * but don't add an extra newline for no reason. */
  if (strchr (string, '\n') == NULL)
    {
      with_nl = g_strdup_printf ("%s\n", string);
      string = with_nl;
      bytes++;
    }
  else
    {
      with_nl = NULL;
    }

  new_num_bytes = bytes + message_queue->num_bytes;
  gtk_text_freeze (text);

  /* Discard entries and count how many bytes to throw away. */
  to_chop = 0;
  while (new_num_entries > message_queue->max_entries
	 || new_num_bytes > message_queue->max_bytes)
    {
      int size;
      GSList *start;
      start = message_queue->first_bytes;
      g_return_if_fail (start != NULL);
      size = GPOINTER_TO_INT (start->data);
      new_num_bytes -= size;
      new_num_entries--;
      to_chop += size;
      message_queue->first_bytes = start->next;
      if (start->next == NULL)
	message_queue->last_bytes = NULL;
      g_slist_free_1 (start);
    }

  /* do the text-pad manipulation */
  gtk_text_set_point (text, 0);
  gtk_text_forward_delete (text, to_chop);
  message_queue->num_bytes -= to_chop;
  gtk_text_set_point (text, message_queue->num_bytes);
  gtk_text_insert (text, NULL, NULL, NULL, string, bytes);
  message_queue->num_bytes += bytes;
  message_queue->num_entries = new_num_entries;

  gtk_adjustment_set_value (message_queue->v_adjustment,
			    message_queue->v_adjustment->upper
			    - message_queue->v_adjustment->page_size);

  gtk_text_thaw (text);

  /* add the size to the size-list */
  message_queue->last_bytes
    = g_slist_append (message_queue->last_bytes, GINT_TO_POINTER (bytes));
  if (message_queue->first_bytes == NULL)
    message_queue->first_bytes = message_queue->last_bytes;
  else
    message_queue->last_bytes = message_queue->last_bytes->next;

  g_free (with_nl);
  g_return_if_fail (message_queue->num_bytes == new_num_bytes);
}

void               gtk_message_queue_destroy(GtkMessageQueue* queue)
{
  gtk_object_sink (GTK_OBJECT (queue->text));
  gtk_object_unref (GTK_OBJECT (queue->text));
  g_slist_free (queue->first_bytes);
  g_free (queue);
}
