/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gtkpixdial.h"
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <stdlib.h>
#include <math.h>

enum {
	MOTION,
	LAST_SIGNAL
};
static guint gtk_pix_dial_signals[LAST_SIGNAL] = { 0 };

static int  gtk_pix_dial_expose_event           (GtkWidget*        widget,
                                              GdkEventExpose*   event);
static int  gtk_pix_dial_configure_event        (GtkWidget*        widget,
                                              GdkEventConfigure*event);
static void gtk_pix_dial_draw                   (GtkWidget*        widget,
                                              GdkRectangle*     area);
static void gtk_pix_dial_draw_default           (GtkWidget*        widget);
static void gtk_pix_dial_draw_focus             (GtkWidget*        widget);
static void gtk_pix_dial_size_request           (GtkWidget*        widget,
                                              GtkRequisition*   event);
static int  gtk_pix_dial_press_event            (GtkWidget*        widget,
                                              GdkEventButton*   event);
static int  gtk_pix_dial_release_event          (GtkWidget*        widget,
                                              GdkEventButton*   event);
static int  gtk_pix_dial_motion_event           (GtkWidget*        widget,
                                              GdkEventMotion*   event);
static void maybe_redraw_image               (GtkPixDial*         dial);
static GtkObjectClass* parent_class;
static void gtk_pix_dial_init(GtkPixDial* pix_dial)
{

  GTK_WIDGET_SET_FLAGS (pix_dial, GTK_CAN_FOCUS | GTK_RECEIVES_DEFAULT);
  gtk_widget_add_events (GTK_WIDGET (pix_dial),
			 GDK_EXPOSURE_MASK
			 | GDK_STRUCTURE_MASK
			 | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
  pix_dial->wrapcount = 1;
}
static void gtk_pix_dial_class_init(GtkPixDialClass* klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);
  widget_class->expose_event = gtk_pix_dial_expose_event;
  widget_class->configure_event = gtk_pix_dial_configure_event;
  widget_class->button_press_event = gtk_pix_dial_press_event;
  widget_class->button_release_event = gtk_pix_dial_release_event;
  widget_class->motion_notify_event = gtk_pix_dial_motion_event;
  widget_class->size_request = gtk_pix_dial_size_request;
  widget_class->draw = gtk_pix_dial_draw;
  widget_class->draw_focus = gtk_pix_dial_draw_focus;
  widget_class->draw_default = gtk_pix_dial_draw_default;
  gtk_pix_dial_signals[MOTION] =
    gtk_signal_new ("motion",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GtkPixDialClass, motion),
		    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, gtk_pix_dial_signals, LAST_SIGNAL);
}

GtkType gtk_pix_dial_get_type()
{
  static GtkType pix_dial_type = 0;
  if (!pix_dial_type)
    {
      static const GtkTypeInfo pix_dial_info = {
	"GtkPixDial",
	sizeof (GtkPixDial),
	sizeof (GtkPixDialClass),
	(GtkClassInitFunc) gtk_pix_dial_class_init,
	(GtkObjectInitFunc) gtk_pix_dial_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GTK_TYPE_EVENT_BOX;
      parent_class = gtk_type_class (parent);
      pix_dial_type = gtk_type_unique (parent, &pix_dial_info);
    }
  return pix_dial_type;
}

static void
gtk_pix_dial_draw_focus (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void
gtk_pix_dial_draw_default (GtkWidget *widget)
{
  gtk_widget_draw (widget, NULL);
}

static void maybe_redraw_image(GtkPixDial* dial)
{
  int new_index;
  gfloat fraction;
  GtkAdjustment *adj = dial->adjustment;
  gdouble value = adj->value;

  if (dial->is_int)
    value = (gdouble)((int)(adj->value + 0.5));

  fraction = (value - adj->lower) / (adj->upper - adj->lower);
  new_index = floor (fraction * dial->num_tiles * dial->wrapcount);
  new_index = new_index % dial->num_tiles;
  new_index = CLAMP (new_index, 0, dial->num_tiles - 1);
  /* We ensure that at full value we show the last picture, not the first. 
   * This distinction is nice for dials driving int args. */ 
  /* TODO: is this nonsense for certain dial configurations?? If the tiles 
   * don't fit evenly, this might be wrong. */
  if (value == dial->adjustment->upper)
    new_index = dial->num_tiles - 1;
  if (dial->images_reversed)
    new_index = dial->num_tiles - 1 - new_index;
  if (new_index != dial->last_image_drawn)
    {
      dial->last_image_drawn = new_index;
      gtk_widget_draw_default (GTK_WIDGET (dial));
    }
}

static void
reset_delta_x(GtkPixDial* dial)
{
  GtkAdjustment *adjustment = dial->adjustment;
  g_return_if_fail (adjustment != NULL);
  gtk_adjustment_set_value (dial->adjustment, dial->orig_value);
}


static void
update_delta_x(GtkPixDial* dial, int delta_x)
{
  GtkAdjustment *adjustment = dial->adjustment;
  gfloat amount = delta_x;
  if (dial->screen_range == 0)
    {
      return;
    }
  g_return_if_fail (adjustment != NULL);
  amount /= dial->screen_range;
  amount *= dial->adjustment->upper - dial->adjustment->lower;
  amount += adjustment->value;
  amount = CLAMP (amount, adjustment->lower, adjustment->upper);
  gtk_adjustment_set_value (dial->adjustment, amount);
}

static void gtk_pix_dial_draw(GtkWidget* widget, GdkRectangle* area)
{
  GtkPixDial *dial = GTK_PIX_DIAL (widget);
  GdkWindow *window = widget->window;
  int window_width = widget->allocation.width;
  int window_height = widget->allocation.height;
  int tile_top, x_margin, y_margin;

  if (!GTK_WIDGET_DRAWABLE (widget))
    return;

  x_margin = (window_width - dial->tile_width) / 2;
  y_margin = (window_height - dial->tile_height) / 2;
  tile_top = dial->tile_height * dial->last_image_drawn;
#if 1
  gdk_draw_pixmap (widget->window, widget->style->fg_gc[0], dial->pixmap, 0, tile_top,	/* src x,y */
		   x_margin, y_margin,	/* dest x,y */
		   dial->tile_width, dial->tile_height);	/* extent */
#else
  gdk_window_copy_area (widget->window, widget->style->fg_gc[0], x_margin, y_margin,	/* dest x,y */
			dial->pixmap, 0, tile_top,	/* src x,y */
			dial->tile_width, dial->tile_height);	/* extent */
#endif
}

static int gtk_pix_dial_press_event(GtkWidget*       widget,
                                 GdkEventButton*  event)
{
  GtkPixDial *dial = GTK_PIX_DIAL (widget);

  if (dial->no_mouse)
    return FALSE;

  if ((event->button == 4) || (event->button == 5))
    {
      gdouble page_size;
      g_return_val_if_fail (dial->adjustment != NULL, FALSE);
      page_size = (dial->adjustment->upper - dial->adjustment->lower) 
		  / dial->num_tiles;
      if (event->button == 4)
	gtk_adjustment_set_value (dial->adjustment, 
				  dial->adjustment->value + page_size);
      else 
	gtk_adjustment_set_value (dial->adjustment, 
				  dial->adjustment->value - page_size);
      return FALSE;
      //return TRUE;  /* trying to keep tooltips from closing... */
    }
  gtk_grab_add (widget);
  dial->last_drag_x = event->x;
  dial->is_dragging = TRUE;
  return FALSE;
  //return TRUE;  /* trying to keep tooltips from closing... */
}

static int gtk_pix_dial_release_event(GtkWidget*       widget,
                                   GdkEventButton*  event)
{
  gboolean clicked = FALSE;
  GtkPixDial *dial = GTK_PIX_DIAL (widget);

  if (dial->no_mouse)
    return FALSE;

  if (event->button == 4)
    return 0;
  else if (event->button == 5)
    return 0;

  gtk_grab_remove (widget);

  if (dial->reset_on_release)
    reset_delta_x (dial);

  dial->is_dragging = FALSE;

  return FALSE;
}

static int gtk_pix_dial_expose_event(GtkWidget*       widget,
                                  GdkEventExpose*  event)
{
  GtkPixDial *dial = GTK_PIX_DIAL (widget);
  gtk_widget_draw_default (GTK_WIDGET (dial));
  return FALSE;
}
static int gtk_pix_dial_configure_event(GtkWidget*         widget,
                                     GdkEventConfigure* event)
{
  GtkPixDial *dial = GTK_PIX_DIAL (widget);
  gtk_widget_draw_default (GTK_WIDGET (dial));
  return FALSE;
}

static void gtk_pix_dial_size_request    (GtkWidget*       widget,
                                       GtkRequisition*  requisition)
{
  GtkPixDial *dial = GTK_PIX_DIAL (widget);
  requisition->width = dial->tile_width;
  requisition->height = dial->tile_height;
}

static int  gtk_pix_dial_motion_event          (GtkWidget*       widget,
                                             GdkEventMotion*  event)
{
  int x, y;
  guint state;
  GtkPixDial *dial;
  dial = GTK_PIX_DIAL (widget);
  if (!dial->is_dragging)
    return 0;

  x = event->x;
  y = event->y;
  state = event->state;
  if (event->is_hint || event->window != widget->window)
    gdk_window_get_pointer (widget->window, &x, &y, &state);

  update_delta_x (dial, x - dial->last_drag_x);
  dial->last_drag_x = x;
  return 0;
}

void gtk_pix_dial_set_adjustment(GtkPixDial*       dial,
                                 GtkAdjustment*    adjustment)
{
  GtkAdjustment *old_adj = dial->adjustment;
  if (old_adj == adjustment)
    return;
  if (old_adj != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (old_adj), dial->connect_id);
      gtk_object_unref (GTK_OBJECT (old_adj));
    }
  dial->adjustment = adjustment;
  if (adjustment == NULL)
    dial->connect_id = 0;
  else
    {
      GtkObject *object = GTK_OBJECT (adjustment);
      gtk_object_ref (object);
      gtk_object_sink (object);
      dial->connect_id
	= gtk_signal_connect_object
	(object,
	 "value_changed",
	 GTK_SIGNAL_FUNC (maybe_redraw_image), GTK_OBJECT (dial));
      update_delta_x (dial, 0);
    }

}


void
gtk_pix_dial_set_pixmap       (GtkPixDial      *dial,
                               GdkPixmap       *dial_pixmap,
                               GdkBitmap       *dial_pixmap_mask)
{
  gint w, h;
  dial->pixmap = dial_pixmap;
  dial->mask = dial_pixmap_mask;
  gdk_window_get_size (dial_pixmap, &w, &h);
  dial->tile_width = w;
  dial->tile_height = h / dial->num_tiles;
  gtk_widget_draw (GTK_WIDGET (dial), NULL);
}
