/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <stdlib.h>
#include <string.h>
#include "gdamskin.h"
#include "gdamcrossfadeskin.h"
#include "gdamdeviceswitcherskin.h"
#include "gdampacker.h"
#include "gdamdefault.h"
#include "gdamdnd.h"
#include "gdamlaunchserverdialog.h"
#include <gdam/component/gdamutil.h>
#include <gdam/component/gdamsongdatabase.h>
#include <gdam/component/gdambuffermanager.h>
#include <gdam/common/gdamautoconf.h>
#include <gdam/common/gdamplugin.h>
#include <gdam/component/gdaminputdevice.h>
#include <gdam/common/gdamwaitpid.h>
#include <gdam/component/gdamhistory.h>
#include <unistd.h>

char* gdam_rcpath = 0;

extern int gdam_producer_mp3_default_buffer_size;

/* TODO: use search path for consistency?? */
static const char* default_rcfiles[] = { 
                                  	"~/.gdam/gdamrc", 
				  	"~/.gdam/default.rc", 
					GDAM_SKIN_PREFIX "/default.rc",
					GDAM_SKIN_PREFIX "/default/default.rc",
					".gdamrc",
					"~/.gdamrc",
				 	"/usr/local/lib/gdam/skins/default.rc",
				  	"/etc/gdam/gdamrc", 
					"default.rc",
					0 };

static void usage(const char* progname)
{
  fprintf (stderr, "usage: %s [--device device_to_open]\n"
	   "\n"
	   "Options:\n"
	   "\t--gdam-server server : server to connect to\n"
	   "\t--device device      : set the device\n"
	   "\t--noladspa           : don't scan for ladspa plugins\n"
	   "\t--rcfile rcfile      : use alternate rc file\n"
	   "\t--rcpath rcpath      : use alternate path when "
	   "searching for rc file\n"
	   "\t--database db        : specify a path to the song "
	   "database [abbrev: -d]\n"
	   "\t--define variable    : define a variable, which "
	   "may turn\n"
	   "\t                       certain features "
	   "on or off\n"
	   "\t--mp3-buffer buflen  : increase the default mp3 "
	   "file buffer (in seconds)\n"
	   "\nPrivate:\n"
	   "\t--mix-class type     : specify a type (instead of "
	   "GdamMix) for the mix\n"
	   "\t--launch-[in]secure  : if needed, launch the gdam-server\n"
	   "\nUndocumented: drag2 drag3\n", progname);
  exit (1);
}

static void open_device(gpointer list_data, gpointer user_data)
{
  gdam_toplevel_open_device (GDAM_TOPLEVEL (user_data),
			     (const char *) list_data, NULL, NULL, NULL);
}

#if 0
static gboolean is_doc_from_glade(xmlDoc* doc)
{
  return TRUE;
}
#endif

#if HAVE_LADSPA_H
static void rescan_ladspa_database()
{
  GtkType type;
  type = gdam_plugin_get_type_by_name ("skin", "GdamLadspaSkin");
  (void) gtk_type_class (type);
}
static void on_finished_scanning_ladspa_plugins(gpointer         user_data,
                                                GdamWaitPidInfo* info)
{
  if (info->exited)
    {
      if (info->d.exit_status == 127)
	{
	  g_warning ("gdamladspapluginscanner was not found "
		     "on your path... no ladspa plugins will "
		     "be available");
	}
      else
	{
	  //g_message("scanning ladspa database");
	  rescan_ladspa_database ();
	}
    }
  else
    {
      g_warning ("gdamladspapluginscanner died from a signal");
    }
}
static void spawn_scan_ladspa_plugins()
{
  int pid;

  pid = fork ();
  if (pid < 0)
    {
      g_warning ("couldn't fork to spawn ladspa scanner... ignoring");
      return;
    }

  if (pid == 0)
    {
      char *arg;
      arg = g_strdup_printf ("%s/.gdam/ladspa.xml", g_get_home_dir ());
      execl ("../component/plugins/gdamladspapluginscanner",
	     "gdamladspapluginscanner", "--append", arg, NULL);
      execlp ("gdamladspapluginscanner",
	      "gdamladspapluginscanner", "--append", arg, NULL);
      _exit (127);
    }

  gdam_wait_pid (pid, on_finished_scanning_ladspa_plugins, NULL, NULL);
}
#endif

typedef struct _MainData MainData;
struct _MainData {
	GdamToplevel* toplevel;
	GdamChannel* channel;
	GdamSongDatabase* db;
	GSList* xml_files;
	const char* rcfile;
	int skins_loaded;
	gboolean done_initializing;

	gboolean launch_server;
	gboolean launch_server_secure;
};

static void on_launch_server(gpointer unused)
{
  (void) unused;
}

static GdamChannelServerLaunch 
do_launch_server_dialog(GdamChannel* channel,
                        gpointer     user_data)
{
  MainData *main_data = user_data;

  if (main_data->launch_server)
    {
      if (main_data->launch_server_secure)
	return GDAM_CHANNEL_SERVER_LAUNCH_SECURE;
      else
	return GDAM_CHANNEL_SERVER_LAUNCH_INSECURE;
    }

  (void) gdam_launch_server_dialog_run (channel, on_launch_server, NULL,	/* on_launch_server_data unneeded */
					NULL);

  return GDAM_CHANNEL_SERVER_LAUNCH_DEFER;
}


static  void   on_connection(GdamChannel*   channel,
	                     gpointer       user_data)
{
  MainData *main_data = user_data;
  GSList *xml_files;
  g_assert (main_data->channel == channel);

  main_data->toplevel = gdam_toplevel_get_global (channel);
  g_list_foreach (main_data->channel->device_list, open_device, 
		   main_data->toplevel);

  if (main_data->xml_files == NULL)
    main_data->xml_files = g_slist_prepend (NULL, "launcher.xml");

  for (xml_files = main_data->xml_files;
       xml_files != NULL; xml_files = xml_files->next)
    {
      const char *xml_file = xml_files->data;
      const char *fname;
      fname = gdam_skin_search_for_xml_file (xml_file);
      if (fname == NULL)
	{
	  g_warning ("couldn't find xml file %s", fname);
	  continue;
	}
      if (gdam_skin_open (fname, channel))
	main_data->skins_loaded++;
    }

  if (main_data->skins_loaded == 0)
    g_error (_("no skins were successfully made:  aborting"));
  main_data->db = gdam_song_database_system ();
  if (g_hash_table_size (main_data->db->name_lookup) == 0)
    {
      const char *fname;
      fname = gdam_skin_search_for_xml_file ("songdbwarning.xml");
      if (fname == NULL)
	{
	  g_warning ("couldn't find xml file %s", fname);
	}
      else
	gdam_skin_open (fname, channel);
    }

}

/* Load certain, known definitions from the user's history. */
/* TODO: should we load every definition?? */
static void
load_stored_configuration ()
{
  char *definitions[10] = {"unstable", "3bandeq", "dj_mixer", "mono_cue", "simple", "builder_macro", "RememberDefinitions", NULL};
  int idx;
  for (idx = 0; definitions[idx] != NULL; idx++)
    {
      const char *val;
      val = gdam_history_get_entry ("SkinDefinitions", definitions[idx]);
      if (val && atoi (val))
	gdam_skin_set_whether_defined (definitions[idx], TRUE);
      else if (val && !atoi (val))
	gdam_skin_set_whether_defined (definitions[idx], FALSE);
    }
}

int main(int argc, char** argv)
{
  //GdamLauncher* launcher;
  int i;
  /* static so it is available in atexit routines */
  static MainData main_data;
  const char *remember;
  gboolean without_ladspa = FALSE;

  memset (&main_data, 0, sizeof (MainData));

  g_set_prgname (argv[0]);

  /* XXX: instead of this block, we should break apart the
   *      argument parsing/connections api of gdam_channel_init
   *      and do all of the parsing prior to connecting.
   *
   *      But for now...
   */
  if (argv[1] != NULL)
    {
      if (strcmp (argv[1], "--help") == 0
	  || strcmp (argv[1], "-h") == 0 || strcmp (argv[1], "-help") == 0)
	{
	  usage (argv[0]);
	  return 0;
	}
    }

  gdk_init (&argc, &argv);
  gtk_init (&argc, &argv);
  if (!gdam_channel_init_full (&argc, &argv, &main_data.channel,
			       do_launch_server_dialog,
			       &main_data, on_connection, &main_data))
    {
      g_error (_("error creating connection to gdam-server"));
      return 1;
    }
  if (argv[1] != NULL)
    {
      if (strcmp (argv[1], "-v") == 0 || strcmp (argv[1], "--version") == 0)
	{
	  printf (_("gdam-launcher version %s.\n"
		    "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
		    "\n"
		    "THIS PROGRAM COMES WITH NO WARRANTY! "
		    "USE AT YOUR OWN RISK.\n"), VERSION);
	  exit (0);
	}
    }
  g_return_val_if_fail (main_data.channel != NULL, 1);
  glade_init ();

  remember = gdam_history_get_entry ("SkinDefinitions", "RememberDefinitions");
  if (remember && atoi (remember))
    {
      load_stored_configuration ();
    }

  for (i = 1; i < argc; i++)
    {
      char *opt = argv[i];
      if (argv[i][0] != '-')
	{
	  main_data.xml_files
	    = g_slist_prepend (main_data.xml_files, argv[i]);
	  continue;
	}
      while (*opt == '-')
	opt++;
      if (strcmp (opt, "h") == 0 || strcmp (opt, "help") == 0)
	usage (argv[0]);
      if (strcmp (opt, "mix-class") == 0)
	{
	  const char *mix_class = argv[++i];
	  if (mix_class == NULL)
	    {
	      g_warning (_("%s must be followed by an argument"),
			 argv[i - 1]);
	      return 1;
	    }
	  gdam_toplevel_server_mix_class_name = mix_class;
	  continue;
	}
      if (strcmp (opt, "rcfile") == 0)
	{
	  main_data.rcfile = argv[++i];
	  continue;
	}
      if (strcmp (opt, "rcpath") == 0)
	{
	  gdam_rcpath = argv[++i];
	  continue;
	}
      if (strcmp (opt, "launch-secure") == 0)
	{
	  main_data.launch_server = TRUE;
	  main_data.launch_server_secure = TRUE;
	  continue;
	}
      if (strcmp (opt, "launch-insecure") == 0)
	{
	  main_data.launch_server = TRUE;
	  main_data.launch_server_secure = FALSE;
	  continue;
	}
      if (strcmp (opt, "drag2") == 0)
	{
	  gdam_drag_mask = GDK_BUTTON2_MASK;
	  continue;
	}
      if (strcmp (opt, "drag3") == 0)
	{
	  gdam_drag_mask = GDK_BUTTON3_MASK;
	  continue;
	}
      if (strcmp (opt, "drag23") == 0)
	{
	  gdam_drag_mask = GDK_BUTTON2_MASK | GDK_BUTTON3_MASK;
	  continue;
	}
      if (strcmp (opt, "define") == 0)
	{
	  ++i;
	  if (argv[i] == NULL)
	    {
	      g_error (_("%s must be followed by an argument"), argv[i - 1]);
	    }
	  gdam_skin_set_whether_defined (argv[i], TRUE);
	  continue;
	}

      if (strcmp (opt, "mp3-buffer") == 0)
	{
	  gdouble secs;
	  ++i;
	  if (argv[i] == NULL)
	    {
	      g_error (_("%s must be followed by an argument"), argv[i - 1]);
	    }
	  secs = strtod (argv[i], NULL);
	  gdam_producer_mp3_default_buffer_size = (44100.0 * secs);
	  if (0 < gdam_producer_mp3_default_buffer_size
	      && gdam_producer_mp3_default_buffer_size < 10240)
	    {
	      gdam_producer_mp3_default_buffer_size = 10240;
	    }
	  continue;
	}

      if (strcmp (opt, "d") == 0 || strcmp (opt, "database") == 0)
	{
	  ++i;
	  if (argv[i] == NULL)
	    {
	      g_error (_("%s must be followed by an argument"), argv[i - 1]);
	    }
	  gdam_song_database_default_filename = argv[i];
	  continue;
	}
      if (strcmp (opt, "noladspa") == 0)
	{
	  without_ladspa = TRUE;
	  continue;
	}
      g_warning ("unknown option: %s", argv[i]);
      usage (argv[0]);
    }
  if (main_data.rcfile == NULL)
    main_data.rcfile = gdam_default_file_find (default_rcfiles);
  if (main_data.rcfile == NULL)
    {
      g_warning (_("Warning: no rcfile found.  " "Colors will be gray."));
    }
  else if (!gdam_file_exists (main_data.rcfile))
    {
      g_warning (_("Warning: rcfile %s not found."), main_data.rcfile);
    }
  else
    {
      g_message (_("Using rcfile %s"), main_data.rcfile);
      gtk_rc_parse (main_data.rcfile);
    }

  if (main_data.channel->device_list == NULL)
    {
      char *device_name = getenv ("GDAM_DEVICE");
      if (device_name)
	main_data.channel->device_list = g_list_append 
						(main_data.channel->device_list,
						 device_name);
      device_name = getenv ("GDAM_DEVICE2");
      if (device_name)
	main_data.channel->device_list = g_list_append 
						(main_data.channel->device_list,
						 device_name);
    }

  if (main_data.channel->device_list == NULL)
    {
      g_message (_("Using /dev/dsp since no --device was given"));
      main_data.channel->device_list = g_list_append 
				   (main_data.channel->device_list, "/dev/dsp");
    }

  (void) gtk_type_class (GDAM_TYPE_PACKER);
  (void) gtk_type_class (GDAM_TYPE_CROSSFADE_SKIN);
  (void) gtk_type_class (GDAM_TYPE_DEVICE_SWITCHER_SKIN);

  /* Get triangle, square, sine waves. */
  gdam_buffer_manager_init (main_data.channel);

  /* Initialize the input device database. */
  if (!gdam_input_device_load_database (NULL))
    {
      g_message (_("no input-device database"));
    }
#if HAVE_LADSPA_H
  if (!without_ladspa)
    spawn_scan_ladspa_plugins ();
  else
    g_message ("skipping ladspa initialization.");
#endif
  gtk_main ();
  return 0;
}
