/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuffereditorskin.h"
#include <gdam/common/gdam-protocol.h>
#include <gtk/gtksignal.h>
#include "../gdamskinutil.h"
#include <gdam/component/gdambuffermanager.h>
#include "../../component/plugins/gdamproducercache.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtklabel.h>
#include <gdam/common/gdamplugin.h>
static gboolean       gdam_buffer_editor_build_new_model
                                                 (GdamSkin      *skin,
                                                  GdamChannel   *channel,
                                                  int            num_options,
                                                  char         **options);

static gboolean gdam_buffer_editor_viewport_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamBufferEditorSkin *editor = GDAM_BUFFER_EDITOR_SKIN (skin);
  if (!GTK_IS_CONTAINER (object))
    {
      g_warning ("cannot add buffer_viewport to non-container");
      return TRUE;
    }
  if (!editor->viewport)
    {
      g_warning ("cannot add non-existant buffer_viewport");
      return TRUE;
    }
  gtk_container_add (GTK_CONTAINER (object),
		     GDAM_SKIN (editor->viewport)->main_widget);
  return TRUE;
}

/* Allow access to viewport's handlers */
static gboolean gdam_buffer_editor_skin_defer_to_viewport_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamBufferEditorSkin *editor = GDAM_BUFFER_EDITOR_SKIN (skin);
  GdamSkin *subskin = GDAM_SKIN (editor->viewport);
  if (!subskin)
    {
      g_warning ("missing subskin in %s", handler_names[0]);
      return FALSE;
    }
  if (!gdam_skin_do_connect (subskin,
			     object,
			     signal_name,
			     num_handler_names - 1, 
			     handler_names + 1))
    {
      g_warning ("gdam_skin_do_connect failed on deferred handler: "
		 "num_args=%d, handler_name=%s, skintype=%s",
		 num_handler_names,
		 handler_names[1],
		 gtk_type_name (GTK_OBJECT (subskin)->klass->type));
      return FALSE;
    }
  return TRUE;
}

static void 
do_load_default_buffer (GdamBufferEditorSkin *editor_skin)
{
  int id;
  if (editor_skin->viewport == NULL)
    return;
  id = gdam_buffer_manager_obtain_id (NULL);
  gdam_buffer_manager_unref (id, NULL);
  gdam_buffer_viewport_set_buffer_id (editor_skin->viewport, id);
}

static void 
do_set_defaults(GdamBufferEditorSkin *editor_skin)
{
  if (editor_skin->viewport == NULL)
    return;
  gdam_buffer_viewport_set_buffer_id (editor_skin->viewport, 
                                      editor_skin->viewport->buffer_id);
}

static void on_buffer_changed (GdamBufferEditorSkin *skin)
{
  int id = skin->viewport->buffer_id;
  skin->read_only = gdam_buffer_manager_is_readonly (id);
}

static void zoom_to_selection(GdamBufferEditorSkin* editor_skin)
{
  GdamBufferViewport *viewport = editor_skin->viewport;
  if (viewport == NULL)
    return;
  if (viewport->has_selection)
    {
      viewport->time_start = viewport->selection_start;
      viewport->time_length = viewport->selection_length;
      gdam_buffer_viewport_changed (viewport);
    }
}

static gboolean gdam_buffer_editor_zoom_to_selection_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamBufferEditorSkin *editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (zoom_to_selection),
			     (GtkObject *) editor_skin);
  return TRUE;
}

static void play_callback(GdamBufferEditorSkin *editor_skin)
{
  GdamProducerCache *cache;
  GdamSkin *skin = GDAM_SKIN (editor_skin);
  GdamBufferViewport *viewport = editor_skin->viewport;
  int samps_per_sec;
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (viewport->buffer_id != 0);
  samps_per_sec = viewport->samples_per_second;
  cache = GDAM_PRODUCER_CACHE (skin->model);
  g_return_if_fail (cache != NULL);
  cache->buffer_id = viewport->buffer_id;
  if (viewport->has_selection)
    {
      cache->start = (int) (viewport->selection_start * samps_per_sec);
      cache->length = (int) (viewport->selection_length * samps_per_sec);
    }
  else
    {
      cache->start = (int) (viewport->time_start * samps_per_sec);
      cache->length = (int) (viewport->time_length * samps_per_sec);
    }
  gdam_producer_play (GDAM_PRODUCER (cache), NULL);
}

typedef struct _CutInfo CutInfo;
struct _CutInfo {
	char* description;
	GdamChannel* channel;
	GdamBufferViewport* viewport;
	GdamBufferEditorSkin *editor_skin;
	gboolean make_buffer;
};

static void successfully_cut(int num_args, GdamArg* args, gpointer user_data)
{
  int buffer_id = 0;
  GdamBufferManagerInfo *info;
  CutInfo *cut_info = user_data;
  GdamBufferViewport *viewport = cut_info->viewport;
  int i;
  for (i = 0; i < num_args; i++)
    {
      if (strcmp (args[i].name, "buffer_id") == 0
	  && args[i].type == GDAM_ARG_TYPE_INT)
	{
	  buffer_id = GDAM_ARG_INT (args + i);
	  continue;
	}
      gdam_debug ("got unknown arg %s as response to buffer_cut",
		  args[i].name);
    }

  gdam_buffer_viewport_buffer_changed (viewport);

  if (buffer_id == 0)
    {
      if (cut_info->make_buffer)
	g_warning ("cut/copy was successful - but didn't get buffer_id");
      g_free (cut_info->description);
      g_free (cut_info);
      return;
    }
  info = g_new0 (GdamBufferManagerInfo, 1);
  info->buffer_id = buffer_id;
  info->channel = cut_info->channel;
  info->description = cut_info->description;
  info->length = -1;
  gdam_buffer_manager_add_buffer (info);
  gdam_buffer_manager_refresh (buffer_id);
  cut_info->editor_skin->last_cut_or_copy = buffer_id;
  g_free (cut_info);
}
static void successfully_copied(int num_args, GdamArg *args, gpointer user_data)
{
  int buffer_id = 0;
  GdamBufferManagerInfo *info;
  CutInfo *cut_info = user_data;
  int i;
  for (i = 0; i < num_args; i++)
    {
      if (strcmp (args[i].name, "buffer_id") == 0
	  && args[i].type == GDAM_ARG_TYPE_INT)
	{
	  buffer_id = GDAM_ARG_INT (args + i);
	  continue;
	}
      gdam_debug ("got unknown arg %s as response to buffer_cut",
		  args[i].name);
    }
  if (buffer_id == 0)
    {
      g_warning ("copy was successful - but didn't get buffer_id");
      g_free (cut_info->description);
      g_free (cut_info);
      return;
    }
  info = g_new0 (GdamBufferManagerInfo, 1);
  info->buffer_id = buffer_id;
  info->channel = cut_info->channel;
  info->description = cut_info->description;
  info->length = -1;
  gdam_buffer_manager_add_buffer (info);
  cut_info->editor_skin->last_cut_or_copy = buffer_id;
  g_free (cut_info);
}
static void failed_to_cut(gpointer user_data)
{
  CutInfo *cut_info = user_data;
  g_free (cut_info->description);
  g_free (cut_info);
}
typedef struct _CutCallbackInfo CutCallbackInfo;
struct _CutCallbackInfo {
	gboolean	      make_buffer;
	gboolean	      cut;
	gboolean	      mute;
	GdamBufferEditorSkin* editor_skin;
};

static void process_callback(CutCallbackInfo* callback_info)
{
  int start, length;
  GdamBufferEditorSkin *editor_skin = callback_info->editor_skin;
  GdamBufferViewport *viewport = editor_skin->viewport;
  int samps_per_sec;
  GdamArg *args;
  CutInfo *cut_info;
  char *description;
  GdamRequest *request;
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (viewport->buffer_id != 0);
  samps_per_sec = viewport->samples_per_second;
  if (!viewport->has_selection)
    {
      return;
    }
  start = (int) (viewport->selection_start * samps_per_sec);
  length = (int) (viewport->selection_length * samps_per_sec);
  args = gdam_arg_new_from_valist (5,
				   "start",
				   GDAM_ARG_TYPE_INT,
				   start,
				   "length",
				   GDAM_ARG_TYPE_INT,
				   length,
				   "make_buffer",
				   GDAM_ARG_TYPE_INT,
				   callback_info->make_buffer,
				   "cut", 
				   GDAM_ARG_TYPE_INT, 
				   callback_info->cut,
				   "mute", 
				   GDAM_ARG_TYPE_INT, 
				   callback_info->mute
				   );
  if (length == -1)
    description = g_strdup_printf ("%s [%4.3g-] of buffer %d",
				   callback_info->cut ? "cut" : "copy",
				   (gdouble) start / samps_per_sec,
				   viewport->buffer_id);
  else
    description = g_strdup_printf ("%s [%4.3g-%4.3g] of buffer %d",
				   callback_info->cut ? "cut" : "copy",
				   (gdouble) start / samps_per_sec,
				   (gdouble) (start + length) / samps_per_sec,
				   viewport->buffer_id);
  cut_info = g_new (CutInfo, 1);
  cut_info->description = description;
  cut_info->channel = GDAM_SKIN (viewport)->channel;
  cut_info->viewport = viewport;
  cut_info->editor_skin = editor_skin;
  cut_info->make_buffer = callback_info->make_buffer;
  request = gdam_request_call_function
    (GDAM_PROTOCOL_BUFFER_ARG,
     viewport->buffer_id,
     "process_region", 5, args, successfully_cut, failed_to_cut, cut_info);
  gdam_channel_append_request (cut_info->channel, request);
  gdam_arg_free_array (5, args);
}

static gboolean gdam_buffer_editor_cut_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  CutCallbackInfo *callback_info = g_new (CutCallbackInfo, 1);
  callback_info->make_buffer = TRUE;
  callback_info->cut = TRUE;
  callback_info->mute = FALSE;
  callback_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (process_callback),
			   NULL, callback_info, g_free, TRUE, FALSE);
  return TRUE;
}
static gboolean gdam_buffer_editor_kill_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  CutCallbackInfo *callback_info = g_new (CutCallbackInfo, 1);
  callback_info->make_buffer = FALSE;
  callback_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (process_callback),
			   NULL, callback_info, g_free, TRUE, FALSE);
  return TRUE;
}


static void mute_callback(GdamBufferEditorSkin* editor_skin)
{
  CutCallbackInfo *callback_info = g_new (CutCallbackInfo, 1);
  callback_info->make_buffer = FALSE;
  callback_info->cut = FALSE;
  callback_info->mute = TRUE;
  callback_info->editor_skin = editor_skin;
  process_callback (callback_info);
  g_free (callback_info);
}


static void copy_callback(GdamBufferEditorSkin* editor_skin)
{
  CutCallbackInfo *callback_info = g_new (CutCallbackInfo, 1);
  callback_info->make_buffer = TRUE;
  callback_info->cut = FALSE;
  callback_info->mute = FALSE;
  callback_info->editor_skin = editor_skin;
  process_callback (callback_info);
  g_free (callback_info);
}


static void
successfully_replaced(int num_arg, GdamArg* args, gpointer user_data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  if (viewport)
    {
      if (viewport->buffer_id)
	gdam_buffer_manager_refresh (viewport->buffer_id);
      gdam_buffer_viewport_buffer_changed (viewport);
    }
  gtk_object_unref (GTK_OBJECT (user_data));
}

static void
failed_to_replaced(gpointer user_data)
{
  gtk_object_unref (GTK_OBJECT (user_data));
}

/* Given a buffer id to replace with, create a request to clear old data
 * within the selection and insert the new buffer id.  If we are preserving
 * alignment, then we make certain the sizes of the to-be-replaced and the
 * replace-with chunks match.  This leaves audio after the affected area
 * in the exact same location. */
static void
do_replace_callback (GdamBufferEditorSkin* editor_skin, int buffer_id,
                     gboolean preserve_alignment)
{
  int start, length;
  int insert_length = -1;
  int insert_buffer_length = -1;
  int samps_per_sec;
  GdamBufferViewport *viewport = editor_skin->viewport;
  GdamBufferManagerInfo *info;
  GdamArg *args;
  GdamRequest *request;
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (viewport->buffer_id != 0);
  g_return_if_fail (buffer_id != 0);
  if (!viewport->has_selection)
    {
      return;
    }
  if (editor_skin->read_only)
    {
      g_message ("the editor is in read-only mode for this buffer");
      return;
    }

  info = gdam_buffer_manager_info_by_id (buffer_id);
  if (info == NULL)
    {
      g_warning ("buffer_editor_replace: manager has no info "
		 "on buffer %d", buffer_id);
      return;
    }
  insert_buffer_length = info->length;
  samps_per_sec = viewport->samples_per_second;
  start = (int) (viewport->selection_start * samps_per_sec);
  length = (int) (viewport->selection_length * samps_per_sec);
  if (preserve_alignment)
    {
      /* If we are replacing with a smaller chunk, replace the size of the
       * smaller chunk, and silence the rest */
      /* TODO: right now, we leave the extra to-replace audio untouched, really
       * we should silence it. */
      if (insert_buffer_length < length)
	{
	  int silence_from = start + insert_length;
	  int silence_len = length - insert_length;
	  length = insert_buffer_length;
	}
      /* If we are replacing with a larger chunk, crop the incoming chunk to
       * the amount of space we have available */
      else
	insert_length = length;
    }
  args = gdam_arg_new_from_valist (4,
				   "start",
				   GDAM_ARG_TYPE_INT,
				   start,
				   "length",
				   GDAM_ARG_TYPE_INT,
				   length,
				   "insert_buffer_id",
				   GDAM_ARG_TYPE_INT, 
				   buffer_id,
				   "insert_length",
				   GDAM_ARG_TYPE_INT,
				   insert_length);
  gtk_object_ref (GTK_OBJECT (viewport));
  request = gdam_request_call_function
    (GDAM_PROTOCOL_BUFFER_ARG,
     viewport->buffer_id,
     "replace_region",
     4, args, successfully_replaced, failed_to_replaced, viewport);
  gdam_channel_append_request (GDAM_SKIN (editor_skin)->channel, request);
  gdam_arg_free_array (4, args);
}

static void 
replace_callback (GdamBufferEditorSkin *editor_skin)
{
  int buffer_id = gdam_buffer_manager_obtain_id (NULL);
  gdam_buffer_manager_unref (buffer_id, NULL);
  do_replace_callback (editor_skin, buffer_id, FALSE);
}

static void 
replace_with_cut (GdamBufferEditorSkin *editor_skin)
{
  int buffer_id = editor_skin->last_cut_or_copy;
  do_replace_callback (editor_skin, buffer_id, FALSE);
}

static void 
replace_and_preserve_callback (GdamBufferEditorSkin *editor_skin)
{
  int buffer_id = gdam_buffer_manager_obtain_id (NULL);
  gdam_buffer_manager_unref (buffer_id, NULL);
  do_replace_callback (editor_skin, buffer_id, TRUE);
}

static void 
replace_and_preserve_with_cut (GdamBufferEditorSkin *editor_skin)
{
  int buffer_id = editor_skin->last_cut_or_copy;
  do_replace_callback (editor_skin, buffer_id, TRUE);
}

typedef struct _InsertInfo InsertInfo;
struct _InsertInfo {
	int			multiplicity;
	gboolean		after_selection;
	gboolean		from_buffer_manager;
	GdamBufferEditorSkin*   editor_skin;
};

static void
successfully_inserted(int num_arg, GdamArg* args, gpointer user_data)
{
  InsertInfo *insert_info = user_data;
  GdamBufferEditorSkin *editor_skin = insert_info->editor_skin;
  GdamBufferViewport *viewport = editor_skin->viewport;
  if (viewport)
    {
      if (viewport->buffer_id)
	gdam_buffer_manager_refresh (viewport->buffer_id);
      gdam_buffer_viewport_buffer_changed (viewport);
    }
  gtk_object_unref (GTK_OBJECT (viewport));
}

static void
failed_to_insert(gpointer user_data)
{
  InsertInfo *insert_info = user_data;
  GdamBufferEditorSkin *editor_skin = insert_info->editor_skin;
  GdamBufferViewport *viewport = editor_skin->viewport;
  gtk_object_unref (GTK_OBJECT (viewport));
}
/* XXX: preserve selection when inserting before / within selection ? */
static void insert_callback(InsertInfo* insert_info)
{
  int start;
  int buffer_id;
  int samps_per_sec;
  GdamBufferEditorSkin *editor_skin = insert_info->editor_skin;
  GdamBufferViewport *viewport = editor_skin->viewport;
  GdamArg *args;
  GdamRequest *request;
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (viewport->buffer_id != 0);
  if (!viewport->has_selection)
    {
      return;
    }
  if (editor_skin->read_only)
    {
      g_message ("the editor is in read-only mode for this buffer");
      return;
    }
  samps_per_sec = viewport->samples_per_second;
  if (insert_info->after_selection)
    start = (int) (samps_per_sec * (viewport->selection_start
				    + viewport->selection_length));
  else
    start = (int) (viewport->selection_start * samps_per_sec);
  if (insert_info->from_buffer_manager)
    {
      buffer_id = gdam_buffer_manager_obtain_id (NULL);
      gdam_buffer_manager_unref (buffer_id, NULL);
    }
  else
    {
      buffer_id = insert_info->editor_skin->last_cut_or_copy;
    }
  args = gdam_arg_new_from_valist (3,
				   "start",
				   GDAM_ARG_TYPE_INT,
				   start,
				   "multiplicity",
				   GDAM_ARG_TYPE_INT,
				   insert_info->multiplicity,
				   "insert_buffer_id",
				   GDAM_ARG_TYPE_INT, buffer_id);
  gtk_object_ref (GTK_OBJECT (viewport));
  request = gdam_request_call_function
    (GDAM_PROTOCOL_BUFFER_ARG,
     viewport->buffer_id,
     "insert_region",
     3, args, successfully_inserted, failed_to_insert, insert_info);
  gdam_channel_append_request (GDAM_SKIN (editor_skin)->channel, request);
  gdam_arg_free_array (3, args);
}
static gboolean gdam_buffer_editor_insert_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  InsertInfo *insert_info = g_new (InsertInfo, 1);
  int i;
  insert_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  insert_info->multiplicity = 1;
  insert_info->after_selection = FALSE;
  insert_info->from_buffer_manager = FALSE;

  for (i = 1; i < num_handler_names; i++)
    {
      char *name = handler_names[i];
      if (strncmp (name, "times=", 6) == 0)
	{
	  int times = atoi (name + 6);
	  if (times > 0)
	    insert_info->multiplicity = times;
	  continue;
	}
      if (strcmp (name, "after") == 0)
	{
	  insert_info->after_selection = TRUE;
	  continue;
	}
      if (strcmp (name, "buffer_manager") == 0)
	{
	  insert_info->from_buffer_manager = TRUE;
	  continue;
	}
      g_message ("unknown handler name for buffer_insert: %s", name);
    }

  gtk_signal_connect_full (object,
			   signal_name,
			   GTK_SIGNAL_FUNC (insert_callback),
			   NULL, insert_info, g_free, TRUE, FALSE);
  return TRUE;
}

/*
 *                                    _
 *  _ __ _____   _____ _ __ ___  __ _| |
 * | '__/ _ \ \ / / _ \ '__/ __|/ _` | |
 * | | |  __/\ V /  __/ |  \__ \ (_| | |
 * |_|  \___| \_/ \___|_|  |___/\__,_|_|
 * 
 */
static void
successfully_reversed(int num_arg, GdamArg* args, gpointer user_data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  gdam_buffer_viewport_buffer_changed (viewport);
  gtk_object_unref (GTK_OBJECT (viewport));
}

static void
failed_to_reverse(gpointer user_data)
{
  GdamBufferViewport *viewport = GDAM_BUFFER_VIEWPORT (user_data);
  gtk_object_unref (GTK_OBJECT (viewport));
}

static void reverse_callback(GdamBufferEditorSkin* editor_skin)
{
  GdamModel *model = GDAM_SKIN (editor_skin)->model;
  guint buffer_id;
  GdamArg *args;
  GdamRequest *request;
  GdamBufferViewport *viewport;
  int start, length;
  int samps_per_sec;

  g_return_if_fail (model != NULL);
  viewport = editor_skin->viewport;
  g_return_if_fail (viewport != NULL);
  buffer_id = editor_skin->viewport->buffer_id;
  if (buffer_id == 0)
    return;
  if (editor_skin->read_only)
    {
      g_message ("the editor is in read-only mode for this buffer");
      return;
    }

  samps_per_sec = viewport->samples_per_second;
  start = (int) (samps_per_sec * viewport->selection_start);
  length = (int) (samps_per_sec * viewport->selection_length);
  args = gdam_arg_new_from_valist (2,
				   "start",
				   GDAM_ARG_TYPE_INT,
				   start,
				   "length", GDAM_ARG_TYPE_INT, length);
  gtk_object_ref (GTK_OBJECT (viewport));
  request = gdam_request_call_function
    (GDAM_PROTOCOL_BUFFER_ARG,
     buffer_id,
     "reverse_region",
     2,
     args, successfully_reversed, failed_to_reverse, editor_skin->viewport);
  gdam_channel_append_request (GDAM_SKIN (editor_skin)->channel, request);
  gdam_arg_free_array (2, args);
}

static void 
selection_to_fit_buffer (GdamBufferEditorSkin *editor_skin, int id)
{
  gdouble length;
  GdamBufferManagerInfo *info;
  GdamBufferViewport *viewport = editor_skin->viewport;
  g_return_if_fail (viewport != NULL);

  if (!viewport->has_selection)
    return;
  if (id == 0)
    return;

  info = gdam_buffer_manager_info_by_id (id);
  if (info == NULL)
    {
      g_warning ("buffer_viewport: manager has no info "
		 "on buffer %d", id);
      return;
    }
  length = (gdouble) info->length / viewport->samples_per_second;

  if (length < 0.0001)
    return;
  gdam_buffer_viewport_set_selection (viewport, viewport->selection_start, length);
}

static void
selection_to_fit_buffer_manager (GdamBufferEditorSkin *editor_skin)
{
  int id = gdam_buffer_get_default ();
  selection_to_fit_buffer (editor_skin, id);
}

static void
selection_to_fit_cut (GdamBufferEditorSkin *editor_skin)
{
  selection_to_fit_buffer (editor_skin, editor_skin->last_cut_or_copy);
}

static void update_start_time_label(GdamBufferViewport* view,
                                    GtkLabel*           to_change)
{
  char buf[1024];
  snprintf (buf, sizeof (buf), "%4.4gs", view->time_start);
  gtk_label_set_text (to_change, buf);
}

static gboolean gdam_buffer_editor_start_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamBufferEditorSkin *editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  if (!GTK_IS_LABEL (object))
    {
      g_message ("start_time_label can only be attached to a label");
      return TRUE;
    }
  if (editor_skin->viewport == NULL)
    return FALSE;
  gtk_signal_connect_while_alive
    (GTK_OBJECT (editor_skin->viewport),
     "view_changed",
     GTK_SIGNAL_FUNC (update_start_time_label), object, object);
  return TRUE;
}


static void update_end_time_label(GdamBufferViewport* view,
                                    GtkLabel*           to_change)
{
  char buf[1024];
  gdouble end = view->time_start + view->time_length;
  snprintf (buf, sizeof (buf), "%4.4gs", end);
  gtk_label_set_text (to_change, buf);
}

static gboolean gdam_buffer_editor_end_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamBufferEditorSkin *editor_skin = GDAM_BUFFER_EDITOR_SKIN (skin);
  if (!GTK_IS_LABEL (object))
    {
      g_message ("end_time_label can only be attached to a label");
      return TRUE;
    }
  if (editor_skin->viewport == NULL)
    return FALSE;
  gtk_signal_connect_while_alive
    (GTK_OBJECT (editor_skin->viewport),
     "view_changed", GTK_SIGNAL_FUNC (update_end_time_label), object, object);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_buffer_editor_viewport",
           gdam_buffer_editor_viewport_handler			},
	{ "gdam_buffer_editor_defer_to_viewport",
           gdam_buffer_editor_skin_defer_to_viewport_handler	},
        { "gdam_buffer_editor_cut",
           gdam_buffer_editor_cut_handler			},
        { "gdam_buffer_editor_kill",
           gdam_buffer_editor_kill_handler			},
        { "gdam_buffer_editor_insert",
           gdam_buffer_editor_insert_handler			},
	{ "gdam_buffer_editor_start_time_label",
	   gdam_buffer_editor_start_time_label_handler		},
	{ "gdam_buffer_editor_end_time_label",
	   gdam_buffer_editor_end_time_label_handler		},
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_buffer_editor_set_default",
           (GdamSkinFunc)do_set_defaults			},
	{ "gdam_buffer_editor_load_default_buffer",
           (GdamSkinFunc)do_load_default_buffer			},
        { "gdam_buffer_editor_play",
           (GdamSkinFunc)play_callback                     	},
        { "gdam_buffer_editor_copy",
           (GdamSkinFunc)copy_callback				},
        { "gdam_buffer_editor_mute",
           (GdamSkinFunc)mute_callback				},
        { "gdam_buffer_editor_replace",
           (GdamSkinFunc)replace_callback			},
        { "gdam_buffer_editor_replace_with_cut",
           (GdamSkinFunc)replace_with_cut                       },
        { "gdam_buffer_editor_replace_and_preserve",
           (GdamSkinFunc)replace_and_preserve_callback		},
        { "gdam_buffer_editor_replace_and_preserve_with_cut",
           (GdamSkinFunc)replace_and_preserve_with_cut          },
        { "gdam_buffer_editor_reverse",
           (GdamSkinFunc)reverse_callback                   	},
        { "gdam_buffer_editor_extend_selection_to_fit_buffer_manager",
           (GdamSkinFunc)selection_to_fit_buffer_manager        },
        { "gdam_buffer_editor_extend_selection_to_fit_cut",
           (GdamSkinFunc)selection_to_fit_cut                   },
};

static void gdam_buffer_editor_skin_init(GdamBufferEditorSkin* editor)
{
}
static void gdam_buffer_editor_skin_class_init(GdamSkinClass* skin_class)
{
  skin_class->build_new_model = gdam_buffer_editor_build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
}


static gboolean       gdam_buffer_editor_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options)
{
  GdamModel *model;
  GdamBufferEditorSkin *editor = GDAM_BUFFER_EDITOR_SKIN (skin);
  GdamBufferViewport *viewport;
  model = GDAM_MODEL (gtk_type_new (GDAM_TYPE_PRODUCER_CACHE));
  model->channel = channel;
  skin->model = model;
  viewport = gdam_buffer_viewport_new_shown (channel);
  gtk_object_ref (GTK_OBJECT (viewport));
  gtk_object_sink (GTK_OBJECT (viewport));
  editor->viewport = viewport;
  gtk_signal_connect_object (GTK_OBJECT (editor->viewport),
			     "buffer_changed", 
			     GTK_SIGNAL_FUNC (on_buffer_changed), 
			     GTK_OBJECT (editor));
  gdam_buffer_viewport_add_markers (viewport,
				    TRUE,		/* is_shown */
				    FALSE,		/* is_under */
				    NULL,		/* font */
				    NULL,		/* gc */
				    NULL,		/* color */
				    NULL,		/* format */
				    1.0);
  return TRUE;
}

GtkType gdam_buffer_editor_skin_get_type()
{
  static GtkType buffer_editor_skin_type = 0;
  if (!buffer_editor_skin_type)
    {
      static const GtkTypeInfo buffer_editor_skin_info = {
	"GdamBufferEditorSkin",
	sizeof (GdamBufferEditorSkin),
	sizeof (GdamBufferEditorSkinClass),
	(GtkClassInitFunc) gdam_buffer_editor_skin_class_init,
	(GtkObjectInitFunc) gdam_buffer_editor_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      buffer_editor_skin_type = gtk_type_unique (parent,
						 &buffer_editor_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return buffer_editor_skin_type;
}
