/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamfaderskin.h"
#include <gdam/component/gdamutil.h>
#include <glade/glade.h>
#include "../gdamskinutil.h"
#include <gtk/gtk.h>
#include <gdam/common/gdambuffer.h>
#include <gdam/client/gdamchannel.h>
#include <gdam/common/gdamsample.h>
#include <gdam/component/gdambuffermanager.h>
#include <stdlib.h>
#include <gdam/common/gdamplugin.h>

#define MAX_GAIN                2.0
#define MAX_RATE                2.0

static void gdam_fader_skin_class_init(GdamFaderSkinClass* klass);

static gboolean adjust_dummy_points  (GdamFaderSkin*          skin,
		                      int                     num_fades,
		                      GdamFaderElement*       fade_array);

/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 


static void gdam_fader_skin_init(GdamFaderSkin* skin)
{
  skin->circle_radius = 2;
  skin->which_dragging = -1;
}

GtkType gdam_fader_skin_get_type()
{
  static GtkType fader_skin_type = 0;
  if (!fader_skin_type)
    {
      static const GtkTypeInfo fader_skin_info = {
	"GdamFaderSkin",
	sizeof (GdamFaderSkin),
	sizeof (GdamFaderSkinClass),
	(GtkClassInitFunc) gdam_fader_skin_class_init,
	(GtkObjectInitFunc) gdam_fader_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      fader_skin_type = gtk_type_unique (GDAM_TYPE_SKIN, &fader_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return fader_skin_type;
}

/*
 *  ____                     _                  _
 * |  _ \ _ __ __ ___      _(_)_ __   __ _     / \   _ __ ___  __ _
 * | | | | '__/ _` \ \ /\ / / | '_ \ / _` |   / _ \ | '__/ _ \/ _` |
 * | |_| | | | (_| |\ V  V /| | | | | (_| |  / ___ \| | |  __/ (_| |
 * |____/|_|  \__,_| \_/\_/ |_|_| |_|\__, | /_/   \_\_|  \___|\__,_|
 *                                   |___/
 */

static void alloc_colors(GdamFaderSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  GdkGC *gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (gc, util_alloc_color_cached (65535, 0, 0));
  skin->circle_gc = gc;
  skin->profile_gc = skin->drawing_area->style->bg_gc[state];
  skin->background = skin->drawing_area->style->fg_gc[state];
}

static int get_y_element_position(GdamFaderSkin*    skin,
                                  GdamFaderElement* element)
{
  int y = element->volume * skin->draw_height / GdamFixed_ONE;
  return skin->draw_height - 1 - y;
}
#if 0
static int get_y_position(GdamFaderSkin*    skin,
                          int               element_index)
{
  GdamFaderElement *element;
  g_return_val_if_fail (0 <= element_index, -1);
  g_return_val_if_fail (element_index < get_fader (skin)->num_fades, -1);
  element = &get_fader (skin)->fade_array[element_index];
  return get_y_element_position (skin, element);
}
#endif

static inline GdamFader* get_fader(GdamFaderSkin* skin)
{
  return GDAM_FADER (GDAM_SKIN (skin)->model);
}

static void get_fade_element_position(GdamFaderSkin*    skin,
			              GdamFaderElement* element,
                                      int*              out_x,
			              int*              out_y)
{
  GdamFader *fader = get_fader (skin);
  int fader_len;
  fader_len = fader->fade_array[fader->num_fades - 1].time + 1;
  *out_x = skin->draw_width * element->time / fader_len;
  *out_y = get_y_element_position (skin, element);
}

static void get_fade_position(GdamFaderSkin*    skin,
			      int               which_fader,
                              int*              out_x,
			      int*              out_y)
{
  GdamFader *fader = get_fader (skin);
  GdamFaderElement *element = &fader->fade_array[which_fader];
  if (fader->num_fades < 2)
    {
      *out_x = *out_y = 0;
      g_return_if_fail (FALSE);
    }
  get_fade_element_position (skin, element, out_x, out_y);
}

static void get_fade_info_from_position(GdamFaderSkin*          skin,
					int                     x_pos,
					int                     y_pos,
					int*			time_out,
					GdamFixed*		gain_out)
{
  GdamFader *fader = get_fader (skin);
  int fader_len;
  if (fader->num_fades < 2)
    {
      *time_out = *gain_out = 0;
      g_return_if_fail (FALSE);
    }
  fader_len = fader->fade_array[fader->num_fades - 1].time + 1;
  *time_out = fader_len * x_pos / skin->draw_width;
  y_pos = skin->draw_height - 1 - y_pos;
  *gain_out = GdamFixed_ONE * y_pos / skin->draw_height;
}

static void copy_backing_area(GdamFaderSkin* skin, const GdkRectangle* area)
{
  gdk_draw_pixmap (skin->drawing_area->window,
		   skin->drawing_area->
		   style->fg_gc[GTK_WIDGET_STATE (skin->drawing_area)],
		   skin->backing_pixmap, area->x, area->y, area->x, area->y,
		   area->width, area->height);
}

/* Adds the two points which define the bottom edge of the polygon. */
static gboolean get_poly_points(GdamFaderSkin* skin, 
                                GdkPoint*      points)
{
  GdamFader *fader = get_fader (skin);
  int num_fades = fader->num_fades;
  int i;
  g_return_val_if_fail (num_fades >= 2, FALSE);
  /* TODO: make these sit at correct_height, invisible? */
  /* these first points are virtuals at the edges. */
  /* If there are no user points, set both to full height. */
  points[0].x = 0;
  points[num_fades + 1].x = skin->draw_width - 1;
  points[0].y = skin->draw_height - 1;
  points[num_fades + 1].y = skin->draw_height - 1;

  for (i = 0; i < num_fades; i++)
    {
      int x, y;
      get_fade_position (skin, i, &x, &y);
      points[i + 1].x = x;
      points[i + 1].y = y;
      g_return_val_if_fail (points[i].x <= points[i + 1].x, FALSE);
    }
  return TRUE;
}

static void redraw_area(GdamFaderSkin* skin, const GdkRectangle* area)
{
  GdamFader *fader = get_fader (skin);
  int num_points, i;
  int circle_radius = skin->circle_radius;
  GdkRectangle rect;
  GdkPoint *points;
  if (area == NULL)
    {
      rect.width = skin->drawing_area->allocation.width;
      rect.height = skin->drawing_area->allocation.height;
      rect.x = rect.y = 0;
      area = &rect;
    }
  //gdam_debug("fader: redraw_area(%d,%d x %d,%d)",
  //area->x, area->y, area->width, area->height);
  if (!skin->circle_gc)
    alloc_colors (skin);
  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->background,
		      TRUE, area->x, area->y, area->width, area->height);

  /* Draw the fader profile (as a polygon). */
  num_points = 2 + fader->num_fades;
  points = g_new (GdkPoint, num_points);
  if (!get_poly_points (skin, points))
    {
      g_warning ("redraw_area (fader): couldn't get polygon points");
      return;
    }
  gdk_draw_polygon (skin->backing_pixmap,
		    skin->profile_gc, TRUE, points, num_points);

  /* Draw the fader click areas. */
  for (i = 1; i < num_points - 1; i++)
    {
      gdk_draw_arc (skin->backing_pixmap,
		    skin->circle_gc,
		    TRUE,
		    points[i].x - circle_radius,
		    points[i].y - circle_radius,
		    circle_radius * 2, circle_radius * 2, 0, 360 * 64);
    }

  copy_backing_area (skin, area);
}

static gint
fader_expose(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             GdamFaderSkin*  skin)
{
  redraw_area (skin, &expose_event->area);
  return 0;
}

/*
 * Returns the index of the fade whose circle is at the given x,y location,
 * or -1 if no fade was found.
 */
static int find_keypoint(GdamFaderSkin* skin, int x, int y)
{
  int radius_squared = (skin->circle_radius + 1) * (skin->circle_radius + 1);
  int i;
  GdamFader *fader = get_fader (skin);
  for (i = 0; i < fader->num_fades; i++)
    {
      int pos_x, pos_y;
      get_fade_position (skin, i, &pos_x, &pos_y);
      pos_x -= x;
      pos_y -= y;
      if (pos_x * pos_x + pos_y * pos_y <= radius_squared)
	return i;
    }
  return -1;
}

static void get_fade_array(GdamFaderSkin*     skin,
                           int*               rv_num_fades,
			   GdamFaderElement** rv_fade_array)
{
  if (skin->preview_fade_array)
    {
      *rv_fade_array = skin->preview_fade_array;
      *rv_num_fades = skin->preview_num_fades;
    }
  else
    {
      *rv_fade_array = get_fader (skin)->fade_array;
      *rv_num_fades = get_fader (skin)->num_fades;
    }
}

/* Find a box which contains the fader's specified keypoints. */
static void compute_container(GdamFaderSkin* skin,
			      int            num,
                              GdamFaderElement* elements,
			      GdkRectangle*  rectangle)
{
  int x, y, i;
  int left, top, right, bottom;
  if (num <= 1)
    return;

  get_fade_element_position (skin, elements, &x, &y);
  left = right = x;
  top = bottom = y;
  for (i = 1; i < num; i++)
    {
      get_fade_element_position (skin, elements + i, &x, &y);
      if (x < left)
	left = x;
      if (x > right)
	right = x;
      if (y < top)
	top = y;
      if (y > bottom)
	bottom = y;
    }
  rectangle->x = left;
  rectangle->y = top;
  rectangle->width = right - left + 1;
  rectangle->height = bottom - top + 1;
}

static void delete_keypoint(GdamFaderSkin*    skin,
                            int               which_fade)
{
  GdkRectangle area;
  GdamFaderElement *fade_array;
  int num_fades;
  get_fade_array (skin, &num_fades, &fade_array);
  if (which_fade <= 0 || which_fade >= num_fades)
    {
      g_warning ("cannot delete the first or last fader keypoint");
      return;
    }
  compute_container (skin, 3, fade_array + which_fade - 1, &area);
  gdam_debug ("deleting fade %d of %d", which_fade, num_fades);
  gdam_fader_remove_fade (get_fader (skin), which_fade);
  redraw_area (skin, &area);
}

/* Returns the index of the left-side of the interval
 * that contains this time. */
static int search_for_interval(int                     time,
                               int                     num,
			       const GdamFaderElement* elements)
{
  int i;
  for (i = 0; i < num - 1; i++)
    if (elements[i].time <= time && time < elements[i + 1].time)
      return i;
  return -1;
}

/* 
 * Update either our preview array, or the actual server array.
 * No graphical updates here...
 */
static void update_fade_array(GdamFaderSkin*          skin,
                              int                     num_fades,
			      GdamFaderElement*       new_array,
			      int                     set_server_side)
{
  gdam_debug ("update_fader_array: num_fades=%d, set_server_side=%d",
	      num_fades, set_server_side);
  if (!adjust_dummy_points (skin, num_fades, new_array))
    return;
  if (set_server_side)
    {
      gdam_fader_set_fades (get_fader (skin), num_fades, new_array);
    }
  else
    {
      GdamFaderElement *copy;
      copy = g_new (GdamFaderElement, num_fades);
      memcpy (copy, new_array, num_fades * sizeof (GdamFaderElement));
      g_free (skin->preview_fade_array);
      skin->preview_num_fades = num_fades;
      skin->preview_fade_array = copy;
    }
}

/*
 * Set the GdamFader's array and redraw the relevant area,
 * and set the server-side array conditionally.
 */
static void
change_fade_keypoint(GdamFaderSkin*        skin,
                     int                   which_fade,
	             int                   new_time,
	             GdamFixed	           new_volume,
		     gboolean              set_server_side)
{
  GdkRectangle area;
  int old_x;
  GdamFaderElement *new_array;
  int array_size;
  int num_fades;
  GdamFaderElement *fade_array;

  get_fade_array (skin, &num_fades, &fade_array);
  old_x = fade_array[which_fade].time;
  array_size = sizeof (GdamFaderElement) * num_fades;
  new_array = (GdamFaderElement *) alloca (array_size);
  memcpy (new_array, fade_array, array_size);

  if (which_fade == 0)
    {
      GdkRectangle area_before, area_after;
      compute_container (skin, 2, new_array, &area_before);
      new_array[0].volume = new_volume;
      compute_container (skin, 2, new_array, &area_after);
      gdk_rectangle_union (&area_after, &area_before, &area);
    }
  else if (which_fade == num_fades - 1)
    {
      GdkRectangle area_before, area_after;
      GdamFaderElement *last = new_array + num_fades - 2;
      compute_container (skin, 2, last, &area_before);
      new_array[which_fade].volume = new_volume;
      compute_container (skin, 2, last, &area_after);
      gdk_rectangle_union (&area_after, &area_before, &area);
    }
  else
    {
      int old_index, new_index;
      int start_keypoint, num_keypoints;
      new_array[which_fade].volume = new_volume;
      new_array[which_fade].time = new_time;
      gdam_fader_sort_fade_array (num_fades, new_array);

      /* Element before the old location of the keypoint. */
      old_index = search_for_interval (old_x, num_fades, new_array);
      new_index = search_for_interval (new_time, num_fades, new_array);

      if (old_index <= new_index)
	{
	  start_keypoint = old_index - 1;
	  num_keypoints = new_index - old_index + 1;
	}
      else
	{
	  start_keypoint = new_index - 1;
	  num_keypoints = old_index - new_index + 1;
	}
      compute_container (skin,
			 num_keypoints, new_array + start_keypoint, &area);
    }
  update_fade_array (skin, num_fades, new_array, set_server_side);
  redraw_area (skin, &area);
}

static void queue_redraw(GdamFaderSkin* skin)
{
  GtkWidget *widget = skin->drawing_area;
  if (widget == NULL)
    return;
  gtk_widget_queue_draw (widget);
}

static void do_phase_synch(GdamFaderSkin* skin)
{
  GdamFader *fader = get_fader (skin);
  gdam_fader_phase_synch (fader);
}

static void
add_fader_point(GdamFaderSkin*        skin,
		int                   new_time,
		GdamFixed	      new_volume,
		gboolean              set_server_side)
{
  GdamFaderElement *new_array;
  int array_size;
  int new_index;
  GdamFaderElement *fade_array;
  int num_fades;
  GdkRectangle area;

  get_fade_array (skin, &num_fades, &fade_array);
  array_size = sizeof (GdamFaderElement) * (num_fades + 1);
  new_array = (GdamFaderElement *) alloca (array_size);
  memcpy (new_array, fade_array, array_size - sizeof (GdamFaderElement));
  new_array[num_fades].volume = new_volume;
  new_array[num_fades].time = new_time;
  gdam_fader_sort_fade_array (num_fades + 1, new_array);
  new_index = search_for_interval (new_time, num_fades + 1, new_array);
  /* BUG: `3' is only the correct number if only one
   *      keypoint exists and time == new_time.
   *
   *      Generally, it should be (2 + multiplicity at that point).
   */
  compute_container (skin, 3, new_array + new_index - 1, &area);
  update_fade_array (skin, num_fades + 1, new_array, set_server_side);
  queue_redraw (skin);
}


/*  This adjusts the volume level of the dummy fade points
 *  so that they match up at the ends of the fade period,
 *  and don't disrupt the fade pattern. */
static gboolean 
adjust_dummy_points  (GdamFaderSkin*          skin,
		      int                     num_fades,
		      GdamFaderElement*       fade_array)
{
  GdamFader *fader = get_fader (skin);

  g_return_val_if_fail (num_fades >= 2, FALSE);
  if (num_fades == 2)
    {
      /* If there are no user points, set both to full height. */
      fade_array[0].volume = GdamFixed_ONE;
      fade_array[1].volume = GdamFixed_ONE;
    }
  else if (num_fades == 3)
    {
      /* If there is one user point, set both to that height. */
      fade_array[0].volume = fade_array[1].volume;
      fade_array[2].volume = fade_array[1].volume;
    }
  else
    {
      int tl, vl, tr, vr;
      int del_t, del_v, mid_t, v_int;
      int fader_len;
      fader_len = fader->fade_array[fader->num_fades - 1].time + 1;
      tl = fade_array[num_fades - 2].time;
      vl = fade_array[num_fades - 2].volume;
      tr = fade_array[1].time + fader_len;
      vr = fade_array[1].volume;
      del_t = tr - tl;
      del_v = vr - vl;
      mid_t = fader_len - tl;
      v_int = del_t ? ((gint64) mid_t * del_v) / del_t + vl : vl + vr / 2;
      fade_array[0].volume = v_int;
      fade_array[num_fades - 1].volume = v_int;
    }
  return TRUE;
}


static gint
fader_button_press(GtkDrawingArea* area,
                   GdkEventButton* button_event,
		   GdamFaderSkin* skin)
{
  int press_x = button_event->x;
  int press_y = button_event->y;
  int width, height;
  int fade;
  int time;
  GdamFixed gain;
  g_return_val_if_fail (skin->which_dragging < 0, 0);
  if ((fade = find_keypoint (skin, press_x, press_y)) >= 0)
    {
      int pos_x, pos_y;
      if (button_event->button == 3)
	{
	  delete_keypoint (skin, fade);
	  return 0;
	}
      get_fade_position (skin, fade, &pos_x, &pos_y);
      skin->drag_click_offset_x = press_x - pos_x;
      skin->drag_click_offset_y = press_y - pos_y;
      skin->which_dragging = fade;
      return 0;
    }

  if (button_event->button == 3)
    return 0;

  width = skin->draw_width;
  height = skin->draw_height;
  get_fade_info_from_position (skin, press_x, press_y, &time, &gain);
  gdam_debug ("add_fader_point: time=%d, gain=%d", time, gain);
  add_fader_point (skin, time, gain, TRUE);
  return 0;
}

static gint
fader_button_release(GtkDrawingArea* area,
                     GdkEventButton* button_event,
		             GdamFaderSkin* skin)
{
  int release_x;
  int release_y;
  int time;
  GdamFixed volume;
  int dragging = skin->which_dragging;
  if (dragging < 0)
    return 0;
  skin->which_dragging = -1;
  release_x = button_event->x - skin->drag_click_offset_x;
  release_y = button_event->y - skin->drag_click_offset_y;
  get_fade_info_from_position (skin, release_x, release_y, &time, &volume);
  change_fade_keypoint (skin, dragging, time, volume, TRUE);
  queue_redraw (skin);
  return 0;
}

#if 0
static gint
fader_button_motion(GtkDrawingArea* area,
                     GdkEventMotion* motion_event,
		     GdamFaderSkin* skin)
{
  int x, y;
  int release_x, release_y;
  int time;
  guint state;
  GdamFixed volume;
  int dragging = skin->which_dragging;
  if (dragging < 0)
    return 0;

  if (motion_event->is_hint)
    gdk_window_get_pointer (motion_event->window, &x, &y, &state);
  else
    {
      x = motion_event->x;
      y = motion_event->y;
      state = motion_event->state;
    }

  x = motion_event->x;
  y = motion_event->y;

  release_x = x - skin->drag_click_offset_x;
  release_y = y - skin->drag_click_offset_y;
  get_fade_info_from_position (skin, release_x, release_y, &time, &volume);
  change_fade_keypoint (skin, dragging, time, volume, FALSE);
  return 0;
}
#endif

static gint
fader_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamFaderSkin* skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  return TRUE;
}

static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamFaderSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 800, 300);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (fader_button_press), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_release_event",
		      GTK_SIGNAL_FUNC (fader_button_release), skin);
#if 0
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (fader_button_motion), skin);
#endif
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event", GTK_SIGNAL_FUNC (fader_expose), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (fader_configure), skin);
  gtk_widget_show (GTK_WIDGET (drawing_area));
  gtk_widget_add_events (GTK_WIDGET (drawing_area),
			 GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect_object (GTK_OBJECT (GDAM_SKIN (skin)->model),
			     "fades_changed",
			     queue_redraw, (GtkObject *) skin);
}


/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */


static gboolean gdam_fader_phase_synch_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name, do_phase_synch, (GtkObject *) skin);
  return TRUE;
}

static gboolean gdam_fader_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
      return TRUE;
    }
  connect_drawing_area (GTK_DRAWING_AREA (object), GDAM_FADER_SKIN (skin));
  return TRUE;
}


static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_fader_drawing_area",
           gdam_fader_drawing_area_handler			},
	{ "gdam_fader_phase_synch",
           gdam_fader_phase_synch_handler			},
};

static void gdam_fader_skin_class_init(GdamFaderSkinClass* klass)
{
  GdamSkinClass *skin_class = GDAM_SKIN_CLASS (klass);
  skin_class->model_type = GDAM_TYPE_FADER;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}
