/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamfeedbackfilterskin.h"
#include "../../component/plugins/gdamfeedbackfilter.h"
#include "../gdamskinutil.h"
#include <gtk/gtksignal.h>
#include <stdlib.h>
#include <gdam/common/gdamplugin.h>

static gboolean gdam_feedback_filter_skin_build_new_model
                                                 (GdamSkin*    skin,
                                                  GdamChannel* channel,
						  int 	       num_options,
					          char**       options);

static void gdam_feedback_filter_skin_init(GdamFeedbackFilterSkin* feedback_filter_skin)
{
}

static void crossfade_adj_changed(GtkAdjustment *adj, 
      			          GdamFeedbackFilterSkin *skin)
{
  GdamFeedbackFilter *filter;
  GdamArg arg;
  float inperc, backperc, top;
  filter = GDAM_FEEDBACK_FILTER (GDAM_SKIN (skin)->model);
  top = 1.0 - skin->shelf_width;
  backperc = adj->value / top;
  if (backperc > 1.0)
    backperc = 1.0;
  if (adj->value < skin->shelf_width)
    inperc = 1.0;
  else
    inperc = 1.0 - ((adj->value - skin->shelf_width) / top);
  arg.name = "feedback_gain";
  arg.type = GDAM_ARG_TYPE_DOUBLE;
  GDAM_ARG_DOUBLE (&arg) = backperc;
  gdam_arg_set_object_arg (GTK_OBJECT (filter), &arg);
  arg.name = "input_gain";
  GDAM_ARG_DOUBLE (&arg) = inperc;
  gdam_arg_set_object_arg (GTK_OBJECT (filter), &arg);
}

static gboolean gdam_feedback_skin_crossfade_adj_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamFeedbackFilterSkin *skin;
  skin = GDAM_FEEDBACK_FILTER_SKIN (_skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  skin->crossfade_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (crossfade_adj_changed), 
		      skin);
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_feedback_skin_crossfade_adj",
          gdam_feedback_skin_crossfade_adj_handler	}
};

static void gdam_feedback_arg_set_shelf_width(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilterSkin *feedback = GDAM_FEEDBACK_FILTER_SKIN (object);
  feedback->shelf_width = GDAM_ARG_DOUBLE (arg);
}
static void gdam_feedback_arg_get_shelf_width(GtkObject* object, GdamArg* arg)
{
  GdamFeedbackFilterSkin *feedback = GDAM_FEEDBACK_FILTER_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = feedback->shelf_width;
}

static void gdam_feedback_filter_skin_class_init(GdamSkinClass* skin_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (skin_class);
  skin_class->build_new_model = gdam_feedback_filter_skin_build_new_model;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_arg_register (object_class,
		     "shelf_width",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_feedback_arg_set_shelf_width,
		     gdam_feedback_arg_get_shelf_width,
		     NULL, "0.0,0.5", GDAM_ARG_VITAL);
}

static void source_was_destroyed(GtkObject* destroyed,
                                 GdamFeedbackFilterSkin* dest)
{
  dest->source = NULL;
  dest->source_destroy_id = 0;
}

static void dest_was_destroyed(GtkObject* destroyed,
                               GdamFeedbackFilterSkin* source)
{
  source->dest = NULL;
  source->dest_destroy_id = 0;
}

void gdam_feedback_filter_skin_connect(GdamFeedbackFilterSkin* dest,
                                       GdamFeedbackFilterSkin* source)
{
  GdamSkin *dest_skin;
  GdamSkin *source_skin;
  GdamFeedbackFilter *dest_model;
  GdamFeedbackFilter *source_model;
  g_return_if_fail (dest != NULL);
  g_return_if_fail (source != NULL);

  if (dest->source != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (dest->source),
			     dest->source_destroy_id);
      dest->source = NULL;
    }

  if (source->dest != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (source->dest),
			     source->dest_destroy_id);
      source->dest = NULL;
    }

  dest->source = source;
  dest->source_destroy_id
    = gtk_signal_connect (GTK_OBJECT (source),
			  "destroy",
			  GTK_SIGNAL_FUNC (source_was_destroyed), dest);

  source->dest = dest;
  source->dest_destroy_id
    = gtk_signal_connect (GTK_OBJECT (dest),
			  "destroy",
			  GTK_SIGNAL_FUNC (dest_was_destroyed), source);
  dest_skin = GDAM_SKIN (dest);
  if (dest_skin->model)
    dest_model = GDAM_FEEDBACK_FILTER (dest_skin->model);
  else
    dest_model = NULL;
  source_skin = GDAM_SKIN (source);
  if (source_skin->model)
    source_model = GDAM_FEEDBACK_FILTER (source_skin->model);
  else
    source_model = NULL;
  if (source_model)
    gdam_feedback_filter_set_dest (source_model, dest_model);
  else if (dest_model)
    gdam_feedback_filter_set_source (dest_model, source_model);
}

GtkType gdam_feedback_filter_skin_get_type()
{
  static GtkType feedback_filter_skin_type = 0;
  if (!feedback_filter_skin_type)
    {
      static const GtkTypeInfo feedback_filter_skin_info = {
	"GdamFeedbackFilterSkin",
	sizeof (GdamFeedbackFilterSkin),
	sizeof (GdamFeedbackFilterSkinClass),
	(GtkClassInitFunc) gdam_feedback_filter_skin_class_init,
	(GtkObjectInitFunc) gdam_feedback_filter_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      feedback_filter_skin_type = gtk_type_unique (parent,
						   &feedback_filter_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return feedback_filter_skin_type;
}

static GdamFeedbackFilterSkin* last_dest = NULL;

static void on_last_dest_destroy() { last_dest = NULL; }

static gboolean gdam_feedback_filter_skin_build_new_model
                                                 (GdamSkin*    skin,
                                                  GdamChannel* channel,
						  int 	       num_options,
					          char**       options)
{
  GdamFeedbackFilterSkin *filter_skin;
  GdamFeedbackFilter *filter;
  GdamModel *model;

  filter_skin = GDAM_FEEDBACK_FILTER_SKIN (skin);
  filter = GDAM_FEEDBACK_FILTER (gtk_type_new (GDAM_TYPE_FEEDBACK_FILTER));
  model = GDAM_MODEL (filter);

  skin->model = model;
  model->channel = channel;

  if (last_dest)
    {
      gdam_feedback_filter_skin_connect (last_dest, filter_skin);
      last_dest = NULL;
    }
  else
    {
      last_dest = filter_skin;
      gtk_signal_connect (GTK_OBJECT (last_dest),
			  "destroy",
			  GTK_SIGNAL_FUNC (on_last_dest_destroy), NULL);
    }
  return TRUE;
}

