/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamladspaskin.h"
#include "../gdamskinutil.h"
#include "../gdambuilderskin.h"
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtklabel.h>
#include <gtk/gtksignal.h>
#include <gdam/component/plugins/gdamladspafilter.h>
#include <gdam/common/gdamsearchpath.h>
#include <math.h>
#include <stdlib.h>

static void gdam_ladspa_skin_init(GdamLadspaSkin* ladspa_skin)
{
}

static gboolean gdam_ladspa_build_new_model(GdamSkin*      skin,
					    GdamChannel*   channel,
					    int            num_options,
	                                    char**         options)
{
  if (num_options == 0)
    {
      g_warning ("LADSPA plugin really *needs* argument " "for plugintype");
      return FALSE;
    }
  skin->model = GDAM_MODEL (gdam_ladspa_filter_new (options[0]));
  if (skin->model == NULL)
    {
      g_warning ("couldn't load LADSPA plugin");
      return FALSE;
    }
  skin->model->channel = channel;
  return TRUE;
}

typedef struct _DoAdjustmentInfo DoAdjustmentInfo;
struct _DoAdjustmentInfo {
	GdamLadspaFilter*	filter;
	GtkAdjustment*          adjustment;
	int			param_index;
	double			value;

	unsigned		is_integer : 1,
				is_logarithmic : 1,
				is_sample_rate_ratio : 1,
				is_inverted : 1;
};

static void
handle_toggle_event(GtkToggleButton* toggle, DoAdjustmentInfo* info)
{
  gdouble value;
  if (gtk_toggle_button_get_active (toggle))
    value = +1.0;
  else
    value = -1.0;
  gdam_ladspa_filter_set_control (info->filter,
				  info->param_index, value, value);
}

static void
handle_control_update (GdamLadspaFilter *ladspa_filter, 
                       int control_index, 
		       DoAdjustmentInfo *info)
{
  GtkAdjustment *adj = info->adjustment;
  gdouble value, ovalue;
  gdouble avalue;
  if (control_index != info->param_index)
    return;
  g_return_if_fail (info->adjustment != NULL);
  avalue = info->adjustment->value;
  ovalue = value = ladspa_filter->controls[control_index * 2];
  if (info->is_logarithmic)
    value = log (value);
  if (info->is_sample_rate_ratio)
    value /= 44100.0;
  if (info->is_inverted)
    value = adj->lower - value + (adj->upper - adj->page_size);
  if (avalue == value)
    return;
  gtk_adjustment_set_value (info->adjustment, value);
}

static void
do_set_parameter (GtkObject *object, DoAdjustmentInfo *info)
{
  gdam_ladspa_filter_set_control (info->filter,
				  info->param_index, 
				  info->value, 
				  info->value);
}

static void
handle_adj_event(GtkAdjustment* adj, DoAdjustmentInfo* info)
{
  gdouble value;
  value = adj->value;
  if (info->is_inverted)
    value = adj->lower - value + (adj->upper - adj->page_size);
  if (info->is_logarithmic)
    value = exp (value);
  if (info->is_sample_rate_ratio)
    value *= 44100.0;
  if (info->is_integer)
    value = (int) (value + 0.5);

  if (value == info->filter->controls[info->param_index * 2])
    return;

  gdam_ladspa_filter_set_control (info->filter,
				  info->param_index, value, value);
}

static void gdam_ladspa_post_connection_handler(GdamSkin*       skin)
{
  GdamLadspaSkin *ladspa_skin;
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *ladspa_filter;
  int i;

  ladspa_skin = GDAM_LADSPA_SKIN (skin);
  if (ladspa_skin->interface_box == NULL)
    {
      return;
    }

  ladspa_filter = GDAM_LADSPA_FILTER (skin->model);
  plugin_type = ladspa_filter->plugin_type;
  g_return_if_fail (plugin_type != NULL);

  for (i = 0; i < plugin_type->num_controls; i++)
    {
      GdamLadspaParameterType *parameter = plugin_type->controls + i;
      GtkWidget *control_widget = NULL;
      DoAdjustmentInfo *info;

      info = g_new (DoAdjustmentInfo, 1);
      info->filter = ladspa_filter;
      info->param_index = i;
      info->is_integer = parameter->is_integer;
      info->is_logarithmic = parameter->is_logarithmic;
      info->is_sample_rate_ratio = parameter->is_sample_rate_ratio;
      info->is_inverted = FALSE;
      info->adjustment = NULL;

      if (parameter->is_toggle)
	{
	  control_widget = gtk_toggle_button_new ();
	  gtk_signal_connect_full
	    (GTK_OBJECT (control_widget),
	     "toggled",
	     GTK_SIGNAL_FUNC (handle_toggle_event),
	     NULL, info, g_free, FALSE, FALSE);
	}
      else if (parameter->has_min_value && parameter->has_max_value)
	{
	  GtkAdjustment *adjustment;
	  gdouble min_value = parameter->min_value;
	  gdouble max_value = parameter->max_value;
	  if (parameter->is_logarithmic)
	    {
	      if (min_value == 0)
		{
		  min_value = -30.0;
		}
	      else if (min_value > 0)
		min_value = log (min_value);
	      else
		g_warning ("minimum was " "negative but logarithmic");
	      if (min_value == 0)
		{
		  min_value = -30.0;
		}
	      else if (max_value > 0)
		max_value = log (max_value);
	      else
		g_warning ("maximum was " "negative but logarithmic");
	    }
	  adjustment = (GtkAdjustment *)
	    gtk_adjustment_new (min_value,
				min_value, max_value, 0.01, 0.01, 0.01);
	  control_widget = gtk_hscrollbar_new (adjustment);
	  info->adjustment = adjustment;
	}
      else
	{
	  GtkAdjustment *adjustment;
	  g_warning ("using range 0 .. 2 since the plugin "
		     "gives no hint...");
	  if (info->is_logarithmic)
	    {
	      g_warning ("   also, cannot support logarithmic"
			 " with no range hints...");
	      info->is_logarithmic = FALSE;
	    }
	  adjustment = (GtkAdjustment *)
	    gtk_adjustment_new (0.0, 0.0, 2.0, 0.01, 0.01, 0.01);
	  control_widget = gtk_hscrollbar_new (adjustment);
	  info->adjustment = adjustment;
	}
      gtk_signal_connect_full
	(GTK_OBJECT (info->adjustment),
	 "value_changed",
	 GTK_SIGNAL_FUNC (handle_adj_event),
	 NULL, info, g_free, FALSE, FALSE);
      gtk_signal_connect_while_alive (GTK_OBJECT (ladspa_filter),
				      "control_updated",
				      GTK_SIGNAL_FUNC (handle_control_update),
				      info,
				      GTK_OBJECT (control_widget));
      {
	GtkWidget *label;
	label = gtk_label_new (parameter->name);
	gtk_widget_show (label);
	gtk_box_pack_start_defaults (ladspa_skin->interface_box, label);
      }
      gtk_widget_show (control_widget);
      gtk_box_pack_start_defaults (ladspa_skin->interface_box,
				   control_widget);
    }
}

static gboolean gdam_ladspa_interface_box_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  g_return_val_if_fail (GTK_IS_BOX (object), TRUE);
  GDAM_LADSPA_SKIN (skin)->interface_box = GTK_BOX (object);
  return TRUE;
}

static gboolean gdam_ladspa_plugin_name_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamLadspaFilter *filter;
  g_return_val_if_fail (GTK_IS_LABEL (object), TRUE);
  filter = GDAM_LADSPA_FILTER (skin->model);
  gtk_label_set_text (GTK_LABEL (object), filter->plugin_type->database_name);
  //GDAM_LADSPA_SKIN(skin)->name_label = GTK_LABEL(object);
  return TRUE;
}

static void
do_control_save (GdamLadspaSkin *skin)
{
  GdamLadspaFilter *filter;
  filter = GDAM_LADSPA_FILTER (GDAM_SKIN (skin)->model);
  gdam_ladspa_filter_save_preset (filter, "default");
}

static void
do_control_load (GdamLadspaSkin *skin)
{
  GdamLadspaFilter *filter;
  filter = GDAM_LADSPA_FILTER (GDAM_SKIN (skin)->model);
  gdam_ladspa_filter_load_preset (filter, "default");
}


/* TODO: simple connect? */
static gboolean 
gdam_ladspa_save_controls_handler (GdamSkin*          skin,
				   GtkObject*         object,
				   const char*        signal_name,
				   int                num_handler_names,
				   char**             handler_names)
{
  gtk_signal_connect_object (object, 
                             signal_name, 
                             GTK_SIGNAL_FUNC (do_control_save), 
			     GTK_OBJECT (skin));
  return TRUE;
}

/* TODO: simple connect? */
static gboolean 
gdam_ladspa_load_controls_handler (GdamSkin*          skin,
				   GtkObject*         object,
				   const char*        signal_name,
				   int                num_handler_names,
				   char**             handler_names)
{
  gtk_signal_connect_object (object, 
                             signal_name, 
                             GTK_SIGNAL_FUNC (do_control_load), 
			     GTK_OBJECT (skin));
  return TRUE;
}


static gboolean gdam_ladspa_parameter_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamLadspaSkin *ladspa_skin;
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *ladspa_filter;
  GdamLadspaParameterType *parameter;
  GtkAdjustment *adjustment;
  DoAdjustmentInfo *info;
  int param_index;
  int i;

  ladspa_skin = GDAM_LADSPA_SKIN (skin);
  ladspa_filter = GDAM_LADSPA_FILTER (skin->model);
  plugin_type = ladspa_filter->plugin_type;
  g_return_val_if_fail (plugin_type != NULL, TRUE);
  adjustment = util_get_widget_adjustment (GTK_WIDGET (object));
  g_return_val_if_fail (GTK_IS_ADJUSTMENT (adjustment), TRUE);

  if (num_handler_names < 2) 
    {
      g_warning ("need parameter index for %s", handler_names[0]);
      return TRUE;
    }

  param_index = atoi (handler_names[1]);

  info = g_new (DoAdjustmentInfo, 1);
  info->filter = ladspa_filter;
  info->param_index = param_index;
  info->adjustment = adjustment;
  info->is_inverted = FALSE;

  /* Other param details get default values from the plugin itself */
  parameter = plugin_type->controls + param_index;
  info->is_integer = parameter->is_integer;
  info->is_logarithmic = parameter->is_logarithmic;
  info->is_sample_rate_ratio = parameter->is_sample_rate_ratio;

  for (i = 2; i < num_handler_names; i++)
    {
      if (g_strcasecmp (handler_names[i], "is_logarithmic") == 0)
	{
	  info->is_logarithmic = TRUE;
	}
      if (g_strcasecmp (handler_names[i], "is_inverted") == 0)
	{
	  info->is_inverted = TRUE;
	}
      else
	g_warning ("unknown handler %s in %s", handler_names[i], 
	           handler_names[0]);
      i++;
    }

  /* TODO: set min/max based on hints? */
  if (parameter->has_min_value && parameter->has_max_value && FALSE)
    {
      gdouble min_value = parameter->min_value;
      gdouble max_value = parameter->max_value;
      if (parameter->is_logarithmic)
	{
	  if (min_value == 0)
	    {
	      min_value = -30.0;
	    }
	  else if (min_value > 0)
	    min_value = log (min_value);
	  else
	    g_warning ("minimum was " "negative but logarithmic");
	  if (min_value == 0)
	    {
	      min_value = -30.0;
	    }
	  else if (max_value > 0)
	    max_value = log (max_value);
	  else
	    g_warning ("maximum was " "negative but logarithmic");
	}
    }

  handle_adj_event (adjustment, info);
  gtk_signal_connect_full (GTK_OBJECT (adjustment), "value_changed",
			   GTK_SIGNAL_FUNC (handle_adj_event),
			   NULL, info, g_free, FALSE, FALSE);
  gtk_signal_connect_while_alive (GTK_OBJECT (ladspa_filter),
      				  "control_updated",
				  GTK_SIGNAL_FUNC (handle_control_update),
				  info,
				  GTK_OBJECT (adjustment));
  return TRUE;
}

static gboolean gdam_ladspa_set_parameter_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamLadspaSkin *ladspa_skin;
  GdamLadspaPluginType *plugin_type;
  GdamLadspaFilter *ladspa_filter;
  GdamLadspaParameterType *parameter;
  DoAdjustmentInfo *info;
  int param_index;
  int i;

  ladspa_skin = GDAM_LADSPA_SKIN (skin);
  ladspa_filter = GDAM_LADSPA_FILTER (skin->model);
  plugin_type = ladspa_filter->plugin_type;
  g_return_val_if_fail (plugin_type != NULL, TRUE);

  if (num_handler_names < 3) 
    {
      g_warning ("need parameter index and value for %s", handler_names[0]);
      return TRUE;
    }

  param_index = atoi (handler_names[1]);

  info = g_new (DoAdjustmentInfo, 1);
  info->filter = ladspa_filter;
  info->param_index = param_index;
  info->value = atof (handler_names[2]);
  info->is_inverted = FALSE;

  /* Other param details get default values from the plugin itself */
  parameter = plugin_type->controls + param_index;
  info->is_integer = parameter->is_integer;
  info->is_logarithmic = parameter->is_logarithmic;
  info->is_sample_rate_ratio = parameter->is_sample_rate_ratio;

  gtk_signal_connect_full (object, signal_name,
			   GTK_SIGNAL_FUNC (do_set_parameter),
			   NULL, info, g_free, FALSE, FALSE);

  return TRUE;
}



static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_ladspa_interface_box",
	   gdam_ladspa_interface_box_handler },
	{ "gdam_ladspa_plugin_name",
	   gdam_ladspa_plugin_name_handler },
	{ "gdam_ladspa_save_controls",
	   gdam_ladspa_save_controls_handler },
	{ "gdam_ladspa_load_controls",
	   gdam_ladspa_load_controls_handler },
	{ "gdam_ladspa_parameter_adjustment",
	   gdam_ladspa_parameter_adjustment_handler },
	{ "gdam_ladspa_set_parameter",
	   gdam_ladspa_set_parameter_handler },
};

static GList* ladspa_skin_list = NULL;

#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <gdam/component/myftw.h>

static int 
ladspa_walk_callback (const char*        file,
                      const struct stat* sb,
		      int                flag)
{
  if (flag == FTW_F)
    if (g_strncasecmp (file+strlen(file)-4, ".xml", 4) == 0)
      ladspa_skin_list = g_list_append (ladspa_skin_list, g_strdup (file));
  return 0;			/*  keep walking  */
}

static void gdam_ladspa_skin_class_init(GdamSkinClass* skin_class)
{
  char **names;
  skin_class->post_connecting_handler = gdam_ladspa_post_connection_handler;
  skin_class->build_new_model = gdam_ladspa_build_new_model;

  (void) GDAM_TYPE_LADSPA_FILTER;
  (void) GDAM_TYPE_BUILDER_SKIN;

  /* Get list of extra ladspa skins */
  {
    char buf[1024];
    char *ladspa_skin_dirs[16];
    int num = 16; /* max num of ladspa skin directories. */
    int i;
    gdam_search_path_all ("ladspa_skins/", NULL, 
	                  GDAM_SEARCH_PATH_FIND_DIRECTORY,
	                  ladspa_skin_dirs, &num);
    for (i = 0; i < num; i++)
      {
	gdam_debug ("sourcing dir %s for ladspa skins", ladspa_skin_dirs[i]);
	ftw (ladspa_skin_dirs[i], ladspa_walk_callback, 3);
      }
  }

  /* Iterate the existing filters and add them
   * to the builder skin. */
  names = gdam_ladspa_plugin_type_get_names ();
  if (names != NULL)
    {
      while (*names)
	{
	  const char *name = *names++;
	  const char *options[2];
	  const char *label_name = name;	/* TODO: change this? */
	  GdamLadspaPluginType *plugin_type;
	  int unique_id;
	  char buf[1024];
	  char *default_skin;
	  GList *cur_skin;

	  plugin_type = gdam_ladspa_plugin_type_get_by_name (name);
	  unique_id = plugin_type->unique_id;
	  options[0] = name;
	  options[1] = NULL;
	  snprintf (buf, sizeof (buf), "ladspa_skins/%d.xml", unique_id);
	  default_skin = gdam_search_path 
	                   (buf, NULL, GDAM_SEARCH_PATH_FIND_FILE);
	  /* add default skin */
	  if (default_skin)
	    {
	      gdam_builder_skin_class_add_extra
		("GdamLadspaSkin", default_skin, label_name, options);
	      g_free (default_skin);
	    }
	  else
	    gdam_builder_skin_class_add_extra
	      ("GdamLadspaSkin", "ladspa_skin.xml", label_name, options);
	  /* look for extra skins which use this filter type.  example filename
	   * "1197_3band.xml" */
	  cur_skin = ladspa_skin_list;
	  while (cur_skin != NULL)
	    {
	      const char *fname = (const char *)cur_skin->data;
	      const char *basename = g_basename (fname);
	      int num_digits = 0;
	      const char *char_ptr = basename;
	      int val = 0;
	      int our_num_digits = strlen (buf + 13) - 4;
	      cur_skin = cur_skin->next;
	      /* If it is our default skin (i.e. nnnn.xml) skip it
	       * (we already handled it above) */
	      if (g_strcasecmp (basename, buf+13) == 0)
		continue;
	      while (char_ptr && isdigit (*char_ptr))
		{
		  char_ptr++;
		  num_digits++;
		}
	      /* If the filename begins with our unique_id, add it */
	      if (num_digits == our_num_digits &&
		  g_strncasecmp (basename, buf+13, our_num_digits) == 0)
		{
		  gdam_builder_skin_class_add_extra
		    ("GdamLadspaSkin", fname, basename, options);
		  gdam_debug ("LadspaSkin: adding extra skin %s", fname);
		}
	    }
	}
    }

  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
}

GtkType gdam_ladspa_skin_get_type()
{
  static GtkType ladspa_skin_type = 0;
  if (!ladspa_skin_type)
    {
      static const GtkTypeInfo ladspa_skin_info = {
	"GdamLadspaSkin",
	sizeof (GdamLadspaSkin),
	sizeof (GdamLadspaSkinClass),
	(GtkClassInitFunc) gdam_ladspa_skin_class_init,
	(GtkObjectInitFunc) gdam_ladspa_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      ladspa_skin_type = gtk_type_unique (parent, &ladspa_skin_info);
    }
  return ladspa_skin_type;
}
