#include "gdammod.h"
#include <gdam/common/gdamdebug.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

static int get_be16(const char* data)
{
  const guint8 *d = (const guint8 *) data;
  int sum = ((int) (d[0]) << 8) | ((int) d[1]);
  return sum;
}

GdamModBuffer* gdam_mod_buffer_parse(FILE* fp, GdamModFile* mod_file)
{
  GdamModBuffer *buffer;
  GdamModVariant variant = mod_file->variant;
  char sample_info[30];
  if (fread (sample_info, sizeof (sample_info), 1, fp) != 1)
    return NULL;

  g_return_val_if_fail (variant == GDAM_MOD_MOD, NULL);

  buffer = g_new (GdamModBuffer, 1);
  {
    char tmp_name[23];
    memcpy (tmp_name, sample_info, 22);
    tmp_name[22] = '\0';
    buffer->sample_name = g_strdup (tmp_name);
  }
  buffer->sample_data = NULL;
  buffer->sample_length = get_be16 (sample_info + 22);
  buffer->finetune = (guint8) sample_info[24];
  buffer->linear_volume = (guint8) sample_info[25];
  buffer->repeat_offset = get_be16 (sample_info + 26);
  buffer->repeat_length = get_be16 (sample_info + 28);
  if (buffer->sample_length < buffer->repeat_offset + buffer->repeat_length)
    {
      g_warning ("repeat range outside of sample range bounds");
      g_free (buffer);
      return NULL;
    }
  /* XXX: it would be good to add every sensible sanity
     check here.  (there isn't much of a header to use
     to even verify we are looking at a .mod file
     otherwise...) */
  return buffer;
}
static void gdam_mod_buffer_destroy(GdamModBuffer* buffer)
{
  g_free (buffer->sample_data);
  g_free (buffer->sample_name);
  g_free (buffer);
}


/* Frequency => period map. */

/*static gdouble period_to_frequency(int period)
{
...}

*/

/* All these numbers were taken from the xm.txt file,
 * but the author clearly didn't understand their
 * derivation.
 */

static int amiga_note_table[96] =
{
  907, 900, 894, 887, 881, 875, 868, 862, 856, 850, 844, 838, 832, 826, 820,
    814, 808, 802, 796, 791, 785, 779, 774, 768, 762, 757, 752, 746, 741, 736,
    730, 725, 720, 715, 709, 704, 699, 694, 689, 684, 678, 675, 670, 665, 660,
    655, 651, 646, 640, 636, 632, 628, 623, 619, 614, 610, 604, 601, 597, 592,
    588, 584, 580, 575, 570, 567, 563, 559, 555, 551, 547, 543, 538, 535, 532,
    528, 524, 520, 516, 513, 508, 505, 502, 498, 494, 491, 487, 484, 480, 477,
    474, 470, 467, 463, 460, 457};


static gdouble amiga_note_to_frequency(int note, int finetune)
{
  gdouble n;
  g_return_val_if_fail (0 <= note && note < 96, 0.0);
  n = amiga_note_table[note];
  return 8363.0 * 1712.0 / n;
}
static gdouble linear_note_to_frequency(int note, int finetune)
{
  /* XXX: clean this up!!! it is totally poorly writtten.
     [from the xm.txt file] */
  gdouble p = (gdouble) (10 * 12 * 16 * 4 - 16 * 4 * note) - 0.5 * finetune;

  return 8363.0 * pow (2.0,
		       floor ((6.0 * 12 * 16 * 4 - p) / (12.0 * 16 * 4)));
}



/*
 *   ___                       _   _  _             _
 *  / __|___ _ _  ___ _ _ __ _| | | || |___ __ _ __| |___ _ _
 * | (_ / -_) ' \/ -_) '_/ _` | | | __ / -_) _` / _` / -_) '_|
 *  \___\___|_||_\___|_| \__,_|_| |_||_\___\__,_\__,_\___|_|
 * 
 *  ___             _
 * | _ \__ _ _ _ __(_)_ _  __ _
 * |  _/ _` | '_(_-< | ' \/ _` |
 * |_| \__,_|_| /__/_|_||_\__, |
 *                        |___/
 */
static gboolean guess_format_from_magic(const char*  magic,
                                        int*         num_buffers_out,
			                int*         num_channels_out)
{
  if (memcmp (magic, "M.K.", 4) == 0
      || memcmp (magic, "FLT4", 4) == 0
      || memcmp (magic, "M!K!", 4) == 0 || memcmp (magic, "4CHN", 4) == 0)
    {
      *num_channels_out = 4;
      *num_buffers_out = 31;
      return TRUE;
    }
  if (memcmp (magic, "6CHN", 4) == 0)
    {
      *num_channels_out = 6;
      *num_buffers_out = 31;
      return TRUE;
    }
  if (memcmp (magic, "8CHN", 4) == 0 || memcmp (magic, "OCTA", 4) == 0)
    {
      *num_channels_out = 8;
      *num_buffers_out = 31;
      return TRUE;
    }

  g_warning ("unknown mod file magic `%c%c%c%c'",
	     magic[0], magic[1], magic[2], magic[3]);
  return FALSE;
}

static gboolean guess_format(FILE*             fp,
			     GdamModVariant*   variant_out,
                             int*              num_buffers_out,
			     int*              num_channels_out)
{
  char magic[4];
  char xm_header[17];

  if (fread (xm_header, sizeof (xm_header), 1, fp) == 1)
    {
      if (memcmp ("Extended module: ", xm_header, 17) == 0
	  || memcmp ("Extended Module: ", xm_header, 17) == 0)
	{
	  *variant_out = GDAM_MOD_XM;
	  return TRUE;
	}
    }
  if (fseek (fp, 1081, SEEK_SET) < 0)
    {
      g_warning ("couldn't seek mod file " "(necessary for parsing)");
      return FALSE;
    }
  if (fread (magic, 4, 1, fp) < 1)
    {
      g_warning ("couldn't read magic from mod file");
      return FALSE;
    }
  *variant_out = GDAM_MOD_MOD;
  return guess_format_from_magic (magic, num_buffers_out, num_channels_out);
}

static int parse_byte(FILE* fp)
{
  int rv = fgetc (fp);
  if (rv == EOF)
    return -1;
  return (guint8) rv;
}
	
static char* memchomp(const char* name, int len)
{
  char *rv;
  while (len > 0 && (name[len - 1] == ' ' || name[len - 1] == 0))
    len--;
  rv = g_new (char, len + 1);
  memcmp (rv, name, len);
  rv[len] = '\0';
  return rv;
}

static GdamModFile* gdam_mod_file_parse_xm_header(FILE* fp)
{
  char module_name[20];
  char tracker_name[20];
  char padding;
  guint8 minor_version, major_version;
  guint32 header_size;
  guint16 song_length;
  guint16 restart_position;
  guint16 num_channels;
  guint16 num_patterns;
  guint16 num_instr;
  guint16 flags;
  guint16 tempo, bpm;
  guint8 pattern_order_table[256];

  GdamModFile *mod_file;

  g_assert (ftell (fp) == 17);

  if (fread (module_name, 20, 1, fp) != 1
      || fread (&padding, 1, 1, fp) != 1
      || fread (tracker_name, 20, 1, fp) != 1
      || fread (&major_version, 1, 1, fp) != 1
      || fread (&minor_version, 1, 1, fp) != 1
      || fread (&header_size, 4, 1, fp) != 1
      || fread (&song_length, 2, 1, fp) != 1
      || fread (&restart_position, 2, 1, fp) != 1
      || fread (&num_channels, 2, 1, fp) != 1
      || fread (&num_patterns, 2, 1, fp) != 1
      || fread (&num_instr, 2, 1, fp) != 1
      || fread (&flags, 2, 1, fp) != 1
      || fread (&tempo, 2, 1, fp) != 1
      || fread (&bpm, 2, 1, fp) != 1
      || fread (pattern_order_table, 256, 1, fp) != 1)
    {
      g_warning ("xm file too short reading header");
      return NULL;
    }


  if (major_version != 1)
    {
      g_warning ("only major versions 1.x are supported by "
		 "this xm parser");
      //return NULL;
    }
  if (minor_version != 3)
    {
      g_warning ("this reader only support XM 1.3, this file is "
		 "%d.%d", major_version, minor_version);
      //return NULL;
    }

  header_size = GUINT32_FROM_BE (header_size);
  song_length = GUINT16_FROM_BE (song_length);
  restart_position = GUINT16_FROM_BE (restart_position);
  num_channels = GUINT16_FROM_BE (num_channels);
  num_patterns = GUINT16_FROM_BE (num_patterns);
  num_instr = GUINT16_FROM_BE (num_instr);
  flags = GUINT16_FROM_BE (flags);
  tempo = GUINT16_FROM_BE (tempo);
  bpm = GUINT16_FROM_BE (bpm);

  g_message ("header_size was %d, current position %d",
	     (int) header_size, (int) ftell (fp));

  mod_file = g_new0 (GdamModFile, 1);
  mod_file->title = memchomp (module_name, 20);
  mod_file->num_buffers = num_instr;
  mod_file->variant = GDAM_MOD_XM;
  mod_file->strict = 1;
  mod_file->pattern_table_size = 256;
  mod_file->pattern_table = g_memdup (pattern_order_table, 256);
  mod_file->num_patterns = num_patterns;
  return mod_file;
}

/* If it returns non-null, it is at the start of the pattern
 * data. */
GdamModFile* gdam_mod_file_parse_header(FILE* fp)
{
  char *song_name;
  GdamModBuffer **mod_buffers;
  int num_buffers = 32, num_channels = 4;
  int song_end_jump_pos;
  GdamModFile *header;
  GdamModVariant variant;
  if (!guess_format (fp, &variant, &num_buffers, &num_channels))
    {
      g_warning ("couldn't figure out this mod file's format");
      return NULL;
    }
  if (variant == GDAM_MOD_XM)
    return gdam_mod_file_parse_xm_header (fp);

  gdam_debug ("MOD parser: %d channels, %d instruments",
	      num_buffers, num_channels);
  if (fseek (fp, 0, SEEK_SET) < 0)
    {
      g_warning ("cannot rewind MOD file, cannot continue");
      return NULL;
    }
  {
    char tmp_buf[21];
    if (fread (tmp_buf, 20, 1, fp) != 1)
      return NULL;
    tmp_buf[20] = '\0';
    song_name = g_strdup (tmp_buf);
  }
  header = g_new0 (GdamModFile, 1);
  header->title = song_name;
  header->num_buffers = num_buffers;
  header->variant = variant;

  mod_buffers = g_new (GdamModBuffer *, num_buffers);

  {
    int i;
    for (i = 0; i < num_buffers; i++)
      {
	mod_buffers[i] = gdam_mod_buffer_parse (fp, header);
	if (mod_buffers[i] == NULL)
	  {
	    int j;
	    g_warning ("error reading mod buffer %d", i);
	    for (j = 0; j < i; j++)
	      gdam_mod_buffer_destroy (mod_buffers[j]);
	    g_free (song_name);
	    g_free (mod_buffers);
	    return NULL;
	  }
      }
    num_buffers = i;
  }

  header->buffers = mod_buffers;

  header->num_patterns = parse_byte (fp);
  song_end_jump_pos = parse_byte (fp);
  if (header->num_patterns < 0 || song_end_jump_pos < 0)
    {
      g_warning ("error reading num patterns");
      gdam_mod_file_destroy (header);
      return NULL;
    }

  header->pattern_table_size = 128;
  header->pattern_table = g_new (guint8, header->pattern_table_size);

  if (fread (header->pattern_table, header->pattern_table_size, 1, fp) != 1)
    {
      g_warning ("error (eof) reading pattern table");
      gdam_mod_file_destroy (header);
      return NULL;
    }
  return header;
}

/*
 *  ____       _   _                    ____                _
 * |  _ \ __ _| |_| |_ ___ _ __ _ __   |  _ \ __ _ _ __ ___(_)_ __   __ _
 * | |_) / _` | __| __/ _ \ '__| '_ \  | |_) / _` | '__/ __| | '_ \ / _` |
 * |  __/ (_| | |_| ||  __/ |  | | | | |  __/ (_| | |  \__ \ | | | | (_| |
 * |_|   \__,_|\__|\__\___|_|  |_| |_| |_|   \__,_|_|  |___/_|_| |_|\__, |
 *                                                                  |___/
 */

static gboolean parse_xm_note(FILE* fp, GdamModNote* note_out, gboolean amiga)
{
  int flags;
  int b, i;
  guint8 args[5];
  b = fgetc (fp);
  args[0] = args[1] = args[2] = args[3] = args[4] = 0;
  if (b == EOF)
    return FALSE;
  b = (guint8) b;
  /* Lame "compression" scheme devised by the imbecile
   * who came up with this crappy format.  Exactly who found
   * any useful savings trimming at most 3 of 6 bytes out of a 
   * pattern element (ie nonsample data...)
   *
   * I guess actually measuring the savings was too much
   * work for that author, when invented a new pseudostandard
   * format.
   *
   * The bits in `flags' are interpreted as which fields to
   * read next, the bits are:
   *
   */
  if ((b & 0x80) == 0x80)
    {
      flags = (b & (~0x80));
    }
  else
    {
      args[0] = b;
      flags = 0x1e;
    }

  for (i = 0; i < 5; i++)
    {
      if ((flags & (1 << i)) != 0)
	{
	  b = fgetc (fp);
	  if (b == EOF)
	    return FALSE;
	  args[i] = (guint8) b;
	}
    }
  note_out->sample = args[1];
  note_out->effect = args[3];
  note_out->effect_arg = args[4];
  if (amiga)
    note_out->hertz = amiga_note_to_frequency (args[0], 0);
  else
    note_out->hertz = linear_note_to_frequency (args[0], 0);

  /* XXX: figure out what to do with ``Volume column byte'' */

  return TRUE;
}
			
static GdamModPattern*
gdam_mod_xm_pattern_parse(FILE* fp, GdamModFile* mod_file)
{
  GdamModPattern *pattern;
  int i;
  gboolean amiga_notes = mod_file->amiga_notes;
  pattern = g_new (GdamModPattern, 1);
  for (i = 0; i < 64; i++)
    {
      if (!parse_xm_note (fp, pattern->notes + i, amiga_notes))
	{
	  g_warning ("end reading note %i in pattern", i);
	  g_free (pattern);
	  return NULL;
	}
    }
  return pattern;
}

static void parse_note_bits(guint8*    note_bits,
                            GdamModNote* note_out)
{
  int sample_number;
  int sample_period;
  int effect;
  int effect_arg;
  sample_number = (note_bits[0] & 0xf0) | (note_bits[1] >> 4);
  sample_period = ((int) (note_bits[0] & 0x0f) << 8) | (int) (note_bits[1]);
  effect = note_bits[2] & 0x0f;
  effect_arg = note_bits[3];

  note_out->sample = sample_number;
  note_out->effect = effect;
  note_out->effect_arg = effect_arg;

  /* UMMM this is probably just wrong.  find refrence */
  note_out->hertz = 8363.0 * 1712.0 / sample_period;
}

GdamModPattern* gdam_mod_pattern_parse(FILE* fp, GdamModFile* mod_file)
{
  GdamModPattern *pattern;
  guint8 raw_data[64 * 4];
  GdamModVariant variant = mod_file->variant;
  int i;
  if (variant == GDAM_MOD_XM)
    return gdam_mod_xm_pattern_parse (fp, mod_file);

  g_return_val_if_fail (variant == GDAM_MOD_MOD, NULL);

  if (fread (raw_data, sizeof (raw_data), 1, fp) < 1)
    {
      g_warning ("eof encountered reading file");
      return NULL;
    }
  pattern = g_new (GdamModPattern, 1);
  for (i = 0; i < 64; i++)
    parse_note_bits (raw_data + i * 4, pattern->notes + i);
  return pattern;
}

/*=====================================================================
 *                       gdam_mod_file_parse
 */

GdamModFile* gdam_mod_file_parse(FILE* fp)
{
  GdamModFile *mod_file;
  int i;
  mod_file = gdam_mod_file_parse_header (fp);
  if (mod_file == NULL)
    return NULL;
  mod_file->patterns = g_new (GdamModPattern *, mod_file->num_patterns);
  for (i = 0; i < mod_file->num_patterns; i++)
    mod_file->patterns[i] = gdam_mod_pattern_parse (fp, mod_file);
  for (i = 0; i < mod_file->num_buffers; i++)
    {
      GdamModBuffer *buf = mod_file->buffers[i];
      if (buf == NULL || buf->sample_length <= 0)
	continue;
      buf->sample_data = g_new (gint8, buf->sample_length);
      if (fread (buf->sample_data, buf->sample_length, 1, fp) < 1)
	{
	  g_warning ("end-of-file encountered in sample %d's data", i);
	  gdam_mod_file_destroy (mod_file);
	  return NULL;
	}
    }
  return mod_file;
}


/*
 *  ____            _                   _   _
 * |  _ \  ___  ___| |_ _ __ _   _  ___| |_(_) ___  _ __
 * | | | |/ _ \/ __| __| '__| | | |/ __| __| |/ _ \| '_ \
 * | |_| |  __/\__ \ |_| |  | |_| | (__| |_| | (_) | | | |
 * |____/ \___||___/\__|_|   \__,_|\___|\__|_|\___/|_| |_|
 * 
 */

void gdam_mod_file_destroy(GdamModFile* file)
{
  g_free (file->title);
  if (file->buffers != NULL)
    {
      int i;
      for (i = 0; i < file->num_buffers; i++)
	gdam_mod_buffer_destroy (file->buffers[i]);
      g_free (file->buffers);
    }
  if (file->patterns != NULL)
    {
      int i;
      for (i = 0; i < file->num_patterns; i++)
	g_free (file->patterns[i]);
    }
}

#if TEST
int main(int argc, char** argv)
{
  FILE *fp;
  int i;
  GdamModFile *mod_file;

  if (argc != 2)
    g_error ("usage: %s modfile", *argv);

  fp = fopen (argv[1], "r");
  if (fp == NULL)
    g_error ("couldn't open %s: %s", argv[1], strerror (errno));

  mod_file = gdam_mod_file_parse (fp);
  if (mod_file == NULL)
    g_error ("couldn't parse mod file %s", argv[1]);
  fclose (fp);

  printf ("mod file %s.\n", mod_file->title);
  for (i = 0; i < mod_file->num_buffers; i++)
    {
      GdamModBuffer *buf = mod_file->buffers[i];
      if (buf == NULL)
	continue;
      printf ("   sample %d: %s: sample length=%d.\n",
	      i, buf->sample_name, buf->sample_length);
    }
  for (i = 0; i < mod_file->num_patterns; i++)
    {
      int j;
      printf ("Pattern %d.\n", i);
      for (j = 0; j < 64; j++)
	{
	  GdamModNote *note = mod_file->patterns[i]->notes + j;
	  printf ("   ");
	  printf ("note %d: sample %d, effect %d, hertz %4.4g\n",
		  i, note->sample, note->effect, note->hertz);
	}
    }
  return 0;
}
#endif
