/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamspectrumscrollskin.h"
#include <gdam/skin/gdamskinutil.h>
#include <gtk/gtkdrawingarea.h>
#include <string.h>
#include <gdam/component/plugins/gdampowerspectrum.h>
#include <gdam/common/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gdam/skin/gdamskinutil.h>
#include <gdam/common/gdamplugin.h>
#include <math.h>

static GtkObjectClass* parent_class = NULL;
static void schedule_query_timeout(GdamSpectrumScrollSkin* skin);

#define NUM_COLORS 16

static void
gdam_spectrum_scroll_skin_init(GdamSpectrumScrollSkin* power_spectrum_skin)
{
  schedule_query_timeout (power_spectrum_skin);
  power_spectrum_skin->frequency_scale = 6.0;
  power_spectrum_skin->frequency_base = 2;
  power_spectrum_skin->normalization_level = 0.3;
  power_spectrum_skin->power_log_base = 10;
}
static void gdam_spectrum_scroll_skin_class_init(GdamSkinClass* skin_class);

GtkType gdam_spectrum_scroll_skin_get_type()
{
  static GtkType power_spectrum_skin_type = 0;
  if (!power_spectrum_skin_type)
    {
      static const GtkTypeInfo power_spectrum_skin_info = {
	"GdamSpectrumScrollSkin",
	sizeof (GdamSpectrumScrollSkin),
	sizeof (GdamSpectrumScrollSkinClass),
	(GtkClassInitFunc) gdam_spectrum_scroll_skin_class_init,
	(GtkObjectInitFunc) gdam_spectrum_scroll_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      parent_class = gtk_type_class (parent);
      power_spectrum_skin_type
	= gtk_type_unique (parent, &power_spectrum_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return power_spectrum_skin_type;
}

static void alloc_gradation(GdkGC** gcs, GdkPixmap* pixmap,
                            int start, int end,
			    float r1, float g1, float b1,
			    float r2, float g2, float b2)
{
  int i;
  for (i = start; i < end; i++)
    {
      float fraction = (float) (i - start) / (end - start);
      float r = r1 * (1.0 - fraction) + r2 * fraction;
      float g = g1 * (1.0 - fraction) + g2 * fraction;
      float b = b1 * (1.0 - fraction) + b2 * fraction;
      int r_int = r * 65535.0;
      int g_int = g * 65535.0;
      int b_int = b * 65535.0;
      GdkGC *gc;
      GdkColor *color;
      gc = gdk_gc_new (pixmap);
      color = util_alloc_color_cached (r_int, g_int, b_int);
      gdk_gc_set_foreground (gc, color);
      gcs[i] = gc;
    }
}

static void alloc_colors(GdamSpectrumScrollSkin* skin)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  int i;
#if 1
  alloc_gradation (skin->gray_family, skin->pixmap,
		   0, NUM_COLORS / 3, 0, 0, 0, 0, 0, 1);
  alloc_gradation (skin->gray_family, skin->pixmap,
		   NUM_COLORS / 3, NUM_COLORS * 2 / 3, 0, 0, 1, 1, 0, 0);
  alloc_gradation (skin->gray_family, skin->pixmap,
		   NUM_COLORS * 2 / 3, NUM_COLORS, 1, 0, 0, 1, 1, 1);
#else
  alloc_gradation (skin->gray_family, skin->pixmap,
		   0, NUM_COLORS, 1, 1, 1, 0, 0, 0);
#endif
}


static void redraw_area(GdamSpectrumScrollSkin* skin,
                        GdkRectangle*          area)
{
  int state = GTK_WIDGET_STATE (skin->drawing_area);
  if (skin->swap_axes)
    {
      /* Draw the earlier part of the pixmap: */
      gdk_draw_pixmap (skin->drawing_area->window,
		       skin->drawing_area->style->fg_gc[state],
		       skin->pixmap,
		       0, skin->start_y,
		       0, 0,
		       skin->draw_width, skin->draw_height - skin->start_y);

      /* Draw the later part of the pixmap: */
      gdk_draw_pixmap (skin->drawing_area->window,
		       skin->drawing_area->style->fg_gc[state],
		       skin->pixmap,
		       0, 0,
		       0, skin->draw_height - skin->start_y,
		       skin->draw_width, skin->start_y);
    }
  else
    {
      /* Draw the earlier part of the pixmap: */
      gdk_draw_pixmap (skin->drawing_area->window,
		       skin->drawing_area->style->fg_gc[state],
		       skin->pixmap,
		       skin->start_x, 0,
		       0, 0,
		       skin->draw_width - skin->start_x, skin->draw_height);

      /* Draw the later part of the pixmap: */
      gdk_draw_pixmap (skin->drawing_area->window,
		       skin->drawing_area->style->fg_gc[state],
		       skin->pixmap,
		       0, 0,
		       skin->draw_width - skin->start_x, 0,
		       skin->start_x, skin->draw_height);
    }
}

static int 
find_color_from_intensity(GdamSpectrumScrollSkin* skin, gdouble intensity)
{
  //int index = log(intensity) / log(skin->power_log_base) * NUM_COLORS;
  int index = log (intensity) / skin->power_log_base * NUM_COLORS;
  return CLAMP (index, 0, NUM_COLORS - 1);
}

static void gdam_spectrum_scroll_set_doubles(GdamSpectrumScrollSkin* skin,
                                            int                    num_doubles,
					    gdouble*               doubles)
{
  GdkSegment *lower;
  int draw_width = skin->draw_width;
  int draw_height = skin->draw_height;
  GArray *points[NUM_COLORS];
  gboolean had_error = FALSE;
  int i;
  int freq_points;
  int start_x = skin->start_x;
  int start_y = skin->start_y;
  int delta;
  gdouble max_double, nf;
  gfloat *y_indices = skin->y_indices;
  g_return_if_fail (y_indices != NULL);

  if (skin->swap_axes)
    freq_points = skin->draw_width;
  else
    freq_points = skin->draw_height;

  for (i = 0; i < NUM_COLORS; i++)
    points[i] = g_array_new (FALSE, FALSE, sizeof (GdkPoint));

  /* Normalization constant. */
  max_double = 0.0;
  for (i = 0; i < num_doubles; i++)
    if (max_double < doubles[i])
      max_double = doubles[i];
  if (max_double < 0.01)
    max_double = 0.01;
  nf = 1 / max_double;
  nf = nf * skin->normalization_level + (1.0 - skin->normalization_level);

  /*  For each pixel, determine color and add to appropriate array. */
  for (i = 0; i < freq_points; i++)
    {
      GdkPoint tmp, tmp2;
      int index;
      int color = 0;
      if (skin->swap_axes)
	{
	  tmp.y = start_y;
	  tmp2.y = start_y + 1;
	  tmp.x = tmp2.x = draw_width - i - 1;
	}
      else
	{
	  tmp.x = start_x;
	  tmp2.x = start_x + 1;
	  tmp.y = tmp2.y = draw_height - i - 1;
	}
      /*  y_indices is a table of frequency coefficients to map from
       *  pixel we're drawing to point within spectrum array which
       *  should provide the coloring for that pixel. */
      index = y_indices[i] * num_doubles;
      if (index < num_doubles)
	color = find_color_from_intensity (skin, doubles[index] * nf);
      else if (!had_error)
	{
	  had_error = TRUE;
	  g_warning ("SEG FAULT in spectrum scrolled.");
	}
      g_array_append_val (points[color], tmp);
    }

  /*  For each color, draw the points of that color.  */
  for (i = 0; i < NUM_COLORS; i++)
    {
      GArray *point_array = points[i];
      if (point_array->len > 0)
	{
	  gdk_draw_points (skin->pixmap,
			   skin->gray_family[i],
			   (GdkPoint *) point_array->data, point_array->len);
	}
      g_array_free (point_array, TRUE);
    }

  delta = 1;
  if (skin->swap_axes && skin->swap_direction)
    {
      skin->start_y -= delta;
      if (skin->start_y < 0)
	skin->start_y = draw_height - delta;
    }
  else if (skin->swap_axes)
    {
      skin->start_y += delta;
      if (skin->start_y >= draw_height - delta + 1)
	skin->start_y = 0;
    }
  else if (skin->swap_direction)
    {
      skin->start_x -= delta;
      if (skin->start_x < 0)
	skin->start_x = draw_width - delta;
    }
  else
    {
      skin->start_x += delta;
      if (skin->start_x >= draw_width - delta + 1)
	skin->start_x = 0;
    }

  redraw_area (skin, NULL);
}

static gint
power_spectrum_expose(GtkDrawingArea* area,
             GdkEventExpose* expose_event,
             GdamSpectrumScrollSkin*  skin)
{
  if (skin->gray_family[0] == NULL)
    alloc_colors (skin);
  redraw_area (skin, &expose_event->area);
  return TRUE;
}
static void 
recompute_y_indices(GdamSpectrumScrollSkin* skin)
{
  int i;
  int freq_points;
  g_free (skin->y_indices);
  if (skin->swap_axes)
    freq_points = skin->draw_width;
  else
    freq_points = skin->draw_height;
  skin->y_indices = g_new (gfloat, freq_points);
  for (i = 0; i < freq_points; i++)
    {
      gfloat fraction = (gfloat) i / freq_points - 1.0;
      fraction *= skin->frequency_scale;
      fraction -= skin->frequency_base;
      skin->y_indices[i] = exp (fraction);
    }
}

static gint
power_spectrum_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamSpectrumScrollSkin* skin)
{
  if (skin->pixmap)
    gdk_pixmap_unref (skin->pixmap);
  skin->pixmap = gdk_pixmap_new (widget->window,
				 widget->allocation.width,
				 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  if (skin->gray_family[0] == NULL)
    alloc_colors (skin);
  recompute_y_indices (skin);
  return TRUE;
}

static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamSpectrumScrollSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 800, 300);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event",
		      GTK_SIGNAL_FUNC (power_spectrum_expose), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (power_spectrum_configure), skin);
  gtk_widget_show (GTK_WIDGET (drawing_area));
  gtk_widget_add_events (GTK_WIDGET (drawing_area), GDK_EXPOSURE_MASK);
}


/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */
static void got_spectrum(int num_args, GdamArg* args, gpointer skin_ptr)
{
  GdamSpectrumScrollSkin *skin = skin_ptr;
  GdamSkin *skin_base = GDAM_SKIN (skin_ptr);
  g_return_if_fail (num_args == 1);
  g_return_if_fail (args->type == GDAM_ARG_TYPE_DOUBLE_ARRAY);

  skin->request_pending = FALSE;

  if (!GTK_OBJECT_DESTROYED (skin)
      && skin_base->main_widget != NULL
      && !GTK_OBJECT_DESTROYED (skin_base->main_widget))
    {
      gdam_spectrum_scroll_set_doubles
	(skin,
	 GDAM_ARG_DOUBLE_ARRAY (args).length,
	 GDAM_ARG_DOUBLE_ARRAY (args).array);
    }

  gtk_object_unref (GTK_OBJECT (skin));
}

static void didnt_get_spectrum(gpointer skin_ptr)
{
  GdamSpectrumScrollSkin *skin = skin_ptr;
  g_warning ("didn't get power spectrum");
  skin->request_pending = FALSE;
  gtk_object_unref (GTK_OBJECT (skin));
}

static gboolean do_query(GdamSpectrumScrollSkin* skin)
{
  GdamModel *model;

  model = GDAM_SKIN (skin)->model;

  if (model == NULL || model->source_id == 0)
    return TRUE;

  if (!skin->request_pending)
    {
      const char *name = "spectrum";
      gtk_object_ref (GTK_OBJECT (skin));
      gdam_channel_append_request
	(model->channel,
	 gdam_request_get_args
	 (model->source_id,
	  1, &name, got_spectrum, didnt_get_spectrum, skin));
      skin->request_pending = TRUE;
    }
  return TRUE;
}

	


static void schedule_query_timeout(GdamSpectrumScrollSkin* skin)
{
  if (skin->query_timeout_id != 0)
    return;
  if (!GTK_OBJECT_DESTROYED (skin))
    {
      skin->query_timeout_id
	= g_timeout_add (30, (GSourceFunc) do_query, skin);
    }
}

static void do_swap_direction(GdamSpectrumScrollSkin* skin){
	skin->swap_direction = !skin->swap_direction;
}

static void do_swap_axes(GdamSpectrumScrollSkin* skin)
{
  skin->swap_axes = !skin->swap_axes;
  if (skin->swap_axes)
    skin->start_y = 0;
  else
    skin->start_x = 0;
}

static gboolean gdam_spectrum_scroll_swap_axes_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name, do_swap_axes, GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_spectrum_scroll_swap_direction_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     do_swap_direction, GTK_OBJECT (skin));
  return TRUE;
}

static gboolean gdam_spectrum_scroll_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
      return TRUE;
    }
  connect_drawing_area (GTK_DRAWING_AREA (object),
			GDAM_SPECTRUM_SCROLL_SKIN (skin));
  return TRUE;
}

static GdamSkinGenericConnect connect_infos[] = { 	
	{ "gdam_spectrum_scroll_drawing_area",
	   gdam_spectrum_scroll_drawing_area_handler },
	{ "gdam_spectrum_scroll_swap_axes",
	   gdam_spectrum_scroll_swap_axes_handler },
	{ "gdam_spectrum_scroll_swap_direction",
	   gdam_spectrum_scroll_swap_direction_handler },
};

static void gdam_spectrum_scroll_skin_destroy(GtkObject* object)

{
  GdamSpectrumScrollSkin *skin = GDAM_SPECTRUM_SCROLL_SKIN (object);
  if (skin->query_timeout_id != 0)
    {
      g_source_remove (skin->query_timeout_id);
      skin->query_timeout_id = 0;
    }
  skin->drawing_area = NULL;
  if (skin->pixmap != NULL)
    {
      gdk_pixmap_unref (skin->pixmap);
      skin->pixmap = NULL;
    }
  (*parent_class->destroy) (object);
}

static void gdam_scrollspec_arg_set_frequency_base(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  ss->frequency_base = GDAM_ARG_DOUBLE (arg);
  recompute_y_indices (ss);
}

static void gdam_scrollspec_arg_get_frequency_base(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ss->frequency_base;
}

static void 
gdam_scrollspec_arg_set_power_log_base(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  ss->power_log_base = GDAM_ARG_DOUBLE (arg);
}

static void 
gdam_scrollspec_arg_get_power_log_base(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ss->power_log_base;
}


static void 
gdam_scrollspec_arg_set_normalization_level(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  ss->normalization_level = GDAM_ARG_DOUBLE (arg);
}

static void 
gdam_scrollspec_arg_get_normalization_level(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ss->normalization_level;
}


static void gdam_scrollspec_arg_set_frequency_scale(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  ss->frequency_scale = GDAM_ARG_DOUBLE (arg);
  recompute_y_indices (ss);
}

static void gdam_scrollspec_arg_get_frequency_scale(GtkObject* object, GdamArg* arg)
{
  GdamSpectrumScrollSkin *ss;
  ss = GDAM_SPECTRUM_SCROLL_SKIN (object);
  GDAM_ARG_DOUBLE (arg) = ss->frequency_scale;
}


static void gdam_spectrum_scroll_skin_class_init(GdamSkinClass* skin_class)
{
  GtkObjectClass *object_class = GTK_OBJECT_CLASS (skin_class);
  skin_class->model_type = GDAM_TYPE_POWER_SPECTRUM;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  object_class->destroy = gdam_spectrum_scroll_skin_destroy;
  gdam_arg_register (object_class,
		     "frequency_scale",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_scrollspec_arg_set_frequency_scale,
		     gdam_scrollspec_arg_get_frequency_scale,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "frequency_base",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_scrollspec_arg_set_frequency_base,
		     gdam_scrollspec_arg_get_frequency_base,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "normalization_level",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_scrollspec_arg_set_normalization_level,
		     gdam_scrollspec_arg_get_normalization_level,
		     NULL, NULL, GDAM_ARG_VITAL);
  gdam_arg_register (object_class,
		     "power_log_base",
		     GDAM_ARG_TYPE_DOUBLE,
		     gdam_scrollspec_arg_set_power_log_base,
		     gdam_scrollspec_arg_get_power_log_base,
		     NULL, NULL, GDAM_ARG_VITAL);
}
