/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencerskin.h"
#include "../gtkpixbutton.h"
#include "../gdamskin-pixwidgets.h"
#include <gdam/component/gdambuffermanager.h>
#include "../gdamskinutil.h"
#include "../../component/plugins/gdammididevice.h"
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtksignal.h>
#include <gtk/gtklabel.h>
#include <gtk/gtklist.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkradiobutton.h>
#include <gdam/common/gdamplugin.h>
#include <gdam/common/gdamarg.h>
#include <gtk/gtkbutton.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <gdam/component/gdamhistory.h>

#define gdam_skin_pop_message(skin, context)
#define gdam_skin_push_message(skin, context, string)


enum {
	SELECTION_MODE_SET,
	SELECTION_MODE_TOGGLE,
	SELECTION_MODE_ADD
};
static GtkObjectClass* parent_class = NULL;

/* Holds info about one element in the sequence. */
typedef struct _GdamBufferInfo GdamBufferInfo;
struct _GdamBufferInfo {
	int		buffer_id;
	int		ref_count;
	char*		buffer_name;
	gdouble         hertz;
	int		render_offset_x, render_offset_y;
};

typedef struct _IntToggleInfo {
  GtkToggleButton   *tbutton;
  GdamSequencerSkin *skin;
  int		     val;
  gboolean	     ignore_button;
} IntToggleInfo;

typedef struct _FloatToggleInfo {
  GtkToggleButton   *tbutton;
  GdamSequencerSkin *skin;
  gdouble	     val;
  gboolean	     ignore_button;
} FloatToggleInfo;

enum _SequencerDisplayMode {
	BY_GAIN,
	BY_RESAMPLE,
	BY_ID,
	BY_LOCAL_ID,
	BY_PATTERN,
	BY_LENGTH,
	BY_START
};

enum _ElementRenderMode {
	RENDER_AS_CIRCLES,
	RENDER_AS_RECTANGLES,
	RENDER_AS_COLOR_SQUARES,
	RENDER_AS_TEE,
	RENDER_AS_PENNANT
};

enum _ElementTextMode {
	TEXT_BUFFER_NAME,
	TEXT_LOCAL_ID
};

enum _SequencerDisplayTool {
	TOOL_ADD,
	TOOL_DELETE,
	TOOL_SELECT,
	TOOL_TRACE_Y_VALUE,
	TOOL_STAMP,
	TOOL_FILL_END_TO_END,
	TOOL_FILL_NEAREST_BEAT,
	TOOL_FILL_NEAREST_AND_STRETCH
};

/* Holds info about all samples with the same buffer id */
typedef struct _GdamSequencerBufferInfo GdamSequencerBufferInfo;
struct _GdamSequencerBufferInfo {
	guint			buffer_id;
	GdamSequencerSkin*	sequencer_skin;
	GtkWidget*		topwidget;
	gdouble			gain_factor;
};

typedef struct _BufIdPtr BufIdPtr;
struct _BufIdPtr {
	GdamSequencerSkin*	sequencer_skin;
	guint			buffer_id;
};
static void delayed_redraw(GdamSequencerSkin* skin);
static void gdam_sequencer_skin_class_init(GdamSkinClass* skin_class);
static GdamSequencerBufferInfo* maybe_track_buffer_id(GdamSequencerSkin* skin,
						      guint         buffer_id);
static void recreate_buffer_list (GdamSequencerSkin *skin);
static GdamSequencerBufferInfo* 
sequence_buffer_info_from_id (GdamSequencerSkin* skin, int id);
static char * get_local_buffer_description (GdamSequencerSkin *skin);

#define MAX_GAIN                3.0

static gboolean       gdam_sequencer_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options);

static int      equal_tempered_table_size;
static gdouble* log_equal_tempered_table = NULL;

static gdouble  min_log_hertz_global;
static gdouble  max_log_hertz_global;

static gdouble* log_equal_tempered_scale()
{
  int i;
  if (log_equal_tempered_table)
    return log_equal_tempered_table;
  equal_tempered_table_size = 12 * 12;
  log_equal_tempered_table = g_new (gdouble, equal_tempered_table_size);
  for (i = 0; i < equal_tempered_table_size; i++)
    log_equal_tempered_table[i] =
      log (440.0 * pow (2.0, (gdouble) (i - 48 - 9) / 12.0));
  min_log_hertz_global = log_equal_tempered_table[0];
  max_log_hertz_global = log_equal_tempered_table[i - 1];
  return log_equal_tempered_table;
}

static void gdam_sequencer_skin_init(GdamSequencerSkin* sequencer_skin)
{
  sequencer_skin->default_resample = 1.0;
  sequencer_skin->default_gain = 1.0;
  sequencer_skin->circle_radius = 10;
  sequencer_skin->selected_elements = g_hash_table_new (NULL, NULL);
  sequencer_skin->selection_mode = SELECTION_MODE_SET;

  sequencer_skin->hidden_buffers = g_hash_table_new (NULL, NULL);

  sequencer_skin->log_resample_grid = log_equal_tempered_scale ();
  sequencer_skin->num_resample_grid = equal_tempered_table_size;
  sequencer_skin->gain_grid = NULL;
  sequencer_skin->id_grid = NULL;
  sequencer_skin->num_gain_grid = 0;
  sequencer_skin->num_id_grid = 0;
  sequencer_skin->max_log_hertz = log (8000);
  sequencer_skin->min_log_hertz = log (110);
  sequencer_skin->local_id_labels = g_array_new (0, TRUE, sizeof (gpointer));
  sequencer_skin->visible_size_from_length = TRUE;
  sequencer_skin->element_render_mode = RENDER_AS_COLOR_SQUARES;
  sequencer_skin->graph_display_mode = BY_LOCAL_ID;
  sequencer_skin->snap_y = TRUE;
}

GtkType gdam_sequencer_skin_get_type()
{
  static GtkType sequencer_skin_type = 0;
  if (!sequencer_skin_type)
    {
      static const GtkTypeInfo sequencer_skin_info = {
	"GdamSequencerSkin",
	sizeof (GdamSequencerSkin),
	sizeof (GdamSequencerSkinClass),
	(GtkClassInitFunc) gdam_sequencer_skin_class_init,
	(GtkObjectInitFunc) gdam_sequencer_skin_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL
      };
      GtkType parent = GDAM_TYPE_SKIN;
      parent_class = gtk_type_class (parent);
      sequencer_skin_type = gtk_type_unique (parent, &sequencer_skin_info);
      gdam_plugin_check_version (__FILE__, GDAM_VERSION);
    }
  return sequencer_skin_type;
}

static GHashTable *get_info_hash (GdamSequencerSkin *skin)
{
  if (skin->buffer_info_hash == NULL)
    skin->buffer_info_hash = g_hash_table_new (NULL, NULL);
  return skin->buffer_info_hash;
}

static GdamBufferInfo* get_buffer_info_from_id(GdamSequencerSkin* skin,  
                                               int id)
{
  GdamBufferInfo *info;
  info = (GdamBufferInfo *) g_hash_table_lookup (get_info_hash (skin),
						 GINT_TO_POINTER (id));
  if (info != NULL)
    {
      info->ref_count++;
      return info;
    }

  info = g_new0 (GdamBufferInfo, 1);
  info->buffer_id = id;
  info->buffer_name = g_strdup_printf ("%d", id);
  /* XXX: we are very sloppy with the ref_count is not managed well */
  info->ref_count = 1;
  info->render_offset_x = info->render_offset_y = -10000;
  info->hertz = 440.0;

  g_hash_table_insert (get_info_hash (skin), GINT_TO_POINTER (id), info);
  return info;
}

void                   deref_buffer           (GdamSequencerSkin *skin,
                                               int 		  id)
{
  GdamBufferInfo *info;
  info =  get_buffer_info_from_id (skin, id);
  if (info->ref_count == 1)
    return;
  if (--(info->ref_count) == 0)
    {
      g_hash_table_remove (get_info_hash (skin), GINT_TO_POINTER (id));
      if (info->buffer_name)
	g_free (info->buffer_name);
      g_free (info);
    }
}

/* Convenience functions for handling local ids */
static gboolean local_id_valid (GdamSequencer *seq, int buffer_id)
{
  guint id = buffer_id;
  g_return_val_if_fail (seq != NULL, FALSE);
  g_return_val_if_fail (seq->local_infos != NULL, FALSE);
  return (id < seq->local_infos->len); 
}

static GdamLocalSampleInfo *
nth_local_info (GdamSequencer *seq, int local_id)
{
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, local_id))
    return NULL;
  linfo = &g_array_index (seq->local_infos, 
                          GdamLocalSampleInfo, 
			  local_id);
  return linfo;
}
static int nth_local_id (GdamSequencer *seq, int local_id)
{
  GdamLocalSampleInfo *linfo = nth_local_info (seq, local_id);
  if (!linfo)
    return 0;
  return linfo->buffer_id;
}
static void set_nth_local_id (GdamSequencer *seq, int index, int id)
{
  GdamLocalSampleInfo *linfo = nth_local_info (seq, index);
  if (!linfo)
    return;
  linfo->buffer_id = id;
}
static void extend_local_ids (GdamSequencer *seq, guint length)
{
  if (length > seq->local_infos->len)
    gdam_sequencer_set_num_local_infos (seq, length);
}
static inline int 
get_element_id (GdamSequencer *sequencer, GdamSequencerElement *element)
{
  if (element->local_id >= 0)
      return nth_local_id (sequencer, element->local_id);
  return element->buffer_id;
}

/*
 *  ____                     _                  _
 * |  _ \ _ __ __ ___      _(_)_ __   __ _     / \   _ __ ___  __ _
 * | | | | '__/ _` \ \ /\ / / | '_ \ / _` |   / _ \ | '__/ _ \/ _` |
 * | |_| | | | (_| |\ V  V /| | | | | (_| |  / ___ \| | |  __/ (_| |
 * |____/|_|  \__,_| \_/\_/ |_|_| |_|\__, | /_/   \_\_|  \___|\__,_|
 *                                   |___/
 */

/* color square rgb table (from visibone.com/colorlab/) 
 *
 */

int num_color_blobs;
int blob_color_array[] = {153,102,153,
			  204,102,204,
			  255,102,255,
			  102,102,153,
			  102,102,204,
			  102,102,255,
			  102,153,153,
			  102,204,204,
			  102,255,255,
			  102,153,102,
			  102,204,102,
			  102,255,102,
			  153,153,102,
			  204,204,102,
			  255,255,102,
			  153,102,102,
			  204,102,102,
			  255,102,102};

#define MIN_INTENSITY 26216
#define MAX_INTENSITY 32768
#define MAX_COLOR 65536
static void alloc_colors(GdamSequencerSkin* skin)
{
  int state;
  int i;
  int g = 0;
  int r = (MAX_COLOR - MIN_INTENSITY) / 4 + MIN_INTENSITY;
  int b = 3 * (MAX_COLOR - MIN_INTENSITY) / 4 + MIN_INTENSITY;
  int delta = MAX_COLOR / GDAM_SEQUENCER_SKIN_NUM_BLOB_COLORS;
  if (skin->backing_pixmap == NULL)
    return;
  if (skin->drawing_area == NULL)
    return;
  state = GTK_WIDGET_STATE (skin->drawing_area);
  skin->circle_gc = skin->drawing_area->style->bg_gc[state];
  skin->background = skin->drawing_area->style->fg_gc[state];
  skin->text_gc = skin->background;
  skin->font = skin->drawing_area->style->font;
  //skin->font = gdk_font_load ("fixed");
  skin->grid_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->grid_gc,
			 util_alloc_color_cached (16384, 16384, 16384));
  skin->selection_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->selection_gc,
			 util_alloc_color_cached (0, 65535, 0));
  skin->circle_boundary_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->circle_boundary_gc,
			 util_alloc_color_cached (65535, 0, 0));
  skin->gridtext_gc = gdk_gc_new (skin->backing_pixmap);
  gdk_gc_set_foreground (skin->gridtext_gc,
			 util_alloc_color_cached (65535, 65535, 65535));

  for (i = 0; i < GDAM_SEQUENCER_SKIN_NUM_BLOB_COLORS; i++)
    {
      r = blob_color_array [3 * i + 0] * 65536 / 255;
      g = blob_color_array [3 * i + 1] * 65536 / 255;
      b = blob_color_array [3 * i + 2] * 65536 / 255;
      skin->blob_colors[i] = gdk_gc_new (skin->backing_pixmap);
      gdk_gc_set_foreground (skin->blob_colors[i],
			     util_alloc_color_cached (r, g, b));
    }
}

typedef struct _RenderInfo RenderInfo;
struct _RenderInfo {
	int x,y,w,h;
	GdamSequencerSkin* skin;
};

static gdouble get_sample_hertz(GdamSequencerSkin* skin,
                                GdamSequencerElement* element)
{
  GdamBufferInfo *info;
  int id = element->buffer_id;
  info = g_hash_table_lookup (get_info_hash (skin), GINT_TO_POINTER (id));
  if (info == NULL)
    return 440.0;
  return info->hertz;
}

static inline GdamSequencer* get_sequencer(GdamSequencerSkin* skin)
{
  return GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
}

static GdamLocalSampleInfo *
maybe_get_local_info (GdamSequencer        *sequencer, 
                      GdamSequencerElement *element)
{
  GdamLocalSampleInfo *linfo= NULL;
  int num_infos = sequencer->local_infos->len;

  if (element->local_id >= 0)
    if (element->local_id < num_infos)
      linfo = &g_array_index (sequencer->local_infos, 
			      GdamLocalSampleInfo,
			      element->local_id);
  return linfo;
}

static int
get_element_possible_length (GdamSequencer        *sequencer, 
                             GdamSequencerElement *subsample_info)
{
  GdamLocalSampleInfo *linfo;
  int possible_length = -1;
  linfo = maybe_get_local_info (sequencer, subsample_info);

  /* If the local info struct limits our length... */
  if (linfo && linfo->output_length >= 0)
    {
      possible_length = linfo->output_length - subsample_info->sample_offset;
      if (possible_length < 0)
	{
	  g_warning ("bad possible length %d from li->len %d el->offset %d",
		     possible_length, linfo->output_length, 
		     subsample_info->sample_offset);
	  possible_length = 0;
	}
    }

  /* negative number means play full length */
  if (possible_length < 0) 
    {
      int id = get_element_id (sequencer, subsample_info);
      GdamBufferManagerInfo *info = gdam_buffer_manager_info_by_id (id);
      if (info == NULL)
	{
	  g_warning ("sequencer_skin: manager has no info "
		     "on buffer %d (element_possible_length lid %d id %d)",
		     id, subsample_info->local_id, subsample_info->buffer_id);
	  possible_length = 2205;
	}
      else
	possible_length = info->length;
    }
  if (linfo && linfo->rate != 0)
    possible_length = (float)possible_length / linfo->rate;
  return possible_length;
}

/* TODO: make this length work a function of sequencer ?? */
/* Returns output length, in output samples.  Never returns -1 */
/* XXX: scale by local rate (check how callers use rv) */
static int
get_element_length (GdamSequencer        *sequencer, 
                    GdamSequencerElement *subsample_info)
{
  GdamLocalSampleInfo *linfo= NULL;
  int length_samples = subsample_info->output_length;
  int id = get_element_id (sequencer, subsample_info);
  int possible_length;

  linfo = maybe_get_local_info (sequencer, subsample_info);
  possible_length = get_element_possible_length (sequencer, subsample_info);

  /* If the element requests a certain length, and it is not possible,
   * then trim the length */
  if (subsample_info->output_length >= 0
      && possible_length < subsample_info->output_length)
    length_samples = possible_length;
  /* If the element requests full length, give it */
  else if (subsample_info->output_length < 0)
    length_samples = possible_length;
  /* If the element requests a possible length, honor it */
  else
    length_samples = subsample_info->output_length;

#if 0
  /* If the local info struct limits our length... */
  if (linfo && linfo->output_length >= 0)
    {
      /* If the element requests a certain length, and it is not possible,
       * then trim the length */
      if (subsample_info->output_length >= 0
	  && possible_length < subsample_info->output_length)
	length_samples = possible_length;
      /* If the element requests full length, give it */
      else if (subsample_info->output_length < 0)
	length_samples = possible_length;
      /* If the element requests a possible length, honor it */
      else
	length_samples = subsample_info->output_length;
    }
#endif

  /* negative number means play full length */
  if (length_samples < 0) 
    {
      GdamBufferManagerInfo *info = gdam_buffer_manager_info_by_id (id);
      if (info == NULL)
	{
	  g_warning ("sequencer_skin: manager has no info "
		     "on buffer %d (get element length local id %d)",
		     id, subsample_info->local_id);
	  length_samples = 2205;
	}
      else
	length_samples = info->length;
    }
  return length_samples;
}

static int get_y_position(GdamSequencerSkin*    skin,
                          GdamSequencerElement* element)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  switch (skin->graph_display_mode)
    {
    case BY_GAIN:
    {
      return skin->draw_height - element->gain * skin->draw_height / MAX_GAIN;
    }
    case BY_RESAMPLE:
      {
	gdouble log_hertz;
	gdouble sample_hertz = get_sample_hertz (skin, element);
	log_hertz = log (element->rate * sample_hertz);
	return skin->draw_height
	  - (log_hertz - skin->min_log_hertz)
	  * skin->draw_height / (skin->max_log_hertz - skin->min_log_hertz);
      }
    case BY_ID:
      {
	int index;
	float perc;
	int id = get_element_id (get_sequencer (skin), element);
	g_return_val_if_fail (skin->buffers != NULL, 0);
	index = g_list_index (skin->buffers, GINT_TO_POINTER (id));
	perc = ((float)index + 0.5) / g_list_length (skin->buffers);
	return skin->draw_height * perc;
      }
    case BY_LOCAL_ID:
      {
	int index;
	double num = get_sequencer (skin)->local_infos->len;
	float perc;
	int my_local_id = element->local_id;
	if (my_local_id < 0)
	  my_local_id = 0;
	if (num == 0)
	  num = 1;
	perc = ((float)my_local_id + 0.5) / num;
	return skin->draw_height * perc;
      }
    case BY_PATTERN:
      {
	int index;
	double num = skin->num_patterns;
	float perc;
	perc = ((float)element->pattern_number + 0.5) / num;
	return skin->draw_height * perc;
      }
    case BY_LENGTH:
      {
	int possible_length;
	float perc = 1.0;
	possible_length = get_element_possible_length (sequencer, element);
	perc = (float) get_element_length (sequencer, element) 
	       / (float) possible_length;
	if (perc < 0 || perc > 1.0)
	  g_warning ("bad perc from possible_length %d el_len %d", 
	             possible_length, get_element_length (sequencer, element));
	return skin->draw_height - skin->draw_height * perc;
      }
    case BY_START:
      {
	int possible_length;
	float perc = 1.0;
	possible_length = get_element_possible_length (sequencer, element);
	perc = element->sample_offset / (float) possible_length;
	if (perc < 0 || perc > 1.0)
	  g_warning ("bad perc from possible_length %d offset %d", 
	             possible_length, element->sample_offset);
	return skin->draw_height - skin->draw_height * perc;
      }
    default:
      g_warning ("unhandled sort method in get_y_position");
    }
  return 0;
}

static void 
set_y_position (GdamSequencerSkin    *skin,
		GdamSequencerElement *element,
		int                   y)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  switch (skin->graph_display_mode)
    {
    case BY_GAIN:
      {
	gdouble gain;
	y = skin->draw_height - y;
	gain = y * MAX_GAIN / skin->draw_height;
	element->gain = gain;
	return;
      }
    case BY_RESAMPLE:
      {
	gdouble log_hertz;
	y = skin->draw_height - y;
	log_hertz = (skin->max_log_hertz - skin->min_log_hertz)
	  * y / skin->draw_height + skin->min_log_hertz;
	element->rate = exp (log_hertz) / get_sample_hertz (skin, element);
	return;
      }
    case BY_ID:
      {
	g_warning ("sort_by_id incomplete in set_y_position");
	return;
	//int index;
	//float perc;
	//int id = get_element_id (get_sequencer (skin), element);
	//g_return_val_if_fail (skin->buffers != NULL, 0);
	//index = g_list_index (skin->buffers, GINT_TO_POINTER (id));
	//perc = ((float)index + 0.5) / g_list_length (skin->buffers);
	//return skin->draw_height * perc;
      }
    case BY_LOCAL_ID:
      {
	double num = get_sequencer (skin)->local_infos->len;
	double perc = (float)y / skin->draw_height;
	/* TODO: round to nearest? */
	element->local_id = (int)(perc * num);
	return;
      }
    case BY_PATTERN:
      {
	double num = skin->num_patterns;
	double perc = (float)y / skin->draw_height;
	/* TODO: round to nearest? */
	element->pattern_number = (int)(perc * num);
	return;
      }
    case BY_LENGTH:
      {
	int possible_length;
	double perc;
	y = skin->draw_height - y;
	perc = (float)y / skin->draw_height;
	possible_length = get_element_possible_length (sequencer, element);
	element->output_length = (int)(perc * possible_length);
	return;
      }
    case BY_START:
      {
	int possible_length;
	double perc;
	y = skin->draw_height - y;
	perc = (float)y / skin->draw_height;
	possible_length = get_element_possible_length (sequencer, element);
	element->sample_offset = (int)(perc * possible_length);
	return;
      }
    default:
      g_warning ("unhandled sort method in set_y_position");
    }
  return;
}

static void get_subsample_position(GdamSequencerElement* element, 
                                   GdamSequencerSkin*    skin,
                                   int*                  out_x,
				   int*                  out_y)
{
  gdouble num_beats = get_sequencer (skin)->num_beats;
  if (num_beats <= 0)
    *out_x = 0;
  else
    {
      gdouble where = element->beat_offset - skin->min_beat;
      where /= (skin->max_beat - skin->min_beat);
      *out_x = skin->draw_width * where;
    }
  *out_y = get_y_position (skin, element);
}

static void 
get_string_offset(GdkFont* font, const char* string,
                  int* offset_x, int* offset_y)
{
  int lbearing, rbearing, width, ascent, descent;
  gdk_string_extents (font, string,
		      &lbearing, &rbearing, &width, &ascent, &descent);
  *offset_x = (lbearing + rbearing) / 2;
  *offset_y = (descent - ascent) / 2;
}


/*  Renders one sample in the sequence as a rectangle whose length corresponds
 *  to the length of the sample.  */
static void 
render_subsample_color_square (GdamSequencerElement *subsample_info,
                               RenderInfo           *render_info)
{
  GdkPixmap *pixmap = render_info->skin->backing_pixmap;
  int x, y, radius, id;
  int length_pixels;
  int color_index;
  GdamBufferInfo *buffer_info;
  GdamSequencerSkin *skin = render_info->skin;
  GdamSequencer *sequencer = get_sequencer (skin);
  GdkGC *gc_inner;
  GdkGC *gc_outer;
  gdouble pixels_per_second = 0;
  char *element_text;

  get_subsample_position (subsample_info, skin, &x, &y);
  radius = skin->circle_radius;
  id = get_element_id (sequencer, subsample_info);

    y -= 1 * radius;


  if (skin->visible_size_from_length) 
    {
      gdouble num_beats = get_sequencer (skin)->num_beats;
      int length_samples = get_element_length (sequencer, subsample_info);

      if (num_beats <= 0)
	g_warning ("invalid num_beats");
      else
	{
	  gdouble bpm = 
	    gdam_beat_info_get_bpm 
	      (&GDAM_MODEL (get_sequencer (skin))->beat_info, 0);
	  gdouble beats_visible = skin->max_beat - skin->min_beat;
	  gdouble seconds_visible = beats_visible * 60.0 / bpm;
	  pixels_per_second = skin->draw_width / seconds_visible;
	}
      length_pixels = (int)(((float)length_samples / 44100.0 / subsample_info->rate) * pixels_per_second);
    }
  else
    length_pixels = 2 * radius;


  /*
   * Does the box around x/y touch the exposed area?
   */
  if (x > render_info->x + render_info->w
      || x + length_pixels < render_info->x
      || y - radius > render_info->y + render_info->h
      || y + radius < render_info->y)
    return;


  /* Determine the coloring */
  color_index = subsample_info->local_id % GDAM_SEQUENCER_SKIN_NUM_BLOB_COLORS;
  if (skin->element_render_mode == RENDER_AS_RECTANGLES)
    {
      gc_inner = skin->circle_gc;
      gc_outer = skin->circle_boundary_gc;

      /* Invert the colors of selected elements */
      if (g_hash_table_lookup (skin->selected_elements, subsample_info))
	{
	  GdkGC *tmp = gc_inner;
	  gc_inner = gc_outer;
	  gc_outer = tmp;
	}
    }
  else
    {
      gc_inner = skin->blob_colors[color_index];
      gc_outer = gc_inner;

      /* Give the element a highlighted outline if it is selected. */
      if (g_hash_table_lookup (skin->selected_elements, subsample_info))
       gc_outer = skin->circle_boundary_gc;

#if 0
      if (!gdam_sequencer_pattern_get_active (sequencer, 
					      subsample_info->pattern_number))
	  gc_outer = gc_inner;
#endif
    }


  /* Draw the outline or handle */
  if (skin->element_render_mode == RENDER_AS_COLOR_SQUARES)
    gdk_draw_rectangle (pixmap,
			gc_outer, FALSE, x, y, length_pixels, 2 * radius);
  else if (skin->element_render_mode == RENDER_AS_RECTANGLES)
    gdk_draw_rectangle (pixmap,
			gc_outer, FALSE, x, y, length_pixels, 2 * radius);
  else if (skin->element_render_mode == RENDER_AS_TEE)
    gdk_draw_rectangle (pixmap, gc_outer, FALSE, x, y, 1, 2 * radius);
  else if (skin->element_render_mode == RENDER_AS_PENNANT)
    {
      GdkPoint points[3];
      points[0].x = x;
      points[0].y = y;
      points[1].x = x;
      points[1].y = y + 2 * radius;
      points[2].x = x + length_pixels;
      points[2].y = y + radius;
      gdk_draw_polygon (pixmap, gc_outer, FALSE, points, 3);
    }
  else
    g_warning ("unhandled render mode %d ??", skin->element_render_mode);

  /*  If the sample's pattern is not active (and  therefore insensitive)
   *  don't fully render it.  */
  if (!gdam_sequencer_pattern_get_active (sequencer, 
					  subsample_info->pattern_number))
    return;

  /*  Render the rest of the element */
  if (skin->element_render_mode == RENDER_AS_COLOR_SQUARES)
    gdk_draw_rectangle (pixmap, gc_inner, TRUE, 
			x + 1, y + 1, length_pixels - 1, (2 * radius) - 1);
  else if (skin->element_render_mode == RENDER_AS_RECTANGLES)
    gdk_draw_rectangle (pixmap, gc_inner, TRUE, 
			x + 1, y + 1, length_pixels - 2, (2 * radius) - 2);
  else if (skin->element_render_mode == RENDER_AS_TEE)
    gdk_draw_rectangle (pixmap, gc_outer, FALSE, 
			x, y + radius, length_pixels, 1);
  else if (skin->element_render_mode == RENDER_AS_PENNANT)
    {
      GdkPoint points[3];
      points[0].x = x;
      points[0].y = y;
      points[1].x = x;
      points[1].y = y + 2 * radius;
      points[2].x = x + length_pixels;
      points[2].y = y + radius;
      gdk_draw_polygon (pixmap, gc_outer, TRUE, points, 3);
    }
  else
    g_warning ("unhandled render mode %d", skin->element_render_mode);

  id = get_element_id (sequencer, subsample_info);
  buffer_info = get_buffer_info_from_id (skin, id);

  element_text = buffer_info->buffer_name;
  if (skin->element_text_mode == TEXT_BUFFER_NAME)
    element_text = buffer_info->buffer_name;
  else if (skin->element_text_mode == TEXT_LOCAL_ID)
    element_text = g_strdup_printf ("%d", subsample_info->local_id);

  /*  Print the element's ID, name, or local id as letter,  at the left edge 
   *  of the rectangle */
  if (buffer_info->render_offset_x < -1000)
    {
      get_string_offset (skin->drawing_area->style->font,
			 buffer_info->buffer_name,
			 &buffer_info->render_offset_x,
			 &buffer_info->render_offset_y);
    }
  if (skin->element_render_mode == RENDER_AS_COLOR_SQUARES)
    gdk_draw_string (pixmap,
		     skin->font,
		     skin->text_gc,
		     x + 1,
		     y + radius - buffer_info->render_offset_y,
		     element_text);
  else if (skin->element_render_mode == RENDER_AS_RECTANGLES ||
           skin->element_render_mode == RENDER_AS_PENNANT)
    gdk_draw_string (pixmap,
		     skin->font,
		     skin->text_gc,
		     x,
		     y + radius - buffer_info->render_offset_y,
		     element_text);
  else if (skin->element_render_mode == RENDER_AS_TEE)
    gdk_draw_string (pixmap,
		     skin->font,
		     gc_outer,
		     x + 3,
		     y + radius - buffer_info->render_offset_y - 4,
		     element_text);

  if (skin->element_text_mode == TEXT_LOCAL_ID)
    g_free (element_text);
}


/*  Renders one sample in the sequence, based on its state.  */
static void 
render_subsample_circle (GdamSequencerElement *subsample_info,
                         RenderInfo           *render_info)
{
  GdkPixmap *pixmap = render_info->skin->backing_pixmap;
  int x, y, radius, id;
  GdamBufferInfo *buffer_info;
  GdamSequencerSkin *skin = render_info->skin;
  GdamSequencer *sequencer = get_sequencer (skin);
  GdkGC *gc_inner;
  GdkGC *gc_outer;
  get_subsample_position (subsample_info, skin, &x, &y);
  radius = skin->circle_radius;

  /*
   * Does the box around x/y touch the exposed area?
   */
  if (x - radius > render_info->x + render_info->w
      || x + radius < render_info->x
      || y - radius > render_info->y + render_info->h
      || y + radius < render_info->y)
    return;

  gc_inner = skin->circle_gc;
  gc_outer = skin->circle_boundary_gc;

  if (g_hash_table_lookup (skin->selected_elements, subsample_info))
    {
      GdkGC *tmp = gc_inner;
      gc_inner = gc_outer;
      gc_outer = tmp;
    }

  gdk_draw_arc (pixmap,
		gc_outer,
		FALSE,
		x - radius + 1,
		y - radius + 1, radius * 2 - 1, radius * 2 - 1, 0, 360 * 64);
  /*  If the sample's pattern is not active (and  therefore insensitive)
   *  don't fully render it.  */
  if (!gdam_sequencer_pattern_get_active (sequencer, 
					  subsample_info->pattern_number))
    return;
  gdk_draw_arc (pixmap,
		gc_inner,
		TRUE,
		x - radius + 1,
		y - radius + 1, radius * 2 - 1, radius * 2 - 1, 0, 360 * 64);

  id = get_element_id (sequencer, subsample_info);
  buffer_info = get_buffer_info_from_id (skin, id);

  if (buffer_info->render_offset_x < -1000)
    {
      get_string_offset (skin->drawing_area->style->font,
			 buffer_info->buffer_name,
			 &buffer_info->render_offset_x,
			 &buffer_info->render_offset_y);
    }

  gdk_draw_string (pixmap,
		   skin->font,
		   skin->text_gc,
		   x - buffer_info->render_offset_x,
		   y - buffer_info->render_offset_y,
		   buffer_info->buffer_name);
}

static void 
render_subsample (GdamSequencerElement *subsample_info,
                  RenderInfo           *render_info)
{
  GdamSequencerSkin *skin = render_info->skin;
  if (skin->in_transit && g_hash_table_lookup (skin->in_transit,
					       subsample_info))
    return;

  if (g_hash_table_lookup (skin->hidden_buffers,
			   GUINT_TO_POINTER (subsample_info->buffer_id)))
    return;

  if (skin->hide_inactive &&
      !gdam_sequencer_pattern_get_active (get_sequencer (skin),
					  subsample_info->pattern_number))
    return;

  if (skin->element_render_mode == RENDER_AS_CIRCLES)
    render_subsample_circle (subsample_info, render_info);
  else if (skin->element_render_mode == RENDER_AS_COLOR_SQUARES ||
           skin->element_render_mode == RENDER_AS_RECTANGLES ||
           skin->element_render_mode == RENDER_AS_PENNANT ||
           skin->element_render_mode == RENDER_AS_TEE)
    render_subsample_color_square (subsample_info, render_info);
  else
    g_warning ("unknown render mode in render_subsample");
}

static inline void
copy_backing_area(GdamSequencerSkin* skin, int x, int y, int w, int h)
{
  GtkWidget *widget = skin->drawing_area;
  GdkGC *gc = widget->style->fg_gc[GTK_WIDGET_STATE (skin->drawing_area)];

  gdk_draw_pixmap (widget->window, gc,
		   skin->backing_pixmap, x, y, x, y, w, h);
}

static void redraw_area(GdamSequencerSkin* skin, int x, int y, int w, int h)
{
  RenderInfo render_info;
  render_info.x = x;
  render_info.y = y;
  render_info.w = w;
  render_info.h = h;
  render_info.skin = skin;
  if (!skin->circle_gc)
    alloc_colors (skin);

  if (skin->backing_pixmap == NULL)
    return;

  /* Fill with background */
  gdk_draw_rectangle (skin->backing_pixmap,
		      skin->background, TRUE, x, y, w, h);

  /* draw horizontal lines */
  if (skin->grid_gc)
    {
      int num_lines;
      gdouble *grid;
      int i;
      gdouble top, bottom;
      if (skin->graph_display_mode == BY_RESAMPLE)
	{
	  grid = skin->log_resample_grid;
	  num_lines = skin->num_resample_grid;
	  top = skin->max_log_hertz;
	  bottom = skin->min_log_hertz;
	}
      else if (skin->graph_display_mode == BY_GAIN)
	{
	  grid = skin->gain_grid;
	  num_lines = skin->num_gain_grid;
	  top = 2.0;
	  bottom = 0.0;
	}
      else
	{
	  grid = skin->id_grid;
	  num_lines = skin->num_id_grid;
	  top = 1.0;
	  bottom = 0.0;
	}
      for (i = 0; i < num_lines; i++)
	{
	  gdouble t = grid[i];
	  if (bottom <= t && t <= top)
	    {
	      int pos = (int) ((t - bottom) * skin->draw_height
			       / (top - bottom));
	      /* reverse y to pixel-style coords */
	      pos = skin->draw_height - 1 - pos;
	      gdk_draw_line (skin->backing_pixmap,
			     skin->grid_gc, x, pos, x + w, pos);
	    }
	}
    }

  /* Draw vertical lines */
  if (skin->grid_gc)
    {
      int num_lines = skin->num_beat_grid;
      int skip = get_sequencer (skin)->num_sub_beats;
      int i;
      int beatcount = 0;
      int interline_width;
      int pos1 = 0;
      int pos2;
      int text_skip = 1;
      gboolean have_pos1 = FALSE;
      gboolean have_pos2 = FALSE;
      if (skip == 0)
	skip = 1;
      
      gdam_debug ("Drawing every %d of %d beatlines", skip, num_lines);
      for (i = 0; i < num_lines; i += skip, beatcount++)
	{
	  gdouble t = skin->beat_grid[i];
	  t -= skin->min_beat;
	  t /= (skin->max_beat - skin->min_beat);
	  if (0 <= t && t <= 1.0)
	    {
	      int pos = t * skin->draw_width;
	      char str[8];
	      int text_width;
	      int text_height;
	      if (!have_pos1)
		{
		  pos1 = pos;
		  have_pos1 = TRUE;
		}
	      else if (!have_pos2)
		{
		  int twidth;
		  pos2 = pos;
		  have_pos2 = TRUE;
		  interline_width = pos2 - pos1;
		  twidth = gdk_string_width (skin->font, "0000");
		  while (twidth / text_skip > interline_width)
		    text_skip *= 2;
		}
	      gdk_draw_line (skin->backing_pixmap,
			     skin->grid_gc, pos, y, pos, y + h);
	      if (beatcount % text_skip == 0)
		{
		  sprintf (str, "%d", beatcount + 1);
		  text_height = gdk_string_height (skin->font, str);
		  text_width = gdk_string_width (skin->font, str);
		  gdk_draw_rectangle (skin->backing_pixmap,
				      skin->background, TRUE, pos+1, 
				      skin->draw_height - text_height, 
				      text_width, text_height);

		  gdk_draw_string (skin->backing_pixmap,
				   skin->font,
				   skin->gridtext_gc,
				   pos+1, 
				   skin->draw_height, 
				   str);
		}
	    }
	}
    }

  recreate_buffer_list (skin);

  /* Draw notes */
  g_list_foreach (get_sequencer (skin)->elements,
		  (GFunc) render_subsample, &render_info);
  copy_backing_area (skin, x, y, w, h);
}

static void 
is_max_pattern (GdamSequencerElement *subsample_info,
                GdamSequencerSkin    *skin)
{
  if (subsample_info->pattern_number + 1 > skin->num_patterns)
    skin->num_patterns = subsample_info->pattern_number + 1;
}

static void 
update_num_patterns (GdamSequencerSkin *skin)
{
  skin->num_patterns = 0;
  g_list_foreach (get_sequencer (skin)->elements,
		  (GFunc) is_max_pattern, skin);
		  //(GFunc) render_subsample, &render_info);
}

#define DIST(a,b)  ( ((a) < (b)) ? ((b) - (a)) : ((a) - (b)) )

/* This function should use a binary search. */
static void match_line(gdouble* adjust, int num, const gdouble* vals)
{
  int closest;
  gdouble closest_dist;
  gdouble saught = *adjust;
  int i;
  if (num <= 0)
    return;
  g_return_if_fail (vals != NULL);

  closest = 0;
  closest_dist = DIST (saught, vals[0]);
  for (i = 1; i < num; i++)
    {
      gdouble dist = DIST (saught, vals[i]);
      if (dist < closest_dist)
	{
	  closest = i;
	  closest_dist = dist;
	}
    }
  *adjust = vals[closest];
}

static void
snap_to_grid(GdamSequencerSkin*      skin,
             int                     x,
	     int                     y,
	     GdamSequencerElement*   element)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  gdouble start_beat = (gdouble) x * (skin->max_beat - skin->min_beat)
    / skin->draw_width + skin->min_beat;
  gdouble temp = start_beat;
  if (sequencer->num_sub_beats > 0)
    match_line (&temp, skin->num_beat_grid, skin->beat_grid);
  gdam_debug ("snap_to_grid: snapped beat from %f to %f", start_beat, temp);
  element->beat_offset = temp;
  gdam_sequencer_update_sequence (sequencer);

  if (!skin->snap_y)
    return;

  /* For the purposes of the following computations
   * and anything involving resample/gain we invert
   * y, so it corresponds to sheet music, math, etc,
   * instead of computers. */
  y = skin->draw_height - y;
  switch (skin->graph_display_mode)
    {
    case BY_GAIN:
      {
	gdouble gain;
	gain = y * MAX_GAIN / skin->draw_height;
	gdam_debug ("snap_to(GAIN): y=%d, draw_height=%d; "
		    "gain=%g", y, skin->draw_height, gain);
	match_line (&gain, skin->num_gain_grid, skin->gain_grid);
	element->gain = gain;
	break;
      }
    case BY_RESAMPLE:
      {
	gdouble log_hertz;
	log_hertz = (skin->max_log_hertz - skin->min_log_hertz)
	  * y / skin->draw_height + skin->min_log_hertz;
	match_line (&log_hertz,
		    skin->num_resample_grid, skin->log_resample_grid);
	element->rate = exp (log_hertz) / get_sample_hertz (skin, element);
	break;
      }
    case BY_ID:
      {
	break;
      }
    case BY_LOCAL_ID:
      {
	break;
      }
    case BY_PATTERN:
      {
	break;
      }
    case BY_LENGTH:
      {
	break;
      }
    case BY_START:
      {
	break;
      }
    default:
      {
	g_warning ("unknown graph display mode");
	g_assert_not_reached ();
      }
    }
}

static gint
sequence_edit_expose(GtkDrawingArea* area,
                     GdkEventExpose* expose_event,
		     GdamSequencerSkin* skin)
{
  redraw_area (skin,
	       expose_event->area.x, expose_event->area.y,
	       expose_event->area.width, expose_event->area.height);
  return 0;
}

/*  Returns a list of all samples whose position on screen places them
 *  within range of a click at x,y.  This function also ignores elements
 *  which aren't part of the current pattern.  TODO: make pattern optional?? */
GList* find_submodel_list_el(GdamSequencerSkin* skin, int x, int y)
{
  GList *at;
  GList *found = NULL;
  GdamSequencer *sequencer = get_sequencer (skin);
  for (at = get_sequencer (skin)->elements; at; at = at->next)
    {
      int pos_x, pos_y;
      GdamSequencerElement *subsample;
      gboolean is_selected;

      is_selected = g_hash_table_lookup (skin->selected_elements,
					 at->data) != NULL;
      if (found != NULL && !is_selected)
	continue;

      subsample = (GdamSequencerElement *) at->data;
      get_subsample_position (subsample, skin, &pos_x, &pos_y);
      if (pos_x - skin->circle_radius <= x
	  && x <= pos_x + skin->circle_radius
	  && pos_y - skin->circle_radius <= y
	  && y <= pos_y + skin->circle_radius
	  && gdam_sequencer_pattern_get_active (sequencer, 
	      					subsample->pattern_number))
	{
	  if (is_selected)
	    return at;
	  found = at;
	}
    }
  return found;
}
GdamSequencerElement* find_submodel(GdamSequencerSkin* skin, int x, int y)
{
  GList *out = find_submodel_list_el (skin, x, y);
  return (GdamSequencerElement *) (out ? out->data : NULL);
}


static void delete_subinfo(GdamSequencerSkin*    skin,
                           GdamSequencerElement* element)
{
  int sample_x, sample_y, radius = skin->circle_radius;

  deref_buffer (skin, element->buffer_id);

  g_hash_table_remove (skin->selected_elements, element);

  get_subsample_position (element, skin, &sample_x, &sample_y);
  gdam_sequencer_remove_element (get_sequencer (skin), element);
  redraw_area (skin, sample_x - radius - 1, sample_y - radius - 1,
	       2 * radius + 3, 2 * radius + 3);
}

static void add_element_ptr(gpointer el, gpointer unused, gpointer arr)
{
  GdamSequencerElement **element = *(GdamSequencerElement ***) arr;
  *element++ = (GdamSequencerElement *) el;
  *(GdamSequencerElement ***) arr = element;
}

static void delete_selected(GdamSequencerSkin*    skin)
{
  GdamSequencerElement **elements_to_kill;
  GdamSequencerElement **tmp;
  GdamSequencer *sequencer = get_sequencer (skin);
  int n = g_hash_table_size (skin->selected_elements);
  int i;
  tmp =
    (GdamSequencerElement **) alloca (sizeof (GdamSequencerElement *) * n);
  elements_to_kill = tmp;
  g_hash_table_foreach (skin->selected_elements, add_element_ptr, &tmp);
  g_assert (elements_to_kill + n == tmp);
  gdam_sequencer_freeze (sequencer);
  for (i = 0; i < n; i++)
    delete_subinfo (skin, elements_to_kill[i]);
  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
}


/*
 *  ____       _           _   _
 * / ___|  ___| | ___  ___| |_(_) ___  _ __     __ _ _ __ ___  __ _
 * \___ \ / _ \ |/ _ \/ __| __| |/ _ \| '_ \   / _` | '__/ _ \/ _` |
 *  ___) |  __/ |  __/ (__| |_| | (_) | | | | | (_| | | |  __/ (_| |
 * |____/ \___|_|\___|\___|\__|_|\___/|_| |_|  \__,_|_|  \___|\__,_|
 * 
 */
static void unselect_all_elements(GdamSequencerSkin* skin)
{
  g_hash_table_destroy (skin->selected_elements);
  skin->selected_elements = g_hash_table_new (NULL, NULL);
  delayed_redraw (skin);
}

static void begin_selection(GdamSequencerSkin* skin, int x, int y)
{
  skin->has_selection = TRUE;
  skin->selection[0].x = skin->selection[1].x = x;
  skin->selection[0].y = skin->selection[1].y = y;
}

static void draw_selection(GdamSequencerSkin* skin)
{
  GtkWidget *widget = skin->drawing_area;
  int x1 = skin->selection[0].x;
  int y1 = skin->selection[0].y;
  int x2 = skin->selection[1].x;
  int y2 = skin->selection[1].y;
  gdk_draw_line (widget->window, skin->selection_gc, x1, y1, x1, y2);
  gdk_draw_line (widget->window, skin->selection_gc, x2, y1, x2, y2);
  gdk_draw_line (widget->window, skin->selection_gc, x1, y1, x2, y1);
  gdk_draw_line (widget->window, skin->selection_gc, x1, y2, x2, y2);
}

static void get_selection_rect(GdamSequencerSkin* skin, GdkRectangle* rect)
{
  int width, height;
  width = skin->selection[1].x - skin->selection[0].x;
  height = skin->selection[1].y - skin->selection[0].y;
  if (width < 0)
    {
      rect->width = -width + 1;
      rect->x = skin->selection[1].x;
    }
  else
    {
      rect->width = width + 1;
      rect->x = skin->selection[0].x;
    }
  if (height < 0)
    {
      rect->height = -height + 1;
      rect->y = skin->selection[1].y;
    }
  else
    {
      rect->height = height + 1;
      rect->y = skin->selection[0].y;
    }
}
static void undraw_selection(GdamSequencerSkin* skin)
{
  GdkRectangle rect;
  get_selection_rect (skin, &rect);

  copy_backing_area (skin, rect.x, rect.y, 1, rect.height);
  copy_backing_area (skin, rect.x + rect.width - 1, rect.y, 1, rect.height);
  if (rect.width > 2)
    {
      copy_backing_area (skin, rect.x + 1, rect.y, rect.width - 2, 1);
      copy_backing_area (skin,
			 rect.x + 1, rect.y + rect.height - 1,
			 rect.width - 2, 1);
    }
}

/*  Called when mouse is moved while making selection.  */
static void update_selection(GdamSequencerSkin* skin, int x, int y)
{
  undraw_selection (skin);
  skin->selection[1].x = x;
  skin->selection[1].y = y;
  draw_selection (skin);
}

static gboolean find_union(gpointer key, gpointer value, gpointer h)
{
  g_hash_table_insert ((GHashTable *) h, key, value);
  return TRUE;
}

static void handle_selected(GdamSequencerSkin*    skin,
                            GdamSequencerElement* element)
{
  GHashTable *selected = skin->selected_elements;
  int mode = skin->selection_mode;
  if (mode == SELECTION_MODE_SET || mode == SELECTION_MODE_ADD)
    {
      g_hash_table_insert (selected, element, element);
    }
  else
    {				/* SELECTION_MODE_TOGGLE */
      if (g_hash_table_lookup (selected, element))
	g_hash_table_remove (selected, element);
      else
	g_hash_table_insert (selected, element, element);
    }
}

/*  Called when mouse is released at end of making selection.  */
static void finish_selection(GdamSequencerSkin* skin, int x, int y)
{
  GHashTable *redraw_table = g_hash_table_new (NULL, NULL);
  GdamSequencer *sequencer;
  GList *elements;
  GdkRectangle rect;
  int radius = skin->circle_radius;
  int old_size = g_hash_table_size (skin->selected_elements);
  int new_size;
  sequencer = get_sequencer (skin);
  elements = sequencer->elements;
  undraw_selection (skin);
  skin->has_selection = FALSE;
  if (skin->selection_mode == SELECTION_MODE_SET)
    g_hash_table_foreach_remove (skin->selected_elements,
				 find_union, redraw_table);
  get_selection_rect (skin, &rect);
  /*  For each element, see if it is within selection boundaries
   *  and in the currently selected pattern.  */
  while (elements != NULL)
    {
      GdamSequencerElement *el = elements->data;
      int sample_x, sample_y;
      get_subsample_position (el, skin, &sample_x, &sample_y);
      if (rect.x <= sample_x + radius
	  && sample_x - radius < rect.x + rect.width
	  && rect.y <= sample_y + radius
	  && sample_y - radius < rect.y + rect.height
	  && gdam_sequencer_pattern_get_active (sequencer, el->pattern_number))
	{
	  g_hash_table_insert (redraw_table, el, el);
	  handle_selected (skin, el);
	}
      elements = elements->next;
    }

  /* TODO: incremental redraw using redraw_table here.... */
  //g_hash_table_foreach(redraw_area, redraw_one_func, skin);

  g_hash_table_destroy (redraw_table);
  new_size = g_hash_table_size (skin->selected_elements);

  {
    char buf[1024];
    gdam_skin_pop_message (GDAM_SKIN (skin), skin->selection_status_context);
    snprintf (buf, sizeof (buf), "Had selected %d elements; "
	      "now %d elements are selected.", old_size, new_size);
    buf[sizeof (buf) - 1] = 0;
    gdam_skin_push_message (GDAM_SKIN (skin),
			    skin->selection_status_context, buf);}

  delayed_redraw (skin);
}

/*
 *  ____                        _
 * |  _ \ _ __ __ _  __ _  __ _(_)_ __   __ _
 * | | | | '__/ _` |/ _` |/ _` | | '_ \ / _` |
 * | |_| | | | (_| | (_| | (_| | | | | | (_| |
 * |____/|_|  \__,_|\__, |\__, |_|_| |_|\__, |
 *                  |___/ |___/         |___/
 */
static void
draw_one_dragged_sample(GdamSequencerSkin* skin, GdamSequencerElement* element)
{
  int x, y;
  int radius = skin->circle_radius;
  get_subsample_position (element, skin, &x, &y);
  x += skin->drag_current_x - skin->drag_start_x;
  y += skin->drag_current_y - skin->drag_start_y;
  gdk_draw_arc (skin->drawing_area->window,
		skin->circle_gc,
		FALSE,
		x - radius + 1,
		y - radius + 1, radius * 2 - 1, radius * 2 - 1, 0, 360 * 64);
}
static void draw_dragged_samples(GdamSequencerSkin* skin)
{
  int i;
  for (i = 0; i < skin->num_dragging; i++)
    draw_one_dragged_sample (skin, skin->dragging + i);
}
static void
undraw_one_dragged_sample(GdamSequencerSkin* skin,
                          GdamSequencerElement* element)
{
  int x, y;
  int radius = skin->circle_radius;
  get_subsample_position (element, skin, &x, &y);
  x += skin->drag_current_x - skin->drag_start_x;
  y += skin->drag_current_y - skin->drag_start_y;
  copy_backing_area (skin,
		     x - radius - 1,
		     y - radius - 1, 2 * radius + 2, 2 * radius + 2);
}
static void undraw_dragged_samples(GdamSequencerSkin* skin)
{
  int i;
  for (i = 0; i < skin->num_dragging; i++)
    undraw_one_dragged_sample (skin, skin->dragging + i);
}

static void copy_element_to_array(gpointer key, gpointer val, gpointer udata)
{
  GdamSequencerElement *element = *(GdamSequencerElement **) udata;
  *element++ = *(GdamSequencerElement *) key;
  *(GdamSequencerElement **) udata = element;
}
static void copy_hash_element(gpointer key, gpointer val, gpointer user_data)
{
  g_hash_table_insert ((GHashTable *) user_data, key, val);
}

static void
begin_dragging(GdamSequencerSkin*    skin,
               GdamSequencerElement* sample_info,
	       guint                 mask)
{
  gboolean copy = (mask & GDK_SHIFT_MASK) == GDK_SHIFT_MASK;

  g_free (skin->dragging);
  if (skin->in_transit != NULL)
    {
      g_hash_table_destroy (skin->in_transit);
      skin->in_transit = NULL;
    }

  /* Set up:
   *      skin->in_transit
   *      skin->num_dragging
   *      skin->dragging. */
  if (g_hash_table_lookup (skin->selected_elements, sample_info))
    {
      int num_dragging = g_hash_table_size (skin->selected_elements);
      GdamSequencerElement *tmp;
      skin->num_dragging = num_dragging;
      skin->dragging = g_new (GdamSequencerElement, num_dragging);
      tmp = skin->dragging;
      g_hash_table_foreach (skin->selected_elements,
			    copy_element_to_array, &tmp);
      if (!copy)
	{
	  skin->in_transit = g_hash_table_new (NULL, NULL);
	  g_hash_table_foreach (skin->selected_elements,
				copy_hash_element, skin->in_transit);
	}
      skin->was_dragging_selected = TRUE;
    }
  else
    {
      skin->num_dragging = 1;
      skin->dragging = g_new (GdamSequencerElement, 1);
      *skin->dragging = *sample_info;
      if (!copy)
	{
	  skin->in_transit = g_hash_table_new (NULL, NULL);
	  g_hash_table_insert (skin->in_transit, sample_info, sample_info);
	}
      skin->was_dragging_selected = FALSE;
    }

  /* TODO: should only redraw the in_transit area. */
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
  draw_dragged_samples (skin);
}
static void remove_element(gpointer el, gpointer unused, gpointer seq)
{
  gdam_sequencer_remove_element (GDAM_SEQUENCER (seq),
				 (GdamSequencerElement *) el);
}

/* When 'cut' is triggered, remove all selected elements and put them
 * in the clipboard.  For 'copy' copy into clipboard, but do not remove. */
static void
do_element_cut_or_copy   (GdamSequencerSkin    *skin,
	                  gboolean              copy,
			  gboolean              merge)
{
  int num_elements_in_clipboard;
  /* TODO: option to merge into clipboard? */

  /* copy selected elements to the clipboard */
  num_elements_in_clipboard = g_hash_table_size (skin->selected_elements);
  if (num_elements_in_clipboard > 0)
    {
      GdamSequencerElement *tmp;

      if (merge)
	{
	  num_elements_in_clipboard += skin->num_elements_in_clipboard;
	  if (skin->clipboard)
	    skin->clipboard = g_renew (GdamSequencerElement, skin->clipboard,
		                       num_elements_in_clipboard);
	  else
	    skin->clipboard = g_new (GdamSequencerElement, 
		                     num_elements_in_clipboard);
	  tmp = skin->clipboard + skin->num_elements_in_clipboard;
	}
      else
	{
	  if (skin->clipboard)
	    g_free (skin->clipboard);
	  skin->clipboard = g_new (GdamSequencerElement, num_elements_in_clipboard);
	  tmp = skin->clipboard;
	}

      g_hash_table_foreach (skin->selected_elements,
			    copy_element_to_array, &tmp);
      skin->num_elements_in_clipboard = num_elements_in_clipboard;

      /* If we are cutting, remove selected elements from sequence */
      if (!copy)
	{
          GdamSequencer *sequencer = get_sequencer (skin);
	  gdam_sequencer_freeze (sequencer);
	  g_hash_table_foreach (skin->selected_elements, remove_element, sequencer);
	  g_hash_table_destroy (skin->selected_elements);
	  skin->selected_elements = g_hash_table_new (NULL, NULL);
	  gdam_sequencer_thaw (sequencer);
	}
    }
  else
    {
      /* TODO: make cut and copy inactive if nothing is selected. */
      g_warning ("nothing selected");
    }

  /* TODO: should only redraw the in_transit area. */
  if (!copy)
    redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
  //draw_dragged_samples (skin);
}

static void
do_element_cut (GdamSequencerSkin    *skin)
{
  do_element_cut_or_copy (skin, FALSE, FALSE);
}

static void
do_element_copy (GdamSequencerSkin    *skin)
{
  do_element_cut_or_copy (skin, TRUE, FALSE);
}

static void
do_element_cut_and_merge (GdamSequencerSkin    *skin)
{
  do_element_cut_or_copy (skin, FALSE, TRUE);
}

static void
do_element_copy_and_merge (GdamSequencerSkin    *skin)
{
  do_element_cut_or_copy (skin, TRUE, TRUE);
}

static void
do_element_clear (GdamSequencerSkin    *skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  gdam_sequencer_freeze (sequencer);
  g_hash_table_foreach (skin->selected_elements, remove_element, sequencer);
  g_hash_table_destroy (skin->selected_elements);
  skin->selected_elements = g_hash_table_new (NULL, NULL);
  gdam_sequencer_thaw (sequencer);
}

static void
select_matching_elements (GdamSequencerSkin *skin, 
                          gboolean match_active,
                          int pattern_number, gboolean match_local_id,
			  int local_id)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  GList *cur_link;

  /* TODO: here or should caller free old selection optionally?? */
  unselect_all_elements (skin);
  gdam_sequencer_freeze (sequencer);

  for (cur_link = sequencer->elements; 
       cur_link != NULL; 
       cur_link = cur_link->next)
    {
      GdamSequencerElement *element = (GdamSequencerElement *)cur_link->data;
      if (match_active && gdam_sequencer_pattern_get_active (sequencer, 
	                    element->pattern_number))
	g_hash_table_insert (skin->selected_elements, element, element);
      else if (!match_active && !match_local_id && 
	       element->pattern_number == pattern_number)
	g_hash_table_insert (skin->selected_elements, element, element);
      else if (!match_active && match_local_id && 
	       element->pattern_number == pattern_number &&
               element->local_id == local_id)
	g_hash_table_insert (skin->selected_elements, element, element);
    }

  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
}

static void
select_matching_pattern_local_id (GdamSequencerSkin *skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  int pattern_number = sequencer->current_pattern;
  int local_id = skin->local_buffer;
  select_matching_elements (skin, FALSE, pattern_number, TRUE, local_id);
}

static void
select_matching_pattern (GdamSequencerSkin *skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  int pattern_number = sequencer->current_pattern;
  select_matching_elements (skin, FALSE, pattern_number, FALSE, 0);
}

static void
select_active (GdamSequencerSkin *skin)
{
  select_matching_elements (skin, TRUE, 0, FALSE, 0);
}

/* copy elements from the clipboard into the current pattern.  Other than
 * pattern, all element attributes are unchanged. */
static void 
do_paste_elements (GdamSequencerSkin *skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  int i;

  gdam_sequencer_freeze (sequencer);
  unselect_all_elements (skin);

  for (i = 0; i < skin->num_elements_in_clipboard; i++)
    {
      GdamSequencerElement *element = skin->clipboard + i;
      int pattern_number;
      int local_id = element->local_id;
      pattern_number = sequencer->current_pattern;
      element->buffer_id = get_element_id (sequencer, element);
      element = gdam_sequencer_add_element (sequencer,
					    element->buffer_id,
					    element->beat_offset,
					    element->rate,
					    element->gain,
					    element->sample_offset,
					    element->output_length);
      g_return_if_fail (element != NULL);
      element->pattern_number = pattern_number;
      element->local_id = local_id;
      /* Pasted elements are selected for easy dragging */
      g_hash_table_insert (skin->selected_elements, element, element);
    }

  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
}


static GdamSequencerElement* 
add_new_element (GdamSequencerSkin* skin, int buffer_id, gdouble offset,
		 gdouble resample, gdouble gain)
{
  GdamSequencerElement *sample_info;
  GdamSequencerBufferInfo *info;
  g_return_val_if_fail (buffer_id != 0, NULL);
  sample_info = gdam_sequencer_add_element (get_sequencer (skin),
					    buffer_id,
					    offset,
					    resample,
					    gain, 
					    0, -1); 
  g_return_val_if_fail (sample_info != NULL, NULL);
  if (buffer_id > 0)
    {
      info = sequence_buffer_info_from_id (skin, buffer_id);
      if (!info) 
	{
	  maybe_track_buffer_id (skin, buffer_id);
	}
      info = sequence_buffer_info_from_id (skin, buffer_id);
      g_return_val_if_fail (info != NULL, NULL);
      sample_info->gain_factor = info->gain_factor;
    }
  else
    {
      sample_info->gain_factor = 1.0;
    }
  /* TODO: local_id here or by caller?? */
  sample_info->pattern_number = get_sequencer (skin)->current_pattern;
  gdam_sequencer_update_sequence (get_sequencer(skin));
  return sample_info;
}

static void finished_dragging(GdamSequencerSkin* skin)
{
  int num_dragging = skin->num_dragging;
  GdamSequencer *sequencer = get_sequencer (skin);
  int i;
  int move_x, move_y;
  gdouble delta_start, frac_y; 
  gdouble delta_gain = 0;
  gdouble delta_length = 0;
  gdouble delta_offset = 0;
  gdouble delta_log_hertz = 0;
  int delta_local_id = 0;
  int delta_pattern = 0;
  gdouble hertz_scale = 1.0;
  gdouble beats_visible;
  move_x = skin->drag_current_x - skin->drag_start_x;
  move_y = skin->drag_current_y - skin->drag_start_y;

  beats_visible = skin->max_beat - skin->min_beat;
  delta_start = move_x * beats_visible / skin->draw_width;
  //delta_start = sequencer->num_beats * move_x / skin->draw_width;
  /* XXX: use set_y_value here?? */
  frac_y = (gdouble) - move_y / skin->draw_height;
  if (skin->graph_display_mode == BY_GAIN)
    {
      delta_gain = MAX_GAIN * frac_y;
    }
  else if (skin->graph_display_mode == BY_RESAMPLE)
    {
      delta_log_hertz = (skin->max_log_hertz - skin->min_log_hertz) * frac_y;
    }
  else if (skin->graph_display_mode == BY_LOCAL_ID)
    {
      double num_local_infos = sequencer->local_infos->len;
      delta_local_id = (int)(-1 * frac_y * num_local_infos);
    }
  else if (skin->graph_display_mode == BY_PATTERN)
    {
      delta_pattern = (int)(-1 * frac_y * skin->num_patterns);
    }
  else if (skin->graph_display_mode == BY_LENGTH)
    {
      /* must be done on a per-element basis */
    }
  else if (skin->graph_display_mode == BY_START)
    {
      /* must be done on a per-element basis */
    }
  else
    {
      g_warning ("unhandled display_mode %d in finished_dragging", 
	         skin->graph_display_mode);

    }

  /* TODO: make this snapping optional. */
  if (sequencer->num_sub_beats > 0)
    {
      delta_start *= sequencer->num_sub_beats;
      delta_start += 0.5;
      delta_start = floor (delta_start);
      delta_start /= sequencer->num_sub_beats;
    }

  /* XXX: this should vary depending on sort mode */
  /* HACK: snap to equal tempered scale. */
  if (skin->snap_y)
  {
    gdouble incr = log (pow (2.0, 1.0 / 12.0));
    //delta_log_hertz += incr / 2.0;
    delta_log_hertz /= incr;
    delta_log_hertz = (int) (delta_log_hertz);
    delta_log_hertz *= incr;
  }

  hertz_scale = exp (delta_log_hertz);
  gdam_debug ("hz %f delta_log %f snap %d", hertz_scale, delta_log_hertz, 
             skin->snap_y);

  gdam_sequencer_freeze (sequencer);
  unselect_all_elements (skin);
  for (i = 0; i < num_dragging; i++)
    {
      GdamSequencerElement *element = skin->dragging + i;
      int pattern_number;
      int local_id = element->local_id + delta_local_id;
      int output_length = element->output_length;
      pattern_number = element->pattern_number + delta_pattern;
      element->buffer_id = get_element_id (sequencer, element);
      if (skin->graph_display_mode == BY_LENGTH)
	{
	  int possible_length;
	  double perc;
	  /* old y value from old length */
	  int new_y = get_y_position (skin, element);
	  new_y += move_y;
	  /* flip vertically */
	  new_y = skin->draw_height - new_y;
	  perc = (float)new_y / skin->draw_height;
	  possible_length = get_element_possible_length (sequencer, element);
	  output_length = (int)(perc * possible_length);
	}
      else if (skin->graph_display_mode == BY_START)
	{
	  int possible_length;
	  possible_length = get_element_possible_length (sequencer, element);
	  delta_offset = (int)(frac_y * possible_length);
	}
      element = gdam_sequencer_add_element (sequencer,
					    element->buffer_id,
					    element->beat_offset +
					    delta_start,
					    element->rate * hertz_scale,
					    element->gain + delta_gain,
					    element->sample_offset +
					    delta_offset,
					    output_length);
      g_return_if_fail (element != NULL);
      element->pattern_number = pattern_number;
      element->local_id = local_id;
      if (skin->was_dragging_selected)
	g_hash_table_insert (skin->selected_elements, element, element);
    }
  if (skin->in_transit)
    {
      g_hash_table_foreach (skin->in_transit, remove_element, sequencer);
      g_hash_table_destroy (skin->in_transit);
      skin->in_transit = NULL;
    }

  g_free (skin->dragging);
  skin->dragging = NULL;
  skin->num_dragging = 0;

  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
}

/*  _              _   _          __  __         
 * | |___  __ __ _| | | |__ _  _ / _|/ _|___ _ _ 
 * | / _ \/ _/ _` | | | '_ \ || |  _|  _/ -_) '_|
 * |_\___/\__\__,_|_| |_.__/\_,_|_| |_| \___|_|  
 *
 *                                              _   
 *  _ __  __ _ _ _  __ _ __ _ ___ _ __  ___ _ _| |_ 
 * | '  \/ _` | ' \/ _` / _` / -_) '  \/ -_) ' \  _|
 * |_|_|_\__,_|_||_\__,_\__, \___|_|_|_\___|_||_\__|
 *                      |___/                       
 */

static void list_buf (gpointer key, gpointer value, gpointer udata)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (udata);
  GdamSequencerBufferInfo *info = (GdamSequencerBufferInfo*)value;
  skin->buffers = g_list_append (skin->buffers, key);
}

static void recreate_buffer_list (GdamSequencerSkin *skin)
{
  if (skin->buffers)
    g_list_free (skin->buffers);
  skin->buffers = NULL;
  if (skin->buffer_id_to_info == NULL)
    {
      skin->buffer_id_to_info = g_hash_table_new (NULL, NULL);
    }
  g_hash_table_foreach (skin->buffer_id_to_info, list_buf, skin);
}

/* XXX: merge with select_matching */
static void select_buffer_by_info     (GdamSequencerBufferInfo* info)
{
  GList *el;
  GdamSequencer *sequencer;
  GHashTable *selected_elements;
  GdamSequencerSkin *sequencer_skin = info->sequencer_skin;
  sequencer = get_sequencer (sequencer_skin);
  unselect_all_elements (sequencer_skin);
  selected_elements = sequencer_skin->selected_elements;
  for (el = sequencer->elements; el != NULL; el = el->next)
    {
      GdamSequencerElement *element = el->data;
      if (element->buffer_id == info->buffer_id)
	g_hash_table_insert (selected_elements, element, element);
    }
  delayed_redraw (sequencer_skin);
}

static void toggle_hide_buffer_by_info(GtkToggleButton*         toggle,
                                       GdamSequencerBufferInfo* info)
{
  gboolean hidden = FALSE;
  GHashTable *hidden_buffers = info->sequencer_skin->hidden_buffers;
  gpointer buffer_id = GUINT_TO_POINTER (info->buffer_id);
  if (GTK_IS_TOGGLE_BUTTON (toggle))
    hidden = GTK_TOGGLE_BUTTON (toggle)->active;
  else if (GTK_IS_PIX_BUTTON (toggle))
    hidden = gtk_pix_button_get_active (GTK_PIX_BUTTON (toggle));
  else
    g_assert_not_reached ();
  if (hidden)
    g_hash_table_insert (hidden_buffers, buffer_id, buffer_id);
  else
    g_hash_table_remove (hidden_buffers, buffer_id);
  delayed_redraw (info->sequencer_skin);
}

static void set_buffer_mute (GdamSequencerSkin* sequencer_skin,
			     int                buffer_id,
			     gboolean           muted,
			     gboolean		is_local_id)
{
  gboolean changed = FALSE;
  GList *el;
  GdamSequencer *sequencer = get_sequencer (sequencer_skin);
  for (el = sequencer->elements; el != NULL; el = el->next)
    {
      GdamSequencerElement *element = el->data;
      if ((!is_local_id && (int)element->buffer_id == buffer_id) ||
           (is_local_id && (int)element->local_id  == buffer_id))
	{
	  changed = TRUE;
	  element->muted = muted;
	}
    }
  if (changed)
    gdam_sequencer_update_sequence (sequencer);
}

static gboolean gdam_util_get_toggle_active (GtkWidget* toggle)
{
  if (GTK_IS_TOGGLE_BUTTON (toggle))
    return GTK_TOGGLE_BUTTON (toggle)->active;
  else if (GTK_IS_PIX_BUTTON (toggle))
    return gtk_pix_button_get_active (GTK_PIX_BUTTON (toggle));
  g_assert_not_reached();
  return FALSE;
}

static void by_local_info_mute_toggled (GtkWidget*     toggle,
					BufIdPtr*      info)
{
  gboolean muted = gdam_util_get_toggle_active (toggle);
  set_buffer_mute (info->sequencer_skin, info->buffer_id, muted, TRUE);
}

static void by_info_mute_toggled(GtkWidget*               toggle,
                                 GdamSequencerBufferInfo* info)
{
  gboolean muted = gdam_util_get_toggle_active (toggle);
  set_buffer_mute (info->sequencer_skin, info->buffer_id, muted, FALSE);
}

static void regain_buffer_by_id (GdamSequencerSkin*  sequencer_skin,
                                 int 		     id,
				 gdouble 	     del_factor,
				 gboolean	     is_local_id)
{
  gboolean changed = FALSE;
  GList *el;
  GdamSequencer *sequencer = get_sequencer (sequencer_skin);
  GdamSequencerBufferInfo *info;
  info = sequence_buffer_info_from_id (sequencer_skin, id);
  g_return_if_fail (info != NULL);
  info->gain_factor *= del_factor;
  for (el = sequencer->elements; el != NULL; el = el->next)
    {
      GdamSequencerElement *element = el->data;
      if ((!is_local_id && (int)element->buffer_id == id) ||
	   (is_local_id && (int)element->local_id  == id))
	{
	  changed = TRUE;
	  element->gain_factor *= del_factor;
	}
    }
  if (changed)
    gdam_sequencer_update_sequence (sequencer);
}

static void transpose_buffer_by_id (GdamSequencerSkin*  sequencer_skin,
                                    int 		id,
				    gdouble 		del_factor,
				    gboolean		is_local_id)
{
  gboolean changed = FALSE;
  GList *el;
  GdamSequencer *sequencer = get_sequencer (sequencer_skin);
  for (el = sequencer->elements; el != NULL; el = el->next)
    {
      GdamSequencerElement *element = el->data;
      if ((!is_local_id && (int)element->buffer_id == id) ||
           (is_local_id && (int)element->local_id  == id))
	{
	  changed = TRUE;
	  element->rate_factor *= del_factor;
	}
    }
  if (changed)
    gdam_sequencer_update_sequence (sequencer);
}

/* Find each element with local id local_id, and set their buffer to new_id */
static void swap_buffer_by_local_id (GdamSequencerSkin  *sequencer_skin,
                                     int 		 local_id,
				     int 		 new_id)
{
  gboolean changed = FALSE;
  GList *el;
  GdamSequencer *sequencer = get_sequencer (sequencer_skin);
  for (el = sequencer->elements; el != NULL; el = el->next)
    {
      GdamSequencerElement *element = el->data;
      if (element->local_id  == local_id)
	{
	  changed = TRUE;
	  element->buffer_id = new_id;
	}
    }
  if (changed)
    gdam_sequencer_update_sequence (sequencer);
}

static void transpose_buffer_up_by_info (GtkWidget               *wid,
                                         GdamSequencerBufferInfo *info)
{
  transpose_buffer_by_id (info->sequencer_skin, info->buffer_id, 2.0, FALSE);
}

static void transpose_buffer_up_by_local_info (GtkWidget *wid,
                                               BufIdPtr  *info)
{
  transpose_buffer_by_id (info->sequencer_skin, info->buffer_id, 2.0, TRUE);
}

static void transpose_buffer_down_by_info (GtkWidget               *wid,
                                           GdamSequencerBufferInfo *info)
{
  transpose_buffer_by_id (info->sequencer_skin, info->buffer_id, 0.5, FALSE);
}

static void transpose_buffer_down_by_local_info (GtkWidget  *wid,
                                                 BufIdPtr   *info)
{
  transpose_buffer_by_id (info->sequencer_skin, info->buffer_id, 0.5, TRUE);
}

static void regain_buffer_down_by_local_info (GtkWidget *wid,
                                              BufIdPtr  *info)
{
  regain_buffer_by_id (info->sequencer_skin, info->buffer_id, 0.5, TRUE);
}

static void regain_buffer_up_by_local_info (GtkWidget *wid,
                                            BufIdPtr  *info)
{
  regain_buffer_by_id (info->sequencer_skin, info->buffer_id, 2.0, TRUE);
}


static void delete_one_buffer(gpointer el, gpointer skin)
{
  gdam_debug ("deleting one buffer.");
  delete_subinfo (GDAM_SEQUENCER_SKIN (skin), (GdamSequencerElement *) el);
}

static void delete_buffer_by_id (GdamSequencerSkin *skin, 
    				 int 	            buffer_id, 
				 gboolean           is_local_id)
{
  GdamSequencer *sequencer;
  GList *at;
  GSList *to_delete = NULL;
  sequencer = get_sequencer (skin);

  gdam_debug ("delete_buffer_by_id called on list of %d elements.",
	      g_list_length (sequencer->elements));

  /* delete the buffer id from the list of sequencer elements. */
  for (at = sequencer->elements; at != NULL; at = at->next)
    {
      GdamSequencerElement *element;
      element = at->data;
      if ((!is_local_id && (int)element->buffer_id == buffer_id) ||
	   (is_local_id && (int)element->local_id  == buffer_id))
	{
	  to_delete = g_slist_prepend (to_delete, element);
	}
    }

  gdam_debug ("delete_buffer_by_id deleting %d elements.",
	      g_slist_length (to_delete));

  gdam_sequencer_freeze (sequencer);
  g_slist_foreach (to_delete, delete_one_buffer, skin);
  gdam_sequencer_thaw (sequencer);
  g_slist_free (to_delete);

  gdam_debug ("delete_buffer_by_id leaving list of %d elements.",
	      g_list_length (sequencer->elements));

  g_hash_table_remove (skin->hidden_buffers,
		       GUINT_TO_POINTER (buffer_id));
  delayed_redraw (skin);
}

static void delete_buffer_by_info(GdamSequencerBufferInfo* info)
{
  delete_buffer_by_id (info->sequencer_skin, info->buffer_id, FALSE);
  if (info->topwidget)
    gtk_widget_destroy (info->topwidget);
}

static void delete_buffer_by_local_info (GtkWidget *widget, BufIdPtr *ptr)
{
  GdamSequencerBufferInfo* info;
  GdamSequencer *sequencer = get_sequencer (ptr->sequencer_skin);
  int buffer_id;

  delete_buffer_by_id (ptr->sequencer_skin, ptr->buffer_id, TRUE);

  /* XXX: destroy topwidget only if no buffers left?? or ever?? */
}

static void on_buffer_topwidget_destroy(GdamSequencerBufferInfo* info)
{
  GdamSequencerSkin *skin = info->sequencer_skin;
  g_hash_table_remove (skin->buffer_id_to_info,
		       GUINT_TO_POINTER (info->buffer_id));
  g_free (info);
}

static GdamSequencerBufferInfo* 
sequence_buffer_info_from_id (GdamSequencerSkin* skin, int id)
{
  guint buffer_id = id;
  GdamSequencerBufferInfo* info;
  //g_return_val_if_fail (skin->buffer_id_to_info != NULL, NULL);
  info = g_hash_table_lookup (skin->buffer_id_to_info,
			      GUINT_TO_POINTER (buffer_id));
  return info;
}

/*  If we aren't already tracking this buffer id, do so.  */
static GdamSequencerBufferInfo* maybe_track_buffer_id(GdamSequencerSkin* skin,
						      guint         buffer_id)
{
  char label_buf[255];
  char filename_label_buf[1024];
  const char* buf_fname;
  GdamSequencerBufferInfo *info;
  GtkWidget *new_widget;
  GtkWidget *hbo;
  GtkWidget *sub_widget;
  GdamBufferManagerInfo *bufinfo;

  if (buffer_id <= 0)
    return NULL;

  if (skin->buffer_id_to_info == NULL)
    {
      skin->buffer_id_to_info = g_hash_table_new (NULL, NULL);
    }

  if (g_hash_table_lookup (skin->buffer_id_to_info,
			   GUINT_TO_POINTER (buffer_id)))
    {
      return g_hash_table_lookup (skin->buffer_id_to_info, 
	                          GUINT_TO_POINTER (buffer_id));
    }

  info = g_new0 (GdamSequencerBufferInfo, 1);

  info->buffer_id = buffer_id;
  info->sequencer_skin = skin;
  info->gain_factor = 1.0;

  gdam_buffer_manager_ref (buffer_id, (GdamModel *) get_sequencer (skin));

  if (skin->buffer_id_holder == NULL)
    {
      g_hash_table_insert (skin->buffer_id_to_info,
			   GUINT_TO_POINTER (buffer_id), info);
      return info;
    }

  info->topwidget = new_widget = gtk_vbox_new (FALSE, 0);
  hbo = gtk_hbox_new (FALSE, 0);

  snprintf (label_buf, sizeof (label_buf), "%d", buffer_id);
  sub_widget = gtk_label_new (label_buf);
  gtk_widget_show (sub_widget);
  hbo = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);
  gtk_box_pack_start_defaults (GTK_BOX (new_widget), hbo);
  gtk_widget_show (hbo);

  bufinfo = gdam_buffer_manager_info_by_id (buffer_id);
  g_return_val_if_fail (bufinfo != NULL, NULL);
  buf_fname = bufinfo->description;
  if (strncmp (buf_fname, "Wav ", 4) == 0)
    {
      buf_fname += 4;
      buf_fname = g_basename (buf_fname);
    }
  snprintf (filename_label_buf, sizeof (filename_label_buf), 
            "%s", buf_fname);
  sub_widget = gtk_label_new (filename_label_buf);
  gtk_widget_show (sub_widget);
  hbo = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);
  gtk_box_pack_start_defaults (GTK_BOX (new_widget), hbo);
  gtk_widget_show (hbo);

  hbo = gtk_hbox_new (FALSE, 0);
  //sub_widget = gtk_button_new_with_label ("Select");
  sub_widget = gtk_pix_button_create_func ("bufferselect_pixbutton", 
      "default/bufferselect.xpm:default/bufferselect.xpm:default/bufferselect.xpm:default/bufferselect.xpm", 
      "", 0,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect_object (GTK_OBJECT (sub_widget),
			     "clicked",
			     GTK_SIGNAL_FUNC (select_buffer_by_info),
			     (GtkObject *) info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  //sub_widget = gtk_toggle_button_new_with_label ("Hide");
  sub_widget = gtk_pix_button_create_func ("bufferhide_pixbutton", 
      "default/bufferhide.xpm:default/bufferhide_active.xpm:default/bufferhide.xpm:default/bufferhide_active.xpm", 
      "", 1,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect (GTK_OBJECT (sub_widget),
		      "toggled",
		      GTK_SIGNAL_FUNC (toggle_hide_buffer_by_info), info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  //sub_widget = gtk_toggle_button_new_with_label ("Mute");
  sub_widget = gtk_pix_button_create_func ("buffermute_pixbutton", 
      "default/buffermute.xpm:default/buffermute_active.xpm:default/buffermute.xpm:default/buffermute_active.xpm", 
      "", 1,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect (GTK_OBJECT (sub_widget),
		      "toggled",
		      GTK_SIGNAL_FUNC (by_info_mute_toggled), info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  //sub_widget = gtk_button_new_with_label ("Pitch Up");
  sub_widget = gtk_pix_button_create_func ("bufferup_pixbutton", 
      "default/bufferup.xpm:default/bufferup.xpm:default/bufferup.xpm:default/bufferup.xpm", 
      "", 0,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect (GTK_OBJECT (sub_widget),
		      "clicked",
		      GTK_SIGNAL_FUNC (transpose_buffer_up_by_info), info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  //sub_widget = gtk_button_new_with_label ("Down");
  sub_widget = gtk_pix_button_create_func ("bufferdown_pixbutton", 
      "default/bufferdown.xpm:default/bufferdown.xpm:default/bufferdown.xpm:default/bufferdown.xpm", 
      "", 0,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect (GTK_OBJECT (sub_widget),
		      "clicked",
		      GTK_SIGNAL_FUNC (transpose_buffer_down_by_info), info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  //sub_widget = gtk_button_new_with_label ("Delete");
  sub_widget = gtk_pix_button_create_func ("bufferdelete_pixbutton", 
      "default/bufferdelete.xpm:default/bufferdelete.xpm:default/bufferdelete.xpm:default/bufferdelete.xpm", 
      "", 0,0);
  gtk_widget_set_usize (sub_widget, 15, 15);
  gtk_signal_connect_object (GTK_OBJECT (sub_widget),
			     "clicked",
			     GTK_SIGNAL_FUNC (delete_buffer_by_info),
			     (GtkObject *) info);
  gtk_widget_show (sub_widget);
  gtk_box_pack_start (GTK_BOX (hbo), sub_widget, FALSE, FALSE, 0);

  gtk_signal_connect_object (GTK_OBJECT (new_widget),
			     "destroy",
			     GTK_SIGNAL_FUNC (on_buffer_topwidget_destroy),
			     (GtkObject *) info);

  gtk_widget_show (hbo);
  gtk_widget_show (new_widget);
  gtk_widget_show (hbo);
  gtk_box_pack_start (GTK_BOX (new_widget), hbo, 0, 0, 0);
  gtk_box_pack_start (skin->buffer_id_holder, new_widget, 0, 0, 5);

  g_hash_table_insert (skin->buffer_id_to_info,
		       GUINT_TO_POINTER (buffer_id), info);

  return info;
}


#if 0
{
  /* TODO: add entry to pattern_list */
  if (skin->pattern_list)
    {
      GList *items = NULL;
      GtkWidget *widget;

      widget = gtk_label_new ("pattern n");
      items = g_list_append (items, widget);
      gtk_list_append_items (skin->pattern_list, items);

    }
}
#endif


/* Given an element, tile it across time.  If stretch-to-fill, then
 * we find the rate necessary to make the buffer last until the next 
 * time it plays (useful for loops) */
static void
tile_sample 	(GdamSequencerSkin      *skin,
             	 GdamSequencerElement 	*sample_info,
		 gboolean	       	 stretch_to_fill)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  gdouble num_beats = sequencer->num_beats;
  gdouble cur_beat;
  gdouble bpm, beat_length, step_size, length_secs;
  int length_samples;
  int buffer_id = nth_local_id (sequencer, sample_info->local_id);
  gdouble elements_per_beat = 1;

  /* XXX: what about length from local buffer info, i.e. when recycling?? */
  length_samples = -1;
  if (length_samples < 0) 
    {
      GdamBufferManagerInfo *info;
      info = gdam_buffer_manager_info_by_id (buffer_id);
      if (info == NULL)
	length_samples = 2205;
      else
	length_samples = info->length;
    }
  length_secs = (float)length_samples / 44100.0;

  /* Figure out how best to tile the buffer across time */
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (sequencer)->beat_info, 0);
  beat_length = 60.0 / bpm;

  while (length_secs * elements_per_beat < 0.75 * beat_length)
    elements_per_beat *= 2;
  while (length_secs * elements_per_beat > 1.5 * beat_length)
    elements_per_beat /= 2;

  step_size = 1.0 / elements_per_beat;
  g_return_if_fail (step_size > 0.01);

  /* If we are stretching to fill, calculate target rate */
  if (stretch_to_fill)
    {
      gdouble target_size = beat_length / elements_per_beat;
      gdouble rate = length_secs / target_size;
      /* adjust rate of the original element */
      sample_info->rate = rate;
    }

  //gdam_buffer_manager_unref (buffer_id, (GdamModel *) sequencer);
  //maybe_track_element (sample_info, skin);


  /* XXX: does this properly wrap around if one is looping from a 
   * sample in the middle of the sequence? */
  cur_beat = sample_info->beat_offset;
  cur_beat += step_size;
  for (;cur_beat < num_beats; cur_beat += step_size)
    {
      GdamSequencerElement *new_sample_info;
      new_sample_info = add_new_element (skin,
					 buffer_id,
					 cur_beat,
					 sample_info->rate,
					 sample_info->gain);
      new_sample_info->pattern_number = sample_info->pattern_number;
      new_sample_info->local_id = sample_info->local_id;
    }

}

/*
 *   ____     _ _    _____                 _
 *  / ___| __| | | _| ____|_   _____ _ __ | |_ ___
 * | |  _ / _` | |/ /  _| \ \ / / _ \ '_ \| __/ __|
 * | |_| | (_| |   <| |___ \ V /  __/ | | | |_\__ \
 *  \____|\__,_|_|\_\_____| \_/ \___|_| |_|\__|___/
 * 
 */

static void maybe_track_element (gpointer data, gpointer udata)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (udata);
  GdamSequencerElement *element = (GdamSequencerElement *)data;
  int id;
  id = get_element_id (get_sequencer (skin), element);
  maybe_track_buffer_id (skin, id);
}

static gint
sequence_edit_button_press (GtkDrawingArea    *area,
                            GdkEventButton    *button_event,
		            GdamSequencerSkin *skin)
{
  int press_x = button_event->x;
  int press_y = button_event->y;
  int width, height;
  GdamSequencerElement *sample_info;
  int buffer_id, local_id, pattern_number;
  gdouble beat;
  gdouble beats_visible;
  GdamSequencer *sequencer = get_sequencer (skin);
  gdam_debug ("sequence_edit_button_press");
  g_return_val_if_fail (skin->dragging == NULL, 0);

  skin->mouse_down = TRUE;

  /*  If the middle button is being used, create a selection.  */
  if (button_event->button == 2)
    {
      int state = button_event->state;
      gboolean is_shifted, is_ctrl;
      begin_selection (skin, press_x, press_y);
      is_ctrl = ((state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK);
      is_shifted = ((state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK);
      gdam_skin_pop_message (GDAM_SKIN (skin), skin->selection_context_id);
      if (is_shifted)
	{
	  gdam_skin_push_message (GDAM_SKIN (skin),
				  skin->selection_context_id,
				  "The framed items will be added to "
				  "the selection set.");
	  skin->selection_mode = SELECTION_MODE_ADD;
	}
      else if (is_ctrl)
	{
	  gdam_skin_push_message (GDAM_SKIN (skin),
				  skin->selection_context_id,
				  "The framed items will be toggled between "
				  "selected and unselected.");
	  skin->selection_mode = SELECTION_MODE_TOGGLE;
	}
      else
	{
	  gdam_skin_push_message (GDAM_SKIN (skin),
				  skin->selection_context_id,
				  "The framed items will be selected; the "
				  "current selection will be cleared.");
	  skin->selection_mode = SELECTION_MODE_SET;
	}
      return TRUE;
    }
  /*  If we clicked on a sample, delete or drag it.  */
  if ((sample_info = find_submodel (skin, press_x, press_y)) != NULL)
    {
      int pos_x, pos_y;
      if (button_event->button == 3)
	{
	  if (g_hash_table_lookup (skin->selected_elements, sample_info))
	    delete_selected (skin);
	  else
	    delete_subinfo (skin, sample_info);
	  return TRUE;
	}
      get_subsample_position (sample_info, skin, &pos_x, &pos_y);
      skin->drag_start_x = press_x;
      skin->drag_start_y = press_y;
      begin_dragging (skin, sample_info, button_event->state);
      return TRUE;
    }

  if (button_event->button == 3)
    return 0;

  /*  Add a new element at the click position.  */
  if (local_id_valid (sequencer, skin->local_buffer))
  {
    buffer_id = nth_local_id (sequencer, skin->local_buffer);
    local_id = skin->local_buffer;
  }
  else
  {
    buffer_id = gdam_buffer_manager_obtain_id (GDAM_MODEL (sequencer));
    local_id = -1;
  }

  if (buffer_id == 0)
    {
      g_message ("no sample has been chosen -- cannot add");
      return 0;
    }

  gdam_sequencer_freeze (sequencer);

  width = skin->draw_width;
  height = skin->draw_height;

  /* TODO: are we sure this position takes into account that we might
   * be viewing only part of the sequence? */
  gdam_debug ("button press: num_beats is %f", sequencer->num_beats);
  beats_visible = skin->max_beat - skin->min_beat;
  beat = press_x * beats_visible / width + skin->min_beat;

  if (skin->graph_display_mode == BY_LOCAL_ID && local_id >= 0)
    {
      double num = get_sequencer (skin)->local_infos->len;
      double perc = (float)press_y / height;
      local_id = (int)(perc * num);
    }
  pattern_number = 0;
  if (skin->graph_display_mode == BY_PATTERN)
    {
      double num = skin->num_patterns;
      double perc = (float)press_y / height;
      pattern_number = (int)(perc * num);
    }

  if (skin->display_tool == TOOL_ADD)
    {
      sample_info = add_new_element (skin,
				     buffer_id,
				     beat,
				     skin->default_resample,
				     skin->default_gain);
      if (skin->graph_display_mode == BY_PATTERN)
	sample_info->pattern_number = pattern_number;
      sample_info->local_id = local_id;

      gdam_buffer_manager_ref (buffer_id, (GdamModel *) sequencer);
      maybe_track_element (sample_info, skin);

      /* Clean up display */
      snap_to_grid (skin, press_x, press_y, sample_info);
      get_subsample_position (sample_info, skin, &press_x, &press_y);
      redraw_area (skin,
		   press_x - skin->circle_radius,
		   press_y - skin->circle_radius,
		   skin->circle_radius * 2 + 1, skin->circle_radius * 2 + 1);
    }
  else if (skin->display_tool == TOOL_FILL_NEAREST_BEAT ||
      skin->display_tool == TOOL_FILL_NEAREST_AND_STRETCH)
    {
      /* Add first element, snap to grid */
      sample_info = add_new_element (skin,
				     buffer_id,
				     beat,
				      /* XXX: snap-to-rate vs. rate-to-fill */
				     skin->default_resample,
				     skin->default_gain);
      if (skin->graph_display_mode == BY_PATTERN)
	sample_info->pattern_number = pattern_number;
      sample_info->local_id = local_id;

      gdam_buffer_manager_ref (buffer_id, (GdamModel *) sequencer);
      maybe_track_element (sample_info, skin);

      snap_to_grid (skin, press_x, press_y, sample_info);
      tile_sample (skin, sample_info, 
	           skin->display_tool == TOOL_FILL_NEAREST_AND_STRETCH 
		     ? TRUE : FALSE);
      redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
    }
  else if (skin->display_tool == TOOL_TRACE_Y_VALUE)
    {
      /* Save current trace location */
      gdouble current_beat, beats_visible, percent;
      percent = button_event->x / skin->draw_width;
      beats_visible = skin->max_beat - skin->min_beat;
      current_beat = percent * beats_visible + skin->min_beat;
      skin->last_trace_beat = current_beat;
    }
  else
    {
      g_warning ("unhandled tool %d", skin->display_tool);
    }

  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
  return TRUE;
}

static gint
sequence_edit_button_motion(GtkDrawingArea* area,
                            GdkEventMotion* button_event,
		            GdamSequencerSkin* skin)
{
  int x, y;
  guint state;
  if (button_event->is_hint)
    gdk_window_get_pointer (button_event->window, &x, &y, &state);
  else
    {
      x = button_event->x;
      y = button_event->y;
      state = button_event->state;
    }

  /*  If we are making a selection, extend it. */
  if (skin->has_selection)
    {
      update_selection (skin, button_event->x, button_event->y);
      return TRUE;
    }
  /* If we are dragging samples, redraw them. */
  if (skin->dragging)
    {
      undraw_dragged_samples (skin);
      skin->drag_current_x = button_event->x;
      skin->drag_current_y = button_event->y;
      draw_dragged_samples (skin);
      return TRUE;
    }
  /* If we are tracing the y axis value, then find all sensitive elements
   * which match the new x axis value, and change their y axis value */
  /* TODO: check for elements at all x values between last and new?? */
  if (skin->mouse_down && skin->display_tool == TOOL_TRACE_Y_VALUE)
    {
      /* TODO x->beatnum convenience function? is_sensitive c.f.? */
      GList *elements;
      GList *el;
      GdamSequencer *sequencer;
      gboolean match = FALSE;
      gdouble current_beat, beats_visible, percent;
      elements = get_sequencer (skin)->elements;
      percent = button_event->x / skin->draw_width;
      beats_visible = skin->max_beat - skin->min_beat;
      current_beat = percent * beats_visible + skin->min_beat;

	/* for each sequencer element, check if it is between the last event
	 * time and the current event time, and that it is manipulateable. 
	 */ 
      for (el=elements; el != NULL; el = el->next)
	{
	  GdamSequencerElement *seqel;
	  seqel = (GdamSequencerElement *)el->data;
	  g_return_val_if_fail (seqel != NULL, TRUE);
	  if (seqel->pattern_number == get_sequencer (skin)->current_pattern)
	    {
	      if ((seqel->beat_offset > skin->last_trace_beat &&
		   seqel->beat_offset <= current_beat) ||
		  (seqel->beat_offset < skin->last_trace_beat &&
		   seqel->beat_offset >= current_beat))
		{
		  set_y_position (skin, seqel, y);
		  match = TRUE;
		}
	    }
	}
      skin->last_trace_beat = current_beat;
      if (match)
	delayed_redraw (skin);
    }
  return TRUE;
}

static gint
sequence_edit_button_release(GtkDrawingArea* area,
                             GdkEventButton* button_event,
		             GdamSequencerSkin* skin)
{
  skin->mouse_down = FALSE;
  if (skin->has_selection)
    {
      finish_selection (skin, button_event->x, button_event->y);
      return 0;
    }
  if (!skin->dragging)
    return 0;

  undraw_dragged_samples (skin);

  skin->drag_current_x = button_event->x;
  skin->drag_current_y = button_event->y;

  finished_dragging (skin);
  return TRUE;
}

static void sequence_size_request(GtkWidget*		wid,
				  GtkRequisition*       requisition,
				  GdamSequencerSkin*    skin)
{
  requisition->width = skin->drawing_area_width;
  requisition->height = skin->drawing_area_height;
}

static void sequence_size_allocate(GtkWidget*		 wid,
				   GtkAllocation*        allocation,
				   GdamSequencerSkin*    skin)
{
  gboolean changed = allocation->width != skin->drawing_area_width
    || allocation->height != skin->drawing_area_height;
  skin->drawing_area_width = allocation->width;
  skin->drawing_area_height = allocation->height;
  if (changed)
    redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static gint
sequence_edit_configure(GtkWidget* widget,
                        GdkEventConfigure* configure_event,
		        GdamSequencerSkin* skin)
{
  if (skin->backing_pixmap)
    gdk_pixmap_unref (skin->backing_pixmap);
  skin->backing_pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height, -1);
  skin->draw_width = widget->allocation.width;
  skin->draw_height = widget->allocation.height;
  return FALSE;
}

static void
arg_set_vertical_sort_mode (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  char *new_mode = GDAM_ARG_STRING (arg);

  if (strcmp (new_mode, "resample") == 0)
    skin->graph_display_mode = BY_RESAMPLE;
  else if (strcmp (new_mode, "gain") == 0)
    skin->graph_display_mode = BY_GAIN;
  else if (strcmp (new_mode, "buffer_id") == 0)
    skin->graph_display_mode = BY_ID;
  else if (strcmp (new_mode, "local_id") == 0)
    skin->graph_display_mode = BY_LOCAL_ID;
  else if (strcmp (new_mode, "length") == 0)
    skin->graph_display_mode = BY_LENGTH;
  else if (strcmp (new_mode, "start") == 0)
    skin->graph_display_mode = BY_START;
  else if (strcmp (new_mode, "pattern") == 0)
    {
      skin->graph_display_mode = BY_PATTERN;
      update_num_patterns (skin);
    }
  else
    g_warning ("unknown vertical sort mode %s", new_mode);
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void
arg_get_vertical_sort_mode (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  char *new_mode = GDAM_ARG_STRING (arg);
  if (skin->graph_display_mode == BY_RESAMPLE)
    new_mode = "resample";
  else if (skin->graph_display_mode == BY_GAIN)
    new_mode = "gain";
  else if (skin->graph_display_mode == BY_ID)
    new_mode = "buffer_id";
  else if (skin->graph_display_mode == BY_LOCAL_ID)
    new_mode = "local_id";
  else if (skin->graph_display_mode == BY_LENGTH)
    new_mode = "length";
  else if (skin->graph_display_mode == BY_START)
    new_mode = "start";
  else if (skin->graph_display_mode == BY_PATTERN)
    new_mode = "pattern";
  else
    {
      g_warning ("unknown vertical sort mode %s", new_mode);
      new_mode = "pattern";
    }
  GDAM_ARG_STRING (arg) = new_mode;
}

/* XXX: these should be connectable... */
static void sort_by_resample(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "resample");
}

static void sort_by_gain(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "gain");
}

static void sort_by_id(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "buffer_id");
}

static void sort_by_local_id(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "local_id");
}

static void sort_by_pattern(GdamSequencerSkin* skin)
{
  update_num_patterns (skin);
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "pattern");
}

static void sort_by_length(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "length");
}

static void sort_by_start(GdamSequencerSkin* skin)
{
  gdam_arg_set (GTK_OBJECT (skin), "vertical_sort_mode", "start");
}

static void 
render_as_circles (GdamSequencerSkin *skin)
{
  skin->element_render_mode = RENDER_AS_CIRCLES;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
render_as_rectangles (GdamSequencerSkin *skin)
{
  skin->element_render_mode = RENDER_AS_RECTANGLES;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
render_as_colored_squares (GdamSequencerSkin *skin)
{
  skin->element_render_mode = RENDER_AS_COLOR_SQUARES;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
render_as_tee (GdamSequencerSkin *skin)
{
  skin->element_render_mode = RENDER_AS_TEE;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
render_as_pennant (GdamSequencerSkin *skin)
{
  skin->element_render_mode = RENDER_AS_PENNANT;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
text_as_buffer_name (GdamSequencerSkin *skin)
{
  skin->element_text_mode = TEXT_BUFFER_NAME;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
text_as_local_id (GdamSequencerSkin *skin)
{
  skin->element_text_mode = TEXT_LOCAL_ID;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
set_tool_fill_nearest (GdamSequencerSkin *skin)
{
  skin->display_tool = TOOL_FILL_NEAREST_BEAT;
}

static void 
set_tool_trace_y_value (GdamSequencerSkin *skin)
{
  skin->display_tool = TOOL_TRACE_Y_VALUE;
}

static void 
set_tool_fill_nearest_and_stretch (GdamSequencerSkin *skin)
{
  skin->display_tool = TOOL_FILL_NEAREST_AND_STRETCH;
}

static void 
set_tool_add (GdamSequencerSkin *skin)
{
  skin->display_tool = TOOL_ADD;
}

static void 
do_hide_inactive (GdamSequencerSkin *skin)
{
  skin->hide_inactive = TRUE;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
do_show_inactive (GdamSequencerSkin *skin)
{
  skin->hide_inactive = FALSE;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
do_size_from_length (GdamSequencerSkin *skin)
{
  skin->visible_size_from_length = TRUE;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void 
undo_size_from_length (GdamSequencerSkin *skin)
{
  skin->visible_size_from_length = FALSE;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
}

static void connect_drawing_area(GtkDrawingArea*    drawing_area,
                                 GdamSequencerSkin* skin)
{
  skin->drawing_area = GTK_WIDGET (drawing_area);
  //gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 800, 600);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_press_event",
		      GTK_SIGNAL_FUNC (sequence_edit_button_press), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "motion_notify_event",
		      GTK_SIGNAL_FUNC (sequence_edit_button_motion), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "button_release_event",
		      GTK_SIGNAL_FUNC (sequence_edit_button_release), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "expose_event",
		      GTK_SIGNAL_FUNC (sequence_edit_expose), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "configure_event",
		      GTK_SIGNAL_FUNC (sequence_edit_configure), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "size_request",
		      GTK_SIGNAL_FUNC (sequence_size_request), skin);
  gtk_signal_connect (GTK_OBJECT (drawing_area),
		      "size_allocate",
		      GTK_SIGNAL_FUNC (sequence_size_allocate), skin);
  gtk_widget_add_events (GTK_WIDGET (drawing_area),
			 GDK_EXPOSURE_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);
  //gtk_widget_show(GTK_WIDGET(drawing_area));
}

/*
 *  __  __ _              _ _
 * |  \/  (_)___  ___ ___| | | __ _ _ __   ___  ___  _   _ ___
 * | |\/| | / __|/ __/ _ \ | |/ _` | '_ \ / _ \/ _ \| | | / __|
 * | |  | | \__ \ (_|  __/ | | (_| | | | |  __/ (_) | |_| \__ \
 * |_|  |_|_|___/\___\___|_|_|\__,_|_| |_|\___|\___/ \__,_|___/
 * 
 * 
 * __      ___ __ __ _ _ __  _ __   ___ _ __ ___
 * \ \ /\ / / '__/ _` | '_ \| '_ \ / _ \ '__/ __|
 *  \ V  V /| | | (_| | |_) | |_) |  __/ |  \__ \
 *   \_/\_/ |_|  \__,_| .__/| .__/ \___|_|  |___/
 *                    |_|   |_|
 */

#if 0
static GdamSequencerSkin* new_blank_skin(GdamSequencer* sequencer)
{

  GdamSequencerSkin *skin = g_new0 (GdamSequencerSkin, 1);
  GDAM_SKIN (skin)->model = GDAM_MODEL (sequencer);
  skin->default_resample = 1.0;
  skin->default_gain = 1.0;
  skin->circle_radius = 10;
  return skin;
}
#endif

static void delayed_rebuffer(GdamSequencerSkin* skin)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
  GList *el;
  g_list_foreach (sequencer->elements, maybe_track_element, skin);
}

static gboolean is_blank(const char* string)
{
  while (*string && isspace (*string))
    string++;
  return *string == 0 ? TRUE : FALSE;
}

static void do_load(GdamSequencerSkin* skin)
{
  const char *filename;
  g_return_if_fail (skin->filename_entry != NULL);
  unselect_all_elements (skin);
  filename = gtk_entry_get_text (skin->filename_entry);
  g_return_if_fail (filename != NULL && !is_blank (filename));
  gdam_sequencer_load (get_sequencer (skin), filename);
  delayed_rebuffer (skin);
}

static void on_note_on(GdamMidiEventNote* note, 
		       GdamSequencerSkin* skin)
{
  GdamSequencer *sequencer = GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
  int buffer_id;
  double resample, gain;
  gboolean must_unref = FALSE;

  g_return_if_fail (local_id_valid (sequencer, skin->local_buffer));
  buffer_id = nth_local_id (sequencer, skin->local_buffer);

  if (!buffer_id)
    {
      buffer_id = gdam_buffer_manager_obtain_id (GDAM_MODEL (sequencer));
      must_unref = TRUE;
    }

  if (buffer_id == 0)
    {
      g_message ("no sample has been chosen -- cannot add from midi");
      return;
    }

  resample = (double) note->hertz / 440.0;
  gain = (double) note->velocity / 128.0;

  gdam_sequencer_add_element_immediate (sequencer,
					buffer_id, resample, gain, 0, -1);
  if (must_unref)
    gdam_buffer_manager_unref (buffer_id, (GdamModel *) sequencer);
  maybe_track_buffer_id (skin, buffer_id);
  delayed_redraw (skin);
}

static void do_midi(GdamSequencerSkin* skin)
{
  guint hook_id;
  GdamInputTrap *trap;
  GdamInputEvent *canon_event;
  GdamMidiDevice *midi_device;

  if (skin->input_hook_id)
    {
      gdam_debug ("sequencer: already have midi input, ignoring...");
      return;
    }
#if 0
  midi = gdam_midi_device_get_unique ("/dev/midi");
  if (midi == NULL)
    {
      g_warning ("Couldn't open /dev/midi");
      return;
    }
  gtk_signal_connect_while_alive (GTK_OBJECT (midi),
				  "note_on",
				  on_note_on, skin, (GtkObject *) skin);
#endif
  midi_device = (GdamMidiDevice *)
    gdam_input_device_get_by_type ("GdamMidiDevice");
  if (midi_device == NULL)
    {
      g_warning ("no midi devices detected");
      return;
    }
  canon_event = gdam_midi_event_note_new (midi_device, 0, 0, 0, TRUE, TRUE);
  trap = gdam_input_trap_find_or_create (canon_event);
  gdam_input_event_destroy (canon_event);
  hook_id = gdam_input_hook_connect (trap,
				     (GdamInputHookFunc) on_note_on,
				     NULL, skin, (GtkObject *) skin);
  skin->input_hook_id = hook_id;
}

static void do_save(GdamSequencerSkin* skin)
{
  const char *filename;
  g_return_if_fail (skin->filename_entry != NULL);
  filename = gtk_entry_get_text (skin->filename_entry);
  g_return_if_fail (filename != NULL && !is_blank (filename));
  gdam_sequencer_save (get_sequencer (skin), filename);
}

static void do_cache(GdamSequencerSkin* skin)
{
  gdam_sequencer_cache (get_sequencer (skin));
}

static void 
do_snap_y (GdamSequencerSkin* skin)
{
  skin->snap_y = TRUE;
}

static void 
do_free_y (GdamSequencerSkin* skin)
{
  skin->snap_y = FALSE;
}

static void 
do_scale_elements (GdamSequencerSkin* skin)
{
  skin->scale_element_lengths = TRUE;
}

static void 
undo_scale_elements (GdamSequencerSkin* skin)
{
  skin->scale_element_lengths = FALSE;
}


/*
 *  ____                   _                ___
 * |  _ \ __ _ _ __  _ __ (_)_ __   __ _   ( _ )
 * | |_) / _` | '_ \| '_ \| | '_ \ / _` |  / _ \/\
 * |  __/ (_| | | | | | | | | | | | (_| | | (_>  <
 * |_|   \__,_|_| |_|_| |_|_|_| |_|\__, |  \___/\/
 *                                 |___/
 *  _____                     _
 * |__  /___   ___  _ __ ___ (_)_ __   __ _
 *   / // _ \ / _ \| '_ ` _ \| | '_ \ / _` |
 *  / /| (_) | (_) | | | | | | | | | | (_| |
 * /____\___/ \___/|_| |_| |_|_|_| |_|\__, |
 *                                    |___/
 */

void gdam_sequencer_skin_display_full (GdamSequencerSkin *skin)
{
  skin->min_beat = 0;
  skin->max_beat = get_sequencer (skin)->num_beats;
  delayed_redraw (skin);
}

static void horizontal_zoom(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  skin->max_beat = skin->min_beat + adj->value;
  if (skin->time_pan_adj)
    {
      skin->time_pan_adj->page_size = adj->value;
      gtk_adjustment_changed (adj);
    }
  delayed_redraw (skin);
}

static void horizontal_pan(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  skin->max_beat += (adj->value - skin->min_beat);
  skin->min_beat = adj->value;
  delayed_redraw (skin);
}

static void vertical_zoom(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  skin->max_log_hertz = skin->min_log_hertz + adj->value;
  if (skin->vertical_pan_adj)
    {
      skin->vertical_pan_adj->page_size = adj->value;
      gtk_adjustment_changed (adj);
    }
  delayed_redraw (skin);
}

static void vertical_pan(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  skin->max_log_hertz += (adj->value - skin->min_log_hertz);
  skin->min_log_hertz = adj->value;
  delayed_redraw (skin);
}

/* 
 *  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */

typedef struct _IntInfo {
	GdamSequencerSkin* skin;
	int num;
} IntInfo;

static void on_beatnum_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);
static void on_tempo_adjusted(GtkAdjustment* adj, gpointer user_data);
static void on_sequencer_set_beat_info(GtkAdjustment* adj, 
				       const GdamBeatInfo* info, 
				       gpointer user_data);
static void on_tempo_string_drop(const char* tempo, gpointer user_data);
static void on_import_loop_by_id_drop (const char *id_string, gpointer data);
static void on_import_locked_local_loop_by_id_drop (const char *id_string, 
                                                    gpointer data);
static void on_buffer_id_drop(const char* id_string, gpointer user_data);
static void on_subbeats_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);
static void on_asynch_step_size_adjusted(GtkAdjustment* adj, 
					 GdamSequencerSkin* skin);
static void on_lag_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin);
static void copy_to_local_buffer_zero (GdamSequencerSkin *skin, guint src);

static void do_select_local_buffer(IntInfo* info)
{
  GdamSequencerSkin* skin = info->skin;
  GdamSequencer* sequencer = GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
  skin->local_buffer = info->num;
  gdam_arg_set (GTK_OBJECT (skin), "current_local_buffer", skin->local_buffer);
}

static void do_pattern_on(IntInfo* info)
{
  GdamSequencerSkin *skin = info->skin;
  GdamSequencer *sequencer = GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
  gdam_sequencer_pattern_on (sequencer, info->num);
  delayed_redraw (skin);
}

static void do_pattern_off(IntInfo* info)
{
  GdamSequencerSkin *skin = info->skin;
  GdamSequencer *sequencer = GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
  gdam_sequencer_pattern_off (sequencer, info->num);
  delayed_redraw (skin);
}

static void do_select_pattern(IntInfo* info)
{
  GdamSequencerSkin *skin = info->skin;
  GdamSequencer *sequencer = GDAM_SEQUENCER (GDAM_SKIN (skin)->model);
  gdam_sequencer_set_pattern (sequencer, info->num);
  delayed_redraw (skin);
}

/* Search the sequence for an element with a certain local id at a certain
 * location.  If pattern is >= 0 then we look for an element with that
 * pattern, else we match an element with the current pattern.
 */
static GdamSequencerElement *match_element (GdamSequencerSkin *skin, 
    					    int 	       local_id, 
					    gdouble            time,
					    int		       pattern)
{
  GList *elements;
  GList *el;
  GdamSequencerElement *match;
  elements = get_sequencer (skin)->elements;
  match = NULL;
    /* for each sequencer element, check if it has our buffer id and the
     * same position as our asynch marker.
     */ 
  for (el=elements; el != NULL && !match; el = el->next)
    {
      GdamSequencerElement *seqel;
      seqel = (GdamSequencerElement *)el->data;
      g_return_val_if_fail (seqel != NULL, NULL);
      if (seqel->local_id == local_id) 
	{
	  if ((pattern < 0 && 
	        seqel->pattern_number == get_sequencer (skin)->current_pattern)
	      || pattern == seqel->pattern_number)
	    {
	      if (seqel->beat_offset > 0.99 * time && 
		  seqel->beat_offset < 1.01 * time)
		match = seqel;
	      else if (!seqel->beat_offset && !time)
		match = seqel;
	    }
	}
    }
  return match;
}

#define TOGGLE_DISPLAY_FOLLOW_TIME 0
#define TOGGLE_DISPLAY_FOLLOW_SAMPLE 1

static void update_sample_toggle (gpointer data, gpointer udata)
{
  IntToggleInfo *ti;
  GdamSequencerSkin *skin;
  GdamSequencer *sequencer;
  gboolean act = FALSE;
  gboolean act2;
  guint buffer_id;
  guint local_id;
  ti = (IntToggleInfo*)data;
  local_id = ti->val;
  skin = GDAM_SEQUENCER_SKIN (udata);
  sequencer = get_sequencer (skin);

  extend_local_ids (sequencer, local_id + 1);
  g_return_if_fail (local_id_valid (sequencer, ti->val));
  buffer_id = nth_local_id (sequencer, ti->val);

  if (skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_TIME)
    {
      /* Try to find an element at this time which has our id */
      act = (int)match_element (skin, local_id, skin->asynch_position, -1);
    }
  else if (skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_SAMPLE)
    {
      /* Check whether our id is the current local sample id */
      if (ti->val >= 0 && ti->val == skin->current_local_sample_id)
	act = TRUE;
    }
  /* Update the toggle's state if needed */
  act2 = gtk_toggle_button_get_active (ti->tbutton);
  if ((act2 && !act) || (act && !act2))
    {
      ti->ignore_button = TRUE;
      gtk_toggle_button_set_active (ti->tbutton, act);
      ti->ignore_button = FALSE;
    }
}


static void update_time_toggle (gpointer data, gpointer udata)
{
  FloatToggleInfo *ti;
  GdamSequencerSkin *skin;
  gboolean act = FALSE;
  gboolean act2;
  gdouble position;
  ti = (FloatToggleInfo*)data;
  skin = GDAM_SEQUENCER_SKIN (udata);
  position = ti->val + skin->base_offset;

  if (skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_TIME)
    {
      /* See if our time matches the asynch position */
      if (position > 0.99 * skin->asynch_position && 
	  position < 1.01 * skin->asynch_position)
	act = TRUE;
      else if (!position && !skin->asynch_position)
	act = TRUE;

    }
  else if (skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_SAMPLE)
    {
      /* Try to find an element at this time which has the current sample id */
      act = (gboolean) match_element (skin, skin->current_local_sample_id, 
	  			      position, -1);
    }

  /* Update the toggle state if needed */
  act2 = gtk_toggle_button_get_active (ti->tbutton);
  if ((act2 && !act) || (act && !act2))
  {
    ti->ignore_button = TRUE;
    gtk_toggle_button_set_active (ti->tbutton, act);
    ti->ignore_button = FALSE;
  }
}

/* Sets the current local id and updates all displays. */
static void set_current_local_sample_id (GdamSequencerSkin *skin, int new_id)
{
  skin->current_local_sample_id = new_id;
  /* update state toggles. */
  if (skin->time_state_toggles)
    g_list_foreach (skin->time_state_toggles, update_time_toggle, skin);
  if (skin->sample_state_toggles)
    g_list_foreach (skin->sample_state_toggles, update_sample_toggle, skin);
}

/* Sets the asynch position to a specific time, and updates all displays. */
static void set_asynch_position (GdamSequencerSkin *skin, gdouble pos)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  gdouble ap;
  skin->asynch_position = pos;
  /* update the display which tells which measure we're in. */
  ap = skin->asynch_position - skin->base_offset;
  /* XXX: although we don't call it in that context (yet) this should
   * handle updating beat_offset for any new position not just +-1 
   */
  if (skin->base_offset_factor && ap >= skin->base_offset_factor)
    {
      skin->base_offset += skin->base_offset_factor;
      if (skin->base_offset_adj)
	gtk_adjustment_set_value (skin->base_offset_adj, 
	    			  skin->base_offset_adj->value + 1);
    }
  else if (skin->base_offset_factor && 
      ap < 0)
    {
      skin->base_offset -= skin->base_offset_factor;
      if (skin->base_offset_adj)
	gtk_adjustment_set_value (skin->base_offset_adj, 
	    			  skin->base_offset_adj->value - 1);
    }
  /* update state toggles. */
  if (skin->time_state_toggles)
    g_list_foreach (skin->time_state_toggles, update_time_toggle, skin);
  if (skin->sample_state_toggles)
    g_list_foreach (skin->sample_state_toggles, update_sample_toggle, skin);
  if (skin->asynch_position_label)
    {
      char txt[16];
      sprintf (txt, "%3.2f", skin->asynch_position);
      gtk_label_set_text (skin->asynch_position_label, txt);
    }
}

/* Sets the asynch position to a specific time relative to base_offset. */
static void set_asynch_local_position (GdamSequencerSkin *skin, gdouble pos)
{
  set_asynch_position (skin,  skin->base_offset + pos);
}

/* Move the asynch time pointer to the next time point (as defined
 * by asynch_step_size) and checks whether we need to wrap.
 */
static void do_step_asynch(GdamSequencerSkin* skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  gdam_debug ("stepping asynchronously.");
  skin->asynch_position += skin->asynch_step_size;
  if (skin->asynch_position >= sequencer->num_beats)
    {
      skin->asynch_position -= sequencer->num_beats;
      skin->base_offset = 0;
      if (skin->base_offset_adj)
	gtk_adjustment_set_value (skin->base_offset_adj, 0);
    }
  set_asynch_position (skin, skin->asynch_position);
}

static void internal_add_asynch(GdamSequencerSkin* skin, int buffer_id)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  GdamSequencerElement *sample_info;
  if (!buffer_id && skin->local_buffer)
    buffer_id = skin->local_buffer;
  else if (!buffer_id)
    {
      g_warning ("missing buffer id in add_asynch");
      buffer_id = gdam_buffer_manager_obtain_id (GDAM_MODEL (sequencer));
    }
  sample_info = add_new_element 
			   (skin, 
			    buffer_id,
			    skin->asynch_position, 
			    skin->default_resample,
			    skin->default_gain); 
  sample_info->local_id = skin->local_buffer;
  if (skin->asynch_autostep)
    {
      gdam_debug ("adding asynch: autostepping.");
      do_step_asynch (skin);
    }
  delayed_redraw (skin);
}

static void do_add_asynch(GdamSequencerSkin* skin){
  internal_add_asynch (skin, 0);
}

static void do_autostep_asynch(GdamSequencerSkin* skin)
{
  gboolean rv;
  gdam_debug ("autostep toggled.");
  if (!skin->autostep_toggle_button)
    return;
  rv = gtk_toggle_button_get_active (skin->autostep_toggle_button);
  skin->asynch_autostep = rv;
}

static void do_add_immediate(IntInfo* info){
	GdamSequencerSkin* skin = info->skin;
	GdamSequencer* sequencer = GDAM_SEQUENCER(GDAM_SKIN(skin)->model);
	int buffer_id;
	gboolean must_unref = FALSE;

	if (info->num > 0 && local_id_valid (sequencer, info->num))
	  buffer_id = nth_local_id (sequencer, info->num);
	else {
		buffer_id = gdam_buffer_manager_obtain_id(
						GDAM_MODEL(sequencer));
		must_unref = TRUE;
	}

	if (buffer_id == 0) {
		g_message("no sample has been chosen -- cannot add immediate");
		return;
	}

	gdam_sequencer_add_element_immediate(sequencer,
					     buffer_id,
					     skin->default_resample,
					     skin->default_gain,
					     0,
					     -1);
	if (must_unref)
		gdam_buffer_manager_unref(buffer_id, (GdamModel*)sequencer);
	delayed_redraw(skin);
}

/* Allow access to viewport's handlers */
static gboolean 
gdam_sequencer_defer_to_cur_viewport_handler
                                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamSequencerSkin *seq_skin = GDAM_SEQUENCER_SKIN (skin);
  GdamSkin *subskin = GDAM_SKIN (seq_skin->cur_local_buffer_viewport);
  if (!subskin)
    {
      g_warning ("missing subskin in %s", handler_names[0]);
      return FALSE;
    }
  if (!gdam_skin_do_connect (subskin,
			     object,
			     signal_name,
			     num_handler_names - 1, 
			     handler_names + 1))
    {
      g_warning ("gdam_skin_do_connect failed on deferred handler: "
		 "num_args=%d, handler_name=%s, skintype=%s",
		 num_handler_names,
		 handler_names[1],
		 gtk_type_name (GTK_OBJECT (subskin)->klass->type));
      return FALSE;
    }
  return TRUE;
}



static gboolean gdam_sequencer_drawing_area_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_DRAWING_AREA (object))
    {
      g_warning ("%s may only be connected to a drawing area",
		 handler_names[0]);
    }
  connect_drawing_area (GTK_DRAWING_AREA (object),
			GDAM_SEQUENCER_SKIN (skin));
  return TRUE;
}

static gboolean gdam_sequencer_tempo_drop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_tempo_string_drop, "gdam-tempo-text", skin);
  return TRUE;
}

static gboolean 
gdam_sequencer_hide_inactive_handler  (GdamSkin          *skin,
				       GtkObject         *object,
				       const char        *signal_name,
				       int                num_handler_names,
				       char             **handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_hide_inactive),
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*do_hide_inactive) (GDAM_SEQUENCER_SKIN (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_show_inactive_handler  (GdamSkin          *skin,
				       GtkObject         *object,
				       const char        *signal_name,
				       int                num_handler_names,
				       char             **handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_show_inactive),
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*do_show_inactive) (GDAM_SEQUENCER_SKIN (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_size_from_length_handler  (GdamSkin          *skin,
				          GtkObject         *object,
				          const char        *signal_name,
				          int                num_handler_names,
				          char             **handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_size_from_length),
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*do_size_from_length) (GDAM_SEQUENCER_SKIN (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_constant_size_handler  (GdamSkin          *skin,
				       GtkObject         *object,
				       const char        *signal_name,
				       int                num_handler_names,
				       char             **handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (undo_size_from_length),
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*undo_size_from_length) (GDAM_SEQUENCER_SKIN (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_sort_by_handler  (GdamSkin          *skin,
				 GtkObject         *object,
				 const char        *signal_name,
				 int                num_handler_names,
				 char             **handler_names)
{
  GtkSignalFunc signal_func;
  signal_func = GTK_SIGNAL_FUNC (sort_by_id);
  if (strstr (handler_names[0], "by_pattern"))
    {
      signal_func = sort_by_pattern;
    }
  if (strstr (handler_names[0], "by_length"))
    {
      signal_func = sort_by_length;
    }
  if (strstr (handler_names[0], "by_start"))
    {
      signal_func = sort_by_start;
    }
  if (strstr (handler_names[0], "by_local_id"))
    {
      signal_func = sort_by_local_id;
    }
  if (strstr (handler_names[0], "by_id"))
    {
      signal_func = sort_by_id;
    }
  if (strstr (handler_names[0], "by_gain"))
    {
      signal_func = sort_by_gain;
    }
  if (strstr (handler_names[0], "by_resample"))
    {
      signal_func = sort_by_resample;
    }
  gtk_signal_connect_object (object,
			     signal_name,
			     signal_func,
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*signal_func) (GTK_OBJECT (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_render_as_handler  (GdamSkin          *skin,
				   GtkObject         *object,
				   const char        *signal_name,
				   int                num_handler_names,
				   char             **handler_names)
{
  GtkSignalFunc signal_func;
  if (num_handler_names < 2)
    {
      g_warning ("need argument render_method (i.e. 'circles') to %s",
	  	 handler_names[0]);
      return TRUE;
    }
  if (g_strcasecmp (handler_names[1], "circles") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (render_as_circles);
    }
  else if (g_strcasecmp (handler_names[1], "rectangles") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (render_as_rectangles);
    }
  else if (g_strcasecmp (handler_names[1], "colored_squares") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (render_as_colored_squares);
    }
  else if (g_strcasecmp (handler_names[1], "tee") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (render_as_tee);
    }
  else if (g_strcasecmp (handler_names[1], "pennant") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (render_as_pennant);
    }
  else 
    {
      g_warning ("unknown render method %s in handler %s", handler_names[1],
	         handler_names[0]);
      return TRUE;
    }
  gtk_signal_connect_object (object,
			     signal_name,
			     signal_func,
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*signal_func) (GTK_OBJECT (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_text_as_handler  (GdamSkin          *skin,
				 GtkObject         *object,
				 const char        *signal_name,
				 int                num_handler_names,
				 char             **handler_names)
{
  GtkSignalFunc signal_func;
  if (num_handler_names < 2)
    {
      g_warning ("need argument text_mode (i.e. 'buffer_name') to %s",
	  	 handler_names[0]);
      return TRUE;
    }
  if (g_strcasecmp (handler_names[1], "buffer_name") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (text_as_buffer_name);
    }
  else if (g_strcasecmp (handler_names[1], "local_id") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (text_as_local_id);
    }
  else 
    {
      g_warning ("unknown text method %s in handler %s", handler_names[1],
	         handler_names[0]);
      return TRUE;
    }
  gtk_signal_connect_object (object,
			     signal_name,
			     signal_func,
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*signal_func) (GTK_OBJECT (skin));
  return TRUE;
}

static gboolean 
gdam_sequencer_display_tool_handler  (GdamSkin          *skin,
				      GtkObject         *object,
				      const char        *signal_name,
				      int                num_handler_names,
				      char             **handler_names)
{
  GtkSignalFunc signal_func;
  if (num_handler_names < 2)
    {
      g_warning ("need argument tool_name (i.e. 'stamp') to %s",
	  	 handler_names[0]);
      return TRUE;
    }
  if (g_strcasecmp (handler_names[1], "add") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (set_tool_add);
    }
  else if (g_strcasecmp (handler_names[1], "fill_nearest") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (set_tool_fill_nearest);
    }
  else if (g_strcasecmp (handler_names[1], "trace_y_value") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (set_tool_trace_y_value);
    }
  else if (g_strcasecmp (handler_names[1], "fill_nearest_and_stretch") == 0)
    {
      signal_func = GTK_SIGNAL_FUNC (set_tool_fill_nearest_and_stretch);
    }
  else 
    {
      g_warning ("unknown display tool %s in handler %s", handler_names[1],
	         handler_names[0]);
      return TRUE;
    }
  gtk_signal_connect_object (object,
			     signal_name,
			     signal_func,
			     (GtkObject *) skin);
  if (GTK_IS_RADIO_BUTTON (object) &&
      gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    (*signal_func) (GTK_OBJECT (skin));
  return TRUE;
}



static gboolean gdam_sequencer_play_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_producer_play_now),
			     (GtkObject *) GDAM_SEQUENCER (skin->model));
  return TRUE;
}

static gboolean gdam_sequencer_asynch_position_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *s_skin = (GdamSequencerSkin *) skin;
  s_skin->asynch_position_label = GTK_LABEL (object);
  return TRUE;
}

static void
solo_pattern (GdamSequencerSkin *skin, int pattern_number)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  int i;

  gdam_sequencer_freeze (sequencer);
  for (i = 0; i < skin->num_patterns; i++)
    if (i != pattern_number)
      gdam_sequencer_pattern_off (sequencer, i);
  gdam_sequencer_pattern_on (sequencer, pattern_number);
  gdam_sequencer_thaw (sequencer);
  delayed_redraw (skin);
}

static gint 
pattern_press (GtkObject *obj, GdkEventButton *event, IntToggleInfo *pi)
{
  /* A button-two click turns off all other patterns, and select the clicked
   * pattern for editing. */
  if (event->button == 2)
    {
      solo_pattern (pi->skin, pi->val);
      return 1;
    }
  return 1;
}

static void pattern_toggled (GtkObject *obj, IntToggleInfo *pi)
{
  GdamSequencer *seq;
  seq = get_sequencer (pi->skin);
  g_return_if_fail (seq != NULL);
  g_return_if_fail (GDAM_IS_SEQUENCER (seq));
  if (pi->ignore_button)
    return;
  if (gtk_toggle_button_get_active (pi->tbutton))
    gdam_sequencer_pattern_on (seq, pi->val);
  else
    gdam_sequencer_pattern_off (seq, pi->val);
}

static void time_toggled (GtkObject *obj, FloatToggleInfo *pi)
{
  gboolean act = gtk_toggle_button_get_active (pi->tbutton);
  if (pi->ignore_button)
    return;

  /* Set our time to be lead, a click always leaves this button active. */
  if (pi->skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_TIME)
    {
      if (!act)
	{
	  pi->ignore_button = TRUE;
	  gtk_toggle_button_set_active (pi->tbutton, TRUE);
	  pi->ignore_button = FALSE;
	}
      set_asynch_local_position (pi->skin, pi->val);
    }
  /* Toggle the existance of a sample of current_local_id at our position. */
  else if (pi->skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_SAMPLE)
    {
      if (act) 
	{
	  GdamSequencerElement *sample_info;
	  GdamSequencer *seq = get_sequencer (pi->skin);
	  int local_id = pi->skin->current_local_sample_id;
	  int buffer_id = 0;

	  if (local_id >= 0 && local_id_valid (seq, local_id))
	    buffer_id = nth_local_id (seq, local_id);

	  /* If an id hasn't been assigned, return and leave unactive */
	  if (!buffer_id)
	    {
	      if (act)
		{
		  pi->ignore_button = TRUE;
		  gtk_toggle_button_set_active (pi->tbutton, FALSE);
		  pi->ignore_button = FALSE;
		}
	      return;
	    }
	  gdam_sequencer_freeze (seq);
	  sample_info = add_new_element (pi->skin,
					 buffer_id, 
					 pi->skin->base_offset + pi->val,
					 pi->skin->default_resample,
					 pi->skin->default_gain);
	  sample_info->local_id = pi->skin->current_local_sample_id;
	  gdam_sequencer_thaw (seq);
	}
      else
	{
	  GdamSequencerElement *match = NULL;
	  match = match_element (pi->skin, pi->skin->current_local_sample_id, 
	                         pi->skin->base_offset + pi->val, -1);
	  if (match)
	    delete_subinfo (pi->skin, match);
	}
    }
}

static void sample_toggled (GtkObject *obj, IntToggleInfo *pi)
{
  GdamSequencer *seq;
  int act;
  int buffer_id = 0;
  seq = get_sequencer (pi->skin);
  g_return_if_fail (seq != NULL);
  g_return_if_fail (GDAM_IS_SEQUENCER (seq));
  if (pi->ignore_button)
    return;
  act =  gtk_toggle_button_get_active (pi->tbutton);

  if (pi->val >= 0 && local_id_valid (seq, pi->val))
    buffer_id = nth_local_id (seq, pi->val);

  /* If an id hasn't been assigned, return and leave unactive */
  if (!buffer_id)
    {
      if (act)
	{
	  pi->ignore_button = TRUE;
	  gtk_toggle_button_set_active (pi->tbutton, FALSE);
	  pi->ignore_button = FALSE;
	}
      return;
    }

  /* Toggle the existance of a sample of our buffer_id at current position. */
  if (pi->skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_TIME)
    {
      if (act) 
	{
	  internal_add_asynch (pi->skin, buffer_id);
	}
      else
	{
	  GdamSequencerElement *match = NULL;
	  match = match_element (pi->skin, pi->val, pi->skin->asynch_position,
	      			 -1);
	  if (match)
	    delete_subinfo (pi->skin, match);
	}
    }
  /* Set our buffer_id to be lead, a click always leaves this button active. */
  else if (pi->skin->toggle_display_mode == TOGGLE_DISPLAY_FOLLOW_SAMPLE)
    {
      if (!act)
	{
	  pi->ignore_button = TRUE;
	  gtk_toggle_button_set_active (pi->tbutton, TRUE);
	  pi->ignore_button = FALSE;
	}
      set_current_local_sample_id (pi->skin, pi->val);
    }
}

static gboolean gdam_sequencer_sample_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin;
  GtkToggleButton *toggle_button;
  IntToggleInfo *ti;
  GdamSequencer *seq;
  int num;
  char *endp;

  seq_skin = GDAM_SEQUENCER_SKIN (skin);
  seq = get_sequencer (seq_skin);
  toggle_button = GTK_TOGGLE_BUTTON (object);

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  ti = g_new0 (IntToggleInfo, 1);
  ti->val = num;
  ti->tbutton = toggle_button;
  ti->skin = seq_skin;
  seq_skin->sample_state_toggles = 
    			g_list_append (seq_skin->sample_state_toggles, ti);
  gtk_signal_connect_while_alive (object, 
				  "toggled", 
				  GTK_SIGNAL_FUNC (sample_toggled),
				  ti,
				  object);
  return TRUE;
}


static gboolean gdam_sequencer_time_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin;
  GtkToggleButton *toggle_button;
  FloatToggleInfo *ti;
  GdamSequencer *seq;
  gdouble num;
  char *endp;

  seq_skin = GDAM_SEQUENCER_SKIN (skin);
  seq = get_sequencer (seq_skin);
  toggle_button = GTK_TOGGLE_BUTTON (object);

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  ti = g_new0 (FloatToggleInfo, 1);
  ti->val = num;
  ti->tbutton = toggle_button;
  ti->skin = seq_skin;
  seq_skin->time_state_toggles = g_list_append (seq_skin->time_state_toggles, 
      						ti);
  gtk_signal_connect_while_alive (object, 
				  "toggled", 
				  GTK_SIGNAL_FUNC (time_toggled),
				  ti,
				  object);
  return TRUE;
}

static gboolean gdam_sequencer_pattern_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin;
  GtkToggleButton *toggle_button;
  IntToggleInfo *pi;
  GdamSequencer *seq;
  int num;
  char *endp;

  seq_skin = GDAM_SEQUENCER_SKIN (skin);
  seq = get_sequencer (seq_skin);
  toggle_button = GTK_TOGGLE_BUTTON (object);

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  pi = g_new0 (IntToggleInfo, 1);
  pi->val = num;
  pi->tbutton = toggle_button;
  pi->skin = seq_skin;
  seq_skin->pattern_toggles = g_list_append (seq_skin->pattern_toggles, pi);
  gtk_signal_connect_while_alive (object, 
				  "toggled", 
				  GTK_SIGNAL_FUNC (pattern_toggled),
				  pi,
				  object);
  gtk_signal_connect_while_alive (object, 
				  "button-press-event", 
				  GTK_SIGNAL_FUNC (pattern_press),
				  pi,
				  object);
  /* XXX: gdam_util_is_active?? (and add gtkitem) */
  if (GTK_IS_TOGGLE_BUTTON (object) 
      && gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (object)))
    {
      gdam_sequencer_pattern_on (seq, num);
    }
  else if (GTK_IS_PIX_BUTTON (object) 
      && gtk_pix_button_get_active (GTK_PIX_BUTTON (object)))
    {
      gdam_sequencer_pattern_on (seq, num);
    }

  return TRUE;
}

static gboolean gdam_sequencer_autostep_asynch_tbutton_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin;
  seq_skin = GDAM_SEQUENCER_SKIN (skin);
  seq_skin->autostep_toggle_button = GTK_TOGGLE_BUTTON (object);
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_autostep_asynch),
			     (GtkObject *) skin);
  return TRUE;
}

static gboolean gdam_sequencer_add_immediate_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  IntInfo *info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      num = 0;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_add_immediate),
			     (GtkObject *) info);
  return TRUE;
}

static gboolean gdam_sequencer_local_buffer_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  guint num;
  IntInfo *info;
  GdamSequencerSkin *sskin = GDAM_SEQUENCER_SKIN(skin);
  char *endp;

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      int snum;
      snum = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
      if (snum < 0)
	g_warning ("%s takes a positive integer", handler_names[0]);
      num = snum;
    }

  if (!GTK_IS_LABEL (object)) 
    {
      g_warning ("%s only valid for labels", handler_names[0]);
      return TRUE;
    }

  if (num >= sskin->local_id_labels->len)
    sskin->local_id_labels = g_array_set_size (sskin->local_id_labels, num+1);
  g_array_index (sskin->local_id_labels, gpointer, num) = GTK_LABEL (object);

  return TRUE;
}


static gboolean gdam_sequencer_drop_local_buffer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  IntInfo *info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_buffer_id_drop,
			       "gdam-buffer-id-text", (GtkObject *) info);
  return TRUE;
}


static void on_cur_buffer_id_drop (const char* id_string, gpointer data);
static gboolean 
drop_cur_local_buffer_handler   (GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_cur_buffer_id_drop,
			       "gdam-buffer-id-text", (GtkObject *) skin);
  return TRUE;
}



static gboolean
gdam_sequencer_drop_loop_handler (GdamSkin          *skin,
				  GtkObject         *object,
				  const char        *signal_name,
				  int                num_handler_names,
				  char             **handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_import_loop_by_id_drop,
			       "gdam-buffer-id-text", skin);
  return TRUE;
}

static gboolean
gdam_sequencer_drop_locked_local_loop_handler 
                                 (GdamSkin          *skin,
				  GtkObject         *object,
				  const char        *signal_name,
				  int                num_handler_names,
				  char             **handler_names)
{
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_import_locked_local_loop_by_id_drop,
			       "gdam-buffer-id-text", skin);
  return TRUE;
}

static gboolean gdam_sequencer_select_local_buffer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  IntInfo *info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_select_local_buffer),
			     (GtkObject *) info);
  gdam_widget_string_dest_set (GTK_WIDGET (object),
			       on_buffer_id_drop,
			       "gdam-buffer-id-text", (GtkObject *) info);
  return TRUE;
}

static void 
do_copy_local_buffer (IntInfo* info)
{
  copy_to_local_buffer_zero (info->skin, (guint)info->num);
}

static gboolean gdam_sequencer_copy_local_buffer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  IntInfo *info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (do_copy_local_buffer),
			     (GtkObject *) info);
  return TRUE;
}


static gboolean gdam_sequencer_pattern_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  int num;
  GtkSignalFunc signal_func = NULL;
  IntInfo *info;
  char *endp;

  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }

  info = g_new (IntInfo, 1);
  info->skin = GDAM_SEQUENCER_SKIN (skin);
  info->num = num;

  if (strcmp (handler_names[0], "gdam_sequencer_select_pattern") == 0)
    signal_func = GTK_SIGNAL_FUNC (do_select_pattern);
  else if (strcmp (handler_names[0], "gdam_sequencer_pattern_on") == 0)
    signal_func = GTK_SIGNAL_FUNC (do_pattern_on);
  else if (strcmp (handler_names[0], "gdam_sequencer_pattern_off") == 0)
    signal_func = GTK_SIGNAL_FUNC (do_pattern_off);
  else
    {
      g_warning ("unhandled handler %s", handler_names[0]);
      return TRUE;
    }

  gtk_signal_connect_object (object,
			     signal_name,
			     signal_func,
			     (GtkObject *) info);
  return TRUE;
}


static gboolean gdam_sequencer_stop_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  gtk_signal_connect_object (object,
			     signal_name,
			     GTK_SIGNAL_FUNC (gdam_producer_stop),
			     (GtkObject *) GDAM_SEQUENCER (skin->model));
  return TRUE;
}

static gboolean gdam_sequencer_set_lag_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_lag_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_set_asynch_step_size_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_asynch_step_size_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_set_sub_beats_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_subbeats_adjusted), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void on_set_num_beats(GdamSequencer* seq, gdouble num_beats, 
			     GtkAdjustment* adj)
  {
    gdam_debug ("on_set_num_beats to %f=%f", num_beats, seq->num_beats);
    gtk_adjustment_set_value(adj, num_beats);
  }

static gboolean gdam_sequencer_beatnum_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_beatnum_adjusted), skin);
  gtk_signal_connect (GTK_OBJECT (get_sequencer (GDAM_SEQUENCER_SKIN (skin))),
		      "set_num_beats",
		      GTK_SIGNAL_FUNC (on_set_num_beats), adj);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static void update_pattern_toggle (gpointer data, gpointer udata)
{
  IntToggleInfo *pi;
  GdamSequencer *seq;
  gboolean act;
  gboolean act2;
  pi = (IntToggleInfo*)data;
  seq = GDAM_SEQUENCER (udata);
  act = gdam_sequencer_pattern_get_active (seq, pi->val);
  act2 = gtk_toggle_button_get_active (pi->tbutton);
  if ((act2 && !act) || (act && !act2))
  {
    pi->ignore_button = TRUE;
    gtk_toggle_button_set_active (pi->tbutton, act);
    pi->ignore_button = FALSE;
  }
}

static void on_toggle_offset_adjusted (GtkAdjustment* adj, gpointer user_data)
{
  GdamSequencerSkin *skin = (GdamSequencerSkin *) user_data;
  GdamSequencer *sequencer = get_sequencer (skin);
  gdouble target_time;
  int old_offset = skin->base_offset;
  int delta_time;
  skin->base_offset = adj->value * skin->base_offset_factor;
  delta_time = skin->base_offset - old_offset;
  target_time = skin->asynch_position + delta_time;
  if (delta_time > 0.01 || delta_time < 0.01) 
    set_asynch_position (skin, target_time);
}

static gboolean gdam_sequencer_toggle_offset_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin = GDAM_SEQUENCER_SKIN (skin);
  GtkAdjustment *adj;
  char *endp;
  int num;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  seq_skin->base_offset_factor = num;
  seq_skin->base_offset_adj = adj;
  seq_skin->base_offset = adj->value * num;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_toggle_offset_adjusted), 
		      skin);
  return TRUE;
}

static gboolean gdam_sequencer_tempo_adjuster_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *seq_skin = GDAM_SEQUENCER_SKIN (skin);
  GtkAdjustment *adj;
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (on_tempo_adjusted), skin);
  gtk_signal_connect_object (GTK_OBJECT (skin->model),
			     "set_beat_info",
			     GTK_SIGNAL_FUNC (on_sequencer_set_beat_info),
			     GTK_OBJECT (adj));
  seq_skin->tempo_adj = adj;
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_h_zoom_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamSequencerSkin *sequencer_skin = GDAM_SEQUENCER_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  sequencer_skin->time_zoom_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (horizontal_zoom), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_h_pan_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamSequencerSkin *sequencer_skin = GDAM_SEQUENCER_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  sequencer_skin->time_pan_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed",
		      GTK_SIGNAL_FUNC (horizontal_pan), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_v_zoom_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamSequencerSkin *sequencer_skin = GDAM_SEQUENCER_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  sequencer_skin->vertical_zoom_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed", GTK_SIGNAL_FUNC (vertical_zoom), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_v_pan_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GtkAdjustment *adj;
  GdamSequencerSkin *sequencer_skin = GDAM_SEQUENCER_SKIN (skin);
  adj = util_get_widget_adjustment (GTK_WIDGET (object));
  sequencer_skin->vertical_pan_adj = adj;
  gtk_signal_connect (GTK_OBJECT (adj),
		      "value_changed", GTK_SIGNAL_FUNC (vertical_pan), skin);
  gdam_delayed_adjustment_init (adj);
  return TRUE;
}

static gboolean gdam_sequencer_filename_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  GdamSequencerSkin *sequencer_skin = GDAM_SEQUENCER_SKIN (skin);
  sequencer_skin->filename_entry = GTK_ENTRY (object);
  return TRUE;
}

static gboolean gdam_sequencer_kill_by_local_id_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (delete_buffer_by_local_info), 
		      info);
  return TRUE;
}



static gboolean gdam_sequencer_transpose_up_by_local_id_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (transpose_buffer_up_by_local_info), 
		      info);
  return TRUE;
}


static gboolean gdam_sequencer_regain_up_by_local_id_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (regain_buffer_up_by_local_info), 
		      info);
  return TRUE;
}

static gboolean gdam_sequencer_regain_down_by_local_id_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (regain_buffer_down_by_local_info), 
		      info);
  return TRUE;
}


static gboolean gdam_sequencer_mute_by_local_id_toggle_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (!GTK_IS_TOGGLE_BUTTON (object) && !GTK_IS_PIX_BUTTON (object))
    {
      g_warning ("handler %s only valid with toggle buttons", handler_names[0]);
      return TRUE;
    }
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (by_local_info_mute_toggled), 
		      info);
  return TRUE;
}


static gboolean gdam_sequencer_transpose_down_by_local_id_handler
				(GdamSkin*          _skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  BufIdPtr *info;
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (_skin);
  char *endp;
  int num;
  if (handler_names[1] == NULL)
    {
      g_warning ("Missing needed arg to %s", handler_names[0]);
      return TRUE;
    }
  else
    {
      num = g_strtod (handler_names[1], &endp);
      if (endp == handler_names[1] || *endp != '\0')
	return TRUE;
    }
  info = g_new0 (BufIdPtr, 1);
  info->buffer_id = num;
  info->sequencer_skin = skin;
  gtk_signal_connect (object,
		      signal_name,
		      GTK_SIGNAL_FUNC (transpose_buffer_down_by_local_info), 
		      info);
  return TRUE;
}

static gboolean gdam_sequencer_buffer_container_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
  if (!GTK_IS_BOX (object))
    {
      g_warning ("arg to %s must be a box", handler_names[0]);
      return TRUE;
    }
  if (GTK_IS_BIN (object))
    {
      g_warning ("arg to %s must not be a bin", handler_names[0]);
      return TRUE;
    }
  GDAM_SEQUENCER_SKIN (skin)->buffer_id_holder = GTK_BOX (object);
  return TRUE;
}

static gboolean 
gdam_sequencer_pattern_list_handler (GdamSkin          *skin,
				     GtkObject         *object,
				     const char        *signal_name,
				     int                num_handler_names,
				     char             **handler_names)
{
  if (!GTK_IS_LIST (object))
    {
      g_warning ("arg to %s must be a list", handler_names[0]);
      return TRUE;
    }
  GDAM_SEQUENCER_SKIN (skin)->pattern_list = GTK_WIDGET (object);
  return TRUE;
}

static gboolean 
gdam_sequencer_pattern_name_entry_handler (GdamSkin          *skin,
				           GtkObject         *object,
				           const char        *signal_name,
				           int                num_handler_names,
				           char             **handler_names)
{
  if (!GTK_IS_ENTRY (object))
    {
      g_warning ("arg to %s must be an entry", handler_names[0]);
      return TRUE;
    }
  GDAM_SEQUENCER_SKIN (skin)->pattern_name_entry = GTK_WIDGET (object);
  return TRUE;
}

static void 
file_sel_cancelled (GdamSequencerSkin *skin)
{
  gtk_object_destroy (GTK_OBJECT (skin->file_sel));
}

static void 
file_sel_destroyed (GdamSequencerSkin *skin)
{
  gtk_object_unref (GTK_OBJECT (skin));
  skin->file_sel = NULL;
}

static void 
file_sel_load (GdamSequencerSkin *skin)
{
  const char *filename;
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (skin->file_sel));
  if (skin->filename_entry)
    gtk_entry_set_text (skin->filename_entry, filename);
  gdam_sequencer_load (get_sequencer (skin), filename);
  gdam_history_set_entry ("Sequencer", "LastSequenceSavePath", filename);
  gtk_object_destroy (GTK_OBJECT (skin->file_sel));
  skin->min_beat = 0;
  skin->max_beat = get_sequencer (skin)->num_beats;
  delayed_redraw (skin);
}

static void 
file_sel_save (GdamSequencerSkin *skin)
{
  const char *filename;
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (skin->file_sel));
  if (skin->filename_entry)
    gtk_entry_set_text (skin->filename_entry, filename);
  gdam_sequencer_save (get_sequencer (skin), filename);
  gdam_history_set_entry ("Sequencer", "LastSequenceSavePath", filename);
  gtk_object_destroy (GTK_OBJECT (skin->file_sel));
}

static void 
make_file_dialog   (GdamSequencerSkin *skin,
		    GtkSignalFunc      sig_func,
		    const char        *title)
{
  const char *last_dir;
  if (skin->file_sel)
    {
      g_warning ("Cannot create file dialog while other dialog is still open");
      return;
    }

  skin->file_sel = (GtkFileSelection *)gtk_file_selection_new (title);

  gtk_object_ref (GTK_OBJECT (skin));
  
  /* XXX: object_unref??  clear skin->file_sel?? */
  gtk_signal_connect_full (GTK_OBJECT (skin->file_sel),
			   "destroy",
			   GTK_SIGNAL_FUNC (file_sel_destroyed),
			   NULL, GTK_OBJECT (skin), 
			   GTK_SIGNAL_FUNC (file_sel_destroyed), 
			   FALSE, FALSE);
  gtk_signal_connect_object (GTK_OBJECT (skin->file_sel->ok_button),
			     "clicked",
			     GTK_SIGNAL_FUNC (sig_func),
			     (GtkObject *) skin);
  gtk_signal_connect_object (GTK_OBJECT (skin->file_sel->cancel_button),
			     "clicked",
			     GTK_SIGNAL_FUNC (file_sel_cancelled),
			     (GtkObject *) skin);
  last_dir = gdam_history_get_entry ("Sequencer", "LastSequenceSavePath");
  if (last_dir)
    gtk_file_selection_set_filename (skin->file_sel, last_dir);
  gtk_widget_show (GTK_WIDGET (skin->file_sel));
}

static void 
gdam_sequencer_skin_make_save_dialog (GdamSequencerSkin *skin)
{
  make_file_dialog (skin, GTK_SIGNAL_FUNC (file_sel_save), "save sequence");
}

static void 
gdam_sequencer_skin_make_load_dialog (GdamSequencerSkin *skin)
{
  make_file_dialog (skin, GTK_SIGNAL_FUNC (file_sel_load), "load sequence");
}

static void 
gdam_sequencer_skin_save (GdamSequencerSkin *skin)
{
  const char *filename;
  g_return_if_fail (skin->filename_entry != NULL);
  filename = gtk_entry_get_text (skin->filename_entry);
  if (filename == NULL || is_blank (filename))
    {
      gdam_sequencer_skin_make_save_dialog (skin);
      return;
    }
  gdam_sequencer_save (get_sequencer (skin), filename);
}

static void recompute_beat_grid(GdamSequencerSkin* skin)
{
  int i;
  gdouble bpm =
    gdam_beat_info_get_bpm (&GDAM_MODEL (get_sequencer (skin))->beat_info,
			    0);
  gdouble one_grid_length;
  GdamSequencer *sequencer = get_sequencer (skin);
  gdam_debug ("recompute_beat_grid: called with %f beats",
	      sequencer->num_beats);
  if (bpm <= 0.0)
    return;
  if (skin->beat_grid)
    g_free (skin->beat_grid);

  skin->num_beat_grid = sequencer->num_beats;
  if (sequencer->num_sub_beats > 0)
    skin->num_beat_grid *= sequencer->num_sub_beats;
  one_grid_length = 1.0;
  if (sequencer->num_sub_beats > 0)
    one_grid_length /= sequencer->num_sub_beats;
  skin->beat_grid = g_new (gdouble, skin->num_beat_grid);
  for (i = 0; i < skin->num_beat_grid; i++)
    skin->beat_grid[i] = one_grid_length * i;
  gdam_debug ("r_b_g: setting sequencer to %f beats", sequencer->num_beats);
  gdam_sequencer_set_num_beats (sequencer, sequencer->num_beats);
}

static int redraw_skin_now(gpointer skin_ptr)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (skin_ptr);
  g_assert (skin->redraw_thread != 0);
  skin->redraw_thread = 0;
  redraw_area (skin, 0, 0, skin->draw_width, skin->draw_height);
  return 0;
}

static void delayed_redraw(GdamSequencerSkin* skin)
{
  if (skin->redraw_thread)
    return;
  skin->redraw_thread = g_timeout_add (100, redraw_skin_now, skin);
}

static void on_lag_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  float new_lag = adj->value;
  GdamSequencer *sequencer = get_sequencer (skin);
  sequencer->immediate_lag = new_lag;
}

static void on_asynch_step_size_adjusted(GtkAdjustment* adj, 
					 GdamSequencerSkin* skin)
{
  skin->asynch_step_size = 1 / adj->value;
}

static void on_subbeats_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
	int new_subbeats = adj->value;
	gboolean redraw = FALSE;
	GdamSequencer* sequencer = get_sequencer(skin);
	gdam_debug("subbeats adjusted: sequencer has %d beats", sequencer->num_beats);
	if ((new_subbeats == 0 && sequencer->num_sub_beats != 0)
	 || (new_subbeats != 0 && sequencer->num_sub_beats == 0))
	 	redraw = TRUE;
	/* TODO:  signal to sequencer?  Recompute sequence here? */
	sequencer->num_sub_beats = new_subbeats;
	if (redraw)
		delayed_redraw(skin);
	recompute_beat_grid(skin);
}

/* TODO: set upper value for base_offset value? */
static void on_beatnum_adjusted(GtkAdjustment* adj, GdamSequencerSkin* skin)
{
  /* TODO:  signal to sequencer?  Recompute sequence here? */
  GdamSequencer *sequencer = get_sequencer (skin);
  gdam_debug ("beatnum_adjusted to %f", adj->value);
  sequencer->num_beats = adj->value;
  gdam_debug ("beatnum_adjusted: set seq to %f beats", sequencer->num_beats);
  //skin->min_beat = 0.0;
  //skin->max_beat = adj->value;
  if (skin->time_zoom_adj)
    {
      skin->time_zoom_adj->value = adj->value;
      skin->time_zoom_adj->upper = adj->value;
    }
  if (skin->time_pan_adj)
    {
      skin->time_pan_adj->value = 0.0;
      //skin->time_pan_adj->upper = skin->time_pan_adj->page_size + adj->value;
      skin->time_pan_adj->upper = adj->value;
    }
  if (skin->time_zoom_adj)
    gtk_adjustment_changed (skin->time_zoom_adj);
  if (skin->time_pan_adj)
    gtk_adjustment_changed (skin->time_pan_adj);
  gdam_debug ("beatnum_adjusted: calling r_b_g with to %f",
	      sequencer->num_beats);
  recompute_beat_grid (skin);
  delayed_redraw (skin);
}

static void on_tempo_adjusted(GtkAdjustment* adj, gpointer user_data)
{
  GdamSequencerSkin *skin = (GdamSequencerSkin *) user_data;
  GdamBeatInfo *new_beat_info;
  GdamBeatInfo *old_beat_info = &GDAM_MODEL (get_sequencer (skin))->beat_info;
  gdouble length_scale_factor = 1.0;
  gdouble old_bpm = gdam_beat_info_get_bpm (old_beat_info, 0);
  if (old_bpm > 0 && adj->value > 0)
    {
      length_scale_factor = old_bpm / adj->value;
      if (skin->scale_element_lengths)
	{
	  GList *el;
	  for (el = get_sequencer (skin)->elements; el != NULL; el = el->next)
	    {
	      GdamSequencerElement *element;
	      element = (GdamSequencerElement *)el->data;
	      if (element->output_length > 0)
		element->output_length *= length_scale_factor;
	    }
	}
    }
  new_beat_info =
    gdam_beat_info_copy (&GDAM_MODEL (get_sequencer (skin))->beat_info);
  gdam_beat_info_set_bpm (new_beat_info, adj->value);
  gdam_model_set_beat_info (GDAM_MODEL (get_sequencer (skin)), new_beat_info);
  gdam_beat_info_destroy (new_beat_info);
  recompute_beat_grid (skin);
  delayed_redraw (skin);
}

static void on_sequencer_set_beat_info(GtkAdjustment* adj, 
				       const GdamBeatInfo* info, 
				       gpointer user_data)
{
  gtk_adjustment_set_value (adj, gdam_beat_info_get_bpm (info, 0));
}

static void 
skin_set_nth_local_buffer (GdamSequencerSkin *skin, int index, int id)
{
  GdamSequencer *seq = get_sequencer (skin);
  //guint index = info->num;
  //guint id = (int) atof (id_string);

  extend_local_ids (seq, index + 1);
  set_nth_local_id (seq, index, id);

  maybe_track_buffer_id (skin, id);
  if (index < (int)skin->local_id_labels->len &&
      g_array_index (skin->local_id_labels, gpointer, index))
    {
      char str[16];
      gpointer ptr = g_array_index (skin->local_id_labels, gpointer, index);
      sprintf (str, "%d", id);
      gtk_label_set_text (GTK_LABEL (ptr), str);
    }
  /* update all local_id elements to use the new buffer */
  swap_buffer_by_local_id (skin, index, id);
  gdam_debug ("setting local buffer %d to %d", index, id);
}

/* Set a certain local id to the value of the string dropped */
static void on_buffer_id_drop(const char* id_string, gpointer data)
{
  IntInfo *info = (IntInfo *) data;
  GdamSequencerSkin *skin = info->skin;
  guint index = info->num;
  guint id = (int) atof (id_string);
  g_return_if_fail (info->num >= 0);
  skin_set_nth_local_buffer (skin, index, id);
}

/* Set the current local buffer's id to the value of the string dropped */
static void on_cur_buffer_id_drop (const char* id_string, gpointer data)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (data);
  int index = skin->local_buffer;
  guint id = (int) atof (id_string);
  skin_set_nth_local_buffer (skin, index, id);
}

/* Import a new cache which is a loop, and create a new pattern which is that
 * buffer tiled and stretched.  For now, we choose the pattern which has
 * the same index as the new local id. */
/* TODO: move to component.  */
static void
import_loop_by_id (GdamSequencerSkin *skin, int buffer_id)
{
  GdamSequencer *seq = get_sequencer (skin);
  GdamSequencerElement *sample_info;
  int num_local_infos = seq->local_infos->len;
  guint index = num_local_infos;

  extend_local_ids (seq, index + 1);
  set_nth_local_id (seq, index, buffer_id);

  maybe_track_buffer_id (skin, buffer_id);
  if (index < skin->local_id_labels->len &&
      g_array_index (skin->local_id_labels, gpointer, index))
    {
      char str[16];
      gpointer ptr = g_array_index (skin->local_id_labels, gpointer, index);
      sprintf (str, "%d", buffer_id);
      gtk_label_set_text (GTK_LABEL (ptr), str);
    }
  /* Add first element, snap to grid */
  sample_info = add_new_element (skin,
				 buffer_id,
				 0,
				  /* XXX: snap-to-rate vs. rate-to-fill */
				 skin->default_resample,
				 skin->default_gain);
  sample_info->pattern_number = index;
  sample_info->local_id = index;

  maybe_track_element (sample_info, skin);

  tile_sample (skin, sample_info, TRUE);
}

/* copy one local buffer to another */
static void 
copy_local_buffer (GdamSequencerSkin *skin, guint src, guint dst)
{
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo, *dst_linfo;
  //guint index = info->num;
  //guint id = (int) atof (id_string);
  //g_return_if_fail (src >= 0);
  //g_return_if_fail (dst >= 0);

  linfo = nth_local_info (seq, src);
  g_return_if_fail (linfo != NULL);
  extend_local_ids (seq, dst + 1);
  set_nth_local_id (seq, dst, linfo->buffer_id);
  dst_linfo = nth_local_info (seq, dst);
  g_return_if_fail (dst_linfo != NULL);
  dst_linfo->rate = linfo->rate;
  dst_linfo->gain = linfo->gain;
  dst_linfo->sample_offset = linfo->sample_offset;
  dst_linfo->output_length = linfo->output_length;
  dst_linfo->trigger_rate = linfo->trigger_rate;

  if (dst < skin->local_id_labels->len &&
      g_array_index (skin->local_id_labels, gpointer, dst))
    {
      char str[16];
      gpointer ptr = g_array_index (skin->local_id_labels, gpointer, dst);
      sprintf (str, "%d", linfo->buffer_id);
      gtk_label_set_text (GTK_LABEL (ptr), str);
    }
  /* update all local_id elements to use the new buffer */
  swap_buffer_by_local_id (skin, dst, linfo->buffer_id);
}

/* copy one local buffer to another */
static void 
copy_to_local_buffer_zero (GdamSequencerSkin *skin, guint src)
{
  copy_local_buffer (skin, src, 0);
}


static gdouble 
buffer_length_secs (int buffer_id)
{
  int length_samples;
  GdamBufferManagerInfo *info = gdam_buffer_manager_info_by_id (buffer_id);
  if (info == NULL)
    length_samples = 2205;
  else
    length_samples = info->length;
  return (float)length_samples / 44100.0;
}

/* Import a new cache which is a loop, and set its rate to exactly fill
 * 4 / 8 / 16 beat.   */
/* TODO: move to component.  */
static void
import_locked_local_loop_by_id (GdamSequencerSkin *skin, int buffer_id)
{
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  int num_local_infos = seq->local_infos->len;
  guint index = num_local_infos;
  gdouble target_size, rate, length_secs, beat_length, bpm, step_size;
  gdouble elements_per_beat = 1;

  extend_local_ids (seq, index + 1);
  set_nth_local_id (seq, index, buffer_id);

  maybe_track_buffer_id (skin, buffer_id);

  if (index < skin->local_id_labels->len &&
      g_array_index (skin->local_id_labels, gpointer, index))
    {
      char str[16];
      gpointer ptr = g_array_index (skin->local_id_labels, gpointer, index);
      sprintf (str, "%d", buffer_id);
      gtk_label_set_text (GTK_LABEL (ptr), str);
    }

  length_secs = buffer_length_secs (buffer_id);

  /* Figure out the target length */
  bpm = gdam_beat_info_get_bpm (&GDAM_MODEL (seq)->beat_info, 0);
  beat_length = 60.0 / bpm;

  while (length_secs * elements_per_beat < 0.75 * beat_length)
    elements_per_beat *= 2;
  while (length_secs * elements_per_beat > 1.5 * beat_length)
    elements_per_beat /= 2;

  step_size = 1.0 / elements_per_beat;
  g_return_if_fail (step_size > 0.01);

  target_size = beat_length / elements_per_beat;
  rate = length_secs / target_size;

  /* set the rate of the local buffer to this magic rate */
  linfo = nth_local_info (seq, index);
  linfo->rate = rate;
}

/* Set a certain local id to the value of the string dropped */
static void 
on_import_loop_by_id_drop (const char *id_string, gpointer data)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (data);
  guint id = (int) atof (id_string);
  import_loop_by_id (skin, id);
}

/* Set a certain local id to the value of the string dropped */
static void
on_import_locked_local_loop_by_id_drop (const char *id_string, gpointer data)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (data);
  guint id = (int) atof (id_string);
  import_locked_local_loop_by_id (skin, id);
}

static void on_tempo_string_drop(const char* tempo, gpointer user_data)
{
  GdamSequencerSkin *skin = (GdamSequencerSkin *) user_data;
  GdamModel *model = GDAM_SKIN (skin)->model;
  GdamBeatInfo new_beat_info = model->beat_info;
  gdam_beat_info_set_bpm (&new_beat_info, atof (tempo));
  gdam_model_set_beat_info (model, &new_beat_info);

}

/* TODO: add this. */
static void on_set_bi(GdamProducer* 		producer,
		      const GdamBeatInfo* 	bi,
		      GdamSequencerSkin* 	skin)
{
  if (skin->tempo_adj)
    {
      //skin->tempo_adj->value = gdam_beat_info_get_bpm(bi,0);
      //gtk_adjustment_value_changed(skin->tempo_adj);
    }
}

static void on_sequence_changed(GdamSequencer* sequencer,
                                GdamSequencerSkin* skin)
{
  delayed_rebuffer (skin);
  delayed_redraw (skin);
  if (skin->pattern_toggles)
    g_list_foreach (skin->pattern_toggles, update_pattern_toggle, sequencer);
  if (skin->sample_state_toggles)
    g_list_foreach (skin->sample_state_toggles, update_sample_toggle, skin);
  if (skin->time_state_toggles)
    g_list_foreach (skin->time_state_toggles, update_time_toggle, skin);
  /* scan for buffer ids, maybe_track. */
  g_list_foreach (sequencer->elements, maybe_track_element, skin);
}

static gboolean 
viewport_handler                (GdamSkin          *skin,
                                 GtkObject         *object,
                                 const char        *signal_name,
                                 int                num_handler_names,
                                 char             **handler_names)
{
  GdamSequencerSkin *seq_skin = GDAM_SEQUENCER_SKIN (skin);
  if (!GTK_IS_CONTAINER (object))
    {
      g_warning ("cannot add buffer_viewport to non-container");
      return TRUE;
    }
  if (!seq_skin->cur_local_buffer_viewport)
    {
      g_warning ("cannot add non-existant buffer_viewport");
      return TRUE;
    }
  gtk_container_add (GTK_CONTAINER (object),
		     GDAM_SKIN (seq_skin->cur_local_buffer_viewport)->main_widget);
  return TRUE;
}


static gboolean       gdam_sequencer_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options)
{
  GdamModel *model;
  GdamSkinClass *skin_class;
  GdamBufferViewport *viewport = gdam_buffer_viewport_new_shown (channel);
  skin_class = GDAM_SKIN_CLASS (GTK_OBJECT (skin)->klass);
  model = GDAM_MODEL (gtk_type_new (skin_class->model_type));
  //model = GDAM_MODEL (gtk_type_new (GDAM_TYPE_SEQUENCER));
  gtk_object_ref (GTK_OBJECT (viewport));
  gtk_object_sink (GTK_OBJECT (viewport));
  gdam_buffer_viewport_set_mode (viewport, "average");
  GDAM_SEQUENCER_SKIN (skin)->cur_local_buffer_viewport = viewport;
  model->channel = channel;
  skin->model = model;
  gtk_signal_connect (GTK_OBJECT (model), "set_beat_info", on_set_bi, skin);
  gtk_signal_connect (GTK_OBJECT (model),
		      "sequence_changed", on_sequence_changed, skin);
  return TRUE;
}
static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_sequencer_drawing_area",
           gdam_sequencer_drawing_area_handler			},
	{ "gdam_sequencer_tempo_drop",
           gdam_sequencer_tempo_drop_handler			},
	{ "gdam_sequencer_size_from_length",
           gdam_sequencer_size_from_length_handler		},
	{ "gdam_sequencer_constant_size",
           gdam_sequencer_constant_size_handler			},
	{ "gdam_sequencer_hide_inactive",
           gdam_sequencer_hide_inactive_handler			},
	{ "gdam_sequencer_outline_inactive",
           gdam_sequencer_show_inactive_handler			},
	{ "gdam_sequencer_sort_by_id",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_local_id",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_pattern",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_resample",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_gain",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_length",
           gdam_sequencer_sort_by_handler			},
	{ "gdam_sequencer_sort_by_start",
           gdam_sequencer_sort_by_handler		        },
	{ "gdam_sequencer_render_as",
           gdam_sequencer_render_as_handler			},
	{ "gdam_sequencer_text_as",
           gdam_sequencer_text_as_handler			},
	{ "gdam_sequencer_display_tool",
           gdam_sequencer_display_tool_handler			},
	   /* TODO: play, stop -> component args? */
	{ "gdam_sequencer_play",
           gdam_sequencer_play_handler				},
	{ "gdam_sequencer_stop",
           gdam_sequencer_stop_handler				},
	{ "gdam_sequencer_add_immediate",
           gdam_sequencer_add_immediate_handler			},
	{ "gdam_sequencer_asynch_position_label",
           gdam_sequencer_asynch_position_label_handler		},
	{ "gdam_sequencer_autostep_asynch_tbutton",
           gdam_sequencer_autostep_asynch_tbutton_handler	},
	{ "gdam_sequencer_pattern_tbutton",
           gdam_sequencer_pattern_tbutton_handler	        },
	{ "gdam_sequencer_time_tbutton",
           gdam_sequencer_time_tbutton_handler	        	},
	{ "gdam_sequencer_sample_tbutton",
           gdam_sequencer_sample_tbutton_handler	       	},
	{ "gdam_sequencer_select_local_buffer",
           gdam_sequencer_select_local_buffer_handler		},
	{ "gdam_sequencer_copy_local_buffer",
           gdam_sequencer_copy_local_buffer_handler		},
	{ "gdam_sequencer_drop_local_buffer",
           gdam_sequencer_drop_local_buffer_handler		},
	{ "gdam_sequencer_drop_cur_local_buffer",
           drop_cur_local_buffer_handler			},
	{ "gdam_sequencer_drop_loop",
           gdam_sequencer_drop_loop_handler			},
	{ "gdam_sequencer_drop_locked_local_loop",
           gdam_sequencer_drop_locked_local_loop_handler	},
	{ "gdam_sequencer_local_buffer_label",
           gdam_sequencer_local_buffer_label_handler		},
	{ "gdam_sequencer_select_pattern",
           gdam_sequencer_pattern_handler			},
	{ "gdam_sequencer_pattern_on",
           gdam_sequencer_pattern_handler			},
	{ "gdam_sequencer_pattern_off",
           gdam_sequencer_pattern_handler			},
	{ "gdam_sequencer_set_sub_beats",
           gdam_sequencer_set_sub_beats_handler			},
	{ "gdam_sequencer_set_asynch_step_size",
           gdam_sequencer_set_asynch_step_size_handler		},
	{ "gdam_sequencer_set_lag",
           gdam_sequencer_set_lag_handler			},
	{ "gdam_sequencer_beatnum_adjuster",
           gdam_sequencer_beatnum_adjuster_handler		},
	{ "gdam_sequencer_tempo_adjuster",
           gdam_sequencer_tempo_adjuster_handler		},
	{ "gdam_sequencer_toggle_offset_adjuster",
           gdam_sequencer_toggle_offset_adjuster_handler	},
	{ "gdam_sequencer_h_zoom",
           gdam_sequencer_h_zoom_handler			},
	{ "gdam_sequencer_h_pan",
           gdam_sequencer_h_pan_handler				},
	{ "gdam_sequencer_v_zoom",
           gdam_sequencer_v_zoom_handler			},
	{ "gdam_sequencer_v_pan",
           gdam_sequencer_v_pan_handler				},
	{ "gdam_sequencer_filename_entry",
           gdam_sequencer_filename_entry_handler		},
	{ "gdam_sequencer_transpose_up_by_local_info",
           gdam_sequencer_transpose_up_by_local_id_handler	},
	{ "gdam_sequencer_transpose_down_by_local_info",
           gdam_sequencer_transpose_down_by_local_id_handler	},
	{ "gdam_sequencer_mute_by_local_id_tbutton",
           gdam_sequencer_mute_by_local_id_toggle_handler       },
	{ "gdam_sequencer_kill_by_local_id",
           gdam_sequencer_kill_by_local_id_handler       	},
	{ "gdam_sequencer_regain_up_by_local_info",
           gdam_sequencer_regain_up_by_local_id_handler		},
	{ "gdam_sequencer_regain_down_by_local_info",
           gdam_sequencer_regain_down_by_local_id_handler	},
	{ "gdam_sequencer_pattern_list",
           gdam_sequencer_pattern_list_handler			},
	{ "gdam_sequencer_pattern_name_entry",
           gdam_sequencer_pattern_name_entry_handler		},
	{ "gdam_sequencer_buffer_container",
           gdam_sequencer_buffer_container_handler		},
	{ "gdam_sequencer_cur_local_buffer_viewport",
           viewport_handler					},
	{ "gdam_sequencer_defer_to_cur_viewport",
	   gdam_sequencer_defer_to_cur_viewport_handler         },
};

static GdamSkinSimpleConnect simple_connect_infos[] = {
	{ "gdam_sequencer_add_asynch",
           (GdamSkinFunc)do_add_asynch				},
	{ "gdam_sequencer_step_asynch",
           (GdamSkinFunc)do_step_asynch				},
	{ "gdam_sequencer_save",
           (GdamSkinFunc)do_save				},
	{ "gdam_sequencer_midi",
           (GdamSkinFunc)do_midi				},
	{ "gdam_sequencer_load",
           (GdamSkinFunc)do_load				},
	{ "gdam_sequencer_cache",
           (GdamSkinFunc)do_cache				},
	{ "gdam_sequencer_load_dialog",
           (GdamSkinFunc)gdam_sequencer_skin_make_load_dialog	},
	{ "gdam_sequencer_save_dialog",
           (GdamSkinFunc)gdam_sequencer_skin_make_save_dialog	},
	{ "gdam_sequencer_skin_save",
           (GdamSkinFunc)gdam_sequencer_skin_save               },
	{ "gdam_sequencer_cut_elements",
           (GdamSkinFunc)do_element_cut               		},
	{ "gdam_sequencer_copy_elements",
           (GdamSkinFunc)do_element_copy               		},
	{ "gdam_sequencer_cut_and_merge_elements",
           (GdamSkinFunc)do_element_cut_and_merge              	},
	{ "gdam_sequencer_copy_and_merge_elements",
           (GdamSkinFunc)do_element_copy_and_merge             	},
	{ "gdam_sequencer_clear_elements",
           (GdamSkinFunc)do_element_clear                	},
	{ "gdam_sequencer_paste_elements",
           (GdamSkinFunc)do_paste_elements               	},
	{ "gdam_sequencer_snap_y",
           (GdamSkinFunc)do_snap_y				},
	{ "gdam_sequencer_free_y",
           (GdamSkinFunc)do_free_y				},
	{ "gdam_sequencer_scale_element_lengths",
           (GdamSkinFunc)do_scale_elements			},
	{ "gdam_sequencer_preserve_element_lengths",
           (GdamSkinFunc)undo_scale_elements			},
	{ "gdam_sequencer_select_matching_pattern",
           (GdamSkinFunc)select_matching_pattern		},
	{ "gdam_sequencer_select_matching_pattern_local_id",
           (GdamSkinFunc)select_matching_pattern_local_id	},
	{ "gdam_sequencer_select_none",
           (GdamSkinFunc)unselect_all_elements	                },
	{ "gdam_sequencer_select_active",
           (GdamSkinFunc)select_active	                	},
};

static void gdam_sequencer_skin_destroy(GtkObject* object)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  if (skin->selected_elements != NULL)
    {
      g_hash_table_destroy (skin->selected_elements);
      skin->selected_elements = NULL;
    }
  if (skin->grid_gc != NULL)
    {
      gdk_gc_unref (skin->grid_gc);
      skin->grid_gc = NULL;
    }
  if (skin->circle_boundary_gc != NULL)
    {
      gdk_gc_unref (skin->circle_boundary_gc);
      skin->circle_boundary_gc = NULL;
    }
  if (skin->blob_colors[0] != NULL)
    {
      int i;
      for (i = 0; i < GDAM_SEQUENCER_SKIN_NUM_BLOB_COLORS; i++)
	{
	  gdk_gc_unref (skin->blob_colors[i]);
	  skin->blob_colors[i] = NULL;
	}
    }
  if (skin->backing_pixmap != NULL)
    {
      gdk_pixmap_unref (skin->backing_pixmap);
      skin->backing_pixmap = NULL;
    }
  if (skin->in_transit != NULL)
    {
      g_hash_table_destroy (skin->in_transit);
      skin->in_transit = NULL;
    }

  (*parent_class->destroy) (object);
}

static void gdam_sequencer_skin_arg_set_toggle_display_mode(GtkObject* object, 
					             	    GdamArg*   arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  skin->toggle_display_mode = GDAM_ARG_INT (arg);
  if (skin->sample_state_toggles)
    g_list_foreach (skin->sample_state_toggles, update_sample_toggle, skin);
  if (skin->time_state_toggles)
    g_list_foreach (skin->time_state_toggles, update_time_toggle, skin);
}

static void gdam_sequencer_skin_arg_get_toggle_display_mode(GtkObject* object, 
					             	    GdamArg*   arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = skin->toggle_display_mode;
}

static int
local_info_length (GdamLocalSampleInfo *linfo)
{
  int l_length = linfo->output_length;
  /* negative number means play full length */
  if (l_length < 0) 
    {
      GdamBufferManagerInfo *info;
      info = gdam_buffer_manager_info_by_id (linfo->buffer_id);
      if (info == NULL)
	{
	  g_warning ("sequencer_skin: manager has no info "
		     "on buffer %d (local info length on %p)", 
		     linfo->buffer_id, linfo);
	  l_length = 2205;
	}
      else
	l_length = info->length;
    }
  return l_length;
}

static void
arg_set_current_local_buffer  (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;

  if (!local_id_valid (seq, GDAM_ARG_INT (arg)))
    return;

  skin->local_buffer = GDAM_ARG_INT (arg);
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  gdam_arg_set (object, "current_local_buffer_rate", linfo->rate);
  gdam_arg_set (object, "current_local_buffer_gain", linfo->gain);
  gdam_arg_set (object, "current_local_buffer_trigger_rate", 
                linfo->trigger_rate);
  gdam_arg_set (object, "current_local_buffer_start",
                (gdouble)linfo->sample_offset / 44100.0);
  gdam_arg_set (object, "current_local_buffer_length",
                (gdouble)linfo->output_length / 44100.0);
  gdam_arg_set (object, "current_local_buffer_description",
		get_local_buffer_description (skin));
  if (skin->cur_local_buffer_viewport)
    {
      int l_length = local_info_length (linfo);
      gdam_buffer_viewport_set_buffer_id (skin->cur_local_buffer_viewport,
	                                  linfo->buffer_id); 
      gdam_buffer_viewport_set_view (skin->cur_local_buffer_viewport,
				     (gdouble)linfo->sample_offset / 44100.0,
				     (gdouble)l_length / 44100.0);
    }
}

static void
arg_get_current_local_buffer  (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GDAM_ARG_INT (arg) = skin->local_buffer;
}

static void
arg_set_current_local_buffer_rate (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  linfo->rate =  GDAM_ARG_DOUBLE (arg);
  gdam_sequencer_update_sequence (seq);
}

static void
arg_get_current_local_buffer_rate (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  GDAM_ARG_DOUBLE (arg) = linfo->rate;
}

static void
arg_set_current_local_buffer_gain (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  linfo->gain =  GDAM_ARG_DOUBLE (arg);
  gdam_sequencer_update_sequence (seq);
}

static void
arg_get_current_local_buffer_gain (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  GDAM_ARG_DOUBLE (arg) = linfo->gain;
}

static void
local_buffer_changed (GdamSequencerSkin *skin, GdamLocalSampleInfo *linfo)
{
  if (skin->cur_local_buffer_viewport)
    {
      int l_length = local_info_length (linfo);
      gdam_buffer_viewport_set_buffer_id (skin->cur_local_buffer_viewport,
	                                  linfo->buffer_id); 
      gdam_buffer_viewport_set_view (skin->cur_local_buffer_viewport,
				     (gdouble)linfo->sample_offset / 44100.0,
				     (gdouble)l_length / 44100.0);
    }
}

static void
arg_set_current_local_buffer_start (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  linfo->sample_offset =  (int)(44100.0 * GDAM_ARG_DOUBLE (arg));
  local_buffer_changed (skin, linfo);
  gdam_sequencer_update_sequence (seq);
}

static void
arg_get_current_local_buffer_start (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  GDAM_ARG_DOUBLE (arg) = (gdouble)linfo->sample_offset / 44100.0;
}

static void
arg_set_current_local_buffer_length (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  linfo->output_length =  (int)(44100.0 * GDAM_ARG_DOUBLE (arg));
  local_buffer_changed (skin, linfo);
  gdam_sequencer_update_sequence (seq);
}

static void
arg_get_current_local_buffer_length (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  GDAM_ARG_DOUBLE (arg) = (gdouble)linfo->output_length / 44100.0;
}

static void
arg_set_current_local_buffer_trigger_rate (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  linfo->trigger_rate =  GDAM_ARG_DOUBLE (arg);
}

static void
arg_get_current_local_buffer_trigger_rate (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *seq = get_sequencer (skin);
  GdamLocalSampleInfo *linfo;
  if (!local_id_valid (seq, skin->local_buffer))
      return;
  linfo = &g_array_index (seq->local_infos,
                          GdamLocalSampleInfo,
			  skin->local_buffer);
  GDAM_ARG_DOUBLE (arg) = linfo->trigger_rate;
}

static void
arg_set_magic_rate (GtkObject* object,
		    GdamArg*   arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GdamSequencer *sequencer = get_sequencer (skin);
  GdamLocalSampleInfo *linfo = nth_local_info (sequencer, skin->local_buffer);
  gdouble rate = gdam_sequencer_get_magic_rate (sequencer, skin->local_buffer);
  gdam_arg_set (object, "current_local_buffer_rate", rate);
  /* XXX: emit _changed ??? */
}

/* Just here so others can trap arg_set */
static void
arg_set_current_local_buffer_description (GtkObject *object, GdamArg *arg)
{
}

/* Returns a newly allocated string */
static char *
get_local_buffer_description (GdamSequencerSkin *skin)
{
  GdamSequencer *sequencer = get_sequencer (skin);
  GdamLocalSampleInfo *linfo = nth_local_info (sequencer, skin->local_buffer);

  if (linfo)
    {
      GdamBufferManagerInfo *bufinfo;
      char *buf_fname;
      bufinfo = gdam_buffer_manager_info_by_id (linfo->buffer_id);
      g_return_val_if_fail (bufinfo != NULL, NULL);
      buf_fname = bufinfo->description;
      /* TODO: extension agnostic (here and elsewhere */
      if (strncmp (buf_fname, "Wav ", 4) == 0)
	buf_fname += 4;
      return g_strdup (g_basename (buf_fname));
    }
  else
    {
      return strdup ("---");
    }
}

static void
arg_get_current_local_buffer_description (GtkObject *object, GdamArg *arg)
{
  GdamSequencerSkin *skin = GDAM_SEQUENCER_SKIN (object);
  GDAM_ARG_STRING (arg) = get_local_buffer_description (skin);
}

static void gdam_sequencer_skin_class_init(GdamSkinClass* skin_class)
{
  skin_class->build_new_model = gdam_sequencer_build_new_model;
  GTK_OBJECT_CLASS (skin_class)->destroy = gdam_sequencer_skin_destroy;
  gdam_skin_class_register_generic_handlers (skin_class,
					     G_N_ELEMENTS (connect_infos),
					     connect_infos);
  gdam_skin_class_register_simple_connections 
    				      (skin_class,
				       G_N_ELEMENTS (simple_connect_infos),
				       simple_connect_infos);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "toggle_display_mode",
		     GDAM_ARG_TYPE_INT,
		     gdam_sequencer_skin_arg_set_toggle_display_mode,
		     gdam_sequencer_skin_arg_get_toggle_display_mode,
		     NULL, "0,1", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer",
		     GDAM_ARG_TYPE_INT,
		     arg_set_current_local_buffer,
		     arg_get_current_local_buffer,
		     NULL, "0,", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_gain",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_current_local_buffer_gain,
		     arg_get_current_local_buffer_gain,
		     NULL, "0,", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_current_local_buffer_rate,
		     arg_get_current_local_buffer_rate,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_trigger_rate",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_current_local_buffer_trigger_rate,
		     arg_get_current_local_buffer_trigger_rate,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_start",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_current_local_buffer_start,
		     arg_get_current_local_buffer_start,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_length",
		     GDAM_ARG_TYPE_DOUBLE,
		     arg_set_current_local_buffer_length,
		     arg_get_current_local_buffer_length,
		     NULL, NULL, 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "find_magic_rate_for_local_buffer",
		     GDAM_ARG_TYPE_INT,
		     arg_set_magic_rate,
		     NULL, NULL, "0,", 0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "vertical_sort_mode",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_vertical_sort_mode,
		     arg_get_vertical_sort_mode,
		     NULL,
	             "resample,gain,buffer_id,local_id,length,start,pattern",
		     0);
  gdam_arg_register (GTK_OBJECT_CLASS (skin_class),
		     "current_local_buffer_description",
		     GDAM_ARG_TYPE_STRING,
		     arg_set_current_local_buffer_description,
		     arg_get_current_local_buffer_description,
		     NULL, NULL, 0);
  skin_class->model_type = GDAM_TYPE_SEQUENCER;
}

